unit AppDirCfg;
{<
  AppDirCfg unit contains TAppDirCfg - non visual components which provide functionality
  for automatic loading and processing of configuration file.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2003)
  @lastmod(Oct 2019)
}

interface

{$INCLUDE _rosi.inc}
uses Windows, Messages, SysUtils, Classes, Forms, StrUtils, ShlObj;

type
  { TAppDirCfg is non visual component which provides functionality
    for automatic loading and processing of txt configuration file.
    Configuration file uses INI file format (Name=Value) and can contain
    specific directory tags which are automatically replaces by directory
    specified by the system. Following symbols are supported:
    %APPDIR%, %SYSTEMDIR%, %WINDIR%, %MYDOC%, %LOCALAPP%, %TEMPDIR%

    Default name of cfg file: ApplicationName.dir  (can be specified by  CfgFileName property)  @br
    Example of cfg file:  @br
    DataDir=%APPDIR%\Data @br
    ReportDir=%APPDIR%\Reports  @br
    DLLDir=%SYSTEMDIR%  @br
    INIDir=%WINDIR%  @br
    }
  TAppDirCfg = class(TComponent)
  private
    Dirs: TStringList;
    FCfgFileName: string;
    FCreateAllDir: boolean;
    procedure SetCfgFileName(Value: string);
  protected
    procedure Loaded; override; {< @exclude }
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { returns current SystemDirectory (e.g. C:\Windows\System\) }
    function SystemDir: string;
    { returns current WindowsDirectory (e.g. C:\Windows\) }
    function WinDir: string;
    { returns MyDocument directory for current user }
    function MyDocumentDir: string;
    { returns current Local Application (e.g. C:\Windows\) }
    function LocalApplDir: string;
    { returns Temporary directory }
    function TemporaryDir: string;
    { returns existency of "Name" item in the cfg file }
    function CfgExists(Name: string): boolean;
    { returns value of "Name" item from cfg file;
      e.g. line in cfgfile: DefaultFormat=0.00 ; Name='DefaultFormat', result will be 0.00 }
    function GetCfg(Name: string): string;
    { returns value of "Name" item from cfg file or "DefaultCfg" if "Name" not found }
    function GetCfgDef(Name, DefaultCfg: string): string;
    { returns value of "DirName" item from cfg file (last character is backslash);
      directory is created if prameter CreateDir is true }
    function GetDir(DirName: string; CreateDir: boolean = true): string;
    { returns value of "DirName" item from cfg file or "DefaultDir" if "DirName" is not found (last character is backslash);
      directory is created if prameter CreateDir is true }
    function GetDirDef(DirName, DefaultDir: string; CreateDir: boolean = true): string;
    { returns special directory in system according parameter FolderType,
      see ShlObj for list of values:
      CSIDL_DESKTOP,
      CSIDL_INTERNET,
      CSIDL_PROGRAMS,
      CSIDL_CONTROLS,
      CSIDL_PRINTERS,
      CSIDL_PERSONAL,
      CSIDL_FAVORITES,
      CSIDL_STARTUP,
      CSIDL_RECENT,
      CSIDL_SENDTO,
      CSIDL_BITBUCKET,
      CSIDL_STARTMENU,
      CSIDL_DESKTOPDIRECTORY,
      CSIDL_DRIVES,
      CSIDL_NETWORK,
      CSIDL_NETHOOD,
      CSIDL_FONTS,
      CSIDL_TEMPLATES,
      CSIDL_COMMON_STARTMENU,
      CSIDL_COMMON_PROGRAMS,
      CSIDL_COMMON_STARTUP,
      CSIDL_COMMON_DESKTOPDIRECTORY,
      CSIDL_APPDATA,
      CSIDL_PRINTHOOD,
      CSIDL_ALTSTARTUP,
      CSIDL_COMMON_ALTSTARTUP,
      CSIDL_COMMON_FAVORITES,
      CSIDL_INTERNET_CACHE,
      CSIDL_COOKIES,
      CSIDL_HISTORY }
    function GetSpecialDir(FolderType: integer): string;
    { reload cfg file from disk (called automatically in Loaded event) }
    procedure LoadCfgFile; virtual;
  published
    { FileName of cfg file, 'ApplicationName.dir' is used if empty }
    property CfgFileName: string read FCfgFileName write SetCfgFileName;
    { if true, application creates all directories from cfg file on startup }
    property CreateAllDir: boolean read FCreateAllDir write FCreateAllDir default false;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}
uses RosiCompGlobalCfg, rstring, rtool, rdlg;

procedure Register;
begin
  RegisterComponents('Rosi PS', [TAppDirCfg]);
end;

constructor TAppDirCfg.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Dirs:=TStringList.Create;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TAppDirCfg.Destroy;
begin
  FreeAndNil(Dirs);
  inherited Destroy;
end;

procedure TAppDirCfg.Loaded;
begin
  inherited Loaded;
  if (csDesigning in ComponentState) then
    Exit;
  if FCfgFileName = '' then
    FCfgFileName:=ChangeFileExt(Application.ExeName, '.dir')
  else
    FCfgFileName:=GetAppDir + FCfgFileName;
  LoadCfgFile;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TAppDirCfg.LoadCfgFile;
var
  i: integer;
  s: string;
begin
  if FileExists(FCfgFileName) then
  begin
    try
      Dirs.LoadFromFile(FCfgFileName);
    except
      on E: Exception do
      begin
        ShowErr(Format(sFileOpenErr, [FCfgFileName]), E);
        Exit;
      end;
    end;
    Dirs.Text:=ReplaceStr('%APPDIR%\', GetAppDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%APPDIR%', GetAppDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%SYSTEMDIR%\', SystemDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%SYSTEMDIR%', SystemDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%WINDIR%\', WinDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%WINDIR%', WinDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%MYDOC%\', MyDocumentDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%MYDOC%', MyDocumentDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%LOCALAPP%\', LocalApplDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%LOCALAPP%', LocalApplDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%TEMPDIR%\', TemporaryDir, Dirs.Text, true, false);
    Dirs.Text:=ReplaceStr('%TEMPDIR%', TemporaryDir, Dirs.Text, true, false);
    if FCreateAllDir then
    begin
      for i:=0 to Dirs.Count - 1 do
      begin
        s:=Dirs.Names[i];
        if StrUtils.EndsText('DIR', s) then
        begin
          s:=Dirs.ValueFromIndex[i];
          if s <> '' then
            ForceDirectories(s);
        end;
      end;
    end;
  end;
end;

procedure TAppDirCfg.SetCfgFileName(Value: string);
begin
  FCfgFileName:=Value;
  LoadCfgFile;
end;

function TAppDirCfg.SystemDir: string;
begin
  Result:=GetSpecialDir(CSIDL_SYSTEM);
end;

function TAppDirCfg.WinDir: string;
begin
  Result:=GetSpecialDir(CSIDL_WINDOWS);
end;

function TAppDirCfg.TemporaryDir: string;
var
  folder: array [0 .. MAX_PATH] of Char;
begin
  GetTempPath(Length(folder), folder);
  Result:=StrPas(folder);
end;

function TAppDirCfg.MyDocumentDir: string;
begin
  Result:=GetSpecialDir(CSIDL_PERSONAL);
end;

function TAppDirCfg.LocalApplDir: string;
begin
  Result:=GetSpecialDir(CSIDL_APPDATA);
end;

function TAppDirCfg.CfgExists(Name: string): boolean;
begin
  Result:=Dirs.IndexOfName(Name) >= 0;
end;

function TAppDirCfg.GetCfg(Name: string): string;
begin
  Result:=Dirs.Values[Name];
end;

function TAppDirCfg.GetCfgDef(Name, DefaultCfg: string): string;
begin
  Result:=GetCfg(Name);
  if Result = '' then
    Result:=DefaultCfg;
end;

function TAppDirCfg.GetDir(DirName: string; CreateDir: boolean = true): string;
begin
  Result:=GetCfg(DirName);
  if Result<>'' then
  begin
    if CreateDir then
      ForceDirectories(Result);
    Result:=SetEndOfPath(Result);
  end;
end;

function TAppDirCfg.GetDirDef(DirName, DefaultDir: string; CreateDir: boolean = true): string;
begin
  Result:=GetDir(DirName, CreateDir);
  if Result = '' then
  begin
    if CreateDir then
      ForceDirectories(DefaultDir);
    Result:=SetEndOfPath(DefaultDir);
  end;
end;

function TAppDirCfg.GetSpecialDir(FolderType: integer): string;
begin
  Result:=GetSystemDir(FolderType);
end;

end.
