unit AsBarcode;

{<
  AsBarcode unit contains TAsBarcode component for generating of 2D barcodes.
  Copyright 1998-2004 Andreas Schmidt and friends

  Version 2.0
  T.Rosinsky fixed many bugs in GetWidth, ShowText, Checksum and many more,
  now only runtime support

  Full functional demo.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2006)
  @lastmod(Apr 2014)
}

interface

{$INCLUDE _rosi.inc}

uses
  Classes, Graphics, {$IFDEF DXE2P} UITypes, {$ENDIF}
  AsBarCode_interface;

type
  // black_half means a black line with 2/5 height (used for PostNet)
  TBarLineType = (white, black, black_half); {< @exclude }

  { TAsBarcode draw barcode graphics to any canvas }
  TAsBarcode = class(TComponent)
  private
    FBarCode: string;
    FBarCodeType: TBarcodeType;
    FShowText: boolean;
    FUseCheckSum: boolean;
    FModul: integer;
    FRatio: double;
    FColor: TColor;
    FColorBar: TColor;
    FLeft: integer;
    FTop: integer;
    FHeight: integer;
    FShowTextFont: TFont;
    FShowTextPosition: TShowTextPosition;
    FAutoFontSize: boolean;

    BCText: string;
    modules: array [0 .. 3] of shortint;
    ShowTextW1, ShowTextH1: integer;

    procedure OneBarProps(code: char; var Width: integer; var lt: TBarLineType);
    procedure DoLines(data: string; Canvas: TCanvas);
    function GetWidthPos(pos: integer): integer;

    function Code_2_5_interleaved: string;
    function Code_2_5_industrial: string;
    function Code_2_5_matrix: string;
    function Code_39: string;
    function Code_39Extended: string;
    function Code_128: string;
    function Code_93: string;
    function Code_93Extended: string;
    function Code_MSI: string;
    function Code_PostNet: string;
    function Code_Codabar: string;
    function Code_EAN8: string;
    function Code_EAN13: string;
    function Code_UPC_A: string;
    function Code_UPC_E0: string;
    function Code_UPC_E1: string;
    function Code_Supp5: string;
    function Code_Supp2: string;

    procedure MakeModules;
    function MakeData: string;
    procedure SetModul(v: integer);
    function GetWidth: integer;
    procedure SetWidth(Value: integer);
    // Additions from Roberto Parola to improve the text output
    procedure SetShowTextFont(const Value: TFont);
    procedure SetTextWH(Canvas: TCanvas; AutoFont: boolean);
    procedure DrawText(Canvas: TCanvas);
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    procedure DrawBarcode(Canvas: TCanvas);
  published
    property BarCode: string read FBarCode write FBarCode;
    property BarCodeType: TBarcodeType read FBarCodeType write FBarCodeType default bcCode_2_5_interleaved;
    property UseCheckSum: boolean read FUseCheckSum write FUseCheckSum default true;
    property Top: integer read FTop write FTop;
    property Left: integer read FLeft write FLeft;
    property Height: integer read FHeight write FHeight;
    property Width: integer read GetWidth write SetWidth stored FALSE;
    { Width of the smallest line in a Barcode }
    property Modul: integer read FModul write SetModul;
    property Ratio: double read FRatio write FRatio;
    property ShowText: boolean read FShowText write FShowText default true;
    property ShowTextFont: TFont read FShowTextFont write SetShowTextFont;
    property ShowTextPosition: TShowTextPosition read FShowTextPosition write FShowTextPosition default stpTopLeft;
    property AutoFontSize: boolean read FAutoFontSize write FAutoFontSize;
    property Color: TColor read FColor write FColor default clWhite;
    property ColorBar: TColor read FColorBar write FColorBar default clBlack;
  end;

implementation

uses Types, SysUtils, math;

type
  TCheckSumMethod = (csmNone, csmModulo10);

const
  BCTypIncWidth1 = [bcCodeEAN8, bcCodeEAN13, bcCodeUPC_A, bcCodeUPC_E0, bcCodeUPC_E1];
  BCTypIncWidth2 = [bcCodeUPC_A, bcCodeUPC_E0, bcCodeUPC_E1];
  BCTypFontSmaller = [bcCodeEAN8, bcCodeUPC_A];

function SumOfDigit(x: integer): integer;
var
  sum: integer;
begin
  sum:=0;
  while x > 0 do
  begin
    sum:=sum + (x mod 10);
    x:=x div 10;
  end;
  Result:=sum;
end;

function CheckSumModulo10(data: string): string;
var
  i, fak, sum: integer;
begin
  sum:=0;
  fak:=Length(data);
  for i:=1 to Length(data) do
  begin
    if (fak mod 2) = 0 then
      sum:=sum + (StrToInt(data[i]) * 1)
    else
      sum:=sum + (StrToInt(data[i]) * 3);
    dec(fak);
  end;
  if (sum mod 10) = 0 then
    Result:=data + '0'
  else
    Result:=data + IntToStr(10 - (sum mod 10));
end;

function DoCheckSumming(data: string; Method: TCheckSumMethod): string;
begin
  case Method of
    csmNone:
      Result:=data;
    csmModulo10:
      Result:=CheckSumModulo10(data);
  end;
end;

function CheckSetLen(data: string; FinalLen: byte; AddZero: boolean; ErrMsg: string = ''): string;
begin
  Result:=data;
  if AddZero then
    Result:=StringOfChar('0', FinalLen - Length(data)) + data;
  if (ErrMsg <> '') and (Length(Result) <> FinalLen) then
    raise Exception.Create(Format('Invalid length for barcode "%s" (must be %d digits, current "%s").', [ErrMsg, FinalLen, Result]));
end;

function SetLenCheckSum(data: string; LenForCheckSum, FinalLen: byte; AddZero, UseCheckSum: boolean; Method: TCheckSumMethod; ErrMsg: string = ''): string;
begin
  if UseCheckSum then
  begin
    data:=CheckSetLen(data, LenForCheckSum, AddZero, ErrMsg);
    data:=DoCheckSumming(data, Method);
    CheckSetLen(data, FinalLen, FALSE, ErrMsg);
  end
  else
    data:=CheckSetLen(data, FinalLen, AddZero, ErrMsg);
  Result:=data;
end;

{
  converts a string from '321' to the internal representation '715'
  i need this function because some pattern tables have a different
  format :

  '00111'
  converts to '05161'
}
function Convert(const s: string): string;
var
  i, v: integer;
begin
  Result:=s; { same Length as Input - string }
  for i:=1 to Length(s) do
  begin
    v:=ord(s[i]) - 1;

    if odd(i) then
      Inc(v, 5);
    Result[i]:=Chr(v);
  end;
end;

constructor TAsBarcode.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FRatio:=2.0;
  FModul:=1;
  FBarCodeType:=bcCodeEAN13;
  FShowText:=true;
  FUseCheckSum:=true;
  FColor:=clWhite;
  FColorBar:=clBlack;
  FShowTextFont:=TFont.Create;
  FShowTextFont.Name:='Arial';
  FShowTextFont.Size:=8;
  FShowTextPosition:=stpTopLeft;
end;

destructor TAsBarcode.Destroy;
begin
  FreeAndNil(FShowTextFont);
  inherited;
end;

procedure TAsBarcode.Assign(Source: TPersistent);
var
  BSource: TAsBarcode;
begin
  if Source is TAsBarcode then
  begin
    BSource:=TAsBarcode(Source);
    FHeight:=BSource.FHeight;
    FBarCode:=BSource.FBarCode;
    FTop:=BSource.FTop;
    FLeft:=BSource.FLeft;
    FModul:=BSource.FModul;
    FRatio:=BSource.FRatio;
    FBarCodeType:=BSource.FBarCodeType;
    FShowText:=BSource.FShowText;
    FShowTextPosition:=BSource.FShowTextPosition;
    FColor:=BSource.FColor;
    FColorBar:=BSource.FColorBar;
  end
  else
    inherited;
end;

procedure TAsBarcode.SetShowTextFont(const Value: TFont);
begin
  FShowTextFont.Assign(Value);
end;

{ set Modul Width }
procedure TAsBarcode.SetModul(v: integer);
begin
  if (v >= 1) and (v < 50) then
    FModul:=v;
end;

{
  calculate the width and the linetype of a sigle bar


  Code   Line-Color      Width               Height
  ------------------------------------------------------------------
  '0'   white           100%                full
  '1'   white           100%*Ratio          full
  '2'   white           150%*Ratio          full
  '3'   white           200%*Ratio          full
  '5'   black           100%                full
  '6'   black           100%*Ratio          full
  '7'   black           150%*Ratio          full
  '8'   black           200%*Ratio          full
  'A'   black           100%                2/5  (used for PostNet)
  'B'   black           100%*Ratio          2/5  (used for PostNet)
  'C'   black           150%*Ratio          2/5  (used for PostNet)
  'D'   black           200%*Ratio          2/5  (used for PostNet)
}
procedure TAsBarcode.OneBarProps(code: char; var Width: integer; var lt: TBarLineType);
begin
  case code of
    '0':
      begin
        Width:=modules[0]; lt:=white;
      end;
    '1':
      begin
        Width:=modules[1]; lt:=white;
      end;
    '2':
      begin
        Width:=modules[2]; lt:=white;
      end;
    '3':
      begin
        Width:=modules[3]; lt:=white;
      end;

    '5':
      begin
        Width:=modules[0]; lt:=black;
      end;
    '6':
      begin
        Width:=modules[1]; lt:=black;
      end;
    '7':
      begin
        Width:=modules[2]; lt:=black;
      end;
    '8':
      begin
        Width:=modules[3]; lt:=black;
      end;

    'A':
      begin
        Width:=modules[0]; lt:=black_half;
      end;
    'B':
      begin
        Width:=modules[1]; lt:=black_half;
      end;
    'C':
      begin
        Width:=modules[2]; lt:=black_half;
      end;
    'D':
      begin
        Width:=modules[3]; lt:=black_half;
      end;
    else
      begin
        { something went wrong  :-( }
        { mistyped pattern table }
        raise Exception.CreateFmt('%s: internal Error', [self.ClassName]);
      end;
    end;
end;

function TAsBarcode.MakeData: string;
var
  i: integer;
begin
  { calculate the with of the different lines (modules) }
  BCText:=FBarCode;
  MakeModules;

  { numeric barcode type ? }
  if BCdata[FBarCodeType].num then
  begin
    BCText:=Trim(FBarCode); { remove blanks }
    for i:=1 to Length(BCText) do
      if (BCText[i] > '9') or (BCText[i] < '0') then
        raise Exception.Create('Barcode must be numeric');
  end;

  { get the pattern of the barcode }
  case FBarCodeType of
    bcCode_2_5_interleaved:
      Result:=Code_2_5_interleaved;
    bcCode_2_5_industrial:
      Result:=Code_2_5_industrial;
    bcCode_2_5_matrix:
      Result:=Code_2_5_matrix;
    bcCode39:
      Result:=Code_39;
    bcCode39Extended:
      Result:=Code_39Extended;
    bcCode128A, bcCode128B, bcCode128C, bcCodeEAN128A, bcCodeEAN128B, bcCodeEAN128C:
      Result:=Code_128;
    bcCode93:
      Result:=Code_93;
    bcCode93Extended:
      Result:=Code_93Extended;
    bcCodeMSI:
      Result:=Code_MSI;
    bcCodePostNet:
      Result:=Code_PostNet;
    bcCodeCodabar:
      Result:=Code_Codabar;
    bcCodeEAN8:
      Result:=Code_EAN8;
    bcCodeEAN13:
      Result:=Code_EAN13;
    bcCodeUPC_A:
      Result:=Code_UPC_A;
    bcCodeUPC_E0:
      Result:=Code_UPC_E0;
    bcCodeUPC_E1:
      Result:=Code_UPC_E1;
    bcCodeUPC_Supp2:
      Result:=Code_Supp2;
    bcCodeUPC_Supp5:
      Result:=Code_Supp5;
    else
      raise Exception.CreateFmt('%s: wrong BarcodeType', [self.ClassName]);
    end;
end;

function TAsBarcode.GetWidth: integer;
var
  data: string;
  i: integer;
  w: integer;
  lt: TBarLineType;
begin
  Result:=0;

  { get barcode pattern }
  data:=MakeData;

  for i:=1 to Length(data) do { examine the pattern string }
  begin
    OneBarProps(data[i], w, lt);
    Inc(Result, w);
  end;
  if FShowText and (FBarCodeType in BCTypIncWidth1) then
    Inc(Result, Round(Result * 0.15));
  if FShowText and (FBarCodeType in BCTypIncWidth2) then
    Inc(Result, Round(Result * 0.1));
end;

function TAsBarcode.GetWidthPos(pos: integer): integer;
var
  data: string;
  i: integer;
  w: integer;
  lt: TBarLineType;
begin
  Result:=0;

  { get barcode pattern }
  data:=MakeData;
  if pos > Length(data) then
    pos:=Length(data);
  for i:=1 to pos do { examine the pattern string }
  begin
    OneBarProps(data[i], w, lt);
    Inc(Result, w);
  end;
  if FShowText and (FBarCodeType in BCTypIncWidth1) then
    Inc(Result, Round(GetWidth * 0.1));
end;

procedure TAsBarcode.SetWidth(Value: integer);
begin
  FModul:=1;
  while GetWidth < Value do
    Inc(FModul);
  if FModul > 1 then
    dec(FModul);
end;

procedure TAsBarcode.MakeModules;
begin
  case FBarCodeType of
    bcCode_2_5_interleaved, bcCode_2_5_industrial, bcCode39, bcCodeEAN8, bcCodeEAN13, bcCode39Extended, bcCodeCodabar, bcCodeUPC_A, bcCodeUPC_E0, bcCodeUPC_E1,
      bcCodeUPC_Supp2, bcCodeUPC_Supp5:

      begin
        if Ratio < 2.0 then
          Ratio:=2.0;
        if Ratio > 3.0 then
          Ratio:=3.0;
      end;

    bcCode_2_5_matrix:
      begin
        if Ratio < 2.25 then
          Ratio:=2.25;
        if Ratio > 3.0 then
          Ratio:=3.0;
      end;
    bcCode128A, bcCode128B, bcCode128C, bcCode93, bcCode93Extended, bcCodeMSI, bcCodePostNet:
      ;
  end;

  modules[0]:=FModul;
  modules[1]:=Round(FModul * FRatio);
  modules[2]:=modules[1] * 3 div 2;
  modules[3]:=modules[1] * 2;
end;

{
  Draw the Barcode

  Parameter :
  'data' holds the pattern for a Barcode.
  A barcode begins always with a black line and
  ends with a black line.

  The white Lines builds the space between the black Lines.

  A black line must always followed by a white Line and vica versa.

  Examples:
  '50505'   // 3 thin black Lines with 2 thin white Lines
  '606'     // 2 fat black Lines with 1 thin white Line
  '5605015' // Error

  data[] : see procedure OneBarProps
}
procedure TAsBarcode.DrawBarcode(Canvas: TCanvas);
var
  data: string;
  SaveFont: TFont;
  SavePen: TPen;
  SaveBrush: TBrush;
begin
  SaveFont:=TFont.Create;
  SavePen:=TPen.Create;
  SaveBrush:=TBrush.Create;

  { get barcode pattern }
  data:=MakeData;

  try
    { store Canvas properties }
    SaveFont.Assign(Canvas.Font);
    SavePen.Assign(Canvas.Pen);
    SaveBrush.Assign(Canvas.Brush);
    SetTextWH(Canvas, FAutoFontSize);

    if FShowText then
      DrawText(Canvas); { show readable Text firstly! }
    DoLines(data, Canvas); { draw the barcode }

    { restore old Canvas properties }
    Canvas.Font.Assign(SaveFont);
    Canvas.Pen.Assign(SavePen);
    Canvas.Brush.Assign(SaveBrush);
  finally
    FreeAndNil(SaveFont);
    FreeAndNil(SavePen);
    FreeAndNil(SaveBrush);
  end;
end;

procedure TAsBarcode.SetTextWH(Canvas: TCanvas; AutoFont: boolean);
var
  SaveFont: TFont;
begin
  with Canvas do
  begin
    SaveFont:=TFont.Create;
    try
      SaveFont.Assign(Font);
      Font.Assign(ShowTextFont);

      if AutoFont then
      begin
        Font.Size:=2;
        while TextWidth(BCText) < Width do
          Font.Size:=Font.Size + 1;
        if (FBarCodeType in BCTypFontSmaller) then
          Font.Size:=Trunc(Font.Size * 0.65)
        else
          Font.Size:=Trunc(Font.Size * 0.75);
        if TextHeight('8') * 2 > Height then
          Font.Size:=Trunc(Font.Size * 0.5);
        ShowTextFont.Assign(Font);
      end;

      ShowTextW1:=TextWidth('8');
      ShowTextH1:=TextHeight('8');

      Font.Assign(SaveFont);
    finally
      FreeAndNil(SaveFont);
    end;
  end;
end;

{
  draw contents and type/name of barcode
  as human readable text at the left
  upper edge of the barcode.

  main use for this procedure is testing.

  note: this procedure changes Pen and Brush
  of the current canvas.

  Modifications from Roberto Parola to improve the text output
  Its useful to print the Text (code) on the barcode, in case the pen
  doesnt read the barcode.
  I didnt implement the EAN8 and EAN13 way to print the code, because
  the first character is outside of the bound of the barcode, and this
  can cause some problems (expecially in a report)
}
procedure TAsBarcode.DrawText(Canvas: TCanvas);
var
  PosX, PosY, w: integer;
  SaveFont: TFont;
  t: string;
begin
  with Canvas do
  begin

    { the fixed font size is a problem, if you
      use very large or small barcodes }
    // then i thought well to modify it

    // I know... you already did it in the DrawBarcode function, and this one is
    // called only by there... but i want to be sure :)
    SaveFont:=TFont.Create;
    try
      Font.Assign(ShowTextFont);
      try
        Pen.Color:=Font.Color;
        Brush.Color:=clWhite;

        // I only consider the Text (code) position.
        // As stated by Andreas Schmidt
        PosX:=FLeft;
        PosY:=FTop;

        if ShowTextPosition in [stpTopLeft, stpBottomLeft] then
          PosX:=FLeft
        else
          if ShowTextPosition in [stpTopRight, stpBottomRight] then
            PosX:=FLeft + Width - TextWidth(BCText)
          else
            // i know, the last IF is useless, but i like it this way
            if ShowTextPosition in [stpTopCenter, stpBottomCenter] then
              PosX:=FLeft + Trunc((Width - TextWidth(BCText)) / 2);

        if ShowTextPosition in [stpTopLeft, stpTopCenter, stpTopRight] then
          PosY:=FTop
        else
          if ShowTextPosition in [stpBottomLeft, stpBottomCenter, stpBottomRight] then
            PosY:=FTop + Height - TextHeight(BCText);

        if FShowText then
        begin
          case FBarCodeType of
            bcCodeEAN8:
              begin
                t:=copy(BCText, 1, 4);
                w:=GetWidthPos(3);
                TextOut(FLeft + w + 1, PosY, t);

                t:=copy(BCText, 5, 4);
                w:=GetWidthPos(23);
                TextOut(FLeft + w + 1, PosY, t);
              end;
            bcCodeEAN13:
              begin
                t:=copy(BCText, 1, 1);
                TextOut(FLeft + 1, PosY, t);

                t:=copy(BCText, 2, 6);
                w:=GetWidthPos(3);
                TextOut(FLeft + w + 1, PosY, t);

                t:=copy(BCText, 8, 6);
                w:=GetWidthPos(31);
                TextOut(FLeft + w + 1, PosY, t);
              end;
            bcCodeUPC_A:
              begin
                t:=copy(BCText, 1, 1);
                TextOut(FLeft + 1, PosY, t);

                t:=copy(BCText, 2, 5);
                w:=GetWidthPos(7);
                TextOut(FLeft + w + 1, PosY, t);

                t:=copy(BCText, 7, 5);
                w:=GetWidthPos(32);
                TextOut(FLeft + w + 1, PosY, t);

                t:=copy(BCText, 12, 1);
                w:=GetWidthPos(59);
                TextOut(FLeft + w + 1, PosY, t);
              end;
            bcCodeUPC_E0, bcCodeUPC_E1:
              begin
                t:=copy(BCText, 1, 1);
                TextOut(FLeft + 1, PosY, t);

                t:=copy(BCText, 2, 6);
                w:=GetWidthPos(4);
                TextOut(FLeft + w + 1, PosY, t);

                t:=copy(BCText, 8, 1);
                w:=GetWidthPos(34);
                TextOut(FLeft + w + 1, PosY, t);
              end
            else
              TextOut(PosX, PosY, BCText); { contents of Barcode }
          end;
        end;
      finally
        Font.Assign(SaveFont);
      end;
    finally
      FreeAndNil(SaveFont);
    end;
  end;
end;

procedure TAsBarcode.DoLines(data: string; Canvas: TCanvas);

var
  i: integer;
  lt: TBarLineType;
  xadd: integer;
  Width, Height: integer;
  a, b, c, d: TPoint; { Edges of a line (we need 4 Point because the line }
begin
  xadd:=0;
  if FShowText and (FBarCodeType in BCTypIncWidth1) then
    Inc(xadd, Round(self.Width * 0.1));

  with Canvas do
  begin
    Pen.Width:=1;

    for i:=1 to Length(data) do { examine the pattern string }
    begin
      OneBarProps(data[i], Width, lt);

      if (lt = black) or (lt = black_half) then
      begin
        Pen.Color:=FColorBar;
      end
      else
      begin
        Pen.Color:=FColor;
      end;
      Brush.Color:=Pen.Color;

      Height:=FHeight;
      if FShowText then
        Height:=Height - ShowTextH1;

      a.x:=FLeft + xadd;
      a.y:=FTop;
      if lt = black_half then
        a.y:=FTop + Height div 2;

      b.x:=a.x;
      b.y:=FTop + Height - 1;

      case FBarCodeType of
        bcCodeEAN8:
          if (i in [1 .. 3, 20 .. 24, 41 .. 43]) then
            b.y:=b.y + ShowTextH1 div 2;
        bcCodeEAN13:
          if (i in [1 .. 3, 28 .. 32, 57 .. 59]) then
            b.y:=b.y + ShowTextH1 div 2;
        bcCodeUPC_A:
          if (i in [1 .. 7, 28 .. 32, 53 .. 59]) then
            b.y:=b.y + ShowTextH1 div 2;
        bcCodeUPC_E0, bcCodeUPC_E1:
          if (i in [1 .. 3, 30 .. 33]) then
            b.y:=b.y + ShowTextH1 div 2;
      end;

      c.x:=a.x + Width - 1; { 23.04.1999 Line was 1 Pixel too wide }
      c.y:=b.y;
      d.x:=a.x + Width - 1; { 23.04.1999 Line was 1 Pixel too wide }
      d.y:=a.y;

      { draw the rectangle }
      Polygon([a, b, c, d]);

      xadd:=xadd + Width;
    end;
  end;
end;

{
  ////////////////////////////// EAN8 /////////////////////////////////////////
}

{ Pattern for Barcode EAN Charset A }
{ L1   S1   L2   S2 }
const
  tabelle_EAN_A: array ['0' .. '9'] of string = (('2605'), { 0 }
    ('1615'), { 1 }
    ('1516'), { 2 }
    ('0805'), { 3 }
    ('0526'), { 4 }
    ('0625'), { 5 }
    ('0508'), { 6 }
    ('0706'), { 7 }
    ('0607'), { 8 }
    ('2506') { 9 }
  );

  { Pattern for Barcode EAN Charset C }
  { S1   L1   S2   L2 }
const
  tabelle_EAN_C: array ['0' .. '9'] of string = (('7150'), { 0 }
    ('6160'), { 1 }
    ('6061'), { 2 }
    ('5350'), { 3 }
    ('5071'), { 4 }
    ('5170'), { 5 }
    ('5053'), { 6 }
    ('5251'), { 7 }
    ('5152'), { 8 }
    ('7051') { 9 }
  );

function TAsBarcode.Code_EAN8: string;
var
  i: integer;
begin
  BCText:=SetLenCheckSum(BCText, 7, 8, true, FUseCheckSum, csmModulo10, 'EAN8');
  Result:='505'; { Startcode }
  for i:=1 to 4 do
    Result:=Result + tabelle_EAN_A[BCText[i]];
  Result:=Result + '05050'; { Center Guard Pattern }
  for i:=5 to 8 do
    Result:=Result + tabelle_EAN_C[BCText[i]];
  Result:=Result + '505'; { Stopcode }
end;

{ ////////////////////////////// EAN13 /////////////////////////////////////// }

{ Pattern for Barcode EAN Zeichensatz B }
{ L1   S1   L2   S2 }
const
  tabelle_EAN_B: array ['0' .. '9'] of string = (('0517'), { 0 }
    ('0616'), { 1 }
    ('1606'), { 2 }
    ('0535'), { 3 }
    ('1705'), { 4 }
    ('0715'), { 5 }
    ('3505'), { 6 }
    ('1525'), { 7 }
    ('2515'), { 8 }
    ('1507') { 9 }
  );

  { Zuordung der Paraitaetsfolgen fr EAN13 }
const
  tabelle_ParityEAN13: array [0 .. 9, 1 .. 6] of char = (('A', 'A', 'A', 'A', 'A', 'A'), { 0 }
    ('A', 'A', 'B', 'A', 'B', 'B'), { 1 }
    ('A', 'A', 'B', 'B', 'A', 'B'), { 2 }
    ('A', 'A', 'B', 'B', 'B', 'A'), { 3 }
    ('A', 'B', 'A', 'A', 'B', 'B'), { 4 }
    ('A', 'B', 'B', 'A', 'A', 'B'), { 5 }
    ('A', 'B', 'B', 'B', 'A', 'A'), { 6 }
    ('A', 'B', 'A', 'B', 'A', 'B'), { 7 }
    ('A', 'B', 'A', 'B', 'B', 'A'), { 8 }
    ('A', 'B', 'B', 'A', 'B', 'A') { 9 }
  );

function TAsBarcode.Code_EAN13: string;
var
  i, LK: integer;
  tmp: String;
begin
  BCText:=SetLenCheckSum(BCText, 12, 13, true, FUseCheckSum, csmModulo10, 'EAN13');
  tmp:=BCText;
  LK:=StrToInt(tmp[1]);
  tmp:=copy(tmp, 2, 12);

  Result:='505'; { Startcode }
  for i:=1 to 6 do
  begin
    case tabelle_ParityEAN13[LK, i] of
      'A':
        Result:=Result + tabelle_EAN_A[tmp[i]];
      'B':
        Result:=Result + tabelle_EAN_B[tmp[i]];
      'C':
        Result:=Result + tabelle_EAN_C[tmp[i]];
    end;
  end;
  Result:=Result + '05050'; { Center Guard Pattern }
  for i:=7 to 12 do
    Result:=Result + tabelle_EAN_C[tmp[i]];
  Result:=Result + '505'; { Stopcode }
end;

{ Pattern for Barcode 2 of 5 }
const
  tabelle_2_5: array ['0' .. '9', 1 .. 5] of char = (('0', '0', '1', '1', '0'), { '0' }
    ('1', '0', '0', '0', '1'), { '1' }
    ('0', '1', '0', '0', '1'), { '2' }
    ('1', '1', '0', '0', '0'), { '3' }
    ('0', '0', '1', '0', '1'), { '4' }
    ('1', '0', '1', '0', '0'), { '5' }
    ('0', '1', '1', '0', '0'), { '6' }
    ('0', '0', '0', '1', '1'), { '7' }
    ('1', '0', '0', '1', '0'), { '8' }
    ('0', '1', '0', '1', '0') { '9' }
  );

function TAsBarcode.Code_2_5_interleaved: string;
var
  i, j: integer;
  c: char;
begin
  Result:='5050'; { Startcode }
  for i:=1 to Length(BCText) div 2 do
  begin
    for j:=1 to 5 do
    begin
      if tabelle_2_5[BCText[i * 2 - 1], j] = '1' then
        c:='6'
      else
        c:='5';
      Result:=Result + c;
      if tabelle_2_5[BCText[i * 2], j] = '1' then
        c:='1'
      else
        c:='0';
      Result:=Result + c;
    end;
  end;
  Result:=Result + '605'; { Stopcode }
end;

function TAsBarcode.Code_2_5_industrial: string;
var
  i, j: integer;
begin
  Result:='606050'; { Startcode }
  for i:=1 to Length(BCText) do
  begin
    for j:=1 to 5 do
    begin
      if tabelle_2_5[BCText[i], j] = '1' then
        Result:=Result + '60'
      else
        Result:=Result + '50';
    end;
  end;
  Result:=Result + '605060'; { Stopcode }
end;

function TAsBarcode.Code_2_5_matrix: string;
var
  i, j: integer;
  c: char;
begin
  Result:='705050'; { Startcode }
  for i:=1 to Length(BCText) do
  begin
    for j:=1 to 5 do
    begin
      if tabelle_2_5[BCText[i], j] = '1' then
        c:='1'
      else
        c:='0';
      { Falls i ungerade ist dann mache Lcke zu Strich }
      if odd(j) then
        c:=Chr(ord(c) + 5);
      Result:=Result + c;
    end;
    Result:=Result + '0'; { Lcke zwischen den Zeichen }
  end;
  Result:=Result + '70505'; { Stopcode }
end;

function TAsBarcode.Code_39: string;

type
  TCode39 = record
    c: char;
    data: array [0 .. 9] of char;
    chk: shortint;
  end;

const
  tabelle_39: array [0 .. 43] of TCode39 = ((c: '0'; data: '505160605'; chk: 0), (c: '1'; data: '605150506'; chk: 1), (c: '2'; data: '506150506'; chk: 2),
    (c: '3'; data: '606150505'; chk: 3), (c: '4'; data: '505160506'; chk: 4), (c: '5'; data: '605160505'; chk: 5), (c: '6'; data: '506160505'; chk: 6),
    (c: '7'; data: '505150606'; chk: 7), (c: '8'; data: '605150605'; chk: 8), (c: '9'; data: '506150605'; chk: 9), (c: 'A'; data: '605051506'; chk: 10),
    (c: 'B'; data: '506051506'; chk: 11), (c: 'C'; data: '606051505'; chk: 12), (c: 'D'; data: '505061506'; chk: 13), (c: 'E'; data: '605061505'; chk: 14),
    (c: 'F'; data: '506061505'; chk: 15), (c: 'G'; data: '505051606'; chk: 16), (c: 'H'; data: '605051605'; chk: 17), (c: 'I'; data: '506051605'; chk: 18),
    (c: 'J'; data: '505061605'; chk: 19), (c: 'K'; data: '605050516'; chk: 20), (c: 'L'; data: '506050516'; chk: 21), (c: 'M'; data: '606050515'; chk: 22),
    (c: 'N'; data: '505060516'; chk: 23), (c: 'O'; data: '605060515'; chk: 24), (c: 'P'; data: '506060515'; chk: 25), (c: 'Q'; data: '505050616'; chk: 26),
    (c: 'R'; data: '605050615'; chk: 27), (c: 'S'; data: '506050615'; chk: 28), (c: 'T'; data: '505060615'; chk: 29), (c: 'U'; data: '615050506'; chk: 30),
    (c: 'V'; data: '516050506'; chk: 31), (c: 'W'; data: '616050505'; chk: 32), (c: 'X'; data: '515060506'; chk: 33), (c: 'Y'; data: '615060505'; chk: 34),
    (c: 'Z'; data: '516060505'; chk: 35), (c: '-'; data: '515050606'; chk: 36), (c: '.'; data: '615050605'; chk: 37), (c: ' '; data: '516050605'; chk: 38),
    (c: '*'; data: '515060605'; chk: 0), (c: '$'; data: '515151505'; chk: 39), (c: '/'; data: '515150515'; chk: 40), (c: '+'; data: '515051515'; chk: 41),
    (c: '%'; data: '505151515'; chk: 42));

  function FindIdx(z: char): integer;
  var
    i: integer;
  begin
    for i:=0 to High(tabelle_39) do
    begin
      if z = tabelle_39[i].c then
      begin
        Result:=i;
        exit;
      end;
    end;
    Result:=-1;
  end;

var
  i, idx: integer;
  iCheckSum: integer;

begin
  iCheckSum:=0;
  { Startcode }
  Result:=tabelle_39[FindIdx('*')].data;
  Result:=Result + '0';

  for i:=1 to Length(BCText) do
  begin
    idx:=FindIdx(BCText[i]);
    if idx < 0 then
      continue;
    Result:=Result + tabelle_39[idx].data + '0';
    Inc(iCheckSum, tabelle_39[idx].chk);
  end;

  { Calculate iCheckSum Data }
  if UseCheckSum then
  begin
    iCheckSum:=iCheckSum mod 43;
    for i:=0 to High(tabelle_39) do
      if iCheckSum = tabelle_39[i].chk then
      begin
        Result:=Result + tabelle_39[i].data + '0';
        break;
      end;
  end;

  { Stopcode }
  Result:=Result + tabelle_39[FindIdx('*')].data;
end;

function TAsBarcode.Code_39Extended: string;

const
  code39x: array [0 .. 127] of string[2] = (('%U'), ('$A'), ('$B'), ('$C'), ('$D'), ('$E'), ('$F'), ('$G'), ('$H'), ('$I'), ('$J'), ('$K'), ('$L'), ('$M'),
    ('$N'), ('$O'), ('$P'), ('$Q'), ('$R'), ('$S'), ('$T'), ('$U'), ('$V'), ('$W'), ('$X'), ('$Y'), ('$Z'), ('%A'), ('%B'), ('%C'), ('%D'), ('%E'), (' '),
    ('/A'), ('/B'), ('/C'), ('/D'), ('/E'), ('/F'), ('/G'), ('/H'), ('/I'), ('/J'), ('/K'), ('/L'), ('/M'), ('/N'), ('/O'), ('0'), ('1'), ('2'), ('3'), ('4'),
    ('5'), ('6'), ('7'), ('8'), ('9'), ('/Z'), ('%F'), ('%G'), ('%H'), ('%I'), ('%J'), ('%V'), ('A'), ('B'), ('C'), ('D'), ('E'), ('F'), ('G'), ('H'), ('I'),
    ('J'), ('K'), ('L'), ('M'), ('N'), ('O'), ('P'), ('Q'), ('R'), ('S'), ('T'), ('U'), ('V'), ('W'), ('X'), ('Y'), ('Z'), ('%K'), ('%L'), ('%M'), ('%N'),
    ('%O'), ('%W'), ('+A'), ('+B'), ('+C'), ('+D'), ('+E'), ('+F'), ('+G'), ('+H'), ('+I'), ('+J'), ('+K'), ('+L'), ('+M'), ('+N'), ('+O'), ('+P'), ('+Q'),
    ('+R'), ('+S'), ('+T'), ('+U'), ('+V'), ('+W'), ('+X'), ('+Y'), ('+Z'), ('%P'), ('%Q'), ('%R'), ('%S'), ('%T'));

var
  save: string;
  i: integer;
begin
  save:=BCText;
  BCText:='';
  for i:=1 to Length(save) do
  begin
    if ord(save[i]) <= 127 then
      BCText:=BCText + string(code39x[ord(save[i])]);
  end;
  Result:=Code_39;
  BCText:=save;
end;

{
  Code 128
}
function TAsBarcode.Code_128: string;
type
  TCode128 = record
    a, b: char;
    c: string[2];
    data: string[6];
  end;

const
  tabelle_128: array [0 .. 102] of TCode128 = ((a: ' '; b: ' '; c: '00'; data: '212222'), (a: '!'; b: '!'; c: '01'; data: '222122'), (a: '"'; b: '"'; c: '02';
      data: '222221'), (a: '#'; b: '#'; c: '03'; data: '121223'), (a: '$'; b: '$'; c: '04'; data: '121322'), (a: '%'; b: '%'; c: '05'; data: '131222'),
    (a: '&'; b: '&'; c: '06'; data: '122213'), (a: ''''; b: ''''; c: '07'; data: '122312'), (a: '('; b: '('; c: '08'; data: '132212'),
    (a: ')'; b: ')'; c: '09'; data: '221213'), (a: '*'; b: '*'; c: '10'; data: '221312'), (a: '+'; b: '+'; c: '11'; data: '231212'), (a: ','; b: ','; c: '12';
      data: '112232'), { 23.10.2001 Stefano Torricella }
    (a: '-'; b: '-'; c: '13'; data: '122132'), (a: '.'; b: '.'; c: '14'; data: '122231'), (a: '/'; b: '/'; c: '15'; data: '113222'), (a: '0'; b: '0'; c: '16';
      data: '123122'), (a: '1'; b: '1'; c: '17'; data: '123221'), (a: '2'; b: '2'; c: '18'; data: '223211'), (a: '3'; b: '3'; c: '19'; data: '221132'),
    (a: '4'; b: '4'; c: '20'; data: '221231'), (a: '5'; b: '5'; c: '21'; data: '213212'), (a: '6'; b: '6'; c: '22'; data: '223112'), (a: '7'; b: '7'; c: '23';
      data: '312131'), (a: '8'; b: '8'; c: '24'; data: '311222'), (a: '9'; b: '9'; c: '25'; data: '321122'), (a: ':'; b: ':'; c: '26'; data: '321221'),
    (a: ';'; b: ';'; c: '27'; data: '312212'), (a: '<'; b: '<'; c: '28'; data: '322112'), (a: '='; b: '='; c: '29'; data: '322211'), (a: '>'; b: '>'; c: '30';
      data: '212123'), (a: '?'; b: '?'; c: '31'; data: '212321'), (a: '@'; b: '@'; c: '32'; data: '232121'), (a: 'A'; b: 'A'; c: '33'; data: '111323'),
    (a: 'B'; b: 'B'; c: '34'; data: '131123'), (a: 'C'; b: 'C'; c: '35'; data: '131321'), (a: 'D'; b: 'D'; c: '36'; data: '112313'), (a: 'E'; b: 'E'; c: '37';
      data: '132113'), (a: 'F'; b: 'F'; c: '38'; data: '132311'), (a: 'G'; b: 'G'; c: '39'; data: '211313'), (a: 'H'; b: 'H'; c: '40'; data: '231113'),
    (a: 'I'; b: 'I'; c: '41'; data: '231311'), (a: 'J'; b: 'J'; c: '42'; data: '112133'), (a: 'K'; b: 'K'; c: '43'; data: '112331'), (a: 'L'; b: 'L'; c: '44';
      data: '132131'), (a: 'M'; b: 'M'; c: '45'; data: '113123'), (a: 'N'; b: 'N'; c: '46'; data: '113321'), (a: 'O'; b: 'O'; c: '47'; data: '133121'),
    (a: 'P'; b: 'P'; c: '48'; data: '313121'), (a: 'Q'; b: 'Q'; c: '49'; data: '211331'), (a: 'R'; b: 'R'; c: '50'; data: '231131'), (a: 'S'; b: 'S'; c: '51';
      data: '213113'), (a: 'T'; b: 'T'; c: '52'; data: '213311'), (a: 'U'; b: 'U'; c: '53'; data: '213131'), (a: 'V'; b: 'V'; c: '54'; data: '311123'),
    (a: 'W'; b: 'W'; c: '55'; data: '311321'), (a: 'X'; b: 'X'; c: '56'; data: '331121'), (a: 'Y'; b: 'Y'; c: '57'; data: '312113'), (a: 'Z'; b: 'Z'; c: '58';
      data: '312311'), (a: '['; b: '['; c: '59'; data: '332111'), (a: '\'; b: '\'; c: '60'; data: '314111'), (a: ']'; b: ']'; c: '61'; data: '221411'),
    (a: '^'; b: '^'; c: '62'; data: '431111'), (a: '_'; b: '_'; c: '63'; data: '111224'), (a: #0; b: '`'; c: '64'; data: '111422'), (a: #1; b: 'a'; c: '65';
      data: '121124'), (a: #2; b: 'b'; c: '66'; data: '121421'), (a: #3; b: 'c'; c: '67'; data: '141122'), (a: #4; b: 'd'; c: '68'; data: '141221'),
    (a: #5; b: 'e'; c: '69'; data: '112214'), (a: #6; b: 'f'; c: '70'; data: '112412'), (a: #7; b: 'g'; c: '71'; data: '122114'), (a: #8; b: 'h'; c: '72';
      data: '122411'), (a: #9; b: 'i'; c: '73'; data: '142112'), (a: #10; b: 'j'; c: '74'; data: '142211'), (a: #11; b: 'k'; c: '75'; data: '241211'),
    (a: #12; b: 'l'; c: '76'; data: '221114'), (a: #13; b: 'm'; c: '77'; data: '413111'), (a: #14; b: 'n'; c: '78'; data: '241112'), (a: #15; b: 'o'; c: '79';
      data: '134111'), (a: #16; b: 'p'; c: '80'; data: '111242'), (a: #17; b: 'q'; c: '81'; data: '121142'), (a: #18; b: 'r'; c: '82'; data: '121241'),
    (a: #19; b: 's'; c: '83'; data: '114212'), (a: #20; b: 't'; c: '84'; data: '124112'), (a: #21; b: 'u'; c: '85'; data: '124211'), (a: #22; b: 'v'; c: '86';
      data: '411212'), (a: #23; b: 'w'; c: '87'; data: '421112'), (a: #24; b: 'x'; c: '88'; data: '421211'), (a: #25; b: 'y'; c: '89'; data: '212141'),
    (a: #26; b: 'z'; c: '90'; data: '214121'), (a: #27; b: '{'; c: '91'; data: '412121'), (a: #28; b: '|'; c: '92'; data: '111143'), (a: #29; b: '}'; c: '93';
      data: '111341'), (a: #30; b: '~'; c: '94'; data: '131141'), (a: #31; b: ' '; c: '95'; data: '114113'), (a: ' '; b: ' '; c: '96'; data: '114311'),
    (a: ' '; b: ' '; c: '97'; data: '411113'), (a: ' '; b: ' '; c: '98'; data: '411311'), (a: ' '; b: ' '; c: '99'; data: '113141'), (a: ' '; b: ' '; c: '  ';
      data: '114131'), (a: ' '; b: ' '; c: '  '; data: '311141'), (a: ' '; b: ' '; c: '  '; data: '411131') { FNC1 }
  );

  StartA = '211412';
  StartB = '211214';
  StartC = '211232';
  Stop = '2331112';

  { find Code 128 Codeset A or B }
  function Find_Code128AB(c: char): integer;
  var
    i: integer;
    v: char;
  begin
    for i:=0 to High(tabelle_128) do
    begin
      if FBarCodeType = bcCode128A then
        v:=tabelle_128[i].a
      else
        v:=tabelle_128[i].b;

      if c = v then
      begin
        Result:=i;
        exit;
      end;
    end;
    Result:=-1;
  end;

{ find Code 128 Codeset C }
  function Find_Code128C(c: string): integer;
  var
    i: integer;
  begin
    for i:=0 to High(tabelle_128) do
    begin
      if string(tabelle_128[i].c) = c then
      begin
        Result:=i;
        exit;
      end;
    end;
    Result:=-1;
  end;

var
  i, j, idx: integer;
  startcode: string;
  iCheckSum: integer;
  codeword_pos: integer;

begin
  case FBarCodeType of
    bcCode128A, bcCodeEAN128A:
      begin
        iCheckSum:=103;
        startcode:=StartA;
      end;
    bcCode128B, bcCodeEAN128B:
      begin
        iCheckSum:=104;
        startcode:=StartB;
      end;
    bcCode128C, bcCodeEAN128C:
      begin
        iCheckSum:=105;
        startcode:=StartC;
      end;
    else
      raise Exception.CreateFmt('%s: wrong BarcodeType in Code_128', [self.ClassName]);
    end;

  Result:=startcode; { Startcode }
  codeword_pos:=1;

  {
    case FBarCodeType of
    bcCodeEAN128A, bcCodeEAN128B, bcCodeEAN128C:
    begin
    //special identifier FNC1 = function code 1 for EAN 128 barcodes
    Result:=Result + tabelle_128[102].data;
    Inc(iCheckSum, 102 * codeword_pos);
    Inc(codeword_pos);
    // if there is no iCheckSum at the end of the string the EAN128 needs one (modulo 10)
    if UseCheckSum then
    BCText:=DoCheckSumming(BCText);
    end;
    end;
    }

  if (FBarCodeType = bcCode128C) or (FBarCodeType = bcCodeEAN128C) then
  begin
    if (Length(BCText) mod 2 <> 0) then
      BCText:='0' + BCText;
    for i:=1 to (Length(BCText) div 2) do
    begin
      j:=(i - 1) * 2 + 1;
      idx:=Find_Code128C(copy(BCText, j, 2));
      if idx < 0 then
        idx:=Find_Code128C('00');
      Result:=Result + string(tabelle_128[idx].data);
      Inc(iCheckSum, idx * codeword_pos);
      Inc(codeword_pos);
    end;
  end
  else
    for i:=1 to Length(BCText) do
    begin
      idx:=Find_Code128AB(BCText[i]);
      if idx < 0 then
        idx:=Find_Code128AB(' ');
      Result:=Result + string(tabelle_128[idx].data);
      Inc(iCheckSum, idx * codeword_pos);
      Inc(codeword_pos);
    end;

  iCheckSum:=iCheckSum mod 103;
  Result:=Result + string(tabelle_128[iCheckSum].data);

  Result:=Result + Stop; { Stopcode }
  Result:=Convert(Result);
end;

function TAsBarcode.Code_93: string;
type
  TCode93 = record
    c: char;
    data: array [0 .. 5] of char;
  end;

const
  tabelle_93: array [0 .. 46] of TCode93 = ((c: '0'; data: '131112'), (c: '1'; data: '111213'), (c: '2'; data: '111312'), (c: '3'; data: '111411'),
    (c: '4'; data: '121113'), (c: '5'; data: '121212'), (c: '6'; data: '121311'), (c: '7'; data: '111114'), (c: '8'; data: '131211'), (c: '9'; data: '141111'),
    (c: 'A'; data: '211113'), (c: 'B'; data: '211212'), (c: 'C'; data: '211311'), (c: 'D'; data: '221112'), (c: 'E'; data: '221211'), (c: 'F'; data: '231111'),
    (c: 'G'; data: '112113'), (c: 'H'; data: '112212'), (c: 'I'; data: '112311'), (c: 'J'; data: '122112'), (c: 'K'; data: '132111'), (c: 'L'; data: '111123'),
    (c: 'M'; data: '111222'), (c: 'N'; data: '111321'), (c: 'O'; data: '121122'), (c: 'P'; data: '131121'), (c: 'Q'; data: '212112'), (c: 'R'; data: '212211'),
    (c: 'S'; data: '211122'), (c: 'T'; data: '211221'), (c: 'U'; data: '221121'), (c: 'V'; data: '222111'), (c: 'W'; data: '112122'), (c: 'X'; data: '112221'),
    (c: 'Y'; data: '122121'), (c: 'Z'; data: '123111'), (c: '-'; data: '121131'), (c: '.'; data: '311112'), (c: ' '; data: '311211'), (c: '$'; data: '321111'),
    (c: '/'; data: '112131'), (c: '+'; data: '113121'), (c: '%'; data: '211131'), (c: '['; data: '121221'), { only used for Extended Code 93 }
    (c: ']'; data: '312111'), { only used for Extended Code 93 }
    (c: '{'; data: '311121'), { only used for Extended Code 93 }
    (c: '}'; data: '122211') { only used for Extended Code 93 }
  );

  { find Code 93 }
  function Find_Code93(c: char): integer;
  var
    i: integer;
  begin
    for i:=0 to High(tabelle_93) do
    begin
      if c = tabelle_93[i].c then
      begin
        Result:=i;
        exit;
      end;
    end;
    Result:=-1;
  end;

var
  i, idx: integer;
  checkC, checkK, { Checksums }
  weightC, weightK: integer;
begin
  Result:='111141'; { Startcode }

  for i:=1 to Length(BCText) do
  begin
    idx:=Find_Code93(BCText[i]);
    if idx < 0 then
      raise Exception.CreateFmt('%s:Code93 bad Data <%s>', [self.ClassName, BCText]);
    Result:=Result + tabelle_93[idx].data;
  end;

  checkC:=0;
  checkK:=0;
  weightC:=1;
  weightK:=2;

  for i:=Length(BCText) downto 1 do
  begin
    idx:=Find_Code93(BCText[i]);
    Inc(checkC, idx * weightC);
    Inc(checkK, idx * weightK);
    Inc(weightC);
    if weightC > 20 then
      weightC:=1;
    Inc(weightK);
    // if weightK > 15 then weightC := 1;
    if weightK > 15 then
      weightK:=1;
  end;

  Inc(checkK, checkC);
  checkC:=checkC mod 47;
  checkK:=checkK mod 47;

  Result:=Result + tabelle_93[checkC].data + tabelle_93[checkK].data;
  Result:=Result + '1111411'; { Stopcode }
  Result:=Convert(Result);
end;

function TAsBarcode.Code_93Extended: string;
const
  code93x: array [0 .. 127] of string[2] = ((']U'), ('[A'), ('[B'), ('[C'), ('[D'), ('[E'), ('[F'), ('[G'), ('[H'), ('[I'), ('[J'), ('[K'), ('[L'), ('[M'),
    ('[N'), ('[O'), ('[P'), ('[Q'), ('[R'), ('[S'), ('[T'), ('[U'), ('[V'), ('[W'), ('[X'), ('[Y'), ('[Z'), (']A'), (']B'), (']C'), (']D'), (']E'), (' '),
    ('{A'), ('{B'), ('{C'), ('{D'), ('{E'), ('{F'), ('{G'), ('{H'), ('{I'), ('{J'), ('{K'), ('{L'), ('{M'), ('{N'), ('{O'), ('0'), ('1'), ('2'), ('3'), ('4'),
    ('5'), ('6'), ('7'), ('8'), ('9'), ('{Z'), (']F'), (']G'), (']H'), (']I'), (']J'), (']V'), ('A'), ('B'), ('C'), ('D'), ('E'), ('F'), ('G'), ('H'), ('I'),
    ('J'), ('K'), ('L'), ('M'), ('N'), ('O'), ('P'), ('Q'), ('R'), ('S'), ('T'), ('U'), ('V'), ('W'), ('X'), ('Y'), ('Z'), (']K'), (']L'), (']M'), (']N'),
    (']O'), (']W'), ('}A'), ('}B'), ('}C'), ('}D'), ('}E'), ('}F'), ('}G'), ('}H'), ('}I'), ('}J'), ('}K'), ('}L'), ('}M'), ('}N'), ('}O'), ('}P'), ('}Q'),
    ('}R'), ('}S'), ('}T'), ('}U'), ('}V'), ('}W'), ('}X'), ('}Y'), ('}Z'), (']P'), (']Q'), (']R'), (']S'), (']T'));

var
  save: string;
  i: integer;
begin
  { CharToOem(PChar(BCText), save); }
  save:=BCText;
  BCText:='';
  for i:=1 to Length(save) do
  begin
    if ord(save[i]) <= 127 then
      BCText:=BCText + string(code93x[ord(save[i])]);
  end;
  Result:=Code_93;
  BCText:=save;
end;

function TAsBarcode.Code_MSI: string;
const
  tabelle_MSI: array ['0' .. '9'] of string[8] = (('51515151'), { '0' }
    ('51515160'), { '1' }
    ('51516051'), { '2' }
    ('51516060'), { '3' }
    ('51605151'), { '4' }
    ('51605160'), { '5' }
    ('51606051'), { '6' }
    ('51606060'), { '7' }
    ('60515151'), { '8' }
    ('60515160') { '9' }
  );

var
  i: integer;
  check_even, check_odd, iCheckSum: integer;
begin
  Result:='60'; { Startcode }
  check_even:=0;
  check_odd:=0;
  for i:=1 to Length(BCText) do
  begin
    if odd(i - 1) then
      check_odd:=check_odd * 10 + StrToInt(BCText[i])
    else
      check_even:=check_even * 10 + StrToInt(BCText[i]);
    Result:=Result + string(tabelle_MSI[BCText[i]]);
  end;

  iCheckSum:=SumOfDigit(check_odd * 2) + check_even;
  iCheckSum:=iCheckSum mod 10;
  if iCheckSum > 0 then
    iCheckSum:=10 - iCheckSum;
  if self.UseCheckSum then
    Result:=Result + string(tabelle_MSI[Chr(ord('0') + iCheckSum)]);
  Result:=Result + '515'; { Stopcode }
end;

function TAsBarcode.Code_PostNet: string;
const
  tabelle_PostNet: array ['0' .. '9'] of string[10] = (('5151A1A1A1'), { '0' }
    ('A1A1A15151'), { '1' }
    ('A1A151A151'), { '2' }
    ('A1A15151A1'), { '3' }
    ('A151A1A151'), { '4' }
    ('A151A151A1'), { '5' }
    ('A15151A1A1'), { '6' }
    ('51A1A1A151'), { '7' }
    ('51A1A151A1'), { '8' }
    ('51A151A1A1') { '9' }
  );
var
  i, iCheckSum: integer;
begin
  Result:='51';
  iCheckSum:=0;
  for i:=1 to Length(BCText) do
  begin
    Result:=Result + string(tabelle_PostNet[BCText[i]]);
    iCheckSum:=iCheckSum + StrToInt(BCText[i]);
  end;
  iCheckSum:=iCheckSum mod 10;
  if iCheckSum > 0 then
    iCheckSum:=10 - iCheckSum;
  if self.UseCheckSum then
    Result:=Result + string(tabelle_PostNet[Chr(ord('0') + iCheckSum)]);
  Result:=Result + '5';
end;

function TAsBarcode.Code_Codabar: string;
type
  TCodabar = record
    c: char;
    data: array [0 .. 6] of char;
  end;

const
  tabelle_cb: array [0 .. 19] of TCodabar = ((c: '1'; data: '5050615'), (c: '2'; data: '5051506'), (c: '3'; data: '6150505'), (c: '4'; data: '5060515'),
    (c: '5'; data: '6050515'), (c: '6'; data: '5150506'), (c: '7'; data: '5150605'), (c: '8'; data: '5160505'), (c: '9'; data: '6051505'),
    (c: '0'; data: '5050516'), (c: '-'; data: '5051605'), (c: '$'; data: '5061505'), (c: ':'; data: '6050606'), (c: '/'; data: '6060506'),
    (c: '.'; data: '6060605'), (c: '+'; data: '5060606'), (c: 'A'; data: '5061515'), (c: 'B'; data: '5151506'), (c: 'C'; data: '5051516'),
    (c: 'D'; data: '5051615'));

  { find Codabar }
  function Find_Codabar(c: char): integer;
  var
    i: integer;
  begin
    for i:=0 to High(tabelle_cb) do
    begin
      if c = tabelle_cb[i].c then
      begin
        Result:=i;
        exit;
      end;
    end;
    Result:=-1;
  end;

var
  i, idx: integer;
begin
  // Result:=tabelle_cb[Find_Codabar('A')].data;
  // Result:=Result + '0';
  for i:=1 to Length(BCText) do
  begin
    idx:=Find_Codabar(BCText[i]);
    Result:=Result + tabelle_cb[idx].data + '0';
  end;
  // Result:=Result + tabelle_cb[Find_Codabar('B')].data;
end;

{ --------------- }

function TAsBarcode.Code_UPC_A: string;
var
  i: integer;
begin
  BCText:=SetLenCheckSum(BCText, 11, 12, true, FUseCheckSum, csmModulo10, 'UPC_A');
  Result:='505'; { Startcode }
  for i:=1 to 6 do
    Result:=Result + tabelle_EAN_A[BCText[i]];
  Result:=Result + '05050'; { Trennzeichen }
  for i:=7 to 12 do
    Result:=Result + tabelle_EAN_C[BCText[i]];
  Result:=Result + '505'; { Stopcode }
end;

{ UPC E Parity Pattern Table , Number System 0 }
const
  tabelle_UPC_E0: array ['0' .. '9', 1 .. 6] of char = (('E', 'E', 'E', 'o', 'o', 'o'), { 0 }
    ('E', 'E', 'o', 'E', 'o', 'o'), { 1 }
    ('E', 'E', 'o', 'o', 'E', 'o'), { 2 }
    ('E', 'E', 'o', 'o', 'o', 'E'), { 3 }
    ('E', 'o', 'E', 'E', 'o', 'o'), { 4 }
    ('E', 'o', 'o', 'E', 'E', 'o'), { 5 }
    ('E', 'o', 'o', 'o', 'E', 'E'), { 6 }
    ('E', 'o', 'E', 'o', 'E', 'o'), { 7 }
    ('E', 'o', 'E', 'o', 'o', 'E'), { 8 }
    ('E', 'o', 'o', 'E', 'o', 'E') { 9 }
  );

function GenCodeUPC_A(UPC_E: string; Leading: char): string;
var
  tmp: string;
begin
  CheckSetLen(UPC_E, 6, FALSE, 'UPC_E');
  case StrToInt(UPC_E[6]) of
    0, 1, 2:
      tmp:=copy(UPC_E, 1, 2) + UPC_E[6] + '00' + '00' + copy(UPC_E, 3, 3);
    3:
      tmp:=copy(UPC_E, 1, 3) + '00' + '000' + copy(UPC_E, 4, 2);
    4:
      tmp:=copy(UPC_E, 1, 4) + '0' + '0000' + copy(UPC_E, 5, 1);
    5 .. 9:
      tmp:=copy(UPC_E, 1, 5) + '0000' + UPC_E[6];
  end;
  tmp:=Leading + tmp;
  Result:=SetLenCheckSum(tmp, 11, 12, FALSE, true, csmModulo10, 'UPC_A');
end;

function TAsBarcode.Code_UPC_E0: string;
var
  i, j: integer;
  tmp: string;
  c: char;
begin
  if Length(BCText) = 6 then
  begin
    // 7th digit can be calculated automatically, see http://en.wikipedia.org/wiki/Universal_Product_Code
    tmp:=GenCodeUPC_A(BCText, '0');
    BCText:=BCText + copy(tmp, 12, 1);
  end;
  tmp:=CheckSetLen(BCText, 7, true, 'UPC_E0');
  c:=tmp[7];
  BCText:='0' + tmp;

  Result:='505'; { Startcode }
  for i:=1 to 6 do
  begin
    if tabelle_UPC_E0[c, i] = 'E' then
    begin
      for j:=1 to 4 do
        Result:=Result + tabelle_EAN_C[tmp[i], 5 - j];
    end
    else
    begin
      Result:=Result + tabelle_EAN_A[tmp[i]];
    end;
  end;
  Result:=Result + '050505'; { Stopcode }
end;

function TAsBarcode.Code_UPC_E1: string;
var
  i, j: integer;
  tmp: String;
  c: char;
begin
  if Length(BCText) = 6 then
  begin
    // 7th digit can be calculated automatically, see http://en.wikipedia.org/wiki/Universal_Product_Code
    tmp:=GenCodeUPC_A(BCText, '1');
    BCText:=BCText + copy(tmp, 12, 1);
  end;
  tmp:=CheckSetLen(BCText, 7, true, 'UPC_E1');
  c:=tmp[7];
  BCText:='1' + tmp;

  Result:='505'; { Startcode }
  for i:=1 to 6 do
  begin
    if tabelle_UPC_E0[c, i] = 'E' then
    begin
      Result:=Result + tabelle_EAN_A[tmp[i]];
    end
    else
    begin
      for j:=1 to 4 do
        Result:=Result + tabelle_EAN_C[tmp[i], 5 - j];
    end;
  end;
  Result:=Result + '050505'; { Stopcode }
end;

{ assist function }
function getSupp(Nr: String): String;
var
  i, fak, sum: integer;
  tmp: String;
begin
  sum:=0;
  tmp:=copy(Nr, 1, Length(Nr) - 1);
  fak:=Length(tmp);
  for i:=1 to Length(tmp) do
  begin
    if (fak mod 2) = 0 then
      sum:=sum + (StrToInt(tmp[i]) * 9)
    else
      sum:=sum + (StrToInt(tmp[i]) * 3);
    dec(fak);
  end;
  sum:=((sum mod 10) mod 10) mod 10;
  Result:=tmp + IntToStr(sum);
end;

function TAsBarcode.Code_Supp5: string;
var
  i, j: integer;
  tmp: String;
  c: char;
begin
  BCText:=CheckSetLen(BCText, 5, true, 'UPC_Supp5');
  tmp:=getSupp(BCText + '0');
  c:=tmp[6];
  Result:='506'; { Startcode }
  for i:=1 to 5 do
  begin
    if tabelle_UPC_E0[c, (6 - 5) + i] = 'E' then
    begin
      for j:=1 to 4 do
        Result:=Result + tabelle_EAN_C[tmp[i], 5 - j];
    end
    else
    begin
      Result:=Result + tabelle_EAN_A[tmp[i]];
    end;
    if i < 5 then
      Result:=Result + '05'; { character delineator }
  end;
end;

function TAsBarcode.Code_Supp2: string;
var
  i, j: integer;
  tmp, mS: String;
begin
  BCText:=CheckSetLen(BCText, 2, true, 'UPC_Supp2');
  i:=StrToInt(BCText);
  case i mod 4 of
    3:
      mS:='EE';
    2:
      mS:='Eo';
    1:
      mS:='oE';
    0:
      mS:='oo';
  end;
  tmp:=getSupp(copy(BCText, 1, 5) + '0');

  Result:='506'; { Startcode }
  for i:=1 to 2 do
  begin
    if mS[i] = 'E' then
    begin
      for j:=1 to 4 do
        Result:=Result + tabelle_EAN_C[tmp[i], 5 - j];
    end
    else
    begin
      Result:=Result + tabelle_EAN_A[tmp[i]];
    end;
    if i < 2 then
      Result:=Result + '05'; { character delineator }
  end;
end;

end.
