unit DBTrackTool;

{<
  DBTrackTool unit contains set of components TTrackRecTarget and TTrackRecSource
  for automated tracking of dataset changes and events (Open, Close, Append, Edit, Delete)
  and saving of old and new values in selected fields of tracked datasets.
  Most of setting can be easily done by design time editor (see RosiDBDesignEditors package).

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2010)
  @lastmod(Feb 2017)
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls,
  Dialogs, DB;

type
  TTrackRecTarget = class;
  TTrackRecSource = class;

  { list of tracking dataset events }
  TTrackingEvent = (teAppend, teEdit, teDelete, teOpen, teClose, teView);
  { set of tracking dataset events }
  TTrackingEvents = set of TTrackingEvent;

  { list of tracking options }
  TTrackingOption = (toTrackNoChanges, toUseFieldName);
  { set of tracking options }
  TTrackingOptions = set of TTrackingOption;

  { event type for user defined datetime value }
  TTrackGetDateTime = procedure(Sender: TTrackRecTarget; var DateTime: TDateTime) of object;
  { event type for user defined message or value }
  TTrackEditMessage = procedure(Sender: TTrackRecSource; var Msg: string) of object;
  { list of possible formats of change information text }
  TChangesFormat = (frmText, frmMemoGrid);

  { non visual component defining dataset which stores tracked records
    of all related TTrackRecSource components }
  TTrackRecTarget = class(TComponent)
  private
    FDisable: boolean;
    FTargetDataSet: TDataSet;
    FFieldDateTime: string;
    FFieldUser: string;
    FFieldAction: string;
    FFieldTableName: string;
    FFieldKeyValue: string;
    FFieldChanges: string;
    FFieldMemo: string;
    FFieldCustomMemo: string;
    FFieldSpecial: string;
    FActiveUser: string;
    FOnGetDateTime: TTrackGetDateTime;
    FChangesFormat: TChangesFormat;
  protected
    procedure Loaded; override;
    procedure SetTargetDataSet(Value: TDataSet);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { main procedure for saving of record, used mainly by related TTrackRecSource }
    procedure SaveTrackRec(Action: TTrackingEvent; TableName, KeyValue: string; Changes, Memo: string;
      Special: integer; CustomMemo: string = '');
  published
    { Link to DataSet used for saving of tracked records }
    property TargetDataSet: TDataSet read FTargetDataSet write SetTargetDataSet;
    { Defines whether tracking is disabled for all related TTrackRecSource }
    property Disable: boolean read FDisable write FDisable;
    { Defines names of DateTime field in target dataset }
    property FieldDateTime: string read FFieldDateTime write FFieldDateTime;
    { Defines names of UserName field in target dataset }
    property FieldUser: string read FFieldUser write FFieldUser;
    { Defines names of Action field in target dataset }
    property FieldAction: string read FFieldAction write FFieldAction;
    { Defines names of TableName field in target dataset }
    property FieldTableName: string read FFieldTableName write FFieldTableName;
    { Defines names of KeyValue field in target dataset }
    property FieldKeyValue: string read FFieldKeyValue write FFieldKeyValue;
    { Defines names of Changes text field in target dataset }
    property FieldChanges: string read FFieldChanges write FFieldChanges;
    { Defines names of Memo field in target dataset }
    property FieldMemo: string read FFieldMemo write FFieldMemo;
    { Defines names of CustomMemo field in target dataset }
    property FieldCustomMemo: string read FFieldCustomMemo write FFieldCustomMemo;
    { Defines names of Special field in target dataset }
    property FieldSpecial: string read FFieldSpecial write FFieldSpecial;
    { Defines user name which is saved in target dataset }
    property ActiveUser: string read FActiveUser write FActiveUser;
    { event used for user definition of DateTime value }
    property OnGetDateTime: TTrackGetDateTime read FOnGetDateTime write FOnGetDateTime;
    { define format of change information text }
    property ChangesFormat: TChangesFormat read FChangesFormat write FChangesFormat;
  end;

  { non visual components defining tracked dataset,
    log is stored by related TTrackRecTarget component }
  TTrackRecSource = class(TComponent)
  private
    FDisable: boolean;
    FTrackedDataSet: TDataSet;
    FTrackRecTarget: TTrackRecTarget;
    FTrackingEvents: TTrackingEvents;
    FTrackingOptions: TTrackingOptions;
    FTrackedFields: TStrings;
    FTrackedSpecialFields: TStrings;
    FKeyFields: string;
    FTableName: string;
    FChangesFormatStr: string;
    FTableOpenStr: string;
    FTableCloseStr: string;
    OldRecValues: array of variant;
    OldRecSpecialValues: array of variant;
    FKeyValues: string;
    FLastAction: TTrackingEvent;
    FLastMsg: string;
    FLastMsgSpec: string;
    FOnEditMessage: TTrackEditMessage;
    FOnGetMemo: TTrackEditMessage;
    FOnGetCustomMemo: TTrackEditMessage;
    oldAfterEdit: TDataSetNotifyEvent;
    oldAfterInsert: TDataSetNotifyEvent;
    oldBeforePost: TDataSetNotifyEvent;
    oldAfterPost: TDataSetNotifyEvent;
    oldBeforeDelete: TDataSetNotifyEvent;
    oldAfterDelete: TDataSetNotifyEvent;
    oldAfterOpen: TDataSetNotifyEvent;
    oldBeforeClose: TDataSetNotifyEvent;
    procedure AfterEdit(DataSet: TDataSet);
    procedure AfterInsert(DataSet: TDataSet);
    procedure BeforePost(DataSet: TDataSet);
    procedure AfterPost(DataSet: TDataSet);
    procedure BeforeDelete(DataSet: TDataSet);
    procedure AfterDelete(DataSet: TDataSet);
    procedure AfterOpen(DataSet: TDataSet);
    procedure BeforeClose(DataSet: TDataSet);
    procedure LinkDataSet(DS: TDataSet);
    procedure UnLinkDataSet(DS: TDataSet);
  protected
    procedure Loaded; override;
    procedure SetTrackedDataSet(Value: TDataSet);
    procedure SetTrackRecTarget(Value: TTrackRecTarget);
    procedure SetTrackedFields(Value: TStrings);
    procedure SetTrackedSpecialFields(Value: TStrings);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure FieldsChanged(Sender: TObject);
    procedure CompareAndStoreChangeMsg;
    procedure SaveTrackRecord;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { store values of active record to memory (OldRecValues) to compare it later with a new values }
    procedure LoadValues;
    { store key field values of active record to memory (FKeyValues) to use it later during log }
    procedure LoadKeyValues;
    { save all dataset changes to the log by TTrackRecTarget }
    procedure SaveChanges;
    { save view log with current state of dataset,
      can be used to store View event even it is not defined in the TrackingEvents }
    procedure SaveViewLog(ForceSaving: boolean = false);
    { indicate last tracking event }
    property LastAction: TTrackingEvent read FLastAction;
    { indicate last tracking message }
    property LastMsg: string read FLastMsg;
    { indicate last tracking special message }
    property LastMsgSpec: string read FLastMsgSpec;
    { contain KeyValues used for storing of Track records }
    property KeyValues: string read FKeyValues write FKeyValues;
  published
    { Link to DataSet which is tracked }
    property TrackedDataSet: TDataSet read FTrackedDataSet write SetTrackedDataSet;
    { Defines whether tracking is disabled for this dataset }
    property Disable: boolean read FDisable write FDisable;
    { Link to TrackRecTarget which defines table for storing of tracking records }
    property TrackRecTarget: TTrackRecTarget read FTrackRecTarget write SetTrackRecTarget;
    { Defines set of events which is tracked }
    property TrackingEvents: TTrackingEvents read FTrackingEvents write FTrackingEvents;
    { Defines tracking options:
      toTrackNoChanges - log is stored even no tracked field was changed
      toUseFieldName - FieldName instead of FieldLabel is used in the log}
    property TrackingOptions: TTrackingOptions read FTrackingOptions write FTrackingOptions;
    { Defines list of field witch will be tracked }
    property TrackedFields: TStrings read FTrackedFields write SetTrackedFields;
    { Defines list of special field witch will be tracked }
    property TrackedSpecialFields: TStrings read FTrackedSpecialFields write SetTrackedSpecialFields;
    { Defines field (or fields seperated by ;) witch contain key value of records }
    property KeyFields: string read FKeyFields write FKeyFields;
    { Defines name of tracked table (used in the log) }
    property TableName: string read FTableName write FTableName;
    { Defines string witch will be used for formating of info changes
      use parameters:
      %f - for name of field
      %o - for old value of field
      %n - for new value of field
      }
    property ChangesFormatStr: string read FChangesFormatStr write FChangesFormatStr;
    { Defines used string after table is opened }
    property TableOpenStr: string read FTableOpenStr write FTableOpenStr;
    { Defines used string after table is closed }
    property TableCloseStr: string read FTableCloseStr write FTableCloseStr;
    { event used for final edition of text with change log }
    property OnEditMessage: TTrackEditMessage read FOnEditMessage write FOnEditMessage;
    { event used for definition of additional memo text }
    property OnGetMemo: TTrackEditMessage read FOnGetMemo write FOnGetMemo;
    { event used for definition of custom memo text }
    property OnGetCustomMemo: TTrackEditMessage read FOnGetCustomMemo write FOnGetCustomMemo;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses RosiCompGlobalCfg, rtool, rdbtool;

const
  ftNotSaveDataType = [ftBytes, ftVarBytes, ftBlob, ftGraphic .. ftTypedBinary, ftOraClob, ftOraBlob, ftArray,
    ftDataSet, ftReference, ftCursor, ftConnection, ftParams, ftStream, ftObject];

procedure Register;
begin
  RegisterComponents('Rosi DB', [TTrackRecTarget, TTrackRecSource]);
end;

// ======================================================================
// TrackRecTarget functions
constructor TTrackRecTarget.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  self.FFieldDateTime:='DateTime';
  self.FFieldUser:='User';
  self.FFieldAction:='Action';
  self.FFieldTableName:='TableName';
  self.FFieldKeyValue:='KeyValue';
  self.FFieldChanges:='Changes';
  self.FFieldMemo:='Memo';
  self.FFieldCustomMemo:='CustomMemo';
  self.FFieldSpecial:='Special';
  FChangesFormat:=frmText;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TTrackRecTarget.Destroy;
begin
  inherited Destroy;
end;

procedure TTrackRecTarget.Loaded;
begin
  inherited Loaded;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TTrackRecTarget.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FTargetDataSet) then
    FTargetDataSet:=nil;
end;

procedure TTrackRecTarget.SetTargetDataSet(Value: TDataSet);
begin
  FTargetDataSet:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

procedure TTrackRecTarget.SaveTrackRec(Action: TTrackingEvent; TableName, KeyValue: string; Changes, Memo: string;
  Special: integer; CustomMemo: string = '');
var
  d: TDateTime;
  f: TField;
begin
   if FDisable then
    Exit;
  if not Assigned(FTargetDataSet) then
  begin
    FDisable:=true;
    ShowMessage('Target dataset for tracking records is not defined. Tracking is turned off.');
    Exit;
  end;
  // load date time value
  d:=Now;
  if Assigned(FOnGetDateTime) then
    FOnGetDateTime(self, d);
  // open tracking table
  if not FTargetDataSet.Active then
    FTargetDataSet.Open;
  // append new record and save info
  FTargetDataSet.Append;
  try
    f:=FTargetDataSet.FindField(FFieldUser);
    if Assigned(f) then
      f.AsString:=FActiveUser;
    f:=FTargetDataSet.FindField(FFieldDateTime);
    if Assigned(f) then
      f.AsDateTime:=d;
    f:=FTargetDataSet.FindField(FFieldAction);
    if Assigned(f) then
      f.AsInteger:=ord(Action);
    f:=FTargetDataSet.FindField(FFieldTableName);
    if Assigned(f) then
      f.AsString:=TableName;
    f:=FTargetDataSet.FindField(FFieldKeyValue);
    if Assigned(f) then
      f.AsString:=KeyValue;
    f:=FTargetDataSet.FindField(FFieldChanges);
    if Assigned(f) then
      f.AsString:=Changes;
    f:=FTargetDataSet.FindField(FFieldMemo);
    if Assigned(f) then
      f.AsString:=Memo;
    f:=FTargetDataSet.FindField(FFieldCustomMemo);
    if Assigned(f) then
      f.AsString:=CustomMemo;
    f:=FTargetDataSet.FindField(FFieldSpecial);
    if Assigned(f) then
      f.AsInteger:=Special;

    FTargetDataSet.Post;
  finally
    if IsEditing(FTargetDataSet) then
      FTargetDataSet.Cancel;
  end;
end;

// ======================================================================
// TTrackRecSource functions
constructor TTrackRecSource.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FTrackingEvents:=[teAppend, teEdit, teDelete];
  FTrackedFields:=TStringList.Create;
  TStringList(FTrackedFields).OnChange:=FieldsChanged;
  FTrackedSpecialFields:=TStringList.Create;
  TStringList(FTrackedSpecialFields).OnChange:=FieldsChanged;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TTrackRecSource.Destroy;
begin
  FreeAndNil(FTrackedFields);
  FreeAndNil(FTrackedSpecialFields);
  inherited Destroy;
end;

procedure TTrackRecSource.Loaded;
begin
  inherited Loaded;
  if not (csDesigning in ComponentState) then
    // link to predefined dataset
    LinkDataSet(FTrackedDataSet);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TTrackRecSource.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FTrackedDataSet) then
    FTrackedDataSet:=nil;
  if (Operation = opRemove) and (AComponent = FTrackRecTarget) then
    FTrackRecTarget:=nil;
end;

procedure TTrackRecSource.SetTrackedDataSet(Value: TDataSet);
var
  old: TDataSet;
begin
  if Value = FTrackedDataSet then
    Exit;
  old:=FTrackedDataSet;
  FTrackedDataSet:=Value;
  if Value <> nil then
    Value.FreeNotification(self);

  // set table name according selected dataset
  if (csDesigning in ComponentState) and not(csLoading in ComponentState) then
    if Value <> nil then
    begin
      if (FTableName = '') or (Assigned(old) and (old.Name = FTableName)) then
        FTableName:=Value.Name;
    end
    else
      FTableName:='';

  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
    Exit;
  // in runtime unlink old dataset and link new dataset
  UnLinkDataSet(old);
  LinkDataSet(Value);
end;

procedure TTrackRecSource.SetTrackRecTarget(Value: TTrackRecTarget);
begin
  FTrackRecTarget:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

procedure TTrackRecSource.SetTrackedFields(Value: TStrings);
begin
  FTrackedFields.Assign(Value);
  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
    Exit;
  LoadValues;
end;

procedure TTrackRecSource.SetTrackedSpecialFields(Value: TStrings);
begin
  FTrackedSpecialFields.Assign(Value);
  if (csDesigning in ComponentState) or (csLoading in ComponentState) then
    Exit;
  LoadValues;
end;

procedure TTrackRecSource.FieldsChanged(Sender: TObject);
begin
  LoadValues;
end;

procedure TTrackRecSource.LinkDataSet(DS: TDataSet);
begin
  if not Assigned(DS) then
    Exit;
  oldAfterEdit:=DS.AfterEdit;
  DS.AfterEdit:=AfterEdit;
  oldAfterInsert:=DS.AfterInsert;
  DS.AfterInsert:=AfterInsert;
  oldBeforePost:=DS.BeforePost;
  DS.BeforePost:=BeforePost;
  oldAfterPost:=DS.AfterPost;
  DS.AfterPost:=AfterPost;
  oldBeforeDelete:=DS.BeforeDelete;
  DS.BeforeDelete:=BeforeDelete;
  oldAfterDelete:=DS.AfterDelete;
  DS.AfterDelete:=AfterDelete;
  oldAfterOpen:=DS.AfterOpen;
  DS.AfterOpen:=AfterOpen;
  oldBeforeClose:=DS.BeforeClose;
  DS.BeforeClose:=BeforeClose;
end;

procedure TTrackRecSource.UnLinkDataSet(DS: TDataSet);
begin
  if not Assigned(DS) then
    Exit;
  DS.AfterEdit:=oldAfterEdit;
  oldAfterEdit:=nil;
  DS.AfterInsert:=oldAfterInsert;
  oldAfterInsert:=nil;
  DS.BeforePost:=oldBeforePost;
  oldBeforePost:=nil;
  DS.AfterPost:=oldAfterPost;
  oldAfterPost:=nil;
  DS.BeforeDelete:=oldBeforeDelete;
  oldBeforeDelete:=nil;
  DS.AfterDelete:=oldAfterDelete;
  oldAfterDelete:=nil;
  DS.AfterOpen:=oldAfterOpen;
  oldAfterOpen:=nil;
  DS.BeforeClose:=oldBeforeClose;
  oldBeforeClose:=nil;
end;

procedure TTrackRecSource.AfterEdit(DataSet: TDataSet);
begin
  FLastAction:=teEdit;
  if (teEdit in FTrackingEvents) then
    LoadValues;
  if Assigned(oldAfterEdit) then
    oldAfterEdit(DataSet);
end;

procedure TTrackRecSource.AfterInsert(DataSet: TDataSet);
begin
  FLastAction:=teAppend;
  if (teAppend in FTrackingEvents) then
    LoadValues;
  if Assigned(oldAfterInsert) then
    oldAfterInsert(DataSet);
end;

procedure TTrackRecSource.BeforePost(DataSet: TDataSet);
begin
  if Assigned(oldBeforePost) then
    oldBeforePost(DataSet);
  CompareAndStoreChangeMsg;
end;

procedure TTrackRecSource.AfterPost(DataSet: TDataSet);
begin
  SaveTrackRecord;
  if Assigned(oldAfterPost) then
    oldAfterPost(DataSet);
end;

procedure TTrackRecSource.BeforeDelete(DataSet: TDataSet);
begin
  if teDelete in FTrackingEvents then
    LoadValues;
  if Assigned(oldBeforeDelete) then
    oldBeforeDelete(DataSet);
end;

procedure TTrackRecSource.AfterDelete(DataSet: TDataSet);
begin
  FLastAction:=teDelete;
  SaveChanges;
  if Assigned(oldAfterDelete) then
    oldAfterDelete(DataSet);
end;

procedure TTrackRecSource.AfterOpen(DataSet: TDataSet);
begin
  FLastAction:=teOpen;
  SaveChanges;
  if Assigned(oldAfterOpen) then
    oldAfterOpen(DataSet);
end;

procedure TTrackRecSource.BeforeClose(DataSet: TDataSet);
begin
  FLastAction:=teClose;
  SaveChanges;
  if Assigned(oldBeforeClose) then
    oldBeforeClose(DataSet);
end;

procedure TTrackRecSource.LoadValues;
var
  i: integer;
  f: TField;
begin
  FLastMsg:='';
  FLastMsgSpec:='';
  SetLength(OldRecValues, FTrackedFields.Count);
  SetLength(OldRecSpecialValues, FTrackedSpecialFields.Count);
  if not Assigned(FTrackedDataSet) or not FTrackedDataSet.Active then
    Exit;
  // store values of active record
  for i:=0 to FTrackedFields.Count - 1 do
  begin
    f:=FTrackedDataSet.FindField(FTrackedFields[i]);
    if Assigned(f) and not(f.DataType in ftNotSaveDataType) then
      OldRecValues[i]:=f.AsVariant
    else
      OldRecValues[i]:='';
  end;
  for i:=0 to FTrackedSpecialFields.Count - 1 do
  begin
    f:=FTrackedDataSet.FindField(FTrackedSpecialFields[i]);
    if Assigned(f) and not(f.DataType in ftNotSaveDataType) then
      OldRecSpecialValues[i]:=f.AsVariant
    else
      OldRecSpecialValues[i]:='';
  end;
  LoadKeyValues;
end;

procedure TTrackRecSource.LoadKeyValues;
var
  i: integer;
  f: TField;
  kf: string;
begin
  // store values from key fields
  if FKeyFields <> '' then
  begin
    i:=1;
    FKeyValues:='';
    repeat
      kf:=rtool.FindStrPart(FKeyFields, ';', i, false);
      if kf = '' then
        Break;
      if i > 1 then
        FKeyValues:=FKeyValues + ';';
      f:=FTrackedDataSet.FindField(kf);
      if Assigned(f) then
        FKeyValues:=FKeyValues + f.AsString;
      Inc(i);
    until false;
  end;
end;

procedure TTrackRecSource.CompareAndStoreChangeMsg;
var
  frm: string;
  i: integer;

  procedure FormatMsg(FieldName: string; Value: variant; var aMsg: string);
  var
    f: TField;
    s, fn: string;
  begin
    f:=FTrackedDataSet.FindField(FieldName);
    if not Assigned(f) or (f.DataType in ftNotSaveDataType) then
      Exit;
    try
      if (Value <> f.AsVariant) or (FLastAction in [teDelete, teAppend, teView]) then
      begin
        s:=frm;
        if (toUseFieldName in TrackingOptions) then
          fn:=f.FieldName
        else
          fn:=f.DisplayLabel;
        s:=rtool.ReplaceStr('%f', fn, s, true, false);
        case FLastAction of
          teAppend:
            s:=rtool.ReplaceStr('%o', '***', s, true, false);
          teView:
            s:=rtool.ReplaceStr('%o', f.AsString, s, true, false);
        else
          s:=rtool.ReplaceStr('%o', VarToStr(Value), s, true, false);
        end;
        case FLastAction of
          teDelete:
            s:=rtool.ReplaceStr('%n', '***', s, true, false);
          teView:
            s:=rtool.ReplaceStr('%n', '', s, true, false);
        else
          s:=rtool.ReplaceStr('%n', f.AsString, s, true, false);
        end;
        if aMsg <> '' then
          aMsg:=aMsg + #13 + #10;
        aMsg:=aMsg + s;
      end;
    except
    end;
  end;

begin
  FLastMsg:='';
  FLastMsgSpec:='';
  if FDisable or not(FLastAction in FTrackingEvents) then
    Exit;
  if not Assigned(FTrackRecTarget) then
    Exit;
  if not Assigned(FTrackedDataSet) or not FTrackedDataSet.Active then
    Exit;
  frm:=FChangesFormatStr;
  if FTrackRecTarget.FChangesFormat = frmMemoGrid then
    frm:='%f' + #9 + '%o' + #9 + '%n'
  else if frm = '' then
    frm:='Field %f: old value "%o", new value: "%n"';
  // create change message
  if FLastAction in [teAppend, teView] then
    LoadKeyValues;
  if FLastAction = teOpen then
  begin
    FLastMsg:=FTableOpenStr;
    if FLastMsg = '' then
      FLastMsg:='Table opened';
  end
  else if FLastAction = teClose then
  begin
    FLastMsg:=FTableCloseStr;
    if FLastMsg = '' then
      FLastMsg:='Table closed';
  end
  else
  begin
    for i:=0 to FTrackedFields.Count - 1 do
      FormatMsg(FTrackedFields[i], OldRecValues[i], FLastMsg);
    for i:=0 to FTrackedSpecialFields.Count - 1 do
      FormatMsg(FTrackedSpecialFields[i], OldRecSpecialValues[i], FLastMsgSpec);
  end;
end;

procedure TTrackRecSource.SaveTrackRecord;
var
  mstr, custmstr: string;
begin
  if (FLastAction <> teView) and ((FLastMsg <> '') or (toTrackNoChanges in FTrackingOptions)) then
  begin
    if Assigned(FOnEditMessage) then
      FOnEditMessage(self, FLastMsg);
    mstr:='';
    if Assigned(FOnGetMemo) then
      FOnGetMemo(self, mstr);
    custmstr:='';
    if Assigned(FOnGetCustomMemo) then
      FOnGetCustomMemo(self, custmstr);
    // call save procedure
    FTrackRecTarget.SaveTrackRec(FLastAction, FTableName, FKeyValues, FLastMsg, mstr, 0, custmstr);
  end;
  if (FLastMsgSpec <> '') then
  begin
    // call save procedure
    FTrackRecTarget.SaveTrackRec(FLastAction, FTableName, FKeyValues, FLastMsgSpec, '', 1, '');
  end;
end;

procedure TTrackRecSource.SaveChanges;
begin
  CompareAndStoreChangeMsg;
  SaveTrackRecord;
end;

procedure TTrackRecSource.SaveViewLog(ForceSaving: boolean = false);
var
  oldAction: TTrackingEvent;
  oldEvents: TTrackingEvents;
begin
  oldAction:=FLastAction;
  FLastAction:=teView;
  oldEvents:=FTrackingEvents;
  try
    if ForceSaving then
      FTrackingEvents:=FTrackingEvents + [teView];
    SaveChanges;
  finally
    FLastAction:=oldAction;
    FTrackingEvents:=oldEvents;
  end;
end;

end.
