unit PropSaveGrids;

{<
  PropSaveGrids unit contains set of components for re/storing
  of TGrid, TDBGrid, TrDBGrid, TrDBRecView and TDataset properties.
  It has to be used together with PropSaveMain and PropSaveFile.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2005)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Grids, DB, dbgrids, dbctrls, PropSaveMain,
{$IFDEF rUseDBComp} rDBGrid, rDBRecView, {$ENDIF} RosiCompGlobalCfg;

type
  { provides re/storing properties of TStringGrid components }
  TGridsPropSave = class(TCustomPropSave)
  private
    FStoreMetrix: boolean;
    FStoreStrings: boolean;
    FStoreRowColCount: boolean;
    procedure SaveStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
    procedure ReadStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
    procedure DelStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { returns whether re/storing of Components from parameter
      is supported by TGridsPropSave }
    function CanSave(Comp: TComponent): boolean; override;
    { returns key for column specified by Idx }
    function GetColumnKey(const Key: string; Idx: integer): string;
  published
    { define whether columns width and rows height is re/stored }
    property StoreMetrix: boolean read FStoreMetrix write FStoreMetrix default true;
    { define whether entered text in cells is re/stored }
    property StoreStrings: boolean read FStoreStrings write FStoreStrings default true;
    { define whether number of columns and rows is re/stored }
    property StoreRowColCount: boolean read FStoreRowColCount write FStoreRowColCount default true;
  end;

  { provides re/storing properties of TDBGrid & TDataset components }
  TDBGridsPropSave = class(TCustomPropSave)
  private
    FStoreColumns: boolean;
    FStoreColumnsProp: boolean;
    FEraseUnusedSection: boolean;
    procedure SaveDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
    procedure ReadDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
    procedure DelDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
    procedure SaveDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
    procedure ReadDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
    procedure DelDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
  protected
    ScaleRatio: double;
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    { list of saved DBColumns }
    SavedDBColumnFieldList: TStrings;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { returns whether re/storing of Components from parameter
      is supported by TDBGridsPropSave }
    function CanSave(Comp: TComponent): boolean; override;
    { returns key for column specified by Idx }
    function GetColumnKey(const Key: string; Idx: integer): string;
    { returns key for field specified by Idx }
    function GetFieldKey(const Key: string; Idx: integer): string;
    { fill list of saved DBColumns }
    function ReadDBColumnList(PSFile: TPropSaveFile; const Key: string): integer;
    { store one DBColumn to PSFile }
    procedure SaveDBColumn(PSFile: TPropSaveFile; const Key: string; C: TColumn; Idx: integer; IsCustomized: boolean = true);
    { read and restore one DBColumn from PSFile }
    procedure ReadDBColumn(PSFile: TPropSaveFile; const Key: string; C: TColumn; Idx: integer);
  published
    { define whether columns position and width is re/stored }
    property StoreColumns: boolean read FStoreColumns write FStoreColumns default true;
    { define whether columns properties (fonts, color, alignmetn...) is re/stored }
    property StoreColumnsProp: boolean read FStoreColumnsProp write FStoreColumnsProp default false;
    { define whether unused sections (columns with higher index) are deleted during reads }
    property EraseUnusedSection: boolean read FEraseUnusedSection write FEraseUnusedSection default true;
  end;
{$IFDEF rUseDBComp}

  { @exclude }
  TReadSavePropsEvent = procedure(G: TrDBGrid; A: TPropSaveAction; PSFile: TPropSaveFile) of object;

  { provides re/storing properties of TrDBGrid, TrDBGrid_MS & TrDBRecView components }
  TrDBGridsPropSave = class(TDBGridsPropSave)
  private
    FStoreFieldsGroup: boolean;
    FStoreFooterRow: boolean;
    FStoreDesignSetting: boolean;
    //FStoreAppearanceSetting: boolean;
    FOnReadSaveProps: TReadSavePropsEvent;
    procedure SaveRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
    procedure ReadRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
    procedure DelRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
    procedure SaveRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
    procedure ReadRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
    procedure DelRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    { returns whether re/storing of Components from parameter is supported by TrDBGridsPropSave }
    function CanSave(Comp: TComponent): boolean; override;
    { returns key for cellparam specified by Idx }
    function GetCellParamKey(const Key: string; Idx: integer): string;
    { read CellParamsCondList from PSFile }
    procedure ReadCellParamsCondList(PSFile: TPropSaveFile; const Key: string; CondList: TCellParamsCondList);
    { store CellParamsCondList to PSFile }
    procedure SaveCellParamsCondList(PSFile: TPropSaveFile; const Key: string; CondList: TCellParamsCondList);
  published
    { define whether fields group definition is re/stored }
    property StoreFieldsGroup: boolean read FStoreFieldsGroup write FStoreFieldsGroup default false;
    { define whether Footer setting is re/stored }
    property StoreFooterRow: boolean read FStoreFooterRow write FStoreFooterRow default false;
    { define whether design setting is re/stored }
    property StoreDesignSetting: boolean read FStoreDesignSetting write FStoreDesignSetting default false;
    { define whether appearance setting is re/stored }
    //property StoreAppearanceSetting: boolean read FStoreAppearanceSetting write FStoreAppearanceSetting default false;
    { event is called after read/store of DBGrid props }
    property OnReadSaveProps: TReadSavePropsEvent read FOnReadSaveProps write FOnReadSaveProps;
  end;
{$ENDIF}

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rtool;

const
  sColCount = 'ColCount';
  sRowCount = 'RowCount';
  sRowHeight = 'RowHeight%d';
  sColumn = 'Column%d';
  sWidth = 'Width';

  sFieldName = 'FieldName';
  sColWidth = 'ColWidth';
  sColVisible = 'ColVisible';

  cTitleLineSubKey = '.Title';
  sColColor = 'ColColor';
  sColFont = 'ColFont';
  sColAlign = 'ColAlign';
  sColTitle = 'ColTitle';
  sColTitleColor = 'ColTitleColor';
  sColTitleFont = 'ColTitleFont';
  sColTitleAlign = 'ColTitleAlign';

  sFieldCount = 'FieldCount';
  sField = 'Field%d';
  sFieldWidth = 'FieldWidth';
  sFieldFormat = 'FieldFormat';
  sFieldVisible = 'FieldVisible';

  sText = 'Text';

  sItemsCount = 'ItemsCount';
  sItemIndex = 'ItemIndex';
  sItemSel = 'Selected%d';
  sItem = 'Item%d';

  sRowLines = 'RowLines';
  sColLines = 'ColLines';
  sShowGradient = 'ShowGradient';
  sSortField = 'SortField';
  sSortDesc = 'SortDesc';
  sSortDirections = 'SortDirections';
  sColumnWidth = 'ColumnWidth';
  sTitleLines = 'TitleLines';
  sRowDataLines = 'RowDataLines';
  sRowDataLineColor = 'RowDataLineColor';
  sRowDataDotLine = 'RowDataDotLine';
  sFixedColumnsCount = 'FixedColumnsCount';
  sFixedColumnsColor = 'FixedColumnsColor';

  sDrawingStyle = 'DrawingStyle';
  sDrawColored = 'DrawColored';
  sColorOddRow = 'ColorOddRow';
  sColorActiveRow = 'ColorActiveRow';
  sColorStartGradient = 'ColorStartGradient';
  sColorEndGradient = 'ColorEndGradient';
  sFontColorFiltered = 'FontColorFiltered';
  sUseCellParamsConditions = 'UseCellParamsConditions';
  sFilterButtons = 'FilterButtons';
  sFilterBar = 'FilterBar';
  sFieldGroups = 'FieldGroups';
  sGridLookStyle = 'LookStyle';

  cFooterBoldText = 'FooterBoldText';
  cFooterCellLinesColor = 'FooterCellLinesColor';
  cFooterColor = 'FooterColor';
  cFooterFieldFooterDefs = 'FooterFieldFooterDefs';
  cFooterFixedColText = 'FooterFixedColText';
  cFooterFont = 'FooterFont';
  cFooterFooterHeight = 'FooterFooterHeight';
  cFooterFooterVisible = 'FooterFooterVisible';
  cFooterParentFont = 'FooterParentFont';
  cFooterShowCellLines = 'FooterShowCellLines';
  cFooterSimpleText = 'FooterSimpleText';
  cFooterSimpleTextAlign = 'FooterSimpleTextAlign';

  cFieldsFieldsAsBoolean = 'FieldsAsBoolean';
  cFieldsFieldsAsColor = 'FieldsAsColor';
  cFieldsFieldsAsDate = 'FieldsAsDate';
  cFieldsFieldsAsGraphics = 'FieldsAsGraphics';
  cFieldsFieldsAsLink = 'FieldsFieldsAsLink';
  cFieldsFieldsAsPassword = 'FieldsAsPassword';
  cFieldsFieldsAutoPickList = 'FieldsAutoPickList';
  cFieldsFieldsDrawDropDown = 'FieldsDrawDropDown';
  cFieldsFieldsFilterButton = 'FieldsFilterButton';
  cTitleHints = 'TitleHints';
  cTitleImages = 'TitleImages';
  cTitlePopupItems = 'TitlePopupItems';
  cDataImages = 'DataImages';

  cFixedColSubKey = '.FixedCol';
  cFixedColParenFont = 'FixedColParentFont';
  cFixedColWidth = 'FixedColWidth';
  cFixedColTitleText = 'FixedColTitleText';
  cFixedColTextMask = 'FixedColTextMask';
  cFixedColTextAlign = 'FixedColTextAlign';

  sFields = 'Fields';
  sSpecFont = 'SpecFont';
  sSpecFields = 'SpecFields';
  sPickLists = 'PickList';
  sOnlyVisibleField = 'OnlyVisibleField';
  sOnlySimpleField = 'OnlySimpleField';
  cFieldsGroupSubKey = '.FieldsGroup';

  sCellParamsCondCount = 'CellParamsCondCount';
  sCellParamsCond = 'CellParamsCond%d';
  sCPField = 'Field';
  sCPOperator = 'Operator';
  sCPValue = 'Value';
  sCPAction = 'Action';
  sCPFontColor = 'FontColor';
  sCPBkColor = 'BkColor';
  sCPBold = 'Bold';
  sCPWholeRow = 'WholeRow';
  sCPUseDataImage = 'UseDataImage';
  sCPImageIndex = 'ImageIndex';
  sCPImageAlignment = 'ImageAlignment';
  sCPShowText = 'ShowText';
  sCPStopIfTrue = 'StopIfTrue';
  sCPDeactive = 'Deactive';

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

procedure Register;
begin
  RegisterComponents('Rosi PS', [TGridsPropSave, TDBGridsPropSave {$IFDEF rUseDBComp}, TrDBGridsPropSave {$ENDIF}]);
end;

function FieldInList(FieldName: string): boolean;
var
  i: integer;
  s: string;
begin
  Result:=false;
  if RosiCompConfig.PropSaveCfg.FieldList_UseNewWidth = '' then
    Exit;
  i:=1;
  repeat
    s:=rtool.FindStrPart(RosiCompConfig.PropSaveCfg.FieldList_UseNewWidth, ';', i, true);
    if s = '' then
      Break;
    if SameText(s, FieldName) then
      Result:=true;
    inc(i);
  until false;
end;

// ==============================================================================
// GridsPropSave functions
constructor TGridsPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FStoreMetrix:=true;
  FStoreStrings:=true;
  FStoreRowColCount:=true;
end;

function TGridsPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TStringGrid);
end;

procedure TGridsPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);
begin
  if (Comp is TStringGrid) then
    case A of
      psaRead:
        ReadStrGrid(PSFile, Key, Comp as TStringGrid);
      psaSave:
        begin
          DelStrGrid(PSFile, Key, Comp as TStringGrid);
          SaveStrGrid(PSFile, Key, Comp as TStringGrid);
        end;
      psaDelete:
        DelStrGrid(PSFile, Key, Comp as TStringGrid);
    end;
end;

function TGridsPropSave.GetColumnKey(const Key: string; Idx: integer): string;
begin
  Result:=Key + '.' + Format(sColumn, [Idx]);
end;

procedure TGridsPropSave.SaveStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
var
  i: integer;
  k: string;
begin
  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    PSFile.WriteInteger(Key, sColorStartGradient, SG.GradientStartColor);
    PSFile.WriteInteger(Key, sColorEndGradient, SG.GradientEndColor);
    PSFile.WriteInteger(Key, sDrawingStyle, ord(SG.DrawingStyle));
  end;

  if FStoreRowColCount then
  begin
    PSFile.WriteInteger(Key, sColCount, SG.ColCount);
    PSFile.WriteInteger(Key, sRowCount, SG.RowCount);
  end;
  if FStoreMetrix then
    for i:=0 to SG.RowCount - 1 do
      PSFile.WriteInteger(Key, Format(sRowHeight, [i]), SG.RowHeights[i]);
  for i:=0 to SG.ColCount - 1 do
  begin
    k:=GetColumnKey(Key, i);
    if FStoreMetrix then
      PSFile.WriteInteger(k, sWidth, SG.ColWidths[i]);
    if FStoreStrings then
      PSFile.WriteStrings(k, SG.Cols[i]);
  end;
end;

procedure TGridsPropSave.ReadStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
var
  i, j: integer;
  k: string;
  L: TStrings;
begin
  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    SG.GradientStartColor:=PSFile.ReadInteger(Key, sColorStartGradient, SG.GradientStartColor);
    SG.GradientEndColor:=PSFile.ReadInteger(Key, sColorEndGradient, SG.GradientEndColor);
    SG.DrawingStyle:=TGridDrawingStyle(PSFile.ReadInteger(Key, sDrawingStyle, ord(SG.DrawingStyle)));
  end;

  if FStoreRowColCount then
  begin
    SG.ColCount:=PSFile.ReadInteger(Key, sColCount, SG.ColCount);
    SG.RowCount:=PSFile.ReadInteger(Key, sRowCount, SG.RowCount);
  end;
  if FStoreMetrix then
    for i:=0 to SG.RowCount - 1 do
      SG.RowHeights[i]:=PSFile.ReadInteger(Key, Format(sRowHeight, [i]), SG.RowHeights[i]);

  L:=TStringList.Create;
  try
    for i:=0 to SG.ColCount - 1 do
    begin
      k:=GetColumnKey(Key, i);
      if FStoreMetrix then
        SG.ColWidths[i]:=PSFile.ReadInteger(k, sWidth, SG.ColWidths[i]);
      if FStoreStrings then
      begin
        PSFile.ReadStrings(k, L);
        for j:=0 to SG.RowCount - 1 do
          if j < L.Count then
            SG.Cells[i, j]:=L[j];
      end;
    end;
  finally
    FreeAndNil(L);
  end;
end;

procedure TGridsPropSave.DelStrGrid(PSFile: TPropSaveFile; const Key: string; SG: TStringGrid);
var
  i: integer;
begin
  for i:=0 to SG.ColCount - 1 do
    PSFile.EraseSection(GetColumnKey(Key, i));
end;

// ==============================================================================
// DBGridsPropSave functions
constructor TDBGridsPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ScaleRatio:=1;
  FStoreColumns:=true;
  FStoreColumnsProp:=false;
  FEraseUnusedSection:=true;
  SavedDBColumnFieldList:=TStringList.Create;
end;

destructor TDBGridsPropSave.Destroy;
begin
  FreeAndNil(SavedDBColumnFieldList);
  inherited;
end;

function TDBGridsPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TDBComboBox) or (Comp is TDBListBox) or (Comp is TCustomDBGrid) or (Comp is TDataSet);
end;

function TDBGridsPropSave.GetColumnKey(const Key: string; Idx: integer): string;
begin
  Result:=Key + '.' + Format(sColumn, [Idx]);
end;

function TDBGridsPropSave.GetFieldKey(const Key: string; Idx: integer): string;
begin
  Result:=Key + '.' + Format(sField, [Idx]);
end;

procedure TDBGridsPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);

  procedure RSDBComboBox(CB: TDBComboBox);
  begin
    case A of
      psaSave:
        PSFile.WriteStrings(Key, CB.Items);
      psaRead:
        begin
          PSFile.ReadStrings(Key, CB.Items);
          try
            if CB.Style = csDropDownList then
              CB.ItemIndex:=CB.Items.IndexOf(CB.DataSource.DataSet.FieldValues[CB.DataField]);
          except
          end;
        end;
      psaDelete:
        PSFile.EraseSection(Key);
    end;
  end;

  procedure RSDBListBox(LB: TDBListBox);
  begin
    case A of
      psaSave:
        PSFile.WriteStrings(Key, LB.Items);
      psaRead:
        PSFile.ReadStrings(Key, LB.Items);
      psaDelete:
        PSFile.EraseSection(Key);
    end;
  end;

begin
  if (Comp is TDBComboBox) then
    RSDBComboBox(Comp as TDBComboBox)
  else
    if (Comp is TDBListBox) then
      RSDBListBox(Comp as TDBListBox)
    else
      if (Comp is TCustomDBGrid) then
      begin
        case A of
          psaRead:
            ReadDBGrid(PSFile, Key, Comp as TCustomDBGrid);
          psaSave:
            SaveDBGrid(PSFile, Key, Comp as TCustomDBGrid);
          psaDelete:
            DelDBGrid(PSFile, Key, Comp as TCustomDBGrid);
        end;
      end
      else
        if (Comp is TDataSet) then
        begin
          case A of
            psaRead:
              ReadDataSet(PSFile, Key, Comp as TDataSet);
            psaSave:
              SaveDataSet(PSFile, Key, Comp as TDataSet);
            psaDelete:
              DelDataSet(PSFile, Key, Comp as TDataSet);
          end;
        end
end;

procedure TDBGridsPropSave.SaveDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
var
  ColIdx: integer;
  HG: THackCustomDBGrid;
begin
  HG:=THackCustomDBGrid(SG);
  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    PSFile.WriteBool(Key, sRowLines, dgRowLines in HG.Options);
    PSFile.WriteBool(Key, sColLines, dgColLines in HG.Options);
    PSFile.WriteInteger(Key, sColorStartGradient, HG.GradientStartColor);
    PSFile.WriteInteger(Key, sColorEndGradient, HG.GradientEndColor);
    PSFile.WriteInteger(Key, sDrawingStyle, ord(HG.DrawingStyle));

    if not HG.ParentFont then
    begin
      PSFile.WriteFont(Key, HG.Font, ScaleRatio);
      PSFile.WriteFont(Key + cTitleLineSubKey, HG.TitleFont, ScaleRatio);
    end
    else
    begin
      PSFile.DeleteFont(Key);
      PSFile.DeleteFont(Key + cTitleLineSubKey);
    end;
  end;
  if FStoreColumns then
  begin
    // save columns
    PSFile.WriteInteger(Key, sColCount, HG.Columns.Count);
    for ColIdx:=0 to HG.Columns.Count - 1 do
      SaveDBColumn(PSFile, Key, HG.Columns[ColIdx], ColIdx, HG.Columns.State=csCustomized);
  end;
end;

procedure TDBGridsPropSave.SaveDBColumn(PSFile: TPropSaveFile; const Key: string; C: TColumn; Idx: integer; IsCustomized: boolean = true);
var
  k: string;
begin
  k:=GetColumnKey(Key, Idx);
  // Save column position
  PSFile.WriteString(k, sFieldName, C.FieldName);
  if C.Visible then
    PSFile.WriteInteger(k, sColWidth, C.Width);
  PSFile.WriteBool(k, sColVisible, C.Visible);

  // Save column props
  if FStoreColumnsProp then
  begin
    if not IsCustomized or ((cvColor in C.AssignedValues) and (C.Color <> C.DefaultColor)) then
      PSFile.WriteInteger(k, sColColor, C.Color)
    else
      PSFile.DeleteKey(k, sColColor);
    if not IsCustomized or (cvFont in C.AssignedValues) then
      PSFile.WriteFont(k, C.Font, ScaleRatio)
    else
      PSFile.DeleteFont(k);
    if not IsCustomized or (cvAlignment in C.AssignedValues) then
      PSFile.WriteInteger(k, sColAlign, ord(C.Alignment))
    else
      PSFile.DeleteKey(k, sColAlign);

    if not IsCustomized or ((cvTitleColor in C.AssignedValues) and (C.Title.Color <> C.Title.DefaultColor)) then
      PSFile.WriteInteger(k, sColTitleColor, C.Title.Color)
    else
      PSFile.DeleteKey(k, sColTitleColor);
    // if (cvTitleFont in C.AssignedValues) then PSFile.WriteFont(k,c.Title.Font, ScaleRatio);
    if not IsCustomized or (cvTitleAlignment in C.AssignedValues) then
      PSFile.WriteInteger(k, sColTitleAlign, ord(C.Title.Alignment))
    else
      PSFile.DeleteKey(k, sColTitleAlign);

    if not IsCustomized or (cvTitleCaption in C.AssignedValues) then
      PSFile.WriteString(k, sColTitle, C.Title.Caption)
    else
      PSFile.DeleteKey(k, sColTitle);

    if C.Field <> nil then
    begin
      PSFile.WriteInteger(k, sFieldWidth, C.Field.DisplayWidth);
      if (C.Field is TNumericField) then
      begin
        if TNumericField(C.Field).DisplayFormat = '' then
          PSFile.DeleteKey(k, sFieldFormat)
        else
          PSFile.WriteString(k, sFieldFormat, TNumericField(C.Field).DisplayFormat);
      end;
      if (C.Field is TDateTimeField) then
      begin
        if TDateTimeField(C.Field).DisplayFormat = '' then
          PSFile.DeleteKey(k, sFieldFormat)
        else
          PSFile.WriteString(k, sFieldFormat, TDateTimeField(C.Field).DisplayFormat);
      end;
      if (C.Field is TSQLTimeStampField) then
      begin
        if TSQLTimeStampField(C.Field).DisplayFormat = '' then
          PSFile.DeleteKey(k, sFieldFormat)
        else
          PSFile.WriteString(k, sFieldFormat, TSQLTimeStampField(C.Field).DisplayFormat);
      end;
    end;
  end;
end;

procedure TDBGridsPropSave.ReadDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
var
  SavedCols, GridCols, scIdx, gcIdx, actIdx: integer;
  k, fn: string;
  C: TColumn;
  b: boolean;
  ChangedColsIdx: array of integer;
  HG: THackCustomDBGrid;

  function MoveIdx(newIdx: integer): integer;
  var
    i: integer;
  begin
    Result:=0;
    if newIdx > High(ChangedColsIdx) then
      newIdx:= High(ChangedColsIdx);
    for i:=0 to newIdx - 1 do
      inc(Result, ChangedColsIdx[i]);
    for i:=newIdx to High(ChangedColsIdx) do
      if ChangedColsIdx[i] = 1 then
        inc(Result)
      else
        Break;
  end;

begin
  HG:=THackCustomDBGrid(SG);
  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    b:=PSFile.ReadBool(Key, sRowLines, dgRowLines in HG.Options);
    if b then
      HG.Options:=HG.Options + [dgRowLines]
    else
      HG.Options:=HG.Options - [dgRowLines];
    b:=PSFile.ReadBool(Key, sColLines, dgColLines in HG.Options);
    if b then
      HG.Options:=HG.Options + [dgColLines]
    else
      HG.Options:=HG.Options - [dgColLines];

    HG.GradientStartColor:=PSFile.ReadInteger(Key, sColorStartGradient, HG.GradientStartColor);
    HG.GradientEndColor:=PSFile.ReadInteger(Key, sColorEndGradient, HG.GradientEndColor);
    HG.DrawingStyle:=TGridDrawingStyle(PSFile.ReadInteger(Key, sDrawingStyle, ord(HG.DrawingStyle)));

    PSFile.ReadFont(Key, HG.Font, ScaleRatio);
    PSFile.ReadFont(Key + cTitleLineSubKey, HG.TitleFont, ScaleRatio);
  end;

  if FStoreColumns then
  begin
    // load saved cols
    SavedCols:=ReadDBColumnList(PSFile, Key);
    // load grid cols and store if col is not saved
    GridCols:=HG.Columns.Count;
    SetLength(ChangedColsIdx, GridCols);
    for gcIdx:=0 to GridCols - 1 do
    begin
      if SavedDBColumnFieldList.IndexOf(HG.Columns[gcIdx].FieldName) < 0 then
        ChangedColsIdx[gcIdx]:=1
      else
        ChangedColsIdx[gcIdx]:=0;
    end;

    for scIdx:=0 to SavedCols - 1 do
    begin
      k:=GetColumnKey(Key, scIdx);
      fn:=SavedDBColumnFieldList[scIdx];
      // find same column in grid (according fieldname)
      actIdx:=-1;
      for gcIdx:=0 to HG.Columns.Count - 1 do
        if HG.Columns[gcIdx].FieldName = fn then
          actIdx:=gcIdx;
      if (actIdx < 0) then
      begin
        // field not found - saved column is not exist already
        if FEraseUnusedSection then
          PSFile.EraseSection(k);
        Continue;
      end;

      // moved columns to new pos
      C:=HG.Columns[actIdx];
      actIdx:=scIdx + MoveIdx(scIdx);
      if actIdx >= GridCols then
        actIdx:=GridCols - 1;
      C.Index:=actIdx;

      ReadDBColumn(PSFile, Key, C, scIdx);

      if FEraseUnusedSection and (scIdx >= HG.Columns.Count) then
        PSFile.EraseSection(k);
    end;
  end;
end;

function TDBGridsPropSave.ReadDBColumnList(PSFile: TPropSaveFile; const Key: string): integer;
var
  i: integer;
begin
  Result:=PSFile.ReadInteger(Key, sColCount, -1);
  SavedDBColumnFieldList.Clear;
  for i:=0 to Result - 1 do
    SavedDBColumnFieldList.Add(PSFile.ReadString(GetColumnKey(Key, i), sFieldName, ''));
end;

procedure TDBGridsPropSave.ReadDBColumn(PSFile: TPropSaveFile; const Key: string; C: TColumn; Idx: integer);
var
  k: string;
  SavedInt: integer;
  SavedStr: string;
begin
  k:=GetColumnKey(Key, Idx);
  C.Width:=PSFile.ReadInteger(k, sColWidth, C.Width);
  C.Visible:=PSFile.ReadBool(k, sColVisible, C.Visible);

  // Read column props
  if FStoreColumnsProp then
  begin
    SavedInt:=PSFile.ReadInteger(k, sColColor, -1);
    if SavedInt <> -1 then
      C.Color:=SavedInt;
    PSFile.ReadFont(k, C.Font, ScaleRatio);
    SavedInt:=PSFile.ReadInteger(k, sColAlign, -1);
    if SavedInt in [ord( Low(TAlignment)) .. ord( High(TAlignment))] then
      C.Alignment:=TAlignment(SavedInt);

    SavedInt:=PSFile.ReadInteger(k, sColTitleColor, -1);
    if SavedInt <> -1 then
      C.Title.Color:=SavedInt;
    // PSFile.ReadFont(k,c.Title.Font,ScaleRatio);
    SavedInt:=PSFile.ReadInteger(k, sColTitleAlign, -1);
    if SavedInt in [ord( Low(TAlignment)) .. ord( High(TAlignment))] then
      C.Title.Alignment:=TAlignment(SavedInt);

    SavedStr:=PSFile.ReadString(k, sColTitle, '');
    if SavedStr <> '' then
      C.Title.Caption:=SavedStr;

    if C.Field <> nil then
    begin
      C.Field.Alignment:=C.Alignment;
      C.Field.DisplayLabel:=C.Title.Caption;
      if not FieldInList(C.Field.FieldName) then
      begin
        C.Field.DisplayWidth:=PSFile.ReadInteger(k, sFieldWidth, C.Field.DisplayWidth);
        if (C.Field is TNumericField) then
          TNumericField(C.Field).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TNumericField(C.Field).DisplayFormat);
        if (C.Field is TDateTimeField) then
          TDateTimeField(C.Field).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TDateTimeField(C.Field).DisplayFormat);
        if (C.Field is TSQLTimeStampField) then
          TSQLTimeStampField(C.Field).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TSQLTimeStampField(C.Field).DisplayFormat);
      end;
    end;
  end;
end;

procedure TDBGridsPropSave.DelDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TCustomDBGrid);
var
  ColIdx: integer;
begin
  PSFile.DeleteFont(Key);
  PSFile.DeleteFont(Key + cTitleLineSubKey);
  for ColIdx:=0 to THackCustomDBGrid(SG).Columns.Count - 1 do
    PSFile.EraseSection(GetColumnKey(Key, ColIdx));
end;

procedure TDBGridsPropSave.SaveDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
var
  i: integer;
  k: string;
  F: TField;
begin
  PSFile.WriteInteger(Key, sFieldCount, DS.Fields.Count);
  for i:=0 to DS.Fields.Count - 1 do
  begin
    F:=DS.Fields[i];
    k:=GetFieldKey(Key, i);
    // Save column position
    PSFile.WriteString(k, sFieldName, F.FieldName);
    PSFile.WriteInteger(k, sFieldWidth, F.DisplayWidth);
    PSFile.WriteBool(k, sFieldVisible, F.Visible);
    // Save column props
    if FStoreColumnsProp then
    begin
      PSFile.WriteInteger(k, sColAlign, ord(F.Alignment));
      PSFile.WriteString(k, sColTitle, F.DisplayLabel);
      if (F is TNumericField) then
        PSFile.WriteString(k, sFieldFormat, TNumericField(F).DisplayFormat);
      if (F is TDateTimeField) then
        PSFile.WriteString(k, sFieldFormat, TDateTimeField(F).DisplayFormat);
      if (F is TSQLTimeStampField) then
        PSFile.WriteString(k, sFieldFormat, TSQLTimeStampField(F).DisplayFormat);
    end;
  end;
end;

procedure TDBGridsPropSave.ReadDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
var
  i, b, p: integer;
  k, fn: string;
  F: TField;
  inList: boolean;
begin
  p:=PSFile.ReadInteger(Key, sFieldCount, -1);
  if p <> DS.Fields.Count then
    for i:=0 to 1000 do
    begin
      k:=GetFieldKey(Key, i);
      if PSFile.SectionExists(k) then
        PSFile.EraseSection(k)
      else
        Break;
    end
    else
      for i:=0 to p - 1 do
      begin
        k:=GetFieldKey(Key, i);
        // Read column position
        fn:=PSFile.ReadString(k, sFieldName, '');
        inList:=FieldInList(fn);
        for b:=0 to p - 1 do
          if DS.Fields[b].FieldName = fn then
            DS.Fields[b].Index:=i;
        F:=DS.Fields[i];
        if not inList then
          F.DisplayWidth:=PSFile.ReadInteger(k, sFieldWidth, F.DisplayWidth);
        F.Visible:=PSFile.ReadBool(k, sFieldVisible, F.Visible);

        // Read column props
        if FStoreColumnsProp then
        begin
          F.Alignment:=TAlignment(PSFile.ReadInteger(k, sColAlign, ord(F.Alignment)));
          F.DisplayLabel:=PSFile.ReadString(k, sColTitle, F.DisplayLabel);
          if (F is TNumericField) then
            TNumericField(F).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TNumericField(F).DisplayFormat);
          if (F is TDateTimeField) then
            TDateTimeField(F).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TDateTimeField(F).DisplayFormat);
          if (F is TSQLTimeStampField) then
            TSQLTimeStampField(F).DisplayFormat:=PSFile.ReadString(k, sFieldFormat, TSQLTimeStampField(F).DisplayFormat);
        end;
      end;
end;

procedure TDBGridsPropSave.DelDataSet(PSFile: TPropSaveFile; const Key: string; DS: TDataSet);
var
  i: integer;
begin
  for i:=0 to DS.Fields.Count - 1 do
    PSFile.EraseSection(GetFieldKey(Key, i));
end;
{$IFDEF rUseDBComp}

// ==============================================================================
// rDBGridsPropSave functions
function TrDBGridsPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:= inherited CanSave(Comp) or (Comp is TrDBGrid) or (Comp is TrDBRecView);
end;

function TrDBGridsPropSave.GetCellParamKey(const Key: string; Idx: integer): string;
begin
  Result:=Key + '.' + Format(sCellParamsCond, [Idx]);
end;

procedure TrDBGridsPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);
begin
  if (Comp is TrDBGrid) then
    ScaleRatio:=TrDBGrid(Comp).ScaleRatio;

  if (Comp is TrDBGrid) and (A = psaRead) then
  begin
    // read and set GridLookStyle before all other props (other props can override standard style)
    if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
      TrDBGrid(Comp).GridLookStyle:=TGridLookStyles(PSFile.ReadInteger(Key, sGridLookStyle, ord(TrDBGrid(Comp).GridLookStyle)));
  end;

  inherited ReadSaveProps(A, PSFile, Key, Comp);
  if (Comp is TrDBGrid) then
  begin
    case A of
      psaRead:
        ReadRDBGrid(PSFile, Key, Comp as TrDBGrid);
      psaSave:
        SaveRDBGrid(PSFile, Key, Comp as TrDBGrid);
      psaDelete:
        DelRDBGrid(PSFile, Key, Comp as TrDBGrid);
    end;
    if Assigned(FOnReadSaveProps) then
      FOnReadSaveProps(Comp as TrDBGrid, A, PSFile);
  end
  else
    if (Comp is TrDBRecView) then
    begin
      case A of
        psaRead:
          ReadRecView(PSFile, Key, Comp as TrDBRecView);
        psaSave:
          SaveRecView(PSFile, Key, Comp as TrDBRecView);
        psaDelete:
          DelRecView(PSFile, Key, Comp as TrDBRecView);
      end;
    end;
end;

procedure TrDBGridsPropSave.SaveRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
begin
  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
    PSFile.WriteInteger(Key, sGridLookStyle, ord(SG.GridLookStyle));

  PSFile.WriteString(Key, sSortField, SG.SortFieldName);
  PSFile.WriteBool(Key, sSortDesc, SG.SortDesc);
  PSFile.WriteString(Key, sSortDirections, SG.SortDirections);
  PSFile.WriteInteger(Key, sColumnWidth, ord(SG.ColumnWidth));
  PSFile.WriteInteger(Key, sRowDataLines, SG.RowLines);
  PSFile.WriteInteger(Key, sTitleLines, SG.TitleLines);
  PSFile.WriteBool(Key, sFilterButtons, SG.OptionsEx2.Filters.FunnelBtn);
  PSFile.WriteBool(Key, sFilterBar, SG.OptionsEx2.Filters.TextBar);

  PSFile.WriteInteger(Key, sRowDataLineColor, SG.OptionsEx2.Appearance.DataRowLineColor);
  PSFile.WriteInteger(Key, sRowDataDotLine, ord(SG.OptionsEx2.Appearance.DataRowLineStyle));

  PSFile.WriteInteger(Key, sFixedColumnsCount, SG.OptionsEx2.FixedCell.FixedCols);
  PSFile.WriteInteger(Key, sFixedColumnsColor, SG.OptionsEx2.FixedCell.FixedColColor);

  if SG.FixedColText.FixedColWidth <> IndicatorWidth then
  begin
    PSFile.WriteBool(Key, cFixedColParenFont, SG.FixedColText.ParentFont);
    if not SG.FixedColText.ParentFont then
      PSFile.WriteFont(Key + cFixedColSubKey, SG.FixedColText.Font, ScaleRatio);
    PSFile.WriteInteger(Key, cFixedColWidth, SG.FixedColText.FixedColWidth);
    PSFile.WriteString(Key, cFixedColTitleText, SG.FixedColText.TitleText);
    PSFile.WriteString(Key, cFixedColTextMask, SG.FixedColText.TextMask);
    PSFile.WriteInteger(Key, cFixedColTextAlign, ord(SG.FixedColText.TextAlign));
  end;

  SaveCellParamsCondList(PSFile, Key, SG.CellParamsCondList);
  PSFile.WriteBool(Key, sUseCellParamsConditions, SG.OptionsEx2.CellCondition.Active);

  if FStoreFieldsGroup then
  begin
    PSFile.WriteBool(Key, sFieldGroups, SG.OptionsEx2.Groups.Active);
    PSFile.WriteStrings(Key + cFieldsGroupSubKey, SG.FieldsGroups);
  end;

  if FStoreFooterRow then
  begin
    PSFile.WriteBool(Key, cFooterBoldText, SG.FooterRow.BoldText);
    PSFile.WriteInteger(Key, cFooterCellLinesColor, SG.FooterRow.CellLinesColor);
    PSFile.WriteInteger(Key, cFooterColor, SG.FooterRow.Color);
    PSFile.WriteStrings(Key + cFooterFieldFooterDefs, SG.FooterRow.FieldFooterDefs);
    PSFile.WriteString(Key, cFooterFixedColText, SG.FooterRow.FixedColText);
    PSFile.WriteFont(Key + cFooterFont, SG.FooterRow.Font);
    PSFile.WriteInteger(Key, cFooterFooterHeight, SG.FooterRow.FooterHeight);
    PSFile.WriteBool(Key, cFooterFooterVisible, SG.FooterRow.FooterVisible);
    PSFile.WriteBool(Key, cFooterParentFont, SG.FooterRow.ParentFont);
    PSFile.WriteBool(Key, cFooterShowCellLines, SG.FooterRow.ShowCellLines);
    PSFile.WriteString(Key, cFooterSimpleText, SG.FooterRow.SimpleText);
    PSFile.WriteInteger(Key, cFooterSimpleTextAlign, ord(SG.FooterRow.SimpleTextAlign));
  end;

  if FStoreDesignSetting then
  begin
    PSFile.WriteStrings(Key + cFieldsFieldsAsBoolean, SG.FieldsAsBoolean);
    PSFile.WriteStrings(Key + cFieldsFieldsAsColor, SG.FieldsAsColor);
    PSFile.WriteStrings(Key + cFieldsFieldsAsDate, SG.FieldsAsDate);
    PSFile.WriteStrings(Key + cFieldsFieldsAsGraphics, SG.FieldsAsGraphics);
    PSFile.WriteStrings(Key + cFieldsFieldsAsLink, SG.FieldsAsLink);
    PSFile.WriteStrings(Key + cFieldsFieldsAsPassword, SG.FieldsAsPassword);
    PSFile.WriteStrings(Key + cFieldsFieldsAutoPickList, SG.FieldsAutoPickList);
    PSFile.WriteStrings(Key + cFieldsFieldsDrawDropDown, SG.FieldsDrawDropDown);
    PSFile.WriteStrings(Key + cFieldsFieldsFilterButton, SG.FieldsFilterButton);
    PSFile.WriteStrings(Key + cTitleHints, SG.TitleHints);
    PSFile.WriteStrings(Key + cTitleImages, SG.TitleImages);
    PSFile.WriteStrings(Key + cTitlePopupItems, SG.TitlePopupItems);
    PSFile.WriteStrings(Key + cDataImages, SG.DataImages);
  end;

  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    PSFile.WriteBool(Key, sDrawColored, SG.OptionsEx2.Appearance.DrawColoredRow);
    PSFile.WriteInteger(Key, sColorOddRow, SG.ColorOddRow);
    PSFile.WriteInteger(Key, sColorActiveRow, SG.ColorActiveRow);
    PSFile.WriteInteger(Key, sFontColorFiltered, SG.FontColorFiltered);
  end;
end;

procedure TrDBGridsPropSave.ReadRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
begin
  SG.SortFieldName:=PSFile.ReadString(Key, sSortField, SG.SortFieldName);
  SG.SortDesc:=PSFile.ReadBool(Key, sSortDesc, SG.SortDesc);
  SG.SortDirections:=PSFile.ReadString(Key, sSortDirections, SG.SortDirections);
  SG.ColumnWidth:=TColumnWidth(PSFile.ReadInteger(Key, sColumnWidth, ord(SG.ColumnWidth)));
  SG.RowLines:=PSFile.ReadInteger(Key, sRowDataLines, SG.RowLines);
  SG.TitleLines:=PSFile.ReadInteger(Key, sTitleLines, SG.TitleLines);

  SG.OptionsEx2.Filters.FunnelBtn:=PSFile.ReadBool(Key, sFilterButtons, SG.OptionsEx2.Filters.FunnelBtn);
  SG.OptionsEx2.Filters.TextBar:=PSFile.ReadBool(Key, sFilterBar, SG.OptionsEx2.Filters.TextBar);

  SG.OptionsEx2.Appearance.DataRowLineColor:=PSFile.ReadInteger(Key, sRowDataLineColor, SG.OptionsEx2.Appearance.DataRowLineColor);
  SG.OptionsEx2.Appearance.DataRowLineStyle:=TPenStyle(PSFile.ReadInteger(Key, sRowDataDotLine, ord(SG.OptionsEx2.Appearance.DataRowLineStyle)));

  SG.OptionsEx2.FixedCell.FixedCols:=PSFile.ReadInteger(Key, sFixedColumnsCount, SG.OptionsEx2.FixedCell.FixedCols);
  SG.OptionsEx2.FixedCell.FixedColColor:=PSFile.ReadInteger(Key, sFixedColumnsColor, SG.OptionsEx2.FixedCell.FixedColColor);

  SG.FixedColText.FixedColWidth:=PSFile.ReadInteger(Key, cFixedColWidth, SG.FixedColText.FixedColWidth);
  if SG.FixedColText.FixedColWidth <> IndicatorWidth then
  begin
    SG.FixedColText.ParentFont:=PSFile.ReadBool(Key, cFixedColParenFont, SG.FixedColText.ParentFont);
    if not SG.FixedColText.ParentFont then
      PSFile.ReadFont(Key + cFixedColSubKey, SG.FixedColText.Font, SG.ScaleRatio);
    SG.FixedColText.TitleText:=PSFile.ReadString(Key, cFixedColTitleText, SG.FixedColText.TitleText);
    SG.FixedColText.TextMask:=PSFile.ReadString(Key, cFixedColTextMask, SG.FixedColText.TextMask);
    SG.FixedColText.TextAlign:=TAlignment(PSFile.ReadInteger(Key, cFixedColTextAlign, ord(SG.FixedColText.TextAlign)));
  end;

  ReadCellParamsCondList(PSFile, Key, SG.CellParamsCondList);
  SG.OptionsEx2.CellCondition.Active:=PSFile.ReadBool(Key, sUseCellParamsConditions, SG.OptionsEx2.CellCondition.Active);

  if FStoreFieldsGroup then
  begin
    SG.OptionsEx2.Groups.Active:=PSFile.ReadBool(Key, sFieldGroups, SG.OptionsEx2.Groups.Active);
    PSFile.ReadStrings(Key + cFieldsGroupSubKey, SG.FieldsGroups);
  end;

  if FStoreFooterRow then
  begin
    SG.FooterRow.BoldText:=PSFile.ReadBool(Key, cFooterBoldText, SG.FooterRow.BoldText);
    SG.FooterRow.CellLinesColor:=PSFile.ReadInteger(Key, cFooterCellLinesColor, SG.FooterRow.CellLinesColor);
    SG.FooterRow.Color:=PSFile.ReadInteger(Key, cFooterColor, SG.FooterRow.Color);
    PSFile.ReadStrings(Key + cFooterFieldFooterDefs, SG.FooterRow.FieldFooterDefs);
    SG.FooterRow.FixedColText:=PSFile.ReadString(Key, cFooterFixedColText, SG.FooterRow.FixedColText);
    PSFile.ReadFont(Key + cFooterFont, SG.FooterRow.Font, SG.ScaleRatio);
    SG.FooterRow.FooterHeight:=PSFile.ReadInteger(Key, cFooterFooterHeight, SG.FooterRow.FooterHeight);
    SG.FooterRow.FooterVisible:=PSFile.ReadBool(Key, cFooterFooterVisible, SG.FooterRow.FooterVisible);
    SG.FooterRow.ParentFont:=PSFile.ReadBool(Key, cFooterParentFont, SG.FooterRow.ParentFont);
    SG.FooterRow.ShowCellLines:=PSFile.ReadBool(Key, cFooterShowCellLines, SG.FooterRow.ShowCellLines);
    SG.FooterRow.SimpleText:=PSFile.ReadString(Key, cFooterSimpleText, SG.FooterRow.SimpleText);
    SG.FooterRow.SimpleTextAlign:=TAlignment(PSFile.ReadInteger(Key, cFooterSimpleTextAlign, ord(SG.FooterRow.SimpleTextAlign)));
  end;

  if FStoreDesignSetting then
  begin
    PSFile.ReadStrings(Key + cFieldsFieldsAsBoolean, SG.FieldsAsBoolean);
    PSFile.ReadStrings(Key + cFieldsFieldsAsColor, SG.FieldsAsColor);
    PSFile.ReadStrings(Key + cFieldsFieldsAsDate, SG.FieldsAsDate);
    PSFile.ReadStrings(Key + cFieldsFieldsAsGraphics, SG.FieldsAsGraphics);
    PSFile.ReadStrings(Key + cFieldsFieldsAsLink, SG.FieldsAsLink);
    PSFile.ReadStrings(Key + cFieldsFieldsAsPassword, SG.FieldsAsPassword);
    PSFile.ReadStrings(Key + cFieldsFieldsAutoPickList, SG.FieldsAutoPickList);
    PSFile.ReadStrings(Key + cFieldsFieldsDrawDropDown, SG.FieldsDrawDropDown);
    PSFile.ReadStrings(Key + cFieldsFieldsFilterButton, SG.FieldsFilterButton);
    PSFile.ReadStrings(Key + cTitleHints, SG.TitleHints);
    PSFile.ReadStrings(Key + cTitleImages, SG.TitleImages);
    PSFile.ReadStrings(Key + cTitlePopupItems, SG.TitlePopupItems);
    PSFile.ReadStrings(Key + cDataImages, SG.DataImages);
  end;

  if not RosiCompConfig.PropSaveCfg.Grids_UseGlobalSetting then
  begin
    SG.OptionsEx2.Appearance.DrawColoredRow:=PSFile.ReadBool(Key, sDrawColored, SG.OptionsEx2.Appearance.DrawColoredRow);
    SG.ColorOddRow:=PSFile.ReadInteger(Key, sColorOddRow, SG.ColorOddRow);
    SG.ColorActiveRow:=PSFile.ReadInteger(Key, sColorActiveRow, SG.ColorActiveRow);
    SG.FontColorFiltered:=PSFile.ReadInteger(Key, sFontColorFiltered, SG.FontColorFiltered);
  end;

  SG.FixedGroupedCol(false);
end;

procedure TrDBGridsPropSave.DelRDBGrid(PSFile: TPropSaveFile; const Key: string; SG: TrDBGrid);
var
  i: integer;
begin
  for i:=0 to SG.CellParamsCondList.Count - 1 do
    PSFile.EraseSection(GetCellParamKey(Key, i));
end;

procedure TrDBGridsPropSave.ReadCellParamsCondList(PSFile: TPropSaveFile; const Key: string; CondList: TCellParamsCondList);
var
  i: integer;
  k: string;
  cp: TCellParamsCond;
  cpcount: integer;
begin
  // load CellParams list
  cpcount:=PSFile.ReadInteger(Key, sCellParamsCondCount, 0);
  if cpcount > 0 then
    CondList.Clear;
  for i:=0 to cpcount - 1 do
  begin
    k:=GetCellParamKey(Key, i);
    cp:=TCellParamsCond.Create;
    cp.FieldName:=PSFile.ReadString(k, sCPField, '');
    cp.CondOperator:=TCellParamsCondOperators(PSFile.ReadInteger(k, sCPOperator, 0));
    cp.Value:=PSFile.ReadString(k, sCPValue, '');
    cp.Action:=TCellParamsCondActions(PSFile.ReadInteger(k, sCPAction, 0));
    cp.FontColor:=PSFile.ReadInteger(k, sCPFontColor, clNone);
    cp.BkColor:=PSFile.ReadInteger(k, sCPBkColor, clNone);
    cp.Bold:=PSFile.ReadBool(k, sCPBold, false);
    cp.WholeRow:=PSFile.ReadBool(k, sCPWholeRow, false);
    cp.ImageDef.UseDataImage:=PSFile.ReadBool(k, sCPUseDataImage, false);
    cp.ImageDef.ImageIndex:=PSFile.ReadInteger(k, sCPImageIndex, -1);
    cp.ImageDef.ImageAlignment:=PSFile.ReadString(k, sCPImageAlignment, '');
    cp.ImageDef.ShowText:=PSFile.ReadString(k, sCPShowText, '');
    cp.StopIfTrue:=PSFile.ReadBool(k, sCPStopIfTrue, false);
    cp.Deactive:=PSFile.ReadBool(k, sCPDeactive, false);
    CondList.Add(cp);
  end;
end;

procedure TrDBGridsPropSave.SaveCellParamsCondList(PSFile: TPropSaveFile; const Key: string; CondList: TCellParamsCondList);
var
  i: integer;
  k: string;
  cp: TCellParamsCond;
begin
  // save CellParams list
  PSFile.WriteInteger(Key, sCellParamsCondCount, CondList.Count);
  for i:=0 to CondList.Count - 1 do
  begin
    k:=GetCellParamKey(Key, i);
    cp:=TCellParamsCond(CondList[i]);
    PSFile.WriteString(k, sCPField, cp.FieldName);
    PSFile.WriteInteger(k, sCPOperator, ord(cp.CondOperator));
    PSFile.WriteString(k, sCPValue, cp.Value);
    PSFile.WriteInteger(k, sCPAction, ord(cp.Action));
    PSFile.WriteInteger(k, sCPFontColor, cp.FontColor);
    PSFile.WriteInteger(k, sCPBkColor, cp.BkColor);
    PSFile.WriteBool(k, sCPBold, cp.Bold);
    PSFile.WriteBool(k, sCPWholeRow, cp.WholeRow);
    PSFile.WriteBool(k, sCPUseDataImage, cp.ImageDef.UseDataImage);
    PSFile.WriteInteger(k, sCPImageIndex, cp.ImageDef.ImageIndex);
    PSFile.WriteString(k, sCPImageAlignment, cp.ImageDef.ImageAlignment);
    PSFile.WriteString(k, sCPShowText, cp.ImageDef.ShowText);
    PSFile.WriteBool(k, sCPStopIfTrue, cp.StopIfTrue);
    PSFile.WriteBool(k, sCPDeactive, cp.Deactive);
  end;

end;

procedure TrDBGridsPropSave.SaveRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
begin
  PSFile.WriteFont(Key + '.' + sColFont, RV.Font, RV.ScaleRatio);
  PSFile.WriteFont(Key + '.' + sSpecFont, RV.SpecFont, RV.ScaleRatio);
  PSFile.WriteStrings(Key + '.' + sFields, RV.Fields);
  PSFile.WriteStrings(Key + '.' + sSpecFields, RV.SpecFields);
  PSFile.WriteStrings(Key + '.' + sPickLists, RV.PickLists);
  // PSFile.WriteBool(Key,sOnlyVisibleField,RV.OnlyVisibleField);
  // PSFile.WriteBool(Key,sOnlySimpleField,RV.OnlySimpleField);
end;

procedure TrDBGridsPropSave.ReadRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
begin
  PSFile.ReadFont(Key + '.' + sColFont, RV.Font, RV.ScaleRatio);
  PSFile.ReadFont(Key + '.' + sSpecFont, RV.SpecFont, RV.ScaleRatio);
  PSFile.ReadStrings(Key + '.' + sFields, RV.Fields);
  PSFile.ReadStrings(Key + '.' + sSpecFields, RV.SpecFields);
  PSFile.ReadStrings(Key + '.' + sPickLists, RV.PickLists);
  // RV.OnlyVisibleField:=PSFile.ReadBool(Key,sOnlyVisibleField,RV.OnlyVisibleField);
  // RV.OnlySimpleField:=PSFile.ReadBool(Key,sOnlySimpleField,RV.OnlySimpleField);
  RV.SetRows;
  RV.Invalidate;
end;

procedure TrDBGridsPropSave.DelRecView(PSFile: TPropSaveFile; const Key: string; RV: TrDBRecView);
begin
end;
{$ENDIF}

end.
