unit PropSaveMain;

{<
  PropSaveMain unit contains set of non visual component which provides
  automatic storing and restoring properties of forms and components.
  It re/stores properties and position of all standard and Rosi components,
  any other components could be easily processed by TCustomPropSave descendant.

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2005)
  @lastmod(May 2019)
}

interface

{$INCLUDE _rosi.inc}

uses Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Math,
  Forms, Dialogs, StdCtrls, ExtCtrls, Menus, IniFiles, ComCtrls, CheckLst,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} StrUtils,
  RosiCompGlobalCfg;

type
  TPropSaveFile = class;
  TPropSaveMain = class;
  TCustomPropSave = class;
  TPositionPropSave = class;
  TStdCtrlPropSave = class;

  TGetEncodingEvent = procedure(var Encoding: TEncoding) of object; {< @exclude }
  TReadValueEvent = procedure(Section, Ident: string; var Value: variant) of object; {< @exclude }
  TWriteValueEvent = procedure(Section, Ident: string; var Value: variant) of object; {< @exclude }

  { Enhanced TMemIniFile which provides compression of file }
  TMemIniFileEx = class(TMemIniFile)
  private
    UseCompress: boolean;
  public
    procedure SetCompress(Compress: boolean; ShowErr: boolean); {< @exclude }
    { store changes to file }
    procedure UpdateFile; override;
    { return MemIni file content as string }
    function GetContentAsString: string;
    { load MemIni file content from string }
    procedure SetContentAsString(const s: string);
  end;

  { Non visual component which provides loading and saving of TMemIniFileEx file,
    it centralize data of all used TPropSaveMain components which uses it as interface to ini data }
  TPropSaveFile = class(TComponent)
  private
    IniFile: TMemIniFileEx;
    FFileName: string;
    FGetIniFileEncoding: TGetEncodingEvent;
    FDisableAll: boolean;
    FCompressFile: boolean;
    FUseAsGlobalFile: boolean;
    FOnLoadContent: TNotifyEvent;
    FOnSaveContent: TNotifyEvent;
    FOnReadValue: TReadValueEvent;
    FOnWriteValue: TWriteValueEvent;
    procedure CreateIniFile(const AFileName: string; Recreate: boolean);
    function SaveFileInt(const AFileName: string; ShowErr: boolean): boolean;
    procedure SetFileName(const Value: string);
    procedure SetCompressFile(Value: boolean);
    procedure SetUseAsGlobalFile(const Value: boolean);
  protected
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Loaded; override; {< @exclude }
    { load ini file from disk }
    procedure LoadFile(const AFileName: string);
    { load ini file from stream }
    procedure LoadFromStream(Source: TStream);
    { save ini file to disk }
    function SaveFile(const AFileName: string): boolean;
    { save ini file to disk, do not show error if saving fail }
    function SaveFileSilent(const AFileName: string): boolean;
    { save ini file to stream }
    procedure SaveToStream(Dest: TStream);
    { define whether section exists in ini file }
    function SectionExists(const Section: string): boolean;
    { define whether value exists in section in ini file }
    function ValueExists(const Section, Ident: string): boolean;
    { read whole section to stringlist }
    procedure ReadSection(const Section: string; Strings: TStrings);
    { write whole section from stringlist }
    procedure WriteSection(const Section: string; Strings: TStrings);
    { erase whole section - all values }
    procedure EraseSection(const Section: string);
    { erase one key value from the section }
    procedure DeleteKey(const Section, Ident: String);
    { erase whole ini file }
    procedure Clear;
    { load IniFile content to List }
    procedure GetStrings(List: TStrings);
    { store List to IniFile content }
    procedure SetStrings(List: TStrings);
    { find full section name according Line, mask with * can be used }
    function GetSectionFromLine(const Line: string): string;

    { return name of INI file }
    function GetFileName: string;
    function ReadVariant(const Section, Ident: string; Default: variant): variant;
    procedure WriteVariant(const Section, Ident: string; Value: variant);
    function ReadString(const Section, Ident, Default: string): string;
    procedure WriteString(const Section, Ident, Value: string);
    function ReadInteger(const Section, Ident: string; Default: Longint): Longint;
    procedure WriteInteger(const Section, Ident: string; Value: Longint);
    function ReadWord(const Section, Ident: string; Default: Word): Word;
    procedure WriteWord(const Section, Ident: string; Value: Word);
    function ReadBool(const Section, Ident: string; Default: boolean): boolean;
    procedure WriteBool(const Section, Ident: string; Value: boolean);
    function ReadDateTime(const Section, Ident: string; Default: TDateTime): TDateTime;
    procedure WriteDateTime(const Section, Ident: string; Value: TDateTime);
    function ReadFloat(const Section, Ident: string; Default: Double): Double;
    procedure WriteFloat(const Section, Ident: string; Value: Double);
    procedure WriteStrings(const Key: string; L: TStrings);
    procedure ReadStrings(const Key: string; L: TStrings);
    procedure DeleteStrings(const Key: string);
    procedure WriteFont(const Key: string; Font: TFont; ScaleRatio: Double = 1);
    procedure ReadFont(const Key: string; Font: TFont; ScaleRatio: Double = 1);
    procedure DeleteFont(const Key: string);
    { compress original file and store as new file }
    procedure DoCompressFile(const OrigFileName, DestFileName: string; DeleteOrigFile: boolean);
  published
    { Defines whether re/storing is completely disabled }
    property DisableAll: boolean read FDisableAll write FDisableAll default false;
    { Defines name of ini file, in runtime use LoadFile and SaveFile to change current name of ini file }
    property FileName: string read FFileName write SetFileName;
    { Defines whether ini file is commpressed, zlib or bzip2 is used according switch in _rosi.inc }
    property CompressFile: boolean read FCompressFile write SetCompressFile default false;
    { Defines whether the component is used as global PropSaveFile for all other shared forms }
    property UseAsGlobalFile: boolean read FUseAsGlobalFile write SetUseAsGlobalFile default false;
    { event is called to load ini content, if not defined then standard ini file is used }
    property OnLoadContent: TNotifyEvent read FOnLoadContent write FOnLoadContent;
    { event is called to save ini content, if not defined then standard ini file is used }
    property OnSaveContent: TNotifyEvent read FOnSaveContent write FOnSaveContent;
    { event is called after reading of any value }
    property OnReadValue: TReadValueEvent read FOnReadValue write FOnReadValue;
    { event is called before writing of any value }
    property OnWriteValue: TWriteValueEvent read FOnWriteValue write FOnWriteValue;
    { event is called during creating of INI file to define used encoding }
    property OnGetIniFileEncoding: TGetEncodingEvent read FGetIniFileEncoding write FGetIniFileEncoding;
  end;

  { list of possible saving events }
  TPropSaveEvents = (pseFormCreateDestroy, pseFormShowHide, pseNone);
  { list of saving actions }
  TPropSaveAction = (psaRead, psaSave, psaDelete);
  { event type used by TPropSaveMain }
  TOnReadSaveProp = procedure(PSFile: TPropSaveFile; Key: string) of object;

  { Main component which provides hook functions and re/stores standard components.
    It also finds all TCustomPropSave descendants on the form linked to PropSaveMain
    and call TCustomPropSave.ReadSaveProps procedure to re/store properties of other controls }
  TPropSaveMain = class(TComponent)
  private
    WasRead: boolean;
    oldOnRead: TNotifyEvent;
    oldOnSave: TNotifyEvent;
    oldOnActivate: TNotifyEvent;
    WasFormPosSaved: boolean;
    FFormKey: string;
    FormPosRect: TRect;
    FPropSaveFile: TPropSaveFile;
    FFormLimitToScreen: boolean;
    FFormSavePos: boolean;
    FPropSaveEvents: TPropSaveEvents;
    FCtrlPosition: TPositionPropSave;
    FStdCtrl: TStdCtrlPropSave;
    FOnSaveProperties: TOnReadSaveProp;
    FOnReadProperties: TOnReadSaveProp;
    procedure FormRead(Sender: TObject);
    procedure FormSave(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure ProcMain(psAction: TPropSaveAction; ForceEnabled: boolean = false);
    procedure SetFormEvent;
  protected
    OwnForm: TForm;
    procedure Loaded; override;
    procedure SetPropSaveFile(Value: TPropSaveFile);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    { indicates whether form was stored in maximized state }
    IsMaximized: boolean;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { define parent component (Form or Frame), needed only in special cases }
    procedure SetParent(AParent: TComponent; UseFormEvent: boolean = false);
    { returns key string for active form }
    function GetFormKey: string;
    { saves all properties on app request }
    procedure SaveProperties;
    { reads all properties on app request }
    procedure ReadProperties;
    { delete all properties on app request }
    procedure DeleteProperties;
    { indicates whether form was already stored }
    function FormIsSaved: boolean;
  published
    { key string for the form, if empty Owner form name is used }
    property FormKey: string read FFormKey write FFormKey;
    { link to TPropSaveFile, usually app use only one PropSaveFile for all forms }
    property PropSaveFile: TPropSaveFile read FPropSaveFile write SetPropSaveFile;
    { defines whether form position is re/stored }
    property FormSavePos: boolean read FFormSavePos write FFormSavePos default true;
    { defines whether form position is limited to screen after restoring,
      ie. when screen resolution was changed }
    property FormLimitToScreen: boolean read FFormLimitToScreen write FFormLimitToScreen default true;
    { PositionPropSave subcomponent, see detail for TPositionPropSave }
    property CtrlPosition: TPositionPropSave read FCtrlPosition;
    { StdCtrlPropSave subcomponent, see detail for TStdCtrlPropSave }
    property StdCtrl: TStdCtrlPropSave read FStdCtrl;
    { defines form events, when re/storing of properties is done,
      if pseNone is set then procedure SaveProperties and ReadProperties has to be called by code }
    property PropSaveEvents: TPropSaveEvents read FPropSaveEvents write FPropSaveEvents default pseFormCreateDestroy;
    { called after standard storing and could be used for storing of user defined properties }
    property OnSaveProperties: TOnReadSaveProp read FOnSaveProperties write FOnSaveProperties;
    { called after standard restoring and could be used for restoring of user defined properties }
    property OnReadProperties: TOnReadSaveProp read FOnReadProperties write FOnReadProperties;
  end;

  { event type used by TCustomPropSave to get KeyString for the component }
  TOnGetKeyString = procedure(Comp: TComponent; var Key: string) of object;
  { event type used by TCustomPropSave to proceed any custom defined component }
  TOnProcComponents = procedure(Sender: TCustomPropSave; psAction: TPropSaveAction) of object;

  { Basic component which provides all functions for re/storing of properties;
    all PropSave components must be inherited from TCustomPropSave }
  TCustomPropSave = class(TComponent)
  private
    FDisable: boolean;
    FComponents: TStrings;
    FDisableComponents: TList;
    FPropSaveMain: TPropSaveMain;
    FOnGetKeyString: TOnGetKeyString;
    FOnProcComponents: TOnProcComponents;
    procedure SetComponents(Value: TStrings);
  protected
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    { Procedure is called by TCustomPropSave for each "Comp" component
      in order to re/storing "Comp" properties, this procedure must be
      overrided by each CustomPropSave descendant }
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); virtual; abstract;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { returns key string for component }
    function GetKey(Comp: TComponent): string;
    { indicates whether re/storing of Components from parameter is supported
      by descendant of TCustomPropSave, must be overridden for all descendand }
    function CanSave(Comp: TComponent): boolean; virtual; abstract;
    { called by MainPropSave in order to proceed re/storing
      properties of all components in "Componets" property }
    procedure ProcComps(psAction: TPropSaveAction);
    { called by ProcComps in order to proceed re/storing
      properties of one particular component }
    procedure ProcComp(psAction: TPropSaveAction; Comp: TComponent; ProcChild: boolean);
    { indicates whether some data exists for component in ini file }
    function PropsIsStored(Comp: TComponent): boolean;
    { list of component names which is excluded from re/storing and stored values is deleted }
    property DisableComponents: TList read FDisableComponents;
  published
    { link to TPropSaveMain which maintain re/storing on form level }
    property PropSaveMain: TPropSaveMain read FPropSaveMain write FPropSaveMain;
    { defines whether re/storing function of this CustomPropSave is disabled }
    property Disable: boolean read FDisable write FDisable default false;
    { defines names of components which is re/stored, for re/storing of all
      children components use asterisk convention, ie. "Form1.*" means that
      all components on form "Form1" is re/stored;
      can be easily set by design time dialog (see RosiDBDesignEditors package) }
    property Components: TStrings read FComponents write SetComponents;
    { event called for each component, can be used to redefine KeyString by user }
    property OnGetKeyString: TOnGetKeyString read FOnGetKeyString write FOnGetKeyString;
    { event called by ProcComps to allow storing of any custom defined component }
    property OnProcComponents: TOnProcComponents read FOnProcComponents write FOnProcComponents;
  end;

  { Object for setting of limit rectangle properties }
  { @exclude }
  TLimitRect = class(TPersistent)
  private
    FLeft, FTop, FRight, FBottom: integer;
    FMinWidthHeight: integer;
  published
    property Left: integer read FLeft write FLeft default 0;
    property Top: integer read FTop write FTop default 0;
    property Right: integer read FRight write FRight default 0;
    property Bottom: integer read FBottom write FBottom default 0;
    property MinWidthHeight: integer read FMinWidthHeight write FMinWidthHeight default 30;
  end;

  { provides re/storing of position of TControl descendand components }
  TPositionPropSave = class(TCustomPropSave)
  private
    FLimitToRect: boolean;
    FLimitRect: TLimitRect;
    function CheckLimit(R: TRect): TRect;
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { indicates whether re/storing of Components from parameter is supported by TPositionPropSave }
    function CanSave(Comp: TComponent): boolean; override;
  published
    { defines whether position of components are limited to "LimitRect" after restoring }
    property LimitToRect: boolean read FLimitToRect write FLimitToRect default false;
    { defines Limit rectangle, component after restoring will be moved to rectangle
      if is out of rectangle }
    property LimitRect: TLimitRect read FLimitRect write FLimitRect;
  end;

  { provides re/storing of properties of standard components }
  TStdCtrlPropSave = class(TCustomPropSave)
  private
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    { indicates whether re/storing of Components from parameter is supported by TStdCtrlPropSave }
    function CanSave(Comp: TComponent): boolean; override;
  published
  end;

  { return link to global TPropSaveFile if defined }
function GetGlobal_PropSaveFile: TPropSaveFile;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, zlib_interface;

const
  sTabReplace = '_#9_';

  sPos = '_XY';
  sLeft = 'Left';
  sRight = 'Right';
  sTop = 'Top';
  sBottom = 'Bottom';
  sWinMax = 'WinMinMax';

  sItemsCount = 'ItemsCount';
  sItemIndex = 'ItemIndex';
  sItemSel = 'Selected%d';
  sItemChecked = 'Checked%d';
  sItem = 'Item%d';

  sText = 'Text';
  sState = 'State';
  sChecked = 'Checked';

  sCharset = 'FontCharset';
  sFontColor = 'FontColor';
  sFontName = 'FontName';
  sFontSize = 'FontSize';
  sFontHeight = 'FontHeight';
  sFontStyle = 'FontStyle';
  sColor = 'Color';
  sDate = 'Date';
  sDateChecked = 'DateChecked';
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi PS', [TPropSaveFile, TPropSaveMain, TPositionPropSave]);
end;

function GetGlobal_PropSaveFile: TPropSaveFile;
begin
  if Assigned(RosiCompGlobalCfg.RosiCompConfig.Global_PropSaveFile) and (RosiCompGlobalCfg.RosiCompConfig.Global_PropSaveFile is TPropSaveFile) then
    Result:=TPropSaveFile(RosiCompGlobalCfg.RosiCompConfig.Global_PropSaveFile)
  else
    Result:=nil;
end;

// ==============================================================================
// TMemIniFileEx functions
procedure TMemIniFileEx.SetCompress(Compress: boolean; ShowErr: boolean); {< @exclude }
begin
  UseCompress:=Compress;
  try
    if UseCompress and (FileName <> '') and FileExists(FileName) then
      SetContentAsString(DeCompressString('', FileName));
  except
    if ShowErr then
      ShowMessage(Format(sFileOpenErr, [FileName]));
  end;
end;

procedure TMemIniFileEx.UpdateFile;
begin
  if UseCompress then
    CompressString(GetContentAsString, FileName)
  else
    inherited UpdateFile;
end;

function TMemIniFileEx.GetContentAsString: string;
var
  L: TStrings;
begin
  L:=TStringList.Create;
  try
    GetStrings(L);
    Result:=L.Text;
  finally
    FreeAndNil(L);
  end;
end;

procedure TMemIniFileEx.SetContentAsString(const s: string);
var
  L: TStrings;
begin
  L:=TStringList.Create;
  try
    L.Text:=s;
    SetStrings(L);
  finally
    FreeAndNil(L);
  end;
end;

// ==============================================================================
// PropSaveFile functions
constructor TPropSaveFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFileName:='';
  FDisableAll:=false;
  FCompressFile:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TPropSaveFile.Destroy;
begin
  if Assigned(IniFile) then
  begin
    if not FDisableAll and not(csDesigning in ComponentState) then
    begin
      if Assigned(FOnSaveContent) then
        FOnSaveContent(Self)
      else
        SaveFile('');
    end;
    FreeAndNil(IniFile);
  end;
  inherited Destroy;
end;

procedure TPropSaveFile.Loaded;
begin
  inherited Loaded;
  if not(csDesigning in ComponentState) then
  begin
    if Assigned(FOnLoadContent) then
    begin
      CreateIniFile('', true);
      FOnLoadContent(Self);
    end
    else
      LoadFile('');
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TPropSaveFile.CreateIniFile(const AFileName: string; Recreate: boolean);
var
  Encoding: TEncoding;
begin
  if Assigned(IniFile) then
  begin
    if Recreate then
      FreeAndNil(IniFile)
    else
      Exit;
  end;
  Encoding:=nil;
  if Assigned(FGetIniFileEncoding) then
    FGetIniFileEncoding(Encoding);
  IniFile:=TMemIniFileEx.Create(AFileName, Encoding);
  IniFile.SetCompress(FCompressFile, true);
end;

procedure TPropSaveFile.SetFileName(const Value: string);
begin
  FFileName:=Value;
  if (FFileName <> '') and Assigned(IniFile) then
  begin
    IniFile.Rename(FFileName, false);
  end;
end;

procedure TPropSaveFile.SetCompressFile(Value: boolean);
begin
  FCompressFile:=Value;
  if Assigned(IniFile) then
    IniFile.UseCompress:=FCompressFile;
end;

procedure TPropSaveFile.SetUseAsGlobalFile(const Value: boolean);
begin
  if FUseAsGlobalFile <> Value then
  begin
    FUseAsGlobalFile:=Value;
    if FUseAsGlobalFile then
    begin
      RosiCompGlobalCfg.RosiCompConfig.Global_PropSaveFile:=Self;
    end;
  end;
end;

procedure TPropSaveFile.DoCompressFile(const OrigFileName, DestFileName: string; DeleteOrigFile: boolean);
var
  oldComp: boolean;
begin
  if FileExists(OrigFileName) then
  begin
    oldComp:=FCompressFile;
    try
      CompressFile:=false;
      LoadFile(OrigFileName);
      CompressFile:=true;
      SaveFile(DestFileName);
      if DeleteOrigFile then
        DeleteFile(OrigFileName);
    finally
      CompressFile:=oldComp;
    end;
  end;
end;

procedure TPropSaveFile.LoadFile(const AFileName: string);
var
  fn: string;
begin
  fn:=AFileName;
  if (fn = '') and Assigned(IniFile) then
    fn:=IniFile.FileName;
  if fn = '' then
  begin
    if FFileName = '' then
    begin
      if FCompressFile then
        fn:=ChangeFileExt(Application.ExeName, '.ini.compress')
      else
        fn:=ChangeFileExt(Application.ExeName, '.ini');
    end
    else
    begin
      fn:=FFileName;
      if ExtractFilePath(fn) = '' then
        fn:=GetAppDir + fn;
    end;
  end;
  CreateIniFile(fn, true);
end;

procedure TPropSaveFile.LoadFromStream(Source: TStream);
var
  s: string;
  Dest: TStringStream;
begin
  Dest:=TStringStream.Create;
  try
    Dest.CopyFrom(Source, Source.Size);
    s:=Dest.DataString;
    if Self.FCompressFile then
      s:=DeCompressString(s);
    CreateIniFile('', false);
    IniFile.SetContentAsString(s);
  finally
    FreeAndNil(Dest);
  end;
end;

function TPropSaveFile.SaveFileInt(const AFileName: string; ShowErr: boolean): boolean;
begin
  Result:=true;
  if not Assigned(IniFile) then
    Exit;
  if AFileName <> '' then
  begin
    FileName:=AFileName;
    // IniFile.Rename(AFileName, false);
  end;
  if FDisableAll then
    Exit;
  try
    if (FileName <> '') and (ExtractFilePath(FileName) = '') then
      IniFile.Rename(GetAppDir + FileName, false);
    IniFile.UpdateFile;
  except
    Result:=false;
    if ShowErr then
      ShowMessage(Format(sFileSaveErr, [GetFileName]));
  end;
end;

procedure TPropSaveFile.SaveToStream(Dest: TStream);
var
  s: string;
  Source: TStringStream;
begin
  if FDisableAll then
    Exit;
  s:=IniFile.GetContentAsString;
  if FCompressFile then
    s:=CompressString(s);
  Source:=TStringStream.Create(s);
  try
    Dest.CopyFrom(Source, Source.Size);
  finally
    FreeAndNil(Source);
  end;
end;

function TPropSaveFile.SaveFileSilent(const AFileName: string): boolean;
begin
  Result:=SaveFileInt(AFileName, false);
end;

function TPropSaveFile.SaveFile(const AFileName: string): boolean;
begin
  Result:=SaveFileInt(AFileName, true);
end;

function TPropSaveFile.GetFileName: string;
begin
  if Assigned(IniFile) then
    Result:=IniFile.FileName
  else
    Result:='';
end;

function TPropSaveFile.SectionExists(const Section: string): boolean;
begin
  Result:=IniFile.SectionExists(Section);
end;

function TPropSaveFile.ValueExists(const Section, Ident: string): boolean;
begin
  Result:=IniFile.ValueExists(Section, Ident);
end;

procedure TPropSaveFile.ReadSection(const Section: string; Strings: TStrings);
begin
  IniFile.ReadSectionValues(Section, Strings);
end;

procedure TPropSaveFile.WriteSection(const Section: string; Strings: TStrings);
var
  i, p: integer;
  s: string;
begin
  for i:=0 to Strings.Count - 1 do
  begin
    s:=Strings[i];
    p:=Pos('=', s);
    if p > 0 then
      IniFile.WriteString(Section, copy(s, 1, p - 1), copy(s, p + 1, MaxInt))
    else
      IniFile.WriteString(Section, s, '');
  end;
end;

procedure TPropSaveFile.EraseSection(const Section: string);
begin
  IniFile.EraseSection(Section);
end;

procedure TPropSaveFile.DeleteKey(const Section, Ident: String);
begin
  IniFile.DeleteKey(Section, Ident);
end;

procedure TPropSaveFile.Clear;
begin
  IniFile.Clear;
end;

procedure TPropSaveFile.GetStrings(List: TStrings);
begin
  IniFile.GetStrings(List);
end;

procedure TPropSaveFile.SetStrings(List: TStrings);
begin
  CreateIniFile('', false);
  IniFile.SetStrings(List);
end;

function TPropSaveFile.GetSectionFromLine(const Line: string): string;
var
  Sections, Vals: TStrings;
  i, j: integer;
  mask: boolean;
begin
  Result:='';
  mask:=EndsText(Line, '*');
  Sections:=TStringList.Create;
  try
    IniFile.ReadSections(Sections);
    for i:=0 to Sections.Count - 1 do
    begin
      Vals:=TStrings(Sections.Objects[i]);
      for j:=0 to Vals.Count - 1 do
        if (mask and (AnsiPosEx(copy(Line, 1, Length(Line) - 1), Vals[j]) = 1)) or SameText(Line, Vals[j]) then
        begin
          Result:=Sections[i];
          Exit;
        end;
    end;
  finally
    FreeAndNil(Sections);
  end;
end;

function TPropSaveFile.ReadVariant(const Section, Ident: string; Default: variant): variant;
begin
  case VarType(Default) of
    varSmallint, varShortInt, varByte, varWord, varLongWord, varInt64, varInteger:
      Result:=IniFile.ReadInteger(Section, Ident, Default);
    varSingle, varDouble, varCurrency:
      Result:=IniFile.ReadFloat(Section, Ident, Default);
    varDate:
      Result:=IniFile.ReadDateTime(Section, Ident, Default);
    varBoolean:
      Result:=IniFile.ReadBool(Section, Ident, Default);
  else
    begin
      Result:=IniFile.ReadString(Section, Ident, Default);
      Result:=rtool.ReplaceStr(sTabReplace, #9, Result, true, true);
    end;
  end;
  if Assigned(FOnReadValue) then
    FOnReadValue(Section, Ident, Result);
end;

procedure TPropSaveFile.WriteVariant(const Section, Ident: string; Value: variant);
begin
  if Assigned(FOnWriteValue) then
    FOnWriteValue(Section, Ident, Value);
  if VarIsNull(Value) then
    IniFile.DeleteKey(Section, Ident)
  else
    case VarType(Value) of
      varSmallint, varShortInt, varByte, varWord, varLongWord, varInt64, varInteger:
        IniFile.WriteInteger(Section, Ident, Value);
      varSingle, varDouble, varCurrency:
        IniFile.WriteFloat(Section, Ident, Value);
      varDate:
        IniFile.WriteDateTime(Section, Ident, Value);
      varBoolean:
        IniFile.WriteBool(Section, Ident, Value);
    else
      begin
        Value:=rtool.ReplaceStr(#9, sTabReplace, Value, true, true);
        IniFile.WriteString(Section, Ident, Value);
      end;
    end;
end;

function TPropSaveFile.ReadString(const Section, Ident, Default: string): string;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteString(const Section, Ident, Value: string);
begin
  WriteVariant(Section, Ident, Value);
end;

function TPropSaveFile.ReadInteger(const Section, Ident: string; Default: Longint): Longint;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteInteger(const Section, Ident: string; Value: Longint);
begin
  WriteVariant(Section, Ident, Value);
end;

function TPropSaveFile.ReadWord(const Section, Ident: string; Default: Word): Word;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteWord(const Section, Ident: string; Value: Word);
begin
  WriteVariant(Section, Ident, Value);
end;

function TPropSaveFile.ReadBool(const Section, Ident: string; Default: boolean): boolean;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteBool(const Section, Ident: string; Value: boolean);
begin
  WriteVariant(Section, Ident, Value);
end;

function TPropSaveFile.ReadDateTime(const Section, Ident: string; Default: TDateTime): TDateTime;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteDateTime(const Section, Ident: string; Value: TDateTime);
begin
  WriteVariant(Section, Ident, Value);
end;

function TPropSaveFile.ReadFloat(const Section, Ident: string; Default: Double): Double;
begin
  Result:=ReadVariant(Section, Ident, Default);
end;

procedure TPropSaveFile.WriteFloat(const Section, Ident: string; Value: Double);
begin
  WriteVariant(Section, Ident, Value);
end;

procedure TPropSaveFile.WriteStrings(const Key: string; L: TStrings);
var
  i: integer;
begin
  WriteInteger(Key, sItemsCount, L.Count);
  for i:=0 to L.Count - 1 do
    WriteString(Key, Format(sItem, [i]), L[i]);
end;

procedure TPropSaveFile.ReadStrings(const Key: string; L: TStrings);
var
  i, c: integer;
begin
  c:=ReadInteger(Key, sItemsCount, -1);
  if c = -1 then
    Exit;
  L.Clear;
  for i:=0 to c - 1 do
    L.Add(ReadString(Key, Format(sItem, [i]), ''));
end;

procedure TPropSaveFile.DeleteStrings(const Key: string);
var
  i, c: integer;
begin
  c:=ReadInteger(Key, sItemsCount, -1);
  for i:=c - 1 downto 0 do
    DeleteKey(Key, Format(sItem, [i]));
  DeleteKey(Key, sItemsCount);
end;

procedure TPropSaveFile.WriteFont(const Key: string; Font: TFont; ScaleRatio: Double = 1);
begin
  WriteInteger(Key, sCharset, Font.Charset);
  WriteInteger(Key, sFontColor, Font.Color);
  WriteString(Key, sFontName, Font.Name);
  WriteInteger(Key, sFontHeight, Round(Font.Height / ScaleRatio));
  WriteInteger(Key, sFontSize, Round(Font.Size / ScaleRatio));
  WriteInteger(Key, sFontStyle, Byte(fsBold in Font.Style) + Byte(fsItalic in Font.Style) * 2 + Byte(fsUnderLine in Font.Style) * 4 + Byte
      (fsStrikeOut in Font.Style) * 8);
end;

procedure TPropSaveFile.ReadFont(const Key: string; Font: TFont; ScaleRatio: Double = 1);
var
  A: integer;
begin
  Font.Charset:=ReadInteger(Key, sCharset, Font.Charset);
  Font.Color:=ReadInteger(Key, sFontColor, Font.Color);
  Font.Name:=ReadString(Key, sFontName, Font.Name);
  if ValueExists(Key, sFontHeight) then
    Font.Height:=Round(ReadInteger(Key, sFontHeight, Font.Height) * ScaleRatio)
  else
    if ValueExists(Key, sFontSize) then
      Font.Size:=Round(ReadInteger(Key, sFontSize, Font.Size) * ScaleRatio);
  A:=ReadInteger(Key, sFontStyle, -1);
  if A = -1 then
    Exit;
  Font.Style:=[];
  if A and 1 > 0 then
    Font.Style:=Font.Style + [fsBold];
  if A and 2 > 0 then
    Font.Style:=Font.Style + [fsItalic];
  if A and 4 > 0 then
    Font.Style:=Font.Style + [fsUnderLine];
  if A and 8 > 0 then
    Font.Style:=Font.Style + [fsStrikeOut];
end;

procedure TPropSaveFile.DeleteFont(const Key: string);
begin
  DeleteKey(Key, sCharset);
  DeleteKey(Key, sFontColor);
  DeleteKey(Key, sFontName);
  DeleteKey(Key, sFontSize);
  DeleteKey(Key, sFontStyle);
end;

// ==============================================================================
// PropSaveMain functions
constructor TPropSaveMain.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FormSavePos:=true;
  FormLimitToScreen:=true;
  WasRead:=false;
  SetParent(AOwner);
  // create subcomponents
  FCtrlPosition:=TPositionPropSave.Create(Self);
  FCtrlPosition.Name:='CtrlPosition';
  FCtrlPosition.PropSaveMain:=Self;
  FCtrlPosition.SetSubComponent(true);
  FStdCtrl:=TStdCtrlPropSave.Create(Self);
  FStdCtrl.Name:='StdCtrl';
  FStdCtrl.PropSaveMain:=Self;
  FStdCtrl.SetSubComponent(true);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TPropSaveMain.Destroy;
begin
  OwnForm:=nil;
  FreeAndNil(FCtrlPosition);
  FreeAndNil(FStdCtrl);
  inherited Destroy;
end;

procedure TPropSaveMain.Loaded;
begin
  inherited Loaded;
  SetFormEvent;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TPropSaveMain.SetParent(AParent: TComponent; UseFormEvent: boolean = false);
begin
  // link owner Form
  if AParent is TForm then
    OwnForm:=TForm(AParent)
  else
    if AParent.Owner is TForm then
      OwnForm:=TForm(AParent.Owner)
    else
      OwnForm:=nil;
  if UseFormEvent then
    SetFormEvent;
end;

procedure TPropSaveMain.SetFormEvent;
begin
  if (OwnForm = nil) or (csDesigning in ComponentState) then
    Exit;
  // hook PropSave functions to Form events
  case FPropSaveEvents of
    pseFormCreateDestroy:
      begin
        oldOnRead:=OwnForm.OnCreate;
        OwnForm.OnCreate:=FormRead;
        oldOnSave:=OwnForm.OnDestroy;
        OwnForm.OnDestroy:=FormSave;
        oldOnActivate:=OwnForm.OnActivate;
        OwnForm.OnActivate:=FormActivate;
      end;
    pseFormShowHide:
      begin
        oldOnRead:=OwnForm.OnShow;
        OwnForm.OnShow:=FormRead;
        oldOnSave:=OwnForm.OnHide;
        OwnForm.OnHide:=FormSave;
        oldOnActivate:=OwnForm.OnActivate;
        OwnForm.OnActivate:=FormActivate;
      end;
  end;
end;

procedure TPropSaveMain.SetPropSaveFile(Value: TPropSaveFile);
begin
  FPropSaveFile:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TPropSaveMain.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FPropSaveFile) then
    FPropSaveFile:=nil;
end;

procedure TPropSaveMain.FormRead(Sender: TObject);
begin
  if not WasRead then
    ProcMain(psaRead);
  WasRead:=true;
  if Assigned(oldOnRead) then
    oldOnRead(Sender);
end;

procedure TPropSaveMain.FormSave(Sender: TObject);
begin
  if Assigned(oldOnSave) then
    oldOnSave(Sender);
  ProcMain(psaSave);
  WasRead:=false;
end;

procedure TPropSaveMain.FormActivate(Sender: TObject);
begin
  if Assigned(OwnForm) then
  begin
    // restore position of Form
    // if (PropSaveEvents = pseFormShowHide) and WasFormPosSaved then
    if WasFormPosSaved then
    begin
      WasFormPosSaved:=false;
      OwnForm.BoundsRect:=FormPosRect;
    end;
    // restore state of Form
    if IsMaximized then
    begin
      IsMaximized:=false;
      OwnForm.WindowState:=wsMaximized;
    end;
  end;
  if Assigned(oldOnActivate) then
    oldOnActivate(Sender);
end;

function TPropSaveMain.GetFormKey: string;
begin
  if FFormKey = '' then
    Result:=OwnForm.Name
  else
    Result:=FFormKey;
end;

function TPropSaveMain.FormIsSaved: boolean;
begin
  Result:=Assigned(FPropSaveFile) and not FPropSaveFile.DisableAll and FPropSaveFile.SectionExists(GetFormKey + sPos);
end;

procedure TPropSaveMain.ProcMain(psAction: TPropSaveAction; ForceEnabled: boolean = false);
var
  i, origW, origH: integer;
  FormPS: TPositionPropSave;
begin
  WasFormPosSaved:=false;
  if not Assigned(OwnForm) or not Assigned(FPropSaveFile) then
    Exit;
  if not ForceEnabled and FPropSaveFile.DisableAll then
    Exit;
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  // read/save form position and state
  if FFormSavePos then
  begin
    WasFormPosSaved:=FormIsSaved;
    FormPS:=TPositionPropSave.Create(Self);
    try
      FormPS.PropSaveMain:=Self;
      FormPS.LimitToRect:=FormLimitToScreen;
      FormPS.FLimitRect.FMinWidthHeight:=120;
      FormPS.FLimitRect.Top:=Screen.DesktopRect.Top;
      FormPS.FLimitRect.Left:=Screen.DesktopRect.Left;
      FormPS.FLimitRect.Right:=Screen.DesktopRect.Right;
      FormPS.FLimitRect.Bottom:=Screen.DesktopRect.Bottom;

      case psAction of
        psaRead:
          begin
            if WasFormPosSaved then
            begin
              // change default pos to limit flickering (only in Create event can be changed)
              if PropSaveEvents = pseFormCreateDestroy then
                OwnForm.Position:=poDesigned;
              if OwnForm.WindowState <> wsNormal then
                OwnForm.WindowState:=wsNormal;
              // read form state
              i:=FPropSaveFile.ReadInteger(GetFormKey + sPos, sWinMax, ord(OwnForm.WindowState));
              IsMaximized:=TWindowState(i) = wsMaximized;
            end;
          end;
        psaSave:
          begin
            // save form state
            FPropSaveFile.WriteInteger(GetFormKey + sPos, sWinMax, ord(OwnForm.WindowState));
          end;
        psaDelete:
          begin
            // save form state
            FPropSaveFile.DeleteKey(GetFormKey + sPos, sWinMax);
          end;
      end;

      // store original width and height
      origW:=OwnForm.Width;
      origH:=OwnForm.Height;
      // do not save if win is maximized
      if (psAction = psaRead) or (OwnForm.WindowState = wsNormal) or not WasFormPosSaved then
        FormPS.ReadSaveProps(psAction, FPropSaveFile, GetFormKey, OwnForm);
      // restore original width and height for dialog (protect size if dialog was resized)
      if (psAction = psaRead) and (OwnForm.WindowState = wsNormal) and (OwnForm.BorderStyle in [bsDialog, bsNone, bsSingle]) then
      begin
        // restore default size
        OwnForm.Width:=origW;
        OwnForm.Height:=origH;
      end;
      // store new position
      FormPosRect:=OwnForm.BoundsRect;
    finally
      FreeAndNil(FormPS);
    end;
  end;

  // process CtrlPosition of PropSaveMain
  CtrlPosition.ProcComps(psAction);
  // process StdCtrl of PropSaveMain
  StdCtrl.ProcComps(psAction);
  // find own CustomPropSave components on Form and process ProcComps for each
  for i:=0 to Owner.ComponentCount - 1 do
    if (Owner.Components[i] is TCustomPropSave) and (TCustomPropSave(Owner.Components[i]).FPropSaveMain = Self) then
    begin
      TCustomPropSave(Owner.Components[i]).ProcComps(psAction);
    end;
  // call User OnRead/OnSave procedures
  if Assigned(FOnReadProperties) and (psAction = psaRead) then
    FOnReadProperties(FPropSaveFile, GetFormKey);
  if Assigned(FOnSaveProperties) and (psAction = psaSave) then
    FOnSaveProperties(FPropSaveFile, GetFormKey);
end;

procedure TPropSaveMain.SaveProperties;
begin
  ProcMain(psaSave, true);
end;

procedure TPropSaveMain.ReadProperties;
begin
  ProcMain(psaRead, true);
end;

procedure TPropSaveMain.DeleteProperties;
begin
  ProcMain(psaDelete, true);
end;

// ==============================================================================
// CustomPropSave functions
constructor TCustomPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FComponents:=TStringList.Create;
  FDisableComponents:=TList.Create;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TCustomPropSave.Destroy;
begin
  FreeAndNil(FComponents);
  FreeAndNil(FDisableComponents);
  inherited Destroy;
end;

procedure TCustomPropSave.Loaded;
begin
  inherited Loaded;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TCustomPropSave.SetComponents(Value: TStrings);
begin
  FComponents.Assign(Value);
end;

procedure TCustomPropSave.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FPropSaveMain) then
    FPropSaveMain:=nil;
end;

function TCustomPropSave.GetKey(Comp: TComponent): string;
var
  k: string;
begin
  Result:='';
  if Comp.Name = '' then
    Exit;
  k:=FPropSaveMain.GetFormKey;
  if Assigned(FOnGetKeyString) then
    FOnGetKeyString(Comp, k)
  else
  begin
    k:=Comp.Name;
    while (Comp <> nil) and (Comp.Owner <> FPropSaveMain.OwnForm) do
    begin
      Comp:=Comp.Owner;
      if Comp <> nil then
        k:=Comp.Name + '.' + k;
    end;
    k:=FPropSaveMain.GetFormKey + '.' + k;
  end;
  Result:=k;
end;

procedure TCustomPropSave.ProcComps(psAction: TPropSaveAction);
var
  i: integer;
  c: TComponent;
  s, cname: string;
  SaveChild: boolean;
begin
  if FDisable then
    Exit;
  // find and read/save all components
  for i:=0 to FComponents.Count - 1 do
  begin
    s:=FComponents.Strings[i];
    SaveChild:=false;
    // test, if also read/save child component, i.e. Panel.*
    if copy(s, Length(s) - 1, 2) = '.*' then
    begin
      s:=copy(s, 1, Length(s) - 2);
      SaveChild:=true;
    end;
    c:=FPropSaveMain.Owner; // .OwnForm;
    while Assigned(c) and (Pos('.', s) > 0) do
    begin
      cname:=rtool.RemoveStrPart(s, '.');
      c:=FPropSaveMain.Owner.FindComponent(cname);
    end;
    if Assigned(c) then
      c:=c.FindComponent(s);
    Assert(not((c = nil) and (not SaveChild)), Format('CustomPropSave: component "%s" not found.', [FComponents.Strings[i]]));
    if c <> nil then
      ProcComp(psAction, c, SaveChild)
    else
      if UpperCase(FPropSaveMain.Owner.Name) = UpperCase(s) then
        ProcComp(psAction, FPropSaveMain.Owner, SaveChild);
  end;
  if Assigned(FOnProcComponents) then
    FOnProcComponents(Self, psAction);
end;

procedure TCustomPropSave.ProcComp(psAction: TPropSaveAction; Comp: TComponent; ProcChild: boolean);
var
  i: integer;
  k: string;
  act: TPropSaveAction;
begin
  if (Comp = nil) or (Comp.Name = '') then
    Exit;
  Assert(ProcChild or CanSave(Comp), Format('CustomPropSave: cannot save component "%s".', [Comp.Name]));
  k:=GetKey(Comp);
  if FDisableComponents.IndexOf(Comp) >= 0 then
    act:=psaDelete
  else
    act:=psAction;
  if act in [psaSave, psaDelete] then
    FPropSaveMain.FPropSaveFile.EraseSection(k);
  // read/save control
  ReadSaveProps(act, FPropSaveMain.FPropSaveFile, k, Comp);
  if ProcChild and (Comp is TWinControl) then
  begin
    // read/save child controls
    for i:=0 to TWinControl(Comp).ControlCount - 1 do
      ProcComp(psAction, TWinControl(Comp).Controls[i], true);
  end;
end;

function TCustomPropSave.PropsIsStored(Comp: TComponent): boolean;
begin
  try
    Result:=FPropSaveMain.FPropSaveFile.SectionExists(GetKey(Comp));
  except
    Result:=false;
  end;
end;

// ==============================================================================
// PositionPropSave functions
constructor TPositionPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLimitRect:=TLimitRect.Create;
  FLimitRect.FMinWidthHeight:=30;
end;

destructor TPositionPropSave.Destroy;
begin
  FreeAndNil(FLimitRect);
  inherited Destroy;
end;

function TPositionPropSave.CheckLimit(R: TRect): TRect;
begin
  if R.Left > FLimitRect.Right - (R.Right - R.Left) then
    R.Left:=FLimitRect.Right - (R.Right - R.Left);
  if R.Left < FLimitRect.Left then
    R.Left:=FLimitRect.Left;
  if R.Right > FLimitRect.Right then
    R.Right:=FLimitRect.Right;
  if R.Right < R.Left + FLimitRect.FMinWidthHeight then
    R.Right:=R.Left + FLimitRect.FMinWidthHeight;
  if R.Top > FLimitRect.Bottom - (R.Bottom - R.Top) then
    R.Top:=FLimitRect.Bottom - (R.Bottom - R.Top);
  if R.Top < FLimitRect.Top then
    R.Top:=FLimitRect.Top;
  if R.Bottom > FLimitRect.Bottom then
    R.Bottom:=FLimitRect.Bottom;
  if R.Bottom < R.Top + FLimitRect.FMinWidthHeight then
    R.Bottom:=R.Top + FLimitRect.FMinWidthHeight;
  Result:=R;
end;

function TPositionPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TControl);
end;

procedure TPositionPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);
var
  c: TControl;
  R, Rn: TRect;
begin
  if not(Comp is TControl) then
    Exit;
  c:=Comp as TControl;
  Key:=Key + sPos;
  case A of
    psaSave:
      begin
        R:=c.BoundsRect;
        PSFile.WriteInteger(Key, sLeft, R.Left);
        PSFile.WriteInteger(Key, sRight, R.Right);
        PSFile.WriteInteger(Key, sTop, R.Top);
        PSFile.WriteInteger(Key, sBottom, R.Bottom);
      end;
    psaRead:
      begin
        R:=c.BoundsRect;
        Rn.Left:=PSFile.ReadInteger(Key, sLeft, R.Left);
        Rn.Right:=PSFile.ReadInteger(Key, sRight, R.Right);
        Rn.Top:=PSFile.ReadInteger(Key, sTop, R.Top);
        Rn.Bottom:=PSFile.ReadInteger(Key, sBottom, R.Bottom);
        if FLimitToRect then
          Rn:=CheckLimit(Rn);
        c.BoundsRect:=Rn;
      end;
    psaDelete:
      begin
        PSFile.DeleteKey(Key, sLeft);
        PSFile.DeleteKey(Key, sRight);
        PSFile.DeleteKey(Key, sTop);
        PSFile.DeleteKey(Key, sBottom);
      end;
  end;
end;

// ==============================================================================
// StdCtrlPropSave functions
function TStdCtrlPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TCustomMemo) or (Comp is TCustomListBox) or (Comp is TComboBox) or (Comp is TCheckBox) or (Comp is TRadioGroup) or (Comp is TRadioButton) or
    (Comp is TCustomEdit) or (Comp is TMenuItem) or (Comp is TTabControl) or (Comp is TPageControl) or (Comp is TColorBox) or (Comp is TDateTimePicker);
end;

procedure TStdCtrlPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);
  procedure RSEdit(E: TCustomEdit);
  begin
    case A of
      psaSave:
        PSFile.WriteString(Key, sText, E.Text);
      psaRead:
        E.Text:=PSFile.ReadString(Key, sText, E.Text);
      psaDelete:
        PSFile.DeleteKey(Key, sText);
    end;
  end;

  procedure RSCheckBox(ChB: TCheckBox);
  begin
    case A of
      psaSave:
        PSFile.WriteInteger(Key, sState, ord(ChB.State));
      psaRead:
        ChB.State:=TCheckBoxState(PSFile.ReadInteger(Key, sState, ord(ChB.State)));
      psaDelete:
        PSFile.DeleteKey(Key, sState);
    end;
  end;

  procedure RSRadioButton(RB: TRadioButton);
  begin
    case A of
      psaSave:
        PSFile.WriteBool(Key, sChecked, RB.Checked);
      psaRead:
        RB.Checked:=PSFile.ReadBool(Key, sChecked, RB.Checked);
      psaDelete:
        PSFile.DeleteKey(Key, sChecked);
    end;
  end;

  procedure RSRadioGroup(RG: TRadioGroup);
  begin
    case A of
      psaSave:
        PSFile.WriteInteger(Key, sItemIndex, RG.ItemIndex);
      psaRead:
        RG.ItemIndex:=PSFile.ReadInteger(Key, sItemIndex, RG.ItemIndex);
      psaDelete:
        PSFile.DeleteKey(Key, sItemIndex);
    end;
  end;

  procedure RSMemo(M: TCustomMemo);
  begin
    case A of
      psaSave:
        PSFile.WriteStrings(Key, M.Lines);
      psaRead:
        PSFile.ReadStrings(Key, M.Lines);
      psaDelete:
        PSFile.DeleteStrings(Key);
    end;
  end;

  procedure RSListBox(LB: TCustomListBox);
  var
    i, n: integer;
  begin
    case A of
      psaSave:
        begin
          if Pos(RosiCompConfig.PropSaveCfg.Suffix_BlockSaveItems, LB.Name) = 0 then
            PSFile.WriteStrings(Key, LB.Items);
          if LB.MultiSelect then
            for i:=0 to LB.Items.Count - 1 do
              PSFile.WriteBool(Key, Format(sItemSel, [i]), LB.Selected[i])
            else
              PSFile.WriteInteger(Key, sItemIndex, LB.ItemIndex);
          if LB is TCheckListBox then
          begin
            for i:=0 to LB.Items.Count - 1 do
              PSFile.WriteBool(Key, Format(sItemChecked, [i]), TCheckListBox(LB).Checked[i]);
          end;
        end;
      psaRead:
        begin
          if Pos(RosiCompConfig.PropSaveCfg.Suffix_BlockSaveItems, LB.Name) = 0 then
            PSFile.ReadStrings(Key, LB.Items);
          if LB.MultiSelect then
            for i:=0 to LB.Items.Count - 1 do
              LB.Selected[i]:=PSFile.ReadBool(Key, Format(sItemSel, [i]), LB.Selected[i])
            else
              LB.ItemIndex:=PSFile.ReadInteger(Key, sItemIndex, LB.ItemIndex);
          if LB is TCheckListBox then
          begin
            for i:=0 to LB.Items.Count - 1 do
              TCheckListBox(LB).Checked[i]:=PSFile.ReadBool(Key, Format(sItemChecked, [i]), TCheckListBox(LB).Checked[i]);
          end;
        end;
      psaDelete:
        begin
          n:=Max(PSFile.ReadInteger(Key, sItemsCount, -1), LB.Items.Count);
          for i:=0 to n do
          begin
            PSFile.DeleteKey(Key, Format(sItemSel, [i]));
            PSFile.DeleteKey(Key, Format(sItemChecked, [i]));
          end;
          PSFile.DeleteKey(Key, sItemIndex);
          PSFile.DeleteStrings(Key);
        end;
    end;
  end;

  procedure RSComboBox(CB: TComboBox);
  begin
    case A of
      psaSave:
        begin
          if Pos(RosiCompConfig.PropSaveCfg.Suffix_BlockSaveItems, CB.Name) = 0 then
            PSFile.WriteStrings(Key, CB.Items);
          if CB.Style = csDropDownList then
            PSFile.WriteInteger(Key, sItemIndex, CB.ItemIndex)
          else
            PSFile.WriteString(Key, sText, CB.Text);
        end;
      psaRead:
        begin
          if Pos(RosiCompConfig.PropSaveCfg.Suffix_BlockSaveItems, CB.Name) = 0 then
            PSFile.ReadStrings(Key, CB.Items);
          if CB.Style = csDropDownList then
            CB.ItemIndex:=PSFile.ReadInteger(Key, sItemIndex, CB.ItemIndex)
          else
            CB.Text:=PSFile.ReadString(Key, sText, CB.Text);
        end;
      psaDelete:
        begin
          PSFile.DeleteKey(Key, sItemIndex);
          PSFile.DeleteKey(Key, sText);
          PSFile.DeleteStrings(Key);
        end;
    end;
  end;

  procedure RSMenuItem(MI: TMenuItem);
  begin
    case A of
      psaSave:
        PSFile.WriteBool(Key, sChecked, MI.Checked);
      psaRead:
        MI.Checked:=PSFile.ReadBool(Key, sChecked, MI.Checked);
      psaDelete:
        PSFile.DeleteKey(Key, sChecked);
    end;
  end;

  procedure RSTabControl(TC: TTabControl);
  begin
    case A of
      psaSave:
        PSFile.WriteInteger(Key, sItemIndex, TC.TabIndex);
      psaRead:
        TC.TabIndex:=PSFile.ReadInteger(Key, sItemIndex, TC.TabIndex);
      psaDelete:
        PSFile.DeleteKey(Key, sItemIndex);
    end;
  end;

  procedure RSPageControl(PC: TPageControl);
  begin
    case A of
      psaSave:
        PSFile.WriteInteger(Key, sItemIndex, PC.ActivePageIndex);
      psaRead:
        PC.ActivePageIndex:=PSFile.ReadInteger(Key, sItemIndex, PC.ActivePageIndex);
      psaDelete:
        PSFile.DeleteKey(Key, sItemIndex);
    end;
  end;

  procedure RSColorBox(CB: TColorBox);
  begin
    case A of
      psaSave:
        PSFile.WriteInteger(Key, sColor, CB.Selected);
      psaRead:
        CB.Selected:=PSFile.ReadInteger(Key, sColor, CB.Selected);
      psaDelete:
        PSFile.DeleteKey(Key, sColor);
    end;
  end;

  procedure RSDateTimePicker(DP: TDateTimePicker);
  var
    oldChecked: boolean;
  begin
    oldChecked:=DP.Checked;
    case A of
      psaSave:
        PSFile.WriteDateTime(Key, sDate, Trunc(DP.Date));
      psaRead:
        DP.Date:=PSFile.ReadDateTime(Key, sDate, Trunc(DP.Date));
      psaDelete:
        PSFile.DeleteKey(Key, sDate);
    end;
    if DP.ShowCheckbox then
    begin
      case A of
        psaSave:
          PSFile.WriteBool(Key, sDateChecked, DP.Checked);
        psaRead:
          DP.Checked:=PSFile.ReadBool(Key, sDateChecked, oldChecked);
        psaDelete:
          PSFile.DeleteKey(Key, sDateChecked);
      end;
    end;
  end;

begin
  if (Comp is TCustomMemo) then
    RSMemo(Comp as TCustomMemo)
  else
    if (Comp is TCustomListBox) then
      RSListBox(Comp as TCustomListBox)
    else
      if (Comp is TComboBox) then
        RSComboBox(Comp as TComboBox)
      else
        if (Comp is TCheckBox) then
          RSCheckBox(Comp as TCheckBox)
        else
          if (Comp is TRadioGroup) then
            RSRadioGroup(Comp as TRadioGroup)
          else
            if (Comp is TRadioButton) then
              RSRadioButton(Comp as TRadioButton)
            else
              if (Comp is TCustomEdit) then
                RSEdit(Comp as TCustomEdit)
              else
                if (Comp is TMenuItem) then
                  RSMenuItem(Comp as TMenuItem)
                else
                  if (Comp is TTabControl) then
                    RSTabControl(Comp as TTabControl)
                  else
                    if (Comp is TPageControl) then
                      RSPageControl(Comp as TPageControl)
                    else
                      if (Comp is TColorBox) then
                        RSColorBox(Comp as TColorBox)
                      else
                        if (Comp is TDateTimePicker) then
                          RSDateTimePicker(Comp as TDateTimePicker);
end;

end.
