unit PropSaveOtherComps;

{<
  PropSaveOtherComps unit contains set of components for re/storing
  properties of other nonstandard components (TrGroupBox, TrImageZoom, TrMemoEx, TrDBMemoEx, ToolBars etc).

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2010)
  @lastmod(Dec 2015)
  }

interface

{$INCLUDE _rosi.inc}

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, PropSaveMain, rImprovedComps, rImageZoom,
{$IFDEF rUseDBComp} rDBComponents, {$ENDIF} RosiCompGlobalCfg;

type
  { provides re/storing properties of rosi components (TrGroupBox, TrImageZoom, TrMemoEx, TrDBMemoEx }
  TRosiCompsPropSave = class(TCustomPropSave)
  private
    FStoreVisibility: boolean;
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { returns whether re/storing of Components from parameter is supported by TRosiCompsPropSave }
    function CanSave(Comp: TComponent): boolean; override;
  published
    property StoreVisibility: boolean read FStoreVisibility write FStoreVisibility default true;
  end;

  { provides re/storing properties of TToolBar component }
  TToolBarPropSave = class(TCustomPropSave)
  private
  protected
    procedure ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { returns whether re/storing of Components from parameter is supported by TToolBarPropSave }
    function CanSave(Comp: TComponent): boolean; override;
  published
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

const
  sVisible = 'Visible';
  sMinimized = 'Minimized';
  sOriginSize = 'OriginSize';
  sButCount = 'ButtonsCount';
  sButton = 'Button%d';
  sShowCaptions = 'ShowCaptions';
  sShowList = 'ShowList';
  sAutoSize = 'AutoSize';
  sAutoFit = 'AutoFit';
  sZoom = 'Zoom';

procedure Register;
begin
  RegisterComponents('Rosi PS', [TRosiCompsPropSave, TToolBarPropSave]);
end;

// ==============================================================================
// RosiCompsPropSave functions
constructor TRosiCompsPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FStoreVisibility:=true;
end;

function TRosiCompsPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TrGroupBox) or (Comp is TrImageZoom) or (Comp is TrMemoEx)
{$IFDEF rUseDBComp} or (Comp is TrDBMemo){$ENDIF};
end;

procedure TRosiCompsPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);

  procedure RSrGroupBox(GB: TrGroupBox);
  begin
    case A of
      psaRead:
        begin
          if FStoreVisibility then
            GB.Visible:=PSFile.ReadBool(Key, sVisible, GB.Visible);
          GB.Minimized:=PSFile.ReadBool(Key, sMinimized, GB.Minimized);
          GB.OriginSize:=PSFile.ReadInteger(Key, sOriginSize, GB.OriginSize);
        end;
      psaSave:
        begin
          if FStoreVisibility then
            PSFile.WriteBool(Key, sVisible, GB.Visible);
          PSFile.WriteBool(Key, sMinimized, GB.Minimized);
          PSFile.WriteInteger(Key, sOriginSize, GB.OriginSize);
        end;
      psaDelete:
        begin
          PSFile.DeleteKey(Key, sVisible);
          PSFile.DeleteKey(Key, sMinimized);
          PSFile.DeleteKey(Key, sOriginSize);
        end;
    end;
  end;

  procedure RSrImageZoom(Img: TrImageZoom);
  begin
    case A of
      psaRead:
        begin
          Img.Zoom:=PSFile.ReadInteger(Key, sZoom, Img.Zoom);
          Img.AutoFit:=PSFile.ReadBool(Key, sAutoFit, Img.AutoFit);
        end;
      psaSave:
        begin
          PSFile.WriteInteger(Key, sZoom, Img.Zoom);
          PSFile.WriteBool(Key, sAutoFit, Img.AutoFit);
        end;
      psaDelete:
        begin
          PSFile.DeleteKey(Key, sZoom);
          PSFile.DeleteKey(Key, sAutoFit);
        end;
    end;
  end;

  procedure RSrMemoEx(Memo: TrMemoEx);
  begin
    case A of
      psaRead:
        begin
          PSFile.ReadFont(Key, Memo.Font);
        end;
      psaSave:
        begin
          if not Memo.ParentFont then
            PSFile.WriteFont(Key, Memo.Font);
        end;
      psaDelete:
        begin
          PSFile.DeleteFont(Key);
        end;
    end;
  end;
{$IFDEF rUseDBComp}
  procedure RSrDBMemoEx(Memo: TrDBMemoEx);
  begin
    case A of
      psaRead:
        begin
          PSFile.ReadFont(Key, Memo.Font);
          PSFile.ReadStrings(Key, Memo.StoredLines);
        end;
      psaSave:
        begin
          if not Memo.ParentFont then
            PSFile.WriteFont(Key, Memo.Font);
          PSFile.WriteStrings(Key, Memo.StoredLines);
        end;
      psaDelete:
        begin
          PSFile.DeleteFont(Key);
          PSFile.DeleteStrings(Key);
        end;
    end;
  end;
{$ENDIF}

begin
  if (Comp is TrGroupBox) then
    RSrGroupBox(Comp as TrGroupBox);
  if (Comp is TrImageZoom) then
    RSrImageZoom(Comp as TrImageZoom);
  if (Comp is TrMemoEx) then
    RSrMemoEx(Comp as TrMemoEx);
{$IFDEF rUseDBComp}
  if (Comp is TrDBMemoEx) then
    RSrDBMemoEx(Comp as TrDBMemoEx);
{$ENDIF}
end;

// ==============================================================================
// ToolBarPropSave functions
constructor TToolBarPropSave.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

function TToolBarPropSave.CanSave(Comp: TComponent): boolean;
begin
  Result:=(Comp is TToolBar);
end;

procedure TToolBarPropSave.ReadSaveProps(A: TPropSaveAction; PSFile: TPropSaveFile; Key: string; Comp: TComponent);

  procedure RSToolBar(TB: TToolBar);
  var
    i: integer;
  begin
    case A of
      psaRead:
        begin
          TB.ShowCaptions:=PSFile.ReadBool(Key, sShowCaptions, TB.ShowCaptions);
          TB.List:=PSFile.ReadBool(Key, sShowList, TB.List);
          for i:=0 to TB.ButtonCount - 1 do
          begin
            TB.Buttons[i].Visible:=PSFile.ReadBool(Key, Format(sButton, [i]), TB.Buttons[i].Visible);
            TB.Buttons[i].AutoSize:=PSFile.ReadBool(Key, sAutoSize, TB.Buttons[i].AutoSize);
          end;
        end;
      psaSave:
        begin
          PSFile.WriteInteger(Key, sButCount, TB.ButtonCount);
          PSFile.WriteBool(Key, sShowCaptions, TB.ShowCaptions);
          PSFile.WriteBool(Key, sShowList, TB.List);
          if TB.ButtonCount > 0 then
            PSFile.WriteBool(Key, sAutoSize, TB.Buttons[0].AutoSize);
          for i:=0 to TB.ButtonCount - 1 do
            PSFile.WriteBool(Key, Format(sButton, [i]), TB.Buttons[i].Visible);
        end;
      psaDelete:
        begin
          PSFile.DeleteKey(Key, sShowCaptions);
          PSFile.DeleteKey(Key, sShowList);
          PSFile.DeleteKey(Key, sAutoSize);
          for i:=0 to TB.ButtonCount - 1 do
            PSFile.DeleteKey(Key, Format(sButton, [i]));
        end;
    end;
  end;

begin
  if (Comp is TToolBar) then
    RSToolBar(Comp as TToolBar);
end;

end.
