unit RTFReport2;

{<
  RTFReport2 unit contains TRTFReport2 - non visual component which provides functionality for
  generating reports according prepared RTF or TXT template with special TAGs;
  it uses this template and fill data into it instead of TAGs;
  RTF template can be modified easily by any RTF or TXT editor (i.g. MS Word).

  See homepage http://www.rosinsky.cz/delphi/ for detail help.

  Updated version compatible with Delphi2010+ and 64bit platform.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Apr 2006)
  @lastmod(Oct 2020)
}

interface

{$INCLUDE _rosi.inc}
{$IFDEF DXE4}
 {$DEFINE SkipBookmark}
{$ENDIF}
{$IFDEF DXE5}
 {$DEFINE SkipBookmark}
{$ENDIF}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, MaskUtils, ComCtrls, StrUtils, Variants, RosiCompGlobalCfg;

const
  maxDS = 25;
  UseUnicodeTag = true;

type
  { list of template format types used by TRTFReport2 }
  TTemplateFormat = (tfAutoDetect, tfRTF, tfText);

  { list of RTF text format types used by TRTFReport2 }
  TRTFTextFormat = (frmBold, frmItalic, frmUnderline);
  { set of RTF text format types used by TRTFReport2 }
  TRTFTextFormats = set of TRTFTextFormat;
  { list of memo or blob format types (used if is not defined in rtf template) used by TRTFReport2 }
  TMemoFormat = (mfText, mfRTF, mfAutoDetect, mfRTFLine, mfPlain);

  { event type called for each record during reporting, used by TRTFReport2 }
  TReportProcessEvent = procedure(Sender: TObject; RecNbr: integer; var Abort: boolean) of object;
  { event type called to get calculated string result, used by TRTFReport2  }
  TUserCalc = procedure(Sender: TObject; var CalcStr: string; var Replace: boolean) of object;
  { event type called to get calculated string result, supports function parameters, used by TRTFReport2 }
  TUserCalcParams = procedure(Sender: TObject; FuncName: string; Params: array of variant; var FuncResult: variant; var Replace: boolean) of object;
  { event type called to get bitmap content for Picture tag }
  TOnGetPictureContent = procedure(Sender: TObject; FuncName: string; Bmp: TBitmap) of object;
  { event type called to modify file content, used by TRTFReport2  }
  TModifyContent = procedure(Sender: TObject; var FileContent: string) of object;
  { event type called instead of loading of template file, used by TRTFReport2 }
  TOnGetTemplateStr = procedure(Sender: TObject; var TemplateStr: string) of object;

  { Object contain calculated value for each field, used by TRTFReport2 }
  TCalcField = class
  private
    Init: boolean;
    DSindex: integer;
    Name: string;
    Expression: string;
    Max: double;
    GrandMax: double;
    Min: double;
    GrandMin: double;
    Sum: double;
    GrandSum: double;
    Count: integer;
    GrandCount: integer;
  end;

  { @exclude }
  TScanInfo = record
    DS: string;
    DSindex: integer;
    WhileCond, SkipCond, Sep: string;
    NewLine, NewPar, NewCol, NewPage, NoIfEmpty, NoFirst: boolean;
    MinRows, MaxRows: integer;
  end;

  { @exclude }
  TSeekInfo = record
    Action: byte;
    OldText, NewText: string;
    SeekFrom, SeekTo: string;
    SetFrmStart, SetFrmStop: string;
    WholeWord: boolean;
  end;

  { non visual component which provides functionality for
    generating reports according prepared RTF template with special TAGs }
  TRTFReport2 = class(TComponent)
  private
    PicCount: integer;
    FReportFile: string;
    FOutputFile: string;
    FDefaultCharset: integer;
    FDefaultMemoFormat: TMemoFormat;
    FDefaultCalcFormat: string;
    FDefaultNumericFormat: string;
    FDefaultBooleanFormat: string;
    FDefaultDateFormat: string;
    FDefaultTimeFormat: string;
    FConvertScientificFormat: boolean;
    FVariables: TStrings;
    FFieldNotFoundValue: string;
    FMemoReplaces: TStrings;
    FExtractEOLAfterIf: boolean;
    FActionAfter: TActionAfter;
    FShowMaximized: boolean;
    FOnReadInfo: TNotifyEvent;
    FOnUserCalc: TUserCalc;
    FOnUserCalcParams: TUserCalcParams;
    FOnGetPictureContent: TOnGetPictureContent;
    FOnReportProcess: TReportProcessEvent;
    FOnGetTemplateStr: TOnGetTemplateStr;
    FTemplateFormat: TTemplateFormat;
    FModifyBeforeSave: TModifyContent;
    FTxtTemplate: boolean;
    FTxtCodePage: integer;
    FAppendToFile: boolean;
    FShowSaveErrorDialog: boolean;
    FShowProgressBar: boolean;
    FExtendedInfo: TStrings;
    procedure SetVariables(Value: TStrings);
    procedure SetMemoReplaces(Value: TStrings);
    procedure DoAbort(Sender: TObject);
  protected
    Aborted: boolean;
    RecNbr, RecNbrDS0, RecCountDS0: integer;
    DS: array [0 .. maxDS] of TDataSet;
    DS_eof: array [0 .. maxDS] of boolean;
    {$IFNDEF SkipBookmark}
    DS_Bookmarks: array [0 .. maxDS] of TBookmark;
    {$ENDIF}
    LCalcField: TList;
    cf: TCalcField;
    RepStr: string;
    SeekInfo: TSeekInfo;
    procedure Loaded; override;
    procedure RaiseErr(ErrType: integer; const ErrStr: string = '');
    procedure SepBracket(const Str: string; var Bracket, Block1, Block2: string);
    function FindTagStr(const TagStr: string; var s: string; WChar: Char; var StartPos: integer; Del: boolean): string;
    function StrToRtf(const Str: string): string;
    function RtfToStr(const Rtf: string): string;
    function GetListPart(var Str: string; const Sep: string): string;
    procedure ClearEOLPart(Sender: TObject; var s: string);
    function ExtractSpecChar(var Str: string): string;
    procedure RemovePar(var Str: string; SkipSpaces: boolean);
    procedure ProcAllSkipPictureTag(var s: string; CallbackProc: TOnGetTemplateStr);

    procedure InitCalcField(DSindex: integer; const CalcStr: string);
    procedure FreeCalcField;
    procedure ProcCalcField(DSindex: integer);
    function GetCalcFieldValue(CalcTypeIndex: integer; const Name: string): double;

    function IsConstantTag(const TagStr: string): boolean;
    function IsNumericTag(const TagStr: string; var Value: double): boolean;
    function IsFieldTag(const TagStr: string): boolean;
    function DatasetCharToIndex(Ch: Char): integer;
    function IndexToDatasetChar(Idx: integer): Char;
    function DSActive(DSindex: integer): boolean;
    function DSNotEOF(DSindex: integer): boolean;
    function GetDSIndex(const TagStr: string): integer;
    function FindFirst(DSindex: integer): boolean;
    function FindNext(DSindex: integer): boolean;
    function FindPrev(DSindex: integer): boolean;
    function FindLast(DSindex: integer): boolean;
    function GetDSBookmark(DSindex: integer): boolean;
    function GotoDSBookmark(DSindex: integer): boolean;
    procedure FreeBookMarks;

    function ListAllFields: string;
    function ProcSepList(const TagStr: string): string;
    function ProcFirstValue(const TagStr: string): string;
    function ProcPicture(const TagStr: string): string;
    function ProcBarcode(const TagStr: string): string;
    function CreateFormatPicture(DSindex: integer; F: TField; const FileName: string; Bmp: TBitmap; FormatStr: string): string;
    function ProcNumToText(const TagStr: string; const TagFormat: string): string;
    procedure ProcDefVar(const TagStr: string);
    function ProcStrPrefix(const TagStr: string): string;
    function ProcCase(const TagStr: string): string;
    procedure ProcExtendedInfo(var Str: string);
    procedure ProcInsertFile(var TagStr: string);

    procedure ProcAll;
    function ProcBlock(const Str: string): string;
    function ProcScan(var s: string): string;
    function DecodeScanLine(var s: string): TScanInfo;
    procedure ProcIf(IIF, ProcAllIf: boolean; var Str: string);
    function CheckExtractLineAfterIf: boolean;

    function ProcRTF(const Str: string): string;
    function ProcTagStr(const TagStr: string; var Repl: boolean): string;
    function GetDBValue(DSindex: integer; const FieldName, FormatStr: string): string;
    function GetVarUserValue(const VarName: string; var Replaced: boolean): string;
    function GetCalcNumValue(const CalcStr: string; ErrFieldAsZero: boolean): double;
    function GetCalcValue(const CalcStr, FormatStr: string; ErrFieldAsZero: boolean): string;
    function GetConditionResult(const CondStr: string): boolean;

    function CheckScientificFormat(const s: string): string;
    procedure ProcFormatText(Sender: TObject; var s: string);
  public
    { define name of printer which will be used if Action=aaPrint }
    PrinterName: string;
    { define number of copies, used if Action=aaPrint }
    PrintCopies: byte;
    { defines whether exception is raised for unknown fields in template }
    RaiseUnknownField: boolean;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { execute creating of report }
    procedure Execute(DataSets: array of TDataSet);
    { can be called in OnModifyBeforeSave event, formats selected text bold, italic, underline }
    procedure FormatText(const TextToFormat: string; FormatTypes: TRTFTextFormats; WholeWordOnly: boolean = false; const SeekFrom: string = '';
      const SeekTo: string = '');
    { can be called in OnModifyBeforeSave event, replace selected text, new string can contain tag to start/stop superscript or subscript
      \^  => start superscript;  \_  => start subscript; \=  => end superscipt or subsript }
    procedure ReplaceText(const OldText, NewText: string; WholeWordOnly: boolean = false; const SeekFrom: string = ''; const SeekTo: string = '');
    { Define whether EOL after IF, IIF, ELSEIF, ELSEIIF is removed }
    property ExtractEOLAfterIf: boolean read FExtractEOLAfterIf write FExtractEOLAfterIf;
    { content of INFO tag in report template,
      INFO tag is used for backward comunication from report to application }
    property ExtendedInfo: TStrings read FExtendedInfo;
    { return value for specific FieldTag }
    function GetFieldTagValue(FieldTag: string): string;
  published
    { Define RTF file name with report template }
    property ReportFile: string read FReportFile write FReportFile;
    { Define output file name }
    property OutputFile: string read FOutputFile write FOutputFile;
    { Define charset number, which will be used in report }
    property DefaultCharset: integer read FDefaultCharset write FDefaultCharset default 0;
    { Define default format for BLOB fields }
    property DefaultMemoFormat: TMemoFormat read FDefaultMemoFormat write FDefaultMemoFormat default mfAutoDetect;
    { Define default format for calculated values }
    property DefaultCalcFormat: string read FDefaultCalcFormat write FDefaultCalcFormat;
    { Define default format for numeric fields }
    property DefaultNumericFormat: string read FDefaultNumericFormat write FDefaultNumericFormat;
    { Define default format for boolean fields }
    property DefaultBooleanFormat: string read FDefaultBooleanFormat write FDefaultBooleanFormat;
    { Define default format for date fields }
    property DefaultDateFormat: string read FDefaultDateFormat write FDefaultDateFormat;
    { Define default format for time fields }
    property DefaultTimeFormat: string read FDefaultTimeFormat write FDefaultTimeFormat;
    { Define whether scientific format is converted to 123 x 10^3  format using superscript }
    property ConvertScientificFormat: boolean read FConvertScientificFormat write FConvertScientificFormat default false;
    { Define list of variables and their values which can be used in
      report template }
    property Variables: TStrings read FVariables write SetVariables;
    { Defines default value used for fields which is not found }
    property FieldNotFoundValue: string read FFieldNotFoundValue write FFieldNotFoundValue;
    { Defines list of strings which is replaced it memo field, e.g.:
      <TAB>=;     - replace #9 by ;
      ;=\line     - replace ; by end of line }
    property MemoReplaces: TStrings read FMemoReplaces write SetMemoReplaces;
    { Define executed action after report is created }
    property ActionAfter: TActionAfter read FActionAfter write FActionAfter default aaShow;
    { Define whether rtf file is opened in maximized window (if rtf editor supports) }
    property ShowMaximized: boolean read FShowMaximized write FShowMaximized default false;
    { define event which is called for each INFO tag }
    property OnReadInfo: TNotifyEvent read FOnReadInfo write FOnReadInfo;
    { define event which is called for getting value of undefined properties }
    property OnUserCalc: TUserCalc read FOnUserCalc write FOnUserCalc;
    { define event which is called for getting value of undefined tags, supports also parameters }
    property OnUserCalcParams: TUserCalcParams read FOnUserCalcParams write FOnUserCalcParams;
    { define event which is called to get bitmap content for Picture tag }
    property OnGetPictureContent: TOnGetPictureContent read FOnGetPictureContent write FOnGetPictureContent;
    { define event which is called for each dataset records
      to show progress of execution }
    property OnReportProcess: TReportProcessEvent read FOnReportProcess write FOnReportProcess;
    { define format type of template (auto, TXT, RTF) }
    property TemplateFormat: TTemplateFormat read FTemplateFormat write FTemplateFormat default tfAutoDetect;
    { define used codepage for inserted data in case of TXT temlate format }
    property TxtCodePage: integer read FTxtCodePage write FTxtCodePage default 0;
    { define whether output is appended to existing file (used only for TXT format) }
    property AppendToFile: boolean read FAppendToFile write FAppendToFile default false;
    { define event which is called just before saving of result }
    property OnModifyBeforeSave: TModifyContent read FModifyBeforeSave write FModifyBeforeSave;
    { define event called instead of loading of template file }
    property OnGetTemplateStr: TOnGetTemplateStr read FOnGetTemplateStr write FOnGetTemplateStr;
    { define whether re-save dialog is shown in case of saving error }
    property ShowSaveErrorDialog: boolean read FShowSaveErrorDialog write FShowSaveErrorDialog default false;
    { indicate whether progress bar is shown }
    property ShowProgressBar: boolean read FShowProgressBar write FShowProgressBar default false;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}
{$I _rosi.inc}

uses {$IFDEF BARCODESUPPORT} AsBarcode_interface, {$ENDIF}
  rtool, rdbtool, rdlg, rstring, rpictool, NumToText;

const
  setDS: TSysCharSet = ['A' .. 'Z'];
  cErrStr: array [0 .. 6] of string = (sTemplateBracketErr, sTemplateFieldErr, sTemplateDatasetErr, sTemplateTagNotFoundErr, sTemplateCalculationErr,
    sTemplateIncorrectTag, sPrinterNotFound);

  // rtf
  cTxtEndPar = #13 + #10;
  cRtfEndPar = '\par ';
  cUsedEndOfPar: array [boolean] of string = (cRtfEndPar, cTxtEndPar);
  cTxtEndPage = #12 + #13;
  cRtfEndPage = '\page ';
  cUsedEndOfPage: array [boolean] of string = (cRtfEndPage, cTxtEndPage);
  cRtfEndLine = '\line ';
  cRtfEndCol = '\column ';
  cRtfHexMark = '\''';
  cRtfUnicodeMark = '\u';
  cRtfCharset = 'fcharset';
  cLinkPicture = '{\field{\*\fldinst { INCLUDEPICTURE "%s" \\* MERGEFORMAT \\d }}{\fldrslt {}}}';
  cInsertPicture2b = '{\pict\picscalex%d\picscaley%d\piccropl0\piccropr0\piccropt0\piccropb0' + '\picw%d\pich%d\picwgoal%d\pichgoal%d\%s %s}';
  cPictureTypeMetafile = 'wmetafile8';
  cPictureTypeJPG = 'jpegblip';
  cPictureTypePNG = 'pngblip';
  cSuperScript = '\super ';
  cSubScript = '\sub ';
  cNoSuperSubScript = '\nosupersub ';

  cRtfViewKind = '\viewkind';

  cFormatSep = ':';
  cParamSep = ';';
  cListSep2 = ',';
  cConstSep = '"';
  cDateSep = '#';
  cTabStr = '<TAB>';

  // tags
  cTagCh = 'X';
  cTag = '\\X\\';

  // scan
  cUsrTagScan = '\\scan';
  cUsrTagScanEntry = '\\scanentry(X)\\';
  cUsrTagScanFooter = '\\scanfooter(X)\\';
  cUsrTagEndScan = '\\endscan(X)\\';
  cScanDS = 'scan';
  cScanWhileCond = 'while';
  cScanSkipCond = 'skipif';
  cScanNewLine = 'newline';
  cScanNewPar = 'newpar';
  cScanNewCol = 'newcol';
  cScanNewPage = 'newpage';
  cScanNewSep = 'sep';
  cScanNoIfEmpty = 'noifempty';
  cScanNoFirst = 'nofirst';
  cScanCalc = 'calc';
  cScanMinRows = 'minrows';
  cScanMaxRows = 'maxrows';

  cScanNext = 'next(';
  cScanPrev = 'prev(';
  cScanFirst = 'first(';
  cScanLast = 'last(';
  cScanGetBookmark = 'getbookmark(';
  cScanGotoBookmark = 'gotobookmark(';

  // if
  cUsrTagIfStart = '\\if';
  cUsrTagElseIf = '\\elseif\\';
  cUsrTagEndIf = '\\endif\\';
  cUsrTagIIfStart = '\\iif';
  cUsrTagElseIIf = '\\elseiif\\';
  cUsrTagEndIIf = '\\endiif\\';
  cUserVarIfExtractLine = 'IF_ExtractLine';

  // case
  cCase = 'case(';

  // calc tag
  cCalcFunc: array [0 .. 9] of string = ('Sum[X]', 'Max[X]', 'Min[X]', 'Count[X]', 'Avg[X]', 'GrandSum[X]', 'GrandMax[X]', 'GrandMin[X]', 'GrandCount[X]',
    'GrandAvg[X]');
  cResetCalc = 'resetcalc';

  // datetime function
  cGetToday = 'Today[]';
  cGetNow = 'Now[]';

  // condition tag
  cCondFunc: array [0 .. 6] of string = ('==', '>=', '<=', '<>', '=', '>', '<');
  cCondAnd = ' and ';
  cCondOr = ' or ';
  cCondNot = 'not ';
  cCondResArr: array [boolean] of string = ('"FALSE"', '"TRUE"');

  // other
  cListAllFields = 'listallfields';
  cSepList = 'seplist';
  cFirstValue = 'firstvalue';
  cLabel = 'label';
  cPicture = 'picture';
  cDefVar = 'defvar';
  cExtendedInfo = '\\info(X)\\';
  cInsertFile = '\\insertfile(X)\\';
  cInsertRTFFile = '\\insertrtffile(X)\\';

  cMemoTextFormat = 'text';
  cMemoRTFFormat = 'rtf';
  cMemoRTFLineFormat = 'rtfline';
  cMemoPlainFormat = 'plain';

  cPicMaxW = 'w=';
  cPicMaxWcm = 'wcm=';
  cPicMaxH = 'h=';
  cPicMaxHcm = 'hcm=';
  cPicEnlarge = 'enlarge';
  cPicNameField = 'namefield=';
  cPicSaveAsBMP = 'saveasbmp';
  cPicSaveAsJPG = 'saveasjpg';
  cPicSaveAsPNG = 'saveaspng';
  cPicInsertInRTF = 'insertinrtf';
  cPicQuality = 'quality=';
  cPicRotation = 'rotate=';

  cBarCode = 'barcode';
  cBCType = 'type';
  cBCShowCode = 'showcode';
  cBCFontName = 'fontname';
  cBCFontSize = 'fontsize';

  cNumToText = 'numtotext';

  // math function
  cMathAbs = 'abs[X]';

  // string routine tags
  cStrPrefix = 'str_';
  cStrPart = 'str_part';
  cStrPartCS = 'str_partcs';
  cStrRepl = 'str_replace';
  cStrReplCS = 'str_replacecs';
  cStrUpper = 'str_upper';
  cStrLower = 'str_lower';
  cStrProper = 'str_proper';
  cStrTrim = 'str_trim';

  // string routine function
  cStrLen = 'strlen[X]';
  cStrPos = 'strpos[X]';
  cStrPosCS = 'strposcs[X]';
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TRTFReport2]);
end;

// ======================================================================
// Public actions
constructor TRTFReport2.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FVariables:=TStringList.Create;
  FMemoReplaces:=TStringList.Create;
  FExtendedInfo:=TStringList.Create;
  LCalcField:=TList.Create;
  FActionAfter:=aaShow;
  FDefaultMemoFormat:=mfAutoDetect;
  PrintCopies:=1;
  RaiseUnknownField:=false;
  FTemplateFormat:=tfAutoDetect;
  FTxtCodePage:=0;
  FConvertScientificFormat:=false;
  FExtractEOLAfterIf:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TRTFReport2.Destroy;
begin
  FreeCalcField;
  FreeAndNil(LCalcField);
  FreeAndNil(FVariables);
  FreeAndNil(FMemoReplaces);
  FreeAndNil(FExtendedInfo);
  inherited Destroy;
end;

procedure TRTFReport2.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

// run report
procedure TRTFReport2.Execute(DataSets: array of TDataSet);
var
  i, m: integer;
  oldCursor: TCursor;
begin
  for i:=0 to maxDS do
  begin
    DS[i]:=nil;
    DS_eof[i]:=false;
  end;
  m:= High(DataSets);
  if m > maxDS then
    raise TReportException.Create('Too many datasets.');
  for i:=0 to m do
    DS[i]:=DataSets[i];
  if DSActive(0) then
    RecCountDS0:=DS[0].RecordCount
  else
    RecCountDS0:=-1;

  PicCount:=0;
  case FTemplateFormat of
    tfAutoDetect:
      FTxtTemplate:=SameText(ExtractFileExt(FReportFile), '.TXT');
    tfRTF:
      FTxtTemplate:=false;
    tfText:
      FTxtTemplate:=true;
  end;

  oldCursor:=crDefault;
  if FShowProgressBar then
    ShowInfo(sCreatingReport)
  else
    oldCursor:=WCursor;
  try
    ProcAll;
  finally
    if FShowProgressBar then
      ShowInfo('')
    else
      ChCursor(oldCursor);
  end;
end;

procedure TRTFReport2.RaiseErr(ErrType: integer; const ErrStr: string = '');
begin
  Raise TReportException.Create(Format(sTemplateFormatErr + #13 + cErrStr[ErrType], [ClearStr('\', ErrStr, false)]));
end;

// ======================================================================
// AUX function
procedure TRTFReport2.SetVariables(Value: TStrings);
begin
  FVariables.Assign(Value);
end;

procedure TRTFReport2.SetMemoReplaces(Value: TStrings);
begin
  FMemoReplaces.Assign(Value);
end;

procedure TRTFReport2.DoAbort(Sender: TObject);
begin
  Aborted:=true;
end;

// separate content of first brackets
procedure TRTFReport2.SepBracket(const Str: string; var Bracket, Block1, Block2: string);
var
  i, b1, b2, bpoc: integer;
  constsep: boolean;
begin
  // check brackets error
  bpoc:=0;
  constsep:=false;
  b1:=0;
  for i:=1 to Length(Str) do
  begin
    if Str[i] = cConstSep then
      constsep:=not constsep;
    if constsep then
      Continue;
    if Str[i] = '(' then
    begin
      Inc(bpoc);
      if b1 = 0 then
        b1:=i;
    end;
    if Str[i] = ')' then
      Dec(bpoc);
  end;
  if (bpoc <> 0) or constsep then
    RaiseErr(0, Str);
  // extract outer bracket
  Bracket:='';
  Block1:='';
  Block2:='';
  // b1:=Pos('(',Str);
  if b1 = 0 then
    Block1:=Str
  else
  begin
    bpoc:=1;
    b2:=0;
    for i:=b1 + 1 to Length(Str) do
    begin
      if Str[i] = cConstSep then
        constsep:=not constsep;
      if constsep then
        Continue;
      if Str[i] = '(' then
        Inc(bpoc);
      if Str[i] = ')' then
        Dec(bpoc);
      if (bpoc = 0) then
      begin
        b2:=i;
        Bracket:=(copy(Str, b1 + 1, b2 - b1 - 1));
        Block1:=(copy(Str, 1, b1 - 1));
        Block2:=(copy(Str, b2 + 1, Length(Str)));
        Break;
      end;
    end;
    if b2 = 0 then
      RaiseErr(0, Str);
  end;
end;

// Find position and content of Tag
function TRTFReport2.FindTagStr(const TagStr: string; var s: string; WChar: Char; var StartPos: integer; Del: boolean): string;
var
  sd, fd: integer;
  p1, p2: string;
begin
  Result:='';
  sd:=StartPos;
  StartPos:=0;
  p1:=copy(TagStr, 1, AnsiPos(WChar, TagStr) - 1);
  if p1 = '' then
    Exit;
  sd:=AnsiPosEx(p1, s, sd, false);
  if sd = 0 then
    Exit;
  p2:=copy(TagStr, AnsiPos(WChar, TagStr) + 1, Length(TagStr));
  if p2 = '' then
    Exit;
  fd:=AnsiPosEx(p2, s, sd + Length(p1), false);
  if fd = 0 then
    Exit;
  Result:=copy(s, sd + Length(p1), fd - sd - Length(p1));
  StartPos:=sd;
  if Del then
    Delete(s, sd, fd - sd + Length(p2));
end;

// Convert 8bit chars to RTF
function TRTFReport2.StrToRtf(const Str: string): string;
var
  i: integer;
  s: ansistring;
begin
  s:='';
  if FTxtTemplate then
    Result:=Str
  else
  begin
    Result:='';
    if UseUnicodeTag then
    begin
      for i:=1 to Length(Str) do
        if ord(Str[i]) < 128 then
          Result:=Result + Str[i]
        else
          Result:=Result + cRtfUnicodeMark + IntToStr(ord(Str[i])) + '?';
    end
    else
    begin
      s:=ansistring(Str);
      for i:=1 to Length(s) do
        if ord(s[i]) < 128 then
          Result:=Result + WideChar(s[i])
        else
          Result:=Result + cRtfHexMark + IntToHex(ord(s[i]), 2);
    end;
  end;
end;

// Convert RTF to 8bit chars
function TRTFReport2.RtfToStr(const Rtf: string): string;
var
  p1, i: integer;
  s: string;
begin
  Result:=Rtf;
  if not FTxtTemplate then
  begin
    repeat
      p1:=AnsiPos(cRtfHexMark, Result);
      if p1 = 0 then
        Break;
      s:=copy(Result, p1 + Length(cRtfHexMark), 2);
      Delete(Result, p1, Length(cRtfHexMark) + 2);
      i:=StrToIntDef('$' + s, 0);
      if i > 0 then
        Insert(string(AnsiChar(Chr(i))), Result, p1);
    until false;
  end;
end;

// proc str functions
function TRTFReport2.GetListPart(var Str: string; const Sep: string): string;

  function GetFirstSep(const Sep, Str: string): integer;
  var
    i, constc, br1c, br2c: integer;
  begin
    // return pos of separator out of constant and brackets
    i:=1;
    constc:=0;
    br1c:=0;
    br2c:=0;
    while i <= Length(Str) do
    begin
      if Str[i] = Sep then
        if (constc = 0) and (br1c = 0) and (br2c = 0) then
          Break;
      case Str[i] of
        cConstSep:
          constc:=(constc + 1) mod 2;
        '(':
          if constc = 0 then
            Inc(br1c);
        ')':
          if constc = 0 then
            Dec(br1c);
        '[':
          if constc = 0 then
            Inc(br2c);
        ']':
          if constc = 0 then
            Dec(br2c);
      end;
      i:=i + 1;
    end;
    if i > Length(Str) then
      i:=0;
    Result:=i;
  end;

var
  pos1, posstart: integer;
  s, b, b1, b2: string;
begin
  // return and extract first paramater in list
  Str:=Trim(Str);
  if copy(Str, 1, 1) = cConstSep then
  begin
    // constant: "str"
    posstart:=2;
    repeat
      pos1:=AnsiPosEx(cConstSep, Str, posstart, false);
      if pos1 = 0 then
        raise Exception.Create(Str);
      if copy(Str, pos1 + 1, 1) = cConstSep then
      begin
        // "" => replace to "
        Str:=copy(Str, 1, pos1) + copy(Str, pos1 + 2, Length(Str));
        posstart:=pos1 + 1;
      end
      else
        Break;
    until false;
    s:=copy(Str, 1, pos1);
    Str:=Trim(copy(Str, pos1 + 1, Length(Str)));
  end
  else
    if AnsiPosEx(cStrPrefix, Str, 1, false) = 1 then
    begin
      // string routine
      SepBracket(Str, b, b1, b2);
      s:=b1 + '(' + b + ')';
      Str:=Trim(b2);
    end
    else
    begin
      // value
      pos1:=GetFirstSep(Sep, Str);
      if pos1 = 0 then
        s:=Str
      else
        s:=Trim(copy(Str, 1, pos1 - 1));
      Str:=Trim(copy(Str, Length(s) + 1, Length(Str)));
    end;
  // remove separator
  if Str <> '' then
  begin
    if copy(Str, 1, 1) <> Sep then
      raise Exception.Create(Str);
    Str:=Trim(copy(Str, 2, Length(Str)));
  end;
  Result:=s;
end;

procedure TRTFReport2.ClearEOLPart(Sender: TObject; var s: string);
var
  i: integer;
begin
  repeat
    i:=Pos(cTxtEndPar, s);
    if i = 0 then
      Break;
    Delete(s, i, Length(cTxtEndPar));
  until false;
end;

// Extract spec chars or internal tags inside tag and return as string
function TRTFReport2.ExtractSpecChar(var Str: string): string;
var
  i, p1, p2: integer;
begin
  Result:='';
  if FTxtTemplate then
    Exit;
  // extract internal tags
  repeat
    p1:=0;
    p2:=0;
    for i:=1 to Length(Str) do
    begin
      if (p1 = 0) and (Str[i] = '\') and (i < Length(Str)) and (Str[i + 1] <> '''') then
        p1:=i;
      if (p1 > 0) and (Str[i] = ' ') then
      begin
        p2:=i;
        Break;
      end;
    end;
    if (p1 = 0) or (p2 = 0) then
      Break;
    Result:=Result + copy(Str, p1, p2 - p1 + 1);
    Delete(Str, p1, p2 - p1 + 1);
  until false;
  // replace spec chars
  p1:=0;
  for i:=1 to Length(Str) do
    if CharInSet(Str[i], ['{', '}']) then
    begin
      if p1 = 0 then
        p1:=i
      else
        p1:=-1;
    end;
  if p1 > 0 then
  begin
    Result:=Result + Str[p1];
    Delete(Str, p1, 1);
  end;
end;

// remove /par if exist at the begining
procedure TRTFReport2.RemovePar(var Str: string; SkipSpaces: boolean);
var
  p: integer;
begin
  p:=AnsiPosEx(cUsedEndOfPar[FTxtTemplate], Str, 1, false);
  if p = 0 then
    Exit;
  if (p = 1) or (SkipSpaces and (Trim(copy(Str, 1, p - 1)) = '')) then
    Str:=copy(Str, p + Length(cUsedEndOfPar[FTxtTemplate]), Length(Str));
end;

procedure TRTFReport2.ProcAllSkipPictureTag(var s: string; CallbackProc: TOnGetTemplateStr);
var
  NewStr, ProcStr, PicStr: string;
  PicStart, PicEnd, PicNestedPos: integer;
begin
  NewStr:='';
  repeat
    PicStart:=Pos('{\pict', s);
    if PicStart = 0 then
    begin
      CallbackProc(Self, s);
      NewStr:=NewStr + s;
      Break;
    end
    else
    begin
      ProcStr:=copy(s, 1, PicStart - 1);
      CallbackProc(Self, ProcStr);
      NewStr:=NewStr + ProcStr;
      PicEnd:=PicStart;
      PicNestedPos:=1;
      repeat
        PicEnd:=PosEx('}', s, PicEnd + 1);
        PicStr:=copy(s, PicStart, PicEnd - PicStart + 1);
        PicNestedPos:=PosEx('{', PicStr, PicNestedPos + 1);
      until (PicNestedPos = 0) or (PicEnd = 0);
      if PicEnd = 0 then
        RaiseErr(5, '{\pict');
      NewStr:=NewStr + copy(s, PicStart, PicEnd - PicStart + 1);
      s:=copy(s, PicEnd + 1, Length(s));
    end;
  until false;
  s:=NewStr;
end;

// ======================================================================
// CALC FIELDS function
// Init CalcFields
procedure TRTFReport2.InitCalcField(DSindex: integer; const CalcStr: string);
var
  part, name, exp: string;
  p, i: integer;
  isexist: boolean;
begin
  // proc CalcStr
  p:=1;
  repeat
    part:=AnsiUpperCase(Trim(FindStrPart(CalcStr, cListSep2, p, false)));
    if part = '' then
      Break;
    // add DSindex for backward compability
    if (Pos('.', part) = 0) and (Pos('=', part) = 0) then
      part:=IndexToDatasetChar(DSindex) + '.' + part;

    // proc one part (FieldName or expression)
    i:=Pos('=', part);
    if i = 0 then
    begin
      name:=part;
      exp:=part;
    end
    else
    begin
      name:=copy(part, 1, i - 1);
      exp:=copy(part, i + 1, Length(part));
    end;
    isexist:=false;
    for i:=0 to LCalcField.Count - 1 do
    begin
      cf:=LCalcField.Items[i];
      if (cf.DSindex = DSindex) and (cf.Name = name) then
      begin
        // already exist, only reset
        isexist:=true;
        cf.Expression:=exp;
        cf.Max:=0;
        cf.Min:=0;
        cf.Sum:=0;
        cf.Count:=0;
      end;
    end;
    if not isexist then
    begin
      // add new calcfield
      cf:=TCalcField.Create;
      cf.Init:=false;
      cf.DSindex:=DSindex;
      cf.Name:=name;
      cf.Expression:=exp;
      LCalcField.Add(cf);
    end;
    Inc(p);
  until false;
end;

// Free all CalcFields
procedure TRTFReport2.FreeCalcField;
var
  i: integer;
begin
  for i:=LCalcField.Count - 1 downto 0 do
  begin
    cf:=LCalcField.Items[i];
    FreeAndNil(cf);
    LCalcField.Delete(i);
  end;
end;

procedure TRTFReport2.ProcCalcField(DSindex: integer);
var
  i: integer;
  r: double;
begin
  for i:=0 to LCalcField.Count - 1 do
    try
      cf:=LCalcField.Items[i];
      if cf.DSindex = DSindex then
      begin
        r:=GetCalcNumValue(cf.Expression, true);
        if not cf.Init then
        begin
          cf.Max:=r;
          cf.Min:=r;
          cf.Sum:=r;
          cf.Count:=1;
          cf.GrandMax:=r;
          cf.GrandMin:=r;
          cf.GrandSum:=r;
          cf.GrandCount:=1;
          cf.Init:=true;
        end
        else
        begin
          if cf.Max < r then
            cf.Max:=r;
          if cf.Min > r then
            cf.Min:=r;
          cf.Sum:=cf.Sum + r;
          cf.Count:=cf.Count + 1;
          if cf.GrandMax < r then
            cf.GrandMax:=r;
          if cf.GrandMin > r then
            cf.GrandMin:=r;
          cf.GrandSum:=cf.GrandSum + r;
          cf.GrandCount:=cf.GrandCount + 1;
        end;
      end;
    except
      RaiseErr(1, cf.Expression);
    end;
end;

// Return value of CalcField
function TRTFReport2.GetCalcFieldValue(CalcTypeIndex: integer; const Name: string): double;
var
  i: integer;
  r: double;
  notfound: boolean;
begin
  notfound:=true;
  r:=0;
  for i:=0 to LCalcField.Count - 1 do
  begin
    cf:=LCalcField.Items[i];
    if (cf.Name = AnsiUpperCase(Name)) then
    begin
      notfound:=false;
      case CalcTypeIndex of
        0:
          r:=cf.Sum;
        1:
          r:=cf.Max;
        2:
          r:=cf.Min;
        3:
          r:=cf.Count;
        4:
          if cf.Count > 0 then
            r:=cf.Sum / cf.Count
          else
            r:=0;
        5:
          r:=cf.GrandSum;
        6:
          r:=cf.GrandMax;
        7:
          r:=cf.GrandMin;
        8:
          r:=cf.GrandCount;
        9:
          if cf.GrandCount > 0 then
            r:=cf.GrandSum / cf.GrandCount
          else
            r:=0;
      else
        notfound:=true;
      end;
    end;
  end;
  if notfound then
    RaiseErr(1, Name);
  Result:=r;
end;

// ======================================================================
// Dataset MOVE function
// Decode dataset index form tag "text(X)"
function TRTFReport2.IsConstantTag(const TagStr: string): boolean;
begin
  Result:=(copy(TagStr, 1, 1) = cConstSep) and (copy(TagStr, Length(TagStr), 1) = cConstSep);
end;

function TRTFReport2.IsNumericTag(const TagStr: string; var Value: double): boolean;
var s: string;
begin
  // Check numeric value
  Result:=false;
  try
    s:=UpdateToDecimalSeparator(TagStr);
    Value:=StrToFloat(s);
    Result:=true;
  except
  end;
end;

function TRTFReport2.IsFieldTag(const TagStr: string): boolean;
begin
  Result:=(Length(TagStr)>2) and CharInSet(TagStr[1], setDS) and (TagStr[2] = '.');
end;

function TRTFReport2.DatasetCharToIndex(Ch: Char): integer;
begin
  Result:=ord(Ch) - ord('A');
end;

function TRTFReport2.IndexToDatasetChar(Idx: integer): Char;
begin
  Result:=Chr(ord('A') + idx);
end;

function TRTFReport2.DSActive(DSindex: integer): boolean;
begin
  Result:=(DSindex in [0 .. maxDS]) and Assigned(DS[DSindex]) and DS[DSindex].Active;
end;

function TRTFReport2.DSNotEOF(DSindex: integer): boolean;
begin
  Result:=(DSindex < 0) or (DSActive(DSindex) and not DS[DSindex].EOF);
end;

function TRTFReport2.GetDSIndex(const TagStr: string): integer;
var
  s, b1, b2: string;
begin
  SepBracket(TagStr, s, b1, b2);
  s:=AnsiUpperCase(Trim(s));
  if s = '' then
    Result:=-1
  else
  begin
    if (Length(s) > 1) or not CharInSet(s[1], setDS) then
      RaiseErr(2, TagStr);
    Result:=DatasetCharToIndex(s[1]);
  end;
end;

function TRTFReport2.FindFirst(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].First;
  Result:=not DS[DSindex].EOF;
  DS_eof[DSindex]:=DS[DSindex].EOF;
end;

function TRTFReport2.FindLast(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Last;
  Result:=not DS[DSindex].BOF;
  DS_eof[DSindex]:=DS[DSindex].BOF;
end;

function TRTFReport2.FindNext(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Next;
  Result:=not DS[DSindex].EOF;
  DS_eof[DSindex]:=DS[DSindex].EOF;
  if not Result then
    Exit;
  Inc(RecNbr);
  if DSindex = 0 then
    Inc(RecNbrDS0);
end;

function TRTFReport2.FindPrev(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Prior;
  Result:=not DS[DSindex].BOF;
  DS_eof[DSindex]:=DS[DSindex].BOF;
  if not Result then
    Exit;
  Inc(RecNbr);
end;

function TRTFReport2.GetDSBookmark(DSindex: integer): boolean;
begin
  Result:=false;
  if not DSActive(DSindex) then
    Exit;
  {$IFNDEF SkipBookmark}
  DS_Bookmarks[DSindex]:=DS[DSindex].GetBookmark;
  {$ENDIF}
end;

function TRTFReport2.GotoDSBookmark(DSindex: integer): boolean;
begin
  Result:=false;
  if not DSActive(DSindex) then
    Exit;
  {$IFNDEF SkipBookmark}
  if not Assigned(DS_Bookmarks[DSindex]) then
    Exit;
  DS[DSindex].GotoBookmark(DS_Bookmarks[DSindex]);
  DS_eof[DSindex]:=DS[DSindex].EOF;
  {$ENDIF}
end;

procedure TRTFReport2.FreeBookMarks;
var
  i: integer;
begin
  {$IFNDEF SkipBookmark}
  for i:=0 to maxDS do
    if Assigned(DS_Bookmarks[i]) and Assigned(DS[i]) then
    begin
      DS[i].FreeBookmark(DS_Bookmarks[i]);
      DS_Bookmarks[i]:=nil;
    end;
  {$ENDIF}
end;

// ======================================================================
// IN-BAND function
// Create list of all datasets fields
function TRTFReport2.ListAllFields: string;
var
  i, j: integer;
  s: string;
begin
  s:='';
  for i:=0 to maxDS do
  begin
    if not Assigned(DS[i]) then
      Continue;
    s:=s + cUsedEndOfPar[FTxtTemplate] + IndexToDatasetChar(i) + ' (' + DS[i].Name + ')' + cUsedEndOfPar[FTxtTemplate];
    for j:=0 to DS[i].FieldCount - 1 do
      s:=s + DS[i].Fields[j].FieldName + '; ';
  end;
  Result:=s + cUsedEndOfPar[FTxtTemplate];
end;

// Separate list values with separator, if value is not empty
function TRTFReport2.ProcSepList(const TagStr: string): string;
var
  i: integer;
  br, part, Sep, res: string;
  Repl, addsep: boolean;
begin
  Result:='';
  SepBracket(TagStr, br, Sep, res);
  addsep:=false;
  // find and clear separator
  Sep:=copy(FindStrFromTo(cConstSep, cConstSep, br, false), 2, 255);
  Sep:=ReplaceStr('*|', '\', Sep, true, false);
  br:=ReplaceStrFromTo(cConstSep, cConstSep, '', br, true, false, false);
  i:=0;
  repeat
    Inc(i);
    part:=FindStrPart(br, cListSep2, i, false);
    if (i = 1) and (part = '') then
      Continue;
    if part = '' then
      Break;
    res:=ProcTagStr(Trim(part), Repl);
    if addsep and (res <> '') then
      Result:=Result + Sep;
    addsep:=addsep or (res <> '');
    Result:=Result + res;
  until false;
end;

// Find and return first not empty value from list
function TRTFReport2.ProcFirstValue(const TagStr: string): string;
var
  i: integer;
  br, part, Sep, res: string;
  Repl: boolean;
begin
  Result:='';
  SepBracket(TagStr, br, Sep, res);
  i:=0;
  repeat
    Inc(i);
    part:=FindStrPart(br, cListSep2, i, false);
    if part = '' then
      Break;
    res:=ProcTagStr(Trim(part), Repl);
    Result:=res;
    if Result <> '' then
      Break;
  until false;
end;

// Insert external picture to report
function TRTFReport2.ProcPicture(const TagStr: string): string;
var
  fn, br, b1, b2, frm: string;
  Repl: boolean;
  B: TBitmap;
begin
  Result:='';
  SepBracket(TagStr, br, b1, b2);
  fn:=FindStrPart(br, cFormatSep, 1, false);
  if b2 <> '' then
    frm:=FindStrPart(b2, cFormatSep, 2, false)
  else
    frm:=FindStrPart(br, cFormatSep, 2, false);

  fn:=GetVarUserValue(fn, Repl);
  if Assigned(FOnGetPictureContent) then
  begin
    B:=TBitmap.Create;
    try
      FOnGetPictureContent(Self, fn, B);
      if B.Width>0 then
      begin
        Result:=CreateFormatPicture(-1, nil, '', B, frm);
        Repl:=false;
      end;
    finally
      B.Free;
    end;
  end;
  if Repl then
    Result:=CreateFormatPicture(-1, nil, fn, nil, frm);
end;
{$IFDEF BARCODESUPPORT}

// Create BarCode picture and insert to report
function TRTFReport2.ProcBarcode(const TagStr: string): string;
var
  i: integer;
  fn, br, b1, b2, frm, FPart: string;
  Repl: boolean;
  TypName, FName: string;
  FSize: integer;
  ShowCode: boolean;
  Bmp: TBitmap;
  bcRect: TRect;
  bcWidth: integer;
  bcType: TBarcodeType;
  bcFont: TFont;
begin
  // proc tag parts
  Result:='';
  frm:=FindStrPart(TagStr, cFormatSep, 2, false);
  fn:=FindStrPart(TagStr, cFormatSep, 1, false);
  SepBracket(fn, br, b1, b2);
  fn:=br;
  if IsConstantTag(fn) then
    fn:=copy(fn, 2, Length(fn) - 2)
  else
  begin
    fn:=ProcTagStr(fn, Repl);
    if not Repl then
    begin
      Result:='?' + fn + '?';
      Exit;
    end;
  end;
  if fn = '' then
    Exit;

  // proc type and format
  TypName:='';
  ShowCode:=false;
  FName:='Verdana';
  FSize:=0;
  i:=0;
  repeat
    Inc(i);
    FPart:=Trim(rtool.FindStrPart(frm, cParamSep, i, false));
    if FPart = '' then
      Break;
    if AnsiPosEx(cBCType, FPart, 1, false) = 1 then
      TypName:=FindStrPart(FPart, '=', 2, false);
    if SameText(cBCShowCode, FPart) then
      ShowCode:=true;
    if AnsiPosEx(cBCFontName, FPart, 1, false) = 1 then
      FName:=FindStrPart(FPart, '=', 2, false);
    if AnsiPosEx(cBCFontSize, FPart, 1, false) = 1 then
      FSize:=StrToIntDef(FindStrPart(FPart, '=', 2, false), 0);
  until false;

  // create barcode picture
  Bmp:=TBitmap.Create;
  try
    bcFont:=TFont.Create;
    try
      try
        // set default size for barcode
        bcRect:=Rect(5, 5, 490, 115);
        Bmp.Width:=500;
        Bmp.Height:=120;
        bcWidth:=2; // too points for single lines
        // set other props
        bcFont.Name:=FName;
        if FSize > 0 then
          bcFont.Size:=FSize;
        // type of barcode
        bcType:=GetBarCodeType(TypName);
        if SameText(TypName, 'CodeEANAuto') then
          case Length(fn) of
            7, 8:
              bcType:=bcCodeEAN8;
            12, 13:
              bcType:=bcCodeEAN13;
          else
            begin
              Result:='Invalid text length (CodeEANAuto): ' + fn;
              Exit;
            end;
          end;
        // check type
        if bcType = bcTypeUndefined then
        begin
          if TypName = '' then
            TypName:='undefined barcode type';
          Result:='?' + TypName + '?';
          Exit;
        end;
        // create barcode
        AsBarcode_interface.DrawBarCode(fn, bcType, Bmp.Canvas, bcRect, bcWidth, bcFont, FSize = 0, ShowCode, stpBottomCenter);
        // set picture size according barcode
        Bmp.Width:=bcRect.Right + 10;
        Bmp.Height:=bcRect.Bottom + 10;
        // insert pic
        Result:=CreateFormatPicture(-1, nil, '', Bmp, frm);
      except
        on E: Exception do
          Result:=E.Message;
      end;
    finally
      FreeAndNil(bcFont);
    end;
  finally
    FreeAndNil(Bmp);
  end;
end;
{$ELSE}

function TRTFReport2.ProcBarcode(const TagStr: string): string;
begin
  Result:='BarCode is not supported.';
end;
{$ENDIF}

// proc graphics BLOB field or external graphics file
function TRTFReport2.CreateFormatPicture(DSindex: integer; F: TField; const FileName: string; Bmp: TBitmap; FormatStr: string): string;
var
  Pic: TPicture;
  Enlarge, Insert, AutoQ: boolean;
  FPart, NameField, temp: string;
  i, MaxW, MaxH, Quality, Rot: integer;
  GrType: TrGraphicType;

  // return name of external file to link
  function GetPicName(Tag: string): string;
  var
    Repl: boolean;
    i: integer;
    Pic: string;
  begin
    Result:='';
    if Tag = '???' then
    begin
      // load filename from OnUserCalc events
      if Assigned(FOnUserCalc) then
      begin
        Repl:=true;
        FOnUserCalc(Self, Tag, Repl);
        if Repl then
          Result:=Tag;
      end;
    end
    else
    begin
      i:=0;
      repeat
        Inc(i);
        Pic:=Trim(rtool.FindStrPart(Tag, '+', i, false));
        if Pic = '' then
          Break;
        if IsConstantTag(Pic) then
          Result:=Result + copy(Pic, 2, Length(Pic) - 2)
        else
          Result:=Result + GetDBValue(DSindex, Pic, '');
      until false;
    end;
  end;

// return rtftag with linked file
  function LinkPic: string;
  var
    fn: string;
  begin
    // set file name for pic
    if NameField <> '' then
      fn:=GetPicName(NameField);
    if fn = '' then
      if DSindex = -1 then
      begin
        Inc(PicCount);
        fn:='ExtPic' + IntToStr(PicCount);
      end
      else
        fn:=IndexToDatasetChar(DSindex) + IntToStr(DS[DSindex].RecNo) + '_' + F.FieldName;
    fn:=fn + rpictool.GetFileTypeExtension(GrType);
    fn:=ExtractFilePath(FOutputFile) + fn;
    // save pic
    SavePictureToFile(Pic, fn, GrType);
    Result:=Format(cLinkPicture, [ExtractFileName(fn)]);
  end;

// return rtftag with wmetafile bits
  function InsertPic: string;
  var
    i: integer;
    picbytes, picrtf: ansistring;
    pictype: string;
    NewW, NewH: integer;
    Stream: TStringStream;

    { Wmf : TMetafile;
      WmfCanvas: TMetafileCanvas;
      RefDC: HDC;
      Len: UINT;
      //Pic: Pointer;
      Pch: PAnsiChar;
      }
  begin
    Result:='';
    // resize according quality and size
    if Quality < 10 then
      Quality:=10;
    if Quality > 100 then
      Quality:=100;
    if AutoQ then
      rpictool.SmoothResizePicture(Pic, MaxW div 20, MaxH div 20, Enlarge)
    else
      if Quality <> 100 then
        rpictool.SmoothResizePicture(Pic, Pic.Width * Quality div 100, Pic.Height * Quality div 100, Enlarge);

    Stream:=TStringStream.Create;
    try
      GrType:=rpictool.SavePictureToStream(Pic, Stream, GrType);
      picbytes:=ansistring(Stream.DataString);
      picrtf:='';
      for i:=1 to Length(picbytes) do
        picrtf:=picrtf + ansistring(IntToHex(ord(picbytes[i]), 2));

      if GrType = gtJPG then
        pictype:=cPictureTypeJPG
      else
        if GrType = gtPNG then
          pictype:=cPictureTypePNG
        else
          if GrType = gtBMP then
            pictype:=cPictureTypeMetafile
          else
            raise Exception.Create(sPictureErr);

      NewW:=Pic.Width * 20;
      if (MaxW > 0) and (NewW > MaxW) or (Enlarge and (NewW < MaxW)) then
        NewW:=MaxW;
      NewH:=MulDiv(NewW, Pic.Height, Pic.Width);
      if (MaxH > 0) and ((NewH > MaxH) or (Enlarge and (NewH < MaxH))) then
      begin
        NewH:=MaxH;
        NewW:=MulDiv(NewH, Pic.Width, Pic.Height);
      end;
      Result:=Format(cInsertPicture2b, [100, 100, Pic.Graphic.Width, Pic.Graphic.Height, NewW, NewH, pictype, picrtf]);
    finally
      FreeAndNil(Stream);
    end;
  end;

begin
  Pic:=TPicture.Create;
  try
    try
      // load pic form field or from file
      if Assigned(F) then // from field
      begin
        GrType:=ReadPictureFromField(F, Pic);
        if GrType = gtError then
          raise Exception.Create(sPictureErr);
      end
      else
        if FileName <> '' then // from external file
        begin
          if not FileExists(FileName) then
            raise Exception.Create(Format(sFileNotFound,[FileName]));
          GrType:=LoadPictureFromFile(Pic, FileName);
          if GrType = gtError then
            raise Exception.Create(sPictureErr);
        end
        else
          if Assigned(Bmp) then // from existing bitmap
          begin
            Pic.Bitmap.Assign(Bmp);
          end
          else
            raise Exception.Create(sInputErr);

      // proc format
      MaxW:=0;
      MaxH:=0;
      Rot:=0;
      Enlarge:=false;
      Insert:=false;
      Quality:=100;
      AutoQ:=false;
      NameField:='';
      i:=0;
      repeat
        Inc(i);
        FPart:=Trim(rtool.FindStrPart(FormatStr, cParamSep, i, false));
        if FPart = '' then
          Break;
        temp:=FindStrPart(FPart, '=', 2, false);
        temp:=UpdateToDecimalSeparator(temp);
        if AnsiPosEx(cPicMaxW, FPart, 1, false) = 1 then
          try
            MaxW:=StrToInt(temp) * 20;
          except
            MaxW:=0;
          end;
        if AnsiPosEx(cPicMaxWcm, FPart, 1, false) = 1 then
          try
            MaxW:=Round(StrToFloat(temp) * 567);
          except
            MaxW:=0;
          end;
        if AnsiPosEx(cPicMaxH, FPart, 1, false) = 1 then
          try
            MaxH:=StrToInt(temp) * 20;
          except
            MaxH:=0;
          end;
        if AnsiPosEx(cPicMaxHcm, FPart, 1, false) = 1 then
          try
            MaxH:=Round(StrToFloat(temp) * 567);
          except
            MaxH:=0;
          end;
        if AnsiPosEx(cPicRotation, FPart, 1, false) = 1 then
          try
            Rot:=StrToInt(temp);
          except
            Rot:=0;
          end;
        if SameText(cPicEnlarge, FPart) then
          Enlarge:=true;
        if AnsiPosEx(cPicNameField, FPart, 1, false) = 1 then
          NameField:=Trim(FindStrPart(FPart, '=', 2, false));
        if SameText(cPicInsertInRTF, FPart) then
          Insert:=true;
        if AnsiPosEx(cPicQuality, FPart, 1, false) = 1 then
          try
            temp:=Trim(FindStrPart(FPart, '=', 2, false));
            if AnsiPosEx('auto', temp, 1, false) = 1 then
              AutoQ:=true
            else
              Quality:=StrToInt(temp);
          except
            Quality:=100;
          end;
        if SameText(cPicSaveAsBMP, FPart) then
          GrType:=gtBMP;
        if rpictool.IsJPGSupported then
          if SameText(cPicSaveAsJPG, FPart) then
            GrType:=gtJPG;
        if rpictool.IsPNGSupported then
          if SameText(cPicSaveAsPNG, FPart) then
            GrType:=gtPNG;
      until false;
      // rotate pic
      if Rot <> 0 then
        rpictool.RotatePicture(Pic, Rot);
      // insert or link pic
      if Insert then
        Result:=InsertPic
      else
      begin
        // resize pic
        if (MaxW > 0) or (MaxH > 0) then
          rpictool.SmoothResizePicture(Pic, MaxW * 4 div 60, MaxH * 4 div 60, Enlarge);
        Result:=LinkPic;
      end;
    except
      on E: Exception do
        Result:=E.Message;
    end;
  finally
    FreeAndNil(Pic);
  end;
end;

function TRTFReport2.ProcNumToText(const TagStr: string; const TagFormat: string): string;
var
  fn, f1, f2, f3: string;
  Repl: boolean;
  val: double;
begin
  // proc tag parts
  Result:='';
  SepBracket(TagStr, fn, f1, f2);
  f1:=Trim(rtool.FindStrPart(TagFormat, cParamSep, 1, false));
  f2:=Trim(rtool.FindStrPart(TagFormat, cParamSep, 2, false));
  f3:=Trim(rtool.FindStrPart(TagFormat, cParamSep, 3, false));

  if IsConstantTag(fn) then
    fn:=copy(fn, 2, Length(fn) - 2)
  else
  begin
    fn:=ProcTagStr(fn, Repl);
    if not Repl then
    begin
      Result:='?' + fn + '?';
      Exit;
    end;
  end;
  fn:=RemoveThousandSeparator(fn);
  if fn = '' then
    Exit;
  try
    val:=StrToFloat(fn);
    if (f3 = '') or SameText(f3, 'CZ') then
      Result:=NumToText.NumToTextCZ(val, ' ', f1, f2)
    else
      Result:=NumToText.NumToTextEN(val, ' ', f1, f2);
  except
    Result:='?' + fn + '?';
  end;
end;

// Define variable
procedure TRTFReport2.ProcDefVar(const TagStr: string);
var
  s, b1, b2: string;
  Repl: boolean;
  i: integer;
begin
  SepBracket(TagStr, s, b1, b2);
  b1:=GetListPart(s, cListSep2);
  if b1 = '' then
    Exit;
  b2:=Trim(GetListPart(s, cListSep2));
  if (b2 = '') or IsConstantTag(b2) then
    s:=ClearStr(cConstSep, b2, false)
  else
    s:=ProcTagStr(Trim(b2), Repl);
  repeat
    i:=FVariables.IndexOfName(b1);
    if i >= 0 then
      FVariables.Delete(i);
  until i = -1;
  FVariables.Add(b1 + '=' + s);
end;

function TRTFReport2.ProcStrPrefix(const TagStr: string): string;

  function ExtractStrConst(s: string): string;
  begin
    s:=copy(s, 2, Length(s) - 2);
    if copy(s, 1, 1) = '''' then
      Result:=copy(s, 2, Length(s) - 2)
    else
      Result:=rtool.ReplaceStr(cTabStr, #9, s, true, false);
  end;

  function ProcStrPart(fparam: string; CS: boolean): string;
  var
    s, p1, p2, p3: string;
    pos1, pos2: integer;
    Repl: boolean;
  begin
    Result:='';
    p1:=GetListPart(fparam, cListSep2);
    p2:=GetListPart(fparam, cListSep2);
    p3:=GetListPart(fparam, cListSep2);
    if fparam <> '' then
      raise Exception.Create(fparam);
    s:=ProcTagStr(p1, Repl);
    if copy(p2, 1, 1) = cConstSep then
    begin
      p2:=ExtractStrConst(p2);
      pos1:=AnsiPosEx(p2, s, 1, CS)
    end
    else
      try
        pos1:=StrToInt(p2);
      except
        pos1:=StrToInt(ProcTagStr(p2, Repl));
      end;
    if copy(p3, 1, 1) = cConstSep then
    begin
      p3:=ExtractStrConst(p3);
      pos2:=AnsiPosEx(p3, s, pos1, CS)
    end
    else
      try
        pos2:=StrToInt(p3) + pos1;
      except
        pos2:=StrToInt(ProcTagStr(p3, Repl)) + pos1;
      end;
    if (pos1 > 0) and (pos2 > 0) then
      Result:=copy(s, pos1, pos2 - pos1);
  end;

  function ProcStrRepl(fparam: string; CS: boolean): string;
  var
    text, old, new: string;
    Repl: boolean;
  begin
    Result:='';
    text:=GetListPart(fparam, cListSep2);
    text:=ProcTagStr(text, Repl);
    while fparam<>'' do
    begin
      old:=GetListPart(fparam, cListSep2);
      new:=GetListPart(fparam, cListSep2);
      if copy(old, 1, 1) = cConstSep then
        old:=ExtractStrConst(old)
      else
        old:=ProcTagStr(old, Repl);
      if copy(new, 1, 1) = cConstSep then
        new:=ExtractStrConst(new)
      else
        new:=ProcTagStr(new, Repl);
      text:=rtool.ReplaceStr(old, new, text, true, CS);
    end;
    Result:=text;
  end;

  function ProcStrCase(fparam: string; idx: integer): string;
  var
    s, p1: string;
    Repl: boolean;
    i: integer;
  begin
    Result:='';
    p1:=GetListPart(fparam, cListSep2);
    if fparam <> '' then
      raise Exception.Create(fparam);
    s:=ProcTagStr(p1, Repl);
    case idx of
      1:
        s:=AnsiUpperCase(s);
      2:
        s:=AnsiLowerCase(s);
      3:
        begin
          s:=AnsiLowerCase(s);
          for i:=1 to Length(s) do
            if (i = 1) or (s[i - 1] = ' ') then
              s[i]:=AnsiUpperCase(s[i])[1];
        end;
      4:
        s:=Trim(s);
    end;
    Result:=s;
  end;

  function ProcStrName(FName, fparam: string): string;
  begin
    if SameText(FName, cStrPart) then
      Result:=ProcStrPart(fparam, false)
    else
      if SameText(FName, cStrPartCS) then
        Result:=ProcStrPart(fparam, true)
      else
        if SameText(FName, cStrRepl) then
          Result:=ProcStrRepl(fparam, false)
        else
          if SameText(FName, cStrReplCS) then
            Result:=ProcStrRepl(fparam, true)
          else
            if SameText(FName, cStrUpper) then
              Result:=ProcStrCase(fparam, 1)
            else
              if SameText(FName, cStrLower) then
                Result:=ProcStrCase(fparam, 2)
              else
                if SameText(FName, cStrProper) then
                  Result:=ProcStrCase(fparam, 3)
                else
                  if SameText(FName, cStrTrim) then
                    Result:=ProcStrCase(fparam, 4)
                  else
                    raise Exception.Create(FName);
  end;

var
  param, name, send: string;
begin
  try
    SepBracket(TagStr, param, name, send);
    if send <> '' then
      raise Exception.Create(send);
    send:=ProcStrName(name, param);
    Result:=StrToRtf(send);
  except
    on E: Exception do
      RaiseErr(4, TagStr + '; ' + Format(sTemplateIncorrectTag, [E.Message]));
  end;
end;

function TRTFReport2.ProcCase(const TagStr: string): string;
var
  i: integer;
  Tag, part, sv, v1, v2: string;
  Repl: boolean;
begin
  Result:='';
  // extract case parameters from bracket
  SepBracket(TagStr, Tag, v1, v2);
  i:=1;
  repeat
    part:=rtool.FindStrPart(Tag, cParamSep, i, true);
    // 1st param as source value
    if i = 1 then
      sv:=part
    else
    begin
      // next params:  compared value, result value
      v1:=rtool.FindStrPart(part, cListSep2, 1, true);
      v2:=rtool.FindStrPart(part, cListSep2, 2, true);
      if SameText(v1, '#ELSE#') or ((v1 <> '') and GetConditionResult(sv + v1)) then
      begin
        Result:=ProcTagStr(v2, Repl);
        Break;
      end;
    end;
    Inc(i);
  until part = '';
end;

procedure TRTFReport2.ProcExtendedInfo(var Str: string);
var
  s, n, v: string;
  i: integer;
begin
  // read all Info tags
  i:=1;
  repeat
    s:=FindTagStr(cExtendedInfo, Str, cTagCh, i, true);
    if s = '' then
      Break;
    FExtendedInfo.Add(s);
  until false;
  // call user defined event
  if Assigned(FOnReadInfo) then
    FOnReadInfo(Self);
  // proc all Info values
  for i:=0 to FExtendedInfo.Count - 1 do
  begin
    s:=FExtendedInfo[i];
    n:=Trim(rtool.FindStrPart(s, '=', 1, true));
    v:=rtool.FindStrPart(s, '=', 2, true);

    if SameText(n, 'SetPrinterName') then
      Self.PrinterName:=v;

    if SameText(n, 'DefaultCalcFormat') then
      Self.FDefaultCalcFormat:=v;
    if SameText(n, 'DefaultNumericFormat') then
      Self.DefaultNumericFormat:=v;
    if SameText(n, 'DefaultBooleanFormat') then
      Self.FDefaultBooleanFormat:=v;
    if SameText(n, 'DefaultDateFormat') then
      Self.FDefaultDateFormat:=v;
    if SameText(n, 'DefaultTimeFormat') then
      Self.FDefaultTimeFormat:=v;

    if SameText(n, 'ConvertScientificFormat') then
      Self.FConvertScientificFormat:=v = '1';
    if SameText(n, 'FieldNotFoundValue') then
      Self.FFieldNotFoundValue:=v;
  end;
end;

procedure TRTFReport2.ProcInsertFile(var TagStr: string);

  procedure InsFile(aTag: string; Convert: boolean);
  var
    s: string;
    i: integer;
    L: TStringList;
  begin
    i:=1;
    repeat
      // read tag
      s:=FindTagStr(aTag, TagStr, cTagCh, i, true);
      if s = '' then
        Break;
      // add or modify directory
      if Pos('\', s) = 0 then
        s:=ExtractFilePath(FReportFile) + ReplaceStr('/', '\', s, true, true);
      // load file content
      if FileExists(s) then
      begin
        L:=TStringList.Create;
        try
          L.Loadfromfile(s);
          s:=L.Text;
        finally
          FreeAndNil(L);
        end;
      end
      else
        s:=Format(sFileNotFound, [s]);
      // pro content
      if Convert then
        s:=ProcBlock(StrToRtf(s));
      Insert(s, TagStr, i);
    until false;
  end;

begin
  InsFile(cInsertFile, true);
  InsFile(cInsertRTFFile, false);
end;

// ======================================================================
// RTF Proccessing function
// proc all (open file, generating report, saving result)
procedure TRTFReport2.ProcAll;
var
{$IFDEF D2009P}
  AStr: ansistring;
{$ELSE}
  AStr: string;
{$ENDIF}
  FS: TFileStream;
  oldPrinterName, err: string;
  i, showtype: integer;

begin
  Aborted:=false;
  RecNbr:=1;
  RecNbrDS0:=-1;
  FExtendedInfo.Clear;
  PrinterName:='';
  if Assigned(FOnReportProcess) then
    FOnReportProcess(Self, 0, Aborted);
  Application.ProcessMessages;

  if Assigned(FOnGetTemplateStr) then
    FOnGetTemplateStr(Self, RepStr)
  else
  begin
    // read report file to string
    try
      FS:=TFileStream.Create(FReportFile, fmOpenRead + fmShareDenyNone);
      try
        SetLength(AStr, FS.Size);
        FS.Read(Pointer(AStr)^, FS.Size);
      finally
        FreeAndNil(FS);
      end;
    except
      raise Exception.Create(Format(sFileOpenErr, [FReportFile]) + #13 + sFileOpenErr2);
    end;
    RepStr:=string(AStr);
  end;

  // remove end of lines
  // skip picture tag => much faster
  if not FTxtTemplate then
    ProcAllSkipPictureTag(RepStr, ClearEOLPart);

  try
    // replace 0 charset with default charset
    RepStr:=ReplaceStr(cRtfCharset + '0', cRtfCharset + IntToStr(DefaultCharset), RepStr, true, false);
    // proc extended info tags
    ProcExtendedInfo(RepStr);
    // proc InsertFile tags
    ProcInsertFile(RepStr);
    // proc IF  (not IIF)
    ProcIf(false, true, RepStr);
    // proc all report
    RepStr:=ProcBlock(RepStr);
  finally
    FreeCalcField;
    FreeBookMarks;
  end;
  Application.ProcessMessages;
  if Aborted then
    Raise Exception.Create(sAborted);

  if Assigned(FModifyBeforeSave) then
  begin
    FModifyBeforeSave(Self, RepStr);
    if RepStr = '' then
      Exit;
  end;

  // save output file
{$IFDEF D2009P}
  AStr:=ansistring(RepStr);
{$ELSE}
  AStr:=string(RepStr);
{$ENDIF}
  repeat
    try
      if FTxtTemplate and FAppendToFile and FileExists(FOutputFile) then
      begin
        FS:=TFileStream.Create(FOutputFile, fmOpenWrite);
        FS.Position:=FS.Size;
      end
      else
        FS:=TFileStream.Create(FOutputFile, fmCreate);
      try
        FS.Write(Pointer(AStr)^, Length(AStr));
      finally
        FreeAndNil(FS);
      end;
      Break;
    except
      err:=Format(sFileSaveErr, [FOutputFile]) + #13 + sFileSaveErr2;
      if FShowSaveErrorDialog then
      begin
        if DlgQ(err + #13 + #13 + sTryAgainQ) <> mrYes then
          Abort;
      end
      else
        raise Exception.Create(err);
    end;
  until false;
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if Assigned(FOnReportProcess) then
    FOnReportProcess(Self, -1, Aborted);
  if FShowMaximized then
    showtype:=SW_SHOWMAXIMIZED
  else
    showtype:=SW_SHOWNORMAL;
  if ActionAfter = aaShow then
    RunFile(FOutputFile, '', '', showtype);
  if ActionAfter = aaPrint then
  begin
    if PrinterName <> '' then
    begin
      oldPrinterName:=GetDefaultPrinterName;
      if not SetDefaultPrinter(PrinterName, true) then
        RaiseErr(6, PrinterName);
    end;
    try
      for i:=1 to PrintCopies do
      begin
        if Assigned(FOnReportProcess) then
          FOnReportProcess(Self, -1, Aborted);
        Application.ProcessMessages;
        if Aborted then
          Raise Exception.Create(sAborted);
        PrintFileSimple(FOutputFile);
        Sleep(1000);
      end;
    finally
      if PrinterName <> '' then
        SetDefaultPrinter(oldPrinterName, true);
    end;
  end;
end;

// Proc block of RTF
function TRTFReport2.ProcBlock(const Str: string): string;
var
  ScanPos, IIfPos, proc: integer;
  s: string;
begin
  Result:='';
  s:=Str;
  repeat
    // test abortion
    Application.ProcessMessages;
    if Assigned(FOnReportProcess) then
      FOnReportProcess(Self, RecNbr, Aborted);
    if Aborted then
      raise Exception.Create(sAborted);
    if FShowProgressBar and (RecNbr mod 10 = 0) then
    begin
      if RecCountDS0 > 0 then
        proc:=Round(100.0 * RecNbrDS0 / RecCountDS0)
      else
        proc:=-1;
      ShowInfoProgressBar(sCreatingReport + #13 + Format(sRecord + ' %d', [RecNbr]), nil, DoAbort, proc);
    end;

    // find start of SCAN and IIF
    ScanPos:=AnsiPosEx(cUsrTagScan, s, 1, false);
    IIfPos:=AnsiPosEx(cUsrTagIIfStart, s, 1, false);
    if (IIfPos > 0) and ((ScanPos = 0) or (IIfPos < ScanPos)) then
    begin
      // IIF first
      // proc text till IIf
      Result:=Result + ProcRTF(copy(s, 1, IIfPos - 1));
      // copy and proc the rest (begining with IIF)
      s:=copy(s, IIfPos, Length(s));
      ProcIf(true, false, s);
    end
    else
    begin
      // Scan first
      if ScanPos = 0 then
        ScanPos:=Length(s) + 1;
      // proc text till scan
      Result:=Result + ProcRTF(copy(s, 1, ScanPos - 1));
      // copy and proc the rest (begining with scan)
      s:=copy(s, ScanPos, Length(s));
      Result:=Result + ProcScan(s);
    end;
  until s = '';
end;

// Proc RTF with SCAN tag (recurse for master/detail report)
function TRTFReport2.ProcScan(var s: string): string;
var
  ScInf: TScanInfo;
  ScanTag, ScanBlock, ScanEntry, ScanFooter: string;
  PageSep: string;
  RowNum: integer;

  // find ScanStr and divide Str to blocks ahead and behind of ScanStr
  function DivideStr(ScanStr, Str: string; var Block1, Block2: string): boolean;
  var
    p: integer;
  begin
    Block1:=Str;
    Block2:='';
    // find ScanStr
    p:=AnsiPosEx(ScanStr, Str, 1, false);
    Result:=p > 0;
    if Result then
    begin
      Block1:=copy(Str, 1, p - 1);
      Block2:=copy(Str, p + Length(ScanStr), Length(Str));
      RemovePar(Block2, true);
    end;
  end;

begin
  Result:='';
  if s = '' then
    Exit;
  ScInf:=DecodeScanLine(s);
  // find endscan
  ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagEndScan, false, false);
  if not DivideStr(ScanTag, s, ScanBlock, s) then
    RaiseErr(3, ScanTag);
  // move to first
  if (ScInf.DSindex >= 0) and not ScInf.NoFirst then
    FindFirst(ScInf.DSindex);
  // scan if not empty
  if (not ScInf.NoIfEmpty) or DSNotEOF(ScInf.DSindex) then
  begin
    // find scanentry
    ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagScanEntry, false, false);
    if DivideStr(ScanTag, ScanBlock, ScanEntry, ScanBlock) then
      Result:=Result + ProcBlock(ScanEntry)
    else
      ScanBlock:=ScanEntry;
    // find scanfooter
    ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagScanFooter, false, false);
    DivideStr(ScanTag, ScanBlock, ScanBlock, ScanFooter);
    PageSep:='';
    RowNum:=0;
    // proc block for each record (recurse)
    repeat
      if not DSNotEOF(ScInf.DSindex) then
        Break;
      if (ScInf.WhileCond <> '') and not GetConditionResult(ScInf.WhileCond) then
        Break;
      if (ScInf.SkipCond <> '') and GetConditionResult(ScInf.SkipCond) then
        Continue;
      if (ScInf.MaxRows > 0) and (RowNum >= ScInf.MaxRows) then
        Break;
      Inc(RowNum);
      Result:=Result + PageSep;
      ProcCalcField(ScInf.DSindex);
      Result:=Result + ProcBlock(ScanBlock);
      PageSep:='';
      if ScInf.NewPage then
        PageSep:=cUsedEndOfPage[FTxtTemplate]
      else
        if ScInf.NewCol then
          PageSep:=cRtfEndCol
        else
          if ScInf.NewPar then
            PageSep:=cRtfEndPar
          else
            if ScInf.NewLine then
              PageSep:=cRtfEndLine;
      PageSep:=ScInf.Sep + PageSep;
    until (ScInf.DSindex >= 0) and not FindNext(ScInf.DSindex);

    if (ScInf.MinRows > 0) then
    begin
      while RowNum < ScInf.MinRows do
      begin
        Inc(RowNum);
        Result:=Result + PageSep;
        Result:=Result + ProcBlock(ScanBlock);
      end;
    end;

    Result:=Result + ProcBlock(ScanFooter);
  end;
end;

// proc Scan command
function TRTFReport2.DecodeScanLine(var s: string): TScanInfo;
var
  scan: string;
  part, b1, b2, br, trimbr: string;
  i: integer;
begin
  Result.DS:='_';
  Result.DSindex:=-1;
  Result.WhileCond:='';
  Result.SkipCond:='';
  Result.NewLine:=false;
  Result.NewPar:=false;
  Result.NewCol:=false;
  Result.NewPage:=false;
  Result.Sep:='';
  Result.NoIfEmpty:=false;
  Result.NoFirst:=false;
  Result.MinRows:=0;
  Result.MaxRows:=0;
  i:=1;
  // load and delete scan line
  scan:=FindTagStr(cTag, s, cTagCh, i, true);
  // proc scan line
  i:=1;
  repeat
    part:=Trim(FindStrPart(scan, cParamSep, i, false));
    if part = '' then
      Break;
    // proc one part of scan line
    if i = 1 then
    begin
      Result.DSindex:=GetDSIndex(part);
      if Result.DSindex >= 0 then
        Result.DS:=IndexToDatasetChar(Result.DSindex)
      else
        Result.DS:='';
    end
    else
    begin
      SepBracket(part, br, b1, b2);
      trimbr:=Trim(br);
      b1:=Trim(b1);
      b2:=Trim(b2);
      if b2 <> '' then
        RaiseErr(5, scan);
      if b1 = '' then
        RaiseErr(5, scan);
      if AnsiSameText(cScanWhileCond, b1) then
        Result.WhileCond:=trimbr
      else
        if AnsiSameText(cScanSkipCond, b1) then
          Result.SkipCond:=trimbr
        else
          if AnsiSameText(cScanNewSep, b1) then
            Result.Sep:=br
          else
            if AnsiSameText(cScanNewLine, b1) then
              Result.NewLine:=true
            else
              if AnsiSameText(cScanNewPar, b1) then
                Result.NewPar:=true
              else
                if AnsiSameText(cScanNewCol, b1) then
                  Result.NewCol:=true
                else
                  if AnsiSameText(cScanNewPage, b1) then
                    Result.NewPage:=true
                  else
                    if AnsiSameText(cScanNoIfEmpty, b1) then
                      Result.NoIfEmpty:=true
                    else
                      if AnsiSameText(cScanNoFirst, b1) then
                        Result.NoFirst:=true
                      else
                        if AnsiSameText(cScanCalc, b1) then
                          InitCalcField(Result.DSindex, trimbr)
                        else
                          if AnsiSameText(cScanMinRows, b1) then
                            Result.MinRows:=StrToIntDef(trimbr, 0)
                          else
                            if AnsiSameText(cScanMaxRows, b1) then
                              Result.MaxRows:=StrToIntDef(trimbr, 0)
                            else
                              RaiseErr(5, scan);
    end;
    Inc(i);
  until false;
  RemovePar(s, true);
end;

// Proc if tags
procedure TRTFReport2.ProcIf(IIF, ProcAllIf: boolean; var Str: string);
var
  StartTag, ElseTag, EndTag: string;
  AllTag, IfTag, cond, b1, b2, res: string;
  StartPos, EndPos, ElsePos, p: integer;

  function FindEndPos(SPos: integer; s: string): integer;
  var
    EPos: integer;
  begin
    EPos:=AnsiPosEx(EndTag, s, SPos, false);
    repeat
      // nested if?
      // try to find another IIF
      SPos:=AnsiPosEx(StartTag, s, SPos + Length(StartTag), false);
      if (SPos = 0) or (SPos > EPos) then
        Break;
      // another IIF inside, find next ENDIIF
      EPos:=AnsiPosEx(EndTag, s, EPos + Length(EndTag), false);
    until EPos = 0;
    if EPos = 0 then
      RaiseErr(3, copy(Str, StartPos, 100) + '-' + EndTag);
    Result:=EPos;
  end;

  function FindElsePos(SPos: integer; s: string): integer;
  var
    ElsePos: integer;
  begin
    ElsePos:=AnsiPosEx(ElseTag, s, SPos, false);
    repeat
      // nested if?
      // try to find another IIF
      // SPos:=AnsiPosEx(StartTag, s, SPos + Length(StartTag), false);
      SPos:=AnsiPosEx(StartTag, s, SPos, false);
      if (SPos = 0) or (SPos > ElsePos) then
        Break;
      // another IIF inside, find end of this IIF
      SPos:=FindEndPos(SPos, s);
      if SPos > ElsePos then
        // ELSEIIF inside another IIF, try to find another ELSEIIF
        ElsePos:=AnsiPosEx(ElseTag, s, SPos, false);
    until ElsePos = 0;
    Result:=ElsePos;
  end;

begin
  // set if tags
  StartTag:=cUsrTagIfStart;
  ElseTag:=cUsrTagElseIf;
  EndTag:=cUsrTagEndIf;
  if IIF then
  begin
    StartTag:=cUsrTagIIfStart;
    ElseTag:=cUsrTagElseIIf;
    EndTag:=cUsrTagEndIIf;
  end;
  // proc if tags
  repeat
    // extract whole string between IF and ENDIF
    StartPos:=AnsiPosEx(StartTag, Str, 1, false);
    if StartPos = 0 then
      Break;
    EndPos:=FindEndPos(StartPos, Str);
    AllTag:=copy(Str, StartPos, EndPos - StartPos);
    // extract conditions
    p:=1;
    IfTag:=FindTagStr(cTag, AllTag, cTagCh, p, true);
    SepBracket(IfTag, cond, b1, b2);
    // proc ELSE tag
    b1:=AllTag;
    b2:='';
    ElsePos:=FindElsePos(1, AllTag);
    if ElsePos > 0 then
    begin
      b1:=copy(AllTag, 1, ElsePos - 1);
      b2:=copy(AllTag, ElsePos + Length(ElseTag), Length(AllTag));
    end;
    // find and insert result
    if GetConditionResult(cond) then
      res:=b1
    else
      res:=b2;

    // remove \par after IF (delete line in tables)
    if CheckExtractLineAfterIf then
      RemovePar(res, false);
    Delete(Str, StartPos, EndPos - StartPos + Length(EndTag));
    Insert(res, Str, StartPos);
    if not ProcAllIf then
      Break;
  until false;
end;

function TRTFReport2.CheckExtractLineAfterIf: boolean;
var
  b: boolean;
  v: string;
begin
  v:=GetVarUserValue(cUserVarIfExtractLine, b);
  Result:=FExtractEOLAfterIf or (b and (v = '1'));
end;

// proc RTF text
function TRTFReport2.ProcRTF(const Str: string): string;
var
  d, tagpos: integer;
  s, TagStr, res, spec: string;
  Repl: boolean;
begin
  // proc IIF
  s:=Str;
  ProcIf(true, true, s);
  // proc other tags
  tagpos:=0;
  repeat
    // find tag
    TagStr:=FindTagStr(cTag, s, cTagCh, tagpos, false);
    if tagpos = 0 then
      Break;
    d:=Pos(cTagCh, cTag) - 1;
    // skip MSWord field format strins and unussual long TAGS (error)
    if Length(TagStr) > 0 then
      try
        if CharInSet(TagStr[1], ['@', '*', '#']) or (Length(TagStr) > 250) then
        begin
          Inc(tagpos, d);
          Continue;
        end;
      except
      end;
    spec:=ExtractSpecChar(TagStr);
    res:=ProcTagStr(TagStr, Repl);
    res:=StrToRtf(res);
    if Repl then
    begin
      // delete parameter and insert value
      FindTagStr(cTag, s, cTagCh, tagpos, true);
      res:=res + spec;
      d:=Length(res);
      Insert(res, s, tagpos);
    end;
    Inc(tagpos, d);
  until false;
  Result:=s;
end;

// Proc Tag and return tag value
// TagStr is command or value
// Value has format X.Field_name:Format or VariableName:Format or =CalcString:Format
function TRTFReport2.ProcTagStr(const TagStr: string; var Repl: boolean): string;
var
  Tag, s, frm, w: string;
  p, wl, wr: integer;

  procedure ExtractFormat;
  var
    br, b1, b2: string;
  begin
    // find format text after bracket or text if bracket not used
    SepBracket(Tag, br, b1, b2);
    if b2 <> '' then
      // after bracket
      frm:=b2
    else
      if br = '' then
        // text without bracket
        frm:=b1
      else
        // bracket without format
        frm:='';

    // find separator in format text
    p:=Pos(cFormatSep, frm);
    if p = 0 then
      frm:=''
    else
    begin
      // extract format
      frm:=copy(frm, p + Length(cFormatSep), Length(frm));
      Tag:=copy(Tag, 1, Length(Tag) - Length(frm) - Length(cFormatSep));
    end;

    // proc format strings
    // :<width or :>width or :<width:format or :>width:format
    wl:=0;
    wr:=0;
    w:='';
    if (frm <> '') and CharInSet(frm[1], ['<', '>']) then
    begin
      // 1st part as width number
      w:=FindStrPart(frm, cFormatSep, 1, false);
      if copy(w, 1, 1) = '<' then
        wl:=StrToIntDef(copy(w, 2, 255), 0)
      else
        if copy(w, 1, 1) = '>' then
          wr:=StrToIntDef(copy(w, 2, 255), 0);
      // 2nd part as format string
      p:=Pos(cFormatSep, frm);
      if p = 0 then
        frm:=''
      else
        frm:=copy(frm, p + Length(cFormatSep), Length(frm));
    end;
  end;

  procedure ProcTagWithFormat;
  begin
    ExtractFormat;
    if Tag <> '' then
    begin
      // SepList
      if AnsiPosEx(cSepList, Tag, 1, false) = 1 then
        s:=ProcSepList(Tag)
      else
        // FirstValue
        if AnsiPosEx(cFirstValue, Tag, 1, false) = 1 then
          s:=ProcFirstValue(Tag)
        else
          // Case
          if AnsiPosEx(cCase, Tag, 1, false) = 1 then
            s:=ProcCase(Tag)
          else
            // NumToText
            if AnsiPosEx(cNumToText, Tag, 1, false) = 1 then
              s:=ProcNumToText(Tag, frm)
            else
              // String routine
              if AnsiPosEx(cStrPrefix, Tag, 1, false) = 1 then
                s:=ProcStrPrefix(Tag)
              else
                if Tag[1] = '=' then
                begin
                  if frm = '' then
                    frm:=FDefaultCalcFormat;
                  s:=GetCalcValue(copy(Tag, 2, Length(Tag)), frm, false)
                end
                else
                begin
                  if IsFieldTag(Tag) then
                    s:=GetDBValue(DatasetCharToIndex(Tag[1]), copy(Tag, 3, Length(Tag)), frm)
                  else
                  begin
                    s:=GetVarUserValue(Tag, Repl);
                    if Repl then
                      s:=ProcRTF(s);
                  end;
                end;
    end;
    // align text according format width
    while Length(s) < wl do
      s:=s + ' ';
    while Length(s) < wr do
      s:=' ' + s;
    // truncate text
    if wr > wl then
      wl:=wr;
    if (wl > 0) and (Length(s) > wl) then
      s:=copy(s, 1, wl);
  end;

begin
  Result:='';
  s:='';
  Tag:=Trim(RtfToStr(TagStr));
  if Length(Tag) < 2 then
  begin
    Repl:=false;
    Exit;
  end;
  Repl:=true;
  // move dataset
  if AnsiPosEx(cScanNext, Tag, 1, false) = 1 then
    FindNext(GetDSIndex(Tag))
  else
    if AnsiPosEx(cScanPrev, Tag, 1, false) = 1 then
      FindPrev(GetDSIndex(Tag))
    else
      if AnsiPosEx(cScanFirst, Tag, 1, false) = 1 then
        FindFirst(GetDSIndex(Tag))
      else
        if AnsiPosEx(cScanLast, Tag, 1, false) = 1 then
          FindLast(GetDSIndex(Tag))
        else
          // bookmark dataset
          if AnsiPosEx(cScanGetBookmark, Tag, 1, false) = 1 then
            GetDSBookmark(GetDSIndex(Tag))
          else
            if AnsiPosEx(cScanGotoBookmark, Tag, 1, false) = 1 then
              GotoDSBookmark(GetDSIndex(Tag))
            else
              // Picture
              if AnsiPosEx(cPicture, Tag, 1, false) = 1 then
                s:=ProcPicture(Tag)
              else
                // BarCode
                if AnsiPosEx(cBarCode, Tag, 1, false) = 1 then
                  s:=ProcBarcode(Tag)
                else
                  // ResetCalc;
                  if AnsiSameText(cResetCalc, Tag) then
                    FreeCalcField
                  else
                    // Def var
                    if AnsiPosEx(cDefVar, Tag, 1, false) = 1 then
                      ProcDefVar(Tag)
                    else
                      // ListAllFields
                      if AnsiSameText(cListAllFields, Tag) then
                        s:=ListAllFields
                      else
                        ProcTagWithFormat;
  // Result:=StrToRtf(s);
  Result:=s;
end;

// Return value of Dataset Field
// FieldName: Name of field or #index
// (FormatStr or DisplayFormat is used for formating)
function TRTFReport2.GetDBValue(DSindex: integer; const FieldName, FormatStr: string): string;
var
  Fld: TField;
  s, frm: string;
  AsText: boolean;

  // proc MEMO and text BLOB field
  function ProcMemo(mstr: string): string;
  var
    RTFBlob: boolean;
    i, j: integer;
    s1, s2: string;
    REdit: TRichEdit;
    Stream: TStringStream;
  begin
    if AnsiSameText(frm, cMemoPlainFormat) or ((frm = '') and (FDefaultMemoFormat = mfPlain)) then
      try
        // convert rtf to plaintext
        REdit:=TRichEdit.CreateParented(HWND_MESSAGE);
        try
          REdit.WordWrap:=false;
          REdit.Width:=8000;
          REdit.PlainText:=false;
          Stream:=TStringStream.Create(Fld.AsString);
          try
            REdit.Lines.LoadFromStream(Stream);
          finally
            FreeAndNil(Stream);
          end;
          REdit.PlainText:=true;
          mstr:=REdit.Lines.Text;
          // replace last #0D#0A
          if copy(mstr, Length(mstr) - 1, 2) = cStrEOL then
            mstr:=copy(mstr, 1, Length(mstr) - 2);
        finally
          FreeAndNil(REdit);
        end;
      except
        on E: Exception do
        begin
          DlgE('RTF to PlainText error: ' + E.Message);
          raise ;
        end;
      end;

    RTFBlob:=AnsiSameText(frm, cMemoRTFFormat) or AnsiSameText(frm, cMemoRTFLineFormat) or
      ((frm = '') and ((FDefaultMemoFormat = mfRTF) or (FDefaultMemoFormat = mfRTFLine) or ((FDefaultMemoFormat = mfAutoDetect) and (Pos(cRtfEndPar, mstr) > 0))
        ));
    if RTFBlob then
    begin
      mstr:=ReplaceStr(cTxtEndPar, '', mstr, true, false);
      for i:=0 to 9 do
        mstr:=ReplaceStr(cRtfViewKind + IntToStr(i), '', mstr, true, false);
    end;
    // replace last #0
    while copy(mstr, Length(mstr), 1) = #0 do
      mstr:=copy(mstr, 1, Length(mstr) - 1);
    if AnsiSameText(frm, cMemoRTFLineFormat) or ((frm = '') and (FDefaultMemoFormat = mfRTFLine)) then
    begin
      // replace last \par
      i:=0;
      repeat
        j:=i;
        i:=rtool.AnsiPosEx(cRtfEndPar, mstr, i + 1, false);
      until i = 0;
      if j > 0 then
        Delete(mstr, j, Length(cRtfEndPar));
    end;
    // memo replaces
    for i:=0 to FMemoReplaces.Count - 1 do
    begin
      s1:=rtool.FindStrPart(FMemoReplaces[i], '=', 1, false);
      s2:=rtool.FindStrPart(FMemoReplaces[i], '=', 2, false);
      s1:=rtool.ReplaceStr(cTabStr, #9, s1, true, false);
      s2:=rtool.ReplaceStr(cTabStr, #9, s2, true, false);
      if (s1) = '' then
        Continue;
      mstr:=ReplaceStr(s1, s2, mstr, true, false);
    end;

    Result:=mstr;
  end;

  function IsImageField: boolean;
  begin
    Result:=rdbtool.IsGraphicField(Fld) or SameText(FVariables.Values[Fld.FieldName], 'image');
    // or (AnsiContainsText(Fld.FieldName, 'picture')) or (AnsiContainsText(Fld.FieldName, 'image'));
  end;

begin
  Result:='unknown dataset';
  if DS[DSindex] = nil then
    Exit;
  Result:='closed dataset';
  if not DSActive(DSindex) then
    Exit;
  // EOF while using NEXT or PREV
  Result:='';
  frm:=FormatStr;
  if DS_eof[DSindex] then
    Exit;
  try
    AsText:=true;
    if FieldName = '#' then
    begin
      Result:=IntToStr(DS[DSindex].RecNo);
      Exit;
    end;
    if FieldName = '##' then
    begin
      Result:=IntToStr(DS[DSindex].RecordCount);
      Exit;
    end;
    if copy(FieldName, 1, 1) = '#' then
      Fld:=DS[DSindex].Fields[StrToInt(copy(FieldName, 2, 255))]
    else
      Fld:=DS[DSindex].FieldByName(FieldName);
    // special format
    if AnsiSameText(frm, cLabel) then
      s:=Fld.DisplayLabel
    else
      if Fld.IsNull then
        s:=''
      else
        case Fld.DataType of
          ftString, ftWideString:
            begin
              if frm = '' then
                s:=Fld.Text
              else
                if frm = '|*' then
                  s:=rtool.ReplaceStr('\', '\\', Fld.AsString)
                else
                  s:=FormatMaskText(frm + ';0;', Fld.AsString);
            end;
          ftSmallint, ftInteger, ftWord, ftBCD, ftAutoInc, ftFloat, ftCurrency:
            begin
              if frm = '' then
                frm:=TNumericField(Fld).DisplayFormat;
              if frm = '' then
                frm:=FDefaultNumericFormat;
              s:=CheckScientificFormat(FormatFloat(frm, Fld.AsFloat));
            end;
          ftBoolean:
            begin
              if frm = '' then
                frm:=(Fld as TBooleanField).DisplayValues;
              if frm = '' then
                frm:=FDefaultBooleanFormat;
              if (frm = '') or (Pos(';', frm) = 0) then
                s:=Fld.Text
              else
                if Fld.AsBoolean then
                  s:=copy(frm, 1, Pos(';', frm) - 1)
                else
                  s:=copy(frm, Pos(';', frm) + 1, Length(frm));
            end;
          ftDate:
            begin
              if frm = '' then
                frm:=TDateTimeField(Fld).DisplayFormat;
              if frm = '' then
                frm:=FDefaultDateFormat;
              if frm = '' then
                s:=DateToStr(Fld.AsDateTime)
              else
                s:=FormatDateTime(frm, Fld.AsDateTime);
            end;
          ftTime:
            begin
              if frm = '' then
                frm:=TDateTimeField(Fld).DisplayFormat;
              if frm = '' then
                frm:=FDefaultTimeFormat;
              if frm = '' then
                s:=TimeToStr(Fld.AsDateTime)
              else
                s:=FormatDateTime(frm, Fld.AsDateTime);
            end;
          ftDateTime:
            begin
              if frm = '' then
                frm:=TDateTimeField(Fld).DisplayFormat;
              if frm = '' then
                frm:=FDefaultDateFormat;
              s:=FormatDateTime(frm, Fld.AsDateTime);
            end;
          ftTimeStamp:
            begin
              if frm = '' then
                frm:=TSQLTimeStampField(Fld).DisplayFormat;
              if frm = '' then
                frm:=FDefaultDateFormat;
              s:=FormatDateTime(frm, Fld.AsDateTime);
            end;
          ftMemo, ftWideMemo, ftFmtMemo, ftGraphic, ftBlob:
            begin
              if (Fld.DataType in [ftMemo, ftWideMemo, ftFmtMemo]) or not IsImageField then
              begin
                s:=ProcMemo(Fld.AsString);
                if Pos('\\', s) > 0 then
                  s:=ProcRTF(s);
              end
              else
              begin
                s:=CreateFormatPicture(DSindex, Fld, '', nil, frm);
                AsText:=false;
              end;
            end;
        else
          s:=Fld.Text;
        end;
    // Replacing EOL with \par (RTF)
    if AsText and not FTxtTemplate then
      s:=ReplaceStr(cTxtEndPar, cRtfEndPar, s, true, false);
  except
    on E: Exception do
    begin
      if RaiseUnknownField then
        RaiseErr(1, FieldName);
      if FFieldNotFoundValue = '' then
        s:='?' + FieldName + '?' + E.Message + '?'
      else
        s:=FFieldNotFoundValue;
    end;
  end;
  if FTxtCodePage <> 0 then
    s:=string(rtool.WideStringToString(s, FTxtCodePage));
  Result:=s;
end;

// Return value of variable
function TRTFReport2.GetVarUserValue(const VarName: string; var Replaced: boolean): string;
var f, p, s: string;
  res: variant;
  val: double;
  arrCount, idx: integer;
  Fld: TField;
  arr: array of variant;
begin
  Result:=VarName;
  Replaced:=false;
  if FVariables.IndexOfName(VarName) >= 0 then
  begin
    Result:=FVariables.Values[VarName];
    Replaced:=true;
  end
  else
    if Assigned(FOnUserCalcParams) then
    begin
      SepBracket(VarName,p,f,s);
      if s<>'' then
        RaiseErr(5, VarName);
      res:='';
      arrCount:=0;
      repeat
        s:=rtool.RemoveStrPart(p,cParamSep);
        if s<>'' then
        begin
          SetLength(arr,arrCount+1);
          Fld:=nil;
          if IsConstantTag(s) then
            arr[arrCount]:=copy(s, 2, Length(s) - 2)
          else
          begin
            if IsFieldTag(s) then
            begin
              idx:=self.DatasetCharToIndex(s[1]);
              if DSActive(idx) then
                Fld:=DS[idx].FindField(copy(s,3,Length(s)));
            end;
            if Assigned(Fld) then
              arr[arrCount]:=Fld.AsVariant
            else
              if IsNumericTag(s, val) then
                arr[arrCount]:=val
              else
                arr[arrCount]:=s;
          end;
          Inc(arrCount);
        end;
      until p='';
      FOnUserCalcParams(Self, f, arr, res, Replaced);
      if not Replaced then
        Result:=VarName
      else
        Result:=res;
    end
    else
    if Assigned(FOnUserCalc) then
    begin
      FOnUserCalc(Self, Result, Replaced);
      if not Replaced then
        Result:=VarName;
    end;
  if FTxtCodePage <> 0 then
    Result:=string(rtool.WideStringToString(Result, FTxtCodePage));
end;

// Return value of calc string
function TRTFReport2.GetCalcNumValue(const CalcStr: string; ErrFieldAsZero: boolean): double;

// Subfunction, return value of Str
  function GetVal(Str: string): double;
  var
    idx, F: integer;
    s, p1, p2: string;
    err, Repl, CS: boolean;
    Fld: TField;
  begin
    Result:=0;
    Str:=Trim(Str);
    if Str = '' then
      Exit;
    // Check numeric value
    err:=not IsNumericTag(Str,Result);
    // Check Field value
    if err then
      try
        if Str[2] = '.' then
        begin
          idx:=DatasetCharToIndex(Str[1]);
          p1:=copy(Str, 3, Length(Str));
          if p1 = '#' then
            Result:=DS[idx].RecNo
          else
            if p1 = '##' then
              Result:=DS[idx].RecordCount
            else
            begin
              Fld:=DS[idx].FieldByName(p1);
              if Fld is TBooleanField then
                Result:=ord(Fld.AsBoolean)
              else
                if ErrFieldAsZero then
                begin
                  try
                    Result:=Fld.AsFloat;
                  except
                    Result:=0;
                  end;
                end
                else
                  Result:=Fld.AsFloat;
            end;
          err:=false;
        end;
      except
      end;
    // Check calc functions
    if err then
      for idx:=0 to 9 do
      begin
        F:=1;
        s:=FindTagStr(cCalcFunc[idx], Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 1) then
        begin
          Result:=GetCalcFieldValue(idx, s);
          err:=false;
          Break;
        end;
      end;
    // Check datetime functions
    if err then
    begin
      if SameText(Str, cGetToday) then
      begin
        Result:=Trunc(Date);
        err:=false;
      end;
      if SameText(Str, cGetNow) then
      begin
        Result:=Frac(Time);
        err:=false;
      end;
    end;
    // check math Abs
    if err then
      try
        s:=FindTagStr(cMathAbs, Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 0) then
        begin
          Result:=Abs(GetCalcNumValue(s, false));
          // if Length(s)>1 then s:=ProcTagStr(s, Repl);
          // Result:=Abs(StrToFloat(s));
          err:=false;
        end;
      except
      end;
    // check string routine "strlen"
    if err then
      try
        s:=FindTagStr(cStrLen, Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 1) then
        begin
          s:=ProcTagStr(s, Repl);
          Result:=Length(s);
          err:=false;
        end;
      except
      end;
    // check string routine "strpos"
    if err then
      try
        s:=FindTagStr(cStrPos, Str, cTagCh, F, false);
        CS:=false;
        if not((F = 1) and (Length(s) > 1)) then
        begin
          s:=FindTagStr(cStrPosCS, Str, cTagCh, F, false);
          CS:=true;
        end;
        if (F = 1) and (Length(s) > 1) then
        begin
          p1:=GetListPart(s, cListSep2);
          p2:=GetListPart(s, cListSep2);
          // if s<>'' then raise Exception.Create(s);
          p1:=ProcTagStr(p1, Repl);
          if copy(p2, 1, 1) = cConstSep then
            p2:=copy(p2, 2, Length(p2) - 2)
          else
            p2:=ProcTagStr(p2, Repl);
          Result:=AnsiPosEx(p2, p1, 1, CS);
          err:=false;
        end;
      except
      end;
    // Check Variable value
    if err then
      try
        Result:=StrToFloat(GetVarUserValue(Str, Repl));
        err:=false;
      except
      end;
    if err then
      RaiseErr(4, Str);
  end;

// Subfunction, return value of CalcStr
  function Calc(CalcStr: string): double;
  var
    Bracket, b1, b2: string;

    // Subfunction, return sign value
    function DoSign(Str, Sign: string; var res: double): boolean;
    var
      p: integer;
      part1, part2: string;
      res1, res2: double;
    begin
      Result:=false;
      p:=Pos(Sign, Str);
      if p = 0 then
        Exit;
      part1:=Trim(copy(Str, 1, p - 1));
      part2:=Trim(copy(Str, p + Length(Sign), Length(Str)));
      if (Sign = '-') and (Length(part1) > 0) and (CharInSet(part1[Length(part1)], ['['])) then
        Exit;
      Result:=true;
      res1:=Calc(part1);
      res2:=Calc(part2);
      if Sign = '+' then
        res:=res1 + res2
      else
        if Sign = '-' then
          res:=res1 - res2
        else
          if Sign = '*' then
            res:=res1 * res2
          else
            if Sign = '/' then
              try
                res:=res1 / res2;
              except
                res:=0;
              end;
    end;

  begin
    Result:=0;
    CalcStr:=Trim(CalcStr);
    if CalcStr = '' then
      Exit;
    // Calculate brackets
    repeat
      SepBracket(CalcStr, Bracket, b1, b2);
      if Bracket = '' then
        Break;
      CalcStr:=b1 + FloatToStrF(Calc(Bracket), ffFixed, 80, 40) + b2;
    until false;
    // Calculate all CalcStr
    if not DoSign(CalcStr, '+', Result) then
      if not DoSign(CalcStr, '-', Result) then
        if not DoSign(CalcStr, '*', Result) then
          if not DoSign(CalcStr, '/', Result) then
            Result:=GetVal(CalcStr);
  end;

begin
  Result:=Calc(CalcStr);
end;

// Return value of calc string
function TRTFReport2.GetCalcValue(const CalcStr, FormatStr: string; ErrFieldAsZero: boolean): string;
var
  v: double;
begin
  v:=GetCalcNumValue(CalcStr, ErrFieldAsZero);
  if (copy(FormatStr, 1, 1) = cDateSep) and (copy(FormatStr, Length(FormatStr), 1) = cDateSep) then
    Result:=FormatDateTime(copy(FormatStr, 2, Length(FormatStr) - 2), v)
  else
  begin
    Result:=CheckScientificFormat(FormatFloat(FormatStr, v));
  end;
end;

// proc condition string
function TRTFReport2.GetConditionResult(const CondStr: string): boolean;
var
  a, i, p: integer;
  br, cond1, cond2: string;
  isnum1, isnum2, Repl: boolean;
  dt: TDateTime;

  function GetCondValue(var Str: string): boolean;
  var
    c, F: string;
  begin
    Result:=false;
    c:=Str;
    try
      // string constant? "test"
      if IsConstantTag(c) then
        Str:=ClearStr(cConstSep, RtfToStr(c), false)
      else
        // string date? #YYYYMMDDhhmmss#
        if (c[1] = cDateSep) and (c[Length(c)] = cDateSep) then
        begin
          Str:=ClearStr(cDateSep, c, false);
          if (Length(Str) = 8) or (Length(Str) = 14) then
          begin
            dt:=EncodeDate(StrToInt(copy(Str, 1, 4)), StrToInt(copy(Str, 5, 2)), StrToInt(copy(Str, 7, 2)));
            if Length(Str) = 14 then
              dt:=dt + EncodeTime(StrToInt(copy(Str, 9, 2)), StrToInt(copy(Str, 11, 2)), StrToInt(copy(Str, 13, 2)), 0);
            Str:=FloatToStrF(dt, ffFixed, 80, 40);
            Result:=true;
          end
          else
            raise Exception.Create('');
        end
        else
        {
          // String routine
          if AnsiPosEx(cStrPrefix,c,1,false)=1 then Str:=ProcStrPrefix(c)
          else
          }
        // last change - calc value
        begin
          Str:=GetCalcValue(c, '', false);
          Result:=true;
        end;
    except
      try
        if IsFieldTag(c) then
        begin
          F:=copy(c, 3, Length(c));
          Str:=GetDBValue(DatasetCharToIndex(c[1]), F, '');
        end
        else
        begin
          Repl:=false;
          Str:=GetVarUserValue(c, Repl);
          if not Repl then
            raise Exception.Create('');
        end;
      except
        RaiseErr(4, c);
      end;
    end;
  end;

begin
  // proc brackets
  SepBracket(CondStr, br, cond1, cond2);
  if br <> '' then
  begin
    Result:=GetConditionResult(cond1 + cCondResArr[GetConditionResult(br)] + cond2);
    Exit;
  end;
  // not sign
  if AnsiPosEx(cCondNot, Trim(CondStr), 1, false) = 1 then
  begin
    Result:=not GetConditionResult(FindStrPart(CondStr, cCondNot, 2, false));
    Exit;
  end;
  // and sign
  p:=AnsiPosEx(cCondAnd, CondStr, 1, false);
  if p > 0 then
  begin
    cond1:=copy(CondStr, 1, p - 1);
    cond2:=copy(CondStr, p + Length(cCondAnd), Length(CondStr));
    Result:=GetConditionResult(cond1) and GetConditionResult(cond2);
    Exit;
  end;
  // or sign
  p:=AnsiPosEx(cCondOr, CondStr, 1, false);
  if p > 0 then
  begin
    cond1:=copy(CondStr, 1, p - 1);
    cond2:=copy(CondStr, p + Length(cCondOr), Length(CondStr));
    Result:=GetConditionResult(cond1) or GetConditionResult(cond2);
    Exit;
  end;

  Result:=false;
  i:=-1;
  // find sign
  for a:=0 to High(cCondFunc) do
  begin
    p:=Pos(cCondFunc[a], CondStr);
    if p > 0 then
      i:=a;
    if i >= 0 then
      Break;
  end;
  if i >= 0 then
  begin
    // sign exist
    cond1:=Trim(copy(CondStr, 1, p - 1));
    cond2:=Trim(copy(CondStr, p + Length(cCondFunc[i]), Length(CondStr)));
    isnum1:=GetCondValue(cond1);
    isnum2:=GetCondValue(cond2);
    if isnum1 and isnum2 then
    begin
      // numeric values
      case i of
        0, 4:
          Result:=StrToFloat(cond1) = StrToFloat(cond2);
        1:
          Result:=StrToFloat(cond1) >= StrToFloat(cond2);
        2:
          Result:=StrToFloat(cond1) <= StrToFloat(cond2);
        3:
          Result:=StrToFloat(cond1) <> StrToFloat(cond2);
        5:
          Result:=StrToFloat(cond1) > StrToFloat(cond2);
        6:
          Result:=StrToFloat(cond1) < StrToFloat(cond2);
      end;
    end
    else
    begin
      // string values
      case i of
        0:
          Result:=cond1 = cond2;
        1:
          Result:=AnsiCompareText(cond1, cond2) >= 0;
        2:
          Result:=AnsiCompareText(cond1, cond2) <= 0;
        3:
          Result:=AnsiCompareText(cond1, cond2) <> 0;
        4:
          Result:=AnsiCompareText(cond1, cond2) = 0;
        5:
          Result:=AnsiCompareText(cond1, cond2) > 0;
        6:
          Result:=AnsiCompareText(cond1, cond2) < 0;
      end;
    end;
  end
  else
  begin
    // sign not exist=boolean field or text result ("FALSE" or "TRUE")
    if CondStr = cCondResArr[false] then
      Result:=false
    else
      if CondStr = cCondResArr[true] then
        Result:=true
      else
        try
          if CondStr[2] = '.' then
          begin
            if CondStr[3] = '0' then
              Result:=DS[DatasetCharToIndex(CondStr[1])].IsEmpty
            else
              if CondStr[3] = '@' then
                Result:=DS_eof[DatasetCharToIndex(CondStr[1])]
              else
                Result:=DS[DatasetCharToIndex(CondStr[1])].FieldByName(copy(CondStr, 3, Length(CondStr))).AsBoolean;
          end
          else
            raise Exception.Create('');
        except
          RaiseErr(4, CondStr);
        end;
  end;
end;

// convert float string in 123E02 format to 123 x 10^2 format
function TRTFReport2.CheckScientificFormat(const s: string): string;
var
  i: integer;
  E, sgn: string;
begin
  Result:=s;
  i:=Pos('E', Result);
  if FConvertScientificFormat and (i > 0) then
  begin
    // extract exponent and sign
    E:=copy(Result, i + 1, Length(Result));
    sgn:='';
    if copy(E, 1, 1) = '-' then
      sgn:='-';
    if (copy(E, 1, 1) = '+') or (copy(E, 1, 1) = '-') then
      E:=copy(E, 2, Length(E));
    // extract leading zeros
    while (Length(E) > 1) and (copy(E, 1, 1) = '0') do
      E:=copy(E, 2, Length(E));
    // create format str
    Result:=copy(Result, 1, i - 1) + ' x 10' + cSuperScript + sgn + E + cNoSuperSubScript;
  end;
end;

// format or replace text (callback procedure)
procedure TRTFReport2.ProcFormatText(Sender: TObject; var s: string);
const
  Sep = '\ .,;()[]_?!"''<>+-*/=';
var
  Block1, Block2, p1, p2: integer;
  Str: string;
  skip: boolean;
begin
  // find block range
  Block1:=1;
  Block2:=Length(s);
  if SeekInfo.SeekFrom <> '' then
  begin
    p1:=rtool.AnsiPosEx(SeekInfo.SeekFrom, s, 1, false);
    if p1 > 0 then
      Block1:=p1;
  end;
  if SeekInfo.SeekTo <> '' then
  begin
    p2:=rtool.AnsiPosEx(SeekInfo.SeekTo, s, Block1, false);
    if p2 > 0 then
      Block2:=p2;
  end;

  // find text inside block
  p1:=Block1;
  repeat
    p1:=rtool.AnsiPosEx(SeekInfo.OldText, s, p1, false);
    if p1 = 0 then
      Break;
    p2:=p1 + Length(SeekInfo.OldText);
    if p2 > Block2 then
      Break;
    // check whole words condition
    skip:=false;
    if SeekInfo.WholeWord then
    begin
      if p1 > 1 then
      begin
        Str:=copy(s, p1 - 1, 1);
        skip:=Pos(Str, Sep) = 0;
      end;
      if not skip then
      begin
        Str:=copy(s, p2, 1);
        skip:=Pos(Str, Sep) = 0;
      end;
    end;

    if skip then
      p1:=p2
    else
    begin
      // do action
      case SeekInfo.Action of
        1:
          begin
            // format
            Str:=SeekInfo.SetFrmStart + copy(s, p1, p2 - p1) + SeekInfo.SetFrmStop;
          end;
        2:
          begin
            // replace
            Str:=SeekInfo.NewText;
          end;
      end;
      s:=copy(s, 1, p1 - 1) + Str + copy(s, p2, Length(s));
      p1:=p1 + Length(Str);
    end;
  until false;
end;

procedure TRTFReport2.FormatText(const TextToFormat: string; FormatTypes: TRTFTextFormats; WholeWordOnly: boolean = false; const SeekFrom: string = '';
  const SeekTo: string = '');
begin
  SeekInfo.Action:=1;
  SeekInfo.OldText:=TextToFormat;
  SeekInfo.NewText:='';
  SeekInfo.WholeWord:=WholeWordOnly;
  SeekInfo.SeekFrom:=SeekFrom;
  SeekInfo.SeekTo:=SeekTo;
  SeekInfo.SetFrmStart:='';
  SeekInfo.SetFrmStop:='';
  if frmBold in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + '\b ';
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '\b0 ';
  end;
  if frmItalic in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + '\i ';
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '\i0 ';
  end;
  if frmUnderline in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + '\ul ';
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '\ul0 ';
  end;
  ProcAllSkipPictureTag(RepStr, ProcFormatText);
end;

procedure TRTFReport2.ReplaceText(const OldText, NewText: string; WholeWordOnly: boolean = false; const SeekFrom: string = ''; const SeekTo: string = '');
var
  i: integer;
  newt: string;
begin
  SeekInfo.Action:=2;
  SeekInfo.OldText:=OldText;
  newt:=NewText;
  // format superscript sign,  \^ starts superscript, \= ends superscript
  i:=Pos('\^', newt);
  if i > 0 then
  begin
    newt:=copy(newt, 1, i - 1) + cSuperScript + copy(newt, i + 2, Length(newt));
    i:=Pos('\=', newt);
    if i > 0 then
      newt:=copy(newt, 1, i - 1) + cNoSuperSubScript + copy(newt, i + 2, Length(newt))
    else
      newt:=newt + cNoSuperSubScript;
  end;
  // format subscript sign,  \_ starts subscript, \= ends subscript
  i:=Pos('\_', newt);
  if i > 0 then
  begin
    newt:=copy(newt, 1, i - 1) + cSubScript + copy(newt, i + 2, Length(newt));
    i:=Pos('\=', newt);
    if i > 0 then
      newt:=copy(newt, 1, i - 1) + cNoSuperSubScript + copy(newt, i + 2, Length(newt))
    else
      newt:=newt + cNoSuperSubScript;
  end;
  SeekInfo.NewText:=newt;

  SeekInfo.WholeWord:=WholeWordOnly;
  SeekInfo.SeekFrom:=SeekFrom;
  SeekInfo.SeekTo:=SeekTo;
  SeekInfo.SetFrmStart:='';
  SeekInfo.SetFrmStop:='';
  ProcAllSkipPictureTag(RepStr, ProcFormatText);
end;

function TRTFReport2.GetFieldTagValue(FieldTag: string): string;
begin
  if IsFieldTag(FieldTag) then
    Result:=GetDBValue(DatasetCharToIndex(FieldTag[1]), copy(FieldTag, 3, Length(FieldTag)), '');
end;

end.
