unit SLKExport;

{<
  SLKExport unit contains TSLKExport - non visual component which provides functionality for
  exporting of database data to SLK file, which can be opened with MS Excel;
  it can easily export data from any dataset or DBGrid.

  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2005)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Dialogs, Db, Graphics, Math, ComCtrls,
  rImprovedComps, DBGrids, rStringGridEd, rDBRecView, rDBTreeView,
  rDBGrid, rDBGrid_MS, RosiCompGlobalCfg;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

  { list of cell format types }
  slkCellFormat = (cfBold, cfItalic, cfUnderline);
  { event is called before and after export, application can export
    own non database data }
  TOnExport = procedure(Sender: TObject; MaxX, MaxY: integer) of object;

  { non visual component which provides functionality for
    exporting of database data to SLK file }
  TSLKExport = class(TComponent)
  private
    FStartRow: integer;
    FStartCol: integer;
    FExportFile: string;
    FSetFontName: string;
    FDefaultDateFormat: string;
    FDefaultTimeFormat: string;
    FDefaultNumFormat: string;
    FSLKIdHeaderLine: string;
    FTrimData: boolean;
    FExportFieldName: boolean;
    FRealFieldNamesAsTitle: boolean;
    FExportOnlyVisibleFields: boolean;
    FExportOnlySelectedRows: boolean;
    FSetColWidth: boolean;
    FMakeSumRow: boolean;
    FExportResult: TStrings;
    FExportStructureOnly: boolean;
    FActionAfter: TActionAfter;
    FOnBeforeExport: TOnExport;
    FOnAfterExport: TOnExport;
    FUseDisableControl: boolean;
    FShowProgressBar: boolean;
    FOnProgressEvent: TProgressEvent;
    procedure SetStartRow(Value: integer);
    procedure SetStartCol(Value: integer);
    procedure ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
    procedure DoAbort(Sender: TObject);
  protected
    SourceType: TSourceType;
    SourceObject: TObject;
    Dataset: TDataSet;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    FN: TStrings;
    X, Y, MaxX: integer;
    RecIdx, RecCount: integer;
    Aborted: boolean;
    procedure Loaded; override;
    function GetField(idx: integer): TField;
    function GetDta(Fld: TField): string;
    function GetExpString(s: string; NumFormat, AddApost: boolean): string;
    procedure ExpLines(arr: array of string);
    procedure AddSeparator(var s: string);
    procedure AddSepData(var s: string; Data: string);
    function SlkFormatXY(sPrev, sPost: string; X, Y: integer): string;
    procedure ExpLabels;
    function ExpData: boolean;
    procedure ExpSum;
    procedure SetWidth;
    procedure SetColFormat;
    procedure SetGlobalFormat;
    procedure ExpTreeItems;
    procedure ExpGridRows;
    procedure ShowProgresInt;
    function ExecuteInt(ASourceType: TSourceType; O: TObject): boolean;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { execute creating export for DBGrid (columns collection is used) }
    function Execute(DBGrid: TCustomDBGrid): boolean; overload;
    { execute creating export for Dataset (fields collection is used) }
    function Execute(Dataset: TDataSet): boolean; overload;
    { execute creating export for one record from rDBRecView (rows/fields collection is used) }
    function Execute(rDBRecView: TrDBRecView): boolean; overload;
    { execute creating export for TreeView (Items collection is used) }
    function Execute(TreeView: TCustomTreeView): boolean; overload;
    { execute creating export for rStringGrid (columns and rows collection is used) }
    function Execute(rStringGridEd: TrStringGridEd): boolean; overload;
    { Exports text to defined cell, used only for user defined export }
    procedure ExportText(const Text: string; X, Y: integer);
    { Exports numeric to defined cell, used only for user defined export }
    procedure ExportNum(Num: extended; X, Y: integer);
    { Sets format of defined cell, used only for user defined export }
    procedure SetCellFormat(CellFormat: slkCellFormat; X, Y: integer);
    { Export one already formatted line in SLK format, used only for user defined export }
    procedure ExportLine(const Text: string);
    { lines with result of exporting }
    property ExportResult: TStrings read FExportResult;
    { define that only structure (name of fields) will be exported }
    property ExportStructureOnly: boolean read FExportStructureOnly write FExportStructureOnly;
  published
    { Defines content of first slk line ID;PWXL;N;E or ID;PCALCOOO32 or similar }
    property SLKIdHeaderLine: string read FSLKIdHeaderLine write FSLKIdHeaderLine;
    { Defines first row for export }
    property StartRow: integer read FStartRow write SetStartRow default 1;
    { Defines first collumn for export }
    property StartCol: integer read FStartCol write SetStartCol default 1;
    { Defines output file name }
    property ExportFile: string read FExportFile write FExportFile;
    { Defines FontName which will be set in output file }
    property SetFontName: string read FSetFontName write FSetFontName;
    { Defines default format for date fields }
    property DefaultDateFormat: string read FDefaultDateFormat write FDefaultDateFormat;
    { Defines default format for time fields }
    property DefaultTimeFormat: string read FDefaultTimeFormat write FDefaultTimeFormat;
    { Defines default format for numeric fields }
    property DefaultNumFormat: string read FDefaultNumFormat write FDefaultNumFormat;
    { Defines whether exported data is trimmed (spaces and EOL is removed from beginning and end) }
    property TrimData: boolean read FTrimData write FTrimData default false;
    { Defines whether firts exported line contains name of Fields }
    property ExportFieldName: boolean read FExportFieldName write FExportFieldName default true;
    { Defines whether firts exported line contains real names of Fields instead of Captions }
    property RealFieldNamesAsTitle: boolean read FRealFieldNamesAsTitle write FRealFieldNamesAsTitle default false;
    { Defines whether visible columns/fields or all columns/fields is exported }
    property ExportOnlyVisibleFields: boolean read FExportOnlyVisibleFields write FExportOnlyVisibleFields default true;
    { Defines whether selected rows from DBGrid is exported }
    property ExportOnlySelectedRows: boolean read FExportOnlySelectedRows write FExportOnlySelectedRows default false;
    { Defines whether output file contains definition of collumn width }
    property SetColWidth: boolean read FSetColWidth write FSetColWidth default true;
    { Defines whether last exported row contains sum }
    property MakeSumRow: boolean read FMakeSumRow write FMakeSumRow default false;

    { Defines action after exporting (None, Show, Print) }
    property ActionAfter: TActionAfter read FActionAfter write FActionAfter default aaShow;
    { Defines event for user defined action }
    property OnBeforeExport: TOnExport read FOnBeforeExport write FOnBeforeExport;
    { Defines event for user defined action }
    property OnAfterExport: TOnExport read FOnAfterExport write FOnAfterExport;

    { indicate whether DisableControls is used during dataset scan }
    property UseDisableControl: boolean read FUseDisableControl write FUseDisableControl default true;
    { indicate whether progress bar is shown }
    property ShowProgressBar: boolean read FShowProgressBar write FShowProgressBar default false;
    { Defines event used periodically during exporting of data }
    property OnProgressEvent: TProgressEvent read FOnProgressEvent write FOnProgressEvent;
  end;

procedure Register; {< @exclude }

implementation

uses rdlg, rtool, rdbtool, rstring;
{$R *.res}

const
  TextApostr = '"';
  cSep = ';';
  cCustomExport = '; CustomExport';
  cRow = '; Row';
  cCell = 'C';
  cValue = 'K';
  cFormat = 'F';

  cSlkEOL = #27#32#58;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TSLKExport]);
end;

constructor TSLKExport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FStartCol:=1;
  FStartRow:=1;
  FSetFontName:='Arial';
  FExportFieldName:=true;
  FRealFieldNamesAsTitle:=false;
  FExportOnlyVisibleFields:=true;
  FExportOnlySelectedRows:=false;
  FSetColWidth:=true;
  FActionAfter:=aaShow;
  FShowProgressBar:=false;
  FUseDisableControl:=true;
  FExportResult:=TStringList.Create;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TSLKExport.Destroy;
begin
  FreeAndNil(FExportResult);
  inherited Destroy;
end;

procedure TSLKExport.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TSLKExport.Execute(DBGrid: TCustomDBGrid): boolean;
begin
  Result:=ExecuteInt(stGrid, DBGrid);
end;

function TSLKExport.Execute(Dataset: TDataSet): boolean;
begin
  Result:=ExecuteInt(stDataset, Dataset);
end;

function TSLKExport.Execute(rDBRecView: TrDBRecView): boolean;
begin
  Result:=ExecuteInt(stRecView, rDBRecView);
end;

function TSLKExport.Execute(TreeView: TCustomTreeView): boolean;
begin
  Result:=ExecuteInt(stTreeView, TreeView);
end;

function TSLKExport.Execute(rStringGridEd: TrStringGridEd): boolean;
begin
  Result:=ExecuteInt(stStringGrid, rStringGridEd);
end;

function TSLKExport.ExecuteInt(ASourceType: TSourceType; O: TObject): boolean;
var
  d: TDataSource;
  FldDef: TGetFldDefs;
begin
  Result:=false;
  if ExportFile = '' then
    Exit;
  if not Assigned(O) then
    Exit;
  SourceType:=ASourceType;
  SourceObject:=O;
  Dataset:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      begin
        DBGrid:=TCustomDBGrid(O);
        d:=DBGrid.DataSource;
      end;
    stRecView:
      begin
        RecView:=TrDBRecView(O);
        d:=RecView.DataSource;
      end;
    stTreeView:
      TreeView:=TCustomTreeView(O);
    stDataset:
      Dataset:=TDataSet(O);
    stStringGrid:
      StringGrid:=TrStringGridEd(O);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.Dataset) or not d.Dataset.Active then
      Exit;
    Dataset:=d.Dataset;
  end;
  if Assigned(Dataset) and not Dataset.Active then
    Exit;

  if FShowProgressBar then
    ShowInfo(sExporting + sOfData);
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psStarting, 0, false, Rect(0, 0, 0, 0), sExporting + sOfData);

  FExportResult.Clear;
  FN:=TStringList.Create;
  try
    FldDef:=GetFldDefs(false, FExportOnlyVisibleFields, true, true, false, false, false);
    GetFldList(SourceObject, FN, FldDef);
    try
      if FSLKIdHeaderLine = '' then
        ExportLine('ID;SLKExport')
      else
        ExportLine(FSLKIdHeaderLine);
      Y:=FStartRow;
      X:=FStartCol;
      SetColFormat;
      if Assigned(FOnBeforeExport) then
        FOnBeforeExport(Self, X, Y);
      if Assigned(Dataset) or (SourceType in [stTreeView, stStringGrid]) then
      begin
        if FSetColWidth then
          SetWidth;
        if FExportFieldName or FExportStructureOnly then
          ExpLabels;
        if not FExportStructureOnly then
        begin
          if not ExpData then
            Exit;
          if FMakeSumRow then
            ExpSum;
        end;
      end;
      if FSetFontName <> 'nil' then
        SetGlobalFormat;
      if Assigned(FOnAfterExport) then
        FOnAfterExport(Self, MaxX, Y);
      ExportLine('E');
      FExportResult.SaveToFile(FExportFile);
{$IFDEF DEMO}
      ShowDemoMsg;
{$ENDIF}
      Result:=true;
      if ActionAfter = aaShow then
        RunFileSimple(ExportFile);
      if ActionAfter = aaPrint then
        PrintFileSimple(ExportFile);
    except
      on E: Exception do
        ShowErr(sExportDataErr, E);
    end;
  finally
    FreeAndNil(FN);
    if FShowProgressBar then
      ShowInfo('');
    if Assigned(OnProgressEvent) then
      OnProgressEvent(Self, psEnding, 100, false, Rect(0, 0, 0, 0), '');
  end;
end;

procedure TSLKExport.ExportLine(const Text: string);
begin
  if Text <> '' then
    FExportResult.Add(Text);
end;

procedure TSLKExport.ExpLines(arr: array of string);
var
  i: integer;
begin
  for i:= Low(arr) to High(arr) do
    ExportLine(arr[i]);
end;

procedure TSLKExport.AddSeparator(var s: string);
begin
  if s <> '' then
    s:=s + cSep;
end;

procedure TSLKExport.AddSepData(var s: string; Data: string);
begin
  if Data = '' then
    Exit;
  AddSeparator(s);
  s:=s + Data;
end;

function TSLKExport.SlkFormatXY(sPrev, sPost: string; X, Y: integer): string;
begin
  Result:=sPrev;
  if X >= 0 then
    AddSepData(Result, Format('X%d', [X]));
  if Y >= 0 then
    AddSepData(Result, Format('Y%d', [Y]));
  AddSepData(Result, sPost);
end;

procedure TSLKExport.ExportText(const Text: string; X, Y: integer);
begin
  ExpLines([cSep, cCustomExport, SlkFormatXY(cCell, cValue + GetExpString(Text, false, true), X, Y)]);
end;

procedure TSLKExport.ExportNum(Num: extended; X, Y: integer);
begin
  ExpLines([cSep, cCustomExport, SlkFormatXY(cCell, cValue + GetExpString(FloatToStr(Num), true, false), X, Y)]);
end;

procedure TSLKExport.SetCellFormat(CellFormat: slkCellFormat; X, Y: integer);
begin
  ExpLines([cSep, cCustomExport, SlkFormatXY(cFormat, Format('SM%d', [ord(CellFormat) + 1]), X, Y)]);
end;

procedure TSLKExport.SetStartRow(Value: integer);
begin
  if Value > 0 then
    FStartRow:=Value;
end;

procedure TSLKExport.SetStartCol(Value: integer);
begin
  if Value > 0 then
    FStartCol:=Value;
end;

function TSLKExport.GetField(idx: integer): TField;
begin
  Result:=Dataset.FindField(FN[idx]);
end;

function TSLKExport.GetDta(Fld: TField): string;
var
  NumFormat: boolean;
  s: string;
begin
  NumFormat:=Fld.DataType in [ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime, ftFMTBcd];
  if (Fld is TSQLTimeStampField) or (Fld is TDateTimeField) then
  begin
    if Fld.AsDateTime < Encodedate(1900, 1, 1) then
      NumFormat:=false;
  end;
  if NumFormat then
    s:=FloatToStr(Fld.AsFloat)
  else
    if DBGrid is TrDBGrid then
      s:=TrDBGrid(DBGrid).GetCellValue(Fld, false)
    else
      s:=Fld.AsString;
  Result:=GetExpString(s, NumFormat, Fld.DataType in (sfApostrFld + [ftBoolean]));
end;

function TSLKExport.GetExpString(s: string; NumFormat, AddApost: boolean): string;
begin
  if FTrimData then
    s:=Trim(s);
  if NumFormat then
    s:=ReplaceStr(GetDecimalSeparator, '.', s, true, false)
  else
    if AddApost then
    begin
      s:=rtool.ReplaceStr(cStrEOL, cSlkEOL, s, true, true);
      s:=rtool.ReplaceStr(#13, cSlkEOL, s, true, true);
      s:=rtool.ReplaceStr(#10, cSlkEOL, s, true, true);
      // s:=rtool.ReplaceStr(#9, '   ', s, true, true);
      if Length(s) > 250 then
      begin
        s:=copy(s, 1, 247);
        if copy(s, Length(s) - 1, 2) = #27#32 then
          s:=copy(s, 1, Length(s) - 2);
        if copy(s, Length(s), 1) = #27 then
          s:=copy(s, 1, Length(s) - 1);
        s:=s + '...';
      end;
      s:=TextApostr + s + TextApostr;
    end;
  Result:=ReplaceStr(';', ';;', s, true, false)
end;

procedure TSLKExport.ExpSum;
var
  c, p, aMin, aMax, Col: integer;
  IsApostr, Tot: boolean;
begin
  Tot:=false;
  aMin:=StartRow;
  if ExportFieldName then
    Inc(aMin);
  aMax:=Y - 1;
  if aMin > aMax then
    Exit;
  p:=X;
  ExpLines([cSep, '; Sum', SlkFormatXY(cCell, '', X, Y)]);
  for c:=0 to FN.Count - 1 do
  begin
    IsApostr:=true;
    case SourceType of
      stGrid, stDataset, stRecView:
        IsApostr:=IsApostrField(GetField(c));
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          if Col >= 0 then
            IsApostr:=StringGrid.Columns[Col].DataType = sgdText;
        end;
    end;
    if IsApostr then
    begin
      if not Tot then
        ExpLines([SlkFormatXY(cCell, cValue + GetExpString('Sum:', false, true), p, -1)]);
      Tot:=true;
    end
    else
      ExpLines([SlkFormatXY(cCell, Format('ESUM(R%dC:R%dC)', [aMin, aMax]), p, -1)]);
    Inc(p);
  end;
end;

procedure TSLKExport.SetWidth;
var
  c, p, dw, Col: integer;
begin
  p:=X;
  ExpLines([cSep, '; Column Width']);
  for c:=0 to FN.Count - 1 do
  begin
    dw:=50;
    case SourceType of
      stGrid, stDataset, stRecView:
        dw:=GetField(c).DisplayWidth;
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          dw:=StringGrid.GetColumnDataWidth(Col);
        end;
    end;
    dw:=max(Min(dw, 50), 5) + 2;
    ExpLines([Format('F;W%d %d %d', [p, p, dw])]);
    Inc(p);
  end;
end;

procedure TSLKExport.SetColFormat;
var
  Frm: string;
  Fld: TField;
  c, Col: integer;
begin
  // default format
  ExportLine('P;PGeneral');
  Frm:=FDefaultDateFormat;
  if Frm = '' then
    Frm:='d/m/yyyy';
  ExportLine('P;P' + Frm);
  Frm:=FDefaultTimeFormat;
  if Frm = '' then
    Frm:='h:mm';
  ExportLine('P;P' + Frm);
  // num fields format
  for c:=0 to FN.Count - 1 do
  begin
    Frm:='';
    case SourceType of
      stGrid, stDataset, stRecView:
        begin
          Fld:=GetField(c);
          if Fld is TNumericField then
          begin
            Frm:=TNumericField(Fld).DisplayFormat;
            if Frm = '' then
              Frm:=FDefaultNumFormat;
          end;
        end;
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          if Col >= 0 then
          begin
            if StringGrid.Columns[Col].DataType <> sgdText then
              Frm:=StringGrid.Columns[Col].NumFormat;
          end;
          if Frm = '' then
            Frm:=FDefaultNumFormat;
        end;
    end;
    if Frm = '' then
      Frm:='General';
    ExportLine('P;P' + Frm);
  end;
  // font format
  if FSetFontName = 'nil' then
    Exit;
  Frm:='P;F' + FSetFontName;
  ExpLines([Frm, Frm + ';SB', Frm + ';SI', Frm + ';SU']);
end;

procedure TSLKExport.SetGlobalFormat;
begin
  ExpLines([cSep, '; Global Formating']);
  if ExportFieldName then
    ExportLine(Format('F;R%d;SM1', [FStartRow]));
  if FMakeSumRow then
    ExportLine(Format('F;R%d;SM1', [Y]));
end;

procedure TSLKExport.ExpLabels;
var
  c, p: integer;
  L: TStrings;
  FldDef: TGetFldDefs;
begin
  L:=TStringList.Create;
  try
    FldDef:=GetFldDefs(not FRealFieldNamesAsTitle, FExportOnlyVisibleFields, true, true, false, false, false);
    if SourceObject is TrDBGrid then
      // use GetCellValue to get groupped column value instead of field value
      GetFldListrDBGrid(TrDBGrid(SourceObject), L, FldDef)
    else
      GetFldList(SourceObject, L, FldDef);
    ExpLines([cSep, '; Title', SlkFormatXY(cCell, '', X, Y)]);
    p:=X;
    for c:=0 to FN.Count - 1 do
    begin
      ExpLines([SlkFormatXY(cCell, cValue + GetExpString(L[c], false, true), p, -1)]);
      Inc(p);
    end;
    Inc(Y);
  finally
    FreeAndNil(L);
  end;
end;

function TSLKExport.ExpData: boolean;
begin
  RecCount:=0;
  Aborted:=false;
  Result:=true;
  case SourceType of
    stGrid, stDataset:
      begin
        RecCount:=Dataset.RecordCount;
        Result:=rdbtool.ScanAllRecords(Dataset, ScanCallBack, FUseDisableControl, FUseDisableControl);
      end;
    stRecView:
      begin
        RecCount:=1;
        ScanCallBack(Dataset, 0, Aborted);
      end;
    stTreeView:
      ExpTreeItems;
    stStringGrid:
      ExpGridRows;
  end;
end;

procedure TSLKExport.ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
var
  c, aX: integer;
  Fld: TField;
  Frm: string;
begin
  // use selected row only?
  if (FExportOnlySelectedRows) and (SourceType = stGrid) then
  begin
    if DBGrid is TrDBGrid_MS then
    begin
      if not TrDBGrid_MS(DBGrid).IsActiveSelected then
        Exit;
    end
    else
      if not THackCustomDBGrid(DBGrid).SelectedRows.CurrentRowSelected then
        Exit;
  end;

  RecIdx:=RecIndex;
  ExpLines([cSep, cRow, SlkFormatXY(cCell, '', X, Y)]);
  aX:=X;
  for c:=0 to FN.Count - 1 do
  begin
    Fld:=GetField(c);
    if not Fld.IsNull then
    begin
      // set format
      if IsDateTimeField(Fld) then
      begin
        if Trunc(Fld.AsDateTime) = 0 then
          Frm:=Format('F;P2;X%d', [aX])
        else
          Frm:=Format('F;P1;X%d', [aX]);
      end
      else
        if Fld is TNumericField then
          Frm:=Format('F;P%d;X%d', [c + 3, aX])
        else
          Frm:=Format('C;X%d', [aX]);
      // write value
      ExpLines([Frm, cCell + cSep + cValue + GetDta(Fld)]);
    end;
    Inc(aX);
  end;
  Inc(Y);

  ShowProgresInt;
  MaxX:=aX - 1;
  if Aborted then
    Abort:=true;
end;

procedure TSLKExport.ExpTreeItems;
  procedure ProcTV(tn: TTreeNode; level: integer);
  var
    aX: integer;
    s: string;
  begin
    while Assigned(tn) do
    begin
      ExpLines([cSep, cRow, SlkFormatXY(cCell, '', X, Y)]);
      aX:=X;
      Inc(aX, level);
      // write value
      s:=GetExpString(tn.Text, false, true);
      ExpLines([Format('C;X%d', [aX]), cCell + cSep + cValue + s]);
      Inc(Y);
      Inc(RecIdx);
      ShowProgresInt;
      if tn.Expanded then
        ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  RecCount:=THackCustomTreeView(TreeView).Items.Count;
  RecIdx:=0;
  ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
end;

procedure TSLKExport.ExpGridRows;
var
  r, c, aX, Col: integer;
  Frm, v: string;
  Column: TrStringGridEdColumn;
begin
  RecCount:=StringGrid.RowCount - StringGrid.FixedRows;
  for r:=StringGrid.FixedRows to StringGrid.RowCount - 1 do
  begin
    RecIdx:=r;
    ExpLines([cSep, cRow, SlkFormatXY(cCell, '', X, Y)]);
    aX:=X;
    for c:=0 to FN.Count - 1 do
    begin
      Col:=StrToIntDef(FN[c], -1);
      if (Col >= 0) and (StringGrid.Cells[Col, r] <> '') then
      begin
        Column:=StringGrid.Columns[Col];
        case Column.DataType of
          sgdNumber, sgdInteger:
            begin
              Frm:=Format('F;P%d;X%d', [c + 3, aX]);
              v:=FloatToStr(StringGrid.NumCells[Col, r]);
            end;
          sgdDate:
            begin
              Frm:=Format('F;P1;X%d', [aX]);
              v:=FloatToStr(StringGrid.DateTimeCells[Col, r]);
            end;
          sgdTime:
            begin
              Frm:=Format('F;P2;X%d', [aX]);
              v:=FloatToStr(StringGrid.DateTimeCells[Col, r]);
            end
          else
          begin
            Frm:=Format('C;X%d', [aX]);
            v:=StringGrid.Cells[Col, r];
          end;
        end;
        // write value
        v:=GetExpString(v, Column.DataType in [sgdNumber, sgdInteger, sgdDate, sgdTime], Column.DataType in [sgdText]);
        ExpLines([Frm, cCell + cSep + cValue + v]);
      end;
      Inc(aX);
    end;
    Inc(Y);
    ShowProgresInt;
    if Aborted then
      Abort;
  end;
end;

procedure TSLKExport.ShowProgresInt;
begin
  if RecCount = 0 then
    Exit;
  if FShowProgressBar and (RecIdx mod 100 = 0) then
    ShowInfoProgressBar(sExporting + sOfData, nil, DoAbort, Round(100.0 * RecIdx / RecCount));
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psRunning, Round(100.0 * RecIdx / RecCount), false, Rect(0, 0, 0, 0), sExporting + sOfData);
end;

procedure TSLKExport.DoAbort(Sender: TObject);
begin
  Aborted:=true;
end;

end.
