unit TXTExport;

{<
  TXTExport unit contains TTXTExport - non visual component which provides functionality for
  exporting of database data to StringList or TXT file;
  it can easily export data from any dataset or DBGrid.

  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2005)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Db, Graphics, Dialogs, ComCtrls,
  rImprovedComps, DBGrids, rStringGridEd, rDBRecView, rDBTreeView,
  rDBGrid, rDBGrid_MS, RosiCompGlobalCfg;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

  { non visual component which provides functionality for
    exporting of database data to StringList or TXT file }
  TTXTExport = class(TComponent)
  private
    FExportFile: string;
    FTextApostrophe: string;
    FFieldSeparator: string;
    FExportResult: TStrings;
    FFillChar: Char;
    FTrimData: boolean;
    FExportFieldName: boolean;
    FRealFieldNamesAsTitle: boolean;
    FExportOnlyVisibleFields: boolean;
    FExportOnlySelectedRows: boolean;
    FAppendToFile: boolean;
    FExportStructureOnly: boolean;
    FExportAllAsText: boolean;
    FActionAfter: TActionAfter;
    FOnBeforeExport: TNotifyEvent;
    FOnAfterExport: TNotifyEvent;
    FUseDisableControl: boolean;
    FShowProgressBar: boolean;
    FOnProgressEvent: TProgressEvent;
    procedure ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
    procedure DoAbort(Sender: TObject);
  protected
    // GDS: TObject;
    SourceType: TSourceType;
    SourceObject: TObject;
    Dataset: TDataSet;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    FN: TStrings;
    RecIdx, RecCount: integer;
    Aborted: boolean;
    procedure Loaded; override;
    function GetField(idx: integer): TField;
    function AlignColData(ColIdx, RowIdx: integer; const Data: string): string;
    procedure AddApostrophe(var s: string);
    procedure AddSeparator(var s: string);
    procedure AddSepData(var s: string; Data: string);
    function PrepareData(s: string): string;
    procedure ExpLabels;
    function ExpData: boolean;
    procedure ExpTreeItems;
    procedure ExpGridRows;
    procedure ShowProgresInt;
    function ExecuteInt(ASourceType: TSourceType; O: TObject): boolean;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { execute creating export for DBGrid (columns collection is used) }
    function Execute(DBGrid: TCustomDBGrid): boolean; overload;
    { execute creating export for Dataset (fields collection is used) }
    function Execute(Dataset: TDataSet): boolean; overload;
    { execute creating export for one record from rDBRecView (rows/fields collection is used) }
    function Execute(rDBRecView: TrDBRecView): boolean; overload;
    { execute creating export for TreeView (Items collection is used) }
    function Execute(TreeView: TCustomTreeView): boolean; overload;
    { execute creating export for rStringGrid (columns and rows collection is used) }
    function Execute(rStringGridEd: TrStringGridEd): boolean; overload;
    { Export one line of text, used only for user defined export }
    procedure ExportLine(const Text: string);
    { lines with result of exporting }
    property ExportResult: TStrings read FExportResult;
    { define that only structure (name of fields) will be exported }
    property ExportStructureOnly: boolean read FExportStructureOnly write FExportStructureOnly;
    { define that all fields are exported as text field (with apostrophe) }
    property ExportAllAsText: boolean read FExportAllAsText write FExportAllAsText;
  published
    { Defines output file name, if empty file is not saved and use ExportResult }
    property ExportFile: string read FExportFile write FExportFile;
    { Defines string which is used as apostrophe for string field }
    property TextApostrophe: string read FTextApostrophe write FTextApostrophe;
    { Defines string which is used as separator }
    property FieldSeparator: string read FFieldSeparator write FFieldSeparator;
    { Defines char which is used for extension of string to needed length }
    property FillChar: Char read FFillChar write FFillChar default ' ';
    { Defines whether exported data is trimmed (spaces and EOL is removed from beginning and end) }
    property TrimData: boolean read FTrimData write FTrimData default false;
    { Defines whether firts exported line contains name of Fields }
    property ExportFieldName: boolean read FExportFieldName write FExportFieldName default true;
    { Defines whether firts exported line contains real names of Fields instead of Captions }
    property RealFieldNamesAsTitle: boolean read FRealFieldNamesAsTitle write FRealFieldNamesAsTitle default false;
    { Defines whether visible columns/fields or all columns/fields is exported }
    property ExportOnlyVisibleFields: boolean read FExportOnlyVisibleFields write FExportOnlyVisibleFields default true;
    { Defines whether selected rows from DBGrid is exported }
    property ExportOnlySelectedRows: boolean read FExportOnlySelectedRows write FExportOnlySelectedRows default false;
    { define whether output is appended to existing file (used only for TXT format) }
    property AppendToFile: boolean read FAppendToFile write FAppendToFile default false;
    { Defines action after exporting (None, Show, Print) }
    property ActionAfter: TActionAfter read FActionAfter write FActionAfter default aaShow;
    { Defines event for user defined action }
    property OnBeforeExport: TNotifyEvent read FOnBeforeExport write FOnBeforeExport;
    { Defines event for user defined action }
    property OnAfterExport: TNotifyEvent read FOnAfterExport write FOnAfterExport;
    { indicate whether DisableControls is used during dataset scan }
    property UseDisableControl: boolean read FUseDisableControl write FUseDisableControl default true;
    { indicate whether progress bar is shown }
    property ShowProgressBar: boolean read FShowProgressBar write FShowProgressBar default false;
    { Defines event used periodically during exporting of data }
    property OnProgressEvent: TProgressEvent read FOnProgressEvent write FOnProgressEvent;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rdlg, rtool, rdbtool, rstring;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TTXTExport]);
end;

constructor TTXTExport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FActionAfter:=aaShow;
  FFieldSeparator:=' ';
  FFillChar:=' ';
  FExportFieldName:=true;
  FRealFieldNamesAsTitle:=false;
  FExportOnlyVisibleFields:=true;
  FExportOnlySelectedRows:=false;
  FExportStructureOnly:=false;
  FExportAllAsText:=false;
  FShowProgressBar:=false;
  FUseDisableControl:=true;
  FExportResult:=TStringList.Create;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TTXTExport.Destroy;
begin
  FreeAndNil(FExportResult);
  inherited Destroy;
end;

procedure TTXTExport.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TTXTExport.Execute(DBGrid: TCustomDBGrid): boolean;
begin
  Result:=ExecuteInt(stGrid, DBGrid);
end;

function TTXTExport.Execute(Dataset: TDataSet): boolean;
begin
  Result:=ExecuteInt(stDataset, Dataset);
end;

function TTXTExport.Execute(rDBRecView: TrDBRecView): boolean;
begin
  Result:=ExecuteInt(stRecView, rDBRecView);
end;

function TTXTExport.Execute(TreeView: TCustomTreeView): boolean;
begin
  Result:=ExecuteInt(stTreeView, TreeView);
end;

function TTXTExport.Execute(rStringGridEd: TrStringGridEd): boolean;
begin
  Result:=ExecuteInt(stStringGrid, rStringGridEd);
end;

function TTXTExport.ExecuteInt(ASourceType: TSourceType; O: TObject): boolean;
var
  d: TDataSource;
  FldDef: TGetFldDefs;
begin
  Result:=false;
  if not Assigned(O) then
    Exit;
  SourceType:=ASourceType;
  SourceObject:=O;
  Dataset:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      begin
        DBGrid:=TCustomDBGrid(O);
        d:=DBGrid.DataSource;
      end;
    stRecView:
      begin
        RecView:=TrDBRecView(O);
        d:=RecView.DataSource;
      end;
    stTreeView:
      TreeView:=TCustomTreeView(O);
    stDataset:
      Dataset:=TDataSet(O);
    stStringGrid:
      StringGrid:=TrStringGridEd(O);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.Dataset) or not d.Dataset.Active then
      Exit;
    Dataset:=d.Dataset;
  end;
  if Assigned(Dataset) and not Dataset.Active then
    Exit;

  if FShowProgressBar then
    ShowInfo(sExporting + sOfData);
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psStarting, 0, false, Rect(0, 0, 0, 0), sExporting + sOfData);

  try
    FExportResult.Clear;
    if FAppendToFile and FileExists(ExportFile) then
      FExportResult.LoadFromFile(ExportFile);
    FN:=TStringList.Create;
    try
      FldDef:=GetFldDefs(false, FExportOnlyVisibleFields, true, false, false, false, false);
      GetFldList(SourceObject, FN, FldDef);
      if Assigned(FOnBeforeExport) then
        FOnBeforeExport(Self);
      if FExportFieldName or FExportStructureOnly then
        ExpLabels;
      if not FExportStructureOnly then
      begin
        if not ExpData then
          Exit;
      end;
      if Assigned(FOnAfterExport) then
        FOnAfterExport(Self);
      Result:=true;
    finally
      FreeAndNil(FN);
    end;
{$IFDEF DEMO}
    ShowDemoMsg;
{$ENDIF}
    if ExportFile = '' then
      Exit;
    // export to file
    try
      FExportResult.SaveToFile(ExportFile);
      if ActionAfter = aaShow then
        RunFileSimple(ExportFile);
      if ActionAfter = aaPrint then
        PrintFileSimple(ExportFile);
    except
      on E: Exception do
      begin
        Result:=false;
        ShowErr(sExportDataErr, E);
      end;
    end;
  finally
    if FShowProgressBar then
      ShowInfo('');
    if Assigned(OnProgressEvent) then
      OnProgressEvent(Self, psEnding, 100, false, Rect(0, 0, 0, 0), '');
  end;
end;

procedure TTXTExport.ExportLine(const Text: string);
begin
  FExportResult.Add(Text);
end;

function TTXTExport.GetField(idx: integer): TField;
begin
  Result:=Dataset.FindField(FN[idx]);
end;

function TTXTExport.AlignColData(ColIdx, RowIdx: integer; const Data: string): string;
var
  Fld: TField;
  // GridCol: TrStringGridEdColumn;
  col, dw: integer;
  Algn: TAlignment;
begin
  Result:=Data;
  if (FFillChar <> #0) then
  begin
    dw:=200;
    Algn:=taLeftJustify;
    case SourceType of
      stGrid, stDataset, stRecView:
        begin
          Fld:=GetField(ColIdx);
          dw:=Fld.DisplayWidth;
          Algn:=Fld.Alignment;
        end;
      stTreeView:
        ;
      stStringGrid:
        begin
          col:=StrToIntDef(FN[ColIdx], -1);
          if col < 0 then
            Exit;
          dw:=StringGrid.GetColumnDataWidth(col);
          Algn:=StringGrid.GetDataAlignment(col, RowIdx);
        end;
    end;
    if Length(Result) > dw then
      Result:=copy(Result, 1, dw - 2) + '..';
    case Algn of
      taRightJustify:
        Result:=FillStrL(Result, FillChar, dw);
      taCenter:
        begin
          Result:=FillStrL(Result, FillChar, Length(Result) + (dw - Length(Result)) div 2);
          Result:=FillStrR(Result, FillChar, dw);
        end;
    else
      Result:=FillStrR(Result, FillChar, dw);
    end;
  end;
end;

procedure TTXTExport.AddApostrophe(var s: string);
begin
  s:=FTextApostrophe + s + FTextApostrophe;
end;

procedure TTXTExport.AddSeparator(var s: string);
begin
  if s <> '' then
    s:=s + FFieldSeparator;
end;

procedure TTXTExport.AddSepData(var s: string; Data: string);
begin
  AddSeparator(s);
  s:=s + Data;
end;

function TTXTExport.PrepareData(s: string): string;
begin
  Result:=s;
  if FTrimData then
  begin
    Result:=Trim(Result);
  end;
end;

procedure TTXTExport.ExpLabels;
var
  c: integer;
  L: TStrings;
  line: string;
  FldDef: TGetFldDefs;
begin
  L:=TStringList.Create;
  try
    FldDef:=GetFldDefs(not FRealFieldNamesAsTitle, FExportOnlyVisibleFields, true, false, false, false, false);
    if SourceObject is TrDBGrid then
      GetFldListrDBGrid(TrDBGrid(SourceObject), L, FldDef)
    else
      GetFldList(SourceObject, L, FldDef);
    line:='';
    for c:=0 to FN.Count - 1 do
      AddSepData(line, AlignColData(c, 0, PrepareData(L[c])));
    ExportLine(line);
  finally
    FreeAndNil(L);
  end;
end;

function TTXTExport.ExpData: boolean;
begin
  RecCount:=0;
  Aborted:=false;
  Result:=true;
  case SourceType of
    stGrid, stDataset:
      begin
        RecCount:=Dataset.RecordCount;
        Result:=rdbtool.ScanAllRecords(Dataset, ScanCallBack, FUseDisableControl, FUseDisableControl);
      end;
    stRecView:
      begin
        RecCount:=0;
        ScanCallBack(Dataset, 0, Aborted);
      end;
    stTreeView:
      ExpTreeItems;
    stStringGrid:
      ExpGridRows;
  end;
end;

procedure TTXTExport.ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
var
  c: integer;
  s: string;
  Fld: TField;
  line: string;
begin
  // use selected row only?
  if FExportOnlySelectedRows and (SourceType = stGrid) then
  begin
    if DBGrid is TrDBGrid_MS then
    begin
      if not TrDBGrid_MS(DBGrid).IsActiveSelected then
        Exit;
    end
    else
      if not THackCustomDBGrid(DBGrid).SelectedRows.CurrentRowSelected then
        Exit;
  end;

  RecIdx:=RecIndex;
  line:='';
  for c:=0 to FN.Count - 1 do
  begin
    Fld:=GetField(c);
    if Assigned(Fld) then
    begin
      if DBGrid is TrDBGrid then
        // use GetCellValue to get groupped column value instead of field value
        s:=TrDBGrid(DBGrid).GetCellValue(Fld, true)
      else
        s:=Fld.DisplayText;
      s:=PrepareData(s);
      if FExportAllAsText or IsApostrField(Fld) then
        AddApostrophe(s);
      AddSepData(line, AlignColData(c, 0, s));
    end;
  end;
  ExportLine(line);
  ShowProgresInt;
  if Aborted then
    Abort:=true;
end;

procedure TTXTExport.ExpTreeItems;
  procedure ProcTV(tn: TTreeNode; level: integer);
  var
    i: integer;
    s, t: string;
  begin
    while Assigned(tn) do
    begin
      s:='';
      for i:=0 to level - 1 do
        s:=s + '  ';
      t:=PrepareData(tn.Text);
      ExportLine(s + t);
      Inc(RecIdx);
      ShowProgresInt;
      if tn.Expanded then
        ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  RecCount:=THackCustomTreeView(TreeView).Items.Count;
  RecIdx:=0;
  ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
end;

procedure TTXTExport.ExpGridRows;
var
  r, c, col: integer;
  s: string;
  line: string;
begin
  RecCount:=StringGrid.RowCount - StringGrid.FixedRows;
  for r:=StringGrid.FixedRows to StringGrid.RowCount - 1 do
  begin
    RecIdx:=r;
    line:='';
    for c:=0 to FN.Count - 1 do
    begin
      col:=StrToIntDef(FN[c], -1);
      if col >= 0 then
      begin
        s:=PrepareData(StringGrid.Cells[col, r]);
        AddSepData(line, AlignColData(c, r, s));
      end;
    end;
    ExportLine(line);
    ShowProgresInt;
    if Aborted then
      Abort;
  end;
end;

procedure TTXTExport.ShowProgresInt;
begin
  if RecCount = 0 then
    Exit;
  if FShowProgressBar and (RecIdx mod 100 = 0) then
    ShowInfoProgressBar(sExporting + sOfData, nil, DoAbort, Round(100.0 * RecIdx / RecCount));
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psRunning, Round(100.0 * RecIdx / RecCount), false, Rect(0, 0, 0, 0), sExporting + sOfData);
end;

procedure TTXTExport.DoAbort(Sender: TObject);
begin
  Aborted:=true;
end;

end.
