unit rDBColorLinkedComp;

{<
  rDBColorLinkedComp unit contains non visual component TrDBColorLinkedComp
  which set font and background color of db avare components
  according rDBGrid colors for selected record and field.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2013)
  @lastmod(Nov 2013)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls,
  Grids, DBGrids, rDBGrid, DB, DBCtrls, Forms, Dialogs, StdCtrls, ComCtrls,
  ExtCtrls, Types, {$IFDEF DXE2P} UITypes, {$ENDIF} Math, Variants,
  rDBComponents, PropSaveMain, PropSaveGrids, RosiCompGlobalCfg;

type
  { event type used by TrDBColorLinkedComp }
  TIsSupportedComponent = procedure(AComp: TComponent; var Result: boolean) of object;
  { event type used by TrDBColorLinkedComp }
  TComponentPropsEvent = procedure(AComp: TComponent; var Field: TField; var BkColor: TColor; var Font: TFont) of object;

  { non visual component which set font and background color of db avare components
    according rDBGrid colors for selected record and field }
  TrDBColorLinkedComp = class(TComponent)
  private
    FLinkedComps: TStrings;
    FLinkAllComps: boolean;
    FLoadAllCompsNow: boolean;
    FLinkDelayed: boolean;
    FLinkDelay: cardinal;
    FKeepFontSize: boolean;
    FKeepBackground: boolean;
    FEnabled: boolean;
    FIsSupportedComponent: TIsSupportedComponent;
    FReadComponentProps: TComponentPropsEvent;
    FSetComponentProps: TComponentPropsEvent;
    OwnForm: TForm;
    AFont: TFont;
    Fld: TField;
    BkColor: TColor;
    Timer: TTimer;
    procedure SetGrid(Value: TrDBGrid);
    procedure SetDataSource(Value: TDataSource);
    function GetDataSource: TDataSource;
    procedure SetPropSaveFile(Value: TPropSaveFile);
    procedure SetLinkedComps(Value: TStrings);
    procedure SetLinkAllComps(Value: boolean);
    procedure SetLoadAllCompsNow(Value: boolean);
    procedure SetEnabled(Value: boolean);
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure AddAllComps(LoadNow: boolean);
    procedure OnTimer(Sender: TObject);
    function IsPropSaveSet: boolean;
  protected
    FDataLink: TrDataLink;
    FGrid: TrDBGrid;
    FPropSaveFile: TPropSaveFile;
    FrDBGridsPropSave: TrDBGridsPropSave;
    FPropSaveKey: string;
    FCellParamsCondList: TCellParamsCondList;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override; {< @exclude }
    function IsSupportedComp(AComp: TComponent): boolean; virtual;
    procedure ReadCompProps(AComp: TComponent); virtual;
    procedure SetCompProps(AComp: TComponent); virtual;
    procedure Loaded; override; {< @exclude }
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { refresh colors of all linked components, called automatically when dataset scrolled or changed }
    procedure ColorLinkedComps;
  published
    { linked rDBGrid which is used as source of colors }
    property rDBGrid: TrDBGrid read FGrid write SetGrid;
    { linked Datasource which is used as source of data in case that rDBGrid is not assigned or created }
    property DataSource: TDataSource read GetDataSource write SetDataSource stored IsPropSaveSet;
    { linked PropSaveFile which is used as source of settings in case that rDBGrid is not assigned or created }
    property PropSaveFile: TPropSaveFile read FPropSaveFile write SetPropSaveFile;
    { define Key used by linked rDBGrid in case that rDBGrid is not assigned or created }
    property PropSaveKey: string read FPropSaveKey write FPropSaveKey stored IsPropSaveSet;
    { list of linked component names, set automatically if LinkAllComps is TRUE }
    property LinkedComps: TStrings read FLinkedComps write SetLinkedComps;
    { define whether all supported components is added automatically }
    property LinkAllComps: boolean read FLinkAllComps write SetLinkAllComps default true;
    { used only for loading of list of all supported components in design time }
    property LoadAllCompsNow: boolean read FLoadAllCompsNow write SetLoadAllCompsNow default false;
    { define whether update of linked components after loading is delayed to avoid flickering }
    property LinkDelayed: boolean read FLinkDelayed write FLinkDelayed default true;
    { define delay in ms for first update after loading }
    property LinkDelay: cardinal read FLinkDelay write FLinkDelay default 250;
    { define whether font size and name of linked components is kept independently of DBGrid font }
    property KeepFontSize: boolean read FKeepFontSize write FKeepFontSize default true;
    { define whether background color of linked components is kept independently of DBGrid color }
    property KeepBackground: boolean read FKeepBackground write FKeepBackground default false;
    { define whether function is Enabled }
    property Enabled: boolean read FEnabled write SetEnabled default true;
    { event can be used to define supported component types }
    property OnIsSupportedComponent: TIsSupportedComponent read FIsSupportedComponent write FIsSupportedComponent;
    { event can be used to read Field, Color and Font of any db avare component }
    property OnReadComponentProps: TComponentPropsEvent read FReadComponentProps write FReadComponentProps;
    { event can be used to set Color and Font of any db avare component }
    property OnSetComponentProps: TComponentPropsEvent read FSetComponentProps write FSetComponentProps;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rtool, rstring, rpictool;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBColorLinkedComp]);
end;

constructor TrDBColorLinkedComp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if AOwner is TForm then
    OwnForm:=TForm(AOwner)
  else
    OwnForm:=nil;
  AFont:=TFont.Create;
  FLinkedComps:=TStringList.Create;
  FDataLink:=TrDataLink.Create(ActiveChange, EditingChange, DataChange);
  FrDBGridsPropSave:=TrDBGridsPropSave.Create(Self);
  FrDBGridsPropSave.StoreColumnsProp:=true;
  FCellParamsCondList:=TCellParamsCondList.Create;
  Timer:=TTimer.Create(Self);
  Timer.Enabled:=false;
  FLinkAllComps:=true;
  FLoadAllCompsNow:=false;
  FLinkDelayed:=true;
  FLinkDelay:=250;
  FKeepFontSize:=true;
  FKeepBackground:=false;
  FEnabled:=true;
end;

destructor TrDBColorLinkedComp.Destroy;
begin
  FreeAndNil(AFont);
  FreeAndNil(FLinkedComps);
  FreeAndNil(FDataLink);
  FreeAndNil(FrDBGridsPropSave);
  FreeAndNil(FCellParamsCondList);
  inherited Destroy;
end;

procedure TrDBColorLinkedComp.Loaded;
begin
  inherited;
  // load component list
  if FLinkAllComps then
    AddAllComps(false);
  // init timer
  if FLinkDelayed and FEnabled and (FLinkDelay > 0) then
  begin
    Timer.OnTimer:=OnTimer;
    Timer.Interval:=FLinkDelay;
    Timer.Enabled:=true;
  end;
end;

procedure TrDBColorLinkedComp.OnTimer(Sender: TObject);
begin
  Timer.Enabled:=false;
  ColorLinkedComps;
end;

procedure TrDBColorLinkedComp.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FGrid) then
      FGrid:=nil;
    if Assigned(FDataLink) and (AComponent = FDataLink.DataSource) then
      FDataLink.DataSource:=nil;
    if (AComponent = FPropSaveFile) then
      FPropSaveFile:=nil;
  end;
end;

procedure TrDBColorLinkedComp.SetGrid(Value: TrDBGrid);
begin
  // link grid
  FGrid:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    // link to grid's datasource and set PropSaveKey
    if Assigned(FGrid) then
    begin
      if Assigned(FGrid.DataSource) then
      begin
        FDataLink.DataSource:=FGrid.DataSource;
        FGrid.DataSource.FreeNotification(Self);
        ActiveChange(Self);
      end;
      try
        FPropSaveKey:=FGrid.Owner.Name + '.' + FGrid.Name;
      except
        FPropSaveKey:='';
      end;
    end;
  end;
end;

procedure TrDBColorLinkedComp.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
  if Assigned(FGrid) and (FGrid.DataSource <> Value) then
    FGrid:=nil;
end;

function TrDBColorLinkedComp.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBColorLinkedComp.SetPropSaveFile(Value: TPropSaveFile);
begin
  FPropSaveFile:=Value;
  if Value <> nil then
    Value.FreeNotification(Self);
  SetGrid(FGrid);
end;

function TrDBColorLinkedComp.IsPropSaveSet;
begin
  Result:=Assigned(FPropSaveFile);
end;

procedure TrDBColorLinkedComp.SetLinkedComps(Value: TStrings);
begin
  FLinkedComps.Assign(Value);
end;

procedure TrDBColorLinkedComp.SetLinkAllComps(Value: boolean);
begin
  FLinkAllComps:=Value;
  if FLinkAllComps then
    AddAllComps(false);
end;

procedure TrDBColorLinkedComp.SetLoadAllCompsNow(Value: boolean);
begin
  FLoadAllCompsNow:=Value;
  if FLoadAllCompsNow then
  begin
    AddAllComps(true);
    Beep;
    FLoadAllCompsNow:=false;
  end;
end;

procedure TrDBColorLinkedComp.SetEnabled(Value: boolean);
begin
  FEnabled:=Value;
  ColorLinkedComps;
end;

procedure TrDBColorLinkedComp.ActiveChange(Sender: TObject);
begin
  if FEnabled then
    ColorLinkedComps;
end;

procedure TrDBColorLinkedComp.DataChange(Sender: TObject);
begin
  if FEnabled then
    ColorLinkedComps;
end;

procedure TrDBColorLinkedComp.EditingChange(Sender: TObject);
begin ;
end;

function TrDBColorLinkedComp.IsSupportedComp(AComp: TComponent): boolean;
begin
  // check standard supported comps
  Result:=(AComp is TDBEdit) or (AComp is TDBMemo) or (AComp is TDBComboBox) or (AComp is TDBLookupComboBox) or (AComp is TDBListBox) or
    (AComp is TDBLookupListBox) or (AComp is TDBRichEdit) or (AComp is TrDBCheckListBox);
  // call user defined event
  if Assigned(FIsSupportedComponent) then
    FIsSupportedComponent(AComp, Result);
end;

procedure TrDBColorLinkedComp.AddAllComps(LoadNow: boolean);
var
  i: integer;
  C: TComponent;
begin
  FLinkedComps.Clear;
  if not LoadNow and (csDesigning in ComponentState) or not Assigned(OwnForm) then
    Exit;
  for i:=0 to OwnForm.ComponentCount - 1 do
  begin
    C:=OwnForm.Components[i];
    if IsSupportedComp(C) then
      FLinkedComps.Add(C.Name);
  end;
end;

procedure TrDBColorLinkedComp.ReadCompProps(AComp: TComponent);
  procedure ReadInt(IntFld: TField; IntFont: TFont; IntColor: TColor);
  begin
    Fld:=IntFld;
    AFont.Assign(IntFont);
    BkColor:=IntColor;
  end;

begin
  Fld:=nil;
  if not IsSupportedComp(AComp) then
    Exit;
  if AComp is TDBEdit then
    with (AComp as TDBEdit) do
      ReadInt(Field, Font, Color);
  if AComp is TDBMemo then
    with (AComp as TDBMemo) do
      ReadInt(Field, Font, Color);
  if AComp is TDBComboBox then
    with (AComp as TDBComboBox) do
      ReadInt(Field, Font, Color);
  if AComp is TDBLookupComboBox then
    with (AComp as TDBLookupComboBox) do
      ReadInt(Field, Font, Color);
  if AComp is TDBListBox then
    with (AComp as TDBListBox) do
      ReadInt(Field, Font, Color);
  if AComp is TDBLookupListBox then
    with (AComp as TDBLookupListBox) do
      ReadInt(Field, Font, Color);
  if AComp is TDBRichEdit then
    with (AComp as TDBRichEdit) do
      ReadInt(Field, Font, Color);
  if AComp is TrDBCheckListBox then
    with (AComp as TrDBCheckListBox) do
      ReadInt(Field, Font, Color);
  if Assigned(FReadComponentProps) then
    FReadComponentProps(AComp, Fld, BkColor, AFont);
end;

procedure TrDBColorLinkedComp.SetCompProps(AComp: TComponent);
begin
  if AComp is TDBEdit then
    with (AComp as TDBEdit) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TDBMemo then
    with (AComp as TDBMemo) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TDBComboBox then
    with (AComp as TDBComboBox) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
      SelLength:=0;
    end;
  if AComp is TDBLookupComboBox then
    with (AComp as TDBLookupComboBox) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TDBListBox then
    with (AComp as TDBListBox) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TDBLookupListBox then
    with (AComp as TDBLookupListBox) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TDBRichEdit then
    with (AComp as TDBRichEdit) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if AComp is TrDBCheckListBox then
    with (AComp as TrDBCheckListBox) do
    begin
      Color:=BkColor;
      Font.Assign(AFont);
    end;
  if Assigned(FSetComponentProps) then
    FSetComponentProps(AComp, Fld, BkColor, AFont);
end;

procedure TrDBColorLinkedComp.ColorLinkedComps;
var
  C: TComponent;
  Col: TColumn;
  i, j: integer;
  loadPS: boolean;
  ImageDef: TDataImageDef;

  procedure UseColProps;
  begin
    // set font and color from column
    if FKeepFontSize then
    begin
      AFont.Color:=Col.Font.Color;
      AFont.Style:=Col.Font.Style;
    end
    else
      AFont.Assign(Col.Font);
    if not FKeepBackground then
      BkColor:=Col.Color;
  end;

begin
  if (csDesigning in ComponentState) or (csLoading in ComponentState) or not Assigned(OwnForm) or Timer.Enabled then
    Exit;

  // load props from mem INI ?
  loadPS:=(not Assigned(FGrid)) and Assigned(FPropSaveFile) and Assigned(FDataLink.DataSet);
  if loadPS then
  begin
    // prepare all objects and load props from mem INI
    FCellParamsCondList.LinkDataset(FDataLink.DataSet);
    FrDBGridsPropSave.ReadCellParamsCondList(FPropSaveFile, FPropSaveKey, FCellParamsCondList);
    FrDBGridsPropSave.ReadDBColumnList(FPropSaveFile, FPropSaveKey);
  end;

  for i:=0 to FLinkedComps.Count - 1 do
  begin
    C:=OwnForm.FindComponent(FLinkedComps[i]);
    // read field and font for component
    ReadCompProps(C);
    if not Assigned(Fld) then
      Continue;

    // set default font and color
    AFont.Color:=clWindowText;
    AFont.Style:=[];
    if not FKeepBackground then
      BkColor:=clWindow;

    // if not enabled set default
    if not FEnabled then
    begin
      SetCompProps(C);
      Continue;
    end;

    if Assigned(FGrid) then
    begin
      // use current Grid props, columns and CellParamsCondList
      // find grid column with same field
      Col:=nil;
      for j:=0 to FGrid.Columns.Count - 1 do
      begin
        if FGrid.Columns[j].Field = Fld then
        begin
          Col:=FGrid.Columns[j];
          Break;
        end;
      end;
      if not Assigned(Col) then
        Continue;

      // read color and font according defined conditions in rDBGrid
      try
        UseColProps;
        FGrid.CellParamsCondList.GetFieldCond(Fld, AFont, BkColor, ImageDef, false, false, true);
      except
      end;
    end
    else
    begin
      if loadPS and Assigned(Fld) then
      begin
        // grid not used = load Column settings and CellParamsCondList from PropSaveFile and use own objects
        Col:=TColumn.Create(nil);
        try
          j:=FrDBGridsPropSave.SavedDBColumnFieldList.IndexOf(Fld.FieldName);
          if j >= 0 then
          begin
            FrDBGridsPropSave.ReadDBColumn(FPropSaveFile, FPropSaveKey, Col, j);
            try
              UseColProps;
              FCellParamsCondList.GetFieldCond(Fld, AFont, BkColor, ImageDef, false, false, true);
            except
            end;
          end;
        finally
          FreeAndNil(Col);
        end;
      end;
    end;

    // set new color and font
    SetCompProps(C);
  end;
end;

end.
