unit rDBGrid_FilterForm;

{<
  rDBGrid_FilterForm form shows list of field values from TrDBGrid dataset
  and allows selection of filtered values.

  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(July 2013)
  @lastmod(May 2018)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus, ImgList,
  Grids, DBGrids, DB, DBCtrls, Forms, Dialogs, StdCtrls, ExtCtrls,
  Types, Math, CheckLst, ComCtrls, ToolWin, rImprovedComps, PropSaveMain,
  RosiCompGlobalCfg;

type
  { form shows list of field values from TrDBGrid dataset
    and allows selection of filtered values }
  TrDBGridFilterForm = class(TForm)
    pClient: TPanel;
    pBtn: TPanel;
    btOk: TButton;
    btCancel: TButton;
    CheckBoxList: TrCheckListBox;
    tbTop: TToolBar;
    tbEnter: TToolButton;
    tbClear: TToolButton;
    tbSep: TToolButton;
    ImgList: TImageList;
    PropSaveFile1: TPropSaveFile;
    chAnd: TCheckBox;
    procedure FormDeactivate(Sender: TObject);
    procedure btCancelClick(Sender: TObject);
    procedure CheckBoxListClickCheck(Sender: TObject);
    procedure btOkClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure tbClearClick(Sender: TObject);
    procedure tbEnterClick(Sender: TObject);
    procedure CheckBoxListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
  private
    Gr: TDBGrid;
    DS: TDataset;
    Aborted, CheckAll: boolean;
    maxw: Integer;
    procedure SetFilter;
    procedure IntSetFilter(FilterStr: string);
    procedure OnAbort(Sender: TObject);
    procedure AddToList(DisplayText, ValueText: string; Checked: boolean);
  public
    DisplayField, ValueField: TField;
    DisplayList, ValueList: TStringList;
    GridCol: TColumn;
    procedure LoadList(DBGrid: TDBGrid; C: TColumn);
  end;

var
  rDBGridFilterForm: TrDBGridFilterForm;

implementation

{$R *.dfm}

uses rstring, rtool, rdlg, rdbtool, rDBGrid, rDBGrid_FilterCondForm;

procedure TrDBGridFilterForm.FormCreate(Sender: TObject);
begin
  DisplayList:=TStringList.Create;
  ValueList:=TStringList.Create;
  DisplayList.Sorted:=false;
  ValueList.Sorted:=false;
end;

procedure TrDBGridFilterForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(DisplayList);
  FreeAndNil(ValueList);
end;

procedure TrDBGridFilterForm.FormDeactivate(Sender: TObject);
begin
  Hide;
end;

procedure TrDBGridFilterForm.btCancelClick(Sender: TObject);
begin
  Hide;
  // reset last FFilterDlgLastColIdx
  TrDBGrid(Gr).FilteredFieldNames:=TrDBGrid(Gr).FilteredFieldNames;
end;

procedure TrDBGridFilterForm.btOkClick(Sender: TObject);
begin
  SetFilter;
end;

// used also as callback for custom event
procedure TrDBGridFilterForm.AddToList(DisplayText, ValueText: string; Checked: boolean);
const
  obj: array [boolean] of integer = (0,1);
var
  w: Integer;
begin
  // DisplayList item contains list text, object contains index of ValueList list, which contains filtered values and state
  DisplayList.AddObject(DisplayText, TObject(ValueList.AddObject(ValueText, TObject(obj[Checked]))));
  if not Checked then
    CheckAll:=false;
  if DisplayText <> sFilterNullValue then
  begin
    // store max width of item
    w:=CheckBoxList.Canvas.TextWidth(DisplayText);
    if w > maxw then
      maxw:=w;
  end;
end;

procedure TrDBGridFilterForm.LoadList(DBGrid: TDBGrid; C: TColumn);
var
  G: TrDBGrid;
  Bk: TBookMark;
  i, RecIdx, RecCount: Integer;
  IsLoaded, oldFiltered, ShowProgress: boolean;
  list, val: string;

  function ProcAllRec(Checked: boolean): boolean;
  begin
    Result:=false;
    DS.First;
    while not DS.EOF do
    begin
      if DisplayField.IsNull then
      begin
        val:=sFilterNullValue;
        list:=val;
      end
      else
      begin
        val:=ValueField.AsString;
        if rdbtool.IsMemoField(DisplayField) then
          list:=copy(DisplayField.AsString, 1, 200)
        else
          list:=DisplayField.DisplayText;
      end;
      if ValueList.IndexOf(val) < 0 then
      begin
        AddToList(list, val, Checked);
        Result:=true;
      end;
      DS.Next;
      Inc(i);
      Inc(RecIdx);
      if ShowProgress and (RecIdx mod 100 = 0) then
        ShowInfoProgressBar(sLoading + sOfData, Gr, OnAbort, Round(100 * RecIdx / RecCount));
      Application.ProcessMessages;
      if Aborted then
        Break;
    end;
  end;

begin
  // init list and comps
  DisplayList.Clear;
  ValueList.Clear;
  CheckBoxList.Items.Clear;
  if not Assigned(DBGrid) or not(DBGrid is TrDBGrid) or not Assigned(C.Field) then
    Exit;
  if not(IsSimpleField(C.Field) or IsMemoField(C.Field)) then
    Exit;
  Gr:=DBGrid;
  GridCol:=C;
  G:=Gr as TrDBGrid;
  DS:=G.DataSource.DataSet;
  DisplayField:=C.Field;
  ValueField:=DisplayField;
  if DisplayField.FieldKind = fkLookup then
    ValueField:=DS.FieldByName(DisplayField.KeyFields);

  CheckAll:=true;
  oldFiltered:=DisplayField.DataSet.Filtered;
  tbEnter.Enabled:=(DisplayField.FieldKind = fkData) and not(DisplayField is TBooleanField);
  tbClear.Enabled:=oldFiltered;
  IsLoaded:=false;
  // use custom event
  if Assigned(G.OnLoadFunnelFilter) then
    G.OnLoadFunnelFilter(self, DS, DisplayField.FieldName, AddToList, IsLoaded);
  if not IsLoaded then
  begin
    // load field values by dataset scan
    Aborted:=false;
    i:=0;
    RecIdx:=0;
    maxw:=0;
    RecCount:=DS.RecordCount;
    ShowProgress:=(RecCount < 0) or (RecCount > 1000);
    if ShowProgress then
      ShowInfoEx(sLoading + sOfData, Gr, OnAbort)
    else
      WCursor;
    Bk:=DS.GetBookmark;
    try
      DS.DisableControls;
      G.SaveRowPos(true);
      G.SaveResetDatasetEvents(grrLoadFilterList);
      try
        // proc filtered only or all record if filter is off
        ProcAllRec(true);
        if DS.Filtered then
        begin
          // proc rest if filter is on
          DS.Filtered:=false;
          RecIdx:=0;
          RecCount:=DS.RecordCount;
          ShowProgress:=(RecCount < 0) or (RecCount > 1000);
          CheckAll:=not ProcAllRec(false);
        end;
        DS.Filtered:=oldFiltered;
        DS.GotoBookmark(Bk);
      finally
        G.RestoreRowPos;
        G.RestoreDatasetEvents(grrLoadFilterList);
        DS.EnableControls;
      end;
      // sort items
      if IsDateTimeField(DisplayField) then
        DisplayList.CustomSort(SortDate)
      else
        if DisplayField is TNumericField then
          DisplayList.CustomSort(SortNum)
        else
          DisplayList.Sort;
    finally
      DS.FreeBookmark(Bk);
      if ShowProgress then
        ShowInfo('');
      NCursor;
    end;
  end;
  // add "Select all" item
  DisplayList.Insert(0, sFilterSelectAll);
  CheckBoxList.Items.Assign(DisplayList);
  // check filtered
  if CheckAll then
    CheckBoxList.CheckAll(cbChecked)
  else
    for i:=1 to CheckBoxList.Items.Count - 1 do
    begin
      CheckBoxList.Checked[i]:=Integer(ValueList.Objects[Integer(DisplayList.Objects[i])]) = 1;
    end;
end;

procedure TrDBGridFilterForm.OnAbort(Sender: TObject);
begin
  Aborted:=true;
end;

procedure TrDBGridFilterForm.CheckBoxListClickCheck(Sender: TObject);
begin
  if CheckBoxList.ItemIndex = 0 then
  begin
    if CheckBoxList.Checked[0] then
      CheckBoxList.CheckAll(cbChecked)
    else
      CheckBoxList.CheckAll(cbUnChecked);
  end
  else
  begin
    if not CheckBoxList.Checked[CheckBoxList.ItemIndex] then
      CheckBoxList.Checked[0]:=false;
  end;
end;

procedure TrDBGridFilterForm.CheckBoxListDrawItem(Control: TWinControl; Index: Integer; Rect: TRect; State: TOwnerDrawState);
var
  s: string;
  Flags: Longint;
begin
  CheckBoxList.Canvas.FillRect(Rect);
  if Index < CheckBoxList.Items.Count then
  begin
    s:=CheckBoxList.Items[Index];
    Flags:=0;
    if RosiCompConfig.rDBGridCfg.FilterListUseColumnAlignment and (s <> sFilterSelectAll) and (s <> sFilterNullValue) then
      case GridCol.Alignment of
        taRightJustify:
          Flags:=DT_RIGHT;
        taCenter:
          Flags:=DT_CENTER;
      else
        Flags:=0;
      end;
    if Flags <> 0 then
      if maxw < Rect.Right - Rect.Left then
        Rect.Right:=Rect.Left + maxw + 2;
    Flags:=DrawTextBiDiModeFlags(Flags or DT_SINGLELINE or DT_VCENTER or DT_NOPREFIX);
    if not UseRightToLeftAlignment then
      Inc(Rect.Left, 2)
    else
      Dec(Rect.Right, 2);
    if RosiCompConfig.rDBGridCfg.FilterListBoldSelectAll and (s = sFilterSelectAll) then
      CheckBoxList.Canvas.Font.Style:=[fsBold];
    DrawText(CheckBoxList.Canvas.Handle, s, Length(s), Rect, Flags);
  end;
end;

procedure TrDBGridFilterForm.SetFilter;
var
  i, vidx: Integer;
  flt, val, cond: string;
  isNullVal: boolean;
begin
  flt:='';
  if not CheckBoxList.Checked[0] then
  begin
    for i:=1 to CheckBoxList.Items.Count - 1 do
      if CheckBoxList.Checked[i] then
      begin
        vidx:=Integer(DisplayList.Objects[i]);
        val:=ValueList[vidx];
        isNullVal:=ValueList[vidx] = sFilterNullValue;
        try
          if not isNullVal and IsDateTimeField(ValueField) then
            val:=GetDateFieldFilterValue(StrToDate(val));
        except
        end;
        cond:=GetFilterCond(ValueField, val, isNullVal);
        if cond <> '' then
        begin
          if flt <> '' then
            flt:=flt + ' OR ';
          flt:=flt + cond;
        end;
      end;
  end;
  IntSetFilter(flt);
end;

procedure TrDBGridFilterForm.tbClearClick(Sender: TObject);
begin
  IntSetFilter('');
end;

procedure TrDBGridFilterForm.tbEnterClick(Sender: TObject);
begin
  btCancel.Click;
  CreateF(TrDBGridFilterCondForm, rDBGridFilterCondForm);
  rDBGridFilterCondForm.BiDiMode:=BiDiMode;
  rDBGridFilterCondForm.InitCond(Gr, DisplayField);
  // read ini state
  rDBGridFilterCondForm.DlgPropSave.PropSaveFile:=PropSaveFile1;
  rDBGridFilterCondForm.DlgPropSave.ReadProperties;
  // set list
  rDBGridFilterCondForm.cbValText1_nostore.Items.Assign(CheckBoxList.Items);
  rDBGridFilterCondForm.cbValText1_nostore.Items.Delete(0);
  if rDBGridFilterCondForm.cbValText1_nostore.Items[0] = sFilterNullValue then
    rDBGridFilterCondForm.cbValText1_nostore.Items.Delete(0);

  if rDBGridFilterCondForm.ShowModal = mrOk then
  begin
    IntSetFilter(rDBGridFilterCondForm.FilterString);
    // store to ini
    rDBGridFilterCondForm.DlgPropSave.SaveProperties;
  end;
  FreeF(rDBGridFilterCondForm);
end;

procedure TrDBGridFilterForm.IntSetFilter(FilterStr: string);
var
  G: TrDBGrid;
  res: boolean;
begin
  G:=Gr as TrDBGrid;
  res:=true;
  if Assigned(G.OnSetColumnFilter) then
  begin
    G.OnSetColumnFilter(Gr, DS, ValueField, FilterStr, '', self, res);
    if res then
    begin
      Hide;
      G.Invalidate;
      if Assigned(G.OnFilterChanged) then
        G.OnFilterChanged(G);
    end;
  end
  else
  begin
    G.SetFilterForField(DisplayField, FilterStr, self.chAnd.Checked);
    Hide;
  end;
end;

end.
