unit rDBGrid_MS;

{<
  rDBGrid_MS unit contains TrDBGrid_MS - enhanced TrDBGrid component which provides
  multiselect function by click on fixed column with definable colors.
  TrDBGrid_MS provide multiselect of range of records by Shift+click
  or Ctrl+Click on fixed column and popup menu on fixed column.

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(May 2009)
  @lastmod(Dec 2020)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Controls, Graphics, Math, Forms,
  Grids, DBGrids, rDBGrid, DB, VDBConsts, Menus,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} Dialogs, RosiCompGlobalCfg;

type
  TrDBGrid_MS = class;

  { class to store list of selected records }
  TrBookmarkList = class(TBookmarkList)
  protected
    FrDBGrid: TrDBGrid_MS;
    function ChangeSelected(ASelected: boolean; Done: boolean = true): boolean; overload;
    function ChangeSelected(ASelected: boolean; var Counter: integer): boolean; overload;
  public
  end;

  { @exclude }
{$IFDEF D2009P}

  TIntBookmark = TBookmark;
{$ELSE}
  TIntBookmark = TBookmarkStr;
{$ENDIF}
  { list of scope types used by ProcScopeRecords procedure to define which records is processed }
  TRecordScope = (rsSelectedRecOnly, rsSelectedOrActiveRec, rsSelectedOrAllRec, rsActiveRecOnly, rsAllRec);
  { event type for getting of possibility to delete selected records }
  TCanDeleteSelected = procedure(Sender: TObject; var CanDelete: boolean) of object;

  { enhanced TrDBGrid component which provides
    multiselect function by click on fixed column with definable colors }
  TrDBGrid_MS = class(TrDBGrid)
  private
    FrBookmarks: TrBookmarkList;
    FFixColPopup: TPopupMenu;
    FUseMultiSelect: boolean;
    FSingleSelect: boolean; // allows single click selection, not shift-click
    FShowMultiselectPopup: boolean;
    FSelectByClickOnRow: boolean;
    FDeselectAllByClickOnRow: boolean;
    LastSelRow: TIntBookmark;
    FSelectedRowsBkColor: TColor;
    FSelectedRowsFontColor: TColor;
    FOnDeleteSelected: TNotifyEvent;
    FDeleteOnlySelected: boolean;
    FSelectingConfirmRecCount: integer;
    FOnSelectingRecords: TNotifyEvent;
    FOnSelectionChanged: TNotifyEvent;
    FOnCanDeleteSelected: TCanDeleteSelected;
    procedure SelOrInverseAll(Sel: boolean);
  protected
    procedure Loaded; override;
    procedure LinkActive(Value: boolean); override;
    procedure CellClick(Column: TColumn); override;
    procedure TitleExClick(Button: TMouseButton; Shift: TShiftState); override;
    procedure FixColClick(Button: TMouseButton; Shift: TShiftState); override;
    procedure DoContextPopup(MousePos: TPoint; var Handled: boolean); override;
    procedure FixColPopupClick(Sender: TObject); virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    function RowIsMultiselected(FromActiveRecord: boolean; FromDataRow: integer): boolean; override;
    procedure SelectionChanged;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { list of bookmarks to selected records }
    property rBookmarks: TrBookmarkList read FrBookmarks;
    { list of bookmarks to selected records - compability only }
    property MyBookmarks: TrBookmarkList read FrBookmarks;
    { return if active record is selected }
    function IsActiveSelected: boolean;
    { select active record }
    procedure SelectActive;
    { deselect active record }
    procedure DeSelectActive;
    { select all records }
    procedure SelectAll;
    { deselect all records }
    procedure DeSelectAll;
    { reverse selection - select unselected and unselect selected records }
    procedure ReverseSelection;
    { delete all selected records, if no record is selected, property
      DeleteOnlySelected is used and active record is delete or nothing is done }
    function DeleteSelected: boolean;
    { indicates and set/reset multiselect function in runtime }
    property UseMultiSelect: boolean read FUseMultiSelect write FUseMultiSelect;
    { indicates and set/reset single select function in runtime }
    property SingleSelect: boolean read FSingleSelect write FSingleSelect;
    { returns count of selected records }
    function GetSelectedRecCount: integer;
    { returns count of selected records according RecordScope }
    function GetScopeRecCount(RecordScope: TRecordScope): integer;
    { proccess all records according RecordScope and call CallBackProc for each of them }
    procedure ProcScopeRecords(RecordScope: TRecordScope; CallBackProc: TNotifyEvent);
    { add bk color for selected records }
    procedure GetCellParams(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef; ActiveRow, Highlight: boolean); override;
    { Popup menu for fix column }
    property FixColPopup: TPopupMenu read FFixColPopup;
  published
    { define whether selection is done by click to any cell }
    property SelectByClickOnRow: boolean read FSelectByClickOnRow write FSelectByClickOnRow default false;
    { define whether selection is done by click to any cell }
    property DeselectAllByClickOnRow: boolean read FDeselectAllByClickOnRow write FDeselectAllByClickOnRow default true;
    { define background color of selected row }
    property SelectedRowsBkColor: TColor read FSelectedRowsBkColor write FSelectedRowsBkColor default clHighlight;
    { define font color of selected row }
    property SelectedRowsFontColor: TColor read FSelectedRowsFontColor write FSelectedRowsFontColor default clHighlightText;
    { define count of selected records for action confirmation to aviod freeze for time consumin actions }
    property SelectingConfirmRecCount: integer read FSelectingConfirmRecCount write FSelectingConfirmRecCount default 1000;
    { define whether popup for fixed column is shown }
    property ShowMultiselectPopup: boolean read FShowMultiselectPopup write FShowMultiselectPopup default true;
    { define if active record is deleted if nothing is selected }
    property DeleteOnlySelected: boolean read FDeleteOnlySelected write FDeleteOnlySelected default false;
    { event is called for deletion of records, if not defined, standard delete method is used }
    property OnDeleteSelected: TNotifyEvent read FOnDeleteSelected write FOnDeleteSelected;
    { event is called during selection of records for each record }
    property OnSelectingRecords: TNotifyEvent read FOnSelectingRecords write FOnSelectingRecords;
    { event is called after selection or deselection of records }
    property OnSelectionChanged: TNotifyEvent read FOnSelectionChanged write FOnSelectionChanged;
    { event is called before showing of fix column popup menu to obtain possibility to delete selected records }
    property OnCanDeleteSelected: TCanDeleteSelected read FOnCanDeleteSelected write FOnCanDeleteSelected;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rdlg;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBGrid_MS]);
end;

// ==============================================================================
// TrBookmarkList
function TrBookmarkList.ChangeSelected(ASelected: boolean; var Counter: integer): boolean;
begin
  Result:=true;
  CurrentRowSelected:=ASelected;
  Inc(Counter);
  if ASelected and Assigned(FrDBGrid.FOnSelectingRecords) then
    FrDBGrid.FOnSelectingRecords(Self);
  if (FrDBGrid.FSelectingConfirmRecCount > 0) and (Counter mod FrDBGrid.FSelectingConfirmRecCount = 0) then
    Result:=DlgQ(Format(sSelectedNumRecContinueQ, [Counter])) = mrYes;
end;

function TrBookmarkList.ChangeSelected(ASelected: boolean; Done: boolean = true): boolean;
var
  i: integer;
begin
  i:=0;
  Result:=ChangeSelected(ASelected, i);
  if Done then
    FrDBGrid.SelectionChanged;
end;

// ==============================================================================
// TrDBGrid_MS
constructor TrDBGrid_MS.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FrBookmarks:=TrBookmarkList.Create(Self);
  FrBookmarks.FrDBGrid:=Self;
  FSelectedRowsBkColor:=clHighlight;
  FSelectedRowsFontColor:=clHighlightText;
  FUseMultiSelect:=false;
  FSingleSelect:=false;
  FSelectByClickOnRow:=false;
  FDeselectAllByClickOnRow:=true;
  FSelectingConfirmRecCount:=1000;
  FShowMultiselectPopup:=true;
  if not(csDesigning in ComponentState) then
  begin
    // title popup
    FFixColPopup:=TPopupMenu.Create(Self);
    FFixColPopup.AutoHotkeys:=maManual;
    FFixColPopup.Items.Add(CreateMenuItem(Self, sSelectAllRecords, 1, FixColPopupClick));
    FFixColPopup.Items.Add(CreateMenuItem(Self, sDeselectAllRecords, 2, FixColPopupClick));
    FFixColPopup.Items.Add(CreateMenuItem(Self, sReverseSelection, 3, FixColPopupClick));
    FFixColPopup.Items.Add(CreateMenuItem(Self, '-', 0, FixColPopupClick));
    FFixColPopup.Items.Add(CreateMenuItem(Self, sDeleteSelectedRecords, 4, FixColPopupClick));
  end;
end;

destructor TrDBGrid_MS.Destroy;
begin
  FreeAndNil(FrBookmarks);
  inherited Destroy;
end;

procedure TrDBGrid_MS.Loaded;
begin
  inherited Loaded;
  if not(csDesigning in ComponentState) then
  begin
    FUseMultiSelect:=(dgMultiSelect in Options);
    if FUseMultiSelect then
      Options:=Options - [dgMultiSelect];
  end;
end;

procedure TrDBGrid_MS.LinkActive(Value: boolean);
begin
  FrBookmarks.LinkActive(Value);
  inherited;
end;

procedure TrDBGrid_MS.SelectionChanged;
begin
  if Assigned(FOnSelectionChanged) then
    FOnSelectionChanged(Self);
end;

// ==============================================================================
// Select/Deselect group
function TrDBGrid_MS.RowIsMultiselected(FromActiveRecord: boolean; FromDataRow: integer): boolean;
var
  OldActive: integer;
begin
  try
    if FromActiveRecord then
      Result:=Datalink.Active and FrBookmarks.CurrentRowSelected
    else
    begin
      Result:=false;
      if (FromDataRow < 0) or not Datalink.Active then
        Exit;
      OldActive:=Datalink.ActiveRecord;
      try
        Datalink.ActiveRecord:=FromDataRow;
        Result:=FrBookmarks.CurrentRowSelected;
      finally
        Datalink.ActiveRecord:=OldActive;
      end;
    end;
  except
    Result:=false;
  end;
end;

function TrDBGrid_MS.IsActiveSelected: boolean;
begin
  try
    Result:=Datalink.Active and FrBookmarks.CurrentRowSelected;
  except
    Result:=false;
  end;
end;

procedure TrDBGrid_MS.SelectActive;
begin
  FrBookmarks.ChangeSelected(true);
  LastSelRow:=FrBookmarks.CurrentRow;
end;

procedure TrDBGrid_MS.DeSelectActive;
begin
  FrBookmarks.ChangeSelected(false);
end;

procedure TrDBGrid_MS.SelOrInverseAll(Sel: boolean);
var
  SelRow: TIntBookmark;
  i: integer;
begin
  // select all or reverse
  FrBookmarks.Refresh;
  SelRow:=FrBookmarks.CurrentRow;
  SaveRowPos;
  Datalink.Datasource.Dataset.DisableControls;
  i:=0;
  WCursor;
  try
    Datalink.Datasource.Dataset.First;
    repeat
      if not FrBookmarks.ChangeSelected(Sel or not FrBookmarks.CurrentRowSelected, i) then
        Break;
      Datalink.Datasource.Dataset.Next;
    until (Datalink.Datasource.Dataset.Eof or Datalink.Datasource.Dataset.Bof);
  finally
    NCursor;
    Datalink.Datasource.Dataset.Bookmark:=SelRow;
    Datalink.Datasource.Dataset.EnableControls;
    RestoreRowPos;
    SelectionChanged;
  end;
end;

procedure TrDBGrid_MS.SelectAll;
begin
  SelOrInverseAll(true);
end;

procedure TrDBGrid_MS.DeSelectAll;
begin
  if FrBookmarks.Count = 0 then
    Exit;
  FrBookmarks.Clear;
  SelectionChanged;
end;

procedure TrDBGrid_MS.ReverseSelection;
begin
  SelOrInverseAll(false);
end;

function TrDBGrid_MS.DeleteSelected: boolean;
var
  Msg: string;
begin
  Result:=false;
  if Assigned(FOnDeleteSelected) then
    FOnDeleteSelected(Self)
  else
    with Datalink.Dataset do
    begin
      if (not ReadOnly) and (not IsEmpty) and CanModify then
      begin
        if DeleteOnlySelected and (FrBookmarks.Count = 0) then
        begin
          DlgW(sNoRecordsSelected);
          Exit;
        end;
        if FrBookmarks.Count > 1 then
          Msg:=sDelMultipleRecordQ
        else
          Msg:=sDelRecordQ;
        if not(dgConfirmDelete in Options) or (DlgQ(Msg) = mrYes) then
        begin
          if FrBookmarks.Count > 0 then
            FrBookmarks.Delete
          else
            Delete;
          DeSelectAll;
          Result:=true;
        end;
      end;
    end;
end;

// ==============================================================================
// Click and event group
procedure TrDBGrid_MS.CellClick;
begin
  inherited;
  // DeselectAll;
end;

procedure TrDBGrid_MS.TitleExClick(Button: TMouseButton; Shift: TShiftState);
begin
  if FUseMultiSelect and OptionsEx2.FixedCell.TitleExBtnSelectAll then
    SelectAll;
  inherited;
end;

procedure TrDBGrid_MS.FixColClick(Button: TMouseButton; Shift: TShiftState);
  procedure SelectRow;
  begin
    FrBookmarks.ChangeSelected(not FrBookmarks.CurrentRowSelected);
    LastSelRow:=FrBookmarks.CurrentRow;
  end;

var
  i, r, prevr: integer;
  NewSelRow: TIntBookmark;
begin
  inherited;
  if not FUseMultiSelect or not Datalink.Active then
    Exit;
  if (ssShift in Shift) and not FSingleSelect then
  begin
    // select range from last selected to actual record
    FrBookmarks.Refresh;
    if not Datalink.Datasource.Dataset.BookmarkValid(LastSelRow) then
      // if (not FMyBookmarks.Find(LastSelRow, r)) then
      SelectRow
    else
    begin
      NewSelRow:=FrBookmarks.CurrentRow;
      SaveRowPos;
      Datalink.Datasource.Dataset.DisableControls;
      try
        i:=0;
        prevr:=0;
        repeat
          if not FrBookmarks.ChangeSelected(true, i) then
            Break;
          r:=FrBookmarks.Compare(FrBookmarks.CurrentRow, LastSelRow);
          if r = 0 then
            Break;
          if (prevr <> 0) and (r <> prevr) then
            // change direction - fix possible bug in the dataset causing endless loop (kbmMemTable)
            Break;
          prevr:=r;
          if r < 0 then
            Datalink.Datasource.Dataset.Next
          else
            Datalink.Datasource.Dataset.Prior;
        until (Datalink.Datasource.Dataset.Eof or Datalink.Datasource.Dataset.Bof);
      finally
        Datalink.Datasource.Dataset.Bookmark:=NewSelRow;
        Datalink.Datasource.Dataset.EnableControls;
        RestoreRowPos;
        SelectionChanged;
      end;
    end;
  end
  else
    SelectRow;
end;

procedure TrDBGrid_MS.FixColPopupClick(Sender: TObject);
var
  Tag: integer;
begin
  if not FUseMultiSelect then
    Exit;
  Tag:=TComponent(Sender).Tag;
  case Tag of
    1:
      SelectAll;
    2:
      DeSelectAll;
    3:
      ReverseSelection;
    4:
      DeleteSelected;
  end;
end;

procedure TrDBGrid_MS.KeyDown(var Key: Word; Shift: TShiftState);
var
  del: boolean;
begin
  if (ssCtrl in Shift) and (Key = VK_DELETE) then
  begin
    del:=not ReadOnly and (dgEditing in Options);
    if Assigned(FOnCanDeleteSelected) then
      FOnCanDeleteSelected(Self, del);
    if del then
    begin
      DeleteSelected;
      Key:=0;
    end;
  end;
  inherited KeyDown(Key, Shift);
  if FUseMultiSelect and (ssShift in Shift) and (Key in [VK_UP, VK_DOWN]) then
  begin
    if Datalink.Datasource.Dataset.BookmarkValid(PrevActiveBookmark) then
      LastSelRow:=PrevActiveBookmark;
    FixColClick(mbLeft, Shift);
  end;
end;

procedure TrDBGrid_MS.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  UpCoord: TGridCoord;
begin
  inherited MouseUp(Button, Shift, X, Y);
  if not FUseMultiSelect then
    Exit;
  UpCoord:=MouseCoord(X, Y);
  if (Button = mbLeft) and (UpCoord.X > FixCol) and (UpCoord.Y > TitleRow) then
  begin
    if FSelectByClickOnRow or (ssCtrl in Shift) or (ssShift in Shift) then
    begin
      if (ssShift in Shift) and Datalink.Datasource.Dataset.BookmarkValid(PrevActiveBookmark) then
        LastSelRow:=PrevActiveBookmark;
      FixColClick(Button, Shift);
    end
    else
      if FDeselectAllByClickOnRow then
        DeSelectAll;
  end;
end;

procedure TrDBGrid_MS.GetCellParams(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef; ActiveRow, Highlight: boolean);
begin
  if (not Highlight) and FrBookmarks.CurrentRowSelected then
  begin
    Background:=FSelectedRowsBkColor;
    AFont.Color:=FSelectedRowsFontColor;
  end;
  inherited;
end;

procedure TrDBGrid_MS.DoContextPopup(MousePos: TPoint; var Handled: boolean);
var
  i: integer;
  del: boolean;
begin
  // popup on fixec col
  if FShowMultiselectPopup and FUseMultiSelect and (OptionsEx2.FixedCell.FixedColBtn) and (MouseCoord(MousePos.X, MousePos.Y).X = FixCol) then
  begin
    del:=not ReadOnly and (dgEditing in Options);
    if Assigned(FOnCanDeleteSelected) then
      FOnCanDeleteSelected(Self, del);
    for i:=FFixColPopup.Items.Count - 1 downto 0 do
    begin
      if FFixColPopup.Items[i].Tag = 4 then
      begin
        FFixColPopup.Items[i].Enabled:=del and (FrBookmarks.Count > 0);
        FFixColPopup.Items[i].Caption:=sDeleteSelectedRecords + Format(' (%d)', [FrBookmarks.Count]);
      end;
    end;
    FFixColPopup.Popup(ClientToScreen(MousePos).X, ClientToScreen(MousePos).Y);
    Handled:=true;
  end;
  inherited;
end;

// ==============================================================================
// Scope record proc group
function TrDBGrid_MS.GetSelectedRecCount: integer;
begin
  Result:=FrBookmarks.Count;
end;

function TrDBGrid_MS.GetScopeRecCount(RecordScope: TRecordScope): integer;
begin
  if not IsActive then
  begin
    Result:=0;
    Exit;
  end;
  Result:=FrBookmarks.Count;
  case RecordScope of
    rsActiveRecOnly:
      if not Datasource.Dataset.IsEmpty then
        Result:=1;
    rsSelectedOrActiveRec:
      if Result = 0 then
        if not Datasource.Dataset.IsEmpty then
          Result:=1;
    rsSelectedOrAllRec:
      if Result = 0 then
        Result:=Datasource.Dataset.RecordCount;
    rsAllRec:
      Result:=Datasource.Dataset.RecordCount;
  end;
end;

procedure TrDBGrid_MS.ProcScopeRecords(RecordScope: TRecordScope; CallBackProc: TNotifyEvent);

  procedure ProcActiveRec;
  begin
    if not Datasource.Dataset.IsEmpty then
      CallBackProc(Self);
  end;

  procedure ProcSelRec;
  var
    i: integer;
  begin
    for i:=0 to FrBookmarks.Count - 1 do
    begin
      Datasource.Dataset.GotoBookmark(FrBookmarks[i]);
      CallBackProc(Self);
    end;
  end;

  procedure ProcAllRec;
  begin
    Datasource.Dataset.First;
    while not Datasource.Dataset.Eof do
    begin
      CallBackProc(Self);
      Datasource.Dataset.Next;
    end;
  end;

begin
  if not IsActive then
    Exit;
  WCursor;
  try
    case RecordScope of
      rsActiveRecOnly:
        ProcActiveRec;
      rsSelectedRecOnly:
        ProcSelRec;
      rsSelectedOrActiveRec:
        if FrBookmarks.Count > 0 then
          ProcSelRec
        else
          ProcActiveRec;
      rsSelectedOrAllRec:
        if FrBookmarks.Count > 0 then
          ProcSelRec
        else
          ProcAllRec;
      rsAllRec:
        ProcAllRec;
    end;
  finally
    NCursor;
  end;
end;

end.
