unit rDBStringGridEd;

{<
  rDBStringGridEd unit contains TrDBStringGridEd - enhanced TrStringGridEd component
  which provides functions for batch loading and saving data from/to dataset.
  It helps to use simple table, multitable or pivot table data showing and editing.

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2015)
  @lastmod(Feb 2021)
}

interface

{$INCLUDE _rosi.inc}

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus,
  Dialogs, Grids, StdCtrls, ComCtrls, Clipbrd, Math, Types, Themes, db,
  rStringGridEd, rDBComponents, rImprovedComps, RosiCompGlobalCfg;

type
  TrDBStringGridEd = class;
  TrDBStringGridEdMultiRow = class;
  TrDBStringGridEdMultiRowClass = class of TrDBStringGridEdMultiRow;

  { list of options used by TrDBStringGridEd }
  TOptionDBEx = (dgSetAutoIncReadOnly, dgSetAutoIncGrayed, dgSaveKeyFieldValue);
  { set of options used by TrDBStringGridEd }
  TOptionsDBEx = set of TOptionDBEx;

  { setting for multirow general options }
  TMultiRowsOptions = class(TPersistent)
  private
    FMultiRowFixedCol: integer;
    FMultiRowDataCol: integer;
    FObjectInnerMargin: integer;
  public
    constructor Create;
  published
    { index of column used as fixed column for multirow mode }
    property MultiRowFixedCol: integer read FMultiRowFixedCol write FMultiRowFixedCol default - 1;
    { index of column used as data column for multirow mode }
    property MultiRowDataCol: integer read FMultiRowDataCol write FMultiRowDataCol default - 1;
    { define additional inner space for objects shown in the cell }
    property ObjectInnerMargin: integer read FObjectInnerMargin write FObjectInnerMargin default 1;
  end;

  { multirow cell type }
  TMultiRowType = (mrtAuto, mrtPlainText, mrtHTML, mrtRTF, mrtColumns, mrtColumnsBlock, mrtImage, mrtObjects, mrtFileObjects);

  { collection item containing all multirows properties }
  TrDBStringGridEdMultiRow = class(TCollectionItem)
  private
    FMultiRowType: TMultiRowType;
    FAlignment: TAlignment;
    FColor: TColor;
    FFont: TFont;
    FParentFont: boolean;
    FCustomDraw: boolean;
    FRowHeightMin: integer;
    FRowHeightMax: integer;
    FRowHeightFreeSpace: integer;
    FWordWrap: boolean;
    FShowFixed: boolean;
    FReadOnly: boolean;
    FFieldName: string;
    FTextMask: TStrings;
    FFixedColTextMask: TStrings;
    FTag: integer;
    procedure SetMultiRowType(Value: TMultiRowType);
    procedure SetAlignment(Value: TAlignment);
    procedure SetColor(Value: TColor);
    procedure SetFont(Value: TFont);
    function IsFontStored: boolean;
    procedure FontChanged(Sender: TObject);
    procedure SetParentFont(Value: boolean);
    procedure SetRowHeightMin(Value: integer);
    procedure SetRowHeightMax(Value: integer);
    procedure SetRowHeightFreeSpace(Value: integer);
    procedure SetWordWrap(Value: boolean);
    procedure SetShowFixed(Value: boolean);
    procedure SetReadOnly(Value: boolean);
    procedure SetFieldName(Value: string);
    procedure SetTextMask(Value: TStrings);
    procedure FTextMaskChanged(Sender: TObject);
    procedure SetFixedColTextMask(Value: TStrings);
    procedure FFixedColTextMaskChanged(Sender: TObject);
  protected
    function GetGrid: TrDBStringGridEd;
  public
    constructor Create(Collection: TCollection); override; {<@exclude }
    destructor Destroy; override; {<@exclude }
    procedure Assign(Source: TPersistent); override;
    property Grid: TrDBStringGridEd read GetGrid; {<@exclude }
    { restore default properties of the multirow }
    procedure RestoreDefaults;
  published
    { defined type of multirow content }
    property MultiRowType: TMultiRowType read FMultiRowType write SetMultiRowType default mrtAuto;
    { define alignment of data in the multirow datacell }
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    { define background color of the multirow datacell }
    property Color: TColor read FColor write SetColor default clNone;
    { define font for the multirow datacell }
    property Font: TFont read FFont write SetFont stored IsFontStored;
    { define usage of parent font for the multirow datacell }
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    { define whether the multirow datacell is drawn by custom code in OnDrawCell event }
    property CustomDraw: boolean read FCustomDraw write FCustomDraw default false;
    { define minimal row height }
    property RowHeightMin: integer read FRowHeightMin write SetRowHeightMin default 0;
    { define maximal row height }
    property RowHeightMax: integer read FRowHeightMax write SetRowHeightMax default 0;
    { define additional row height as free space }
    property RowHeightFreeSpace: integer read FRowHeightFreeSpace write SetRowHeightFreeSpace default 0;
    { define whether long string should be wraped in teh multirow datacell }
    property WordWrap: boolean read FWordWrap write SetWordWrap default false;
    { define whether the multirow datacell is shown as fixed }
    property ShowFixed: boolean read FShowFixed write SetShowFixed default false;
    { define that multirow datacell cannot be modified }
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
    { define FieldName as content for the multirow datacell }
    property FieldName: string read FFieldName write SetFieldName;
    { define Text mask as content for the multirow datacell, tags <<FieldName>> can be used to insert dataset value }
    property TextMask: TStrings read FTextMask write SetTextMask;
    { define Text mask as content for the multirow fixedcell, tags <<FieldName>> can be used to insert dataset value }
    property FixedColTextMask: TStrings read FFixedColTextMask write SetFixedColTextMask;
    { tag property for any custom value }
    property Tag: integer read FTag write FTag default 0;
  end;

  { collection of TrDBStringGridEdMultiRow used by TrStringGridEd }
  TrDBStringGridEdMultiRows = class(TCollection)
  private
    FGrid: TrDBStringGridEd;
    function GetMultiRow(Index: integer): TrDBStringGridEdMultiRow;
    procedure SetMultiRow(Index: integer; Value: TrDBStringGridEdMultiRow);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(Grid: TrDBStringGridEd; MultiRowClass: TrDBStringGridEdMultiRowClass);
    function Add: TrDBStringGridEdMultiRow;
    procedure RestoreDefaults;
    property Grid: TrDBStringGridEd read FGrid;
    property Items[Index: integer]: TrDBStringGridEdMultiRow read GetMultiRow write SetMultiRow; default;
  end;

  { event type used by TrDBStringGridEd before post }
  TDBBeforePostEvent = procedure(Sender: TObject; Dataset: TDataset; ARow, ACol: Longint; KeyValue: integer) of object;
  { event type used by TrDBStringGridEd to load and store field value }
  TDBLoadStoreCellField = procedure(Sender: TObject; Dataset: TDataset; ARow, ACol: Longint; F: Tfield) of object;
  { event type used by TrDBStringGridEd to load and store field value }
  TDBUpdateCellContent = procedure(Sender: TObject; Dataset: TDataset; ARow, ACol: Longint; MultiRowDef: TrDBStringGridEdMultiRow; var CellText: string)
    of object;

  { enhanced TrStringGridEd component which provides functions
    for batch loading and saving data from/to dataset.
    It helps to use simple table, multitable or pivot table data showing and editing }
  TrDBStringGridEd = class(TrStringGridEd)
  private
    FMultiRowsOptions: TMultiRowsOptions;
    FMultiRowsDef: TrDBStringGridEdMultiRows;
    FMultiRowsPreview: boolean;
    FOptionsDBEx: TOptionsDBEx;
    FKeyFieldName: string;
    FKeyValue: integer;
    FCreateTableCol: boolean;
    FWarningList: TStrings;
    FRecsUpdatedCount: Longint;
    FCellsUpdatedCount: Longint;
    FDBUpdateCellContent: TDBUpdateCellContent;
    FDBLoadCellFromField: TDBLoadStoreCellField;
    FDBStoreCellToField: TDBLoadStoreCellField;
    FDBBeforePost: TDBBeforePostEvent;
    FDBUpdateError: TUpdateErrorEvent;
    procedure SetMultiRowsPreview(Value: boolean);
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure SetCreateTableCol(Value: boolean);
    function GetRowKeyValue(ARow: Longint): integer;
    procedure SetRowKeyValue(ARow: Longint; const Value: integer);
    function GetColKeyValue(ACol: Longint): integer;
    procedure SetColKeyValue(ACol: Longint; const Value: integer);
    procedure UpdateState(Sender: TObject);
    function HandleError(DS: TDataset; E: EDatabaseError; UpdateKind: TUpdateKind): TUpdateAction;
    procedure TestObjectInCell(ACol, ARow: Longint; DblClick: boolean);
  protected
    FDataLink: TrDataLink;
    FIsTableMultiRows: boolean;
    function CanEditModify: boolean; override;
    function CanEditShow: boolean; override;
    procedure CellClick(ACol, ARow: Longint); override;
    procedure DblClick; override;
    procedure DoFinishEditing(ACol, ARow: Longint; Forced: boolean); override;
    procedure DrawCellEx(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState; BkColor: TColor; DrawFocus: boolean); override;
    function GetCellWordWrap(ACol, ARow: Longint): boolean; override;
    procedure DrawColumnsInCell(R: TRect; s: string; Block: boolean; Align: TAlignment; Wrap: boolean);
    procedure DrawObjectsInCell(var R: TRect; ACol, ARow: Longint; CalcSize, TestObjectClick, DblClick: boolean);
    procedure ResetUpdatedCount;
    procedure LoadCellFromField(Dataset: TDataset; ACol, ARow: Longint; F: Tfield);
    function StoreCellToField(Dataset: TDataset; ACol, ARow: Longint; F: Tfield): boolean;
    function PostRowData(Dataset: TDataset; KeyField: Tfield; UpdateKind: TUpdateKind; ARow: Longint; ColumnGroup, Action: integer): boolean;
    procedure TryPostRowData(Dataset: TDataset; KeyField: Tfield; UpdateKind: TUpdateKind; ARow: Longint; ColumnGroup, Action: integer);
    function CalcCellHeight(ACol, ARow: Longint; UseLimits: boolean): integer; override;
    function EvaluateMultiRowType(CurrentType: TMultiRowType; Text: string): TMultiRowType;
    function GetActiveCellMultiRowType: TMultiRowType;
    procedure ShowMultiRowPreview;
    procedure LoadMultiRowRecInt(StartRow: Longint; NewRec: boolean; KeyFld: TField = nil; DS: TDataset = nil);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }

    { return edit style of data in cell }
    function GetEditStyleEx(ACol, ARow: Longint): TrStringGridEdEditStyle; override;
    { define grid columns properties from Field list of linked datasource, only visible and non graphics fields are used }
    procedure CreateTableColumns(SkipFixedCol, SetColWidth: boolean; TitleAlignment: TAlignment = taCenter);
    { reset grid and load all rows from linked datasource to grid, KeyField value is used to keep ID for each record }
    procedure LoadTableRows;
    { post all changes to linked dataset, KeyField value is used to locate modified record }
    function SaveTableRows(SaveNewRec: boolean = true; DeleteEmptyRec: boolean = true): Longint;

    { use defined columns+rows and load data from Dataset to grid,
      KeyField value is used to locate row for each record, Column FieldName is used to locate column for each field }
    procedure LoadTableData(Dataset: TDataset; KeyField: Tfield; ColumnGroup: integer = 0; ClearCurrentData: boolean = true; LoadAsNewValues: boolean = false);
    { post all changes to Dataset,
      KeyField value is used to locate row for each record, Column FieldName is used to locate column for each field }
    function SaveTableData(Dataset: TDataset; KeyField: Tfield; ColumnGroup: integer = 0; DeleteEmptyRec: boolean = true): integer;

    { define grid columns properties from records of Dataset, for each record one Column is created,
      ColumnKeyField value is used to keep ID for each column record }
    procedure LoadPivotColumn(Dataset: TDataset; ColumnKeyField, ColumnDisplayField: Tfield; FromColumnIdx: Longint; ColumnGroup: integer = 0);
    { load data from Dataset to defined Pivot table,
      RowKeyField value is used to locate row for each record, ColumnKeyField value is used to locate column for each record,
      ValueField value is stored to cell }
    procedure LoadPivotData(Dataset: TDataset; RowKeyField, ColumnKeyField, ValueField: Tfield; ColumnGroup: integer = 0; ClearCurrentData: boolean = true;
      LoadAsNewValues: boolean = false);
    { post all changes to Dataset,
      RowKeyField and ColumnKeyField value is used to locate/create of record for each changed cell,
      ValueField is updated from cell }
    function SavePivotData(Dataset: TDataset; RowKeyField, ColumnKeyField, ValueField: Tfield; ColumnGroup: integer = 0; DeleteEmptyRec: boolean = true)
      : Longint;

    { load multirows for each record from linked datasource to grid,
      KeyField value is used to keep ID for each multirow group }
    procedure LoadTableMultiRows;
    { post all multirow changes to linked dataset, KeyField value is used to locate modified record }
    function SaveTableMultiRows(SaveNewRec: boolean = true): Longint;
    { add or insert new empty rows for a new multirow record }
    procedure AddNewRecMultiRows(InsertToTop: boolean = false; InsertInCurrentPos: boolean = false);

    { define whether LoadTableMultiRows was used to load data }
    function IsTableMultiRows: boolean;
    { set row heights for all multirows }
    procedure RefreshMultiRowHeights;
    { returns multirow definition for the specific row }
    function GetMultiRowDef(ARow: Longint): TrDBStringGridEdMultiRow;
    { returns multirow definition for the specific cell }
    function GetCellMultiRowDef(ACol, ARow: Longint): TrDBStringGridEdMultiRow;
    { returns multirow content type for the cell }
    function GetCellMultiRowType(ACol, ARow: Longint): TMultiRowType;
    { returns column index of the column used as multirow fixed col }
    function GetMultiRowFixedCol: integer;
    { returns column index of the column used as multirow data col }
    function GetMultiRowDataCol: integer;

    { define number of modified rows }
    property RecsUpdatedCount: Longint read FRecsUpdatedCount;
    { define number of modified cells }
    property CellsUpdatedCount: Longint read FCellsUpdatedCount;
    { define list of warnings during loading data }
    property WarningList: TStrings read FWarningList;

    { get or set Key Value for defined Row }
    property RowKeyValue[ARow: Longint]: integer read GetRowKeyValue write SetRowKeyValue;
    { get or set Key Value for defined Column }
    property ColKeyValue[ACol: Longint]: integer read GetColKeyValue write SetColKeyValue;
    { return index of row with defined key value }
    function IndexOfRowKeyValue(KeyValue: integer): Longint;
    { return index of column with defined key value }
    function IndexOfColKeyValue(KeyValue: integer): Longint;
  published
    { multirow options }
    property MultiRowsOptions: TMultiRowsOptions read FMultiRowsOptions write FMultiRowsOptions;
    { multirow collection, for detail see TrDBStringGridEdMultiRows }
    property MultiRowsDef: TrDBStringGridEdMultiRows read FMultiRowsDef write FMultiRowsDef;
    { defines preview of multirow definition in the design time }
    property MultiRowsPreview: boolean read FMultiRowsPreview write SetMultiRowsPreview default false;
    { list of new options }
    property OptionsDBEx: TOptionsDBEx read FOptionsDBEx write FOptionsDBEx default[dgSetAutoIncGrayed, dgSaveKeyFieldValue];
    { link to connected dataset }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { define name of KeyField which contain ID of each record (should be integer value) }
    property KeyFieldName: string read FKeyFieldName write FKeyFieldName;
    { used in design time to load column definition from linked datasource }
    property CreateTableColNow: boolean read FCreateTableCol write SetCreateTableCol default false;
    { event type used by TrDBStringGridEd to load and store field value for multirow cell }
    property DBUpdateCellContent: TDBUpdateCellContent read FDBUpdateCellContent write FDBUpdateCellContent;
    { event used for user defined loading data from field to cell }
    property DBLoadCellFromField: TDBLoadStoreCellField read FDBLoadCellFromField write FDBLoadCellFromField;
    { event used for user defined storing data from cell to field }
    property DBStoreCellToField: TDBLoadStoreCellField read FDBStoreCellToField write FDBStoreCellToField;
    { event used for modification of dataset before post changes }
    property DBBeforePostEvent: TDBBeforePostEvent read FDBBeforePost write FDBBeforePost;
    { event for error handling }
    property OnDBUpdateError: TUpdateErrorEvent read FDBUpdateError write FDBUpdateError;
  end;

procedure Register; {< @exclude }

implementation

uses rstring, rtool, rdbtool, rpictool, rfiletool, rHTMLDrawText;

const
  KeyFieldNotFoundMsg = 'KeyField is not defined. Update is not possible.';
  WarningMsg = 'KeyValue not found: %d';
  cObjInnerMargin = 3;
  cObjSpace = 6;
  cObjIconSize = 16;

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBStringGridEd]);
end;

// =============================================================================
// TMultiRow
constructor TMultiRowsOptions.Create;
begin
  FMultiRowFixedCol:=-1;
  FMultiRowDataCol:=-1;
  FObjectInnerMargin:=1;
end;

constructor TrDBStringGridEdMultiRow.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FFont:=TFont.Create;
  FFont.OnChange:=FontChanged;
  FParentFont:=true;
  FTextMask:=TStringList.Create;
  TStringList(FTextMask).OnChange:=FTextMaskChanged;
  FFixedColTextMask:=TStringList.Create;
  TStringList(FFixedColTextMask).OnChange:=FFixedColTextMaskChanged;
  FColor:=clNone;
  FCustomDraw:=false;
  RestoreDefaults;
end;

destructor TrDBStringGridEdMultiRow.Destroy;
begin
  FreeAndNil(FFixedColTextMask);
  FreeAndNil(FTextMask);
  FreeAndNil(FFont);
  inherited Destroy;
end;

procedure TrDBStringGridEdMultiRow.Assign(Source: TPersistent);
var
  s: TrDBStringGridEdMultiRow;
begin
  if Assigned(Source) and (Source is TrDBStringGridEdMultiRow) then
  begin
    s:=Source as TrDBStringGridEdMultiRow;
    FMultiRowType:=s.FMultiRowType;
    FAlignment:=s.FAlignment;
    FColor:=s.FColor;
    FFont.Assign(s.Font);
    ParentFont:=s.FParentFont;
    FCustomDraw:=s.FCustomDraw;
    FRowHeightMin:=s.FRowHeightMin;
    FRowHeightMax:=s.FRowHeightMax;
    FRowHeightFreeSpace:=s.RowHeightFreeSpace;
    FWordWrap:=s.FWordWrap;
    FShowFixed:=s.FShowFixed;
    FReadOnly:=s.FReadOnly;
    FFieldName:=s.FFieldName;
    FTextMask:=s.FTextMask;
    FFixedColTextMask:=s.FFixedColTextMask;
    FTag:=s.FTag;
  end;
end;

procedure TrDBStringGridEdMultiRow.RestoreDefaults;
begin
  // set default values for one Column
  FMultiRowType:=mrtAuto;
  FAlignment:=taLeftJustify;
  FColor:=clNone;
  ParentFont:=true;
  FCustomDraw:=false;
  FRowHeightMin:=0;
  FRowHeightMax:=0;
  FRowHeightFreeSpace:=0;
  FWordWrap:=false;
  FShowFixed:=false;
  FReadOnly:=false;
  // FFieldName:='';
  // FTextMask:='';
  // FFixedColTextMask:='';
  FTag:=0;
  Changed(false);
end;

function TrDBStringGridEdMultiRow.GetGrid: TrDBStringGridEd;
begin
  if Assigned(Collection) and (Collection is TrDBStringGridEdMultiRows) then
    Result:=TrDBStringGridEdMultiRows(Collection).Grid
  else
    Result:=nil;
end;

procedure TrDBStringGridEdMultiRow.SetMultiRowType(Value: TMultiRowType);
begin
  FMultiRowType:=Value;
  Grid.ShowMultiRowPreview;
end;

procedure TrDBStringGridEdMultiRow.SetAlignment(Value: TAlignment);
begin
  if (Value = FAlignment) then
    Exit;
  FAlignment:=Value;
  Changed(false);
end;

procedure TrDBStringGridEdMultiRow.SetColor(Value: TColor);
begin
  if (Value = FColor) then
    Exit;
  FColor:=Value;
  Changed(false);
end;

procedure TrDBStringGridEdMultiRow.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  Changed(false);
end;

procedure TrDBStringGridEdMultiRow.SetParentFont(Value: boolean);
begin
  if Value then
    FFont.Assign(Grid.Font);
  FParentFont:=Value;
end;

function TrDBStringGridEdMultiRow.IsFontStored: boolean;
begin
  Result:=not FParentFont;
end;

procedure TrDBStringGridEdMultiRow.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  Grid.RefreshMultiRowHeights;
end;

procedure TrDBStringGridEdMultiRow.SetRowHeightMin(Value: integer);
begin
  FRowHeightMin:=Value;
  Grid.RefreshMultiRowHeights;
end;

procedure TrDBStringGridEdMultiRow.SetRowHeightMax(Value: integer);
begin
  FRowHeightMax:=Value;
  Grid.RefreshMultiRowHeights;
end;

procedure TrDBStringGridEdMultiRow.SetRowHeightFreeSpace(Value: integer);
begin
  FRowHeightFreeSpace:=Value;
  Grid.RefreshMultiRowHeights;
end;

procedure TrDBStringGridEdMultiRow.SetWordWrap(Value: boolean);
begin
  if (Value = FWordWrap) then
    Exit;
  FWordWrap:=Value;
  Grid.RefreshMultiRowHeights;
end;

procedure TrDBStringGridEdMultiRow.SetShowFixed(Value: boolean);
begin
  if (Value = FShowFixed) then
    Exit;
  FShowFixed:=Value;
  Changed(false);
end;

procedure TrDBStringGridEdMultiRow.SetReadOnly(Value: boolean);
begin
  if (Value = FReadOnly) then
    Exit;
  FReadOnly:=Value;
end;

procedure TrDBStringGridEdMultiRow.SetFieldName(Value: string);
begin
  if (Value = FFieldName) then
    Exit;
  FFieldName:=Value;
  Grid.ShowMultiRowPreview;
end;

procedure TrDBStringGridEdMultiRow.SetTextMask(Value: TStrings);
begin
  FTextMask.Assign(Value);
  Grid.ShowMultiRowPreview;
end;

procedure TrDBStringGridEdMultiRow.FTextMaskChanged(Sender: TObject);
begin
  Grid.ShowMultiRowPreview;
end;

procedure TrDBStringGridEdMultiRow.SetFixedColTextMask(Value: TStrings);
begin
  FFixedColTextMask.Assign(Value);
  Grid.ShowMultiRowPreview;
end;

procedure TrDBStringGridEdMultiRow.FFixedColTextMaskChanged(Sender: TObject);
begin
  Grid.ShowMultiRowPreview;
end;

// =============================================================================
// TMultiRows Collection
constructor TrDBStringGridEdMultiRows.Create(Grid: TrDBStringGridEd; MultiRowClass: TrDBStringGridEdMultiRowClass);
begin
  inherited Create(MultiRowClass);
  FGrid:=Grid;
end;

function TrDBStringGridEdMultiRows.Add: TrDBStringGridEdMultiRow;
begin
  Result:=TrDBStringGridEdMultiRow( inherited Add);
end;

function TrDBStringGridEdMultiRows.GetMultiRow(Index: integer): TrDBStringGridEdMultiRow;
begin
  Result:=TrDBStringGridEdMultiRow( inherited Items[Index]);
end;

procedure TrDBStringGridEdMultiRows.SetMultiRow(Index: integer; Value: TrDBStringGridEdMultiRow);
begin
  Items[Index].Assign(Value);
end;

function TrDBStringGridEdMultiRows.GetOwner: TPersistent;
begin
  Result:=FGrid;
end;

procedure TrDBStringGridEdMultiRows.RestoreDefaults;
var
  i: integer;
begin
  // set default for all columns
  BeginUpdate;
  try
    for i:=0 to Count - 1 do
      Items[i].RestoreDefaults;
  finally
    EndUpdate;
  end;
end;

procedure TrDBStringGridEdMultiRows.Update(Item: TCollectionItem);
begin
  // column props was updated => redraw
  Grid.ShowMultiRowPreview;
end;

// =============================================================================
// Grid
constructor TrDBStringGridEd.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, UpdateState);
  FWarningList:=TStringList.Create;
  // Create MultiRows
  FMultiRowsOptions:=TMultiRowsOptions.Create;
  MultiRowsDef:=TrDBStringGridEdMultiRows.Create(self, TrDBStringGridEdMultiRow);
  FOptionsDBEx:=[dgSetAutoIncGrayed, dgSaveKeyFieldValue];
  FIsTableMultiRows:=false;
end;

destructor TrDBStringGridEd.Destroy;
begin
  FreeAndNil(FWarningList);
  FreeAndNil(FMultiRowsDef);
  FMultiRowsOptions.Free;
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

function TrDBStringGridEd.GetEditStyleEx(ACol, ARow: Longint): TrStringGridEdEditStyle;
var
  mrDef: TrDBStringGridEdMultiRow;
  Fld: Tfield;
begin
  Result:=Columns[ACol].ButtonStyle;
  mrDef:=GetCellMultiRowDef(ACol, ARow);
  if Assigned(mrDef) then
  begin
    Fld:=nil;
    if Assigned(FDataLink.Dataset) then
      Fld:=FDataLink.Dataset.FindField(mrDef.FFieldName);
    case mrDef.FMultiRowType of
      mrtAuto, mrtPlainText:
        if Fld is TBooleanField then
          Result:=sgbCheckbox
        else
          if rdbtool.IsDateTimeField(Fld) then
            Result:=sgbDateTimePicker;
      mrtHTML:
        ;
      mrtRTF:
        Result:=sgbRichText;
      mrtColumns:
        ;
      mrtColumnsBlock:
        ;
      mrtImage:
        ;
      mrtObjects:
        ;
      mrtFileObjects:
        ;
    end;
  end;
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(self, ACol, ARow, Result);
end;

function TrDBStringGridEd.CanEditModify: boolean;
var
  mrDef: TrDBStringGridEdMultiRow;
begin
  Result:= inherited CanEditModify;
  if Result then
  begin
    mrDef:=GetCellMultiRowDef(Col, Row);
    if Assigned(mrDef) then
      if mrDef.FReadOnly or (mrDef.FMultiRowType in [mrtImage, mrtObjects, mrtFileObjects]) then
        Result:=false;
  end;
end;

function TrDBStringGridEd.CanEditShow: boolean;
var
  mrDef: TrDBStringGridEdMultiRow;
begin
  Result:= inherited CanEditShow;
  if Result then
  begin
    mrDef:=GetCellMultiRowDef(Col, Row);
    if Assigned(mrDef) then
      if mrDef.FReadOnly or (mrDef.FMultiRowType in [mrtImage, mrtObjects, mrtFileObjects]) then
        Result:=false;
    // hide editor if edit is disable - MUST!!! in other case active cell has strange content
    if not Result then
      HideEditor;
  end;
end;

procedure TrDBStringGridEd.TestObjectInCell(ACol, ARow: Longint; DblClick: boolean);
var
  mrDef: TrDBStringGridEdMultiRow;
  R: TRect;
begin
  mrDef:=GetCellMultiRowDef(ACol, ARow);
  if Assigned(mrDef) and (mrDef.FMultiRowType in [mrtObjects, mrtFileObjects]) then
  begin
    R:=CellRect(ACol, ARow);
    InflateRect(R, -1 - InnerMargin, -1);
    // set canvas format
    SetCanvasStyle(ACol, ARow, Canvas);
    if not mrDef.FParentFont then
      Canvas.Font.Assign(mrDef.FFont);
    DrawObjectsInCell(R, ACol, ARow, false, true, DblClick);
  end;
end;

procedure TrDBStringGridEd.CellClick(ACol, ARow: Longint);
begin
  inherited;
  TestObjectInCell(ACol, ARow, false);
end;

procedure TrDBStringGridEd.DblClick;
begin
  inherited;
  TestObjectInCell(Col, Row, true);
end;

procedure TrDBStringGridEd.DoFinishEditing(ACol, ARow: Longint; Forced: boolean);
begin
  inherited;
end;

procedure TrDBStringGridEd.SetMultiRowsPreview(Value: boolean);
begin
  if Value <> FMultiRowsPreview then
  begin
    FMultiRowsPreview:=Value;
    if FMultiRowsPreview then
      ShowMultiRowPreview
    else
    begin
      FIsTableMultiRows:=false;
      ClearEntireTable(true);
      RowCount:=FixedRows + 1;
    end;
    Invalidate;
  end;
end;

function TrDBStringGridEd.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBStringGridEd.SetDataSource(Value: TDataSource);
begin
  if Value <> FDataLink.DataSource then
  begin
    FDataLink.DataSource:=Value;
    UpdateState(nil);
  end;
end;

procedure TrDBStringGridEd.UpdateState(Sender: TObject);
begin
end;

procedure TrDBStringGridEd.SetCreateTableCol(Value: boolean);
begin
  if Value then
    CreateTableColumns(true, true);
end;

function TrDBStringGridEd.GetRowKeyValue(ARow: Longint): integer;
begin
  Result:=integer(Objects[0, ARow]);
end;

procedure TrDBStringGridEd.SetRowKeyValue(ARow: Longint; const Value: integer);
begin
  Objects[0, ARow]:=TObject(Value);
end;

function TrDBStringGridEd.GetColKeyValue(ACol: Longint): integer;
begin
  Result:=integer(Objects[ACol, 0]);
end;

procedure TrDBStringGridEd.SetColKeyValue(ACol: Longint; const Value: integer);
begin
  Objects[ACol, 0]:=TObject(Value);
end;

function TrDBStringGridEd.IndexOfRowKeyValue(KeyValue: integer): Longint;
begin
  for Result:=FixedRows to RowCount - 1 do
    if Objects[0, Result] = TObject(KeyValue) then
      Exit;
  Result:=-1;
end;

function TrDBStringGridEd.IndexOfColKeyValue(KeyValue: integer): Longint;
begin
  Result:=Rows[0].IndexOfObject(TObject(KeyValue));
end;

function TrDBStringGridEd.HandleError(DS: TDataset; E: EDatabaseError; UpdateKind: TUpdateKind): TUpdateAction;
begin
  Result:=uaFail;
  if Assigned(FDBUpdateError) then
    FDBUpdateError(DS, E, UpdateKind, Result);
  case Result of
    uaAbort:
      Abort;
    uaSkip:
      DS.Cancel;
    uaApplied:
      DS.Cancel;
  end;
end;

procedure TrDBStringGridEd.ResetUpdatedCount;
begin
  FRecsUpdatedCount:=0;
  FCellsUpdatedCount:=0;
end;

procedure TrDBStringGridEd.LoadCellFromField(Dataset: TDataset; ACol, ARow: Longint; F: Tfield);
begin
  if not Assigned(F) then // or not F.Visible then
    Exit;
  if Assigned(FDBLoadCellFromField) then
    FDBLoadCellFromField(self, Dataset, ARow, ACol, F)
  else
  begin
    if F.IsNull then
      Cells[ACol, ARow]:=''
    else
      if F is TNumericField then
        NumCells[ACol, ARow]:=F.AsFloat
      else
        if F is TBooleanField then
          BoolCells[ACol, ARow]:=F.AsBoolean
        else
          StringCells[ACol, ARow]:=F.AsString;
  end;
end;

function TrDBStringGridEd.StoreCellToField(Dataset: TDataset; ACol, ARow: Longint; F: Tfield): boolean;
begin
  Result:=false;
  if Assigned(F) and (ChangedCells.IndexOf(GetChangedCellStr(ACol, ARow)) >= 0) then
  begin
    if Assigned(FDBStoreCellToField) then
      FDBStoreCellToField(self, Dataset, ARow, ACol, F)
    else
    begin
      if Trim(Cells[ACol, ARow]) = '' then
        F.Clear
      else
      begin
        if F is TNumericField then
          F.AsFloat:=NumCells[ACol, ARow]
        else
          if F is TBooleanField then
            F.AsBoolean:=BoolCells[ACol, ARow]
          else
            F.AsString:=Cells[ACol, ARow];
      end;
    end;
    Result:=true;
  end;
end;

function TrDBStringGridEd.PostRowData(Dataset: TDataset; KeyField: Tfield; UpdateKind: TUpdateKind; ARow: Longint; ColumnGroup, Action: integer): boolean;
var
  c, CelUpdTemp: Longint;
  F: Tfield;
begin
  // Action values: 0 = do all; 1 = start modify only; 2 = post only
  Result:=true;
  CelUpdTemp:=0;
  try
    if UpdateKind = ukDelete then
    begin
      Dataset.Delete;
      Inc(FRecsUpdatedCount);
    end
    else
    begin
      if Action in [0, 1] then
      begin
        case UpdateKind of
          ukModify:
            Dataset.Edit;
          ukInsert:
            begin
              Dataset.Append;
              if (dgSaveKeyFieldValue in FOptionsDBEx) and Assigned(KeyField) and not KeyField.ReadOnly then
                KeyField.AsInteger:=GetRowKeyValue(ARow);
            end;
        end;
      end;
      if Action in [0] then
      begin
        for c:=0 to ColCount - 1 do
        begin
          F:=Dataset.FindField(Columns[c].FieldName);
          if (ColumnGroup = 0) or (ColumnGroup = Columns[c].ColumnGroup) then
            if StoreCellToField(Dataset, c, ARow, F) then
              Inc(CelUpdTemp);
        end;
      end;
      if Action in [0, 2] then
      begin
        if Assigned(FDBBeforePost) then
          FDBBeforePost(self, Dataset, ARow, -1, FKeyValue);
        Dataset.Post;
        // store New KeyValue
        if (UpdateKind = ukInsert) and Assigned(KeyField) then
          try
            RowKeyValue[ARow]:=KeyField.AsInteger;
          except
          end;

        Inc(FRecsUpdatedCount);
        Inc(FCellsUpdatedCount, CelUpdTemp);
      end;
    end;
  except
    on E: EDatabaseError do
    begin
      case HandleError(Dataset, E, UpdateKind) of
        uaFail:
          raise ;
        uaRetry:
          Result:=false;
      end;
    end
    else
      raise ;
  end;
end;

procedure TrDBStringGridEd.TryPostRowData(Dataset: TDataset; KeyField: Tfield; UpdateKind: TUpdateKind; ARow: Longint; ColumnGroup, Action: integer);
begin
  repeat
  until PostRowData(Dataset, KeyField, UpdateKind, ARow, ColumnGroup, Action);
end;

procedure TrDBStringGridEd.CreateTableColumns(SkipFixedCol, SetColWidth: boolean; TitleAlignment: TAlignment = taCenter);
var
  i, c, w: Longint;
  F: Tfield;
begin
  BeginUpdate;
  try
    if Assigned(FDataLink.Dataset) then
    begin
      ColCount:=FixedCols + FDataLink.Dataset.FieldCount + 1;
      // set first col idx
      if SkipFixedCol then
        c:=FixedCols
      else
        c:=0;

      for i:=0 to FDataLink.Dataset.FieldCount - 1 do
      begin
        F:=FDataLink.Dataset.Fields[i];
        if not F.Visible or rdbtool.IsGraphicField(F) then
          Continue;
        // set column props
        Columns[c].RestoreDefaults;
        Columns[c].TitleCaption:=F.DisplayLabel;
        Columns[c].FieldName:=F.FieldName;
        Columns[c].TitleAlignment:=TitleAlignment;
        if SetColWidth then
        begin
          Canvas.Font:=Columns[c].TitleFont;
          w:=F.DisplayWidth * Canvas.TextWidth('a');
          w:=Max(50, Min(w, 200));
          ColWidths[c]:=w;
        end;
        Columns[c].TitleWordWrap:=true;

        case F.DataType of
          ftString, ftWideString:
            begin
              Columns[c].DataType:=sgdText;
              Columns[c].Alignment:=taLeftJustify;
            end;
          ftSmallint, ftInteger, ftWord, ftBCD, ftLargeint, ftLongWord, ftShortint, ftByte, ftAutoInc:
            begin
              Columns[c].DataType:=sgdInteger;
              Columns[c].Alignment:=taRightJustify;
            end;
          ftFloat, ftCurrency, ftExtended, ftSingle:
            begin
              Columns[c].DataType:=sgdNumber;
              Columns[c].Alignment:=taRightJustify;
            end;
          ftBoolean:
            Columns[c].ButtonStyle:=sgbCheckbox;
          ftDate, ftDateTime, ftTimeStamp:
            begin
              Columns[c].DataType:=sgdDate;
              Columns[c].ButtonStyle:=sgbDateTimePicker;
            end;
          ftTime:
            Columns[c].DataType:=sgdTime;
          ftMemo, ftFmtMemo, ftWideMemo:
            Columns[c].WordWrap:=true;
        end;
        if F.DataType = ftAutoInc then
        begin
          if dgSetAutoIncReadOnly in FOptionsDBEx then
            Columns[c].ReadOnly:=true;
          if dgSetAutoIncGrayed in FOptionsDBEx then
            Columns[c].Font.Color:=clGrayText;
        end;
        if F is TNumericField then
          Columns[c].NumFormat:=TNumericField(F).DisplayFormat;

        Inc(c);
      end;
      ColCount:=c;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TrDBStringGridEd.LoadTableRows;
var
  R, c: Longint;
  F: Tfield;
begin
  ClearEntireTable(true);
  SelectCell(FixedCols, FixedRows);
  if not FDataLink.Active then
    Exit;
  FDataLink.Dataset.Cancel;
  R:=FDataLink.Dataset.RecordCount;
  if R = 0 then
    Inc(R);
  RowCount:=FixedRows + R;
  FDataLink.Dataset.DisableControls;
  BeginUpdate;
  try
    R:=FixedRows;
    FDataLink.Dataset.First;
    while not FDataLink.Dataset.Eof do
    begin
      F:=FDataLink.Dataset.FindField(KeyFieldName);
      if Assigned(F) then
        RowKeyValue[R]:=F.AsInteger;
      for c:=0 to ColCount - 1 do
      begin
        F:=FDataLink.Dataset.FindField(Columns[c].FieldName);
        LoadCellFromField(FDataLink.Dataset, c, R, F);
      end;
      FDataLink.Dataset.Next;
      Inc(R);
    end;
  finally
    EndUpdate;
    FDataLink.Dataset.EnableControls;
  end;
  ResetChangeLog;
end;

function TrDBStringGridEd.SaveTableRows(SaveNewRec: boolean = true; DeleteEmptyRec: boolean = true): Longint;
var
  i, R: Longint;
  KeyF: Tfield;
begin
  Result:=0;
  ResetUpdatedCount;
  if not FDataLink.Active then
    Exit;
  FDataLink.Dataset.Cancel;
  KeyF:=FDataLink.Dataset.FindField(KeyFieldName);
  if not Assigned(KeyF) then
    raise Exception.Create(KeyFieldNotFoundMsg);

  for i:=0 to self.ChangedRows.Count - 1 do
  begin
    R:=StrToInt(ChangedRows[i]);
    FKeyValue:=RowKeyValue[R];

    if FKeyValue = 0 then
    begin
      if SaveNewRec and not IsRowEmpty(R) then
        TryPostRowData(FDataLink.Dataset, KeyF, ukInsert, R, 0, 0);
    end
    else
      if LocateRecord(KeyF, FKeyValue) then
      begin
        if not IsRowEmpty(R) then
          TryPostRowData(FDataLink.Dataset, KeyF, ukModify, R, 0, 0)
        else
        begin
          if DeleteEmptyRec then
            TryPostRowData(FDataLink.Dataset, KeyF, ukDelete, R, 0, 0);
        end;
      end
      else
        raise Exception.Create(Format(WarningMsg, [FKeyValue]));
  end;
  ResetChangeLog;
  Result:=FRecsUpdatedCount;
end;

procedure TrDBStringGridEd.LoadTableData(Dataset: TDataset; KeyField: Tfield; ColumnGroup: integer = 0; ClearCurrentData: boolean = true;
  LoadAsNewValues: boolean = false);
var
  R, c: Longint;
  F: Tfield;
begin
  if not Dataset.Active then
    Exit;
  Dataset.DisableControls;
  BeginUpdate(LoadAsNewValues);
  try
    if ClearCurrentData then
      for c:=0 to ColCount - 1 do
        if (ColumnGroup = Columns[c].ColumnGroup) then
          for R:=FixedRows to RowCount - 1 do
            Cells[c, R]:='';

    FWarningList.Clear;
    Dataset.First;
    while not Dataset.Eof do
    begin
      R:=IndexOfRowKeyValue(KeyField.AsInteger);
      if R < FixedRows then
        FWarningList.Add(Format(WarningMsg, [KeyField.AsInteger]))
      else
      begin
        for c:=0 to ColCount - 1 do
        begin
          F:=Dataset.FindField(Columns[c].FieldName);
          if (ColumnGroup = 0) or (ColumnGroup = Columns[c].ColumnGroup) then
            LoadCellFromField(Dataset, c, R, F);
        end;
      end;
      Dataset.Next;
    end;
  finally
    EndUpdate;
    Dataset.EnableControls;
  end;
  if not LoadAsNewValues then
    ResetChangeLog;
end;

function TrDBStringGridEd.SaveTableData(Dataset: TDataset; KeyField: Tfield; ColumnGroup: integer = 0; DeleteEmptyRec: boolean = true): Longint;
var
  i, R: Longint;
  RowEmpty: boolean;
begin
  Result:=0;
  ResetUpdatedCount;
  if not Dataset.Active then
    Exit;
  Dataset.Cancel;
  // if not Assigned(KeyF) then
  // raise Exception.Create('KeyField is not defined. Update is not possible.');

  for i:=0 to self.ChangedRows.Count - 1 do
  begin
    R:=StrToInt(ChangedRows[i]);
    FKeyValue:=RowKeyValue[R];
    RowEmpty:=IsRowEmpty(R, true, ColumnGroup);

    if LocateRecord(KeyField, FKeyValue) then
    begin
      if not RowEmpty then
        TryPostRowData(Dataset, KeyField, ukModify, R, ColumnGroup, 0)
      else
      begin
        if DeleteEmptyRec then
          TryPostRowData(Dataset, KeyField, ukDelete, R, ColumnGroup, 0);
      end;
    end
    else
      // add new rec
      if not RowEmpty then
        TryPostRowData(Dataset, KeyField, ukInsert, R, ColumnGroup, 0);
  end;
  ResetChangeLog;
  Result:=FRecsUpdatedCount;
end;

procedure TrDBStringGridEd.LoadPivotColumn(Dataset: TDataset; ColumnKeyField, ColumnDisplayField: Tfield; FromColumnIdx: Longint; ColumnGroup: integer = 0);
var
  c: Longint;
begin
  if not Dataset.Active then
    Exit;
  Dataset.DisableControls;
  BeginUpdate;
  try
    ColCount:=FromColumnIdx + Dataset.RecordCount;
    c:=FromColumnIdx;
    Dataset.First;
    while not Dataset.Eof do
    begin
      if c > FromColumnIdx then
        Columns[c].Assign(Columns[FromColumnIdx]);
      ColKeyValue[c]:=ColumnKeyField.AsInteger;
      Columns[c].TitleCaption:=ColumnDisplayField.DisplayText;
      Columns[c].ColumnGroup:=ColumnGroup;
      Dataset.Next;
      Inc(c);
    end;
  finally
    EndUpdate;
    Dataset.EnableControls;
  end;
  ResetChangeLog;
end;

procedure TrDBStringGridEd.LoadPivotData(Dataset: TDataset; RowKeyField, ColumnKeyField, ValueField: Tfield; ColumnGroup: integer = 0;
  ClearCurrentData: boolean = true; LoadAsNewValues: boolean = false);
var
  c, R, i: Longint;
begin
  if not Dataset.Active then
    Exit;
  Dataset.DisableControls;
  BeginUpdate(LoadAsNewValues);
  try
    // clear current data
    if ClearCurrentData then
      for c:=0 to ColCount - 1 do
        if (ColumnGroup = Columns[c].ColumnGroup) then
          for R:=FixedRows to RowCount - 1 do
            Cells[c, R]:='';

    FWarningList.Clear;
    Dataset.First;
    while not Dataset.Eof do
    begin
      R:=self.IndexOfRowKeyValue(RowKeyField.AsInteger);
      c:=-1;
      for i:=0 to ColCount - 1 do
        if (ColumnGroup = Columns[i].ColumnGroup) and (ColKeyValue[i] = ColumnKeyField.AsInteger) then
        begin
          c:=i;
          Break;
        end;
      if R < 0 then
        FWarningList.Add(Format(WarningMsg, [RowKeyField.AsInteger]))
      else
        if c < 0 then
          FWarningList.Add(Format(WarningMsg, [ColumnKeyField.AsInteger]))
        else
          LoadCellFromField(Dataset, c, R, ValueField);
      Dataset.Next;
    end;
  finally
    EndUpdate;
    Dataset.EnableControls;
  end;
  if not LoadAsNewValues then
    ResetChangeLog;
end;

function TrDBStringGridEd.SavePivotData(Dataset: TDataset; RowKeyField, ColumnKeyField, ValueField: Tfield; ColumnGroup: integer = 0;
  DeleteEmptyRec: boolean = true): Longint;
var
  i, c, R: Longint;
  RowKey, ColKey: integer;
  Add, RowEmpty: boolean;
begin
  Result:=0;
  ResetUpdatedCount;
  if not Dataset.Active then
    Exit;
  Dataset.Cancel;
  // if not Assigned(KeyF) then
  // raise Exception.Create('KeyField is not defined. Update is not possible.');

  for i:=0 to self.ChangedCells.Count - 1 do
  begin
    c:=StrToInt(rtool.FindStrPart(ChangedCells[i], ';', 1));
    R:=StrToInt(rtool.FindStrPart(ChangedCells[i], ';', 2));

    RowKey:=self.RowKeyValue[R];
    ColKey:=self.ColKeyValue[c];
    RowEmpty:=Trim(Cells[c, R]) = '';

    Add:=true;
    Dataset.First;
    while not Dataset.Eof do
    begin
      if (RowKeyField.AsInteger = RowKey) and (ColumnKeyField.AsInteger = ColKey) then
      begin
        Add:=false;
        if RowEmpty then
        begin
          if DeleteEmptyRec then
            TryPostRowData(Dataset, nil, ukDelete, R, ColumnGroup, 0)
        end
        else
        begin
          TryPostRowData(Dataset, nil, ukModify, R, ColumnGroup, 1);
          StoreCellToField(Dataset, c, R, ValueField);
          TryPostRowData(Dataset, nil, ukModify, R, ColumnGroup, 2);
        end;
        Break;
      end;
      Dataset.Next;
    end;

    // add new rec
    if Add and not RowEmpty then
    begin
      TryPostRowData(Dataset, nil, ukInsert, R, ColumnGroup, 1);
      RowKeyField.AsInteger:=RowKey;
      ColumnKeyField.AsInteger:=ColKey;
      StoreCellToField(Dataset, c, R, ValueField);
      TryPostRowData(Dataset, nil, ukInsert, R, ColumnGroup, 2);
    end;
  end;

  ResetChangeLog;
  Result:=FRecsUpdatedCount;
end;

// =============================================================================
// MultiRow
function TrDBStringGridEd.GetCellWordWrap(ACol, ARow: Longint): boolean;
var
  mrDef: TrDBStringGridEdMultiRow;
begin
  // inherited result
  Result:= inherited GetCellWordWrap(ACol, ARow);
  // result for multirow
  mrDef:=GetCellMultiRowDef(ACol, ARow);
  if Assigned(mrDef) then
    Result:=mrDef.FWordWrap;
end;

procedure TrDBStringGridEd.DrawCellEx(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState; BkColor: TColor; DrawFocus: boolean);
var
  Frm: integer;
  mrDef: TrDBStringGridEdMultiRow;
  s: string;
  R: TRect;

  procedure DrwBmpPlace(Bmp: TBitmap);
  var
    h, w, rh, rw: integer;
  begin
    h:=Bmp.Height + 2;
    if h < DefaultRowHeight then
      h:=DefaultRowHeight;
    if mrDef.RowHeightMax > 0 then
      h:=Min(mrDef.RowHeightMax, h);
    if mrDef.RowHeightMin > 0 then
      h:=Max(mrDef.RowHeightMin, h);
    if h <> RowHeights[ARow] then
      RowHeights[ARow]:=h;

    h:=ARect.Bottom - ARect.Top;
    w:=ARect.Right - ARect.Left;

    if (Bmp.Height <= h) and (Bmp.Width <= w) then
      rpictool.DrawBmpAlign(Canvas, ARect, Bmp, mrDef.FAlignment)
    else
    begin
      if h / Bmp.Height < w / Bmp.Width then
      begin
        rh:=h;
        rw:=Round(Bmp.Width * h / Bmp.Height);
      end
      else
      begin
        rh:=Round(Bmp.Height * w / Bmp.Width);
        rw:=w;
      end;
      R:=ARect;
      // R.Top:=ARect.Top + (h - rh) div 2;
      R.Bottom:=R.Top + rh;
      case mrDef.FAlignment of
        taLeftJustify:
          R.Right:=R.Left + rw;
        taRightJustify:
          R.Left:=R.Right - rw;
        taCenter:
          begin
            R.Left:=R.Left + (w - rw) div 2;
            R.Right:=R.Left + rw;
          end;
      end;
      rpictool.DrawBmpCenter(Canvas, R, Bmp, true);
    end;
  end;

  procedure DrwBmp;
  var
    Fld: Tfield;
    B: TBitmap;
  begin
    Fld:=FDataLink.Dataset.FindField(KeyFieldName);
    if Assigned(Fld) then
      if rdbtool.LocateRecord(Fld, self.GetRowKeyValue(ARow)) then
      begin
        Fld:=FDataLink.Dataset.FindField(mrDef.FFieldName);
        if Assigned(Fld) then
        begin
          B:=TBitmap.Create;
          try
            try
              ReadBitmapFromField(Fld, B);
              DrwBmpPlace(B);
            except
              s:=sPictureCorrupted;
              DrawText(Canvas.Handle, PChar(s), Length(s), ARect, GetTextFormatSetting(false));
            end;
          finally
            FreeAndNil(B);
          end;
        end;
      end;
  end;

begin
  // get def
  mrDef:=GetCellMultiRowDef(ACol, ARow);
  // extra draw for multirow data cell
  if Assigned(mrDef) and Columns[ACol].CustomDraw then
  begin
    R:=ARect;
    InflateRect(ARect, -1, -1);
    // inflate Background rect if focus is drawn
    if DrawFocus then
      R:=ARect;

    // set canvas format
    SetCanvasStyle(ACol, ARow, Canvas);
    if not mrDef.FParentFont then
      Canvas.Font.Assign(mrDef.FFont);

    // show background if needed
    if mrDef.FShowFixed then
    begin
      AState:=AState + [gdFixed]; // -[gdSelected];
      DrawCellBackground(R, Canvas.Brush.Color, AState, ACol, ARow);
    end
    else
    begin
      if mrDef.FColor <> clNone then
        DrawCellBackground(R, mrDef.FColor, AState, ACol, ARow);
    end;

    Canvas.Brush.Style:=bsClear;
    // do not draw cell content if custom draw
    if mrDef.FCustomDraw then
      Exit;

    // set alignment
    case mrDef.FAlignment of
      taRightJustify:
        Frm:=DT_RIGHT;
      taCenter:
        Frm:=DT_CENTER
      else
        Frm:=DT_LEFT;
    end;
    Frm:=Frm or GetTextFormatSetting(mrDef.FWordWrap);

    // draw cell content
    s:=Cells[ACol, ARow];
    R:=ARect;
    InflateRect(ARect, -InnerMargin, 0);
    case EvaluateMultiRowType(mrDef.FMultiRowType, s) of
      mrtPlainText:
        begin
          InflateRect(ARect, -1, 0);
          DrawText(Canvas.Handle, PChar(s), Length(s), ARect, Frm);
        end;
      mrtHTML:
        begin
          rHTMLDrawText.rHTML_DrawText(self, Canvas, s, ARect, false, true, mrDef.WordWrap, false, mrDef.FAlignment, tlCenter, 0);
        end;
      mrtRTF:
        begin
          if Assigned(FRichText) then
            DrawRichText(R, s, BkColor)
          else
          begin
            InflateRect(ARect, -1, 0);
            DrawText(Canvas.Handle, PChar(s), Length(s), ARect, GetTextFormatSetting(true));
          end;
        end;
      mrtColumns:
        DrawColumnsInCell(ARect, s, false, mrDef.Alignment, mrDef.WordWrap);
      mrtColumnsBlock:
        DrawColumnsInCell(ARect, s, true, mrDef.Alignment, mrDef.WordWrap);
      mrtObjects, mrtFileObjects:
        DrawObjectsInCell(ARect, ACol, ARow, false, false, false);
      mrtImage:
        if csDesigning in ComponentState then
          DrawColumnsInCell(ARect, s, true, mrDef.Alignment, mrDef.WordWrap)
        else
          DrwBmp;
    end;
  end;
end;

function TrDBStringGridEd.CalcCellHeight(ACol, ARow: Longint; UseLimits: boolean): integer;
var
  h: integer;
  R: TRect;
  s: string;
  mrDef: TrDBStringGridEdMultiRow;
  UseInherited: boolean;
begin
  mrDef:=GetMultiRowDef(ARow);
  UseInherited:=not Assigned(mrDef);
  if Assigned(mrDef) then
  begin
    SetRect(R, 0, 0, ColWidths[ACol], DefaultRowHeight);
    // reduce width (diference between CellRect and R)
    R.Right:=R.Right - 4 - 2 * InnerMargin;
    // set canvas format
    SetCanvasStyle(ACol, ARow, Canvas);
    if not mrDef.FParentFont then
      Canvas.Font.Assign(mrDef.FFont);
    // calculate height
    case EvaluateMultiRowType(mrDef.FMultiRowType, s) of
      mrtHTML:
        rHTMLDrawText.rHTML_DrawText(self, self.Canvas, s, R, true, true, mrDef.WordWrap, false);
      mrtColumns, mrtColumnsBlock:
        DrawText(Canvas.Handle, PChar(s), Length(s), R, GetTextFormatSetting(false) or DT_CALCRECT);
      mrtObjects, mrtFileObjects:
        DrawObjectsInCell(R, ACol, ARow, true, false, false);
    else
      UseInherited:=true;
    end;
  end;
  if UseInherited then
    h:= inherited CalcCellHeight(ACol, ARow, UseLimits)
  else
    h:=R.Bottom - R.Top + 2;
  // check min, max, free space
  if UseLimits then
  begin
    Inc(h, mrDef.FRowHeightFreeSpace);
    if (mrDef.RowHeightMax > 0) then
      h:=Min(mrDef.RowHeightMax, h);
    if mrDef.RowHeightMin > 0 then
      h:=Max(mrDef.RowHeightMin, h);
  end;
  Result:=h;
end;

procedure TrDBStringGridEd.RefreshMultiRowHeights;
begin
  // SetMultiRowHeights;
  Invalidate;
end;

procedure TrDBStringGridEd.DrawColumnsInCell(R: TRect; s: string; Block: boolean; Align: TAlignment; Wrap: boolean);
var
  L: TStrings;
  c, i, cW: integer;
  cR: TRect;
begin
  // draw each line as one part of the cell
  L:=TStringList.Create;
  try
    L.Text:=s;
    c:=L.Count;
    if c = 0 then
      Exit;
    cW:=(R.Right - R.Left) div c;
    cR:=R;
    for i:=0 to c - 1 do
    begin
      s:=L[i];
      cR.Left:=R.Left + cW * i;
      cR.Right:=cR.Left + cW;
      if Block then
      begin
        if i = 0 then
          Align:=taLeftJustify
        else
          if i = c - 1 then
            Align:=taRightJustify
          else
            Align:=taCenter;
      end;
      rHTMLDrawText.rHTML_DrawText(self, Canvas, s, cR, false, true, Wrap, false, Align, tlCenter, 0);
    end;
  finally
    L.Free;
  end;
end;

procedure TrDBStringGridEd.DrawObjectsInCell(var R: TRect; ACol, ARow: Longint; CalcSize, TestObjectClick, DblClick: boolean);
var
  L: TStrings;
  s: string;
  c, i, cW, cH, aTop, aLeft: integer;
  tR, rIcon: TRect;
  BColor: TColor;
  B: TBitmap;
  useIcon, FileObject: boolean;
begin
  if not CalcSize then
    Canvas.Pen.Color:=Canvas.Font.Color;
  BColor:=Canvas.Brush.Color;
  FileObject:=GetCellMultiRowType(ACol, ARow) = mrtFileObjects;
  // draw each line as one object
  L:=TStringList.Create;
  try
    L.Text:=Cells[ACol, ARow];
    c:=L.Count;
    if c = 0 then
      Exit;
    // get text height - same for all items
    tR:=R;
    rHTMLDrawText.rHTML_DrawTextSimple(self, Canvas, 'Wq', tR, true, true);
    cH:=tR.Bottom - tR.Top + 2 * (cObjInnerMargin + FMultiRowsOptions.FObjectInnerMargin);
    // init 1st obj pos
    aTop:=R.Top + cObjSpace;
    aLeft:=R.Left;
    // for each object
    for i:=0 to c - 1 do
    begin
      s:=L[i];
      tR:=R;
      // get text width
      rHTMLDrawText.rHTML_DrawText(self, Canvas, s, tR, true, true, false, false, taCenter, tlCenter, 0);
      cW:=tR.Right - tR.Left + 2 * (cObjInnerMargin + FMultiRowsOptions.FObjectInnerMargin);
      useIcon:=FileObject and (ExtractFileExt(s) <> '');
      if useIcon then
        Inc(cW, cObjIconSize + cObjInnerMargin + FMultiRowsOptions.FObjectInnerMargin);
      if (aLeft > R.Left) and (aLeft + cW > R.Right) then
      begin
        // move to next line
        Inc(aTop, cH + cObjSpace);
        aLeft:=R.Left;
      end;
      // calc obj pos+size
      tR:=Rect(aLeft, aTop, Min(aLeft + cW, R.Right), aTop + cH);

      if not CalcSize then
      begin
        if TestObjectClick then
        begin
          if PtInRect(tR, MouseDownCellPoint) then
          begin
            // call events
            if DblClick then
              DoGeneralEventParam(geCellObjectDblClick, ACol, ARow, TObject(i), s)
            else
              DoGeneralEventParam(geCellObjectClick, ACol, ARow, TObject(i), s);
            Exit;
          end;
        end
        else
        begin
          // change bk for selected object
          if PtInRect(tR, MouseDownCellPoint) then
            Canvas.Brush.Color:=clBtnFace
          else
            Canvas.Brush.Color:=BColor;
          Canvas.RoundRect(tR, cH div 2, cH div 2);
          if useIcon then
          begin
            B:=TBitmap.Create;
            try
              B.Canvas.Brush.Color:=Canvas.Brush.Color;
              LoadAssociatedIconAsBitmap(B, s);
              rIcon:=tR;
              rIcon.Left:=tR.Left + 1;
              rIcon.Right:=rIcon.Left + cObjIconSize + 2 * (cObjInnerMargin + +FMultiRowsOptions.FObjectInnerMargin);
              InflateRect(rIcon, 0, -1);
              rpictool.DrawBmpCenter(Canvas, rIcon, B, false);
            finally
              B.Free;
            end;
            tR.Left:=tR.Left + cObjIconSize + cObjInnerMargin + FMultiRowsOptions.FObjectInnerMargin;
          end;
          rHTMLDrawText.rHTML_DrawText(self, Canvas, s, tR, false, true, false, false, taCenter, tlCenter, 0);
        end;
      end;
      // move to next pos
      Inc(aLeft, cW + cObjSpace);
    end;
    if CalcSize then
      R.Bottom:=tR.Bottom + cObjSpace;
  finally
    L.Free;
  end;
end;

function TrDBStringGridEd.IsTableMultiRows: boolean;
begin
  Result:=FIsTableMultiRows;
end;

function TrDBStringGridEd.GetMultiRowDef(ARow: Longint): TrDBStringGridEdMultiRow;
begin
  Result:=nil;
  if (ARow >= FixedRows) and (FMultiRowsDef.Count > 0) then
    Result:=FMultiRowsDef.Items[(ARow - FixedRows) mod FMultiRowsDef.Count];
end;

function TrDBStringGridEd.GetCellMultiRowDef(ACol, ARow: Longint): TrDBStringGridEdMultiRow;
begin
  Result:=nil;
  if IsTableMultiRows and (ACol = GetMultiRowDataCol) and (ARow >= FixedRows) then
    Result:=GetMultiRowDef(ARow);
end;

function TrDBStringGridEd.EvaluateMultiRowType(CurrentType: TMultiRowType; Text: string): TMultiRowType;
begin
  Result:=CurrentType;
  if CurrentType = mrtAuto then
  begin
    if AnsiPos('{\rtf', Text) = 1 then
      Result:=mrtRTF
    else
      if AnsiPos('</', Text) > 0 then
        Result:=mrtHTML
      else
        Result:=mrtPlainText;
  end;
end;

function TrDBStringGridEd.GetCellMultiRowType(ACol, ARow: Longint): TMultiRowType;
var
  mrDef: TrDBStringGridEdMultiRow;
begin
  Result:=mrtPlainText;
  mrDef:=GetCellMultiRowDef(ACol, ARow);
  if Assigned(mrDef) then
    Result:=EvaluateMultiRowType(mrDef.FMultiRowType, Cells[ACol, ARow]);
end;

function TrDBStringGridEd.GetActiveCellMultiRowType: TMultiRowType;
begin
  Result:=GetCellMultiRowType(Col, Row);
end;

function TrDBStringGridEd.GetMultiRowFixedCol: integer;
begin
  Result:=FMultiRowsOptions.FMultiRowFixedCol;
  if Result < 0 then
    Result:=FixedCols - 1;
end;

function TrDBStringGridEd.GetMultiRowDataCol: integer;
begin
  Result:=FMultiRowsOptions.FMultiRowDataCol;
  if Result < 0 then
    Result:=FixedCols;
end;

procedure TrDBStringGridEd.ShowMultiRowPreview;
var
  ARow, mrIdx: Longint;
  mrDef: TrDBStringGridEdMultiRow;

  function ReplTag(s: string): string;
  begin
    s:=ReplaceStr('<<', '[', s);
    s:=ReplaceStr('>>', ']', s);
    Result:=s;
  end;

begin
  if FMultiRowsPreview and (csDesigning in ComponentState) then
  begin
    ClearEntireTable(true);
    SelectCell(FixedCols, FixedRows);
    Columns[GetMultiRowDataCol].CustomDraw:=true;
    FIsTableMultiRows:=true;

    RowCount:=Max(FixedRows + 1, FixedRows + FMultiRowsDef.Count);
    for mrIdx:=0 to FMultiRowsDef.Count - 1 do
    begin
      mrDef:=FMultiRowsDef.Items[mrIdx];
      ARow:=FixedRows + mrIdx;
      if GetMultiRowFixedCol >= 0 then
        StringCells[GetMultiRowFixedCol, ARow]:=rtool.CleanLastEOL(ReplTag(mrDef.FFixedColTextMask.Text));
      if mrDef.FTextMask.Count > 0 then
        StringCells[GetMultiRowDataCol, ARow]:=rtool.CleanLastEOL(ReplTag(mrDef.FTextMask.Text))
      else
        if mrDef.FFieldName <> '' then
          StringCells[GetMultiRowDataCol, ARow]:='[' + mrDef.FFieldName + ']';
    end;
    ResetChangeLog;
  end;
end;

procedure TrDBStringGridEd.LoadMultiRowRecInt(StartRow: Longint; NewRec: boolean; KeyFld: TField = nil; DS: TDataset = nil);
var
  ARow, mrIdx: Longint;
  mrDef: TrDBStringGridEdMultiRow;
  Fld: Tfield;
  s: string;
begin
  ARow:=StartRow;
  for mrIdx:=0 to FMultiRowsDef.Count - 1 do
  begin
    if Assigned(KeyFld) then
      RowKeyValue[ARow]:=KeyFld.AsInteger
    else
      RowKeyValue[ARow]:=0;
    mrDef:=FMultiRowsDef.Items[mrIdx];
    if GetMultiRowFixedCol >= 0 then
    begin
      s:=rtool.CleanLastEOL(mrDef.FFixedColTextMask.Text);
      if Assigned(DS) then
        s:=rdbtool.FillTextMask(s, DS);
      if Assigned(FDBUpdateCellContent) then
        FDBUpdateCellContent(self, DS, ARow, GetMultiRowFixedCol, mrDef, s);
      StringCells[GetMultiRowFixedCol, ARow]:=s;
    end;
    s:='';
    if NewRec then
      StringCells[GetMultiRowDataCol, ARow]:=s
    else
    if mrDef.FTextMask.Count > 0 then
    begin
      s:=rtool.CleanLastEOL(mrDef.FTextMask.Text);
      if Assigned(DS) then
        s:=rdbtool.FillTextMask(s, DS);
      if Assigned(FDBUpdateCellContent) then
        FDBUpdateCellContent(self, DS, ARow, GetMultiRowDataCol, mrDef, s);
      StringCells[GetMultiRowDataCol, ARow]:=s;
    end
    else
      if mrDef.FFieldName <> '' then
      begin
        if Assigned(DS) then
        begin
          Fld:=DS.FindField(mrDef.FFieldName);
          if Assigned(Fld) then
          begin
            if mrDef.FMultiRowType <> mrtImage then
              LoadCellFromField(DS, GetMultiRowDataCol, ARow, Fld)
          end
          else
            StringCells[GetMultiRowDataCol, ARow]:=Format('Field "%s" not found.', [mrDef.FFieldName]);
        end;
        { else
          if not AddNewRec then
          StringCells[GetMultiRowDataCol, ARow]:=mrDef.FFieldName;
          }
      end;
    Inc(ARow);
  end;
end;

procedure TrDBStringGridEd.AddNewRecMultiRows(InsertToTop: boolean = false; InsertInCurrentPos: boolean = false);
var
  ARow: Longint;
begin
  if InsertToTop then
  begin
    ARow:=FixedRows;
    InsertRowEx(ARow, FMultiRowsDef.Count, false);
  end
  else
    if InsertInCurrentPos then
    begin
      ARow:=((Row - FixedRows) div FMultiRowsDef.Count) * FMultiRowsDef.Count + FixedRows;
      InsertRowEx(ARow, FMultiRowsDef.Count, false);
    end
    else
    begin
      ARow:=RowCount;
      RowCount:=RowCount + FMultiRowsDef.Count;
    end;
  LoadMultiRowRecInt(ARow, true);
  Row:=ARow;
end;

procedure TrDBStringGridEd.LoadTableMultiRows;
var
  ARow: Longint;
  KeyFld: Tfield;
begin
  // clear grid and init data column
  ClearEntireTable(true);
  SelectCell(FixedCols, FixedRows);
  Columns[GetMultiRowDataCol].CustomDraw:=true;
  FIsTableMultiRows:=true;

  if Assigned(FDataLink.Dataset) then
  begin
    // DS used, load all records
    if not FDataLink.Active then
      Exit;
    FDataLink.Dataset.Cancel;
    ARow:=FDataLink.Dataset.RecordCount;
    if ARow = 0 then
    begin
      RowCount:=FixedRows + 1;
      Exit;
    end;
    RowCount:=FixedRows + ARow * FMultiRowsDef.Count;
    KeyFld:=FDataLink.Dataset.FindField(KeyFieldName);
    FDataLink.Dataset.DisableControls;
    BeginUpdate;
    try
      ARow:=FixedRows;
      FDataLink.Dataset.First;
      while not FDataLink.Dataset.Eof do
      begin
        LoadMultiRowRecInt(ARow, false, KeyFld, FDataLink.Dataset);
        Inc(ARow, FMultiRowsDef.Count);
        FDataLink.Dataset.Next;
      end;
    finally
      EndUpdate;
      FDataLink.Dataset.EnableControls;
    end;
  end
  else
  begin
    // DS not used, load once
    RowCount:=FixedRows + FMultiRowsDef.Count;
    BeginUpdate;
    try
      LoadMultiRowRecInt(FixedRows, false);
    finally
      EndUpdate;
    end;
  end;
  ResetChangeLog;
end;

function TrDBStringGridEd.SaveTableMultiRows(SaveNewRec: boolean = true): Longint;
var
  R, c, mrIdx: Longint;
  KeyF, ValueField: Tfield;
  mrDef: TrDBStringGridEdMultiRow;
  IsEditing: boolean;
  UpdKind: TUpdateKind;
begin
  Result:=0;
  ResetUpdatedCount;
  if not FDataLink.Active then
    Exit;
  FDataLink.Dataset.Cancel;
  KeyF:=FDataLink.Dataset.FindField(KeyFieldName);
  if not Assigned(KeyF) then
    raise Exception.Create(KeyFieldNotFoundMsg);

  c:=GetMultiRowDataCol;
  IsEditing:=false;
  // all rows
  for R:=FixedRows to RowCount - 1 do
  begin
    mrIdx:=(R - FixedRows) mod FMultiRowsDef.Count;
    if mrIdx = 0 then
    begin
      // first row in the multirow block
      FKeyValue:=RowKeyValue[R];
    end;
    if FKeyValue = 0 then
    begin
      if not SaveNewRec then
        Continue;
    end;
    UpdKind:=ukModify;
    if IsCellChanged(c, R) then
    begin
      if not IsEditing then
      begin
        if (FKeyValue = 0) then
          UpdKind:=ukInsert
        else
          if not LocateRecord(KeyF, FKeyValue) then
            raise Exception.Create(Format(WarningMsg, [FKeyValue]));
        TryPostRowData(FDataLink.Dataset, nil, UpdKind, R, 0, 1);
        IsEditing:=true;
      end;
      // update field
      mrDef:=GetCellMultiRowDef(c, R);
      ValueField:=FDataLink.Dataset.FindField(mrDef.FFieldName);
      StoreCellToField(FDataLink.Dataset, c, R, ValueField);
    end;

    if mrIdx = FMultiRowsDef.Count - 1 then
    begin
      // last row in the multirow block - store record
      if IsEditing then
        try
          TryPostRowData(FDataLink.Dataset, nil, UpdKind, R, 0, 2);
        finally
          IsEditing:=false;
        end;
    end;
  end;
  ResetChangeLog;
  Result:=FRecsUpdatedCount;
end;

end.
