unit rDBTreeView;

{<
  rDBTreeView unit contains TrDBTreeView - database version of standard TTreeView component
  which provides function for loading datasets values to tree structure.

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2012)
  @lastmod(Jan 2019)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, StdCtrls,
  Forms, DB, Clipbrd, ComCtrls, CommCtrl, StrUtils, Dialogs, ExtCtrls,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} Math, Variants,
  RosiCompGlobalCfg;

type
  TrDBTreeView = class;
  TDataItemDefs = class;
  TNodeObject = class;

  { list of possible DropItem options used by TrDBTreeView }
  TDragDropItemOption = (dnoSameLevel, dnoAllLevels, dnoSameParent, dnoSameParentType, dnoOtherControl, dnoMoveItem);
  { set of possible DropItem options used by TrDBTreeView }
  TDragDropItemOptions = set of TDragDropItemOption;
  { event used as notification after some node is added or refreshed used by TrDBTreeView }
  TTVItemObjectDefinedEvent = procedure(Sender: TObject; Node: TTreeNode; NodeObject: TNodeObject; Dataset: TDataset) of object;

  { DataItemDefinition object - define titletext, recordtext and properties for one datasource used by TrDBTreeView }
  TDataItemDef = class(TCollectionItem)
  private
    FDataSource: TDataSource;
    FParentKeyFieldName: string;
    FKeyFieldName: string;
    FListAllRecords: boolean;
    FMultilevelMode: boolean;
    FMultilevelTopLevelFilter: string;
    FShowTitle: boolean;
    FExpandNode: boolean;
    FCanEditText: boolean;
    FDragDropMode: TDragDropItemOptions;
    FMaxRecords: integer;
    FTitleText: string;
    FTitleImageIndex: integer;
    FTitleBkColor: TColor;
    FTitleFontColor: TColor;
    FTitleStyle: TFontStyles;
    FRecordTextMask: string;
    FRecordImageIndex: integer;
    FRecordSelectedImageIndex: integer;
    FRecordExpandedImageIndex: integer;
    FRecordStateImageIndex: integer;
    FRecordBkColor: TColor;
    FRecordFontColor: TColor;
    FRecordChangedFontColor: TColor;
    FRecordStyle: TFontStyles;
    FEmptyDatasetText: string;
    FFillLeftBorder: integer;
    FFillRightBorder: integer;
    FSubDataItemDefs: TDataItemDefs;
    procedure SetSubDataItemDefs(Value: TDataItemDefs);
    procedure SetDataSource(Value: TDataSource);
    procedure SetListAllRecords(Value: boolean);
    procedure SetShowTitle(Value: boolean);
    procedure SetExpandNode(Value: boolean);
    procedure SetTitleText(const Value: string);
    procedure SetTitleImageIndex(Value: integer);
    procedure SetTitleBkColor(Value: TColor);
    procedure SetTitleFontColor(Value: TColor);
    procedure SetTitleStyle(Value: TFontStyles);
    procedure SetRecordTextMask(const Value: string);
    procedure SetRecordImageIndex(Value: integer);
    procedure SetRecordBkColor(Value: TColor);
    procedure SetRecordFontColor(Value: TColor);
    procedure SetRecordStyle(Value: TFontStyles);
    procedure SetFillLeftBorder(Value: integer);
    procedure SetFillRightBorder(Value: integer);
  protected
    function GetTreeView: TrDBTreeView;
    procedure RefreshTreeView;
  public
    constructor Create(Collection: TCollection); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    { list of child DataItemDefinition - is shown as child treenode }
    property SubDataItemDefs: TDataItemDefs read FSubDataItemDefs write SetSubDataItemDefs;
    { datasource used for data fields }
    property DataSource: TDataSource read FDataSource write SetDataSource;
    { KeyField is used for saving of KeyValue for each node }
    property KeyFieldName: string read FKeyFieldName write FKeyFieldName;
    { ParentKeyFieldName define field which contains id of parent node (used for multilevel mode) }
    property ParentKeyFieldName: string read FParentKeyFieldName write FParentKeyFieldName;
    { text used for empty dataset node }
    property EmptyDatasetText: string read FEmptyDatasetText write FEmptyDatasetText;
    { define that all records will be loaded and displayed in treeview }
    property ListAllRecords: boolean read FListAllRecords write SetListAllRecords default true;
    { define that sublevels will be created too for the same dataset - ParentKeyFieldName is used to find child records }
    property MultilevelMode: boolean read FMultilevelMode write FMultilevelMode default false;
    { define filter string for first level for multilevel - usually ParentID=0 or ParentID=null }
    property MultilevelTopLevelFilter: string read FMultilevelTopLevelFilter write FMultilevelTopLevelFilter;
    { define maximum number of records in one node, last node contain ... }
    property MaxRecords: integer read FMaxRecords write FMaxRecords default 0;
    { define whether node is expanded after loading }
    property ExpandNode: boolean read FExpandNode write SetExpandNode default true;
    { define whether extra node with title is used }
    property ShowTitle: boolean read FShowTitle write SetShowTitle default true;
    { define title text, can contain mask for loadind of field value - <<FieldName>> }
    property TitleText: string read FTitleText write SetTitleText;
    { define image index for title node }
    property TitleImageIndex: integer read FTitleImageIndex write SetTitleImageIndex default - 1;
    { define background color for title node }
    property TitleBkColor: TColor read FTitleBkColor write SetTitleBkColor default clNone;
    { define font color for title node }
    property TitleFontColor: TColor read FTitleFontColor write SetTitleFontColor default clNone;
    { define font style for title node }
    property TitleStyle: TFontStyles read FTitleStyle write SetTitleStyle default[];
    { define text for each record, can contain mask for loadind of field value - <<FieldName>> }
    property RecordTextMask: string read FRecordTextMask write SetRecordTextMask;
    { define image index for each record node }
    property RecordImageIndex: integer read FRecordImageIndex write SetRecordImageIndex default - 1;
    { define image index for each record node }
    property RecordSelectedImageIndex: integer read FRecordSelectedImageIndex write FRecordSelectedImageIndex default - 1;
    { define image index for each record node }
    property RecordExpandedImageIndex: integer read FRecordExpandedImageIndex write FRecordExpandedImageIndex default - 1;
    { define image index for each record node }
    property RecordStateImageIndex: integer read FRecordStateImageIndex write FRecordStateImageIndex default - 1;
    { define background color for each record node }
    property RecordBkColor: TColor read FRecordBkColor write SetRecordBkColor default clNone;
    { define font color for each record node }
    property RecordFontColor: TColor read FRecordFontColor write SetRecordFontColor default clNone;
    { define font color for chnaged record node }
    property RecordChangedFontColor: TColor read FRecordChangedFontColor write FRecordChangedFontColor default clNone;
    { define font style for each record node }
    property RecordStyle: TFontStyles read FRecordStyle write SetRecordStyle default[];
    { define left border for background and underline }
    property FillLeftBorder: integer read FFillLeftBorder write SetFillLeftBorder default 0;
    { define right border for background and underline }
    property FillRightBorder: integer read FFillRightBorder write SetFillRightBorder default 0;
    { define whether items can be edited }
    property CanEditText: boolean read FCanEditText write FCanEditText default false;
    { define whether and how items can be moved }
    property DragDropMode: TDragDropItemOptions read FDragDropMode write FDragDropMode default[];
  end;

  TDataItemDefClass = class of TDataItemDef; {< @exclude }

  { DataItemDefinition collection - define collection of TDataItemDef used by TrDBTreeView }
  TDataItemDefs = class(TCollection)
  private
    FrDBTreeView: TrDBTreeView;
    function GetDataItemDef(Index: integer): TDataItemDef;
    procedure SetDataItemDef(Index: integer; Value: TDataItemDef);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(rDBTreeView: TrDBTreeView; DataItemDefClass: TDataItemDefClass); {< @exclude }
    function Add: TDataItemDef; {< @exclude }
    property rDBTreeView: TrDBTreeView read FrDBTreeView; {< @exclude }
    property Items[Index: integer]: TDataItemDef read GetDataItemDef write SetDataItemDef; default; {< @exclude }
  end;

  { NodeObject datastructure stored dataset and visual information for each node used by TrDBTreeView }
  TNodeObject = class
  public
    { indicate that item is initialized already, used internaly }
    IsInitialized: boolean;
    { indicate that item is loaded already, used internaly }
    NodeIsLoaded: boolean;
    { indicate that item is Title }
    IsTitle: boolean;
    { indicate that item is info item added for Empty Datase }
    IsEmptyDataset: boolean;
    { contain RecNo of record }
    RecNo: integer;
    { link to KeyField object }
    KeyField: TField;
    { contain Value of KeyField }
    KeyValue: variant;
    { contain background color }
    BkColor: TColor;
    { contain font color }
    FontColor: TColor;
    { contain font color for changed items }
    ChangedFontColor: TColor;
    { contain font styles }
    Style: TFontStyles;
    { link to ParentDataItemDef object, can be used to read all itemdef parameters }
    ParentDataItemDef: TDataItemDef;
    { link to DataItemDef object, can be used to read all itemdef parameters }
    DataItemDef: TDataItemDef;
    { indicate that item text was edited }
    ItemChanged: boolean;
    { indicate that item was moved by drag-and-drop }
    ItemMoved: boolean;
  end;

  { extended class of TTreeNode contaning TNodeObject useb by TrDBTreeView }
  TrTreeNode = class(TTreeNode)
  protected
    NodeObject: TNodeObject;
  public
    constructor Create(AOwner: TTreeNodes); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  end;

  { @exclude }
  TrDBRecordLink = class(TDataLink)
  private
    FrDBTreeView: TrDBTreeView;
  protected
    procedure ActiveChanged; override;
    procedure DataSetChanged; override;
    procedure EditingChanged; override;
    procedure UpdateData; override;
  public
    constructor Create(rDBTreeView: TrDBTreeView);
  end;

  { database version of standard TTreeView component
    which provides function for loading datasets values to tree structure. }
  TrDBTreeView = class(TTreeView)
  private
    IsRefreshing: boolean;
    FUpdateCount: Integer;
    FDataLink: TrDBRecordLink;
    FDataItemDefs: TDataItemDefs;
    FDataItemDesignTime: boolean;
    FLoadFirstLevelOnly: boolean;
    oldTVEditedEvent: TTVEditedEvent;
    lastEditedNode: TTreeNode;
    FOnItemEdited: TTVChangedEvent;
    FOnItemMoved: TTVChangedEvent;
    FOnItemObjectDefined: TTVItemObjectDefinedEvent;
    FScrollTimer: TTimer;
    FLoadDataTimer: TTimer;
    FLoadDataDelay: cardinal;
    function GetMasterDataSource: TDataSource;
    procedure SetMasterDataSource(Value: TDataSource);
    procedure SetDataItemDefs(Value: TDataItemDefs);
    procedure SetImageIndex(Node: TTreeNode; idx: integer);
    procedure ShowSubDataSources(ItemDef: TDataItemDef; Node: TTreeNode; Expanding: boolean);
    function AddItem(ParentItemDef, ItemDef: TDataItemDef; ParentNode: TTreeNode; ShowSubData, IsTitle, Expanding: boolean; ConstantText: string = '')
      : TTreeNode;
    procedure LoadDataItemDef(ParentItemDef, ItemDef: TDataItemDef; ParentNode: TTreeNode; Expanding: boolean);
    procedure SetDataItemDesignTime(Value: boolean);
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMMouseWheel(var Message: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure IntTVEditedEvent(Sender: TObject; Node: TTreeNode; var S: string);
    procedure IntScrollTimer(Sender: TObject);
    procedure IntLoadTimer(Sender: TObject);
  protected
    procedure Loaded; override; {< @exclude }
    procedure Notification(AComponent: TComponent; Operation: TOperation); override; {< @exclude }
    function CreateNode: TTreeNode; override; {< @exclude }
    function CanEdit(Node: TTreeNode): boolean; override;
    procedure Edit(const Item: TTVItem); override;
    function CanExpand(Node: TTreeNode): boolean; override;
    function IsCustomDrawn(Target: TCustomDrawTarget; Stage: TCustomDrawStage): boolean; override;
    function CustomDrawItem(Node: TTreeNode; State: TCustomDrawState; Stage: TCustomDrawStage; var PaintImages: boolean): boolean; override;
    procedure Change(Node: TTreeNode); override;
    procedure DragOver(Source: TObject; X, Y: integer; State: TDragState; var Accept: boolean); override;
    procedure Collapse(Node: TTreeNode); override;
    procedure Expand(Node: TTreeNode); override;
    property UpdateCount: Integer read FUpdateCount;
  public
    constructor Create(AOwner: TComponent); override; { @exclude }
    destructor Destroy; override; { @exclude }
    procedure BeginUpdate; virtual;
    procedure EndUpdate; virtual;
    procedure DragDrop(Source: TObject; X, Y: integer); override; { @exclude }
    { return datastructure for selected node, Node.Data can be used too }
    function GetSelectedData: TNodeObject;
    { return datastructure for node according index, use AbsoluteIndex to get Node index }
    function GetNodeData(NodeAbsoluteIndex: integer): TNodeObject;
    { go through all items and call CallBackProc for each changed or moved item }
    procedure ProcessAllItems(UseChanged, UseMoved: boolean; CallBackProc: TTVChangedEvent; ResetState: boolean);
    { try to find node according KeyField and value }
    function FindNode(KeyField: TField; KeyValue: variant): TTreeNode;
    { reload all nodes }
    procedure RefreshData(Delayed: boolean = false; KeepSelectedNode: boolean = false);
    { reload text of node, it uses current record, so dataset must be moved to correct record before calling }
    procedure RefreshNode(Node: TTreeNode);
    { create child nodes, it uses current dataset records, so number of items is defined by dataset }
    procedure CreateChildNodes(ParentNode: TTreeNode; MultilevelNode, RemoveCurrentNodes: boolean);
    { create child nodes, it uses current dataset records, so number of items is defined by dataset }
    procedure CreateSiblingNode(SiblingNode: TTreeNode);
  published
    { define datasource used for automatic loading of data }
    property MasterDataSource: TDataSource read GetMasterDataSource write SetMasterDataSource;
    { define collection of DataItemDefinition }
    property DataItemDefs: TDataItemDefs read FDataItemDefs write SetDataItemDefs;
    { define whether data items is shown in design time and stored to DFM as items }
    property DataItemDesignTime: boolean read FDataItemDesignTime write SetDataItemDesignTime default true;
    { define whether only first level of items is loaded and child data is loaded after expand of item }
    property LoadFirstLevelOnly: boolean read FLoadFirstLevelOnly write FLoadFirstLevelOnly default false;
    { event is called after edit text of item }
    property OnItemEdited: TTVChangedEvent read FOnItemEdited write FOnItemEdited;
    { event is called after drag-and-drop for moved item }
    property OnItemMoved: TTVChangedEvent read FOnItemMoved write FOnItemMoved;
    { event is called when new node is created and NodeObject is defined }
    property OnItemObjectDefined: TTVItemObjectDefinedEvent read FOnItemObjectDefined write FOnItemObjectDefined;
    { define delay before Masterdatasource scroll and loading to treeview }
    property LoadDataDelay: cardinal read FLoadDataDelay write FLoadDataDelay default 0;
  end;

procedure Register; {< @exclude }

implementation

uses rstring, rtool, rdbtool;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 30 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBTreeView]);
end;

const
  ScrollBorder = 16;

  // ==============================================================================
  // TDataItemDef class
constructor TDataItemDef.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  // create subdata defs
  FSubDataItemDefs:=TDataItemDefs.Create(GetTreeView, TDataItemDef);
  // init all props to default
  FListAllRecords:=true;
  FMultilevelMode:=false;
  FShowTitle:=true;
  FExpandNode:=true;
  FCanEditText:=false;
  FDragDropMode:=[];
  FMaxRecords:=0;
  FTitleImageIndex:=-1;
  FTitleBkColor:=clNone;
  FTitleFontColor:=clNone;
  FTitleStyle:=[];
  FRecordImageIndex:=-1;
  FRecordSelectedImageIndex:=-1;
  FRecordExpandedImageIndex:=-1;
  FRecordStateImageIndex:=-1;
  FRecordBkColor:=clNone;
  FRecordFontColor:=clNone;
  FRecordChangedFontColor:=clNone;
  FRecordStyle:=[];
  FFillLeftBorder:=0;
  FFillRightBorder:=0;
end;

destructor TDataItemDef.Destroy;
begin
  FreeAndNil(FSubDataItemDefs);
  inherited Destroy;
end;

procedure TDataItemDef.SetSubDataItemDefs(Value: TDataItemDefs);
begin
  FSubDataItemDefs.Assign(Value);
end;

procedure TDataItemDef.SetDataSource(Value: TDataSource);
var
  tv: TrDBTreeView;
begin
  // remove and add FreeNotification to TreeView component in order to
  // update DataItemDef if dataset is removed
  // (only component can be used, collection do not support freenotification)
  tv:=GetTreeView;
  if Assigned(FDataSource) and Assigned(tv) then
    FDataSource.RemoveFreeNotification(tv);
  FDataSource:=Value;
  if Assigned(Value) and Assigned(tv) then
    Value.FreeNotification(tv);
  RefreshTreeView;
end;

procedure TDataItemDef.SetListAllRecords(Value: boolean);
begin
  FListAllRecords:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetShowTitle(Value: boolean);
begin
  FShowTitle:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetExpandNode(Value: boolean);
begin
  FExpandNode:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetTitleText(const Value: string);
begin
  FTitleText:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetTitleImageIndex(Value: integer);
begin
  FTitleImageIndex:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetTitleBkColor(Value: TColor);
begin
  FTitleBkColor:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetTitleFontColor(Value: TColor);
begin
  FTitleFontColor:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetTitleStyle(Value: TFontStyles);
begin
  FTitleStyle:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetRecordTextMask(const Value: string);
begin
  FRecordTextMask:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetRecordImageIndex(Value: integer);
begin
  FRecordImageIndex:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetRecordBkColor(Value: TColor);
begin
  FRecordBkColor:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetRecordFontColor(Value: TColor);
begin
  FRecordFontColor:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetRecordStyle(Value: TFontStyles);
begin
  FRecordStyle:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetFillLeftBorder(Value: integer);
begin
  FFillLeftBorder:=Value;
  RefreshTreeView;
end;

procedure TDataItemDef.SetFillRightBorder(Value: integer);
begin
  FFillRightBorder:=Value;
  RefreshTreeView;
end;

function TDataItemDef.GetTreeView: TrDBTreeView;
begin
  if Assigned(Collection) and (Collection is TDataItemDefs) then
    Result:=TDataItemDefs(Collection).FrDBTreeView
  else
    Result:=nil;
end;

procedure TDataItemDef.RefreshTreeView;
var
  tv: TrDBTreeView;
begin
  tv:=GetTreeView;
  if Assigned(tv) then
    tv.RefreshData(true);
end;

// ==============================================================================
// TDataItemDefs class
constructor TDataItemDefs.Create(rDBTreeView: TrDBTreeView; DataItemDefClass: TDataItemDefClass);
begin
  inherited Create(DataItemDefClass);
  FrDBTreeView:=rDBTreeView;
end;

function TDataItemDefs.Add: TDataItemDef;
begin
  BeginUpdate;
  Result:=TDataItemDef( inherited Add);
  EndUpdate;
end;

procedure TrDBTreeView.SetDataItemDefs(Value: TDataItemDefs);
begin
  FDataItemDefs.Assign(Value);
end;

function TDataItemDefs.GetDataItemDef(Index: integer): TDataItemDef;
begin
  Result:=TDataItemDef( inherited Items[Index]);
end;

procedure TDataItemDefs.SetDataItemDef(Index: integer; Value: TDataItemDef);
begin
  Items[Index].Assign(Value);
end;

function TDataItemDefs.GetOwner: TPersistent;
begin
  Result:=FrDBTreeView;
end;

procedure TDataItemDefs.Update(Item: TCollectionItem);
begin
  if (FrDBTreeView = nil) or (csLoading in FrDBTreeView.ComponentState) then
    Exit;
  FrDBTreeView.RefreshData(true);
end;

// ==============================================================================
// TrDBRecordLink class
constructor TrDBRecordLink.Create(rDBTreeView: TrDBTreeView);
begin
  inherited Create;
  FrDBTreeView:=rDBTreeView;
  VisualControl:=true;
end;

procedure TrDBRecordLink.ActiveChanged;
begin
  inherited ActiveChanged;
  // if Active then
  FrDBTreeView.RefreshData(true);
end;

procedure TrDBRecordLink.DataSetChanged;
begin
  inherited DataSetChanged;
  FrDBTreeView.RefreshData(true);
end;

procedure TrDBRecordLink.EditingChanged;
begin
  inherited EditingChanged;
end;

procedure TrDBRecordLink.UpdateData;
begin
  inherited UpdateData;
  // not needed FrDBTreeView.RefreshData;
end;

// ==============================================================================
// TrTreeNode class
constructor TrTreeNode.Create(AOwner: TTreeNodes);
begin
  inherited Create(AOwner);
  NodeObject:=TNodeObject.Create;
  NodeObject.IsInitialized:=false;
  // ShowMessage('Node created');
end;

destructor TrTreeNode.Destroy;
begin
  // ShowMessage('Node destroyed');
  if Assigned(NodeObject) then
    try
      FreeAndNil(NodeObject);
    except
    end;
  inherited Destroy;
end;

// ==============================================================================
// TrDBTreeView class
constructor TrDBTreeView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FUpdateCount:=0;
  FLoadFirstLevelOnly:=false;
  IsRefreshing:=false;
  FDataItemDesignTime:=true;
  FDataLink:=TrDBRecordLink.Create(self);
  FDataItemDefs:=TDataItemDefs.Create(self, TDataItemDef);
  lastEditedNode:=nil;
  FScrollTimer:=TTimer.Create(self);
  FScrollTimer.Enabled:=false;
  FScrollTimer.Interval:=200;
  FScrollTimer.OnTimer:=IntScrollTimer;
  FLoadDataTimer:=TTimer.Create(self);
  FLoadDataTimer.Enabled:=false;
  FLoadDataTimer.Interval:=0;
  FLoadDataTimer.OnTimer:=IntLoadTimer;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBTreeView.Destroy;
begin
  FUpdateCount:=1;
  Items.Clear;
  FreeAndNil(FDataItemDefs);
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBTreeView.Loaded;
begin
  inherited;
  oldTVEditedEvent:=OnEdited;
  OnEdited:=IntTVEditedEvent;
  // ShowMessage('Loaded event');
  RefreshData(false);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBTreeView.Notification(AComponent: TComponent; Operation: TOperation);
  procedure RemoveDataSourceLink(ItemDefs: TDataItemDefs);
  var
    i: integer;
  begin
    for i:=0 to ItemDefs.Count - 1 do
    begin
      if ItemDefs[i].DataSource = AComponent then
        ItemDefs[i].DataSource:=nil;
      RemoveDataSourceLink(ItemDefs[i].FSubDataItemDefs);
    end;
  end;

begin
  inherited Notification(AComponent, Operation);
  // check all linked SubDataSources and remove DataSource
  if (Operation = opRemove) and (AComponent is TDataSource) and Assigned(FDataItemDefs) then
    RemoveDataSourceLink(FDataItemDefs);
end;

procedure TrDBTreeView.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TrDBTreeView.EndUpdate;
begin
  Dec(FUpdateCount);
  RefreshData(true);
end;

function TrDBTreeView.CreateNode: TTreeNode;
var
  LClass: TTreeNodeClass;
begin
  LClass:=TrTreeNode;
  if Assigned(OnCreateNodeClass) then
    OnCreateNodeClass(self, LClass);
  Result:=LClass.Create(Items);
end;

function TrDBTreeView.GetMasterDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBTreeView.SetMasterDataSource(Value: TDataSource);
begin
  FDataLink.DataSource:=Value;
  // if Assigned(FDataLink.DataSource) then RefreshData;
end;

procedure TrDBTreeView.SetDataItemDesignTime(Value: boolean);
begin
  FDataItemDesignTime:=Value;
  RefreshData(true);
end;

procedure TrDBTreeView.IntLoadTimer(Sender: TObject);
begin
  FLoadDataTimer.Enabled:=false;
  RefreshData(false);
end;

procedure TrDBTreeView.SetImageIndex(Node: TTreeNode; idx: integer);
begin
  Node.ImageIndex:=idx;
  Node.SelectedIndex:=Node.ImageIndex;
  Node.StateIndex:=Node.ImageIndex;
  Node.ExpandedImageIndex:=Node.ImageIndex;
end;

procedure TrDBTreeView.ShowSubDataSources(ItemDef: TDataItemDef; Node: TTreeNode; Expanding: boolean);
var
  i: integer;
begin
  // recurse LoadDataItemDef for SubDataItems
  for i:=0 to ItemDef.FSubDataItemDefs.Count - 1 do
    LoadDataItemDef(ItemDef, ItemDef.FSubDataItemDefs.Items[i], Node, Expanding);
  if ItemDef.FExpandNode and Assigned(Node) then
    Node.Expand(false);
end;

function TrDBTreeView.AddItem(ParentItemDef, ItemDef: TDataItemDef; ParentNode: TTreeNode; ShowSubData, IsTitle, Expanding: boolean; ConstantText: string = '')
  : TTreeNode;
var
  m: string;
  tn: TrTreeNode;
  ds: TDataset;
begin
  tn:=TrTreeNode(Items.AddChild(ParentNode, ''));
  // copy node object to Data to keep backward compability
  if not(csDesigning in ComponentState) then
    tn.Data:=tn.NodeObject;
  // init node object structure
  tn.NodeObject.IsInitialized:=true;
  tn.NodeObject.ItemChanged:=false;
  tn.NodeObject.ItemMoved:=false;
  tn.NodeObject.NodeIsLoaded:=Expanding or not FLoadFirstLevelOnly;
  tn.NodeObject.ParentDataItemDef:=ParentItemDef;
  tn.NodeObject.DataItemDef:=ItemDef;
  tn.NodeObject.IsTitle:=IsTitle;
  if IsTitle then
  begin
    tn.NodeObject.BkColor:=ItemDef.TitleBkColor;
    tn.NodeObject.FontColor:=ItemDef.TitleFontColor;
    tn.NodeObject.Style:=ItemDef.TitleStyle;
  end
  else
  begin
    tn.NodeObject.BkColor:=ItemDef.RecordBkColor;
    tn.NodeObject.FontColor:=ItemDef.RecordFontColor;
    tn.NodeObject.ChangedFontColor:=ItemDef.RecordChangedFontColor;
    tn.NodeObject.Style:=ItemDef.RecordStyle;
  end;

  ds:=nil;
  if Assigned(ItemDef.DataSource) then
    ds:=ItemDef.DataSource.Dataset;
  if Assigned(ds) and ds.Active and not IsTitle then
  begin
    tn.NodeObject.KeyField:=ds.FindField(ItemDef.KeyFieldName);
    tn.NodeObject.RecNo:=ds.RecNo;
    tn.NodeObject.IsEmptyDataset:=ds.IsEmpty and not IsTitle;
  end
  else
  begin
    tn.NodeObject.KeyField:=nil;
    tn.NodeObject.RecNo:=-1;
    tn.NodeObject.IsEmptyDataset:=false;
  end;
  if Assigned(tn.NodeObject.KeyField) then
    tn.NodeObject.KeyValue:=tn.NodeObject.KeyField.AsVariant
  else
    tn.NodeObject.KeyValue:=varEmpty;

  // add node
  if ConstantText = '' then
  begin
    if IsTitle then
      m:=ItemDef.FTitleText
    else
      m:=ItemDef.FRecordTextMask;
    if Assigned(ds) and ds.Active then
      m:=FillTextMask(m, ds);
  end
  else
    m:=ConstantText;
  tn.Text:=m;

  if IsTitle then
    SetImageIndex(tn, ItemDef.FTitleImageIndex)
  else
  begin
    SetImageIndex(tn, ItemDef.FRecordImageIndex);
    if ItemDef.FRecordSelectedImageIndex >= 0 then
      tn.SelectedIndex:=ItemDef.FRecordSelectedImageIndex;
    if ItemDef.FRecordExpandedImageIndex >= 0 then
      tn.ExpandedImageIndex:=ItemDef.FRecordExpandedImageIndex;
    if ItemDef.FRecordStateImageIndex >= 0 then
      tn.StateIndex:=ItemDef.FRecordStateImageIndex;
  end;
  if Assigned(FOnItemObjectDefined) then
    FOnItemObjectDefined(self, tn, tn.NodeObject, ds);
  if ShowSubData then
    ShowSubDataSources(ItemDef, tn, Expanding);
  Result:=tn;
end;

procedure TrDBTreeView.LoadDataItemDef(ParentItemDef, ItemDef: TDataItemDef; ParentNode: TTreeNode; Expanding: boolean);
var
  tn: TTreeNode;
  rec: integer;
  ds: TDataset;

  function AddTitleNode(AParent: TTreeNode): TTreeNode;
  begin
    Result:=AParent;
    if ItemDef.FShowTitle and (ItemDef.FTitleText <> '') then
      Result:=AddItem(ParentItemDef, ItemDef, AParent, false, true, Expanding);
  end;

  function AddTitleAndItem(AParent: TTreeNode; ShowSubData: boolean; ConstantText: string = ''): TTreeNode;
  begin
    AddTitleNode(AParent);
    Result:=AddItem(ParentItemDef, ItemDef, AParent, ShowSubData, false, Expanding, ConstantText);
  end;

  function SetMultilevelFitler: boolean;
  var
    KeyField, ParentFld: TField;
    pid: variant;
    fs, err: string;
  begin
    Result:=false;
    err:='';
    try
      // check Multilevel limitations - limitation for subdata already solved
      // if ItemDef.SubDataItemDefs.Count > 0 then err:='Node with MultilevelMode cannot contain SubDataItems.';
      ParentFld:=ds.FindField(ItemDef.ParentKeyFieldName);
      if ParentFld = nil then
        err:=Format('ParentKeyField "%s" not found.', [ItemDef.ParentKeyFieldName]);
      KeyField:=ds.FindField(ItemDef.KeyFieldName);
      if KeyField = nil then
        err:=Format('KeyField "%s" not found.', [ItemDef.KeyFieldName]);
      if (ItemDef.FMultilevelTopLevelFilter = '') and not(ParentFld is TNumericField) then
        err:='ParentKeyFieldName has to be numeric if MultilevelTopLevelFilter is not defined.';

      // filter records according parent node Key value
      if err = '' then
      begin
        fs:='';
        pid:=0;
        if Assigned(ParentNode) then
          pid:=TrTreeNode(ParentNode).NodeObject.KeyValue
        else
          fs:=ItemDef.FMultilevelTopLevelFilter;
        if fs = '' then
          fs:=Format('%s=%s', [ItemDef.ParentKeyFieldName, VarToStrDef(pid, '0')]);
        ds.Filter:=fs;
        ds.Filtered:=true;
      end;
    except
      on E: Exception do
        err:=E.Message;
    end;
    if err = '' then
      Result:=true
    else
      AddTitleAndItem(ParentNode, true, err);
  end;

  procedure ResetMultilevelFitler;
  begin
    ds.Filtered:=false;
    ds.Filter:='';
    ds.First;
  end;

  procedure ProcDataset;
  var
    atn: TTreeNode;
  begin
    if ds.IsEmpty then
    begin
      if ItemDef.FEmptyDatasetText <> '' then
        AddItem(ParentItemDef, ItemDef, ParentNode, false, false, Expanding, ItemDef.FEmptyDatasetText);
    end
    else
      if not ItemDef.FListAllRecords then
      begin
        if (ItemDef.FRecordTextMask = '') then
          ShowSubDataSources(ItemDef, tn, Expanding)
        else
          AddItem(ParentItemDef, ItemDef, ParentNode, true, false, Expanding);
      end
      else
      begin
        // ds.DisableControls = cannot use, master-detail is not working if disabled
        if FLoadFirstLevelOnly and not Expanding and Assigned(ParentNode) then
        begin
          // add dummy item to allow expanding, items will be loaded after expand
          AddItem(ParentItemDef, ItemDef, ParentNode, false, false, Expanding, 'to be loaded...');
        end
        else
        begin
          // set filter for multilevel
          if ItemDef.FMultilevelMode then
            if not SetMultilevelFitler then
              Exit;
          // proc all records
          ds.First;
          rec:=0;
          while not ds.Eof do
          begin
            // check max number of rec
            if (ItemDef.FMaxRecords > 0) and (ItemDef.MaxRecords <= rec) then
            begin
              AddItem(ParentItemDef, ItemDef, ParentNode, false, false, Expanding, '...');
              Break;
            end;
            AddItem(ParentItemDef, ItemDef, ParentNode, true, false, Expanding);
            ds.Next;
            Inc(rec);
          end;

          if ItemDef.FMultilevelMode then
          begin
            // reset filter
            ResetMultilevelFitler;
            // load next level for each sibling nodes with same KeyField
            if Assigned(ParentNode) then
              atn:=(ParentNode.getFirstChild)
            else
              atn:=(Items.GetFirstNode);
            while Assigned(atn) do
            begin
              if not TrTreeNode(atn).NodeObject.IsTitle and (TrTreeNode(atn).NodeObject.KeyField = ds.FindField(ItemDef.KeyFieldName)) then
              begin
                LoadDataItemDef(ParentItemDef, ItemDef, atn, Expanding);
                if ItemDef.ExpandNode then
                  atn.Expand(true);
              end;
              atn:=atn.getNextSibling;
            end;
          end;
        end;
        // ds.EnableControls;
      end;
  end;

begin
  tn:=nil;
  ds:=nil;
  if not Assigned(ItemDef.DataSource) then
  begin
    // without dataset - add title only and subdata
    tn:=AddTitleNode(ParentNode);
    ShowSubDataSources(ItemDef, tn, Expanding);
  end
  else
  begin
    ds:=ItemDef.DataSource.Dataset;
    if not Assigned(ds) then
      AddTitleAndItem(ParentNode, false, 'Dataset not defined.')
    else
      if (csDesigning in ComponentState) then
      begin
        // show definition in design mode
        tn:=AddTitleNode(ParentNode);
        if (not ItemDef.FListAllRecords) and (ItemDef.FRecordTextMask = '') then
          ShowSubDataSources(ItemDef, tn, Expanding)
        else
          AddItem(ParentItemDef, ItemDef, ParentNode, true, false, Expanding, ItemDef.FRecordTextMask);
      end
      else
        if not ds.Active then
          AddTitleAndItem(ParentNode, true, 'Dataset is closed.')
        else
        begin
          // add title if exist
          AddTitleNode(ParentNode);
          // show dataset records
          ProcDataset;
        end;
  end;
end;

procedure TrDBTreeView.RefreshData(Delayed: boolean = false; KeepSelectedNode: boolean = false);
var
  i: integer;
  SelFld, RootFld: TField;
  SelVal, RootVal: variant;
  no: TNodeObject;
  tn: TTreeNode;
begin
  if (csLoading in ComponentState) or IsRefreshing or (FUpdateCount<>0) then
    Exit;
  if Delayed and (FLoadDataDelay > 0) then
  begin
    FLoadDataTimer.Enabled:=false;
    FLoadDataTimer.Interval:=FLoadDataDelay;
    FLoadDataTimer.Enabled:=true;
    Exit;
  end;

  IsRefreshing:=true;
  // store selected node
  tn:=Selected;
  no:=GetSelectedData;
  RootFld:=nil;
  SelFld:=nil;
  if KeepSelectedNode and Assigned(tn) and Assigned(no) then
  begin
    SelFld:=no.KeyField;
    SelVal:=no.KeyValue;
    // find root node
    while tn.Parent <> nil do
      tn:=tn.Parent;
    no:=TrTreeNode(tn).NodeObject;
    RootFld:=no.KeyField;
    RootVal:=no.KeyValue;
  end;

  // reaload all nodes
  Items.Clear;
  Items.BeginUpdate;
  try
    for i:=0 to FDataItemDefs.Count - 1 do
      LoadDataItemDef(nil, FDataItemDefs.Items[i], nil, false);
  finally
    Items.EndUpdate;
    IsRefreshing:=false;
  end;

  // preselect first item
  if Items.Count > 0 then
    tn:=Items[0]
  else
    tn:=nil;
  // find stored node
  if KeepSelectedNode and Assigned(SelFld) then
  begin
    // find root node
    tn:=FindNode(RootFld, RootVal);
    if Assigned(tn) and (RootFld <> SelFld) then
    begin
      // expand and find stored node
      tn.Expand(false);
      tn:=FindNode(SelFld, SelVal);
    end;
  end;
  // select node
  if Assigned(tn) then
    Select(tn);
  if (csDesigning in ComponentState) and not FDataItemDesignTime then
    Items.Clear;
end;

procedure TrDBTreeView.RefreshNode(Node: TTreeNode);
var
  tnobj: TNodeObject;
  ds: TDataset;
  m: string;
begin
  if Assigned(Node) then
  begin
    tnobj:=TrTreeNode(Node).NodeObject;
    ds:=tnobj.DataItemDef.FDataSource.Dataset;
    if Assigned(ds) and ds.Active then
    begin
      if tnobj.IsTitle then
        m:=tnobj.DataItemDef.FTitleText
      else
        m:=tnobj.DataItemDef.FRecordTextMask;
      m:=FillTextMask(m, ds);
      Node.Text:=m;
    end;
    if Assigned(FOnItemObjectDefined) then
      FOnItemObjectDefined(self, Node, tnobj, ds);
  end;
end;

procedure TrDBTreeView.CreateChildNodes(ParentNode: TTreeNode; MultilevelNode, RemoveCurrentNodes: boolean);
var
  tnobj: TNodeObject;
begin
  if Assigned(ParentNode) then
  begin
    tnobj:=TrTreeNode(ParentNode).NodeObject;
    if RemoveCurrentNodes then
      ParentNode.DeleteChildren;
    if MultilevelNode then
    begin
      LoadDataItemDef(tnobj.ParentDataItemDef, tnobj.DataItemDef, ParentNode, true);
      if tnobj.DataItemDef.FExpandNode and Assigned(ParentNode) then
        ParentNode.Expand(false);
    end
    else
      ShowSubDataSources(tnobj.DataItemDef, ParentNode, true);
  end;
end;

procedure TrDBTreeView.CreateSiblingNode(SiblingNode: TTreeNode);
var
  tnobj: TNodeObject;
begin
  if Assigned(SiblingNode) then
  begin
    tnobj:=TrTreeNode(SiblingNode).NodeObject;
    LoadDataItemDef(tnobj.ParentDataItemDef, tnobj.DataItemDef, SiblingNode.Parent, true);
  end;
end;

function TrDBTreeView.FindNode(KeyField: TField; KeyValue: variant): TTreeNode;
var
  i: integer;
  tn: TrTreeNode;
begin
  Result:=nil;
  if IsRefreshing then
    Exit;
  for i:=0 to Items.Count - 1 do
  begin
    tn:=TrTreeNode(Items[i]);
    if tn.NodeObject.IsInitialized then
    begin
      if not(tn.NodeObject.IsTitle) and (tn.NodeObject.KeyField = KeyField) and (tn.NodeObject.KeyValue = KeyValue) then
      begin
        Result:=tn;
        Break;
      end;
    end;
  end;
end;

procedure TrDBTreeView.ProcessAllItems(UseChanged: boolean; UseMoved: boolean; CallBackProc: TTVChangedEvent; ResetState: boolean);
var
  i: integer;
  tn: TrTreeNode;
begin
  for i:=0 to Items.Count - 1 do
  begin
    tn:=TrTreeNode(Items[i]);
    if tn.NodeObject.IsInitialized then
    begin
      if (UseChanged and tn.NodeObject.ItemChanged) or (UseMoved and tn.NodeObject.ItemMoved) then
      begin
        CallBackProc(self, tn);
        if ResetState then
        begin
          if UseChanged then
            tn.NodeObject.ItemChanged:=false;
          if UseMoved then
            tn.NodeObject.ItemMoved:=false;
        end;
      end;
    end;
  end;
  Invalidate;
end;

function TrDBTreeView.CanExpand(Node: TTreeNode): boolean;
var
  i: integer;
  ParentNodeObject: TNodeObject;
  recfound: boolean;
begin
  if not(csDesigning in ComponentState) and FLoadFirstLevelOnly and Assigned(Node) and TrTreeNode(Node).NodeObject.IsInitialized then
  begin
    ParentNodeObject:=TrTreeNode(Node).NodeObject;
    // check if loading on expand is necessary
    if not ParentNodeObject.NodeIsLoaded then
    begin
      IsRefreshing:=true;
      Items.BeginUpdate;
      try
        // locate master record by KeyField or RecNo
        if (ParentNodeObject.KeyField <> nil) then
          recfound:=LocateRecord(ParentNodeObject.KeyField, ParentNodeObject.KeyValue)
        else
          try
            ParentNodeObject.DataItemDef.DataSource.Dataset.RecNo:=ParentNodeObject.RecNo;
            recfound:=true;
          except
            recfound:=false;
          end;
        if recfound then
        begin
          // delete dummy items
          Node.DeleteChildren;
          // load all items for node
          if ParentNodeObject.DataItemDef.FMultilevelMode then
            LoadDataItemDef(ParentNodeObject.ParentDataItemDef, ParentNodeObject.DataItemDef, Node, true)
          else
            for i:=0 to ParentNodeObject.DataItemDef.FSubDataItemDefs.Count - 1 do
              LoadDataItemDef(ParentNodeObject.DataItemDef, ParentNodeObject.DataItemDef.FSubDataItemDefs.Items[i], Node, true);
          // set as loaded
          ParentNodeObject.NodeIsLoaded:=true;
        end;
      finally
        Items.EndUpdate;
        IsRefreshing:=false;
      end;
    end;
  end;
  Result:= inherited CanExpand(Node);
end;

function TrDBTreeView.IsCustomDrawn(Target: TCustomDrawTarget; Stage: TCustomDrawStage): boolean;
begin
  Result:= inherited;
  // force CustomDraw for Items
  if Target = dtItem then
    Result:=true;
end;

function TrDBTreeView.CustomDrawItem(Node: TTreeNode; State: TCustomDrawState; Stage: TCustomDrawStage; var PaintImages: boolean): boolean;
var
  NodeRect: TRect;
  IsSelected: boolean;
  tnobj: TNodeObject;
  FontColor, BkColor: TColor;

  function SetColorIfNone(AColor, AColorIfNone: TColor): TColor;
  begin
    if AColor = clNone then
      Result:=AColorIfNone
    else
      Result:=AColor;
  end;

begin
  tnobj:=TrTreeNode(Node).NodeObject;
  // skip custom draw for saved node data from DFM
  if tnobj.IsInitialized then
  try
    // IsSelected:=(cdsSelected in State) or (cdsHot in State) or (cdsDropHilited in State);
    IsSelected:=(cdsSelected in State);
    NodeRect:=Node.DisplayRect(false);
    if Assigned(tnobj.DataItemDef) then
    begin
      NodeRect.Left:=NodeRect.Left + tnobj.DataItemDef.FFillLeftBorder;
      NodeRect.Right:=NodeRect.Right - tnobj.DataItemDef.FFillRightBorder;
    end;

    if (Stage = cdPrePaint) then
    begin
      // set Font style + color
      Canvas.Font.Style:=[];
      FontColor:=Canvas.Font.Color;
      if fsBold in tnobj.Style then
        Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
      if fsItalic in tnobj.Style then
        Canvas.Font.Style:=Canvas.Font.Style + [fsItalic];
      if not(cdsFocused in State) or not IsSelected then
        Canvas.Font.Color:=SetColorIfNone(tnobj.FontColor, FontColor);
      if tnobj.ItemChanged or tnobj.ItemMoved then
        Canvas.Font.Color:=SetColorIfNone(tnobj.ChangedFontColor, FontColor);
      if tnobj.IsEmptyDataset then
        Canvas.Font.Color:=clGrayText;

      // set Brush and draw bk
      BkColor:=Canvas.Brush.Color;
      if not tnobj.IsEmptyDataset then
        Canvas.Brush.Color:=SetColorIfNone(tnobj.BkColor, BkColor);
      if (Canvas.Brush.Color <> BkColor) and (not(cdsFocused in State) or not IsSelected) then
        Canvas.FillRect(NodeRect);
      if not GetThemeEnabled then
        Canvas.Brush.Color:=BkColor;
    end;

    if (Stage = cdPostPaint) then
    begin
      // draw line if Underline
      if fsUnderline in tnobj.Style then
      begin
        Canvas.Pen.Color:=tnobj.FontColor;
        Canvas.MoveTo(NodeRect.Left, NodeRect.Bottom - 1);
        Canvas.LineTo(NodeRect.Right, NodeRect.Bottom - 1);
      end;
    end;
  except
    Beep;
  end;
  Result:= inherited CustomDrawItem(Node, State, Stage, PaintImages);
end;

procedure TrDBTreeView.WMHScroll(var Message: TWMHScroll);
begin
  // refresh all items - own drawn items is not scrolled properly
  inherited;
  Invalidate;
end;

procedure TrDBTreeView.WMMouseWheel(var Message: TWMMouseWheel);
begin
  inherited;
  // refresh all items - own drawn items is not Wheeled properly if VScroll is not shown
  if (GetWindowlong(Handle, GWL_STYLE) and WS_HSCROLL) <> 0 then
    if ((GetWindowlong(Handle, GWL_STYLE) and WS_VSCROLL) = 0) then
      Invalidate;
end;

procedure TrDBTreeView.Change(Node: TTreeNode);
var
  OldOnChange: TTVChangedEvent;
begin
  // skip standard OnChange and call only if not refreshing
  OldOnChange:=OnChange;
  OnChange:=nil;
  inherited;
  OnChange:=OldOnChange;
  if (not IsRefreshing) and not(csLoading in ComponentState) then
    if Assigned(OnChange) then
      OnChange(self, Node);
end;

function TrDBTreeView.GetSelectedData: TNodeObject;
begin
  Result:=nil;
  if Assigned(Selected) then
    Result:=TrTreeNode(Selected).NodeObject;
end;

function TrDBTreeView.GetNodeData(NodeAbsoluteIndex: integer): TNodeObject;
begin
  Result:=nil;
  if NodeAbsoluteIndex in [0 .. Items.Count - 1] then
    Result:=TrTreeNode(Items[NodeAbsoluteIndex]).NodeObject;
end;

function TrDBTreeView.CanEdit(Node: TTreeNode): boolean;
var
  tnobj: TNodeObject;
begin
  // define if item can be edited - disable for Title and if CanEditText=false
  tnobj:=TrTreeNode(Node).NodeObject;
  Result:=tnobj.IsInitialized and not tnobj.IsTitle and tnobj.DataItemDef.CanEditText;
  if Assigned(OnEditing) then
    OnEditing(self, Node, Result);
end;

procedure TrDBTreeView.Edit(const Item: tagTVITEMW);
var
  tnobj: TNodeObject;
begin
  lastEditedNode:=nil;
  inherited;
  if Assigned(lastEditedNode) then
  begin
    tnobj:=TrTreeNode(lastEditedNode).NodeObject;
    if tnobj.IsInitialized then
      tnobj.ItemChanged:=true;
    if Assigned(FOnItemEdited) then
      FOnItemEdited(self, lastEditedNode);
  end;
end;

procedure TrDBTreeView.IntTVEditedEvent(Sender: TObject; Node: TTreeNode; var S: string);
begin
  if Assigned(oldTVEditedEvent) then
    oldTVEditedEvent(Sender, Node, S);
  lastEditedNode:=Node;
end;

procedure TrDBTreeView.IntScrollTimer(Sender: TObject);
begin
  if not Mouse.IsDragging then
    FScrollTimer.Enabled:=false
  else
    FScrollTimer.Interval:=25 + (Abs(FScrollTimer.Tag)) * 10;
  if FScrollTimer.Tag > 0 then
    SendMessage(Handle, WM_VSCROLL, SB_LINEUP, 0)
  else
    SendMessage(Handle, WM_VSCROLL, SB_LINEDOWN, 0);
end;

procedure TrDBTreeView.DragOver(Source: TObject; X: integer; Y: integer; State: TDragState; var Accept: boolean);
var
  stn, sptn, ttn: TrTreeNode;
begin
  // define it item can be dropped according DragDropMode of source item
  Accept:=false;
  ttn:=TrTreeNode(GetNodeAt(X, Y));
  stn:=nil;
  if (Source is TrDBTreeView) then
    stn:=TrTreeNode(TrDBTreeView(Source).Selected);
  sptn:=nil;
  if Assigned(stn) then
    sptn:=TrTreeNode(stn.Parent);
  if Assigned(stn) and stn.NodeObject.IsInitialized and Assigned(ttn) and ttn.NodeObject.IsInitialized then
    try
      Accept:=((dnoOtherControl in stn.NodeObject.DataItemDef.DragDropMode) or (Source = self)) and
        ((dnoSameParent in stn.NodeObject.DataItemDef.DragDropMode) or (stn.Parent <> ttn)) and
        ((dnoAllLevels in stn.NodeObject.DataItemDef.DragDropMode) or ((dnoSameLevel in stn.NodeObject.DataItemDef.DragDropMode) and (stn.Level - 1 = ttn.Level)
          ) or ((dnoSameParentType in stn.NodeObject.DataItemDef.DragDropMode) and Assigned(sptn) and (sptn.NodeObject.KeyField = ttn.NodeObject.KeyField)));
    except
    end;
  if Assigned(OnDragOver) then
    OnDragOver(self, Source, X, Y, State, Accept);

  if Y < ScrollBorder then
    FScrollTimer.Tag:=Max(Y, 1)
  else
    if (Height - Y) < ScrollBorder then
      FScrollTimer.Tag:=Min(Y - Height + 2, -1)
    else
      FScrollTimer.Tag:=0;
  FScrollTimer.Enabled:=FScrollTimer.Tag <> 0;
end;

procedure TrDBTreeView.DragDrop(Source: TObject; X: integer; Y: integer);
var
  ttn: TrTreeNode;
  i: integer;
  L: TList;

  procedure MoveItem(stn: TrTreeNode);
  begin
    if Assigned(stn) and stn.NodeObject.IsInitialized and Assigned(ttn) and ttn.NodeObject.IsInitialized then
    begin
      if (dnoMoveItem in stn.NodeObject.DataItemDef.DragDropMode) then
      begin
        stn.MoveTo(ttn, naAddChild);
        stn.NodeObject.ItemMoved:=true;
        InvalidateSelectionsRects;
        if Assigned(FOnItemMoved) then
          FOnItemMoved(self, ttn);
      end;
    end;
  end;

begin
  // move item if dnoMoveItem is set for source item
  if Source <> self then
    Exit;
  ttn:=TrTreeNode(GetNodeAt(X, Y));
  if not Assigned(ttn) then
    Exit;
  ttn.Expand(false);
  if SelectionCount < 2 then
    MoveItem(TrTreeNode(Selected))
  else
  begin
    L:=TList.Create;
    try
      for i:=0 to SelectionCount - 1 do
        L.Add(Selections[i]);
      // select master to protect child expand
      Select(ttn);
      // keep node orders - bug in standard comp
      MoveItem(L[0]);
      for i:=L.Count - 1 downto 1 do
        MoveItem(L[i]);
      // reselect moved nodes
      for i:=0 to L.Count - 1 do
        self.Select(L[i], [ssCtrl]);
    finally
      FreeAndNil(L);
    end;
  end;
  inherited;
end;

procedure TrDBTreeView.Collapse(Node: TTreeNode);
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  inherited;
end;

procedure TrDBTreeView.Expand(Node: TTreeNode);
begin
{$IFDEF DEMO}
  // ShowDemoMsg;
{$ENDIF}
  inherited;
end;

end.
