unit rHTMLLabel;

{<
  rHTMLLabel unit contains TrHTMLLabel - enhanced TLabel component which display caption
  with more font styles and colors according HTML tags in caption string,
  it can also display hyperlinks (anchor) and can handle click on it.

  Supported HTML tags:
  <B>, </B>, <I>, </I>, <U>, </U>, <BIG>, </BIG>, <SMALL>, </SMALL>,
  <A HREF=target>  for automatic launching of URL target (http, mailto, file etc.),
  <A NAME=param> for calling of user defined events, </A>,
  <COLOR=Red>, <COLOR=#FFFFFF> (RGB value), </COLOR> for setting of font color,
  <FONT COLOR=Red>, <FONT COLOR=#FFFFFF> (RGB value), </FONT> for setting of font color,
  <BGCOLOR=Gray>, </BGCOLOR> for setting of background color
  <BR> for line break,
  <TAB> for tabs grid
  <INDENT> for defining left indent for wraped lines
  <VAR>, </VAR> - defines parameter
  <NOWORDWRAP> - change label properties  WordWrap to false
  (force drawing in one line)

  Caption can contain parameters (between <VAR> and </VAR> tags), for each
  parameter OnGetParameter events is called .

  TrHTMLLabel can display caption with 3D effect and beveled.
  For long caption can be used new memo property - CaptionStrings.

  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2006)
  @lastmod(Mar 2020)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, StdCtrls, Forms,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} Dialogs, ShellApi, Variants,
  RosiCompGlobalCfg;

type
  { enhanced TLabel component which provides functionality
    for displaying caption with more font styles and colors according
    HTML tags in caption string }
  TrHTMLLabel = class(TLabel)
  private
    ACaptionAsHint: boolean;
    AColorEnabled: boolean;
    BeforeAnchorFont: TFontRecall;
    OriginalBrush: TBrush;
    actAnchorInfo: TAnchorInfo;
    actAnchorItem: integer;
    AnchorInfoList: TList;
    IsAnchorDraw: boolean;
    LastAnchorText: string;
    FCaptionStrings: TStrings;
    FCaptionSkipEnter: boolean;
    FCaptionSimpleText: boolean;
    FFontColor: TColor;
    FShowText3D: boolean;
    FShadowColor: TColor;
    FAnchorColor: TColor;
    FShadowSize: integer;
    FShowBeveled: boolean;
    FInnerSpace: byte;
    FTabGridWidth: byte;
    FOnAnchorClick: TAnchorClickEvent;
    FOnGetParameter: TGetParameter;
    procedure FCaptionStringsChanged(Sender: TObject);
    procedure FreeAnchorInfo;
    function FindFirstFrm(const s: string): integer;
    function ProcFrm(var s: string; Calc: boolean): integer;
    procedure SetCaptionStrings(Value: TStrings);
    procedure SetCaptionSkipEnter(Value: boolean);
    procedure SetCaptionSimpleText(Value: boolean);
    procedure SetFontColor(Value: TColor);
    procedure SetAnchorColor(Value: TColor);
    procedure SetShowText3D(Value: boolean);
    procedure SetShadowColor(Value: TColor);
    procedure SetShadowSize(Value: integer);
    procedure SetShowBeveled(Value: boolean);
    procedure SetInnerSpace(Value: byte);
    procedure SetTabGridWidth(Value: byte);
    procedure CMMouseLeave(var AMsg: TMessage); message CM_MOUSELEAVE;
  protected
    procedure Loaded; override;
    function GetLabelText: string; override;
    procedure SetAutoSize(Value: boolean); override;
    procedure DoDrawText(var Rect: TRect; Flags: Longint); override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure Click; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    { CaptionStrings are used as Caption, can contain more lines
      and can be edited easily in design time }
    property CaptionStrings: TStrings read FCaptionStrings write SetCaptionStrings;
    { Define whether line breaks in CaptionStrings will be ignored }
    property CaptionSkipEnter: boolean read FCaptionSkipEnter write SetCaptionSkipEnter default false;
    { Define whether CaptionStrings is shown as is without HTML processing }
    property CaptionSimpleText: boolean read FCaptionSimpleText write SetCaptionSimpleText default false;
    { Define fotn color instead of font - it allows to use parent font }
    property FontColor: TColor read FFontColor write SetFontColor default clNone;
    { Define color of shadow in case of 3D efect }
    property AnchorColor: TColor read FAnchorColor write SetAnchorColor default clNone;
    { Define whether label has 3D efect, see ShadowColor and ShadowSize too }
    property ShowText3D: boolean read FShowText3D write SetShowText3D default false;
    { Define color of shadow in case of 3D efect }
    property ShadowColor: TColor read FShadowColor write SetShadowColor default clBtnShadow;
    { Define size of shadow in case of 3D efect }
    property ShadowSize: integer read FShadowSize write SetShadowSize default 1;
    { Define whether bevel is drawn arounf the caption }
    property ShowBeveled: boolean read FShowBeveled write SetShowBeveled default false;
    { Define space between border and text }
    property InnerSpace: byte read FInnerSpace write SetInnerSpace default 0;
    { Define size of empty space if TAB key or tag <TAB> is used in text }
    property TabGridWidth: byte read FTabGridWidth write SetTabGridWidth default 30;
    { Define event when user click to Anchor }
    property OnAnchorClick: TAnchorClickEvent read FOnAnchorClick write FOnAnchorClick;
    { Define event when user click to Anchor defined by tag <A NAME=paramname ... }
    property OnGetParameter: TGetParameter read FOnGetParameter write FOnGetParameter;
  end;

  { enhanced TLabel component which provides functionality
    for launching of any URL target }
  TrURLLabel = class(TLabel)
  private
    FFontColor: TColor;
    FHotColor: TColor;
    FURLCommand: string;
    MouseOver: boolean;
    procedure SetFontColor(Value: TColor);
    procedure SetHotColor(Value: TColor);
    procedure SetURLCommand(const Value: string); {< @exclude }
    procedure CMMouseEnter(var AMsg: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var AMsg: TMessage); message CM_MOUSELEAVE;
  protected
    procedure DoDrawText(var Rect: TRect; Flags: Longint); override;
    procedure Loaded; override;
    procedure Click; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  published
    property FontColor: TColor read FFontColor write SetFontColor default clHighLight;
    property HotColor: TColor read FHotColor write SetHotColor default clHotLight;
    { define url command if caption contain any other text }
    property URLCommand: string read FURLCommand write SetURLCommand;
  end;

procedure Register; {< @exclude }

const
  FrmNoWordWrap = '<NOWORDWRAP>';

implementation

uses rstring, rtool;
{$R *.res}

const
  FrmStrCount = 20;
  FrmStr: array [1 .. FrmStrCount] of string = ('<B>', '</B>', '<I>', '</I>', '<U>', '</U>', '<BIG>', '</BIG>', '<SMALL>', '</SMALL>', '<A', '</A>', '<COLOR=',
    '</COLOR>', '<FONT COLOR=', '</FONT>', '<BGCOLOR=', '</BGCOLOR>', '<TAB>', '<INDENT>');

  FrmAnchorStartNum = 11;
  FrmAnchorStopNum = 12;
  FrmColor1StartNum = 13;
  FrmColor1StopNum = 14;
  FrmColor2StartNum = 15;
  FrmColor2StopNum = 16;
  FrmBgColorStartNum = 17;
  FrmBgColorStopNum = 18;
  FrmTabNum = 19;
  FrmIndentNum = 20;

  FrmCR = '<BR>';
  FrmStop = '>';
  FrmParamStart = '<VAR>';
  FrmParamStop = '</VAR>';
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi', [TrHTMLLabel, TrURLLabel]);
end;

constructor TrHTMLLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FTabGridWidth:=30;
  FCaptionStrings:=TStringList.Create;
  TStringList(FCaptionStrings).OnChange:=FCaptionStringsChanged;
  AnchorInfoList:=TList.Create;
  actAnchorItem:=-1;
  FFontColor:=clNone;
  FAnchorColor:=clNone;
  FShadowColor:=clBtnShadow;
  FShadowSize:=1;
  BeforeAnchorFont:=nil;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrHTMLLabel.Destroy;
begin
  FreeAnchorInfo;
  FreeAndNil(FCaptionStrings);
  FreeAndNil(AnchorInfoList);
  inherited Destroy;
end;

procedure TrHTMLLabel.Loaded;
begin
  inherited Loaded;
  ACaptionAsHint:=Hint = '';
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrHTMLLabel.FCaptionStringsChanged(Sender: TObject);
begin
  AdjustBounds;
  Invalidate;
end;

procedure TrHTMLLabel.FreeAnchorInfo;
var
  i: integer;
begin
  for i:=0 to AnchorInfoList.Count - 1 do
  begin
    actAnchorInfo:=AnchorInfoList.Items[i];
    FreeAndNil(actAnchorInfo);
  end;
  AnchorInfoList.Clear;
  IsAnchorDraw:=false;
  LastAnchorText:='';
end;

procedure TrHTMLLabel.SetCaptionStrings(Value: TStrings);
begin
  FCaptionStrings.Assign(Value);
  Caption:='';
end;

procedure TrHTMLLabel.SetCaptionSkipEnter(Value: boolean);
begin
  FCaptionSkipEnter:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetCaptionSimpleText(Value: boolean);
begin
  FCaptionSimpleText:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetFontColor(Value: TColor);
begin
  FFontColor:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetAnchorColor(Value: TColor);
begin
  FAnchorColor:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetShowText3D(Value: boolean);
begin
  FShowText3D:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetShadowColor(Value: TColor);
begin
  FShadowColor:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetShadowSize(Value: integer);
begin
  FShadowSize:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetShowBeveled(Value: boolean);
begin
  FShowBeveled:=Value;
  Invalidate;
end;

procedure TrHTMLLabel.SetInnerSpace(Value: byte);
begin
  FInnerSpace:=Value;
  AdjustBounds;
  Invalidate;
end;

procedure TrHTMLLabel.SetTabGridWidth(Value: byte);
begin
  if Value < 1 then
    Value:=1;
  FTabGridWidth:=Value;
  AdjustBounds;
  Invalidate;
end;

procedure TrHTMLLabel.SetAutoSize(Value: boolean);
begin
  inherited SetAutoSize(Value);
end;

function TrHTMLLabel.FindFirstFrm(const s: string): integer;
var
  a, p, m: integer;
begin
  Result:=0;
  if not FCaptionSimpleText then
  begin
    p:=Length(s) + 1;
    for a:=1 to FrmStrCount do
    begin
      m:=AnsiPosEx(FrmStr[a], s, 1, false);
      if (m > 0) and (m < p) then
        p:=m;
    end;
    if p <= Length(s) then
      Result:=p;
  end;
end;

function TrHTMLLabel.ProcFrm(var s: string; Calc: boolean): integer;
var
  a: integer;

  function ExtractFrmParam(FrmStr: string): string;
  var
    i, l: integer;
  begin
    Result:='';
    i:=AnsiPosEx(FrmStop, s, 1, false);
    if i = 0 then
      Exit;
    l:=Length(FrmStr);
    Result:=copy(s, l + 1, i - l - 1);
    s:=FrmStr + copy(s, i + 1, Length(s));
  end;

  procedure ProcAnchorStop;
  begin
    LastAnchorText:='';
    if Assigned(BeforeAnchorFont) then
      FreeAndNil(BeforeAnchorFont);
    if IsAnchorDraw and not Calc then
      IsAnchorDraw:=false;
  end;

  procedure ProcAnchorStart;
  var
    temp: string;
    p: integer;
  begin
    ProcAnchorStop;
    // Find anchor info
    LastAnchorText:=copy(s, 1, AnsiPosEx(FrmStop, s, 1, false) + Length(FrmStop) - 1);
    if LastAnchorText = '' then
      Exit;
    temp:=ExtractFrmParam(FrmStr[FrmAnchorStartNum]);
    // Set anchor font and color
    BeforeAnchorFont:=TFontRecall.Create(Canvas.Font);
    Canvas.Font.Style:=Canvas.Font.Style + [fsUnderline];
    if AColorEnabled then
    begin
      if AnchorInfoList.Count = actAnchorItem then
        Canvas.Font.Color:=clHotLight
      else
        Canvas.Font.Color:=clHighLight;
      if FAnchorColor <> clNone then
        Canvas.Font.Color:=FAnchorColor;
    end;
    // Save anchor info
    if not Calc then
    begin
      IsAnchorDraw:=true;
      actAnchorInfo:=TAnchorInfo.Create;
      AnchorInfoList.Add(actAnchorInfo);
      p:=Pos('=', temp);
      if p = 0 then
      begin
        actAnchorInfo.LinkType:='';
        actAnchorInfo.LinkName:=Trim(temp);
      end
      else
      begin
        actAnchorInfo.LinkType:=AnsiUpperCase(Trim(copy(temp, 1, p - 1)));
        actAnchorInfo.LinkName:=Trim(copy(temp, p + 1, Length(temp)));
      end;
    end;
  end;

  function GetColorFromHex(HexCode: string): Longint;
  var
    Red, Blue: Longint;
  begin
    try
      Result:=StrToInt('$' + HexCode); { but bytes are backwards! }
      Red:=Result and $FF;
      Blue:=Result and $FF0000;
      Result:=(Result and $00FF00) + (Red shl 16) + (Blue shr 16);
    except
      Result:=0;
    end;
  end;

  procedure ProcColorStart(idx: integer);
  var
    temp: string;
    Color: Longint;
  begin
    temp:=ExtractFrmParam(FrmStr[idx]);
    if AColorEnabled then
    begin
      if Pos('#', temp) = 1 then
        Canvas.Font.Color:=GetColorFromHex(copy(temp, 2, 6))
      else
        if IdentToColor('cl' + temp, Color) then
          Canvas.Font.Color:=Color;
    end;
  end;

  procedure ProcColorStop;
  begin
    Canvas.Font.Color:=GetStyleFontColor_LabelText(Font.Color);
    if FFontColor <> clNone then
      Canvas.Font.Color:=FFontColor;
  end;

  procedure ProcBgColorStart(idx: integer);
  var
    temp: string;
    Color: Longint;
  begin
    temp:=ExtractFrmParam(FrmStr[idx]);
    if AColorEnabled then
    begin
      if Pos('#', temp) = 1 then
        Canvas.Brush.Color:=GetColorFromHex(copy(temp, 2, 6))
      else
        if IdentToColor('cl' + temp, Color) then
          Canvas.Brush.Color:=Color;
    end;
  end;

  procedure ProcBgColorStop;
  begin
    if Assigned(OriginalBrush) then
      Canvas.Brush.Assign(OriginalBrush);
  end;

begin
  Result:=0;
  for a:=1 to FrmStrCount do
  begin
    if AnsiPosEx(FrmStr[a], s, 1, false) = 1 then
      with Canvas.Font do
      begin
        case a of
          1:
            Style:=Style + [fsBold];
          2:
            Style:=Style - [fsBold];
          3:
            Style:=Style + [fsItalic];
          4:
            Style:=Style - [fsItalic];
          5:
            Style:=Style + [fsUnderline];
          6:
            Style:=Style - [fsUnderline];
          7:
            Size:=Size + 2;
          8:
            Size:=Font.Size;
          9:
            Size:=Size - 2;
          10:
            Size:=Font.Size;
          FrmAnchorStartNum:
            ProcAnchorStart;
          FrmAnchorStopNum:
            ProcAnchorStop;
          FrmColor1StartNum, FrmColor2StartNum:
            ProcColorStart(a);
          FrmColor1StopNum, FrmColor2StopNum:
            ProcColorStop;
          FrmBgColorStartNum:
            ProcBgColorStart(a);
          FrmBgColorStopNum:
            ProcBgColorStop;
        end;
        s:=copy(s, Length(FrmStr[a]) + 1, Length(s));
        Result:=a;
        Break;
      end;
  end;
end;

function TrHTMLLabel.GetLabelText: string;
var
  s, param: string;
begin
  s:=TrimRight(FCaptionStrings.Text);
  if s = '' then
    s:=Caption;
  // proceed parameters
  repeat
    param:=FindStrFromTo(FrmParamStart, FrmParamStop, s, false);
    if param = '' then
      Break;
    param:=copy(param, Length(FrmParamStart) + 1, Length(param));
    if (csDesigning in ComponentState) then
      param:='[' + param + ']'
    else
      if Assigned(FOnGetParameter) then
        FOnGetParameter(Self, param)
      else
        param:='';
    param:=ClearStr(FrmParamStart, param, false);
    param:=ClearStr(FrmParamStop, param, false);
    s:=ReplaceStrFromTo(FrmParamStart, FrmParamStop, param, s, true, false, false);
  until false;
  // replace nowrap tag
  if AnsiPosEx(FrmNoWordWrap, s, 1, false) > 0 then
  begin
    s:=ReplaceStr(FrmNoWordWrap, '', s, true, false);
    if WordWrap then
    begin
      WordWrap:=false;
      AdjustBounds;
      s:=s + '<B>'; // ensure extend drawing
    end;
  end;
  // replace spec chars
  // s:=ReplaceStr(FrmTAB,#9,s,true,false);
  s:=ReplaceStr(#9, '<TAB>', s, true, false);
  s:=ReplaceStr(#10, '', s, true, false);
  if FCaptionSkipEnter then
  begin
    // proc "<br> enter" tag as <br> only
    s:=ReplaceStr(FrmCR + #13, FrmCR, s, true, false);
    // proc "enter" tag as space
    s:=ReplaceStr(#13, ' ', s, true, false);
  end;
  // change <br> to new line
  s:=ReplaceStr(FrmCR, #13, s, true, false);
  // replace empty lines with space
  if copy(s, 1, 1) = #13 then
    s:=' ' + s;
  while Pos(#13#13, s) > 0 do
    s:=ReplaceStr(#13#13, #13 + ' ' + #13, s, true, false);
  // if not (csDesigning in ComponentState) and CaptionAsHint then Hint:=s;
  Result:=s;
end;

procedure TrHTMLLabel.DoDrawText(var Rect: TRect; Flags: Longint);
const
  Deviders = ' ._-/\@#$&|';
var
  Text: string;
  LineCount, X, W, H, MaxW, Indent, NextIndent: integer;
  Line, Rest: string;
  Rc, Rd: TRect;
  oldFont: TFontRecall;
  oldPen: TPenRecall;
  CalcOnly: boolean;

  function DrawTextPart(var R: TRect; F: Longint; Wrap: boolean; s: string): string;
  var
    a, DivPos, RestW: integer;
  begin
    // cannot calculate empty string - add space
    if s = '' then
      s:=' ';
    RestW:=Width - R.Left - 2 - 2 * FInnerSpace; // rest width till end of line
    if Wrap and (Canvas.TextWidth(s) > RestW) then
    begin
      // find position for wrapping
      DivPos:=-1;
      a:=1;
      while a <= Length(s) do
      begin
        if Pos(s[a], Deviders) > 0 then
          DivPos:=a;
        if Canvas.TextWidth(copy(s, 1, a)) > RestW then
          Break;
        Inc(a);
      end;
      if (a = 1) and (R.Left <= FInnerSpace + 1 + Indent) then
        Result:='' // too small width to wrap
      else
      begin
        if (a - 1 = Length(s)) then
          DivPos:=a // no wrap, use whole line
        else
          if (DivPos = -1) then // no separator found
            if (R.Left > FInnerSpace + 1 + Indent) then
              DivPos:=0 // not first word, wrap all
            else
              DivPos:=a - 1; // wrap last letter
        Result:=copy(s, DivPos + 1, Length(s)); // wraped part
        s:=Trim(copy(s, 1, DivPos));
        // if s='' then s:=' ';
      end;
    end
    else
      Result:='';
    // Save anchor start point
    if IsAnchorDraw then
      actAnchorInfo.Rect:=R;
    if s <> '' then
      DrawText(Canvas.Handle, PChar(s), Length(s), R, F);
    // Save anchor stop point
    if IsAnchorDraw then
    begin
      actAnchorInfo.Rect.Right:=actAnchorInfo.Rect.Left + Canvas.TextWidth(s);
      actAnchorInfo.Rect.Bottom:=actAnchorInfo.Rect.Top + Canvas.TextHeight(s);
    end;
    Inc(W, R.Right - R.Left);
    R.Left:=R.Left + Canvas.TextWidth(s);
  end;

  function ProcLine(var R: TRect; F: Longint; Calc, Wrap: boolean): string;
  var
    p: integer;
    s: string;
  begin
    if Calc then
      F:=F or DT_CALCRECT;
    Result:='';
    s:=Line;
    { for p:=1 to FrmStrCount do
      if CompareText(s, FrmStr[p]) = 0 then
      s:=s + ' ';
      // check only anchor on line
      if (AnsiPosEx('<A', s, 1, false) = 1) and (AnsiPosEx('</A>', s, 1, false) = Length(s) - 3) then
      s:=s + ' ';
      }
    repeat
      p:=FindFirstFrm(s);
      if p = 0 then
      begin
        Result:=DrawTextPart(R, F, Wrap, s);
        Break;
      end;
      if p > 1 then
        Result:=DrawTextPart(R, F, Wrap, copy(s, 1, p - 1));
      s:=copy(s, p, Length(s));
      if Result <> '' then // wraped ?
      begin
        Result:=Result + s; // add rest of line
        Break;
      end;
      case ProcFrm(s, Calc) of
        FrmTabNum:
          begin
            // proc TAB tag
            R.Left:=((R.Left div FTabGridWidth) + 1) * FTabGridWidth;
            W:=R.Left;
          end;
        FrmIndentNum:
          if Calc then
            NextIndent:=R.Left - 1;
      end;
    until false;
    // if (Result='') and (s<>'') then Result:=DrawTextPart(R,F,Wrap,s);
    // set Line for drawing
    Line:=copy(Line, 1, Length(Line) - Length(Result));
    if LastAnchorText <> '' then
    begin
      // End of line without </A>; (one anchor can not be on two lines)
      // add start tag to next line and add stop tag to this line
      if Result <> '' then
        Result:=LastAnchorText + Result;
      s:=FrmStr[FrmAnchorStopNum];
      ProcFrm(s, Calc);
    end;
  end;

begin
  CalcOnly:=(Flags and DT_CALCRECT) <> 0;
  Text:=GetLabelText;
  if Text = '' then
    Exit;
  AColorEnabled:=true;
  // set ellipsis Flags
  { if EllipsisPosition=epEndEllipsis then Flags:=Flags or DT_END_ELLIPSIS
    else
    if EllipsisPosition=epPathEllipsis then Flags:=Flags or DT_PATH_ELLIPSIS
    else
    if EllipsisPosition=epWordEllipsis then Flags:=Flags or DT_WORD_ELLIPSIS;
    if (FEllipsisType in [etEnd,etPath]) then
    if not (csDesigning in ComponentState) and CaptionAsHint then Hint:=Text;
    }

  if FShowBeveled and (not CalcOnly) then
  begin
    // Draw bevel
    oldPen:=TPenRecall.Create(Canvas.Pen);
    try
      Canvas.Pen.Color:=clBtnShadow;
      Canvas.PolyLine([Point(0, Height - 1), Point(0, 0), Point(Width - 1, 0)]);
      Canvas.Pen.Color:=clBtnHighlight;
      Canvas.PolyLine([Point(Width - 1, 0), Point(Width - 1, Height - 1), Point(0, Height - 1)]);
    finally
      FreeAndNil(oldPen);
    end;
  end;

  if (not FShowText3D) and (FindFirstFrm(Text) = 0) and (FFontColor = clNone) then
  begin
    // substract space
    InflateRect(Rect, -FInnerSpace, -FInnerSpace);
    // Without formating => standard DoDrawText
    inherited DoDrawText(Rect, Flags);
    // add space
    InflateRect(Rect, FInnerSpace, FInnerSpace);
    if not(csDesigning in ComponentState) and ACaptionAsHint and not AutoSize then
      if (Canvas.TextWidth(Text) > Width + 3) then
        Hint:=Text
      else
        Hint:='';
    Exit;
  end;

  // Exteded formating
  // Cancel WordWrap and Ellipsis for formated text
  // Alignment is done not by Flags
  Flags:=Flags and (not(DT_WORDBREAK + DT_RIGHT + DT_CENTER + DT_END_ELLIPSIS + DT_PATH_ELLIPSIS + DT_WORD_ELLIPSIS));
  Flags:=Flags or DT_NOPREFIX or DT_EXPANDTABS;
  Flags:=DrawTextBiDiModeFlags(Flags);

  // substract space
  InflateRect(Rect, -FInnerSpace, -FInnerSpace);
  Canvas.Font:=Font;
  Canvas.Font.Color:=GetStyleFontColor_LabelText(Canvas.Font.Color);
  if FFontColor <> clNone then
    Canvas.Font.Color:=FFontColor;
  OriginalBrush:=TBrush.Create;
  OriginalBrush.Assign(Canvas.Brush);
  try
    Rd:=Rect;
    LineCount:=1;
    MaxW:=0;
    H:=0;
    FreeAnchorInfo;
    repeat
      // Find and proc one line
      Rest:=FindStrPart(Text, #13, LineCount, false);
      if Rest = '' then
        Break;
      Indent:=0;
      NextIndent:=0;
      repeat
        // calc width of the line and proccess WordWrap
        if Rest = '' then
          Break;
        Line:=Rest;
        Rc:=Rect;
        W:=Indent;
        Rc.Left:=W;
        if CalcOnly then
          Rest:=ProcLine(Rc, Flags, true, WordWrap)
        else
        begin
          // Reset font back after proc line for Drawing
          oldFont:=TFontRecall.Create(Canvas.Font);
          try
            Rest:=ProcLine(Rc, Flags, true, WordWrap);
          finally
            FreeAndNil(oldFont);
          end;
        end;
        // store maxWidth and Height
        if W > MaxW then
          MaxW:=W;
        // if Rc.Bottom - Rc.Top = 0 then ShowMessage('0');

        Inc(H, Rc.Bottom - Rc.Top);

        if not CalcOnly then
        begin
          // draw the line
          case Alignment of
            taCenter:
              begin
                X:=(Rect.Right - Rect.Left - W) div 2 + Rect.Left;
                if X < Rect.Left + 1 + Indent then
                  X:=Rect.Left + 1 + Indent;
              end;
            taRightJustify:
              X:=Rect.Right - W - 1;
            taLeftJustify:
              X:=Rect.Left + 1 + Indent;
          else
            X:=0;
          end;
          Rd.Left:=X;

          if (not Enabled) or FShowText3D then
          begin
            // Draw shadow
            AColorEnabled:=false;
            oldFont:=TFontRecall.Create(Canvas.Font);
            try
              OffsetRect(Rd, FShadowSize, FShadowSize);
              if not Enabled then
                Canvas.Font.Color:=clBtnHighlight
              else
                Canvas.Font.Color:=FShadowColor; // clBtnShadow;
              ProcLine(Rd, Flags, false, false);
            finally
              FreeAndNil(oldFont);
            end;
            OffsetRect(Rd, -FShadowSize, -FShadowSize);
            Rd.Left:=X;
            if not Enabled then
              Canvas.Font.Color:=clBtnShadow;
          end;
          AColorEnabled:=Enabled;
          ProcLine(Rd, Flags, false, false);
          // Set Rect for next line
          Rd.Top:=Rect.Top + H;
        end;
        if Alignment = taLeftJustify then
          Indent:=NextIndent;
      until false;
      Inc(LineCount);
    until false;
    // Set calculated bounds
    if not WordWrap then
      Rect.Right:=Rect.Left + MaxW + 2 + FInnerSpace;
    Rect.Bottom:=Rect.Top + H + 2 + FInnerSpace;
  finally
    Canvas.Brush.Assign(OriginalBrush);
    FreeAndNil(OriginalBrush);
  end;
  if Assigned(BeforeAnchorFont) then
    FreeAndNil(BeforeAnchorFont);
end;

procedure TrHTMLLabel.MouseMove(Shift: TShiftState; X, Y: integer);
var
  oldAct, i: integer;
  R: TRect;
begin
  oldAct:=actAnchorItem;
  actAnchorItem:=-1;
  // Find active anchor
  for i:=0 to AnchorInfoList.Count - 1 do
  begin
    actAnchorInfo:=AnchorInfoList.Items[i];
    if (X > actAnchorInfo.Rect.Left) and (X < actAnchorInfo.Rect.Right) and (Y > actAnchorInfo.Rect.Top) and (Y < actAnchorInfo.Rect.Bottom) then
    begin
      actAnchorItem:=i;
      Break;
    end;
  end;
  if actAnchorItem <> -1 then
    Screen.Cursor:=crHandPoint
  else
    Screen.Cursor:=Cursor;
  if oldAct <> actAnchorItem then
  begin
    if (oldAct <> -1) and (oldAct < AnchorInfoList.Count) then
    begin
      R:=TAnchorInfo(AnchorInfoList.Items[oldAct]).Rect;
      OffsetRect(R, Left, Top);
      InvalidateRect(Parent.Handle, @R, false);
      Update;
    end;
    if (actAnchorItem <> -1) then
    begin
      R:=actAnchorInfo.Rect;
      OffsetRect(R, Left, Top);
      InvalidateRect(Parent.Handle, @R, false);
      Update;
    end;
  end;
  inherited;
end;

procedure TrHTMLLabel.CMMouseLeave(var AMsg: TMessage);
begin
  Screen.Cursor:=Cursor;
  if actAnchorItem <> -1 then
  begin
    actAnchorItem:=-1;
    Repaint;
  end;
end;

procedure TrHTMLLabel.Click;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if actAnchorItem <> -1 then
  begin
    Screen.Cursor:=Cursor;
    actAnchorInfo:=AnchorInfoList.Items[actAnchorItem];
    if actAnchorInfo.LinkType = 'HREF' then
      ShellExecute(Application.Handle, 'open', PChar(actAnchorInfo.LinkName), nil, nil, SW_NORMAL)
    else
      if Assigned(FOnAnchorClick) then
        FOnAnchorClick(Self, actAnchorInfo.LinkName);
  end;
  inherited;
end;

// ==============================================================================
// TrURLLabel
constructor TrURLLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFontColor:=clHighLight;
  FHotColor:=clHotLight;
  Cursor:=crHandPoint;
  MouseOver:=false;
  ShowAccelChar:=false;
end;

procedure TrURLLabel.Loaded;
begin
  inherited Loaded;
  if not(csDesigning in ComponentState) and (Hint = '') then
    Hint:=FURLCommand;
  ParentFont:=true;
end;

procedure TrURLLabel.SetFontColor(Value: TColor);
begin
  FFontColor:=Value;
  Invalidate;
end;

procedure TrURLLabel.SetHotColor(Value: TColor);
begin
  FHotColor:=Value;
  Invalidate;
end;

procedure TrURLLabel.SetURLCommand(const Value: string);
begin
  if Hint = FURLCommand then
    Hint:=Value;
  FURLCommand:=Value;
end;

procedure TrURLLabel.CMMouseEnter(var AMsg: TMessage);
begin
  inherited;
  MouseOver:=true;
  Invalidate;
end;

procedure TrURLLabel.CMMouseLeave(var AMsg: TMessage);
begin
  inherited;
  MouseOver:=false;
  Invalidate;
end;

procedure TrURLLabel.DoDrawText(var Rect: TRect; Flags: Longint);
var
  Text: string;
begin
  Text:=GetLabelText;
  if (Flags and DT_CALCRECT <> 0) and (Text = '') then
    Text:=Text + ' ';
  if not ShowAccelChar then
    Flags:=Flags or DT_NOPREFIX;
  Flags:=DrawTextBiDiModeFlags(Flags);
  Canvas.Font:=Font;
  Canvas.Font.Style:=Canvas.Font.Style + [fsUnderline];
  if MouseOver then
    Canvas.Font.Color:=FHotColor
  else
    Canvas.Font.Color:=FFontColor;
  if not Enabled then
  begin
    OffsetRect(Rect, 1, 1);
    Canvas.Font.Color:=clBtnHighlight;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
    OffsetRect(Rect, -1, -1);
    Canvas.Font.Color:=clBtnShadow;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
  end
  else
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Flags);
end;

procedure TrURLLabel.Click;
begin
  inherited;
  if FURLCommand <> '' then
    ShellExecute(Application.Handle, 'open', PChar(FURLCommand), nil, nil, SW_NORMAL);
end;

end.
