unit rHTMLListView;

{<
  rHTMLListView unit contains TrHTMLListView - enhanced TCustomListView component
  which supports definable size of items and icons, background colors and fonts, HTML labels.
  It provides many new properties and functions for loading of items
  from database of from directory.

  Version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2015)
  @lastmod(Jan 2016)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, StdCtrls,
  Forms, DB, ComCtrls, CommCtrl, ComStrs, Dialogs, Variants,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} Math, Themes, UxTheme, ImgList,
  RosiCompGlobalCfg;

type
  TrHTMLListView = class;

  { new extended ListView Item object used by TrHTMLListView }
  TrHTMLListItem = class(TListItem)
  public
    { define html text, used instead of standard Item.Text property }
    HTMLText: string;
    { define hint text, if empty hint show full HTMLText if truncated }
    HintText: string;
    { indicate rect area for image }
    ImageRect: TRect;
    { indicate rect area for text }
    TextRect: TRect;
    { indicate taht HTMLText was truncated in item }
    TextTruncated: boolean;
    { indicate then image was not loaded yet }
    ImageNotLoaded: boolean;
    { indicate that image is not defined }
    ImageNotDefined: boolean;
    { indicate that image is corrupted }
    ImageError: boolean;
    { define image index from StateImages which is used instead of standard image }
    StateImageIndex: integer;
    { if true redefine item as transparent }
    ForceTransparent: boolean;
    { contain bitmap for Item - used for LoadFileImages or LoadDBImages or by user,
      if exist is used instead of Image list }
    ItemBitmap: TBitmap;
    { define ID value for Item - filename, KeyField value or any other value set by user }
    ItemID: variant;
    { define data value for Item - any value set by user, not used by TrHTMLListView }
    ItemVarData: variant;
    { define object value for Item - any value set by user, not used by TrHTMLListView }
    ItemObject: TObject;
    constructor Create(AOwner: TListItems); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Assign(Source: TPersistent); override; {< @exclude }
  end;

  { class for new Items design properties (size, colors, position) used by TrHTMLListView}
  TItemOptionsEx = class(TComponent)
  private
    FListView: TrHTMLListView;
    FCustomDraw: boolean;

    FFullItemWidth: integer;
    FFullItemHeight: integer;
    FFullItemMargin: integer;
    FFullIconItemVSpace: integer;

    FItemFrameType: TFrameType;
    FItemFrameColor: TColor;
    FItemFrameColorFocused: TColor;
    FItemBkColor: TColor;
    FItemBkColorSelected: TColor;
    FItemBkColorHot: TColor;

    FImages: TCustomImageList;
    FImageMargin: integer;
    FImageFrameType: TFrameType;
    FImageFrameColor: TColor;
    FImageBkColor: TColor;
    FImageStretch: boolean;
    FImageTransparent: boolean;

    FStateImages: TCustomImageList;
    FStateNotLoadedIndex: integer;
    FStateNotDefinedIndex: integer;
    FStateErrorIndex: integer;

    FTextWidth: integer;
    FTextHeight: integer;
    FTextLines: integer;
    FTextWordWrap: boolean;
    FTextAlignment: TAlignment;
    FTextLayout: TTextLayout;
    FTextMargin: integer;
    FTextFrameType: TFrameType;
    FTextFrameColor: TColor;
    FTextBkColor: TColor;
    FTextTabGridWidth: byte;

    procedure SetCustomDraw(Value: boolean);

    procedure SetFullItemWidth(Value: integer);
    procedure SetFullItemHeight(Value: integer);
    procedure SetFullIconItemVSpace(Value: integer);
    procedure SetFullItemMargin(Value: integer);

    procedure SetItemFrameType(Value: TFrameType);
    procedure SetItemFrameColor(Value: TColor);
    procedure SetItemFrameColorFocused(Value: TColor);
    procedure SetItemBkColor(Value: TColor);
    procedure SetItemBkColorSelected(Value: TColor);
    procedure SetItemBkColorHot(Value: TColor);

    procedure SetImages(Value: TCustomImageList);
    procedure SetImageMargin(Value: integer);
    procedure SetImageFrameType(Value: TFrameType);
    procedure SetImageFrameColor(Value: TColor);
    procedure SetImageBkColor(Value: TColor);
    procedure SetImageStretch(Value: boolean);
    procedure SetImageTransparent(Value: boolean);

    procedure SetStateImages(Value: TCustomImageList);

    procedure SetTextWidth(Value: integer);
    procedure SetTextHeight(Value: integer);
    procedure SetTextLines(Value: integer);
    procedure SetTextAlignment(Value: TAlignment);
    procedure SetTextLayout(Value: TTextLayout);
    procedure SetTextMargin(Value: integer);
    procedure SetTextFrameType(Value: TFrameType);
    procedure SetTextFrameColor(Value: TColor);
    procedure SetTextBkColor(Value: TColor);
    procedure SetTextTabGridWidth(Value: byte);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  published
    { define whether new design is used }
    property CustomDraw: boolean read FCustomDraw write SetCustomDraw default true;

    { define width of item }
    property FullItemWidth: integer read FFullItemWidth write SetFullItemWidth default 64;
    { define height of item }
    property FullItemHeight: integer read FFullItemHeight write SetFullItemHeight default 64;
    { define additional vertical space for vsIcon style (negative number is possible) }
    property FullIconItemVSpace: integer read FFullIconItemVSpace write SetFullIconItemVSpace default 0;
    { define margin between items }
    property FullItemMargin: integer read FFullItemMargin write SetFullItemMargin default 0;

    { define frame type around item rect }
    property ItemFrameType: TFrameType read FItemFrameType write SetItemFrameType default ft3D;
    { define color of frame around item rect }
    property ItemFrameColor: TColor read FItemFrameColor write SetItemFrameColor default clWindowFrame;
    { define color of frame around item rect for focused item }
    property ItemFrameColorFocused: TColor read FItemFrameColorFocused write SetItemFrameColorFocused default clHotLight;
    { define background color of item rect }
    property ItemBkColor: TColor read FItemBkColor write SetItemBkColor default clNone;
    { define background color of item rect for all selected items }
    property ItemBkColorSelected: TColor read FItemBkColorSelected write SetItemBkColorSelected default clNone;
    { define background color of item rect }
    property ItemBkColorHot: TColor read FItemBkColorHot write SetItemBkColorHot default clNone;

    { define ImageList for items }
    property Images: TCustomImageList read FImages write SetImages;
    { define margin between image and item border }
    property ImageMargin: integer read FImageMargin write SetImageMargin default 4;
    { define frame type around image rect }
    property ImageFrameType: TFrameType read FImageFrameType write SetImageFrameType default ftNone;
    { define color of frame around image rect }
    property ImageFrameColor: TColor read FImageFrameColor write SetImageFrameColor default clWindowFrame;
    { define background color of image rect }
    property ImageBkColor: TColor read FImageBkColor write SetImageBkColor default clNone;
    { define whether image from ImageList is resized to image rect }
    property ImageStretch: boolean read FImageStretch write SetImageStretch default true;
    { define whether image from ImageList is shown transparently }
    property ImageTransparent: boolean read FImageTransparent write SetImageTransparent default true;

    { define ImageList for special state images }
    property StateImages: TCustomImageList read FStateImages write SetStateImages;
    { define image index from StateImages in case that image is not loaded yet }
    property StateNotLoadedIndex: integer read FStateNotLoadedIndex write FStateNotLoadedIndex default - 1;
    { define image index from StateImages in case that image is not defined or empty }
    property StateNotDefinedIndex: integer read FStateNotDefinedIndex write FStateNotDefinedIndex default - 1;
    { define image index from StateImages in case that image is corrupted }
    property StateErrorIndex: integer read FStateErrorIndex write FStateErrorIndex default - 1;

    { define width of text rect, whole item width is used if TextWidth=0 }
    property TextWidth: integer read FTextWidth write SetTextWidth default 0;
    { define height of text rect, whole item height is used if TextHeight=0 }
    property TextHeight: integer read FTextHeight write SetTextHeight default 0;
    { define height of text rect as needed space for defined text lines, whole item height is used if TextLines=0 }
    property TextLines: integer read FTextLines write SetTextLines default 0;
    { define whether text is wrapped if needed }
    property TextWordWrap: boolean read FTextWordWrap write FTextWordWrap default true;
    { define horizontal alignment in text rect }
    property TextAlignment: TAlignment read FTextAlignment write SetTextAlignment default taCenter;
    { define vertical alignment in text rect }
    property TextLayout: TTextLayout read FTextLayout write SetTextLayout default tlCenter;
    { define margin between text rect and item border }
    property TextMargin: integer read FTextMargin write SetTextMargin default 4;
    { define frame type around text rect }
    property TextFrameType: TFrameType read FTextFrameType write SetTextFrameType default ftNone;
    { define color of frame around text rect }
    property TextFrameColor: TColor read FTextFrameColor write SetTextFrameColor default clWindowFrame;
    { define background color of text rect }
    property TextBkColor: TColor read FTextBkColor write SetTextBkColor default clNone;
    { define width of Tabulator grid }
    property TextTabGridWidth: byte read FTextTabGridWidth write SetTextTabGridWidth default 30;
  end;

  { @exclude }
  TLVGetItemParamsEvent = procedure(Sender: TObject; Item: TrHTMLListItem; var ItemBkColor: TColor; var ItemFrameType: TFrameType; var ItemFrameColor: TColor;
    TextFont: TFont; var TextBkColor: TColor; var TextFrameType: TFrameType; var TextFrameColor: TColor) of object;
  { @exclude }
  TLVGetItemContentEvent = procedure(Sender: TObject; Item: TrHTMLListItem; var HTMLText: string; Image: TBitmap) of object;
  { @exclude }
  TLVDrawHTMLItemEvent = procedure(ACanvas: TCanvas; Item: TrHTMLListItem; ItemRect: TRect) of object;
  { @exclude }
  TLVClickItemEvent = procedure(Sender: TObject; Item: TrHTMLListItem; Button: TMouseButton; Shift: TShiftState; ClickToText, ClickToImage: boolean) of object;

  { enhanced TCustomListView component which supports definable size of items and icons,
    background colors and fonts, HTML labels.
    It provides many new properties and functions for loading of items from database of from directory }
  TrHTMLListView = class(TCustomListView)
  private
    FItemOptionsEx: TItemOptionsEx;
    FIntImageList: TCustomImageList;
    FOnGetItemParams: TLVGetItemParamsEvent;
    FOnGetItemContent: TLVGetItemContentEvent;
    FOnCustomDrawHTMLItemEnd: TLVDrawHTMLItemEvent;
    FOnClickItem: TLVClickItemEvent;
    procedure UpdateIconItemSize(RefSize: boolean);
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure CNNotify(var Message: TWMNotifyLV); message CN_NOTIFY;
  protected
    FLoadingHTMLMask: string;
    FLoadingHintMask: string;
    FLoadingLastItemIndex: integer;
    FLoadingAborted: boolean;
    procedure LoadingAbortEvent(Sender: TObject);
    procedure Loaded; override;
    procedure SetViewStyle(Value: TViewStyle); override;
    function CreateListItem: TListItem; override;
    function IsCustomDrawn(Target: TCustomDrawTarget; Stage: TCustomDrawStage): boolean; override;
    function CustomDrawItem(Item: TListItem; State: TCustomDrawState; Stage: TCustomDrawStage): boolean; override;
    procedure DrawItemEx(ACanvas: TCanvas; R: TRect; Item: TrHTMLListItem; State: TCustomDrawState); virtual;
    procedure DrawItemImageEx(ACanvas: TCanvas; R: TRect; Item: TrHTMLListItem; State: TCustomDrawState; IconBmp: TBitmap); virtual;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure LoadImage(ItemIndex: integer; FromFile: boolean; KeyField, ImageField: TField);
    function GetMaskText(Item: TrHTMLListItem; MaskText: string): string;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { clear all items and init internal loading }
    procedure InitLoadingItems(ClearItems: boolean = false);
    { return item as TrHTMLListItem }
    function GetHTMLItem(ItemIndex: integer): TrHTMLListItem;
    { add new item with defined text and hint }
    function AddHTMLItem(HTMLText, HintText: string; ImageIndex: integer = -1): TrHTMLListItem;
    { add new item with defined text, hint and data }
    function AddHTMLItemData(HTMLText, HintText: string; ItemBitmap: TBitmap; ItemID, ItemVarData: variant; ItemObject: TObject): TrHTMLListItem;
    { update and refresh item text and hint }
    procedure UpdateHTMLItem(ItemIndex: integer; HTMLText, HintText: string);
    { update and refresh item bitmap and data }
    procedure UpdateHTMLItemData(ItemIndex: integer; ItemBitmap: TBitmap; ItemID, ItemVarData: variant; ItemObject: TObject);
    { default compare according HTML text }
    procedure CompareEvent(Sender: TObject; Item1, Item2: TListItem; Data: integer; var Compare: integer);
    { draw frame of defined type and color }
    procedure DrawFrame(ACanvas: TCanvas; var R: TRect; FrameType: TFrameType; aColor: TColor);
    { draw standard theme background of item according state }
    procedure DrawThemeBk(ACanvas: TCanvas; R: TRect; State: TCustomDrawState);
    { add Items as file list from specified directory according file mask,
      set item text according HtmlTextFormat (<<File>> will be replaced by filename)
      set ItemID to file name including path }
    procedure LoadFileList(FilePath, FileMask: string; HtmlTextMask: string = ''; SortByFileName: boolean = false; LoadImages: boolean = false);
    { add images for loaded file list, can be called repeatedly by timer or OnIdle event,
      set ItemVarData to string Width;Height
      return false if all items are loaded }
    function LoadFileImages(LoadAllItems: boolean = false; LoadSelected: boolean = false): boolean;
    { load records from dataset, KeyField define key value for each record,
      set item text according HtmlTextMask (<<FieldName>> will be replaced by field value),
      set hint text according HintTextMask (<<FieldName>> will be replaced by field value),
      set ItemID to record key value }
    procedure LoadDBList(Dataset: TDataset; KeyField, ImageField: TField; HtmlTextMask: string = ''; HintTextMask: string = ''; LoadImages: boolean = false);
    { add images for loaded db records, can be called repeatedly by timer or OnIdle event,
      set ItemVarData to string Width;Height
      return false if all items are loaded }
    function LoadDBImages(Dataset: TDataset; KeyField, ImageField: TField; LoadAllItems: boolean = false; LoadSelected: boolean = false): boolean;
  published
    property Action;
    property Align;
    property AllocBy;
    property Anchors;
    property BevelEdges;
    property BevelInner;
    property BevelOuter;
    property BevelKind default bkNone;
    property BevelWidth;
    property BiDiMode;
    property BorderStyle;
    property BorderWidth;
    // property Checkboxes;
    property Color;
    property Columns;
    property ColumnClick;
    property Constraints;
    property Ctl3D;
    property DoubleBuffered;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property Font;
    property FlatScrollBars;
    property FullDrag;
    property GridLines;
    // property Groups;
    property HideSelection;
    property HotTrack;
    property HotTrackStyles;
    property HoverTime;
    property IconOptions;
    property Items;
    // property LargeImages;
    property MultiSelect;
    property OwnerData;
    property OwnerDraw;
    // property GroupHeaderImages;
    // property GroupView default False;
    property ReadOnly default false;
    property RowSelect;
    property ParentBiDiMode;
    property ParentColor default false;
    property ParentDoubleBuffered;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowColumnHeaders;
    property ShowWorkAreas;
    property ShowHint;
    // property SmallImages;
    property SortType;
    // property StateImages;
    property TabOrder;
    property TabStop default true;
    property Touch;
    property ViewStyle;
    property Visible;
    property OnAdvancedCustomDraw;
    property OnAdvancedCustomDrawItem;
    property OnAdvancedCustomDrawSubItem;
    property OnChange;
    property OnChanging;
    property OnClick;
    property OnColumnClick;
    property OnColumnDragged;
    property OnColumnRightClick;
    property OnCompare;
    property OnContextPopup;
    property OnCustomDraw;
    property OnCustomDrawItem;
    property OnCustomDrawSubItem;
    property OnCreateItemClass;
    property OnData;
    property OnDataFind;
    property OnDataHint;
    property OnDataStateChange;
    property OnDblClick;
    property OnDeletion;
    property OnDrawItem;
    property OnEdited;
    property OnEditing;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGesture;
    property OnGetImageIndex;
    // property OnGetSubItemImage;
    property OnDragDrop;
    property OnDragOver;
    property OnInfoTip;
    property OnInsert;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseActivate;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnSelectItem;
    // property OnItemChecked;
    property OnStartDock;
    property OnStartDrag;
    { defines all new Items design properties (size, colors, position) }
    property ItemOptionsEx: TItemOptionsEx read FItemOptionsEx write FItemOptionsEx;
    { event is called to modify all design properties of item }
    property OnGetItemParams: TLVGetItemParamsEvent read FOnGetItemParams write FOnGetItemParams;
    { event is called to modify text and image of item }
    property OnGetItemContent: TLVGetItemContentEvent read FOnGetItemContent write FOnGetItemContent;
    { event is called to add any custom drawing to item }
    property OnCustomDrawHTMLItemEnd: TLVDrawHTMLItemEvent read FOnCustomDrawHTMLItemEnd write FOnCustomDrawHTMLItemEnd;
    { event is called after click to any item }
    property OnClickItem: TLVClickItemEvent read FOnClickItem write FOnClickItem;
  end;

procedure Register; {< @exclude }

implementation

uses rstring, rpictool, rtool, rfiletool, rdbtool, rdlg, rHTMLDrawText;
{$R *.res}

const
  cTextInnerSpace = 2;
  cMaskFileName = '<<File>>';
  cMaskFileW = '<<W>>';
  cMaskFileH = '<<H>>';
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi', [TrHTMLListView]);
end;

// ==============================================================================
// TIconOptionsEx

constructor TrHTMLListItem.Create(AOwner: TListItems);
begin
  inherited;
  HTMLText:='';
  HintText:='';
  TextTruncated:=false;
  ImageNotLoaded:=false;
  ImageNotDefined:=false;
  ImageError:=false;
  StateImageIndex:=-1;
  ForceTransparent:=false;
  ItemBitmap:=nil;
  ItemID:=null;
  ItemVarData:=null;
  ItemObject:=nil;
end;

destructor TrHTMLListItem.Destroy;
begin
  if Assigned(ItemBitmap) then
    FreeAndNil(ItemBitmap);
  inherited;
end;

procedure TrHTMLListItem.Assign(Source: TPersistent);
begin
  inherited Assign(Source);
  if Source is TrHTMLListItem then
    with Source as TrHTMLListItem do
    begin
      Self.HTMLText:=HTMLText;
      Self.HintText:=HintText;
      Self.ImageRect:=ImageRect;
      Self.TextRect:=TextRect;
      Self.TextTruncated:=TextTruncated;
      Self.ImageNotLoaded:=ImageNotLoaded;
      Self.ImageNotDefined:=ImageNotDefined;
      Self.ImageError:=ImageError;
      Self.StateImageIndex:=StateImageIndex;
      Self.ForceTransparent:=ForceTransparent;
      if Assigned(ItemBitmap) then
      begin
        Self.ItemBitmap:=TBitmap.Create;
        Self.ItemBitmap.Assign(ItemBitmap);
      end;
      Self.ItemID:=ItemID;
      Self.ItemVarData:=ItemVarData;
      Self.ItemObject:=ItemObject;
    end;
end;

constructor TItemOptionsEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  if not(AOwner is TrHTMLListView) then
    raise Exception.Create(sInvalidOwner);
  FListView:=AOwner as TrHTMLListView;
  Name:='ItemOptionsEx';
  Self.SetSubComponent(true);

  FCustomDraw:=true;
  FFullItemWidth:=64;
  FFullItemHeight:=64;
  FFullItemMargin:=0;

  FItemFrameType:=ft3D;
  FItemFrameColor:=clWindowFrame;
  FItemFrameColorFocused:=clHotLight;
  FItemBkColor:=clNone;
  FItemBkColorSelected:=clNone;
  FItemBkColorHot:=clNone;

  FImageMargin:=4;
  FImageFrameType:=ftNone;
  FImageFrameColor:=clWindowFrame;
  FImageBkColor:=clNone;
  FImageStretch:=true;
  FImageTransparent:=true;

  FStateNotLoadedIndex:=-1;
  StateNotDefinedIndex:=-1;
  FStateErrorIndex:=-1;

  FTextWidth:=0;
  FTextHeight:=0;
  FTextLines:=0;
  FTextWordWrap:=true;
  FTextAlignment:=taCenter;
  FTextLayout:=tlCenter;
  FTextMargin:=4;
  FTextFrameType:=ftNone;
  FTextFrameColor:=clWindowFrame;
  FTextBkColor:=clNone;
  FTextTabGridWidth:=30;
end;

procedure TItemOptionsEx.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FImages then
      FImages:=nil;
    if AComponent = FStateImages then
      FStateImages:=nil;
  end;
end;

procedure TItemOptionsEx.SetCustomDraw(Value: boolean);
begin
  FCustomDraw:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetFullItemWidth(Value: integer);
begin
  if Value <> FFullItemWidth then
  begin
    FFullItemWidth:=Value;
    FListView.UpdateIconItemSize(true);
  end;
end;

procedure TItemOptionsEx.SetFullItemHeight(Value: integer);
begin
  if Value <> FFullItemHeight then
  begin
    FFullItemHeight:=Value;
    FListView.UpdateIconItemSize(true);
  end;
end;

procedure TItemOptionsEx.SetFullIconItemVSpace(Value: integer);
begin
  if Value <> FFullIconItemVSpace then
  begin
    FFullIconItemVSpace:=Value;
    FListView.UpdateIconItemSize(false);
  end;
end;

procedure TItemOptionsEx.SetFullItemMargin(Value: integer);
begin
  if Value <> FFullItemMargin then
  begin
    FFullItemMargin:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetItemFrameType(Value: TFrameType);
begin
  FItemFrameType:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetItemFrameColor(Value: TColor);
begin
  FItemFrameColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetItemFrameColorFocused(Value: TColor);
begin
  FItemFrameColorFocused:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetItemBkColor(Value: TColor);
begin
  FItemBkColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetItemBkColorSelected(Value: TColor);
begin
  FItemBkColorSelected:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetItemBkColorHot(Value: TColor);
begin
  FItemBkColorHot:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetImages(Value: TCustomImageList);
begin
  if FImages <> Value then
  begin
    FImages:=Value;
    if FImages <> nil then
      FImages.FreeNotification(Self);
    FListView.Invalidate;
    FListView.IconOptions.WrapText:=false;
  end;
end;

procedure TItemOptionsEx.SetImageMargin(Value: integer);
begin
  if Value <> FImageMargin then
  begin
    FImageMargin:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetImageFrameType(Value: TFrameType);
begin
  FImageFrameType:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetImageFrameColor(Value: TColor);
begin
  FImageFrameColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetImageBkColor(Value: TColor);
begin
  FImageBkColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetImageStretch(Value: boolean);
begin
  FImageStretch:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetImageTransparent(Value: boolean);
begin
  FImageTransparent:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetStateImages(Value: TCustomImageList);
begin
  if FStateImages <> Value then
  begin
    FStateImages:=Value;
    if FStateImages <> nil then
      FStateImages.FreeNotification(Self);
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextWidth(Value: integer);
begin
  if Value <> FTextWidth then
  begin
    FTextWidth:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextHeight(Value: integer);
begin
  if Value <> FTextHeight then
  begin
    FTextHeight:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextLines(Value: integer);
begin
  if Value <> FTextLines then
  begin
    FTextLines:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextAlignment(Value: TAlignment);
begin
  if Value <> FTextAlignment then
  begin
    FTextAlignment:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextLayout(Value: TTextLayout);
begin
  if Value <> FTextLayout then
  begin
    FTextLayout:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextMargin(Value: integer);
begin
  if Value <> FTextMargin then
  begin
    FTextMargin:=Value;
    FListView.Invalidate;
  end;
end;

procedure TItemOptionsEx.SetTextFrameType(Value: TFrameType);
begin
  FTextFrameType:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetTextFrameColor(Value: TColor);
begin
  FTextFrameColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetTextBkColor(Value: TColor);
begin
  FTextBkColor:=Value;
  FListView.Invalidate;
end;

procedure TItemOptionsEx.SetTextTabGridWidth(Value: byte);
begin
  FTextTabGridWidth:=Value;
  FListView.Invalidate;
end;

// ==============================================================================
// TrHTMLListView

constructor TrHTMLListView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItemOptionsEx:=TItemOptionsEx.Create(Self);
  FIntImageList:=TCustomImageList.Create(Self);
  FIntImageList.SetSubComponent(true);
  FIntImageList.Name:='IntImageList';
  InitLoadingItems(false);
  ReadOnly:=true;
  OnCompare:=CompareEvent;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrHTMLListView.Destroy;
begin
  FreeAndNil(FIntImageList);
  FreeAndNil(FItemOptionsEx);
  inherited Destroy;
end;

procedure TrHTMLListView.Loaded;
begin
  inherited;
  if HandleAllocated then
  begin
    LargeImages:=FIntImageList;
    SmallImages:=FIntImageList;
    UpdateIconItemSize(true);
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrHTMLListView.InitLoadingItems(ClearItems: boolean = false);
begin
  Items.Clear;
  FLoadingHTMLMask:='';
  FLoadingHintMask:='';
  FLoadingAborted:=false;
  FLoadingLastItemIndex:=-1;
end;

function TrHTMLListView.CreateListItem: TListItem;
var
  LClass: TListItemClass;
begin
  LClass:=TrHTMLListItem;
  if Assigned(OnCreateItemClass) then
    OnCreateItemClass(Self, LClass);
  Result:=LClass.Create(Items);
end;

procedure TrHTMLListView.CompareEvent(Sender: TObject; Item1: TListItem; Item2: TListItem; Data: integer; var Compare: integer);
begin
  Compare:=lstrcmp(PChar(TrHTMLListItem(Item1).HTMLText), PChar(TrHTMLListItem(Item2).HTMLText));
end;

function TrHTMLListView.GetHTMLItem(ItemIndex: integer): TrHTMLListItem;
begin
  if (ItemIndex >= 0) and (ItemIndex < Items.Count) then
    Result:=TrHTMLListItem(Items[ItemIndex])
  else
    Result:=nil;
end;

function TrHTMLListView.AddHTMLItem(HTMLText: string; HintText: string; ImageIndex: integer = -1): TrHTMLListItem;
begin
  Result:=AddHTMLItemData(HTMLText, HintText, nil, null, null, nil);
  Result.ImageIndex:=ImageIndex;
end;

function TrHTMLListView.AddHTMLItemData(HTMLText: string; HintText: string; ItemBitmap: TBitmap; ItemID: variant; ItemVarData: variant; ItemObject: TObject)
  : TrHTMLListItem;
var
  Item: TrHTMLListItem;
begin
  Item:=TrHTMLListItem(CreateListItem);
  Item.HTMLText:=HTMLText;
  Item.HintText:=HintText;
  Result:=TrHTMLListItem(Items.AddItem(Item));
  UpdateHTMLItemData(Result.Index, ItemBitmap, ItemID, ItemVarData, ItemObject);
end;

procedure TrHTMLListView.UpdateHTMLItem(ItemIndex: integer; HTMLText: string; HintText: string);
var
  Item: TrHTMLListItem;
begin
  if (ItemIndex >= 0) and (ItemIndex < Items.Count) then
  begin
    Item:=TrHTMLListItem(Items[ItemIndex]);
    Item.HTMLText:=HTMLText;
    Item.HintText:=HintText;
    InvalidateRect(Handle, Item.DisplayRect(drBounds), true);
  end;
end;

procedure TrHTMLListView.UpdateHTMLItemData(ItemIndex: integer; ItemBitmap: TBitmap; ItemID, ItemVarData: variant; ItemObject: TObject);
var
  Item: TrHTMLListItem;
begin
  if (ItemIndex >= 0) and (ItemIndex < Items.Count) then
  begin
    Item:=TrHTMLListItem(Items[ItemIndex]);
    if Assigned(ItemBitmap) then
    begin
      if not Assigned(Item.ItemBitmap) then
        Item.ItemBitmap:=TBitmap.Create;
      Item.ItemBitmap.Width:=0;
      Item.ItemBitmap.Assign(ItemBitmap);
    end;
    Item.ItemID:=ItemID;
    Item.ItemVarData:=ItemVarData;
    Item.ItemObject:=ItemObject;
    InvalidateRect(Handle, Item.DisplayRect(drBounds), true);
  end;
end;

procedure TrHTMLListView.SetViewStyle(Value: TViewStyle);
var
  TempLV: TrHTMLListView;
begin
  if Value <> ViewStyle then
  begin
    if HandleAllocated and not(csDesigning in ComponentState) then
    begin
      // store all items to another listview
      TempLV:=TrHTMLListView.Create(Owner);
      try
        TempLV.Visible:=false;
        TempLV.Parent:=Parent;
        TempLV.Items.Assign(Items);
        // set style
        inherited;
        // restore items
        Items.Assign(TempLV.Items);
      finally
        FreeAndNil(TempLV);
      end;
    end
    else
      inherited;
  end;
  if HandleAllocated then
    UpdateIconItemSize(true);
end;

procedure TrHTMLListView.UpdateIconItemSize(RefSize: boolean);
begin
  if RefSize then
  begin
    // update image size of internal image list
    if (ViewStyle = vsIcon) and Assigned(LargeImages) then
    begin
      LargeImages.Width:=FItemOptionsEx.FFullItemWidth;
      LargeImages.Height:=FItemOptionsEx.FFullItemHeight;
    end
    else
      if Assigned(SmallImages) then
      begin
        // reset LargeImages to raise SmallImages update event (not raised if Large=Small)
        LargeImages:=nil;
        SmallImages.Width:=FItemOptionsEx.FFullItemWidth div 2;
        SmallImages.Height:=FItemOptionsEx.FFullItemHeight;
        LargeImages:=SmallImages;
      end;
  end;

  if (ViewStyle = vsIcon) and Assigned(LargeImages) then
    ListView_SetIconSpacing(Handle, LargeImages.Width + 12, LargeImages.Height + 28 + FItemOptionsEx.FFullIconItemVSpace)
  else
    ListView_SetIconSpacing(Handle, 0, 0);
  Invalidate;
end;

function TrHTMLListView.IsCustomDrawn(Target: TCustomDrawTarget; Stage: TCustomDrawStage): boolean;
begin
  Result:= inherited;
  if (Stage = cdPrePaint) and (Target = dtItem) and (ViewStyle <> vsReport) then
    Result:=FItemOptionsEx.FCustomDraw;
end;

procedure TrHTMLListView.CMHintShow(var Message: TCMHintShow);
var
  Item: TrHTMLListItem;
  s, h: string;
begin
  // find item and set hint text
  Item:=TrHTMLListItem(GetItemAt(Message.HintInfo.CursorPos.X, Message.HintInfo.CursorPos.Y));
  if Assigned(Item) then
    try
      s:=Item.HintText;
      if s = '' then
        s:=Item.HTMLText;
      if s = '' then
        s:=Item.Caption;
      h:=s;
      DoInfoTip(Item, h);
      if (h = s) and (Item.HintText = '') and not Item.TextTruncated then
        h:='';
      // Set hint string in message
      Message.HintInfo^.HintStr:=h;
    except
    end;
  inherited;
end;

procedure TrHTMLListView.CNNotify(var Message: TWMNotifyLV);
begin
  inherited;
  with Message do
    case NMHdr.code of
      LVN_GETINFOTIPA, LVN_GETINFOTIPW:
        // activate hint window
        Application.ActivateHint(Mouse.CursorPos);
    end;
end;

procedure TrHTMLListView.DrawFrame(ACanvas: TCanvas; var R: TRect; FrameType: TFrameType; aColor: TColor);
var
  rTemp: TRect;
begin
  if (FrameType = ftNone) or ((FrameType <> ft3D) and (aColor = clNone)) then
    Exit;
  // draw frame and reduce Rect according frame width
  ACanvas.Pen.Width:=1;
  ACanvas.Brush.Color:=aColor;
  rTemp:=R;
  case FrameType of
    ftSingleLine:
      BiDi_FrameRect(ACanvas, rTemp);
    ftDoubleLine:
      begin
        BiDi_FrameRect(ACanvas, rTemp);
        InflateRect(rTemp, -1, -1);
        BiDi_FrameRect(ACanvas, rTemp);
      end;
    ft3D:
      begin
        BiDi_FrameRect(ACanvas, rTemp, clBtnHighlight);
        Inc(rTemp.Left);
        Inc(rTemp.Top);
        BiDi_FrameRect(ACanvas, rTemp);
        OffsetRect(rTemp, -1, -1);
        BiDi_FrameRect(ACanvas, rTemp, clBtnShadow);
      end;
  end;
  if FrameType in [ftSingleLine] then
    InflateRect(R, -1, -1);
  if FrameType in [ftDoubleLine, ft3D] then
    InflateRect(R, -2, -2);
end;

procedure TrHTMLListView.DrawThemeBk(ACanvas: TCanvas; R: TRect; State: TCustomDrawState);
var
  ThemeData: HTHEME;
  aStateId: integer;
begin
  if (cdsSelected in State) then
    aStateId:=LIS_SELECTED
  else
    if (cdsHot in State) then
      aStateId:=LIS_HOT
    else
      aStateId:=LIS_NORMAL;
  ThemeData:=OpenThemeData(Handle, VSCLASS_LISTVIEW);
  // InflateRect(R,1,1);
  DrawThemeBackground(ThemeData, ACanvas.Handle, LVP_LISTITEM, aStateId, R, nil);
  CloseThemeData(ThemeData);
end;

function TrHTMLListView.CustomDrawItem(Item: TListItem; State: TCustomDrawState; Stage: TCustomDrawStage): boolean;
var
  rItem: TRect;
  B: TBitmap;
begin
  Result:= inherited CustomDrawItem(Item, State, Stage);
  if Result and (Stage = cdPrePaint) then
  begin
    Result:=false;
    // check and set full item size
    rItem:=Item.DisplayRect(drBounds);
    if Min(rItem.Right - rItem.Left, rItem.Bottom - rItem.Top) > FItemOptionsEx.FFullItemMargin then
    begin
      InflateRect(rItem, -FItemOptionsEx.FFullItemMargin, -FItemOptionsEx.FFullItemMargin);
      // draw to temporary bitmap to reduce flickering
      B:=TBitmap.Create;
      try
        B.Width:=rItem.Right - rItem.Left;
        B.Height:=rItem.Bottom - rItem.Top;
        DrawItemEx(B.Canvas, Rect(0, 0, B.Width, B.Height), TrHTMLListItem(Item), State);
        if Assigned(FOnCustomDrawHTMLItemEnd) then
          FOnCustomDrawHTMLItemEnd(B.Canvas, TrHTMLListItem(Item), Rect(0, 0, B.Width, B.Height));
        Canvas.Draw(rItem.Left, rItem.Top, B);
      finally
        FreeAndNil(B);
      end;
    end;
  end;
end;

procedure TrHTMLListView.DrawItemEx(ACanvas: TCanvas; R: TRect; Item: TrHTMLListItem; State: TCustomDrawState);
var
  Text: string;
  rItem, rImg, rText: TRect;
  ItemBkColor: TColor;
  ItemFrameType: TFrameType;
  ItemFrameColor: TColor;
  TextBkColor: TColor;
  TextFrameType: TFrameType;
  TextFrameColor: TColor;
  IconBmp: TBitmap;
  txtw, txth: integer;

  procedure IntFillRect(var aRect: TRect; aColor: TColor);
  begin
    if aColor = clNone then
      Exit;
    ACanvas.Brush.Style:=bsSolid;
    ACanvas.Brush.Color:=aColor;
    ACanvas.FillRect(aRect);
  end;

  procedure IntDrawBk(FullRect: TRect; var aRect: TRect);
  var
    aColor: TColor;
  begin
    if (cdsSelected in State) then
      aColor:=FItemOptionsEx.FItemBkColorSelected
    else
      if (cdsHot in State) then
        aColor:=FItemOptionsEx.FItemBkColorHot
      else
      begin
        aColor:=ItemBkColor;
        if aColor = clNone then
          aColor:=Color;
      end;
    if aColor <> clNone then
      IntFillRect(aRect, aColor)
    else
      DrawThemeBk(ACanvas, FullRect, State);
  end;

  procedure LoadBmpFromList(ImgList: TCustomImageList; ImgIndex: integer);
  begin
    IconBmp.Width:=ImgList.Width;
    IconBmp.Height:=ImgList.Height;
    IconBmp.Canvas.Brush.Style:=bsSolid;
    IconBmp.Canvas.Brush.Color:=clWhite;
    IconBmp.Canvas.FillRect(Rect(0, 0, IconBmp.Width, IconBmp.Height));
    ImgList.GetBitmap(ImgIndex, IconBmp);
  end;

  procedure SetBmpItemState(StateIndex: integer; StateText: string);
  var
    R: TRect;
  begin
    if Assigned(FItemOptionsEx.FStateImages) and (StateIndex >= 0) then
      LoadBmpFromList(FItemOptionsEx.FStateImages, StateIndex)
    else
    begin
      // draw text to Bitmap
      R:=Rect(0, 0, 0, 0);
      // set font and bk color
      IconBmp.Canvas.Font.Assign(Font);
      if FItemOptionsEx.FImageBkColor <> clNone then
        IconBmp.Canvas.Brush.Color:=FItemOptionsEx.FImageBkColor
      else
        if FItemOptionsEx.FItemBkColor <> clNone then
          IconBmp.Canvas.Brush.Color:=FItemOptionsEx.FItemBkColor
        else
          IconBmp.Canvas.Brush.Color:=Color;
      // IconBmp.Canvas.Brush.Style:=bsClear;
      rHTMLDrawText.rHTML_DrawTextSimple(nil, IconBmp.Canvas, StateText, R, true);
      IconBmp.Width:=R.Right;
      IconBmp.Height:=R.Bottom;
      rHTMLDrawText.rHTML_DrawTextSimple(nil, IconBmp.Canvas, StateText, R, false);
      Item.ForceTransparent:=true;
    end;
  end;

  function GetTextHeight: integer;
  begin
    Result:=ACanvas.TextHeight('Wg') * Max(FItemOptionsEx.FTextLines, 1) + 2 * cTextInnerSpace;
    if TextFrameType <> ftNone then
      Inc(Result);
    if TextFrameType in [ft3D, ftDoubleLine] then
      Inc(Result);
  end;

begin
  IconBmp:=TBitmap.Create;
  try
    // set default props of item
    ACanvas.Font.Assign(Font);
    ItemBkColor:=FItemOptionsEx.FItemBkColor;
    ItemFrameType:=FItemOptionsEx.FItemFrameType;
    ItemFrameColor:=FItemOptionsEx.FItemFrameColor;
    TextBkColor:=FItemOptionsEx.FTextBkColor;
    TextFrameType:=FItemOptionsEx.FTextFrameType;
    TextFrameColor:=FItemOptionsEx.FTextFrameColor;
    // call event to modify default props
    if Assigned(FOnGetItemParams) then
      FOnGetItemParams(Self, Item, ItemBkColor, ItemFrameType, ItemFrameColor, ACanvas.Font, TextBkColor, TextFrameType, TextFrameColor);

    // TEXT content
    Text:=Item.HTMLText;
    if (Text = '') then
      Text:=Item.Caption;

    // IconBmp content
    Item.ForceTransparent:=false;
    if Assigned(Item.ItemBitmap) and (Item.ItemBitmap.Width > 0) then
      IconBmp.Assign(Item.ItemBitmap)
    else
      if Item.ImageNotLoaded then
        SetBmpItemState(FItemOptionsEx.FStateNotLoadedIndex, sLoading)
      else
        if Item.ImageError then
          SetBmpItemState(FItemOptionsEx.FStateErrorIndex, sPictureCorrupted)
        else
          if Item.ImageNotDefined then
            SetBmpItemState(FItemOptionsEx.FStateNotDefinedIndex, sPictureEmpty)
          else
            if Assigned(FItemOptionsEx.StateImages) and (Item.StateImageIndex >= 0) then
              LoadBmpFromList(FItemOptionsEx.StateImages, Item.StateImageIndex)
            else
              if Assigned(FItemOptionsEx.Images) then
                LoadBmpFromList(FItemOptionsEx.Images, Item.ImageIndex);
    IconBmp.Transparent:=Item.ForceTransparent or FItemOptionsEx.FImageTransparent;
    // call event
    if Assigned(FOnGetItemContent) then
      FOnGetItemContent(Self, Item, Text, IconBmp);

    // ITEM frame and bk
    rItem:=R;
    DrawFrame(ACanvas, rItem, ItemFrameType, ItemFrameColor);
    IntDrawBk(R, rItem);

    // TEXT rect
    rText:=rItem;
    InflateRect(rText, -FItemOptionsEx.TextMargin, -FItemOptionsEx.TextMargin);
    txtw:=FItemOptionsEx.TextWidth;
    txth:=FItemOptionsEx.TextHeight;
    if (ViewStyle = vsIcon) then
    begin
      // reduce text rect width
      if (txtw <> 0) and (txtw < rText.Right - rText.Left) then
      begin
        rText.Left:=rText.Left + (rText.Right - rText.Left - txtw) div 2;
        rText.Right:=rText.Left + txtw;
      end;
      // set text rect height
      if txth = 0 then
        txth:=GetTextHeight;
      rText.Top:=Max(rItem.Top + FItemOptionsEx.TextMargin, rText.Bottom - txth - FItemOptionsEx.TextMargin);
    end
    else
    begin
      // set text rect width
      if (txtw = 0) then
        txtw:=(rItem.Right - rItem.Left - FItemOptionsEx.TextMargin) div 2;
      rText.Left:=Max(rItem.Left + FItemOptionsEx.TextMargin, rText.Right - txtw - FItemOptionsEx.TextMargin);
      // reduce text rect height
      if (txth = 0) and (FItemOptionsEx.FTextLines > 0) then
        txth:=GetTextHeight;
      if txth > 0 then
      begin
        rText.Top:=rText.Top + (rText.Bottom - rText.Top - txth) div 2;
        rText.Bottom:=rText.Top + txth;
      end;
    end;

    // IMG rect
    rImg:=rItem;
    InflateRect(rImg, -FItemOptionsEx.ImageMargin, -FItemOptionsEx.ImageMargin);
    if (ViewStyle = vsIcon) then
      rImg.Bottom:=rText.Top - FItemOptionsEx.ImageMargin
    else
      rImg.Right:=rText.Left - FItemOptionsEx.ImageMargin;

    // TEXT frame and bk
    DrawFrame(ACanvas, rText, TextFrameType, TextFrameColor);
    IntFillRect(rText, TextBkColor);
    // TEXT content
    ACanvas.Brush.Style:=bsClear;
    Item.TextTruncated:=rHTMLDrawText.rHTML_DrawText(nil, ACanvas, Text, rText, false, true, FItemOptionsEx.FTextWordWrap, false,
      FItemOptionsEx.FTextAlignment, FItemOptionsEx.FTextLayout, cTextInnerSpace, FItemOptionsEx.FTextTabGridWidth) <> 0;

    if (rImg.Bottom - rImg.Top > 4) and (rImg.Right - rImg.Left > 4) then
    begin
      // IMAGE frame and bk
      DrawFrame(ACanvas, rImg, FItemOptionsEx.FImageFrameType, FItemOptionsEx.FImageFrameColor);
      IntFillRect(rImg, FItemOptionsEx.FImageBkColor);
      DrawItemImageEx(ACanvas, rImg, Item, State, IconBmp);
    end;

    // ITEM focused frame
    if (cdsFocused in State) then
      DrawFrame(ACanvas, R, ftDoubleLine, FItemOptionsEx.FItemFrameColorFocused);

    // store text rect and image rect
    Item.TextRect:=rText;
    OffsetRect(Item.TextRect, FItemOptionsEx.FFullItemMargin, FItemOptionsEx.FFullItemMargin);
    Item.ImageRect:=rImg;
    OffsetRect(Item.ImageRect, FItemOptionsEx.FFullItemMargin, FItemOptionsEx.FFullItemMargin);
  finally
    FreeAndNil(IconBmp);
  end;
end;

procedure TrHTMLListView.DrawItemImageEx(ACanvas: TCanvas; R: TRect; Item: TrHTMLListItem; State: TCustomDrawState; IconBmp: TBitmap);
begin
  if Assigned(IconBmp) and (IconBmp.Width > 0) then
  begin
    if FItemOptionsEx.FImageStretch then
      rpictool.SmoothResizeBitmap(IconBmp, R.Right - R.Left, R.Bottom - R.Top, false);
    IconBmp.Transparent:=Item.ForceTransparent or FItemOptionsEx.FImageTransparent;
    rpictool.DrawBmpCenter(ACanvas, R, IconBmp);
  end;
end;

procedure TrHTMLListView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  Item: TrHTMLListItem;
  rItem: TRect;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if Assigned(FOnClickItem) then
  begin
    Item:=TrHTMLListItem(Self.GetItemAt(X, Y));
    if Assigned(Item) then
    begin
      rItem:=Item.DisplayRect(drBounds);
      Dec(X, rItem.Left);
      Dec(Y, rItem.Top);
      FOnClickItem(Self, Item, Button, Shift, PtInRect(Item.TextRect, Point(X, Y)), PtInRect(Item.ImageRect, Point(X, Y)));
    end;
  end;
end;

procedure TrHTMLListView.LoadImage(ItemIndex: integer; FromFile: boolean; KeyField, ImageField: TField);
var
  Item: TrHTMLListItem;
  s: string;
  GrType: TrGraphicType;
  R: TRect;
begin
  GrType:=gtNone;

  // add and init item
  Item:=TrHTMLListItem(Items[ItemIndex]);
  Item.ImageNotLoaded:=false;
  Item.ImageNotDefined:=false;
  Item.ImageError:=false;
  R:=Item.DisplayRect(drBounds);
  if not Assigned(Item.ItemBitmap) then
    Item.ItemBitmap:=TBitmap.Create;
  Item.ItemBitmap.Width:=0;

  if FromFile then
  begin
    // load image from filename
    s:=VarToStrDef(Item.ItemID, '');
    if (s <> '') and FileExists(s) then
      GrType:=rpictool.LoadBitmapFromFile(Item.ItemBitmap, s, false)
    else
      Item.ImageNotDefined:=true;
  end
  else
  begin
    // load image from record
    if LocateRecord(KeyField, Item.ItemID) then
      GrType:=rpictool.ReadBitmapFromField(ImageField, Item.ItemBitmap)
    else
      Item.ImageNotDefined:=true;
  end;

  if not Item.ImageNotDefined then
  begin
    case GrType of
      gtError:
        Item.ImageError:=true;
      gtNone:
        Item.ImageNotDefined:=true;
    else
      begin
        // store original image size to ItemVarData
        s:=Format('%d;%d', [Item.ItemBitmap.Width, Item.ItemBitmap.Height]);
        Item.ItemVarData:=s;
        // resize image to double item size to save memory
        rpictool.SmoothResizeBitmap(Item.ItemBitmap, (R.Right - R.Left) * 2, (R.Bottom - R.Top) * 2, false);
        // update HTML text according mask
        if FLoadingHTMLMask <> '' then
        begin
          s:=GetMaskText(Item, FLoadingHTMLMask);
          if (not FromFile) then
            s:=rdbtool.FillTextMask(s, KeyField.Dataset);
          Item.HTMLText:=s;
        end;
        if (not FromFile) and (FLoadingHintMask <> '') then
        begin
          s:=GetMaskText(Item, FLoadingHintMask);
          s:=rdbtool.FillTextMask(s, KeyField.Dataset);
          Item.HintText:=s;
        end;
      end;
    end;
  end;
  InvalidateRect(Handle, R, true);
end;

function TrHTMLListView.GetMaskText(Item: TrHTMLListItem; MaskText: string): string;
var
  s, v: string;
begin
  s:=MaskText;
  s:=rtool.ReplaceStr(cMaskFileName, ExtractFileName(VarToStrDef(Item.ItemID, '')), s);
  v:=VarToStrDef(Item.ItemVarData, '');
  s:=rtool.ReplaceStr(cMaskFileW, rtool.FindStrPart(v, ';', 1), s);
  s:=rtool.ReplaceStr(cMaskFileH, rtool.FindStrPart(v, ';', 2), s);
  Result:=s;
end;

procedure TrHTMLListView.LoadFileList(FilePath, FileMask: string; HtmlTextMask: string = ''; SortByFileName: boolean = false; LoadImages: boolean = false);
var
  i: integer;
  L: TStrings;
  Item: TrHTMLListItem;
  oldCursor: TCursor;
begin
  oldCursor:=WCursor;
  if HtmlTextMask = '' then
    HtmlTextMask:=cMaskFileName;
  FLoadingHTMLMask:=HtmlTextMask;
  L:=TStringList.Create;
  try
    // load filenames to list
    rfiletool.LoadFileList(L, rtool.SetEndOfPath(FilePath) + FileMask);
    if SortByFileName then
      TStringList(L).Sort;
    Items.BeginUpdate;
    try
      for i:=0 to L.Count - 1 do
      begin
        Item:=TrHTMLListItem(Items.Add);
        Item.ImageIndex:=-1;
        Item.ImageNotLoaded:=true;
        Item.ItemID:=rtool.SetEndOfPath(FilePath) + L[i];
        Item.HTMLText:=GetMaskText(Item, FLoadingHTMLMask);
      end;
    finally
      Items.EndUpdate;
    end;
  finally
    FreeAndNil(L);
    ChCursor(oldCursor);
  end;
  if LoadImages then
    LoadFileImages(true);
end;

function TrHTMLListView.LoadFileImages(LoadAllItems: boolean = false; LoadSelected: boolean = false): boolean;
var
  i: integer;
begin
  if LoadAllItems then
  begin
    // load image for all items
    FLoadingAborted:=false;
    rdlg.ShowInfoProgressBar(sLoading + '...', Self, LoadingAbortEvent, 0);
    try
      for i:=0 to Items.Count - 1 do
      begin
        FLoadingLastItemIndex:=i;
        LoadImage(i, true, nil, nil);
        if i mod 10 = 0 then
          rdlg.ShowInfoProgressBar(sLoading, Self, LoadingAbortEvent, Round(i * 100 / Items.Count));
        if FLoadingAborted then
          Break;
      end;
    finally
      rdlg.ShowInfo('');
    end;
  end
  else
    if LoadSelected then
    begin
      if Assigned(Selected) and TrHTMLListItem(Selected).ImageNotLoaded then
        LoadImage(Selected.Index, true, nil, nil);
    end
    else
    begin
      // load image for one next item
      Inc(FLoadingLastItemIndex);
      if FLoadingLastItemIndex < 0 then
        FLoadingLastItemIndex:=0;
      if FLoadingLastItemIndex < Items.Count then
        LoadImage(FLoadingLastItemIndex, true, nil, nil);
    end;
  Result:=FLoadingLastItemIndex >= Items.Count - 1;
end;

procedure TrHTMLListView.LoadDBList(Dataset: TDataset; KeyField, ImageField: TField; HtmlTextMask: string = ''; HintTextMask: string = '';
  LoadImages: boolean = false);
var
  Item: TrHTMLListItem;
  oldCursor: TCursor;
begin
  if not Assigned(Dataset) or not Dataset.Active or not Assigned(KeyField) then
    Exit;
  oldCursor:=WCursor;
  FLoadingHTMLMask:=HtmlTextMask;
  FLoadingHintMask:=HintTextMask;
  Items.BeginUpdate;
  try
    Dataset.First;
    while not Dataset.Eof do
    begin
      Item:=TrHTMLListItem(Items.Add);
      if HtmlTextMask = '' then
        Item.HTMLText:=KeyField.DisplayText
      else
        Item.HTMLText:=rdbtool.FillTextMask(GetMaskText(Item, FLoadingHTMLMask), Dataset);
      Item.HintText:=rdbtool.FillTextMask(GetMaskText(Item, FLoadingHintMask), Dataset);
      Item.ImageIndex:=-1;
      Item.ImageNotLoaded:=true;
      Item.ItemID:=KeyField.AsVariant;
      Dataset.Next
    end;
  finally
    Items.EndUpdate;
    ChCursor(oldCursor);
  end;
  if LoadImages then
    LoadDBImages(Dataset, KeyField, ImageField, true);
end;

function TrHTMLListView.LoadDBImages(Dataset: TDataset; KeyField, ImageField: TField; LoadAllItems: boolean = false; LoadSelected: boolean = false): boolean;
var
  i: integer;
begin
  Result:=false;
  if not Assigned(Dataset) or not Dataset.Active or not Assigned(KeyField) or not Assigned(ImageField) then
    Exit;
  if LoadAllItems then
  begin
    FLoadingAborted:=false;
    rdlg.ShowInfoProgressBar(sLoading + '...', Self, LoadingAbortEvent, 0);
    // Dataset.DisableControls;
    try
      for i:=0 to Items.Count - 1 do
      begin
        FLoadingLastItemIndex:=i;
        LoadImage(i, false, KeyField, ImageField);
        if i mod 10 = 0 then
          rdlg.ShowInfoProgressBar(sLoading, Self, LoadingAbortEvent, Round(i * 100 / Items.Count));
        if FLoadingAborted then
          Break;
        Dataset.Next;
      end;
    finally
      // Dataset.EnableControls;
      rdlg.ShowInfo('');
    end;
  end
  else
    if LoadSelected then
    begin
      if Assigned(Selected) and TrHTMLListItem(Selected).ImageNotLoaded then
        LoadImage(Selected.Index, false, KeyField, ImageField);
    end
    else
    begin
      Inc(FLoadingLastItemIndex);
      if FLoadingLastItemIndex < 0 then
        FLoadingLastItemIndex:=0;
      if FLoadingLastItemIndex < Items.Count then
        LoadImage(FLoadingLastItemIndex, false, KeyField, ImageField);
    end;
  Result:=FLoadingLastItemIndex >= Items.Count - 1;
end;

procedure TrHTMLListView.LoadingAbortEvent(Sender: TObject);
begin
  FLoadingAborted:=true;
end;

end.
