unit rHintWindow;

{<
  rHintWindow unit contain TrHTMLHint - non visual component and function
  providing enhanced hint window with images and HTML support.

  Version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2015)
  @lastmod(Feb 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Classes, Controls, Types, Windows, Graphics, Forms, SysUtils,
  StdCtrls, ExtCtrls, StrUtils, Math, rImprovedComps, RosiCompGlobalCfg;

type
  { HintWindow exception class used by TrHTMLHint }
  ErHTMLHintError = class(Exception)
  end;

  { non visual component which define HTML Hint window properties,
    placed on Main form activate new HTML hint window }
  TrHTMLHint = class(TComponent)
  private
    FHintFont: TFont;
    FParentFont: boolean;
    FMinWidth: Integer;
    FMaxWidth: Integer;
    FMinHeight: Integer;
    FMaxHeight: Integer;
    FBorderLeft: byte;
    FBorderRight: byte;
    FBorderTop: byte;
    FBorderBottom: byte;
    FTabGridWidth: byte;
    FImages: TImageList;
    FLeftBarWidth: byte;
    FLeftBarColor: TColor;
    FLeftBarImageIndex: Integer;
    FRightBarWidth: byte;
    FRightBarColor: TColor;
    FRightBarImageIndex: Integer;
    FRichText: TrRichEdit;
    FHintImage: TPicture;
    FHintSkipEnter: Boolean;
    FHintPlainText: Boolean;
    FUseStyleFontColor: Boolean;
    FUseAsGlobalHint: Boolean;
    FOnPaint: TNotifyEvent;
    procedure SetParentFont(Value: boolean);
    procedure SetHintFont(const Value: TFont);
    function HintFontStored: boolean;
    procedure HintFontChanged(Sender: TObject);
    procedure SetHintImage(const Value: TPicture);
    procedure SetUseAsGlobalHint(const Value: Boolean);
  protected
    procedure Loaded; override; {< @exclude }
    procedure Notification(AComponent: TComponent; AOperation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property HintFont: TFont read FHintFont write SetHintFont stored HintFontStored;
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    property MinWidth: Integer read FMinWidth write FMinWidth default 0;
    property MaxWidth: Integer read FMaxWidth write FMaxWidth default 0;
    property MinHeight: Integer read FMinHeight write FMinHeight default 0;
    property MaxHeight: Integer read FMaxHeight write FMaxHeight default 0;
    property BorderLeft: byte read FBorderLeft write FBorderLeft default 2;
    property BorderRight: byte read FBorderRight write FBorderRight default 2;
    property BorderTop: byte read FBorderTop write FBorderTop default 2;
    property BorderBottom: byte read FBorderBottom write FBorderBottom default 2;
    property TabGridWidth: byte read FTabGridWidth write FTabGridWidth default 90;
    property Images: TImageList read FImages write FImages;
    property LeftBarWidth: byte read FLeftBarWidth write FLeftBarWidth default 16;
    property LeftBarColor: TColor read FLeftBarColor write FLeftBarColor default clSilver;
    property LeftBarImageIndex: Integer read FLeftBarImageIndex write FLeftBarImageIndex default - 1;
    property RightBarWidth: byte read FRightBarWidth write FRightBarWidth default 0;
    property RightBarColor: TColor read FRightBarColor write FRightBarColor default clSilver;
    property RightBarImageIndex: Integer read FRightBarImageIndex write FRightBarImageIndex default - 1;
    property HintImage: TPicture read FHintImage write SetHintImage;
    property HintSkipEnter: Boolean read FHintSkipEnter write FHintSkipEnter default false;
    property HintPlainText: Boolean read FHintPlainText write FHintPlainText default false;
    property OnPaint: TNotifyEvent read FOnPaint write FOnPaint;
    property UseStyleFontColor: Boolean read FUseStyleFontColor write FUseStyleFontColor default true;
    property UseAsGlobalHint: Boolean read FUseAsGlobalHint write SetUseAsGlobalHint default false;
  end;

  { @exclude }
  TrHTMLHintWindow = class(THintWindow)
  private
    HintComponent: TrHTMLHint;
    bl, br, bt, bb, LBw, RBw: byte;
    LBColor, RBColor: TColor;
    Bimg: TImageList;
    LBImgIdx, RBImgIdx, TabW, ImagePos: Integer;
    LabelRect, ImageRect: TRect;
    LabelWrap, SkipEnter, PlainText, RichText: Boolean;
    TextHint: string;
    OnP: TNotifyEvent;
    function FindHintComponent: TrHTMLHint;
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ActivateHint(Rect: TRect; const AHint: string); override;
  published
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rpictool, rtool, rHTMLDrawText;

procedure Register;
begin
  RegisterComponents('Rosi', [TrHTMLHint]);
end;

constructor TrHTMLHint.Create(AOwner: TComponent);
var
  I, Instances: Integer;
begin
  inherited Create(AOwner);
  if not(Owner is TForm) then
    raise ErHTMLHintError.Create('rHTMLHint component must be placed on a form.');

  Instances:=0;
  for I:=0 to Owner.ComponentCount - 1 do
    if (Owner.Components[I] is TrHTMLHint) then
      Inc(Instances);
  if (Instances > 1) then
    raise ErHTMLHintError.Create('There can be only one instance of TrHTMLHint.');

  if not(csDesigning in ComponentState) then
  begin
    // set hint window class
    HintWindowClass:=TrHTMLHintWindow;
    Application.ShowHint:=not Application.ShowHint;
    Application.ShowHint:=not Application.ShowHint;
    FRichText:=TrRichEdit.Create(Self);
    FRichText.Visible:=false;
    //FRichText.Parent:=self;
    FRichText.BorderStyle:=bsNone;
  end;

  FUseStyleFontColor:=true;
  FHintImage:=TPicture.Create;
  FHintFont:=TFont.Create;
  FHintFont.Assign(TForm(Owner).Font);
  FHintFont.OnChange:=HintFontChanged;
  FParentFont:=true;
  FMaxWidth:=0;
  FBorderLeft:=2;
  FBorderRight:=2;
  FBorderTop:=2;
  FBorderBottom:=2;
  FTabGridWidth:=90;
  FLeftBarWidth:=16;
  FLeftBarColor:=clSilver;
  FLeftBarImageIndex:=-1;
  FRightBarWidth:=0;
  FRightBarColor:=clSilver;
  FRightBarImageIndex:=-1;
  FHintSkipEnter:=false;
  FHintPlainText:=false;
  FUseAsGlobalHint:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrHTMLHint.Destroy;
begin
  FreeAndNil(FHintImage);
  FreeAndNil(FHintFont);
  inherited Destroy;
end;

procedure TrHTMLHint.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrHTMLHint.Notification(AComponent: TComponent; AOperation: TOperation);
begin
  if (AOperation = opRemove) and (AComponent = FImages) then
    FImages:=nil;
  inherited;
end;

procedure TrHTMLHint.SetHintFont(const Value: TFont);
begin
  FHintFont.Assign(Value);
end;

procedure TrHTMLHint.SetParentFont(Value: Boolean);
begin
  if Value then
    FHintFont.Assign(TForm(Owner).Font);
  FParentFont:=Value;
end;

function TrHTMLHint.HintFontStored: boolean;
begin
  Result:=not FParentFont;
end;

procedure TrHTMLHint.HintFontChanged(Sender: TObject);
begin
  FParentFont:=false;
end;

procedure TrHTMLHint.SetHintImage(const Value: TPicture);
begin
  FHintImage.Assign(Value);
end;

procedure TrHTMLHint.SetUseAsGlobalHint(const Value: Boolean);
begin
  if FUseAsGlobalHint <> Value then
  begin
    FUseAsGlobalHint:=Value;
    if FUseAsGlobalHint then
    begin
      RosiCompGlobalCfg.RosiCompConfig.Global_HTMLHintComponent:=Self;
      RosiCompGlobalCfg.RosiCompConfig.Global_HTMLHintPicture:=FHintImage;
    end;
  end;
end;

constructor TrHTMLHintWindow.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
end;

destructor TrHTMLHintWindow.Destroy;
begin
  inherited Destroy;
end;

procedure TrHTMLHintWindow.ActivateHint(Rect: TRect; const AHint: string);
var
  minw, maxw, minh, maxh, imgw, imgh, actw, acth: Integer;
  koef: double;
  HTMLMode: TrHTMLMode;
begin
  if Assigned(RosiCompConfig.Global_HTMLHintComponent) and (RosiCompConfig.Global_HTMLHintComponent is TrHTMLHint) then
    HintComponent:=TrHTMLHint(RosiCompConfig.Global_HTMLHintComponent)
  else
    HintComponent:=FindHintComponent;

  // check hint length
  if Length(AHint) > 1000 then
    TextHint:=copy(AHint, 1, 1000) + '...'
  else
    TextHint:=AHint;

  // init all settings
  ImagePos:=0;
  imgw:=0;
  imgh:=0;
  if Assigned(HintComponent) then
  begin
    if HintComponent.ParentFont then
      Canvas.Font.Assign(TForm(HintComponent.Owner).Font)
    else
      Canvas.Font.Assign(HintComponent.FHintFont);
    //DPI_ScaleFontHeight(Canvas.Font, Screen.PixelsPerInch, 96);

    bl:=HintComponent.FBorderLeft;
    br:=HintComponent.FBorderRight;
    bt:=HintComponent.FBorderTop;
    bb:=HintComponent.FBorderBottom;
    LBw:=HintComponent.FLeftBarWidth;
    LBColor:=HintComponent.FLeftBarColor;
    LBImgIdx:=HintComponent.FLeftBarImageIndex;
    RBw:=HintComponent.FRightBarWidth;
    RBColor:=HintComponent.FRightBarColor;
    RBImgIdx:=HintComponent.FRightBarImageIndex;
    Bimg:=HintComponent.FImages;
    minw:=HintComponent.FMinWidth;
    maxw:=HintComponent.FMaxWidth;
    minh:=HintComponent.FMinHeight;
    maxh:=HintComponent.FMaxHeight;
    OnP:=HintComponent.FOnPaint;
    TabW:=HintComponent.FTabGridWidth;
    ImagePos:=PosEx(rHTMLHintImageTag, AHint);
    if (ImagePos > 0) and Assigned(HintComponent.FHintImage) then
    begin
      TextHint:=rtool.ReplaceStr(rHTMLHintImageTag, '', TextHint);
      imgw:=HintComponent.FHintImage.Width;
      imgh:=HintComponent.FHintImage.Height;
    end;
    SkipEnter:=HintComponent.FHintSkipEnter;
    PlainText:=HintComponent.FHintPlainText;
    RichText:=PosEx(rHTMLHintRichTextTag, AHint)>0;
    if RichText then
      TextHint:=rtool.ReplaceStr(rHTMLHintRichTextTag, '', AHint);
  end
  else
  begin
    bl:=2;
    br:=2;
    bt:=2;
    bb:=2;
    LBw:=16;
    LBColor:=clSilver;
    LBImgIdx:=-1;
    RBw:=0;
    RBColor:=clSilver;
    RBImgIdx:=-1;
    Bimg:=nil;
    minw:=0;
    maxw:=0;
    minh:=0;
    maxh:=0;
    OnP:=nil;
    TabW:=90;
    SkipEnter:=false;
    PlainText:=false;
  end;
  Inc(br, 2);

  if PlainText then
    HTMLMode:=hmPlainText
  else
    if SkipEnter then
      HTMLMode:=hmHTMLSkipEnterAndLink
    else
      HTMLMode:=hmHTMLSkipLink;
  // calc text rect
  LabelRect:=Types.Rect(bl + LBw, bt, bl + LBw, bt);
  ImageRect:=Types.Rect(bl + LBw, bt, bl + LBw + imgw, bt + imgh);
  LabelWrap:=false;
  if Trim(TextHint) <> '' then
    rHTMLDrawText.rHTML_DrawText(nil, Self.Canvas, TextHint, LabelRect, true, true, LabelWrap, false, taLeftJustify, tlTop, 0, TabW, HTMLMode);

  // check min max width
  Rect.Right:=Rect.Left + Max(minw, Max(LabelRect.Right, ImageRect.Right) + br + RBw);
  if maxw = 0 then
    maxw:=Screen.Width - 100;
  if (maxw > 0) and (Rect.Right - Rect.Left > maxw) then
  begin
    // ensure min width of label
    if maxw < LabelRect.Left + 60 + RBw then
      maxw:=LabelRect.Left + 60 + RBw;
    Rect.Right:=Rect.Left + maxw;
    actw:=Rect.Right - Rect.Left - br - RBw;
    // recalc new text rect with word wrap
    if LabelRect.Right > actw then
    begin
      LabelRect.Right:=actw;
      LabelWrap:=true;
      if Trim(TextHint) <> '' then
        rHTMLDrawText.rHTML_DrawText(nil, Self.Canvas, TextHint, LabelRect, true, true, LabelWrap, false, taLeftJustify, tlTop, 0, TabW, HTMLMode);
    end;
    // resize ImageRect
    if ImageRect.Right > actw then
    begin
      koef:=(actw - ImageRect.Left) / (ImageRect.Right - ImageRect.Left);
      ImageRect.Right:=actw;
      ImageRect.Bottom:=ImageRect.Top + Round((ImageRect.Bottom - ImageRect.Top) * koef);
    end;
  end;
  if (Trim(TextHint) <> '') and RichText then
  begin
    HintComponent.FRichText.Parent:=Self;
    HintComponent.FRichText.RtfText:=TextHint;
    HintComponent.FRichText.CalculateTextHeight(LabelRect);
  end;

  // align text and image
  if (ImagePos = 1) then
  begin
    LabelRect.Bottom:=ImageRect.Bottom + 2 + LabelRect.Bottom - LabelRect.Top;
    LabelRect.Top:=ImageRect.Bottom + 2;
  end
  else
    if (ImagePos > 1) then
    begin
      ImageRect.Bottom:=LabelRect.Bottom + 2 + ImageRect.Bottom - ImageRect.Top;
      ImageRect.Top:=LabelRect.Bottom + 2;
    end;

  // calc height
  Rect.Bottom:=Rect.Top + Max(minh, Max(LabelRect.Bottom, ImageRect.Bottom) + bb);
  // check min max height
  if maxh = 0 then
    maxh:=Screen.Height - 100;
  if (maxh > 0) and (Rect.Bottom - Rect.Top > maxh) then
  begin
    Rect.Bottom:=Rect.Top + maxh;
    acth:=Rect.Bottom - Rect.Top - bb;
    // resize image
    if ImageRect.Bottom > acth then
    begin
      // add space for label below
      if (ImagePos = 1) then
        Dec(acth, LabelRect.Bottom - LabelRect.Top);
      if acth <= ImageRect.Top then
      begin
        // not enough space - disable image
        ImagePos:=0;
        LabelRect.Top:=ImageRect.Top;
      end
      else
      begin
        koef:=(acth - ImageRect.Top) / (ImageRect.Bottom - ImageRect.Top);
        ImageRect.Bottom:=acth;
        ImageRect.Right:=ImageRect.Left + Round((ImageRect.Right - ImageRect.Left) * koef);
      end;
      // update label pos
      if (ImagePos = 1) then
        LabelRect.Top:=ImageRect.Bottom + 2;
      LabelRect.Bottom:=Rect.Bottom - Rect.Top - bb;
      // recalc max width
      Rect.Right:=Rect.Left + Max(minw, Max(LabelRect.Right, ImageRect.Right) + br + RBw);
    end;
  end;

  UpdateBoundsRect(Rect);
  inherited ActivateHint(Rect, '');
end;

procedure TrHTMLHintWindow.Paint;
var
  Bmp: TBitmap;
  R: TRect;
  HTMLMode: TrHTMLMode;
begin
  inherited;
  // draw text
  Canvas.Brush.Style:=bsClear;
  if Assigned(HintComponent) then
  begin
    if HintComponent.ParentFont then
      Canvas.Font.Assign(TForm(HintComponent.Owner).Font)
    else
      Canvas.Font.Assign(HintComponent.FHintFont);
    if HintComponent.FUseStyleFontColor then
      Canvas.Font.Color:=GetStyleFontColor_LabelText(Canvas.Font.Color);
    //DPI_ScaleFontHeight(Canvas.Font, Screen.PixelsPerInch, 96);
  end;
  if PlainText then
    HTMLMode:=hmPlainText
  else
    if SkipEnter then
      HTMLMode:=hmHTMLSkipEnterAndLink
    else
      HTMLMode:=hmHTMLSkipLink;
  if RichText then
  begin
    HintComponent.FRichText.Parent:=Self;
    HintComponent.FRichText.RtfText:=TextHint;
    HintComponent.FRichText.DrawToCanvas(LabelRect, 0, 0, Self.Canvas, true);
  end
  else
    rHTMLDrawText.rHTML_DrawText(nil, Self.Canvas, TextHint, LabelRect, false, true, LabelWrap, false, taLeftJustify, tlTop, 0, TabW, HTMLMode);
  // draw image
  if (ImagePos > 0) and Assigned(HintComponent) then
    rpictool.DrawPictureCenter(Canvas, ImageRect, HintComponent.HintImage, true, false);
  // draw left border
  if LBw > 0 then
  begin
    R.Left:=0;
    R.Top:=0;
    R.Right:=LBw - 1;
    R.Bottom:=Height;
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=LBColor;
    Canvas.Pen.Color:=LBColor;
    Canvas.Rectangle(R);
    if Assigned(Bimg) and (LBImgIdx >= 0) and (LBImgIdx < Bimg.Count) then
    begin
      Bmp:=TBitmap.Create;
      try
        Bmp.Canvas.Brush.Color:=LBColor;
        Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
        Bmp.Transparent:=true;
        Bimg.GetBitmap(LBImgIdx, Bmp);
        R.Bottom:=R.Top + R.Right - R.Left;
        if Bmp.Height > R.Bottom - R.Top then
          R.Bottom:=R.Top + Bmp.Height + 4;
        rpictool.DrawBmpCenter(Self.Canvas, R, Bmp);
      finally
        FreeAndNil(Bmp);
      end;
    end;
  end;
  // draw right border
  if RBw > 0 then
  begin
    R.Left:=Width - RBw;
    R.Top:=0;
    R.Right:=Width;
    R.Bottom:=Height;
    Canvas.Brush.Style:=bsSolid;
    Canvas.Brush.Color:=RBColor;
    Canvas.Pen.Color:=RBColor;
    Canvas.Rectangle(R);
    if Assigned(Bimg) and (RBImgIdx >= 0) and (RBImgIdx < Bimg.Count) then
    begin
      Bmp:=TBitmap.Create;
      try
        Bmp.Canvas.Brush.Color:=RBColor;
        Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
        Bmp.Transparent:=true;
        Bimg.GetBitmap(RBImgIdx, Bmp);
        R.Bottom:=R.Top + R.Right - R.Left;
        if Bmp.Height > R.Bottom - R.Top then
          R.Bottom:=R.Top + Bmp.Height + 4;
        rpictool.DrawBmpCenter(Self.Canvas, R, Bmp);
      finally
        FreeAndNil(Bmp);
      end;
    end;
  end;
  // call user event
  if Assigned(OnP) then
    OnP(Self);
end;

function TrHTMLHintWindow.FindHintComponent: TrHTMLHint;
var
  I: Integer;
begin
  Result:=nil;
  if Assigned(Application.MainForm) then
    with Application.MainForm do
      for I:=0 to ComponentCount - 1 do
        if Components[I] is TrHTMLHint then
        begin
          Result:=TrHTMLHint(Components[I]);
          Break;
        end;
end;

end.
