unit rMapiMail;

{<
  rMapiMail unit contains TrMapiEmail - non visual component which provides all function
  for sending of email by MAPI interface, single or batch sending is supported.


  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2014)
  @lastmod(Jul 2019)
}

interface

{$WARN SYMBOL_DEPRECATED OFF}

uses Classes, Graphics, Windows, SysUtils, StrUtils, Forms, Controls,
  Variants, OleCtrls, ComObj;

const
  // batch mail application constant
  smRunFile = 'SendMail.exe';
  smIniFile = 'SendMail.ini';
  smSendTo = 'SendTo';
  smCopyTo = 'CopyTo';
  smBlindCopyTo = 'BlindCopyTo';
  smSubject = 'Subject';
  smBody = 'Body';
  smAttachedFile = 'AttachedFile';
  smShowDialog = 'ShowDialog';
  smUseOutlook = 'UseOutlook';
  smNextMail = 'NextMail';
  smConfirmCounter = 'ConfirmCounter';
  smUseNameOnlyPrefix = 'NAME:';

type
  { list of RunMail sending types }
  TRunMailType = (rmInternal, rmExternalAppOnly, rmExternalAppPrefered);
  { list of RunMail MAPI version types }
  TRunMailMapiVersion = (rmAuto, rmAnsi, rmUnicode);
  { list of RunMail MAPI / Outlook preferences }
  TRunMailMapiOutlook = (rmMapiOnly, rmOutlookOnly, rmMapiPreferred, rmOutlookPreferred);

  { outlook options }
  TOutlookOption = (ooUseRunningApplication, ooShowEmailModal);
  { outlook options set }
  TOutlookOptions = set of TOutlookOption;

  { non visual components which provides all function for sending of email
    by MAPI interface, single or batch sending is supported }
  TrMapiEmail = class(TComponent)
  private
    FBatchConfirmCount: byte;
    FSendTo: TStrings;
    FCopyTo: TStrings;
    FBlindCopyTo: TStrings;
    FSubject: string;
    FBody: TStrings;
    FAttachments: TStrings;
    FShowErrorDlg: boolean;
    FRemoveRecepientsDuplicity: boolean;
    FMailIniCfg: TStrings;
    FBatchMailList: TStrings;
    FMailAppDirectory: string;
    FAddressPrefix: string;
    FSendAutomatically: boolean;
    FUseOutlook: boolean;
    FOutlookOptions: TOutlookOptions;
    procedure SetSendTo(Value: TStrings);
    procedure SetCopyTo(Value: TStrings);
    procedure SetBlindCopyTo(Value: TStrings);
    procedure SetBody(Value: TStrings);
    procedure SetAttachments(Value: TStrings);
  protected
    function GetNameAddress(const s: string; var name, address: string): boolean;
    procedure CleanStrings(L: TStrings; CheckFileName: boolean = false);
    procedure CleanAll;
    procedure AddMailList(UseOutlook: boolean);
    function ProcMapiMail: boolean;
    function ProcMapiMailW: boolean;
    function ProcOutlookOLE: boolean;
    function RunExternalMailTool(L: TStrings): boolean;
    function ProcBatchMail: boolean;
    function GetMailAppDir: string;
    function GetFirstHTMLAttachmentIndex: integer;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { check if external application exists }
    function ExternalAppExists: boolean;
    { used by external mail application to load ini file and send emails }
    procedure ExternalAppLoadSendMail(IniFileName: string = ''; DeleteAfterSend: boolean = true);
    { init all mail variables }
    procedure InitMail;
    { send mail via mailto command (limited body lenght and without attachment) }
    function SendSimpleMail: boolean;
    { send mail via MAPI }
    function SendMapiMail(RunMailType: TRunMailType = rmExternalAppPrefered; RunMailVersion: TRunMailMapiVersion = rmAuto): boolean;
    { send mail via Outlook OLE }
    function SendOutlookMail(RunMailType: TRunMailType = rmExternalAppPrefered): boolean;
    { send mail via mailto command or MAPI or Outlook according mail content }
    function SendAutoSelect(MapiOrOutlook: TRunMailMapiOutlook = rmOutlookPreferred): boolean;
    { add mail to batch list }
    procedure AddToBatchMail(UseOutlook: boolean = false);
    { init batch list }
    procedure InitBatchMail;
    { send batch emails }
    function SendBatchMail(RunMailType: TRunMailType = rmExternalAppPrefered): boolean;
    { return list of mails in batch list }
    property BatchMailList: TStrings read FBatchMailList;
  published
    { define number of sent mail when Continue confirm dialog is shown }
    property BatchConfirmCount: byte read FBatchConfirmCount write FBatchConfirmCount default 10;
    { define recipient addresses }
    property SendTo: TStrings read FSendTo write SetSendTo;
    { define copy recipient addresses }
    property CopyTo: TStrings read FCopyTo write SetCopyTo;
    { define blind copy recipient addresses }
    property BlindCopyTo: TStrings read FBlindCopyTo write SetBlindCopyTo;
    { define mail subject text }
    property Subject: string read FSubject write FSubject;
    { define mail body text }
    property Body: TStrings read FBody write SetBody;
    { define list of attachments }
    property Attachments: TStrings read FAttachments write SetAttachments;
    { define whether error dialog is show }
    property ShowErrorDlg: boolean read FShowErrorDlg write FShowErrorDlg default false;
    { define location of SendMail.exe as external app }
    property MailAppDirectory: string read FMailAppDirectory write FMailAppDirectory;
    { define automatic removal of duplicity in the email addresses }
    property RemoveRecepientsDuplicity: boolean read FRemoveRecepientsDuplicity write FRemoveRecepientsDuplicity default true;
    { define whether email is sent without any prompt (if MAPI client allows it) }
    property SendAutomatically: boolean read FSendAutomatically write FSendAutomatically default false;
    { define options for Outlook email }
    property OutlookOptions: TOutlookOptions read FOutlookOptions write FOutlookOptions default [ooUseRunningApplication];
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rdlg, Mapi, MapiW;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi', [TrMapiEmail]);
end;

constructor TrMapiEmail.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBatchConfirmCount:=10;
  FShowErrorDlg:=false;
  FSendTo:=TStringList.Create;
  FCopyTo:=TStringList.Create;
  FBlindCopyTo:=TStringList.Create;
  FBody:=TStringList.Create;
  FAttachments:=TStringList.Create;
  FBatchMailList:=TStringList.Create;
  FMailIniCfg:=TStringList.Create;
  FRemoveRecepientsDuplicity:=true;
  FSendAutomatically:=false;
  FOutlookOptions:=[ooUseRunningApplication];
  // FAddressPrefix:='';
  FAddressPrefix:='SMTP:';
end;

destructor TrMapiEmail.Destroy;
begin
  FreeAndNil(FSendTo);
  FreeAndNil(FCopyTo);
  FreeAndNil(FBlindCopyTo);
  FreeAndNil(FBody);
  FreeAndNil(FAttachments);
  FreeAndNil(FBatchMailList);
  FreeAndNil(FMailIniCfg);
  inherited Destroy;
end;

procedure TrMapiEmail.SetSendTo(Value: TStrings);
begin
  FSendTo.Assign(Value);
end;

procedure TrMapiEmail.SetCopyTo(Value: TStrings);
begin
  FCopyTo.Assign(Value);
end;

procedure TrMapiEmail.SetBlindCopyTo(Value: TStrings);
begin
  FBlindCopyTo.Assign(Value);
end;

procedure TrMapiEmail.SetBody(Value: TStrings);
begin
  FBody.Assign(Value);
end;

procedure TrMapiEmail.SetAttachments(Value: TStrings);
begin
  FAttachments.Assign(Value);
end;

function TrMapiEmail.GetNameAddress(const s: string; var name, address: string): boolean;
begin
  name:='';
  address:='';
  if rtool.AnsiPosEx(smUseNameOnlyPrefix, s) = 1 then
    name:=Trim(copy(s, Length(smUseNameOnlyPrefix) + 1, Length(s)))
  else
  begin
    name:=Trim(s);
    address:=name;
  end;
  Result:=(name <> '');
end;

procedure TrMapiEmail.CleanStrings(L: TStrings; CheckFileName: boolean = false);
var
  i: integer;
begin
  for i:=L.Count - 1 downto 0 do
  begin
    L[i]:=Trim(L[i]);
    if L[i] = '' then
      L.Delete(i)
    else
      if CheckFileName and not FileExists(L[i]) then
      begin
        DlgW(Format(sFileNotFound, [L[i]]));
        L.Delete(i);
      end;
  end;
end;

procedure TrMapiEmail.CleanAll;
begin
  CleanStrings(FSendTo);
  CleanStrings(FCopyTo);
  CleanStrings(FBlindCopyTo);
  CleanStrings(FAttachments, true);
end;

function TrMapiEmail.ExternalAppExists: boolean;
begin
  Result:=FileExists(GetMailAppDir + smRunFile);
end;

function TrMapiEmail.GetFirstHTMLAttachmentIndex: integer;
var
  i: Integer;
  ext: string;
begin
  Result:=-1;
  for i := 0 to FAttachments.Count - 1 do
  begin
    ext:=ExtractFileExt(FAttachments[i]);
    if SameText(ext,'.HTM') or SameText(ext,'.HTML') then
    begin
      Result:=i;
      Exit;
    end;
  end;
end;

procedure TrMapiEmail.ExternalAppLoadSendMail(IniFileName: string = ''; DeleteAfterSend: boolean = true);
begin
  if IniFileName = '' then
    IniFileName:=GetMailAppDir + smIniFile;
  try
    BatchMailList.LoadFromFile(IniFileName);
    SendBatchMail(rmInternal);
    if DeleteAfterSend then
      DeleteFile(IniFileName);
  except
    on E: Exception do
      DlgE(E.Message);
  end;
end;

function TrMapiEmail.ProcMapiMail: boolean;
type
  TMapiRecipDescArray = array [0 .. 0] of TMapiRecipDesc;
  PMapiRecipDescArray = ^TMapiRecipDescArray;
  TMapiFileDescArray = array [0 .. 0] of TMapiFileDesc;
  PMapiFileDescArray = ^TMapiFileDescArray;
var
  dwRet: Cardinal;
  // MAPI_Session: Cardinal;
  MapiMessage: TMapiMessage;
  Recip: PMapiRecipDescArray;
  Attachments: PMapiFileDescArray;
  WndList: Pointer;
  idx: integer;
  FileName: string;
  MapiFlag: integer;

  procedure AddRecip(Adr: TStrings; RecipClass: Cardinal);
  var
    i: integer;
    n, a: string;
  begin
    for i:=0 to Adr.Count - 1 do
      if GetNameAddress(Adr.Strings[i], n, a) then
      begin
        Recip^[idx].ulRecipClass:=RecipClass;
        if n = '' then
          Recip^[idx].lpszName:=nil
        else
          Recip^[idx].lpszName:=StrNew(PAnsiChar(ansistring(n)));
        if a = '' then
          Recip^[idx].lpszAddress:=nil
        else
          Recip^[idx].lpszAddress:=StrNew(PAnsiChar(ansistring(FAddressPrefix + a)));
        Inc(idx);
      end;
  end;

  procedure RemoveDuplic(L: TStrings);
  var
    LTemp: TStrings;
    i: integer;
  begin
    LTemp:=TStringList.Create;
    try
      for i:=0 to L.Count - 1 do
      begin
        if Trim(L[i]) = '' then
          Continue;
        if LTemp.IndexOf(Trim(L[i])) < 0 then
          LTemp.Add(Trim(L[i]));
      end;
      L.Assign(LTemp);
    finally
      FreeAndNil(LTemp);
    end;
  end;

begin
  { dwRet:=MapiLogon(Application.Handle, PAnsiChar(''), PAnsiChar(''), MAPI_LOGON_UI or MAPI_NEW_SESSION, 0, @MAPI_Session);
    if (dwRet <> SUCCESS_SUCCESS) then
    begin
    MessageBox(Application.Handle, PChar('Error while trying to send email'), PChar('Error'), MB_ICONERROR or MB_OK);
    Exit;
    end;
    }
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  CleanAll;
  FillChar(MapiMessage, SizeOf(MapiMessage), #0);
  Recip:=nil;
  Attachments:=nil;
  try
    // addresses
    if FRemoveRecepientsDuplicity then
    begin
      RemoveDuplic(FSendTo);
      RemoveDuplic(FCopyTo);
      RemoveDuplic(FBlindCopyTo);
    end;
    MapiMessage.nRecipCount:=FSendTo.Count + FCopyTo.Count + FBlindCopyTo.Count;
    if MapiMessage.nRecipCount > 0 then
    begin
      GetMem(Recip, MapiMessage.nRecipCount * SizeOf(TMapiRecipDesc));
      FillChar(Recip^, MapiMessage.nRecipCount * SizeOf(TMapiRecipDesc), 0);
      idx:=0;
      AddRecip(FSendTo, MAPI_TO);
      AddRecip(FCopyTo, MAPI_CC);
      AddRecip(FBlindCopyTo, MAPI_BCC);
      MapiMessage.lpRecips:=PMapiRecipDesc(Recip);
    end;

    // subject and body
    MapiMessage.lpszSubject:=StrNew(PAnsiChar(ansistring(FSubject)));
    MapiMessage.lpszNoteText:=StrNew(PAnsiChar(ansistring(FBody.Text)));

    // attachments
    MapiMessage.nFileCount:=FAttachments.Count;
    if MapiMessage.nFileCount > 0 then
    begin
      GetMem(Attachments, SizeOf(TMapiFileDesc) * MapiMessage.nFileCount);
      FillChar(Attachments^, SizeOf(TMapiFileDesc) * MapiMessage.nFileCount, 0);
      for idx:=0 to MapiMessage.nFileCount - 1 do
      begin
        FileName:=FAttachments[idx];
        Attachments^[idx].ulReserved:=0;
        Attachments^[idx].flFlags:=0;
        Attachments^[idx].nPosition:=ULONG($FFFFFFFF);
        Attachments^[idx].lpszPathName:=StrNew(PAnsiChar(ansistring(FileName)));
        Attachments^[idx].lpszFileName:=StrNew(PAnsiChar(ansistring(ExtractFileName(FileName))));
        Attachments^[idx].lpFileType:=nil;
      end;
      MapiMessage.lpFiles:=PMapiFileDesc(Attachments);
    end;

    // run mail
    WCursor;
    WndList:=DisableTaskWindows(0);
    try
      if FSendAutomatically then
        MapiFlag:=0
      else
        MapiFlag:=MAPI_DIALOG;
      dwRet:=MapiSendMail(0, Application.Handle, MapiMessage, MapiFlag, 0);
      Result:=dwRet = SUCCESS_SUCCESS;
      if (dwRet > MAPI_USER_ABORT) and FShowErrorDlg then
      begin
        DlgE(sMailErr + Format(' (%d)', [dwRet]));
      end;
    finally
      EnableTaskWindows(WndList);
      NCursor;
    end;

  finally
    // free memory
    if MapiMessage.nRecipCount > 0 then
      for idx:=0 to MapiMessage.nRecipCount - 1 do
      begin
        StrDispose(Recip^[idx].lpszName);
        StrDispose(Recip^[idx].lpszAddress);
      end;
    if Assigned(MapiMessage.lpszSubject) then
      StrDispose(MapiMessage.lpszSubject);
    if Assigned(MapiMessage.lpszNoteText) then
      StrDispose(MapiMessage.lpszNoteText);
    if MapiMessage.nFileCount > 0 then
      for idx:=0 to MapiMessage.nFileCount - 1 do
      begin
        StrDispose(Attachments^[idx].lpszPathName);
        StrDispose(Attachments^[idx].lpszFileName);
      end;
    if Assigned(Recip) then
      FreeMem(Recip, MapiMessage.nRecipCount * SizeOf(TMapiRecipDesc));
    if Assigned(Attachments) then
      FreeMem(Attachments, SizeOf(TMapiFileDesc) * MapiMessage.nFileCount);
  end;
end;

function TrMapiEmail.ProcMapiMailW: boolean;
type
  TMapiRecipDescArray = array [0 .. 0] of TMapiRecipDescW;
  PMapiRecipDescArray = ^TMapiRecipDescArray;
  TMapiFileDescArray = array [0 .. 0] of TMapiFileDescW;
  PMapiFileDescArray = ^TMapiFileDescArray;
var
  dwRet: Cardinal;
  // MAPI_Session: Cardinal;
  MapiMessageW: TMapiMessageW;
  Recip: PMapiRecipDescArray;
  Attachments: PMapiFileDescArray;
  WndList: Pointer;
  idx: integer;
  FileName: string;
  MapiFlag: integer;

  procedure AddRecip(Adr: TStrings; RecipClass: Cardinal);
  var
    i: integer;
    n, a: string;
  begin
    for i:=0 to Adr.Count - 1 do
      if GetNameAddress(Adr.Strings[i], n, a) then
      begin
        Recip^[idx].ulRecipClass:=RecipClass;
        if n = '' then
          Recip^[idx].lpszName:=nil
        else
          Recip^[idx].lpszName:=StrNew(PWideChar(n));
        if a = '' then
          Recip^[idx].lpszAddress:=nil
        else
          Recip^[idx].lpszAddress:=StrNew(PWideChar(FAddressPrefix + a));
        Inc(idx);
      end;
  end;

begin
  { dwRet:=MapiLogon(Application.Handle, PAnsiChar(''), PAnsiChar(''), MAPI_LOGON_UI or MAPI_NEW_SESSION, 0, @MAPI_Session);
    if (dwRet <> SUCCESS_SUCCESS) then
    begin
    MessageBox(Application.Handle, PChar('Error while trying to send email'), PChar('Error'), MB_ICONERROR or MB_OK);
    Exit;
    end;
    }
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  CleanAll;
  FillChar(MapiMessageW, SizeOf(MapiMessageW), #0);
  Recip:=nil;
  Attachments:=nil;
  try
    // addresses
    MapiMessageW.nRecipCount:=FSendTo.Count + FCopyTo.Count + FBlindCopyTo.Count;
    if MapiMessageW.nRecipCount > 0 then
    begin
      GetMem(Recip, MapiMessageW.nRecipCount * SizeOf(TMapiRecipDescW));
      FillChar(Recip^, MapiMessageW.nRecipCount * SizeOf(TMapiRecipDescW), 0);
      idx:=0;
      AddRecip(FSendTo, MAPI_TO);
      AddRecip(FCopyTo, MAPI_CC);
      AddRecip(FBlindCopyTo, MAPI_BCC);
      MapiMessageW.lpRecips:=PMapiRecipDescW(Recip);
    end;

    // subject and body
    MapiMessageW.lpszSubject:=StrNew(PWideChar((FSubject)));
    MapiMessageW.lpszNoteText:=StrNew(PWideChar((FBody.Text)));

    // attachments
    MapiMessageW.nFileCount:=FAttachments.Count;
    if MapiMessageW.nFileCount > 0 then
    begin
      GetMem(Attachments, SizeOf(TMapiFileDescW) * MapiMessageW.nFileCount);
      FillChar(Attachments^, SizeOf(TMapiFileDescW) * MapiMessageW.nFileCount, 0);
      for idx:=0 to MapiMessageW.nFileCount - 1 do
      begin
        FileName:=FAttachments[idx];
        Attachments^[idx].ulReserved:=0;
        Attachments^[idx].flFlags:=0;
        Attachments^[idx].nPosition:=ULONG($FFFFFFFF);
        Attachments^[idx].lpszPathName:=StrNew(PWideChar((FileName)));
        Attachments^[idx].lpszFileName:=StrNew(PWideChar((ExtractFileName(FileName))));
        Attachments^[idx].lpFileType:=nil;
      end;
      MapiMessageW.lpFiles:=PMapiFileDescW(Attachments);
    end;

    // run mail
    WCursor;
    WndList:=DisableTaskWindows(0);
    try
      if FSendAutomatically then
        MapiFlag:=0
      else
        MapiFlag:=MAPI_DIALOG;
      dwRet:=MapiSendMailW(0, Application.Handle, MapiMessageW, MapiFlag, 0);
      Result:=dwRet = SUCCESS_SUCCESS;
      if (dwRet > MAPI_USER_ABORT) and FShowErrorDlg then
      begin
        DlgE(sMailErr + Format(' (%d)', [dwRet]));
      end;
    finally
      EnableTaskWindows(WndList);
      NCursor;
    end;

  finally
    // free memory
    for idx:=0 to MapiMessageW.nRecipCount - 1 do
    begin
      StrDispose(Recip^[idx].lpszName);
      StrDispose(Recip^[idx].lpszAddress);
    end;
    if Assigned(MapiMessageW.lpszSubject) then
      StrDispose(MapiMessageW.lpszSubject);
    if Assigned(MapiMessageW.lpszNoteText) then
      StrDispose(MapiMessageW.lpszNoteText);
    for idx:=0 to MapiMessageW.nFileCount - 1 do
    begin
      StrDispose(Attachments^[idx].lpszPathName);
      StrDispose(Attachments^[idx].lpszFileName);
    end;
    if Assigned(Recip) then
      FreeMem(Recip, MapiMessageW.nRecipCount * SizeOf(TMapiRecipDescW));
    if Assigned(Attachments) then
      FreeMem(Attachments, SizeOf(TMapiFileDescW) * MapiMessageW.nFileCount);
  end;
end;

function TrMapiEmail.ProcOutlookOLE;
const
  olMailItem = 0;
var
  Outlook: OLEVariant;
  MailItem: Variant;
  // MailInspector: Variant;
  AppOk: boolean;
  i, HtmlAtt: integer;
  L: TStrings;

  function GetAddrList(L: TStrings): string;
  var
    i: integer;
  begin
    Result:='';
    for i:=0 to L.Count - 1 do
      Result:=Result + Trim(L[i]) + ';';
  end;

begin
  Result:=false;
  AppOk:=false;
  CleanAll;
  WCursor;
  try
    if ooUseRunningApplication in FOutlookOptions then
      try
        Outlook:=GetActiveOleObject('Outlook.Application');
        AppOk:=true;
      except
      end;
    if not AppOk then
      try
        Outlook:=CreateOleObject('Outlook.Application');
      except
        on E: Exception do
        begin
          if ShowErrorDlg then
            ShowErr(sMailErr, E);
          Exit;
        end;
      end;
    try
      try
        MailItem:=Outlook.CreateItem(olMailItem);
        MailItem.Subject:=Subject;
        MailItem. To :=GetAddrList(SendTo);
        MailItem.CC:=GetAddrList(CopyTo);
        MailItem.BCC:=GetAddrList(BlindCopyTo);
        HtmlAtt:=-1;
        // recognize HTML body
        if ContainsText(Body.Text, '</html>') then
          MailItem.HTMLBody:=Body.Text
        else
          MailItem.Body:=Body.Text;
        if (Body.Text='') then
        begin
          // find HTML attachment
          HtmlAtt:=GetFirstHTMLAttachmentIndex;
          if HtmlAtt>=0 then
          begin
            // used HTML attachment as HTML body if body is empty
            L:=TStringList.Create;
            try
              L.LoadFromFile(Attachments[HtmlAtt]);
              MailItem.HTMLBody:=L.Text;
            finally
              L.Free;
            end;
          end;
        end;

        for i:=0 to Attachments.Count - 1 do
          if i<>HtmlAtt then
            MailItem.Attachments.Add(Attachments[i]);
        // MailInspector:=MailItem.GetInspector;
        // MailInspector.display(false); // true means modal
        if FSendAutomatically then
          MailItem.Send
        else
          MailItem.Display(ooShowEmailModal in FOutlookOptions);
        Result:=true;
      except
        on E: Exception do
          if ShowErrorDlg then
            ShowErr(sMailErr, E);
      end;
    finally
      Outlook:=Unassigned;
    end;
  finally
    NCursor;
  end;
end;

function TrMapiEmail.GetMailAppDir: string;
begin
  if FMailAppDirectory = '' then
    Result:=GetAppDir
  else
    Result:=FMailAppDirectory;
end;

procedure TrMapiEmail.InitMail;
begin
  FSendTo.Clear;
  FCopyTo.Clear;
  FBlindCopyTo.Clear;
  FSubject:='';
  FBody.Clear;
  FAttachments.Clear;
  FMailIniCfg.Clear;
  FUseOutlook:=false;
end;

procedure TrMapiEmail.AddMailList(UseOutlook: boolean);
var
  i: integer;
begin
  CleanAll;
  FMailIniCfg.Clear;
  for i:=0 to FSendTo.Count - 1 do
    FMailIniCfg.Add(smSendTo + '=' + FSendTo[i]);
  for i:=0 to FCopyTo.Count - 1 do
    FMailIniCfg.Add(smCopyTo + '=' + FCopyTo[i]);
  for i:=0 to FBlindCopyTo.Count - 1 do
    FMailIniCfg.Add(smBlindCopyTo + '=' + FBlindCopyTo[i]);
  FMailIniCfg.Add(smSubject + '=' + FSubject);
  for i:=0 to FBody.Count - 1 do
    FMailIniCfg.Add(smBody + '=' + FBody[i]);
  for i:=0 to FAttachments.Count - 1 do
    FMailIniCfg.Add(smAttachedFile + '=' + FAttachments[i]);
  if FSendAutomatically then
    FMailIniCfg.Add(smShowDialog + '=0');
  if UseOutlook then
    FMailIniCfg.Add(smUseOutlook + '=1');
end;

function TrMapiEmail.RunExternalMailTool(L: TStrings): boolean;
begin
  Result:=false;
  try
    L.SaveToFile(GetMailAppDir + smIniFile);
    Result:=RunFileSimple(GetMailAppDir + smRunFile) > 32;
  except
    on E: Exception do
    begin
      if FShowErrorDlg then
        ShowErr(sMailErr, E, false);
    end;
  end;
end;

function TrMapiEmail.ProcBatchMail: boolean;
var
  i, mProc, mCount: integer;
  MailToSend, Aborted: boolean;

  function DelQM(s: string): string;
  begin
    Result:=rtool.ClearStr('?', s);
  end;

  function Send(Batch: boolean): boolean;
  var
    procinfo: string;
  begin
    if FUseOutlook then
      Result:=self.SendOutlookMail(rmInternal)
    else
      Result:=SendMapiMail(rmInternal);
    FSendAutomatically:=false;
    procinfo:=Format(' (%d z %d)', [mProc + 1, mCount]);
    if Batch and not Result and (mCount > 1) then
    begin
      case rdlg.DlgQUserDef(DelQM(sTryAgainQ), DelQM(sContinueQ), DelQM(sTerminateQ), [sMailErr + procinfo]) of
        mrYes:
          begin
            Result:=Send(Batch);
            Exit;
          end;
        mrCancel:
          Aborted:=true;
      end;
    end;
    Inc(mProc);
    InitMail;
    MailToSend:=false;
    if Result then
    begin
      if (FBatchConfirmCount > 0) and (mProc mod FBatchConfirmCount = 0) then
        Aborted:=DlgQ2(sMailGenInfo + procinfo, sContinueQ) <> mrYes;
    end;
  end;

  procedure ProcLine(s: string);
  var
    n, v: string;
  begin
    if Trim(s) = smNextMail then
      Send(true)
    else
    begin
      n:=Trim(copy(s, 1, pos('=', s) - 1));
      v:=(copy(s, pos('=', s) + 1, Length(s)));
      if v <> '' then
        MailToSend:=true;
      if n = smSendTo then
        FSendTo.Add(v);
      if n = smCopyTo then
        FCopyTo.Add(v);
      if n = smBlindCopyTo then
        FBlindCopyTo.Add(v);
      if n = smSubject then
      begin
        if FSubject <> '' then
          FSubject:=FSubject + ' ' + v
        else
          FSubject:=v;
      end;
      if n = smBody then
        FBody.Add(v);
      if n = smAttachedFile then
        FAttachments.Add(v);
      if n = smShowDialog then
        FSendAutomatically:=v = '0';
      if n = smUseOutlook then
        FUseOutlook:=v = '1';
      if n = smConfirmCounter then
        FBatchConfirmCount:=StrToIntDef(v, 10);
    end;
  end;

begin
  InitMail;
  MailToSend:=false;
  Aborted:=false;
  mProc:=0;
  mCount:=0;
  // count # of emails
  for i:=0 to FBatchMailList.Count - 1 do
    if Trim(FBatchMailList[i]) = smNextMail then
      Inc(mCount);

  FSendAutomatically:=false;
  for i:=0 to FBatchMailList.Count - 1 do
  begin
    ProcLine(FBatchMailList[i]);
    if Aborted then
      Break;
  end;
  if not Aborted and MailToSend then
    Send(false);
  Result:=Aborted;
end;

function TrMapiEmail.SendSimpleMail: boolean;
var
  i: integer;
  Adr, bd, s: string;
begin
  if FSendTo.Count > 0 then
    Adr:=FSendTo[0]
  else
    Adr:='nobody';
  bd:='';
  for i:=0 to FBody.Count - 1 do
    bd:=bd + FBody[i] + '%0A';
  s:=Format('mailto:%s?subject=%s&body=%s', [Adr, FSubject, bd]);
  s:=ReplaceStr(' ', '%20', s, true, false);
  WCursor;
  try
    Result:=RunFileSimple(s) > 32;
  finally
    NCursor;
  end;
end;

function TrMapiEmail.SendMapiMail(RunMailType: TRunMailType = rmExternalAppPrefered; RunMailVersion: TRunMailMapiVersion = rmAuto): boolean;
begin
  if (RunMailType = rmExternalAppOnly) or ((RunMailType = rmExternalAppPrefered) and ExternalAppExists) then
  begin
    AddMailList(false);
    Result:=RunExternalMailTool(FMailIniCfg);
    FMailIniCfg.Clear;
  end
  else
  begin
    case RunMailVersion of
      rmAnsi:
        Result:=ProcMapiMail;
      rmUnicode:
        Result:=ProcMapiMailW;
    else
      begin
        if CheckMapiUnicode then // CheckWin32Version(6, 2) then
          Result:=ProcMapiMailW
        else
          Result:=ProcMapiMail;
      end;
    end;
  end;
end;

function TrMapiEmail.SendOutlookMail(RunMailType: TRunMailType = rmExternalAppPrefered): boolean;
begin
  if (RunMailType = rmExternalAppOnly) or ((RunMailType = rmExternalAppPrefered) and ExternalAppExists) then
  begin
    AddMailList(true);
    Result:=RunExternalMailTool(FMailIniCfg);
    FMailIniCfg.Clear;
  end
  else
    Result:=ProcOutlookOLE;
end;

function TrMapiEmail.SendAutoSelect(MapiOrOutlook: TRunMailMapiOutlook = rmOutlookPreferred): boolean;
begin
  if (FSendTo.Count = 1) and (FCopyTo.Count = 0) and (FBlindCopyTo.Count = 0) and (FAttachments.Count = 0) and (Length(FBody.Text) < 250)
    and not FSendAutomatically then
    Result:=SendSimpleMail
  else
  begin
    case MapiOrOutlook of
      rmMapiOnly:
        Result:=SendMapiMail;
      rmOutlookOnly:
        Result:=SendOutlookMail;
      rmMapiPreferred:
      begin
        Result:=SendMapiMail;
        if not Result then
          Result:=SendOutlookMail;
      end;
      rmOutlookPreferred:
      begin
        Result:=SendOutlookMail;
        if not Result then
          Result:=SendMapiMail;
      end;
      else
        Result:=false;
    end;
  end;
end;

procedure TrMapiEmail.InitBatchMail;
begin
  FBatchMailList.Clear;
end;

procedure TrMapiEmail.AddToBatchMail(UseOutlook: boolean = false);
begin
  AddMailList(UseOutlook);
  FBatchMailList.AddStrings(FMailIniCfg);
  FBatchMailList.Add(smNextMail);
end;

function TrMapiEmail.SendBatchMail(RunMailType: TRunMailType = rmExternalAppPrefered): boolean;
begin
  if (RunMailType = rmExternalAppOnly) or ((RunMailType = rmExternalAppPrefered) and ExternalAppExists) then
    Result:=RunExternalMailTool(FBatchMailList)
  else
    Result:=ProcBatchMail;
  FBatchMailList.Clear;
end;

end.
