unit rStringGridEd;

{<
  rStringGridEd  unit contains TrStringGridEd - enhanced StringGrid component which provides
  similar function as DBGrid - introduce Columns property for setting
  and storing of properties for each column.

  List of new functions:
  - introduce Columns property for setting and storing of properties for each column
  (Columns items are created automatically according ColCount value)
  - provides new types of editors = DropDown, DropDownList, DropDownColor,
  DropDownCheckList, Checkbox, DatePicker, Ellipsis, Button
  - can show any column as boolean checkbox with defineable TRUE FALSE values
  - support multiline title, cell text and editors
  - each column has definable Datatype, NumFormat, MinMax values
  - shows complete text as hint for truncated text in cell and for memo field
  - introduces event for getting properties for each cell individually
  (datatype, editorstyle, canedit, numformat)
  - supports automatic addition of new rows
  - supports auto height for rows and width for columns
  - supports range selection in Edit mode with row/column selection by fixed cell click
  - supports drag and drop
  - supports Copy/Paste function in MSExcel format
  - supports data sorting
  - introduces event for click on fixed cells
  - introduces event for getting of PickList items
  - introduces event for setting Font and Color of any cell
  - introduces event for custom hint on any cell
  - provide 3D effect and gradient fill for fixed cells


  Hot keys for editors:
  F2 - show / hide editor;
  F4 - show/hide dropdownlist or click to ellipsis button
  Alt+Down - show dropdownlist or click to ellipsis button;
  Space - change checkbox;
  Enter - select item from Listbox

  Updated version compatible with Delphi2010+ and 64bit platform.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Apr 2009)
  @lastmod(Feb 2021)
}

interface

{$INCLUDE _rosi.inc}
// FPopupCalendar is used instead of TDateTimePicker if compiler switch TrPopupMonthCalendar is defined
{$DEFINE TrPopupMonthCalendar}

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Menus,
  Dialogs, Grids, StdCtrls, ComCtrls, Clipbrd, Math, Types, DateUtils,
{$IFDEF DX101P} Rtti, {$ENDIF}
{$IFDEF DXE2P} UITypes, {$ENDIF} Themes, rImprovedComps, RosiCompGlobalCfg;

const
  FieldListSep = ';';

type
  TrStringGridEd = class;
  TrStringGridEdColumn = class;
  TrStringGridEdColumnClass = class of TrStringGridEdColumn;
  TrStringGridEdInplaceEdit = class;
  { list of editor types used by TrStringGridEd }
  TrStringGridEdEditStyle = (sgbNone, sgbEllipsis, sgbEllipsisOnly, sgbDropDown, sgbDropDownList, sgbDropDownColor, sgbCheckbox, sgbDateTimePicker,
    sgbDropDownCheckList, sgbDropDownCheckListOnly, sgbDropDownDataImage, sgbButton, sgbRichText, sgbCustomEditor);
  { list of data types used by TrStringGridEd }
  TrStringGridEdDataType = (sgdText, sgdNumber, sgdInteger, sgdDate, sgdTime, sgdDateTime, sgdDataBar, sgdStatusBar);
  { list of data statuses used by TrStringGridEd }
  TrStringGridEdDataState = (sdsOk, sdsEmpty, sdsInvalid, sdsOutOfLimit);
  { list of invalid data actions used by TrStringGridEd }
  TrStringGridEdInvalidDataAction = (sgaNone, sgaHighlight, sgaBeep);
  { list of enhanced Option properties used by TrStringGridEd }
  TOptionEx = (dgDragCell, dgSelectRange, dgAlwaysShowSelection, dgEditArrowKeyMoveCell, dgShowTitleBold, dgColorChangedCells, dgSortAfterTitleClick,
    dgSortFixedCol, dgSortShowProgress, dgResizeLastColumn, dgNextRowAfterReturn, dgCopyPasteHiddenColumns, dgSkipCellHighlight);
  { set of enhanced Option properties used by TrStringGridEd }
  TOptionsEx = set of TOptionEx;

  { data image record used by TrStringGridEd and TrDBGrid }
  TDataImageDef = record
    UseDataImage: boolean;
    ImageIndex: integer;
    ImageAlignment: string;
    ShowText: string;
    SkipTextDrawing: boolean;
  end;

  { DataBar class used by TrStringGridEdColumn }
  TrDataBar = class(TPersistent)
  private
    FMinValue: double;
    FMaxValue: double;
    FColorFrom: TColor;
    FColorTo: TColor;
    FText: TrDataBarText;
    function IsMaxValueStored: boolean;
  public
    constructor Create(Owner: TrStringGridEdColumn);
  published
    { starting value for databar (0%) }
    property MinValue: double read FMinValue write FMinValue;
    { ending value for databar (100%) }
    property MaxValue: double read FMaxValue write FMaxValue stored IsMaxValueStored;
    { gradient start color }
    property ColorFrom: TColor read FColorFrom write FColorFrom default clNone;
    { gradient end color }
    property ColorTo: TColor read FColorTo write FColorTo default clNone;
    { databar text type (value, percent or nothing) }
    property Text: TrDataBarText read FText write FText default dbtValue;
  end;

  { StatusBar class used by TrStringGridEdColumn }
  TrStatusBar = class(TPersistent)
  private
    FSteps: integer;
    FColorAll: TColor;
    FColorDone: TColor;
    FIconType: TrStatusBarType;
  public
    constructor Create(Owner: TrStringGridEdColumn);
  published
    { number of steps for status bar }
    property Steps: integer read FSteps write FSteps default 5;
    { color for all steps }
    property ColorAll: TColor read FColorAll write FColorAll default clNone;
    { color for done steps }
    property ColorDone: TColor read FColorDone write FColorDone default clNone;
    { status bar type (icon type) }
    property IconType: TrStatusBarType read FIconType write FIconType default sbtDots;
  end;

  { collection item containing all column properties }
  TrStringGridEdColumn = class(TCollectionItem)
  private
    FReadOnly: boolean;
    FVisible: boolean;
    FCustomDraw: boolean;
    FWidth: integer;
    FButtonStyle: TrStringGridEdEditStyle;
    FDataType: TrStringGridEdDataType;
    FDataImageList: TImageList;
    FNumFormat: string;
    FInvalidDataAction: TrStringGridEdInvalidDataAction;
    FInvalidDataColor: TColor;
    FOutOfLimitDataColor: TColor;
    FMinValue: double;
    FMaxValue: double;
    FAlignment: TAlignment;
    FCheckBoxAlignment: TAlignment;
    FAutoAlignment: boolean;
    FColor: TColor;
    FFont: TFont;
    FParentFont: boolean;
    FPickList: TStringList;
    FDropDownRows: integer;
    FDropDownListCanDelete: boolean;
    FTitleCaption: string;
    FTitleAlignment: TAlignment;
    FTitleFont: TFont;
    FTitleParentFont: boolean;
    FTitleColor: TColor;
    FDataBar: TrDataBar;
    FStatusBar: TrStatusBar;
    FCheckBoxTrueValue: string;
    FCheckBoxFalseValue: string;
    FCheckBoxDblClick: boolean;
    FCheckListSeparator: string;
    FCheckListCaseSensitive: boolean;
    FWordWrap: boolean;
    FTitleWordWrap: boolean;
    FFieldName: string;
    FColumnGroup: integer;
    FDropDownDrawItemEvent: TDrawItemEvent;
    FDropDownMeasureItemEvent: TMeasureItemEvent;
    FDropDownGetSize: TCanResizeEvent;
    procedure SetReadOnly(Value: boolean);
    procedure SetVisible(Value: boolean);
    procedure SetDataImageList(Value: TImageList);
    procedure SetButtonStyle(Value: TrStringGridEdEditStyle);
    procedure SetAlignment(Value: TAlignment);
    procedure SetCheckBoxAlignment(Value: TAlignment);
    procedure SetColor(Value: TColor);
    procedure SetFont(Value: TFont);
    procedure SetPickList(Value: TStringList);
    procedure SetTitleCaption(const Value: string);
    procedure SetTitleAlignment(Value: TAlignment);
    procedure SetTitleFont(Value: TFont);
    procedure SetParentFont(Value: boolean);
    procedure SetTitleParentFont(Value: boolean);
    procedure SetTitleColor(Value: TColor);
    procedure SetWidth(Value: integer);
    function GetWidth: integer;
    function IsFontStored: boolean;
    function IsTitleFontStored: boolean;
    procedure FontChanged(Sender: TObject);
    procedure TitleFontChanged(Sender: TObject);
    procedure SetCustomDraw(Value: boolean);
    procedure SetWordWrap(Value: boolean);
    procedure SetTitleWordWrap(Value: boolean);
    function CheckBoxTrueValueStored: boolean;
    function CheckBoxFalseValueStored: boolean;
    function CheckListSeparatorStored: boolean;
  protected
    function GetGrid: TrStringGridEd;
  public
    constructor Create(Collection: TCollection); override; {<@exclude }
    destructor Destroy; override; {<@exclude }
    procedure Assign(Source: TPersistent); override;
    property Grid: TrStringGridEd read GetGrid; {<@exclude }
    { restore default properties column }
    procedure RestoreDefaults;
    { check if checklist text contain value }
    function CheckListContainValue(Text, Value: string): boolean;
    { add value to checklist text }
    function CheckListAddValue(Text, Value: string): string;
    { remove value from checklist text }
    function CheckListRemoveValue(Text, Value: string): string;
    { clean up checklist text  - remove values outside of picklist and sort by picklist }
    function CheckListCleanUpText(Text: string; SortAsPickList: boolean = false): string;
    { clean up checklist text from the row and update cell text }
    function CheckListCleanUpRow(ARow: longint; SortAsPickList: boolean = false): string;
  published
    { define edit style of column }
    property ButtonStyle: TrStringGridEdEditStyle read FButtonStyle write SetButtonStyle default sgbNone;
    { define type of data in column  - text, int, float, datetime }
    property DataType: TrStringGridEdDataType read FDataType write FDataType default sgdText;
    { link to ImageList for images in data cells }
    property DataImageList: TImageList read FDataImageList write SetDataImageList;
    { define numeric format for data in column }
    property NumFormat: string read FNumFormat write FNumFormat;
    { define min limit for numeric data, DataType must be int or number }
    property MinValue: double read FMinValue write FMinValue;
    { define max limit for numeric data, DataType must be int or number }
    property MaxValue: double read FMaxValue write FMaxValue;
    { define action if invalid data is entered }
    property InvalidDataAction: TrStringGridEdInvalidDataAction read FInvalidDataAction write FInvalidDataAction default sgaHighlight;
    { define font color for invalid data }
    property InvalidDataColor: TColor read FInvalidDataColor write FInvalidDataColor default clRed;
    { define font color for data out of min max limit }
    property OutOfLimitDataColor: TColor read FOutOfLimitDataColor write FOutOfLimitDataColor default clBlack;
    { define number of items in dropdown listbox }
    property DropDownRows: integer read FDropDownRows write FDropDownRows default 12;
    { define whether cell value can be deleted for DropDownList + DropDownColor style }
    property DropDownListCanDelete: boolean read FDropDownListCanDelete write FDropDownListCanDelete default false;
    { define saved value if checkbox is checked }
    property CheckBoxTrueValue: string read FCheckBoxTrueValue write FCheckBoxTrueValue stored CheckBoxTrueValueStored;
    { define saved value if checkbox is unchecked }
    property CheckBoxFalseValue: string read FCheckBoxFalseValue write FCheckBoxFalseValue stored CheckBoxFalseValueStored;
    { define whether double click must be used for changing of checkbox }
    property CheckBoxDblClick: boolean read FCheckBoxDblClick write FCheckBoxDblClick default false;
    { define separators between values from checklist }
    property CheckListSeparator: string read FCheckListSeparator write FCheckListSeparator stored CheckListSeparatorStored;
    { define whether values in check list is case sensitive }
    property CheckListCaseSensitive: boolean read FCheckListCaseSensitive write FCheckListCaseSensitive default false;
    { define alignment of data in column }
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    { define alignment of data in column }
    property CheckboxAlignment: TAlignment read FCheckBoxAlignment write SetCheckBoxAlignment default taCenter;
    { define alignment according datatype of each cell }
    property AutoAlignment: boolean read FAutoAlignment write FAutoAlignment default false;
    { define background color for column }
    property Color: TColor read FColor write SetColor default clWindow;
    { define font for all column except title }
    property Font: TFont read FFont write SetFont stored IsFontStored;
    { define whether column is drawn only in OnDrawCell }
    property CustomDraw: boolean read FCustomDraw write SetCustomDraw default false;
    { define usage of parent font for column }
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    { define list of values for dropdown listbox }
    property PickList: TStringList read FPickList write SetPickList;
    { define that column cannot be modified }
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
    { define that column is visible or hidden (width=-1) }
    property Visible: boolean read FVisible write SetVisible default true;
    { define caption of column }
    property TitleCaption: string read FTitleCaption write SetTitleCaption;
    { define title font of column }
    property TitleFont: TFont read FTitleFont write SetTitleFont stored IsTitleFontStored;
    { define usage of parent font for column title }
    property TitleParentFont: boolean read FTitleParentFont write SetTitleParentFont default true;
    { define background color for column title }
    property TitleColor: TColor read FTitleColor write SetTitleColor default clNone;
    { define allignment of column title }
    property TitleAlignment: TAlignment read FTitleAlignment write SetTitleAlignment default taLeftJustify;
    { define whether long string should be wraped in column }
    property WordWrap: boolean read FWordWrap write SetWordWrap default false;
    { define whether long string in title row should be wraped in column }
    property TitleWordWrap: boolean read FTitleWordWrap write SetTitleWordWrap default false;
    { define progress bar properties }
    property DataBar: TrDataBar read FDataBar write FDataBar;
    { define status bar properties }
    property StatusBar: TrStatusBar read FStatusBar write FStatusBar;
    { define Width of column }
    property Width: integer read GetWidth write SetWidth default 64;
    { define FieldName for column in case of Database support, used in rStringGridEdDB }
    property FieldName: string read FFieldName write FFieldName;
    { define any number for Column Grouping, used mainly for DB support but can be used for any other purpose }
    property ColumnGroup: integer read FColumnGroup write FColumnGroup default 0;
    { define event for owner draw on dropdown listbox }
    property OnDropDownDrawItemEvent: TDrawItemEvent read FDropDownDrawItemEvent write FDropDownDrawItemEvent;
    { define event for item heights if owner draw on dropdown listbox is used }
    property OnDropDownMeasureItemEvent: TMeasureItemEvent read FDropDownMeasureItemEvent write FDropDownMeasureItemEvent;
    { define event for definning of dropdown listbox size }
    property OnDropDownGetSize: TCanResizeEvent read FDropDownGetSize write FDropDownGetSize;
  end;

  { collection of TrStringGridEdColumn used by TrStringGridEd }
  TrStringGridEdColumns = class(TCollection)
  private
    FGrid: TrStringGridEd;
    function GetColumn(Index: integer): TrStringGridEdColumn;
    procedure SetColumn(Index: integer; Value: TrStringGridEdColumn);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(Grid: TrStringGridEd; ColumnClass: TrStringGridEdColumnClass);
    function Add: TrStringGridEdColumn;
    procedure RestoreDefaults;
    property Grid: TrStringGridEd read FGrid;
    property Items[Index: integer]: TrStringGridEdColumn read GetColumn write SetColumn; default;
  end;

  { @exclude }
  TPopupListbox = class(TCustomListbox)
  private
    FSearchText: String;
    FSearchTickCount: Cardinal;
    IsColor: boolean;
    IsCheckList: boolean;
    IsDataImage: boolean;
    AColumn: TrStringGridEdColumn;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure MeasureItem(Index: integer; var Height: integer); override;
    function UseOwnDraw: boolean;
    procedure GetItemHeight(var H: integer);
  end;

  { enhanced TInplaceEdit used by TrStringGridEd }
  TrStringGridEdInplaceEdit = class(TInplaceEdit)
  private
    FEllipsisBtnWidth: integer;
    FActBtnWidth: integer;
    FEditStyle: TrStringGridEdEditStyle;
    FListVisible: boolean;
    FRichEditorVisible: boolean;
    FCustomEditorVisible: boolean;
    FTracking: boolean;
    FPressed: boolean;
    FMouseInControl: boolean;
    FDataType: TrStringGridEdDataType;
    FNumFormat: string;
    FOrigDateTime: TDateTime;
    function GetDateCalendarVisible: boolean;
    procedure CreateHideCalendar;
    procedure SetEditStyle(Value: TrStringGridEdEditStyle);
    procedure SetDateTime(DateTimeValue: TDateTime);
    procedure PopupCallbackSelectDate(Sender: TObject);
    procedure StopTracking;
    procedure TrackButton(X, Y: integer);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    procedure WMCancelMode(var Message: TMessage); message WM_CancelMode;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
    procedure WMLButtonDblClk(var Message: TWMLButtonDblClk); message wm_LButtonDblClk;
    procedure WMPaint(var Message: TWMPaint); message wm_Paint;
    procedure WMSetCursor(var Message: TWMSetCursor); message WM_SetCursor;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure WMSetFocus(var Message: TWMSetFocus); message WM_SETFOCUS;
  protected
    FPickList: TPopupListbox;
    FPopupCalendar: TrPopupMonthCalendar;
    FRichEditor: TrRichEdit;
    FCustomEditor: TWinControl;
    FCustomEditCoord: TGridCoord;
    FActiveList: TWinControl;
{$IFNDEF TrPopupMonthCalendar}
    FDatePickerVisible: boolean;
    FDateTimePicker: TWinControl;
    function CreateDateTimePicker: TWinControl; virtual;
    procedure SetDateTimePickerDate(ADateTime: TDateTime); virtual;
    function GetDateTimePickerDate: TDateTime; virtual;
    procedure ShowDateTimePicker(Rect: TRect); virtual;
    procedure DateTimePickerClose(Sender: TObject);
    procedure DateTimePickerChange(Sender: TObject);
    procedure RecreateDatePicker;
{$ENDIF}
    procedure ShowRichEditor(Rect: TRect); virtual;
    procedure HideRichEditor(Accept: boolean); virtual;
    procedure RichEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure RichEditPopupMenu(Sender: TObject);
    function CreateCustomEditor: TWinControl; virtual;
    procedure SetCustomEditorValue(AValue: string); virtual;
    function GetCustomEditorValue: string; virtual;
    procedure ShowCustomEditor(Rect: TRect); virtual;
    procedure HideCustomEditor(Accept: boolean); virtual;
    procedure BoundsChanged; override;
    procedure CloseUp(Accept, HideList: boolean);
    procedure DoDropDownKeys(var Key: Word; Shift: TShiftState);
    procedure KeyPress(var Key: Char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure PaintWindow(DC: HDC); override;
    procedure UpdateContents; override;
    procedure WndProc(var Message: TMessage); override;
    property EditStyle: TrStringGridEdEditStyle read FEditStyle write SetEditStyle;
    property ActiveList: TWinControl read FActiveList write FActiveList;
    function EditCanModify: boolean; override;
  public
    constructor Create(Owner: TComponent); override;
    destructor Destroy; override;
    procedure DropDown;
    property Font;
    property PickList: TPopupListbox read FPickList;
    function OwnerGrid: TCustomGrid;
  end;

  { list of possible action for CustomEditor events }
  TCustomEditorAction = (ceaSetValue, ceaGetValue, ceaShow, ceaHide);
  { event type for creating of CustomEditor }
  TCustomEditorCreate = procedure(Sender: TObject; var CustomEditor: TWinControl) of object;
  { event type for CustomEditor action }
  TCustomEditorEvent = procedure(Sender: TObject; CustomEditor: TWinControl; Action: TCustomEditorAction; var Value: string; var Done: boolean) of object;

  { event type for TrStringGridEd cell }
  TrStringGridEdCellEvent = procedure(Sender: TObject; Col, Row: longint) of object;
  { event type for TrStringGridEd Editor event }
  TrStringGridEdEditorEvent = procedure(Sender: TObject; Col, Row: longint; ActionID: string) of object;
  { event type for TrStringGridEd dropdown list }
  TrStringGridEdDropdownEvent = procedure(Sender: TObject; Col, Row: longint; var PickList: TStringList) of object;
  { event type for TrStringGridEd cell parameters }
  TGetCellParamsEvent = procedure(Sender: TObject; Col, Row: longint; AFont: TFont; var Background: TColor; var Highlight: boolean) of object;
  { event type for TrStringGridEd cell extended parameters }
  TGetCellParamsEventEx = procedure(Sender: TObject; Col, Row: longint; AFont: TFont; var Background: TColor; var Highlight: boolean;
    var DataImageDef: TDataImageDef) of object;
  { event type for TrStringGridEd hint }
  TGetHintText = procedure(Sender: TStringGrid; X, Y: integer; var Hint: string) of object;
  { event type for TrStringGridEd data type }
  TGetDataTypeEvent = procedure(Sender: TObject; Col, Row: longint; var DataType: TrStringGridEdDataType) of object;
  { event type for TrStringGridEd style type }
  TGetEditStyleEvent = procedure(Sender: TObject; Col, Row: longint; var EditStyle: TrStringGridEdEditStyle) of object;
  { event type for TrStringGridEd edit }
  TCanEditEvent = procedure(Sender: TObject; Col, Row: longint; var CanEdit: boolean) of object;
  { event type for TrStringGridEd cell numeric format }
  TGetNumFormat = procedure(Sender: TObject; Col, Row: longint; var NumFormat: string) of object;
  { event type for TrStringGridEd ellipsis format }
  TGetEllipsisButtonFormat = procedure(Sender: TObject; Col, Row: longint; var ButtonWidth: integer; var Glyph: TBitmap) of object;
  { event type for TrStringGridEd column restore }
  TOnColumnRestoreDefault = procedure(Column: TrStringGridEdColumn) of object;
  { event type for TrStringGridEd progress event }
  TOnProgress = procedure(Sender: TObject; ActionType: string; Percent: double; var Aborted: boolean) of object;

  { general event type used by TrStringGridEd }
  TGeneralEventType = (geChangeScaleBefore, geChangeScaleAfter, geSetColumnWidth, geGetDataBarParam, geGetStatusBarParam, geCellObjectClick,
    geCellObjectDblClick, gePasteDataBefore, gePasteDataAfter);

  { general event record used as parameter for TGeneralEventType }
  TGeneralEventRec = record
    Col, Row: longint;
    ParamObject: TObject;
    ParamValue: variant;
  end;

  { event is called in many events to add posibility of any custom code and to skip rest of the standard event used by TrStringGridEd }
  TGeneralEvent = procedure(Sender: TObject; EventType: TGeneralEventType; var EventParam: TGeneralEventRec; var ContinueEvent: boolean) of object;

  TCellPosition = (cpInvalid, cpFixedColRow, cpFixedRow, cpFixedCol, cpNotFixed);

  TCustomGridHelper = class helper for TCustomGrid
  public
    procedure InvalidateRectEx(ARect: TGridRect);
  end;

  { enhanced TStringGrid component which introduce Columns property for setting
    and storing of properties for each column }
  TrStringGridEd = class(TStringGrid)
  private
    FCustomGlyphs: TrCustomGlyphs;
    FCheckBoxWidth: integer;
    FirstInvalidateEditor: boolean;
    FInnerMargin: integer;
    FTitleRowHeight: integer;
    FChangedCellsBkColor: TColor;
    FSelectedCellBkColor: TColor;
    FInRangeCellsBkColor: TColor;
    FUseChangeLog: boolean;
    FColWidthUpdating: boolean;
    MMoveCoord: TGridCoord;
    MDownCoord: TGridCoord;
    MDownCellPoint: TPoint;
    FEditCoord: TGridCoord;
    FGetHintText: TGetHintText;
    FAutoRowHeight: boolean;
    FAutoRowMaxHeight: integer;
    FAutoAddNewRow: boolean;
    FBlockClick: boolean;
    FSkipCheckDropDownList: boolean;
    FChangedRows: TStrings;
    FChangedCells: TStrings;
    FIniKeyColumn: integer;
    FIniValueColumn: integer;
    FIniGroupRows: TStrings;
    FOnDrawCellBackground: TDrawCellEvent;
    FOnEllipsisClicked: TrStringGridEdCellEvent;
    FOnPicklistDropdown: TrStringGridEdDropdownEvent;
    FOnEditorEvent: TrStringGridEdEditorEvent;
    FOnGetCellParams: TGetCellParamsEvent;
    FOnGetCellParamsEx: TGetCellParamsEventEx;
    FOnGetDataType: TGetDataTypeEvent;
    FOnGetNumFormat: TGetNumFormat;
    FOnGetEditStyle: TGetEditStyleEvent;
    FOnCanEdit: TCanEditEvent;
    FColumns: TrStringGridEdColumns;
{$IFNDEF TrPopupMonthCalendar}
    FDateTimePickerCreate: TDateTimePickerCreate;
    FDateTimePickerEvent: TDateTimePickerEvent;
{$ENDIF}
    FCustomEditorCreate: TCustomEditorCreate;
    FCustomEditorEvent: TCustomEditorEvent;
    // FOnMouseWheelUp: TMouseWheelUpDownEvent;
    // FOnMouseWheelDown: TMouseWheelUpDownEvent;
    FOnColumnRestoreDefault: TOnColumnRestoreDefault;
    FOnFinishCellEditing: TrStringGridEdCellEvent;
    FOnEditorKeyPress: TKeyPressEvent;
    FOldEditedValue: string;
    FOptionsEx: TOptionsEx;
    FOnCellClick: TFixedCellClickEvent;
    FOnGetEllipsisButtonFormat: TGetEllipsisButtonFormat;
    FOnProgress: TOnProgress;
    FOnGeneralEvent: TGeneralEvent;
    FGeneralEventRec: TGeneralEventRec;
    FRichTextTopMargin: integer;
    procedure SetCustomGlyphs(const Value: TrCustomGlyphs);
    procedure SetTitleRowHeight(const Value: integer);
    procedure SetAutoRowHeight(Value: boolean);
    function GetColCount: longint;
    procedure SetColCount(Col: longint);
    function CheckboxIsChecked(ACol, ARow: longint): boolean;
    procedure ChangeCheckbox;
    procedure EllipsisClick;
    procedure ButtonClick;
    function GetCoordPosition(X, Y: integer): TCellPosition; overload;
    function GetCoordPosition(Coord: TGridCoord): TCellPosition; overload;
    function GetStringCells(ACol, ARow: longint): string;
    procedure SetStringCells(ACol, ARow: longint; const Value: string);
    function GetNumCells(ACol, ARow: longint): double;
    procedure SetNumCells(ACol, ARow: longint; const Value: double);
    function GetIntCells(ACol, ARow: longint): integer;
    procedure SetIntCells(ACol, ARow: longint; const Value: integer);
    function GetBoolCells(ACol, ARow: longint): boolean;
    procedure SetBoolCells(ACol, ARow: longint; const Value: boolean);
    function GetDateTimeCells(ACol, ARow: longint): TDateTime;
    procedure SetDateTimeCells(ACol, ARow: longint; const Value: TDateTime);
    function GetIniValue(const IniKey: string): string;
    procedure SetIniValue(const IniKey, Value: string);
    procedure SetInnerMargin(Value: integer);
    procedure SetOptionsEx(const Value: TOptionsEx);
    procedure SetAutoAddNewRow(const Value: boolean);
    procedure SetChangedCellsBkColor(const Value: TColor);
    procedure SetSelectedCellBkColor(const Value: TColor);
    procedure SetInRangeCellsBkColor(const Value: TColor);
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure DoSelRangeProc(SelRange: TGridRect; Act: integer);
    procedure PasteSelRangeInt(fromCol, fromRow: longint; CanExpandGrid, SkipAllChecks, SelectPasted: boolean; const PasteData: string);
    function GetSelectedRange: TGridRect;
    procedure SetSelectedRange(Value: TGridRect);
    function GetRangeCellCount(Range: TGridRect): integer;
    function IsRangeSelected: boolean;
    procedure InvalidateSelectedRange;
    procedure SetRichTextTopMargin(Value: integer);
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure WMHScroll(var Msg: TWMVScroll); message WM_HSCROLL;
    procedure WMVScroll(var Msg: TWMVScroll); message WM_VSCROLL;
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
  protected
    SelRangeStart, SelRangeStop: TGridCoord;
    SelRangeText: string;
    LastSortColIndex: integer;
    LastSortDescend: boolean;
    FScaleRatio: double;
    FAborted: boolean;
    FCellChanged: integer;
    FUpdateLock: integer;
    FStartEditKey: Word;
    FRichText: TrRichEdit;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    function CreateEditor: TInplaceEdit; override;
    procedure HideEditorEx;
    procedure DoFinishEditing(ACol, ARow: longint; Forced: boolean); virtual;
    procedure ColWidthsChanged; override;
    procedure RowHeightsChanged; override;
    function CanEditModify: boolean; override;
    function CanEditModifyEx(AllowDropDownList: boolean): boolean;
    function CanEditShow: boolean; override;
    function CanEditCell(ACol, ARow: longint): boolean;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean; override;
    procedure KeyPress(var Key: Char); override;
    function SelectCell(ACol, ARow: longint): boolean; override;
    procedure ColumnMoved(FromIndex, ToIndex: longint); override;
    procedure Paint; override;
    procedure DrawCell(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState); override;
    procedure DrawCellEx(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState; BkColor: TColor; DrawFocus: boolean); virtual;
    procedure DrawCellBackground(const ARect: TRect; AColor: TColor; AState: TGridDrawState; ACol, ARow: integer); override;
    procedure DrawRichText(ARect: TRect; Text: string; BkColor: TColor); virtual;
    procedure SetCanvasStyle(ACol, ARow: longint; ACanvas: TCanvas);
    function GetTextFormatSetting(WordWrap: boolean): integer;
    function GetCellWordWrap(ACol, ARow: longint): boolean; virtual;
    procedure SetCellStyle(ACol, ARow: longint; ACanvas: TCanvas; var AState: TGridDrawState; var Highlighted: boolean; var DataImageDef: TDataImageDef);
    procedure GetCellParams(ACol, ARow: longint; AFont: TFont; var Background: TColor; var Highlight: boolean; var DataImageDef: TDataImageDef); virtual;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure SetEditText(ACol, ARow: longint; const Value: string); override;
    function GetEditText(ACol, ARow: longint): string; override;
    procedure DblClick; override;
    procedure EditorKeyPress(var Key: Char); virtual;
    function GetCellText(ACol, ARow: longint): string;
    procedure SetDefaultRowHeight; virtual;
    function CalcMaxRowHeight(ARow: longint; UseLimits: boolean): integer; virtual;
    function CalcCellHeight(ACol, ARow: longint; UseLimits: boolean): integer; virtual;
    procedure FixedCellClick(ACol, ARow: longint); override;
    procedure CellClick(ACol, ARow: longint); virtual;
    procedure ResetSelRange;
    function CheckLastRow: boolean; virtual;
    procedure CheckLastRows;
    function GetCheckboxAlignment(ACol, ARow: longint): TAlignment;
    function GetChangedCellStr(ACol, ARow: longint): string;
    function IsCellChanged(ACol, ARow: longint): boolean;
    procedure UpdateChangedCell(UpdateRows: boolean; ChangedStartIdx, NumberOfItems: longint);
    function CompareValues(v1, v2: string; Column: TrStringGridEdColumn; CompareType: longint): TValueRelationship; virtual;
    function CompareCellData(ACol, ARow1, ARow2, CompVal: longint): TValueRelationship; virtual;
    procedure QuickSort(List: TStrings; L, R: integer; Column: TrStringGridEdColumn; CompareType: longint; var Proc: longint);
    procedure DoAbort(Sender: TObject);
    function DoGeneralEvent(EventType: TGeneralEventType): boolean;
    function DoGeneralEventParam(EventType: TGeneralEventType; Col, Row: longint; ParamObject: TObject; ParamValue: variant): boolean;
  public
    constructor Create(Owner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Loaded; override; {< @exclude }
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: integer); override; {< @exclude }
    procedure CheckColumns; {< @exclude }
    { start updating of the cell data - blocks refreshing after each update, call EndUpdate after all updates }
    procedure BeginUpdate(UseChangeLog: boolean = false);
    { finish updating of the cell data, call after all updates if BeginUpdate is used }
    procedure EndUpdate;
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { returns grid coord of last Mouse down (click or dblclick) }
    property MouseDownCoord: TGridCoord read MDownCoord;
    { returns point coord of the last Mouse down in the selected cell (reset to (0,0) after cell selection) }
    property MouseDownCellPoint: TPoint read MDownCellPoint;
    { access to internal inplace editor }
    function InplaceEditorEx: TrStringGridEdInplaceEdit;
    { read and set cells as string value }
    property StringCells[ACol, ARow: longint]: string read GetStringCells write SetStringCells;
    { read and set cells as double value }
    property NumCells[ACol, ARow: longint]: double read GetNumCells write SetNumCells;
    { read and set cells as integer value }
    property IntCells[ACol, ARow: longint]: integer read GetIntCells write SetIntCells;
    { read and set cells as boolean value }
    property BoolCells[ACol, ARow: longint]: boolean read GetBoolCells write SetBoolCells;
    { read and set cells as boolean value }
    property DateTimeCells[ACol, ARow: longint]: TDateTime read GetDateTimeCells write SetDateTimeCells;
    { return state of data in cell - empty, ok, invalid, outoflimit }
    function IsValidData(ACol, ARow: longint): TrStringGridEdDataState;
    { return type of data in cell - text, int, double, datetime }
    function GetDataType(ACol, ARow: longint): TrStringGridEdDataType; virtual;
    { return alignment of data in cell }
    function GetDataAlignment(ACol, ARow: longint): TAlignment; virtual;
    { return edit style of data in cell }
    function GetEditStyleEx(ACol, ARow: longint): TrStringGridEdEditStyle; virtual;
    { return numeric format of data in cell }
    function GetNumFormat(ACol, ARow: longint): string; virtual;
    { set Canvas style for data in cell }
    procedure GetCellStyle(ACol, ARow: longint; ACanvas: TCanvas);
    { return or set selection range coordinates }
    property SelectedRange: TGridRect read GetSelectedRange write SetSelectedRange;
    { return whether cell is inside of range select }
    function CellInSelectedRange(ACol, ARow: longint): boolean;
    { clear content of selected range }
    procedure ClearSelRange;
    { copy selected range to clipboard, columns are separated by TABs and prepared to paste in Excel }
    function CopySelRange: string;
    { paste values from clipboard to cells starting by selected cell (columns separated by TABs) }
    procedure PasteSelRange(CanExpandGrid: boolean; const PasteData: string = '');
    { copy entire table to clipboard, returns clipboard text }
    function CopyEntireTable(IncludingFixedRows: boolean = true; IncludingFixedColumns: boolean = true): string;
    { paste values from clipboard to grid from cell 0,0  (columns separated by TABs) }
    procedure PasteEntireTable(CanExpandGrid: boolean; const PasteData: string = '');
    { clear all cells }
    procedure ClearEntireTable(IncludingFixedCells: boolean = false);
    { store entire table to file, columns are separated by TABs and prepared for using in Excel }
    procedure SaveToFile(const FileName: string; Encoding: TEncoding = nil);
    { load entire table from file (columns separated by TABs) }
    procedure LoadFromFile(const FileName: string; SkipAllChecks: boolean = false; Encoding: TEncoding = nil);

    { find row index of the Key }
    function FindIniRow(const IniKey: string): longint;
    { load or set Value for the Key }
    property IniValues[const IniKey: string]: string read GetIniValue write SetIniValue;
    { store 2 columns to ini file if defined and/or return inifile text content (Key=Value) }
    function SaveToIniFile(const IniFileName: string; Encoding: TEncoding = nil): string;
    { load 2 columns from ini file or from inifile text content (Key=Value),
      increase row count if needed }
    procedure LoadFromIniFile(const IniFileName: string; const IniFileContent: string = ''; Encoding: TEncoding = nil);
    { load ValueColumn from ini file (Key=Value), do not change KeyColumn and row count }
    procedure LoadValuesFromIniFile(const IniFileName: string; const IniFileContent: string = ''; Encoding: TEncoding = nil);
    { return whether row is defined as Group row for INI grid }
    function IsIniGroupRow(ARow: longint): boolean;

    { set column width according data width }
    procedure SetColumnAutoFit(ColIdx: longint);
    { set all columns width according data width }
    procedure SetColumnsAutoFit;
    { get maximal length or Field datawidth for the column }
    function GetColumnDataWidth(ColIdx: longint; MaxRowsToTest: longint = 10000): integer; virtual;
    { simulate click to ellipsis button }
    procedure DoEllipsisClick;
    { hide editor and finish editing }
    procedure FinishEditing;
    { store list of Row index which was modified }
    property ChangedRows: TStrings read FChangedRows;
    { store list of cells coordinates which was modified; format: ColIdx;RowIdx }
    property ChangedCells: TStrings read FChangedCells;
    { reset list of changed Rows and Cells }
    procedure ResetChangeLog;
    { return content of active cell }
    function ActiveCellText: string;
    { return true if all cells in row is empty }
    function IsRowEmpty(ARow: longint; SkipFixedCol: boolean = true; ColumnGroup: integer = 0): boolean;
    { move cursor to first selectable cell }
    procedure SelectFirstCell(minCol: longint = 0; minRow: longint = 0);

    { move row to new position }
    procedure MoveRowEx(FromIndex, ToIndex: longint);
    { insert new empty row to define index position }
    procedure InsertRowEx(RowIndex: longint; NumberOfRows: integer = 1; Activate: boolean = true);
    { delete whole row }
    procedure DeleteRowEx(RowIndex: longint; NumberOfRows: integer = 1);
    { move column to new position }
    procedure MoveColumnEx(FromIndex, ToIndex: longint);
    { insert new empty column to define index position }
    procedure InsertColumnEx(ColIndex: longint; NumberOfCols: integer = 1; Activate: boolean = true);
    { delete whole column }
    procedure DeleteColumnEx(ColIndex: longint; NumberOfCols: integer = 1);
    { sort data/rows according selected column }
    procedure SortData(ColIndex: integer; Desc: boolean = false; ResetSelection: boolean = false);
    { top margin of RitchText in the cell }
    property RichTextTopMargin: integer read FRichTextTopMargin write SetRichTextTopMargin;
  published
    // property OnMouseWheelUp: TMouseWheelUpDownEvent read FOnMouseWheelUp write FOnMouseWheelUp;
    // property OnMouseWheelDown: TMouseWheelUpDownEvent read FOnMouseWheelDown write FOnMouseWheelDown;
    { define link to rCustomGlyphs component which can define all glyphs }
    property CustomGlyphs: TrCustomGlyphs read FCustomGlyphs write SetCustomGlyphs;
    { list of new options }
    property OptionsEx: TOptionsEx read FOptionsEx write SetOptionsEx default[];
    { define number of columns, columns collection is updated automatically }
    property ColCount: longint read GetColCount write SetColCount default 5;
    { columns collection, for detail see TrStringGridEdColumns }
    property Columns: TrStringGridEdColumns read FColumns write FColumns;

    { define index of Key column for INI function }
    property IniKeyColumn: integer read FIniKeyColumn write FIniKeyColumn default 0;
    { define index of Value column for INI function }
    property IniValueColumn: integer read FIniValueColumn write FIniValueColumn default 1;

    { define inner space between cell border and text in cell }
    property InnerMargin: integer read FInnerMargin write SetInnerMargin default 0;
    { Defines number of lines in title row }
    property TitleRowHeight: integer read FTitleRowHeight write SetTitleRowHeight default 0;
    { define whether row of column is set automatically, can be used with textwrap property of some column }
    property AutoRowHeight: boolean read FAutoRowHeight write SetAutoRowHeight default false;
    { define maximal row height during automatic setting (used if AutoRowHeight is set) }
    property AutoRowMaxHeight: integer read FAutoRowMaxHeight write FAutoRowMaxHeight default 0;
    { define whether new row is automatically added if last row is edited }
    property AutoAddNewRow: boolean read FAutoAddNewRow write SetAutoAddNewRow default false;
    { define background color of changed cells (used if dgColorChangedCells is set to true) }
    property ChangedCellsBkColor: TColor read FChangedCellsBkColor write SetChangedCellsBkColor default clInfoBk;
    { define background color of active cell with focus, default highlight is used if set to clNone }
    property SelectedCellBkColor: TColor read FSelectedCellBkColor write SetSelectedCellBkColor default clNone;
    { define background color of cell in the selected range, default highlight is used if set to clNone }
    property InRangeCellsBkColor: TColor read FInRangeCellsBkColor write SetInRangeCellsBkColor default clNone;
{$IFNDEF TrPopupMonthCalendar}
    { event type for creating of DateTimePicker }
    property DateTimePickerCreate: TDateTimePickerCreate read FDateTimePickerCreate write FDateTimePickerCreate;
    { event type for DateTimePicker action }
    property DateTimePickerEvent: TDateTimePickerEvent read FDateTimePickerEvent write FDateTimePickerEvent;
{$ENDIF}
    { event type for creating of CustomEditor }
    property CustomEditorCreate: TCustomEditorCreate read FCustomEditorCreate write FCustomEditorCreate;
    { event type for CustomEditor action }
    property CustomEditorEvent: TCustomEditorEvent read FCustomEditorEvent write FCustomEditorEvent;
    { event is called to draw cell background directly }
    property OnDrawCellBackground: TDrawCellEvent read FOnDrawCellBackground write FOnDrawCellBackground;
    { event for user defined setting of hint string }
    property OnGetHintText: TGetHintText read FGetHintText write FGetHintText;
    { event for user defined action after user click to any non fixed cell }
    property OnCellClick: TFixedCellClickEvent read FOnCellClick write FOnCellClick;
    { event for user defined action after click to ellipsis button }
    property OnEllipsisClicked: TrStringGridEdCellEvent read FOnEllipsisClicked write FOnEllipsisClicked;
    { event for user defined setting of values after click on drop down button }
    property OnPicklistDropdown: TrStringGridEdDropdownEvent read FOnPicklistDropdown write FOnPicklistDropdown;
    { universal event called in different events of inplace editor }
    property OnEditorEvent: TrStringGridEdEditorEvent read FOnEditorEvent write FOnEditorEvent;
    { event for user defined setting of individual cell params - font, color }
    property OnGetCellParams: TGetCellParamsEvent read FOnGetCellParams write FOnGetCellParams;
    { event for user defined setting of individual cell extended params - font, color, image }
    property OnGetCellParamsEx: TGetCellParamsEventEx read FOnGetCellParamsEx write FOnGetCellParamsEx;
    { event for user defined setting of datatype for individual cell }
    property OnGetDataType: TGetDataTypeEvent read FOnGetDataType write FOnGetDataType;
    { event for user defined setting of numformat for individual cell }
    property OnGetNumFormat: TGetNumFormat read FOnGetNumFormat write FOnGetNumFormat;
    { event for user defined setting of editstyle for individual cell }
    property OnGetEditStyle: TGetEditStyleEvent read FOnGetEditStyle write FOnGetEditStyle;
    { event for user defined setting whether individual cell can be edited }
    property OnCanEdit: TCanEditEvent read FOnCanEdit write FOnCanEdit;
    { event for user defined action after editing of cell is finished }
    property OnFinishCellEditing: TrStringGridEdCellEvent read FOnFinishCellEditing write FOnFinishCellEditing;
    { event for user defined action after press of key in editor }
    property OnEditorKeyPress: TKeyPressEvent read FOnEditorKeyPress write FOnEditorKeyPress;
    { event for user defined action for setting of default values }
    property OnColumnRestoreDefault: TOnColumnRestoreDefault read FOnColumnRestoreDefault write FOnColumnRestoreDefault;
    { event is called to define format of Ellipsis button in the cell }
    property OnGetEllipsisButtonFormat: TGetEllipsisButtonFormat read FOnGetEllipsisButtonFormat write FOnGetEllipsisButtonFormat;
    { event is called during sorting of data }
    property OnProgress: TOnProgress read FOnProgress write FOnProgress;
    { event is called in many events to add posibility of any custom code and to skip rest of the standard event }
    property OnGeneralEvent: TGeneralEvent read FOnGeneralEvent write FOnGeneralEvent;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rpictool, rdlg;

const
  cIntTimerID = 2099;
  cesShowBtn = [sgbEllipsis, sgbEllipsisOnly, sgbDropDown, sgbDropDownList, sgbDropDownColor, sgbDateTimePicker, sgbDropDownCheckList,
    sgbDropDownCheckListOnly, sgbDropDownDataImage];
  cesDropDown = [sgbDropDown, sgbDropDownList, sgbDropDownColor, sgbDropDownCheckList, sgbDropDownCheckListOnly, sgbDropDownDataImage];
  cesReadOnly = [sgbEllipsisOnly, sgbDropDownList, sgbDropDownColor, sgbDropDownCheckListOnly, sgbButton];
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 30 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi', [TrStringGridEd]);
end;

// =============================================================================
// TColumn
constructor TrDataBar.Create(Owner: TrStringGridEdColumn);
begin
  FMinValue:=0;
  FMaxValue:=100;
  FColorFrom:=clNone;
  FColorTo:=clNone;
  FText:=dbtValue;
end;

function TrDataBar.IsMaxValueStored: boolean;
begin
  Result:=FMaxValue <> 100;
end;

constructor TrStatusBar.Create(Owner: TrStringGridEdColumn);
begin
  self.FSteps:=5;
  self.FColorAll:=clNone;
  self.FColorDone:=clNone;
  self.FIconType:=sbtDots;
end;

constructor TrStringGridEdColumn.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FDataBar:=TrDataBar.Create(self);
  FStatusBar:=TrStatusBar.Create(self);
  FFont:=TFont.Create;
  FFont.OnChange:=FontChanged;
  FTitleFont:=TFont.Create;
  FTitleFont.OnChange:=TitleFontChanged;
  FParentFont:=true;
  FTitleParentFont:=true;
  FTitleColor:=clNone;
  FPickList:=TStringList.Create;
  FInvalidDataColor:=clRed;
  FInvalidDataAction:=sgaHighlight;
  FVisible:=true;
  FCustomDraw:=false;
  RestoreDefaults;
end;

destructor TrStringGridEdColumn.Destroy;
begin
  FreeAndNil(FDataBar);
  FreeAndNil(FStatusBar);
  FreeAndNil(FFont);
  FreeAndNil(FTitleFont);
  FreeAndNil(FPickList);
  inherited Destroy;
end;

procedure TrStringGridEdColumn.Assign(Source: TPersistent);
var
  s: TrStringGridEdColumn;
begin
  if Assigned(Source) and (Source is TrStringGridEdColumn) then
  begin
    s:=Source as TrStringGridEdColumn;
    FReadOnly:=s.FReadOnly;
    FVisible:=s.FVisible;
    FButtonStyle:=s.FButtonStyle;
    FDataType:=s.FDataType;
    FNumFormat:=s.FNumFormat;
    FMinValue:=s.FMinValue;
    FMaxValue:=s.FMaxValue;
    FInvalidDataAction:=s.FInvalidDataAction;
    FInvalidDataColor:=s.FInvalidDataColor;
    FOutOfLimitDataColor:=s.FOutOfLimitDataColor;
    FDropDownRows:=s.FDropDownRows;
    FDropDownListCanDelete:=s.FDropDownListCanDelete;
    FCheckBoxTrueValue:=s.FCheckBoxTrueValue;
    FCheckBoxFalseValue:=s.FCheckBoxFalseValue;
    FCheckBoxDblClick:=s.FCheckBoxDblClick;
    FCheckListSeparator:=s.FCheckListSeparator;
    FCheckListCaseSensitive:=s.FCheckListCaseSensitive;
    FAlignment:=s.FAlignment;
    FCheckBoxAlignment:=s.FCheckBoxAlignment;
    FAutoAlignment:=s.FAutoAlignment;
    FColor:=s.FColor;
    FFont.Assign(s.Font);
    ParentFont:=s.FParentFont;
    FPickList.Assign(s.FPickList);
    FReadOnly:=s.FReadOnly;
    FVisible:=s.Visible;
    FTitleCaption:=s.FTitleCaption;
    FTitleFont.Assign(s.FTitleFont);
    TitleParentFont:=s.FTitleParentFont;
    FTitleColor:=s.FTitleColor;
    FTitleAlignment:=s.FTitleAlignment;
    FWordWrap:=s.FWordWrap;
    FTitleWordWrap:=s.FTitleWordWrap;
    Width:=s.FWidth;
    FFieldName:=s.FFieldName;
    FColumnGroup:=s.FColumnGroup;
    FDropDownDrawItemEvent:=s.FDropDownDrawItemEvent;
    FDropDownMeasureItemEvent:=s.FDropDownMeasureItemEvent;
    FDropDownGetSize:=s.FDropDownGetSize;
  end;
end;

procedure TrStringGridEdColumn.RestoreDefaults;
begin
  // set default values for one Column
  FReadOnly:=false;
  FDropDownRows:=12;
  FButtonStyle:=sgbNone;
  FPickList.Clear;
  FAlignment:=taLeftJustify;
  FAutoAlignment:=false;
  FTitleAlignment:=taLeftJustify;
  FCheckBoxAlignment:=taCenter;
  FColor:=clWindow;
  FTitleCaption:='';
  FCheckBoxTrueValue:='1';
  FCheckBoxFalseValue:='0';
  FCheckBoxDblClick:=false;
  FCheckListSeparator:=',';
  FCheckListCaseSensitive:=false;
  FDataType:=sgdText;
  FNumFormat:='';
  FInvalidDataAction:=sgaHighlight;
  FInvalidDataColor:=clRed;
  FMinValue:=0;
  FMaxValue:=0;
  ParentFont:=true;
  TitleParentFont:=true;
  Width:=90;
  if Assigned(Grid) then
  begin
    Width:=Grid.DefaultColWidth;
    FColor:=Grid.Color;
    if Assigned(Grid.FOnColumnRestoreDefault) then
      Grid.FOnColumnRestoreDefault(self);
  end;
end;

function TrStringGridEdColumn.CheckListContainValue(Text, Value: string): boolean;
var
  s: string;
begin
  // sgbDropDownCheckList, sgbDropDownCheckListOnly
  repeat
    s:=rtool.RemoveStrPart(Text, FCheckListSeparator);
    if FCheckListCaseSensitive then
      Result:=(s <> '') and (s = Value)
    else
      Result:=(s <> '') and SameText(s, Value);
    if Result or (Text = '') then
      Break;
  until (false);
end;

function TrStringGridEdColumn.CheckListAddValue(Text: string; Value: string): string;
begin
  if Text = '' then
    Result:=Value
  else
    if CheckListContainValue(Text, Value) then
      Result:=Text
    else
      Result:=Text + FCheckListSeparator + Value;
end;

function TrStringGridEdColumn.CheckListRemoveValue(Text: string; Value: string): string;
var
  s: string;
  b: boolean;
begin
  Result:='';
  repeat
    s:=rtool.RemoveStrPart(Text, FCheckListSeparator);
    if FCheckListCaseSensitive then
      b:=s = Value
    else
      b:=SameText(s, Value);
    if not b then
      Result:=CheckListAddValue(Result, s);
    if (Text = '') then
      Break;
  until false;
end;

function TrStringGridEdColumn.CheckListCleanUpText(Text: string; SortAsPickList: boolean = false): string;
var
  i: integer;
  s: string;
begin
  if not(FButtonStyle in [sgbDropDownCheckList, sgbDropDownCheckListOnly]) then
  begin
    Result:=Text;
    Exit;
  end;
  Result:='';
  if SortAsPickList then
  begin
    for i:=0 to FPickList.Count - 1 do
    begin
      if CheckListContainValue(Text, FPickList[i]) then
        Result:=CheckListAddValue(Result, FPickList[i]);
    end;
  end
  else
  begin
    repeat
      s:=rtool.RemoveStrPart(Text, FCheckListSeparator);
      if FPickList.IndexOf(s) >= 0 then
        Result:=CheckListAddValue(Result, s);
      if (Text = '') then
        Break;
    until false;
  end;
end;

function TrStringGridEdColumn.CheckListCleanUpRow(ARow: longint; SortAsPickList: boolean = false): string;
var
  v, newv: string;
begin
  if (ARow >= Grid.FixedRows) and (ARow < Grid.RowCount) then
  begin
    v:=Grid.Cells[self.Index, ARow];
    newv:=CheckListCleanUpText(v, SortAsPickList);
    if newv <> v then
      Grid.Cells[self.Index, ARow]:=newv;
  end;
end;

function TrStringGridEdColumn.GetGrid: TrStringGridEd;
begin
  if Assigned(Collection) and (Collection is TrStringGridEdColumns) then
    Result:=TrStringGridEdColumns(Collection).Grid
  else
    Result:=nil;
end;

procedure TrStringGridEdColumn.SetReadOnly(Value: boolean);
begin
  if (Value = FReadOnly) then
    Exit;
  FReadOnly:=Value;
end;

procedure TrStringGridEdColumn.SetVisible(Value: boolean);
begin
  if (Value = FVisible) then
    Exit;
  FVisible:=Value;
  if FVisible then
    Grid.ColWidths[Index]:=FWidth
  else
  begin
    FWidth:=Grid.ColWidths[Index];
    Grid.ColWidths[Index]:=-1;
  end;
end;

procedure TrStringGridEdColumn.SetDataImageList(Value: TImageList);
begin
  FDataImageList:=Value;
  if Value <> nil then
    Value.FreeNotification(Grid);
  Changed(false);
end;

procedure TrStringGridEdColumn.SetPickList(Value: TStringList);
begin
  FPickList.Assign(Value);
end;

procedure TrStringGridEdColumn.SetButtonStyle(Value: TrStringGridEdEditStyle);
begin
  if Value = FButtonStyle then
    Exit;
  FButtonStyle:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetAlignment(Value: TAlignment);
begin
  if (Value = FAlignment) then
    Exit;
  FAlignment:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetCheckBoxAlignment(Value: TAlignment);
begin
  if (Value = FCheckBoxAlignment) then
    Exit;
  FCheckBoxAlignment:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetColor(Value: TColor);
begin
  if (Value = FColor) then
    Exit;
  FColor:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TrStringGridEdColumn.SetTitleFont(Value: TFont);
begin
  FTitleFont.Assign(Value);
end;

procedure TrStringGridEdColumn.SetTitleAlignment(Value: TAlignment);
begin
  if (Value = FTitleAlignment) then
    Exit;
  FTitleAlignment:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetTitleCaption(const Value: string);
begin
  if (Value = FTitleCaption) then
    Exit;
  FTitleCaption:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetParentFont(Value: boolean);
begin
  if Value then
    FFont.Assign(Grid.Font);
  FParentFont:=Value;
end;

procedure TrStringGridEdColumn.SetTitleParentFont(Value: boolean);
begin
  if Value then
    FTitleFont.Assign(Grid.Font);
  FTitleParentFont:=Value;
end;

procedure TrStringGridEdColumn.SetTitleColor(Value: TColor);
begin
  if (Value = FTitleColor) then
    Exit;
  FTitleColor:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetWidth(Value: integer);
begin
  FWidth:=Value;
  Grid.ColWidths[Index]:=Value;
end;

function TrStringGridEdColumn.GetWidth: integer;
begin
  Result:=FWidth;
  // if Result=0 then Result:=Grid.ColWidths[Index];
end;

function TrStringGridEdColumn.IsFontStored: boolean;
begin
  Result:=not FParentFont;
end;

function TrStringGridEdColumn.IsTitleFontStored: boolean;
begin
  Result:=not FTitleParentFont;
end;

procedure TrStringGridEdColumn.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  Changed(false);
end;

procedure TrStringGridEdColumn.TitleFontChanged(Sender: TObject);
begin
  FTitleParentFont:=false;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetCustomDraw(Value: boolean);
begin
  FCustomDraw:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetWordWrap(Value: boolean);
begin
  FWordWrap:=Value;
  Changed(false);
end;

procedure TrStringGridEdColumn.SetTitleWordWrap(Value: boolean);
begin
  FTitleWordWrap:=Value;
  Changed(false);
end;

function TrStringGridEdColumn.CheckBoxTrueValueStored: boolean;
begin
  Result:=FCheckBoxTrueValue <> '1';
end;

function TrStringGridEdColumn.CheckBoxFalseValueStored: boolean;
begin
  Result:=FCheckBoxFalseValue <> '0';
end;

function TrStringGridEdColumn.CheckListSeparatorStored: boolean;
begin
  Result:=FCheckListSeparator <> ',';
end;

// =============================================================================
// TColumns Collection
constructor TrStringGridEdColumns.Create(Grid: TrStringGridEd; ColumnClass: TrStringGridEdColumnClass);
begin
  inherited Create(ColumnClass);
  FGrid:=Grid;
end;

function TrStringGridEdColumns.Add: TrStringGridEdColumn;
begin
  Result:=TrStringGridEdColumn( inherited Add);
end;

function TrStringGridEdColumns.GetColumn(Index: integer): TrStringGridEdColumn;
begin
  Result:=TrStringGridEdColumn( inherited Items[Index]);
end;

procedure TrStringGridEdColumns.SetColumn(Index: integer; Value: TrStringGridEdColumn);
begin
  Items[Index].Assign(Value);
end;

function TrStringGridEdColumns.GetOwner: TPersistent;
begin
  Result:=FGrid;
end;

procedure TrStringGridEdColumns.RestoreDefaults;
var
  i: integer;
begin
  // set default for all columns
  BeginUpdate;
  try
    for i:=0 to Count - 1 do
      Items[i].RestoreDefaults;
  finally
    EndUpdate;
  end;
end;

procedure TrStringGridEdColumns.Update(Item: TCollectionItem);
begin
  // column props was updated => redraw
  Grid.Invalidate;
end;

// =============================================================================
// InplaceEdit ListBox  (copied from DBGrid)
procedure TPopupListbox.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style:=Style or WS_BORDER;
    ExStyle:=WS_EX_TOOLWINDOW or WS_EX_TOPMOST;
    WindowClass.Style:=CS_SAVEBITS;
  end;
end;

procedure TPopupListbox.CreateWnd;
begin
  inherited CreateWnd;
  Windows.SetParent(Handle, 0);
  CallWindowProc(DefWndProc, Handle, WM_SETFOCUS, 0, 0);
end;

procedure TPopupListbox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  // another F4 close dropdown
  if Key = VK_F4 then
    TrStringGridEdInplaceEdit(Owner).CloseUp(false, true);
  inherited KeyDown(Key, Shift);
end;

procedure TPopupListbox.KeyPress(var Key: Char);
var
  TickCount: Cardinal;
begin
  // space key change checklist item
  if IsCheckList and (Key = ' ') then
    TrStringGridEdInplaceEdit(Owner).CloseUp(true, false);
  // item lookup
  case Key of
    #8, #27:
      FSearchText:='';
    #32 .. #255:
      begin
        TickCount:=GetTickCount;
        if TickCount - FSearchTickCount > 2000 then
          FSearchText:='';
        FSearchTickCount:=TickCount;
        if Length(FSearchText) < 32 then
          FSearchText:=FSearchText + Key;
        SendMessage(Handle, LB_SelectString, wParam(-1), lParam(PChar(FSearchText)));
        Key:=#0;
      end;
  end;
  inherited KeyPress(Key);
end;

procedure TPopupListbox.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  TrStringGridEdInplaceEdit(Owner).CloseUp((X >= 0) and (Y >= 0) and (X < Width) and (Y < Height), not IsCheckList);
end;

procedure TPopupListbox.DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState);
var
  FColor, BkColor: TColor;
  Bmp: TBitmap;
  c: TColor;
  R: TRect;
  idx: integer;
  Details: TThemedElementDetails;
  btnStyle: TThemedButton;
  btnState: Cardinal;
  s, t: string;
begin
  if UseOwnDraw then
  begin
    FColor:=Font.Color;
    BkColor:=Color;

    R:=Rect;
    s:=Items[Index];
    if not(odSelected in State) then
      Canvas.Brush.Color:=GetStyleBkColor_ListBox(BkColor);
    Canvas.FillRect(R);

    if IsColor then
    begin
      // draw colored rect
      R.Right:=R.Left + R.Bottom - R.Top;
      idx:=integer(Items.Objects[Index]);
      c:=rpictool.GetColorValue(idx);
      InflateRect(R, -1, -1);
      BiDi_FrameRect(Canvas, R, clBlack);
      InflateRect(R, -1, -1);
      Canvas.Brush.Color:=c;
      if c <> clNone then
        Canvas.FillRect(R);
    end;

    if IsCheckList then
    begin
      // draw checkbox
      R.Right:=R.Left + R.Bottom - R.Top;
      t:=TrStringGridEdInplaceEdit(Owner).Text;
      if GetThemeEnabled then
      begin
        if AColumn.CheckListContainValue(t, s) then
          btnStyle:=tbCheckBoxCheckedNormal
        else
          btnStyle:=tbCheckBoxUnCheckedNormal;
        Details:=GetThemeServices.GetElementDetails(btnStyle);
        InflateRect(R, -1, -1);
        GetThemeServices.DrawElement(Canvas.Handle, Details, R);
      end
      else
      begin
        btnState:=DFCS_BUTTONCHECK;
        if AColumn.CheckListContainValue(t, s) then
          btnState:=btnState or DFCS_CHECKED;
        DrawFrameControl(Canvas.Handle, R, DFC_BUTTON, btnState);
      end;
    end;

    Canvas.Font.Color:=FColor;
    Canvas.Font.Color:=GetStyleFontColor_ListBox(FColor, odSelected in State);
    if IsDataImage then
    begin
      if Assigned(AColumn.FDataImageList) and (s <> '') then
      begin
        // draw colored rect
        idx:=StrToIntDef(s, -1);
        if (idx >= 0) and (idx < AColumn.FDataImageList.Count) then
        begin
          s:='';
          // draw bk
          Canvas.TextRect(Rect, Rect.Left, Rect.Top, ' ');
          InflateRect(R, -1, -1);
          // display image
          Bmp:=TBitmap.Create;
          try
            Bmp.Canvas.Brush.Color:=Canvas.Brush.Color;
            Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
            Bmp.Transparent:=AColumn.FDataImageList.tag = 0;
            AColumn.FDataImageList.GetBitmap(idx, Bmp);
            rpictool.DrawBmpCenter(Canvas, R, Bmp);
          finally
            FreeAndNil(Bmp);
          end;
        end;
      end;
    end;

    // draw text
    Canvas.Brush.Style:=bsClear;
    R:=Rect;
    R.Left:=R.Left + R.Bottom - R.Top;
    if s <> '' then
      Canvas.TextRect(R, R.Left + 2, R.Top + 1, s);
    Canvas.Brush.Color:=BkColor;
  end
  else
    inherited DrawItem(Index, Rect, State);
end;

procedure TPopupListbox.MeasureItem(Index: integer; var Height: integer);
begin
  inherited MeasureItem(Index, Height);
  GetItemHeight(Height);
end;

function TPopupListbox.UseOwnDraw: boolean;
begin
  Result:=IsColor or IsCheckList or IsDataImage;
end;

procedure TPopupListbox.GetItemHeight(var H: integer);
begin
  if IsColor or IsCheckList then
    H:=Canvas.TextHeight('Wg') + 2;
  if IsDataImage and Assigned(AColumn.FDataImageList) then
    H:=AColumn.FDataImageList.Height + 2;
end;

// =============================================================================
// InplaceEditor (copied from DBGrid and modified + enhanced)
constructor TrStringGridEdInplaceEdit.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FEllipsisBtnWidth:=GetSystemMetrics(SM_CXVSCROLL);
  FActBtnWidth:=FEllipsisBtnWidth;
  FEditStyle:=sgbEllipsis;
  if not(csDesigning in ComponentState) then
  begin
    // create PopupCalendar - use SELF as Owner
    FPopupCalendar:=TrPopupMonthCalendar.CreateEx(self, PopupCallbackSelectDate);
  end;
end;

destructor TrStringGridEdInplaceEdit.Destroy;
begin
{$IFNDEF TrPopupMonthCalendar}
  if Assigned(FDateTimePicker) then
    FreeAndNil(FDateTimePicker);
{$ENDIF}
  if Assigned(FRichEditor) then
    FreeAndNil(FRichEditor);
  if Assigned(FCustomEditor) then
    FreeAndNil(FCustomEditor);
  inherited Destroy;
end;

function TrStringGridEdInplaceEdit.OwnerGrid: TCustomGrid;
begin
  Result:=Grid;
end;

procedure TrStringGridEdInplaceEdit.BoundsChanged;
var
  R: TRect;
  M, th: integer;
  wrap: boolean;
begin
  if Assigned(Grid) then
  begin
    with TrStringGridEd(Grid) do
    begin
      CheckColumns;
      wrap:=GetCellWordWrap(Col, Row); // Columns[Col].WordWrap;
      M:=InnerMargin;
      Canvas.Font:=self.Font;
      th:=Canvas.TextHeight('Wg');
    end;
    // add inner space
    SetRect(R, 2 + M, 2, Width - 2 - M, Height);
    if not wrap and (R.Bottom - R.Top > th) then
    begin
      // center editor vertically
      R.Top:=(R.Bottom - R.Top - th) div 2 + 1;
      R.Bottom:=R.Top + th;
    end;
  end;
  if FEditStyle in cesShowBtn then
    Dec(R.Right, FActBtnWidth);
  SendMessage(Handle, EM_SETRECTNP, 0, lParam(@R));
  SendMessage(Handle, EM_SCROLLCARET, 0, 0);
  if SysLocale.Fareast then
    SetImeCompositionWindow(Font, R.Left, R.Top);
end;

// ==============================================================================
// DateTimePicker block
// FPopupCalendar is used instead of TDateTimePicker if compiler switch TrPopupMonthCalendar is defined
function TrStringGridEdInplaceEdit.GetDateCalendarVisible: boolean;
begin
{$IFDEF TrPopupMonthCalendar}
  Result:=FPopupCalendar.IsShown
{$ELSE}
  Result:=FDatePickerVisible;
{$ENDIF}
end;

procedure TrStringGridEdInplaceEdit.CreateHideCalendar;
begin
{$IFDEF TrPopupMonthCalendar}
  FPopupCalendar.HideCalendar;
  FActiveList:=FPopupCalendar;
{$ELSE}
  RecreateDatePicker;
  FActiveList:=FDateTimePicker;
{$ENDIF}
end;

procedure TrStringGridEdInplaceEdit.PopupCallbackSelectDate(Sender: TObject);
var
  dt: TDateTime;
begin
  dt:=Trunc(FPopupCalendar.Date);
  if FDataType = sgdDateTime then
    dt:=dt + TimeOf(FOrigDateTime);
  SetDateTime(dt);
end;
{$IFNDEF TrPopupMonthCalendar}

function TrStringGridEdInplaceEdit.CreateDateTimePicker: TWinControl;
begin
  // Init DateTimePicker
  Result:=nil;
  if Assigned(TrStringGridEd(Grid).FDateTimePickerCreate) then
    TrStringGridEd(Grid).FDateTimePickerCreate(Grid, Result, DateTimePickerChange, DateTimePickerClose)
  else
  begin
    Result:=TDateTimePicker.Create(self);
    TDateTimePicker(Result).OnChange:=DateTimePickerChange;
    TDateTimePicker(Result).OnCloseUp:=DateTimePickerClose;
  end;
  if Assigned(Result) then
  begin
    Result.SetSubComponent(true);
    Result.Visible:=false;
    Result.Parent:=Grid;
    Result.BoundsRect:=Rect(0, 0, 0, 0);
  end;
end;

procedure TrStringGridEdInplaceEdit.SetDateTimePickerDate(ADateTime: TDateTime);
var
  Done: boolean;
begin
  Done:=false;
  if Assigned(TrStringGridEd(Grid).FDateTimePickerEvent) then
    TrStringGridEd(Grid).FDateTimePickerEvent(Grid, FDateTimePicker, dpaSetDate, ADateTime, Done);
  if not Done then
  begin
    if (FDateTimePicker is TDateTimePicker) then
      TDateTimePicker(FDateTimePicker).Date:=DateOf(ADateTime);
  end;
end;

function TrStringGridEdInplaceEdit.GetDateTimePickerDate: TDateTime;
var
  Done: boolean;
begin
  Result:=Date;
  Done:=false;
  if Assigned(TrStringGridEd(Grid).FDateTimePickerEvent) then
    TrStringGridEd(Grid).FDateTimePickerEvent(Grid, FDateTimePicker, dpaGetDate, Result, Done);
  if not Done then
  begin
    if (FDateTimePicker is TDateTimePicker) then
      Result:=DateOf(TDateTimePicker(FDateTimePicker).Date);
  end;
end;

procedure TrStringGridEdInplaceEdit.ShowDateTimePicker(Rect: TRect);
var
  dt: TDateTime;
  Done: boolean;
begin
  if not Assigned(FDateTimePicker) then
    Exit;
  dt:=Date;
  Done:=false;
  FDateTimePicker.BoundsRect:=Rect;
  if Assigned(TrStringGridEd(Grid).FDateTimePickerEvent) then
    TrStringGridEd(Grid).FDateTimePickerEvent(Grid, FDateTimePicker, dpaShow, dt, Done);
  if not Done then
  begin
    FDateTimePicker.Show;
    try
      FDateTimePicker.SetFocus;
      // open picker
{$IFDEF DXE2P}
      if TStyleManager.IsCustomStyleActive then
      begin
        // refresh height of picker to get correct calendar size for style
        SendMessage(FDateTimePicker.Handle, CM_FONTCHANGED, 0, 0);
        SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
        // resize back to cell size
        FDateTimePicker.BoundsRect:=Rect;
      end
      else
{$ENDIF}
        SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
    except
      FDateTimePicker.Hide;
    end;
  end;
end;

procedure TrStringGridEdInplaceEdit.DateTimePickerChange(Sender: TObject);
var
  ListValue: string;
  dt: TDateTime;
begin
  if EditCanModify then
  begin
    dt:=GetDateTimePickerDate;
    if FDataType = sgdDateTime then
      dt:=dt + TimeOf(FOrigDateTime);
    ListValue:=DateTimeToFormattedStr(dt, FNumFormat);
    // ListValue:=ListValue + ' '+TimeToStr(TimeOf(dt));
    Perform(WM_SETTEXT, 0, lParam(string(ListValue)));
    Modified:=true;
    with TrStringGridEd(Grid) do
    begin
      SetEditText(Col, Row, ListValue);
      if not(goAlwaysShowEditor in Options) then
        HideEditor;
      DoFinishEditing(Col, Row, false);
      FBlockClick:=true;
    end;
  end;
end;

procedure TrStringGridEdInplaceEdit.DateTimePickerClose(Sender: TObject);
var
  dt: TDateTime;
  Done: boolean;
begin
  if Assigned(FDateTimePicker) then
  begin
    dt:=Date;
    Done:=false;
    if Assigned(TrStringGridEd(Grid).FDateTimePickerEvent) then
      TrStringGridEd(Grid).FDateTimePickerEvent(Grid, FDateTimePicker, dpaHide, dt, Done);
    if not Done then
      FDateTimePicker.Hide;
  end;
  FDatePickerVisible:=false;
  try
    Grid.SetFocus;
  except
  end;
  TrStringGridEd(Grid).FBlockClick:=true;
end;

procedure TrStringGridEdInplaceEdit.RecreateDatePicker;
begin
  // free is needed to hide opened picker, hide is not working
  if Assigned(FDateTimePicker) and FDatePickerVisible then
    FreeAndNil(FDateTimePicker);
  if not Assigned(FDateTimePicker) then
    FDateTimePicker:=CreateDateTimePicker;
  FDatePickerVisible:=false;
end;
{$ENDIF}

// ==============================================================================
// RichEditor block
procedure TrStringGridEdInplaceEdit.ShowRichEditor(Rect: TRect);
var
  k: Word;
begin
  if not Assigned(FRichEditor) then
    Exit;
  // hide standard inplace before showing of RichEdit editor to aviod calling after edit and replacement of the new value from custom edit (MUST)
  Hide;
  // set rich edit
  with TrStringGridEd(Grid) do
  begin
    k:=FStartEditKey;
    Rect.Top:=Rect.Top + FRichTextTopMargin;
    FRichEditor.BoundsRect:=Rect;
    FCustomEditCoord.X:=Col;
    FCustomEditCoord.Y:=Row;
    FRichEditor.RtfText:=GetEditText(Col, Row);
    FRichEditor.Color:=Color;
  end;

  FRichEditor.Show;
  try
    FRichEditor.SetFocus;
    FRichEditorVisible:=true;
  except
    FRichEditor.Hide;
  end;

  FRichEditor.SelStart:=FRichEditor.GetTextLen;
  FRichEditor.Perform(EM_SCROLLCARET, 0, 0);
  if (k <> VK_F2) then
    FRichEditor.SelectAll;
  if (k <> 0) and (k <> VK_F2) then
    PostMessage(FRichEditor.Handle, WM_CHAR, k, 0);
end;

procedure TrStringGridEdInplaceEdit.HideRichEditor(Accept: boolean);
begin
  if Assigned(FRichEditor) and FRichEditorVisible then
  begin
    if Accept then
      TrStringGridEd(Grid).StringCells[FCustomEditCoord.X, FCustomEditCoord.Y]:=FRichEditor.RtfText;
    FRichEditorVisible:=false;
    FRichEditor.Hide;
    try
      Grid.SetFocus;
    except
    end;
  end;
end;

procedure TrStringGridEdInplaceEdit.RichEditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Key = VK_ESCAPE then
    HideRichEditor(false);
  if Shift = [ssCtrl] then
  begin
    if (Key = VK_RETURN) then
      HideRichEditor(true);
  end;
end;

procedure TrStringGridEdInplaceEdit.RichEditPopupMenu(Sender: TObject);
begin
  case TMenuItem(Sender).tag of
    1001:
      HideRichEditor(true);
    1002:
      HideRichEditor(false);
  end;
end;

// ==============================================================================
// CustomEditor block
function TrStringGridEdInplaceEdit.CreateCustomEditor: TWinControl;
begin
  // Init CustomEditor
  Result:=nil;
  if Assigned(TrStringGridEd(Grid).FCustomEditorCreate) then
    TrStringGridEd(Grid).FCustomEditorCreate(Grid, Result);
  if Assigned(Result) then
  begin
    Result.SetSubComponent(true);
    Result.Visible:=false;
    Result.Parent:=Grid;
    Result.BoundsRect:=Rect(0, 0, 0, 0);
  end;
end;

procedure TrStringGridEdInplaceEdit.SetCustomEditorValue(AValue: string);
var
  Done: boolean;
begin
  Done:=false;
  if Assigned(TrStringGridEd(Grid).FCustomEditorEvent) then
    TrStringGridEd(Grid).FCustomEditorEvent(Grid, FCustomEditor, ceaSetValue, AValue, Done);
  if not Done then
  begin
    if FCustomEditor is TCustomMemo then
      TCustomMemo(FCustomEditor).Lines.Text:=AValue
    else
      if FCustomEditor is TCustomEdit then
        TCustomEdit(FCustomEditor).Text:=AValue;
  end;
end;

function TrStringGridEdInplaceEdit.GetCustomEditorValue: string;
var
  Done: boolean;
begin
  Result:='';
  Done:=false;
  if Assigned(TrStringGridEd(Grid).FCustomEditorEvent) then
    TrStringGridEd(Grid).FCustomEditorEvent(Grid, FCustomEditor, ceaGetValue, Result, Done);
  if not Done then
  begin
    if FCustomEditor is TCustomMemo then
      Result:=TCustomMemo(FCustomEditor).Lines.Text
    else
      if FCustomEditor is TCustomEdit then
        Result:=TCustomEdit(FCustomEditor).Text;
  end;
end;

procedure TrStringGridEdInplaceEdit.ShowCustomEditor(Rect: TRect);
var
  s: string;
  Done: boolean;
begin
  if not Assigned(FCustomEditor) then
    Exit;
  s:='';
  Done:=false;
  // hide standard inplace before showing of custom editor to aviod calling after edit and replacement of the new value from custom edit (MUST)
  Hide;
  // setup custom editor
  FCustomEditor.BoundsRect:=Rect;
  with TrStringGridEd(Grid) do
  begin
    FCustomEditCoord.X:=Col;
    FCustomEditCoord.Y:=Row;
    SetCustomEditorValue(GetEditText(Col, Row));
    if Assigned(FCustomEditorEvent) then
      FCustomEditorEvent(Grid, FCustomEditor, ceaShow, s, Done);
  end;
  if not Done then
  begin
    // show editor
    // SetWindowPos(FCustomEditor.Handle, HWND_TOP, Rect.Left, Rect.Top, Rect.Right - Rect.Left, Rect.Bottom - Rect.Top,
    // SWP_SHOWWINDOW or SWP_NOREDRAW);
    FCustomEditor.Show;
    try
      FCustomEditor.SetFocus;
    except
      FCustomEditor.Hide;
    end;
  end;
  FCustomEditorVisible:=true;
end;

procedure TrStringGridEdInplaceEdit.HideCustomEditor(Accept: boolean);
var
  s: string;
  Done: boolean;
begin
  if Assigned(FCustomEditor) and FCustomEditorVisible then
  begin
    s:='';
    Done:=false;
    if Accept then
      TrStringGridEd(Grid).StringCells[FCustomEditCoord.X, FCustomEditCoord.Y]:=GetCustomEditorValue;
    if Assigned(TrStringGridEd(Grid).FCustomEditorEvent) then
      TrStringGridEd(Grid).FCustomEditorEvent(Grid, FCustomEditor, ceaHide, s, Done);
    FCustomEditorVisible:=false;
    if not Done then
      FCustomEditor.Hide;
    try
      Grid.SetFocus;
    except
    end;
  end;
end;

// ==============================================================================
// Common Editor block
function TrStringGridEdInplaceEdit.EditCanModify: boolean;
begin
  if (EditStyle in [sgbDropDownList, sgbDropDownColor, sgbDropDownCheckListOnly, sgbDropDownDataImage]) then
    Result:=true
  else
    Result:= inherited EditCanModify;
end;

procedure TrStringGridEdInplaceEdit.SetDateTime(DateTimeValue: TDateTime);
begin
  if DateTimeValue = 0 then
    Text:=''
  else
    Text:=DateTimeToFormattedStr(DateTimeValue, FNumFormat);
  SelectAll;
end;

procedure TrStringGridEdInplaceEdit.DoDropDownKeys(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_UP, VK_DOWN:
      if ssAlt in Shift then
      begin
        if FListVisible or GetDateCalendarVisible then
          CloseUp(true, true)
        else
          DropDown;
        Key:=0;
      end;
    VK_RETURN, VK_ESCAPE:
      if (FListVisible or GetDateCalendarVisible) and not(ssAlt in Shift) then
      begin
        CloseUp(Key = VK_RETURN, true);
        Key:=0;
      end;
  end;
end;

procedure TrStringGridEdInplaceEdit.DropDown;
var
  p: TPoint;
  i, w, maxw, H, Y: integer;
  Gr: TrStringGridEd;
  Column: TrStringGridEdColumn;
  str: TStringList;
  R: boolean;
begin
  if not Assigned(FActiveList) then
    Exit;
  if (FEditStyle <> sgbDateTimePicker) and not FListVisible then
  begin
    // show listbox
    Gr:=TrStringGridEd(Grid);
    Gr.CheckColumns;
    // set all props of listbox
    Column:=Gr.Columns[Gr.Col];
    FPickList.AColumn:=Column;
    FPickList.Color:=self.Color;
    FPickList.Font:=self.Font;
    FPickList.Style:=lbStandard;
    // set items
    if FPickList.IsColor and (Column.PickList.Count = 0) then
      GetColorList(Column.PickList, true, true, true);
    if FPickList.IsDataImage then
    begin
      Column.PickList.Clear;
      if Assigned(Column.FDataImageList) then
        for i:=0 to Column.FDataImageList.Count - 1 do
          Column.PickList.Add(IntToStr(i));
    end;
    FPickList.Items:=Column.PickList;
    if Assigned(Gr.FOnPicklistDropdown) then
    begin
      str:=TStringList.Create;
      try
        str.Assign(FPickList.Items);
        Gr.FOnPicklistDropdown(self, Gr.Col, Gr.Row, str);
        FPickList.Items.Assign(str);
      finally
        FreeAndNil(str);
      end;
    end;
    // set list width according items
    FPickList.Width:=self.Width;
    maxw:=FPickList.ClientWidth;
    for i:=0 to FPickList.Items.Count - 1 do
    begin
      w:=FPickList.Canvas.TextWidth(FPickList.Items[i]);
      if w > maxw then
        maxw:=w + 10;
    end;
    if FPickList.UseOwnDraw then
      Inc(maxw, FPickList.ItemHeight);
    FPickList.ClientWidth:=maxw;
    // set list height
    H:=FPickList.ItemHeight;
    FPickList.GetItemHeight(H);
    if FPickList.Items.Count >= Column.DropDownRows then
      FPickList.Height:=Column.DropDownRows * H + 2
    else
      FPickList.Height:=FPickList.Items.Count * H + 2;
    // load size from event
    if Assigned(Column.FDropDownGetSize) then
    begin
      w:=FPickList.Width;
      H:=FPickList.Height;
      R:=true;
      Column.FDropDownGetSize(Column, w, H, R);
      if R then
      begin
        FPickList.Width:=w;
        FPickList.Height:=H;
      end;
    end;
    // ListBox OnDraw function
    if Assigned(Column.FDropDownDrawItemEvent) or FPickList.UseOwnDraw then
      FPickList.Style:=lbOwnerDrawVariable;
    FPickList.OnDrawItem:=Column.FDropDownDrawItemEvent;
    FPickList.OnMeasureItem:=Column.FDropDownMeasureItemEvent;
    // select active item
    if self.Text = '' then
      FPickList.ItemIndex:=-1
    else
      FPickList.ItemIndex:=FPickList.Items.IndexOf(self.Text);

    if Assigned(Gr.FOnEditorEvent) then
      Gr.FOnEditorEvent(self, Gr.Col, Gr.Row, 'PickListBeforeShow');
    // set screen possition
    p:=Parent.ClientToScreen(Point(Left, Top));
    Y:=p.Y + Height;
    if Y + FPickList.Height > Screen.Height then
      Y:=p.Y - FActiveList.Height;
    SetWindowPos(FPickList.Handle, HWND_TOP, p.X, Y, 0, 0, SWP_NOSIZE or SWP_NOACTIVATE or SWP_SHOWWINDOW);
    FListVisible:=true;
    if Assigned(Gr.FOnEditorEvent) then
      Gr.FOnEditorEvent(self, Gr.Col, Gr.Row, 'PickListAfterShow');
    Invalidate;
    Windows.SetFocus(Handle);
  end;

  if (FEditStyle = sgbDateTimePicker) and not GetDateCalendarVisible then
  begin
    // show date picker
    FOrigDateTime:=FormattedStrToDateTimeDef(Text, FNumFormat, Date, FDataType = sgdDate);
{$IFDEF TrPopupMonthCalendar}
    FPopupCalendar.ShowCalendar(FOrigDateTime);
{$ELSE}
    SetDateTimePickerDate(FOrigDateTime);
    ShowDateTimePicker(BoundsRect);
    FDatePickerVisible:=true;
{$ENDIF}
  end;
end;

procedure TrStringGridEdInplaceEdit.CloseUp(Accept, HideList: boolean);
var
  ListValue: string;
  c: TColor;
  cidx: integer;
  Dlg: TColorDialog;
  Gr: TrStringGridEd;
begin
  if FListVisible then
  begin
    Gr:=TrStringGridEd(Grid);
    if GetCapture <> 0 then
      SendMessage(GetCapture, WM_CancelMode, 0, 0);
    cidx:=-1;
    // load selected item and color index
    if (FPickList.ItemIndex >= 0) and (FPickList.ItemIndex < FPickList.Count) then
    begin
      ListValue:=FPickList.Items[FPickList.ItemIndex];
      try
        cidx:=integer(FPickList.Items.Objects[FPickList.ItemIndex]);
      except
      end;
    end
    else
      Accept:=false;
    if HideList then
    begin
      SetWindowPos(FActiveList.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_HIDEWINDOW);
      FListVisible:=false;
      if Assigned(Gr.FOnEditorEvent) then
        Gr.FOnEditorEvent(self, Gr.Col, Gr.Row, 'PickListAfterHide');
    end;
    Invalidate;
    if Accept and EditCanModify then
    begin
      // open ColorDialog for custom
      if FPickList.IsColor then
      begin
        if cidx = ColorDefCustomIdx then
        begin
          Dlg:=TColorDialog.Create(self);
          try
            Dlg.Color:=rpictool.StringToColor(self.Text);
            Dlg.Options:=[cdFullOpen, cdAnyColor];
            if not Dlg.Execute(self.Handle) then
              Exit;
            c:=Dlg.Color;
            ListValue:='$' + IntToHex(ColorToRGB(c), 8);
          finally
            FreeAndNil(Dlg);
          end;
        end;
      end;

      // add or remove value from text
      if FPickList.IsCheckList then
      begin
        if FPickList.AColumn.CheckListContainValue(Text, ListValue) then
          ListValue:=FPickList.AColumn.CheckListRemoveValue(Text, ListValue)
        else
          ListValue:=FPickList.AColumn.CheckListAddValue(Text, ListValue);
      end;

      { Here we store the new value directly in the edit control so that
        we bypass the CMTextChanged method on TCustomMaskedEdit.  This
        preserves the old value so that we can restore it later by calling
        the Reset method. }
      // set value
      Perform(WM_SETTEXT, 0, lParam(ListValue));
      Modified:=true;
      with TrStringGridEd(Grid) do
      begin
        SetEditText(Col, Row, ListValue);
        if HideList and not(goAlwaysShowEditor in Options) then
          HideEditor;
        DoFinishEditing(Col, Row, false);
      end;
      if not HideList then
        FPickList.Invalidate;
    end;
  end;

  if FPopupCalendar.IsShown then
  begin
    if GetCapture <> 0 then
      SendMessage(GetCapture, WM_CancelMode, 0, 0);
    if Accept and EditCanModify then
      PopupCallbackSelectDate(nil);
    FPopupCalendar.HideCalendar;
  end;
  // hide do not work for Picker, wait for CloseUp event
  { if FDatePickerVisible then
    begin
    Grid.SetFocus;
    FDatePicker.Hide;
    FDatePickerVisible:=false;
    end;
    }
end;

procedure TrStringGridEdInplaceEdit.KeyPress(var Key: Char);
var
  Sep: TSysCharSet;
begin
  // press editor key during DatePicker visibility => hide picker and go back to editor
  if (Key <> #0) and GetDateCalendarVisible then
    CreateHideCalendar;
  TrStringGridEd(Grid).EditorKeyPress(Key);
  if (Key >= #32) and (EditStyle in [sgbDropDownList, sgbDropDownColor, sgbDropDownCheckListOnly]) and not FListVisible then
  begin
    DropDown;
    FPickList.FSearchTickCount:=0;
    FPickList.KeyPress(Key);
  end;
  if (Key >= #32) and (FDataType in [sgdDate, sgdDateTime]) and not GetDateCalendarVisible then
  begin
    // shortcut for date
    case Key of
      'T', 't', '*':
        begin
          if FDataType = sgdDate then
            SetDateTime(Date)
          else
            SetDateTime(Now);
          Key:=#0;
        end;
      'S', 's':
        begin
          SetDateTime(DateUtils.StartOfTheMonth(Date));
          Key:=#0;
        end;
      'E', 'e':
        begin
          SetDateTime(DateUtils.EndOfTheMonth(Date));
          Key:=#0;
        end;
    end;
    Sep:=['.', ',', '/', '-'];
    if CharInSet(Key, Sep) then
      Key:=GetDateSeparator;
  end;
  if Key <> #0 then
    inherited KeyPress(Key);
end;

procedure TrStringGridEdInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);

  procedure SendToParent;
  begin
    TrStringGridEd(Grid).KeyDown(Key, Shift);
    Key:=0;
  end;

  function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, longint(@Result.StartPos), longint(@Result.EndPos));
  end;

  function CaretPos: integer;
  var
    p: TPoint;
  begin
    Windows.GetCaretPos(p);
    Result:=SendMessage(Handle, EM_CHARFROMPOS, 0, MakeLong(p.X, p.Y));
  end;

  function RightSide: boolean;
  begin
    with Selection do
      Result:=(CaretPos = GetTextLen) and ((StartPos = 0) or (EndPos = StartPos)) and (EndPos = GetTextLen);
  end;

  function LeftSide: boolean;
  begin
    with Selection do
      Result:=(CaretPos = 0) and (StartPos = 0) and ((EndPos = 0) or (EndPos = GetTextLen));
  end;

begin
  if FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYDOWN, Key, 0);
    Key:=0;
  end;

  // quick open/close dropdown for F4
  if (Key = VK_F4) and (Shift = []) and (EditStyle in cesShowBtn) then
  begin
    if EditStyle in [sgbEllipsis, sgbEllipsisOnly] then
      TrStringGridEd(Grid).EllipsisClick
    else
      if FListVisible then
        CloseUp(false, true)
      else
        if GetDateCalendarVisible then
          CreateHideCalendar
        else
          DropDown;
    Key:=0;
  end;

  if (EditStyle in [sgbEllipsis, sgbEllipsisOnly]) and (Key = VK_DOWN) and (Shift = [ssAlt]) then
    TrStringGridEd(Grid).EllipsisClick;
  case Key of
    VK_LEFT:
      if (dgEditArrowKeyMoveCell in TrStringGridEd(Grid).OptionsEx) and LeftSide then
        SendToParent;
    VK_RIGHT:
      if (dgEditArrowKeyMoveCell in TrStringGridEd(Grid).OptionsEx) and RightSide then
        SendToParent;
  end;
  if Key <> 0 then
    inherited KeyDown(Key, Shift);
end;

procedure TrStringGridEdInplaceEdit.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYUP, Key, 0);
    Key:=0;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrStringGridEdInplaceEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  if (Button = mbLeft) then
  begin
    if (FEditStyle in cesShowBtn) and PtInRect(Rect(Width - FActBtnWidth, 0, Width, Height), Point(X, Y)) then
    begin
      if FListVisible or GetDateCalendarVisible then
        CloseUp(false, true)
      else
      begin
        MouseCapture:=true;
        FTracking:=true;
        TrackButton(X, Y);
        if Assigned(FActiveList) then
          DropDown;
      end;
    end;
  end;
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TrStringGridEdInplaceEdit.MouseMove(Shift: TShiftState; X, Y: integer);
var
  ListPos: TPoint;
  MousePos: TSmallPoint;
begin
  if FTracking then
  begin
    TrackButton(X, Y);
    if FListVisible then
    begin
      ListPos:=FActiveList.ScreenToClient(ClientToScreen(Point(X, Y)));
      if PtInRect(FActiveList.ClientRect, ListPos) then
      begin
        StopTracking;
        MousePos:=PointToSmallPoint(ListPos);
        SendMessage(FActiveList.Handle, WM_LBUTTONDOWN, 0, integer(MousePos));
        Exit;
      end;
    end;
  end;
  inherited MouseMove(Shift, X, Y);
end;

procedure TrStringGridEdInplaceEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  WasPressed: boolean;
begin
  WasPressed:=FPressed;
  StopTracking;
  if (Button = mbLeft) and (FEditStyle in [sgbEllipsis, sgbEllipsisOnly]) and WasPressed then
    TrStringGridEd(Grid).EllipsisClick;
  inherited MouseUp(Button, Shift, X, Y);
end;

procedure TrStringGridEdInplaceEdit.StopTracking;
begin
  if FTracking then
  begin
    TrackButton(-1, -1);
    FTracking:=false;
    MouseCapture:=false;
  end;
end;

procedure TrStringGridEdInplaceEdit.TrackButton(X, Y: integer);
var
  NewState: boolean;
  R: TRect;
begin
  SetRect(R, ClientWidth - FActBtnWidth, 0, ClientWidth, ClientHeight);
  NewState:=PtInRect(R, Point(X, Y));
  if FPressed <> NewState then
  begin
    FPressed:=NewState;
    InvalidateRect(Handle, @R, false);
  end;
end;

procedure TrStringGridEdInplaceEdit.SetEditStyle(Value: TrStringGridEdEditStyle);
begin
  if Value = FEditStyle then
    Exit;
  FEditStyle:=Value;
  if FEditStyle in cesDropDown then
  begin
    if not Assigned(FPickList) then
    begin
      FPickList:=TPopupListbox.Create(self);
      FPickList.Visible:=false;
      FPickList.Parent:=self;
      FPickList.IntegralHeight:=false;
      FPickList.ItemHeight:=11;
    end;
    FPickList.IsColor:=FEditStyle = sgbDropDownColor;
    FPickList.IsCheckList:=FEditStyle in [sgbDropDownCheckList, sgbDropDownCheckListOnly];
    FPickList.IsDataImage:=FEditStyle = sgbDropDownDataImage;
    FActiveList:=FPickList;
  end
  else
    if FEditStyle in [sgbDateTimePicker] then
    begin
      CreateHideCalendar;
    end
    else
      FActiveList:=nil;

  if FEditStyle = sgbRichText then
  begin
    if not Assigned(FRichEditor) then
    begin
      FRichEditor:=TrRichEdit.Create(Grid);
      FRichEditor.PlainText:=false;
      FRichEditor.ScrollBars:=ssVertical;
      FRichEditor.BorderStyle:=bsNone;
      FRichEditor.OnKeyDown:=RichEditKeyDown;
      if Assigned(FRichEditor.PopupMenu) then
      begin
        // Insert edit popup menus
        FRichEditor.PopupMenu.Items.Insert(0, rtool.CreateMenuItem(FRichEditor.PopupMenu, sConfirmChanges, TextToShortCut('Ctrl+Enter'), 1001,
            RichEditPopupMenu));
        FRichEditor.PopupMenu.Items.Insert(1, rtool.CreateMenuItem(FRichEditor.PopupMenu, sCancelChanges + ' (Esc)', 1002, RichEditPopupMenu));
        FRichEditor.PopupMenu.Items.Insert(2, rtool.CreateMenuItem(FRichEditor.PopupMenu, '-', 0, RichEditPopupMenu));
      end;
      FRichEditor.SetSubComponent(true);
      FRichEditor.Visible:=false;
      FRichEditor.Parent:=Grid;
      FRichEditor.BoundsRect:=Rect(0, 0, 0, 0);
    end;
    FRichEditorVisible:=false;
  end;

  if FEditStyle = sgbCustomEditor then
  begin
    if not Assigned(FCustomEditor) then
      FCustomEditor:=CreateCustomEditor;
    FCustomEditorVisible:=false;
  end;
  Repaint;
end;

procedure TrStringGridEdInplaceEdit.UpdateContents;
begin
  inherited UpdateContents;
  if Assigned(Grid) then
    with TrStringGridEd(Grid) do
    begin
      CheckColumns;
      EditStyle:=GetEditStyleEx(Col, Row);
      FNumFormat:=GetNumFormat(Col, Row);
      FDataType:=GetDataType(Col, Row);
    end;
end;

procedure TrStringGridEdInplaceEdit.PaintWindow(DC: HDC);
var
  R: TRect;
  Flags: integer;
  w, X, Y: integer;
  Details: TThemedElementDetails;
  Bmp: TBitmap;
  Gr: TrStringGridEd;
  DrawBtn: boolean;
begin
  if FEditStyle in cesShowBtn then
  begin
    FActBtnWidth:=FEllipsisBtnWidth;
    SetRect(R, Width - FActBtnWidth, 0, Width, Height);
    Flags:=0;
    if FEditStyle in cesDropDown + [sgbDateTimePicker] then
    begin
      if GetThemeEnabled then
      begin
        if FPressed then
          Details:=GetThemeServices.GetElementDetails(tcDropDownButtonPressed)
        else
          if FMouseInControl then
            Details:=GetThemeServices.GetElementDetails(tcDropDownButtonHot)
          else
            Details:=GetThemeServices.GetElementDetails(tcDropDownButtonNormal);
        GetThemeServices.DrawElement(DC, Details, R);
      end
      else
      begin
        if FPressed then
          Flags:=DFCS_FLAT or DFCS_PUSHED;
        DrawFrameControl(DC, R, DFC_SCROLL, Flags or DFCS_SCROLLCOMBOBOX);
      end;
    end;
    Bmp:=nil;
    if FEditStyle in [sgbEllipsis, sgbEllipsisOnly] then
    begin
      DrawBtn:=true;
      if Assigned(Grid) and (Grid is TrStringGridEd) then
      begin
        // get glyph and button width
        Gr:=TrStringGridEd(Grid);
        Bmp:=rGlyphs.BmpFromIndex(rGlyphs.EllipsisIndex, Gr.FScaleRatio);
        if Assigned(Gr.FOnGetEllipsisButtonFormat) then
        begin
          Gr.FOnGetEllipsisButtonFormat(Gr, Gr.Col, Gr.Row, FActBtnWidth, Bmp);
          DrawBtn:=FActBtnWidth > 0;
          if not DrawBtn then
          begin
            if Assigned(Bmp) and (Bmp.Width > 0) then
              FActBtnWidth:=Bmp.Width + 4
            else
              FActBtnWidth:=FEllipsisBtnWidth;
          end;
          SetRect(R, Width - FActBtnWidth, 0, Width, Height);
        end;
      end;
      if DrawBtn then
      begin
        // show button
        if GetThemeEnabled then
        begin
          if FPressed then
            Details:=GetThemeServices.GetElementDetails(tbPushButtonPressed)
          else
            if FMouseInControl then
              Details:=GetThemeServices.GetElementDetails(tbPushButtonHot)
            else
              Details:=GetThemeServices.GetElementDetails(tbPushButtonNormal);
          GetThemeServices.DrawElement(DC, Details, R);
        end
        else
        begin
          if FPressed then
            Flags:=BF_FLAT;
          DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_MIDDLE or Flags);
        end;
      end;
      if Assigned(Bmp) and (Bmp.Width > 0) then
      begin
        rpictool.DrawBmpCenter(DC, R, Bmp, true);
      end
      else
      begin
        X:=R.Left + ((R.Right - R.Left) shr 1) - 1 + Ord(FPressed);
        Y:=R.Top + ((R.Bottom - R.Top) shr 1) - 1 + Ord(FPressed);
        w:=FActBtnWidth shr 3;
        if w = 0 then
          w:=1;
        PatBlt(DC, X, Y, w, w, BLACKNESS);
        PatBlt(DC, X - (w * 2), Y, w, w, BLACKNESS);
        PatBlt(DC, X + (w * 2), Y, w, w, BLACKNESS);
      end;
    end;
    ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
  end;
  inherited PaintWindow(DC);
end;

procedure TrStringGridEdInplaceEdit.CMCancelMode(var Message: TCMCancelMode);
begin
  inherited;
  if (Message.Sender <> self) and (Message.Sender <> FActiveList) then
    CloseUp(false, true);
  if not Assigned(FPopupCalendar) then
    Exit;
  if not FPopupCalendar.IsShown or (Message.Sender = self) or (Message.Sender = FPopupCalendar) then
    Exit;
  FPopupCalendar.HideCalendar;
end;

procedure TrStringGridEdInplaceEdit.WMCancelMode(var Message: TMessage);
begin
  StopTracking;
  inherited;
end;

procedure TrStringGridEdInplaceEdit.WMKillFocus(var Message: TMessage);
begin
  if SysLocale.Fareast then
  begin
    ImeName:=Screen.DefaultIme;
    ImeMode:=imDontCare;
  end;
  inherited;
  CloseUp(false, true);
  if Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
end;

procedure TrStringGridEdInplaceEdit.WMLButtonDblClk(var Message: TWMLButtonDblClk);
begin
  with Message do
    if (FEditStyle in cesShowBtn) and PtInRect(Rect(Width - FActBtnWidth, 0, Width, Height), Point(XPos, YPos)) then
      Exit;
  inherited;
end;

procedure TrStringGridEdInplaceEdit.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrStringGridEdInplaceEdit.WMSetCursor(var Message: TWMSetCursor);
var
  p: TPoint;
begin
  GetCursorPos(p);
  if (FEditStyle in cesShowBtn) and PtInRect(Rect(Width - FActBtnWidth, 0, Width, Height), ScreenToClient(p)) then
    Windows.SetCursor(LoadCursor(0, idc_Arrow))
  else
    inherited;
end;

procedure TrStringGridEdInplaceEdit.WndProc(var Message: TMessage);
begin
  case Message.Msg of
    WM_KEYDOWN, WM_SYSKEYDOWN, WM_CHAR:
      if EditStyle in cesDropDown + [sgbDateTimePicker] then
        with TWMKey(Message) do
        begin
          DoDropDownKeys(CharCode, KeyDataToShiftState(KeyData));
          if (CharCode <> 0) and FListVisible then
          begin
            with TMessage(Message) do
              SendMessage(FActiveList.Handle, Msg, wParam, lParam);
            Exit;
          end;
        end
  end;
  inherited;
end;

procedure TrStringGridEdInplaceEdit.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  if (FEditStyle in cesShowBtn) and GetThemeEnabled and not FMouseInControl then
  begin
    FMouseInControl:=true;
    InvalidateRect(Handle, Rect(Width - FActBtnWidth, 0, Width, Height), false);
  end;
end;

procedure TrStringGridEdInplaceEdit.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  if (FEditStyle in cesShowBtn) and GetThemeEnabled and FMouseInControl then
  begin
    FMouseInControl:=false;
    InvalidateRect(Handle, Rect(Width - FActBtnWidth, 0, Width, Height), false);
  end;
end;

procedure TrStringGridEdInplaceEdit.WMPaste(var Message: TMessage);
var
  s: string;
begin
  inherited;
  if not EditCanModify then
    Exit;
  s:=Clipboard.AsText;
  with TrStringGridEd(Grid) do
  begin
    case GetDataType(Col, Row) of
      sgdNumber, sgdDataBar, sgdInteger, sgdStatusBar:
        s:=rtool.CleanNumericString(s);
      sgdDate, sgdTime, sgdDateTime:
        s:=rtool.CleanDateTimeString(s);
    else
      s:=rtool.CleanLastEOL(s);
    end;
    if s <> Clipboard.AsText then
    begin
      EditText:=s;
      SetCursor(Length(s));
    end;
  end;
end;

procedure TrStringGridEdInplaceEdit.WMSetFocus(var Message: TWMSetFocus);
var
  R: TRect;
  M: integer;
begin
  inherited;
  R:=BoundsRect;
  M:=TrStringGridEd(Grid).InnerMargin;
  InflateRect(R, -M, -1);
  // show rich editor instead of inplace
  if (EditStyle = sgbRichText) then
    ShowRichEditor(R);
  // show custom editor instead of inplace
  if (EditStyle = sgbCustomEditor) then
    ShowCustomEditor(R);
end;

// =============================================================================
procedure TCustomGridHelper.InvalidateRectEx(ARect: TGridRect);
begin
  with self do
    InvalidateRect(ARect);
end;

// TrStringGridEd
constructor TrStringGridEd.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FScaleRatio:=1;
  FSkipCheckDropDownList:=false;
  FTitleRowHeight:=0;
  FIniKeyColumn:=0;
  FIniValueColumn:=1;
  FInnerMargin:=0;
  FBlockClick:=false;
  Options:=[goFixedVertLine, goFixedHorzLine, goVertLine, goHorzLine, goColSizing];
  FOptionsEx:=[];
  FChangedCellsBkColor:=clInfoBk;
  FSelectedCellBkColor:=clNone;
  FInRangeCellsBkColor:=clNone;
  DefaultDrawing:=false;
  DefaultRowHeight:=17;
  DefaultColWidth:=90;
  FCellChanged:=-1;
  FChangedRows:=TStringList.Create;
  FChangedCells:=TStringList.Create;
  FIniGroupRows:=TStringList.Create;
  FRichTextTopMargin:=1;
  // Create columns
  Columns:=TrStringGridEdColumns.Create(self, TrStringGridEdColumn);
  CheckColumns;
  LastSortColIndex:=-1;
  LastSortDescend:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(self, rDBComps_ComponentCreatedEventID);
  if not(csDesigning in ComponentState) then
  begin
    FRichText:=TrRichEdit.Create(self);
    FRichText.Visible:=false;
    FRichText.Parent:=self;
    FRichText.BorderStyle:=bsNone;
  end;
end;

destructor TrStringGridEd.Destroy;
begin
  FUpdateLock:=1;
  FreeAndNil(FChangedRows);
  FreeAndNil(FChangedCells);
  FreeAndNil(FIniGroupRows);
  FreeAndNil(FColumns);
  inherited Destroy;
end;

procedure TrStringGridEd.Loaded;
var
  i: integer;
begin
  // block loading of ColWidths, use Columns.Width instead
  FColWidthUpdating:=true;
  try
    inherited Loaded;
    // load col width
    for i:=0 to Columns.Count - 1 do
      if (i < ColCount) and (Columns[i].FWidth > 0) then
        ColWidths[i]:=Columns[i].FWidth;
  finally
    FColWidthUpdating:=false;
  end;
  if not(csDesigning in ComponentState) then
  begin
    // create editor to be able to catch keypress (for checkbox)
    ResetSelRange;
    FirstInvalidateEditor:=true;
    InvalidateEditor;
    FirstInvalidateEditor:=false;
    // SelectCell(Col, Row);
    if AutoRowHeight then
      Options:=Options - [goRowSizing];
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(self, rDBComps_ComponentLoadedEventID);
end;

procedure TrStringGridEd.SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer);
var
  i, w: integer;
begin
  inherited;
  if (dgResizeLastColumn in FOptionsEx) and Assigned(Columns) and (Columns.Count = ColCount) then
  begin
    w:=0;
    for i:=0 to ColCount - 2 do
      Inc(w, Columns[i].Width);
    w:=self.ClientWidth - w - 8;
    if w > 30 then
      Columns[ColCount - 1].Width:=w;
  end;
end;

procedure TrStringGridEd.Notification(AComponent: TComponent; Operation: TOperation);
var
  c: integer;
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) then
  begin
    if (AComponent = FCustomGlyphs) then
    begin
      FCustomGlyphs:=nil;
      Invalidate;
    end;
    if (csDestroying in ComponentState) then
      Exit;
    // check columns
    for c:=0 to Columns.Count - 1 do
      if (AComponent = Columns[c].FDataImageList) then
      begin
        Columns[c].FDataImageList:=nil;
        Invalidate;
      end;
  end;
end;

procedure TrStringGridEd.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
var
  i: integer;
begin
  if not DoGeneralEvent(geChangeScaleBefore) then
    Exit;
  FScaleRatio:=FScaleRatio * M / D;
  // scale all col widths, block ColWidthsChanged event
  FColWidthUpdating:=true;
  try
    inherited;
    for i:=0 to Columns.Count - 1 do
      if (i < ColCount) and (Columns[i].FWidth > 0) then
        Columns[i].Width:=MulDiv(Columns[i].FWidth, M, D);
  finally
    FColWidthUpdating:=false;
  end;
  TitleRowHeight:=MulDiv(TitleRowHeight, M, D);
  // update font for each column
  for i:=0 to Columns.Count - 1 do
  begin
    if not Columns[i].FParentFont then
      rtool.DPI_ScaleFontHeight(Columns[i].Font, M, D);
    if not Columns[i].FTitleParentFont then
      rtool.DPI_ScaleFontHeight(Columns[i].TitleFont, M, D);
  end;
  if not DoGeneralEvent(geChangeScaleAfter) then
    Exit;
end;

procedure TrStringGridEd.SetCustomGlyphs(const Value: TrCustomGlyphs);
begin
  FCustomGlyphs:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
  Invalidate;
end;

procedure TrStringGridEd.SetTitleRowHeight(const Value: integer);
begin
  FTitleRowHeight:=Value;
  if (FTitleRowHeight > 0) then
    RowHeights[0]:=FTitleRowHeight
  else
    RowHeights[0]:=DefaultRowHeight;
end;

procedure TrStringGridEd.SetAutoRowHeight(Value: boolean);
begin
  FAutoRowHeight:=Value;
  if FAutoRowHeight then
    Options:=Options - [goRowSizing];
  Invalidate;
end;

procedure TrStringGridEd.SetInnerMargin(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FInnerMargin:=Value;
  Invalidate;
end;

function TrStringGridEd.GetChangedCellStr(ACol, ARow: longint): string;
begin
  Result:=IntToStr(ACol) + ';' + IntToStr(ARow);
end;

function TrStringGridEd.IsCellChanged(ACol, ARow: longint): boolean;
begin
  Result:=FChangedCells.IndexOf(GetChangedCellStr(ACol, ARow)) >= 0;
end;

procedure TrStringGridEd.SetOptionsEx(const Value: TOptionsEx);
begin
  FOptionsEx:=Value;
  Invalidate;
end;

procedure TrStringGridEd.SetAutoAddNewRow(const Value: boolean);
begin
  FAutoAddNewRow:=Value;
  if not(csLoading in ComponentState) and FAutoAddNewRow then
    CheckLastRows;
end;

procedure TrStringGridEd.SetChangedCellsBkColor(const Value: TColor);
begin
  FChangedCellsBkColor:=Value;
  Invalidate;
end;

procedure TrStringGridEd.SetSelectedCellBkColor(const Value: TColor);
begin
  FSelectedCellBkColor:=Value;
  Invalidate;
end;

procedure TrStringGridEd.SetInRangeCellsBkColor(const Value: TColor);
begin
  FInRangeCellsBkColor:=Value;
  Invalidate;
end;

procedure TrStringGridEd.SetRichTextTopMargin(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FRichTextTopMargin:=Value;
  Invalidate;
end;

procedure TrStringGridEd.CMFontChanged(var Message: TMessage);
var
  i: longint;
begin
  inherited;
  for i:=0 to Columns.Count - 1 do
  begin
    if Columns[i].FParentFont then
    begin
      Columns[i].FFont.Assign(Font);
      Columns[i].FParentFont:=true;
    end;
    if Columns[i].FTitleParentFont then
    begin
      Columns[i].FTitleFont.Assign(Font);
      Columns[i].FTitleParentFont:=true;
    end;
  end;
  SetDefaultRowHeight;
end;

function TrStringGridEd.GetEditStyleEx(ACol, ARow: longint): TrStringGridEdEditStyle;
begin
  Result:=Columns[ACol].ButtonStyle;
  if Assigned(FOnGetEditStyle) then
    FOnGetEditStyle(self, ACol, ARow, Result);
end;

function TrStringGridEd.GetNumFormat(ACol, ARow: longint): string;
begin
  Result:=Columns[ACol].NumFormat;
  if Assigned(FOnGetNumFormat) then
    FOnGetNumFormat(self, ACol, ARow, Result);
end;

function TrStringGridEd.GetCoordPosition(X, Y: integer): TCellPosition;
begin
  if (X < 0) or (Y < 0) or (X >= ColCount) or (Y >= RowCount) then
    Result:=cpInvalid
  else
    if (Y < FixedRows) and (X < FixedCols) then
      Result:=cpFixedColRow
    else
      if (Y < FixedRows) then
        Result:=cpFixedRow
      else
        if (X < FixedCols) then
          Result:=cpFixedCol
        else
          Result:=cpNotFixed;
end;

function TrStringGridEd.GetCoordPosition(Coord: TGridCoord): TCellPosition;
begin
  Result:=GetCoordPosition(Coord.X, Coord.Y);
end;

procedure TrStringGridEd.CheckColumns;
begin
  // add or delete columns according ColCount
  if GetColCount <> Columns.Count then
  begin
    while GetColCount > Columns.Count do
      Columns.Add;
    while GetColCount < Columns.Count do
      Columns.Delete(Columns.Count - 1);
  end;
end;

procedure TrStringGridEd.EllipsisClick;
begin
  if Assigned(OnEllipsisClicked) then
    OnEllipsisClicked(self, Col, Row);
end;

procedure TrStringGridEd.ButtonClick;
var
  btnEnabled: boolean;
begin
  btnEnabled:=Enabled;
  if btnEnabled and Assigned(FOnCanEdit) then
    FOnCanEdit(self, Col, Row, btnEnabled);
  if btnEnabled and Assigned(OnEllipsisClicked) then
    OnEllipsisClicked(self, Col, Row);
end;

function TrStringGridEd.GetColCount: longint;
begin
  Result:= inherited ColCount;
end;

procedure TrStringGridEd.SetColCount(Col: longint);
begin
  inherited ColCount:=Col;
  CheckColumns;
end;

function TrStringGridEd.CreateEditor: TInplaceEdit;
begin
  FStartEditKey:=0;
  Result:=TrStringGridEdInplaceEdit.Create(self);
  if Assigned(FOnEditorEvent) then
    FOnEditorEvent(Result, Col, Row, 'EditorCreated');
end;

procedure TrStringGridEd.HideEditorEx;
begin
  if Assigned(InplaceEditor) then
  begin
    InplaceEditorEx.HideRichEditor(true);
    InplaceEditorEx.HideCustomEditor(true);
  end;
  HideEditor;
end;

function TrStringGridEd.InplaceEditorEx: TrStringGridEdInplaceEdit;
begin
  if Assigned(InplaceEditor) then
    Result:=InplaceEditor as TrStringGridEdInplaceEdit
  else
    Result:=nil;
end;

procedure TrStringGridEd.ColWidthsChanged;
var
  i: integer;
begin
  inherited;
  // update Columns.Width property
  if Assigned(Columns) and not FColWidthUpdating then
    for i:=0 to Columns.Count - 1 do
      if (i < ColCount) and Columns[i].Visible then
        Columns[i].FWidth:=ColWidths[i];
end;

procedure TrStringGridEd.RowHeightsChanged;
begin
  inherited;
  SetTitleRowHeight(FTitleRowHeight);
end;

function TrStringGridEd.CanEditModify: boolean;
begin
  CheckColumns;
  Result:= inherited CanEditModify;
  // disable editing for ReadOnly Column or DropDownList
  if Columns[Col].ReadOnly then
    Result:=false;
  if (not FSkipCheckDropDownList) and ((GetEditStyleEx(Col, Row) in cesReadOnly)) then
    Result:=false;
  if IsIniGroupRow(Row) then
    Result:=false;
  if Result and Assigned(FOnCanEdit) then
    FOnCanEdit(self, Col, Row, Result);
end;

function TrStringGridEd.CanEditModifyEx(AllowDropDownList: boolean): boolean;
begin
  if AllowDropDownList then
    FSkipCheckDropDownList:=true;
  try
    Result:=CanEditModify and ([goRowSelect, goEditing] * Options = [goEditing]);
  finally
    FSkipCheckDropDownList:=false;
  end;
end;

function TrStringGridEd.CanEditShow: boolean;
begin
  if FirstInvalidateEditor then
  begin
    // create editor during Loading
    FirstInvalidateEditor:=false;
    Result:=true;
    Exit;
  end;
  Result:= inherited CanEditShow;
  // disable editor for checkbox and readonly columns
  if Assigned(Columns) and (Col < Columns.Count) then
    if (Columns[Col].ReadOnly) or (GetEditStyleEx(Col, Row) in [sgbCheckbox, sgbButton]) then
      Result:=false;
  if IsIniGroupRow(Row) then
    Result:=false;
  // call user defined event
  if Result and Assigned(FOnCanEdit) then
    FOnCanEdit(self, Col, Row, Result);
  // hide editor if edit is disable - MUST!!! in other case active cell has strange content
  if not Result then
    HideEditor;
end;

function TrStringGridEd.CanEditCell(ACol, ARow: longint): boolean;
begin
  Result:=true;
  if Assigned(Columns) and (ACol < Columns.Count) then
    if (Columns[ACol].ReadOnly) or (GetEditStyleEx(ACol, ARow) = sgbCheckbox) then
      Result:=false;
  if Result and Assigned(FOnCanEdit) then
    FOnCanEdit(self, ACol, ARow, Result);
end;

procedure TrStringGridEd.CMHintShow(var Message: TCMHintShow);
var
  s, hnt: string;
  EdStyle: TrStringGridEdEditStyle;
begin
  if GetCoordPosition(MMoveCoord) <> cpInvalid then
    try
      s:=GetCellText(MMoveCoord.X, MMoveCoord.Y);
      hnt:=s;
      if Assigned(FGetHintText) then
        FGetHintText(self, MMoveCoord.X, MMoveCoord.Y, hnt);
      // if no custom hint - show only if truncated = text height > current row height
      if (hnt = s) then
      begin
        if CalcCellHeight(MMoveCoord.X, MMoveCoord.Y, false) - 4 <= RowHeights[MMoveCoord.Y] then
          hnt:=''
        else
        begin
          // get style
          EdStyle:=GetEditStyleEx(MMoveCoord.X, MMoveCoord.Y);
          // modify hint for rich text
          if Assigned(FRichText) and (EdStyle = sgbRichText) then
          begin
            if Assigned(RosiCompGlobalCfg.RosiCompConfig.Global_HTMLHintComponent) then
              // rHTMLHint is used = add RTF tag
              hnt:=hnt + RosiCompGlobalCfg.rHTMLHintRichTextTag
            else
            begin
              // use plain text only
              FRichText.RtfText:=hnt;
              hnt:=FRichText.Lines.Text;
            end;
          end;
        end;
      end;
      Message.HintInfo^.HintStr:=hnt;
    except
      // skip exception for hint
    end;
  inherited;
end;

procedure TrStringGridEd.FixedCellClick(ACol, ARow: longint);
var
  CellPos: TCellPosition;
begin
  CellPos:=GetCoordPosition(ACol, ARow);
  // entirecolumns or entirerow select
  if (dgSelectRange in FOptionsEx) and not FBlockClick then
  begin
    InvalidateSelectedRange;
    SelRangeStart.X:=ACol;
    SelRangeStart.Y:=ARow;
    SelRangeStop:=SelRangeStart;
    case CellPos of
      cpFixedColRow:
        begin
          Row:=FixedRows;
          Col:=FixedCols;
          SelRangeStart.X:=FixedCols;
          SelRangeStop.X:=ColCount - 1;
          SelRangeStart.Y:=FixedRows;
          SelRangeStop.Y:=RowCount - 1;
          Invalidate;
        end;
      cpFixedRow:
        begin
          Row:=FixedRows;
          Col:=ACol;
          SelRangeStart.Y:=FixedRows;
          SelRangeStop.Y:=RowCount - 1;
          InvalidateCol(ACol);
        end;
      cpFixedCol:
        begin
          Row:=ARow;
          Col:=FixedCols;
          SelRangeStart.X:=FixedCols;
          SelRangeStop.X:=ColCount - 1;
          InvalidateRow(ARow);
        end;
    end;
  end;
  if (dgSortAfterTitleClick in FOptionsEx) and (ARow < FixedRows) and ((ACol >= FixedCols) or (dgSortFixedCol in FOptionsEx)) then
  begin
    if LastSortColIndex = ACol then
      LastSortDescend:=not LastSortDescend
    else
      LastSortDescend:=false;
    LastSortColIndex:=ACol;
    SortData(ACol, LastSortDescend, false);
  end;
  inherited;
end;

procedure TrStringGridEd.CellClick(ACol, ARow: longint);
begin
{$IFDEF DEMO}
  ShowDemoMsg;
  ResetSelRange;
{$ENDIF}
  if Assigned(FOnCellClick) then
    FOnCellClick(self, ACol, ARow);
end;

procedure TrStringGridEd.DblClick;
begin
  inherited;
end;

procedure TrStringGridEd.WMTimer(var Msg: TWMTimer);
begin
  if Msg.TimerID = cIntTimerID then
  begin
    KillTimer(Handle, cIntTimerID);
    EllipsisClick;
  end
  else
    inherited;
end;

procedure TrStringGridEd.WMHScroll(var Msg: TWMVScroll);
begin
  MDownCellPoint:=Point(0, 0);
  inherited;
end;

procedure TrStringGridEd.WMVScroll(var Msg: TWMVScroll);
begin
  MDownCellPoint:=Point(0, 0);
  inherited;
end;

procedure TrStringGridEd.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  R: TRect;
  EdStyle: TrStringGridEdEditStyle;
  c: integer;
  NewCellSelected: boolean;
begin
  MDownCoord:=MouseCoord(X, Y);
  MDownCellPoint:=Point(X, Y);
  NewCellSelected:=(MDownCoord.X <> Col) or (MDownCoord.Y <> Row);
  // finish editing if click to another cell including fixed cell (BUG in standard Grid)
  if NewCellSelected then
    HideEditorEx;

  // proc right-click to select cell before context menu
  if (Button = mbRight) and (MDownCoord.X >= 0) and (MDownCoord.Y >= 0) then
  begin
    if (MDownCoord.X >= FixedCols) and (MDownCoord.Y >= FixedRows) then
      // select cell (simulate click by left button)
      inherited MouseDown(mbLeft, Shift, X, Y)
    else
      // call FixedCellClick to select range on right click (before popup menu)
      if (GetHotTrackCoord(FHotTrackCell).X <> -1) and (GetHotTrackCoord(FHotTrackCell).Y <> -1) and
        (((goFixedColClick in Options) and (MDownCoord.X < FixedCols)) or ((goFixedRowClick in Options) and (MDownCoord.Y < FixedRows))) then
        FixedCellClick(MDownCoord.X, MDownCoord.Y);
  end;

  inherited MouseDown(Button, Shift, X, Y);
  // do not read mouse position again - could be auto scrolled
  // DO NOT MDownCoord:=MouseCoord(X, Y);

  // ensure all columns are  created
  CheckColumns;

  if (ssLeft in Shift) and (GetCoordPosition(MDownCoord) = cpNotFixed) then
  begin
    CellClick(MDownCoord.X, MDownCoord.Y);
    // check again limits as number of cols and rows can be changed in the CellClick event
    if (GetCoordPosition(MDownCoord) = cpNotFixed) and CanEditModifyEx(true) then
    begin
      EdStyle:=GetEditStyleEx(MDownCoord.X, MDownCoord.Y);
      // try checkbox
      if (EdStyle = sgbCheckbox) and (Columns[MDownCoord.X].FCheckBoxDblClick = (ssDouble in Shift)) then
      begin
        // Actual Cell Rect - reduce to checkbox size (11x11)
        R:=CellRect(MDownCoord.X, MDownCoord.Y);
        case GetCheckboxAlignment(MDownCoord.X, MDownCoord.Y) of
          taLeftJustify:
            R.Right:=R.Left + FCheckBoxWidth + 4;
          taRightJustify:
            R.Left:=R.Right - FCheckBoxWidth - 4;
        else
          while R.Right - R.Left > FCheckBoxWidth + 4 do
            InflateRect(R, -1, 0);
        end;
        while R.Bottom - R.Top > FCheckBoxWidth + 4 do
          InflateRect(R, 0, -1);
        if PtInRect(R, Point(X, Y)) then
          ChangeCheckbox;
      end;

      if (ssDouble in Shift) then
      begin
        ShowEditor;
        if (EdStyle in cesDropDown + [sgbDateTimePicker]) then
          InplaceEditorEx.DropDown;
        // use timer to delay ellipsis click after grid MouseUp event,
        // in other case modal dialog block it and grid would stay in strange state
        if EdStyle in [sgbEllipsis, sgbEllipsisOnly] then
          DoEllipsisClick;
      end;
    end;

    // range select
    if (dgSelectRange in FOptionsEx) and NewCellSelected and not FBlockClick then
    begin
      InvalidateSelectedRange;
      // Shift means extend selection = do not change SelRangeStart
      if not(ssShift in Shift) then
        SelRangeStart:=MDownCoord;
      SelRangeStop:=MDownCoord;
      InvalidateSelectedRange;
    end;

    // drag and drop support
    if (dgDragCell in FOptionsEx) then
      BeginDrag(false);
  end;

  // column autofit after dbl click on column border
  if (ssDouble in Shift) and (GetCoordPosition(MDownCoord) = cpFixedRow) then
  begin
    c:=-1;
    R:=CellRect(MDownCoord.X, MDownCoord.Y);
    if X >= R.Right - 3 then
      c:=MDownCoord.X;
    if X <= R.Left + 3 then
      c:=MDownCoord.X - 1;
    if c >= 0 then
    begin
      HideEditorEx;
      SetColumnAutoFit(c);
      FBlockClick:=true;
    end;
  end;
end;

procedure TrStringGridEd.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  ResetRng: boolean;
begin
  ResetRng:=FBlockClick or (FGridState in [ { gsRowSizing, gsColSizing, } gsRowMoving, gsColMoving]);
  inherited MouseUp(Button, Shift, X, Y);
  // drag and drop support
  if (dgDragCell in FOptionsEx) and Dragging then
    EndDrag(false);
  FBlockClick:=false;
  if ResetRng then
    ResetSelRange;
  // click to cell Button
  if (GetCoordPosition(MDownCoord) = cpNotFixed) and (MMoveCoord.X = MDownCoord.X) and (MMoveCoord.Y = MDownCoord.Y) then
    if GetEditStyleEx(MDownCoord.X, MDownCoord.Y) = sgbButton then
    begin
      ButtonClick;
      InvalidateCell(MDownCoord.X, MDownCoord.Y);
    end;
  MDownCoord.X:=-1;
  MDownCoord.Y:=-1;
end;

procedure TrStringGridEd.MouseMove(Shift: TShiftState; X, Y: integer);
var
  NewCoord: TGridCoord;
  Reshow: boolean;
begin
  NewCoord:=MouseCoord(X, Y);
  Reshow:=(NewCoord.X <> MMoveCoord.X) or (NewCoord.Y <> MMoveCoord.Y);
  // invalide previous undermouse cell
  if Reshow then
    InvalidateCell(MMoveCoord.X, MMoveCoord.Y);
  MMoveCoord:=NewCoord;
  // invalide new undermouse cell
  if Reshow then
    InvalidateCell(MMoveCoord.X, MMoveCoord.Y);
  // Activate hint if mouse enter new cell
  if Reshow then
    Application.ActivateHint(Mouse.CursorPos);
  inherited MouseMove(Shift, X, Y);
  // finish editing for Sizing or Moving (BUG in standard Grid)
  if (FGridState in [gsRowSizing, gsColSizing, gsRowMoving, gsColMoving]) then
  begin
    HideEditorEx;
    Exit;
  end;
  // update range selected cells
  if Reshow and (dgSelectRange in FOptionsEx) and (ssLeft in Shift) and (GetCoordPosition(NewCoord) = cpNotFixed) then
  begin
    // click during DatePicker do not select new cell - do not set range
    if not EditorMode and not InplaceEditorEx.GetDateCalendarVisible and not FBlockClick then
    begin
      InvalidateSelectedRange;
      SelRangeStop:=NewCoord;
      InvalidateSelectedRange;
    end;
  end;
end;

function TrStringGridEd.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result:=false;
  // default scroll for editors
  if not(csDesigning in ComponentState) then
    if InplaceEditorEx.FListVisible or InplaceEditorEx.FRichEditorVisible or InplaceEditorEx.FCustomEditorVisible then
      Exit;
  Result:= inherited DoMouseWheelDown(Shift, MousePos);
end;

function TrStringGridEd.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): boolean;
begin
  Result:=false;
  // default scroll for editors
  if not(csDesigning in ComponentState) then
    if InplaceEditorEx.FListVisible or InplaceEditorEx.FRichEditorVisible or InplaceEditorEx.FCustomEditorVisible then
      Exit;
  Result:= inherited DoMouseWheelUp(Shift, MousePos);
end;

function TrStringGridEd.CheckboxIsChecked(ACol, ARow: longint): boolean;
begin
  Result:=SameText(Cells[ACol, ARow], Columns[ACol].FCheckBoxTrueValue);
end;

procedure TrStringGridEd.ChangeCheckbox;
begin
  if not CanEditModifyEx(false) then
    Exit;
  if CheckboxIsChecked(Col, Row) then
    SetEditText(Col, Row, Columns[Col].FCheckBoxFalseValue)
  else
    SetEditText(Col, Row, Columns[Col].FCheckBoxTrueValue);
  // hide editor after change
  HideEditorEx;
  DoFinishEditing(Col, Row, false);
end;

function TrStringGridEd.GetCheckboxAlignment(ACol, ARow: longint): TAlignment;
begin
  Result:=Columns[ACol].FCheckBoxAlignment;
end;

procedure TrStringGridEd.ColumnMoved(FromIndex, ToIndex: longint);
begin
  inherited ColumnMoved(FromIndex, ToIndex);
  // move Columns props accroding new columns
  Columns[FromIndex].Index:=ToIndex;
end;

procedure TrStringGridEd.Paint;
begin
  inherited;
  // draw backgroung for edited cell
  // BUG - in standard Grid is not drawn and causes black background and flickering
  if EditorMode then
    DrawCellBackground(CellRect(Col, Row), Color, [], Col, Row);
end;

function TrStringGridEd.GetTextFormatSetting(WordWrap: boolean): integer;
const
  Format1 = DT_VCENTER or DT_SINGLELINE or DT_NOPREFIX or DT_END_ELLIPSIS;
  Format2 = DT_VCENTER or DT_NOPREFIX or DT_WORDBREAK;
begin
  if WordWrap then
    Result:=Format2
  else
    Result:=Format1;
end;

function TrStringGridEd.GetCellWordWrap(ACol, ARow: longint): boolean;
begin
  Result:=((ARow < FixedRows) and Columns[ACol].TitleWordWrap) or ((ARow >= FixedRows) and Columns[ACol].WordWrap);
end;

procedure TrStringGridEd.SetCanvasStyle(ACol, ARow: longint; ACanvas: TCanvas);
begin
  ACanvas.Font.Color:=Font.Color;
  if ACol < Columns.Count then
  begin
    ACanvas.Brush.Color:=Columns[ACol].Color;
    if (ACol < FixedCols) and (Columns[ACol].Color = Color) then
      ACanvas.Brush.Color:=FixedColor;
    if ARow < FixedRows then
    begin
      if Columns[ACol].TitleColor = clNone then
        ACanvas.Brush.Color:=FixedColor
      else
        ACanvas.Brush.Color:=Columns[ACol].TitleColor;
    end;
    if ARow < FixedRows then
    begin
      ACanvas.Font.Assign(Columns[ACol].TitleFont);
      if dgShowTitleBold in FOptionsEx then
        ACanvas.Font.Style:=ACanvas.Font.Style + [fsBold];
    end
    else
      ACanvas.Font.Assign(Columns[ACol].Font);
  end;
end;

procedure TrStringGridEd.SetCellStyle(ACol, ARow: longint; ACanvas: TCanvas; var AState: TGridDrawState; var Highlighted: boolean;
  var DataImageDef: TDataImageDef);
var
  AColor: TColor;
begin
  SetCanvasStyle(ACol, ARow, ACanvas);
  AColor:=ACanvas.Brush.Color;
  // aColor:=Columns[ACol].Color;
  if (dgColorChangedCells in FOptionsEx) and IsCellChanged(ACol, ARow) then
    AColor:=FChangedCellsBkColor;
  if not Enabled then
    ACanvas.Font.Color:=clGrayText;
  if IsIniGroupRow(ARow) then
  begin
    Include(AState, gdFixed);
    ACanvas.Font.Style:=ACanvas.Font.Style + [fsBold];
  end;
  GetCellParams(ACol, ARow, ACanvas.Font, AColor, Highlighted, DataImageDef);
  ACanvas.Brush.Color:=AColor;
  if not Highlighted then
  begin
    // check data types
    if Columns[ACol].InvalidDataAction <> sgaNone then
      case IsValidData(ACol, ARow) of
        sdsOutOfLimit:
          ACanvas.Font.Color:=Columns[ACol].OutOfLimitDataColor;
        sdsInvalid:
          ACanvas.Font.Color:=Columns[ACol].InvalidDataColor;
      end;
  end;
end;

procedure TrStringGridEd.GetCellParams(ACol, ARow: longint; AFont: TFont; var Background: TColor; var Highlight: boolean; var DataImageDef: TDataImageDef);
begin
  if Assigned(FOnGetCellParams) then
    FOnGetCellParams(self, ACol, ARow, AFont, Background, Highlight);
  if Assigned(FOnGetCellParamsEx) then
    FOnGetCellParamsEx(self, ACol, ARow, AFont, Background, Highlight, DataImageDef);
end;

function TrStringGridEd.GetDataType(ACol, ARow: longint): TrStringGridEdDataType;
begin
  Result:=Columns[ACol].DataType;
  if Assigned(FOnGetDataType) then
    FOnGetDataType(self, ACol, ARow, Result);
end;

function TrStringGridEd.GetDataAlignment(ACol, ARow: longint): TAlignment;
begin
  Result:=Columns[ACol].Alignment;
  if Columns[ACol].FAutoAlignment then
  begin
    case GetDataType(ACol, ARow) of
      sgdText:
        Result:=taLeftJustify;
    else
      Result:=taRightJustify;
    end;
  end;
end;

procedure TrStringGridEd.GetCellStyle(ACol, ARow: longint; ACanvas: TCanvas);
var
  dummyHighlighted: boolean;
  dummyState: TGridDrawState;
  dummyDataImageDef: TDataImageDef;
begin
  dummyHighlighted:=false;
  dummyState:=[];
  SetCellStyle(ACol, ARow, ACanvas, dummyState, dummyHighlighted, dummyDataImageDef);
end;

function TrStringGridEd.GetCellText(ACol, ARow: longint): string;
begin
  Result:=Cells[ACol, ARow];
  if (ARow = 0) and (FixedRows > 0) then
    if Columns[ACol].TitleCaption <> '' then
      Result:=Columns[ACol].TitleCaption;
end;

procedure TrStringGridEd.SetDefaultRowHeight;
var
  H: integer;
begin
  if AutoRowHeight then
  begin
    Canvas.Font:=Font;
    H:=Canvas.TextHeight('Wg') + 2;
    if goHorzLine in Options then
      Inc(H, 2);
    DefaultRowHeight:=H;
  end;
end;

function TrStringGridEd.CalcMaxRowHeight(ARow: longint; UseLimits: boolean): integer;
var
  c: longint;
  H: integer;
begin
  Result:=DefaultRowHeight;
  for c:=0 to ColCount - 1 do
  begin
    // if Columns[c].FCustomDraw then
    // Continue;
    H:=CalcCellHeight(c, ARow, UseLimits);
    if H > Result then
      Result:=H;
    if UseLimits and (FAutoRowMaxHeight > 0) then
      if Result >= FAutoRowMaxHeight then
      begin
        Result:=FAutoRowMaxHeight;
        Exit;
      end;
  end;
end;

function TrStringGridEd.CalcCellHeight(ACol, ARow: longint; UseLimits: boolean): integer;
var
  s: string;
  R: TRect;
begin
  Result:=0;
  s:=GetCellText(ACol, ARow);
  if s = '' then
    Exit;
  SetRect(R, 0, 0, ColWidths[ACol], DefaultRowHeight);
  SetCanvasStyle(ACol, ARow, Canvas);
  // reduce width (diference between CellRect and ARect)
  R.Right:=R.Right - 4 - 2 * FInnerMargin;
  // calc height
  if Assigned(FRichText) and (GetEditStyleEx(ACol, ARow) = sgbRichText) then
  begin
    FRichText.RtfText:=s;
    FRichText.CalculateTextHeight(R);
    R.Bottom:=R.Bottom + FRichTextTopMargin;
  end
  else
    DrawText(Canvas.Handle, PChar(s), Length(s), R, GetTextFormatSetting(GetCellWordWrap(ACol, ARow)) or DT_CALCRECT);
  Result:=R.Bottom - R.Top + 2;
end;

procedure TrStringGridEd.DrawCellBackground(const ARect: TRect; AColor: TColor; AState: TGridDrawState; ACol, ARow: integer);
begin
  inherited;
  if Assigned(FOnDrawCellBackground) then
    FOnDrawCellBackground(self, ACol, ARow, ARect, AState);
end;

procedure TrStringGridEd.DrawRichText(ARect: TRect; Text: string; BkColor: TColor);
begin
  if Assigned(FRichText) then
  begin
    if BkColor <> clNone then
      FRichText.Color:=BkColor
    else
      FRichText.Color:=Canvas.Brush.Color;
    FRichText.RtfText:=Text;
    FRichText.DrawToCanvas(ARect, InnerMargin, FRichTextTopMargin, Canvas);
  end;
end;

procedure TrStringGridEd.DrawCell(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState);
var
  CellValue: string;
  H, Frm, Flags: integer;
  Alignment: TAlignment;
  EdStyle: TrStringGridEdEditStyle;
  DataType: TrStringGridEdDataType;
  myFontColor, myBrushColor, AColor, StyleFontColor, SelBkColor: TColor;
  FullR, TextR: TRect;
  Selected, Highlighted, InSelRange, FocusDrawn, IsColor, IsDataImage, ShowValue: boolean;
  Bmp: TBitmap;
  DataImageDef: TDataImageDef;
  Details: TThemedElementDetails;

  procedure DrawCellText(R: TRect; s: string);
  begin
    DrawText(Canvas.Handle, PChar(s), Length(s), R, Frm);
  end;

  procedure DrawCellBmp(R: TRect; b: TBitmap; Resize: boolean; AlignCfg: string = '');
  var
    Transparent: boolean;
  begin
    // InflateRect(R, -1, -1);
    if Resize then
    begin
      Transparent:=b.Transparent;
      rpictool.SmoothResizeBitmap(b, R.Right - R.Left, R.Bottom - R.Top, false);
      b.Transparent:=Transparent;
    end;
    if AlignCfg <> '' then
    begin
      if SameText(copy(AlignCfg, 1, 1), 'L') then
        R.Right:=R.Left + b.Width;
      if SameText(copy(AlignCfg, 1, 1), 'R') then
        R.Left:=R.Right - b.Width;
    end;
    DrawBmpCenter(Canvas, R, b);
  end;

  procedure DrawDataImage;
  var
    PicAlign: string;
  begin
    PicAlign:=DataImageDef.ImageAlignment;
    if DataImageDef.ShowText <> '' then
    begin
      // display text value instead of graphics
      if DataImageDef.ShowText = '+' then
        DataImageDef.ShowText:=GetCellText(ACol, ARow) // draw field value
      else
        if DataImageDef.ShowText[1] = '''' then
          DataImageDef.ShowText:=copy(DataImageDef.ShowText, 2, 255); // remove first '
      DrawCellText(ARect, DataImageDef.ShowText);
      // automatic alignment of picture with text
      if SameText(copy(PicAlign, 1, 1), 'A') then
      begin
        case Columns[ACol].Alignment of
          taLeftJustify:
            PicAlign:='R';
          taRightJustify, taCenter:
            PicAlign:='L';
        end;
      end;
    end;
    if Assigned(Columns[ACol].FDataImageList) and DataImageDef.UseDataImage and (DataImageDef.ImageIndex >= 0) and
      (DataImageDef.ImageIndex < Columns[ACol].FDataImageList.Count) then
    begin
      // display image
      Bmp:=TBitmap.Create;
      try
        Bmp.Canvas.Brush.Color:=Canvas.Brush.Color;
        Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
        Bmp.Transparent:=Columns[ACol].FDataImageList.tag = 0;
        Columns[ACol].FDataImageList.GetBitmap(DataImageDef.ImageIndex, Bmp);
        DrawCellBmp(ARect, Bmp, true, PicAlign);
      finally
        FreeAndNil(Bmp);
      end;
    end;
    ShowValue:=false;
  end;

  procedure DrawProgressBar(sval: string);
  var
    v, perc, vmin, vmax: double;
    cFrom, cTo: TColor;
    i: integer;
    vType: TrDataBarText;
    cfg, newcfg: string;
  begin
    vmin:=Columns[ACol].FDataBar.FMinValue;
    vmax:=Columns[ACol].FDataBar.FMaxValue;
    vType:=Columns[ACol].FDataBar.FText;
    cFrom:=Columns[ACol].FDataBar.FColorFrom;
    cTo:=Columns[ACol].FDataBar.FColorTo;

    cfg:=Format('%g,%g,%d,%s,%s', [vmin, vmax, Ord(vType), ColorToString(cFrom), ColorToString(cTo)]);
    DoGeneralEventParam(geGetDataBarParam, ACol, ARow, nil, cfg);
    newcfg:=FGeneralEventRec.ParamValue;
    if newcfg = '' then
      Exit;
    if cfg <> newcfg then
    begin
      vmin:=StrToFloatDef(UpdateToDecimalSeparator(rtool.RemoveStrPart(newcfg, FieldListSep)), 0);
      vmax:=StrToFloatDef(UpdateToDecimalSeparator(rtool.RemoveStrPart(newcfg, FieldListSep)), 100);
      i:=StrToIntDef(rtool.RemoveStrPart(newcfg, FieldListSep), 0);
      if i > Ord( High(TrDataBarText)) then
        vType:= High(TrDataBarText)
      else
        vType:=TrDataBarText(i);
      cFrom:=StringToColor(rtool.RemoveStrPart(newcfg, FieldListSep));
      cTo:=StringToColor(rtool.RemoveStrPart(newcfg, FieldListSep));
    end;

    v:=StrToFloatDef(RemoveThousandSeparator(sval), 0);
    perc:=rpictool.DrawDataBar(Canvas, ARect, v, vmin, vmax, cFrom, cTo);
    Canvas.Brush.Style:=bsClear;
    ShowValue:=vType = dbtValue;
    if vType = dbtPercent then
      DrawCellText(TextR, FormatFloat('0', perc * 100) + '%');
  end;

  procedure DrawStatusBar(sval: string);
  var
    cAll, cDone: TColor;
    v, vsteps, i: integer;
    vType: TrStatusBarType;
    cfg, newcfg: string;
  begin
    vsteps:=Columns[ACol].FStatusBar.FSteps;
    vType:=Columns[ACol].FStatusBar.IconType;
    cAll:=Columns[ACol].FStatusBar.FColorAll;
    cDone:=Columns[ACol].FStatusBar.FColorDone;

    cfg:=Format('%d,%d,%s,%s', [vsteps, Ord(vType), ColorToString(cAll), ColorToString(cDone)]);
    DoGeneralEventParam(geGetStatusBarParam, ACol, ARow, nil, cfg);
    newcfg:=FGeneralEventRec.ParamValue;
    if newcfg = '' then
      Exit;
    if cfg <> newcfg then
    begin
      vsteps:=StrToIntDef(rtool.RemoveStrPart(newcfg, FieldListSep), 5);
      i:=StrToIntDef(rtool.RemoveStrPart(newcfg, FieldListSep), 0);
      if i > Ord( High(TrStatusBarType)) then
        vType:= High(TrStatusBarType)
      else
        vType:=TrStatusBarType(i);
      cAll:=StringToColor(rtool.RemoveStrPart(newcfg, FieldListSep));
      cDone:=StringToColor(rtool.RemoveStrPart(newcfg, FieldListSep));
    end;

    v:=Round(StrToFloatDef(RemoveThousandSeparator(sval), 0));
    rpictool.DrawStatusBar(Canvas, ARect, v, vsteps, vType, cAll, cDone);
    Canvas.Brush.Style:=bsClear;
    ShowValue:=false;
  end;

  procedure DrawButton;
  var
    btnPressed, btnEnabled: boolean;
    R: TRect;
  begin
    // show button
    btnEnabled:=Enabled;
    if btnEnabled and Assigned(FOnCanEdit) then
      FOnCanEdit(self, ACol, ARow, btnEnabled);
    btnPressed:=btnEnabled and (MDownCoord.X = ACol) and (MDownCoord.Y = ARow) and (MMoveCoord.X = ACol) and (MMoveCoord.Y = ARow);
    R:=ARect;
    if GetThemeEnabled then
    begin
      if not btnEnabled then
        Details:=GetThemeServices.GetElementDetails(tbPushButtonDisabled)
      else
        if btnPressed then
          Details:=GetThemeServices.GetElementDetails(tbPushButtonPressed)
        else
          if (MMoveCoord.X = ACol) and (MMoveCoord.Y = ARow) then
            Details:=GetThemeServices.GetElementDetails(tbPushButtonHot)
          else
            Details:=GetThemeServices.GetElementDetails(tbPushButtonNormal);
      GetThemeServices.DrawElement(Canvas.Handle, Details, ARect);
    end
    else
    begin
      if btnPressed then
      begin
        Flags:=BF_FLAT;
        OffsetRect(R, 1, 1);
      end
      else
        Flags:=0;
      DrawEdge(Canvas.Handle, ARect, EDGE_RAISED, BF_RECT or BF_MIDDLE or Flags);
    end;
    // draw button text
    Canvas.Brush.Style:=bsClear;
    CellValue:=GetCellText(ACol, ARow);

    Canvas.Font.Color:=GetStyleFontColor_ButtonText(Canvas.Font.Color, not btnEnabled);
    DrawText(Canvas.Handle, PChar(CellValue), Length(CellValue), R, GetTextFormatSetting(false) or DT_CENTER);
    // draw focus
    if FocusDrawn and btnEnabled then
    begin
      R:=ARect;
      InflateRect(R, -3, -3);
      Canvas.Brush.Color:=clWindow;
      BiDi_DrawFocusRect(Canvas, R);
    end;
  end;

  procedure DrawCheckbox;
  var
    Checked: boolean;
    CheckBmp: TBitmap;
  begin
    Checked:=CheckboxIsChecked(ACol, ARow);
    // show checkbox
    CheckBmp:=rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(Checked), FScaleRatio, FCustomGlyphs);
    FCheckBoxWidth:=CheckBmp.Width;
    StyleFontColor:=GetStyleFontColor_GridText(Canvas.Font.Color);
    if Highlighted or IsSameColor(StyleFontColor, clWhite, 85) then
    begin
      DrawBmpAlign(Canvas, ARect, CheckBmp, GetCheckboxAlignment(ACol, ARow), 1);
    end
    else
    begin
      Bmp:=TBitmap.Create;
      try
        // change color of selected image to font color
        BitmapChangeColor(CheckBmp, Bmp, StyleFontColor);
        DrawBmpAlign(Canvas, ARect, Bmp, GetCheckboxAlignment(ACol, ARow), 1);
      finally
        FreeAndNil(Bmp);
      end;
    end;
  end;

  procedure SetTextFrmRect;
  var
    R: TRect;
    wrap: boolean;
  begin
    CellValue:=GetCellText(ACol, ARow);
    case Alignment of
      taRightJustify:
        Frm:=DT_RIGHT;
      taCenter:
        Frm:=DT_CENTER
      else
        Frm:=DT_LEFT;
    end;
    wrap:=GetCellWordWrap(ACol, ARow);
    Frm:=Frm or GetTextFormatSetting(wrap);
    if wrap then
    begin
      // move to vertical centered
      R:=ARect;
      DrawText(Canvas.Handle, PChar(CellValue), Length(CellValue), R, Frm or DT_CALCRECT);
      H:=R.Bottom - R.Top;
      if H < ARect.Bottom - ARect.Top then
        ARect.Top:=ARect.Top + (ARect.Bottom - ARect.Top - H) div 2;
    end;
    TextR:=ARect;
  end;

begin
  if FUpdateLock > 0 then
    Exit;
  FullR:=ARect;
  CheckColumns;
  if (SelRangeStart.X < FixedCols) or (SelRangeStart.Y < FixedRows) or (SelRangeStop.X < FixedCols) or (SelRangeStop.Y < FixedRows) then
    ResetSelRange;
  if FAutoRowHeight and (ARow > 0) and ((ACol = 0) or (FEditCoord.Y = ARow)) then
  begin
    // set row height according the most height column in row
    H:=CalcMaxRowHeight(ARow, true);
    if H <> RowHeights[ARow] then
    begin
      RowHeights[ARow]:=H;
      Exit; // auto refresh after changing rowheight
    end;
  end;
  if RowHeights[ARow] < 1 then
    Exit;

  DataImageDef.UseDataImage:=false;
  DataImageDef.ImageIndex:=-1;
  DataImageDef.ImageAlignment:='';
  DataImageDef.ShowText:='';
  DataImageDef.SkipTextDrawing:=false;

  EdStyle:=GetEditStyleEx(ACol, ARow);
  DataType:=GetDataType(ACol, ARow);
  InSelRange:=CellInSelectedRange(ACol, ARow);
  Selected:=((gdSelected in AState) and ((gdFocused in AState) or (dgAlwaysShowSelection in FOptionsEx)));
  Highlighted:=not(dgSkipCellHighlight in FOptionsEx) and Selected;
  if FInRangeCellsBkColor = clNone then
    Highlighted:=Highlighted or InSelRange;
  if FSelectedCellBkColor <> clNone then
    Highlighted:=false;

  // set special style
  myFontColor:=Canvas.Font.Color;
  myBrushColor:=Canvas.Brush.Color;
  SetCellStyle(ACol, ARow, Canvas, AState, Highlighted, DataImageDef);
  // special color defined?
  if myFontColor = Canvas.Font.Color then
    myFontColor:=clNone
  else
    myFontColor:=Canvas.Font.Color;
  if myBrushColor = Canvas.Brush.Color then
    myBrushColor:=clNone
  else
    myBrushColor:=Canvas.Brush.Color;

  // set Bk color for selected or in range
  SelBkColor:=Canvas.Brush.Color;
  if Highlighted then
    SelBkColor:=clHighLight
  else
    if InSelRange then
    begin
      if (FInRangeCellsBkColor <> clNone) then
        SelBkColor:=FInRangeCellsBkColor
      else
        if (FSelectedCellBkColor <> clNone) then
          SelBkColor:=FSelectedCellBkColor;
    end;
  if Selected and (FSelectedCellBkColor <> clNone) then
    SelBkColor:=FSelectedCellBkColor;

  // draw Bk
  if Highlighted and (EdStyle <> sgbButton) then
    DrawCellHighlight(ARect, AState, ACol, ARow)
  else
    DrawCellBackground(ARect, SelBkColor, AState, ACol, ARow);

  // reuse special font color if defined
  if myFontColor <> clNone then
    Canvas.Font.Color:=myFontColor;
  if myBrushColor <> clNone then
    Canvas.Brush.Color:=myBrushColor;

  // draw focus rect ?
  FocusDrawn:=Focused and (ARow = Row) and (ACol >= FixedCols) and ((ACol = Col) or (goRowSelect in Options));
  if (EdStyle <> sgbButton) then
  begin
    if FocusDrawn then
    begin
      BiDi_FrameRect(Canvas, ARect, clWindow);
      BiDi_DrawFocusRect(Canvas, ARect);
    end;
    InflateRect(ARect, -1, -1);
  end;

  if ARow < FixedRows then
  begin
    // fixed row
    Alignment:=Columns[ACol].TitleAlignment;
    // set inner margin and format
    InflateRect(ARect, -1 - FInnerMargin, 0);
    SetTextFrmRect;
    // draw text
    Canvas.Brush.Style:=bsClear;
    DrawCellText(TextR, CellValue);
  end
  else
    if not Columns[ACol].FCustomDraw then
    begin
      // data row
      Alignment:=GetDataAlignment(ACol, ARow);
      if (EdStyle = sgbButton) then
        DrawButton
      else
        if (EdStyle = sgbCheckbox) then
          DrawCheckbox
        else
          if (EdStyle = sgbRichText) then
            DrawRichText(ARect, GetCellText(ACol, ARow), SelBkColor)
          else
            if not DataImageDef.SkipTextDrawing then
            begin
              IsColor:=EdStyle = sgbDropDownColor;
              IsDataImage:=(EdStyle = sgbDropDownDataImage) and Assigned(Columns[ACol].FDataImageList);
              // set inner margin
              InflateRect(ARect, -1 - FInnerMargin, 0);
              // show aligned text with wordwrap support
              SetTextFrmRect;
              ShowValue:=true;
              if IsColor and (CellValue <> '') then
              begin
                // show color box
                AColor:=rpictool.StringToColor(CellValue);
                Canvas.Brush.Color:=AColor;
                Canvas.Brush.Style:=bsSolid;
                TextR.Right:=TextR.Left + TextR.Bottom - TextR.Top;
                InflateRect(TextR, -1, -1);
                if AColor <> clNone then
                  Canvas.FillRect(TextR);
                TextR:=ARect;
                TextR.Left:=TextR.Left + TextR.Bottom - TextR.Top + 2;
              end;
              if IsDataImage and (CellValue <> '') then
              begin
                DataImageDef.ImageIndex:=StrToIntDef(CellValue, -1);
                DataImageDef.UseDataImage:=(DataImageDef.ImageIndex >= 0) and (DataImageDef.ImageIndex < Columns[ACol].FDataImageList.Count);
                DataImageDef.ImageAlignment:='C';
                DataImageDef.ShowText:='';
              end;
              Canvas.Brush.Style:=bsClear;
              if DataImageDef.UseDataImage then
                DrawDataImage;
              if (DataType = sgdDataBar) and (CellValue <> '') then
                DrawProgressBar(CellValue);
              if (DataType = sgdStatusBar) and (CellValue <> '') then
                DrawStatusBar(CellValue);
              if ShowValue then
                DrawCellText(TextR, CellValue);
            end;
    end;

  DrawCellEx(ACol, ARow, FullR, AState, SelBkColor, FocusDrawn);

  Canvas.Font.Color:=Font.Color;
  inherited DrawCell(ACol, ARow, FullR, AState);
  // draw edge for classic style
  if (gdFixed in AState) and ([goVertLine, goHorzLine] * Options = [goVertLine, goHorzLine]) and Ctl3D
{$IFDEF D2010P}
  and (FInternalDrawingStyle = gdsClassic) and not(gdPressed in AState)
{$ENDIF} then
  begin
    // InflateRect(FullRect, 1, 1);
    DrawEdge(Canvas.Handle, FullR, BDR_RAISEDINNER, BF_BOTTOMRIGHT);
    DrawEdge(Canvas.Handle, FullR, BDR_RAISEDINNER, BF_TOPLEFT);
  end;
end;

procedure TrStringGridEd.DrawCellEx(ACol, ARow: longint; ARect: TRect; AState: TGridDrawState; BkColor: TColor; DrawFocus: boolean);
begin
  // nothing to do
end;

function TrStringGridEd.SelectCell(ACol, ARow: longint): boolean;
begin
  MDownCellPoint:=Point(0, 0);
  HideEditorEx;
  DoFinishEditing(Col, Row, false);
  Result:= inherited SelectCell(ACol, ARow);

  // if FBlockClick then Invalidate;
  { set editor font according column font }
  if Assigned(InplaceEditor) then
  begin
    InplaceEditorEx.Font.Assign(Columns[ACol].Font);
    InplaceEditorEx.EditStyle:=self.GetEditStyleEx(ACol, ARow);
  end;
  if (goAlwaysShowEditor in Options) then
    ShowEditor;
end;

procedure TrStringGridEd.DoEnter;
begin
  inherited;
  FCellChanged:=0;
end;

procedure TrStringGridEd.DoExit;
begin
  HideEditorEx;
  DoFinishEditing(Col, Row, false);
  inherited;
end;

function TrStringGridEd.GetRangeCellCount(Range: TGridRect): integer;
begin
  Result:=(Range.Right - Range.Left + 1) * (Range.Bottom - Range.Top + 1);
end;

function TrStringGridEd.IsRangeSelected: boolean;
begin
  Result:=(dgSelectRange in FOptionsEx) and (GetRangeCellCount(GetSelectedRange) > 1);
end;

function TrStringGridEd.GetSelectedRange: TGridRect;
begin
  Result.Left:=Min(SelRangeStart.X, SelRangeStop.X);
  Result.Top:=Min(SelRangeStart.Y, SelRangeStop.Y);
  Result.Right:=Max(SelRangeStart.X, SelRangeStop.X);
  Result.Bottom:=Max(SelRangeStart.Y, SelRangeStop.Y);
end;

procedure TrStringGridEd.SetSelectedRange(Value: TGridRect);
begin
  InvalidateSelectedRange;
  SelRangeStart.X:=Max(FixedCols, Min(Value.Left, Value.Right));
  SelRangeStart.Y:=Max(FixedRows, Min(Value.Top, Value.Bottom));
  SelRangeStop.X:=Min(ColCount - 1, Max(Value.Left, Value.Right));
  SelRangeStop.Y:=Min(RowCount - 1, Max(Value.Top, Value.Bottom));
  InvalidateSelectedRange;
end;

procedure TrStringGridEd.ResetSelRange;
var
  SelRect: TGridRect;
begin
  SelRect.Left:=Col;
  SelRect.Right:=Col;
  SelRect.Top:=Row;
  SelRect.Bottom:=Row;
  SetSelectedRange(SelRect);
end;

procedure TrStringGridEd.InvalidateSelectedRange;
begin
  if IsRangeSelected then
    InvalidateRectEx(GetSelectedRange);
end;

function TrStringGridEd.CellInSelectedRange(ACol, ARow: longint): boolean;
var
  SelRect: TGridRect;
begin
  if IsRangeSelected then
  begin
    SelRect:=GetSelectedRange;
    Result:=(ACol >= SelRect.Left) and (ACol <= SelRect.Right) and (ARow >= SelRect.Top) and (ARow <= SelRect.Bottom);
  end
  else
    Result:=false;
end;

procedure TrStringGridEd.ClearEntireTable(IncludingFixedCells: boolean = false);
var
  c, R, minc, minr: longint;
begin
  SelectFirstCell;
  SelRangeText:='';
  if IncludingFixedCells then
  begin
    minc:=0;
    minr:=0;
  end
  else
  begin
    minc:=FixedCols;
    minr:=FixedRows;
  end;
  for R:=minr to RowCount - 1 do
  begin
    for c:=minc to ColCount - 1 do
    begin
      Cells[c, R]:='';
      Objects[c, R]:=nil;
    end;
  end;
  CheckLastRows;
end;

procedure TrStringGridEd.SaveToFile(const FileName: string; Encoding: TEncoding = nil);
var
  L: TStrings;
begin
  L:=TStringList.Create;
  try
    L.Text:=CopyEntireTable(true, true);
    L.SaveToFile(FileName, Encoding);
  finally
    FreeAndNil(L);
  end;
end;

procedure TrStringGridEd.LoadFromFile(const FileName: string; SkipAllChecks: boolean = false; Encoding: TEncoding = nil);
var
  L: TStrings;
begin
  L:=TStringList.Create;
  try
    L.LoadFromFile(FileName, Encoding);
    if L.Count >= FixedRows then
      PasteSelRangeInt(0, 0, true, SkipAllChecks, false, L.Text);
  finally
    FreeAndNil(L);
  end;
end;

procedure TrStringGridEd.DoSelRangeProc(SelRange: TGridRect; Act: integer);
var
  R, c: longint;
  sr, sc: string;
  oldCursor: TCursor;
begin
  oldCursor:=WCursor;
  try
    SelRangeText:='';
    for R:=SelRange.Top to SelRange.Bottom do
    begin
      sr:='';
      for c:=SelRange.Left to SelRange.Right do
      begin
        if Act = -1 then
        begin
          if CanEditCell(c, R) then
          begin
            Cells[c, R]:='';
            DoFinishEditing(c, R, true);
          end;
        end
        else
        begin
          if (dgCopyPasteHiddenColumns in FOptionsEx) or (Columns[c].Visible) then
          begin
            if sr <> '' then
              sr:=sr + #9;
            sc:=Cells[c, R];
            if (sc = '') and (R = 0) then
              sc:=Columns[c].TitleCaption;
            // proc special chars in cell text
            if (sc = '') or (Pos(#10, sc) > 0) or (Pos('"', sc) > 0) or (Pos(#9, sc) > 0) then
            begin
              sc:=ReplaceStr(cStrEOL, #10, sc, true, true);
              sc:=ReplaceStr('"', '""', sc, true, true);
              sc:='"' + sc + '"';
            end;
            sr:=sr + sc;
          end;
        end;
      end;
      if R > SelRange.Top then
        SelRangeText:=SelRangeText + cStrEOL;
      SelRangeText:=SelRangeText + sr;
    end;
    if Act = -1 then
      CheckLastRow
      // SetEditText(SelRange.Left, SelRange.Top, '')
    else
      Clipboard.AsText:=SelRangeText;
  finally
    ChCursor(oldCursor);
  end;
end;

function TrStringGridEd.CopyEntireTable(IncludingFixedRows: boolean = true; IncludingFixedColumns: boolean = true): string;
var
  SelRange: TGridRect;
begin
  if IncludingFixedRows then
    SelRange.Top:=0
  else
    SelRange.Top:=self.FixedRows;
  if IncludingFixedColumns then
    SelRange.Left:=0
  else
    SelRange.Left:=self.FixedCols;
  SelRange.Right:=ColCount - 1;
  SelRange.Bottom:=RowCount - 1;
  DoSelRangeProc(SelRange, 0);
  Result:=SelRangeText;
end;

procedure TrStringGridEd.ClearSelRange;
begin
  DoSelRangeProc(GetSelectedRange, -1);
  CheckLastRows;
end;

function TrStringGridEd.CopySelRange: string;
begin
  DoSelRangeProc(GetSelectedRange, 0);
  Result:=SelRangeText;
end;

procedure TrStringGridEd.PasteSelRangeInt(fromCol, fromRow: longint; CanExpandGrid, SkipAllChecks, SelectPasted: boolean; const PasteData: string);
var
  R, c, maxc: longint;
  sr, sc, srrest, screst: string;
  oldCursor: TCursor;
begin
  if PasteData <> '' then
    SelRangeText:=PasteData
  else
    try
      SelRangeText:=Clipboard.AsText;
    except
      Exit;
    end;
  if (SelRangeText = '') then
    Exit;
  if not SkipAllChecks and not([goRowSelect, goEditing] * Options = [goEditing]) then
    Exit;
  if not DoGeneralEvent(gePasteDataBefore) then
    Exit;
  oldCursor:=WCursor;
  try
    srrest:=SelRangeText;
    R:=fromRow;
    maxc:=fromCol;
    repeat
      // extract row text
      sr:=rtool.RemoveStrPart(srrest, cStrEOL, false);
      screst:=sr;
      c:=fromCol;
      repeat
        // extract cell text
        sc:=rtool.RemoveStrPart(screst, #9, false);
        // proc special chars in cell text
        if (copy(sc, 1, 1) = '"') and (copy(sc, Length(sc), 1) = '"') then
        begin
          sc:=copy(sc, 2, Length(sc) - 2);
          sc:=ReplaceStr('""', '"', sc, true, true);
          sc:=ReplaceStr(#10, cStrEOL, sc, true, true);
        end;
        // check visibility
        if not(dgCopyPasteHiddenColumns in FOptionsEx) then
          while (c < ColCount) and not(Columns[c].Visible) do
            Inc(c);
        if CanExpandGrid and (R >= RowCount) then
          RowCount:=R + 1;
        if CanExpandGrid and (c >= ColCount) then
          ColCount:=c + 1;
        if (R < RowCount) and (c < ColCount) then
        begin
          case GetDataType(c, R) of
            sgdNumber, sgdDataBar, sgdInteger, sgdStatusBar:
              sc:=rtool.CleanNumericString(sc);
            sgdDate, sgdTime, sgdDateTime:
              sc:=rtool.CleanDateTimeString(sc);
          end;
          if SkipAllChecks then
            Cells[c, R]:=sc
          else
          begin
            if CanEditCell(c, R) then
            begin
              Cells[c, R]:=sc;
              if (c >= FixedCols) and (R >= FixedRows) then
                DoFinishEditing(c, R, true);
            end;
          end;
          if c > maxc then
            maxc:=c;
        end;
        Inc(c);
      until screst = '';
      Inc(R);
    until srrest = '';
    CheckLastRow;
    if SelectPasted then
    begin
      SelRangeStart.X:=fromCol;
      SelRangeStart.Y:=fromRow;
      SelRangeStop.X:=maxc;
      SelRangeStop.Y:=R - 1;
    end;
    Invalidate;
  finally
    ChCursor(oldCursor);
  end;
  if not DoGeneralEvent(gePasteDataAfter) then
    Exit;
end;

procedure TrStringGridEd.PasteEntireTable(CanExpandGrid: boolean; const PasteData: string = '');
begin
  PasteSelRangeInt(0, 0, CanExpandGrid, false, false, PasteData);
end;

procedure TrStringGridEd.PasteSelRange(CanExpandGrid: boolean; const PasteData: string = '');
begin
  PasteSelRangeInt(Col, Row, CanExpandGrid, false, true, PasteData);
end;

function TrStringGridEd.FindIniRow(const IniKey: string): longint;
begin
  Result:=Cols[FIniKeyColumn].IndexOf(IniKey);
end;

function TrStringGridEd.GetIniValue(const IniKey: string): string;
var
  R: longint;
begin
  R:=FindIniRow(IniKey);
  if R >= FixedRows then
    Result:=Cells[FIniValueColumn, R]
  else
    Result:='';
end;

procedure TrStringGridEd.SetIniValue(const IniKey, Value: string);
var
  R: longint;
begin
  R:=FindIniRow(IniKey);
  if R >= FixedRows then
    Cells[FIniValueColumn, R]:=Value;
end;

function TrStringGridEd.SaveToIniFile(const IniFileName: string; Encoding: TEncoding = nil): string;
var
  L: TStrings;
  R: longint;
begin
  L:=TStringList.Create;
  try
    for R:=FixedRows to RowCount - 1 do
    begin
      if Cells[FIniKeyColumn, R] <> '' then
        if IsIniGroupRow(R) then
          L.Add('[' + Cells[FIniKeyColumn, R] + ']')
        else
          L.Add(Cells[FIniKeyColumn, R] + L.NameValueSeparator + Cells[FIniValueColumn, R]);
    end;
    Result:=L.Text;
    if IniFileName <> '' then
      L.SaveToFile(IniFileName, Encoding);
  finally
    FreeAndNil(L);
  end;
end;

procedure TrStringGridEd.LoadFromIniFile(const IniFileName: string; const IniFileContent: string = ''; Encoding: TEncoding = nil);
var
  L: TStrings;
  R: longint;
  s: string;
begin
  FIniGroupRows.Clear;
  L:=TStringList.Create;
  try
    if IniFileName <> '' then
      L.LoadFromFile(IniFileName, Encoding)
    else
      L.Text:=IniFileContent;
    if L.Count > RowCount - FixedRows then
      RowCount:=FixedRows + L.Count;
    for R:=0 to L.Count - 1 do
    begin
      s:=Trim(L[R]);
      if (copy(s, 1, 1) = '[') and (copy(s, Length(s), 1) = ']') then
      begin
        Cells[FIniKeyColumn, R + FixedRows]:=copy(s, 2, Length(s) - 2);
        FIniGroupRows.Add(Cells[FIniKeyColumn, R + FixedRows]);
      end
      else
      begin
        Cells[FIniKeyColumn, R + FixedRows]:=L.Names[R];
        Cells[FIniValueColumn, R + FixedRows]:=L.ValueFromIndex[R];
      end;
    end;
  finally
    FreeAndNil(L);
  end;
end;

procedure TrStringGridEd.LoadValuesFromIniFile(const IniFileName: string; const IniFileContent: string = ''; Encoding: TEncoding = nil);
var
  L: TStrings;
  R: longint;
begin
  L:=TStringList.Create;
  try
    if IniFileName <> '' then
      L.LoadFromFile(IniFileName, Encoding)
    else
      L.Text:=IniFileContent;
    for R:=FixedRows to RowCount - 1 do
    begin
      if L.IndexOfName(Cells[FIniKeyColumn, R]) >= 0 then
        Cells[FIniValueColumn, R]:=L.Values[Cells[FIniKeyColumn, R]];
    end;
  finally
    FreeAndNil(L);
  end;
end;

function TrStringGridEd.IsIniGroupRow(ARow: longint): boolean;
begin
  Result:=Assigned(FIniGroupRows) and (FIniGroupRows.Count > 0) and (FIniGroupRows.IndexOf(Cells[FIniKeyColumn, ARow]) >= 0);
end;

function TrStringGridEd.IsValidData(ACol, ARow: longint): TrStringGridEdDataState;
var
  s: string;
  f: double;
begin
  Result:=sdsOk;
  s:=Cells[ACol, ARow];
  if s = '' then
    Result:=sdsEmpty
  else
    if GetCoordPosition(ACol, ARow) = cpNotFixed then
    begin
      f:=0;
      try
        // remove space and non-break space for numbers
        case GetDataType(ACol, ARow) of
          sgdNumber, sgdDataBar:
            f:=StrToFloat(RemoveThousandSeparator(s));
          sgdInteger, sgdStatusBar:
            f:=StrToInt(RemoveThousandSeparator(s));
          sgdDate:
            begin
              f:=FormattedStrToDateTimeDef(s, GetNumFormat(ACol, ARow), 0, true);
              if f = 0 then
                Result:=sdsInvalid;
            end;
          sgdDateTime:
            begin
              f:=FormattedStrToDateTimeDef(s, GetNumFormat(ACol, ARow), 0, false);
              if f = 0 then
                Result:=sdsInvalid;
            end;
          sgdTime:
            f:=StrToTime(s);
        else
          Exit;
        end;
      except
        Result:=sdsInvalid;
      end;
      if (Result = sdsOk) and ((Columns[ACol].MinValue <> 0) or (Columns[ACol].MaxValue <> 0)) then
        if (f < Columns[ACol].MinValue) or (f > Columns[ACol].MaxValue) then
          Result:=sdsOutOfLimit;
    end;
end;

procedure TrStringGridEd.SetEditText(ACol, ARow: longint; const Value: string);
var
  oldValue: string;
begin
  oldValue:=Cells[ACol, ARow];
  inherited SetEditText(ACol, ARow, Value);
  FEditCoord.X:=ACol;
  FEditCoord.Y:=ARow;
  if (Value <> '-') and (IsValidData(ACol, ARow) = sdsInvalid) then
    if Columns[ACol].InvalidDataAction = sgaBeep then
      Beep;
  if (ACol = Col) and (ARow = Row) and (oldValue <> Value) then
    Inc(FCellChanged);
  CheckLastRow;
end;

function TrStringGridEd.GetEditText(ACol, ARow: longint): string;
begin
  FOldEditedValue:=Cells[ACol, ARow];
  Result:= inherited GetEditText(ACol, ARow);
end;

procedure TrStringGridEd.DoFinishEditing(ACol, ARow: longint; Forced: boolean);
var
  DataType: TrStringGridEdDataType;
  NumFormat: string;
  dt: TDateTime;
begin
  if Forced or (FCellChanged > 0) then
  begin
    try
      DataType:=GetDataType(ACol, ARow);
      NumFormat:=GetNumFormat(ACol, ARow);
      if (DataType in [sgdNumber, sgdInteger, sgdDataBar, sgdStatusBar]) then
        if IsValidData(ACol, ARow) in [sdsOk, sdsOutOfLimit] then
          SetEditText(ACol, ARow, FormatFloat(NumFormat, GetNumCells(ACol, ARow)));
      if (DataType in [sgdDate, sgdTime, sgdDateTime]) then
        if IsValidData(ACol, ARow) in [sdsOk, sdsOutOfLimit] then
        begin
          dt:=FormattedStrToDateTimeDef(GetStringCells(ACol, ARow), GetNumFormat(ACol, ARow), Date, DataType = sgdDate);
          if DataType = sgdDate then
            dt:=DateOf(dt);
          if DataType = sgdTime then
            dt:=TimeOf(dt);
          SetEditText(ACol, ARow, DateTimeToFormattedStr(dt, NumFormat, DataType = sgdTime));
        end;
      if (FUpdateLock = 0) or FUseChangeLog then
      begin
        if (FChangedRows.IndexOf(IntToStr(ARow)) < 0) then
          FChangedRows.Add(IntToStr(ARow));
        if not IsCellChanged(ACol, ARow) then
          FChangedCells.Add(GetChangedCellStr(ACol, ARow));
      end;
    except
    end;
    if (FUpdateLock = 0) and Assigned(FOnFinishCellEditing) then
      FOnFinishCellEditing(self, ACol, ARow);
  end;
  FCellChanged:=0;
end;

procedure TrStringGridEd.EditorKeyPress(var Key: Char);
begin
  if not CanEditShow then
    Key:=#0;
  if Ord(Key) = VK_ESCAPE then
  begin
    HideEditorEx;
    SetEditText(Col, Row, FOldEditedValue);
    if (goAlwaysShowEditor in Options) then
      ShowEditor;
  end;
  if Ord(Key) = VK_RETURN then
  begin
    // Key:=#0;
    if EditorMode then
    begin
      HideEditorEx;
      DoFinishEditing(Col, Row, false);
      ShowEditor;
    end;
    if (dgNextRowAfterReturn in FOptionsEx) and (Row < RowCount - 1) then
      Row:=Row + 1;
  end;
  if Assigned(FOnEditorKeyPress) then
    FOnEditorKeyPress(self, Key);
end;

function TrStringGridEd.CheckLastRow: boolean;
var
  e1, e2: boolean;
begin
  CheckLastRow:=false;
  // add new row or delete last empty row
  if (FUpdateLock > 0) or not FAutoAddNewRow then
    Exit;
  if RowCount < 3 then
    e1:=false
  else
    e1:=IsRowEmpty(RowCount - 2);
  e2:=IsRowEmpty(RowCount - 1);
  if e1 and e2 then
  begin
    RowCount:=RowCount - 1;
    CheckLastRow:=true;
  end
  else
    if not e2 then
    begin
      RowCount:=RowCount + 1;
      CheckLastRow:=true;
    end;
end;

procedure TrStringGridEd.CheckLastRows;
begin
  while CheckLastRow do
    ;
end;

procedure TrStringGridEd.BeginUpdate(UseChangeLog: boolean = false);
begin
  Inc(FUpdateLock);
  FUseChangeLog:=UseChangeLog;
end;

procedure TrStringGridEd.EndUpdate;
begin
  if FUpdateLock > 0 then
    Dec(FUpdateLock);
  CheckLastRows;
  Invalidate;
end;

procedure TrStringGridEd.WMChar(var Msg: TWMChar);
begin
  if (goEditing in Options) and (CharInSet(Char(Msg.CharCode), [^H]) or (Char(Msg.CharCode) >= #32)) then
    FStartEditKey:=Msg.CharCode;
  inherited;
  FStartEditKey:=0;
end;

procedure TrStringGridEd.KeyDown(var Key: Word; Shift: TShiftState);
var
  ebs: TrStringGridEdEditStyle;
  s: string;

  procedure DoDelete;
  begin
    if CanEditModifyEx(Columns[Col].FDropDownListCanDelete) then
    begin
      SetEditText(Col, Row, '');
      DoFinishEditing(Col, Row, true);
      if (dgSelectRange in FOptionsEx) then
        ClearSelRange;
    end;
  end;

  procedure DoCopy;
  begin
    Clipboard.AsText:=Cells[Col, Row];
    if (dgSelectRange in FOptionsEx) then
      CopySelRange;
  end;

begin
  if (Shift = []) and (Key = VK_F2) then
    FStartEditKey:=VK_F2;
  inherited KeyDown(Key, Shift);
  FStartEditKey:=0;
  // Ctrl+C or Ctrl+Insert - copy to clipboard
  if (CharInSet(chr(Key), ['c', 'C']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssCtrl])) then
    DoCopy;
  // Ctrl+X - Cut to clipboard
  if (Shift = [ssCtrl]) and CharInSet(chr(Key), ['x', 'X']) then
  begin
    DoCopy;
    DoDelete;
  end;
  // Ctrl+V or Shift+Insert- paste from clipboard
  if (CharInSet(chr(Key), ['v', 'V']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssShift])) then
  begin
    if CanEditModifyEx(false) then
    begin
      if (dgSelectRange in FOptionsEx) then
        PasteSelRange(FAutoAddNewRow)
      else
      begin
        s:=Clipboard.AsText;
        case GetDataType(Col, Row) of
          sgdNumber, sgdDataBar, sgdInteger, sgdStatusBar:
            s:=rtool.CleanNumericString(s);
          sgdDate, sgdTime, sgdDateTime:
            s:=rtool.CleanDateTimeString(s);
        else
          s:=rtool.CleanLastEOL(s);
        end;
        SetEditText(Col, Row, s);
        DoFinishEditing(Col, Row, true);
      end;
    end;
  end;
  // delete
  if (Shift = []) and (Key = VK_Delete) then
    DoDelete;
  // change Checkbox on Space key
  if Key = 32 then
  begin
    case GetEditStyleEx(Col, Row) of
      sgbCheckbox:
        ChangeCheckbox;
      sgbButton:
        ButtonClick;
    end;
  end;

  // F4 - click to edditbutton
  if (Key = VK_F4) and CanEditModifyEx(true) then
  begin
    ebs:=GetEditStyleEx(Col, Row);
    if ebs in [sgbEllipsis, sgbEllipsisOnly] then
    begin
      ShowEditor;
      EllipsisClick;
    end
    else
      if (ebs in cesShowBtn) then
      begin
        ShowEditor;
        if Assigned(InplaceEditor) then
          InplaceEditorEx.DropDown;
      end;
  end;

  if (dgSelectRange in FOptionsEx) then
  begin
    case Key of
      VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT, VK_PRIOR, VK_NEXT, VK_HOME, VK_END:
        if (ssShift in Shift) then
        begin
          // select range
          InvalidateSelectedRange;
          SelRangeStop.X:=Col;
          SelRangeStop.Y:=Row;
          InvalidateSelectedRange;
        end
        else
        begin
          // set selected cell as StartSelRange
          if not IsRangeSelected then
            ResetSelRange;
        end;
      VK_ESCAPE:
        if IsRangeSelected then
          ResetSelRange;
    end;
  end;
end;

procedure TrStringGridEd.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
end;

function TrStringGridEd.GetStringCells(ACol, ARow: longint): string;
begin
  Result:=Cells[ACol, ARow];
end;

procedure TrStringGridEd.SetStringCells(ACol, ARow: longint; const Value: string);
begin
  if Cells[ACol, ARow] <> Value then
  begin
    SetEditText(ACol, ARow, Value);
    DoFinishEditing(ACol, ARow, true);
  end;
end;

function TrStringGridEd.GetNumCells(ACol, ARow: longint): double;
var
  s: string;
begin
  s:=GetStringCells(ACol, ARow);
  try
    if s = '' then
      s:='0';
    Result:=StrToFloat(RemoveThousandSeparator(s));
  except
    Col:=ACol;
    Row:=ARow;
    raise ;
  end;
end;

procedure TrStringGridEd.SetNumCells(ACol, ARow: longint; const Value: double);
begin
  SetStringCells(ACol, ARow, FormatFloat(GetNumFormat(ACol, ARow), Value));
end;

function TrStringGridEd.GetIntCells(ACol, ARow: longint): integer;
var
  s: string;
begin
  s:=GetStringCells(ACol, ARow);
  try
    if s = '' then
      s:='0';
    Result:=StrToInt(s);
  except
    Col:=ACol;
    Row:=ARow;
    raise ;
  end;
end;

procedure TrStringGridEd.SetIntCells(ACol, ARow: longint; const Value: integer);
begin
  SetNumCells(ACol, ARow, Value);
end;

function TrStringGridEd.GetBoolCells(ACol, ARow: longint): boolean;
begin
  Result:=CheckboxIsChecked(ACol, ARow);
end;

procedure TrStringGridEd.SetBoolCells(ACol, ARow: longint; const Value: boolean);
begin
  if Value then
    SetEditText(ACol, ARow, Columns[ACol].FCheckBoxTrueValue)
  else
    SetEditText(ACol, ARow, Columns[ACol].FCheckBoxFalseValue);
end;

function TrStringGridEd.GetDateTimeCells(ACol, ARow: longint): TDateTime;
begin
  Result:=FormattedStrToDateTimeDef(GetStringCells(ACol, ARow), GetNumFormat(ACol, ARow), Date, GetDataType(ACol, ARow) = sgdDate);
end;

procedure TrStringGridEd.SetDateTimeCells(ACol, ARow: longint; const Value: TDateTime);
begin
  SetEditText(ACol, ARow, DateTimeToFormattedStr(Value, GetNumFormat(ACol, ARow), GetDataType(ACol, ARow) = sgdTime));
end;

procedure TrStringGridEd.SetColumnAutoFit(ColIdx: longint);
var
  R, w, maxw, maxmaxw: longint;
begin
  if (ColIdx >= FixedCols) and (ColIdx < ColCount) then
  begin
    if Columns[ColIdx].ButtonStyle = sgbRichText then
      Exit;
    Canvas.Font.Assign(Columns[ColIdx].FTitleFont);
    maxw:=Max(Canvas.TextWidth(Columns[ColIdx].TitleCaption), 40);
    maxmaxw:=Width * 2 div 3;
    Canvas.Font.Assign(Columns[ColIdx].FFont);
    for R:=0 to RowCount - 1 do
    begin
      w:=Canvas.TextWidth(Cells[ColIdx, R]);
      if w > maxw then
        maxw:=w;
      if w > maxmaxw then
      begin
        maxw:=maxmaxw;
        Break;
      end;
    end;
    if Columns[ColIdx].FButtonStyle in cesShowBtn then
      Inc(maxw, GetSystemMetrics(SM_CXVSCROLL));
    if not DoGeneralEventParam(geSetColumnWidth, ColIdx, 0, nil, maxw) then
      Exit;
    maxw:=FGeneralEventRec.ParamValue;
    Columns[ColIdx].Width:=maxw + 12;
  end;
end;

procedure TrStringGridEd.SetColumnsAutoFit;
var
  c: longint;
begin
  for c:=FixedCols to ColCount - 1 do
    SetColumnAutoFit(c);
end;

function TrStringGridEd.GetColumnDataWidth(ColIdx: longint; MaxRowsToTest: longint = 10000): integer;
var
  R, w, maxw: integer;
  // col: TrStringGridEdColumn;
begin
  Result:=-1;
  if (ColIdx < 0) or (ColIdx >= ColCount) then
    Exit;
  maxw:=10;
  // col:=Columns[ColIdx];
  for R:=FixedRows to Min(RowCount - 1, MaxRowsToTest) do
  begin
    w:=Length(Cells[ColIdx, R]);
    if w > maxw then
      maxw:=w;
  end;
  Result:=maxw;
end;

procedure TrStringGridEd.DoEllipsisClick;
begin
  SetTimer(Handle, cIntTimerID, 120, nil); // EllipsisClick;
end;

procedure TrStringGridEd.FinishEditing;
begin
  HideEditorEx;
  DoFinishEditing(Col, Row, false);
end;

function TrStringGridEd.ActiveCellText: string;
begin
  Result:=Cells[Col, Row];
end;

function TrStringGridEd.IsRowEmpty(ARow: longint; SkipFixedCol: boolean = true; ColumnGroup: integer = 0): boolean;
var
  c, minc: integer;
begin
  Result:=true;
  if (ARow < FixedRows) or (ARow >= RowCount) then
    Exit;
  if SkipFixedCol then
    minc:=FixedCols
  else
    minc:=0;
  for c:=minc to ColCount - 1 do
    if (Trim(Cells[c, ARow]) <> '') and ((ColumnGroup = 0) or (ColumnGroup = Columns[c].ColumnGroup)) then
    begin
      Result:=false;
      Break;
    end;
end;

procedure TrStringGridEd.ResetChangeLog;
begin
  FinishEditing;
  FChangedRows.Clear;
  FChangedCells.Clear;
  Invalidate;
end;

procedure TrStringGridEd.SelectFirstCell(minCol: longint = 0; minRow: longint = 0);
begin
  FinishEditing;
  Col:=Min(Max(FixedCols, minCol), ColCount - 1);
  Row:=Min(Max(FixedRows, minRow), RowCount - 1);
  ResetSelRange;
end;

procedure TrStringGridEd.UpdateChangedCell(UpdateRows: boolean; ChangedStartIdx, NumberOfItems: longint);
var
  i, R, c: integer;
  s, t: string;
  L: TStrings;
begin
  L:=TStringList.Create;
  try
    // row changes
    L.Clear;
    if UpdateRows then
    begin
      for i:=0 to FChangedRows.Count - 1 do
      begin
        s:=FChangedRows[i];
        R:=StrToIntDef(s, 0);
        if R >= ChangedStartIdx then
        begin
          Inc(R, NumberOfItems);
          s:=IntToStr(R);
          // delete line if deleted ( NumberOfRows < 0)
          if R - ChangedStartIdx < 0 then
            s:='';
        end;
        if s <> '' then
          L.Add(s);
      end;
      FChangedRows.Assign(L);
    end;
    // cell changes
    L.Clear;
    for i:=0 to FChangedCells.Count - 1 do
    begin
      s:=FChangedCells[i];
      t:=s;
      c:=StrToIntDef(rtool.RemoveStrPart(t, ';'), 0);
      R:=StrToIntDef(t, 0);
      if UpdateRows then
      begin
        if R >= ChangedStartIdx then
        begin
          Inc(R, NumberOfItems);
          s:=GetChangedCellStr(c, R);
          // delete line if deleted ( NumberOfRows < 0)
          if R - ChangedStartIdx < 0 then
            s:='';
        end;
      end
      else
      begin
        if c >= ChangedStartIdx then
        begin
          Inc(c, NumberOfItems);
          s:=GetChangedCellStr(c, R);
          // delete line if deleted ( NumberOfRows < 0)
          if c - ChangedStartIdx < 0 then
            s:='';
        end;
      end;
      if s <> '' then
        L.Add(s);
    end;
    FChangedCells.Assign(L);
  finally
    L.Free;
  end;
end;

procedure TrStringGridEd.MoveRowEx(FromIndex: integer; ToIndex: integer);
begin
  FinishEditing;
  ResetSelRange;
  MoveRow(FromIndex, ToIndex);
end;

procedure TrStringGridEd.InsertRowEx(RowIndex: integer; NumberOfRows: integer = 1; Activate: boolean = true);
var
  oldRow, R: longint;
begin
  if NumberOfRows < 1 then
    Exit;
  oldRow:=Row;
  if RowIndex < FixedRows then
    RowIndex:=FixedRows;
  RowCount:=RowCount + NumberOfRows;
  for R:=1 to NumberOfRows do
    MoveRowEx(RowCount - 1, RowIndex);
  Row:=RowIndex;
  if oldRow > RowIndex then
    Inc(oldRow);
  if not Activate then
    Row:=oldRow;
  UpdateChangedCell(true, RowIndex, NumberOfRows);
end;

procedure TrStringGridEd.DeleteRowEx(RowIndex: longint; NumberOfRows: integer = 1);
var
  oldRow, R: longint;
begin
  if NumberOfRows < 1 then
    Exit;
  if (RowIndex >= FixedRows) and (RowIndex < RowCount) then
  begin
    oldRow:=Row;
    for R:=1 to NumberOfRows do
    begin
      MoveRowEx(RowIndex, RowCount - 1);
      Rows[RowCount - 1].Clear;
    end;
    RowCount:=Max(RowCount - NumberOfRows, FixedRows + 1);
    if oldRow > RowCount - 1 then
      oldRow:=RowCount - 1;
    Row:=oldRow;
    CheckLastRows;
  end;
  UpdateChangedCell(true, RowIndex, -NumberOfRows);
end;

procedure TrStringGridEd.MoveColumnEx(FromIndex: integer; ToIndex: integer);
begin
  FinishEditing;
  ResetSelRange;
  MoveColumn(FromIndex, ToIndex);
end;

procedure TrStringGridEd.InsertColumnEx(ColIndex: integer; NumberOfCols: integer = 1; Activate: boolean = true);
var
  oldCol, c: integer;
begin
  if NumberOfCols < 1 then
    Exit;
  oldCol:=Col;
  while ColIndex < FixedCols do
    Inc(ColIndex);
  ColCount:=ColCount + 1;
  for c:=1 to NumberOfCols do
    MoveColumnEx(ColCount - 1, ColIndex);
  Col:=ColIndex;
  if oldCol > ColIndex then
    Inc(oldCol);
  if not Activate then
    Col:=oldCol;
  UpdateChangedCell(false, ColIndex, NumberOfCols);
end;

procedure TrStringGridEd.DeleteColumnEx(ColIndex: integer; NumberOfCols: integer = 1);
var
  oldCol, c: integer;
begin
  if NumberOfCols < 1 then
    Exit;
  if (ColIndex >= FixedCols) and (ColIndex < ColCount) then
  begin
    oldCol:=Col;
    for c:=1 to NumberOfCols do
    begin
      MoveColumnEx(ColIndex, ColCount - 1);
      Cols[ColCount - 1].Clear;
    end;
    ColCount:=Max(ColCount - NumberOfCols, FixedCols + 1);
    if oldCol > ColCount - 1 then
      oldCol:=ColCount - 1;
    Col:=oldCol;
  end;
  UpdateChangedCell(false, ColIndex, -NumberOfCols);
end;

function TrStringGridEd.CompareValues(v1, v2: string; Column: TrStringGridEdColumn; CompareType: longint): TValueRelationship;
begin
  Result:=0;
  if v1 = v2 then
    Exit;
  if v1 = '' then
    Result:=CompareType
  else
    if v2 <> '' then
      try
        case Column.DataType of
          sgdText:
            Result:=AnsiCompareStr(v1, v2);
          sgdNumber, sgdInteger, sgdDataBar, sgdStatusBar:
            Result:=CompareValue(StrToFloat(RemoveThousandSeparator(v1)), StrToFloat(RemoveThousandSeparator(v2)));
          sgdDate:
            Result:=CompareDateTime(FormattedStrToDateTimeDef(v1, Column.NumFormat, 0, true), FormattedStrToDateTimeDef(v2, Column.NumFormat, 0, true));
          sgdDateTime:
            Result:=CompareDateTime(FormattedStrToDateTimeDef(v1, Column.NumFormat, 0, false), FormattedStrToDateTimeDef(v2, Column.NumFormat, 0, false));
          sgdTime:
            Result:=CompareDateTime(StrToTime(v1), StrToTime(v2));
        end;
      except
      end;
end;

function TrStringGridEd.CompareCellData(ACol, ARow1, ARow2, CompVal: longint): TValueRelationship;
begin
  Result:=CompareValues(Cells[ACol, ARow1], Cells[ACol, ARow2], Columns[ACol], CompVal);
end;

procedure TrStringGridEd.QuickSort(List: TStrings; L, R: integer; Column: TrStringGridEdColumn; CompareType: longint; var Proc: longint);
var
  i, J, p: integer;
begin
  repeat
    i:=L;
    J:=R;
    p:=(L + R) shr 1;
    repeat
      while CompareValues(List[i], List[p], Column, CompareType) * CompareType < 0 do
        Inc(i);
      while CompareValues(List[J], List[p], Column, CompareType) * CompareType > 0 do
        Dec(J);
      if i <= J then
      begin
        if i <> J then
        begin
          // exchange list and rows
          List.Exchange(i, J);
          if i < J then
          begin
            RowMoved(J, i);
            RowMoved(i + 1, J);
          end
          else
          begin
            RowMoved(i, J);
            RowMoved(J + 1, i);
          end;
          // progress
          if (Proc >= 0) then
          begin
            if (Proc mod 1000 = 0) then
              rdlg.ShowInfoProgressBar(sLoading + sOfData, self, DoAbort, 50 * Proc div RowCount);
            if Assigned(FOnProgress) then
              FOnProgress(self, 'sort', 50 * Proc div RowCount, FAborted);
            Inc(Proc);
          end;
        end;
        if p = i then
          p:=J
        else
          if p = J then
            p:=i;
        Inc(i);
        Dec(J);
      end;
      if FAborted then
        Break;
    until i > J;
    if FAborted then
      Break;
    if L < J then
      QuickSort(List, L, J, Column, CompareType, Proc);
    L:=i;
  until i >= R;
end;

procedure TrStringGridEd.SortData(ColIndex: integer; Desc: boolean = false; ResetSelection: boolean = false);
var
  lastRow, comp, Proc: longint;
  showProgress: boolean;
  L: TStringList;
begin
  FinishEditing;
  if (ColIndex < ColCount) then
  begin
    lastRow:=RowCount - 1;
    if AutoAddNewRow then
      Dec(lastRow);
    if Desc then
      comp:=-1
    else
      comp:=1;
    showProgress:=(dgSortShowProgress in FOptionsEx) and (lastRow > 1000);
    if showProgress then
      Proc:=0
    else
      Proc:=-1;

    FAborted:=false;

    L:=TStringList.Create;
    try
      L.BeginUpdate;
      BeginUpdate;
      try
        L.Assign(Cols[ColIndex]);
        QuickSort(L, FixedRows, lastRow, Columns[ColIndex], comp, Proc);
      finally
        EndUpdate;
        L.EndUpdate;
        if showProgress then
          rdlg.ShowInfo('');
        if Assigned(FOnProgress) then
          FOnProgress(self, 'end', 100, FAborted);
      end;
    finally
      L.Free;
    end;
  end;
  if ColIndex >= FixedCols then
  begin
    Col:=ColIndex;
    Row:=FixedRows;
  end;
  if ResetSelection then
    ResetSelRange;
end;

procedure TrStringGridEd.DoAbort(Sender: TObject);
begin
  FAborted:=true;
end;

function TrStringGridEd.DoGeneralEvent(EventType: TGeneralEventType): boolean;
begin
  Result:=DoGeneralEventParam(EventType, -1, -1, nil, VarEmpty);
end;

function TrStringGridEd.DoGeneralEventParam(EventType: TGeneralEventType; Col, Row: longint; ParamObject: TObject; ParamValue: variant): boolean;
begin
  Result:=true;
  // FGeneralEventRec.Rect:=R;
  FGeneralEventRec.Col:=Col;
  FGeneralEventRec.Row:=Row;
  FGeneralEventRec.ParamObject:=ParamObject;
  FGeneralEventRec.ParamValue:=ParamValue;
  if Assigned(FOnGeneralEvent) then
    FOnGeneralEvent(self, EventType, FGeneralEventRec, Result);
end;

end.
