unit rXLSExport;

{<
  rXLSExport unit contains TrXLSExport - non visual component which provides functionality for
  exporting of any data directly to XLS file using OLE;
  it can easily export data from any dataset or DBGrid.

  Updated version compatible with Delphi2010+.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Feb 2014)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Dialogs, Controls, Db, Math, ComCtrls,
  Variants, ComObj, Graphics, Clipbrd,
  rImprovedComps, DBGrids, rStringGridEd, rDBRecView, rDBTreeView,
  rDBGrid, rDBGrid_MS, RosiCompGlobalCfg;

const
  // XlFileFormat Enumeration (Excel)
  // see http://msdn.microsoft.com/en-us/library/office/ff198017(v=office.15).aspx for full list
  xlSYLK = 2;
  xlCSV = 6;
  xlCSVWindows = 23;
  xlCSVMSDOS = 24;
  xlTextWindows = 20;
  xlUnicodeText = 42;
  xlExcel8 = 56;
  xlExcel9795 = 43;
  xlExcel12 = 50;
  xlOpenXMLWorkbook = 51;
  xlOpenXMLWorkbookMacroEnabled = 52;
  xlXMLSpreadsheet = 46;
  xlOpenDocumentSpreadsheet = 60;
  xlWorkbookNormal = -4143;
  xlTemplate = 17;

  // Cell Borders
  xlHairline = 1;
  xlMedium = -4138;
  xlThick = 4;
  xlThin = 2;

  xlAllBorders = 0;
  xlDiagonalDown = 5;
  xlDiagonalUp = 6;
  xlEdgeBottom = 9;
  xlEdgeLeft = 7;
  xlEdgeRight = 10;
  xlEdgeTop = 8;
  xlInsideHorizontal = 12;
  xlInsideVertical = 11;

  // Line type
  xlContinuous = 1;
  xlDash = -4115;
  xlDashDot = 4;
  xlDashDotDot = 5;
  xlDot = -4118;
  xlDouble = -4119;
  xlLineStyleNone = -4142;
  xlSlantDashDot = 13;

  // Constants for enum XlPageOrientation
  xlLandscape = $00000002;
  xlPortrait = $00000001;

  // ratio between width/height and pixels
  cColWidthPixelRatioDefault = 9.15;
  cRowHeightPixelRatioDefault = 1.65;

  cFldTypeNormal = 0;
  cFldTypePicture = 1;
  cFldTypeColor = 2;
  cFldTypeDataBar = 3;
  cFldTypeColorScale = 4;
  cFldTypeIconSet = 5;

  // IconSets
  xl3Arrows = 1;
  xl3ArrowsGray = 2;
  xl3Flags = 3;
  xl3Signs = 6;
  xl3Stars = 18;
  xl3Symbols = 7;
  xl3Symbols2 = 8;
  xl3TrafficLights1 = 4;
  xl3TrafficLights2 = 5;
  xl3Triangles = 19;
  xl4Arrows = 9;
  xl4ArrowsGray = 10;
  xl4CRV = 12;
  xl4RedToBlack = 11;
  xl4TrafficLights = 13;
  xl5Arrows = 14;
  xl5ArrowsGray = 15;
  xl5Boxes = 20;
  xl5CRV = 16;
  xl5Quarters = 17;
  xlCustomSet = -1;

  // international
  xlDayLeadingZero = $0000002A;
  xlDecimalSeparator = $00000003;
  xlMonthLeadingZero = $00000029;
  xl4DigitYears = $0000002B;
  xlDayCode = $00000015;
  xlMonthCode = $00000014;
  xlYearCode = $00000013;
  xlDateSeparator = $00000011;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;

  THackCustomTreeView = class(TCustomTreeView)
  end;

  { Defines types of cell format }
  TXLSCellFormat = (cfNone, cfBold, cfItalic, cfUnderline);
  { Defines types of cell format }
  TXLSCellFormats = set of TXLSCellFormat;

  { Defines options for ExportTable function }
  TXLSExportOption = (xleExportTitles, xleRealFieldNamesAsTitle, xleBoldTitles, xleAddSumRow, xleStringFieldAsText, xleSetCellFormat, xleSetCellBorders,
    xleWrapMemoField, xleSetColumnWidth, xleSetRowHeight, xleFreezeTitles, xleSetAutoFilter, xleSetPageSetup, xleOnlyVisibleFields, xleOnlySelectedRows,
    xleExportPictures, xleAddAsNewSheet, xleAddToExistingRows, xleTrimData);
  { Defines options for ExportTable function }
  TXLSExportOptions = set of TXLSExportOption;

  TXLSExportColor = (xleSetBkColor, xleSetFontColor, xleHideText);
  TXLSExportColors = set of TXLSExportColor;

  { Defines vertical alignment possibilty }
  TXLSVerticalAlignment = (vaNone, vaTop, vaCenter, vaBottom);
  { Defines event for custom definition of field with picture }
  TGetFieldType = procedure(Sender: TObject; Field: TField; var FieldType: integer) of object;
  { Defines event for custom preparation of the picture before exporting }
  TPreparePictureEvent = procedure(Sender: TObject; Row, Column: integer; Field: TField; Picture: TPicture) of object;

  { setting for Databar used by TrXLSExport event }
  TDatabarSetting = record
    MinValue: double;
    MaxValue: double;
    FillColor: TColor;
    BorderColor: TColor;
    ShowValue: boolean;
  end;

  { Defines event for custom preparation of the Databar before exporting }
  TPrepareDatabarEvent = procedure(Sender: TObject; Field: TField; var DatabarSetting: TDatabarSetting) of object;

  { setting for ColorScale used by TrXLSExport event }
  TColorScaleSetting = record
    UseMidValue: boolean;
    MinValue: double;
    MidValue: double;
    MaxValue: double;
    MinColor: TColor;
    MidColor: TColor;
    MaxColor: TColor;
  end;

  { Defines event for custom preparation of the ColorScale before exporting }
  TPrepareColorScaleEvent = procedure(Sender: TObject; Field: TField; var ColorScaleSetting: TColorScaleSetting) of object;

  { setting for IconSet used by TrXLSExport event }
  TIconSetSetting = record
    IconSetType: integer;
    UseThresholds: boolean;
    ThresholdValues: array [1 .. 4] of double;
    ShowIconOnly: boolean;
  end;

  { Defines event for custom preparation of the IconSet before exporting }
  TPrepareIconSetEvent = procedure(Sender: TObject; Field: TField; var IconSetSetting: TIconSetSetting) of object;

  { non visual component which provides functionality for
    exporting of any data directly to XLS file using OLE }
  TrXLSExport = class(TComponent)
  private
    FApp: OleVariant;
    FAppOpened: boolean;
    FWorkbook: OleVariant;
    FWorksheet, FOrigWorksheet: OleVariant;
    FRange: OleVariant;
    FDataRowCount: integer;
    FDataColCount: integer;
    FOptions: TXLSExportOptions;
    FOptionColors: TXLSExportColors;
    FVerticalAlignment: TXLSVerticalAlignment;
    FStartRow: integer;
    FStartCol: integer;
    FExportFile: string;
    FSetFontName: string;
    FSetFontSize: integer;
    FMaxColWidth: integer;
    FMaxRowHeight: integer;
    FPageZoom: integer;
    FWindowZoom: integer;
    FDefaultDateFormat: string;
    FDefaultTimeFormat: string;
    FDefaultNumFormat: string;
    FExportStructureOnly: boolean;
    FActionAfter: TActionAfter;
    FPictureMaxWidth: integer;
    FPictureMaxHeight: integer;
    FPictureCellMargin: integer;
    FPictureUseShadow: boolean;
    FGetFieldType: TGetFieldType;
    FOnPreparePictureEvent: TPreparePictureEvent;
    FOnPrepareDatabarEvent: TPrepareDatabarEvent;
    FOnPrepareColorScaleEvent: TPrepareColorScaleEvent;
    FOnPrepareIconSetEvent: TPrepareIconSetEvent;
    FOnBeforeExport: TNotifyEvent;
    FOnAfterExport: TNotifyEvent;
    FUseDisableControl: boolean;
    FShowProgressBar: boolean;
    FOnProgressEvent: TProgressEvent;
    FColWidthPixelRatio: double;
    FRowHeightPixelRatio: double;
    procedure SetStartRow(Value: integer);
    procedure SetStartCol(Value: integer);
    procedure ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
    procedure DoAbort(Sender: TObject);
    function PasteRowData: boolean;
  protected
    // GDS: TObject;
    SourceType: TSourceType;
    SourceObject: TObject;
    Dataset: TDataSet;
    DBGrid: TCustomDBGrid;
    RecView: TrDBRecView;
    TreeView: TCustomTreeView;
    StringGrid: TrStringGridEd;
    FN, FAsColor: TStrings;
    RowIdx, RecExported, PicExported, RecIdx, RecCount, FirstDataRow, LastDataRow: integer;
    Aborted: boolean;
    AppNotExist: boolean;
    Pic: TPicture;
    procedure Loaded; override;
    function GetField(idx: integer): TField;
    function UsePictures: boolean;
    function FieldAsPicture(Fld: TField): boolean;
    function UseColors: boolean;
    function FieldAsColor(Fld: TField): boolean;
    function FieldAsDatabar(Fld: TField): boolean;
    procedure LoadDatabarProps(Fld: TField; var DatabarSetting: TDatabarSetting);
    function FieldAsColorScale(Fld: TField): boolean;
    procedure LoadColorScaleProps(Fld: TField; var ColorScaleSetting: TColorScaleSetting);
    function FieldAsIconSet(Fld: TField): boolean;
    procedure LoadIconSetProps(Fld: TField; var IconSetSetting: TIconSetSetting);
    function PrepareData(s: string): string;
    procedure ExpLabels;
    function ExpData: boolean;
    procedure ExpSum(rows: integer);
    procedure SetColFormat;
    procedure SetColProps;
    procedure SetRowProps;
    procedure ExpTreeItems;
    procedure ExpGridRows;
    procedure ShowProgresInt;
    function ExportDBTableInt(ASourceType: TSourceType; O: TObject; AutoCloseApp: boolean; const ExportToSheetName: string): boolean;
    function ExportPicture(Field: TField; Row, Col: integer): string; virtual;
  public
    { Data array contain data which was read from cells or which should be stored to cells }
    Data: Variant;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }

    { test if Excel App is available }
    function IsAppExist: boolean;
    { run Excel and open new or existing file }
    function OpenFile(NewFile: boolean; const OpenFileName: string = ''; CanReadOnly: boolean = true): boolean;
    { close Excel and optionally save file }
    function SaveCloseFile(CloseApp, SaveFile: boolean; const SaveFileName: string = ''; FileFormat: integer = 0): boolean;
    { Disconnect from current excel application, Excel keep opened }
    procedure Disconnect;
    { set visibility of Excel application }
    procedure SetVisibility(Visible: boolean);
    { return local NumericFormat string for short date in Excel }
    function GetShortDateFormat: string;
    { return local NumericFormat string for short date in Excel }
    function ConvertDateFormat(FieldDateFormat: string): string;

    { return number of sheets }
    function GetSheetsCount: integer;
    { add new sheet and set its name or select active sheet according SheetName }
    function SelectSheet(NewSheet: boolean; const SheetName: string = ''; SheetIndex: integer = 0): boolean;
    { delete all unselected sheets or delete one sheet according SheetName }
    function DeleteSheet(AllUnselected: boolean; const SheetName: string = ''): boolean;
    { remember actual active sheet }
    procedure RememberActiveSheet;
    { select active sheet after open or after call RememberActiveSheet }
    procedure SelectRememberSheet;
    { select sheet by name defined in range name and return range name without sheet: sheet!range }
    function ActivateRangeSheet(var RangeName: string): boolean;

    { set Range object }
    procedure SetCurrentRange(Row, Col, RowCount, ColCount: integer);
    { set starting row for export, select first empty row if AddToExisting is true }
    procedure SetStartPos(ClearExistingRows: boolean = true; AddToExistingRows: boolean = false);
    { return value of one cell }
    function GetCellValue(Row, Col: integer): Variant;
    { set value of one cell }
    procedure SetCellValue(Row, Col: integer; Value: Variant; const NumberFormat: string = '');
    { set value of RangeName }
    procedure SetRangeValue(RangeName: string; Value: Variant; const NumberFormat: string = '');

    { prepare Data array according RangeName size }
    procedure SetDataSizeFromRange(RangeName: string);
    { prepare Data array according RowCount and ColCount }
    procedure SetDataSize(RowCount, ColCount: integer);
    { read values from cells to Data array,
      StartRow and StartCol + DataRowCount and DataColCount is used to define source range }
    function CopyCellsToData: boolean;
    { copy values from cells to Clipboard,
      StartRow and StartCol + DataRowCount and DataColCount is used to define source range }
    function CopyCellsToClipboard: boolean;
    { read values from UsedRange of selected Sheet to Data array, size is set according UsedRange }
    function CopyUsedRangeToData: boolean;
    { read values from UsedRange of selected Sheet to Clipboard }
    function CopyUsedRangeToClipboard: boolean;
    { read values from UsedRange of selected Sheet to Data array, size is set according UsedRange }
    function CopyCurrentRegionToData: boolean;
    { read values from UsedRange of selected Sheet to Clipboard }
    function CopyCurrentRegionToClipboard: boolean;
    { write values from Data array to cells,
      StartRow and StartCol + DataRowCount and DataColCount is used to define target range }
    function PasteDataToCells(SetAutoFilter: boolean = false): boolean;
    { paste values from Clipboard to cells,
      StartRow and StartCol is used to define target range }
    function PasteClipboardToCells(SetAutoFilter: boolean = false): boolean;

    { switch on FreezePanes in selected sheet, if Row and Col is 0 FreezePanes is switched off }
    procedure FreezePanes(Row, Col: integer);
    { set width of one column }
    procedure SetColumnWidth(Col: integer; ColWidth: integer);
    { set height of one row }
    procedure SetRowHeight(Row: integer; RowHeight: integer);
    { set automatic width of one column }
    procedure ColumnAutoFit(Col: integer; MaxColWidth: integer);
    { set automatic height of one row }
    procedure RowAutoFit(Row: integer; MaxRowHeight: integer);
    { set automatic width of all columns and rows in selected sheet }
    procedure UsedRangeAutoFit(MaxColWidth: integer = 0; MaxRowHeight: integer = 0; SetColumnWidth: boolean = true; SetRowHeight: boolean = true);
    { set Number format and/or Font for cells in square defined by Row/Col/Count }
    procedure FormatCells(Row, Col, RowCount, ColCount: integer; const NumberFormat: string; const FontName: string = ''; FontSize: integer = 0;
      Frm: TXLSCellFormats = [cfNone]);
    { set Number format and/or Font for cells in square defined by Row/Col/Count }
    procedure BorderCells(Row, Col, RowCount, ColCount: integer; Border: integer = xlAllBorders; LineColor: TColor = clNone; LineWidth: integer = xlThin;
      LineStyle: integer = xlContinuous);
    { set condition format as Databar for cells in square defined by Row/Col/Count }
    procedure DatabarCells(Row, Col, RowCount, ColCount: integer; DatabarSetting: TDatabarSetting);
    { set condition format as ColorScale for cells in square defined by Row/Col/Count }
    procedure ColorScaleCells(Row, Col, RowCount, ColCount: integer; ColorScaleSetting: TColorScaleSetting);
    { set condition format as IconSet for cells in square defined by Row/Col/Count }
    procedure IconSetCells(Row, Col, RowCount, ColCount: integer; IconSetSetting: TIconSetSetting);

    { set page zoom or number of pages for printout }
    procedure SetPageZoom(Zoom: integer; FitToPagesTall: integer = 1; FitToPagesWide: integer = 1; Orientation: byte = 0);
    { set window zoom }
    procedure SetWindowZoom(Zoom: integer);

    { execute creating export for DBGrid (columns collection is used) }
    function ExportDBTable(DBGrid: TCustomDBGrid; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean; overload;
    { execute creating export for Dataset (fields collection is used) }
    function ExportDBTable(Dataset: TDataSet; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean; overload;
    { execute creating export for one record from rDBRecView (rows/fields collection is used) }
    function ExportDBTable(rDBRecView: TrDBRecView; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean; overload;
    { execute creating export for TreeView (Items collection is used) }
    function ExportDBTable(TreeView: TCustomTreeView; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean; overload;
    { execute creating export for rStringGrid (columns and rows collection is used) }
    function ExportDBTable(rStringGridEd: TrStringGridEd; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean; overload;

    { indicate size of Data array }
    property DataRowCount: integer read FDataRowCount;
    { indicate size of Data array }
    property DataColCount: integer read FDataColCount;
    { return Excel application OLE object }
    property App: OleVariant read FApp;
    { indicate whether excel app is opened by component }
    property AppOpened: boolean read FAppOpened;
    { return active Workbook OLE object }
    property Workbook: OleVariant read FWorkbook;
    { return active Worksheet OLE object }
    property Worksheet: OleVariant read FWorksheet;
    { return current Range OLE object }
    property Range: OleVariant read FRange;
    { define that only structure (name of fields) will be exported }
    property ExportStructureOnly: boolean read FExportStructureOnly write FExportStructureOnly;
    { define pixels to column width ratio }
    property ColWidthPixelRatio: double read FColWidthPixelRatio write FColWidthPixelRatio;
    { define pixels to row height ratio }
    property RowHeightPixelRatio: double read FRowHeightPixelRatio write FRowHeightPixelRatio;
  published
    { Defines first row for export }
    property StartRow: integer read FStartRow write SetStartRow default 1;
    { Defines first collumn for export }
    property StartCol: integer read FStartCol write SetStartCol default 1;
    { Defines output file name, if missing new file is opened }
    property ExportFile: string read FExportFile write FExportFile;
    { Defines FontName which will be set in output file }
    property SetFontName: string read FSetFontName write FSetFontName;
    { Defines FontSize which will be set in output file }
    property SetFontSize: integer read FSetFontSize write FSetFontSize default 0;
    { Defines default format for date fields }
    property DefaultDateFormat: string read FDefaultDateFormat write FDefaultDateFormat;
    { Defines default format for time fields }
    property DefaultTimeFormat: string read FDefaultTimeFormat write FDefaultTimeFormat;
    { Defines default format for numeric fields }
    property DefaultNumFormat: string read FDefaultNumFormat write FDefaultNumFormat;
    { Defines maximal width of columns after AutoFit }
    property MaxColWidth: integer read FMaxColWidth write FMaxColWidth default 100;
    { Defines maximal height of rows after AutoFit }
    property MaxRowHeight: integer read FMaxRowHeight write FMaxRowHeight default 30;
    { Defines vertical alignment of all rows }
    property VerticalAlignment: TXLSVerticalAlignment read FVerticalAlignment write FVerticalAlignment default vaNone;
    { Defines standard page zoom }
    property PageZoom: integer read FPageZoom write FPageZoom default 0;
    { Defines standard window zoom }
    property WindowZoom: integer read FWindowZoom write FWindowZoom default 0;

    { Defines maximal width of the picture, bigger picture will be resized }
    property PictureMaxWidth: integer read FPictureMaxWidth write FPictureMaxWidth default 200;
    { Defines maximal height of the picture, bigger picture will be resized }
    property PictureMaxHeight: integer read FPictureMaxHeight write FPictureMaxHeight default 200;
    { Defines margin around the picture in the cell }
    property PictureCellMargin: integer read FPictureCellMargin write FPictureCellMargin default 5;
    { Defines usage of shadow around the picture in the cell }
    property PictureUseShadow: boolean read FPictureUseShadow write FPictureUseShadow default false;
    { Defines options for ExportDBTable function }
    property Options: TXLSExportOptions read FOptions write FOptions default[xleExportTitles, xleBoldTitles, xleOnlyVisibleFields, xleAddSumRow,
      xleSetCellFormat, xleSetColumnWidth, xleSetRowHeight];
    property OptionColors: TXLSExportColors read FOptionColors write FOptionColors default[];
    { Defines action after exporting (None, Show, Print) }
    property ActionAfter: TActionAfter read FActionAfter write FActionAfter default aaShow;
    { Defines event for user defined action }
    property OnBeforeExport: TNotifyEvent read FOnBeforeExport write FOnBeforeExport;
    { Defines event for user defined action }
    property OnAfterExport: TNotifyEvent read FOnAfterExport write FOnAfterExport;
    { indicate whether DisableControls is used during dataset scan }
    property UseDisableControl: boolean read FUseDisableControl write FUseDisableControl default true;
    { indicate whether progress bar is shown }
    property ShowProgressBar: boolean read FShowProgressBar write FShowProgressBar default false;
    { Defines events for custom specification of field type (used for exporting of Pictures, Colors, databar, ColorScales }
    property GetFieldType: TGetFieldType read FGetFieldType write FGetFieldType;
    { Defines event used periodically during exporting of data }
    property OnProgressEvent: TProgressEvent read FOnProgressEvent write FOnProgressEvent;
    { Defines event for custom preparation of the picture before exporting }
    property OnPreparePictureEvent: TPreparePictureEvent read FOnPreparePictureEvent write FOnPreparePictureEvent;
    { Defines event for custom preparation of the databar format before exporting }
    property OnPrepareDatabarEvent: TPrepareDatabarEvent read FOnPrepareDatabarEvent write FOnPrepareDatabarEvent;
    { Defines event for custom preparation of the ColorScale format before exporting }
    property OnPrepareColorScaleEvent: TPrepareColorScaleEvent read FOnPrepareColorScaleEvent write FOnPrepareColorScaleEvent;
    { Defines event for custom preparation of the IconSet format before exporting }
    property OnPrepareIconSetEvent: TPrepareIconSetEvent read FOnPrepareIconSetEvent write FOnPrepareIconSetEvent;
  end;

procedure Register; {< @exclude }

implementation

uses rdlg, rtool, rdbtool, rpictool, rstring;
{$R *.res}
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrXLSExport]);
end;

constructor TrXLSExport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Pic:=TPicture.Create;
  AppNotExist:=false;
  FStartCol:=1;
  FStartRow:=1;
  FActionAfter:=aaShow;
  FMaxColWidth:=100;
  FMaxRowHeight:=30;
  FShowProgressBar:=false;
  FUseDisableControl:=true;
  FOptions:=[xleExportTitles, xleBoldTitles, xleOnlyVisibleFields, xleAddSumRow, xleSetCellFormat, xleSetColumnWidth, xleSetRowHeight];
  FPictureMaxWidth:=200;
  FPictureMaxHeight:=200;
  FPictureCellMargin:=5;
  FPictureUseShadow:=false;
  FColWidthPixelRatio:=cColWidthPixelRatioDefault;
  FRowHeightPixelRatio:=cRowHeightPixelRatioDefault;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrXLSExport.Destroy;
begin
  FreeAndNil(Pic);
  if FAppOpened then
    try
      if FApp.Visible then
        // disconnect to keep excel opened
        Disconnect
      else
        SaveCloseFile(true, false);
    except
    end;
  inherited Destroy;
end;

procedure TrXLSExport.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrXLSExport.IsAppExist: boolean;
begin
  Result:=FAppOpened;
  if not Result and not AppNotExist then
  begin
    try
      FApp:=CreateOLEObject('Excel.Application');
      FApp.DisplayAlerts:=false;
      FApp.Quit;
      FApp:=Unassigned;
      Result:=true;
    except
      AppNotExist:=true;
    end;
  end;
end;

function TrXLSExport.OpenFile(NewFile: boolean; const OpenFileName: string = ''; CanReadOnly: boolean = true): boolean;
begin
  Result:=false;
  if not FAppOpened then
  begin
    try
      FApp:=CreateOLEObject('Excel.Application');
      FAppOpened:=true;
    except
      ShowMessage(sMSExcelNotFound);
      Exit;
    end;
  end;

  try
    if NewFile or not FileExists(OpenFileName) then
      FWorkbook:=FApp.Workbooks.Add
    else
      FWorkbook:=FApp.Workbooks.Open(OpenFileName, Notify:=CanReadOnly);
    FWorksheet:=FWorkbook.ActiveSheet;
    RememberActiveSheet;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.SaveCloseFile(CloseApp: boolean; SaveFile: boolean; const SaveFileName: string = ''; FileFormat: integer = 0): boolean;
begin
  Result:=false;
  if FAppOpened then
  begin
    if SaveFile then
    begin
      try
        // set format according extension
        if (FileFormat = 0) then
        begin
          if SameText(ExtractFileExt(SaveFileName), '.xlsx') then
            FileFormat:=xlOpenXMLWorkbook;
          if SameText(ExtractFileExt(SaveFileName), '.xlsm') then
            FileFormat:=xlOpenXMLWorkbookMacroEnabled;
        end;
        if (SaveFileName <> '') and not SameText(FWorkbook.Name, ExtractFileName(SaveFileName)) then
        begin
          if FileFormat = 0 then
            FWorkbook.SaveAs(SaveFileName, xlWorkbookNormal)
          else
            FWorkbook.SaveAs(SaveFileName, FileFormat);
        end
        else
          FWorkbook.Save;
      except
        on E: Exception do
        begin
          ShowErr(sMSExcelError, E);
          Exit;
        end;
      end;
    end;
    if CloseApp then
    begin
      FApp.DisplayAlerts:=false;
      FApp.Quit;
      Disconnect;
    end;
    Result:=true;
  end;
end;

procedure TrXLSExport.Disconnect;
begin
  FAppOpened:=false;
  FApp:=Unassigned;
  FWorkbook:=Unassigned;
  FWorksheet:=Unassigned;
  FRange:=Unassigned;
end;

procedure TrXLSExport.SetVisibility(Visible: boolean);
begin
  if FAppOpened then
  begin
    FApp.Visible:=Visible;
  end;
end;

function TrXLSExport.GetSheetsCount: integer;
begin
  if FAppOpened then
    Result:=FWorkbook.WorkSheets.Count
  else
    Result:=0;
end;

procedure TrXLSExport.RememberActiveSheet;
begin
  FOrigWorksheet:=FWorkbook.ActiveSheet;
end;

procedure TrXLSExport.SelectRememberSheet;
begin
  FOrigWorksheet.Select;
end;

function TrXLSExport.SelectSheet(NewSheet: boolean; const SheetName: string = ''; SheetIndex: integer = 0): boolean;
var
  i: integer;
begin
  Result:=false;
  try
    if NewSheet then
    begin
      FWorksheet:=FWorkbook.WorkSheets.Add;
      if SheetName <> '' then
        FWorksheet.Name:=SheetName;
      Result:=true;
    end
    else
    begin
      for i:=1 to FWorkbook.WorkSheets.Count do
        if SameText(FWorkbook.WorkSheets[i].Name, SheetName) or (i = SheetIndex) then
        begin
          FWorksheet:=FWorkbook.WorkSheets[i];
          FWorksheet.Select;
          Result:=true;
        end;
    end;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.DeleteSheet(AllUnselected: boolean; const SheetName: string = ''): boolean;
var
  i: integer;
  Alerts: boolean;
begin
  Result:=false;
  Alerts:=FApp.DisplayAlerts;
  FApp.DisplayAlerts:=false;
  try
    try
      for i:=FWorkbook.WorkSheets.Count downto 1 do
      begin
        if SameText(FWorkbook.WorkSheets[i].Name, SheetName) or (AllUnselected and not SameText(FWorkbook.WorkSheets[i].Name, FWorksheet.Name)) then
        begin
          FWorkbook.WorkSheets[i].Delete;
          Result:=true;
        end;
      end;
    except
      on E: Exception do
        ShowErr(sMSExcelError, E);
    end;
  finally
    FApp.DisplayAlerts:=Alerts;
  end;
end;

procedure TrXLSExport.SetCurrentRange(Row, Col, RowCount, ColCount: integer);
begin
  try
    FRange:=FWorksheet.Range[FWorksheet.Cells[Row, Col], FWorksheet.Cells[Row - 1 + RowCount, Col - 1 + ColCount]];
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.SetStartPos(ClearExistingRows: boolean = true; AddToExistingRows: boolean = false);
begin
  try
    FStartRow:=1;
    // FStartCol:=1;
    FRange:=FWorksheet.UsedRange;
    if ClearExistingRows then
    begin
      FRange.Clear;
      Exit;
    end;
    if AddToExistingRows and ((FRange.rows.Count > 1) or (FRange.Columns.Count > 1) or not VarIsEmpty(FWorksheet.Cells[FRange.Row, FRange.Column].Value)) then
      FStartRow:=FRange.Row + FRange.rows.Count;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.GetCellValue(Row, Col: integer): Variant;
begin
  try
    Result:=FWorksheet.Cells[Row, Col].Value;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.SetCellValue(Row, Col: integer; Value: Variant; const NumberFormat: string = '');
begin
  try
    if VarIsStr(Value) then
      FWorksheet.Cells[Row, Col].Value:=VarToStr(Value)
    else
      FWorksheet.Cells[Row, Col].Value:=Value;
    if NumberFormat <> '' then
      FWorksheet.Cells[Row, Col].NumberFormat:=NumberFormat;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.ActivateRangeSheet(var RangeName: string): boolean;
var
  shname: string;
begin
  Result:=true;
  if Pos('!', RangeName) > 0 then
  begin
    shname:=rtool.RemoveStrPart(RangeName, '!');
    Result:=SelectSheet(false, shname);
  end;
end;

procedure TrXLSExport.SetRangeValue(RangeName: string; Value: Variant; const NumberFormat: string = '');
begin
  try
    FWorksheet.Range[RangeName]:=Value;
    if NumberFormat <> '' then
      FWorksheet.Range[RangeName].NumberFormat:=NumberFormat;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.SetDataSizeFromRange(RangeName: string);
begin
  FRange:=FWorksheet.Range[RangeName];
  FStartRow:=FRange.Row;
  FStartCol:=FRange.Column;
  SetDataSize(FRange.rows.Count, FRange.Columns.Count);
end;

procedure TrXLSExport.SetDataSize(RowCount, ColCount: integer);
begin
  FDataRowCount:=RowCount;
  FDataColCount:=ColCount;
  Data:=VarArrayCreate([1, RowCount, 1, ColCount], varVariant);
end;

function TrXLSExport.CopyCellsToData: boolean;
begin
  Result:=false;
  try
    SetCurrentRange(FStartRow, FStartCol, FDataRowCount, FDataColCount);
    Data:=FRange.Value;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.CopyCellsToClipboard: boolean;
begin
  Result:=false;
  try
    SetCurrentRange(FStartRow, FStartCol, FDataRowCount, FDataColCount);
    FRange.Copy;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.PasteDataToCells(SetAutoFilter: boolean = false): boolean;
begin
  Result:=false;
  try
    SetCurrentRange(FStartRow, FStartCol, FDataRowCount, FDataColCount);
    FRange.Value:=Data;
    if SetAutoFilter then
      FRange.Autofilter;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.PasteClipboardToCells(SetAutoFilter: boolean = false): boolean;
begin
  Result:=false;
  try
    FRange:=FWorksheet.Cells[FStartRow, FStartCol];
    FRange.PasteSpecial;
    if SetAutoFilter then
      FRange.Autofilter;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.CopyUsedRangeToData: boolean;
begin
  Result:=false;
  try
    FRange:=FWorksheet.UsedRange;
    FStartRow:=FRange.Row;
    FStartCol:=FRange.Column;
    SetDataSize(FRange.rows.Count, FRange.Columns.Count);
    CopyCellsToData;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.CopyUsedRangeToClipboard: boolean;
begin
  Result:=false;
  try
    FRange:=FWorksheet.UsedRange;
    FRange.Copy;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.CopyCurrentRegionToData: boolean;
begin
  Result:=false;
  try
    FRange:=FWorksheet.Cells[FStartRow, FStartCol].CurrentRegion;
    SetDataSize(FRange.rows.Count, FRange.Columns.Count);
    CopyCellsToData;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

function TrXLSExport.CopyCurrentRegionToClipboard: boolean;
begin
  Result:=false;
  try
    FRange:=FWorksheet.Cells[FStartRow, FStartCol].CurrentRegion;
    FRange.Copy;
    Result:=true;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.FreezePanes(Row, Col: integer);
begin
  try
    FApp.ActiveWindow.FreezePanes:=false;
    if (Row > 0) or (Col > 0) then
    begin
      FWorksheet.Cells[Row + 1, Col + 1].Select;
      FApp.ActiveWindow.FreezePanes:=true;
    end;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.SetColumnWidth(Col: integer; ColWidth: integer);
begin
  FWorksheet.Columns[Col].ColumnWidth:=ColWidth;
end;

procedure TrXLSExport.SetRowHeight(Row: integer; RowHeight: integer);
begin
  FWorksheet.rows[Row].RowHeight:=RowHeight;
end;

procedure TrXLSExport.ColumnAutoFit(Col: integer; MaxColWidth: integer);
begin
  FWorksheet.Columns[Col].EntireColumn.AutoFit;
  if (MaxColWidth > 0) and (FWorksheet.Columns[Col].ColumnWidth > MaxColWidth) then
    FWorksheet.Columns[Col].ColumnWidth:=MaxColWidth;
end;

procedure TrXLSExport.RowAutoFit(Row: integer; MaxRowHeight: integer);
begin
  FWorksheet.rows[Row].EntireRow.AutoFit;
  if (MaxRowHeight > 0) and (FWorksheet.rows[Row].RowHeight > MaxRowHeight) then
    FWorksheet.rows[Row].RowHeight:=MaxRowHeight;
end;

procedure TrXLSExport.UsedRangeAutoFit(MaxColWidth: integer = 0; MaxRowHeight: integer = 0; SetColumnWidth: boolean = true; SetRowHeight: boolean = true);
var
  i: integer;
begin
  FRange:=FWorksheet.UsedRange;
  if SetColumnWidth then
  begin
    FRange.EntireColumn.AutoFit;
    if (MaxColWidth > 0) then
      for i:=1 to FRange.Columns.Count do
        if (FRange.Columns[i].ColumnWidth > MaxColWidth) then
          FRange.Columns[i].ColumnWidth:=MaxColWidth;
  end;
  if SetRowHeight then
  begin
    FRange.EntireRow.AutoFit;
    if (MaxRowHeight > 0) then
      for i:=1 to FRange.rows.Count do
        if (FRange.rows[i].RowHeight > MaxRowHeight) then
          FRange.rows[i].RowHeight:=MaxRowHeight;
  end;
end;

procedure TrXLSExport.FormatCells(Row, Col, RowCount, ColCount: integer; const NumberFormat: string; const FontName: string = ''; FontSize: integer = 0;
  Frm: TXLSCellFormats = [cfNone]);
begin
  try
    SetCurrentRange(Row, Col, RowCount, ColCount);
    if NumberFormat <> '' then
      FRange.NumberFormat:=NumberFormat;
    if FontName <> '' then
      FRange.Font.Name:=FontName;
    if FontSize <> 0 then
      FRange.Font.Size:=FontSize;
    if not(cfNone in Frm) then
    begin
      FRange.Font.Bold:=cfBold in Frm;
      FRange.Font.Italic:=cfItalic in Frm;
      FRange.Font.Underline:=cfUnderline in Frm;
    end;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.DatabarCells(Row, Col, RowCount, ColCount: integer; DatabarSetting: TDatabarSetting);
const
  xlConditionValueNumber = 0;
  {
    xlConditionValueAutomaticMax	7  The longest data bar is proportional to the maximum value in the range.
    xlConditionValueAutomaticMin	6  The shortest data bar is proportional to the minimum value in the range.
    xlConditionValueFormula	      4  Formula is used.
    xlConditionValueHighestValue	2  Highest value from the list of values.
    xlConditionValueLowestValue	  1  Lowest value from the list of values.
    xlConditionValueNone	       -1  No conditional value.
    xlConditionValueNumber	      0  Number is used.
    xlConditionValuePercent	      3  Percentage is used.
    xlConditionValuePercentile	  5  Percentile is used.
    }
  xlDataBarBorderSolid = 1;
  {
    xlDataBarBorderNone	0	The data bar has no border.
    xlDataBarBorderSolid	1	The data bar has a solid border.
    }

var
  Db: OleVariant;
begin
  try
    SetCurrentRange(Row, Col, RowCount, ColCount);
    Db:=FRange.FormatConditions.AddDataBar;
    if (DatabarSetting.MinValue <> 0) or (DatabarSetting.MaxValue <> 0) then
    begin
      Db.MinPoint.Modify(xlConditionValueNumber, DatabarSetting.MinValue);
      Db.MaxPoint.Modify(xlConditionValueNumber, DatabarSetting.MaxValue);
    end;
    if DatabarSetting.FillColor <> clNone then
      Db.BarColor.Color:=DatabarSetting.FillColor;
    if DatabarSetting.BorderColor <> clNone then
    begin
      Db.BarBorder. Type :=xlDataBarBorderSolid;
      Db.BarBorder.Color.Color:=DatabarSetting.BorderColor;
    end;
    Db.ShowValue:=DatabarSetting.ShowValue;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.ColorScaleCells(Row, Col, RowCount, ColCount: integer; ColorScaleSetting: TColorScaleSetting);
const
  xlConditionValueNumber = 0;
var
  Db: OleVariant;
  i: integer;
  SetValue: boolean;

  procedure SetItem(Value: double; Color: TColor);
  begin
    if SetValue then
    begin
      Db.ColorScaleCriteria[i]. Type :=xlConditionValueNumber;
      Db.ColorScaleCriteria[i].Value:=Value;
    end;
    if Color <> clNone then
      Db.ColorScaleCriteria[i].FormatColor.Color:=Color;
  end;

begin
  try
    SetCurrentRange(Row, Col, RowCount, ColCount);
    SetValue:=(ColorScaleSetting.MinValue <> 0) or (ColorScaleSetting.MaxValue <> 0);
    if ColorScaleSetting.UseMidValue then
      i:=3
    else
      i:=2;
    Db:=FRange.FormatConditions.AddColorScale(i);
    i:=1;
    SetItem(ColorScaleSetting.MinValue, ColorScaleSetting.MinColor);
    Inc(i);
    if ColorScaleSetting.UseMidValue then
    begin
      SetItem(ColorScaleSetting.MidValue, ColorScaleSetting.MidColor);
      Inc(i);
    end;
    SetItem(ColorScaleSetting.MaxValue, ColorScaleSetting.MaxColor);
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.IconSetCells(Row, Col, RowCount, ColCount: integer; IconSetSetting: TIconSetSetting);
const
  xlConditionValueNumber = 0;
  xlGreater = 5;
  xlGreaterEqual = 7;
var
  Db, iconset: OleVariant;
  i: integer;
begin
  try
    SetCurrentRange(Row, Col, RowCount, ColCount);
    Db:=FRange.FormatConditions.AddIconSetCondition;
    if IconSetSetting.IconSetType > 0 then
      try
        // i:=FWorkbook.IconSets.count;
        iconset:=FWorkbook.IconSets.Item[IconSetSetting.IconSetType];
        Db.iconset:=iconset;
        // Db.IconSet:=FWorkbook.IconSets.Item[IconSetSetting.IconSetType];
      except
        // no error
      end;
    if IconSetSetting.UseThresholds then
    begin
      for i:=2 to Min(Db.IconCriteria.Count, 5) do
      begin
        try
          Db.IconCriteria[i]. Type :=xlConditionValueNumber;
          Db.IconCriteria[i].Operator:=xlGreaterEqual;
          Db.IconCriteria[i].Value:=IconSetSetting.ThresholdValues[i - 1];
        except
          // no error
        end;
      end;
    end;
    Db.ShowIconOnly:=IconSetSetting.ShowIconOnly;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.BorderCells(Row, Col, RowCount, ColCount: integer; Border: integer = xlAllBorders; LineColor: TColor = clNone;
  LineWidth: integer = xlThin; LineStyle: integer = xlContinuous);
begin
  SetCurrentRange(Row, Col, RowCount, ColCount);
  if Border = xlAllBorders then
  begin
    FRange.Borders.LineStyle:=LineStyle;
    if LineColor <> clNone then
      FRange.Borders.Color:=LineColor;
    FRange.Borders.Weight:=LineWidth;
  end
  else
  begin
    FRange.Borders.Item[Border].LineStyle:=LineStyle;
    if LineColor <> clNone then
      FRange.Borders.Item[Border].Color:=LineColor;
    FRange.Borders.Item[Border].Weight:=LineWidth;
  end;
end;

procedure TrXLSExport.SetPageZoom(Zoom: integer; FitToPagesTall: integer = 1; FitToPagesWide: integer = 1; Orientation: byte = 0);
begin
  try
    if Zoom > 0 then
      FWorksheet.PageSetup.Zoom:=Zoom
    else
    begin
      FWorksheet.PageSetup.Zoom:=false;
      FWorksheet.PageSetup.FitToPagesTall:=FitToPagesTall;
      FWorksheet.PageSetup.FitToPagesWide:=FitToPagesWide;
    end;
    if Orientation > 0 then
      FWorksheet.PageSetup.Orientation:=Orientation;
  except
    on E: Exception do
      ShowErr(sMSExcelError, E);
  end;
end;

procedure TrXLSExport.SetWindowZoom(Zoom: integer);
begin
  if Zoom <> 0 then
    FApp.ActiveWindow.Zoom:=Zoom;
end;

function TrXLSExport.GetShortDateFormat: string;
  function GetVal(Code, Len: integer): string;
  begin
    Result:=StringOfChar(VarToStr(FApp.International[Code])[1], Len);
  end;

var
  d, m, y: integer;
  DateSep: string;
begin
  Result:='';
  if not FAppOpened then
    Exit;
  d:=1;
  m:=1;
  y:=2;
  if FApp.International[xlDayLeadingZero] then
    Inc(d);
  if FApp.International[xlMonthLeadingZero] then
    Inc(m);
  if FApp.International[xl4DigitYears] then
    Inc(y, 2);
  DateSep:=VarToStr(FApp.International[xlDateSeparator]);
  Result:=Format('%1:s%0:s%2:s%0:s%3:s', [DateSep, GetVal(xlDayCode, d), GetVal(xlMonthCode, m), GetVal(xlYearCode, y)]);
end;

function TrXLSExport.ConvertDateFormat(FieldDateFormat: string): string;
begin
  Result:=FieldDateFormat;
  if (Result = '') or not FAppOpened then
    Exit;
  Result:=rtool.ReplaceStr('d', VarToStr(FApp.International[xlDayCode]), Result);
  Result:=rtool.ReplaceStr('m', VarToStr(FApp.International[xlMonthCode]), Result);
  Result:=rtool.ReplaceStr('y', VarToStr(FApp.International[xlYearCode]), Result);
  Result:=rtool.ReplaceStr('/', VarToStr(FApp.International[xlDateSeparator]), Result);
end;

procedure TrXLSExport.SetColFormat;
var
  c, Col: integer;
  Fld: TField;
  Frm: string;
begin
  for c:=0 to FN.Count - 1 do
  begin
    Frm:='';
    case SourceType of
      stGrid, stDataset, stRecView:
        begin
          Fld:=GetField(c);
          if Fld is TNumericField then
          begin
            Frm:=TNumericField(Fld).DisplayFormat;
            if (Frm = '') and (Fld is TFloatField) then
              Frm:=FDefaultNumFormat;
            Frm:=UpdateToDecimalSeparator(Frm);
          end
          else
            if Fld is TTimeField then
            begin
              Frm:=TTimeField(Fld).DisplayFormat;
              if (Frm = '') then
                Frm:=FDefaultTimeFormat;
            end
            else
              if (Fld is TDateTimeField) or (Fld is TSQLTimeStampField) then
              begin
                if (Fld is TDateTimeField) then
                  Frm:=ConvertDateFormat(TDateTimeField(Fld).DisplayFormat)
                else
                  Frm:=ConvertDateFormat(TSQLTimeStampField(Fld).DisplayFormat);
                if (Frm = '') then
                  Frm:=ConvertDateFormat(FDefaultDateFormat);
                if (Frm = '') then
                  Frm:=GetShortDateFormat;
              end;
        end;
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          if Col >= 0 then
          begin
            case StringGrid.Columns[Col].DataType of
              sgdNumber, sgdInteger:
                begin
                  Frm:=StringGrid.Columns[Col].NumFormat;
                  if (Frm = '') then
                    Frm:=FDefaultNumFormat;
                  Frm:=UpdateToDecimalSeparator(Frm);
                end;
              sgdDate:
                Frm:=ConvertDateFormat(FDefaultDateFormat);
              sgdTime:
                Frm:=FDefaultTimeFormat;
            end;
          end;
        end;
    end;
    if Frm <> '' then
      FormatCells(FStartRow, FStartCol + c, RowIdx, 1, Frm);
  end;
end;

procedure TrXLSExport.SetColProps;
var
  c, Col: integer;
  Fld: TField;
  DatabarSetting: TDatabarSetting;
  ColorScaleSetting: TColorScaleSetting;
  IconSetSetting: TIconSetSetting;
begin
  for c:=0 to FN.Count - 1 do
  begin
    case SourceType of
      stGrid, stDataset, stRecView:
        begin
          Fld:=GetField(c);
          if (xleSetColumnWidth in FOptions) and not FieldAsPicture(Fld) then
            ColumnAutoFit(FStartCol + c, FMaxColWidth);
          if (xleWrapMemoField in FOptions) and rdbtool.IsMemoField(Fld) then
            FWorksheet.Columns[FStartCol + c].WrapText:=true;

          if FieldAsDatabar(Fld) then
          begin
            Self.LoadDatabarProps(Fld, DatabarSetting);
            Self.DatabarCells(FirstDataRow, FStartCol + c, LastDataRow - FirstDataRow + 1, 1, DatabarSetting);
          end;
          if FieldAsColorScale(Fld) then
          begin
            Self.LoadColorScaleProps(Fld, ColorScaleSetting);
            Self.ColorScaleCells(FirstDataRow, FStartCol + c, LastDataRow - FirstDataRow + 1, 1, ColorScaleSetting);
          end;
          if FieldAsIconSet(Fld) then
          begin
            Self.LoadIconSetProps(Fld, IconSetSetting);
            Self.IconSetCells(FirstDataRow, FStartCol + c, LastDataRow - FirstDataRow + 1, 1, IconSetSetting);
          end;
        end;
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          if Col >= 0 then
          begin
            if (xleSetColumnWidth in FOptions) then
              ColumnAutoFit(FStartCol + c, FMaxColWidth);
          end;
        end;
    end;
  end;
end;

procedure TrXLSExport.SetRowProps;
var
  i: integer;
begin
  for i:=Self.FStartRow to FStartRow + RowIdx do
  begin
    if (xleSetRowHeight in FOptions) and not UsePictures then
      RowAutoFit(i, FMaxRowHeight);
    case FVerticalAlignment of
      vaTop:
        FWorksheet.rows[i].VerticalAlignment:=-4160;
      vaCenter:
        FWorksheet.rows[i].VerticalAlignment:=-4108;
      vaBottom:
        FWorksheet.rows[i].VerticalAlignment:=-4107;
    end;
  end;
end;

function TrXLSExport.ExportDBTable(DBGrid: TCustomDBGrid; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean;
begin
  Result:=ExportDBTableInt(stGrid, DBGrid, AutoCloseApp, ExportToSheetName);
end;

function TrXLSExport.ExportDBTable(Dataset: TDataSet; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean;
begin
  Result:=ExportDBTableInt(stDataset, Dataset, AutoCloseApp, ExportToSheetName);
end;

function TrXLSExport.ExportDBTable(rDBRecView: TrDBRecView; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean;
begin
  Result:=ExportDBTableInt(stRecView, rDBRecView, AutoCloseApp, ExportToSheetName);
end;

function TrXLSExport.ExportDBTable(TreeView: TCustomTreeView; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean;
begin
  Result:=ExportDBTableInt(stTreeView, TreeView, AutoCloseApp, ExportToSheetName);
end;

function TrXLSExport.ExportDBTable(rStringGridEd: TrStringGridEd; AutoCloseApp: boolean = true; const ExportToSheetName: string = ''): boolean;
begin
  Result:=ExportDBTableInt(stStringGrid, rStringGridEd, AutoCloseApp, ExportToSheetName);
end;

function TrXLSExport.ExportDBTableInt(ASourceType: TSourceType; O: TObject; AutoCloseApp: boolean; const ExportToSheetName: string): boolean;
var
  d: TDataSource;
  FldDef: TGetFldDefs;
  c: integer;
begin
  Result:=false;
  if not Assigned(O) then
    Exit;
  SourceType:=ASourceType;
  SourceObject:=O;
  Dataset:=nil;
  d:=nil;
  case SourceType of
    stGrid:
      begin
        DBGrid:=TCustomDBGrid(O);
        d:=DBGrid.DataSource;
      end;
    stRecView:
      begin
        RecView:=TrDBRecView(O);
        d:=RecView.DataSource;
      end;
    stTreeView:
      TreeView:=TCustomTreeView(O);
    stDataset:
      Dataset:=TDataSet(O);
    stStringGrid:
      StringGrid:=TrStringGridEd(O);
  end;
  if SourceType in [stGrid, stRecView] then
  begin
    if not Assigned(d) or not Assigned(d.Dataset) or not d.Dataset.Active then
      Exit;
    Dataset:=d.Dataset;
  end;
  if Assigned(Dataset) and not Dataset.Active then
    Exit;

  if FShowProgressBar then
    ShowInfo(sExporting + sOfData);
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psStarting, 0, false, Rect(0, 0, 0, 0), sExporting + sOfData);

  FN:=TStringList.Create;
  FAsColor:=TStringList.Create;
  try
    if not OpenFile(ExportFile = '', ExportFile) then
      Exit;
    // test only: SetVisibility(true);

    // proc sheet
    if xleAddAsNewSheet in FOptions then
      Self.SelectSheet(true)
    else
      if ExportToSheetName <> '' then
        if not SelectSheet(false, ExportToSheetName) then
        begin
          DlgW(sNotFound + ': ' + ExportToSheetName);
          Exit;
        end;

    // proc fields
    FldDef:=GetFldDefs(false, xleOnlyVisibleFields in FOptions, true, true, UsePictures, false, false);
    GetFldList(SourceObject, FN, FldDef);
    if UseColors then
    begin
      for c:=0 to FN.Count - 1 do
      begin
        case SourceType of
          stGrid, stDataset, stRecView:
            if FieldAsColor(GetField(c)) then
              FAsColor.Add(FN[c]);
          stTreeView:
            ;
          stStringGrid:
            ;
        end;
      end;
    end;
    try
      // init possition for export and copy data
      if xleAddToExistingRows in FOptions then
        SetStartPos(false, true)
      else
        FWorksheet.UsedRange.Clear;
      if Assigned(FOnBeforeExport) then
        FOnBeforeExport(Self);

      // set Data array size for one row
      SetDataSize(1, Max(1, FN.Count));
      RecCount:=0;
      RecExported:=0;
      PicExported:=0;
      RowIdx:=0;
      FirstDataRow:=0;
      LastDataRow:=0;

      if (xleExportTitles in FOptions) or FExportStructureOnly then
        ExpLabels;
      if not FExportStructureOnly then
      begin
        // export all records
        FirstDataRow:=FStartRow + RowIdx;
        if not ExpData then
          Exit;
        LastDataRow:=FStartRow + RowIdx - 1;
        // export summary row
        if (xleAddSumRow in FOptions) and (RecExported > 0) then
          ExpSum(RecExported);
      end;

      // set numeric format
      if (xleSetCellFormat in FOptions) then
        SetColFormat;
      // set font for whole range
      if (xleSetCellFormat in FOptions) then
        FormatCells(FStartRow, FStartCol, RowIdx, FDataColCount, '', FSetFontName, FSetFontSize);
      // set border for whole range
      if (xleSetCellBorders in FOptions) then
        BorderCells(FStartRow, FStartCol, RowIdx, FDataColCount);

      // format title row
      if (xleExportTitles in FOptions) and (xleBoldTitles in FOptions) then
        FormatCells(FStartRow, FStartCol, 1, FDataColCount, '', '', 0, [cfBold]);
      // format sum row
      if (xleAddSumRow in FOptions) and (RecExported > 0) and (xleBoldTitles in FOptions) then
        FormatCells(FStartRow + RowIdx - 1, FStartCol, 1, FDataColCount, '', '', 0, [cfBold]);
      // set col props
      SetColProps;
      // set row props
      SetRowProps;
      // set zoom
      if (xleSetPageSetup in FOptions) then
      begin
        SetWindowZoom(FWindowZoom);
        if FPageZoom > 0 then
          SetPageZoom(FPageZoom);
      end;

      // freeze panes
      if (xleExportTitles in FOptions) and (xleFreezeTitles in FOptions) then
        FreezePanes(FStartRow, 0);
      // autofilter
      if (xleSetAutoFilter in FOptions) then
        FWorksheet.Cells[FStartRow, FStartCol].Autofilter;
{$IFDEF DEMO}
      ShowDemoMsg;
{$ENDIF}
      if Assigned(FOnAfterExport) then
        FOnAfterExport(Self);

      // print file
      if FActionAfter = aaPrint then
        FWorksheet.Printout;
      // save file
      if ExportFile <> '' then
        if not SaveCloseFile(false, true, ExportFile) then
          Exit;
      Result:=true;

      if (FActionAfter = aaShow) then
        // keep opened and visible
        SetVisibility(true)
      else
        if AutoCloseApp then
          SaveCloseFile(true, false);
    except
      on E: Exception do
        ShowErr(sExportDataErr, E);
    end;
  finally
    FreeAndNil(FN);
    FreeAndNil(FAsColor);
    if FShowProgressBar then
      ShowInfo('');
    if Assigned(OnProgressEvent) then
      OnProgressEvent(Self, psEnding, 100, false, Rect(0, 0, 0, 0), '');
    if not Result then
      SaveCloseFile(true, false);
  end;
end;

procedure TrXLSExport.SetStartRow(Value: integer);
begin
  if Value > 0 then
    FStartRow:=Value;
end;

procedure TrXLSExport.SetStartCol(Value: integer);
begin
  if Value > 0 then
    FStartCol:=Value;
end;

function TrXLSExport.GetField(idx: integer): TField;
begin
  Result:=Dataset.FindField(FN[idx]);
end;

function TrXLSExport.UsePictures: boolean;
begin
  Result:=xleExportPictures in FOptions;
end;

function TrXLSExport.FieldAsPicture(Fld: TField): boolean;
var
  FldType: integer;
begin
  FldType:=cFldTypeNormal;
  if IsGraphicField(Fld) then
    FldType:=cFldTypePicture;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
    if TrDBGrid(DBGrid).FieldsAsGraphics.IndexOf(Fld.FieldName) >= 0 then
      FldType:=cFldTypePicture;
  if Assigned(FGetFieldType) then
    FGetFieldType(Self, Fld, FldType);
  Result:=(FldType = cFldTypePicture);
end;

function TrXLSExport.UseColors: boolean;
begin
  Result:=(xleSetBkColor in FOptionColors) or (xleSetFontColor in FOptionColors);
end;

function TrXLSExport.FieldAsColor(Fld: TField): boolean;
var
  FldType: integer;
begin
  FldType:=cFldTypeNormal;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
    if TrDBGrid(DBGrid).FieldsAsColor.IndexOf(Fld.FieldName) >= 0 then
      FldType:=cFldTypeColor;
  if Assigned(FGetFieldType) then
    FGetFieldType(Self, Fld, FldType);
  Result:=(FldType = cFldTypeColor);
end;

function TrXLSExport.FieldAsDatabar(Fld: TField): boolean;
var
  FldType: integer;
begin
  FldType:=cFldTypeNormal;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
    if TrDBGrid(DBGrid).FieldsAsDataBar.IndexOfName(Fld.FieldName) >= 0 then
      FldType:=cFldTypeDataBar;
  if Assigned(FGetFieldType) then
    FGetFieldType(Self, Fld, FldType);
  Result:=(FldType = cFldTypeDataBar);
end;

procedure TrXLSExport.LoadDatabarProps(Fld: TField; var DatabarSetting: TDatabarSetting);
var
  dbt: TrDataBarText;
  cTo: TColor;
begin
  DatabarSetting.MinValue:=0;
  DatabarSetting.MaxValue:=0;
  DatabarSetting.FillColor:=clNone;
  DatabarSetting.BorderColor:=clNone;
  DatabarSetting.ShowValue:=true;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
  begin
    TrDBGrid(DBGrid).Fields_GetDataBarSettings(Fld.FieldName, DatabarSetting.MinValue, DatabarSetting.MaxValue, dbt, DatabarSetting.FillColor, cTo);
    if dbt = dbtNothing then
      DatabarSetting.ShowValue:=false;
  end;
  if Assigned(FOnPrepareDatabarEvent) then
    FOnPrepareDatabarEvent(Self, Fld, DatabarSetting);
end;

function TrXLSExport.FieldAsColorScale(Fld: TField): boolean;
var
  FldType: integer;
begin
  FldType:=cFldTypeNormal;
  if Assigned(FGetFieldType) then
    FGetFieldType(Self, Fld, FldType);
  Result:=(FldType = cFldTypeColorScale);
end;

procedure TrXLSExport.LoadColorScaleProps(Fld: TField; var ColorScaleSetting: TColorScaleSetting);
begin
  ColorScaleSetting.MinValue:=0;
  ColorScaleSetting.MidValue:=0;
  ColorScaleSetting.MaxValue:=0;
  ColorScaleSetting.MinColor:=clNone;
  ColorScaleSetting.MidColor:=clNone;
  ColorScaleSetting.MaxColor:=clNone;
  if Assigned(FOnPrepareColorScaleEvent) then
    FOnPrepareColorScaleEvent(Self, Fld, ColorScaleSetting);
end;

function TrXLSExport.FieldAsIconSet(Fld: TField): boolean;
var
  FldType: integer;
begin
  FldType:=cFldTypeNormal;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
    if TrDBGrid(DBGrid).FieldsAsStatusBar.IndexOfName(Fld.FieldName) >= 0 then
      FldType:=cFldTypeIconSet;
  if Assigned(FGetFieldType) then
    FGetFieldType(Self, Fld, FldType);
  Result:=(FldType = cFldTypeIconSet);
end;

procedure TrXLSExport.LoadIconSetProps(Fld: TField; var IconSetSetting: TIconSetSetting);
var
  sbt: TrStatusBarType;
  cFrom, cTo: TColor;
  i: integer;
begin
  IconSetSetting.IconSetType:=0;
  IconSetSetting.UseThresholds:=false;
  for i:=1 to 5 do
    IconSetSetting.ThresholdValues[i]:=0;
  IconSetSetting.ShowIconOnly:=false;
  if Assigned(Fld) and Assigned(DBGrid) and (DBGrid is TrDBGrid) then
  begin
    TrDBGrid(DBGrid).Fields_GetStatusBarSettings(Fld.FieldName, i, sbt, cFrom, cTo);
    case sbt of
      sbtProgressBar:
        IconSetSetting.IconSetType:=xl5Quarters;
      sbtDots:
        IconSetSetting.IconSetType:=xl5Quarters;
      sbtCircles:
        IconSetSetting.IconSetType:=xl5Quarters;
      sbtEllipses:
        IconSetSetting.IconSetType:=xl5Quarters;
      sbtSignalStrength:
        if i < 5 then
          IconSetSetting.IconSetType:=xl4CRV
        else
          IconSetSetting.IconSetType:=xl5CRV;
      sbtPie:
        IconSetSetting.IconSetType:=xl5Quarters;
    end;
  end;
  if Assigned(FOnPrepareIconSetEvent) then
    FOnPrepareIconSetEvent(Self, Fld, IconSetSetting);
end;

function TrXLSExport.PasteRowData: boolean;
var
  c: integer;
begin
  Result:=false;
  try
    // try to paste whole row
    SetCurrentRange(FStartRow + RowIdx, FStartCol, 1, FDataColCount);
    FRange.Value:=Data;
    Result:=true;
  except
    // error => try paste each cell individually
    for c:=FStartCol to FStartCol + FDataColCount - 1 do
    begin
      try
        SetCurrentRange(FStartRow + RowIdx, c, 1, 1);
        FRange.Value:=Data[1, c];
      except
        on E: Exception do
          if DlgQ2(sMSExcelError + #13#10 + E.Message + #13#10 + Format('Line %d: %s', [RowIdx, Data[1, c]]), sContinueQ) <> mrYes then
          begin
            Aborted:=true;
            Break;
          end;
      end;
    end;
  end;
  Inc(RowIdx);
end;

function TrXLSExport.PrepareData(s: string): string;
begin
  Result:=s;
  if xleTrimData in FOptions then
  begin
    Result:=Trim(Result);
  end;
end;

procedure TrXLSExport.ExpLabels;
var
  i: integer;
  L: TStrings;
  FldDef: TGetFldDefs;
begin
  L:=TStringList.Create;
  try
    FldDef:=GetFldDefs(not(xleRealFieldNamesAsTitle in FOptions), xleOnlyVisibleFields in FOptions, true, true, UsePictures, false, false);
    if SourceObject is TrDBGrid then
      GetFldListrDBGrid(TrDBGrid(SourceObject), L, FldDef)
    else
      GetFldList(SourceObject, L, FldDef);
    for i:=0 to FN.Count - 1 do
      Data[1, i + 1]:=L[i];
  finally
    FreeAndNil(L);
  end;
  PasteRowData;
end;

function TrXLSExport.ExpData: boolean;
var
  i: integer;
  ExistDrawings: integer;
begin
  RecCount:=0;
  Aborted:=false;
  Result:=true;
  ExistDrawings:=FWorksheet.DrawingObjects.Count;
  case SourceType of
    stGrid, stDataset:
      begin
        RecCount:=Dataset.RecordCount;
        Result:=rdbtool.ScanAllRecords(Dataset, ScanCallBack, FUseDisableControl, FUseDisableControl);
      end;
    stRecView:
      begin
        RecCount:=1;
        ScanCallBack(Dataset, 0, Aborted);
      end;
    stTreeView:
      ExpTreeItems;
    stStringGrid:
      ExpGridRows;
  end;
  // move all new exported pictures
  if UsePictures then
  begin
    for i:=ExistDrawings + 1 to FWorksheet.DrawingObjects.Count do
    begin
      if (FPictureCellMargin > 0) then
      begin
        FWorksheet.DrawingObjects[i].Left:=FWorksheet.DrawingObjects[i].Left + FPictureCellMargin / FRowHeightPixelRatio;
        FWorksheet.DrawingObjects[i].Top:=FWorksheet.DrawingObjects[i].Top + FPictureCellMargin / FRowHeightPixelRatio;
      end;
      if FPictureUseShadow then
        FWorksheet.DrawingObjects[i].Shadow:=true;
      // FWorksheet.DrawingObjects[i].Shadow.Transparency:=0.5;    Transparency is not supported
      // FWorksheet.DrawingObjects[i].SoftEdge.Type:=2;       SoftEdge is not supported
    end;
  end;
end;

procedure TrXLSExport.ExpSum(rows: integer);
var
  c, Col: integer;
  Fld: TField;
  IsApostr, IsGraphic, tot: boolean;
begin
  tot:=false;
  for c:=0 to FN.Count - 1 do
  begin
    Data[1, c + 1]:='';
    IsApostr:=true;
    IsGraphic:=false;
    case SourceType of
      stGrid, stDataset, stRecView:
        begin
          Fld:=GetField(c);
          IsApostr:=IsApostrField(Fld);
          IsGraphic:=FieldAsPicture(Fld);
        end;
      stTreeView:
        ;
      stStringGrid:
        begin
          Col:=StrToIntDef(FN[c], -1);
          if Col >= 0 then
            IsApostr:=StringGrid.Columns[Col].DataType = sgdText;
        end;
    end;
    if IsApostr then
    begin
      if not tot then
        Data[1, c + 1]:='Sum:';
      tot:=true;
    end
    else
      if not IsGraphic then
        Data[1, c + 1]:=Format('=SUM(R[-%d]C:R[-1]C)', [rows]);
  end;
  PasteRowData;
end;

function TrXLSExport.ExportPicture(Field: TField; Row, Col: integer): string;
var
  w, h: integer;
begin
  Result:='';
  if IsGraphicField(Field) then
  begin
    case rpictool.ReadPictureFromField(Field, Pic) of
      gtNone:
        Pic.Bitmap.Width:=0;
      gtError:
        Result:=rstring.sPictureErr;
    end;
  end
  else
  begin
    Pic.Bitmap.Width:=0;
    if not Assigned(FOnPreparePictureEvent) then
      Result:='Use OnPreparePictureEvent to define picture';
  end;
  if (Result = '') then
  begin
    try
      if Assigned(FOnPreparePictureEvent) then
        // call event to prepare picture size
        FOnPreparePictureEvent(Self, Row, Col, Field, Pic);
      if (Pic.Width = 0) then
        Exit;
      // resize
      rpictool.SmoothResizePicture(Pic, FPictureMaxWidth, FPictureMaxHeight, false);
      // set col width for first row or extend col width
      w:=Pic.Width + 2 * FPictureCellMargin;
      if (PicExported = 0) or (FWorksheet.Columns[Col].ColumnWidth * FColWidthPixelRatio < w) then
        FWorksheet.Columns[Col].ColumnWidth:=w / FColWidthPixelRatio;
      // extend row height
      h:=Pic.Height + 2 * FPictureCellMargin;
      if FWorksheet.rows[Row].RowHeight * FRowHeightPixelRatio < w then
        FWorksheet.rows[Row].RowHeight:=h / FRowHeightPixelRatio;
      // copy picture to clipboard
      Clipboard.Open;
      try
        Clipboard.Assign(Pic);
      finally
        Clipboard.Close;
      end;
      // have to wait some time, in other case excel show OLE error (?)
      Sleep(50);
      // paste to the cell
      SetCurrentRange(Row, Col, 1, 1);
      FRange.Select;
      FWorksheet.Paste;
      // clear clipboard
      Clipboard.Open;
      try
        Clipboard.Clear;
      finally
        Clipboard.Close;
      end;
      Inc(PicExported);
    except
      // store error text to the cell
      on E: Exception do
        Result:=sMSExcelError + ' ' + E.Message;
    end;
  end;
end;

procedure TrXLSExport.ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
var
  c: integer;
  Fld: TField;
  AsStr: boolean;
  Color: TColor;
  FldStr: string;
begin
  // use selected row only?
  if (xleOnlySelectedRows in FOptions) and (SourceType = stGrid) then
  begin
    if DBGrid is TrDBGrid_MS then
    begin
      if not TrDBGrid_MS(DBGrid).IsActiveSelected then
        Exit;
    end
    else
      if not THackCustomDBGrid(DBGrid).SelectedRows.CurrentRowSelected then
        Exit;
  end;

  RecIdx:=RecIndex;
  for c:=0 to FN.Count - 1 do
  begin
    Data[1, c + 1]:='';
    Fld:=GetField(c);
    if FieldAsPicture(Fld) then
      Data[1, c + 1]:=ExportPicture(Fld, FStartRow + RowIdx, FStartCol + c)
    else
    begin
      AsStr:=false;
      if DBGrid is TrDBGrid then
        // use GetCellValue to get groupped column value instead of field value
        FldStr:=TrDBGrid(DBGrid).GetCellValue(Fld, false)
      else
        FldStr:=Fld.AsString;
      if (FldStr <> '') then
      begin
        AsStr:=(xleStringFieldAsText in FOptions) and (Fld is TStringField);
        if not AsStr then
          AsStr:=Copy(FldStr, 1, 1) = '='; // CharInSet(Fld.AsString[1], ['=', '+', '-', '/', '*']);
      end;
      if AsStr then
        Data[1, c + 1]:='''' + PrepareData(FldStr)
      else
        if Fld.IsNull then
          Data[1, c + 1]:=null
        else
          if (Fld is TSQLTimeStampField) or (Fld is TDateTimeField) then
          begin
            if Fld.AsDateTime < Encodedate(1900, 1, 1) then
              Data[1, c + 1]:=FldStr
            else
              Data[1, c + 1]:=Fld.AsDateTime;
          end
          else
            if Fld is TFMTBCDField then
              Data[1, c + 1]:=Fld.AsFloat
            else
              if Fld is TStringField then
                Data[1, c + 1]:=PrepareData(FldStr)
              else
                Data[1, c + 1]:=Fld.AsVariant;
    end;
    if UseColors and (FAsColor.IndexOf(Fld.FieldName) >= 0) then
    begin
      Color:=rpictool.StringToColor(Fld.AsString);
      if Color <> clNone then
      begin
        SetCurrentRange(FStartRow + RowIdx, FStartCol + c, 1, 1);
        if (xleSetBkColor in FOptionColors) then
          FRange.Interior.Color:=Color;
        if (xleSetFontColor in FOptionColors) then
          FRange.Font.Color:=Color;
        if (xleHideText in FOptionColors) then
          Data[1, c + 1]:='';
      end;
    end;
  end;
  PasteRowData;
  Inc(RecExported);
  ShowProgresInt;
  if Aborted then
    Abort:=true;
end;

procedure TrXLSExport.ExpTreeItems;
  procedure ProcTV(tn: TTreeNode; level: integer);
  begin
    while Assigned(tn) do
    begin
      SetDataSize(1, level + 1);
      Data[1, level + 1]:=PrepareData(tn.Text);
      PasteRowData;
      Inc(RecExported);
      Inc(RecIdx);
      ShowProgresInt;
      if tn.Expanded then
        ProcTV(tn.GetFirstChild, level + 1);
      tn:=tn.GetNextSibling;
    end;
  end;

begin
  RecCount:=THackCustomTreeView(TreeView).Items.Count;
  RecIdx:=0;
  ProcTV(THackCustomTreeView(TreeView).Items.GetFirstNode, 0);
end;

procedure TrXLSExport.ExpGridRows;
var
  r, c, Col: integer;
  v: Variant;
  Column: TrStringGridEdColumn;
begin
  RecCount:=StringGrid.RowCount - StringGrid.FixedRows;
  for r:=StringGrid.FixedRows to StringGrid.RowCount - 1 do
  begin
    RecIdx:=r;
    for c:=0 to FN.Count - 1 do
    begin
      Col:=StrToIntDef(FN[c], -1);
      v:='';
      if (Col >= 0) and (StringGrid.Cells[Col, r] <> '') then
      begin
        Column:=StringGrid.Columns[Col];
        case Column.DataType of
          sgdNumber, sgdInteger:
            v:=StringGrid.NumCells[Col, r];
          sgdDate, sgdTime:
            v:=StringGrid.DateTimeCells[Col, r];
        else
          v:=PrepareData(StringGrid.Cells[Col, r]);
        end;
      end;
      Data[1, c + 1]:=v;
    end;
    PasteRowData;
    Inc(RecExported);
    ShowProgresInt;
    if Aborted then
      Abort;
  end;
end;

procedure TrXLSExport.ShowProgresInt;
begin
  if RecCount = 0 then
    Exit;
  if FShowProgressBar and (RecIdx mod 100 = 0) then
    ShowInfoProgressBar(sExporting + sOfData, nil, DoAbort, Round(100.0 * RecIdx / RecCount));
  if Assigned(OnProgressEvent) then
    OnProgressEvent(Self, psRunning, Round(100.0 * RecIdx / RecCount), false, Rect(0, 0, 0, 0), sExporting + sOfData);
end;

procedure TrXLSExport.DoAbort(Sender: TObject);
begin
  Aborted:=true;
end;

end.
