unit rXMLReport;

{<
  rXMLReport unit contains TrXMLReport - non visual component which provides functionality for
  generating reports according prepared DOCX or ODT template with special TAGs;
  it uses this template and fill data into it instead of TAGs;
  DOCX or ODT template can be modified easily by any editor
  (MS Word, OpenOffice, Google docs).

  See homepage http://www.rosinsky.cz/delphi/ for detail help.

  Updated version compatible with Delphi2010+ and 64bit platform.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2014)
  @lastmod(Dec 2020)
}

interface

{$INCLUDE _rosi.inc}
{$IFDEF DXE4}
 {$DEFINE SkipBookmark}
{$ENDIF}
{$IFDEF DXE5}
 {$DEFINE SkipBookmark}
{$ENDIF}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, MaskUtils, StrUtils, Masks, Math,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF}
  rpictool, RosiCompGlobalCfg;

const
  maxDS = 25;
  cDOCXBold = '<w:b/>';
  cDOCXItalic = '<w:i/>';
  cDOCXUnderline = '<w:u w:val="single"/>';
  cDOCXStrikeOut = '<w:strike/>';
  cInlineFormatConvert = '###convert###';

type
  { Defines default template format }
  TTemplateFormat = (tfAutoDetect, tfDOCX, tfODT);

  { list of supported RTF format types }
  TRtfSupportedFormat = (rtfFontName, rtfFontSize, rtfBold, rtfItalic, rtfUnderline, rtfUnderlineDouble, rtfStrike, rtfStrikeDouble);
  { set of supported RTF format types }
  TRtfSupportedFormats = set of TRtfSupportedFormat;

  { list of XML text format types }
  TXMLTextFormat = (frmBold, frmItalic, frmUnderline);
  { set of XML text format types }
  TXMLTextFormats = set of TXMLTextFormat;

  { event type called for each record during reporting }
  TReportProcessEvent = procedure(Sender: TObject; RecNbr: integer; var Abort: boolean) of object;
  { event type called to get calculated string result }
  TUserCalc = procedure(Sender: TObject; var CalcStr: string; var Replace: boolean) of object;
  { event type called to get calculated string result, supports function parameters }
  TUserCalcParams = procedure(Sender: TObject; FuncName: string; Params: array of variant; var FuncResult: variant; var Replace: boolean) of object;
  { event type called to get bitmap content for Picture tag }
  TOnGetPictureContent = procedure(Sender: TObject; FuncName: string; Bmp: TBitmap) of object;
  { event type called to modify file content }
  TModifyContent = procedure(Sender: TObject; FileName: string; var Content: string) of object;

  { @exclude }
  TTableCellFormatRec = record
    HAlign, VAlign: string;
    RotateText: boolean;
    BkColor: TColor;
    Font: TFont;
  end;

  { event type called to get row height for named tables }
  TGetRowHeight = procedure(Sender: TObject; ID: string; RowIndex: integer; var RowHeightMM: integer) of object;
  { event type called to get column width for named tables }
  TGetColumnWidth = procedure(Sender: TObject; ID: string; Field: TField; ColIndex: integer; var ColWidthMM: integer; var AbsoluteWidth: boolean) of object;
  { event type called to get cell parameters for named tables }
  TGetCellParams = procedure(Sender: TObject; ID: string; Field: TField; RowIndex, ColIndex: integer; var Text: string;
    var TableCellFormatRec: TTableCellFormatRec) of object;

  { @exclude }
  TTableContentRec = record
    TableName: string;
    TableXMLContent: string;
    CellXMLContent: string;
    PlainText: string;
  end;

  { Object contain calculated value for each field }
  TCalcField = class
  private
    Init: boolean;
    DSindex: integer;
    Name: string;
    Expression: string;
    Max: double;
    GrandMax: double;
    Min: double;
    GrandMin: double;
    Sum: double;
    GrandSum: double;
    Count: integer;
    GrandCount: integer;
  end;

  { @exclude }
  TScanInfo = record
    DS: string;
    DSindex: integer;
    WhileCond, SkipCond, Sep: string;
    NewLine, NewPar, NewCol, NewPage, NoIfEmpty, NoFirst, TblScan: boolean;
    TblRowCond: array of string;
    MinRows, MaxRows: integer;
  end;

  { @exclude }
  TSeekInfo = record
    Action: byte;
    OldText, NewText: string;
    SeekFrom, SeekTo: string;
    SetFrmStart, SetFrmStop: string;
    WholeWord: boolean;
  end;

  { @exclude }
  TRunInfo = record
    RunBeginPos: integer;
    RunEndPos: integer;
    RunXML: string;
    RunPrFormatFontSize: string;
    RunPrFormatFontName: string;
  end;

  { rXMLReport is non visual component which provides functionality for
    generating reports according prepared DOCX or ODT template with special TAGs }
  TrXMLReport = class(TComponent)
  private
    PicCount: integer;
    FileList: TStrings;
    RelItemStr: string;
    XMLFileName: string;
    XMLFormat: TTemplateFormat;
    FReportFile: string;
    FOutputFile: string;
    FTemplateFormat: TTemplateFormat;
    FDefaultCalcFormat: string;
    FDefaultNumericFormat: string;
    FDefaultBooleanFormat: string;
    FDefaultDateFormat: string;
    FDefaultTimeFormat: string;
    FJoinTouchingTables: boolean;
    FConvertInlineFormatting: boolean;
    FConvertScientificFormat: boolean;
    TryScientificFormat: boolean;
    FVariables: TStrings;
    FFieldNotFoundValue: string;
    FMemoReplaces: TStrings;
    FInlineFormatting: TStrings;
    FActionAfter: TActionAfter;
    FShowMaximized: boolean;
    FLoadTableNamePrefix: string;
    FOnReadInfo: TNotifyEvent;
    FOnUserCalc: TUserCalc;
    FOnUserCalcParams: TUserCalcParams;
    FOnGetPictureContent: TOnGetPictureContent;
    FOnReportProcess: TReportProcessEvent;
    FOnGetTemplateStr: TModifyContent;
    FOnGetRowHeight: TGetRowHeight;
    FOnGetColumnWidth: TGetColumnWidth;
    FOnGetCellParams: TGetCellParams;
    FModifyBeforeSave: TModifyContent;
    FShowSaveErrorDialog: boolean;
    FShowProgressBar: boolean;
    FExtendedInfo: TStrings;
    FRtfSupportedFormats: TRtfSupportedFormats;
    FRtfCustomFormats: TStrings;
    procedure SetVariables(Value: TStrings);
    procedure SetMemoReplaces(Value: TStrings);
    procedure SetInlineFormatting(Value: TStrings);
    procedure SetRtfCustomFormats(Value: TStrings);
    procedure DoAbort(Sender: TObject);
  protected
    Aborted: boolean;
    RecNbr, RecNbrDS0, RecCountDS0: integer;
    DS: array [0 .. maxDS] of TDataset;
    DS_eof: array [0 .. maxDS] of boolean;
    {$IFNDEF SkipBookmark}
    DS_Bookmarks: array [0 .. maxDS] of TBookmark;
    {$ENDIF}
    LCalcField: TList;
    cf: TCalcField;
    RepStr: string;
    SeekInfo: TSeekInfo;
    procedure Loaded; override;
    procedure RaiseErr(ErrType: integer; const ErrStr: string = '');
    procedure SetReportFormat(FileName: string);
    procedure CheckReportFormat(Frm: TTemplateFormat; Tag: string);

    function StrToXML(const Str: string): string;
    function XMLToStr(const XML: string): string;
    procedure SepBracket(const Str: string; var Bracket, Block1, Block2: string);
    function FindTagStr(const TagStr: string; var s: string; WChar: Char; var StartPos: integer; Del: boolean): string;
    function GetListPart(var Str: string; const Sep: string): string;
    procedure ClearEOLPart(var s: string);
    procedure ExtractIntXMLTags(var s: string);
    procedure ClearBookmarkPart(var s: string);
    procedure ClearEmptyParPart(var s: string);
    procedure JoinTables(var s: string);

    procedure InitCalcField(DSindex: integer; const CalcStr: string);
    procedure FreeCalcField;
    procedure ProcCalcField(DSindex: integer);
    function GetCalcFieldValue(CalcTypeIndex: integer; const Name: string): double;

    function IsConstantTag(const TagStr: string): boolean;
    function IsNumericTag(const TagStr: string; var Value: double): boolean;
    function IsFieldTag(const TagStr: string): boolean;
    function DatasetCharToIndex(Ch: Char): integer;
    function IndexToDatasetChar(Idx: integer): Char;
    function DSActive(DSindex: integer): boolean;
    function DSNotEOF(DSindex: integer): boolean;
    function GetDSIndex(const TagStr: string): integer;
    function CheckDSIndex(DSindex: integer): string;
    function FindFirst(DSindex: integer): boolean;
    function FindNext(DSindex: integer): boolean;
    function FindPrev(DSindex: integer): boolean;
    function FindLast(DSindex: integer): boolean;
    function GetDSBookmark(DSindex: integer): boolean;
    function GotoDSBookmark(DSindex: integer): boolean;
    procedure FreeBookMarks;

    function ListAllFields: string;
    function ProcSepList(const TagStr: string): string;
    function ProcFirstValue(const TagStr: string): string;
    function ProcPicture(const TagStr: string): string;
    function ProcBarcode(const TagStr: string): string;
    function CreateFormatPicture(DSindex: integer; F: TField; const FileName: string; Bmp: TBitmap; const FormatStr: string): string;
    function ProcNumToText(const TagStr: string): string;
    procedure ProcDefVar(const TagStr: string);
    function ProcStrPrefix(const TagStr: string): string;
    function ProcCase(const TagStr: string): string;
    procedure ProcExtendedInfo(var TagStr: string);
    procedure ProcInsertFile(var TagStr: string);
    procedure ProcInsertTable(var TblStr: string);

    procedure ProcAllFiles;
    function ProcXMLFiles(ZipFileName: string; ReadOnly: boolean): boolean;
    function AddImageFile(Pic: TPicture; GrType: TrGraphicType): string;
    procedure UpdateRelFile;

    procedure ProcXMLFileContent;
    procedure UpdateXMLTags;
    procedure ReadWriteTablesContent(Write: boolean);

    function ProcBlock(const Str: string): string;
    function ProcScan(var s: string): string;
    function DecodeScanLine(var s: string): TScanInfo;
    procedure ProcIf(IIF, ProcAllIf: boolean; var s: string);
    function ProcXML(const Str: string): string;
    function ProcTagStr(const TagStr: string; var Replaced: boolean): string;
    function ParseRTFFormat(const RTFStr: string): string;
    function GetDBValue(DSindex: integer; const FieldName, FormatStr: string): string;
    function GetVarUserValue(const VarName: string; var Replaced: boolean): string;
    function GetCalcValue(const CalcStr, FormatStr: string; ErrFieldAsZero: boolean): string;
    function GetConditionResult(const CondStr: string): boolean;

    function CheckScientificFormat(const s: string): string;
    function GetRunXMLFormat(StartPos: integer; const s: string): TRunInfo;
    procedure ProcFormatText(var s: string); virtual;
    procedure CheckInlineFormat(var s: string); virtual;
  public
    { define name of printer which will be used if Action=aaPrint }
    PrinterName: string;
    { define number of copies, used if Action=aaPrint }
    PrintCopies: byte;
    { defines whether exception is raised for unknown fields in template }
    RaiseUnknownField: boolean;
    { defines content of all named tables in DOCX file,
      set by NamedTables_LoadContent or NamedTables_DefineContent,
      read and used during report generation by method Execute }
    NamedTables_Content: array of TTableContentRec;

    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { execute creating of report }
    procedure Execute(DataSets: array of TDataset);
    { can be called in OnModifyBeforeSave event, formats selected text bold, italic, underline }
    procedure FormatText(const TextToFormat: string; FormatTypes: TXMLTextFormats; WholeWordOnly: boolean = false; const SeekFrom: string = '';
      const SeekTo: string = '');
    { can be called in OnModifyBeforeSave event, replace selected text, new string can contain tag to start/stop superscript or subscript
      \^  => start superscript;  \_  => start subscript; \=  => end superscipt or subsript }
    procedure ReplaceText(const OldText, NewText: string; WholeWordOnly: boolean = false; const SeekFrom: string = ''; const SeekTo: string = '');

    { load all named table content with TableNamePrefix to NamedTables_Content variable }
    function NamedTables_LoadContent(const FileName: string; const TableNamePrefix: string = ''): integer;
    { add or modify item of NamedTables_Content variable for TableNameID }
    procedure NamedTables_DefineContent(const TableNameID: string; const CellXMLContent: string; const PlainText: string);
    { clear all items from NamedTables_Content variable }
    procedure NamedTables_ClearContent;
    { content of INFO tag in report template,
      INFO tag is used for backward comunication from report to application }
    property ExtendedInfo: TStrings read FExtendedInfo;
  published
    { Define DOCX or ODT file name with report template }
    property ReportFile: string read FReportFile write FReportFile;
    { Define output file name }
    property OutputFile: string read FOutputFile write FOutputFile;
    { define format type of template (auto, DOCX, ODT) }
    property TemplateFormat: TTemplateFormat read FTemplateFormat write FTemplateFormat default tfAutoDetect;
    { Define default format for calculated values }
    property DefaultCalcFormat: string read FDefaultCalcFormat write FDefaultCalcFormat;
    { Define default format for numeric fields }
    property DefaultNumericFormat: string read FDefaultNumericFormat write FDefaultNumericFormat;
    { Define default format for boolean fields }
    property DefaultBooleanFormat: string read FDefaultBooleanFormat write FDefaultBooleanFormat;
    { Define default format for date fields }
    property DefaultDateFormat: string read FDefaultDateFormat write FDefaultDateFormat;
    { Define default format for time fields }
    property DefaultTimeFormat: string read FDefaultTimeFormat write FDefaultTimeFormat;
    { Define whether tables without any text between is redefined as one table (influences columns width in ODT) }
    property JoinTouchingTables: boolean read FJoinTouchingTables write FJoinTouchingTables default false;
    { Define whether scientific format is converted to 123 x 10^3  format using superscript, only for DOCX }
    property ConvertScientificFormat: boolean read FConvertScientificFormat write FConvertScientificFormat default false;
    { Define whether inline format is converted to XML format accordign InlineFormatting property, only for DOCX }
    property ConvertInlineFormatting: boolean read FConvertInlineFormatting write FConvertInlineFormatting default false;
    { Define list of tag used as formatting tag in text (StartTag #9 EndTag #9 FormatTag), only for DOCX }
    property InlineFormatting: TStrings read FInlineFormatting write SetInlineFormatting;
    { Define list of variables and their values which can be used in report template }
    property Variables: TStrings read FVariables write SetVariables;
    { Defines default value used for fields which is not found }
    property FieldNotFoundValue: string read FFieldNotFoundValue write FFieldNotFoundValue;
    { Defines list of strings which is replaced it memo field, e.g.:
      <TAB>=;     - replace #9 by ;
      ;=\line     - replace ; by end of line }
    property MemoReplaces: TStrings read FMemoReplaces write SetMemoReplaces;
    { Define executed action after report is created }
    property ActionAfter: TActionAfter read FActionAfter write FActionAfter default aaShow;
    { Define whether DOCX file is opened in maximized window (if DOCX editor supports) }
    property ShowMaximized: boolean read FShowMaximized write FShowMaximized default false;
    { define event which is called for each INFO tag }
    property OnReadInfo: TNotifyEvent read FOnReadInfo write FOnReadInfo;
    { define event which is called for getting value of undefined tags }
    property OnUserCalc: TUserCalc read FOnUserCalc write FOnUserCalc;
    { define event which is called for getting value of undefined tags, supports also parameters }
    property OnUserCalcParams: TUserCalcParams read FOnUserCalcParams write FOnUserCalcParams;
    { define event which is called to get bitmap content for Picture tag }
    property OnGetPictureContent: TOnGetPictureContent read FOnGetPictureContent write FOnGetPictureContent;
    { define event which is called for each dataset records
      to show progress of execution }
    property OnReportProcess: TReportProcessEvent read FOnReportProcess write FOnReportProcess;
    { define event which is called just before saving of result }
    property OnModifyBeforeSave: TModifyContent read FModifyBeforeSave write FModifyBeforeSave;
    { define event called instead of loading of template file }
    property OnGetTemplateStr: TModifyContent read FOnGetTemplateStr write FOnGetTemplateStr;
    { define whether re-save dialog is shown in case of saving error }
    property ShowSaveErrorDialog: boolean read FShowSaveErrorDialog write FShowSaveErrorDialog default false;
    { indicate whether progress bar is shown }
    property ShowProgressBar: boolean read FShowProgressBar write FShowProgressBar default false;

    { define event used by InserTable tag, event is called once for each row }
    property OnGetRowHeight: TGetRowHeight read FOnGetRowHeight write FOnGetRowHeight;
    { define event used by InserTable tag, event is called once for each column }
    property OnGetColumnWidth: TGetColumnWidth read FOnGetColumnWidth write FOnGetColumnWidth;
    { define event used by InserTable tag, event is called for each cell to define content and formats }
    property OnGetCellParams: TGetCellParams read FOnGetCellParams write FOnGetCellParams;

    { define rtf formats of rtf memo field which is converted to XML format }
    property RtfSupportedFormats: TRtfSupportedFormats read FRtfSupportedFormats write FRtfSupportedFormats default[];
    { Defines list of custom rtf formats (rtfformat = xml format, e.g.:
      \uldb=<w:u w:val="double"/>
      \strike=<w:strike/> }
    property RtfCustomFormats: TStrings read FRtfCustomFormats write SetRtfCustomFormats;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}
{$I _rosi.inc}

uses {$IFDEF BARCODESUPPORT} AsBarcode_interface, {$ENDIF}
  rtool, rdbtool, rdlg, rstring, rzip_interface, NumToText,
  u_rtfparser;

const
  emusPerInch = 914400;
  emusPerCm = 360000;
  emusPerPix = 9525;

  setDS: TSysCharSet = ['A' .. 'Z'];
  cErrStr: array [0 .. 7] of string = (sTemplateBracketErr, sTemplateFieldErr, sTemplateDatasetErr, sTemplateTagNotFoundErr, sTemplateCalculationErr,
    sTemplateIncorrectTag, sPrinterNotFound, sTemplateInsertTableErr);

  // XML
  cTxtEndPar = #13 + #10;
  cXMLEndPar: array [TTemplateFormat] of string = ('', '</w:t><w:br/><w:t xml:space="preserve">', '<text:line-break/>');
  cXMLEndPage: array [TTemplateFormat] of string = ('', '</w:t><w:br w:type="page"/><w:t xml:space="preserve">', '<text:soft-page-break/>');
  cXMLEndCol: array [TTemplateFormat] of string = ('', '</w:t><w:br w:type="column"/><w:t xml:space="preserve">', '');
  cXMLTab: array [TTemplateFormat] of string = ('', '</w:t><w:tab/><w:t xml:space="preserve">', '<text:tab/>');
  cXMLSimpleTagEnd = '/>';

  cXMLParBeginAll: array [TTemplateFormat] of string = ('', '<w:p><w:r><w:t>', '');
  cXMLParBeginPartStart: array [TTemplateFormat] of string = ('', '<w:p><w:pPr>', '<text:p text:style-name=');
  cXMLParBeginPartEnd: array [TTemplateFormat] of string = ('', '</w:pPr>', '>');
  cXMLParBeginPartCont: array [TTemplateFormat] of string = ('', '<w:r><w:t>', '');
  cXMLParEndAll: array [TTemplateFormat] of string = ('', '</w:t></w:r></w:p>', '</text:p>');
  cXMLTagRemove1: array [TTemplateFormat] of string = ('', '</w:t></w:r><w:r><w:t>', '');

  cXMLTableTagBegin: array [TTemplateFormat] of string = ('', '<w:tbl>', '<table:table');
  cXMLTableTagEnd: array [TTemplateFormat] of string = ('', '</w:tbl>', '</table:table>');
  cXMLTableRowBegin: array [TTemplateFormat] of string = ('', '<w:tr>', '<table:table-row');
  cXMLTableRowEnd: array [TTemplateFormat] of string = ('', '</w:tr>', '</table:table-row>');
  cXMLTableCellBegin: array [TTemplateFormat] of string = ('', '<w:tc>', '<table:table-cell');
  cXMLTableCellEnd: array [TTemplateFormat] of string = ('', '</w:tc>', '</table:table-cell>');

  cXMLTableJoinTagStart: array [TTemplateFormat] of string = ('', '</w:tbl><w:tbl>', '</table:table><table:table');
  cXMLTableJoinTagEnd: array [TTemplateFormat] of string = ('', '<w:tbl>', '</table:table-columns>');

  cXMLBookmarkStart = '<w:bookmarkStart';
  cXMLBookmarkEnd = '<w:bookmarkEnd';

  cXMLRelClose: array [TTemplateFormat] of string = ('', '</Relationships>', '</manifest:manifest>');
  cDOCXRelImage = '<Relationship Id="%s" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image" Target="%s"/>';
  cODTRelImage = '<manifest:file-entry manifest:full-path="%s" manifest:media-type="image/jpeg"/>';
  cXMLRelImage: array [TTemplateFormat] of string = ('', cDOCXRelImage, cODTRelImage);
  cDOCXRelEmpty = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?><Relationships xmlns="http://schemas.openxmlformats.org/package/2006/relationships"></Relationships>';

  // format parameters : cx, cy, relation ID, cx, cy
  cDOCXPicTag = '</w:t></w:r><w:r><w:drawing><wp:inline distT="0" distB="0" distL="0" distR="0"><wp:extent cx="%d" cy="%d"/><wp:docPr id="1" name="Pic1"/>' +
    '<wp:cNvGraphicFramePr><a:graphicFrameLocks xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main" noChangeAspect="1"/></wp:cNvGraphicFramePr>' +
    '<a:graphic xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/main">' +
    '<a:graphicData uri="http://schemas.openxmlformats.org/drawingml/2006/picture">' +
    '<pic:pic xmlns:pic="http://schemas.openxmlformats.org/drawingml/2006/picture">' +
    '<pic:nvPicPr><pic:cNvPr id="0" name="Image"/><pic:cNvPicPr/></pic:nvPicPr>' +
    '<pic:blipFill><a:blip r:embed="%s"></a:blip><a:stretch><a:fillRect/></a:stretch></pic:blipFill>' +
    '<pic:spPr><a:xfrm><a:off x="0" y="0"/><a:ext cx="%d" cy="%d"/></a:xfrm><a:prstGeom prst="rect"><a:avLst/></a:prstGeom>' +
    '</pic:spPr></pic:pic></a:graphicData></a:graphic></wp:inline></w:drawing></w:r><w:r><w:t>';
  cODTPicTag = '<draw:frame draw:style-name="a0" draw:name="Img1" text:anchor-type="as-char" ' +
    'svg:x="0cm" svg:y="0cm" svg:width="%scm" svg:height="%scm" style:rel-width="scale" style:rel-height="scale">' +
    '<draw:image xlink:href="%s" xlink:type="simple" xlink:show="embed" xlink:actuate="onLoad"/><svg:title/><svg:desc/></draw:frame>';
  cXMLPicTag: array [TTemplateFormat] of string = ('', cDOCXPicTag, cODTPicTag);

  cDOCXDefContentFile = '[Content_Types].xml';
  cDOCXDefContentPNG = '<Default Extension="png" ContentType="image/png"/>';
  cDOCXDefContentJPG = '<Default Extension="jpg" ContentType="image/jpeg"/>';
  cDOCXDefContentEnd = '</Types>';

  cDOCXBlockFormat = '</w:t></w:r><w:r><w:rPr>%s</w:rPr><w:t xml:space="preserve">';
  cDOCXBlockNoFormat = '</w:t></w:r><w:r><w:t xml:space="preserve">';
  cDOCXSuperScript = '<w:vertAlign w:val="superscript"/>';
  cDOCXSubScript = '<w:vertAlign w:val="subscript"/>';

  cDOCXParStart = '<w:p>';
  cDOCXParEnd = '</w:p>';
  cDOCXRunStart = '<w:r>';
  cDOCXRunEnd = '</w:r>';
  cDOCXTextStart = '<w:t>';
  cDOCXTextEnd = '</w:t>';
  cXMLNonBrakeSpace = #$C2#$A0;

  cTblGridWStart = '<w:tblGrid>';
  cTblGridWEnd = '</w:tblGrid>';
  cTblGridWidth = '<w:gridCol w:w="%d"/>';
  cTblGridW1cm = 566;

  cRowFormatParam = '<w:trPr><w:trHeight w:val="%d"/></w:trPr>';

  cCellFormatStart = '<w:tcPr>';
  cCellFormatEnd = '</w:tcPr>';
  cCellFormatParam = '<w:tcPr>%s</w:tcPr>';
  cCellAlignParam = '<w:vAlign w:val="%s"/>';
  cCellRotateFormat = '<w:textDirection w:val="btLr"/>';
  // bk color
  cCellColorParam = '<w:shd w:val="clear" w:color="auto" w:fill="%s"/>';

  cParFormatParam = '<w:pPr>%s</w:pPr>';
  cParAlignParam = '<w:jc w:val="%s"/>';

  cRunFormatParam = '<w:rPr>%s</w:rPr>';
  // font color, font, size
  cRunColorParam = '<w:color w:val="%s"/>';
  cRunFontNameParam = '<w:rFonts w:ascii="%s" w:hAnsi="%s" w:cs="%s"/>';
  cRunFontSizeParam = '<w:sz w:val="%d"/><w:szCs w:val="%d"/>';

  cFormatSep = ':';
  cParamSep = ';';
  cListSep2 = ',';
  cConstSep = '"';
  cDateSep = '#';
  cTabStr = '<TAB>';

  // tags
  cTagCh = 'X';
  cTag = '&lt;&lt;X&gt;&gt;';

  // scan
  cUsrTagScan = '&lt;&lt;scan';
  cUsrTagScanEntry = '&lt;&lt;scanentry(X)&gt;&gt;';
  cUsrTagScanFooter = '&lt;&lt;scanfooter(X)&gt;&gt;';
  cUsrTagEndScan = '&lt;&lt;endscan(X)&gt;&gt;';
  cScanDS = 'scan';
  cScanWhileCond = 'while';
  cScanSkipCond = 'skipif';
  cScanNewLine = 'newline';
  cScanNewPar = 'newpar';
  cScanNewCol = 'newcol';
  cScanNewPage = 'newpage';
  cScanNewSep = 'sep';
  cScanNoIfEmpty = 'noifempty';
  cScanNoFirst = 'nofirst';
  cScanCalc = 'calc';
  cScanTblRow = 'tblrow';
  cScanMinRows = 'minrows';
  cScanMaxRows = 'maxrows';

  cScanNext = 'next(';
  cScanPrev = 'prev(';
  cScanFirst = 'first(';
  cScanLast = 'last(';
  cScanGetBookmark = 'getbookmark(';
  cScanGotoBookmark = 'gotobookmark(';

  cInsertTable = 'inserttable(X)';

  // if
  cUsrTagIfStart = '&lt;&lt;if';
  cUsrTagElseIf = '&lt;&lt;elseif&gt;&gt;';
  cUsrTagEndIf = '&lt;&lt;endif&gt;&gt;';
  cUsrTagIIfStart = '&lt;&lt;iif';
  cUsrTagElseIIf = '&lt;&lt;elseiif&gt;&gt;';
  cUsrTagEndIIf = '&lt;&lt;endiif&gt;&gt;';

  // case
  cCase = 'case(';

  // calc tag
  cCalcFunc: array [0 .. 9] of string = ('Sum[X]', 'Max[X]', 'Min[X]', 'Count[X]', 'Avg[X]', 'GrandSum[X]', 'GrandMax[X]', 'GrandMin[X]', 'GrandCount[X]',
    'GrandAvg[X]');
  cResetCalc = 'resetcalc';

  // datetime function
  cGetToday = 'Today[]';
  cGetNow = 'Now[]';

  // condition tag
  cCondFunc: array [0 .. 6] of string = ('==', '>=', '<=', '<>', '=', '>', '<');
  cCondAnd = ' and ';
  cCondOr = ' or ';
  cCondNot = 'not ';
  cCondResArr: array [boolean] of string = ('"FALSE"', '"TRUE"');

  // other
  cListAllFields = 'listallfields';
  cSepList = 'seplist';
  cFirstValue = 'firstvalue';
  cLabel = 'label';
  cPicture = 'picture';
  cDefVar = 'defvar';

  cExtendedInfo = '&lt;&lt;info(X)&gt;&gt;';
  cInsertFile = '&lt;&lt;insertfile(X)&gt;&gt;';
  cInsertXMLFile = '&lt;&lt;insertxmlfile(X)&gt;&gt;';

  cPicMaxW = 'w=';
  cPicMaxWcm = 'wcm=';
  cPicMaxWinch = 'winch=';
  cPicMaxH = 'h=';
  cPicMaxHcm = 'hcm=';
  cPicMaxHinch = 'hinch=';
  cPicEnlarge = 'enlarge';
  cPicNameField = 'namefield=';
  cPicSaveAsJPG = 'saveasjpg';
  cPicSaveAsPNG = 'saveaspng';
  // cPicInsertInRTF = 'insertinrtf';
  cPicQuality = 'quality=';
  cPicRotation = 'rotate=';

  cBarCode = 'barcode';
  cBCType = 'type';
  cBCShowCode = 'showcode';
  cBCFontName = 'fontname';
  cBCFontSize = 'fontsize';

  cNumToText = 'numtotext';

  // math function
  cMathAbs = 'abs[X]';

  // string routine tags
  cStrPrefix = 'str_';
  cStrPart = 'str_part';
  cStrPartCS = 'str_partcs';
  cStrRepl = 'str_replace';
  cStrReplCS = 'str_replacecs';
  cStrUpper = 'str_upper';
  cStrLower = 'str_lower';
  cStrProper = 'str_proper';
  cStrTrim = 'str_trim';

  // string routine function
  cStrLen = 'strlen[X]';
  cStrPos = 'strpos[X]';
  cStrPosCS = 'strposcs[X]';

  // RTF inlineformat
  cRtfInlineStart = '\*format_begin*\';
  cRtfInlineEnd = '\*format_end*\';
  cRtfInlineFont = '\font*';

  // TRtfSupportedFormat = (rtfFontName, rtfFontSize, rtfBold, rtfItalic, rtfUnderline, rtfUnderlineDouble, rtfStrike, rtfStrikeDouble);
  cRtfInlineFrm_Rtf1: array [TRtfSupportedFormat] of string = ('\f#', '\fs#', '\b', '\i', '\ul', '\uldb', '\strike', '\striked1');
  cRtfInlineFrm_Rtf2: array [TRtfSupportedFormat] of string = ('', '', '\b0', '\i0', '\ulnone', '\ulnone', '\strikenone', '\strikenone');
  cRtfInlineFrm_XML: array [TRtfSupportedFormat] of string = ('<w:rFonts w:ascii="%s" w:hAnsi="%s" w:cs="%s"/>', '<w:sz w:val="%d"/><w:szCs w:val="%d"/>',
    '<w:b/>', '<w:i/>', '<w:u w:val="single"/>', '<w:u w:val="double"/>', '<w:strike/>', '<w:dstrike/>');
  // alignment
  // L.Add('\qr');
  // L.Add('\qc');
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 10 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrXMLReport]);
end;

// ======================================================================
// Public actions
constructor TrXMLReport.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FileList:=TStringList.Create;
  FVariables:=TStringList.Create;
  FMemoReplaces:=TStringList.Create;
  FInlineFormatting:=TStringList.Create;
  FExtendedInfo:=TStringList.Create;
  LCalcField:=TList.Create;
  FTemplateFormat:=tfAutoDetect;
  FActionAfter:=aaShow;
  PrintCopies:=1;
  RaiseUnknownField:=false;
  FConvertScientificFormat:=false;
  FConvertInlineFormatting:=false;
  FJoinTouchingTables:=false;
  FLoadTableNamePrefix:='';
  FRtfSupportedFormats:=[];
  FRtfCustomFormats:=TStringList.Create;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrXMLReport.Destroy;
begin
  FreeCalcField;
  FreeAndNil(FileList);
  FreeAndNil(LCalcField);
  FreeAndNil(FVariables);
  FreeAndNil(FMemoReplaces);
  FreeAndNil(FInlineFormatting);
  FreeAndNil(FExtendedInfo);
  FreeAndNil(FRtfCustomFormats);
  inherited Destroy;
end;

procedure TrXMLReport.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

// run report
procedure TrXMLReport.Execute(DataSets: array of TDataset);
var
  i, m: integer;
  oldCursor: TCursor;
begin
  for i:=0 to maxDS do
  begin
    DS[i]:=nil;
    DS_eof[i]:=false;
  end;
  m:= High(DataSets);
  if m > maxDS then
    raise TReportException.Create('Too many datasets.');
  for i:=0 to m do
    DS[i]:=DataSets[i];
  if DSActive(0) then
    RecCountDS0:=DS[0].RecordCount
  else
    RecCountDS0:=-1;

  PicCount:=0;
  SetReportFormat(FReportFile);

  oldCursor:=crDefault;
  if FShowProgressBar then
    ShowInfo(sCreatingReport)
  else
    oldCursor:=WCursor;
  try
    ProcAllFiles;
  finally
    if FShowProgressBar then
      ShowInfo('')
    else
      ChCursor(oldCursor);
  end;
end;

procedure TrXMLReport.RaiseErr(ErrType: integer; const ErrStr: string = '');
var
  s: string;
begin
  case ErrType of
    0 .. 10:
      s:=Format(sTemplateFormatErr + #13 + cErrStr[ErrType], [XMLToStr(ErrStr)]);
    -2:
      s:=sCreatingReportErr + #13 + 'Unknown template type.';
    -3:
      s:=sTemplateFormatErr + #13 + 'Invalid table structure.';
  else
    s:=sCreatingReportErr + #13 + ErrStr;
  end;
  raise TReportException.Create(s);
end;

procedure TrXMLReport.SetReportFormat(FileName: string);
begin
  if FTemplateFormat = tfAutoDetect then
  begin
    if SameText(ExtractFileExt(FileName), '.docx') then
      XMLFormat:=tfDOCX;
    if SameText(ExtractFileExt(FileName), '.odt') then
      XMLFormat:=tfODT;
  end
  else
    XMLFormat:=FTemplateFormat;
end;

procedure TrXMLReport.CheckReportFormat(Frm: TTemplateFormat; Tag: string);
begin
  if XMLFormat <> Frm then
    RaiseErr(-3, Format('Tag "%s" is not supproted in this template format.', [Tag]));
end;

// ======================================================================
// AUX function
procedure TrXMLReport.SetVariables(Value: TStrings);
begin
  FVariables.Assign(Value);
end;

procedure TrXMLReport.SetMemoReplaces(Value: TStrings);
begin
  FMemoReplaces.Assign(Value);
end;

procedure TrXMLReport.SetInlineFormatting(Value: TStrings);
begin
  FInlineFormatting.Assign(Value);
end;

procedure TrXMLReport.SetRtfCustomFormats(Value: TStrings);
begin
  FRtfCustomFormats.Assign(Value);
end;

procedure TrXMLReport.DoAbort(Sender: TObject);
begin
  Aborted:=true;
end;

// separate content of first brackets
procedure TrXMLReport.SepBracket(const Str: string; var Bracket, Block1, Block2: string);
var
  i, b1, b2, bpoc: integer;
  constsep: boolean;
begin
  // check brackets error
  bpoc:=0;
  constsep:=false;
  b1:=0;
  for i:=1 to Length(Str) do
  begin
    if Str[i] = cConstSep then
      constsep:=not constsep;
    if constsep then
      Continue;
    if Str[i] = '(' then
    begin
      Inc(bpoc);
      if b1 = 0 then
        b1:=i;
    end;
    if Str[i] = ')' then
      Dec(bpoc);
  end;
  if (bpoc <> 0) or constsep then
    RaiseErr(0, Str);
  // extract outer bracket
  Bracket:='';
  Block1:='';
  Block2:='';
  // b1:=Pos('(',Str);
  if b1 = 0 then
    Block1:=Str
  else
  begin
    bpoc:=1;
    b2:=0;
    for i:=b1 + 1 to Length(Str) do
    begin
      if Str[i] = cConstSep then
        constsep:=not constsep;
      if constsep then
        Continue;
      if Str[i] = '(' then
        Inc(bpoc);
      if Str[i] = ')' then
        Dec(bpoc);
      if (bpoc = 0) then
      begin
        b2:=i;
        Bracket:=(copy(Str, b1 + 1, b2 - b1 - 1));
        Block1:=(copy(Str, 1, b1 - 1));
        Block2:=(copy(Str, b2 + 1, Length(Str)));
        Break;
      end;
    end;
    if b2 = 0 then
      RaiseErr(0, Str);
  end;
end;

// Find position and content of Tag
function TrXMLReport.FindTagStr(const TagStr: string; var s: string; WChar: Char; var StartPos: integer; Del: boolean): string;
var
  pos1, pos2: integer;
  part1, part2: string;
begin
  Result:='';
  pos1:=StartPos;
  StartPos:=0;
  part1:=copy(TagStr, 1, AnsiPos(WChar, TagStr) - 1);
  if part1 = '' then
    Exit;
  pos1:=AnsiPosEx(part1, s, pos1, false);
  if pos1 = 0 then
    Exit;
  part2:=copy(TagStr, AnsiPos(WChar, TagStr) + 1, Length(TagStr));
  if part2 = '' then
    Exit;
  pos2:=AnsiPosEx(part2, s, pos1 + Length(part1), false);
  if pos2 = 0 then
    Exit;
  Result:=copy(s, pos1 + Length(part1), pos2 - pos1 - Length(part1));
  StartPos:=pos1;
  if Del then
    Delete(s, pos1, pos2 - pos1 + Length(part2));
end;

// Convert 8bit chars to XML
// Character                Name	   Entity Reference	Character Reference	Numeric Reference
// Ampersand	              &amp;	   &	   &#38;#38;
// Left angle bracket	      &lt;	   <	   &#38;#60;
// Right angle bracket	    &gt;	   >	   &#62;
// Straight quotation mark	&quot;	 "	   &#39;
// Apostrophe	              &apos;	 '	   &#34;
function TrXMLReport.StrToXML(const Str: string): string;
var
  i: integer;
  s: string;
  UTFstr: RawByteString;
begin
  s:=ReplaceStr('&', '&amp;', Str, true, true);
  s:=ReplaceStr('<', '&lt;', s, true, true);
  s:=ReplaceStr('>', '&gt;', s, true, true);
  s:=ReplaceStr('"', '&quot;', s, true, true);
  s:=ReplaceStr('''', '&apos;', s, true, true);
  s:=ReplaceStr(#9, cXMLTab[XMLFormat], s, true, true);
  s:=ReplaceStr(cTxtEndPar, cXMLEndPar[XMLFormat], s, true, true);
  UTFstr:=(UTF8Encode(s));
  Result:='';
  for i:=1 to Length(UTFstr) do
    Result:=Result + string(AnsiChar(UTFstr[i]));
end;

// Convert XML to 8bit chars
function TrXMLReport.XMLToStr(const XML: string): string;
var
  s: string;
begin
  s:=ReplaceStr('&amp;', '&', XML, true, true);
  s:=ReplaceStr('&lt;', '<', s, true, true);
  s:=ReplaceStr('&gt;', '>', s, true, true);
  s:=ReplaceStr('&quot;', '"', s, true, true);
  s:=ReplaceStr('&apos;', '''', s, true, true);
  s:=UTF8ToString(RawByteString(s));
  Result:=s;
end;

function TrXMLReport.GetListPart(var Str: string; const Sep: string): string;

  function GetFirstSep(Sep, Str: string): integer;
  var
    i, constc, br1c, br2c: integer;
  begin
    // return pos of separator out of constant and brackets
    i:=1;
    constc:=0;
    br1c:=0;
    br2c:=0;
    while i <= Length(Str) do
    begin
      if Str[i] = Sep then
        if (constc = 0) and (br1c = 0) and (br2c = 0) then
          Break;
      case Str[i] of
        cConstSep:
          constc:=(constc + 1) mod 2;
        '(':
          if constc = 0 then
            Inc(br1c);
        ')':
          if constc = 0 then
            Dec(br1c);
        '[':
          if constc = 0 then
            Inc(br2c);
        ']':
          if constc = 0 then
            Dec(br2c);
      end;
      i:=i + 1;
    end;
    if i > Length(Str) then
      i:=0;
    Result:=i;
  end;

var
  p1, posstart: integer;
  s, brack, Block1, Block2: string;
begin
  // return and extract first paramater in list
  Str:=Trim(Str);
  if copy(Str, 1, 1) = cConstSep then
  begin
    // constant: "str"
    posstart:=2;
    repeat
      p1:=AnsiPosEx(cConstSep, Str, posstart, false);
      if p1 = 0 then
        RaiseErr(5, Str);
      if copy(Str, p1 + 1, 1) = cConstSep then
      begin
        // "" => replace to "
        Str:=copy(Str, 1, p1) + copy(Str, p1 + 2, Length(Str));
        posstart:=p1 + 1;
      end
      else
        Break;
    until false;
    s:=copy(Str, 1, p1);
    Str:=Trim(copy(Str, p1 + 1, Length(Str)));
  end
  else
    if AnsiPosEx(cStrPrefix, Str, 1, false) = 1 then
    begin
      // string routine
      SepBracket(Str, brack, Block1, Block2);
      s:=Block1 + '(' + brack + ')';
      Str:=Trim(Block2);
    end
    else
    begin
      // value
      p1:=GetFirstSep(Sep, Str);
      if p1 = 0 then
        s:=Str
      else
        s:=Trim(copy(Str, 1, p1 - 1));
      Str:=Trim(copy(Str, Length(s) + 1, Length(Str)));
    end;
  // remove separator
  if Str <> '' then
  begin
    if copy(Str, 1, 1) <> Sep then
      RaiseErr(5, Str);
    Str:=Trim(copy(Str, 2, Length(Str)));
  end;
  Result:=s;
end;

procedure TrXMLReport.ClearEOLPart(var s: string);
var
  i: integer;
begin
  repeat
    i:=Pos(cTxtEndPar, s);
    if i = 0 then
      Break;
    Delete(s, i, Length(cTxtEndPar));
  until false;
end;

procedure TrXMLReport.ExtractIntXMLTags(var s: string);
var
  i, oLen, p1, p2: integer;
  t, pref, suf, newt, itg: string;
begin
  i:=1;
  repeat
    // find any tag
    t:=FindTagStr(cTag, s, cTagCh, i, false);
    if i = 0 then
      Break;
    // init
    oLen:=Length(t) + Length(cTag) - Length(cTagCh);
    pref:='';
    suf:='';
    newt:='';
    // extract XML tags inside of << and >>
    repeat
      p1:=Pos('<', t);
      p2:=Pos('>', t);
      if (p1=0) or (p2<p1) then
      begin
        newt:=newt + t;
        Break;
      end;
      itg:=copy(t, p1, p2-p1+1);
      if (p1=1) and (newt='') then
      begin
        pref:=pref + itg;
      end
      else
      begin
        newt:=newt + copy(t,1,p1-1);
        suf:=suf+itg;
      end;
      t:=copy(t, p2 + 1, oLen);
    until false;

    // replace original tag by new tag
    if t <> newt then
    begin
      // delete old tag
      Delete(s, i, oLen);
      // replace spaces before and after by Non-brake spaces (trailing spaces are not used by default)
      if copy(s, i - 1, 1) = ' ' then
      begin
        Delete(s, i - 1, 1);
        Insert(cXMLNonBrakeSpace, s, i - 1);
        Inc(i, Length(cXMLNonBrakeSpace) - 1);
      end;
      if copy(s, i, 1) = ' ' then
      begin
        Delete(s, i, 1);
        Insert(cXMLNonBrakeSpace, s, i);
      end;
      // generate new tag with extracted prefix and suffix
      newt:=pref + rtool.ReplaceStr(cTagCh, newt, cTag) + suf;
      Insert(newt, s, i);
    end;
    // move index
    Inc(i, oLen);
  until false;
end;

{
  procedure TrXMLReport.ClearBookmarkPart(var s: string);
  var
  p1, p2: integer;
  Str: string;
  begin
  // remove <w:bookmarkStart w:id="0" w:name="_GoBack"/>
  repeat
  p1:=Pos(cXMLBookmarkStart, s);
  if p1 = 0 then
  Break;
  Str:=copy(s, p1, 200);
  p2:=Pos(cXMLSimpleTagEnd, Str);
  if p2 > 0 then
  Delete(s, p1, p2 + 1);
  until false;
  // remove <w:bookmarkEnd w:id="0"/>
  repeat
  p1:=Pos(cXMLBookmarkEnd, s);
  if p1 = 0 then
  Break;
  Str:=copy(s, p1, 200);
  p2:=Pos(cXMLSimpleTagEnd, Str);
  if p2 > 0 then
  Delete(s, p1, p2 + 1);
  until false;
  // remove rest of tags
  repeat
  p1:=Pos(cXMLTagRemove1[XMLFormat], s);
  if p1 = 0 then
  Break;
  Delete(s, p1, Length(cXMLTagRemove1[XMLFormat]));
  until false;
  end;
}

procedure TrXMLReport.ClearBookmarkPart(var s: string);
const
  bb = '</w:t></w:r>';
var
  p1, p2, p: integer;
  Str: string;
begin
  // remove <w:bookmarkStart w:id="0" w:name="_GoBack"/>  from </w:t> to <w:t>
  repeat
    p:=Pos(cXMLBookmarkStart, s);
    if (p < Length(bb)) or not SameText(copy(s, p - Length(bb), Length(bb)), bb) then
      Break;
    p1:=p - Length(bb);
    p:=PosEx(cXMLBookmarkEnd, s, p);
    if p = 0 then
      Break;
    p:=PosEx(cXMLSimpleTagEnd, s, p);
    if p = 0 then
      Break;
    p:=p + Length(cXMLSimpleTagEnd);
    p2:=PosEx('<w:t>', s, p);
    if (p2 = 0) or (p2 - p > 5) then
      Break;
    // DlgI(copy(s, p1, p2 - p1 + 5));
    Delete(s, p1, p2 - p1 + 5);
  until false;

  // remove <w:bookmarkStart w:id="0" w:name="_GoBack"/> only
  repeat
    p1:=Pos(cXMLBookmarkStart, s);
    if p1 = 0 then
      Break;
    Str:=copy(s, p1, 200);
    p2:=Pos(cXMLSimpleTagEnd, Str);
    if p2 > 0 then
      Delete(s, p1, p2 + 1);
  until false;
  // remove <w:bookmarkEnd w:id="0"/>
  repeat
    p1:=Pos(cXMLBookmarkEnd, s);
    if p1 = 0 then
      Break;
    Str:=copy(s, p1, 200);
    p2:=Pos(cXMLSimpleTagEnd, Str);
    if p2 > 0 then
      Delete(s, p1, p2 + 1);
  until false;

  // remove rest of tags
  repeat
    p1:=Pos(cXMLTagRemove1[XMLFormat], s);
    if p1 = 0 then
      Break;
    Delete(s, p1, Length(cXMLTagRemove1[XMLFormat]));
  until false;

end;

procedure TrXMLReport.ClearEmptyParPart(var s: string);
var
  p1, p2, t1, t2: integer;
  fs: string;
begin
  if cXMLParBeginAll[XMLFormat] <> '' then
  begin
    repeat
      fs:=cXMLParBeginAll[XMLFormat] + cXMLParEndAll[XMLFormat];
      p1:=Pos(fs, s);
      if p1 = 0 then
        Break;
      Delete(s, p1, Length(fs));
    until false;
  end;

  if cXMLParBeginPartStart[XMLFormat] <> '' then
  begin
    p1:=1;
    repeat
      t1:=PosEx(cXMLTableTagBegin[XMLFormat], s, p1);
      p1:=PosEx(cXMLParBeginPartStart[XMLFormat], s, p1);
      if p1 = 0 then
        Break;
      // skip table content = block removing empty par in cell
      if (t1 > 0) and (p1 > t1) then
      begin
        t2:=PosEx(cXMLTableTagEnd[XMLFormat], s, t1);
        p1:=t2;
        Continue;
      end;

      p2:=PosEx(cXMLParBeginPartEnd[XMLFormat], s, p1);
      if p2 = 0 then
        Break;
      fs:=cXMLParBeginPartCont[XMLFormat] + cXMLParEndAll[XMLFormat];
      if copy(s, p2 + Length(cXMLParBeginPartEnd[XMLFormat]), Length(fs)) = fs then
      begin
        // dlgi(copy(s,p1,p2 - p1 +  Length(cXMLParBeginPartEnd[XMLFormat]) + Length(fs)));
        Delete(s, p1, p2 - p1 + Length(cXMLParBeginPartEnd[XMLFormat]) + Length(fs));
      end
      else
        p1:=p2 + Length(cXMLParBeginPartEnd[XMLFormat]);
    until false;
  end;
end;

procedure TrXMLReport.JoinTables(var s: string);
var
  p1, p2: integer;
begin
  p1:=1;
  repeat
    p1:=PosEx(cXMLTableJoinTagStart[XMLFormat], s, p1);
    if p1 = 0 then
      Break;
    p2:=PosEx(cXMLTableJoinTagEnd[XMLFormat], s, p1);
    if p2 = 0 then
      Break;
    Delete(s, p1, p2 - p1 + Length(cXMLTableJoinTagEnd[XMLFormat]));
  until false;
end;

// ======================================================================
// CALC FIELDS function
// Init CalcFields
procedure TrXMLReport.InitCalcField(DSindex: integer; const CalcStr: string);
var
  part, name, exp: string;
  p1, p2: integer;
  isexist: boolean;
begin
  // proc CalcStr
  p1:=1;
  repeat
    part:=AnsiUpperCase(Trim(FindStrPart(CalcStr, cListSep2, p1, false)));
    if part = '' then
      Break;
    // add DSindex for backward compability
    if (Pos('.', part) = 0) and (Pos('=', part) = 0) then
      part:=IndexToDatasetChar(DSindex) + '.' + part;

    // proc one part (FieldName or expression)
    p2:=Pos('=', part);
    if p2 = 0 then
    begin
      name:=part;
      exp:=part;
    end
    else
    begin
      name:=copy(part, 1, p2 - 1);
      exp:=copy(part, p2 + 1, Length(part));
    end;
    isexist:=false;
    for p2:=0 to LCalcField.Count - 1 do
    begin
      cf:=LCalcField.Items[p2];
      if (cf.DSindex = DSindex) and (cf.Name = name) then
      begin
        // already exist, only reset
        isexist:=true;
        cf.Expression:=exp;
        cf.Max:=0;
        cf.Min:=0;
        cf.Sum:=0;
        cf.Count:=0;
      end;
    end;
    if not isexist then
    begin
      // add new calcfield
      cf:=TCalcField.Create;
      cf.Init:=false;
      cf.DSindex:=DSindex;
      cf.Name:=name;
      cf.Expression:=exp;
      LCalcField.Add(cf);
    end;
    Inc(p1);
  until false;
end;

// Free all CalcFields
procedure TrXMLReport.FreeCalcField;
var
  i: integer;
begin
  for i:=LCalcField.Count - 1 downto 0 do
  begin
    cf:=LCalcField.Items[i];
    FreeAndNil(cf);
    LCalcField.Delete(i);
  end;
end;

procedure TrXMLReport.ProcCalcField(DSindex: integer);
var
  i: integer;
  r: double;
begin
  for i:=0 to LCalcField.Count - 1 do
    try
      cf:=LCalcField.Items[i];
      if cf.DSindex = DSindex then
      begin
        r:=StrToFloat(GetCalcValue(cf.Expression, '', true));
        if not cf.Init then
        begin
          cf.Max:=r;
          cf.Min:=r;
          cf.Sum:=r;
          cf.Count:=1;
          cf.GrandMax:=r;
          cf.GrandMin:=r;
          cf.GrandSum:=r;
          cf.GrandCount:=1;
          cf.Init:=true;
        end
        else
        begin
          if cf.Max < r then
            cf.Max:=r;
          if cf.Min > r then
            cf.Min:=r;
          cf.Sum:=cf.Sum + r;
          cf.Count:=cf.Count + 1;
          if cf.GrandMax < r then
            cf.GrandMax:=r;
          if cf.GrandMin > r then
            cf.GrandMin:=r;
          cf.GrandSum:=cf.GrandSum + r;
          cf.GrandCount:=cf.GrandCount + 1;
        end;
      end;
    except
      RaiseErr(1, cf.Expression);
    end;
end;

// Return value of CalcField
function TrXMLReport.GetCalcFieldValue(CalcTypeIndex: integer; const Name: string): double;
var
  i: integer;
  r: double;
  notfound: boolean;
begin
  notfound:=true;
  r:=0;
  for i:=0 to LCalcField.Count - 1 do
  begin
    cf:=LCalcField.Items[i];
    if (cf.Name = AnsiUpperCase(Name)) then
    begin
      notfound:=false;
      case CalcTypeIndex of
        0:
          r:=cf.Sum;
        1:
          r:=cf.Max;
        2:
          r:=cf.Min;
        3:
          r:=cf.Count;
        4:
          if cf.Count > 0 then
            r:=cf.Sum / cf.Count
          else
            r:=0;
        5:
          r:=cf.GrandSum;
        6:
          r:=cf.GrandMax;
        7:
          r:=cf.GrandMin;
        8:
          r:=cf.GrandCount;
        9:
          if cf.GrandCount > 0 then
            r:=cf.GrandSum / cf.GrandCount
          else
            r:=0;
      else
        notfound:=true;
      end;
    end;
  end;
  if notfound then
    RaiseErr(1, Name);
  Result:=r;
end;

// ======================================================================
// Dataset MOVE function
// Decode dataset index form tag "text(X)"
function TrXMLReport.IsConstantTag(const TagStr: string): boolean;
begin
  Result:=(copy(TagStr, 1, 1) = cConstSep) and (copy(TagStr, Length(TagStr), 1) = cConstSep);
end;

function TrXMLReport.IsNumericTag(const TagStr: string; var Value: double): boolean;
var
  s: string;
begin
  // Check numeric value
  Result:=false;
  try
    s:=UpdateToDecimalSeparator(TagStr);
    Value:=StrToFloat(s);
    Result:=true;
  except
  end;
end;

function TrXMLReport.IsFieldTag(const TagStr: string): boolean;
begin
  Result:=(Length(TagStr) > 2) and CharInSet(TagStr[1], setDS) and (TagStr[2] = '.');
end;

function TrXMLReport.DatasetCharToIndex(Ch: Char): integer;
begin
  Result:=ord(Ch) - ord('A');
end;

function TrXMLReport.IndexToDatasetChar(Idx: integer): Char;
begin
  Result:=Chr(ord('A') + Idx);
end;

function TrXMLReport.DSActive(DSindex: integer): boolean;
begin
  Result:=(DSindex in [0 .. maxDS]) and Assigned(DS[DSindex]) and DS[DSindex].Active;
end;

function TrXMLReport.DSNotEOF(DSindex: integer): boolean;
begin
  Result:=(DSindex < 0) or (DSActive(DSindex) and not DS[DSindex].EOF);
end;

function TrXMLReport.GetDSIndex(const TagStr: string): integer;
var
  br, Block1, Block2: string;
begin
  SepBracket(TagStr, br, Block1, Block2);
  br:=AnsiUpperCase(Trim(br));
  if br = '' then
    Result:=-1
  else
  begin
    if (Length(br) > 1) or not CharInSet(br[1], setDS) then
      RaiseErr(2, TagStr);
    Result:=DatasetCharToIndex(br[1]);
  end;
end;

function TrXMLReport.CheckDSIndex(DSindex: integer): string;
begin
  Result:='';
  if DS[DSindex] = nil then
    Result:='unknown dataset'
  else
    if not DS[DSindex].Active then
      Result:='closed dataset';
end;

function TrXMLReport.FindFirst(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].First;
  Result:=not DS[DSindex].EOF;
  DS_eof[DSindex]:=DS[DSindex].EOF;
end;

function TrXMLReport.FindLast(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Last;
  Result:=not DS[DSindex].BOF;
  DS_eof[DSindex]:=DS[DSindex].BOF;
end;

function TrXMLReport.FindNext(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Next;
  Result:=not DS[DSindex].EOF;
  DS_eof[DSindex]:=DS[DSindex].EOF;
  if not Result then
    Exit;
  Inc(RecNbr);
  if DSindex = 0 then
    Inc(RecNbrDS0);
end;

function TrXMLReport.FindPrev(DSindex: integer): boolean;
begin
  Result:=false;
  DS_eof[DSindex]:=true;
  if not DSActive(DSindex) then
    Exit;
  DS[DSindex].Prior;
  Result:=not DS[DSindex].BOF;
  DS_eof[DSindex]:=DS[DSindex].BOF;
  if not Result then
    Exit;
  Inc(RecNbr);
end;

function TrXMLReport.GetDSBookmark(DSindex: integer): boolean;
begin
  Result:=false;
  if not DSActive(DSindex) then
    Exit;
  {$IFNDEF SkipBookmark}
  DS_Bookmarks[DSindex]:=DS[DSindex].GetBookmark;
  {$ENDIF}
end;

function TrXMLReport.GotoDSBookmark(DSindex: integer): boolean;
begin
  Result:=false;
  if not DSActive(DSindex) then
    Exit;
  {$IFNDEF SkipBookmark}
  if not Assigned(DS_Bookmarks[DSindex]) then
    Exit;
  DS[DSindex].GotoBookmark(DS_Bookmarks[DSindex]);
  DS_eof[DSindex]:=DS[DSindex].EOF;
  {$ENDIF}
end;

procedure TrXMLReport.FreeBookMarks;
var
  i: integer;
begin
  {$IFNDEF SkipBookmark}
  for i:=0 to maxDS do
    if Assigned(DS_Bookmarks[i]) and Assigned(DS[i]) then
    begin
      DS[i].FreeBookmark(DS_Bookmarks[i]);
      DS_Bookmarks[i]:=nil;
    end;
  {$ENDIF}
end;

// ======================================================================
// IN-BAND function
// Create list of all datasets fields
function TrXMLReport.ListAllFields: string;
var
  i, j: integer;
  s: string;
begin
  s:='';
  for i:=0 to maxDS do
  begin
    if not Assigned(DS[i]) then
      Continue;
    s:=s + cTxtEndPar + IndexToDatasetChar(i) + ' (' + DS[i].Name + ')' + cTxtEndPar;
    for j:=0 to DS[i].FieldCount - 1 do
      s:=s + DS[i].Fields[j].FieldName + '; ';
  end;
  Result:=s + cTxtEndPar;
end;

// Separate list values with separator, if value is not empty
function TrXMLReport.ProcSepList(const TagStr: string): string;
var
  i: integer;
  br, part, Sep, res: string;
  Repl, addsep: boolean;
begin
  Result:='';
  SepBracket(TagStr, br, Sep, res);
  addsep:=false;
  // find and clear separator
  Sep:=copy(FindStrFromTo(cConstSep, cConstSep, br, false), 2, 255);
  Sep:=ReplaceStr('*|', cTxtEndPar, Sep, true, false);
  br:=ReplaceStrFromTo(cConstSep, cConstSep, '', br, true, false, false);
  i:=0;
  repeat
    Inc(i);
    part:=FindStrPart(br, cListSep2, i, false);
    if (i = 1) and (part = '') then
      Continue;
    if part = '' then
      Break;
    res:=ProcTagStr(Trim(part), Repl);
    if addsep and (res <> '') then
      Result:=Result + Sep;
    addsep:=addsep or (res <> '');
    Result:=Result + res;
  until false;
end;

// Find and return first not empty value from list
function TrXMLReport.ProcFirstValue(const TagStr: string): string;
var
  i: integer;
  br, part, Sep, res: string;
  Repl: boolean;
begin
  Result:='';
  SepBracket(TagStr, br, Sep, res);
  i:=0;
  repeat
    Inc(i);
    part:=FindStrPart(br, cListSep2, i, false);
    if part = '' then
      Break;
    res:=ProcTagStr(Trim(part), Repl);
    Result:=res;
    if Result <> '' then
      Break;
  until false;
end;

// Insert external picture to report
function TrXMLReport.ProcPicture(const TagStr: string): string;
var
  fn, br, Block1, Block2, Frm: string;
  Repl: boolean;
  B: TBitmap;
begin
  Result:='';
  SepBracket(TagStr, br, Block1, Block2);
  fn:=FindStrPart(br, cFormatSep, 1, false);
  if Block2 <> '' then
    Frm:=FindStrPart(Block2, cFormatSep, 2, false)
  else
    Frm:=FindStrPart(br, cFormatSep, 2, false);
  fn:=GetVarUserValue(fn, Repl);
  if Assigned(FOnGetPictureContent) then
  begin
    B:=TBitmap.Create;
    try
      FOnGetPictureContent(Self, fn, B);
      if B.Width>0 then
      begin
        Result:=CreateFormatPicture(-1, nil, '', B, frm);
        Repl:=false;
      end;
    finally
      B.Free;
    end;
  end;
  if Repl then
    Result:=CreateFormatPicture(-1, nil, fn, nil, Frm);
end;
{$IFDEF BARCODESUPPORT}

// Create BarCode picture and insert to report
function TrXMLReport.ProcBarcode(const TagStr: string): string;
var
  i: integer;
  fn, br, Block1, Block2, Frm, FPart: string;
  Repl: boolean;
  TypName, FName: string;
  FSize: integer;
  ShowCode: boolean;
  Bmp: TBitmap;
  bcType: TBarcodeType;
  bcRect: TRect;
  bcWidth: integer;
  bcFont: TFont;
begin
  // proc tag parts
  Result:='';
  Frm:=FindStrPart(TagStr, cFormatSep, 2, false);
  fn:=FindStrPart(TagStr, cFormatSep, 1, false);
  SepBracket(fn, br, Block1, Block2);
  fn:=br;
  if IsConstantTag(fn) then
    fn:=copy(fn, 2, Length(fn) - 2)
  else
  begin
    fn:=ProcTagStr(fn, Repl);
    if not Repl then
    begin
      Result:='?' + fn + '?';
      Exit;
    end;
  end;
  if fn = '' then
    Exit;

  // proc type and format
  TypName:='';
  ShowCode:=false;
  FName:='Verdana';
  FSize:=0;
  i:=0;
  repeat
    Inc(i);
    FPart:=Trim(rtool.FindStrPart(Frm, cParamSep, i, false));
    if FPart = '' then
      Break;
    if AnsiPosEx(cBCType, FPart, 1, false) = 1 then
      TypName:=FindStrPart(FPart, '=', 2, false);
    if SameText(cBCShowCode, FPart) then
      ShowCode:=true;
    if AnsiPosEx(cBCFontName, FPart, 1, false) = 1 then
      FName:=FindStrPart(FPart, '=', 2, false);
    if AnsiPosEx(cBCFontSize, FPart, 1, false) = 1 then
      FSize:=StrToIntDef(FindStrPart(FPart, '=', 2, false), 0);
  until false;

  // create barcode picture
  Bmp:=TBitmap.Create;
  try
    bcFont:=TFont.Create;
    try
      try
        // set default size for barcode
        bcRect:=Rect(5, 5, 490, 115);
        Bmp.Width:=500;
        Bmp.Height:=120;
        bcWidth:=2; // too points for single lines
        // set other props
        bcFont.Name:=FName;
        if FSize > 0 then
          bcFont.Size:=FSize;
        // type of barcode
        bcType:=GetBarCodeType(TypName);
        if SameText(TypName, 'CodeEANAuto') then
          case Length(fn) of
            7, 8:
              bcType:=bcCodeEAN8;
            12, 13:
              bcType:=bcCodeEAN13;
          else
            begin
              Result:='Invalid text length (CodeEANAuto): ' + fn;
              Exit;
            end;
          end;
        // check type
        if bcType = bcTypeUndefined then
        begin
          if TypName = '' then
            TypName:='undefined barcode type';
          Result:='?' + TypName + '?';
          Exit;
        end;
        // create barcode
        AsBarcode_interface.DrawBarCode(fn, bcType, Bmp.Canvas, bcRect, bcWidth, bcFont, FSize = 0, ShowCode, stpBottomCenter);
        // set picture size according barcode
        Bmp.Width:=bcRect.Right + 10;
        Bmp.Height:=bcRect.Bottom + 10;
        // insert pic
        Result:=CreateFormatPicture(-1, nil, '', Bmp, Frm);
      except
        on E: Exception do
          Result:=E.Message;
      end;
    finally
      FreeAndNil(bcFont);
    end;
  finally
    FreeAndNil(Bmp);
  end;
end;
{$ELSE}

function TrXMLReport.ProcBarcode(const TagStr: string): string;
begin
  Result:='BarCode is not supported.';
end;
{$ENDIF}

// proc graphics BLOB field or external graphics file or created Bmp
function TrXMLReport.CreateFormatPicture(DSindex: integer; F: TField; const FileName: string; Bmp: TBitmap; const FormatStr: string): string;
var
  Pic: TPicture;
  Enlarge, AutoQ: boolean;
  FPart, NameField, temp: string;
  i, MaxW, MaxH, Quality, Rot: integer;
  GrType: TrGraphicType;
  EnumToCmRatio: double;

  function FormatSizeCm(v: integer): string;
  begin
    Result:=IntToStr(v div 1000) + '.' + IntToStr(v mod 1000);
  end;

  function InsertPic: string;
  var
    NewW, NewH: integer;
    ID: string;
  begin
    Result:='';
    // resize according quality and size
    if Quality < 10 then
      Quality:=10;
    if Quality > 100 then
      Quality:=100;
    if AutoQ then
      rpictool.SmoothResizePicture(Pic, MaxW div 20, MaxH div 20, Enlarge)
    else
      if Quality <> 100 then
        rpictool.SmoothResizePicture(Pic, Pic.Width * Quality div 100, Pic.Height * Quality div 100, Enlarge);

    NewW:=Round(Pic.Width * emusPerPix * EnumToCmRatio);

    if (MaxW > 0) and (NewW > MaxW) or (Enlarge and (NewW < MaxW)) then
      NewW:=MaxW;
    NewH:=MulDiv(NewW, Pic.Height, Pic.Width);
    if (MaxH > 0) and ((NewH > MaxH) or (Enlarge and (NewH < MaxH))) then
    begin
      NewH:=MaxH;
      NewW:=MulDiv(NewH, Pic.Width, Pic.Height);
    end;
    ID:=AddImageFile(Pic, GrType);
    case XMLFormat of
      tfDOCX:
        // format parameters : cx, cy, relation ID, cx, cy  (in enum)
        Result:=Format(cXMLPicTag[XMLFormat], [NewW, NewH, ID, NewW, NewH]);
      tfODT:
        // format parameters : cx, cy, image filename, (in cm)
        Result:=Format(cXMLPicTag[XMLFormat], [FormatSizeCm(NewW), FormatSizeCm(NewH), ID]);
    else
      Result:='';
    end;
  end;

  function GetPartNum(param: string; Ratio: double): integer;
  var
    s: string;
  begin
    try
      s:=FindStrPart(param, '=', 2, false);
      s:=UpdateToDecimalSeparator(s);
      Result:=Round(StrToFloat(s) * Ratio);
    except
      Result:=0;
    end;
  end;

begin
  EnumToCmRatio:=1;
  if XMLFormat = tfODT then
    EnumToCmRatio:=1000 / emusPerCm;

  Pic:=TPicture.Create;
  try
    try
      // load pic form field or from file
      if Assigned(F) then // from field
      begin
        GrType:=ReadPictureFromField(F, Pic);
        if GrType = gtError then
          raise Exception.Create(sPictureErr);
      end
      else
        if FileName <> '' then // from external file
        begin
          if not FileExists(FileName) then
            raise Exception.Create(Format(sFileNotFound, [FileName]));
          GrType:=LoadPictureFromFile(Pic, FileName);
          if GrType = gtError then
            raise Exception.Create(sPictureErr);
        end
        else
          if Assigned(Bmp) then // from existing bitmap
          begin
            Pic.Bitmap.Assign(Bmp);
            GrType:=gtBMP;
          end
          else
            raise Exception.Create(sInputErr);
      // BMP is not supported in XML, change to PNG
      if GrType = gtBMP then
        GrType:=gtPNG;

      // proc format
      MaxW:=0;
      MaxH:=0;
      Rot:=0;
      Enlarge:=false;
      // Insert:=true;
      Quality:=100;
      AutoQ:=false;
      NameField:='';
      i:=0;

      repeat
        Inc(i);
        FPart:=Trim(rtool.FindStrPart(FormatStr, cParamSep, i, false));
        if FPart = '' then
          Break;
        if AnsiPosEx(cPicMaxW, FPart, 1, false) = 1 then
          MaxW:=GetPartNum(FPart, emusPerPix * EnumToCmRatio);
        if AnsiPosEx(cPicMaxWcm, FPart, 1, false) = 1 then
          MaxW:=GetPartNum(FPart, emusPerCm * EnumToCmRatio);
        if AnsiPosEx(cPicMaxWinch, FPart, 1, false) = 1 then
          MaxW:=GetPartNum(FPart, emusPerInch * EnumToCmRatio);
        if AnsiPosEx(cPicMaxH, FPart, 1, false) = 1 then
          MaxH:=GetPartNum(FPart, emusPerPix * EnumToCmRatio);
        if AnsiPosEx(cPicMaxHcm, FPart, 1, false) = 1 then
          MaxH:=GetPartNum(FPart, emusPerCm * EnumToCmRatio);
        if AnsiPosEx(cPicMaxHinch, FPart, 1, false) = 1 then
          MaxH:=GetPartNum(FPart, emusPerInch * EnumToCmRatio);

        if AnsiPosEx(cPicRotation, FPart, 1, false) = 1 then
          try
            temp:=FindStrPart(FPart, '=', 2, false);
            temp:=UpdateToDecimalSeparator(temp);
            Rot:=StrToInt(temp);
          except
            Rot:=0;
          end;
        if SameText(cPicEnlarge, FPart) then
          Enlarge:=true;
        if AnsiPosEx(cPicNameField, FPart, 1, false) = 1 then
          NameField:=Trim(FindStrPart(FPart, '=', 2, false));
        // if SameText(cPicInsertInRTF, FPart) then Insert:=true;
        if AnsiPosEx(cPicQuality, FPart, 1, false) = 1 then
          try
            temp:=Trim(FindStrPart(FPart, '=', 2, false));
            if AnsiPosEx('auto', temp, 1, false) = 1 then
              AutoQ:=true
            else
              Quality:=StrToInt(temp);
          except
            Quality:=100;
          end;
        if rpictool.IsJPGSupported then
          if SameText(cPicSaveAsJPG, FPart) then
            GrType:=gtJPG;
        if rpictool.IsPNGSupported then
          if SameText(cPicSaveAsPNG, FPart) then
            GrType:=gtPNG;
      until false;
      // rotate pic
      if Rot <> 0 then
        rpictool.RotatePicture(Pic, Rot);
      // insert or link pic
      // if Insert then
      Result:=InsertPic;
      {
        else
        begin
        // resize pic
        if (MaxW > 0) or (MaxH > 0) then
        rpictool.SmoothResizePicture(Pic, MaxW * 4 div 60, MaxH * 4 div 60, Enlarge);
        Result:=LinkPic;
        end;
        }
    except
      on E: Exception do
        Result:=E.Message;
    end;
  finally
    FreeAndNil(Pic);
  end;
end;

// convert numeric value to text (words)
function TrXMLReport.ProcNumToText(const TagStr: string): string;
var
  fn, br, f1, f2, f3, Frm: string;
  Repl: boolean;
  val: double;
begin
  // proc tag parts
  Result:='';
  Frm:=FindStrPart(TagStr, cFormatSep, 2, false);
  fn:=FindStrPart(TagStr, cFormatSep, 1, false);
  SepBracket(fn, br, f1, f2);
  fn:=br;
  f1:=Trim(rtool.FindStrPart(Frm, cParamSep, 1, false));
  f2:=Trim(rtool.FindStrPart(Frm, cParamSep, 2, false));
  f3:=Trim(rtool.FindStrPart(Frm, cParamSep, 3, false));

  if IsConstantTag(fn) then
    fn:=copy(fn, 2, Length(fn) - 2)
  else
  begin
    fn:=ProcTagStr(fn, Repl);
    if not Repl then
    begin
      Result:='?' + fn + '?';
      Exit;
    end;
  end;
  if fn = '' then
    Exit;
  try
    val:=StrToFloat(fn);
    if (f3 = '') or SameText(f3, 'CZ') then
      Result:=NumToText.NumToTextCZ(val, ' ', f1, f2)
    else
      Result:=NumToText.NumToTextEN(val, ' ', f1, f2);
  except
    Result:='?' + fn + '?';
  end;
end;

// Define variable
procedure TrXMLReport.ProcDefVar(const TagStr: string);
var
  s, b1, b2: string;
  Repl: boolean;
  i: integer;
begin
  SepBracket(TagStr, s, b1, b2);
  b1:=GetListPart(s, cListSep2);
  if b1 = '' then
    Exit;
  b2:=Trim(GetListPart(s, cListSep2));
  if (b2 = '') or IsConstantTag(b2) then
    s:=ClearStr(cConstSep, b2, false)
  else
    s:=ProcTagStr(Trim(b2), Repl);
  repeat
    i:=FVariables.IndexOfName(b1);
    if i >= 0 then
      FVariables.Delete(i);
  until i = -1;
  FVariables.Add(b1 + '=' + s);
end;

// proc string functions (Str_part etc)
function TrXMLReport.ProcStrPrefix(const TagStr: string): string;

  function ExtractStrConst(s: string): string;
  begin
    s:=copy(s, 2, Length(s) - 2);
    if copy(s, 1, 1) = '''' then
      Result:=copy(s, 2, Length(s) - 2)
    else
      Result:=rtool.ReplaceStr(cTabStr, #9, s, true, false);
  end;

  function ProcStrPart(fparam: string; CS: boolean): string;
  var
    s, p1, p2, p3: string;
    pos1, pos2: integer;
    Repl: boolean;
  begin
    Result:='';
    p1:=GetListPart(fparam, cListSep2);
    p2:=GetListPart(fparam, cListSep2);
    p3:=GetListPart(fparam, cListSep2);
    if fparam <> '' then
      raise Exception.Create(fparam);
    s:=ProcTagStr(p1, Repl);
    if copy(p2, 1, 1) = cConstSep then
    begin
      p2:=ExtractStrConst(p2);
      pos1:=AnsiPosEx(p2, s, 1, CS);
    end
    else
      try
        pos1:=StrToInt(p2);
      except
        pos1:=StrToInt(ProcTagStr(p2, Repl));
      end;
    if copy(p3, 1, 1) = cConstSep then
    begin
      p3:=ExtractStrConst(p3);
      pos2:=AnsiPosEx(p3, s, pos1, CS)
    end
    else
      try
        pos2:=StrToInt(p3) + pos1;
      except
        pos2:=StrToInt(ProcTagStr(p3, Repl)) + pos1;
      end;
    if (pos1 > 0) and (pos2 > 0) then
      Result:=copy(s, pos1, pos2 - pos1);
  end;

  function ProcStrRepl(fparam: string; CS: boolean): string;
  var
    Text, old, new: string;
    Repl: boolean;
  begin
    Result:='';
    Text:=GetListPart(fparam, cListSep2);
    Text:=ProcTagStr(Text, Repl);
    while fparam <> '' do
    begin
      old:=GetListPart(fparam, cListSep2);
      new:=GetListPart(fparam, cListSep2);
      if copy(old, 1, 1) = cConstSep then
        old:=ExtractStrConst(old)
      else
        old:=ProcTagStr(old, Repl);
      if copy(new, 1, 1) = cConstSep then
        new:=ExtractStrConst(new)
      else
        new:=ProcTagStr(new, Repl);
      Text:=rtool.ReplaceStr(old, new, Text, true, CS);
    end;
    Result:=Text;
  end;

  function ProcStrCase(fparam: string; Idx: integer): string;
  var
    s, p1: string;
    Repl: boolean;
    i: integer;
  begin
    Result:='';
    p1:=GetListPart(fparam, cListSep2);
    if fparam <> '' then
      raise Exception.Create(fparam);
    s:=ProcTagStr(p1, Repl);
    case Idx of
      1:
        s:=AnsiUpperCase(s);
      2:
        s:=AnsiLowerCase(s);
      3:
        begin
          s:=AnsiLowerCase(s);
          for i:=1 to Length(s) do
            if (i = 1) or (s[i - 1] = ' ') then
              s[i]:=AnsiUpperCase(s[i])[1];
        end;
      4:
        s:=Trim(s);
    end;
    Result:=s;
  end;

  function ProcStrName(FName, fparam: string): string;
  begin
    if SameText(FName, cStrPart) then
      Result:=ProcStrPart(fparam, false)
    else
      if SameText(FName, cStrPartCS) then
        Result:=ProcStrPart(fparam, true)
      else
        if SameText(FName, cStrRepl) then
          Result:=ProcStrRepl(fparam, false)
        else
          if SameText(FName, cStrReplCS) then
            Result:=ProcStrRepl(fparam, true)
          else
            if SameText(FName, cStrUpper) then
              Result:=ProcStrCase(fparam, 1)
            else
              if SameText(FName, cStrLower) then
                Result:=ProcStrCase(fparam, 2)
              else
                if SameText(FName, cStrProper) then
                  Result:=ProcStrCase(fparam, 3)
                else
                  if SameText(FName, cStrTrim) then
                    Result:=ProcStrCase(fparam, 4)
                  else
                    raise Exception.Create(FName);
  end;

var
  param, name, send: string;
begin
  try
    SepBracket(TagStr, param, name, send);
    if send <> '' then
      raise Exception.Create(send);
    Result:=ProcStrName(name, param);
  except
    on E: Exception do
      RaiseErr(4, TagStr + '; ' + Format(sTemplateIncorrectTag, [E.Message]));
  end;
end;

// proc CASE tag
function TrXMLReport.ProcCase(const TagStr: string): string;
var
  i: integer;
  Tag, part, sv, v1, v2: string;
  Repl: boolean;
begin
  Result:='';
  // extract case parameters from bracket
  SepBracket(TagStr, Tag, v1, v2);
  i:=1;
  repeat
    part:=rtool.FindStrPart(Tag, cParamSep, i, true);
    // 1st param as source value
    if i = 1 then
      sv:=part
    else
    begin
      // next params:  compared value, result value
      v1:=rtool.FindStrPart(part, cListSep2, 1, true);
      v2:=rtool.FindStrPart(part, cListSep2, 2, true);
      if SameText(v1, '#ELSE#') or ((v1 <> '') and GetConditionResult(sv + v1)) then
      begin
        Result:=ProcTagStr(v2, Repl);
        Break;
      end;
    end;
    Inc(i);
  until part = '';
end;

// proc Info tag
procedure TrXMLReport.ProcExtendedInfo(var TagStr: string);
var
  s, n, v: string;
  i: integer;
begin
  // read all Info tags
  i:=1;
  repeat
    s:=FindTagStr(cExtendedInfo, TagStr, cTagCh, i, true);
    if s = '' then
      Break;
    s:=XMLToStr(s);
    FExtendedInfo.Add(s);
  until false;
  // call user defined event
  if Assigned(FOnReadInfo) then
    FOnReadInfo(Self);
  // proc all Info values
  for i:=0 to FExtendedInfo.Count - 1 do
  begin
    s:=FExtendedInfo[i];
    n:=Trim(rtool.FindStrPart(s, '=', 1, true));
    v:=rtool.FindStrPart(s, '=', 2, true);

    if SameText(n, 'SetPrinterName') then
      Self.PrinterName:=v;

    if SameText(n, 'DefaultCalcFormat') then
      Self.FDefaultCalcFormat:=v;
    if SameText(n, 'DefaultNumericFormat') then
      Self.DefaultNumericFormat:=v;
    if SameText(n, 'DefaultBooleanFormat') then
      Self.FDefaultBooleanFormat:=v;
    if SameText(n, 'DefaultDateFormat') then
      Self.FDefaultDateFormat:=v;
    if SameText(n, 'DefaultTimeFormat') then
      Self.FDefaultTimeFormat:=v;

    if SameText(n, 'JoinTouchingTables') then
      Self.FJoinTouchingTables:=v = '1';
    if SameText(n, 'ConvertScientificFormat') then
      Self.FConvertScientificFormat:=v = '1';
    if SameText(n, 'ConvertInlineFormatting') then
      Self.FConvertInlineFormatting:=v = '1';
    if SameText(n, 'InlineFormatting') then
      Self.FInlineFormatting.Add(ReplaceStr('#9', #9, v, true, true));
    if SameText(n, 'FieldNotFoundValue') then
      Self.FFieldNotFoundValue:=v;
  end;
end;

procedure TrXMLReport.ProcInsertFile(var TagStr: string);

  procedure InsFile(aTag: string; Convert: boolean);
  var
    s: string;
    i: integer;
    L: TStringList;
  begin
    i:=1;
    repeat
      // read tag
      s:=FindTagStr(aTag, TagStr, cTagCh, i, true);
      if s = '' then
        Break;
      // add or modify directory
      if Pos('\', s) = 0 then
        s:=ExtractFilePath(FReportFile) + ReplaceStr('/', '\', s, true, true);
      // load file content
      if FileExists(s) then
      begin
        L:=TStringList.Create;
        try
          L.Loadfromfile(s);
          s:=L.Text;
        finally
          FreeAndNil(L);
        end;
      end
      else
        s:=Format(sFileNotFound, [s]);
      // pro content
      if Convert then
        s:=ProcBlock(StrToXML(s));
      Insert(s, TagStr, i);
    until false;
  end;

begin
  InsFile(cInsertFile, true);
  InsFile(cInsertXMLFile, false);
end;

procedure TrXMLReport.ProcInsertTable(var TblStr: string);
const
  cDefault = 'default';
  cRight = 'right';
  cCenter = 'center';
  cTop = 'top';
  cBottom = 'bottom';
  cParFrmAlignment: array [TAlignment] of string = ('', cRight, cCenter);

type
  TDefineFormat = record
    TitleHeight, RowHeight, MinRows, MaxRows, AddSumRows: integer;
    NoIfEmpty, NoFirst, TitleRotate, DataWidth, SizeWidth: boolean;
    ID, TitleHAlign, TitleVAlign, DataHAlign, DataVAlign: string;
    WhileCond, SkipCond: string;
  end;

  TColWRec = record
    TitleW: integer;
    ColW: integer;
    AbsW: boolean;
  end;

var
  DefFormat: TDefineFormat;
  Rec: TTableCellFormatRec;

  procedure InitFormat(Rot: boolean);
  begin
    Rec.Font.Name:=cDefault;
    Rec.Font.Size:=0;
    Rec.Font.Color:=clNone;
    Rec.Font.Style:=[];
    Rec.BkColor:=clNone;
    Rec.RotateText:=Rot;
  end;

  function TColorToHex(Color: TColor): string;
  begin
    Result:=IntToHex(GetRValue(Color), 2) + IntToHex(GetGValue(Color), 2) + IntToHex(GetBValue(Color), 2);
  end;

  function AddRow(Str: string; RowNum: integer): string;
  var
    h: integer;
  begin
    if RowNum <= 0 then
      h:=DefFormat.TitleHeight
    else
      h:=DefFormat.RowHeight;
    if Assigned(FOnGetRowHeight) then
      FOnGetRowHeight(Self, DefFormat.ID, RowNum, h);
    Result:=cXMLTableRowBegin[XMLFormat];
    // row height to row Pr (trPr)
    if h > 0 then
      Result:=Result + Format(cRowFormatParam, [Round(h * cTblGridW1cm / 10)]);
    Result:=Result + Str + cXMLTableRowEnd[XMLFormat];
  end;

  function AddCell(Str: string; RowNum, ColNum: integer; F: TField): string;
  var
    c, p, r, t, Frm: string;
  begin
    InitFormat(false);
    if RowNum <= 0 then
      Rec.RotateText:=DefFormat.TitleRotate;
    if Assigned(FOnGetCellParams) then
      FOnGetCellParams(Self, DefFormat.ID, F, RowNum, ColNum, Str, Rec);
    if Str = '' then
      Str:=' ';
    Str:=StrToXML(Str);
    // default VAlign = center
    if Rec.VAlign = '' then
      Rec.VAlign:=cCenter;

    // vertical align and bkcolor to cell (tcPr)
    if Rec.RotateText then
      Frm:=Rec.HAlign
    else
      Frm:=Rec.VAlign;
    if SameText(Frm, cCenter) then
      Frm:=cCenter
    else
      if SameText(Frm, cBottom) then
        Frm:=cBottom
      else
        if SameText(Frm, cRight) then
          Frm:=cBottom
        else
          Frm:=cTop;
    if Frm <> '' then
      Frm:=Format(cCellAlignParam, [Frm]);
    if Rec.RotateText then
      Frm:=Frm + cCellRotateFormat;
    if Rec.BkColor <> clNone then
      Frm:=Frm + Format(cCellColorParam, [TColorToHex(Rec.BkColor)]);
    c:=Format(cCellFormatParam, [Frm]);

    // Horizontal aling to par (pPr)
    if Rec.RotateText then
      Frm:=Rec.VAlign
    else
      Frm:=Rec.HAlign;
    if SameText(Frm, cCenter) then
      Frm:=cCenter
    else
      if SameText(Frm, cRight) then
        Frm:=cRight
      else
        if SameText(Frm, cTop) then
          Frm:=cRight
        else
          Frm:='';
    if Frm <> '' then
      Frm:=Format(cParAlignParam, [Frm]);
    p:=cDOCXParStart;
    if Frm <> '' then
      p:=p + Format(cParFormatParam, [Frm]);

    // font color to run
    Frm:='';
    if Assigned(Rec.Font) then
    begin
      if Rec.Font.Name <> cDefault then
        Frm:=Frm + Format(cRunFontNameParam, [Rec.Font.Name, Rec.Font.Name, Rec.Font.Name]);
      if Rec.Font.Size <> 0 then
        Frm:=Frm + Format(cRunFontSizeParam, [Rec.Font.Size * 2, Rec.Font.Size * 2, Rec.Font.Size * 2]);
      if Rec.Font.Color <> clNone then
        Frm:=Frm + Format(cRunColorParam, [TColorToHex(Rec.Font.Color)]);
      if fsBold in Rec.Font.Style then
        Frm:=Frm + cDOCXBold;
      if fsItalic in Rec.Font.Style then
        Frm:=Frm + cDOCXItalic;
      if fsUnderline in Rec.Font.Style then
        Frm:=Frm + cDOCXUnderline;
      if fsStrikeOut in Rec.Font.Style then
        Frm:=Frm + cDOCXStrikeOut;
    end;
    r:=cDOCXRunStart;
    if Frm <> '' then
      r:=r + Format(cRunFormatParam, [Frm]);

    // text
    t:=cDOCXTextStart;

    // create whole cell tag
    Result:=cXMLTableCellBegin[XMLFormat] + c + p + r + t + Str + cXMLParEndAll[XMLFormat] + cXMLTableCellEnd[XMLFormat];
  end;

  function IsExportFld(F: TField): boolean;
  begin
    Result:=F.Visible and (IsSimpleField(F) or IsMemoField(F));
  end;

var
  TagStr, Params, part, res: string;
  RowNum, DSIdx, TagPos, FldCount, ColCount, SumColW, SumColAbsW, TableW: integer;
  ColWArr: array of TColWRec;
  Fld: TField;
  s, s1, s2, sr: string;
  i, ci, p1, p2, w: integer;
  // AbsW: boolean;
begin
  Rec.Font:=TFont.Create;
  try
    InitFormat(false);
    TagPos:=0;
    TagStr:=FindTagStr(cTag, TblStr, cTagCh, TagPos, false);
    TagStr:=XMLToStr(TagStr);
    // get dataset
    DSIdx:=GetDSIndex(TagStr);
    if DSIdx < 0 then
      RaiseErr(2, TagStr);

    // init params
    DefFormat.TitleHeight:=0;
    DefFormat.RowHeight:=0;
    DefFormat.MinRows:=0;
    DefFormat.MaxRows:=0;
    DefFormat.AddSumRows:=0;

    DefFormat.NoIfEmpty:=false;
    DefFormat.NoFirst:=false;
    DefFormat.TitleRotate:=false;
    DefFormat.DataWidth:=false;
    DefFormat.SizeWidth:=false;

    DefFormat.ID:='';
    DefFormat.TitleHAlign:='';
    DefFormat.TitleVAlign:='';
    DefFormat.DataHAlign:='';
    DefFormat.DataVAlign:='';

    DefFormat.WhileCond:='';
    DefFormat.SkipCond:='';

    // remove first part
    Params:=TagStr;
    Trim(rtool.RemoveStrPart(Params, cParamSep));
    // proc params
    while Params <> '' do
    begin
      part:=Trim(rtool.RemoveStrPart(Params, cParamSep));
      if part = '' then
        Continue;
      SepBracket(part, s, s1, s2);
      s:=Trim(s);
      s1:=Trim(s1);
      s2:=Trim(s2);
      if s2 <> '' then
        RaiseErr(5, part);
      if s1 = '' then
        RaiseErr(5, part);
      if AnsiSameText('id', s1) then
        DefFormat.ID:=s
      else
        if AnsiSameText(cScanWhileCond, s1) then
          DefFormat.WhileCond:=s
        else
          if AnsiSameText(cScanSkipCond, s1) then
            DefFormat.SkipCond:=s
          else
            if AnsiSameText(cScanNoIfEmpty, s1) then
              DefFormat.NoIfEmpty:=true
            else
              if AnsiSameText(cScanNoFirst, s1) then
                DefFormat.NoFirst:=true
              else
                if AnsiSameText(cScanMinRows, s1) then
                  DefFormat.MinRows:=StrToIntDef(s, 0)
                else
                  if AnsiSameText(cScanMaxRows, s1) then
                    DefFormat.MaxRows:=StrToIntDef(s, 0)
                  else
                    if AnsiSameText('addsumrows', s1) then
                      DefFormat.AddSumRows:=StrToIntDef(s, 0)
                    else
                      if AnsiSameText('titlehalign', s1) then
                        DefFormat.TitleHAlign:=s
                      else
                        if AnsiSameText('titlevalign', s1) then
                          DefFormat.TitleVAlign:=s
                        else
                          if AnsiSameText('datahalign', s1) then
                            DefFormat.DataHAlign:=s
                          else
                            if AnsiSameText('datavalign', s1) then
                              DefFormat.DataVAlign:=s
                            else
                              if AnsiSameText('titlerotate', s1) then
                                DefFormat.TitleRotate:=true
                              else
                                if AnsiSameText('titleheight', s1) then
                                  DefFormat.TitleHeight:=StrToIntDef(s, 0)
                                else
                                  if AnsiSameText('rowheight', s1) then
                                    DefFormat.RowHeight:=StrToIntDef(s, 0)
                                  else
                                    if AnsiSameText('datawidth', s1) then
                                      DefFormat.DataWidth:=true
                                    else
                                      if AnsiSameText('sizewidth', s1) then
                                        DefFormat.SizeWidth:=true
                                      else
                                        RaiseErr(5, part);
    end;

    // find and remove current table row
    p1:=PosEx(cXMLTableRowBegin[XMLFormat], TblStr);
    p2:=PosEx(cXMLTableRowEnd[XMLFormat], TblStr);
    if (p1 = 0) or (p2 < p1) then
      RaiseErr(7, TagStr);
    Inc(p2, Length(cXMLTableRowEnd[XMLFormat]));
    Delete(TblStr, p1, p2 - p1);
    // check another row
    if (PosEx(cXMLTableRowBegin[XMLFormat], TblStr) > 0) or (PosEx(cXMLTableRowEnd[XMLFormat], TblStr) > 0) then
      RaiseErr(7, TagStr);

    // check DS active
    res:=CheckDSIndex(DSIdx);
    if res <> '' then
    begin
      Insert(AddRow(AddCell(res, -1, 1, nil), -1), TblStr, p1);
      Exit;
    end;
    // check empty
    if DefFormat.NoIfEmpty and not DSNotEOF(DSIdx) then
    begin
      TblStr:='';
      Exit;
    end;
    // go to first
    if not DefFormat.NoFirst then
      FindFirst(DSIdx);
    // calc num of columns
    FldCount:=DS[DSIdx].FieldCount;
    ColCount:=0;
    for i:=0 to FldCount - 1 do
      if IsExportFld(DS[DSIdx].Fields[i]) then
        Inc(ColCount);
    // init
    res:='';
    RowNum:=0;
    SetLength(ColWArr, ColCount);
    ci:=0;
    // generate title row
    sr:='';
    for i:=0 to FldCount - 1 do
    begin
      Fld:=DS[DSIdx].Fields[i];
      if IsExportFld(Fld) then
      begin
        s:=Fld.DisplayLabel;
        if DefFormat.TitleHAlign = '' then
        begin
          if DefFormat.TitleRotate then
            Rec.HAlign:=cParFrmAlignment[taCenter]
          else
            Rec.HAlign:=cParFrmAlignment[Fld.Alignment];
        end
        else
          Rec.HAlign:=DefFormat.TitleHAlign;
        Rec.VAlign:=DefFormat.TitleVAlign;
        sr:=sr + AddCell(s, RowNum, ci + 1, Fld);
        if DefFormat.SizeWidth then
        begin
          ColWArr[ci].TitleW:=Fld.DisplayWidth;
          ColWArr[ci].ColW:=Fld.DisplayWidth;
        end
        else
        begin
          ColWArr[ci].TitleW:=Length(s);
          ColWArr[ci].ColW:=1;
        end;
        ColWArr[ci].AbsW:=false;
        Inc(ci);
      end;
    end;
    // calculate auto title height
    if DefFormat.TitleRotate and (DefFormat.TitleHeight = 0) then
    begin
      for i:=0 to ColCount - 1 do
      begin
        if DefFormat.TitleHeight < ColWArr[i].TitleW then
          DefFormat.TitleHeight:=ColWArr[i].TitleW;
      end;
      DefFormat.TitleHeight:=Round(DefFormat.TitleHeight * 2);
    end;
    res:=AddRow(sr, 0);
    // generate row for each record
    repeat
      if not DSNotEOF(DSIdx) then
        Break;
      if (DefFormat.WhileCond <> '') and not GetConditionResult(DefFormat.WhileCond) then
        Break;
      if (DefFormat.SkipCond <> '') and GetConditionResult(DefFormat.SkipCond) then
        Continue;
      if (DefFormat.MaxRows > 0) and (RowNum >= DefFormat.MaxRows) then
        Break;
      // ProcCalcField(ScInf.DSindex);
      Inc(RowNum);
      ci:=0;
      sr:='';
      for i:=0 to FldCount - 1 do
      begin
        Fld:=DS[DSIdx].Fields[i];
        if IsExportFld(Fld) then
        begin
          if IsMemoField(Fld) then
            s:=ParseRTFFormat(Fld.AsString)
          else
            s:=Fld.DisplayText;
          if DefFormat.DataHAlign = '' then
            Rec.HAlign:=cParFrmAlignment[Fld.Alignment]
          else
            Rec.HAlign:=DefFormat.DataHAlign;
          Rec.VAlign:=DefFormat.DataVAlign;
          sr:=sr + AddCell(s, RowNum, ci + 1, Fld);
          if DefFormat.DataWidth and (ColWArr[ci].ColW < Length(s)) then
            ColWArr[ci].ColW:=Length(s);
          Inc(ci);
        end;
      end;
      res:=res + AddRow(sr, RowNum);
    until not FindNext(DSIdx);

    // add empty rows
    if (DefFormat.MinRows > 0) or (DefFormat.AddSumRows > 0) then
    begin
      if DefFormat.MinRows < RowNum then
        DefFormat.MinRows:=RowNum;

      while RowNum < DefFormat.MinRows + DefFormat.AddSumRows do
      begin
        Inc(RowNum);
        ci:=0;
        sr:='';
        for i:=0 to FldCount - 1 do
        begin
          Fld:=DS[DSIdx].Fields[i];
          if IsExportFld(Fld) then
          begin
            s:='';
            if DefFormat.DataHAlign = '' then
              Rec.HAlign:=cParFrmAlignment[Fld.Alignment]
            else
              Rec.HAlign:=DefFormat.DataHAlign;
            Rec.VAlign:=DefFormat.DataVAlign;
            sr:=sr + AddCell(s, RowNum, ci + 1, Fld);
            if DefFormat.DataWidth and (ColWArr[ci].ColW < Length(s)) then
              ColWArr[ci].ColW:=Length(s);
            Inc(ci);
          end;
        end;
        res:=res + AddRow(sr, RowNum);
      end;
    end;

    // insert table
    Insert(res, TblStr, p1);

    // set columns width
    p1:=PosEx(cTblGridWStart, TblStr);
    p2:=PosEx(cTblGridWEnd, TblStr);
    if (p1 > 0) and (p2 > p1) then
    begin
      Inc(p1, Length(cTblGridWStart));
      // get width of original one column
      s:=copy(TblStr, p1, p2 - p1);
      s:=rtool.FindStrFromTo(cConstSep, cConstSep, s, true, true);
      TableW:=StrToIntDef(s, 0);
      if TableW > 0 then
      begin
        // remove original width
        Delete(TblStr, p1, p2 - p1);
        res:='';

        ci:=0;
        for i:=0 to FldCount - 1 do
        begin
          Fld:=DS[DSIdx].Fields[i];
          if IsExportFld(Fld) then
          begin
            // adjust current col widths
            if not DefFormat.TitleRotate and (ColWArr[ci].TitleW > ColWArr[ci].ColW) then
              Inc(ColWArr[ci].ColW, (ColWArr[ci].TitleW - ColWArr[ci].ColW) div 2);
            Inc(ColWArr[ci].ColW, 3);
            // limit col widths to range 6..50
            ColWArr[ci].ColW:=Min(Max(ColWArr[ci].ColW, 6), 50);
            // call event
            if Assigned(FOnGetColumnWidth) then
              FOnGetColumnWidth(Self, DefFormat.ID, Fld, ci + 1, ColWArr[ci].ColW, ColWArr[ci].AbsW);
            Inc(ci);
          end;
        end;

        // calc width sum
        SumColW:=0;
        SumColAbsW:=0;
        for ci:=0 to ColCount - 1 do
        begin
          if ColWArr[ci].AbsW then
            Inc(SumColAbsW, ColWArr[ci].ColW)
          else
            Inc(SumColW, ColWArr[ci].ColW);
        end;
        // recalc AbsW to points
        SumColAbsW:=Round(SumColAbsW * cTblGridW1cm / 10);
        // calc new widths
        for ci:=0 to ColCount - 1 do
        begin
          if ColWArr[ci].AbsW then
            // set absolute width
            w:=Round(ColWArr[ci].ColW * cTblGridW1cm / 10)
          else
            if SumColW > 0 then
              // calc relative width
              w:=Round((TableW - SumColAbsW) * ColWArr[ci].ColW / SumColW)
            else
              w:=0;
          w:=Max(w, cTblGridW1cm div 2);
          res:=res + Format(cTblGridWidth, [w]);
        end;
        Insert(res, TblStr, p1);
      end;
    end;

  finally
    SetLength(ColWArr, 0);
    FreeAndNil(Rec.Font);
  end;
end;

// ======================================================================
// XML Proccessing function
// proc all (open file, generating report, saving result)
procedure TrXMLReport.ProcAllFiles;
var
  err, oldPrinterName: string;
  i, showtype: integer;
begin
  Aborted:=false;
  RecNbr:=1;
  RecNbrDS0:=-1;
  FExtendedInfo.Clear;
  PrinterName:='';
  if Assigned(FOnReportProcess) then
    FOnReportProcess(Self, 0, Aborted);
  Application.ProcessMessages;

  // copy report file to output file
  if not FileExists(FReportFile) then
    raise Exception.Create(Format(sFileOpenErr, [FReportFile]) + #13 + sFileOpenErr2);

  repeat
    if CopyFile(PChar(FReportFile), PChar(FOutputFile), false) then
      Break;
    err:=Format(sFileSaveErr, [FOutputFile]) + #13 + sFileSaveErr2;
    if FShowSaveErrorDialog then
    begin
      if DlgQ(err + #13 + #13 + sTryAgainQ) <> mrYes then
        Abort;
    end
    else
      RaiseErr(-1, err);
  until false;

  // proc XML  in output file
  if not ProcXMLFiles(FOutputFile, false) then
    Exit;
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if Assigned(FOnReportProcess) then
    FOnReportProcess(Self, -1, Aborted);
  // show or print result
  if FShowMaximized then
    showtype:=SW_SHOWMAXIMIZED
  else
    showtype:=SW_SHOWNORMAL;
  if ActionAfter = aaShow then
    RunFile(FOutputFile, '', '', showtype);
  if ActionAfter = aaPrint then
  begin
    if PrinterName <> '' then
    begin
      oldPrinterName:=GetDefaultPrinterName;
      if not SetDefaultPrinter(PrinterName, true) then
        RaiseErr(6, PrinterName);
    end;
    try
      for i:=1 to PrintCopies do
      begin
        if Assigned(FOnReportProcess) then
          FOnReportProcess(Self, -1, Aborted);
        Application.ProcessMessages;
        if Aborted then
          Raise Exception.Create(sAborted);
        PrintFileSimple(FOutputFile);
        Sleep(1000);
      end;
    finally
      if PrinterName <> '' then
        SetDefaultPrinter(oldPrinterName, true);
    end;
  end;
end;

// load list of XML files to update
function TrXMLReport.ProcXMLFiles(ZipFileName: string; ReadOnly: boolean): boolean;
const
  FileMaskList: array [TTemplateFormat] of string = ('', 'word/header*.xml;word/footer*.xml;word/document.xml', 'content.xml;styles.xml');

  function UseXMLFile: boolean;
  var
    i: integer;
    fn: string;
  begin
    i:=1;
    Result:=false;
    repeat
      fn:=rtool.FindStrPart(FileMaskList[XMLFormat], cParamSep, i, true);
      if fn = '' then
        Break;
      Result:=Masks.MatchesMask(XMLFileName, fn);
      if Result then
        Break;
      Inc(i);
    until false;
  end;

var
  FileIdx: integer;
  ListStr: string;
begin
  Result:=false;
  rzip.Load;
  if rzip.Loaded then
  begin
    try
      // open zipped file
      if rzip.OpenArchive(ZipFileName) <> 0 then
        RaiseErr(-1, rzip.GetLastErrorMsg);
      // Load list of files
      if rzip.GetFileList(ListStr) <> 0 then
        RaiseErr(-1, rzip.GetLastErrorMsg);
      FileList.Text:=ListStr;

      // proc files
      for FileIdx:=0 to FileList.Count - 1 do
      begin
        XMLFileName:=FileList[FileIdx];
        RelItemStr:='';

        if UseXMLFile then
        begin
          if Assigned(FOnGetTemplateStr) then
            FOnGetTemplateStr(Self, XMLFileName, RepStr)
          else
          begin
            // load XML content
            if rzip.ExtractToString(XMLFileName, RepStr) <> 0 then
              RaiseErr(-1, rzip.GetLastErrorMsg);
          end;
          // proc XML file
          ProcXMLFileContent;
          Application.ProcessMessages;
          if Aborted then
            Raise Exception.Create(sAborted);

          if not ReadOnly then
          begin
            if Assigned(FModifyBeforeSave) then
            begin
              FModifyBeforeSave(Self, XMLFileName, RepStr);
              if RepStr = '' then
                Continue;
            end;
            // update XML content
            if rzip.AddFromString(FileList[FileIdx], RepStr) <> 0 then
              RaiseErr(-1, rzip.GetLastErrorMsg);
            // update rel file if image used
            if RelItemStr <> '' then
              UpdateRelFile;
          end;
        end;
      end;
      rzip.CloseArchive(true);
      Result:=true;
    finally
      rzip.Unload;
    end;
  end;
end;

// add image file to zipped file, generate and return REL ID or image filename
function TrXMLReport.AddImageFile(Pic: TPicture; GrType: TrGraphicType): string;
var
  ID, fn: string;
  mStream: TMemoryStream;
  Buffer: TBytes;
begin
  Inc(PicCount);
  mStream:=TMemoryStream.Create;
  try
    // update picture type and load to stream
    GrType:=rpictool.SavePictureToStream(Pic, mStream, GrType);
    // update REL string
    fn:='media/image' + IntToStr0(PicCount, 6) + GetFileTypeExtension(GrType);
    case XMLFormat of
      tfDOCX:
        ID:='xRepId' + IntToStr0(PicCount, 6);
      tfODT:
        ID:=fn;
    else
      RaiseErr(-2);
    end;
    Result:=ID;

    RelItemStr:=RelItemStr + Format(cXMLRelImage[XMLFormat], [ID, fn]);
    // store image to zipped file
    if XMLFormat = tfDOCX then
      fn:='word/' + fn;
    SetLength(Buffer, mStream.Size);
    mStream.Position:=0;
    mStream.Read(Buffer[0], mStream.Size);
    rzip.AddFromBuffer(fn, @Buffer[0], mStream.Size);
  finally
    FreeAndNil(mStream);
  end;
end;

procedure TrXMLReport.UpdateRelFile;
var
  fn, Str, newct: string;
  pos1: integer;
begin
  // load REL file according XML file name, use empty header if not exist for DOCX
  case XMLFormat of
    tfDOCX:
      fn:='word/_rels/' + ClearStr('word/', XMLFileName, false) + '.rels';
    tfODT:
      fn:='META-INF/manifest.xml';
  else
    RaiseErr(-2);
  end;

  if rzip.ExtractToString(fn, Str) <> 0 then
  begin
    case XMLFormat of
      tfDOCX:
        Str:=cDOCXRelEmpty;
      tfODT:
        RaiseErr(-1, rzip.GetLastErrorMsg);
    end;
  end;

  // add new relations
  pos1:=AnsiPosEx(cXMLRelClose[XMLFormat], Str);
  if pos1 > 0 then
    Str:=copy(Str, 1, pos1 - 1) + RelItemStr + copy(Str, pos1, Length(Str));
  // update REL XML file
  if rzip.AddFromString(fn, Str) <> 0 then
    RaiseErr(-1, rzip.GetLastErrorMsg);

  if XMLFormat = tfDOCX then
  begin
    // check and add content type definition for JPG + PNG
    fn:=cDOCXDefContentFile;
    if rzip.ExtractToString(fn, Str) <> 0 then
      RaiseErr(-1, rzip.GetLastErrorMsg);
    // check content
    newct:='';
    if AnsiPosEx(cDOCXDefContentJPG, Str) = 0 then
      newct:=newct + cDOCXDefContentJPG;
    if AnsiPosEx(cDOCXDefContentPNG, Str) = 0 then
      newct:=newct + cDOCXDefContentPNG;
    pos1:=AnsiPosEx(cDOCXDefContentEnd, Str);
    if pos1 > 0 then
      Str:=copy(Str, 1, pos1 - 1) + newct + copy(Str, pos1, Length(Str));
    // update CONTENT XML file
    if rzip.AddFromString(fn, Str) <> 0 then
      RaiseErr(-1, rzip.GetLastErrorMsg);
  end;
end;

procedure TrXMLReport.ProcXMLFileContent;
begin
  if FLoadTableNamePrefix = '' then
  begin
    UpdateXMLTags;
    if Length(NamedTables_Content) > 0 then
      ReadWriteTablesContent(true);
  end
  else
    ReadWriteTablesContent(false);
end;

procedure TrXMLReport.UpdateXMLTags;
begin
  if InlineFormatting.Text = cInlineFormatConvert then
    Exit;
  try
    // remove end of lines
    ClearEOLPart(RepStr);
    // extract internal XML tags
    ExtractIntXMLTags(RepStr);
    // remove bookmarks tag
    ClearBookmarkPart(RepStr);
    // proc extended info tags
    ProcExtendedInfo(RepStr);
    // proc InsertFile tags
    ProcInsertFile(RepStr);
    // proc IF (not IIF)
    ProcIf(false, true, RepStr);
    // proc all report
    RepStr:=ProcBlock(RepStr);
    // remove empty paragraphs
    ClearEmptyParPart(RepStr);
    // join tables without text between
    if FJoinTouchingTables then
      JoinTables(RepStr);
    // replace spec. chars by format
    CheckInlineFormat(RepStr);
  finally
    FreeCalcField;
    FreeBookMarks;
  end;
end;

procedure TrXMLReport.ReadWriteTablesContent(Write: boolean);
const
  TblCaptTagStart = '<w:tblCaption w:val="';
var
  i, p, t1, t2, TagPos, ArrIdx, Tbl1, Tbl2: integer;
  TblStr, RowStr, CellStr, s: string;
  InTag, AddEOL: boolean;

  function GetBeginEndPart(Str, sBegin, send: string; FromPos: integer; InclEndTag: boolean): string;
  begin
    // find nearest start
    p:=0;
    t1:=0;
    repeat
      p:=PosEx(sBegin, Str, p + 1);
      if (FromPos = 1) then
        t1:=p;
      if (FromPos = 1) or (p = 0) or (p > FromPos) then
        Break;
      if p < FromPos then
        t1:=p;
    until false;
    if t1 = 0 then
      RaiseErr(-3);
    // find nearest end
    t2:=PosEx(send, Str, FromPos);
    if t2 = 0 then
      RaiseErr(-3);
    if InclEndTag then
      Inc(t2, Length(send));
    // get table content
    Result:=copy(Str, t1, t2 - t1);
  end;

  function GetCellContent(Str: string): string;
  begin
    t1:=PosEx(cXMLTableCellBegin[XMLFormat], Str);
    if (t1 = 0) then
      RaiseErr(-3);
    Inc(t1, Length(cXMLTableCellBegin[XMLFormat]));
    t2:=PosEx(cCellFormatStart, Str);
    if t2 = t1 then
    begin
      t1:=PosEx(cCellFormatEnd, Str);
      if (t1 = 0) then
        RaiseErr(-3);
      Inc(t1, Length(cCellFormatEnd));
    end;
    t2:=PosEx(cXMLTableCellEnd[XMLFormat], Str, t1);
    Result:=copy(Str, t1, t2 - t1);
  end;

begin
  if Write then
  begin
    for i:=0 to High(NamedTables_Content) do
    begin
      TagPos:=PosEx(TblCaptTagStart + NamedTables_Content[i].TableName, RepStr);
      if TagPos = 0 then
        Continue;
      // get table content
      TblStr:=GetBeginEndPart(RepStr, cXMLTableTagBegin[XMLFormat], cXMLTableTagEnd[XMLFormat], TagPos, true);
      Tbl1:=t1;
      Tbl2:=t2;
      // find current table row
      RowStr:=GetBeginEndPart(TblStr, cXMLTableRowBegin[XMLFormat], cXMLTableRowEnd[XMLFormat], 1, true);
      // find current table cell
      CellStr:=GetBeginEndPart(RowStr, cXMLTableCellBegin[XMLFormat], cXMLTableCellEnd[XMLFormat], 1, true);
      // find cell content
      CellStr:=GetCellContent(CellStr);

      s:=NamedTables_Content[i].CellXMLContent;
      if s = '' then
      begin
        s:=StrToXML(NamedTables_Content[i].PlainText);
        s:=cXMLParBeginAll[XMLFormat] + s + cXMLParEndAll[XMLFormat];
      end;
      if s <> '' then
      begin
        TblStr:=ReplaceStr(CellStr, s, TblStr);
        Delete(RepStr, Tbl1, Tbl2 - Tbl1);
        Insert(TblStr, RepStr, Tbl1);
      end;
    end;
  end
  else
  begin
    SetLength(NamedTables_Content, 0);
    ArrIdx:=0;
    TagPos:=1;
    repeat
      s:=TblCaptTagStart;
      if FLoadTableNamePrefix <> '*' then
        s:=s + FLoadTableNamePrefix;
      TagPos:=PosEx(s, RepStr, TagPos);
      if TagPos = 0 then
        Break;
      // get table content
      TblStr:=GetBeginEndPart(RepStr, cXMLTableTagBegin[XMLFormat], cXMLTableTagEnd[XMLFormat], TagPos, true);
      TagPos:=t2;
      // find current table row
      RowStr:=GetBeginEndPart(TblStr, cXMLTableRowBegin[XMLFormat], cXMLTableRowEnd[XMLFormat], 1, true);
      // find current table cell
      CellStr:=GetBeginEndPart(RowStr, cXMLTableCellBegin[XMLFormat], cXMLTableCellEnd[XMLFormat], 1, true);
      // find cell content
      CellStr:=GetCellContent(CellStr);
      //CellStr:=TblStr;
      // get plain text
      s:='';
      InTag:=false;
      AddEOL:=false;
      for i:=1 to Length(CellStr) do
      begin
        if CellStr[i] = '<' then
        begin
          InTag:=true;
          if PosEx(cDOCXParEnd, CellStr, i) = i then
            AddEOL:=true;
        end;
        if not InTag then
        begin
          if AddEOL then
            s:=s + cTxtEndPar;
          AddEOL:=false;
          s:=s + CellStr[i];
        end;
        if CellStr[i] = '>' then
          InTag:=false;
      end;
      s:=Self.XMLToStr(s);

      // store to array
      SetLength(NamedTables_Content, ArrIdx + 1);
      NamedTables_Content[ArrIdx].TableName:=rtool.FindStrFromTo(TblCaptTagStart, cConstSep, TblStr, false, true);
      NamedTables_Content[ArrIdx].TableXMLContent:=TblStr;
      NamedTables_Content[ArrIdx].CellXMLContent:=CellStr;
      NamedTables_Content[ArrIdx].PlainText:=s;
      Inc(ArrIdx);
    until false;
  end;
end;

// Proc block of XML
function TrXMLReport.ProcBlock(const Str: string): string;
var
  ScanPos, IIfPos, proc: integer;
  s: string;
begin
  Result:='';
  s:=Str;
  repeat
    // test abortion
    Application.ProcessMessages;
    if Assigned(FOnReportProcess) then
      FOnReportProcess(Self, RecNbr, Aborted);
    if Aborted then
      raise Exception.Create(sAborted);
    if FShowProgressBar and (RecNbr mod 10 = 0) then
    begin
      if RecCountDS0 > 0 then
        proc:=Round(100.0 * RecNbrDS0 / RecCountDS0)
      else
        proc:=-1;
      ShowInfoProgressBar(sCreatingReport + #13 + Format(sRecord + ' %d', [RecNbr]), nil, DoAbort, proc);
    end;

    // find start of SCAN and IIF
    ScanPos:=AnsiPosEx(cUsrTagScan, s, 1, false);
    IIfPos:=AnsiPosEx(cUsrTagIIfStart, s, 1, false);
    if (IIfPos > 0) and ((ScanPos = 0) or (IIfPos < ScanPos)) then
    begin
      // IIF first
      // proc text till IIf
      Result:=Result + ProcXML(copy(s, 1, IIfPos - 1));
      // copy and proc the rest (begining with IIF)
      s:=copy(s, IIfPos, Length(s));
      ProcIf(true, false, s);
    end
    else
    begin
      // Scan first
      if ScanPos = 0 then
        ScanPos:=Length(s) + 1;
      // proc text till scan
      Result:=Result + ProcXML(copy(s, 1, ScanPos - 1));
      // copy and proc the rest (begining with scan)
      s:=copy(s, ScanPos, Length(s));
      Result:=Result + ProcScan(s);
    end;
  until s = '';
end;

// Proc XML with SCAN tag (recurse for master/detail report)
function TrXMLReport.ProcScan(var s: string): string;
var
  ScInf: TScanInfo;
  ScanTag, ScanBlock, ScanEntry, ScanFooter: string;
  PageSep: string;
  ResStr: string;
  TblDefStart, TblDefEnd: string;
  TblRows: array of string;
  RowNum: integer;

  // find ScanStr and divide Str to blocks ahead and behind of ScanStr
  function DivideStr(ScanStr, Str: string; var Block1, Block2: string): boolean;
  var
    p: integer;
  begin
    Block1:=Str;
    Block2:='';
    // find ScanStr
    p:=AnsiPosEx(ScanStr, Str, 1, false);
    Result:=p > 0;
    if Result then
    begin
      Block1:=copy(Str, 1, p - 1);
      Block2:=copy(Str, p + Length(ScanStr), Length(Str));
      // RemovePar(Block2, true);
    end;
  end;

  procedure LoadTblRows;
  var
    i, p1, p2, r1, r2: integer;
    rows: string;
  begin
    SetLength(TblRows, 0);
    // find start of first row and store as table def
    p1:=AnsiPosEx(cXMLTableRowBegin[XMLFormat], ScanBlock, 1, false);
    if p1 = 0 then
      Exit;
    TblDefStart:=copy(ScanBlock, 1, p1 - 1);
    // find enf of table and store rest as table end
    p2:=AnsiPosEx(cXMLTableTagEnd[XMLFormat], ScanBlock, p1, false);
    if p2 = 0 then
      RaiseErr(5, cXMLTableTagEnd[XMLFormat]);
    TblDefEnd:=copy(ScanBlock, p2, Length(ScanBlock));
    // extract rows definition
    rows:=copy(ScanBlock, p1, p2 - p1);
    i:=1;
    r1:=1;
    repeat
      r1:=AnsiPosEx(cXMLTableRowBegin[XMLFormat], rows, r1, false);
      if r1 = 0 then
        Break;
      r2:=AnsiPosEx(cXMLTableRowEnd[XMLFormat], rows, r1, false);
      if r2 = 0 then
        RaiseErr(5, cXMLTableRowEnd[XMLFormat]);
      SetLength(TblRows, i);
      TblRows[i - 1]:=copy(rows, r1, r2 - r1 + Length(cXMLTableRowEnd[XMLFormat]));
      Inc(i);
      r1:=r2 + Length(cXMLTableRowEnd[XMLFormat]);
    until false;
  end;

  function ProcTblRows(Header, Footer: boolean): string;
  var
    i: integer;
    res: boolean;

    function IsRowHeader: boolean;
    begin
      Result:=SameText(ScInf.TblRowCond[i], 'h') or SameText(ScInf.TblRowCond[i], 'header');
    end;
    function IsRowFooter: boolean;
    begin
      Result:=SameText(ScInf.TblRowCond[i], 'f') or SameText(ScInf.TblRowCond[i], 'footer');
    end;
    function IsRowHidden: boolean;
    begin
      Result:=SameText(ScInf.TblRowCond[i], '0') or SameText(ScInf.TblRowCond[i], 'hide');
    end;
    function IsRowShow: boolean;
    begin
      Result:=(Trim(ScInf.TblRowCond[i]) = '') or SameText(ScInf.TblRowCond[i], '1');
    end;

  begin
    Result:='';
    for i:=0 to High(TblRows) do
    begin
      if i <= High(ScInf.TblRowCond) then
      begin
        if Header then
          res:=IsRowHeader
        else
          if Footer then
            res:=IsRowFooter
          else
            if IsRowHeader or IsRowFooter or IsRowHidden then
              res:=false
            else
              res:=IsRowShow or GetConditionResult(ScInf.TblRowCond[i]);
      end
      else
        res:=not(Header or Footer);

      if res then
        Result:=Result + ProcBlock(TblRows[i]);
    end;
  end;

begin
  Result:='';
  ResStr:='';
  if s = '' then
    Exit;
  ScInf:=DecodeScanLine(s);
  // find endscan
  ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagEndScan, false, false);
  if not DivideStr(ScanTag, s, ScanBlock, s) then
    RaiseErr(3, ScanTag);
  // move to first
  if (ScInf.DSindex >= 0) and not ScInf.NoFirst then
    FindFirst(ScInf.DSindex);
  // scan if not empty
  if (not ScInf.NoIfEmpty) or DSNotEOF(ScInf.DSindex) then
  begin
    // find scanentry
    ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagScanEntry, false, false);
    if DivideStr(ScanTag, ScanBlock, ScanEntry, ScanBlock) then
      Result:=Result + ProcBlock(ScanEntry)
    else
      ScanBlock:=ScanEntry;
    // find scanfooter
    ScanTag:=ReplaceStr(cTagCh, ScInf.DS, cUsrTagScanFooter, false, false);
    DivideStr(ScanTag, ScanBlock, ScanBlock, ScanFooter);
    PageSep:='';
    RowNum:=0;

    if ScInf.TblScan then
    begin
      // load whole table and extract def and rows
      LoadTblRows;
      // add table def + header table row
      ResStr:=TblDefStart + ProcTblRows(true, false);
    end;

    // proc block for each record (recurse)
    repeat
      if not DSNotEOF(ScInf.DSindex) then
        Break;
      if (ScInf.WhileCond <> '') and not GetConditionResult(ScInf.WhileCond) then
        Break;
      if (ScInf.SkipCond <> '') and GetConditionResult(ScInf.SkipCond) then
        Continue;
      if (ScInf.MaxRows > 0) and (RowNum >= ScInf.MaxRows) then
        Break;
      Inc(RowNum);

      ProcCalcField(ScInf.DSindex);
      if ScInf.TblScan then
      begin
        // add table rows according conditions
        ResStr:=ResStr + ProcTblRows(false, false);
      end
      else
      begin
        // add all block
        ResStr:=ResStr + PageSep + ProcBlock(ScanBlock);
        PageSep:='';
        if ScInf.NewPage then
          PageSep:=cXMLEndPage[XMLFormat]
        else
          if ScInf.NewCol then
            PageSep:=cXMLEndCol[XMLFormat]
          else
            if ScInf.NewPar then
              PageSep:=cXMLEndPar[XMLFormat]
            else
              if ScInf.NewLine then
                PageSep:=cXMLEndPar[XMLFormat];
        PageSep:=ScInf.Sep + PageSep;
      end;
    until (ScInf.DSindex >= 0) and not FindNext(ScInf.DSindex);

    if (ScInf.MinRows > 0) then
    begin
      while RowNum < ScInf.MinRows do
      begin
        Inc(RowNum);
        if ScInf.TblScan then
          ResStr:=ResStr + ProcTblRows(false, false)
        else
          ResStr:=ResStr + PageSep + ProcBlock(ScanBlock);
      end;
    end;

    if ScInf.TblScan then
    begin
      // add footer table row and end of table definition
      ResStr:=ResStr + ProcTblRows(false, true) + TblDefEnd;
    end;
    Result:=Result + ResStr + ProcBlock(ScanFooter);
  end;
end;

// proc Scan command
function TrXMLReport.DecodeScanLine(var s: string): TScanInfo;
var
  scan: string;
  part, b1, b2, br, trimbr: string;
  i, r: integer;
begin
  Result.DS:='_';
  Result.DSindex:=-1;
  Result.WhileCond:='';
  Result.SkipCond:='';
  Result.NewLine:=false;
  Result.NewPar:=false;
  Result.NewCol:=false;
  Result.NewPage:=false;
  Result.Sep:='';
  Result.NoIfEmpty:=false;
  Result.NoFirst:=false;
  Result.TblScan:=false;
  Result.MinRows:=0;
  Result.MaxRows:=0;
  i:=1;
  // load and delete scan line
  scan:=FindTagStr(cTag, s, cTagCh, i, true);
  scan:=XMLToStr(scan);

  // proc scan line
  i:=1;
  repeat
    part:=Trim(FindStrPart(scan, cParamSep, i, false));
    if part = '' then
      Break;
    // proc one part of scan line
    if i = 1 then
    begin
      Result.DSindex:=GetDSIndex(part);
      if Result.DSindex >= 0 then
        Result.DS:=IndexToDatasetChar(Result.DSindex)
      else
        Result.DS:='';
    end
    else
    begin
      SepBracket(part, br, b1, b2);
      trimbr:=Trim(br);
      b1:=Trim(b1);
      b2:=Trim(b2);
      if b2 <> '' then
        RaiseErr(5, scan);
      if b1 = '' then
        RaiseErr(5, scan);
      if AnsiSameText(cScanWhileCond, b1) then
        Result.WhileCond:=trimbr
      else
        if AnsiSameText(cScanSkipCond, b1) then
          Result.SkipCond:=trimbr
        else
          if AnsiSameText(cScanNewSep, b1) then
            Result.Sep:=br
          else
            if AnsiSameText(cScanNewLine, b1) then
              Result.NewLine:=true
            else
              if AnsiSameText(cScanNewPar, b1) then
                Result.NewPar:=true
              else
                if AnsiSameText(cScanNewCol, b1) then
                  Result.NewCol:=true
                else
                  if AnsiSameText(cScanNewPage, b1) then
                    Result.NewPage:=true
                  else
                    if AnsiSameText(cScanNoIfEmpty, b1) then
                      Result.NoIfEmpty:=true
                    else
                      if AnsiSameText(cScanNoFirst, b1) then
                        Result.NoFirst:=true
                      else
                        if AnsiSameText(cScanCalc, b1) then
                          InitCalcField(Result.DSindex, trimbr)
                        else
                          if AnsiSameText(cScanMinRows, b1) then
                            Result.MinRows:=StrToIntDef(trimbr, 0)
                          else
                            if AnsiSameText(cScanMaxRows, b1) then
                              Result.MaxRows:=StrToIntDef(trimbr, 0)
                            else
                              if AnsiSameText(cScanTblRow, copy(b1, 1, Length(cScanTblRow))) then
                              begin
                                Result.TblScan:=true;
                                r:=StrToIntDef(copy(b1, Length(cScanTblRow) + 1, Length(b1)), 0);
                                if r < 1 then
                                  RaiseErr(5, scan);
                                SetLength(Result.TblRowCond, r);
                                Result.TblRowCond[r - 1]:=trimbr;
                              end
                              else
                                RaiseErr(5, scan);
    end;
    Inc(i);
  until false;
  // RemovePar(s, true);
end;

// Proc if tags
procedure TrXMLReport.ProcIf(IIF, ProcAllIf: boolean; var s: string);
var
  StartTag, ElseTag, EndTag: string;
  AllTag, IfTag, cond, b1, b2, res: string;
  StartPos, EndPos, ElsePos, p: integer;

  function FindEndPos(SPos: integer; sPart: string): integer;
  var
    EPos: integer;
  begin
    EPos:=AnsiPosEx(EndTag, sPart, SPos, false);
    repeat
      // nested if?
      // try to find another IIF
      SPos:=AnsiPosEx(StartTag, sPart, SPos + Length(StartTag), false);
      if (SPos = 0) or (SPos > EPos) then
        Break;
      // another IIF inside, find next ENDIIF
      EPos:=AnsiPosEx(EndTag, sPart, EPos + Length(EndTag), false);
    until EPos = 0;
    if EPos = 0 then
      RaiseErr(3, EndTag);
    Result:=EPos;
  end;

  function FindElsePos(SPos: integer; s: string): integer;
  var
    ElsePos: integer;
  begin
    ElsePos:=AnsiPosEx(ElseTag, s, SPos, false);
    repeat
      // nested if?
      // try to find another IIF
      // SPos:=AnsiPosEx(StartTag, s, SPos + Length(StartTag), false);
      SPos:=AnsiPosEx(StartTag, s, SPos, false);
      if (SPos = 0) or (SPos > ElsePos) then
        Break;
      // another IIF inside, find end of this IIF
      SPos:=FindEndPos(SPos, s);
      if SPos > ElsePos then
        // ELSEIIF inside another IIF, try to find another ELSEIIF
        ElsePos:=AnsiPosEx(ElseTag, s, SPos, false);
    until ElsePos = 0;
    Result:=ElsePos;
  end;

begin
  // set if tags
  StartTag:=cUsrTagIfStart;
  ElseTag:=cUsrTagElseIf;
  EndTag:=cUsrTagEndIf;
  if IIF then
  begin
    StartTag:=cUsrTagIIfStart;
    ElseTag:=cUsrTagElseIIf;
    EndTag:=cUsrTagEndIIf;
  end;
  // proc if tags
  repeat
    // extract whole string between IF and ENDIF
    StartPos:=AnsiPosEx(StartTag, s, 1, false);
    if StartPos = 0 then
      Break;
    EndPos:=FindEndPos(StartPos, s);
    AllTag:=copy(s, StartPos, EndPos - StartPos);
    // extract conditions
    p:=1;
    IfTag:=FindTagStr(cTag, AllTag, cTagCh, p, true);
    IfTag:=XMLToStr(IfTag);
    SepBracket(IfTag, cond, b1, b2);
    // proc ELSE tag
    b1:=AllTag;
    b2:='';
    ElsePos:=FindElsePos(1, AllTag);
    if ElsePos > 0 then
    begin
      b1:=copy(AllTag, 1, ElsePos - 1);
      b2:=copy(AllTag, ElsePos + Length(ElseTag), Length(AllTag));
    end;
    // find and insert result
    if GetConditionResult(cond) then
      res:=b1
    else
      res:=b2;

    Delete(s, StartPos, EndPos - StartPos + Length(EndTag));
    Insert(res, s, StartPos);
    if not ProcAllIf then
      Break;
  until false;
end;

// proc RTF text
function TrXMLReport.ProcXML(const Str: string): string;
var
  d, TagPos, p, t1, t2: integer;
  s, TagStr, res: string;
  Repl: boolean;
begin
  // proc IIF
  s:=Str;
  ProcIf(true, true, s);
  // proc other tags
  TagPos:=0;
  repeat
    TryScientificFormat:=false;
    // find tag
    TagStr:=FindTagStr(cTag, s, cTagCh, TagPos, false);
    TagStr:=Trim(XMLToStr(TagStr));
    if TagPos = 0 then
      Break;
    d:=Pos(cTagCh, cTag) - 1;

    // check InsertTable tag
    if AnsiPosEx(copy(cInsertTable, 1, AnsiPos(cTagCh, cInsertTable) - 1), TagStr) = 1 then
    begin
      CheckReportFormat(tfDOCX, cInsertTable);
      // find nearest start of table
      p:=0;
      t1:=0;
      repeat
        p:=PosEx(cXMLTableTagBegin[XMLFormat], s, p + 1);
        if (p = 0) or (p > TagPos) then
          Break;
        if p < TagPos then
          t1:=p;
      until false;
      if t1 = 0 then
        RaiseErr(7, cInsertTable);
      // find nearest end of table
      t2:=PosEx(cXMLTableTagEnd[XMLFormat], s, TagPos);
      if t2 = 0 then
        RaiseErr(7, cInsertTable);
      Inc(t2, Length(cXMLTableTagEnd[XMLFormat]));
      // proc InsertTable tag
      res:=copy(s, t1, t2 - t1);
      ProcInsertTable(res);
      // replace whole XML table by new res
      Delete(s, t1, t2 - t1);
      Insert(res, s, t1);
      d:=Length(res);
      TagPos:=t1 + d;
    end
    else
    begin
      res:=ProcTagStr(TagStr, Repl);
      // skip conversion for image
      if copy(cXMLPicTag[XMLFormat], 1, 50) <> copy(res, 1, 50) then
        res:=StrToXML(res);
      // replace 1Exx by scientific format
      if TryScientificFormat then
        res:=CheckScientificFormat(res);
      if Repl then
      begin
        // delete parameter and insert value
        FindTagStr(cTag, s, cTagCh, TagPos, true);
        d:=Length(res);
        Insert(res, s, TagPos);
      end;
      Inc(TagPos, d);
    end;
  until false;
  Result:=s;
end;

// Proc Tag and return tag value
// TagStr is command or value
// Value has format X.Field_name:Format or VariableName:Format or =CalcString:Format
function TrXMLReport.ProcTagStr(const TagStr: string; var Replaced: boolean): string;
var
  Tag, s, Frm, w: string;
  wl, wr: integer;

  procedure ExtractFormat;
  var
    p1: integer;
    br, b1, b2: string;
  begin
    // find format text after bracket or text if bracket not used
    SepBracket(Tag, br, b1, b2);
    if b2 <> '' then
      // after bracket
      Frm:=b2
    else
      if br = '' then
        // text without bracket
        Frm:=b1
      else
        // bracket without format
        Frm:='';

    // find separator in format text
    p1:=Pos(cFormatSep, Frm);
    if p1 = 0 then
      Frm:=''
    else
    begin
      // extract format
      Frm:=copy(Frm, p1 + Length(cFormatSep), Length(Frm));
      Tag:=copy(Tag, 1, Length(Tag) - Length(Frm) - Length(cFormatSep));
    end;

    // proc format strings
    // :<width or :>width or :<width:format or :>width:format
    wl:=0;
    wr:=0;
    w:='';
    if (Frm <> '') and CharInSet(Frm[1], ['<', '>']) then
    begin
      // 1st part as width number
      w:=FindStrPart(Frm, cFormatSep, 1, false);
      if copy(w, 1, 1) = '<' then
        wl:=StrToIntDef(copy(w, 2, 255), 0)
      else
        if copy(w, 1, 1) = '>' then
          wr:=StrToIntDef(copy(w, 2, 255), 0);
      // 2nd part as format string
      p1:=Pos(cFormatSep, Frm);
      if p1 = 0 then
        Frm:=''
      else
        Frm:=copy(Frm, p1 + Length(cFormatSep), Length(Frm));
    end;
  end;

  procedure ProcTagWithFormat;
  begin
    ExtractFormat;
    if Tag <> '' then
    begin
      // SepList
      if AnsiPosEx(cSepList, Tag, 1, false) = 1 then
        s:=ProcSepList(Tag)
      else
        // FirstValue
        if AnsiPosEx(cFirstValue, Tag, 1, false) = 1 then
          s:=ProcFirstValue(Tag)
        else
          // Case
          if AnsiPosEx(cCase, Tag, 1, false) = 1 then
            s:=ProcCase(Tag)
          else
            // NumToText
            if AnsiPosEx(cNumToText, Tag, 1, false) = 1 then
              s:=ProcNumToText(Tag)
            else
              // String routine
              if AnsiPosEx(cStrPrefix, Tag, 1, false) = 1 then
                s:=ProcStrPrefix(Tag)
              else
                if Tag[1] = '=' then
                begin
                  if Frm = '' then
                    Frm:=FDefaultCalcFormat;
                  s:=GetCalcValue(copy(Tag, 2, Length(Tag)), Frm, false)
                end
                else
                begin
                  if IsFieldTag(Tag) then
                    s:=GetDBValue(DatasetCharToIndex(Tag[1]), copy(Tag, 3, Length(Tag)), Frm)
                  else
                  begin
                    s:=GetVarUserValue(Tag, Replaced);
                    if Replaced then
                      s:=ProcXML(s);
                  end;
                end;
    end;
    // align text according format width
    while Length(s) < wl do
      s:=s + ' ';
    while Length(s) < wr do
      s:=' ' + s;
    // truncate text
    if wr > wl then
      wl:=wr;
    if (wl > 0) and (Length(s) > wl) then
      s:=copy(s, 1, wl);
  end;

begin
  Result:='';
  s:='';
  Tag:=TagStr;
  Replaced:=false;
  if Length(Tag) < 2 then
    Exit;
  Replaced:=true;
  // move dataset
  if AnsiPosEx(cScanNext, Tag, 1, false) = 1 then
    FindNext(GetDSIndex(Tag))
  else
    if AnsiPosEx(cScanPrev, Tag, 1, false) = 1 then
      FindPrev(GetDSIndex(Tag))
    else
      if AnsiPosEx(cScanFirst, Tag, 1, false) = 1 then
        FindFirst(GetDSIndex(Tag))
      else
        if AnsiPosEx(cScanLast, Tag, 1, false) = 1 then
          FindLast(GetDSIndex(Tag))
        else
          // bookmark dataset
          if AnsiPosEx(cScanGetBookmark, Tag, 1, false) = 1 then
            GetDSBookmark(GetDSIndex(Tag))
          else
            if AnsiPosEx(cScanGotoBookmark, Tag, 1, false) = 1 then
              GotoDSBookmark(GetDSIndex(Tag))
            else
              // Picture
              if AnsiPosEx(cPicture, Tag, 1, false) = 1 then
                s:=ProcPicture(Tag)
              else
                // BarCode
                if AnsiPosEx(cBarCode, Tag, 1, false) = 1 then
                  s:=ProcBarcode(Tag)
                else
                  // ResetCalc
                  if AnsiSameText(cResetCalc, Tag) then
                    FreeCalcField
                  else
                    // Def var
                    if AnsiPosEx(cDefVar, Tag, 1, false) = 1 then
                      ProcDefVar(Tag)
                    else
                      // ListAllFields
                      if AnsiSameText(cListAllFields, Tag) then
                        s:=ListAllFields
                      else
                        ProcTagWithFormat;
  Result:=s;
end;

function TrXMLReport.ParseRTFFormat(const RTFStr: string): string;
var
  Parser: TrRTFParser;
  L: TStrings;
  i, j: integer;
  t, s, Frm: string;
  RtfFrm: TRtfSupportedFormat;

  procedure AddFrm;
  begin
    // use actual format
    if Frm = '' then
      Result:=Result + s
    else
      Result:=Result + cRtfInlineStart + Frm + cRtfInlineEnd + s;
    s:='';
    // reset actual format
    Frm:='';
  end;

begin
  if not AnsiStartsText('{\rtf1', RTFStr) then
  begin
    // no rtf fomat, copy whole text
    Result:=RTFStr;
    Exit;
  end;
  Result:='';
  L:=TStringList.Create;
  try
    // define supported tags
    L.Add('\par');
    L.Add('\line');
    // rtf supported formats
    if XMLFormat = tfDOCX then
    begin
      for RtfFrm:= Low(TRtfSupportedFormat) to High(TRtfSupportedFormat) do
        if RtfFrm in FRtfSupportedFormats then
        begin
          L.Add(cRtfInlineFrm_Rtf1[RtfFrm]);
          L.Add(cRtfInlineFrm_Rtf2[RtfFrm]);
        end;
      // add user defined format
      for i:=0 to FRtfCustomFormats.Count - 1 do
      begin
        t:=Trim(FRtfCustomFormats.Names[i]);
        if t <> '' then
          L.Add(t);
      end;
    end;

    // parse rtf content
    Parser:=TrRTFParser.Create(Self);
    try
      Parser.RtfContent:=RTFStr;
      Parser.ParseRtfContent(L);

      s:='';
      Frm:='';
      // parse all elements
      for i:=0 to Parser.RtfElements.Count - 1 do
      begin
        t:=Parser.RtfElements[i];
        if t <> '' then
        begin
          if t[1] = '\' then
          begin
            // special tag
            if AnsiStartsText('\f', t) then
              // font id (f0,f1...)
              for j:=0 to Parser.FontTable.Count - 1 do
                if SameText(t, '\' + Parser.FontTable.Names[j]) then
                begin
                  t:=cRtfInlineFont + Parser.FontTable.ValueFromIndex[j];
                  Break;
                end;
            if SameText(t, '\par') or SameText(t, '\line') then
              s:=s + cTxtEndPar
            else
            begin
              // add format of previous text
              if s <> '' then
                AddFrm;
              // add to current format
              Frm:=Frm + t + cParamSep;
            end;
          end
          else
          begin
            // add plain text
            s:=s + t;
          end;
        end;
      end;
      if s <> '' then
        AddFrm;

    finally
      FreeAndNil(Parser);
    end;
  finally
    FreeAndNil(L);
  end;
end;

// Return value of Dataset Field
// FieldName: Name of field or #index
// (FormatStr or DisplayFormat is used for formating)
function TrXMLReport.GetDBValue(DSindex: integer; const FieldName, FormatStr: string): string;
var
  Fld: TField;
  s, Frm: string;

  // proc MEMO and text BLOB field
  function ProcMemo(mstr: string): string;
  var
    i: integer;
    s1, s2: string;
    // REdit: TRichEdit;
    // Stream: TStringStream;
  begin
    mstr:=ParseRTFFormat(mstr);
    {
      if AnsiStartsText('{\rtf1', mstr) then
      begin
      try
      // convert rtf to plaintext
      REdit:=TRichEdit.CreateParented(HWND_MESSAGE);
      try
      REdit.WordWrap:=false;
      REdit.Width:=8000;
      REdit.PlainText:=false;
      Stream:=TStringStream.Create(Fld.AsString);
      try
      REdit.Lines.LoadFromStream(Stream);
      finally
      Stream.Free;
      end;
      REdit.PlainText:=true;
      mstr:=REdit.Lines.Text;
      finally
      REdit.Free;
      end;
      except
      on E: Exception do
      begin
      DlgE('RTF to PlainText error: ' + E.Message);
      raise ;
      end;
      end;
      }
    // replace last #0
    while copy(mstr, Length(mstr), 1) = #0 do
      mstr:=copy(mstr, 1, Length(mstr) - 1);
    // memo replaces
    for i:=0 to FMemoReplaces.Count - 1 do
    begin
      s1:=rtool.FindStrPart(FMemoReplaces[i], '=', 1, false);
      s2:=rtool.FindStrPart(FMemoReplaces[i], '=', 2, false);
      s1:=rtool.ReplaceStr('<TAB>', #9, s1, true, false);
      s2:=rtool.ReplaceStr('<TAB>', #9, s2, true, false);
      if (s1) = '' then
        Continue;
      mstr:=ReplaceStr(s1, s2, mstr, true, false);
    end;

    Result:=mstr;
  end;

  function IsImageField: boolean;
  begin
    Result:=rdbtool.IsGraphicField(Fld) or SameText(FVariables.Values[Fld.FieldName], 'image');
    // or (AnsiContainsText(Fld.FieldName, 'picture')) or (AnsiContainsText(Fld.FieldName, 'image'));
  end;

begin
  TryScientificFormat:=false;
  Result:=CheckDSIndex(DSindex);
  if Result <> '' then
    Exit;
  // EOF while using NEXT or PREV
  Frm:=FormatStr;
  if DS_eof[DSindex] then
    Exit;
  try
    // AsText:=true;
    if FieldName = '#' then
    begin
      Result:=IntToStr(DS[DSindex].RecNo);
      Exit;
    end;
    if FieldName = '##' then
    begin
      Result:=IntToStr(DS[DSindex].RecordCount);
      Exit;
    end;
    if copy(FieldName, 1, 1) = '#' then
      Fld:=DS[DSindex].Fields[StrToInt(copy(FieldName, 2, 255))]
    else
      Fld:=DS[DSindex].FieldByName(FieldName);
    // special format
    if AnsiSameText(Frm, cLabel) then
      s:=Fld.DisplayLabel
    else
      if Fld.IsNull then
        s:=''
      else
        case Fld.DataType of
          ftString, ftWideString:
            begin
              if Frm = '' then
                s:=Fld.Text
              else
                s:=FormatMaskText(Frm + ';0;', Fld.AsString);
            end;
          ftSmallint, ftInteger, ftWord, ftBCD, ftAutoInc, ftFloat, ftCurrency:
            begin
              if Frm = '' then
                Frm:=TNumericField(Fld).DisplayFormat;
              if Frm = '' then
                Frm:=FDefaultNumericFormat;
              s:=FormatFloat(Frm, Fld.AsFloat);
              TryScientificFormat:=true;
            end;
          ftBoolean:
            begin
              if Frm = '' then
                Frm:=TBooleanField(Fld).DisplayValues;
              if Frm = '' then
                Frm:=FDefaultBooleanFormat;
              if (Frm = '') or (Pos(';', Frm) = 0) then
                s:=Fld.Text
              else
                if Fld.AsBoolean then
                  s:=copy(Frm, 1, Pos(';', Frm) - 1)
                else
                  s:=copy(Frm, Pos(';', Frm) + 1, Length(Frm));
            end;
          ftDate:
            begin
              if Frm = '' then
                Frm:=TDateTimeField(Fld).DisplayFormat;
              if Frm = '' then
                Frm:=FDefaultDateFormat;
              if Frm = '' then
                s:=DateToStr(Fld.AsDateTime)
              else
                s:=FormatDateTime(Frm, Fld.AsDateTime);
            end;
          ftTime:
            begin
              if Frm = '' then
                Frm:=TDateTimeField(Fld).DisplayFormat;
              if Frm = '' then
                Frm:=FDefaultTimeFormat;
              if Frm = '' then
                s:=TimeToStr(Fld.AsDateTime)
              else
                s:=FormatDateTime(Frm, Fld.AsDateTime);
            end;
          ftDateTime:
            begin
              if Frm = '' then
                Frm:=TDateTimeField(Fld).DisplayFormat;
              if Frm = '' then
                Frm:=FDefaultDateFormat;
              s:=FormatDateTime(Frm, Fld.AsDateTime);
            end;
          ftTimeStamp:
            begin
              if Frm = '' then
                Frm:=TSQLTimeStampField(Fld).DisplayFormat;
              if Frm = '' then
                Frm:=FDefaultDateFormat;
              s:=FormatDateTime(Frm, Fld.AsDateTime);
            end;
          ftMemo, ftWideMemo, ftFmtMemo, ftGraphic, ftBlob:
            begin
              if (Fld.DataType in [ftMemo, ftWideMemo, ftFmtMemo]) or not IsImageField then
              begin
                s:=ProcMemo(Fld.AsString);
                if Pos('<<', s) > 0 then
                  s:=ProcXML(s);
              end
              else
              begin
                s:=CreateFormatPicture(DSindex, Fld, '', nil, Frm);
              end;
            end;
        else
          s:=Fld.Text;
        end;
    // Replacing EOL with  XMLEndPar - moved to StrToXML
    // if AsText and not FTxtTemplate then
    // s:=ReplaceStr(cTxtEndPar, cXMLEndPar, s, true, false);
  except
    on E: Exception do
    begin
      if RaiseUnknownField then
        RaiseErr(1, FieldName);
      if FFieldNotFoundValue = '' then
        s:='?' + IndexToDatasetChar(DSindex) + '.' + FieldName + '? (' + E.Message + ')'
      else
        s:=FFieldNotFoundValue;
    end;
  end;
  Result:=s;
end;

// Return value of variable
function TrXMLReport.GetVarUserValue(const VarName: string; var Replaced: boolean): string;
var
  F, p, s: string;
  res: variant;
  val: double;
  arrCount, Idx: integer;
  Fld: TField;
  arr: array of variant;
begin
  Result:=VarName;
  Replaced:=false;
  if FVariables.IndexOfName(VarName) >= 0 then
  begin
    Result:=FVariables.Values[VarName];
    Replaced:=true;
  end
  else
    if Assigned(FOnUserCalcParams) then
    begin
      SepBracket(VarName, p, F, s);
      if s <> '' then
        RaiseErr(5, VarName);
      res:='';
      arrCount:=0;
      repeat
        s:=rtool.RemoveStrPart(p, cParamSep);
        if s <> '' then
        begin
          SetLength(arr, arrCount + 1);
          Fld:=nil;
          if IsConstantTag(s) then
            arr[arrCount]:=copy(s, 2, Length(s) - 2)
          else
          begin
            if IsFieldTag(s) then
            begin
              Idx:=Self.DatasetCharToIndex(s[1]);
              if DSActive(Idx) then
                Fld:=DS[Idx].FindField(copy(s, 3, Length(s)));
            end;
            if Assigned(Fld) then
              arr[arrCount]:=Fld.AsVariant
            else
              if IsNumericTag(s, val) then
                arr[arrCount]:=val
              else
                arr[arrCount]:=s;
          end;
          Inc(arrCount);
        end;
      until p = '';
      FOnUserCalcParams(Self, F, arr, res, Replaced);
      if not Replaced then
        Result:=VarName
      else
        Result:=res;
    end
    else
      if Assigned(FOnUserCalc) then
      begin
        FOnUserCalc(Self, Result, Replaced);
        if not Replaced then
          Result:=VarName;
      end;
end;

// Return value of calc string
function TrXMLReport.GetCalcValue(const CalcStr, FormatStr: string; ErrFieldAsZero: boolean): string;

// Subfunction, return value of Str
  function GetVal(Str: string): double;
  var
    Idx, F: integer;
    s, p1, p2: string;
    err, Repl, CS: boolean;
    Fld: TField;
  begin
    Result:=0;
    Str:=Trim(Str);
    if Str = '' then
      Exit;
    // Check numeric value
    err:=not IsNumericTag(Str, Result);
    // Check Field value
    if err then
      try
        if Str[2] = '.' then
        begin
          Idx:=DatasetCharToIndex(Str[1]);
          p1:=copy(Str, 3, Length(Str));
          if p1 = '#' then
            Result:=DS[Idx].RecNo
          else
            if p1 = '##' then
              Result:=DS[Idx].RecordCount
            else
            begin
              Fld:=DS[Idx].FieldByName(p1);
              if Fld is TBooleanField then
                Result:=ord(Fld.AsBoolean)
              else
                if ErrFieldAsZero then
                begin
                  try
                    Result:=Fld.AsFloat;
                  except
                    Result:=0;
                  end;
                end
                else
                  Result:=Fld.AsFloat;
            end;
          err:=false;
        end;
      except
      end;
    // Check calc functions
    if err then
      for Idx:=0 to 9 do
      begin
        F:=1;
        s:=FindTagStr(cCalcFunc[Idx], Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 1) then
        begin
          Result:=GetCalcFieldValue(Idx, s);
          err:=false;
          Break;
        end;
      end;
    // Check datetime functions
    if err then
    begin
      if SameText(Str, cGetToday) then
      begin
        Result:=Trunc(Date);
        err:=false;
      end;
      if SameText(Str, cGetNow) then
      begin
        Result:=Frac(Time);
        err:=false;
      end;
    end;
    // check math Abs
    if err then
      try
        s:=FindTagStr(cMathAbs, Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 0) then
        begin
          if Length(s) > 1 then
            s:=ProcTagStr(s, Repl);
          Result:=Abs(StrToFloat(s));
          err:=false;
        end;
      except
      end;
    // check string routine "strlen"
    if err then
      try
        s:=FindTagStr(cStrLen, Str, cTagCh, F, false);
        if (F = 1) and (Length(s) > 1) then
        begin
          s:=ProcTagStr(s, Repl);
          Result:=Length(s);
          err:=false;
        end;
      except
      end;
    // check string routine "strpos"
    if err then
      try
        s:=FindTagStr(cStrPos, Str, cTagCh, F, false);
        CS:=false;
        if not((F = 1) and (Length(s) > 1)) then
        begin
          s:=FindTagStr(cStrPosCS, Str, cTagCh, F, false);
          CS:=true;
        end;
        if (F = 1) and (Length(s) > 1) then
        begin
          p1:=GetListPart(s, cListSep2);
          p2:=GetListPart(s, cListSep2);
          // if s<>'' then raise Exception.Create(s);
          p1:=ProcTagStr(p1, Repl);
          if copy(p2, 1, 1) = cConstSep then
            p2:=copy(p2, 2, Length(p2) - 2)
          else
            p2:=ProcTagStr(p2, Repl);
          Result:=AnsiPosEx(p2, p1, 1, CS);
          err:=false;
        end;
      except
      end;
    // Check Variable value
    if err then
      try
        Result:=StrToFloat(GetVarUserValue(Str, Repl));
        err:=false;
      except
      end;
    if err then
      RaiseErr(4, Str);
  end;

// Subfunction, return value of CalcStr
  function Calc(CalcStr: string): double;
  var
    Bracket, b1, b2: string;

    // Subfunction, return sign value
    function DoSign(Str, Sign: string; var res: double): boolean;
    var
      p: integer;
      part1, part2: string;
      res1, res2: double;
    begin
      Result:=false;
      p:=Pos(Sign, Str);
      if p = 0 then
        Exit;
      part1:=Trim(copy(Str, 1, p - 1));
      part2:=Trim(copy(Str, p + Length(Sign), Length(Str)));
      if (Sign = '-') and (Length(part1) > 0) and (CharInSet(part1[Length(part1)], ['['])) then
        Exit;
      Result:=true;
      res1:=Calc(part1);
      res2:=Calc(part2);
      if Sign = '+' then
        res:=res1 + res2
      else
        if Sign = '-' then
          res:=res1 - res2
        else
          if Sign = '*' then
            res:=res1 * res2
          else
            if Sign = '/' then
              try
                res:=res1 / res2;
              except
                res:=0;
              end;
    end;

  begin
    Result:=0;
    CalcStr:=Trim(CalcStr);
    if CalcStr = '' then
      Exit;
    // Calculate brackets
    repeat
      SepBracket(CalcStr, Bracket, b1, b2);
      if Bracket = '' then
        Break;
      CalcStr:=b1 + FloatToStrF(Calc(Bracket), ffFixed, 80, 40) + b2;
    until false;
    // Calculate all CalcStr
    if not DoSign(CalcStr, '+', Result) then
      if not DoSign(CalcStr, '-', Result) then
        if not DoSign(CalcStr, '*', Result) then
          if not DoSign(CalcStr, '/', Result) then
            Result:=GetVal(CalcStr);
  end;

begin
  if (copy(FormatStr, 1, 1) = cDateSep) and (copy(FormatStr, Length(FormatStr), 1) = cDateSep) then
    Result:=FormatDateTime(copy(FormatStr, 2, Length(FormatStr) - 2), Calc(CalcStr))
  else
  begin
    Result:=FormatFloat(FormatStr, Calc(CalcStr));
    TryScientificFormat:=true;
  end;
end;

// proc condition string
function TrXMLReport.GetConditionResult(const CondStr: string): boolean;
var
  i, p1, FuncIdx: integer;
  br, cond1, cond2: string;
  IsNum1, IsNum2, Repl: boolean;
  dt: TDateTime;

  function GetCondValue(var Str: string): boolean;
  var
    s, FldName: string;
  begin
    Result:=false;
    s:=Str;
    try
      // string constant? "test"
      if IsConstantTag(s) then
        Str:=ClearStr(cConstSep, s, false)
      else
        // string date? #YYYYMMDDhhmmss#
        if (s[1] = cDateSep) and (s[Length(s)] = cDateSep) then
        begin
          Str:=ClearStr(cDateSep, s, false);
          if (Length(Str) = 8) or (Length(Str) = 14) then
          begin
            dt:=EncodeDate(StrToInt(copy(Str, 1, 4)), StrToInt(copy(Str, 5, 2)), StrToInt(copy(Str, 7, 2)));
            if Length(Str) = 14 then
              dt:=dt + EncodeTime(StrToInt(copy(Str, 9, 2)), StrToInt(copy(Str, 11, 2)), StrToInt(copy(Str, 13, 2)), 0);
            Str:=FloatToStrF(dt, ffFixed, 80, 40);
            Result:=true;
          end
          else
            raise Exception.Create('');
        end
        else
        {
          // String routine
          if AnsiPosEx(cStrPrefix,s,1,false)=1 then Str:=ProcStrPrefix(s)
          else
          }
        // last change - calc value
        begin
          Str:=GetCalcValue(s, '', false);
          Result:=true;
        end;
    except
      try
        if IsFieldTag(s) then
        begin
          FldName:=copy(s, 3, Length(s));
          Str:=GetDBValue(DatasetCharToIndex(s[1]), FldName, '');
        end
        else
        begin
          Repl:=false;
          Str:=GetVarUserValue(s, Repl);
          if not Repl then
            raise Exception.Create('');
        end;
      except
        RaiseErr(4, s);
      end;
    end;
  end;

begin
  // proc brackets
  SepBracket(CondStr, br, cond1, cond2);
  if br <> '' then
  begin
    Result:=GetConditionResult(cond1 + cCondResArr[GetConditionResult(br)] + cond2);
    Exit;
  end;
  // not sign
  if AnsiPosEx(cCondNot, Trim(CondStr), 1, false) = 1 then
  begin
    Result:=not GetConditionResult(FindStrPart(CondStr, cCondNot, 2, false));
    Exit;
  end;
  // and sign
  p1:=AnsiPosEx(cCondAnd, CondStr, 1, false);
  if p1 > 0 then
  begin
    cond1:=copy(CondStr, 1, p1 - 1);
    cond2:=copy(CondStr, p1 + Length(cCondAnd), Length(CondStr));
    Result:=GetConditionResult(cond1) and GetConditionResult(cond2);
    Exit;
  end;
  // or sign
  p1:=AnsiPosEx(cCondOr, CondStr, 1, false);
  if p1 > 0 then
  begin
    cond1:=copy(CondStr, 1, p1 - 1);
    cond2:=copy(CondStr, p1 + Length(cCondOr), Length(CondStr));
    Result:=GetConditionResult(cond1) or GetConditionResult(cond2);
    Exit;
  end;

  Result:=false;
  i:=-1;
  // find sign
  for FuncIdx:=0 to High(cCondFunc) do
  begin
    p1:=Pos(cCondFunc[FuncIdx], CondStr);
    if p1 > 0 then
      i:=FuncIdx;
    if i >= 0 then
      Break;
  end;
  if i >= 0 then
  begin
    // sign exist
    cond1:=Trim(copy(CondStr, 1, p1 - 1));
    cond2:=Trim(copy(CondStr, p1 + Length(cCondFunc[i]), Length(CondStr)));
    IsNum1:=GetCondValue(cond1);
    IsNum2:=GetCondValue(cond2);
    if IsNum1 and IsNum2 then
    begin
      // numeric values
      case i of
        0, 4:
          Result:=StrToFloat(cond1) = StrToFloat(cond2);
        1:
          Result:=StrToFloat(cond1) >= StrToFloat(cond2);
        2:
          Result:=StrToFloat(cond1) <= StrToFloat(cond2);
        3:
          Result:=StrToFloat(cond1) <> StrToFloat(cond2);
        5:
          Result:=StrToFloat(cond1) > StrToFloat(cond2);
        6:
          Result:=StrToFloat(cond1) < StrToFloat(cond2);
      end;
    end
    else
    begin
      // string values
      case i of
        0:
          Result:=cond1 = cond2;
        1:
          Result:=AnsiCompareText(cond1, cond2) >= 0;
        2:
          Result:=AnsiCompareText(cond1, cond2) <= 0;
        3:
          Result:=AnsiCompareText(cond1, cond2) <> 0;
        4:
          Result:=AnsiCompareText(cond1, cond2) = 0;
        5:
          Result:=AnsiCompareText(cond1, cond2) > 0;
        6:
          Result:=AnsiCompareText(cond1, cond2) < 0;
      end;
    end;
  end
  else
  begin
    // sign not exist=boolean field or text result ("FALSE" or "TRUE")
    if CondStr = cCondResArr[false] then
      Result:=false
    else
      if CondStr = cCondResArr[true] then
        Result:=true
      else
        try
          if CondStr[2] = '.' then
          begin
            if CondStr[3] = '0' then
              Result:=DS[DatasetCharToIndex(CondStr[1])].IsEmpty
            else
              if CondStr[3] = '@' then
                Result:=DS_eof[DatasetCharToIndex(CondStr[1])]
              else
                Result:=DS[DatasetCharToIndex(CondStr[1])].FieldByName(copy(CondStr, 3, Length(CondStr))).AsBoolean;
          end
          else
            raise Exception.Create('');
        except
          RaiseErr(4, CondStr);
        end;
  end;
end;

// -----------------------------------------------------------
// Format Text
// convert float string in 123E02 format to 123 x 10^2 format
function TrXMLReport.CheckScientificFormat(const s: string): string;
var
  i: integer;
  E, sgn: string;
begin
  Result:=s;
  case XMLFormat of
    tfDOCX:
      begin
        i:=Pos('E', Result);
        if FConvertScientificFormat and (i > 0) then
        begin
          // extract exponent and sign
          E:=copy(Result, i + 1, Length(Result));
          sgn:='';
          if copy(E, 1, 1) = '-' then
            sgn:='-';
          if (copy(E, 1, 1) = '+') or (copy(E, 1, 1) = '-') then
            E:=copy(E, 2, Length(E));
          // extract leading zeros
          while (Length(E) > 1) and (copy(E, 1, 1) = '0') do
            E:=copy(E, 2, Length(E));
          // create format str
          Result:=copy(Result, 1, i - 1) + ' x 10';
          Result:=Result + Format(cDOCXBlockFormat, [cDOCXSuperScript]) + sgn + E + cDOCXBlockNoFormat;
        end;
      end;
  end;
end;

// find one run and extract rPr formats
function TrXMLReport.GetRunXMLFormat(StartPos: integer; const s: string): TRunInfo;
var
  spr, srun, sdef: string;
  prun1, prun2, ppr1, ppr2: integer;
begin
  Result.RunBeginPos:=0;
  Result.RunEndPos:=0;
  Result.RunXML:='';
  Result.RunPrFormatFontSize:='';
  Result.RunPrFormatFontName:='';
  if StartPos = 0 then
    Exit;
  // find run
  prun1:=StartPos;
  while prun1 > 0 do
  begin
    // started tag without > coz can contain inner elements
    // eg. <w:r w:rsidDel="00000000" w:rsidR="00000000" w:rsidRPr="00000000">
    if SameText(copy(s, prun1, 4), '<w:r') then
      Break;
    Dec(prun1);
  end;
  if prun1 < 1 then
    Exit;
  prun2:=AnsiPosEx('</w:r>', s, prun1, false);
  if prun2 = 0 then
    Exit;
  prun2:=prun2 + 5;
  srun:=copy(s, prun1, prun2 - prun1 + 1);

  // find rPr
  ppr1:=AnsiPos('<w:rPr>', srun);
  if ppr1 > 0 then
  begin
    ppr2:=AnsiPos('</w:rPr>', srun);
    if ppr2 > 0 then
    begin
      spr:=copy(srun, ppr1, ppr2 - ppr1);
      // find defs inside rPr
      sdef:=rtool.FindStrFromTo('<w:rFonts', '/>', spr, false);
      if sdef <> '' then
        Result.RunPrFormatFontName:=sdef + '/>';
      sdef:=rtool.FindStrFromTo('<w:sz', '/>', spr, false);
      if sdef <> '' then
        Result.RunPrFormatFontSize:=sdef + '/>';
      sdef:=rtool.FindStrFromTo('<w:szCs', '/>', spr, false);
      if sdef <> '' then
        Result.RunPrFormatFontSize:=Result.RunPrFormatFontSize + sdef + '/>';
    end;
  end;
  Result.RunBeginPos:=prun1;
  Result.RunEndPos:=prun2;
  Result.RunXML:=srun;
end;

// format text according formatting tags inside of text
procedure TrXMLReport.CheckInlineFormat(var s: string);
var
  i, p1, p2, CheckCounter: integer;
  fln, f1, f2, f3, ActFontSize, ActFontName: string;
  RtfFrm: TRtfSupportedFormat;
  used: boolean;

  function ReplaceOneFormat(FrmStart, FrmStop, XMLFrm: string; ReplaceAll: boolean): integer;
  var
    space, FontTag, XmlTag: string;
    pfrmg, pfrmrun: integer;
    ri: TRunInfo;
  begin
    Result:=0;
    repeat
      pfrmg:=Pos(FrmStart, s);
      if pfrmg = 0 then
        Break;
      // Get run info
      ri:=GetRunXMLFormat(pfrmg, s);
      if ri.RunXML = '' then
        Break;
      // replace format
      pfrmrun:=Pos(FrmStart, ri.RunXML);
      // extra run for spaces
      space:='';
      while (pfrmrun > 1) and CharInSet(ri.RunXML[pfrmrun - 1], [' ']) do
      begin
        Dec(pfrmrun);
        space:=space + ' ';
      end;
      if space <> '' then
        ri.RunXML:=copy(ri.RunXML, 1, pfrmrun - 1) + cDOCXBlockNoFormat + space + copy(ri.RunXML, pfrmrun + Length(space), Length(ri.RunXML));

      // font formats
      if ActFontName = '' then
        FontTag:=ri.RunPrFormatFontName
      else
        FontTag:=ActFontName;
      if ActFontSize = '' then
        FontTag:=FontTag + ri.RunPrFormatFontSize
      else
        FontTag:=FontTag + ActFontSize;
      // insert begin of new run with format
      if XMLFrm + FontTag = '' then
        XmlTag:=cDOCXBlockNoFormat
      else
        XmlTag:=cDOCXBlockFormat;
      ri.RunXML:=ReplaceStr(FrmStart, Format(XmlTag, [XMLFrm + FontTag]), ri.RunXML, false, false);
      // insert end of run and start new run with font format
      if FontTag = '' then
        XmlTag:=cDOCXBlockNoFormat
      else
        XmlTag:=cDOCXBlockFormat;
      ri.RunXML:=ReplaceStr(FrmStop, Format(XmlTag, [FontTag]), ri.RunXML, false, false);
      // replace old run by new with format
      s:=copy(s, 1, ri.RunBeginPos - 1) + ri.RunXML + copy(s, ri.RunEndPos + 1, Length(s));

      Inc(Result);
      if not ReplaceAll or (Result > 100000) then
        Break;
    until false;
    // s:=ReplaceStr(FrmStop, cDOCXBlockNoFormat, s, false, false);
  end;

begin
  if XMLFormat = tfDOCX then
  begin
    // proc all RTF inline formats
    CheckCounter:=0;
    p1:=1;
    ActFontSize:='';
    ActFontName:='';
    repeat
      p1:=rtool.AnsiPosEx(cRtfInlineStart, s, p1);
      if p1 = 0 then
        Break;
      p2:=rtool.AnsiPosEx(cRtfInlineEnd, s, p1);
      if p2 = 0 then
        Break;
      f1:=copy(s, p1 + Length(cRtfInlineStart), p2 - p1 - Length(cRtfInlineStart));
      f2:=f1;
      fln:='';
      repeat
        f3:=rtool.RemoveStrPart(f2, cParamSep);
        if f3 = '' then
          Break;
        used:=false;
        // own format first coz can override standard format
        for i:=0 to FRtfCustomFormats.Count - 1 do
          if SameText(f3, Trim(FRtfCustomFormats.Names[i])) then
          begin
            fln:=fln + FRtfCustomFormats.ValueFromIndex[i];
            used:=true;
            Break;
          end;
        if not used then
          // supported rtf formats
          for RtfFrm:= Low(TRtfSupportedFormat) to High(TRtfSupportedFormat) do
          begin
            if SameText(f3, cRtfInlineFrm_Rtf1[RtfFrm]) then
            begin
              fln:=fln + cRtfInlineFrm_XML[RtfFrm];
              used:=true;
              Break;
            end;
          end;
        if not used then
          // font size
          if Pos('\fs', f3) = 1 then
          begin
            i:=StrToIntDef(copy(f3, 4, 255), 0);
            if i > 0 then
            begin
              ActFontSize:=Format(cRtfInlineFrm_XML[rtfFontSize], [i, i]);
              used:=true;
            end;
          end;
        if not used then
          // font name
          if Pos(cRtfInlineFont, f3) = 1 then
          begin
            f3:=Trim(copy(f3, 7, 255));
            if f3 <> '' then
              ActFontName:=Format(cRtfInlineFrm_XML[rtfFontName], [f3, f3, f3]);
          end;
      until false;
      // replace format, break if not replaced
      if ReplaceOneFormat(cRtfInlineStart + f1 + cRtfInlineEnd, '', fln, false) = 0 then
        Break;
      Inc(CheckCounter);
      if CheckCounter > 100000 then
        Break;
    until false;

    // proc custom inline format
    for i:=0 to InlineFormatting.Count - 1 do
    begin
      fln:=InlineFormatting[i];
      f1:=rtool.FindStrPart(fln, #9, 1, true);
      f2:=rtool.FindStrPart(fln, #9, 2, true);
      f3:=rtool.FindStrPart(fln, #9, 3, true);
      if f3 <> '' then
        ReplaceOneFormat(f1, f2, f3, true);
    end;
  end;
end;

// format or replace text (callback procedure)
procedure TrXMLReport.ProcFormatText(var s: string);
const
  Sep = '\ .,;()[]_?!"''<>+-*/=';
var
  Block1, Block2, p1, p2: integer;
  Str, sp, sfrm: string;
  skip: boolean;
  ri: TRunInfo;
begin
  if XMLFormat = tfDOCX then
  begin
    // find block range
    Block1:=1;
    Block2:=Length(s);
    if SeekInfo.SeekFrom <> '' then
    begin
      p1:=rtool.AnsiPosEx(SeekInfo.SeekFrom, s, 1, false);
      if p1 > 0 then
        Block1:=p1;
    end;
    if SeekInfo.SeekTo <> '' then
    begin
      p2:=rtool.AnsiPosEx(SeekInfo.SeekTo, s, Block1, false);
      if p2 > 0 then
        Block2:=p2;
    end;

    // find text inside block
    p1:=Block1;
    repeat
      p1:=rtool.AnsiPosEx(SeekInfo.OldText, s, p1, false);
      if p1 = 0 then
        Break;
      p2:=p1 + Length(SeekInfo.OldText);
      if p2 > Block2 then
        Break;
      // check whole words condition
      skip:=false;
      if SeekInfo.WholeWord then
      begin
        if p1 > 1 then
        begin
          Str:=copy(s, p1 - 1, 1);
          skip:=Pos(Str, Sep) = 0;
        end;
        if not skip then
        begin
          Str:=copy(s, p2, 1);
          skip:=Pos(Str, Sep) = 0;
        end;
      end;

      if skip then
        p1:=p2
      else
      begin
        // do action
        case SeekInfo.Action of
          1:
            begin
              // format
              ri:=GetRunXMLFormat(p1, s);
              sfrm:=ri.RunPrFormatFontName + ri.RunPrFormatFontSize;
              Str:=Format(cDOCXBlockFormat, [sfrm + SeekInfo.SetFrmStart]) + copy(s, p1, p2 - p1) + Format(cDOCXBlockFormat, [sfrm + SeekInfo.SetFrmStop]);
              // extra run for spaces
              sp:='';
              while (p1 > 1) and CharInSet(s[p1 - 1], [' ']) do
              begin
                Dec(p1);
                sp:=sp + ' ';
              end;
              if sp <> '' then
                Str:=Format(cDOCXBlockFormat, [sfrm]) + sp + Str;
            end;
          2:
            begin
              // replace
              Str:=SeekInfo.NewText;
            end;
        end;
        s:=copy(s, 1, p1 - 1) + Str + copy(s, p2, Length(s));
        p1:=p1 + Length(Str);
      end;
    until false;
  end;
end;

procedure TrXMLReport.FormatText(const TextToFormat: string; FormatTypes: TXMLTextFormats; WholeWordOnly: boolean = false; const SeekFrom: string = '';
  const SeekTo: string = '');
begin
  SeekInfo.Action:=1;
  SeekInfo.OldText:=TextToFormat;
  SeekInfo.NewText:='';
  SeekInfo.WholeWord:=WholeWordOnly;
  SeekInfo.SeekFrom:=SeekFrom;
  SeekInfo.SeekTo:=SeekTo;
  SeekInfo.SetFrmStart:='';
  SeekInfo.SetFrmStop:='';
  if frmBold in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + cDOCXBold;
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '';
  end;
  if frmItalic in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + cDOCXItalic;
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '';
  end;
  if frmUnderline in FormatTypes then
  begin
    SeekInfo.SetFrmStart:=SeekInfo.SetFrmStart + cDOCXUnderline;
    SeekInfo.SetFrmStop:=SeekInfo.SetFrmStop + '';
  end;
  ProcFormatText(RepStr);
end;

procedure TrXMLReport.ReplaceText(const OldText, NewText: string; WholeWordOnly: boolean = false; const SeekFrom: string = ''; const SeekTo: string = '');
var
  i: integer;
  newt: string;
begin
  SeekInfo.Action:=2;
  SeekInfo.OldText:=OldText;
  newt:=NewText;

  // format superscript sign,  \^ starts superscript, \= ends superscript
  i:=Pos('\^', newt);
  if i > 0 then
  begin
    newt:=copy(newt, 1, i - 1) + Format(cDOCXBlockFormat, [cDOCXSuperScript]) + copy(newt, i + 2, Length(newt));
    i:=Pos('\=', newt);
    if i > 0 then
      newt:=copy(newt, 1, i - 1) + cDOCXBlockNoFormat + copy(newt, i + 2, Length(newt))
    else
      newt:=newt + cDOCXBlockNoFormat;
  end;
  // format subscript sign,  \_ starts subscript, \= ends subscript
  i:=Pos('\_', newt);
  if i > 0 then
  begin
    newt:=copy(newt, 1, i - 1) + Format(cDOCXBlockFormat, [cDOCXSubScript]) + copy(newt, i + 2, Length(newt));
    i:=Pos('\=', newt);
    if i > 0 then
      newt:=copy(newt, 1, i - 1) + cDOCXBlockNoFormat + copy(newt, i + 2, Length(newt))
    else
      newt:=newt + cDOCXBlockNoFormat;
  end;
  SeekInfo.NewText:=newt;

  SeekInfo.WholeWord:=WholeWordOnly;
  SeekInfo.SeekFrom:=SeekFrom;
  SeekInfo.SeekTo:=SeekTo;
  SeekInfo.SetFrmStart:='';
  SeekInfo.SetFrmStop:='';
  ProcFormatText(RepStr);
end;

function TrXMLReport.NamedTables_LoadContent(const FileName: string; const TableNamePrefix: string = ''): integer;
begin
  Result:=0;
  SetReportFormat(FileName);
  CheckReportFormat(tfDOCX, 'TableContent');
  FLoadTableNamePrefix:=TableNamePrefix;
  if FLoadTableNamePrefix = '' then
    FLoadTableNamePrefix:='*';
  try
    if ProcXMLFiles(FileName, true) then
      Result:=Length(NamedTables_Content);
  finally
    FLoadTableNamePrefix:='';
  end;
end;

procedure TrXMLReport.NamedTables_DefineContent(const TableNameID: string; const CellXMLContent: string; const PlainText: string);
var
  i, Idx: integer;
begin
  Idx:=-1;
  for i:=0 to High(NamedTables_Content) do
    if SameText(TableNameID, NamedTables_Content[i].TableName) then
    begin
      Idx:=i;
      Break;
    end;
  if Idx < 0 then
  begin
    SetLength(NamedTables_Content, Length(NamedTables_Content) + 1);
    Idx:= High(NamedTables_Content);
  end;
  NamedTables_Content[Idx].TableName:=TableNameID;
  NamedTables_Content[Idx].CellXMLContent:=CellXMLContent;
  NamedTables_Content[Idx].PlainText:=PlainText;
end;

procedure TrXMLReport.NamedTables_ClearContent;
begin
  SetLength(NamedTables_Content, 0);
end;

end.
