unit rdbtool;

{<
  rdbtool unit contains set of usefull database functions used by all other units.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Dec 2011)
  @lastmod(Nov 2020)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Math,
  Db, DBConsts, Grids, DBGrids, Menus, Forms, Variants,
{$IFDEF rUseDBComp} rDBRecView, rStringGridEd, {$ENDIF} RosiCompGlobalCfg;

const
  sfSimpleFld: Set of TFieldType = ([ftString, ftWideString, ftSmallint, ftInteger, ftWord, ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
    ftAutoInc, ftLargeint, ftTimeStamp, ftFMTBcd, ftLongWord, ftShortint, ftByte, ftExtended, ftSingle, ftGuid]);
  sfApostrFld: Set of TFieldType = ([ftString, ftWideString, ftDate, ftTime, ftDateTime, ftTimeStamp, ftMemo, ftFmtMemo, ftWideMemo, ftBlob, ftGuid]);
  // sBooleanString: array [boolean] of string = ('false', 'true');

type
  {< @exclude }
{$IFDEF D2006}
  // use widestrings in Delphi2006
  TSQLStrings = TWideStrings;
{$ELSE}
  TSQLStrings = TStrings;
{$ENDIF}

  { list of possible filter join conditions }
  TFilterJoin = (fjNone, fjAnd, fjOr);
  { list of possible filter compare types }
  TrLikeCompareType = (lcAnyPos, lcBegin, lcSame, lcAllWords, lcAnyWords);

  { return localized hint string }
function GetSearchHintText(SearchText, NewKey: string; SearchingType: TSearchingType; Found, AllowFilter, Filtered: boolean): string;
{ return Dataset from object of type Dataset, Datasource or DBGrid }
function GetDataSet(GridOrDSet: TObject): TDataSet;
{ return localized text of Dataset state }
function GetDSStateStr(GridOrDSet: TObject): string;
{ indicate that Dataset is TClientDataset }
function IsClientDataset(Dataset: TDataSet): boolean;
{ indicate that Dataset is Advantage Database }
function IsAdvantageDatabaseDataset(Dataset: TDataSet): boolean;
{ assign colors from one grid to another }
procedure AssignGridProps(Source, Dest: TCustomDBGrid);

{ return whether object is in editing state }
function IsEditing(GridOrDSet: TObject): boolean;
{ return whether Field is ftData type }
function IsDataField(Fld: TField): boolean;
{ return whether Field is simple field (text, number, date) }
function IsSimpleField(Fld: TField): boolean;
{ return whether Field needs appostrophe in Filter }
function IsApostrField(Fld: TField): boolean;
{ return whether Field is Memo field,
  result depends also on RosiCompConfig.rDBToolCfg.BlobField_ftBlobAsMemo config value }
function IsMemoField(Fld: TField): boolean;
{ return whether Field is Graphics field }
function IsGraphicField(Fld: TField): boolean;
{ return whether Field is DateTime type }
function IsDateTimeField(Fld: TField): boolean;
{ return whether Field is Date type }
function IsDateField(Fld: TField): boolean;
{ return whether Field is Time type }
function IsTimeField(Fld: TField): boolean;
{ return defined text value for Boolean Field and defined State }
function GetBooleanFieldTextValue(Fld: TField; State: boolean): string;
{ return state according defined TextValue for Boolean Field }
function GetBooleanFieldState(Fld: TField; TextValue: string): boolean;

{ set FldsDef according defined parameters }
function GetFldDefs(GetLabels, VisibleFieldOnly, SimpleField, MemoField, PictureField, AllField, DataFieldOnly: boolean): TGetFldDefs;
{ load list of field names or labels from the object according FldsDef }
function GetFldListDBGrid(DBGrid: TCustomDBGrid; List: TStrings; FldsDef: TGetFldDefs): integer;
function GetFldListDataset(Dataset: TDataSet; List: TStrings; FldsDef: TGetFldDefs): integer;
{$IFDEF rUseDBComp}
function GetFldListRecView(RecView: TrDBRecView; List: TStrings; FldsDef: TGetFldDefs): integer;
function GetFldListStringGrid(StringGrid: TrStringGridEd; List: TStrings; FldsDef: TGetFldDefs): integer;
{$ENDIF}
{ load list of field names or labels from the object according FldsDef }
function GetFldList(O: TObject; List: TStrings; FldsDef: TGetFldDefs): integer;
{ load list of field values from the object according FldsDef }
function GetFldListValues(Dataset: TDataSet; const FldList: string; const ListSeparator: string = ';'; const ConstMarks: string = '"'): string;

{ show dialog with number of records }
function ShowRecordCount(GridOrDSet: TObject): boolean;
{ locate record according Field and Value, return if record was found }
function LocateRecord(F: TField; Value: Variant; CaseSensitive: boolean = false): boolean;
{ locate record according Field and text Value, return if record was found }
function LocateRecordText(F: TField; const Value: string): boolean;

{ replace all Tags in Mask text by Dataset field values;
  possible tags:
  <<FieldName>> or <<FieldName:format>>
  <<FieldName:20>> or <<FieldName:L20>> - aling left to 20 chars
  <<FieldName:R20>> - aling right to 20 chars
  <<FieldName:C20>> - aling center to 20 chars
  <<FieldName:T20>> - limit and aling left to 20 chars
  <<FieldName:TR20>> - limit and aling right to 20 chars
  <<FieldName:TC20>> - limit and aling center to 20 chars
  <<#CRLF>> - replace by #13#10 chars
  <<#xx>> - replace by char(xx)
  <<#StrPart(FieldName,From,Count)>>
  }
function FillTextMask(const Mask: string; Dataset: TDataSet; const MaskStrStart: string = '<<'; const MaskStrStop: string = '>>';
  const ReplaceNullValue: string = ''; SkipNonExistingFields: boolean = false): string;

{ add or insert new record and select first visible column }
procedure AddRecordEx(G: TCustomDBGrid; Insert: boolean = false);
{ duplicate current record for DBGrid or Dataset,
  all fields which tag = TagToDuplic is duplicated }
procedure DuplicRecord(GridOrDSet: TObject; TagToDuplic: integer; DuplicBeforeAfterInsert: boolean);

{ return line which starts by defined KeyWord }
function GetSQLLine(SQL: TSQLStrings; const KeyWord: string): string;
{ find and return condition from line which starts by WHERE }
function GetSQLWhere(SQL: TSQLStrings): string;
{ find and return first table name from sql command }
function GetSQLTableName(SQL: TSQLStrings): string;
{ find line which starts by defined KeyWord and replace it by NewLine }
function SetSQLLine(SQL: TSQLStrings; const KeyWord, NewLine: string): boolean;
{ find line which starts by ORDER and replace it by new sorting from FieldName and Desc }
function SetSQLOrder(SQL: TSQLStrings; const FieldName: string; Desc: boolean): boolean;
{ find line which starts by WHERE and replace it by new Condition }
function SetSQLWhere(SQL: TSQLStrings; const Condition: string): boolean; overload;
{ find line which starts by WHERE and replace it by new condition according FormatCondition and Value }
function SetSQLWhere(SQL: TSQLStrings; const FormatCondition: string; Value: Variant): boolean; overload;
{ find line which starts by WHERE and replace it by new condition according Field and its current value }
function SetSQLWhere(SQL: TSQLStrings; Field: TField): boolean; overload;
{ find line which starts by WHERE and replace it by new condition according Field and Value }
function SetSQLWhere(SQL: TSQLStrings; Field: TField; Value: Variant): boolean; overload;
{ find line which starts by WHERE and add another condition with OR }
function AddSQLWhereOR(SQL: TSQLStrings; const Condition: string): boolean;
{ find line which starts by WHERE and add another condition with AND }
function AddSQLWhereAND(SQL: TSQLStrings; const Condition: string): boolean;

{ make sql condition according Field and its current value }
function MakeSQLCond(Field: TField): string; overload;
{ make sql condition according Field and Value }
function MakeSQLCond(Field: TField; Value: Variant): string; overload;
{ make sql condition according according FormatCondition and Value }
function MakeSQLCond(FormatCondition: string; Value: Variant): string; overload;
{ join all conditions by OR }
function MakeSQLCondOR(const Params: array of string): string;
{ join all conditions by AND }
function MakeSQLCondAND(const Params: array of string): string;

{ make sql condition according Field and its current value, null field is converted to 0 or empty string }
function MakeSQLCondIsNull(Field: TField): string; overload;
{ make sql condition according Field and Value, null field is converted to 0 or empty string }
function MakeSQLCondIsNull(Field: TField; Value: Variant): string; overload;
{ make sql LIKE condition according FieldName, Text and CompareType }
function MakeSQLCondLike(const FieldName, Text: string; CompareType: TrLikeCompareType): string;

{ find DBGrid column according FieldName and return Column Title }
function FieldNameToCaption(Grid: TCustomDBGrid; const FieldName: string): string; overload;
{ find Dataset Field according FieldName and return DisplayLabel }
function FieldNameToCaption(Dataset: TDataSet; const FieldName: string): string; overload;
{ for each part from FieldList find DBGrid column according FieldName and return ColumnTitle list }
function FieldListToCaptionList(Grid: TCustomDBGrid; const FieldList, Separator: string): string; overload;
{ for each part from FieldList find Dataset Field according FieldName and return DisplayLabel list }
function FieldListToCaptionList(Dataset: TDataSet; const FieldList, Separator: string): string; overload;

{ read file content and store to the blob field }
procedure FileToBlob(Fld: TField; const FileName: string);
{ read blob field content and store to the file }
procedure BlobToFile(Fld: TField; const FileName: string);

type
  { callback procedure used by ScanAllRecords }
  TScanCallBackProc = procedure(Dataset: TDataSet; RecIndex: integer; var Abort: boolean) of object;
  { scan all record in Dataset and call CallBack procedure for each one, reset all events before scaning }
function ScanAllRecords(Dataset: TDataSet; CallBack: TScanCallBackProc; DisableControls: boolean = true; BlockEvents: boolean = true;
  UseBookmark: boolean = true): boolean;

{ scan all dataset and load captions to List, List.Object contain IdField integer value converted to Object }
function LoadRecordList(CaptionField, IdField: TField; L: TStrings): integer;

{ generate INSERT or UPDATE sql script for active or all record with defined list of fields }
procedure GenerateSQLScript(SqlCmd: TSQLStrings; DS: TDataSet; const TableName: string; KeyFieldList, FieldList: TStrings; GenerateInsert, AllRecords: boolean;
  AddAfterEachRecord: string = '');
{ generate and save INSERT or UPDATE sql script for active or all record with defined list of fields }
procedure GenerateSaveSQLScript(SqlCmd: TSQLStrings; DS: TDataSet; const TableName: string; KeyFieldList, FieldList: TStrings;
  GenerateInsert, AllRecords: boolean; FileName: string);

{ return TableName with quotes if needed }
function GetQuotedTableName(TableName: string): string;
{ return FieldName with quotes if needed }
function GetQuotedFieldName(FieldName: string): string;

// Dataset Filter tools (moved from rDBFilter)
{ Get filter WildCard char regarding used dataset type and global setting (% or *) }
function GetFilterWildCardChar(Dataset: TDataSet; SQLFormat: boolean = false): Char;
{ Get filter LIKE operator string regarding used dataset type and global setting (LIKE or any other value) }
function GetFilterLikeOperator(Dataset: TDataSet; SQLFormat: boolean = false): string;
{ Get filter equal string for null field regarding used dataset type and global setting (IS or =) }
function GetFilterNullEqualStr(Dataset: TDataSet): string;
{ set Filter string for Dataset, try to keep current record if possible }
function SetFilterString(Dataset: TDataSet; const Filter: string; ShowError: boolean): boolean;
{ clear filter string for dataset or grid, try to keep current record if possible }
procedure ClearFilter(const GridOrDataSet: TObject);
{ make Filter condition according Field and Value }
function GetFilterCond(Field: TField; const Value: string; IsNull: boolean; Condition: string = ''): string;
{ make Filter condition according Field and its current value }
function FilterByField(Field: TField; FilterJoin: TFilterJoin = fjNone): boolean;
{ make Filter condition according DBGrid selected cell (selected Field and its current value) }
function FilterBySel(const Grid: TCustomDBGrid; FilterJoin: TFilterJoin = fjNone): boolean;
{ negate active filter }
function NegateFilter(const GridOrDataSet: TObject): boolean;
{ return Numeric Field value string prepared for filter }
function GetNumericFieldFilterValue(Value: extended): string;
{ return Date Field value string prepared for filter }
function GetDateFieldFilterValue(Value: TDateTime): string;
{ make filter condition }
function MakeFilterCond(FldName, Cond, Value: string; AddSep: string = ''): string;
{ make filter condition }
function MakeFilterBetweenCond(FldName, ValueFrom, ValueTo: string; AddSep: string = ''): string;

implementation

uses rstring, rdlg, rtool;

type
  THackCustomDBGrid = class(TCustomDBGrid)
  end;


function GetSearchHintText(SearchText, NewKey: string; SearchingType: TSearchingType; Found, AllowFilter, Filtered: boolean): string;
var
  s: string;
begin
  Result:=SearchText;
  case SearchingType of
    stFindFirst:
      ;
    stFindNext:
      Result:=Result + ' >>';
    stFindPrev:
      Result:='<< ' + Result;
  end;
  s:='';
  if not Found then
  begin
    case SearchingType of
      stFindFirst:
        begin
          if NewKey = #0 then
            NewKey:=' ';
          s:=Format(' (%s %s)', [SearchText + NewKey, LowerCase(sNotFound)]);
        end;
      stFindNext:
        s:=Format(' (%s)', [LowerCase(sNotFound)]);
      stFindPrev:
        s:=Format(' (%s)', [LowerCase(sNotFound)]);
    end;
  end;
  Result:=Result + s;
  if AllowFilter and (SearchText <> '') then
  begin
    if Filtered then
      s:=Format('* %s, %s', [sFiltered, sFilterResetHint])
    else
      s:=Format('* %s', [sFilterSetHint]);
    Result:=Result + #13#10 + s;
  end;
end;

function GetDataSet(GridOrDSet: TObject): TDataSet;
begin
  Result:=nil;
  try
    if (GridOrDSet is TCustomDBGrid) and Assigned(TCustomDBGrid(GridOrDSet).DataSource) then
      Result:=TCustomDBGrid(GridOrDSet).DataSource.Dataset;
{$IFDEF rUseDBComp}
    if (GridOrDSet is TrDBRecView) and Assigned(TrDBRecView(GridOrDSet).DataSource) then
      Result:=TrDBRecView(GridOrDSet).DataSource.Dataset;
{$ENDIF}
    if GridOrDSet is TDataSource then
      Result:=TDataSource(GridOrDSet).Dataset;
    if GridOrDSet is TDataSet then
      Result:=GridOrDSet as TDataSet;
  except
    Result:=nil;
  end;
end;

function GetDSStateStr(GridOrDSet: TObject): string;
var
  DS: TDataSet;
begin
  Result:='';
  DS:=GetDataSet(GridOrDSet);
  if Assigned(DS) then
    try
      case DS.State of
        dsInsert:
          Result:=sInserting;
        dsEdit:
          Result:=sEditing;
        dsBrowse:
          Result:=sBrowsing;
        dsInActive:
          Result:=sInactive;
      else
        Result:='';
      end;
    except
      Result:='';
    end;
end;

function IsClientDataset(Dataset: TDataSet): boolean;
begin
  Result:=SameText(Dataset.ClassName, 'TClientDataset');
end;

function IsAdvantageDatabaseDataset(Dataset: TDataSet): boolean;
begin
  Result:=SameText(Dataset.ClassName, 'TAdsTable') or SameText(Dataset.ClassName, 'TAdsQuery') or SameText(Dataset.ClassName, 'TAdsStoredProcedure');
end;

procedure AssignGridProps(Source, Dest: TCustomDBGrid);
begin
  THackCustomDBGrid(Dest).Color:=THackCustomDBGrid(Source).Color;
  THackCustomDBGrid(Dest).Font.Assign(THackCustomDBGrid(Source).Font);
  THackCustomDBGrid(Dest).TitleFont.Assign(THackCustomDBGrid(Source).TitleFont);
  THackCustomDBGrid(Dest).DataSource:=THackCustomDBGrid(Source).DataSource;
  THackCustomDBGrid(Dest).Columns.Assign(THackCustomDBGrid(Source).Columns);
  // set internal color (by default set in Paint, but is not called if visible=false
  THackCustomDBGrid(Dest).FInternalColor:=THackCustomDBGrid(Dest).Color;
end;

function IsEditing(GridOrDSet: TObject): boolean;
var
  DS: TDataSet;
begin
  Result:=false;
  DS:=GetDataSet(GridOrDSet);
  if Assigned(DS) then
    Result:=(DS.State in [dsEdit, dsInsert]);
end;

function IsDataField(Fld: TField): boolean;
begin
  if Fld = nil then
    Result:=false
  else
    Result:=Fld.FieldKind = fkData;
end;

function IsSimpleField(Fld: TField): boolean;
begin
  if Fld = nil then
    Result:=false
  else
    Result:=Fld.DataType in sfSimpleFld;
end;

function IsApostrField(Fld: TField): boolean;
begin
  if Fld = nil then
    Result:=false
  else
    Result:=Fld.DataType in sfApostrFld;
end;

function IsMemoField(Fld: TField): boolean;
var
  bt: set of TBlobType;
begin
  if RosiCompConfig.rDBToolCfg.BlobField_ftBlobAsMemo then
    bt:=[ftBlob]
  else
    bt:=[];
{$IFDEF D2006P}
  Result:=(Fld is TMemoField) or (Fld is TWideMemoField);
  if Fld is TBlobField then
    if (TBlobField(Fld).BlobType in (bt + [ftMemo, ftFmtMemo, ftWideMemo])) then
      Result:=true;
{$ELSE}
  Result:=Fld is TMemoField;
  if Fld is TBlobField then
    if (TBlobField(Fld).BlobType in (bt + [ftMemo, ftFmtMemo])) then
      Result:=true;
{$ENDIF}
end;

function IsGraphicField(Fld: TField): boolean;
begin
  Result:=Fld is TGraphicField;
  if (Fld is TBlobField) and not IsMemoField(Fld) then
    if (TBlobField(Fld).BlobType in [ftTypedBinary, ftGraphic]) or (TBlobField(Fld).GraphicHeader) then
      Result:=true;
end;

function IsDateTimeField(Fld: TField): boolean;
begin
  Result:=(Fld is TDateTimeField) or (Fld is TSQLTimeStampField);
end;

function IsDateField(Fld: TField): boolean;
begin
  Result:=IsDateTimeField(Fld) and not IsTimeField(Fld);
end;

function IsTimeField(Fld: TField): boolean;
begin
  Result:=(Fld is TTimeField);
end;

function GetBooleanFieldTextValue(Fld: TField; State: boolean): string;
var
  p: integer;
  dv: string;
begin
  Result:='';
  if Fld is TBooleanField then
  begin
    dv:=TBooleanField(Fld).DisplayValues;
    if dv = '' then
    begin
      if State then
        Result:=STextTrue
      else
        Result:=STextFalse;
    end
    else
    begin
      p:=Pos(';', dv);
      if p = 0 then
        p:=256;
      if State then
        Result:=Copy(dv, 1, p - 1)
      else
        Result:=Copy(dv, p + 1, 255);
    end;
  end;
end;

function GetBooleanFieldState(Fld: TField; TextValue: string): boolean;
begin
  Result:=SameText(GetBooleanFieldTextValue(Fld, true), TextValue);
end;

function GetFldDefs(GetLabels, VisibleFieldOnly, SimpleField, MemoField, PictureField, AllField, DataFieldOnly: boolean): TGetFldDefs;
begin
  Result:=[];
  if GetLabels then
    Include(Result, fdGetLabels);
  if VisibleFieldOnly then
    Include(Result, fdVisibleFieldOnly);
  if SimpleField then
    Include(Result, fdSimpleField);
  if MemoField then
    Include(Result, fdMemoField);
  if PictureField then
    Include(Result, fdGraphicField);
  if AllField then
    Include(Result, fdAllField);
  if DataFieldOnly then
    Include(Result, fdDataFieldOnly);
end;

function GetFldListDBGrid(DBGrid: TCustomDBGrid; List: TStrings; FldsDef: TGetFldDefs): integer;
var
  i: integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    with THackCustomDBGrid(DBGrid) do
      for i:=0 to Columns.Count - 1 do
        if ((not(fdVisibleFieldOnly in FldsDef)) or (Columns[i].Visible)) and (not(fdDataFieldOnly in FldsDef) or IsDataField(Columns[i].Field)) and
          ((fdAllField in FldsDef) or ((fdSimpleField in FldsDef) and IsSimpleField(Columns[i].Field)) or
            ((fdMemoField in FldsDef) and IsMemoField(Columns[i].Field)) or ((fdGraphicField in FldsDef) and IsGraphicField(Columns[i].Field))) then
          if fdGetLabels in FldsDef then
            List.Add(Columns[i].Title.Caption)
          else
            List.Add(Columns[i].FieldName);
    Result:=List.Count;
  finally
    List.EndUpdate;
  end;
end;

function GetFldListDataset(Dataset: TDataSet; List: TStrings; FldsDef: TGetFldDefs): integer;
var
  i: integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    with Dataset do
      for i:=0 to FieldCount - 1 do
        if ((not(fdVisibleFieldOnly in FldsDef)) or (Fields[i].Visible)) and (not(fdDataFieldOnly in FldsDef) or IsDataField(Fields[i])) and
          ((fdAllField in FldsDef) or ((fdSimpleField in FldsDef) and IsSimpleField(Fields[i])) or ((fdMemoField in FldsDef) and IsMemoField(Fields[i])) or
            ((fdGraphicField in FldsDef) and IsGraphicField(Fields[i]))) then
          if fdGetLabels in FldsDef then
            List.Add(Fields[i].DisplayLabel)
          else
            List.Add(Fields[i].FieldName);
    Result:=List.Count;
  finally
    List.EndUpdate;
  end;
end;
{$IFDEF rUseDBComp}

function GetFldListRecView(RecView: TrDBRecView; List: TStrings; FldsDef: TGetFldDefs): integer;
var
  i: integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    with RecView do
      for i:=0 to GetFieldCount - 1 do
        if (not(fdDataFieldOnly in FldsDef) or IsDataField(GetField(i))) and
          ((fdAllField in FldsDef) or ((fdSimpleField in FldsDef) and IsSimpleField(GetField(i))) or ((fdMemoField in FldsDef) and IsMemoField(GetField(i))) or
            ((fdGraphicField in FldsDef) and IsGraphicField(GetField(i)))) then
          if fdGetLabels in FldsDef then
            List.Add(GetCaption(i))
          else
            List.Add(GetField(i).FieldName);
    Result:=List.Count;
  finally
    List.EndUpdate;
  end;
end;

function GetFldListStringGrid(StringGrid: TrStringGridEd; List: TStrings; FldsDef: TGetFldDefs): integer;
var
  i: integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    with StringGrid do
      for i:=0 to Columns.Count - 1 do
        if ((not(fdVisibleFieldOnly in FldsDef)) or (Columns[i].Visible)) then
          if fdGetLabels in FldsDef then
            List.Add(Columns[i].TitleCaption)
          else
            List.Add(IntToStr(i));
    Result:=List.Count;
  finally
    List.EndUpdate;
  end;
end;
{$ENDIF}

function GetFldList(O: TObject; List: TStrings; FldsDef: TGetFldDefs): integer;
begin
  Result:=0;
  if O is TCustomDBGrid then
    Result:=GetFldListDBGrid(TCustomDBGrid(O), List, FldsDef);
  if O is TDataSet then
    Result:=GetFldListDataset(TDataSet(O), List, FldsDef);
{$IFDEF rUseDBComp}
  if O is TrDBRecView then
    Result:=GetFldListRecView(TrDBRecView(O), List, FldsDef);
  if O is TrStringGridEd then
    Result:=GetFldListStringGrid(TrStringGridEd(O), List, FldsDef);
{$ENDIF}
end;

function GetFldListValues(Dataset: TDataSet; const FldList: string; const ListSeparator: string = ';'; const ConstMarks: string = '"'): string;
var
  i: integer;
  s, v: string;
begin
  Result:='';
  i:=1;
  repeat
    s:=rtool.FindStrPart(FldList, ListSeparator, i, true);
    if s = '' then
      Break;
    s:=Trim(s);
    if (Copy(s, 1, Length(ConstMarks)) = ConstMarks) and (Copy(s, Length(s), Length(ConstMarks)) = ConstMarks) then
      v:=Copy(s, 1 + Length(ConstMarks), Length(s) - 2 * Length(ConstMarks))
    else
      v:=Dataset.FieldByName(s).Text;
    Result:=Result + v;
    Inc(i);
  until false;
end;

function ShowRecordCount(GridOrDSet: TObject): boolean;
var
  DS: TDataSet;
begin
  Result:=false;
  DS:=GetDataSet(GridOrDSet);
  if Assigned(DS) then
  begin
    DlgI(Format(sSelRecords, [DS.RecordCount]));
    Result:=true;
  end;
end;

function LocateRecord(F: TField; Value: Variant; CaseSensitive: boolean = false): boolean;

  function CheckFieldValue: boolean;
  begin
    if F is TStringField then
    begin
      if CaseSensitive then
        Result:=F.AsString = Value
      else
        Result:=SameText(F.AsString, Value);
    end
    else
      Result:=F.AsVariant = Value;
  end;

var
  opt: TLocateOptions;
begin
  Result:=false;
  if Assigned(F) and not VarIsNull(Value) then
  begin
    if F.Dataset.Filtered and (Pos('ADO', F.Dataset.ClassName) > 0) then
    begin
      // BUG for ADO - Locate on Filtered dataset not working properly
      F.Dataset.First;
      while not F.Dataset.Eof do
      begin
        if CheckFieldValue then
          Break;
        F.Dataset.Next;
      end;
    end
    else
    begin
      if CaseSensitive then
        opt:=[]
      else
        opt:=[loCaseInsensitive];
      F.Dataset.Locate(F.FieldName, Value, opt);
    end;
    Result:=CheckFieldValue;
  end;
end;

function LocateRecordText(F: TField; const Value: string): boolean;
begin
  Result:=LocateRecord(F, Value, false);
end;

function FillTextMask(const Mask: string; Dataset: TDataSet; const MaskStrStart: string = '<<'; const MaskStrStop: string = '>>';
  const ReplaceNullValue: string = ''; SkipNonExistingFields: boolean = false): string;

  function GetFieldString(s: string): string;
  var
    F: TField;
    FldName, Frm, Align, v: string;
    w: integer;
    limit: boolean;
  begin
    // extract field and format
    FldName:=rtool.FindStrPart(s, ':', 1);
    Frm:=rtool.FindStrPart(s, ':', 2);
    // find field
    F:=Dataset.FindField(FldName);
    if Assigned(F) then
    begin
      if F.IsNull and (ReplaceNullValue <> '') then
        v:=ReplaceNullValue
      else
        if IsMemoField(F) then
          v:=F.AsString
        else
          v:=F.DisplayText;
      // use format
      if Frm <> '' then
      begin
        // first T means limit to width
        limit:=SameText(Copy(Frm, 1, 1), 'T');
        if limit then
          Frm:=Copy(Frm, 2, Length(Frm));
        Align:='';
        if Copy(Frm, 1, 1) > '9' then
        begin
          Align:=UpperCase(Copy(Frm, 1, 1));
          Frm:=Copy(Frm, 2, Length(Frm));
        end;
        w:=StrToIntDef(Frm, 0);
        if w > 0 then
        begin
          if limit then
            v:=Copy(v, 1, w);
          if Align = 'R' then
            v:=StringOfChar(' ', w - Length(v)) + v
          else
            if Align = 'C' then
            begin
              v:=StringOfChar(' ', (w - Length(v)) div 2) + v;
              v:=v + StringOfChar(' ', w - Length(v));
            end
            else
              v:=v + StringOfChar(' ', w - Length(v));
        end;
      end;
    end
    else
      if SkipNonExistingFields then
        v:='<??<' + FldName + '>??>'
      else
        v:='?' + FldName + '?';
    Result:=v;
  end;

var
  Tag, fn, res, p: string;
  p1, p2: integer;
begin
  Result:='';
  if Assigned(Dataset) and Dataset.Active then
  begin
    Result:=Mask;
    repeat
      { replace all <<FieldName>> mask by field value
        possible format <<FieldName>>
        <<FieldName:20>> or <<FieldName:L20>> - aling left to 20 chars
        <<FieldName:R20>> - aling right to 20 chars
        <<FieldName:C20>> - aling center to 20 chars
        <<FieldName:T20>> - limit and aling left to 20 chars
        <<FieldName:TR20>> - limit and aling right to 20 chars
        <<FieldName:TC20>> - limit and aling center to 20 chars
        <<#CRLF>> - replace by #13#10 chars
        <<#xx>> - replace by char(xx)
        <<#StrPart(FieldName,From,Count)>>
        }
      Tag:=rtool.FindStrFromTo(MaskStrStart, MaskStrStop, Result, false, true);
      if Tag = '' then
        Break;
      res:='';
      if Copy(Tag, 1, 1) = '#' then
      begin
        fn:=Copy(Tag, 2, Length(Tag));
        p1:=StrToIntDef(fn, 0);
        if p1 > 0 then
          res:=chr(p1)
        else
          if SameText(fn, 'CRLF') then
            res:=cStrEOL
          else
            if AnsiPosEx('StrPart(', fn) = 1 then
            begin
              // tag from brackets
              fn:=rtool.FindStrFromTo('(', ')', fn, true, true);
              // 1st param - field
              res:=GetFieldString(rtool.FindStrPart(fn, ',', 1));
              // 2nd - from index or substr
              p:=rtool.FindStrPart(fn, ',', 2);
              p1:=StrToIntDef(p, 0);
              if p1 = 0 then
                p1:=AnsiPosEx(p, res);
              // 3rd - length or substr
              p:=rtool.FindStrPart(fn, ',', 3);
              p2:=StrToIntDef(p, 0);
              if p2 = 0 then
                p2:=AnsiPosEx(p, res, p1) - p1;
              if (p1 > 0) and (p2 > 0) then
                res:=Copy(res, p1, p2);
            end
            else
              res:='?' + Tag + '?';
      end
      else
      begin
        res:=GetFieldString(Tag);
      end;
      // replace tag by value
      Result:=rtool.ReplaceStr(MaskStrStart + Tag + MaskStrStop, res, Result, true, false);
    until false;
    if SkipNonExistingFields then
    begin
      Result:=rtool.ReplaceStr('<??<', '<<', Result);
      Result:=rtool.ReplaceStr('>??>', '>>', Result);
    end;
  end;
end;

procedure AddRecordEx(G: TCustomDBGrid; Insert: boolean = false);
begin
  if G = nil then
    Exit;
  G.SelectedIndex:=0;
  while (not THackCustomDBGrid(G).Columns[G.SelectedIndex].Visible) and (G.SelectedIndex < THackCustomDBGrid(G).Columns.Count - 1) do
    G.SelectedIndex:=G.SelectedIndex + 1;
  if Insert then
    G.DataSource.Dataset.Insert
  else
    G.DataSource.Dataset.Append;
end;

procedure DuplicRecord(GridOrDSet: TObject; TagToDuplic: integer; DuplicBeforeAfterInsert: boolean);
var
  i, max: integer;
  arr: array of Variant;
  D: TDataSet;
  oldev: TDataSetNotifyEvent;
begin
  D:=GetDataSet(GridOrDSet);
  if (D = nil) or (not D.Active) then
    Exit;
  oldev:=D.AfterInsert;
  if DuplicBeforeAfterInsert then
    D.AfterInsert:=nil;
  try
    max:=D.FieldCount;
    SetLength(arr, max);
    for i:=0 to max - 1 do
      arr[i]:=D.Fields[i].AsVariant;
    if not D.Eof then
      D.Next;
    if GridOrDSet is TCustomDBGrid then
    begin
      AddRecordEx(GridOrDSet as TCustomDBGrid, not D.Eof)
    end
    else
      if D.Eof then
        D.Append
      else
        D.Insert;
    for i:=0 to max - 1 do
      try
        if (D.Fields[i].Tag and TagToDuplic) <> 0 then
          D.Fields[i].AsVariant:=arr[i];
      except
      end;
  finally
    D.AfterInsert:=oldev;
  end;
  if DuplicBeforeAfterInsert and Assigned(oldev) then
    oldev(D);
end;

function GetSQLLine(SQL: TSQLStrings; const KeyWord: string): string;
var
  a: integer;
  s: string;
begin
  Result:='';
  for a:=0 to SQL.Count - 1 do
  begin
    s:=AnsiUpperCase(Trim(SQL.Strings[a]));
    if Pos(AnsiUpperCase(KeyWord), s) = 1 then
    begin
      Result:=SQL.Strings[a];
    end;
  end;
end;

function GetSQLWhere(SQL: TSQLStrings): string;
const
  w = 'WHERE ';
begin
  Result:=GetSQLLine(SQL, w);
  Result:=Copy(Result, Length(w) + 1, Length(Result));
end;

function GetSQLTableName(SQL: TSQLStrings): string;
var
  s: string;
  p: integer;
begin
  Result:='';
  p:=rtool.AnsiPosEx('FROM ', SQL.Text, 1, false);
  if p = 0 then
    Exit;
  s:=Copy(SQL.Text, p + 5, 255);
  s:=Trim(s);
  p:=AnsiPos(' ', s);
  if p > 0 then
    s:=Copy(s, 1, p - 1);
  p:=AnsiPos(cStrEOL, s);
  if p > 0 then
    s:=Copy(s, 1, p - 1);
  Result:=s;
end;

function SetSQLLine(SQL: TSQLStrings; const KeyWord, NewLine: string): boolean;
var
  a: integer;
  s: string;
begin
  Result:=false;
  for a:=0 to SQL.Count - 1 do
  begin
    s:=AnsiUpperCase(Trim(SQL.Strings[a]));
    if Pos(AnsiUpperCase(KeyWord), s) = 1 then
    begin
      SQL.Strings[a]:=NewLine;
      Result:=true;
    end;
  end;
end;

function SetSQLOrder(SQL: TSQLStrings; const FieldName: string; Desc: boolean): boolean;
const
  O = 'ORDER BY ';
var
  s: string;
begin
  s:=O + FieldName;
  if Desc then
    s:=s + ' DESC';
  Result:=SetSQLLine(SQL, O, s);
end;

function SetSQLWhere(SQL: TSQLStrings; const Condition: string): boolean;
const
  w = 'WHERE ';
begin
  Result:=SetSQLLine(SQL, w, w + '(' + Condition + ')');
end;

function SetSQLWhere(SQL: TSQLStrings; const FormatCondition: string; Value: Variant): boolean;
begin
  Result:=SetSQLWhere(SQL, MakeSQLCond(FormatCondition, Value));
end;

function SetSQLWhere(SQL: TSQLStrings; Field: TField): boolean;
begin
  Result:=SetSQLWhere(SQL, MakeSQLCond(Field));
end;

function SetSQLWhere(SQL: TSQLStrings; Field: TField; Value: Variant): boolean; overload;
begin
  Result:=SetSQLWhere(SQL, MakeSQLCond(Field, Value));
end;

function AddSQLWhereOR(SQL: TSQLStrings; const Condition: string): boolean;
begin
  Result:=SetSQLWhere(SQL, MakeSQLCondOR([GetSQLWhere(SQL), Condition]));
end;

function AddSQLWhereAND(SQL: TSQLStrings; const Condition: string): boolean;
begin
  Result:=SetSQLWhere(SQL, MakeSQLCondAND([GetSQLWhere(SQL), Condition]));
end;

function MakeSQLCond(Field: TField): string;
begin
  if Field is TBooleanField then
  begin
    if TBooleanField(Field).AsBoolean then
      Result:=MakeSQLCond(Field, RosiCompConfig.rDBToolCfg.SQLScript_BooleanTrueValue)
    else
      Result:=MakeSQLCond(Field, RosiCompConfig.rDBToolCfg.SQLScript_BooleanFalseValue);
  end
  else
    Result:=MakeSQLCond(Field, Field.AsVariant);
end;

function MakeSQLCond(Field: TField; Value: Variant): string;
var
  s: string;
begin
  s:=VarToStrDef(Value, 'null');
  if IsApostrField(Field) and not VarIsNull(Value) then
    s:=RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes + s + RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes
  else
    if s = '' then
      s:='0';
  Result:=Field.FieldName + '=' + s;
end;

function MakeSQLCond(FormatCondition: string; Value: Variant): string;
var
  w: string;
begin
  if Pos('%d', FormatCondition) > 0 then
    w:=Format(FormatCondition, [integer(Value)])
  else
    if Pos('%s', FormatCondition) > 0 then
      w:=Format(FormatCondition, [string(Value)])
    else
      w:=Format(FormatCondition, [string(Value)]);
  Result:=w;
end;

function MakeSQLCondInt(LogOperator: string; Params: array of string): string;
var
  i: integer;
  s: string;
begin
  Result:='';
  for i:=0 to High(Params) do
  begin
    s:=Params[i];
    if s = '' then
      Continue;
    if Result <> '' then
      Result:=Result + ' ' + LogOperator + ' ';
    Result:=Result + '(' + s + ')';
  end;
end;

function MakeSQLCondIsNull(Field: TField): string; overload;
begin
  Result:=MakeSQLCondIsNull(Field, Field.AsVariant);
end;

function MakeSQLCondIsNull(Field: TField; Value: Variant): string; overload;
var
  F, s: string;
begin
  if IsApostrField(Field) then
    s:=''''''
  else
    s:='0';
  F:=Format('IsNull(%s,%s)', [Field.FieldName, s]);

  // if VarIsNull(Value) then Result:=Field.FieldName+' IS NULL'
  // else
  begin
    s:=VarToStr(Value);
    if IsApostrField(Field) then
      s:=RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes + s + RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes
    else
      if s = '' then
        s:='0';
    Result:=F + '=' + s;
  end;
end;

function MakeSQLCondLike(const FieldName, Text: string; CompareType: TrLikeCompareType): string;
const
  sqlUnivLike = '(%s %s ''%s'')';
var
  sText, sWord, LikeCh, LikeOper: string;
begin
  Result:='';
  sText:=ReplaceStr('''', '''''', Text);
  LikeCh:=GetFilterWildCardChar(nil, true);
  LikeOper:=GetFilterLikeOperator(nil, true);
  case CompareType of
    lcAnyPos:
      Result:=Format(sqlUnivLike, [FieldName, LikeOper, LikeCh + sText + LikeCh]);
    lcBegin:
      Result:=Format(sqlUnivLike, [FieldName, LikeOper, sText + LikeCh]);
    lcSame:
      Result:=Format(sqlUnivLike, [FieldName, LikeOper, sText]);
    lcAllWords, lcAnyWords:
      begin
        repeat
          sWord:=RemoveStrPart(sText, ' ', true);
          if sWord <> '' then
          begin
            if CompareType = lcAllWords then
              Result:=MakeSQLCondAND([Result, Format(sqlUnivLike, [FieldName, LikeOper, LikeCh + sWord + LikeCh])])
            else
              Result:=MakeSQLCondOR([Result, Format(sqlUnivLike, [FieldName, LikeOper, LikeCh + sWord + LikeCh])]);
          end;
          if sText = '' then
            Break;
        until false;
      end;
  end;
end;

function MakeSQLCondOR(const Params: array of string): string;
begin
  Result:=MakeSQLCondInt('OR', Params);
end;

function MakeSQLCondAND(const Params: array of string): string;
begin
  Result:=MakeSQLCondInt('AND', Params);
end;

function FieldNameToCaption(Grid: TCustomDBGrid; const FieldName: string): string;
var
  i: integer;
begin
  Result:='';
  with THackCustomDBGrid(Grid) do
    for i:=0 to Columns.Count - 1 do
      if SameText(Columns[i].FieldName, FieldName) then
      begin
        Result:=Columns[i].Title.Caption;
        Break;
      end;
end;

function FieldNameToCaption(Dataset: TDataSet; const FieldName: string): string;
var
  i: integer;
begin
  Result:='';
  for i:=0 to Dataset.Fields.Count - 1 do
    if SameText(Dataset.Fields[i].FieldName, FieldName) then
    begin
      Result:=Dataset.Fields[i].DisplayLabel;
      Break;
    end;
end;

function FieldListToCaptionList(Grid: TCustomDBGrid; const FieldList, Separator: string): string; overload;
var
  p: integer;
  s: string;
begin
  p:=1;
  Result:='';
  repeat
    s:=rtool.FindStrPart(FieldList, Separator, p, true);
    if s = '' then
      Break;
    if Result <> '' then
      Result:=Result + Separator;
    Result:=Result + Trim(FieldNameToCaption(Grid, s));
    Inc(p);
  until false;
end;

function FieldListToCaptionList(Dataset: TDataSet; const FieldList, Separator: string): string; overload;
var
  p: integer;
  s: string;
begin
  p:=1;
  Result:='';
  repeat
    s:=rtool.FindStrPart(FieldList, ';', p, true);
    if s = '' then
      Break;
    if Result <> '' then
      Result:=Result + Separator;
    Result:=Result + Trim(FieldNameToCaption(Dataset, s));
    Inc(p);
  until false;
end;

function ScanAllRecords(Dataset: TDataSet; CallBack: TScanCallBackProc; DisableControls: boolean = true; BlockEvents: boolean = true;
  UseBookmark: boolean = true): boolean;
var
  Bk: Tbookmark;
  oldAfterScroll, oldBeforeScroll: TDataSetNotifyEvent;
  r: integer;
  Abort: boolean;
  oldCursor: TCursor;
begin
  Result:=true;
  if not(Assigned(Dataset) and Dataset.Active) then
    Exit;
  oldCursor:=WCursor;
  oldAfterScroll:=nil;
  oldBeforeScroll:=nil;
  Bk:=Dataset.GetBookmark;
  try
    r:=0;
    Abort:=false;
    if DisableControls then
      Dataset.DisableControls;
    if BlockEvents then
    begin
      oldBeforeScroll:=Dataset.BeforeScroll;
      oldAfterScroll:=Dataset.AfterScroll;
      Dataset.BeforeScroll:=nil;
      Dataset.AfterScroll:=nil;
    end;
    try
      Dataset.First;
      while not Dataset.Eof do
      begin
        Inc(r);
        CallBack(Dataset, r, Abort);
        if Abort then
          Break;
        Dataset.Next;
      end;
      if UseBookmark then
        Dataset.GotoBookmark(Bk);
    finally
      if BlockEvents then
      begin
        Dataset.BeforeScroll:=oldBeforeScroll;
        Dataset.AfterScroll:=oldAfterScroll;
      end;
      if DisableControls then
        Dataset.EnableControls;
    end;
  finally
    Dataset.FreeBookmark(Bk);
    ChCursor(oldCursor);
    Result:=not Abort;
  end;
end;

type
  TScanObj = class(TObject)
    L: TStrings;
    CaptionFld, IdFld: TField;
    procedure ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
  end;

procedure TScanObj.ScanCallBack(Dataset: TDataSet; RecIndex: integer; var Abort: boolean);
begin
  if Assigned(IdFld) then
    L.AddObject(CaptionFld.AsString, TObject(IdFld.AsInteger))
  else
    L.AddObject(CaptionFld.AsString, TObject(-1));
end;

function LoadRecordList(CaptionField, IdField: TField; L: TStrings): integer;
var
  s: TScanObj;
begin
  L.Clear;
  s:=TScanObj.Create;
  try
    s.L:=L;
    s.CaptionFld:=CaptionField;
    s.IdFld:=IdField;
    ScanAllRecords(CaptionField.Dataset, s.ScanCallBack);
  finally
    FreeAndNil(s);
  end;
  Result:=L.Count;
end;

procedure GenerateSQLScript(SqlCmd: TSQLStrings; DS: TDataSet; const TableName: string; KeyFieldList, FieldList: TStrings; GenerateInsert, AllRecords: boolean;
  AddAfterEachRecord: string);

  procedure Add1Rec;
  var
    F: TField;
    First, IsKey: boolean;
    i: integer;
    cmd, tbl, wh, fldssum, flds, valssum, vals: string;

    procedure AddSep;
    begin
      if First then
        First:=false
      else
      begin
        flds:=flds + ', ';
        vals:=vals + ', ';
      end;
    end;

    function FrmFieldValue(F: TField; AddFieldName: boolean): string;
    var
      v: string;
    begin
      if F.IsNull then
        v:='null'
      else
      begin
        if F is TNumericField then
        begin
          v:=GetNumericFieldFilterValue(F.AsFloat);
        end
        else
          if F is TBooleanField then
          begin
            if F.AsBoolean then
              v:=RosiCompConfig.rDBToolCfg.SQLScript_BooleanTrueValue
            else
              v:=RosiCompConfig.rDBToolCfg.SQLScript_BooleanFalseValue;
          end
          else
            if IsDateTimeField(F) then
            begin
              v:=FormatDateTime('yyyy-mm-dd hh:nn:ss', F.AsDateTime);
              v:=Format('CONVERT(DATETIME, ''%s'', 102)', [v]);
            end
            else
            begin
              v:=F.AsString;
              if IsApostrField(F) then
              begin
                // double each appostroph
                v:=rtool.ReplaceStr('''', '''''', v, true, true);
                v:=RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes + v + RosiCompConfig.rDBToolCfg.SQLScript_StringFieldQuotes;
                v:=rtool.ReplaceStr(cStrEOL, ''' + char(13) + char(10) + ''', v, true, true);
              end;
            end;
      end;
      if AddFieldName then
        v:=rdbtool.GetQuotedFieldName(F.FieldName) + '=' + v;
      Result:=v;
    end;

  begin
    First:=true;
    fldssum:='';
    valssum:='';
    flds:='';
    vals:='';
    wh:='';
    for i:=0 to DS.FieldCount - 1 do
    begin
      F:=DS.Fields[i];
      if not((F.FieldKind = fkData) and (IsSimpleField(F) or IsMemoField(F)) and F.Visible) then
        Continue;

      IsKey:=Assigned(KeyFieldList) and (KeyFieldList.IndexOf(F.FieldName) >= 0);
      if not IsKey and Assigned(FieldList) and (FieldList.IndexOf(F.FieldName) < 0) then
        Continue;

      if GenerateInsert then
      begin
        if not IsKey then
        begin
          AddSep;
          flds:=flds + GetQuotedFieldName(F.FieldName);
          vals:=vals + FrmFieldValue(F, false);
        end;
      end
      else
      begin
        if IsKey then
        begin
          if wh <> '' then
            wh:=wh + ' AND ';
          wh:=wh + FrmFieldValue(F, true);
        end
        else
        begin
          AddSep;
          flds:=flds + FrmFieldValue(F, true);
        end;
        vals:='';
      end;
      if Length(flds) > 150 then
      begin
        fldssum:=fldssum + flds + cStrEOL + '  ';
        flds:='';
      end;
      if Length(vals) > 150 then
      begin
        valssum:=valssum + vals + cStrEOL + '  ';
        vals:='';
      end;
    end;
    fldssum:=fldssum + flds;
    valssum:=valssum + vals;

    if GenerateInsert then
      cmd:='INSERT INTO %s (%s) ' + cStrEOL + 'VALUES (%s)'
    else
      cmd:='UPDATE %s SET %s ' + cStrEOL + 'WHERE %s%s';

    tbl:=GetQuotedTableName(TableName);
    cmd:=Format(cmd, [tbl, fldssum, valssum, wh]);
    // if AddAfterEachRecord<>'' then
    cmd:=cmd + AddAfterEachRecord;
    SqlCmd.Add(cmd);
    // SqlCmd.Add(rDBToolSqlCommandSeparator);
  end;

var
  oldCursor: TCursor;
begin
  if AllRecords then
  begin
    DS.DisableControls;
    SqlCmd.BeginUpdate;
    oldCursor:=Wcursor;
    try
      DS.First;
      while not DS.Eof do
      begin
        Add1Rec;
        DS.Next;
      end;
    finally
      DS.EnableControls;
      SqlCmd.EndUpdate;
      ChCursor(oldCursor);
    end;
  end
  else
    Add1Rec;
end;

procedure GenerateSaveSQLScript(SqlCmd: TSQLStrings; DS: TDataSet; const TableName: string; KeyFieldList, FieldList: TStrings;
  GenerateInsert, AllRecords: boolean; FileName: string);
var
  L: TStringList;
begin
  L:=TStringList.Create;
  try
    if FileName = '' then
      FileName:=GetAppDir + TableName + '.sql';
    GenerateSQLScript(L, DS, TableName, KeyFieldList, FieldList, GenerateInsert, AllRecords);
    L.SaveToFile(FileName);
  finally
    FreeAndNil(L);
  end;
end;

function GetFilterWildCardChar(Dataset: TDataSet; SQLFormat: boolean = false): Char;
begin
  if SQLFormat then
    Result:=RosiCompConfig.rDBFilterCfg.SQLWildCardChar
  else
    if rdbtool.IsClientDataset(Dataset) then
      Result:='%'
    else
      Result:=RosiCompConfig.rDBFilterCfg.FilterWildCardChar;
end;

function GetFilterLikeOperator(Dataset: TDataSet; SQLFormat: boolean = false): string;
begin
  if SQLFormat then
    Result:=RosiCompConfig.rDBFilterCfg.SQLLikeOperator
  else
    if rdbtool.IsClientDataset(Dataset) then
      Result:='LIKE'
    else
      Result:=RosiCompConfig.rDBFilterCfg.FilterLikeOperator;
end;

function GetFilterNullEqualStr(Dataset: TDataSet): string;
begin
  if rdbtool.IsClientDataset(Dataset) then
    Result:='IS'
  else
    Result:=RosiCompConfig.rDBFilterCfg.NullEqualStr;
end;

function SetFilterString(Dataset: TDataSet; const Filter: string; ShowError: boolean): boolean;
// Try to set filter; show SQL cursor
var
  oldCursor: TCursor;
  B: Tbookmark;
  F: string;
begin
  Result:=false;
  if Dataset = nil then
    Exit;
  try
    oldCursor:=SQLCursor;
    B:=Dataset.GetBookmark;
    Dataset.DisableControls;
    try
      F:=Trim(Filter);
      if F <> '' then
        Dataset.Filter:=F;
      Dataset.Filtered:=F <> '';
      Result:=true;
      if Dataset.Active and not Dataset.IsEmpty then
        try
          if Dataset.BookmarkValid(B) then
            Dataset.GotoBookmark(B)
          else
            Dataset.First;
        except
        end;
    finally
      Dataset.FreeBookmark(B);
      ChCursor(oldCursor);
      Dataset.EnableControls;
    end;
  except
    on E: Exception do
      if ShowError then
        ShowErr(sExpressionErr, E, false);
  end;
end;

procedure ClearFilter(const GridOrDataSet: TObject);
begin
  SetFilterString(GetDataSet(GridOrDataSet), '', true);
end;

function GetFilterCond(Field: TField; const Value: string; IsNull: boolean; Condition: string = ''): string;
var
  val, sep, Cond: string;
begin
  Result:='';
  if (Field = nil) or (Field.Dataset = nil) then
    Exit;
  if not Field.Dataset.Active then
    Exit;
  try
    // If Lookup field, filter by Keyfield
    if Field.FieldKind = fkLookup then
      Field:=Field.Dataset.FieldByName(Field.KeyFields);
    if not(IsSimpleField(Field) or IsMemoField(Field)) then
      Exit;
    if Field.DataType in sfApostrFld then
      sep:=''''
    else
      sep:='';
    if Field is TBooleanField then
      val:=RosiCompConfig.rDBFilterCfg.BooleanString[GetBooleanFieldState(Field, Value) or SameText(Value, RosiCompConfig.rDBFilterCfg.BooleanString[true]) or
        (Value = '1')
      { or SameText(copy(Value, 1, 1), 'Y') or SameText(copy(Value, 1, 1), 'A') } ]
    else
      if (Field is TNumericField) and (GetDecimalSeparator <> RosiCompConfig.rDBFilterCfg.DecimalSeparator) then
        val:=ReplaceStr(GetDecimalSeparator, RosiCompConfig.rDBFilterCfg.DecimalSeparator, Value)
      else
        val:=Value;
    if sep <> '' then
      val:=ReplaceStr('''', '''''', val, true, true);
    if Condition <> '' then
      Cond:=Condition
    else
      if (IsClientDataset(Field.Dataset) and IsMemoField(Field)) then
        Cond:='LIKE'
      else
        Cond:='=';
    if IsNull then
    begin
      sep:=' ';
      val:='NULL';
      Cond:=GetFilterNullEqualStr(Field.Dataset);
    end;
    Result:=MakeFilterCond(GetQuotedFieldName(Field.FieldName), Cond, val, sep);
  except
  end;
end;

function FilterByField(Field: TField; FilterJoin: TFilterJoin = fjNone): boolean;
// Set filter according value of field
var
  val, newf, oldf: string;
begin
  Result:=false;
  // If Lookup field, filter by Keyfield
  if Field.FieldKind = fkLookup then
    Field:=Field.Dataset.FieldByName(Field.KeyFields);
  if Field is TBooleanField then
    val:=RosiCompConfig.rDBFilterCfg.BooleanString[Field.AsBoolean]
  else
    if IsDateField(Field) then
      val:=GetDateFieldFilterValue(Field.AsDateTime)
    else
      val:=Field.AsString;
  newf:=GetFilterCond(Field, val, Field.IsNull);
  if newf = '' then
    Exit;
  try
    oldf:=Trim(Field.Dataset.Filter);
    if oldf <> '' then
      case FilterJoin of
        fjOr:
          oldf:='(' + oldf + ') OR ';
        fjAnd:
          oldf:='(' + oldf + ') AND ';
      else
        oldf:='';
      end;
    Result:=SetFilterString(Field.Dataset, oldf + newf, true);
  except
  end;
end;

function FilterBySel(const Grid: TCustomDBGrid; FilterJoin: TFilterJoin = fjNone): boolean;
begin
  if Assigned(Grid) then
    Result:=FilterByField(Grid.SelectedField, FilterJoin)
  else
    Result:=false;
end;

function NegateFilter(const GridOrDataSet: TObject): boolean;
var
  F: string;
  D: TDataSet;
begin
  Result:=false;
  D:=GetDataSet(GridOrDataSet);
  if D = nil then
    Exit;
  F:=Trim(D.Filter);
  if F = '' then
    Exit;
  F:='NOT (' + F + ')';
  Result:=SetFilterString(D, F, true);
end;

function GetQuotedTableName(TableName: string): string;
begin
  Result:=TableName;
  if (Pos(' ', Result) > 0) or (Pos('_', Result) > 0) or RosiCompConfig.rDBToolCfg.IsReservedKeyword(Result) then
    Result:=RosiCompConfig.rDBFilterCfg.TableQuoteStart + Result + RosiCompConfig.rDBFilterCfg.TableQuoteStop;
end;

function GetQuotedFieldName(FieldName: string): string;
begin
  Result:=FieldName;
  if (Pos(' ', Result) > 0) or (Pos('_', Result) > 0) or RosiCompConfig.rDBToolCfg.IsReservedKeyword(Result) then
    Result:=RosiCompConfig.rDBFilterCfg.FieldQuoteStart + Result + RosiCompConfig.rDBFilterCfg.FieldQuoteStop;
end;

function GetNumericFieldFilterValue(Value: extended): string;
begin
  Result:=FloatToStr(Value);
  if (GetDecimalSeparator <> '.') then
    Result:=ReplaceStr(GetDecimalSeparator, '.', Result);
end;

function GetDateFieldFilterValue(Value: TDateTime): string;
begin
  if (RosiCompConfig.rDBFilterCfg.DateFormat <> '') then
    Result:=FormatDateTime(RosiCompConfig.rDBFilterCfg.DateFormat, Value)
  else
    Result:=DateToStr(Value);
end;

function MakeFilterCond(FldName, Cond, Value: string; AddSep: string = ''): string;
begin
  // if Cond='LIKE' then
  Cond:=' ' + Cond + ' ';
  Result:='(' + FldName + Cond + AddSep + Value + AddSep + ')';
end;

function MakeFilterBetweenCond(FldName, ValueFrom, ValueTo: string; AddSep: string = ''): string;
begin
  Result:='(' + FldName + '>=' + AddSep + ValueFrom + AddSep + ') AND (' + FldName + '<=' + AddSep + ValueTo + AddSep + ')';
end;

procedure FileToBlob(Fld: TField; const FileName: string);
const
  CopySize = 512000;
var
  bs: TStream;
  fs: TFileStream;
  Copied: Int64;
begin
  fs:=TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    // load file to blob
    bs:=Fld.Dataset.CreateBlobStream(Fld, bmWrite);
    try
      repeat
        Copied:=Min(CopySize, fs.Size - fs.Position);
        bs.CopyFrom(fs, Copied);
      until Copied < CopySize;
    finally
      FreeAndNil(bs);
    end;
  finally
    FreeAndNil(fs);
  end;
end;

procedure BlobToFile(Fld: TField; const FileName: string);
const
  CopySize = 512000;
var
  bs: TStream;
  fs: TFileStream;
  Copied: Int64;
begin
  fs:=TFileStream.Create(FileName, fmCreate or fmShareDenyWrite);
  try
    // load file to blob
    bs:=Fld.Dataset.CreateBlobStream(Fld, bmRead);
    try
      repeat
        Copied:=Min(CopySize, fs.Size - fs.Position);
        fs.CopyFrom(bs, Copied);
      until Copied < CopySize;
    finally
      FreeAndNil(bs);
    end;
  finally
    FreeAndNil(fs);
  end;
end;

end.
