unit rdlg;

{<
  rDlg form provides set of Yes/No, Ok/Cancel and input dialogs with HTML format captions,
  new grafics icons and buttons.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jan 2006)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  StdCtrls, ImgList, Buttons, ExtCtrls, Dialogs, ComCtrls, pngimage,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF}
  RosiCompGlobalCfg;

type
  { @exclude }
  TrDlgForm = class(TForm)
    imgI: TImage;
    imgW: TImage;
    imgQ: TImage;
    imgE: TImage;
    ImgDlgBtn: TImageList;
  private
  public
  end;

  { list of possible dialog types used by rDlg }
  TDialogType = (dtNone, dtI, dtQ, dtW, dtE);
  { list of possible button types used by rDlg }
  TButtonsType = (btOk, btOkCancel, btYesNo, btYesNoCancel, btYesNoUserDef, btYesNoCancelUserDef);

{ show Information dialog with one html message and Ok button }
function DlgI(const InfoText: string; WordWrap: boolean = true): integer;
{ show Question dialog with one html message and Yes No button }
function DlgQ(const InfoText: string; WordWrap: boolean = true): integer;
{ show Question dialog with one html message and Yes No Cancel button }
function DlgQCancel(const InfoText: string; WordWrap: boolean = true): integer;
{ show Warning dialog with one html message and Ok button }
function DlgW(const InfoText: string; WordWrap: boolean = true): integer;
{ show Error dialog with one html message and Ok button }
function DlgE(const InfoText: string; WordWrap: boolean = true): integer;

{ show Information dialog with two html messages and Ok button }
function DlgI2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
{ show Question dialog with two html messages and Yes No button }
function DlgQ2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
{ show Question dialog with two html messages and Yes No Cancel button }
function DlgQCancel2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
{ show Warning dialog with two html messages and Ok button }
function DlgW2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
{ show Error dialog with two html messages and Ok button }
function DlgE2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;

{ show dialog with 2 or 3 buttons with user defined text and any number of messages }
function DlgQUserDef(const Button1, Button2, Button3: string; const Text: array of string): integer;

{ show user defined with any number of html messages
  Params contain array of strings to display or helpcontext, if string begins:
  ^C - this string will be used as caption of window
  ^B - this string will be displayed bold;
  ^I - -"- italic;
  ^U - -"- underline;
  ^L - -"- large font;
  ^X - -"- extra large font; }
function ShowDlg(const DialogType: TDialogType; const ButtonsType: TButtonsType; const Params: array of Variant; WordWrap: boolean = true): integer;

{ show input dialog for string input }
function InputStr(const Caption, Info: string; var Input: string): boolean;
{ show input dialog for integer input }
function InputInt(const Caption, Info: string; var Input: integer): boolean;
{ show input dialog for password input }
function InputPassword(const Info: string; var Psw: string): boolean;

{ show input dialog for string input with definable width and mask char }
function ShowInput(const Caption, Info: string; var Input: string; const Width: integer; const MaskChar: Char): boolean;
{ show input dialog for string input with definable width, mask char and ComboBox with items }
function ShowInputCB(const Caption, Info: string; var Input: string; const Width: integer; const MaskChar: Char; CBItems: TStrings; ListOnly: boolean): boolean;

{ show small info window with text message and without any button or link }
procedure ShowInfo(const InfoText: string; const CenterIn: TWinControl = nil);
{ show small info window with text message and Abort link }
procedure ShowInfoEx(const InfoText: string; const CenterIn: TWinControl; OnAbort: TNotifyEvent);
{ show small info window with text message, progress bar and Abort link }
procedure ShowInfoProgressBar(const InfoText: string; const CenterIn: TWinControl; OnAbort: TNotifyEvent; ProgressProcent: integer);

{ show error dialog with info message and exception description text }
procedure ShowErr(const Info: string; E: Exception; EndProgram: boolean = false);

implementation

{$R *.dfm}

uses rstring, rtool, rHTMLLabel;

const
  gap = 10;
  btnW = 100;
  btnH = 38;
  cBtnNum: array [0 .. 5] of integer = (1, 2, 2, 3, 2, 3);
  cDlgCapStr: array [0 .. 4] of string = ('', sInformation, sQuestion, sWarning, sError);
  cBtnCapStr: array [1 .. 18] of string = (sBtOk, '', '', sBtOk, sBtCancel, '', sBtYes, sBtNo, '', sBtYes, sBtNo, sBtCancel, '', '', '', '', '', '');
  cBtnImgIdx: array [1 .. 18] of integer = (0, 0, 0, 0, 1, 0, 0, 2, 0, 0, 2, 1, -1, -1, -1, -1, -1, -1);

var
  InfoForm: TForm;

function DlgI(const InfoText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtI, btOk, [InfoText], WordWrap);
end;

function DlgQ(const InfoText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtQ, btYesNo, [InfoText], WordWrap);
end;

function DlgQCancel(const InfoText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtQ, btYesNoCancel, [InfoText], WordWrap);
end;

function DlgW(const InfoText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtW, btOk, [InfoText], WordWrap);
end;

function DlgE(const InfoText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtE, btOk, [InfoText], WordWrap);
end;

function DlgI2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtI, btOk, ['^B' + BoldText, NormalText], WordWrap);
end;

function DlgQ2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtQ, btYesNo, ['^B' + BoldText, NormalText], WordWrap);
end;

function DlgQCancel2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtQ, btYesNoCancel, ['^B' + BoldText, NormalText], WordWrap);
end;

function DlgW2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtW, btOk, ['^B' + BoldText, NormalText], WordWrap);
end;

function DlgE2(const BoldText, NormalText: string; WordWrap: boolean = true): integer;
begin
  Result:=ShowDlg(dtE, btOk, ['^B' + BoldText, NormalText], WordWrap);
end;

function DlgQUserDef(const Button1, Button2, Button3: string; const Text: array of string): integer;
var
  v: array of Variant;
  i: integer;
begin
  SetLength(v, Length(Text) + 3);
  for i:=0 to High(Text) do
    v[i]:=Text[i];
  i:=Length(Text);
  v[i]:='@1' + Button1;
  Inc(i);
  v[i]:='@2' + Button2;
  if Button3 = '' then
    Result:=ShowDlg(dtQ, btYesNoUserDef, v)
  else
  begin
    Inc(i);
    v[i]:='@3' + Button3;
    Result:=ShowDlg(dtQ, btYesNoCancelUserDef, v);
  end;
end;

function ShowDlg(const DialogType: TDialogType; const ButtonsType: TButtonsType; const Params: array of Variant; WordWrap: boolean = true): integer;
var
  F: TrDlgForm;
  i: TImage;
  B: array [1 .. 3] of TButton;
  L: TrHTMLLabel;
  a, aBut, aLeft, aTop, aButW: integer;
  s: string;
  oldCursor: TCursor;

  function UsePrefixShow(F: TFont; SkipBtnPrefix: boolean): string;
  begin
    if SkipBtnPrefix then
      s:=copy(s, 3, Length(s));
    repeat
      if Length(s) < 2 then
        Break;
      if s[1] = '^' then
      begin
        case s[2] of
          'B':
            F.Style:=F.Style + [fsBold];
          'I':
            F.Style:=F.Style + [fsItalic];
          'U':
            F.Style:=F.Style + [fsUnderline];
          'L':
            F.Size:=F.Size + 2;
          'X':
            F.Size:=F.Size + 4;
        end;
        s:=copy(s, 3, Length(s));
      end
      else
        Break;
    until false;
    Result:=Trim(s);
  end;

begin
  // inicialization
  aLeft:=gap;
  aTop:=gap;

  // Form
  F:=TrDlgForm.Create(GetActiveOwner);
  // F.Position:=poScreenCenter;
  F.ClientWidth:=DPI_ScaleSize(400);
  F.ClientHeight:=DPI_ScaleSize(600);
  F.BorderStyle:=bsDialog;
  F.Caption:=cDlgCapStr[ord(DialogType)];
  F.Font.Name:=RosiCompConfig.rDBToolCfg.rDlg_DefaultFont;
  F.Font.Size:=RosiCompConfig.rDBToolCfg.rDlg_DefaultSize;
  rtool.DPI_SetFontHeight(F.Font);

  // Buttons create only
  aBut:=cBtnNum[ord(ButtonsType)];
  for a:=1 to aBut do
  begin
    B[a]:=TButton.Create(F);
    B[a].Parent:=F;
    B[a].Caption:=cBtnCapStr[a + ord(ButtonsType) * 3];
    B[a].Images:=F.ImgDlgBtn;
    B[a].ImageIndex:=cBtnImgIdx[a + ord(ButtonsType) * 3];
    B[a].ImageMargins.Left:=4;
    B[a].WordWrap:=true;
  end;

  // Image
  if DialogType <> dtNone then
  begin
    case ord(DialogType) of
      1:
        i:=F.imgI;
      2:
        i:=F.imgQ;
      3:
        i:=F.imgW;
      4:
        i:=F.imgE;
    else
      i:=nil;
    end;
    if Assigned(i) then
    begin
      i.Left:=aLeft;
      Inc(aLeft, i.Width + gap);
      i.Top:=aTop;
      Inc(aTop, 8);
      i.Show;
    end;
  end;

  // Labels
  for a:=0 to High(Params) do
    case VarType(Params[a]) of
      varString {$IFDEF D2009P}, varUString {$ENDIF}:
        begin
          s:=Params[a] + '   ';
          // caption?
          if copy(s, 1, 2) = '^C' then
            F.Caption:=Trim(copy(s, 3, Length(s)))
          else
            if copy(s, 1, 2) = '@1' then
              B[1].Caption:=UsePrefixShow(B[1].Font, true)
            else
              if copy(s, 1, 2) = '@2' then
                B[2].Caption:=UsePrefixShow(B[2].Font, true)
              else
                if copy(s, 1, 2) = '@3' then
                  B[3].Caption:=UsePrefixShow(B[3].Font, true)
                else
                begin
                  L:=TrHTMLLabel.Create(F);
                  L.Parent:=F;
                  L.WordWrap:=WordWrap;
                  L.Left:=aLeft;
                  L.Top:=aTop;
                  L.Width:=F.ClientWidth - L.Left - gap;
                  L.Caption:=UsePrefixShow(L.Font, false);
                  Inc(aTop, L.Height + gap);
                  // check width
                  if L.Width > Screen.DesktopRect.Right then
                    L.Width:=Screen.DesktopRect.Right - aLeft - 2 * gap;
                  if L.Width > F.ClientWidth - L.Left - gap then
                    F.ClientWidth:=L.Left + gap + L.Width;
                  L.OnAnchorClick:=RosiCompConfig.rDBToolCfg.rDlg_AnchorClick;
                end;
        end;
      varInteger:
        F.HelpContext:=Params[a];
    end;

  // Buttons
  aTop:=aTop + gap;
  if aTop < DPI_ScaleSize(60) then
    aTop:=DPI_ScaleSize(60);
  aButW:=btnW;
  if ButtonsType in [btYesNoUserDef, btYesNoCancelUserDef] then
    aButW:=(F.ClientWidth - (5 * gap)) div aBut;
  for a:=1 to aBut do
  begin
    aLeft:=((F.ClientWidth - (2 * gap) - aBut * aButW) div (aBut + 1)) * a + gap + aButW * (a - 1);
    B[a].SetBounds(aLeft, aTop, aButW, btnH);
  end;

  Inc(aTop, B[aBut].Height + gap);
  case ButtonsType of
    btOk:
      begin
        B[1].Cancel:=true;
        B[1].Default:=true;
        B[1].ModalResult:=mrOK;
      end;
    btOkCancel:
      begin
        B[1].Default:=true;
        B[1].ModalResult:=mrOK;
        B[2].Cancel:=true;
        B[2].ModalResult:=mrCancel;
      end;
    btYesNo, btYesNoUserDef:
      begin
        B[1].ModalResult:=mrYes;
        B[2].ModalResult:=mrNo;
        F.BorderIcons:=[];
      end;
    btYesNoCancel, btYesNoCancelUserDef:
      begin
        B[1].ModalResult:=mrYes;
        B[2].ModalResult:=mrNo;
        B[3].Cancel:=true;
        B[3].ModalResult:=mrCancel;
      end;
  end;

  F.ClientHeight:=aTop;
  oldCursor:=NCursor;
  Result:=F.ShowModal;
  ChCursor(oldCursor);
  FreeAndNil(F);
end;

function InputStr(const Caption, Info: string; var Input: string): boolean;
begin
  Result:=ShowInput(Caption, Info, Input, 350, #0);
end;

function InputInt(const Caption, Info: string; var Input: integer): boolean;
var
  s: string;
begin
  repeat
    s:=IntToStr(Input);
    Result:=ShowInput(Caption, Info, s, 240, #0);
    if Result then
      try
        Input:=StrToInt(s);
      except
        DlgE2(sInputErr, sFormatNumberErr);
        s:='e';
      end;
  until s <> 'e';
end;

function InputPassword(const Info: string; var Psw: string): boolean;
begin
  Result:=ShowInput(sInput, Info, Psw, 240, '*');
end;

function ShowInput(const Caption, Info: string; var Input: string; const Width: integer; const MaskChar: Char): boolean;
begin
  Result:=ShowInputCB(Caption, Info, Input, Width, MaskChar, nil, false);
end;

function ShowInputCB(const Caption, Info: string; var Input: string; const Width: integer; const MaskChar: Char; CBItems: TStrings; ListOnly: boolean): boolean;
var
  F: TrDlgForm;
  L: TrHTMLLabel;
  E: TEdit;
  CB: TComboBox;
  B1, B2: TButton;
  aLeft, aTop: integer;
  oldCursor: TCursor;
begin
  // inicialization
  aLeft:=gap;
  aTop:=gap;

  // Form
  F:=TrDlgForm.Create(GetActiveOwner);
  // F.Position:=poScreenCenter;
  F.ClientWidth:=DPI_ScaleSize(Width);
  F.ClientHeight:=DPI_ScaleSize(600);
  F.BorderStyle:=bsDialog;
  F.Caption:=Caption;
  F.Font.Name:=RosiCompConfig.rDBToolCfg.rDlg_DefaultFont;
  F.Font.Size:=RosiCompConfig.rDBToolCfg.rDlg_DefaultSize;
  rtool.DPI_SetFontHeight(F.Font);

  // Label
  L:=TrHTMLLabel.Create(F);
  L.OnAnchorClick:=RosiCompConfig.rDBToolCfg.rDlg_AnchorClick;
  L.Parent:=F;
  L.WordWrap:=true;
  L.Left:=aLeft;
  L.Top:=aTop;
  L.Width:=F.ClientWidth - L.Left - gap;
  L.Caption:=Info;
  Inc(aTop, L.Height + gap);
  // check width
  if L.Width > F.ClientWidth - L.Left - gap then
    F.ClientWidth:=L.Left + gap + L.Width;

  E:=nil;
  CB:=nil;
  if CBItems = nil then
  begin
    // Edit
    E:=TEdit.Create(F);
    E.Parent:=F;
    E.Left:=aLeft;
    E.Width:=F.ClientWidth - E.Left - gap;
    E.Top:=aTop;
    Inc(aTop, E.Height + gap);
    E.PasswordChar:=MaskChar;
    E.Text:=Input;
    E.SelectAll;
  end
  else
  begin
    // ComboBox
    CB:=TComboBox.Create(F);
    CB.Parent:=F;
    CB.Left:=aLeft;
    CB.Width:=F.ClientWidth - CB.Left - gap;
    CB.Top:=aTop;
    Inc(aTop, CB.Height + gap);
    CB.DropDownCount:=12;
    CB.Items.Assign(CBItems);
    if ListOnly then
    begin
      CB.Style:=csDropDownList;
      CB.ItemIndex:=CBItems.IndexOf(Input);
    end
    else
    begin
      CB.Text:=Input;
      CB.SelectAll;
    end;
  end;

  // Buttons
  B1:=TButton.Create(F);
  B1.Parent:=F;
  B1.Top:=aTop;
  B1.Left:=aLeft;
  B1.Height:=btnH;
  B1.Width:=btnW;
  B1.Caption:=sBtOk;
  B1.Images:=F.ImgDlgBtn;
  B1.ImageIndex:=0;
  B1.Default:=true;
  B1.ModalResult:=mrOK;
  B1.ImageMargins.Left:=4;
  B2:=TButton.Create(F);
  B2.Parent:=F;
  B2.Top:=aTop;
  B2.Left:=B1.Left + B1.Width + 2 * gap;
  B2.Height:=btnH;
  B2.Width:=btnW;
  B2.Caption:=sBtCancel;
  B2.Images:=F.ImgDlgBtn;
  B2.ImageIndex:=1;
  B2.Cancel:=true;
  B2.ModalResult:=mrCancel;
  B2.ImageMargins.Left:=4;
  Inc(aTop, B1.Height + gap);

  F.ClientHeight:=aTop;
  oldCursor:=NCursor;
  Result:=F.ShowModal = mrOK;
  ChCursor(oldCursor);
  if Result then
    if CBItems = nil then
      Input:=E.Text
    else
      Input:=CB.Text;
  FreeAndNil(F);
end;

procedure ShowInfo(const InfoText: string; const CenterIn: TWinControl = nil);
begin
  ShowInfoProgressBar(InfoText, CenterIn, nil, -1);
end;

procedure ShowInfoEx(const InfoText: string; const CenterIn: TWinControl; OnAbort: TNotifyEvent);
begin
  ShowInfoProgressBar(InfoText, CenterIn, OnAbort, -1);
end;

procedure ShowInfoProgressBar(const InfoText: string; const CenterIn: TWinControl; OnAbort: TNotifyEvent; ProgressProcent: integer);
var
  p: TPoint;
  Panel: TPanel;
  LInfo: TrHTMLLabel;
  PBar: TProgressBar;
  LAbort: TrURLLabel;
  i: integer;
begin
  // Hide?
  if (InfoText = '') then
  begin
    NCursor;
    if InfoForm = nil then
      Exit;
    InfoForm.Hide;
    FreeAndNil(InfoForm);
    Exit;
  end;
  // Show
  if InfoForm = nil then
  begin
    // Create form and controls
    InfoForm:=TForm.Create(GetActiveOwner);
    InfoForm.BorderIcons:=[];
    InfoForm.BorderStyle:=bsToolWindow;
    InfoForm.Position:=poOwnerFormCenter;
    InfoForm.Caption:=sInformation;
    InfoForm.Font.Name:=RosiCompConfig.rDBToolCfg.rDlg_DefaultFont;
    InfoForm.Font.Size:=RosiCompConfig.rDBToolCfg.rDlg_DefaultSize;
    rtool.DPI_SetFontHeight(InfoForm.Font);
    InfoForm.Height:=DPI_ScaleSize(120);
    InfoForm.Width:=DPI_ScaleSize(300);

    Panel:=TPanel.Create(InfoForm);
    Panel.Parent:=InfoForm;
    Panel.Align:=alClient;
    Panel.BevelWidth:=2;

    LInfo:=TrHTMLLabel.Create(InfoForm);
    LInfo.Parent:=Panel;
    LInfo.Align:=alClient;
    LInfo.Alignment:=taCenter;
    LInfo.Layout:=tlCenter;
    LInfo.WordWrap:=true;

    PBar:=TProgressBar.Create(InfoForm);
    PBar.Visible:=false;
    PBar.Parent:=Panel;
    PBar.Align:=alBottom;
    PBar.AlignWithMargins:=true;
    PBar.Min:=0;
    PBar.Max:=100;
    PBar.Position:=0;
    PBar.Step:=1;

    LAbort:=TrURLLabel.Create(InfoForm);
    LAbort.Visible:=false;
    LAbort.Parent:=Panel;
    LAbort.Caption:=sAbort;
  end;

  if Assigned(CenterIn) then
  begin
    InfoForm.Position:=poDesigned;
    if CenterIn is TForm then
    begin
      InfoForm.PopupParent:=CenterIn as TForm;
      p:=CenterIn.BoundsRect.TopLeft;
    end
    else
      p:=CenterIn.ClientToScreen(Point(0, 0)); // CenterIn.BoundsRect.TopLeft);
    InfoForm.Left:=p.X + (CenterIn.Width - InfoForm.Width) div 2;
    InfoForm.Top:=p.Y + (CenterIn.Height - InfoForm.Height) div 2;
  end;

  Panel:=(InfoForm.Controls[0] as TPanel);
  LInfo:=nil;
  PBar:=nil;
  LAbort:=nil;
  for i:=0 to Panel.ControlCount - 1 do
  begin
    if Panel.Controls[i] is TrHTMLLabel then
      LInfo:=(Panel.Controls[i] as TrHTMLLabel);
    if Panel.Controls[i] is TrURLLabel then
      LAbort:=(Panel.Controls[i] as TrURLLabel);
    if Panel.Controls[i] is TProgressBar then
      PBar:=(Panel.Controls[i] as TProgressBar);
  end;
  LInfo.Caption:=InfoText;
  if (ProgressProcent >= 0) then
    SetProgressBar(PBar, ProgressProcent, 100);
  InfoForm.Show;
  if Assigned(OnAbort) then
  begin
    PBar.Visible:=true;
    PBar.Margins.Right:=LAbort.Width + 16;
    LAbort.Left:=PBar.Left + PBar.Width + 8;
    LAbort.Top:=(PBar.Height - LAbort.Height) div 2 + PBar.Top;
    LAbort.OnClick:=OnAbort;
    NCursor;
  end
  else
    WCursor;
  PBar.Visible:=ProgressProcent >= 0;
  LAbort.Visible:=Assigned(OnAbort);
  InfoForm.Update;
  Application.ProcessMessages;
end;

// display error message
procedure ShowErr(const Info: string; E: Exception; EndProgram: boolean);
begin
  // ShowInfo('',nil);
  if Application.Terminated then
    Exit;
  MessageBeep(MB_ICONEXCLAMATION);
  if Info <> '' then
    DlgE2(Info, sErrorDescription + #13 + E.Message)
  else
    DlgE(sErrorDescription + #13 + E.Message);
  if EndProgram then
    Application.Terminate;
end;

initialization

InfoForm:=nil;

end.
