unit rfiletool;

{<
  rfiletool is set of file and system functions used by other components.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Jun 2016)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Clipbrd, ShlObj, ShellApi, ActiveX, ComObj, Axctrls, RosiCompGlobalCfg;

{ fill list with all files according DirWildCard }
function LoadFileList(FileList: TStrings; const DirWildCard: string; AddDirToFileName: Boolean = false): integer;
{ fill list with all files according DirWildCard including files from subdirectories }
function LoadFileListSubDir(FileList: TStrings; const Dir: string; const FileWildCard: string): integer;

{ Get file size as INT64 }
function GetFileSize(const FileName: string): Int64;
{ Get file type description for defined extension }
function GetGenericFileType(const AExtension: string): string;

{ copy whole directory using standard progress dialog }
function Sys_CopyDir(const fromDir, toDir: string; Confirmation: Boolean = true): Boolean;
{ delete whole directory using standard progress dialog }
function Sys_DeleteDir(const Dir: string; Confirmation: Boolean = true): Boolean;
{ copy file list using standard progress dialog }
function Sys_CopyFiles(const FileArray: array of string; const toDir: string): Boolean; overload;
{ copy file list using standard progress dialog }
function Sys_CopyFiles(const FileList: TStrings; const toDir: string): Boolean; overload;
{ paste files from clipboard using standard progress dialog }
function Sys_PasteFilesFromClipboard(const toDir: string): integer;
{ paste files from data object (e.g. Email) using standard progress dialog }
function Sys_PasteFilesFromDataObject(Obj: IDataObject; const toDir: string): integer;
{ copy file list to clipboard using standard progress dialog }
procedure Sys_CopyFilesToClipboard(const FileArray: array of string); overload;
{ copy file list to clipboard using standard progress dialog }
procedure Sys_CopyFilesToClipboard(const FileList: TStrings); overload;
{ delete file list (move to recykle bin) }
function Sys_DeleteFilesToBin(const FileArray: array of string): Boolean; overload;
{ delete file list (move to recykle bin) }
function Sys_DeleteFilesToBin(const FileList: TStrings): Boolean; overload;

{ return handle of icon associated with the file type }
function GetAssociatedIconHandle(FileName: string): HIcon;
{ load icon associated with the file type to bitmap }
procedure LoadAssociatedIconAsBitmap(Bmp: TBitmap; FileName: string);

implementation

uses rstring, rtool, rdlg;

function LoadFileList(FileList: TStrings; const DirWildCard: string; AddDirToFileName: Boolean = false): integer;
var
  F: TSearchRec;
begin
  FileList.Clear;
  if FindFirst(DirWildCard, faAnyFile, F) = 0 then
  begin
    repeat
      if AddDirToFileName then
        FileList.Add(ExtractFilePath(DirWildCard) + (F.Name))
      else
        FileList.Add(F.Name);
    until FindNext(F) <> 0;
    FindClose(F);
  end;
  Result:=FileList.Count;
end;

function LoadFileListSubDir(FileList: TStrings; const Dir: string; const FileWildCard: string): integer;
  procedure AddDir(ADir: string);
  var
    F: TSearchRec;
  begin
    ADir:=rtool.SetEndOfPath(ADir);
    // load files
    if FindFirst(ADir + FileWildCard, faAnyFile, F) = 0 then
    begin
      repeat
        if (F.Attr and faDirectory) <> 0 then
          Continue;
        FileList.Add(ADir + F.Name);
      until FindNext(F) <> 0;
      FindClose(F);
    end;
    // load dirs
    if FindFirst(ADir + '*.*', faDirectory, F) = 0 then
    begin
      repeat
        if (F.Name = '.') or (F.Name = '..') then
          Continue;
        if (F.Attr and faDirectory) <> 0 then
          AddDir(ADir + F.Name);
      until FindNext(F) <> 0;
      FindClose(F);
    end;
  end;

begin
  FileList.Clear;
  AddDir(Dir);
  Result:=FileList.Count;
end;

function GetFileSize(const FileName: string): Int64;
var
  sr: TSearchRec;
begin
{$WARNINGS OFF}
  if SysUtils.FindFirst(FileName, faAnyFile, sr) = 0 then
    Result:=Int64(sr.FindData.nFileSizeHigh) shl Int64(32) + Int64(sr.FindData.nFileSizeLow)
  else
    Result:=-1;
  SysUtils.FindClose(sr);
{$WARNINGS ON}
end;

function GetGenericFileType(const AExtension: string): string;
var
  AInfo: TSHFileInfo;
begin
  SHGetFileInfo(PChar('*' + AExtension), FILE_ATTRIBUTE_NORMAL, AInfo, SizeOf(AInfo), SHGFI_TYPENAME or SHGFI_USEFILEATTRIBUTES);
  Result:=AInfo.szTypeName;
end;

function Sys_CopyDir(const fromDir, toDir: string; Confirmation: Boolean = true): Boolean;
var
  fos: TSHFileOpStruct;
begin
  ZeroMemory(@fos, SizeOf(fos));
  with fos do
  begin
    wFunc:=FO_COPY;
    if Confirmation then
      fFlags:=0
    else
      fFlags:=FOF_NOCONFIRMATION;
    pFrom:=PChar(fromDir + #0);
    pTo:=PChar(toDir + #0)
  end;
  Result:=(ShFileOperation(fos) = 0);
  if Result then
    if fos.fAnyOperationsAborted then
      Result:=false;
end;

function Sys_DeleteDir(const Dir: string; Confirmation: Boolean = true): Boolean;
var
  fos: TSHFileOpStruct;
begin
  ZeroMemory(@fos, SizeOf(fos));
  with fos do
  begin
    wFunc:=FO_DELETE;
    if Confirmation then
      fFlags:=0
    else
      fFlags:=FOF_NOCONFIRMATION;
    pFrom:=PChar(Dir + #0);
    pTo:=nil;
  end;
  Result:=(ShFileOperation(fos) = 0);
  if Result then
    if fos.fAnyOperationsAborted then
      Result:=false;
end;

function Sys_CopyFiles_Int(FileListZeroSep, toDir: string): Boolean;
var
  fos: TSHFileOpStruct;
begin
  ZeroMemory(@fos, SizeOf(fos));
  with fos do
  begin
    wFunc:=FO_COPY;
    // fFlags := FOF_FILESONLY;
    pFrom:=PChar(FileListZeroSep + #0);
    pTo:=PChar(toDir);
  end;
  Result:=(ShFileOperation(fos) = 0);
end;

function Sys_CopyFiles(const FileArray: array of string; const toDir: string): Boolean;
var
  fn: string;
  i: integer;
begin
  fn:='';
  for i:=0 to High(FileArray) do
    fn:=fn + FileArray[i] + #0;
  Result:=Sys_CopyFiles_Int(fn, toDir);
end;

function Sys_CopyFiles(const FileList: TStrings; const toDir: string): Boolean;
var
  fn: string;
  i: integer;
begin
  fn:='';
  for i:=0 to FileList.Count - 1 do
    fn:=fn + FileList[i] + #0;
  Result:=Sys_CopyFiles_Int(fn, toDir);
end;
{$IFOPT R+}
{$DEFINE RANGEON}
{$R-}
{$ELSE}
{$UNDEF RANGEON}
{$ENDIF}

function Sys_PasteFilesFromDataObject(Obj: IDataObject; const toDir: string): integer;

  function FillStreamFromIStorage(const Storage: IStorage; Stream: TStream): Boolean;
  var
    LockBytes: ILockBytes;
    HGlob: HGLOBAL;
    NewStorage: IStorage;
    intfStream: IStream;
    OleStream: TOLEStream;
  begin
    Result:=true;
    HGlob:=GlobalAlloc(GMEM_MOVEABLE, 0);
    try
      // TODO : Check if the above memory is leaked.
      OleCheck(CreateILockBytesOnHGlobal(HGlob, true, LockBytes));
      OleCheck(StgCreateDocfileOnILockBytes(LockBytes, STGM_CREATE or STGM_READWRITE or STGM_SHARE_EXCLUSIVE
          { or STGM_DIRECT } , 0, NewStorage));
      OleCheck(Storage.CopyTo(0, nil, nil, NewStorage));
      OleCheck(Storage.Commit(STGC_DEFAULT));
      OleCheck(GetHGlobalFromILockBytes(LockBytes, HGlob));
      OleCheck(CreateStreamOnHGlobal(HGlob, true, intfStream));
      OleStream:=TOLEStream.Create(intfStream);
      Stream.CopyFrom(OleStream, OleStream.Size);
      intfStream:=nil;
    except
      GlobalFree(HGlob);
      raise ;
    end;
  end;

  procedure ProcFileContents(Index: UINT; FileName: string);
  var
    Frm: FORMATETC;
    Medium: STGMEDIUM;
    FS: TFileStream;
    OleStream: TOLEStream;
    FullFn: string;
    Storage: IStorage;
    oldCursor: TCursor;
  begin
    Frm.cfFormat:=RegisterClipboardFormat(CFSTR_FILECONTENTS);
    Frm.ptd:=nil;
    Frm.dwAspect:=DVASPECT_CONTENT;
    Frm.lindex:=Index;
    Frm.tymed:=TYMED_HGLOBAL or TYMED_ISTREAM or TYMED_ISTORAGE;
    if Failed(Obj.GetData(Frm, Medium)) then
      Exit;
    try
      FullFn:=rtool.SetEndOfPath(toDir) + FileName;
      case Medium.tymed of
        TYMED_ISTREAM:
          begin
            if not FileExists(FullFn) or (DlgQ2(Format(sFileExist, [FileName]), sOverwriteQ) = mrYes) then
            begin
              FS:=TFileStream.Create(FullFn, fmCreate);
              oldCursor:=WCursor;
              try
                OleStream:=TOLEStream.Create(IStream(Medium.stm));
                FS.CopyFrom(OleStream, OleStream.Size);
              finally
                FreeAndNil(FS);
                ChCursor(oldCursor);
              end;
              Inc(Result);
            end;
          end;
        TYMED_ISTORAGE:
          begin
            Storage:=IStorage(Medium.stg);
            FS:=TFileStream.Create(FullFn, fmCreate);
            oldCursor:=WCursor;
            try
              FillStreamFromIStorage(Storage, FS);
            finally
              FreeAndNil(FS);
              ChCursor(oldCursor);
            end;
            Inc(Result);
          end;
      end;
    except
      on E: Exception do
        ShowErr(Format(rstring.sFileSaveErr, [FullFn]), E, false);
    end;
    ReleaseStgMedium(Medium);
  end;

  procedure ProcAnsiFiles(Group: PFileGroupDescriptorA);
  var
    i: UINT;
    FileName: AnsiString;
  begin
    for i:=0 to Group.cItems - 1 do
    begin
      FileName:=Group.fgd[i].cFileName;
      ProcFileContents(i, string(FileName));
    end;
  end;

  procedure ProcUnicodeFiles(Group: PFileGroupDescriptorW);
  var
    i: UINT;
    FileName: WideString;
  begin
    for i:=0 to Group.cItems - 1 do
    begin
      FileName:=Group.fgd[i].cFileName;
      ProcFileContents(i, FileName);
    end;
  end;

var
  Frm: TFormatEtc;
  Medium: TStgMedium;
  Group: Pointer;
begin
  Result:=0;
  Frm.cfFormat:=RegisterClipboardFormat(CFSTR_FILEDESCRIPTOR);
  Frm.ptd:=nil;
  Frm.dwAspect:=DVASPECT_CONTENT;
  Frm.lindex:=-1;
  Frm.tymed:=TYMED_HGLOBAL;

  if Failed(Obj.GetData(Frm, Medium)) then
    Exit;
  if toDir = '' then
  begin
    Result:=1;
    Exit;
  end;

  Group:=GlobalLock(Medium.HGLOBAL);
  try
    if Frm.cfFormat = RegisterClipboardFormat(CFSTR_FILEDESCRIPTORW) then
      ProcUnicodeFiles(PFileGroupDescriptorW(Group))
    else
      ProcAnsiFiles(PFileGroupDescriptorA(Group));
  finally
    GlobalUnlock(Medium.HGLOBAL);
  end;
  ReleaseStgMedium(Medium);
end;
{$IFDEF RANGEON}
{$R+}
{$UNDEF RANGEON}
{$ENDIF}

function Sys_PasteFilesFromClipboard(const toDir: string): integer;
var
  F: THandle;
  folder: Array [0 .. MAX_PATH] of Char;
  i, numFiles: integer;
  fn: array of string;
  TryPasteStream: Boolean;
  Obj: IDataObject;
begin
  Result:=0;
  TryPasteStream:=false;
  Clipboard.Open;
  try
    F:=Clipboard.GetAsHandle(CF_HDROP);
    if F <> 0 then
    begin
      numFiles:=DragQueryFile(F, $FFFFFFFF, nil, 0);
      SetLength(fn, numFiles);
      for i:=0 to numFiles - 1 do
      begin
        folder[0]:=#0;
        DragQueryFile(F, i, folder, SizeOf(folder));
        fn[i]:=(folder);
      end;
      if Sys_CopyFiles(fn, toDir) then
        Result:=numFiles;
    end
    else
      TryPasteStream:=true;
  finally
    Clipboard.Close;
  end;
  if TryPasteStream then
  begin
    OleGetClipboard(Obj);
    if not Assigned(Obj) then
      Exit;
    Result:=Sys_PasteFilesFromDataObject(Obj, toDir);
  end;
end;

procedure Sys_CopyFilesToClipboard_Int(FileListZeroSep: string);
var
  DropFiles: PDropFiles;
  HGLOBAL: THandle;
  iLen: integer;
begin
  FileListZeroSep:=FileListZeroSep + #0;
  iLen:=Length(FileListZeroSep);
  HGLOBAL:=GlobalAlloc(GMEM_SHARE or GMEM_MOVEABLE or GMEM_ZEROINIT, SizeOf(TDropFiles) + ((iLen) * SizeOf(Char)));
  if (HGLOBAL = 0) then
    raise Exception.Create('Could not allocate memory.');
  try
    DropFiles:=GlobalLock(HGLOBAL);
    if (DropFiles = nil) then
      raise Exception.Create('Could not access allocated memory.');
    try
      DropFiles^.pFiles:=SizeOf(TDropFiles);
      // if you need the code to be compatible with pre-2009 versions
      // of Delphi, then replace the below line with this instead:
      //
      // DropFiles^.fWide := (SizeOf(Char) = SizeOf(WideChar));
      //
      DropFiles^.fWide:=true;
      Move(FileListZeroSep[1], (PByte(DropFiles) + SizeOf(TDropFiles))^, iLen * SizeOf(Char));
    finally
      GlobalUnlock(HGLOBAL);
    end;
    Clipboard.SetAsHandle(CF_HDROP, HGLOBAL);
  except
    GlobalFree(HGLOBAL);
  end;
end;

procedure Sys_CopyFilesToClipboard(const FileArray: array of string);
var
  i: integer;
  fn: string;
begin
  fn:='';
  for i:=0 to High(FileArray) do
    fn:=fn + FileArray[i] + #0;
  if fn = '' then
    Exit;
  Sys_CopyFilesToClipboard_Int(fn);
end;

procedure Sys_CopyFilesToClipboard(const FileList: TStrings); overload;
var
  i: integer;
  fn: string;
begin
  fn:='';
  for i:=0 to FileList.Count - 1 do
    fn:=fn + FileList[i] + #0;
  if fn = '' then
    Exit;
  Sys_CopyFilesToClipboard_Int(fn);
end;

function Sys_DeleteFilesToBin_Int(FileListZeroSep: string): Boolean;
var
  fos: TSHFileOpStruct;
begin
  ZeroMemory(@fos, SizeOf(fos));
  with fos do
  begin
    wFunc:=FO_DELETE;
    pFrom:=PChar(FileListZeroSep + #0);
    fFlags:=FOF_ALLOWUNDO;
  end;
  Result:=(ShFileOperation(fos) = 0);
end;

function Sys_DeleteFilesToBin(const FileArray: array of string): Boolean;
var
  fn: string;
  i: integer;
begin
  fn:='';
  for i:=0 to High(FileArray) do
    fn:=fn + FileArray[i] + #0;
  Result:=Sys_DeleteFilesToBin_Int(fn);
end;

function Sys_DeleteFilesToBin(const FileList: TStrings): Boolean; overload;
var
  fn: string;
  i: integer;
begin
  fn:='';
  for i:=0 to FileList.Count - 1 do
    fn:=fn + FileList[i] + #0;
  Result:=Sys_DeleteFilesToBin_Int(fn);
end;

function GetAssociatedIconHandle(FileName: string): HIcon;
var
  Extension: string;
  FileInfo: SHFILEINFO;
begin
    Extension:='*' + ExtractFileExt(FileName);
    {
      // Get description type
      SHGetFileInfo(PChar(Extension),
      FILE_ATTRIBUTE_NORMAL,
      FileInfo,
      SizeOf(FileInfo),
      SHGFI_TYPENAME or SHGFI_USEFILEATTRIBUTES
      );
      FDescriptions.Add(FileInfo.szTypeName);
      }

    // Get icon and copy into ImageList
    SHGetFileInfo(PChar(Extension), FILE_ATTRIBUTE_NORMAL, FileInfo, SizeOf(FileInfo),
      SHGFI_ICON or SHGFI_SMALLICON or SHGFI_SYSICONINDEX or SHGFI_USEFILEATTRIBUTES);
    Result:=FileInfo.HIcon;
end;

procedure LoadAssociatedIconAsBitmap(Bmp: TBitmap; FileName: string);
var
  Icon: TIcon;
begin
  Icon:=TIcon.Create;
  try
    Icon.Handle:=GetAssociatedIconHandle(FileName);
    Bmp.Width:=Icon.Width;
    Bmp.Height:=Icon.Height;
    Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
    Bmp.Canvas.Draw(0, 0, Icon);
  finally
    Icon.Free;
  end;
end;

end.
