unit rpictool;

{<
  rpictool unit contains set of graphic functions working with BMP, JPG, PNG and color + gradients.

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2010)
  @lastmod(May 2018)
}

interface

{$INCLUDE _rosi.inc}

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Math, Dialogs,
  GraphUtil,
{$IFDEF JPG_support} jpeg, {$ENDIF}
{$IFDEF PNG_support} pngimage, {$ENDIF}
  db, rstring, RosiCompGlobalCfg;

type
  { record type for one color definition }
  TColorDefMap = record
    Value: TColor;
    Name: string;
    LocalizedName: string;
  end;

  { list of graphics types }
  TrGraphicType = (gtNone, gtError, gtBMP, gtJPG, gtPNG, gtEMF, gtOther);

const
  { define number of steps for gradient background }
  cGradientStandardSteps = 100;
  { define standard JPEG quality, used by WritePictureToField,
    0..100   100=best quality but large file }
  JPEGDefaultQuality = 100;

  { define name array of graphics types }
  GraphicTypeLabel: array [TrGraphicType] of string = ('None', 'Error', 'BMP', 'JPG', 'PNG', 'EMF', 'Other');

  { define array of color records }
  ColorsDefArr: array [0 .. 20] of TColorDefMap = ((Value: clBlack; Name: sColorBlack; LocalizedName: sColorBlackLoc), (Value: clMaroon; Name: sColorMaroon;
      LocalizedName: sColorMaroonLoc), (Value: clGreen; Name: sColorGreen; LocalizedName: sColorGreenLoc), (Value: clOlive; Name: sColorOlive;
      LocalizedName: sColorOliveLoc), (Value: clNavy; Name: sColorNavy; LocalizedName: sColorNavyLoc), (Value: clPurple; Name: sColorPurple;
      LocalizedName: sColorPurpleLoc), (Value: clTeal; Name: sColorTeal; LocalizedName: sColorTealLoc), (Value: clGray; Name: sColorGray;
      LocalizedName: sColorGrayLoc), (Value: clSilver; Name: sColorSilver; LocalizedName: sColorSilverLoc), (Value: clRed; Name: sColorRed;
      LocalizedName: sColorRedLoc), (Value: clLime; Name: sColorLime; LocalizedName: sColorLimeLoc), (Value: clYellow; Name: sColorYellow;
      LocalizedName: sColorYellowLoc), (Value: clBlue; Name: sColorBlue; LocalizedName: sColorBlueLoc), (Value: clFuchsia; Name: sColorFuchsia;
      LocalizedName: sColorFuchsiaLoc), (Value: clAqua; Name: sColorAqua; LocalizedName: sColorAquaLoc), (Value: clWhite; Name: sColorWhite;
      LocalizedName: sColorWhiteLoc),

    (Value: clMoneyGreen; Name: sColorMoneyGreen; LocalizedName: sColorMoneyGreenLoc), (Value: clSkyBlue; Name: sColorSkyBlue; LocalizedName: sColorSkyBlueLoc)
      , (Value: clCream; Name: sColorCream; LocalizedName: sColorCreamLoc), (Value: clMedGray; Name: sColorMedGray; LocalizedName: sColorMedGrayLoc),

    (Value: clNone; Name: sColorCustom; LocalizedName: sColorCustomLoc));

  ColorDefStandardCount = 16;
  ColorDefExtendedCount = 4;
  ColorDefCustomIdx = 20;

  // color functions

{ load list of Colors to list }
function GetColorList(L: TStrings; AddExtendedColor, AddCustomColor, LocalizedName: boolean): integer;
{ return color value according color index }
function GetColorValue(idx: integer): TColor;
{ return color name or localozed name according color index }
function GetColorName(idx: integer; LocalizedName: boolean): string;
{ return color index according color value }
function GetColorIdx(Color: TColor): integer; overload;
{ return color index according color name }
function GetColorIdx(const ColorName: string): integer; overload;
{ return color name or hex string value according color value }
function ColorToString(Color: TColor; LocalizedName: boolean = false): string;
{ return color value according color string }
function StringToColor(const ColorStr: string): TColor;
{ compare two colors according provided threshold}
function IsSameColor(Color1, Color2: TColor; Threshold_Proc: integer = 80): boolean;

{ set font orientation (rotate font) }
procedure FontGetRotated(Font: TFont; Orientation: integer = 900);

{ modify Rect according to Canvas Orientation }
procedure BiDi_ModifyRect(Canvas: TCanvas; var ARect: TRect);
{ draw FrameRect according to Canvas Orientation }
procedure BiDi_FrameRect(Canvas: TCanvas; ARect: TRect; AColor: TColor = clNone);
{ draw FocusRect according to Canvas Orientation }
procedure BiDi_DrawFocusRect(Canvas: TCanvas; ARect: TRect);

{ draw vertical gradient to canvas }
procedure DrawGradientV(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor);
{ draw horizontal gradient to canvas }
procedure DrawGradientH(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor);
{ draw any gradient to canvas with definable number of steps }
procedure DrawGradient(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor; Vertical: boolean; MaxSteps: integer);

{ draw ProgressBar to the rect (used by StringGrid and DBGrid) }
function DrawDataBar(Canvas: TCanvas; ARect: TRect; Value, ValMin, ValMax: double; ColorFrom, ColorTo: TColor): double;
{ draw StatusBar to the rect (used by StringGrid and DBGrid) }
procedure DrawStatusBar(Canvas: TCanvas; ARect: TRect; Value, Steps: integer; BarType: TrStatusBarType; ColorAll, ColorDone: TColor);

{ animate resizing of Control }
procedure AnimateResize(W: TControl; NewRect: TRect);

{ assig picture to Bitmap or Draw to bitmap if Picture contains Metafile }
procedure GetBitmapFromPicture(B: TBitmap; P: TPicture);
{ draw picture to center of rectangle }
procedure DrawPictureCenter(Canvas: TCanvas; R: TRect; P: TPicture); overload;
{ draw picture to center of rectangle, definable strech and transparency }
procedure DrawPictureCenter(Canvas: TCanvas; R: TRect; P: TPicture; Stretch, Transparent: boolean); overload;
{ draw bitmap aligned of rectangle }
procedure DrawBmpAlign(Canvas: TCanvas; R: TRect; Bmp: TBitmap; Alignment: TAlignment; Margin: integer = 0);
{ draw bitmap to center of rectangle }
procedure DrawBmpCenter(Canvas: TCanvas; R: TRect; Bmp: TBitmap); overload;
{ draw bitmap to center of rectangle, definable strech and transparency }
procedure DrawBmpCenter(Canvas: TCanvas; R: TRect; Bmp: TBitmap; Stretch: boolean); overload;
{ draw bitmap to center of rectangle by HDC instead of Canvas, definable transparency }
procedure DrawBmpCenter(DC: HDC; R: TRect; Bmp: TBitmap; Transparent: boolean); overload;
{ draw picture to rectangle as tiles }
procedure DrawPictureTiles(Canvas: TCanvas; R: TRect; P: TPicture; Transparent: boolean);
{ draw bitmap to rectangle as tiles }
procedure DrawBmpTiles(Canvas: TCanvas; R: TRect; Bmp: TBitmap);

// resize and rotate BMP
{ resize bitmap to max width or height }
procedure SmoothResizeBitmap(Bmp: TBitmap; MaxW, MaxH: integer; SmallEnlarge: boolean);
{ resize one bitmap to another bitmap }
procedure SmoothResizeToBitmap(Src, Dst: TBitmap);
{ resize picture to max width or height }
procedure SmoothResizePicture(P: TPicture; MaxW, MaxH: integer; SmallEnlarge: boolean);
{ rotate bitmap by defined angle }
procedure RotateBitmap(Bmp: TBitmap; Angle: double);
{ rotate one bitmap to another bitmap by defined angle }
procedure RotateToBitmap(SrcBmp: TBitmap; DestBmp: TBitmap; Angle: double);
{ rotate picture by defined angle }
procedure RotatePicture(P: TPicture; Angle: double);

{ convert bitmap to icon }
procedure ConvertBmpToIco(Bmp: TBitmap; Icon: TIcon; IconSize: integer);
{ change bitmap pixels of one color to another color }
procedure BitmapChangeColor(SourceBmp, DestBmp: TBitmap; NewColor: TColor; OldColor: TColor = clBlack);

// returns state according conditional compilation
{ indicate whether JPG is supported }
function IsJPGSupported: boolean;
{ indicate whether PNG is supported }
function IsPNGSupported: boolean;
{ indicate whether GraphicType is not Empty or Error }
function IsGraphicTypeValid(GraphicType: TrGraphicType): boolean;

// read/write picture from/to db field
{ read field content and stored to Picture, return graphics type }
function ReadPictureFromField(F: TField; P: TPicture): TrGraphicType;
{ read field content and stored to Bitmap, return graphics type }
function ReadBitmapFromField(F: TField; Bmp: TBitmap): TrGraphicType;
{ write picture content to field in defined graphics type }
function WritePictureToField(F: TField; P: TPicture; FieldType: TrGraphicType): boolean;
{ write bitmap content to field in defined graphics type }
function WriteBitmapToField(F: TField; Bmp: TBitmap; FieldType: TrGraphicType): boolean;

// load/save picture from/to file
{ read file content and stored to Picture, return graphics type }
function LoadPictureFromFile(P: TPicture; const FileName: string; ShowErrorMessage: boolean = true): TrGraphicType;
{ read file content and stored to bitmap, return graphics type }
function LoadBitmapFromFile(Bmp: TBitmap; const FileName: string; ShowErrorMessage: boolean = true): TrGraphicType;
{ write picture content to stream in defined graphics type, return final graphics type }
function SavePictureToStream(P: TPicture; Stream: TStream; FileType: TrGraphicType): TrGraphicType;
{ write bitmap content to stream in defined graphics type, return final graphics type }
function SaveBitmapToStream(Bmp: TBitmap; Stream: TStream; FileType: TrGraphicType): TrGraphicType;
{ write picture content to file in defined graphics type, return final graphics type }
function SavePictureToFile(P: TPicture; const FileName: string; FileType: TrGraphicType): TrGraphicType;
{ write bitmap content to file in defined graphics type, return final graphics type }
function SaveBitmapToFile(Bmp: TBitmap; const FileName: string; FileType: TrGraphicType): TrGraphicType;

// open/save dialog function
{ return list of FileType filters used by Open/Save dialogs }
function GetFileTypeFilterList(UseAllFormats: boolean): string;
{ return file extension for defined GraphicType }
function GetFileTypeExtension(GraphicType: TrGraphicType): string;
{ return GraphicType according filename extension }
function GetFileTypeFromFileName(const FileName: string): TrGraphicType;
{ select graphic file from disk and load content to picture, return GraphicType }
function OpenDlg_PictureFromFile(P: TPicture; const DefaultDir: string; DefaultGrType: TrGraphicType): TrGraphicType;
{ select file name and save picture content to file, return real GraphicType }
function SaveDlg_PictureToFile(P: TPicture; const DefaultDir: string; DefaultGrType: TrGraphicType): TrGraphicType;
{ initialize and optionaly execute OpenDialog }
function OpenDlg_GraphicFiles(D: TOpenDialog; MultiSelect, Execute: boolean; const DefaultDir: string; DefaultGrType: TrGraphicType): boolean;

// JPEG specific function
{ read size of jpeg file }
function GetJPEGDimensions(const FileName: string; var X, Y: Word): boolean;
{ write picture content to field in jpg format and defined quality }
function WritePictureToJPGField(F: TField; P: TPicture; Quality: integer): boolean;
{ write bitmap content to field in jpg format and defined quality }
function WriteBitmapToJPGField(F: TField; Bmp: TBitmap; Quality: integer): boolean;
{ write picture content to jpg file in defined quality }
function SavePictureToJPGFile(P: TPicture; const FileName: string; Quality: integer): boolean;
{ write bitmap content to jpg file in defined quality }
function SaveBitmapToJPGFile(Bmp: TBitmap; const FileName: string; Quality: integer): boolean;

implementation

uses rdlg, rDialogsEx;

function GetColorList(L: TStrings; AddExtendedColor, AddCustomColor, LocalizedName: boolean): integer;
  function GetName(idx: integer): string;
  begin
    if LocalizedName then
      Result:=ColorsDefArr[idx].LocalizedName
    else
      Result:=ColorsDefArr[idx].Name;
  end;

var
  i: integer;
begin
  L.Clear;
  for i:=0 to ColorDefStandardCount - 1 do
    L.AddObject(GetName(i), TObject(i));
  if AddExtendedColor then
    for i:=ColorDefStandardCount to ColorDefStandardCount + ColorDefExtendedCount - 1 do
      L.AddObject(GetName(i), TObject(i));
  if AddCustomColor then
    L.AddObject(GetName(ColorDefCustomIdx), TObject(ColorDefCustomIdx));
  Result:=L.Count;
end;

function GetColorValue(idx: integer): TColor;
begin
  Result:=clNone;
  if idx in [0 .. High(ColorsDefArr)] then
    Result:=ColorsDefArr[idx].Value;
end;

function GetColorName(idx: integer; LocalizedName: boolean): string;
begin
  Result:='';
  if idx in [0 .. High(ColorsDefArr)] then
    if LocalizedName then
      Result:=ColorsDefArr[idx].LocalizedName
    else
      Result:=ColorsDefArr[idx].Name;
end;

function GetColorIdx(Color: TColor): integer;
var
  i: integer;
begin
  Result:=-1;
  for i:=0 to High(ColorsDefArr) do
    if ColorsDefArr[i].Value = Color then
    begin
      Result:=i;
      Exit;
    end;
end;

function GetColorIdx(const ColorName: string): integer;
var
  i: integer;
begin
  Result:=-1;
  for i:=0 to High(ColorsDefArr) do
    if SameText(ColorsDefArr[i].Name, ColorName) or SameText(ColorsDefArr[i].LocalizedName, ColorName) then
    begin
      Result:=i;
      Exit;
    end;
end;

function ColorToString(Color: TColor; LocalizedName: boolean = false): string;
var
  idx: integer;
begin
  idx:=GetColorIdx(Color);
  if idx >= 0 then
    Result:=GetColorName(idx, LocalizedName)
  else
    Result:='$' + IntToHex(ColorToRGB(Color), 8);
end;

function StringToColor(const ColorStr: string): TColor;
var
  idx: integer;
begin
  idx:=GetColorIdx(ColorStr);
  if idx >= 0 then
    Result:=GetColorValue(idx)
  else
    Result:=StrToIntDef(ColorStr, clNone);
end;

function IsSameColor(Color1, Color2: TColor; Threshold_Proc: integer = 80): boolean;
var dif: double;
begin
  Color1:=ColorToRGB(Color1);
  Color2:=ColorToRGB(Color2);
  dif:=sqrt(sqr(GetRValue(Color1)-GetRValue(Color2))+sqr(GetGValue(Color1)-GetGValue(Color2))+sqr(GetBValue(Color1)-GetBValue(Color2)));
  dif:=100*dif/sqrt(sqr(255)*3);
  Result:=dif<100-Threshold_Proc;
end;

procedure FontGetRotated(Font: TFont; Orientation: integer = 900);
var
  lf: TLogFont;
  tf: TFont;
begin
  tf:=TFont.Create;
  try
    tf.Assign(Font);
    tf.Name:=Font.Name;
    GetObject(tf.Handle, sizeof(lf), @lf);
    lf.lfEscapement:=Orientation;
    lf.lfOrientation:=Orientation;
    tf.Handle:=CreateFontIndirect(lf);
    Font.Assign(tf);
  finally
    FreeAndNil(tf);
  end;
end;

procedure BiDi_ModifyRect(Canvas: TCanvas; var ARect: TRect);
var hold: integer;
begin
  if (Canvas.CanvasOrientation = coRightToLeft) then
  begin
    hold:=ARect.Left;
    ARect.Left:=ARect.Right;
    ARect.Right:=hold;
  end;
end;

procedure BiDi_FrameRect(Canvas: TCanvas; ARect: TRect; AColor: TColor = clNone);
begin
  BiDi_ModifyRect(Canvas, ARect);
  if AColor<>clNone then
    Canvas.Brush.Color:=AColor;
  Canvas.FrameRect(ARect);
end;

procedure BiDi_DrawFocusRect(Canvas: TCanvas; ARect: TRect);
begin
  BiDi_ModifyRect(Canvas, ARect);
  Canvas.DrawFocusRect(ARect);
end;

procedure DrawGradientV(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor);
begin
  DrawGradient(Canvas, ARect, FromColor, ToColor, true, cGradientStandardSteps);
end;

procedure DrawGradientH(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor);
begin
  DrawGradient(Canvas, ARect, FromColor, ToColor, false, cGradientStandardSteps);
end;

procedure DrawGradient(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor; Vertical: boolean; MaxSteps: integer);
var
  Dir: TGradientDirection;
begin
  if Vertical then
    Dir:=gdVertical
  else
    Dir:=gdHorizontal;
  GraphUtil.GradientFillCanvas(Canvas, FromColor, ToColor, ARect, Dir);
end;

{
procedure DrawGradient(Canvas: TCanvas; ARect: TRect; FromColor, ToColor: TColor; Vertical: boolean; MaxSteps: integer);
var
  dr, dg, db: double;
  r1, r2, g1, g2, b1, b2: byte;
  R, G, B: byte;
  Steps, cnt, i: integer;
  DrawRect: TRect;
begin
  FromColor:=ColorToRGB(FromColor);
  ToColor:=ColorToRGB(ToColor);
  if (FromColor = ToColor) or (ARect.Right - ARect.Left < 3) or (ARect.Bottom - ARect.Top < 3) then
  begin
    Canvas.Brush.Color:=FromColor;
    Canvas.FillRect(ARect);
    Exit;
  end;

  r1:=GetRValue(FromColor);
  g1:=GetGValue(FromColor);
  b1:=GetBValue(FromColor);
  r2:=GetRValue(ToColor);
  g2:=GetGValue(ToColor);
  b2:=GetBValue(ToColor);
  if Vertical then
    cnt:=(ARect.Bottom - ARect.Top)
  else
    cnt:=(ARect.Right - ARect.Left);
  if cnt < MaxSteps then
    Steps:=cnt
  else
    Steps:=MaxSteps;
  dr:=(r2 - r1) / Steps;
  dg:=(g2 - g1) / Steps;
  db:=(b2 - b1) / Steps;
  for i:=0 to Steps - 1 do
  begin
    R:=r1 + Ceil(dr * i);
    G:=g1 + Ceil(dg * i);
    B:=b1 + Ceil(db * i);
    if Vertical then
    begin
      DrawRect.Left:=ARect.Left;
      DrawRect.Right:=ARect.Right;
      DrawRect.Top:=ARect.Top + Ceil(cnt * i / Steps);
      DrawRect.Bottom:=DrawRect.Top + Ceil(cnt * (i + 1) / Steps);
      if DrawRect.Bottom > ARect.Bottom then
        DrawRect.Bottom:=ARect.Bottom;
    end
    else
    begin
      DrawRect.Top:=ARect.Top;
      DrawRect.Bottom:=ARect.Bottom;
      DrawRect.Left:=ARect.Left + Ceil(cnt * i / Steps);
      DrawRect.Right:=DrawRect.Left + Ceil(cnt * (i + 1) / Steps);
      if DrawRect.Right > ARect.Right then
        DrawRect.Right:=ARect.Right;
    end;
    Canvas.Brush.Color:=RGB(R, G, B);
    Canvas.FillRect(DrawRect);
  end;
end;
}

procedure MakeSquare(var R: TRect);
begin
  if (R.Right - R.Left) > (R.Bottom - R.Top) then
  begin
    R.Left:=R.Left + (R.Right - R.Left - (R.Bottom - R.Top)) div 2;
    R.Right:=R.Left + R.Bottom - R.Top;
  end;
  if (R.Right - R.Left) < (R.Bottom - R.Top) then
  begin
    R.Top:=R.Top + (R.Bottom - R.Top - (R.Right - R.Left)) div 2;
    R.Bottom:=R.Top + (R.Right - R.Left);
  end;
end;

function DrawDataBar(Canvas: TCanvas; ARect: TRect; Value, ValMin, ValMax: double; ColorFrom, ColorTo: TColor): double;
var
  perc: double;
begin
  if ColorFrom = clNone then
    ColorFrom:=clSkyBlue;
  if ColorTo = clNone then
    ColorTo:=clWhite;
  Value:=Min(Max(Value, ValMin), ValMax);
  if (ValMin = ValMax) then
    perc:=0
  else
    perc:=(Value - ValMin) / (ValMax - ValMin);
  if perc > 0 then
  begin
    InflateRect(ARect, -1, -1);
    ARect.Right:=ARect.Left + Round((ARect.Right - ARect.Left) * perc);
    Canvas.Brush.Color:=ColorFrom;
    Canvas.Brush.Style:=bsSolid;
    if ColorFrom = ColorTo then
      Canvas.FillRect(ARect)
    else
    begin
      rpictool.DrawGradientH(Canvas, ARect, ColorFrom, ColorTo);
      BiDi_FrameRect(Canvas, ARect, ColorFrom);
    end;
  end;
  Result:=perc;
end;

procedure DrawStatusBar(Canvas: TCanvas; ARect: TRect; Value, Steps: integer; BarType: TrStatusBarType; ColorAll, ColorDone: TColor);
var
  sr: TRect;
  i, H, W: integer;
  radius: double;
begin
  Value:=Min(Max(Value, 0), Steps);
  if ColorAll = clNone then
    ColorAll:=clSilver;
  if ColorDone = clNone then
    ColorDone:=clTeal;
  InflateRect(ARect, -1, -1);
  H:=ARect.Bottom - ARect.Top;
  W:=ARect.Right - ARect.Left;
  Canvas.Brush.Style:=bsSolid;
  for i:=1 to Steps do
  begin
    if Value >= i then
      Canvas.Brush.Color:=ColorDone
    else
      Canvas.Brush.Color:=ColorAll;
    Canvas.Pen.Color:=ColorAll;
    sr:=ARect;
    case BarType of
      sbtProgressBar:
        begin
          // progress bar
          sr.Right:=sr.Left + Round(W * Value / Steps);
          Canvas.Brush.Color:=ColorDone;
          Canvas.FillRect(sr);
          sr.Left:=sr.Right;
          sr.Right:=ARect.Right;
          Canvas.Brush.Color:=ColorAll;
          Canvas.FillRect(sr);
          BiDi_FrameRect(Canvas, ARect, ColorAll);
          Break;
        end;
      sbtDots:
        begin
          // dots
          InflateRect(sr, -1, -1);
          sr.Left:=ARect.Left + (W div Steps) * (i - 1) + 1;
          sr.Right:=sr.Left + (W div Steps) - 3;
          Canvas.FillRect(sr);
        end;
      sbtCircles:
        begin
          // circles
          InflateRect(sr, -1, -1);
          sr.Left:=ARect.Left + (W div Steps) * (i - 1) + 1;
          sr.Right:=sr.Left + (W div Steps) - 2;
          MakeSquare(sr);
          Canvas.Ellipse(sr);
        end;
      sbtEllipses:
        begin
          // ellipses
          InflateRect(sr, -1, -1);
          sr.Left:=ARect.Left + (W div Steps) * (i - 1) + 1;
          sr.Right:=sr.Left + (W div Steps) - 3;
          Canvas.Ellipse(sr);
        end;
      sbtSignalStrength:
        begin
          // signal
          sr.Left:=ARect.Left + (W - Steps * 3) div 2 + (i - 1) * 3;
          sr.Right:=sr.Left + 2;
          sr.Top:=sr.Bottom - H * i div Steps;
          Canvas.FillRect(sr);
        end;
      sbtPie:
        begin
          // pie
          MakeSquare(sr);
          radius:=(sr.Right - sr.Left) / 2;
          Canvas.Brush.Color:=ColorAll;
          if Value < Steps then
            Canvas.Ellipse(sr);
          Canvas.Brush.Color:=ColorDone;
          if Value = Steps then
            Canvas.Ellipse(sr)
          else
            if Value > 0 then
              Canvas.Pie(sr.Left, sr.Bottom, sr.Right, sr.Top, Round(sr.Left + radius + radius * cos(2.0 * pi * Value / Steps - pi / 2)), Round
                  (sr.Top + radius + radius * sin(2.0 * pi * Value / Steps - pi / 2)), sr.Left + (sr.Right - sr.Left) div 2, sr.Top);
          Break;
        end;
    end;
  end;
  Canvas.Brush.Style:=bsClear;
end;

procedure AnimateResize(W: TControl; NewRect: TRect);
var
  i, Steps: integer;
  dx1, dx2, dy1, dy2: double;
  OldRect: TRect;
begin
  Steps:=10;
  OldRect:=W.BoundsRect;
  dx1:=(NewRect.Left - OldRect.Left) / Steps;
  dx2:=(NewRect.Right - OldRect.Right) / Steps;
  dy1:=(NewRect.Top - OldRect.Top) / Steps;
  dy2:=(NewRect.Bottom - OldRect.Bottom) / Steps;
  for i:=1 to Steps do
  begin
    NewRect.Left:=Round(OldRect.Left + i * dx1);
    NewRect.Right:=Round(OldRect.Right + i * dx2);
    NewRect.Top:=Round(OldRect.Top + i * dy1);
    NewRect.Bottom:=Round(OldRect.Bottom + i * dy2);
    W.BoundsRect:=NewRect;
    W.Update;
    Sleep(10);
  end;
end;

procedure GetBitmapFromPicture(B: TBitmap; P: TPicture);
begin
  if P.Graphic is TMetaFile then
  begin
    B.Width:=P.Graphic.Width;
    B.Height:=P.Graphic.Height;
    B.Canvas.Draw(0, 0, P.Graphic);
  end
  else
    B.Assign(P.Graphic);
end;

procedure DrawPictureCenter(Canvas: TCanvas; R: TRect; P: TPicture);
begin
  DrawPictureCenter(Canvas, R, P, false, false);
end;

procedure DrawPictureCenter(Canvas: TCanvas; R: TRect; P: TPicture; Stretch, Transparent: boolean);
var
  B: TBitmap;
begin
  if P.Width = 0 then
    Exit;
  B:=TBitmap.Create;
  try
    GetBitmapFromPicture(B, P);
    B.Transparent:=Transparent;
    DrawBmpCenter(Canvas, R, B, Stretch);
  finally
    FreeAndNil(B);
  end;
end;

procedure DrawBmpAlign(Canvas: TCanvas; R: TRect; Bmp: TBitmap; Alignment: TAlignment; Margin: integer = 0);
begin
  if R.Right - R.Left - 2 * Margin > Bmp.Width then
  begin
    case Alignment of
      taLeftJustify:
        R.Right:=R.Left + Bmp.Width + 2 * Margin;
      taRightJustify:
        R.Left:=R.Right - Bmp.Width - 2 * Margin;
    end;
  end;
  DrawBmpCenter(Canvas, R, Bmp);
end;

procedure DrawBmpCenter(Canvas: TCanvas; R: TRect; Bmp: TBitmap);
begin
  DrawBmpCenter(Canvas, R, Bmp, false);
end;

procedure DrawBmpCenter(Canvas: TCanvas; R: TRect; Bmp: TBitmap; Stretch: boolean);
var
  W, H: integer;
begin
  if (Bmp.Width = 0) or (R.Right - R.Left < 3) or (R.Bottom - R.Top < 3) then
    Exit;
  if Stretch then
    Canvas.StretchDraw(R, Bmp)
  else
  {
    // canvas draw is slow, use BitBlt directly to speed up
    if (Bmp.Width <= R.Right - R.Left) and (Bmp.Height <= R.Bottom - R.Top) then
    begin
    // draw center
    Canvas.Draw((R.Right - R.Left - Bmp.Width) div 2 + R.Left, (R.Bottom - R.Top - Bmp.Height) div 2 + R.Top, Bmp)
    end
    else
    begin
    // crop bitmap to center rect
    B:=TBitmap.Create;
    try
    B.Width:=Min(Bmp.Width, R.Right - R.Left);
    B.Height:=Min(Bmp.Height, R.Bottom - R.Top);
    BitBlt(B.Canvas.Handle, 0, 0, B.Width, B.Height, Bmp.Canvas.Handle, Max(0, (Bmp.Width - B.Width) div 2), Max(0, (Bmp.Height - B.Height) div 2),
    SRCCOPY);
    B.Transparent:=Bmp.Transparent;
    // draw center
    Canvas.Draw((R.Right - R.Left - B.Width) div 2 + R.Left, (R.Bottom - R.Top - B.Height) div 2 + R.Top, B);
    finally
    B.Free;
    end;
    end;
    { }
  begin
    W:=Min(Bmp.Width, R.Right - R.Left);
    H:=Min(Bmp.Height, R.Bottom - R.Top);
    if Bmp.Transparent then
      TransparentBlt(Canvas.Handle, Max((R.Right - R.Left - Bmp.Width) div 2 + R.Left, R.Left), Max((R.Bottom - R.Top - Bmp.Height) div 2 + R.Top, R.Top), W,
        H, Bmp.Canvas.Handle, Max(0, (Bmp.Width - W) div 2), Max(0, (Bmp.Height - H) div 2), W, H, Bmp.TransparentColor)
    else
      BitBlt(Canvas.Handle, Max((R.Right - R.Left - Bmp.Width) div 2 + R.Left, R.Left), Max((R.Bottom - R.Top - Bmp.Height) div 2 + R.Top, R.Top), W, H,
        Bmp.Canvas.Handle, Max(0, (Bmp.Width - W) div 2), Max(0, (Bmp.Height - H) div 2), SRCCOPY);
  end;
end;

procedure DrawBmpCenter(DC: HDC; R: TRect; Bmp: TBitmap; Transparent: boolean); overload;
var
  X, Y: integer;
begin
  X:=(R.Right - R.Left - Bmp.Width) div 2 + R.Left;
  Y:=(R.Bottom - R.Top - Bmp.Height) div 2 + R.Top;
  if Transparent then
    TransparentBlt(DC, X, Y, Bmp.Width, Bmp.Height, Bmp.Canvas.Handle, 0, 0, Bmp.Width, Bmp.Height, Bmp.Canvas.Pixels[0, Bmp.Height - 1])
  else
    BitBlt(DC, X, Y, Bmp.Width, Bmp.Height, Bmp.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure DrawPictureTiles(Canvas: TCanvas; R: TRect; P: TPicture; Transparent: boolean);
var
  B: TBitmap;
begin
  if P.Width = 0 then
    Exit;
  B:=TBitmap.Create;
  try
    GetBitmapFromPicture(B, P);
    B.Transparent:=Transparent;
    DrawBmpTiles(Canvas, R, B);
  finally
    FreeAndNil(B);
  end;
end;

procedure DrawBmpTiles(Canvas: TCanvas; R: TRect; Bmp: TBitmap);
var
  X, Y: integer;
begin
  if Bmp.Width = 0 then
    Exit;
  for X:=0 to (R.Right - R.Left) div Bmp.Width do
    for Y:=0 to (R.Bottom - R.Top) div Bmp.Height do
      Canvas.Draw(R.Left + Bmp.Width * X, R.Top + Bmp.Height * Y, Bmp);
end;

procedure SmoothResizeBitmap(Bmp: TBitmap; MaxW, MaxH: integer; SmallEnlarge: boolean);
var
  NewB: TBitmap;
  NewW, NewH: integer;
begin
  NewW:=Bmp.Width;
  if (MaxW > 0) and ((NewW > MaxW) or SmallEnlarge) then
    NewW:=MaxW;
  NewH:=MulDiv(NewW, Bmp.Height, Bmp.Width);
  if (MaxH > 0) and ((NewH > MaxH) or ((MaxW = 0) and SmallEnlarge)) then
  begin
    NewH:=MaxH;
    NewW:=MulDiv(NewH, Bmp.Width, Bmp.Height);
  end;
  if NewW <> Bmp.Width then
  begin
    NewB:=TBitmap.Create;
    try
      NewB.Width:=NewW;
      NewB.Height:=NewH;
      SmoothResizeToBitmap(Bmp, NewB);
      Bmp.Assign(NewB);
    finally
      FreeAndNil(NewB);
    end;
  end;
end;

procedure SmoothResizeToBitmap(Src, Dst: TBitmap);
type
  TRGBArray = array [Word] of TRGBTriple;
  pRGBArray = ^TRGBArray;
var
  X, Y: integer;
  xP, yP: integer;
  xP2, yP2: integer;
  SrcLine1, SrcLine2: pRGBArray;
  t3: integer;
  z, z2, iz2: integer;
  DstLine: pRGBArray;
  DstGap: integer;
  w1, w2, w3, w4: integer;
begin
  Src.PixelFormat:=pf24Bit;
  Dst.PixelFormat:=pf24Bit;

  if (Src.Width = Dst.Width) and (Src.Height = Dst.Height) then
    Dst.Assign(Src)
  else
  begin
    DstLine:=Dst.ScanLine[0];
    DstGap:=integer(Dst.ScanLine[1]) - integer(DstLine);

    xP2:=MulDiv(pred(Src.Width), $10000, Dst.Width);
    yP2:=MulDiv(pred(Src.Height), $10000, Dst.Height);
    yP:=0;

    for Y:=0 to pred(Dst.Height) do
    begin
      xP:=0;

      SrcLine1:=Src.ScanLine[yP shr 16];

      if (yP shr 16 < pred(Src.Height)) then
        SrcLine2:=Src.ScanLine[succ(yP shr 16)]
      else
        SrcLine2:=Src.ScanLine[yP shr 16];

      z2:=succ(yP and $FFFF);
      iz2:=succ((not yP) and $FFFF);
      for X:=0 to pred(Dst.Width) do
      begin
        t3:=xP shr 16;
        z:=xP and $FFFF;
        w2:=MulDiv(z, iz2, $10000);
        w1:=iz2 - w2;
        w4:=MulDiv(z, z2, $10000);
        w3:=z2 - w4;
        DstLine[X].rgbtRed:=(SrcLine1[t3].rgbtRed * w1 + SrcLine1[t3 + 1].rgbtRed * w2 + SrcLine2[t3].rgbtRed * w3 + SrcLine2[t3 + 1].rgbtRed * w4) shr 16;
        DstLine[X].rgbtGreen:=(SrcLine1[t3].rgbtGreen * w1 + SrcLine1[t3 + 1].rgbtGreen * w2 +

            SrcLine2[t3].rgbtGreen * w3 + SrcLine2[t3 + 1].rgbtGreen * w4) shr 16;
        DstLine[X].rgbtBlue:=(SrcLine1[t3].rgbtBlue * w1 + SrcLine1[t3 + 1].rgbtBlue * w2 + SrcLine2[t3].rgbtBlue * w3 + SrcLine2[t3 + 1].rgbtBlue * w4) shr 16;
        Inc(xP, xP2);
      end;
      Inc(yP, yP2);
      DstLine:=pRGBArray(integer(DstLine) + DstGap);
    end;
  end;
end;

procedure SmoothResizePicture(P: TPicture; MaxW, MaxH: integer; SmallEnlarge: boolean);
var
  B: TBitmap;
begin
  if P.Width = 0 then
    Exit;
  B:=TBitmap.Create;
  try
    GetBitmapFromPicture(B, P);
    B.Transparent:=true;
    SmoothResizeBitmap(B, MaxW, MaxH, SmallEnlarge);
    P.Graphic.Assign(B);
  finally
    FreeAndNil(B);
  end;
end;

procedure IntRotateBitmap(var hBitmapDC: THandle; var lWidth: Longint; var lHeight: Longint; lRadians: double);
var
  i: Longint; // loop counter
  J: Longint; // loop counter
  hNewBitmapDC: THandle; // DC of the new bitmap
  hNewBitmap: THandle; // handle to the new bitmap
  lSine: double; // sine used in rotation
  lCosine: double; // cosine used in rotation
  X1: Longint; // used in calculating new
  // bitmap dimensions
  X2: Longint; // used in calculating new
  // bitmap dimensions
  X3: Longint; // used in calculating new
  // bitmap dimensions
  Y1: Longint; // used in calculating new
  // bitmap dimensions
  Y2: Longint; // used in calculating new
  // bitmap dimensions
  Y3: Longint; // used in calculating new
  // bitmap dimensions
  lMinX: Longint; // used in calculating new
  // bitmap dimensions
  lMaxX: Longint; // used in calculating new
  // bitmap dimensions
  lMinY: Longint; // used in calculating new
  // bitmap dimensions
  lMaxY: Longint; // used in calculating new
  // bitmap dimensions
  lNewWidth: Longint; // width of new bitmap
  lNewHeight: Longint; // height of new bitmap
  lSourceX: Longint; // x pixel coord we are blitting
  // from the source  image
  lSourceY: Longint; // y pixel coord we are blitting
  // from the source image

begin
  // create a compatible DC from the one just brought
  // into this function
  hNewBitmapDC:=CreateCompatibleDC(hBitmapDC);
  // compute the sine/cosinse of the radians used to
  // rotate this image
  lRadians:=lRadians;
  lSine:=sin(lRadians);
  lCosine:=cos(lRadians);
  // compute the size of the new bitmap being created
  Dec(lWidth);
  Dec(lHeight);
  X1:=Round(-lHeight * lSine);
  Y1:=Round(lHeight * lCosine);
  X2:=Round(lWidth * lCosine - lHeight * lSine);
  Y2:=Round(lHeight * lCosine + lWidth * lSine);
  X3:=Round(lWidth * lCosine);
  Y3:=Round(lWidth * lSine);
  Inc(lWidth);
  Inc(lHeight);
  // figure out the max/min size of the new bitmap
  lMinX:=Min(0, Min(X1, Min(X2, X3)));
  lMinY:=Min(0, Min(Y1, Min(Y2, Y3)));
  lMaxX:=Max(X1, Max(X2, X3));
  lMaxY:=Max(Y1, Max(Y2, Y3));
  // set the new bitmap width/height
  lNewWidth:=lMaxX - lMinX + 1;
  lNewHeight:=lMaxY - lMinY + 1;
  // create a new bitmap based upon the new width/height of the
  // rotated bitmap
  hNewBitmap:=CreateCompatibleBitmap(hBitmapDC, lNewWidth, lNewHeight);
  // attach the new bitmap to the new device context created
  // above before constructing the rotated bitmap
  SelectObject(hNewBitmapDC, hNewBitmap);
  // loop through and translate each pixel to its new location.
  // this is using a standard rotation algorithm
  For i:=0 To lNewHeight do
  begin
    For J:=0 To lNewWidth do
    begin
      lSourceX:=Round((J + lMinX) * lCosine + (i + lMinY) * lSine);
      lSourceY:=Round((i + lMinY) * lCosine - (J + lMinX) * lSine);
      If (lSourceX >= 0) And (lSourceX < lWidth) And (lSourceY >= 0) And (lSourceY < lHeight) Then
        BitBlt(hNewBitmapDC, J, i, 1, 1, hBitmapDC, lSourceX, lSourceY, SRCCOPY)
      else
        SetPixel(hNewBitmapDC, J, i, clWhite);
    end;
  end;

  // reset the new bitmap width and height
  lWidth:=lNewWidth;
  lHeight:=lNewHeight;
  // return the DC to the new bitmap
  hBitmapDC:=hNewBitmapDC;
  // destroy the bitmap created
  DeleteObject(hNewBitmap);
end;

procedure RotateBitmap(Bmp: TBitmap; Angle: double);
var
  B: TBitmap;
begin
  B:=TBitmap.Create;
  try
    RotateToBitmap(Bmp, B, Angle);
    Bmp.Width:=B.Width;
    Bmp.Height:=B.Height;
    Bmp.Assign(B);
  finally
    FreeAndNil(B);
  end;
end;

procedure RotateToBitmap(SrcBmp: TBitmap; DestBmp: TBitmap; Angle: double);
var
  DC: THandle;
  H, W: integer;
begin
  DC:=SrcBmp.Canvas.Handle;
  H:=SrcBmp.Height;
  W:=SrcBmp.Width;
  IntRotateBitmap(DC, W, H, pi * (Angle) / 180);
  DestBmp.Width:=W;
  DestBmp.Height:=H;
  BitBlt(DestBmp.Canvas.Handle, 0, 0, W, H, DC, 0, 0, SRCCOPY);
end;

procedure RotatePicture(P: TPicture; Angle: double);
var
  B: TBitmap;
begin
  if P.Width = 0 then
    Exit;
  B:=TBitmap.Create;
  try
    GetBitmapFromPicture(B, P);
    RotateBitmap(B, Angle);
    P.Graphic.Assign(B);
  finally
    FreeAndNil(B);
  end;
end;

procedure ConvertBmpToIco(Bmp: TBitmap; Icon: TIcon; IconSize: integer);
var
  TempBmp: TBitmap;
  ImageList: TImageList;
begin
  TempBmp:=TBitmap.Create;
  ImageList:=TImageList.CreateSize(IconSize, IconSize);
  try
    TempBmp.Width:=IconSize;
    TempBmp.Height:=IconSize;
    TempBmp.Canvas.Brush.Color:=clBtnFace;
    TempBmp.Canvas.FillRect(TempBmp.Canvas.ClipRect);
    DrawBmpCenter(TempBmp.Canvas, Rect(0, 0, IconSize, IconSize), Bmp);
    ImageList.AddMasked(TempBmp, TempBmp.TransparentColor);
    ImageList.GetIcon(0, Icon);
  finally
    FreeAndNil(TempBmp);
    FreeAndNil(ImageList);
  end;
end;

procedure BitmapChangeColor(SourceBmp, DestBmp: TBitmap; NewColor: TColor; OldColor: TColor = clBlack);
var
  DrawR: TRect;
begin
  DestBmp.Transparent:=SourceBmp.Transparent;
  DestBmp.Width:=SourceBmp.Width;
  DestBmp.Height:=SourceBmp.Height;
  DrawR:=Classes.Rect(0, 0, DestBmp.Width, DestBmp.Height);
  // replace white by nearly white to keep transparency
  if NewColor=clWhite then
    NewColor:=RGB(220,220,220);
  DestBmp.Canvas.Brush.Color:=NewColor;
  DestBmp.Canvas.BrushCopy(DrawR, SourceBmp, DrawR, OldColor);
end;

// ========================================================================
// database function
function IsJPGSupported: boolean;
begin
{$IFDEF JPG_support}
  Result:=true;
{$ELSE}
  Result:=false;
{$ENDIF}
end;

function IsPNGSupported: boolean;
begin
{$IFDEF PNG_support}
  Result:=true;
{$ELSE}
  Result:=false;
{$ENDIF}
end;

function IsGraphicTypeValid(GraphicType: TrGraphicType): boolean;
begin
  Result:=not(GraphicType in [gtNone, gtError]);
end;

function ReadPictureFromField(F: TField; P: TPicture): TrGraphicType;
var
  BmpImage: TBitmap;
  WMFImage: TMetaFile;
{$IFDEF JPG_support} JPEGImage: TJPEGImage; {$ENDIF}
{$IFDEF PNG_support} pngimage: TPNGImage; {$ENDIF}
begin
  if F.IsNull then
  begin
    Result:=gtNone;
    Exit;
  end;
  Result:=gtError;
  // try BMP
  BmpImage:=TBitmap.Create;
  try
    try
      BmpImage.Assign(F);
      P.Assign(BmpImage);
      Result:=gtBMP;
    except
    end;
  finally
    FreeAndNil(BmpImage);
  end;
{$IFDEF JPG_support}
  if Result = gtError then
  begin
    // try JPG
    JPEGImage:=TJPEGImage.Create;
    try
      try
        JPEGImage.Assign(F);
        P.Assign(JPEGImage);
        Result:=gtJPG;
      except
      end;
    finally
      FreeAndNil(JPEGImage);
    end;
  end;
{$ENDIF}
{$IFDEF PNG_support}
  if Result = gtError then
  begin
    // try PNG
    pngimage:=TPNGImage.Create;
    try
      try
        pngimage.Assign(F);
        P.Assign(pngimage);
        Result:=gtPNG;
      except
      end;
    finally
      FreeAndNil(pngimage);
    end;
  end;
{$ENDIF}
  if Result = gtError then
  begin
    // try WMF
    WMFImage:=TMetaFile.Create;
    try
      try
        WMFImage.Assign(F);
        P.Assign(WMFImage);
        Result:=gtEMF;
      except
      end;
    finally
      FreeAndNil(WMFImage);
    end;
  end;
end;

function ReadBitmapFromField(F: TField; Bmp: TBitmap): TrGraphicType;
var
  P: TPicture;
begin
  Bmp.Width:=0;
  P:=TPicture.Create;
  try
    Result:=ReadPictureFromField(F, P);
    case Result of
      gtBMP:
        Bmp.Assign(P.Bitmap);
      gtJPG, gtPNG, gtEMF, gtOther:
        GetBitmapFromPicture(Bmp, P);
    end;
  finally
    FreeAndNil(P);
  end;
end;

function WritePictureToField_Int(F: TField; P: TPicture; FieldType: TrGraphicType; JpegQ: integer): boolean;
var
  BmpImage: TBitmap;
{$IFDEF JPG_support} JPEGImage: TJPEGImage; {$ENDIF}
{$IFDEF PNG_support} pngimage: TPNGImage; {$ENDIF}
begin
  Result:=false;
  try
    BmpImage:=TBitmap.Create;
    try
      case FieldType of
        gtNone:
          F.Clear;
        gtBMP:
          begin
            GetBitmapFromPicture(BmpImage, P);
            if BmpImage.Empty then
              F.Clear
            else
              F.Assign(BmpImage);
          end;
{$IFDEF JPG_support}
        gtJPG:
          begin
            JPEGImage:=TJPEGImage.Create;
            try
              try
                JPEGImage.Assign(P.Graphic);
              except
                GetBitmapFromPicture(BmpImage, P);
                JPEGImage.Assign(BmpImage);
              end;
              if JPEGImage.Empty then
                F.Clear
              else
              begin
                JPEGImage.CompressionQuality:=JpegQ;
                if JPEGImage.CompressionQuality <> 100 then
                  JPEGImage.Compress;
                F.Assign(JPEGImage);
              end;
            finally
              FreeAndNil(JPEGImage);
            end;
          end;
{$ENDIF}
{$IFDEF PNG_support}
        gtPNG:
          begin
            pngimage:=TPNGImage.Create;
            try
              try
                pngimage.Assign(P.Graphic);
              except
                GetBitmapFromPicture(BmpImage, P);
                pngimage.Assign(BmpImage);
              end;
              if pngimage.Empty then
                F.Clear
              else
                F.Assign(pngimage);
            finally
              FreeAndNil(pngimage);
            end;
          end;
{$ENDIF}
        gtEMF, gtOther:
          begin
            GetBitmapFromPicture(BmpImage, P);
            F.Assign(BmpImage);
          end;
      else
        raise Exception.Create(sPictureErr);
      end;
      Result:=true;
    finally
      FreeAndNil(BmpImage);
    end;
  except
    on E: Exception do
      ShowErr('', E);
  end;
end;

function WritePictureToField(F: TField; P: TPicture; FieldType: TrGraphicType): boolean;
begin
  Result:=WritePictureToField_Int(F, P, FieldType, JPEGDefaultQuality);
end;

function WriteBitmapToField(F: TField; Bmp: TBitmap; FieldType: TrGraphicType): boolean;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  try
    P.Assign(Bmp);
    Result:=WritePictureToField(F, P, FieldType);
  finally
    FreeAndNil(P);
  end;
end;

function SavePictureToStream_Int(P: TPicture; Stream: TStream; FileType: TrGraphicType; JpegQ: integer): TrGraphicType;
var
  BmpImage: TBitmap;
{$IFDEF JPG_support} JPEGImage: TJPEGImage; {$ENDIF}
{$IFDEF PNG_support} pngimage: TPNGImage; {$ENDIF}
begin
  Result:=gtNone;
  if FileType = gtNone then
    FileType:=gtBMP;
  // if FileType=gtError then Exit;
  try
    BmpImage:=TBitmap.Create;
    try
      case FileType of
        gtBMP:
          begin
            BmpImage.Assign(P.Graphic);
            BmpImage.SaveToStream(Stream);
          end;
{$IFDEF JPG_support}
        gtJPG:
          begin
            JPEGImage:=TJPEGImage.Create;
            try
              try
                JPEGImage.Assign(P.Graphic);
              except
                BmpImage.Assign(P.Graphic);
                JPEGImage.Assign(BmpImage);
              end;
              JPEGImage.CompressionQuality:=JpegQ;
              if JPEGImage.CompressionQuality <> 100 then
                JPEGImage.Compress;
              JPEGImage.SaveToStream(Stream);
            finally
              FreeAndNil(JPEGImage);
            end;
          end;
{$ENDIF}
{$IFDEF PNG_support}
        gtPNG:
          begin
            pngimage:=TPNGImage.Create;
            try
              try
                pngimage.Assign(P.Graphic);
              except
                BmpImage.Assign(P.Graphic);
                pngimage.Assign(BmpImage);
              end;
              pngimage.SaveToStream(Stream);
            finally
              FreeAndNil(pngimage);
            end;
          end;
{$ENDIF}
        gtEMF, gtOther:
          begin
            try
              P.Graphic.SaveToStream(Stream);
            except
              BmpImage.Assign(P.Graphic);
              BmpImage.SaveToStream(Stream);
            end;
          end;
      else
        raise Exception.Create(sPictureErr);
      end;
      Result:=FileType;
    finally
      FreeAndNil(BmpImage);
    end;
  except
    on E: Exception do
      ShowErr('', E);
  end;
end;

function SavePictureToStream(P: TPicture; Stream: TStream; FileType: TrGraphicType): TrGraphicType;
begin
  Result:=SavePictureToStream_Int(P, Stream, FileType, JPEGDefaultQuality);
end;

function SaveBitmapToStream(Bmp: TBitmap; Stream: TStream; FileType: TrGraphicType): TrGraphicType;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  try
    P.Assign(Bmp);
    Result:=SavePictureToStream(P, Stream, FileType);
  finally
    FreeAndNil(P);
  end;
end;

function SavePictureToFile(P: TPicture; const FileName: string; FileType: TrGraphicType): TrGraphicType;
var
  Stream: TStream;
begin
  Stream:=TFileStream.Create(FileName, fmCreate);
  RosiCompConfig.rDBToolCfg.rpictool_LastFileName:=FileName;
  try
    Result:=SavePictureToStream(P, Stream, FileType);
  finally
    FreeAndNil(Stream);
  end;
end;

function SaveBitmapToFile(Bmp: TBitmap; const FileName: string; FileType: TrGraphicType): TrGraphicType;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  try
    P.Assign(Bmp);
    Result:=SavePictureToFile(P, FileName, FileType);
    RosiCompConfig.rDBToolCfg.rpictool_LastFileName:=FileName;
  finally
    FreeAndNil(P);
  end;
end;

function LoadPictureFromFile(P: TPicture; const FileName: string; ShowErrorMessage: boolean = true): TrGraphicType;
var
  BmpImage: TBitmap;
{$IFDEF JPG_support} JPEGImage: TJPEGImage; {$ENDIF}
{$IFDEF PNG_support} pngimage: TPNGImage; {$ENDIF}
  FileType: TrGraphicType;
begin
  Result:=gtError;
  if not FileExists(FileName) then
    Exit;
  FileType:=GetFileTypeFromFileName(FileName);
  // if FileType=gtError then Exit;
  try
    case FileType of
      gtBMP:
        begin
          BmpImage:=TBitmap.Create;
          try
            BmpImage.LoadFromFile(FileName);
            P.Assign(BmpImage);
          finally
            FreeAndNil(BmpImage);
          end;
        end;
{$IFDEF JPG_support}
      gtJPG:
        begin
          JPEGImage:=TJPEGImage.Create;
          try
            JPEGImage.LoadFromFile(FileName);
            P.Assign(JPEGImage);
          finally
            FreeAndNil(JPEGImage);
          end;
        end;
{$ENDIF}
{$IFDEF PNG_support}
      gtPNG:
        begin
          pngimage:=TPNGImage.Create;
          try
            pngimage.LoadFromFile(FileName);
            P.Assign(pngimage);
          finally
            FreeAndNil(pngimage);
          end;
        end;
{$ENDIF}
      gtEMF, gtOther:
        begin
          P.LoadFromFile(FileName);
        end;
    else
      raise Exception.Create(sPictureErr + cStrEOL + Format(sFileOpenErr, [FileName]));
    end;
    Result:=FileType;
    RosiCompConfig.rDBToolCfg.rpictool_LastFileName:=FileName;
  except
    on E: Exception do
      if ShowErrorMessage then
        ShowErr('', E)
      else
        Result:=gtError;
  end;
end;

function LoadBitmapFromFile(Bmp: TBitmap; const FileName: string; ShowErrorMessage: boolean = true): TrGraphicType;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  Bmp.Width:=0;
  try
    Result:=LoadPictureFromFile(P, FileName, ShowErrorMessage);
    if Result <> gtError then
      try
        GetBitmapFromPicture(Bmp, P);
      except
        on E: Exception do
          if ShowErrorMessage then
            ShowErr('', E)
          else
            Result:=gtError;
      end;
  finally
    FreeAndNil(P);
  end;
end;

function GetFileTypeFilterList(UseAllFormats: boolean): string;
var
  s: string;
begin
  Result:=sFileTypeBMP;
  s:='';
{$IFDEF JPG_support}
  Result:=Result + '|' + sFileTypeJPG;
  s:=sFileTypeBMPJPG;
{$ENDIF}
{$IFDEF PNG_support}
  Result:=Result + '|' + sFileTypePng;
  s:=sFileTypeBMPJPGPNG;
{$ENDIF}
  if UseAllFormats and (s <> '') then
    Result:=s + '|' + Result;
end;

function GetFileTypeExtension(GraphicType: TrGraphicType): string;
begin
  case GraphicType of
    gtBMP:
      Result:='.bmp';
    gtJPG:
      Result:='.jpg';
    gtPNG:
      Result:='.png';
    gtEMF:
      Result:='.emf';
  else
    Result:='';
  end;
end;

function GetFileTypeFromFileName(const FileName: string): TrGraphicType;
var
  E: string;
begin
  E:=ExtractFileExt(FileName);
  if SameText(E, '.bmp') then
    Result:=gtBMP
  else
    if SameText(E, '.jpg') or SameText(E, '.jpeg') then
      Result:=gtJPG
    else
      if SameText(E, '.png') then
        Result:=gtPNG
      else
        if SameText(E, '.emf') then
          Result:=gtEMF
        else
          Result:=gtOther;
end;

function OpenDlg_PictureFromFile(P: TPicture; const DefaultDir: string; DefaultGrType: TrGraphicType): TrGraphicType;
var
  D: TrOpenDialog;
  fn: string;
begin
  Result:=gtNone;
  D:=TrOpenDialog.Create(nil);
  try
    // D.Options:=D.Options+[ofNoChangeDir];
    if DefaultDir <> '' then
      D.InitialDir:=DefaultDir;
    D.DefaultExt:=GetFileTypeExtension(DefaultGrType);
    D.Filter:=GetFileTypeFilterList(true);
    D.FilterIndex:=0;
    if not D.Execute then
      Exit;
    fn:=D.FileName;
    if ExtractFileExt(fn) = '' then
      fn:=fn + GetFileTypeExtension(DefaultGrType);
    Result:=LoadPictureFromFile(P, fn);
  finally
    FreeAndNil(D);
  end;
end;

function OpenDlg_GraphicFiles(D: TOpenDialog; MultiSelect, Execute: boolean; const DefaultDir: string; DefaultGrType: TrGraphicType): boolean;
begin
  if not Assigned(D) then
    D:=TOpenDialog.Create(nil);
  // D.Options:=D.Options+[ofNoChangeDir];
  if MultiSelect then
    D.Options:=D.Options + [ofAllowMultiSelect];
  if DefaultDir <> '' then
    D.InitialDir:=DefaultDir;
  D.DefaultExt:=GetFileTypeExtension(DefaultGrType);
  D.Filter:=GetFileTypeFilterList(true);
  D.FilterIndex:=0;
  if Execute then
    Result:=D.Execute
  else
    Result:=true;
end;

function SaveDlg_PictureToFile(P: TPicture; const DefaultDir: string; DefaultGrType: TrGraphicType): TrGraphicType;
var
  D: TrSaveDialog;
  fn: string;
begin
  Result:=gtNone;
  D:=TrSaveDialog.Create(nil);
  try
    D.Options:=D.Options + [ofOverwritePrompt];
    if DefaultDir <> '' then
      D.InitialDir:=DefaultDir;
    D.DefaultExt:=GetFileTypeExtension(DefaultGrType);
    D.Filter:=GetFileTypeFilterList(false);
    D.FilterIndex:=Ord(DefaultGrType) - 1;
    if not D.Execute then
      Exit;
    fn:=D.FileName;
    if ExtractFileExt(fn) = '' then
      fn:=fn + GetFileTypeExtension(DefaultGrType);
    Result:=SavePictureToFile(P, fn, GetFileTypeFromFileName(fn));
  finally
    FreeAndNil(D);
  end;
end;

// JPEG specific function
function WritePictureToJPGField(F: TField; P: TPicture; Quality: integer): boolean;
begin
  Result:=WritePictureToField_Int(F, P, gtJPG, Quality);
end;

function WriteBitmapToJPGField(F: TField; Bmp: TBitmap; Quality: integer): boolean;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  try
    P.Assign(Bmp);
    Result:=WritePictureToField_Int(F, P, gtJPG, Quality);
  finally
    FreeAndNil(P);
  end;
end;

function SavePictureToJPGFile(P: TPicture; const FileName: string; Quality: integer): boolean;
var
  Stream: TStream;
begin
  Stream:=TFileStream.Create(FileName, fmCreate);
  try
    Result:=SavePictureToStream_Int(P, Stream, gtJPG, Quality) = gtJPG;
    RosiCompConfig.rDBToolCfg.rpictool_LastFileName:=FileName;
  finally
    FreeAndNil(Stream);
  end;
end;

function SaveBitmapToJPGFile(Bmp: TBitmap; const FileName: string; Quality: integer): boolean;
var
  P: TPicture;
begin
  P:=TPicture.Create;
  try
    P.Assign(Bmp);
    Result:=SavePictureToJPGFile(P, FileName, Quality);
  finally
    FreeAndNil(P);
  end;
end;

function GetJPEGDimensions(const FileName: string; var X, Y: Word): boolean;
var
  SegmentPos: integer;
  SOIcount: integer;
  B: byte;
begin
  Result:=false;
  with TFileStream.Create(FileName, fmOpenRead or fmShareDenyNone) do
  begin
    try
      Position:=0;
      Read(X, 2);
      if (X <> $D8FF) then
        Exit;
      SOIcount:=0;
      Position:=0;
      while (Position + 7 < Size) do
      begin
        Read(B, 1);
        if (B = $FF) then
        begin
          Read(B, 1);
          if (B = $D8) then
            Inc(SOIcount);
          if (B = $DA) then
            Break;
        end; { if }
      end; { while }
      if (B <> $DA) then
        Exit;
      SegmentPos:=-1;
      Position:=0;
      while (Position + 7 < Size) do
      begin
        Read(B, 1);
        if (B = $FF) then
        begin
          Read(B, 1);
          if (B in [$C0, $C1, $C2]) then
          begin
            SegmentPos:=Position;
            Dec(SOIcount);
            if (SOIcount = 0) then
              Break;
          end; { if }
        end; { if }
      end; { while }
      if (SegmentPos = -1) then
        Exit;
      if (Position + 7 > Size) then
        Exit;
      Position:=SegmentPos + 3;
      Read(Y, 2);
      Read(X, 2);
      X:=Swap(X);
      Y:=Swap(Y);
      Result:=true;
    finally
      Free;
    end; { try }
  end; { with }
end; { JPEGDimensions }

end.
