unit rtool;

{<
  rtool contains set of variety functions
  (strings, datetimes, forms, cursor, printers, system routines).

  Updated version compatible with Delphi2010+.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Nov 2005)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ComCtrls, Menus, CheckLst, Printers, WinSpool, ActnList, Grids,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF}
  Math, StrUtils, DateUtils, ShlObj, ActiveX, ShellApi, Themes, RosiCompGlobalCfg;

const
  rcEditSep = 0;
  rcEditUndo = 1;
  rcEditCut = 2;
  rcEditCopy = 3;
  rcEditPaste = 4;
  rcEditDelete = 5;
  rcEditSelectAll = 10;
  rcEditFind = 11;
  rcEditFindNext = 12;
  rcEditSelectFont = 13;
  rcMemoWordWrap = 20;
  rcMemoLoadFile = 21;
  rcMemoSaveFile = 22;

  { converts WideString with specified codePage to AnsiString }
function WideStringToString(const ws: WideString; codePage: Word): AnsiString;
{ converts AnsiString with specified codePage to WideString }
function StringToWideString(const s: AnsiString; codePage: Word): WideString;

{ provides decimal separator setting in RAD version independent way }
function GetDecimalSeparator: Char;
{ provides thousand separator setting in RAD version independent way }
function GetThousandSeparator: Char;
{ provides date separator setting in RAD version independent way }
function GetDateSeparator: Char;
{ provides time separator setting in RAD version independent way }
function GetTimeSeparator: Char;
{ update string representing numeric value with any separators to string with current separators }
function UpdateToDecimalSeparator(const s: string): string;
{ update string representing date value with any separators to string with current date separator }
function UpdateToDateSeparator(const s: string): string;
{ update string representing time value with any separators to string with current time separator }
function UpdateToTimeSeparator(const s: string): string;
{ remove thousand separator from teh input string }
function RemoveThousandSeparator(const s: string): string;
{ remove specified characters from the input string }
function CleanString(const s, CharsToRemove: string): string;
{ remove end of line (10 or 13) from the input string }
function CleanLastEOL(const s: string): string;
{ remove space, white spaces, tabs and end of line from the input string }
function CleanNumericString(const s: string): string;
{ remove space, white spaces, tabs and end of line from the input string }
function CleanDateTimeString(const s: string): string;

{ Converts the given string to a date and time value.
  Input string can contain a date, date followed by a time or time only.
  The date and time parts is decoded according to given Format string
  (same specifiers as for StrToDateTime is used so can be used as reverse function of StrToDateTime).
  System FormatSettings is used }
function ScanFormattedDateTime(DateTimeStr, Format: string; var ADate: TDateTime): Boolean; overload;
{ Converts the given string to a date and time value.
  Input string can contain a date, date followed by a time or time only.
  The date and time parts is decoded according to given Format string
  (same specifiers as for StrToDateTime is used so can be used as reverse function of StrToDateTime).
  Specific FormatSettings is used }
function ScanFormattedDateTime(DateTimeStr, Format: string; var ADate: TDateTime; const FormatSettings: TFormatSettings): Boolean; overload;
{ Converts the given string to a date and time value.
  Standard StrToDateTimeDef is used if Format is empty, in other case ScanFormattedDateTime is used }
function FormattedStrToDateTimeDef(DateTimeStr, Format: string; Default: TDateTime; DatePartOnly: Boolean): TDateTime;
{ Converts Date, Time or DateTime to string according format (if provided) or according to DateTime value }
function DateTimeToFormattedStr(ADateTime: TDateTime; Format: string; TimeOnly: Boolean = false): string;
{ Converts Text to Date with limits, current month is used if day only is entered }
function StrToDateEx(Text: string; var AsDate: TDateTime; Default: TDateTime; MinDate: TDateTime = 0; MaxDate: TDateTime = 0): Boolean;

{ Support for Themes and Style - get StyleService and used Font and Background colors for specific components }
{$IFDEF DXE2P}
{ return TCustomStyleServices for RAD version XE2+ }
function GetThemeServices: TCustomStyleServices;
{ return font color for provided TStyleFont,
  if style is not active then return provided AColor, for RAD version XE2+ }
function GetStyleFontColor(AColor: TColor; AStyleFont: TStyleFont): TColor;
{ return background color for provided TStyleColor,
  if style is not active then return provided AColor, for RAD version XE2+ }
function GetStyleBkColor(AColor: TColor; AStyleColor: TStyleColor = scGrid): TColor;
{$ELSE}
{ return TThemeServices for RAD version lower then XE2 }
function GetThemeServices: TThemeServices;
{$ENDIF}
{ return is Theme/Style is active (RAD version independent) }
function GetThemeEnabled: Boolean;

{ return TLabel font color for active style or AColor if style is not used (RAD version independent) }
function GetStyleFontColor_LabelText(AColor: TColor): TColor;
{ return TButton font color for active style or AColor if style is not used (RAD version independent) }
function GetStyleFontColor_ButtonText(AColor: TColor; Disabled: Boolean = false): TColor;
{ return TComboBox font color for active style or AColor if style is not used (RAD version independent) }
function GetStyleFontColor_ComboBox(AColor: TColor): TColor;
{ return TListBox font color for active style or AColor if style is not used (RAD version independent) }
function GetStyleFontColor_ListBox(AColor: TColor; Selected: Boolean = false): TColor;
{ return TGrid text font color for active style or AColor if style is not used (RAD version independent) }
function GetStyleFontColor_GridText(AColor: TColor; Fixed: Boolean = false; Hot: Boolean = false; Pressed: Boolean = false): TColor;

{ return Gradient base background color for active style or AColor if style is not used (RAD version independent) }
function GetStyleBkColor_GradientBase(AColor: TColor): TColor;
{ return TComboBox background color for active style or AColor if style is not used (RAD version independent) }
function GetStyleBkColor_ComboBox(AColor: TColor): TColor;
{ return TListBox background color for active style or AColor if style is not used (RAD version independent) }
function GetStyleBkColor_ListBox(AColor: TColor): TColor;
{ return TGrid background color for active style or AColor if style is not used (RAD version independent) }
function GetStyleBkColor_Grid(AColor: TColor): TColor;
{ return TGrid border background color for active style or AColor if style is not used (RAD version independent) }
function GetStyleColor_GridCellBorder(AColor: TColor): TColor;

{ return TGridCoord from THotTrackCellInfo (RAD version independent) }
function GetHotTrackCoord(HotTrack: THotTrackCellInfo): TGridCoord;
{ set THotTrackCellInfo from X, Y (RAD version independent) }
procedure SetHotTrackCoord(var HotTrack: THotTrackCellInfo; X, Y: integer); overload;
{ set THotTrackCellInfo from TGridCoord (RAD version independent) }
procedure SetHotTrackCoord(var HotTrack: THotTrackCellInfo; GridCoord: TGridCoord); overload;
{ compare THotTrackCellInfo with X, Y (RAD version independent) }
function CheckHotTrackCoord(HotTrack: THotTrackCellInfo; X, Y: integer): Boolean; overload;
{ compare THotTrackCellInfo with TGridCoord (RAD version independent) }
function CheckHotTrackCoord(HotTrack: THotTrackCellInfo; GridCoord: TGridCoord): Boolean; overload;

// DPi + scale
{ set default font height }
procedure DPI_SetFontHeight(AFont: TFont);
{ scale current font height }
procedure DPI_ScaleFontHeight(AFont: TFont; M, D: integer); overload;
{ scale current font height }
procedure DPI_ScaleFontHeight(AFont: TFont; ScaleRatio: double); overload;
{ scale default size }
function DPI_ScaleSize(Size: integer): integer;

{ fill str from left }
function FillStrL(const s: string; By: Char; Len: integer): string;
{ fill str from right }
function FillStrR(const s: string; By: Char; Len: integer): string;
{ find substring from start position }
function AnsiPosEx(const Substr, s: string; From: integer = 1; CaseSensitive: Boolean = false): integer;
{ find index-th part of input string separated by separator }
function FindStrPart(const s, Separator: string; Sequence: byte; CaseSensitive: Boolean = false): string;
{ find first part of input string separated by separator, return it and remove from input }
function RemoveStrPart(var s: string; const Separator: string; CaseSensitive: Boolean = false): string;
{ find part of input string identificated by StartStr and StopStr }
function FindStrFromTo(const SubstrStart, SubstrStop, s: string; CaseSensitive: Boolean = false; ExtractSubstrStart: Boolean = false): string;
{ clear all substr from str }
function ClearStr(const Substr, s: string; CaseSensitive: Boolean = false): string;
{ replace oldSubStr by newSubStr in s }
function ReplaceStr(const OldSubstr, NewSubstr, s: string; All: Boolean = true; CaseSensitive: Boolean = false): string;
{ replace oldSubStr identificated by StartStr and StopStr by newSubStr in s }
function ReplaceStrFromTo(const OldSubstrStart, OldSubstrStop, NewSubstr, s: string; ReplaceSubstrStop: Boolean = true; All: Boolean = true;
  CaseSensitive: Boolean = false): string;
{ add string to separated list }
procedure AddToSepList(var List: string; const StrToAdd: string; const Sep: string = ', ');
{ check if string is wholeword = separated from other string by any chars from separators }
function ContainWholeWord(const Substr, s: string; Separators: TSysCharSet; StartPos: integer = 0): Boolean;
{ check if string is wholeword = items are separated from by separator }
function ExistsInSepList(const Substr, s: string; Separator: string; TrimSpaces: Boolean = true): Boolean;

{ conversion variable to string (for saving in BLOB) }
function VarToString(PointToVar: Pointer; LengthVar: integer): string;
{ conversion string to variable }
function StringToVar(PointToVar: Pointer; LengthVar: integer; s: string): Boolean;
{ get one line from string separated by EOL }
function GetLine(const StrList: string; Index: integer): string;
{ set one line in string separated by EOL }
function SetLine(const StrList, Line: string; Index: integer): string;
{ conversion int to str with 0 on left side }
function IntToStr0(Int: integer; Len: integer): string;

{ not used, use DayOf... version from DateUtils.pas
  function GetDay(Date: TDateTime): word;
  function GetMonth(Date: TDateTime): word;
  function GetYear(Date: TDateTime): word;
  function DaysPerMonth(Year,Month: word): word;
  function DaysPerMonthD(Date: TDateTime): word;
}

{ form actions }

{ create a form instance (if not already assigned) and add it to the CreatedFormAdrList }
function CreateF(const FormClass: TFormClass; var Ref; AddToFormList: Boolean = true): Boolean;
{ free a form instance (if already assigned) and remove it from the CreatedFormAdrList }
function FreeF(var Ref): Boolean;
{ create a form instance (if not already assigned), update to normal (not minimized) and not hidden and show as modal, then free a form instance if it was created before }
function CreateShowFM(const FormClass: TFormClass; var Ref): integer;
{ create a form instance (if not already assigned), update to normal (not minimized) show as non-modal }
procedure CreateShowF(const FormClass: TFormClass; var Ref);
{ free all forms from the CreatedFormAdrList, it means created by CreateF function }
procedure FreeAllF;
{ returns active form if exist, in other case returns Application }
function GetActiveOwner: TComponent;
{ send message WM_NEXTDLGCTL to move focus to nest contral }
procedure FocusNextControl(Form: TForm);
{ calls standard WinHelp procedure }
function ShowWinHelp(Form: TForm; Command: Word; Data: integer; var CallHelp: Boolean): Boolean;
{ open file defined by Form.HelpFile (by associated application) }
function ShowExtHelp(Form: TForm; Command: Word; Data: integer; var CallHelp: Boolean): Boolean;

{ create and return new menu item object with properties from params }
function CreateMenuItem(Owner: TComponent; Caption: string; Tag: integer; OnClick: TNotifyEvent): TMenuItem; overload;
{ create and return new menu item object with properties from params }
function CreateMenuItem(Owner: TComponent; Caption: string; ShortCut: Word; Tag: integer; OnClick: TNotifyEvent): TMenuItem; overload;
{ create and return new menu item object with properties from params }
function CreateMenuItem(Owner: TComponent; Caption, Hint: string; ShortCut: Word; Tag, ImageIndex: integer; OnClick: TNotifyEvent; Action: TAction): TMenuItem;
  overload;
{ define standard popup menu list used for any edit components
  (undo, cut, copy, paste, select all, find, font, save, load) }
procedure DefineEditPopupMenuItems(PM: TPopupMenu; OnClick: TNotifyEvent);

{ universal raising error procedure }
procedure RaiseErr(ErrDesc: string);

{ set cursor to specified type and return previous cursor }
function ChCursor(const TypCursor: TCursor): TCursor;
{ set cursor to crHourGlass and return previous cursor }
function WCursor: TCursor;
{ set cursor to crDefault and return previous cursor }
function NCursor: TCursor;
{ set cursor to crSQLWait and return previous cursor }
function SQLCursor: TCursor;
{ update progress bar status (value) }
procedure SetProgressBar(PB: TProgressBar; APos, Max: integer);

{ mouse actions }
function ProcMouseWheel(var WheelDelta: integer; var Key: Word): Boolean;

{ negate bool value and return a new value }
function NegBool(var b: Boolean): Boolean;
{ negate check value of the TMenuItem and return a new value }
function NegMenuItem(Item: TMenuItem): Boolean;
{ negate check value of the TCheckBox and return a new value }
function NegCheckBox(Box: TCheckBox): Boolean;

{ open document or run file }
function RunFileSimple(const FileName: string): THandle;
{ open document or run file }
function RunFile(const FileName, Params, DefaultDir: string; ShowCmd: integer; ShowErr: Boolean = true): THandle;
{ open and print document }
function PrintFileSimple(const FileName: string): THandle;
{ open and print document }
function PrintFile(const FileName, Params, DefaultDir: string; ShowCmd: integer): THandle;

{ move item in TListBox from old to new index }
function MoveListBox(L: TListBox; old, new: integer): Boolean;
{ move item in TCheckListBox from old to new index }
function MoveCheckListBox(L: TCheckListBox; old, new: integer): Boolean;
{ compare two numeric items from the List, used by TList.CustomSort }
function SortNum(List: TStringList; Index1, Index2: integer): integer;
{ compare two date items from the List, used by TList.CustomSort }
function SortDate(List: TStringList; Index1, Index2: integer): integer;
{ add item to List if not exist }
function AddItemIfNotExist(List: TStrings; Text: string): Boolean;

// printer functions
{ fill list by printer names }
procedure GetPrinterNames(List: TStrings);
{ return default printer name }
function GetDefaultPrinterName: string;
{ set default printer }
function SetDefaultPrinter(const PrinterName: String; SendInfoMsg: Boolean): Boolean;
{ show printer dialog }
procedure SetupPrinterDlg(const PrinterName: String);

{ Get temporary directory deprecated 'use GetTempDir instead' }
function GetTempDirectory: string; deprecated 'use GetTempDir instead';
{ Get temporary directory }
function GetTempDir: string;
{ returns special directory in system according parameter FolderType,
  see ShlObj for list of values }
function GetSystemDir(FolderType: integer): string;
{ return path of exe program }
function GetAppPath: string; deprecated 'use GetAppDir instead';
{ return path of exe program }
function GetAppDir: string;
{ set end of the Path with backslash }
function SetEndOfPath(const Path: string): string;

{ set font style, color or size according Param string:
  N=normal
  B=bold
  I=italic
  U=underline
  Red, Blue = color
  $00xxxxxx = color
  8, 9, 10, 12, 14 ... = size }
procedure UseFontFormatParams(AFont: TFont; Param: string); overload;
procedure UseFontFormatParams(AFont: TFont; Param: string; var NormalStyleDefined: Boolean); overload;
{ update Alignment setting according to TAlignmentDefault setting }
procedure UpdateAlignmentFromAlignmentDefault(var Alignment: TAlignment; AlignmentDefault: TAlignmentDefault);
{ update FontStyle setting according to TFontStylesDefault setting }
procedure UpdateFontStylesFromFontStylesDefault(var FontStyles: TFontStyles; FontStylesDefault: TFontStylesDefault);

{ check if ABitIndex is set in AValueToCheck }
function IsBitSet(const AValueToCheck, ABitIndex: integer): Boolean;
{ set ABitIndex in AValueToAlter }
function SetBit(const AValueToAlter, ABitIndex: integer): integer;
{ reset ABitIndex in AValueToAlter }
function ResetBit(const AValueToAlter, ABitIndex: integer): integer;

implementation

uses rdlg, rstring;

var
  CreatedFormAdrList: TList;

function WideStringToString(const ws: WideString; codePage: Word): AnsiString;
var
  L: integer;
begin
  if ws = '' then
    Result:=''
  else
  begin
    L:=WideCharToMultiByte(codePage, WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR, @ws[1], -1, nil, 0, nil, nil);
    SetLength(Result, L - 1);
    if L > 1 then
      WideCharToMultiByte(codePage, WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR, @ws[1], -1, @Result[1], L - 1, nil, nil);
  end;
end;

function StringToWideString(const s: AnsiString; codePage: Word): WideString;
var
  L: integer;
begin
  if s = '' then
    Result:=''
  else
  begin
    L:=MultiByteToWideChar(codePage, MB_PRECOMPOSED, PAnsiChar(@s[1]), -1, nil, 0);
    SetLength(Result, L - 1);
    if L > 1 then
      MultiByteToWideChar(codePage, MB_PRECOMPOSED, PAnsiChar(@s[1]), -1, PWideChar(@Result[1]), L - 1);
  end;
end;

function GetDecimalSeparator: Char;
begin
{$IFDEF DXEP}
  Result:=FormatSettings.DecimalSeparator;
{$ELSE}
  Result:=DecimalSeparator;
{$ENDIF}
end;

function GetThousandSeparator: Char;
begin
{$IFDEF DXEP}
  Result:=FormatSettings.ThousandSeparator;
{$ELSE}
  Result:=ThousandSeparator;
{$ENDIF}
end;

function GetDateSeparator: Char;
begin
{$IFDEF DXEP}
  Result:=FormatSettings.DateSeparator;
{$ELSE}
  Result:=DateSeparator;
{$ENDIF}
end;

function GetTimeSeparator: Char;
begin
{$IFDEF DXEP}
  Result:=FormatSettings.TimeSeparator;
{$ELSE}
  Result:=TimeSeparator;
{$ENDIF}
end;

function UpdateToDecimalSeparator(const s: string): string;
var
  i, pdot, pcom: integer;
  thsep: string;
begin
  pdot:=Pos('.', s);
  pcom:=Pos(',', s);
  thsep:='';
  if (pdot > 0) and (pcom > 0) then
  begin
    if pdot < pcom then
      thsep:='.'
    else
      thsep:=',';
  end;
  Result:='';
  for i:=1 to Length(s) do
    if s[i] = thsep then
      Result:=Result + GetThousandSeparator
    else
      if Pos(s[i], '.,') > 0 then
        Result:=Result + GetDecimalSeparator
      else
        Result:=Result + s[i];
end;

function _UpdateToSeparator(const s, Sep, chars: string): string;
var
  i: integer;
begin
  Result:='';
  for i:=1 to Length(s) do
    if Pos(s[i], chars) > 0 then
      Result:=Result + Sep
    else
      Result:=Result + s[i];
end;

function UpdateToDateSeparator(const s: string): string;
begin
  Result:=_UpdateToSeparator(s, GetDateSeparator, '.,;+-*/:_');
end;

function UpdateToTimeSeparator(const s: string): string;
begin
  Result:=_UpdateToSeparator(s, GetTimeSeparator, '.,;+-*/:_');
end;

function RemoveThousandSeparator(const s: string): string;
begin
  Result:=ReplaceStr(GetThousandSeparator, '', s);
  Result:=ReplaceStr(' ', '', Result);
end;

function CleanString(const s, CharsToRemove: string): string;
var
  i: integer;
begin
  Result:=s;
  for i:=1 to Length(CharsToRemove) do
    Result:=ClearStr(CharsToRemove[i], Result, false);
end;

function CleanLastEOL(const s: string): string;
var
  ch: string;
begin
  Result:=s;
  while (Result <> '') do
  begin
    ch:=copy(Result, Length(Result), 1);
    if (ch = #10) or (ch = #13) then
      Result:=copy(Result, 1, Length(Result) - 1)
    else
      Break;
  end;
end;

function CleanNumericString(const s: string): string;
begin
  Result:=CleanString(s, ' ' + #9 + #10 + #13 + #160);
end;

function CleanDateTimeString(const s: string): string;
begin
  Result:=CleanString(s, ' ' + #9 + #10 + #13 + #160);
end;

procedure ScanDateTime_GetSep(var s, Sep, mask: string);
var
  i: integer;
begin
  // find position of separator
  for i:=1 to Length(s) do
    if Pos(s[i], ' .-/:*_[]{}()') > 0 then
    begin
      // return mask till sep
      mask:=copy(s, 1, i - 1);
      // return sep
      Sep:=s[i];
      // return rest after sep
      s:=copy(s, i + 1, 255);
      Exit;
    end;
  // use rest
  mask:=s;
  Sep:='';
  s:='';
end;

procedure ScanDateTime_GetPart(var s, Sep, p: string);
var
  i: integer;
begin
  i:=Pos(Sep, s);
  if (i = 0) and (Sep = '/') then
  begin
    i:=Pos(GetDateSeparator, s);
    if i > 0 then
      Sep:=GetDateSeparator;
  end;
  if i > 0 then
  begin
    // return part till sep
    p:=copy(s, 1, i - 1);
    // return rest after sep
    s:=copy(s, i + 1, 255);
    Exit;
  end;
  // sep not found, use rest
  p:=s;
  s:='';
end;

function ScanFormattedDateTime(DateTimeStr, Format: string; var ADate: TDateTime; const FormatSettings: TFormatSettings): Boolean; overload;
const
  arrMax = 10;
var
  i: integer;
  part, Sep, mask: array [1 .. arrMax] of string;
  D, M, Y: integer;
  h, n, s, z: integer;
  res: TDateTime;

  function GetNumPart(ch: Char; MaxLen: integer = 2; FoundIdx: integer = 1; TryMonth: Boolean = false): integer;
  var
    i, idx: integer;
  begin
    Result:=-1;
    idx:=0;
    // find mask index
    for i:=1 to arrMax do
    begin
      // maxLen is used to ignore ddd for weekdate name
      if (Length(mask[i]) <= MaxLen) and ContainsText(mask[i], ch) then
      begin
        // FoundIdx is used to find 2nd instance of M for minutes
        if FoundIdx = 1 then
        begin
          idx:=i;
          Break;
        end
        else
          Dec(FoundIdx);
      end;
    end;
    if idx > 0 then
    begin
      // get part from mask index
      Result:=StrToIntDef(part[idx], 0);
      if (Result = 0) and TryMonth then
      begin
        // find month names
        for i:=1 to 12 do
          if SameText(part[idx], FormatSettings.ShortMonthNames[i]) or SameText(part[idx], FormatSettings.LongMonthNames[i]) then
          begin
            Result:=i;
            Break;
          end;
      end;
    end;
  end;

  function IsPM: Boolean;
  var
    i: integer;
  begin
    Result:=false;
    for i:=1 to arrMax do
      if SameText(part[i], 'pm') or SameText(part[i], FormatSettings.TimePMString) then
        Result:=true;
  end;

begin
  Result:=false;
  Format:=ClearStr('"', Format);
  // find all separators and mask
  for i:=1 to arrMax do
    ScanDateTime_GetSep(Format, Sep[i], mask[i]);
  // find all parts according the separators
  for i:=1 to arrMax do
    ScanDateTime_GetPart(DateTimeStr, Sep[i], part[i]);

  // get date values
  D:=GetNumPart('d');
  M:=GetNumPart('m', 4, 1, true);
  Y:=GetNumPart('y', 4);
  if (D >= 0) then
  begin
    // use current year if not defined
    if Y = 0 then
      Y:=YearOf(Date);
    // correct 2 digit year
    if Y < 100 then
      Y:=Y + YearOf(Date) div 100 * 100;
  end;
  // get time values
  h:=GetNumPart('h');
  n:=GetNumPart('n');
  // try to use m instead of n
  if (n < 0) then
    if (D < 0) and (Y < 0) then
      n:=GetNumPart('m')
    else
      n:=GetNumPart('m', 4, 2);
  s:=GetNumPart('s');
  z:=GetNumPart('z', 3);

  try
    res:=0;
    if D >= 0 then
    begin
      res:=EncodeDate(Y, M, D);
      Result:=true;
    end;
    if h >= 0 then
    begin
      if IsPM then
        Inc(h, 12);
      res:=res + EncodeTime(h, Max(n, 0), Max(s, 0), Max(z, 0));
      Result:=true;
    end;
    if Result then
      ADate:=res;
  except
    Result:=false;
  end;
end;

function ScanFormattedDateTime(DateTimeStr, Format: string; var ADate: TDateTime): Boolean; overload;
{$IFNDEF DXEP}
var
  fs: TFormatSettings;
{$ENDIF}
begin
{$IFDEF DXEP}
  Result:=ScanFormattedDateTime(DateTimeStr, Format, ADate, FormatSettings);
{$ELSE}
  GetLocaleFormatSettings(LOCALE_USER_DEFAULT, fs);
  Result:=ScanFormattedDateTime(DateTimeStr, Format, ADate, fs);
{$ENDIF}
end;

function FormattedStrToDateTimeDef(DateTimeStr, Format: string; Default: TDateTime; DatePartOnly: Boolean): TDateTime;
begin
  Format:=Trim(Format);
  if Format = '' then
  begin
    if not DatePartOnly or not StrToDateEx(DateTimeStr, Result, Default) then
      Result:=StrToDateTimeDef(DateTimeStr, Default);
  end
  else
    if not ScanFormattedDateTime(DateTimeStr, Format, Result) then
      Result:=Default;
end;

function DateTimeToFormattedStr(ADateTime: TDateTime; Format: string; TimeOnly: Boolean = false): string;
begin
  Format:=Trim(Format);
  if Format = '' then
  begin
    if TimeOnly then
      Result:=TimeToStr(ADateTime)
    else
      if TimeOf(ADateTime) = 0 then
        Result:=DateToStr(ADateTime)
      else
        Result:=DateTimeToStr(ADateTime);
  end
  else
    Result:=FormatDateTime(Format, ADateTime);
end;

function StrToDateEx(Text: string; var AsDate: TDateTime; Default: TDateTime; MinDate: TDateTime = 0; MaxDate: TDateTime = 0): Boolean;
var
  ADate: TDateTime;
  Y, M, D: Word;
begin
  Result:=true;
  try
    if Text[Length(Text)] = GetDateSeparator then
      Text:=copy(Text, 1, Length(Text) - 1);
    if Pos(GetDateSeparator, Text) > 0 then
      ADate:=StrToDateDef(ClearStr(' ', Text, true), Default)
    else
    begin
      DecodeDate(Date, Y, M, D);
      D:=StrToInt(copy(Text, 1, 2));
      if Length(Text) > 2 then
        M:=StrToInt(copy(Text, 3, 2));
      if Length(Text) > 4 then
      begin
        Y:=StrToInt(copy(Text, 5, 4));
        if Y < 100 then
          Y:=2000 + Y;
      end;
      ADate:=EncodeDate(Y, M, D);
    end;
    if MaxDate > 0 then
      if ADate > MaxDate then
        ADate:=MaxDate;
    if MinDate > 0 then
      if ADate < MinDate then
        ADate:=MinDate;
    AsDate:=ADate;
  except
    Result:=false;
    // DlgE2(sInputErr, sFormatDateErr);
  end;
end;
{$IFDEF DXE2P}

function GetThemeServices: TCustomStyleServices;
begin
  Result:=StyleServices;
end;

function GetThemeEnabled: Boolean;
begin
  Result:=StyleServices.Enabled;
end;

function StyleIsActive: Boolean;
begin
  Result:=StyleServices.Enabled and StyleServices.Available and not StyleServices.IsSystemStyle;
end;

function GetStyleFontColor(AColor: TColor; AStyleFont: TStyleFont): TColor;
begin
  Result:=AColor;
  if StyleIsActive then
    Result:=StyleServices.GetStyleFontColor(AStyleFont);
end;

function GetStyleBkColor(AColor: TColor; AStyleColor: TStyleColor = scGrid): TColor;
begin
  Result:=AColor;
  if StyleIsActive then
    Result:=StyleServices.GetStyleColor(AStyleColor);
end;

function GetStyleFontColor_LabelText(AColor: TColor): TColor;
begin
  Result:=GetStyleFontColor(AColor, sfTextLabelNormal);
end;

function GetStyleFontColor_ButtonText(AColor: TColor; Disabled: Boolean = false): TColor;
begin
  if Disabled then
    Result:=GetStyleFontColor(AColor, sfButtonTextDisabled)
  else
    Result:=GetStyleFontColor(AColor, sfButtonTextNormal);
end;

function GetStyleFontColor_ComboBox(AColor: TColor): TColor;
begin
  Result:=GetStyleFontColor(AColor, sfComboBoxItemNormal);
end;

function GetStyleFontColor_ListBox(AColor: TColor; Selected: Boolean = false): TColor;
begin
  if Selected then
    Result:=GetStyleFontColor(AColor, sfListItemTextSelected)
  else
    Result:=GetStyleFontColor(AColor, sfListItemTextNormal);
end;

function GetStyleFontColor_GridText(AColor: TColor; Fixed: Boolean = false; Hot: Boolean = false; Pressed: Boolean = false): TColor;
begin
  if Fixed then
  begin
    Result:=GetStyleFontColor(AColor, sfGridItemFixedNormal);
    if Hot then
      Result:=GetStyleFontColor(AColor, sfGridItemFixedHot);
    if Pressed then
      Result:=GetStyleFontColor(AColor, sfGridItemFixedPressed);
  end
  else
    Result:=GetStyleFontColor(AColor, sfGridItemNormal);
end;

function GetStyleBkColor_GradientBase(AColor: TColor): TColor;
begin
  Result:=GetStyleBkColor(AColor, scGenericGradientBase);
end;

function GetStyleBkColor_ComboBox(AColor: TColor): TColor;
begin
  Result:=GetStyleBkColor(AColor, scComboBox);
end;

function GetStyleBkColor_ListBox(AColor: TColor): TColor;
begin
  Result:=GetStyleBkColor(AColor, scListBox);
end;

function GetStyleBkColor_Grid(AColor: TColor): TColor;
begin
  Result:=GetStyleBkColor(AColor, scGrid);
end;

function GetStyleColor_GridCellBorder(AColor: TColor): TColor;
begin
  Result:=AColor;
  if StyleIsActive then
    StyleServices.GetElementColor(StyleServices.GetElementDetails(tgCellNormal), ecBorderColor, Result);
end;
{$ELSE}

function GetThemeServices: TThemeServices;
begin
  Result:=ThemeServices;
end;

function GetThemeEnabled: Boolean;
begin
  Result:=ThemeServices.ThemesEnabled;
end;

function GetStyleFontColor_LabelText(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleFontColor_ButtonText(AColor: TColor; Disabled: Boolean = false): TColor;
begin
  Result:=AColor;
end;

function GetStyleFontColor_ComboBox(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleFontColor_ListBox(AColor: TColor; Selected: Boolean = false): TColor;
begin
  Result:=AColor;
end;

function GetStyleFontColor_GridText(AColor: TColor; Fixed: Boolean = false; Hot: Boolean = false; Pressed: Boolean = false): TColor;
begin
  Result:=AColor;
end;

function GetStyleBkColor_GradientBase(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleBkColor_ComboBox(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleBkColor_ListBox(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleBkColor_Grid(AColor: TColor): TColor;
begin
  Result:=AColor;
end;

function GetStyleColor_GridCellBorder(AColor: TColor): TColor;
begin
  Result:=AColor;
end;
{$ENDIF}

function GetHotTrackCoord(HotTrack: THotTrackCellInfo): TGridCoord;
begin
{$IFDEF DX102P}
  Result:=HotTrack.Rect.TopLeft;
{$ELSE}
  Result:=HotTrack.Coord;
{$ENDIF}
end;

procedure SetHotTrackCoord(var HotTrack: THotTrackCellInfo; X, Y: integer);
begin
{$IFDEF DX102P}
  HotTrack.Rect.Left:=X;
  HotTrack.Rect.Right:=X;
  HotTrack.Rect.Top:=Y;
  HotTrack.Rect.Bottom:=Y;
{$ELSE}
  HotTrack.Coord.X:=X;
  HotTrack.Coord.Y:=Y;
{$ENDIF}
end;

procedure SetHotTrackCoord(var HotTrack: THotTrackCellInfo; GridCoord: TGridCoord); overload;
begin
  SetHotTrackCoord(HotTrack, GridCoord.X, GridCoord.Y);
end;

function CheckHotTrackCoord(HotTrack: THotTrackCellInfo; X, Y: integer): Boolean;
begin
  Result:=(GetHotTrackCoord(HotTrack).X = X) and (GetHotTrackCoord(HotTrack).Y = Y);
end;

function CheckHotTrackCoord(HotTrack: THotTrackCellInfo; GridCoord: TGridCoord): Boolean; overload;
begin
  Result:=CheckHotTrackCoord(HotTrack, GridCoord.X, GridCoord.Y);
end;

procedure DPI_SetFontHeight(AFont: TFont);
begin
  AFont.Height:=Round(-11 * Screen.PixelsPerInch / 96);
end;

procedure DPI_ScaleFontHeight(AFont: TFont; M, D: integer);
begin
  AFont.Height:=MulDiv(AFont.Height, M, D);
end;

procedure DPI_ScaleFontHeight(AFont: TFont; ScaleRatio: double); overload;
begin
  AFont.Height:=Round(AFont.Height * ScaleRatio);
end;

function DPI_ScaleSize(Size: integer): integer;
begin
  Result:=Round(Size * Screen.PixelsPerInch / 96);
end;

function FillStrL(const s: string; By: Char; Len: integer): string;
var
  a: integer;
begin
  Result:=s;
  for a:=1 to Len - Length(Result) do
    Result:=By + Result;
end;

function FillStrR(const s: string; By: Char; Len: integer): string;
var
  a: integer;
begin
  Result:=s;
  for a:=1 to Len - Length(Result) do
    Result:=Result + By;
end;

function AnsiPosEx(const Substr, s: string; From: integer = 1; CaseSensitive: Boolean = false): integer;
begin
  if From < 1 then
    From:=1;
  if CaseSensitive then
    Result:=PosEx(Substr, s, From)
  else
    Result:=PosEx(AnsiUpperCase(Substr), AnsiUpperCase(s), From);
end;

function FindStrPart(const s, Separator: string; Sequence: byte; CaseSensitive: Boolean = false): string;
var
  i, p1, p2: integer;
begin
  Result:='';
  if Trim(s) = '' then
    Exit;
  // find start pos of sequence
  p1:=1;
  for i:=2 to Sequence do
  begin
    p1:=AnsiPosEx(Separator, s, p1, CaseSensitive);
    if p1 = 0 then
      Exit;
    Inc(p1, Length(Separator));
  end;
  // find end pos of next sequence
  p2:=AnsiPosEx(Separator, s, p1, CaseSensitive);
  if p2 = 0 then
    p2:=Length(s) + 1;
  Result:=copy(s, p1, p2 - p1);
end;

function RemoveStrPart(var s: string; const Separator: string; CaseSensitive: Boolean = false): string;
var
  p: integer;
begin
  Result:='';
  // if Trim(s) = '' then Exit;
  p:=AnsiPosEx(Separator, s, 1, CaseSensitive);
  if p > 0 then
  begin
    Result:=copy(s, 1, p - 1);
    s:=copy(s, p + Length(Separator), Length(s));
  end
  else
  begin
    Result:=s;
    s:='';
  end;
end;

function FindStrFromTo(const SubstrStart, SubstrStop, s: string; CaseSensitive: Boolean = false; ExtractSubstrStart: Boolean = false): string;
var
  p1, p2: integer;
begin
  p1:=AnsiPosEx(SubstrStart, s, 1, CaseSensitive);
  p2:=AnsiPosEx(SubstrStop, s, p1 + Length(SubstrStart), CaseSensitive);
  if (p1 = 0) or (p2 = 0) then
    Result:=''
  else
  begin
    Result:=copy(s, p1, p2 - p1);
    if ExtractSubstrStart then
      Result:=copy(Result, Length(SubstrStart) + 1, Length(Result));
  end;
end;

function ClearStr(const Substr, s: string; CaseSensitive: Boolean = false): string;
var
  p, DelCount, SubStrLen: integer;
  aSubStr, aStr: string;
begin
  // proc Case before and use directly PosEx to keep it fast
  Result:=s;
  if CaseSensitive then
  begin
    aSubStr:=Substr;
    aStr:=s;
  end
  else
  begin
    aSubStr:=AnsiUpperCase(Substr);
    aStr:=AnsiUpperCase(s);
  end;
  p:=1;
  DelCount:=0;
  SubStrLen:=Length(aSubStr);
  repeat
    p:=PosEx(aSubStr, aStr, p);
    if p = 0 then
      Break;
    Delete(Result, p - DelCount, SubStrLen);
    Inc(DelCount, SubStrLen);
    Inc(p, SubStrLen);
  until false;
end;

function ReplaceStr(const OldSubstr, NewSubstr, s: string; All: Boolean = true; CaseSensitive: Boolean = false): string;
var
  p: integer;
begin
  Result:=s;
  p:=1;
  repeat
    p:=AnsiPosEx(OldSubstr, Result, p, CaseSensitive);
    if p = 0 then
      Break;
    Delete(Result, p, Length(OldSubstr));
    Insert(NewSubstr, Result, p);
    if not All then
      Break;
    Inc(p, Length(NewSubstr));
  until false;
end;

function ReplaceStrFromTo(const OldSubstrStart, OldSubstrStop, NewSubstr, s: string; ReplaceSubstrStop: Boolean = true; All: Boolean = true;
  CaseSensitive: Boolean = false): string;
var
  p1, p2: integer;
begin
  Result:=s;
  p1:=1;
  repeat
    p1:=AnsiPosEx(OldSubstrStart, Result, p1, CaseSensitive);
    if p1 = 0 then
      Break;
    p2:=AnsiPosEx(OldSubstrStop, Result, p1 + Length(OldSubstrStart), CaseSensitive);
    if p2 = 0 then
      Break;
    if ReplaceSubstrStop then
      Inc(p2, Length(OldSubstrStop));
    Delete(Result, p1, p2 - p1);
    Insert(NewSubstr, Result, p1);
    if not All then
      Break;
    Inc(p1, Length(NewSubstr));
  until false;
end;

procedure AddToSepList(var List: string; const StrToAdd: string; const Sep: string = ', ');
begin
  if Trim(StrToAdd) = '' then
    Exit;
  if List <> '' then
    List:=List + Sep;
  List:=List + StrToAdd;
end;

function ContainWholeWord(const Substr, s: string; Separators: TSysCharSet; StartPos: integer = 0): Boolean;
var
  p: integer;
begin
  Result:=false;
  if StartPos = 0 then
    p:=1
  else
    p:=StartPos;
  repeat
    p:=AnsiPosEx(Substr, s, p);
    if (p = 0) or ((StartPos > 0) and (p <> StartPos)) then
      Break;
    Result:=(p = 1) or (CharInSet(s[p - 1], Separators));
    Inc(p, Length(Substr));
    Result:=Result and ((p = Length(s) + 1) or (CharInSet(s[p], Separators)));
  until Result or (StartPos > 0);
end;

function ExistsInSepList(const Substr, s: string; Separator: string; TrimSpaces: Boolean = true): Boolean;
var
  w, p: string;
begin
  Result:=false;
  w:=s;
  repeat
    p:=rtool.RemoveStrPart(w, Separator);
    if TrimSpaces then
      p:=Trim(p);
    if SameText(p, Substr) then
    begin
      Result:=true;
      Exit;
    end;
  until w = '';
end;

function VarToString(PointToVar: Pointer; LengthVar: integer): string;
var
  p: ^byte;
  a: integer;
begin
  p:=PointToVar;
  Result:='';
  for a:=1 to LengthVar do
  begin
    Result:=Result + chr(p^);
    Inc(p);
  end;
end;

function StringToVar(PointToVar: Pointer; LengthVar: integer; s: string): Boolean;
var
  p: ^byte;
  a: integer;
begin
  Result:=false;
  if Length(s) <> LengthVar then
    Exit;
  p:=PointToVar;
  for a:=1 to LengthVar do
  begin
    p^:=ord(s[a]);
    Inc(p);
  end;
  Result:=true;
end;

function GetLine(const StrList: string; Index: integer): string;
var
  L: TStringList;
begin
  Result:='';
  L:=TStringList.Create;
  try
    L.Text:=StrList;
    if L.Count > Index then
      Result:=L.Strings[Index];
  finally
    FreeAndNil(L);
  end;
end;

function SetLine(const StrList, Line: string; Index: integer): string;
var
  L: TStringList;
begin
  Result:='';
  L:=TStringList.Create;
  try
    L.Text:=StrList;
    while L.Count <= Index do
      L.Add('');
    L.Strings[Index]:=Line;
    Result:=L.Text;
  finally
    FreeAndNil(L);
  end;
end;

function IntToStr0(Int: integer; Len: integer): string;
begin
  Result:=FillStrL(IntToStr(Int), '0', Len);
end;

// Form functions
function CreateF(const FormClass: TFormClass; var Ref; AddToFormList: Boolean = true): Boolean;
var
  oldCursor: TCursor;
begin
  Result:=false;
  if Assigned(TForm(Ref)) then
    Exit;
  oldCursor:=WCursor;
  try
    Application.CreateForm(FormClass, Ref);
    if AddToFormList then
      CreatedFormAdrList.Add(Addr(Ref));
    Result:=true;
  finally
    ChCursor(oldCursor);
  end;
end;

function FreeF(var Ref): Boolean;
var
  i: integer;
begin
  Result:=false;
  if not Assigned(TForm(Ref)) then
    Exit;
  i:=CreatedFormAdrList.IndexOf(Addr(Ref));
  if i >= 0 then
    CreatedFormAdrList.Delete(i);
  FreeAndNil(TForm(Ref));
  // TForm(Ref).Free;
  // TForm(Ref):=nil;
  Result:=true;
end;

function CreateShowFM(const FormClass: TFormClass; var Ref): integer;
var
  temp: Boolean;
begin
  temp:=CreateF(FormClass, Ref);
  if TForm(Ref).WindowState = wsMinimized then
    TForm(Ref).WindowState:=wsNormal;
  if TForm(Ref).Visible then
    TForm(Ref).Hide;
  Result:=TForm(Ref).ShowModal;
  if temp then
    FreeF(Ref);
end;

procedure CreateShowF(const FormClass: TFormClass; var Ref);
begin
  CreateF(FormClass, Ref);
  if TForm(Ref).WindowState = wsMinimized then
    TForm(Ref).WindowState:=wsNormal;
  TForm(Ref).Show;
end;

procedure FreeAllF;
var
  i: integer;
  p: PPointer;
begin
  for i:=CreatedFormAdrList.Count - 1 downto 0 do
    try
      p:=CreatedFormAdrList.Items[i];
      FreeF(TForm(p^));
    except
      Beep;
    end;
  CreatedFormAdrList.Clear;
end;

function GetActiveOwner: TComponent;
begin
  if Assigned(Screen.ActiveForm) and (Screen.ActiveForm.FormStyle in [fsNormal, fsStayOnTop]) then
    Result:=Screen.ActiveForm
  else
    Result:=Application;
end;

function CreateMenuItem(Owner: TComponent; Caption: string; Tag: integer; OnClick: TNotifyEvent): TMenuItem; overload;
begin
  Result:=CreateMenuItem(Owner, Caption, Caption, 0, Tag, -1, OnClick, nil);
end;

function CreateMenuItem(Owner: TComponent; Caption: string; ShortCut: Word; Tag: integer; OnClick: TNotifyEvent): TMenuItem; overload;
begin
  Result:=CreateMenuItem(Owner, Caption, Caption, ShortCut, Tag, -1, OnClick, nil);
end;

{ create and return new menu item with properties from params }
function CreateMenuItem(Owner: TComponent; Caption, Hint: string; ShortCut: Word; Tag, ImageIndex: integer; OnClick: TNotifyEvent; Action: TAction): TMenuItem;
  overload;
var
  M: TMenuItem;
begin
  M:=TMenuItem.Create(Owner);
  M.AutoHotkeys:=maManual;
  M.Caption:=Caption;
  M.Hint:=Hint;
  M.ShortCut:=ShortCut;
  M.Tag:=Tag;
  M.ImageIndex:=ImageIndex;
  // if Assigned(OnClick) and (OnClick is TNotifyEvent) then
  M.OnClick:=OnClick;
  if Assigned(Action) then
    M.Action:=Action;
  Result:=M;
end;

procedure DefineEditPopupMenuItems(PM: TPopupMenu; OnClick: TNotifyEvent);
var
  mi: TMenuItem;
begin
  PM.Items.Add(CreateMenuItem(PM, sUndo, rcEditUndo, OnClick));
  PM.Items.Add(CreateMenuItem(PM, '-', rcEditSep, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sCut, rcEditCut, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sCopy, rcEditCopy, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sPaste, rcEditPaste, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sDelete, rcEditDelete, OnClick));
  PM.Items.Add(CreateMenuItem(PM, '-', rcEditSep, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sSelectAll, TextToShortCut('Ctrl+A'), rcEditSelectAll, OnClick));
  PM.Items.Add(CreateMenuItem(PM, '-', rcEditSep, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sFind, TextToShortCut('Ctrl+F'), rcEditFind, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sFindNext, TextToShortCut('Ctrl+Shift+R'), rcEditFindNext, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sChangeFont, rcEditSelectFont, OnClick));
  mi:=CreateMenuItem(PM, sWordWrap, rcMemoWordWrap, OnClick);
  mi.AutoCheck:=true;
  PM.Items.Add(mi);
  PM.Items.Add(CreateMenuItem(PM, '-', rcEditSep, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sLoadFromFile, rcMemoLoadFile, OnClick));
  PM.Items.Add(CreateMenuItem(PM, sSaveToFile, rcMemoSaveFile, OnClick));
end;

procedure RaiseErr(ErrDesc: string);
var
  s: string;
begin
  s:=sRaiseErr;
  if ErrDesc <> '' then
    s:=s + #13 + sErrorDescription + #13 + ErrDesc;
  raise Exception.Create(s);
end;

procedure FocusNextControl(Form: TForm);
begin
  SendMessage(Form.Handle, WM_NEXTDLGCTL, 0, 0);
end;

function ShowWinHelp(Form: TForm; Command: Word; Data: integer; var CallHelp: Boolean): Boolean;
begin
  Result:=false;
  if Form.HelpFile = '' then
    Exit;
  if (Command = HELP_CONTEXTPOPUP) and (Data < 100) then
    Result:=WinHelp(Form.Handle, PChar(Form.HelpFile), HELP_CONTEXT, Data)
  else
    Result:=WinHelp(Form.Handle, PChar(Form.HelpFile), Command, Data);
  CallHelp:=false;
end;

function ShowExtHelp(Form: TForm; Command: Word; Data: integer; var CallHelp: Boolean): Boolean;
begin
  CallHelp:=false;
  if (Command = HELP_CONTEXTPOPUP) then
    Result:=(RunFile(RosiCompConfig.Global_HelpDirectory + Form.HelpFile, '', '', SW_SHOW) >= 32)
  else
    Result:=true;
end;

// Cursor functions
function ChCursor(const TypCursor: TCursor): TCursor;
begin
  Result:=Screen.Cursor;
  Screen.Cursor:=TypCursor;
end;

function WCursor: TCursor;
begin
  Result:=ChCursor(crHourGlass);
end;

function NCursor: TCursor;
begin
  Result:=ChCursor(crDefault);
end;

function SQLCursor: TCursor;
begin
  Result:=ChCursor(crSQLWait);
end;

procedure SetProgressBar(PB: TProgressBar; APos, Max: integer);
begin
  if PB.Max <> Max then
    PB.Max:=Max;
  if PB.Position <> APos then
  begin
    PB.Position:=APos + 1;
    // have to set -1 to fix BUG in Windows Themes
    PB.Position:=APos;
  end;
end;

// Bool function
function NegBool(var b: Boolean): Boolean;
begin
  b:=not b;
  Result:=b;
end;

function NegMenuItem(Item: TMenuItem): Boolean;
begin
  Item.Checked:=not Item.Checked;
  Result:=Item.Checked;
end;

function NegCheckBox(Box: TCheckBox): Boolean;
begin
  Box.Checked:=not Box.Checked;
  Result:=Box.Checked;
end;

// Shell functions
function RunFile(const FileName, Params, DefaultDir: string; ShowCmd: integer; ShowErr: Boolean = true): THandle;
var
  h: THandle;
begin
  if Application.MainForm = nil then
    h:=Application.Handle
  else
    h:=Application.MainForm.Handle;
  Result:=ShellExecute(h, nil, PChar(FileName), PChar(Params), PChar(DefaultDir), ShowCmd);
  if Result <= 32 then
    if ShowErr then
      DlgE(Format(sFileOpenErr, [FileName]))
    else
      raise Exception.Create(Format(sFileOpenErr, [FileName]));
end;

function RunFileSimple(const FileName: string): THandle;
begin
  Result:=RunFile(FileName, '', '', SW_SHOW, true);
end;

function PrintFile(const FileName, Params, DefaultDir: string; ShowCmd: integer): THandle;
var
  h: THandle;
begin
  if Application.MainForm = nil then
    h:=Application.Handle
  else
    h:=Application.MainForm.Handle;
  Result:=ShellExecute(h, 'print', PChar(FileName), PChar(Params), PChar(DefaultDir), ShowCmd);
  if Result <= 32 then
    if ShowDlg(dtE, btYesNo, ['^C' + sError, '^B' + Format(sFilePrintErr, [FileName]), Format(sFileOpenQ, [FileName])]) = mrYes then
      RunFile(FileName, Params, DefaultDir, ShowCmd);
end;

function PrintFileSimple(const FileName: string): THandle;
begin
  Result:=PrintFile(FileName, '', '', SW_SHOW);
end;

// listbox actions
function MoveListBox(L: TListBox; old, new: integer): Boolean;
begin
  Result:=false;
  if (new = old) or (new < 0) or (old < 0) or (new > L.Items.Count - 1) then
    Exit;
  L.Items.Move(old, new);
  if (new <= L.TopIndex) and (L.TopIndex > 0) then
    L.TopIndex:=L.TopIndex - 1;
  if new >= L.TopIndex + (L.Height div L.ItemHeight) - 1 then
    L.TopIndex:=L.TopIndex + 1;
  if L.MultiSelect then
    L.Selected[new]:=true
  else
    L.ItemIndex:=new;
  Result:=true;
end;

function MoveCheckListBox(L: TCheckListBox; old, new: integer): Boolean;
var
  oldState: TCheckBoxState;
begin
  Result:=false;
  if (new = old) or (new < 0) or (old < 0) or (new > L.Items.Count - 1) then
    Exit;
  oldState:=L.State[old];
  L.Items.Move(old, new);
  L.State[new]:=oldState;
  if (new <= L.TopIndex) and (L.TopIndex > 0) then
    L.TopIndex:=L.TopIndex - 1;
  if new >= L.TopIndex + (L.Height div L.ItemHeight) - 1 then
    L.TopIndex:=L.TopIndex + 1;
  L.ItemIndex:=new;
  Result:=true;
end;

function SortNum(List: TStringList; Index1, Index2: integer): integer;
var
  s1, s2: string;
  c1, c2: double;
begin
  Result:=0;
  s1:=ClearStr(' ', List[Index1]);
  s2:=ClearStr(' ', List[Index2]);
  if s1 = '' then
    Result:=-1
  else
    if s2 = '' then
      Result:=1
    else
    begin
      c1:=StrToFloatDef(s1, Low(integer));
      c2:=StrToFloatDef(s2, Low(integer));
      if c1 < c2 then
        Result:=-1;
      if c1 > c2 then
        Result:=1;
    end;
end;

function SortDate(List: TStringList; Index1, Index2: integer): integer;
var
  d1, d2: double;
begin
  Result:=0;
  if List[Index1] = '' then
    Result:=-1
  else
    if List[Index2] = '' then
      Result:=1
    else
    begin
      d1:=StrToDateTimeDef(List[Index1], 0);
      d2:=StrToDateTimeDef(List[Index2], 0);
      if d1 < d2 then
        Result:=-1;
      if d1 > d2 then
        Result:=1;
    end;
end;

function ProcMouseWheel(var WheelDelta: integer; var Key: Word): Boolean;
begin
  Result:=WheelDelta <> 0;
  if WheelDelta > 0 then
  begin
    Key:=VK_UP;
    Dec(WheelDelta, WHEEL_DELTA);
    if WheelDelta < 0 then
      WheelDelta:=0;
  end;
  if WheelDelta < 0 then
  begin
    Key:=VK_DOWN;
    Inc(WheelDelta, WHEEL_DELTA);
    if WheelDelta > 0 then
      WheelDelta:=0;
  end;
end;

function AddItemIfNotExist(List: TStrings; Text: string): Boolean;
begin
  Text:=Trim(Text);
  Result:=(Text <> '') and (List.IndexOf(Text) < 0);
  if Result then
    List.Add(Text);
end;

// printer functions
procedure GetPrinterNames(List: TStrings);
begin
  List.Clear;
  List.Assign(Printer.Printers);
end;

function GetDefaultPrinterName: string;
var
  W2KSDP: function(pszBuffer: PWideChar; var pcchBuffer: integer): BOOL;
stdcall;
w :
array [0 .. 1023] of Char;
i :
integer;
begin
  if (Win32Platform = VER_PLATFORM_WIN32_NT) and (Win32MajorVersion >= 5) then
  begin
    @W2KSDP:=GetProcAddress(GetModuleHandle(winspl), 'GetDefaultPrinterW');
    if @W2KSDP = nil then
      RaiseLastOSError;
    i:=Length(w);
    if not W2KSDP(w, i) then
      RaiseLastOSError;
    Result:=w;
  end
  else
  begin
    GetProfileString('Windows', 'device', '', w, 255);
    Result:=copy(w, 1, Pos(',', w) - 1);
  end;
end;

function SetDefaultPrinter(const PrinterName: String; SendInfoMsg: Boolean): Boolean;
const
  MaxBuf = 1023;
  cs1 = 'Windows';
  cs2 = 'Device';
  cs3 = 'Devices';
  cs4 = #0;
var
  W2KSDP: function(pszPrinter: PWideChar): Boolean;
stdcall;
Buf :
array [0 .. MaxBuf] of Char;
PrinterInfo :
string;
res :
integer;
begin
  Result:=false;
  if (Win32Platform = VER_PLATFORM_WIN32_NT) and (Win32MajorVersion >= 5) then
  begin
    @W2KSDP:=GetProcAddress(GetModuleHandle(winspl), 'SetDefaultPrinterW');
    if @W2KSDP = nil then
      RaiseLastOSError;
    if not W2KSDP(PWideChar(PrinterName)) then
      RaiseLastOSError;
    Result:=true;
  end
  else
  begin
    res:=GetProfileString(cs3, PChar(PrinterName), #0, Buf, MaxBuf);
    if (res > 0) and (Trim(Buf) <> '') then
    begin
      PrinterInfo:=PrinterName + ',' + Buf;
      while GetProfileString(cs1, cs2, cs4, Buf, MaxBuf) > 0 do
        WriteProfileString(cs1, cs2, #0);
      WriteProfileString(cs1, cs2, PChar(PrinterInfo));
      if SendInfoMsg then
        case Win32Platform of
          VER_PLATFORM_WIN32_NT:
            SendMessage(HWND_BROADCAST, WM_WININICHANGE, 0, lParam(PChar(cs1)));
          VER_PLATFORM_WIN32_WINDOWS:
            SendMessage(HWND_BROADCAST, WM_SETTINGCHANGE, 0, lParam(PChar(cs1)));
        end;
      Result:=true;
    end;
  end;
end;

procedure SetupPrinterDlg(const PrinterName: String);
var
  p: array [0 .. 255] of WideChar;
begin
  StringToWideChar(PrinterName, p, 255);
  SHInvokePrinterCommand(Application.Handle, PRINTACTION_PROPERTIES, p, '', true);
end;

function GetTempDirectory: String;
begin
  Result:=GetTempDir;
end;

function GetTempDir: String;
var
  folder: array [0 .. MAX_PATH] of Char;
begin
  GetTempPath(Length(folder), folder);
  Result:=StrPas(folder);
end;

function GetSystemDir(FolderType: integer): string;
var
  folder: array [0 .. MAX_PATH] of Char;
  shellMalloc: IMalloc;
  ppidl: PItemIdList;
begin
  ppidl:=nil;
  try
    if SHGetMalloc(shellMalloc) = NOERROR then
    begin
      SHGetSpecialFolderLocation(Application.Handle, FolderType, ppidl);
      if not SHGetPathFromIDList(ppidl, folder) then
        raise Exception.Create('SHGetPathFromIDList failed : invalid ppidl');
      Result:=SetEndOfPath(folder);
    end;
  finally
    if ppidl <> nil then
      shellMalloc.Free(ppidl);
  end;
end;

function GetAppPath: string;
begin
  Result:=GetAppDir;
end;

function GetAppDir: string;
begin
  Result:=SetEndOfPath(ExtractFilePath(AnsiLowerCase(Application.ExeName)));
end;

function SetEndOfPath(const Path: string): string;
begin
  Result:=Path;
  if Path = '' then
    Exit;
  if Path[Length(Path)] <> '\' then
    Result:=Path + '\';
end;

procedure UseFontFormatParams(AFont: TFont; Param: string);
var
  sd: Boolean;
begin
  UseFontFormatParams(AFont, Param, sd);
end;

procedure UseFontFormatParams(AFont: TFont; Param: string; var NormalStyleDefined: Boolean); overload;
var
  C, n: Longint;
begin
  if Param = 'N' then
  begin
    AFont.Style:=[];
    NormalStyleDefined:=true;
  end
  else
    // B=bold
    if Param = 'B' then
      AFont.Style:=AFont.Style + [fsBold]
    else
      // I=italic
      if Param = 'I' then
        AFont.Style:=AFont.Style + [fsItalic]
      else
        // u=underline
        if Param = 'U' then
          AFont.Style:=AFont.Style + [fsUnderline]
        else
          if IdentToColor('cl' + Param, C) then
            AFont.Color:=C
          else
          begin
            // try numeric = size or color
            n:=StrToIntDef(Param, 0);
            if n > 0 then
            begin
              if copy(Param, 1, 1) = '$' then
                AFont.Color:=n
              else
                AFont.Size:=n;
            end;
          end;
end;

procedure UpdateAlignmentFromAlignmentDefault(var Alignment: TAlignment; AlignmentDefault: TAlignmentDefault);
begin
  case AlignmentDefault of
    tadDefault:
      ;
    tadLeft:
      Alignment:=taLeftJustify;
    tadCenter:
      Alignment:=taCenter;
    tadRight:
      Alignment:=taRightJustify;
  end;
end;

procedure UpdateFontStylesFromFontStylesDefault(var FontStyles: TFontStyles; FontStylesDefault: TFontStylesDefault);
begin
  if fsdNormal in FontStylesDefault then
    FontStyles:=[];
  if fsdBold in FontStylesDefault then
    Include(FontStyles, fsBold);
  if fsdItalic in FontStylesDefault then
    Include(FontStyles, fsItalic);
  if fsdUnderline in FontStylesDefault then
    Include(FontStyles, fsUnderline);
end;

// Check if the bit at ABitIndex position is 1 (true) or 0 (false)
function IsBitSet(const AValueToCheck, ABitIndex: integer): Boolean;
begin
  Result:=AValueToCheck and (1 shl ABitIndex) <> 0;
end;

// set the bit at ABitIndex position to 1
function SetBit(const AValueToAlter, ABitIndex: integer): integer;
begin
  Result:=AValueToAlter or (1 shl ABitIndex);
end;

// set the bit at ABitIndex position to 0
function ResetBit(const AValueToAlter, ABitIndex: integer): integer;
begin
  Result:=AValueToAlter and (not(1 shl ABitIndex));
end;

initialization

CreatedFormAdrList:=TList.Create;

finalization

FreeAndNil(CreatedFormAdrList);

end.
