unit rDBFind;

{<
  rDBFind provides universal dialog for record lookup in DBGrid or Dataset.
  Supports lookup in Master-Detail datasets.

  Full functional demo
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Sep 2002)
  @lastmod(Apr 2017)
}

interface

{$INCLUDE _rosi.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, db, dbgrids, ExtCtrls, PropSaveMain, Variants,
  ImgList, rHTMLLabel, rImprovedComps, rdbdlg_baseform;

type
  { Callback procedure for changing of master-detail relationship during searching }
  TChangeMasterDetailProc = procedure(DataSet: TDataSet; MasterSource: TDataSource; ShowAll: boolean) of object;

  { define record for master-detail searching }
  TFindDefDetailRec = record
    MasterKeyFieldName: string;
    DetailDataset: TDataSet;
    DetailKeyFieldName: string;
    DetailListFieldNames: string; // list of fields with ; separator
    DetailLookupFieldName: string;
    LookupDataset: TDataSet;
    LookupKeyFieldName: string;
  end;

  { define array for master-detail searching }
  TFindDefDetailArr = array of TFindDefDetailRec;

  { show DBFind dialog for DBGrid (Columns collection is used) }
function FindRec(const DBGrid: TCustomDBGrid; SortList: boolean = false): boolean; overload;
{ show DBFind dialog for Dataset (Fields collection is used) }
function FindRec(const DataSet: TDataSet; SortList: boolean = false): boolean; overload;

{ show DBFind dialog for detail DBGrid which use master-detail relationship }
function FindRecMD(const DBGrid: TCustomDBGrid; MasterSource: TDataSource; const MasterField, DetailField: string; ChangeMasterDetailProc: TChangeMasterDetailProc;
  SortList: boolean = false): boolean; overload;
{ show DBFind dialog for detail Dataset which use master-detail relationship }
function FindRecMD(const DataSet: TDataSet; MasterSource: TDataSource; const MasterField, DetailField: string; ChangeMasterDetailProc: TChangeMasterDetailProc;
  SortList: boolean = false): boolean; overload;

{ show DBFind dialog for DBGrid and list of details datasets }
function FindRecDetails(const DBGrid: TCustomDBGrid; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean = false): boolean; overload;
{ show DBFind dialog for Dataset and list of details datasets }
function FindRecDetails(const DataSet: TDataSet; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean = false): boolean; overload;

{ only for backward compability }
function FindRecSorted(const GridOrDataSet: TObject): boolean;
{ only for backward compability }
function FindRecMDSorted(const GridOrDataSet: TObject; MasterSource: TDataSource; const MasterField, DetailField: string;
  ChangeMasterDetailProc: TChangeMasterDetailProc): boolean;

type
  { universal dialog for searching in grid or database }
  TrDBFindForm = class(TrDBDlgBaseForm)
    lbLabels: TrListBox;
    Label1: TLabel;
    Label2: TLabel;
    rgOpt: TRadioGroup;
    cbCase: TCheckBox;
    btUp: TrButton;
    btDown: TrButton;
    cbDelQuotes: TCheckBox;
    btSort: TrURLLabel;
    eText: TComboBox;
    btClearHist: TrHTMLLabel;
    chShowMsg: TCheckBox;
    chCloseFound: TCheckBox;
    pBottom: TPanel;
    btFind: TrButton;
    btFindNext: TrButton;
    btFindPrev: TrButton;
    btClose: TrButton;
    btSelAll: TrURLLabel;
    lblResult: TrHTMLLabel;
    mInfo: TrMemoEx;
    procedure btFindClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure btUpClick(Sender: TObject);
    procedure btSortClick(Sender: TObject);
    procedure btClearHistAnchorClick(Sender: TObject; LinkName: String);
    procedure btSelAllClick(Sender: TObject);
    procedure rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
    procedure rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
    procedure rDlgPropSaveStdCtrlGetKeyString(Comp: TComponent; var Key: string);
    procedure FormResize(Sender: TObject);
  private
    GDS: TComponent;
    FN: TStrings;
    FoundIndex: Integer;
    FoundFieldText: string;
    SeekField: TField;
    // detail def variables
    DefDetailArr: TFindDefDetailArr;
    DetailDataset, DetailLookupDataset: TDataSet;
    DetailLookupField, LookupKeyField: TField;
    // master-detail variables
    MDMasterDataset: TDataSource;
    MDMasterField, MDDetailField: string;
    MDChangeProc: TChangeMasterDetailProc;
    MDShowAllDetail: boolean;
    Sorted: boolean;
    function GetPSKey(Key, Suffix: string): string;
    function TryFind(Backward: boolean): boolean;
    procedure SortList;
    procedure SetDetailFilters(MasterDS: TDataSet);
    procedure ResetDetailFilters;
    procedure SetDetailDSField(ItemIdx: Integer);
  public
  end;

var
  rDBFindForm: TrDBFindForm;

implementation

{$R *.DFM}
{$I _rosi.inc}

uses RosiCompGlobalCfg, rtool, rdbtool, rstring, rdlg;

const
  sItemsCount = 'ItemsCount';
  sItem = 'Item%d';
  sOrder = 'Order';
  sSelected = 'Selected';

  PSTextIndiv = true;
  PSCompareIndiv = true;

function FindRecInternal(const GridOrDataSet: TObject; MasterSource: TDataSource; const MasterField, DetailField: string;
  ChangeMasterDetailProc: TChangeMasterDetailProc; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean): boolean; forward;

function FindRec(const DBGrid: TCustomDBGrid; SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DBGrid, nil, '', '', nil, nil, SortList);
end;

function FindRec(const DataSet: TDataSet; SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DataSet, nil, '', '', nil, nil, SortList);
end;

function FindRecMD(const DBGrid: TCustomDBGrid; MasterSource: TDataSource; const MasterField, DetailField: string; ChangeMasterDetailProc: TChangeMasterDetailProc;
  SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DBGrid, MasterSource, MasterField, DetailField, ChangeMasterDetailProc, nil, SortList);
end;

function FindRecMD(const DataSet: TDataSet; MasterSource: TDataSource; const MasterField, DetailField: string; ChangeMasterDetailProc: TChangeMasterDetailProc;
  SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DataSet, MasterSource, MasterField, DetailField, ChangeMasterDetailProc, nil, SortList);
end;

function FindRecDetails(const DBGrid: TCustomDBGrid; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DBGrid, nil, '', '', nil, FindDefDetailArr, SortList);
end;

function FindRecDetails(const DataSet: TDataSet; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean = false): boolean; overload;
begin
  Result:=FindRecInternal(DataSet, nil, '', '', nil, FindDefDetailArr, SortList);
end;

function FindRecSorted(const GridOrDataSet: TObject): boolean;
begin
  Result:=FindRecInternal(GridOrDataSet, nil, '', '', nil, nil, true);
end;

function FindRecMDSorted(const GridOrDataSet: TObject; MasterSource: TDataSource; const MasterField, DetailField: string;
  ChangeMasterDetailProc: TChangeMasterDetailProc): boolean;
begin
  Result:=FindRecInternal(GridOrDataSet, MasterSource, MasterField, DetailField, ChangeMasterDetailProc, nil, true);
end;

function FindRecInternal(const GridOrDataSet: TObject; MasterSource: TDataSource; const MasterField, DetailField: string;
  ChangeMasterDetailProc: TChangeMasterDetailProc; const FindDefDetailArr: TFindDefDetailArr; SortList: boolean): boolean;
begin
  Result:=false;
  if not IsLinkedAndActive(GridOrDataSet) then
    Exit;
  rDBFindForm:=TrDBFindForm.Create(GetActiveOwner);
  try
    rDBFindForm.GDS:=GridOrDataSet as TComponent;
    rDBFindForm.DefDetailArr:=FindDefDetailArr;
    rDBFindForm.MDMasterDataset:=MasterSource;
    rDBFindForm.MDMasterField:=MasterField;
    rDBFindForm.MDDetailField:=DetailField;
    rDBFindForm.MDChangeProc:=ChangeMasterDetailProc;
    rDBFindForm.cbDelQuotes.Visible:=RosiCompConfig.rDBFindCfg.ShowDelQuotes;
    if not RosiCompConfig.rDBFindCfg.ShowDelQuotes then
      rDBFindForm.cbDelQuotes.Checked:=false;
    rDBFindForm.Sorted:=SortList;
    Result:=rDBFindForm.ShowModal = mrOk;
  finally
    FreeF(rDBFindForm);
  end;
end;

procedure TrDBFindForm.FormCreate(Sender: TObject);
begin
  inherited;
  FN:=TStringList.Create;
  rgOpt.ItemIndex:=RosiCompConfig.rDBFindCfg.LastOptIndex;
  cbCase.Checked:=RosiCompConfig.rDBFindCfg.LastCaseSens;
  eText.Text:=RosiCompConfig.rDBFindCfg.LastFindStr;
  btFindNext.Enabled:=eText.Text <> '';
  MDShowAllDetail:=false;
end;

procedure TrDBFindForm.FormShow(Sender: TObject);
var
  Idx: Integer;
  flds, s: string;
  F: TField;

  procedure AddDetailItem;
  begin
    // add Detail Index as object of item
    FN.AddObject(F.FieldName, TObject(Idx + 1));
    lbLabels.Items.Add(F.DisplayLabel);
  end;

begin
  inherited;
  lblResult.Font.Style:=[fsBold];
  lblResult.Caption:='';
  mInfo.Lines.Clear;
  // load field list
  GetFldList(GDS, FN, [fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  GetFldList(GDS, lbLabels.Items, [fdGetLabels, fdVisibleFieldOnly, fdSimpleField, fdMemoField]);
  // add detail field list
  if Assigned(DefDetailArr) then
    for Idx:=0 to High(DefDetailArr) do
      if Assigned(DefDetailArr[Idx].DetailDataset) and DefDetailArr[Idx].DetailDataset.Active then
      begin
        flds:=DefDetailArr[Idx].DetailListFieldNames;
        repeat
          s:=Trim(rtool.RemoveStrPart(flds, ';'));
          // detail dataset field ?
          F:=DefDetailArr[Idx].DetailDataset.FindField(s);
          if Assigned(F) then
            AddDetailItem
          else
          begin
            // lookup dataset field ?
            if Assigned(DefDetailArr[Idx].LookupDataset) and DefDetailArr[Idx].LookupDataset.Active then
              F:=DefDetailArr[Idx].LookupDataset.FindField(s);
            if Assigned(F) then
              AddDetailItem;
          end;
        until Trim(flds) = '';
      end;
  // init list
  lbLabels.ItemIndex:=0;
  for Idx:=lbLabels.Items.Count - 1 downto 0 do
    lbLabels.Selected[Idx]:=true;
  if Sorted then
    SortList;
  rDlgPropSave.ReadProperties;
  lbLabels.ItemIndex:=0;
  eText.SetFocus;
end;

procedure TrDBFindForm.FormClose(Sender: TObject; var Action: TCloseAction);
var
  DS: TDataSet;
  FoundField: TField;
  FoundVal: variant;
  DetailVal: variant;
  DetailValArr: array of variant;
  p, num: Integer;
  FldName: string;
begin
  DS:=GetDataSet(GDS);
  FoundField:=nil;
  // was Md?
  if MDShowAllDetail then
  begin
    if FoundIndex > -1 then
    begin
      // save found value
      FoundField:=DS.FindField(FN[FoundIndex]);
      FoundVal:=FoundField.AsVariant;
      // find rec in master
      repeat
        p:=Pos(';', MDDetailField);
        if p > 0 then
          FldName:=copy(MDDetailField, 1, p - 1)
        else
          FldName:=MDDetailField;
        DetailVal:=DS.FindField(Trim(FldName)).AsVariant;
        num:=Length(DetailValArr);
        SetLength(DetailValArr, num + 1);
        DetailValArr[num]:=DetailVal;
        if p = 0 then
          Break;
        MDDetailField:=copy(MDDetailField, p + 1, 255);
      until false;
      if Length(DetailValArr) = 1 then
        MDMasterDataset.DataSet.Locate(MDMasterField, DetailVal, [])
      else
        MDMasterDataset.DataSet.Locate(MDMasterField, DetailValArr, []);
    end;
    // call MD proc - show again only detail
    WCursor;
    try
      if Assigned(MDChangeProc) then
        MDChangeProc(DS, MDMasterDataset, false);
    finally
      NCursor;
    end;
    // find again value in detail
    if FoundIndex > -1 then
      while not DS.Eof do
      begin
        if FoundField.AsVariant = FoundVal then
          Break;
        DS.Next;
      end;
  end;
  rDlgPropSave.SaveProperties;
  inherited;
end;

procedure TrDBFindForm.FormDestroy(Sender: TObject);
begin
  FN.Free;
  RosiCompConfig.rDBFindCfg.LastOptIndex:=rgOpt.ItemIndex;
  RosiCompConfig.rDBFindCfg.LastCaseSens:=cbCase.Checked;
  RosiCompConfig.rDBFindCfg.LastFindStr:=eText.Text;
end;

procedure TrDBFindForm.FormResize(Sender: TObject);
begin
  inherited;
  btClearHist.Left:=eText.Left + eText.Width - btClearHist.Width;
  btSort.Left:=btUp.Left - btSort.Width - 8;
  btSelAll.Left:=lbLabels.Left + lbLabels.Height - btSelAll.Width;
end;

procedure TrDBFindForm.SetDetailFilters(MasterDS: TDataSet);
var
  Idx: Integer;
  Def: TFindDefDetailRec;
  FD, FM: TField;
  s: string;
begin
  // set filter for all details DS
  if Assigned(DefDetailArr) then
    for Idx:=0 to High(DefDetailArr) do
    begin
      Def:=DefDetailArr[Idx];
      if Assigned(Def.DetailDataset) and Def.DetailDataset.Active then
      begin
        FD:=Def.DetailDataset.FindField(Def.DetailKeyFieldName);
        if not Assigned(FD) then
          raise Exception.Create(Format('Detail KeyField "%s" not found.', [Def.DetailKeyFieldName]));
        FM:=MasterDS.FindField(Def.MasterKeyFieldName);
        if not Assigned(FM) then
          raise Exception.Create(Format('Master KeyField "%s" not found.', [Def.MasterKeyFieldName]));
        s:=rdbtool.GetFilterCond(FD, FM.AsString, false);
        Def.DetailDataset.Filter:=s;
        Def.DetailDataset.Filtered:=true;
        Def.DetailDataset.First;
      end;
    end;
end;

procedure TrDBFindForm.ResetDetailFilters;
var
  Idx: Integer;
  Def: TFindDefDetailRec;
begin
  // reset filter for all details DS
  if Assigned(DefDetailArr) then
    for Idx:=0 to High(DefDetailArr) do
    begin
      Def:=DefDetailArr[Idx];
      if Assigned(Def.DetailDataset) and Def.DetailDataset.Active then
      begin
        Def.DetailDataset.Filtered:=false;
      end;
    end;
end;

procedure TrDBFindForm.SetDetailDSField(ItemIdx: Integer);
var
  DefIdx: Integer;
begin
  SeekField:=nil;
  DetailDataset:=nil;
  DetailLookupDataset:=nil;
  DetailLookupField:=nil;
  LookupKeyField:=nil;
  DefIdx:=Integer(FN.Objects[ItemIdx]) - 1;
  if DefIdx < 0 then
  begin
    // standard master dataset field
    SeekField:=GetDataSet(GDS).FindField(FN[ItemIdx]);
  end
  else
  begin
    DetailDataset:=DefDetailArr[DefIdx].DetailDataset;
    if Assigned(DetailDataset) then
    begin
      DetailDataset.First;
      // field from Detail Dataset
      SeekField:=DetailDataset.FindField(FN[ItemIdx]);
      if not Assigned(SeekField) then
      begin
        // field from lookup dataset
        DetailLookupField:=DetailDataset.FindField(DefDetailArr[DefIdx].DetailLookupFieldName);
        DetailLookupDataset:=DefDetailArr[DefIdx].LookupDataset;
        if Assigned(DetailLookupField) and Assigned(DetailLookupDataset) then
        begin
          LookupKeyField:=DetailLookupDataset.FindField(DefDetailArr[DefIdx].LookupKeyFieldName);
          if Assigned(LookupKeyField) then
            SeekField:=DetailLookupDataset.FindField(FN[ItemIdx]);
        end;
      end;
    end;
  end;
end;

function TrDBFindForm.TryFind(Backward: boolean): boolean;
var
  i: Integer;
  FindText, s: string;
  DS: TDataSet;
  ok: boolean;
  oldAfterScroll: TDataSetNotifyEvent;
begin
  Result:=false;
  ok:=false;
  FindText:=eText.Text;
  if not cbCase.Checked then
    FindText:=AnsiUpperCase(FindText);
  DS:=GetDataSet(GDS);
  DS.DisableControls;
  oldAfterScroll:=DS.AfterScroll;
  DS.AfterScroll:=nil;
  SQLCursor;
  try
    while not((not Backward and DS.Eof) or (Backward and DS.Bof)) do
    begin
      SetDetailFilters(DS);
      for i:=0 to lbLabels.Items.Count - 1 do
        if lbLabels.Selected[i] then
        begin
          // set seek field and detail+lookup datasets
          SetDetailDSField(i);
          if SeekField = nil then
            Continue;
          repeat
            if Assigned(DetailLookupDataset) then
            begin
              // locate record in lookup dataset according DetailDataset value
              if not LocateRecord(LookupKeyField, DetailLookupField.AsVariant) then
                Break;
            end;
            s:=SeekField.AsString;
            if not cbCase.Checked then
              s:=AnsiUpperCase(s);
            case rgOpt.ItemIndex of
              0:
                ok:=FindText = s;
              1:
                ok:=Pos(FindText, s) = 1;
              2:
                ok:=Pos(FindText, s) > 0;
            end;
            if ok then
            begin
              Result:=true;
              FoundIndex:=i;
              FoundFieldText:=SeekField.AsString;
              Exit;
            end;
            if Assigned(DetailDataset) then
            begin
              // anoter record for detail DS
              DetailDataset.Next;
              if DetailDataset.Eof then
                Break;
            end
            else
              Break;
          until false;
        end;
      // anoter record for detail DS
      if Backward then
        DS.Prior
      else
        DS.Next;
    end;
  finally
    ResetDetailFilters;
    DS.EnableControls;
    DS.AfterScroll:=oldAfterScroll;
    if Assigned(oldAfterScroll) then
      oldAfterScroll(DS);
    NCursor;
  end;
  if not Result then
    FoundIndex:=-1;
end;

procedure TrDBFindForm.btFindClick(Sender: TObject);
var
  r, e1, e2: boolean;
  DS: TDataSet;
  L: TStringList;
begin
  if cbDelQuotes.Checked then
    eText.Text:=ClearStr('"', eText.Text, true);
  if eText.Items.IndexOf(eText.Text) < 0 then
  begin
    L:=TStringList.Create;
    try
      L.Assign(eText.Items);
      L.Add(eText.Text);
      L.Sort;
      eText.Items.Assign(L);
    finally
      L.Free;
    end;
  end;

  DS:=GetDataSet(GDS);
  // if MD - call MDProc with showall
  if Assigned(MDMasterDataset) and not MDShowAllDetail then
  begin
    MDShowAllDetail:=true;
    WCursor;
    try
      if Assigned(MDChangeProc) then
        MDChangeProc(DS, MDMasterDataset, true);
    finally
      NCursor;
    end;
  end;

  if Sender = btFind then
    DS.First
  else
    if Sender = btFindNext then
      DS.Next
    else
      DS.Prior;
  btFind.Enabled:=false;
  e1:=btFindNext.Enabled;
  e2:=btFindPrev.Enabled;
  btFindNext.Enabled:=false;
  btFindPrev.Enabled:=false;
  lblResult.Caption:=sLookingFor;
  mInfo.Lines.Clear;
  mInfo.HighlightWords.Clear;
  Update;
  r:=TryFind(Sender = btFindPrev);
  if r then
  begin
    // lblResult.Caption:=Format('%s: %s = "%s"', [sFound, lbLabels.Items[FoundIndex], FoundValue]);
    lblResult.Caption:=Format('%s. %s:', [sFound, lbLabels.Items[FoundIndex]]);
    mInfo.Lines.Text:=FoundFieldText;
    if not SameText(FoundFieldText, eText.Text) then
    begin
      mInfo.HighlightWords.Text:=eText.Text;
      mInfo.SelStart:=AnsiPosEx(eText.Text, FoundFieldText);
      mInfo.SelLength:=0;
    end;
    // rtool.ReplaceStr(eText.Text,'<B>'+eText.Text+'</B>',FoundFieldText);
    if chCloseFound.Checked then
      Close;
  end
  else
  begin
    lblResult.Caption:=sNotFound;
    if chShowMsg.Checked then
      DlgW(sNotFound);
  end;
  eText.SelectAll;
  eText.SetFocus;
  btFind.Enabled:=true;
  btFindNext.Enabled:=r or ((Sender = btFindPrev) and e1);
  btFindPrev.Enabled:=r or ((Sender <> btFindPrev) and e2);
  if r and ((Sender = btFind) or (Sender = btFindNext)) then
    btFindNext.SetFocus;
  if r and (Sender = btFindPrev) then
    btFindPrev.SetFocus;
  if r and Assigned(RosiCompConfig.rDBFindCfg.OnFoundEvent) then
    RosiCompConfig.rDBFindCfg.OnFoundEvent(GDS);
end;

function TrDBFindForm.GetPSKey(Key, Suffix: string): string;
begin
  Result:=Key + '.' + GDS.Owner.Name + '.' + GDS.Name;
  if Suffix <> '' then
    Result:=Result + '.' + Suffix;
end;

procedure TrDBFindForm.rDlgPropSaveReadProperties(PSFile: TPropSaveFile; Key: string);
var
  i, n, c: Integer;
  L: TStringList;
begin
  if RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields then
    Exit;
  L:=TStringList.Create;
  try
    // load order of items
    PSFile.ReadStrings(GetPSKey(Key, sOrder), L);
    c:=0;
    for i:=0 to L.Count - 1 do
    begin
      n:=FN.IndexOf(L[i]);
      if n >= 0 then
      begin
        FN.Move(n, c);
        lbLabels.Items.Move(n, c);
        Inc(c);
      end;
    end;
    // load and select saved fieldname
    PSFile.ReadStrings(GetPSKey(Key, sSelected), L);
    if L.Count > 0 then
      for i:=0 to lbLabels.Count - 1 do
        lbLabels.Selected[i]:=false;
    for i:=0 to L.Count - 1 do
    begin
      n:=FN.IndexOf(L[i]);
      if n >= 0 then
        lbLabels.Selected[n]:=true;
    end;
  finally
    L.Free;
  end;
end;

procedure TrDBFindForm.rDlgPropSaveSaveProperties(PSFile: TPropSaveFile; Key: string);
var
  i: Integer;
  L: TStringList;
begin
  if RosiCompConfig.rDBToolCfg.rDlg_BlockStoringGridFields then
    Exit;
  L:=TStringList.Create;
  try
    // save order of items
    PSFile.WriteStrings(GetPSKey(Key, sOrder), FN);
    // save fieldname of selected fields
    for i:=0 to lbLabels.Items.Count - 1 do
      if lbLabels.Selected[i] then
        L.Add(FN[i]);
    PSFile.EraseSection(GetPSKey(Key, sSelected));
    PSFile.WriteStrings(GetPSKey(Key, sSelected), L);
  finally
    L.Free;
  end;
end;

procedure TrDBFindForm.rDlgPropSaveStdCtrlGetKeyString(Comp: TComponent; var Key: string);
var
  indiv: boolean;
begin
  if (Comp = self.eText) then
    indiv:=PSTextIndiv
  else
    if (Comp = self.rgOpt) then
      indiv:=PSCompareIndiv
    else
      indiv:=false;
  if indiv then
    Key:=GetPSKey(Key, '');
  Key:=Key + '.' + Comp.Name;
end;

procedure TrDBFindForm.btUpClick(Sender: TObject);
var
  i, n: Integer;
  s: boolean;
begin
  lbLabels.SetFocus;
  i:=lbLabels.ItemIndex;
  s:=lbLabels.Selected[i];
  if (i >= 0) and (i < lbLabels.Items.Count) then
  begin
    if Sender = btUp then
      n:=i - 1
    else
      n:=i + 1;
    if (n < 0) or (n >= lbLabels.Items.Count) then
      Exit;
    lbLabels.Items.Move(i, n);
    FN.Move(i, n);
    lbLabels.ItemIndex:=n;
    lbLabels.Selected[n]:=s;
  end;
end;

procedure TrDBFindForm.SortList;
var
  i, j, min: Integer;
  s: boolean;
begin
  for i:=0 to lbLabels.Items.Count - 2 do
  begin
    min:=i;
    for j:=i + 1 to lbLabels.Items.Count - 1 do
    begin
      if AnsiCompareText(lbLabels.Items[j], lbLabels.Items[min]) < 0 then
        min:=j;
    end;
    if min <> i then
    begin
      s:=lbLabels.Selected[min];
      lbLabels.Items.Move(min, i);
      FN.Move(min, i);
      lbLabels.Selected[i]:=s;
    end;
  end;
  lbLabels.ItemIndex:=0;
end;

procedure TrDBFindForm.btSelAllClick(Sender: TObject);
begin
  lbLabels.SelectAll;
  lbLabels.ItemIndex:=0;
end;

procedure TrDBFindForm.btSortClick(Sender: TObject);
begin
  SortList;
end;

procedure TrDBFindForm.btClearHistAnchorClick(Sender: TObject; LinkName: String);
begin
  eText.Items.Clear;
end;

end.
