unit rDBComponents;

{<
  rDBComponents units contains enhanced DB components which can replace
  standard DB components without loosing of set properties.

  List of components in unit:
  TrDBRecordSelection, TrDBMoveButton, TrDBToolButton,
  TrDBEdit, TrDBEditDate, TrDBMemo, TrDBMemoEx, TrDBMemoCB, TrDBCheckBox,
  TrDBListBox, TrDBComboBox, TrDBLookupListBox, TrDBLookupComboBox,
  TrDBDateTimePicker, TrDBImage, TrDBLabel, TrDBBlob,
  TrDBCheckListBox, TrDBCheckListBoxEx

  rDBComponents provide folowing functionality:                                 @br
  - can show Label in any possition, caption is DisplayLabel of field           @br
  - change background color and disable dropdown action
  (for listbox and combobox) if dataset is not in editing mode components       @br
  - rDBMoveButton provide function for moving in dataset records                @br
  - rDBDateTimePicker support of null value (show additional checkbox)          @br
  - rDBImage support JPG and PNG format, LoadFromFile and SaveToFile procedures,
  Copy/Paste to/from clipboard, introduce Zoom functions with scrollbars
  and buttons, introduce ImageEdit window for showing of image in independent
  window with function for editation of brightness and contrast                 @br
  - mousewheel is working (for listbox and combobox)                            @br
  - some bugs of standard components is fixed                                   @br

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2006)
  @lastmod(Jan 2021)
}
interface

{$INCLUDE _rosi.inc}
{$DEFINE UseHTMLLabel}
{$DEFINE DoNotUseGlyphInDesign}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus,
  StdCtrls, ComCtrls, Buttons, Forms, DB, DBCtrls, CheckLst, ExtCtrls,
  Dialogs, Clipbrd, ImgList, Variants, Math, CommCtrl, Themes,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} DBConsts, Contnrs, StrUtils, DateUtils,
{$IFDEF DX101P} Rtti, {$ENDIF}
{$IFDEF UseHTMLLabel} rHTMLLabel, {$ENDIF}
  rImageZoom, rpictool, rImprovedComps, RosiCompGlobalCfg;

{ @exclude }
const
  rDBImageBtnCount = 4;

type
  { @exclude }
  TrDataLink = class(TDataLink)
  private
    FActiveChEvent: TNotifyEvent;
    FEditingChEvent: TNotifyEvent;
    FDataSetChEvent: TNotifyEvent;
  protected
    procedure ActiveChanged; override;
    procedure EditingChanged; override;
    procedure DataSetChanged; override;
  public
    constructor Create(ActiveChEvent, EditingChEvent, DataSetChEvent: TNotifyEvent);
    function Editing: boolean;
  end;

  { list of rDBMoveButton types }
  TrDBButtonType = (btFirst, btPrior, btNext, btLast, btAdd, btInsert, btDuplic, btEdit, btDelete, btPost, btCancel, btEnaIfEdit, btEnaIfNew, btEnaIfBrowse,
    btEnaIfNotEmpty, btEnaIfCanModify, btEnaIfBrowseCanEdit, btEnaIfBrowseNotEmpty);

  { Button which provides mooving and editing actions for linked dataset,
    enability of button is automaticaly set according ButtonType and Dataset state }
  TrDBMoveButton = class(TrButton)
  private
    FButtonType: TrDBButtonType;
    FEnabledAll: boolean;
    FSkipDBAction: boolean;
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure UpdateState(Sender: TObject);
    procedure SetEnabledAll(Value: boolean);
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Click; override; {< @exclude }
  published
    { link to connected dataset }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { type of button - define enability and action after click,
      btEnaIf types do not do any action but their enability depends on database state }
    property ButtonType: TrDBButtonType read FButtonType write FButtonType default btEnaIfEdit;
    { set global enable property for button, if set to false then button is always disabled }
    property EnabledAll: boolean read FEnabledAll write SetEnabledAll default true;
    { define if database action is skipped and click event is called only }
    property SkipDBAction: boolean read FSkipDBAction write FSkipDBAction default false;
  end;

  { ToolButton which provides mooving and editing actions for linked dataset,
    enability of button is automaticaly set according ButtonType and Dataset state }
  TrDBToolButton = class(TToolButton)
  private
    FButtonType: TrDBButtonType;
    FEnabledAll: boolean;
    FSkipDBAction: boolean;
    function GetDataSource: TDataSource;
    procedure SetDataSource(Value: TDataSource);
    procedure UpdateState(Sender: TObject);
    procedure SetEnabledAll(Value: boolean);
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Click; override; {< @exclude }
  published
    { link to connected dataset }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { type of button - define enability and action after click,
      btEnaIf types do not do any action but their enability depends on database state }
    property ButtonType: TrDBButtonType read FButtonType write FButtonType default btEnaIfEdit;
    { set global enable property for button, if set to false then button is always disabled }
    property EnabledAll: boolean read FEnabledAll write SetEnabledAll default true;
    { define if database action is skipped and click event is called only }
    property SkipDBAction: boolean read FSkipDBAction write FSkipDBAction default false;
    property Font;
    property ParentFont;
  end;

  { list of BoundLabel positions }
  TLabelPos = (lpTopLeft, lpTopCenter, lpTopRight, lpLeftTop, lpLeftCenter, lpLeftBottom);

  { TLabel subcomponent used by rDBComponents as field caption }
  TDBBoundLabel = class(TCustomLabel)
  private
    FField: TField;
    FOwnCaption: string;
    FLabelFieldCaption: string;
    FFieldCaptionSuffix: string;
    FPosition: TLabelPos;
    oldVisible: boolean;
    FShowRequiredBold: boolean;
    FShowRequiredMark: boolean;
    FRequiredMarkColor: TColor;
    FRequiredMarkSpace: integer;
    procedure SetPosition(Value: TLabelPos);
    procedure SetShowRequiredBold(Value: boolean);
    procedure SetShowRequiredMark(Value: boolean);
    procedure SetRequiredMarkColor(Value: TColor);
    procedure SetRequiredMarkSpace(Value: integer);
    procedure SetFieldCaptionSuffix(Value: string);
    function IsFieldCaptionSuffixStored: boolean;
  protected
    procedure RequestAlign; override;
    procedure RefreshPosition(RecalcBounds: boolean = false); {< @exclude }
    procedure SetOwnCaption(const Value: string);
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure DoDrawText(var Rect: TRect; Flags: Longint); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    procedure SetField(DataSource: TDataSource; FieldName: string); {< @exclude }
    procedure SetLabelField(DataSource: TDataSource; FieldName: string); {< @exclude }
    procedure RefreshCaption(RecalcBounds: boolean = false); {< @exclude }
    procedure ChangeVisible(AVisible: boolean); {< @exclude }
    procedure StoreVisible; {< @exclude }
  published
    { user defined caption text for component, if empty DisplayLabel of connected field is used }
    property OwnCaption: string read FOwnCaption write SetOwnCaption;
    { define whether boundlabel is bold for required field }
    property ShowRequiredBold: boolean read FShowRequiredBold write SetShowRequiredBold default false;
    { define whether asterisk is shown for required field }
    property ShowRequiredMark: boolean read FShowRequiredMark write SetShowRequiredMark default true;
    { define color of asterisk shown for required field }
    property RequiredMarkColor: TColor read FRequiredMarkColor write SetRequiredMarkColor default clRed;
    { define space between caption and asterisk shown for required field }
    property RequiredMarkSpace: integer read FRequiredMarkSpace write SetRequiredMarkSpace default 2;
    { define possition of bound label }
    property Position: TLabelPos read FPosition write SetPosition default lpTopLeft;
    { define suffix after field caption }
    property FieldCaptionSuffix: string read FFieldCaptionSuffix write SetFieldCaptionSuffix stored IsFieldCaptionSuffixStored;
    property Color;
    property Font;
    property ParentColor;
    property ParentFont;
    property ShowAccelChar;
    property Transparent stored false;
    property Visible;
    property OnClick;
    property OnDblClick;
    property Left stored false;
    property Top stored false;
    property Width stored false;
    property Height stored false;
  end;

  { TrHTMLLabel or TLabel subcomponent, show additional user defined text in some db component }
  TBoundLabel = class({$IFDEF UseHTMLLabel} TrHTMLLabel {$ELSE} TLabel {$ENDIF})
  private
    FPosition: TLabelPos;
    procedure SetPosition(Value: TLabelPos);
  protected
    procedure RequestAlign; override;
    procedure RefreshPosition;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { define possition of bound label }
    property Position: TLabelPos read FPosition write SetPosition default lpTopLeft;
    property Left stored false;
    property Top stored false;
    property Width stored false;
    property Height stored false;
  end;

  { list of bound buttons position for TrDBImage }
  TBtnPos = (bpTopLeft, bpTopRight, bpBottomLeft, bpBottomRight, bpLeftTop, bpLeftBottom, bpRightTop, bpRightBottom);

  { subcomponent used by TrDBImage for showing of supplementary buttons }
  TBoundButtons = class(TComponent)
  private
    FCaptions: string;
    FVisible: boolean;
    FFlat: boolean;
    FTransparent: boolean;
    FPosition: TBtnPos;
    FSize: integer;
    FShowGlyphs: boolean;
    FImages: TImageList;
    FImageChangeLink: TChangeLink;
    procedure SetCaptions(const Value: string);
    procedure SetVisible(Value: boolean);
    procedure SetFlat(Value: boolean);
    procedure SetTransparent(Value: boolean);
    procedure SetPosition(Value: TBtnPos);
    procedure SetSize(Value: integer);
    procedure SetShowGlyphs(Value: boolean);
    procedure SetImages(Value: TImageList);
    procedure ImageListChange(Sender: TObject);
  protected
    RefreshEvent: TNotifyEvent;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    { define list of buttons captions separated by ; }
    property Captions: string read FCaptions write SetCaptions;
    { define whether supplementary buttons for rDBImage is visible }
    property Visible: boolean read FVisible write SetVisible default false;
    { define appearance of supplementary buttons for rDBImage }
    property Flat: boolean read FFlat write SetFlat default true;
    { define appearance of supplementary buttons for rDBImage }
    property Transparent: boolean read FTransparent write SetTransparent default true;
    { define position of supplementary buttons for rDBImage }
    property Position: TBtnPos read FPosition write SetPosition default bpBottomLeft;
    { define size of supplementary buttons for rDBImage }
    property Size: integer read FSize write SetSize default 23;
    { define whether glyps is shown in supplementary buttons for rDBImage }
    property ShowGlyphs: boolean read FShowGlyphs write SetShowGlyphs default false;
    { define list to ImageList with glyphs for supplementary buttons,
      if list is not defined standard glyphs from rDBComponents.res is used }
    property Images: TImageList read FImages write SetImages;
  end;

  { SpeedButton subcomponent used by some components for showing of supplementary buttons }
  TBoundSpeedButton = class(TSpeedButton)
  private
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Left stored false;
    property Top stored false;
  end;

  { TListBox subcomponent used by other TrDBMemoCB }
  TBoundListBox = class(TCustomListBox)
  private
    procedure CNCommand(var Message: TWMCommand); message CN_COMMAND;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure KeyPress(var Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property Style;
    property BevelEdges;
    property BevelInner;
    property BevelKind default bkNone;
    property BevelOuter;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property Enabled;
    property Items;
    property PopupMenu;
    property ScrollWidth;
    property ShowHint;
    property Sorted;
    property Visible;
    property OnDblClick;
    property OnDrawItem;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  { TButtonedEdit subcomponent used by TrDBCheckListBox }
  TBoundButtonedEdit = class(TButtonedEdit)
  private
    FImages: TImageList;
    FTimer: TTimer;
    FFindDelay: cardinal;
    FOnFindEvent: TNotifyEvent;
    procedure FindKeyPress(Sender: TObject; var Key: Char);
    procedure FindChange(Sender: TObject);
    procedure LeftBtnClick(Sender: TObject);
    procedure RightBtnClick(Sender: TObject);
    procedure FindTimer(Sender: TObject);
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    procedure Loaded; override;
  public
    constructor CreateLinkEvent(AOwner: TComponent; OnFindEvent: TNotifyEvent);
    destructor Destroy; override; {< @exclude }
    procedure RefreshPosition(W: TControl; HorizontalSpace: integer); {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  published
    { define delay between text update and find event }
    property FindDelay: cardinal read FFindDelay write FFindDelay default 250;
    property Left stored false;
    property Top stored false;
    property Width stored false;
    property Height stored false;
  end;

  { @exclude }
  THackEditButton = class(TEditButton)
  end;

  TrDBEdit = class;

  { internal button subcomponent used by TrDBEdit }
  TrDBEditButton = class(TPersistent)
  strict private
  type
    TButtonState = (bsNormal, bsHot, bsPushed);

    TGlyph = class(TCustomControl)
    private
      FButton: TrDBEditButton;
      FState: TButtonState;
    protected
      procedure Click; override;
      procedure CreateWnd; override;
      procedure Paint; override;
      procedure WndProc(var Message: TMessage); override;
    public
      constructor Create(AButton: TrDBEditButton); reintroduce; virtual;
    end;
  protected type
    TButtonPosition = (bpLeft, bpRight);
    TButtonClickAction = (caClick, caHTTP, caMailTo, caFile);
  strict private
    FDisabledImageIndex: TImageIndex;
    FDropDownMenu: TPopupMenu;
    FEditControl: TrDBEdit;
    FGlyph: TGlyph;
    FHotImageIndex: TImageIndex;
    FImageIndex: TImageIndex;
    FPosition: TButtonPosition;
    FButtonClickAction: TButtonClickAction;
    FPressedImageIndex: TImageIndex;
    function GetEnabled: boolean;
    function GetCustomHint: TCustomHint;
    function GetHint: string;
    function GetImages: TImageList;
    function GetVisible: boolean;
    procedure SetDisabledImageIndex(const Value: TImageIndex);
    procedure SetEnabled(const Value: boolean);
    procedure SetCustomHint(const Value: TCustomHint);
    procedure SetHint(const Value: string);
    procedure SetHotImageIndex(const Value: TImageIndex);
    procedure SetImageIndex(const Value: TImageIndex);
    procedure SetPressedImageIndex(const Value: TImageIndex);
    procedure SetVisible(const Value: boolean);
  protected
    function GetOwner: TPersistent; override;
    procedure UpdateBounds; dynamic;
    property EditControl: TrDBEdit read FEditControl;
    property Glyph: TGlyph read FGlyph;
    property Images: TImageList read GetImages;
    property Position: TButtonPosition read FPosition;
  public
    constructor Create(EditControl: TrDBEdit; APosition: TButtonPosition); reintroduce; virtual;
    destructor Destroy; override;
  published
    property CustomHint: TCustomHint read GetCustomHint write SetCustomHint;
    property DisabledImageIndex: TImageIndex read FDisabledImageIndex write SetDisabledImageIndex default - 1;
    property DropDownMenu: TPopupMenu read FDropDownMenu write FDropDownMenu;
    property Enabled: boolean read GetEnabled write SetEnabled default true;
    property Hint: string read GetHint write SetHint;
    property HotImageIndex: TImageIndex read FHotImageIndex write SetHotImageIndex default - 1;
    property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default - 1;
    property PressedImageIndex: TImageIndex read FPressedImageIndex write SetPressedImageIndex default - 1;
    property ButtonClickAction: TButtonClickAction read FButtonClickAction write FButtonClickAction default caClick;
    property Visible: boolean read GetVisible write SetVisible default false;
  end;

  { enhanced TDBEdit with bound label and left + right edit button }
  TrDBEdit = class(TDBEdit)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FDefaultPrefix: string;
    FCanvas: TControlCanvas;
    FImages: TImageList;
    FImageChangeLink: TChangeLink;
    FLeftButton: TrDBEditButton;
    FRightButton: TrDBEditButton;
    procedure SetDefaultPrefix(const Value: string);
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure ImageListChange(Sender: TObject);
    procedure SetImages(const Value: TImageList);
    function GetOnLeftButtonClick: TNotifyEvent;
    function GetOnRightButtonClick: TNotifyEvent;
    procedure SetLeftButton(const Value: TrDBEditButton);
    procedure SetRightButton(const Value: TrDBEditButton);
    procedure SetOnLeftButtonClick(const Value: TNotifyEvent);
    procedure SetOnRightButtonClick(const Value: TNotifyEvent);
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure DoOwnPaint(MsgDC: HDC; Align: TAlignment);
    function GetTextMargins: TRect;
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DoExit; override;
    procedure UpdateEditMargins; dynamic;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
    procedure DefaultHandler(var Message); override; {< @exclude }
    { provide additional validation of the entered value (clean white spaces) }
    procedure ValidateEdit; override;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define bound label component }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { image list used for LeftButton and RightButton }
    property Images: TImageList read FImages write SetImages;
    { Left internal Button }
    property LeftButton: TrDBEditButton read FLeftButton write SetLeftButton;
    { Right internal Button }
    property RightButton: TrDBEditButton read FRightButton write SetRightButton;
    { click to Left internal Button }
    property OnLeftButtonClick: TNotifyEvent read GetOnLeftButtonClick write SetOnLeftButtonClick;
    { click to Right internal Button }
    property OnRightButtonClick: TNotifyEvent read GetOnRightButtonClick write SetOnRightButtonClick;
  end;

  { enhanced TrDBEdit allows fast entering of the date value by keyboard or by Month calendar;
    supported input formats - dd, dd/mm, dd/mm/yy, dd/mm/yyyy
    shortcuts - arrows Up/Down (optionaly Ctrl+Shift) to move date/month/year
    t = today, s = start of month, e = end of month, d,p = dropdown month calendar (if right button is visible) }
  TrDBEditDate = class(TrDBEdit)
  private
    FMinDate: TDateTime;
    FMaxDate: TDateTime;
    OrigText: string;
    FUpdateAfterChange: boolean;
    procedure SetDate(DateValue: TDateTime);
    function GetDate: TDateTime;
    function CheckDate(UpdateDataLink: boolean): boolean;
    procedure UpdateDatalinkValue;
    procedure FRightButtonClick(Sender: TObject);
    procedure PopupCallbackSelectDate(Sender: TObject);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
  protected
    FPopupCalendar: TrPopupMonthCalendar;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure DoEnter; override;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    function DoDropDown: boolean; virtual;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    procedure UpdateState(Sender: TObject); override;
  published
    { Minimal date value which can be entered }
    property MinDate: TDateTime read FMinDate write FMinDate;
    { Maximal date value which can be entered }
    property MaxDate: TDateTime read FMaxDate write FMaxDate;
    { Get or Set actual Date value }
    property DateValue: TDateTime read GetDate write SetDate;
    { define wheather record is update directly after enter or step change instead of after exit }
    property UpdateAfterChange: boolean read FUpdateAfterChange write FUpdateAfterChange default false;
  end;

  TDBMemoOption = (moAllowSetFont, moAllowWordWrap, moAllowLoadSaveFile);
  TDBMemoOptions = set of TDBMemoOption;

  { enhanced TDBMemo with bound label }
  TrDBMemo = class(TDBMemo)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FDefaultPrefix: string;
    FPopupMenu: TPopupMenu;
    FOptionsEx: TDBMemoOptions;
    procedure SetDefaultPrefix(const Value: string);
    procedure PopupMenuOpen(Sender: TObject);
    procedure PopupMenuClick(Sender: TObject);
  protected
    FScaleRatio: double;
    FSearchText: string;
    FSearchIgnoreCase: boolean;
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
    { search and select text if found }
    function SearchText(const Search: string = ''; FromStart: boolean = true; IgnoreCase: boolean = true): boolean;
    { search and select next text occurrence }
    function SearchNext: boolean;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define additional options - list of items in the popup menu:
      moAllowSetFont, moAllowWordWrap, moAllowLoadSaveFile }
    property OptionsEx: TDBMemoOptions read FOptionsEx write FOptionsEx default[moAllowSetFont, moAllowLoadSaveFile];
  end;

  { enhanced TDBListBox with bound label }
  TrDBListBox = class(TDBListBox)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FBoundLabel: rDBComponents.TBoundLabel;
    FDefaultPrefix: string;
    procedure SetDefaultPrefix(const Value: string);
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
  end;

  { enhanced TDBComboBox with bound label }
  TrDBComboBox = class(TDBComboBox)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FBoundLabel: rDBComponents.TBoundLabel;
    FDefaultPrefix: string;
    FDropDownAutoWidth: boolean;
    FDropDownMaxWidth: integer;
    FUpdateAfterClick: boolean;
    FDisplayedItems: TDisplayedItems;
    FGetListItemProps: TGetListItemProps;
    FNullValueKey: TShortCut;
    WndRecreated: boolean;
    IntTimer: TTimer;
    FUpdateTimerEvent: TNotifyEvent;
    FUpdateTimerDelay: cardinal;
    FOnUpdateLoadItems: TComboLoadItemsEvent;
    FLastKeyDown: Word;
    FJustClosed: boolean;
    FItemHeightEx: integer;
    procedure SetDefaultPrefix(const Value: string);
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure ListDefined(Sender: TObject);
    function UseOwnDraw: boolean;
    procedure IntTimerEvent(Sender: TObject);
    procedure SetItemHeightEx(Value: integer);
  protected
    FDataLink: TrDataLink;
    FScaleRatio: double;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure EditWndProc(var Message: TMessage); override;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Change; override;
    procedure Click; override;
    procedure WndProc(var Message: TMessage); override;
    procedure DropDown; override;
    procedure DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure CloseUp; override;
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
    { reset after change timer }
    procedure ResetTimer(SetActive, CallEventNow: boolean);
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { defines whether width of dropdown list is set according items length }
    property DropDownAutoWidth: boolean read FDropDownAutoWidth write FDropDownAutoWidth default true;
    { defines maximal width of dropdown list }
    property DropDownMaxWidth: integer read FDropDownMaxWidth write FDropDownMaxWidth default 600;
    { define wheather record is update directly after click instead of after exit }
    property UpdateAfterClick: boolean read FUpdateAfterClick write FUpdateAfterClick default false;
    { define additional item columns }
    property DisplayedItems: TDisplayedItems read FDisplayedItems write FDisplayedItems;
    { event for user definition of item properties }
    property OnGetListItemProps: TGetListItemProps read FGetListItemProps write FGetListItemProps;
    { define shortcut to set NULL value to the field }
    property NullValueKey: TShortCut read FNullValueKey write FNullValueKey default 0;
    { define after update timer interval }
    property UpdateTimerDelay: cardinal read FUpdateTimerDelay write FUpdateTimerDelay default 0;
    { event called with defined delay after update }
    property OnUpdateTimer: TNotifyEvent read FUpdateTimerEvent write FUpdateTimerEvent;
    { event called to load items during OnUpdateTimer event }
    property OnUpdateLoadItems: TComboLoadItemsEvent read FOnUpdateLoadItems write FOnUpdateLoadItems;
    { contain last pressed key }
    property LastKeyDown: Word read FLastKeyDown;
    { define item height for fixed draw before scaling }
    property ItemHeightEx: integer read FItemHeightEx write SetItemHeightEx default cDBComboBoxItemHeightDefault;
  end;

  { enhanced TDBLookupListBox with bound label }
  TrDBLookupListBox = class(TDBLookupListBox)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FBoundLabel: rDBComponents.TBoundLabel;
    FDefaultPrefix: string;
    FOnMouseClick: TNotifyEvent;
    procedure SetDefaultPrefix(const Value: string);
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define event after mouse up }
    property OnMouseClick: TNotifyEvent read FOnMouseClick write FOnMouseClick;
  end;

  { event for user definition of lookup dropdown list width }
  TGetLookupListItemHeight = procedure(Sender: TObject; Canvas: TCanvas; KeyField: TField; ListFields: string; var Height: integer) of object;
  { event for user definition of lookup dropdown list items }
  TGetLookupListItemProps = procedure(Sender: TObject; Canvas: TCanvas; RowIndex, ColIndex: integer; Selected: boolean; KeyField, ListField: TField;
    var Width: integer; Height: integer; var Text: string; ShowBmp: TBitmap; var DrawSeparatorTop: boolean; var DrawSeparatorBottom: boolean) of object;

  { @exclude }
  TrDBLookupControlHelper = class helper for TDBLookupControl
  protected
    function GetLookupMode: boolean;
  public
    function GetKeyField: TField;
  end;

  { @exclude }
  TrDBLookupListBoxHelper = class helper for TDBLookupListBox
  protected
    function GetRowCount: integer;
    function GetKeySelected: boolean;
    function GetRecordCount: integer;
    function GetRecordIndex: integer;
  end;

  { TPopupDataList for TrDBLookupComboBox to access Paint method }
  TrPopupDataList = class(TPopupDataList)
  private
    FItemHeight: integer;
    FItemIncreaseHeight: integer;
    FColWidth: array of integer;
    FItemIncreaseWidth: integer;
    FGetListItemProps: TGetLookupListItemProps;
    FProcessSearchKey: TKeyPressEvent;
  protected
    function GetTextHeight: integer; override;
    procedure Paint; override;
    procedure ProcessSearchKey(Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    property ItemHeight: integer read FItemHeight write FItemHeight;
    property ItemIncreaseHeight: integer read FItemIncreaseHeight write FItemIncreaseHeight;
    property ItemIncreaseWidth: integer read FItemIncreaseWidth write FItemIncreaseWidth;
    property GetListItemProps: TGetLookupListItemProps read FGetListItemProps write FGetListItemProps;
    property OnProcessSearchKey: TKeyPressEvent read FProcessSearchKey write FProcessSearchKey;
    function GetMaxItemWidth: integer;
    property Canvas;
  end;

  { @exclude }
  { helper for TDBLookupComboBox to access TPopupDataList }
  TrDBLookupComboBoxHelper = class helper for {$IFDEF DXE2P} TCustomDBLookupComboBox {$ELSE} TDBLookupComboBox {$ENDIF}
    protected function GetPopupList: TPopupDataList;
    procedure SetPopupList(L: TPopupDataList);
  end;

  { enhanced TDBLookupComboBox with bound label }
  TrDBLookupComboBox = class(TDBLookupComboBox)
  private
    FDataListEx: TrPopupDataList;
    FldValArr: array of string;
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FBoundLabel: rDBComponents.TBoundLabel;
    FDefaultPrefix: string;
    FEmptyDisplayValue: string;
    FDropDownButtonWidth: integer;
    FDropDownAutoWidth: boolean;
    FDropDownMaxWidth: integer;
    FDropDownItemHeight: integer;
    FDropDownItemIncreaseHeight: integer;
    FDropDownItemIncreaseWidth: integer;
    FListFieldIndexMax: integer;
    OwnHintWin: THintWindow;
    FTimer: TTimer;
    FSearchingResetInterval: integer;
    FSearchingPrevNext: TSearchingType;
    FGetListItemProps: TGetLookupListItemProps;
    FAllowLookupFilter: boolean;
    FAutoDropDown: boolean;
    function GetMaxItemWidth: integer;
    procedure SetDefaultPrefix(const Value: string);
    procedure WMKeyDown(var Message: TWMKeyDown); message WM_KEYDOWN;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure FOnTimer(Sender: TObject);
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure Paint; override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure ProcessSearchKey(Key: Char); override;
    procedure ListLinkDataChanged; override;
    procedure DoExit; override;
    procedure SearchNext(Previous: boolean);
    function SearchLookupFilter(ResetFilter: boolean): boolean;
    procedure StopSearching;
    function IsSearching: boolean;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DataListProcessSearchKey(Sender: TObject; var Key: Char);
    procedure DataListMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    procedure SetDataSource(ADataSource: TDataSource; ADataField: string); {< @exclude }
    procedure SetFocus; override; {< @exclude }
    procedure DropDown; override; {< @exclude }
    procedure CloseUp(Accept: boolean); override;
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
    { return default item height of standard lookup list }
    function DropDownDefaultItemHeight: integer;
    function GetSelectedText: string;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { defines whether width of dropdown list is set according items length }
    property DropDownAutoWidth: boolean read FDropDownAutoWidth write FDropDownAutoWidth default true;
    { defines maximal width of dropdown list }
    property DropDownMaxWidth: integer read FDropDownMaxWidth write FDropDownMaxWidth default 600;
    { defines absolute item height of dropdown list, default is used if DropDownItemHeight=0 }
    property DropDownItemHeight: integer read FDropDownItemHeight write FDropDownItemHeight default 0;
    { defines increase of standard item height of dropdown list }
    property DropDownItemIncreaseHeight: integer read FDropDownItemIncreaseHeight write FDropDownItemIncreaseHeight default 2;
    { defines increase of item column width if DropDownAutoWidth=true }
    property DropDownItemIncreaseWidth: integer read FDropDownItemIncreaseWidth write FDropDownItemIncreaseWidth default 0;
    { defines max ListField index which is show in edit window (own paint method is used) }
    property ListFieldIndexMax: integer read FListFieldIndexMax write FListFieldIndexMax default 0;
    { defines text value shown if KeyValue is empty }
    property EmptyDisplayValue: string read FEmptyDisplayValue write FEmptyDisplayValue;
    { event for user definition of lookup dropdown item properties }
    property OnGetListItemProps: TGetLookupListItemProps read FGetListItemProps write FGetListItemProps;
    { define interval without any keypress before reset searching text }
    property SearchingResetInterval: integer read FSearchingResetInterval write FSearchingResetInterval default 2500;
    { allow filtering of data for the lookup field }
    property AllowLookupFilter: boolean read FAllowLookupFilter write FAllowLookupFilter default false;
    { allow automatic dropdown after key press }
    property AutoDropDown: boolean read FAutoDropDown write FAutoDropDown default false;
  end;

  { enhanced TDBCheckBox providing readonly color and caption loaded
    from selected field }
  TrDBCheckBox = class(TDBCheckBox)
  private
    FShowFieldCaption: boolean;
    FReadOnlyColor: TColor;
    FDefaultPrefix: string;
    FShowDisabledIfReadOnly: boolean;
    FUpdateAfterClick: boolean;
    procedure SetDefaultPrefix(const Value: string);
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure UpdateCaption;
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    function GetDataField: string;
    procedure SetDataField(const Value: string);
    procedure SetShowFieldCaption(const Value: boolean);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetDataSource(DS: TDataSource); {< @exclude }
    procedure Click; override; {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); virtual;
  published
    property DataField: string read GetDataField write SetDataField;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define whether DisplayLabel from selected field is used as caption }
    property ShowFieldCaption: boolean read FShowFieldCaption write SetShowFieldCaption;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define wheather checkbox is shown disabled in readonly mode }
    property ShowDisabledIfReadOnly: boolean read FShowDisabledIfReadOnly write FShowDisabledIfReadOnly default false;
    { define wheather record is update directly after click instead of after exit }
    property UpdateAfterClick: boolean read FUpdateAfterClick write FUpdateAfterClick default false;
  end;

  { checkbox subcomponent used by TrDBTimePicker }
  TCheckBoxOnly = class(TCheckBox)
  private
    FInteriorColor: TColor;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
  protected
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    procedure SetPos(W: TWinControl; ScaleRatio: double); {< @exclude }
    procedure SetInteriorColor(C: TColor); {< @exclude }
    procedure Click; override; {< @exclude }
  end;

  { DB version of TDateTimePicker with bound label and null value checkbox }
  TrDBDateTimePicker = class(TDateTimePicker)
  private
    oldColor: TColor;
    FReadOnlyColor: TColor;
    FReadOnly: boolean;
    FDataAllowNull: boolean;
    FEnabledAll: boolean;
    FFocused: boolean;
    FChanging: boolean;
    FCheckBox: TCheckBoxOnly;
    FAutoSelectDayPart: boolean;
    FDBEditLabel: TDBBoundLabel;
    oldFormat: string;
    FFormatEmpty: string;
    FDefaultPrefix: string;
    procedure SetDefaultPrefix(const Value: string);
    function GetIsChecked: boolean;
    procedure SetSubComp;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CheckClick(Sender: TObject);
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure CNNotify(var Message: TWMNotifyDT); message CN_NOTIFY;
  protected
    FDataLink: TFieldDataLink;
    FScaleRatio: double;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetEnabledAll(Value: boolean);
    procedure SetDataAllowNull(Value: boolean);
    procedure SetParent(AParent: TWinControl); override;
    procedure SetFocused(Value: boolean);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure Change; override;
    procedure SetIntFormat;
    function CanEdit: boolean;
    procedure DoExit; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { define whether component is in read only mode }
    property ReadOnly: boolean read FReadOnly write FReadOnly default false;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define whether checkbox is shown and null value can be entered }
    property DataAllowNull: boolean read FDataAllowNull write SetDataAllowNull default false;
    { define whether all component including label is enable }
    property EnabledAll: boolean read FEnabledAll write SetEnabledAll default true;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { return whether null checkbox is checked }
    property IsChecked: boolean read GetIsChecked;
    { define format for empty (null) date }
    property FormatEmpty: string read FFormatEmpty write FFormatEmpty;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define whether day part is automatically selected for next enter }
    property AutoSelectDayPart: boolean read FAutoSelectDayPart write FAutoSelectDayPart default false;
  end;

  { DB version of TrImageZoom, provides zoom function with scrollbars
    and drag and move function }
  TrDBImage = class(TrImageZoom)
  private
    FDBEditLabel: TDBBoundLabel;
    FBoundButtons: TBoundButtons;
    FBoundToolBar: TToolBar;
    FBtnArr: array [1 .. rDBImageBtnCount] of TToolButton;
    Aborted: boolean;
    FShowConvertMenu: boolean;
    FPictureIfEmpty: TPicture;
    FDefaultPrefix: string;
    FMaxImageSize: integer;
    procedure SetDefaultPrefix(const Value: string);
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure RefBoundButtons(Sender: TObject);
    procedure ClickBoundButtons(Sender: TObject);
    procedure IntSaveToField(FieldType: TrGraphicType);
    procedure SetShowConvertMenu(Value: boolean);
    procedure OnAbort(Sender: TObject);
    procedure SetPictureIfEmpty(Value: TPicture);
  protected
    FDataLink: TFieldDataLink;
    procedure Loaded; override;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure ZoomChanged; override;
    function CanModify: boolean; override;
    function CanEdit: boolean;
    procedure OnPopupMenu(Sender: TObject); override;
    procedure PopupClick(Sender: TObject); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { cut image and insert to clipboard, works if dataset is in edit state or autoedit is on }
    function CutImageToClipboard: boolean; override;
    { paste image from clipboard, works if dataset is in edit state or autoedit is on }
    function PasteImageFromClipboard: boolean; override;
    { load image from file (BMP, JPG, PNG format is supported),
      works if dataset is in edit state or autoedit is on }
    function LoadFromFile(const FileName: string): boolean; override;
    { change resolution of image, works if dataset is in edit state or autoedit is on }
    function ResizePic(ratio: double): boolean; override;
    { show image in independent window supporting edit functions }
    function ShowEditorWindow: boolean; override;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bound buttons - ZoomIn, ZoomOit, Fit, Edit }
    property BoundButtons: TBoundButtons read FBoundButtons write FBoundButtons;
    { define whether convert menu is shown }
    property ShowConvertMenu: boolean read FShowConvertMenu write SetShowConvertMenu default true;
    { define maximum size of picture - bigger picture is resize
      when user load or copy image to field }
    property MaxImageSize: integer read FMaxImageSize write FMaxImageSize default 0;
    { define picture which is used as background for empty field }
    property PictureIfEmpty: TPicture read FPictureIfEmpty write SetPictureIfEmpty;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
  end;

  { DBVersion of TLabel, show db text from selected field as caption }
  TrDBLabel = class({$IFDEF UseHTMLLabel} TrHTMLLabel {$ELSE} TLabel {$ENDIF})
  private
    FDefaultPrefix: string;
    procedure SetDefaultPrefix(const Value: string);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
  protected
    FDataLink: TFieldDataLink;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
  end;

  { definition of possible formats in rDBMemoEx }
  TFormatType = (ftRTF, ftHTML);

  { enhanced TrDBMemo component which provides highlighting of words in text
    according list of words or highlighting of text between start and stop tag,
    contain Format and Edit buttons with menu and user defined text in header }
  TrDBMemoEx = class(TrDBMemo)
  private
    FColIncrement: integer;
    FRowIncrement: integer;
    FHighlightWords: TStringList;
    FHighlightWholeWords: boolean;
    FHighlightFromStr: string;
    FHighlightToStr: string;
    FHighlightedFontColor: TColor;
    FHighlightedBrushColor: TColor;
    FHighlightFromToEx: TStringList;
    FPageBreaksCharNumber: string;
    FButtonFrm: TBoundSpeedButton;
    FButtonEdit: TBoundSpeedButton;
    FBoundLabel: rDBComponents.TBoundLabel;
    FShowButtonFrm: boolean;
    FShowButtonEdit: boolean;
    FFormatPopupMenu: TPopupMenu;
    FStoredLines: TStrings;
    FFormatType: TFormatType;
    procedure SetHighlightWords(Value: TStringList);
    procedure SetHighlightWholeWords(Value: boolean);
    procedure SetHighlightFromStr(const Value: string);
    procedure SetHighlightToStr(const Value: string);
    procedure SetHighlightedFontColor(Value: TColor);
    procedure SetHighlightedBrushColor(Value: TColor);
    procedure SetHighlightFromToEx(Value: TStringList);
    procedure HighlightWordsChanged(Sender: TObject);
    procedure LoadSBPos;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure SetShowButtonFrm(Value: boolean);
    procedure SetShowButtonEdit(Value: boolean);
    // procedure SetBoundLabelRightMargin(Value: integer);
    procedure SetStoredLines(Value: TStrings);
    procedure RefreshBtPos;
    procedure RefreshBtState;
    procedure PopupClick(Sender: TObject);
  protected
    FScaleRatio: double;
    CloseOnCancel: boolean;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure Change; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure ButtonClick(Sender: TObject); virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure FormatInt(Tag: integer); virtual;
    procedure CreateFormatPopupMenu;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); override;
    procedure FormatBold;
    procedure FormatItalic;
    procedure FormatUnderline;
    procedure FormatHTMLUnnumberedList;
    procedure FormatHTMLNumberedList;
    procedure FormatNewPage;
    procedure FormatHTMLNewLine;
    procedure FormatHTMLNewLinesAuto;
    procedure FormatHTMLHorLine;
    procedure FormatHTMLLink;
    procedure FormatClear;
    { show text in independent window supporting edit functions }
    procedure ShowEditWin;
  published
    { list of hightlighted words, font color and background color can be defined in line on 2nd and 3rd place with TAB separator }
    property HighlightWords: TStringList read FHighlightWords write SetHighlightWords;
    { define whether only whole words is highlighted }
    property HighlightWholeWords: boolean read FHighlightWholeWords write SetHighlightWholeWords default false;
    { define start tag for hightlighting, text between start and stop tag is highlighted }
    property HighlightFromStr: string read FHighlightFromStr write SetHighlightFromStr;
    { define stop tag for hightlighting, text between start and stop tag is highlighted }
    property HighlightToStr: string read FHighlightToStr write SetHighlightToStr;
    { define font color of highlighted text }
    property HighlightedFontColor: TColor read FHighlightedFontColor write SetHighlightedFontColor default clWindowText;
    { define background color of highlighted text }
    property HighlightedBrushColor: TColor read FHighlightedBrushColor write SetHighlightedBrushColor default clMoneyGreen;
    { enhanced from-to hightlighting, list of FromStr + ToStr + font color + background color
      items is defined in line with TAB separator }
    property HighlightFromToEx: TStringList read FHighlightFromToEx write SetHighlightFromToEx;
    { define properties of bounded Format button }
    property ButtonFrm: TBoundSpeedButton read FButtonFrm;
    { define properties of bounded Edit button }
    property ButtonEdit: TBoundSpeedButton read FButtonEdit;
    { define visibility of Format button }
    property ShowButtonFrm: boolean read FShowButtonFrm write SetShowButtonFrm default false;
    { define visibility of Edit button }
    property ShowButtonEdit: boolean read FShowButtonEdit write SetShowButtonEdit default true;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define stored lines, can be used as default value for new record etc. }
    property StoredLines: TStrings read FStoredLines write SetStoredLines;
    { define number of chars used for showing of pagebreaks inside of edit area }
    property PageBreaksCharNumber: string read FPageBreaksCharNumber write FPageBreaksCharNumber;
    { define type of formatting used by FormatButton }
    property FormatType: TFormatType read FFormatType write FFormatType default ftRTF;
  end;

  { enhanced TrDBMemo with combobox function }
  TrDBMemoCB = class(TrDBMemo)
  private
    FButtonInsert: TBoundSpeedButton;
    FButtonAdd: TBoundSpeedButton;
    FListBox: TBoundListBox;
    FShowButtonInsert: boolean;
    FShowButtonAdd: boolean;
    FInserting: boolean;
    FShowListBox: TNotifyEvent;
    FNoDuplicLine: boolean;
    procedure SetShowButtonInsert(Value: boolean);
    procedure SetShowButtonAdd(Value: boolean);
    procedure RefreshBtPos;
    procedure RefreshBtState;
  protected
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure ButtonClick(Sender: TObject); virtual;
    procedure ListBoxClick(Sender: TObject); virtual;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { update current state of component (called automatically) }
    procedure UpdateState(Sender: TObject); override;
  published
    { define properties of bounded Insert button }
    property ButtonInsert: TBoundSpeedButton read FButtonInsert;
    { define properties of bounded Add button }
    property ButtonAdd: TBoundSpeedButton read FButtonAdd;
    { define properties of bounded ListBox }
    property ListBox: TBoundListBox read FListBox;
    { define visibility of Insert button }
    property ShowButtonInsert: boolean read FShowButtonInsert write SetShowButtonInsert default true;
    { define visibility of Add button }
    property ShowButtonAdd: boolean read FShowButtonAdd write SetShowButtonAdd default true;
    { event called before showing of bounded listbox }
    property OnShowListBox: TNotifyEvent read FShowListBox write FShowListBox;
    { define  whrether existing line can be added }
    property NoDuplicLine: boolean read FNoDuplicLine write FNoDuplicLine default false;
  end;

  { class to store one item in TrDBCheckListBox }
  TCheckItem = class
  public
    Id: integer;
    Text: string;
    Checked: boolean;
    constructor Create(ItemId: integer; const ItemText: string; IsChecked: boolean);
  end;

  { class to store list of TCheckItem in TrDBCheckListBox }
  TCheckItemList = class(TObjectList)
  protected
    FDuplicIdArr: array of integer;
  public
    function CheckItem(Index: integer): TCheckItem;
    procedure AddCheckItem(ItemId: integer; const ItemText: string; IsChecked: boolean);
    procedure SetItemChecked(ItemId: integer; IsChecked: boolean);
    function GetItemChecked(ItemId: integer): boolean;
    procedure ResetItemsChecked;
    function CheckedCount: integer;
    procedure StoreItemsForDuplication;
    procedure RestoreDuplicatedItems;
  end;

  { list of possible items order for TrDBCheckListBox }
  TCheckItemsStyle = (chsAllItems, chsCheckedFirst, chsCheckedOnly);
  { list of possible items visibility for TrDBCheckListBox }
  TFindItemsStyle = (fisNone, fisFilter, fisSelect);
  { event type for TrDBCheckListBox }
  TLoadSaveItemList = procedure(Sender: TObject; Items: TCheckItemList) of object;

  { DB version of TCheckListBox for n:m relationship }
  TrDBCheckListBox = class(TCheckListBox)
  private
    oldColor: TColor;
    oldEditState: boolean;
    FChanging: boolean;
    FPostponedSaving: boolean;
    FPostponeSaveNewRecord: boolean;
    FReadOnlyColor: TColor;
    FDBEditLabel: TDBBoundLabel;
    FBoundLabel: rDBComponents.TBoundLabel;
    FFindEdit: TBoundButtonedEdit;
    FFindEditSpace: integer;
    FFindItemsStyle: TFindItemsStyle;
    oldFindText: string;
    FReadOnly: boolean;
    FOnLoadItemList: TLoadSaveItemList;
    FOnLoadItemValues: TLoadSaveItemList;
    FOnSaveItemValues: TLoadSaveItemList;
    FDefaultPrefix: string;
    FCheckItemsStyle: TCheckItemsStyle;
    FCheckboxUpdateMode: TCheckboxUpdateMode;
    procedure SetDefaultPrefix(const Value: string);
    function CanEdit: boolean;
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure UpdateData(Sender: TObject);
    procedure SetReadOnly(Value: boolean);
    procedure SetCheckItemsStyle(Value: TCheckItemsStyle);
    procedure SetFindEditSpace(Value: integer);
    procedure SetFindItemsStyle(Value: TFindItemsStyle);
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
  protected
    FDataLink: TFieldDataLink;
    FAllItems: TCheckItemList;
    procedure Loaded; override;
    procedure ClickCheck; override;
    function GetDataField: string;
    function GetDataSource: TDataSource;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    function GetField: TField;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure ChangeFindText(Sender: TObject);
    function CanModifyCheckItem(Index: integer): boolean; virtual;
    procedure DblClick; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { update state of all items according dataset state }
    procedure UpdateChecklistState; virtual;
    { load list of available items to FAllItemsList }
    procedure DoLoadItemList; virtual;
    { load and set state (check/uncheck) of all items }
    procedure DoLoadItemValues; virtual;
    { store state (check/uncheck) of all items }
    procedure DoSaveItemValues; virtual;
    { cancel all changes and load previous state }
    procedure DoCancelItemValues; virtual;
    { indicates count of checked items }
    function CheckedCount: integer;
    { show list of items in CheckListBox according filter state }
    procedure FilterItemsList; virtual;
    { select item in the list of items in CheckListBox according filter text }
    procedure SelectItemsList; virtual;
    { returns linked Field }
    property Field: TField read GetField;
    { programmatically set state of one item }
    procedure SetItemState(AItemIndex: integer; AChecked: boolean);
    { programmatically set state of all items }
    procedure SetAllItemsState(AChecked: boolean);
    { store state (check/uncheck) of all items for duplication to another record }
    procedure StoreItemsForDuplication;
    { restore state (check/uncheck) of all items from state stored by StoreItemsForDuplication }
    procedure RestoreDuplicatedItems;
    { get list of checked items }
    function GetCheckedItemsList(Separator: string; UseItemObjectAsInt: boolean = false): string;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReadOnly: boolean read FReadOnly write SetReadOnly default false;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define color for not edit state }
    property ReadOnlyColor: TColor read FReadOnlyColor write FReadOnlyColor default clBtnFace;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of bounded label }
    property BoundLabel: rDBComponents.TBoundLabel read FBoundLabel;
    { define properties of bound find editor }
    property FindEdit: TBoundButtonedEdit read FFindEdit;
    { define horizontal space between checkbox and find edit }
    property FindEditSpace: integer read FFindEditSpace write SetFindEditSpace default 4;
    { define using of Find edit box below listbox }
    property FindItemsStyle: TFindItemsStyle read FFindItemsStyle write SetFindItemsStyle default fisNone;
    { define order of items in listbox }
    property CheckItemsStyle: TCheckItemsStyle read FCheckItemsStyle write SetCheckItemsStyle default chsAllItems;
    { event for loading of item list }
    property OnLoadItemList: TLoadSaveItemList read FOnLoadItemList write FOnLoadItemList;
    { event for loading of item values }
    property OnLoadItemValues: TLoadSaveItemList read FOnLoadItemValues write FOnLoadItemValues;
    { event for saving of item values }
    property OnSaveItemValues: TLoadSaveItemList read FOnSaveItemValues write FOnSaveItemValues;
    { indicate that values should be saved after saving of record (used for new record without ID value before post) }
    property PostponedSaving: boolean read FPostponedSaving;
    { define whether postponed saving should be used for new record (to avoid database error in relationship }
    property PostponeSaveNewRecord: boolean read FPostponeSaveNewRecord write FPostponeSaveNewRecord default false;
    { define mode for checkbox update }
    property CheckboxUpdateMode: TCheckboxUpdateMode read FCheckboxUpdateMode write FCheckboxUpdateMode default chmCheckboxClick;
  end;

  { enhanced TrDBCheckListBox component for n:m relationship which provides
    links to ListDataSource and ValueDataSource for automatic work with checklistbox }
  TrDBCheckListBoxEx = class(TrDBCheckListBox)
  private
    function GetListSource: TDataSource;
    procedure SetListSource(Value: TDataSource);
    function GetValuesSource: TDataSource;
    procedure SetValuesSource(Value: TDataSource);
  protected
    FListLink: TDataLink;
    FValuesLink: TDataLink;
    FListKeyFieldName: string;
    FListDisplayFieldName: string;
    FValuesDataKeyField: string;
    FValuesListKeyField: string;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { load list of available items to FAllItemsList }
    procedure DoLoadItemList; override;
    { load and set state (check/uncheck) of all items }
    procedure DoLoadItemValues; override;
    { store state (check/uncheck) of all items }
    procedure DoSaveItemValues; override;
  published
    { define datasource with List of Items in listbox }
    property ListSource: TDataSource read GetListSource write SetListSource;
    { define Key (ID) Field from ListSource }
    property ListKeyField: string read FListKeyFieldName write FListKeyFieldName;
    { define displayed Fields from ListSource, Fields or "constants" are separated by ; }
    property ListDisplayField: string read FListDisplayFieldName write FListDisplayFieldName;
    { define datasource with records matching DataSource and ListSource }
    property ValuesSource: TDataSource read GetValuesSource write SetValuesSource;
    { define Key (ID) Field for DataSource key value }
    property ValuesDataKeyField: string read FValuesDataKeyField write FValuesDataKeyField;
    { define Key (ID) Field for ListSource key value }
    property ValuesListKeyField: string read FValuesListKeyField write FValuesListKeyField;
  end;

  { @exclude }
  { TLabel subcomponent used by TrDBBlob }
  TChildLabel = class(TCustomLabel)
  protected
    function GetLabelText: string; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  published
    property Align;
    property Alignment;
    property EllipsisPosition;
    property Caption;
    property Color;
    property Font;
    property ParentColor;
    property ParentFont;
    property Transparent;
  end;

  { @exclude }
  { Button subcomponent used by rDBBlob }
  TChildBtn = class(TButton)
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  end;

  { event type used by TrDBBlob }
  TBlobFileEvent = procedure(Sender: TObject; const FileName: string; FileSize: Int64) of object;

  { DB component for working with Blob fields,
    contains function for loading, saving and opening of any content in blob }
  TrDBBlob = class(TPanel)
  private
    FDBEditLabel: TDBBoundLabel;
    FInfoLabel: TChildLabel;
    FBtnOpen: TChildBtn;
    FBtnLoad: TChildBtn;
    FBtnSave: TChildBtn;
    FBtnClear: TChildBtn;
    FBtnMargin: integer;
    FBtnSpace: integer;
    FPopupMenu: TPopupMenu;
    FFileFilter: string;
    FFileSizeLimitMB: double;
    FOnLoadBlob: TBlobFileEvent;
    FOnOpenBlob: TBlobFileEvent;
    FOpenByApp: string;
    FOpenParams: string;
    FOpenTempDir: string;
    FPrepareBlobDataset: TNotifyEvent;
    FShowProgressWindow: boolean;
    Aborted: boolean;
    FDefaultPrefix: string;

    procedure SetDefaultPrefix(const Value: string);
    procedure RefBoundButtons(Sender: TObject);
    procedure ClickBoundButtons(Sender: TObject);
    function CanEdit: boolean;
    procedure DataChange(Sender: TObject);
    procedure SetBtnMargin(Value: integer);
    procedure SetBtnSpace(Value: integer);
    procedure OnAbort(Sender: TObject);
    function IsLinked: boolean;
    procedure OnPopupMenu(Sender: TObject);
    function GetPopupItem(Tag: integer): TMenuItem;
  protected
    FDataLinkBlob: TFieldDataLink;
    FDataLinkInfo: TFieldDataLink;
    procedure Loaded; override;
    function GetDataFieldBlob: string;
    function GetDataSourceBlob: TDataSource;
    procedure SetDataFieldBlob(const Value: string);
    procedure SetDataSourceBlob(Value: TDataSource);
    function GetDataFieldInfo: string;
    function GetDataSourceInfo: TDataSource;
    procedure SetDataFieldInfo(const Value: string);
    procedure SetDataSourceInfo(Value: TDataSource);
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { save blob content to disk and run }
    procedure OpenBlob;
    { save blob to file }
    function SaveAsFile(const FileName: string; ProgressInfo: string = ''; InfoStayVisible: boolean = false): boolean;
    { load blob from file }
    function LoadFromFile(const FileName: string; ProgressInfo: string = ''; InfoStayVisible: boolean = false): boolean;
    { clear blob field and info field, works if dataset is in edit state or autoedit is on }
    procedure ClearBlob;
    { rename info field, works if dataset is in edit state or autoedit is on }
    procedure RenameBlob;
    { return text info with name and size of blob }
    function BlobInfo: string;
  published
    { datalink to blob field containing data }
    property DataFieldBlob: string read GetDataFieldBlob write SetDataFieldBlob;
    { datalink to blob field containing data }
    property DataSourceBlob: TDataSource read GetDataSourceBlob write SetDataSourceBlob;
    { datalink to info field containing filename and filetype }
    property DataFieldInfo: string read GetDataFieldInfo write SetDataFieldInfo;
    { datalink to info field containing filename and filetype }
    property DataSourceInfo: TDataSource read GetDataSourceInfo write SetDataSourceInfo;
    { define prefix for auto-named function (name of component is set according datafield) }
    property DefaultPrefix: string read FDefaultPrefix write SetDefaultPrefix;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
    { define properties of info label }
    property InfoLabel: TChildLabel read FInfoLabel;
    { define properties of Open button }
    property BtnOpen: TChildBtn read FBtnOpen;
    { define properties of Load button }
    property BtnLoad: TChildBtn read FBtnLoad;
    { define properties of Save button }
    property BtnSave: TChildBtn read FBtnSave;
    { define properties of Clear button }
    property BtnClear: TChildBtn read FBtnClear;
    { define margin for buttons }
    property BtnMargin: integer read FBtnMargin write SetBtnMargin;
    { define space between buttons }
    property BtnSpace: integer read FBtnSpace write SetBtnSpace;
    { define application used to open blob content }
    property OpenByApp: string read FOpenByApp write FOpenByApp;
    { define parameters used to open blob content }
    property OpenParams: string read FOpenParams write FOpenParams;
    { define directory used to store blob content for opening }
    property OpenTempDir: string read FOpenTempDir write FOpenTempDir;
    { define file filter used in OpenFile dialog }
    property FileFilter: string read FFileFilter write FFileFilter;
    { define limit for size of loaded file }
    property FileSizeLimitMB: double read FFileSizeLimitMB write FFileSizeLimitMB;
    { event is called after file is selected before start loading of file content }
    property OnLoadBlob: TBlobFileEvent read FOnLoadBlob write FOnLoadBlob;
    { event is called after blob is stored and before file is open }
    property OnOpenBlob: TBlobFileEvent read FOnOpenBlob write FOnOpenBlob;
    { event is called each time before blob dataset is accessed -
      user can open and load data in this event }
    property OnPrepareBlobDataset: TNotifyEvent read FPrepareBlobDataset write FPrepareBlobDataset;
    { define whether progress window is shown during time consuming operations }
    property ShowProgressWindow: boolean read FShowProgressWindow write FShowProgressWindow default true;
  end;

  { list of possible loading event used by TrDBRecordSelection }
  TLoadListEvent = (leAfterOpen, leAfterDropDownFirst, leAfterDropDownAlways, leNever);
  { list of possible state used by TrDBRecordSelection }
  TEditingState = (esDisable, esReadOnly, esAskBeforeMove, esPostBeforeMove);

  { DB Combobox for showing and selecting of active record,
    show active record and drop down show list of all records with possibility to select active record }
  TrDBRecordSelection = class(TCustomComboBox)
  private
    FRecIdField: string;
    FListFields: string;
    FLoadListEvent: TLoadListEvent;
    FEditingState: TEditingState;
    FIsLoadingItems: boolean;
    FItemsLoaded: boolean;
    FDBEditLabel: TDBBoundLabel;
    procedure SetDataSource(Value: TDataSource);
    function GetDataSource: TDataSource;
    procedure SetRecIdField(const Value: string);
    procedure SetListFields(const Value: string);
    procedure ActiveChange(Sender: TObject);
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    procedure Resort;
  protected
    FDataLink: TrDataLink;
    procedure Loaded; override;
    procedure DropDown; override;
    procedure LoadActiveRec;
    procedure SetActiveRec;
    function GetRecordListId: integer;
    function GetRecordListText: string;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure KeyPress(var Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Click; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { load list of records (called automatically) }
    function LoadList: integer;
  published
    property Align;
    property AutoComplete default true;
    property AutoCompleteDelay default 500;
    property AutoDropDown default false;
    property AutoCloseUp default false;
    property BevelEdges;
    property BevelInner;
    property BevelKind default bkNone;
    property BevelOuter;
    property Style; { Must be published before Items }
    property Anchors;
    property BiDiMode;
    property Color;
    property Constraints;
    property Ctl3D;
    property DoubleBuffered;
    property DragCursor;
    property DragKind;
    property DragMode;
    property DropDownCount;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property ItemHeight;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentDoubleBuffered;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property TextHint;
    property Touch;
    property Visible;
    property OnChange;
    property OnClick;
    property OnCloseUp;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGesture;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnSelect;
    property OnStartDock;
    property OnStartDrag;
    { define datasource to select record }
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    { define ID field name which identifies record, must be numeric field }
    property RecIdField: string read FRecIdField write SetRecIdField;
    { define list of fields and constants for each item, field separator=;  constant separator="
      e.g. ProductName;";  (id=";ProductId;")" }
    property ListFields: string read FListFields write SetListFields;
    { define event when the list is loaded }
    property LoadListEvent: TLoadListEvent read FLoadListEvent write FLoadListEvent default leAfterOpen;
    { define action if editing is active and use select another record }
    property EditingState: TEditingState read FEditingState write FEditingState default esDisable;
    { define properties of bound label }
    property DBEditLabel: TDBBoundLabel read FDBEditLabel;
  end;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rfiletool, rdbtool, rdlg, rDialogsEx, rImageEdit;

const
  DBLabelTopSpacing = 1;
  DBLabelLeftSpacing = 3;
  DBCheckBoxWidth = 16;
  // ListBoxItemDefHeight = 15;

  cFormatBold = 1;
  cFormatItalic = 2;
  cFormatUnderline = 3;
  cFormatAsUnnumberedList = 11;
  cFormatAsNumberedList = 12;
  cFormatRTFNewPage = 21;
  cFormatHTMLNewLine = 22;
  cFormatHTMLNewLineAuto = 23;
  cFormatHTMLHorLine = 24;
  cFormatHTMLLink = 31;
  cFormatClear = 99;

  cPicConvertMenu = 40;
  sPicConvertToJPG1 = 41;
  sPicConvertToJPGAll = 42;
  sPicConvertToBMP1 = 43;
  sPicConvertToBMPAll = 44;

  cBlobNone = 0;
  cBlobOpen = 1;
  cBlobLoad = 2;
  cBlobSave = 3;
  cBlobClear = 4;
  cBlobInfo = 10;
  cBlobRename = 11;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 50 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBMoveButton, TrDBToolButton, TrDBEdit, TrDBEditDate, TrDBMemo, TrDBMemoEx, TrDBMemoCB, TrDBListBox, TrDBLookupListBox,
    TrDBComboBox, TrDBLookupComboBox, TrDBCheckBox, TrDBDateTimePicker, TrDBLabel, TrDBImage, TrDBBlob, TrDBCheckListBox, TrDBCheckListBoxEx,
    TrDBRecordSelection]);
end;

procedure SetDBCompName(C: TComponent; DefaultPrefix, FieldName: string);
var
  I: integer;
  suf: string;
begin
  if not(csDesigning in C.ComponentState) or (DefaultPrefix = '') then
    Exit;
  if FieldName = '' then
    FieldName:='NoDataField';
  I:=0;
  repeat
    try
      if I = 0 then
        suf:=''
      else
        suf:=IntToStr(I);
      C.Name:=DefaultPrefix + FieldName + suf;
      Exit;
    except
      Inc(I);
    end;
  until false;
end;

procedure SetVisibilityFromField(Sender: TControl; DataSource: TDataSource; DataField: string);
var
  F: TField;
begin
  if Assigned(DataSource) and Assigned(DataSource.DataSet) then
  begin
    F:=DataSource.DataSet.FindField(DataField);
    if Assigned(F) then
    begin
      if F.Visible and RosiCompConfig.rDBComponentsCfg.VisibilityFromField_AutoShow then
        Sender.Visible:=true;
      if (not F.Visible) and RosiCompConfig.rDBComponentsCfg.VisibilityFromField_AutoHide then
        Sender.Visible:=false;
    end;
  end;
end;

function GetButtonState(FDataLink: TrDataLink; ButtonType: TrDBButtonType): boolean;
begin
  Result:=false;
  if not Assigned(FDataLink) or not FDataLink.Active then
    Exit;
  try
    case ButtonType of
      btFirst:
        Result:=not FDataLink.GetBOF and not FDataLink.Editing;
      btPrior:
        Result:=not FDataLink.GetBOF and not FDataLink.Editing;
      btNext:
        Result:=not FDataLink.GetEOF and not FDataLink.Editing;
      btLast:
        Result:=not FDataLink.GetEOF and not FDataLink.Editing;

      btAdd, btInsert:
        Result:=not IsEditing(FDataLink.DataSet) and FDataLink.DataSet.CanModify;
      btDuplic:
        Result:=not IsEditing(FDataLink.DataSet) and FDataLink.DataSet.CanModify and not FDataLink.DataSet.IsEmpty;
      btEdit, btDelete:
        Result:=not IsEditing(FDataLink.DataSet) and FDataLink.DataSet.CanModify and not FDataLink.DataSet.IsEmpty;
      btPost, btCancel:
        Result:=IsEditing(FDataLink.DataSet);

      btEnaIfEdit:
        Result:=IsEditing(FDataLink.DataSet);
      btEnaIfNew:
        Result:=FDataLink.DataSet.State = dsInsert;
      btEnaIfCanModify:
        Result:=FDataLink.DataSet.CanModify;
      btEnaIfBrowse:
        Result:=not IsEditing(FDataLink.DataSet);
      btEnaIfNotEmpty:
        Result:=not FDataLink.DataSet.IsEmpty;
      btEnaIfBrowseNotEmpty:
        Result:=not IsEditing(FDataLink.DataSet) and not FDataLink.DataSet.IsEmpty;
      btEnaIfBrowseCanEdit:
        Result:=not IsEditing(FDataLink.DataSet) and FDataLink.DataSet.CanModify;

    end;
  except
    Result:=false;
  end;
end;

procedure DBButtonClick(FDataLink: TrDataLink; ButtonType: TrDBButtonType);
begin
  case ButtonType of
    btFirst:
      FDataLink.DataSet.First;
    btPrior:
      FDataLink.DataSet.Prior;
    btNext:
      FDataLink.DataSet.Next;
    btLast:
      FDataLink.DataSet.Last;
    btPost:
      FDataLink.DataSet.Post;
    btCancel:
      FDataLink.DataSet.Cancel;
    btAdd:
      FDataLink.DataSet.Append;
    btInsert:
      FDataLink.DataSet.Insert;
    btEdit:
      FDataLink.DataSet.Edit;
    btDelete:
      FDataLink.DataSet.Delete;
  end;
end;

// ==============================================================================
// TrDataLink class
constructor TrDataLink.Create(ActiveChEvent, EditingChEvent, DataSetChEvent: TNotifyEvent);
begin
  inherited Create;
  FActiveChEvent:=ActiveChEvent;
  FEditingChEvent:=EditingChEvent;
  FDataSetChEvent:=DataSetChEvent;
  VisualControl:=true;
end;

procedure TrDataLink.ActiveChanged;
begin
  inherited ActiveChanged;
  if Assigned(FActiveChEvent) then
    FActiveChEvent(Self);
end;

procedure TrDataLink.EditingChanged;
begin
  inherited EditingChanged;
  if Assigned(FEditingChEvent) then
    FEditingChEvent(Self);
end;

procedure TrDataLink.DataSetChanged;
begin
  inherited DataSetChanged;
  if Assigned(FDataSetChEvent) then
    FDataSetChEvent(Self);
end;

function TrDataLink.Editing: boolean;
begin
  Result:=(not Assigned(DataSource)) or IsEditing(DataSource);
end;

// ==============================================================================
// TrDBMoveButton
constructor TrDBMoveButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, UpdateState);
  FButtonType:=btEnaIfEdit;
  FEnabledAll:=true;
  FSkipDBAction:=false;
end;

destructor TrDBMoveButton.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBMoveButton.Loaded;
begin
  inherited Loaded;
  UpdateState(nil);
end;

function TrDBMoveButton.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBMoveButton.SetDataSource(Value: TDataSource);
begin
  if Value <> FDataLink.DataSource then
  begin
    FDataLink.DataSource:=Value;
    UpdateState(nil);
  end;
end;

procedure TrDBMoveButton.SetEnabledAll(Value: boolean);
begin
  if Value <> FEnabledAll then
  begin
    FEnabledAll:=Value;
    UpdateState(nil);
  end;
end;

procedure TrDBMoveButton.UpdateState(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  Enabled:=GetButtonState(FDataLink, FButtonType) and FEnabledAll;
end;

procedure TrDBMoveButton.Click;
begin
  if not FSkipDBAction then
    DBButtonClick(FDataLink, FButtonType);
  inherited;
end;

// ==============================================================================
// TrDBToolButton
constructor TrDBToolButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEnabledAll:=true;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, UpdateState);
  FButtonType:=btEnaIfEdit;
end;

destructor TrDBToolButton.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBToolButton.Loaded;
begin
  inherited Loaded;
  UpdateState(nil);
end;

function TrDBToolButton.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBToolButton.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource:=Value;
  UpdateState(nil);
end;

procedure TrDBToolButton.SetEnabledAll(Value: boolean);
begin
  if Value <> FEnabledAll then
  begin
    FEnabledAll:=Value;
    UpdateState(nil);
  end;
end;

procedure TrDBToolButton.UpdateState(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  Enabled:=GetButtonState(FDataLink, FButtonType) and FEnabledAll;
end;

procedure TrDBToolButton.Click;
begin
  if not FSkipDBAction then
    DBButtonClick(FDataLink, FButtonType);
  inherited;
end;

// ==============================================================================
// TDBBoundLabel
constructor TDBBoundLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Name:='SubDBLabel';
  SetSubComponent(true);
  FField:=nil;
  FLabelFieldCaption:='';
  FPosition:=lpTopLeft;
  Transparent:=true;
  FShowRequiredBold:=false;
  FShowRequiredMark:=true;
  FRequiredMarkColor:=clRed;
  FRequiredMarkSpace:=2;
  FFieldCaptionSuffix:=':';
  if AOwner is TWinControl then
    FocusControl:=AOwner as TWinControl;
end;

procedure TDBBoundLabel.SetPosition(Value: TLabelPos);
begin
  FPosition:=Value;
  RefreshCaption;
end;

procedure TDBBoundLabel.SetField(DataSource: TDataSource; FieldName: string);
begin
  FField:=nil;
  try
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
      FField:=DataSource.DataSet.FindField(FieldName);
  except
  end;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetLabelField(DataSource: TDataSource; FieldName: string);
var
  F: TField;
begin
  FLabelFieldCaption:='';
  try
    if Assigned(DataSource) and Assigned(DataSource.DataSet) then
    begin
      F:=DataSource.DataSet.FindField(FieldName);
      if Assigned(F) then
        FLabelFieldCaption:=F.DisplayLabel;
    end;
  except
  end;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetOwnCaption(const Value: string);
begin
  FOwnCaption:=Value;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetShowRequiredBold(Value: boolean);
begin
  FShowRequiredBold:=Value;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetShowRequiredMark(Value: boolean);
begin
  FShowRequiredMark:=Value;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetRequiredMarkColor(Value: TColor);
begin
  FRequiredMarkColor:=Value;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetRequiredMarkSpace(Value: integer);
begin
  FRequiredMarkSpace:=Value;
  RefreshCaption(true);
end;

procedure TDBBoundLabel.SetFieldCaptionSuffix(Value: string);
begin
  FFieldCaptionSuffix:=Value;
  RefreshCaption(true);
end;

function TDBBoundLabel.IsFieldCaptionSuffixStored: boolean;
begin
  Result:=FFieldCaptionSuffix <> ':';
end;

procedure TDBBoundLabel.RequestAlign;
begin
  inherited;
  RefreshPosition;
end;

procedure TDBBoundLabel.RefreshCaption(RecalcBounds: boolean = false);
var
  capt: string;
begin
  capt:='';
  if Visible or not(csDesigning in ComponentState) then
    try
      if FOwnCaption <> '' then
        capt:=FOwnCaption
      else
        if FLabelFieldCaption <> '' then
          capt:=FLabelFieldCaption + FFieldCaptionSuffix
        else
          if Assigned(FField) then
            capt:=FField.DisplayLabel + FFieldCaptionSuffix
          else
            if (csDesigning in ComponentState) then
              capt:='FieldLabel';
    except
      capt:=Name;
    end;
  try
    Caption:=capt;
    RefreshPosition(RecalcBounds);
  except
  end;
end;

procedure TDBBoundLabel.DoDrawText(var Rect: TRect; Flags: Longint);
var
  CalcOnly: boolean;
  W: integer;
  F: TFontRecall;
begin
  if not(Assigned(FField) and FField.Required and FShowRequiredBold) then
    inherited DoDrawText(Rect, Flags);
  if Assigned(FField) and FField.Required then
  begin
    CalcOnly:=(Flags and DT_CALCRECT) <> 0;
    F:=TFontRecall.Create(Canvas.Font);
    try
      if FShowRequiredBold then
      begin
        Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
        W:=Canvas.TextWidth(GetLabelText);
        if CalcOnly then
          Rect.Right:=Rect.Left + W;
        if not CalcOnly then
          Canvas.TextRect(Rect, Rect.Left, Rect.Top, GetLabelText);
      end;
      if FShowRequiredMark then
      begin
        Canvas.Font.Color:=FRequiredMarkColor;
        Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
        W:=Canvas.TextWidth('*');
        if CalcOnly then
          Rect.Right:=Rect.Right + W + FRequiredMarkSpace;
        if not CalcOnly then
          Canvas.TextOut(Rect.Right - W, Rect.Top, '*');
      end;
    finally
      FreeAndNil(F);
    end;
  end;
end;

procedure TDBBoundLabel.RefreshPosition(RecalcBounds: boolean = false);
var
  W: TControl;
  ml, mr, mt, mb: integer;
  dx: integer;
begin
  if Owner is TControl then
    W:=TControl(Owner)
  else
    Exit;
  if AlignWithMargins then
  begin
    ml:=Margins.Left;
    mr:=Margins.Right;
    mt:=Margins.Top;
    mb:=Margins.Bottom;
  end
  else
  begin
    ml:=0;
    mr:=0;
    mt:=0;
    mb:=0;
  end;
  dx:=0;
  if W is TrDBDateTimePicker then
    with TrDBDateTimePicker(W) do
      if DataAllowNull then
        dx:=-Round(DBCheckBoxWidth * FScaleRatio);
  if RecalcBounds then
    AdjustBounds;
  case FPosition of
    lpTopLeft:
      SetBounds(W.Left + dx + ml, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpTopCenter:
      SetBounds((W.Width - Width - ml - mr) div 2 + W.Left + dx + ml, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpTopRight:
      SetBounds(W.Left + W.Width - Width + dx - mr, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpLeftTop:
      SetBounds(W.Left - Width - DBLabelLeftSpacing + dx - ml, W.Top + mt, Width, Height);
    lpLeftCenter:
      SetBounds(W.Left - Width - DBLabelLeftSpacing + dx - ml, (W.Height - Height) div 2 + W.Top, Width, Height);
    lpLeftBottom:
      SetBounds(W.Left - Width - DBLabelLeftSpacing + dx - ml, W.Top + W.Height - Height - mb, Width, Height);
  end;
end;

procedure TDBBoundLabel.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  RefreshCaption;
end;

procedure TDBBoundLabel.ChangeVisible(AVisible: boolean);
begin
  if AVisible = Visible then
    Exit;
  if AVisible then
    Visible:=oldVisible
  else
  begin
    oldVisible:=Visible;
    Visible:=false;
  end;
end;

procedure TDBBoundLabel.StoreVisible;
begin
  oldVisible:=Visible;
end;

// ==============================================================================
// TBoundLabel
constructor TBoundLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
end;

procedure TBoundLabel.SetPosition(Value: TLabelPos);
begin
  FPosition:=Value;
  if Assigned(Owner) and (Owner is TControl) then
    RefreshPosition;
end;

procedure TBoundLabel.RequestAlign;
begin
  inherited;
  RefreshPosition;
end;

procedure TBoundLabel.RefreshPosition;
var
  W: TControl;
  ml, mr, mt, mb: integer;
begin
  if Owner is TControl then
    W:=TControl(Owner)
  else
    Exit;
  if AlignWithMargins then
  begin
    ml:=Margins.Left;
    mr:=Margins.Right;
    mt:=Margins.Top;
    mb:=Margins.Bottom;
  end
  else
  begin
    ml:=0;
    mr:=0;
    mt:=0;
    mb:=0;
  end;
  case FPosition of
    lpTopLeft:
      SetBounds(W.Left + ml, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpTopCenter:
      SetBounds((W.Width - Width - ml - mr) div 2 + W.Left + ml, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpTopRight:
      SetBounds(W.Left + W.Width - Width - mr, W.Top - Height - DBLabelTopSpacing - mt, Width, Height);
    lpLeftTop:
      SetBounds(W.Left - Width - DBLabelLeftSpacing - ml, W.Top + mt, Width, Height);
    lpLeftCenter:
      SetBounds(W.Left - Width - DBLabelLeftSpacing - ml, (W.Height - Height) div 2 + W.Top, Width, Height);
    lpLeftBottom:
      SetBounds(W.Left - Width - DBLabelLeftSpacing - ml, W.Top + W.Height - Height - mb, Width, Height);
  end;
end;

// ==============================================================================
// TBoundButtons
constructor TBoundButtons.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Name:='SubBoundButtons';
  SetSubComponent(true);
  FCaptions:='Fit;+;-;Edit';
  FVisible:=false;
  FShowGlyphs:=false;
  FPosition:=bpBottomLeft;
  FFlat:=true;
  FTransparent:=true;
  FSize:=23;
  FImageChangeLink:=TChangeLink.Create;
  FImageChangeLink.OnChange:=ImageListChange;
end;

destructor TBoundButtons.Destroy;
begin
  FreeAndNil(FImageChangeLink);
  inherited;
end;

procedure TBoundButtons.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FImages then
      Images:=nil;
  end;
end;

procedure TBoundButtons.SetCaptions(const Value: string);
begin
  if Value <> FCaptions then
  begin
    FCaptions:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetVisible(Value: boolean);
begin
  if Value <> FVisible then
  begin
    FVisible:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetFlat(Value: boolean);
begin
  if Value <> FFlat then
  begin
    FFlat:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetTransparent(Value: boolean);
begin
  if Value <> FTransparent then
  begin
    FTransparent:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetPosition(Value: TBtnPos);
begin
  if Value <> FPosition then
  begin
    FPosition:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetSize(Value: integer);
begin
  if Value <> FSize then
  begin
    FSize:=Value;
    TrDBImage(Owner).FBoundToolBar.ButtonHeight:=FSize;
    TrDBImage(Owner).FBoundToolBar.ButtonWidth:=FSize;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetShowGlyphs(Value: boolean);
begin
  if Value <> FShowGlyphs then
  begin
    FShowGlyphs:=Value;
    RefreshEvent(Self);
  end;
end;

procedure TBoundButtons.SetImages(Value: TImageList);
begin
  if FImages <> nil then
    FImages.UnRegisterChanges(FImageChangeLink);
  FImages:=Value;
  if FImages <> nil then
  begin
    FImages.RegisterChanges(FImageChangeLink);
    FImages.FreeNotification(Self);
  end;
  RefreshEvent(Self);
end;

procedure TBoundButtons.ImageListChange(Sender: TObject);
begin
  RefreshEvent(Self);
end;

// ==============================================================================
// TDBBoundEdit
constructor TBoundButtonedEdit.CreateLinkEvent(AOwner: TComponent; OnFindEvent: TNotifyEvent);
var
  B: TBitmap;
begin
  inherited Create(AOwner);
  Name:='SubEdit';
  SetSubComponent(true);
  Text:='';
  FScaleRatio:=1;

  FImages:=TImageList.Create(Self);
  FImages.ColorDepth:=cd32Bit;
  FImages.Name:='BtnImages';
  B:=rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('BTNSEEK'), FScaleRatio);
  FImages.AddMasked(B, B.Canvas.Pixels[0, 0]);
  B:=rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('BTNERASE'), FScaleRatio);
  FImages.AddMasked(B, B.Canvas.Pixels[0, 0]);

  FFindDelay:=250;
  FTimer:=TTimer.Create(Self);
  FTimer.Enabled:=false;
  FTimer.Interval:=FFindDelay;
  FTimer.OnTimer:=FindTimer;

  OnLeftButtonClick:=LeftBtnClick;
  OnRightButtonClick:=RightBtnClick;
  OnChange:=FindChange;
  OnKeyPress:=FindKeyPress;
  FOnFindEvent:=OnFindEvent;
end;

destructor TBoundButtonedEdit.Destroy;
begin
  FreeAndNil(FImages);
  FreeAndNil(FTimer);
  inherited Destroy;
end;

procedure TBoundButtonedEdit.Loaded;
begin
  inherited;
  if not Assigned(Images) then
  begin
    // use resource if user not use own images
    Images:=FImages;
    LeftButton.ImageIndex:=0;
    RightButton.ImageIndex:=1;
    LeftButton.Visible:=true;
    RightButton.Visible:=true;
  end;
end;

procedure TBoundButtonedEdit.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
end;

procedure TBoundButtonedEdit.RefreshPosition(W: TControl; HorizontalSpace: integer);
begin
  if Visible then
    SetBounds(W.Left, W.Top + W.Height + HorizontalSpace, W.Width, Height)
  else
    SetBounds(W.Left + 1, W.Top + W.Height - Height, 0, Height)
end;

procedure TBoundButtonedEdit.FindKeyPress(Sender: TObject; var Key: Char);
begin
  if ord(Key) = VK_RETURN then
  begin
    FindTimer(nil);
    Key:=#0;
  end;
  if ord(Key) = VK_ESCAPE then
  begin
    RightBtnClick(nil);
    Key:=#0;
  end;
end;

procedure TBoundButtonedEdit.FindChange(Sender: TObject);
begin
  if FFindDelay = 0 then
    FindTimer(nil)
  else
  begin
    FTimer.Enabled:=false;
    FTimer.Interval:=FFindDelay;
    FTimer.Enabled:=true;
  end;
end;

procedure TBoundButtonedEdit.LeftBtnClick(Sender: TObject);
begin
  FindTimer(nil);
end;

procedure TBoundButtonedEdit.RightBtnClick(Sender: TObject);
begin
  Text:='';
  if FFindDelay > 0 then
    FindTimer(nil);
end;

procedure TBoundButtonedEdit.FindTimer(Sender: TObject);
begin
  FTimer.Enabled:=false;
  if Assigned(FOnFindEvent) then
    FOnFindEvent(Self);
end;

// ==============================================================================
// other bounded components
constructor TBoundSpeedButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
end;

constructor TBoundListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
end;

procedure TBoundListBox.CNCommand(var Message: TWMCommand);
begin
  // do not call click when change selection
  case Message.NotifyCode of
    LBN_SELCHANGE:
      begin
        inherited Changed;
        // Click;
      end;
    LBN_DBLCLK:
      DblClick;
  end;
end;

procedure TBoundListBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
  Click;
end;

procedure TBoundListBox.KeyPress(var Key: Char);
begin
  inherited;
  if Key = #13 then
    Click;
  if Key = #27 then
  begin
    ItemIndex:=-1;
    Visible:=false;
  end;
end;

{ TEditButton.TGlyph }

constructor TrDBEditButton.TGlyph.Create(AButton: TrDBEditButton);
begin
  inherited Create(AButton.FEditControl);
  FButton:=AButton;
  FState:=bsNormal;
  Parent:=FButton.FEditControl;
  Visible:=false;
end;

procedure TrDBEditButton.TGlyph.Click;
var
  s: string;
begin
  s:=FButton.EditControl.Text;
  case FButton.ButtonClickAction of
    caClick:
      begin
        // Replicate from TControl to set Sender to owning TButtonedEdit control
        if Assigned(OnClick) and (Action <> nil) and not DelegatesEqual(@OnClick, @Action.OnExecute) then
          OnClick(FButton.EditControl)
        else
          if not(csDesigning in ComponentState) and (ActionLink <> nil) then
            ActionLink.Execute(FButton.EditControl)
          else
            if Assigned(OnClick) then
              OnClick(FButton.EditControl);
      end;
    caHTTP:
      if s <> '' then
        RunFileSimple(rRunHTTP + s);
    caMailTo:
      if s <> '' then
        RunFileSimple(rRunMailTo + s);
    caFile:
      if s <> '' then
        RunFileSimple(rRunFile + s);
  end;
end;

procedure TrDBEditButton.TGlyph.CreateWnd;
begin
  inherited;
  if Visible then
    FButton.FEditControl.UpdateEditMargins;
end;

procedure TrDBEditButton.TGlyph.Paint;
var
  LIndex: integer;
begin
  inherited;
  if (FButton.Images <> nil) and Visible then
  begin
    LIndex:=FButton.ImageIndex;
    if Enabled then
    begin
      case FState of
        bsHot:
          if FButton.HotImageIndex <> -1 then
            LIndex:=FButton.HotImageIndex;
        bsPushed:
          if FButton.PressedImageIndex <> -1 then
            LIndex:=FButton.PressedImageIndex;
      end;
    end
    else
      if FButton.DisabledImageIndex <> -1 then
        LIndex:=FButton.DisabledImageIndex;
    if LIndex <> -1 then
      FButton.Images.Draw(Canvas, 0, 0, LIndex);
  end;
end;

procedure TrDBEditButton.TGlyph.WndProc(var Message: TMessage);
var
  LPoint: TPoint;
begin
  if (Message.Msg = WM_CONTEXTMENU) and (FButton.EditControl.PopupMenu = nil) then
    Exit;

  inherited;
  case Message.Msg of
    CM_MOUSEENTER:
      FState:=bsHot;
    CM_MOUSELEAVE:
      FState:=bsNormal;
    WM_LBUTTONDOWN:
      if FButton.FDropDownMenu <> nil then
      begin
        if not(csDesigning in Parent.ComponentState) then
        begin
          LPoint:=ClientToScreen(Point(0, FButton.EditControl.Height));
          FButton.FDropDownMenu.Popup(LPoint.X, LPoint.Y);
        end;
      end
      else
        FState:=bsPushed;
    WM_LBUTTONUP:
      FState:=bsHot;
    CM_VISIBLECHANGED:
      FButton.UpdateBounds;
  else
    Exit;
  end;
  Invalidate;
end;

{ TrDBEditButton }

constructor TrDBEditButton.Create(EditControl: TrDBEdit; APosition: TButtonPosition);
begin
  inherited Create;
  FEditControl:=EditControl;
  FGlyph:=TGlyph.Create(Self);
  FHotImageIndex:=-1;
  FImageIndex:=-1;
  FPosition:=APosition;
  FPressedImageIndex:=-1;
  FDisabledImageIndex:=-1;
  FButtonClickAction:=caClick;
end;

destructor TrDBEditButton.Destroy;
begin
  FGlyph.Parent.RemoveControl(FGlyph);
  FreeAndNil(FGlyph);
  inherited;
end;

function TrDBEditButton.GetEnabled: boolean;
begin
  Result:=FGlyph.Enabled;
end;

function TrDBEditButton.GetCustomHint: TCustomHint;
begin
  Result:=FGlyph.CustomHint;
end;

function TrDBEditButton.GetHint: string;
begin
  Result:=FGlyph.Hint;
end;

function TrDBEditButton.GetImages: TImageList;
begin
  Result:=FEditControl.Images;
end;

function TrDBEditButton.GetOwner: TPersistent;
begin
  Result:=FEditControl;
end;

function TrDBEditButton.GetVisible: boolean;
begin
  Result:=FGlyph.Visible;
end;

procedure TrDBEditButton.SetDisabledImageIndex(const Value: TImageIndex);
begin
  if Value <> FDisabledImageIndex then
  begin
    FDisabledImageIndex:=Value;
    if not Enabled then
      FGlyph.Invalidate;
  end;
end;

procedure TrDBEditButton.SetEnabled(const Value: boolean);
begin
  if Value <> FGlyph.Enabled then
  begin
    FGlyph.Enabled:=Value;
    FGlyph.Invalidate;
  end;
end;

procedure TrDBEditButton.SetCustomHint(const Value: TCustomHint);
begin
  if Value <> FGlyph.CustomHint then
    FGlyph.CustomHint:=Value;
end;

procedure TrDBEditButton.SetHint(const Value: string);
begin
  if Value <> FGlyph.Hint then
    FGlyph.Hint:=Value;
end;

procedure TrDBEditButton.SetHotImageIndex(const Value: TImageIndex);
begin
  if Value <> FHotImageIndex then
  begin
    FHotImageIndex:=Value;
    if FGlyph.FState = bsHot then
      FGlyph.Invalidate;
  end;
end;

procedure TrDBEditButton.SetImageIndex(const Value: TImageIndex);
begin
  if Value <> FImageIndex then
  begin
    FImageIndex:=Value;
    if FGlyph.FState = bsNormal then
      FGlyph.Invalidate;
  end;
end;

procedure TrDBEditButton.SetPressedImageIndex(const Value: TImageIndex);
begin
  if Value <> FPressedImageIndex then
  begin
    FPressedImageIndex:=Value;
    if FGlyph.FState = bsPushed then
      FGlyph.Invalidate;
  end;
end;

procedure TrDBEditButton.SetVisible(const Value: boolean);
begin
  if Value <> FGlyph.Visible then
  begin
    FGlyph.Visible:=Value;
    FEditControl.UpdateEditMargins;
  end;
end;

procedure TrDBEditButton.UpdateBounds;
var
  EdgeSize, NewLeft: integer;
begin
  if FGlyph <> nil then
  begin
    if Images <> nil then
    begin
      FGlyph.Width:=Images.Width;
      FGlyph.Height:=Images.Height;
    end
    else
    begin
      FGlyph.Width:=0;
      FGlyph.Height:=0;
    end;
    FGlyph.Top:=0;
    NewLeft:=FGlyph.Left;

    if not GetThemeEnabled then
      FGlyph.Top:=1;

    case FPosition of
      bpLeft:
        begin
          if GetThemeEnabled then
            NewLeft:=0
          else
            NewLeft:=1;
        end;
      bpRight:
        begin
          NewLeft:=FEditControl.Width - FGlyph.Width;
          if FEditControl.BorderStyle <> bsNone then
            Dec(NewLeft, 4);
          if FEditControl.BevelKind <> bkNone then
          begin
            EdgeSize:=0;
            if FEditControl.BevelInner <> bvNone then
              Inc(EdgeSize, FEditControl.BevelWidth);
            if FEditControl.BevelOuter <> bvNone then
              Inc(EdgeSize, FEditControl.BevelWidth);
            if beRight in FEditControl.BevelEdges then
              Dec(NewLeft, EdgeSize);
            if beLeft in FEditControl.BevelEdges then
              Dec(NewLeft, EdgeSize);
          end;
          if not GetThemeEnabled then
            Dec(NewLeft);
        end;
    end;

    if (not FEditControl.Ctl3D) and (FEditControl.BorderStyle <> bsNone) then
    begin
      FGlyph.Top:=2;
      Inc(NewLeft, 2);
    end;

    FGlyph.Left:=NewLeft;

    if (csDesigning in FEditControl.ComponentState) and not Visible then
      FGlyph.Width:=0;

    if Assigned(FEditControl) and Assigned(Images) then
      FGlyph.Top:=(FEditControl.ClientHeight - Images.Height) div 2;
  end;
end;

// ==============================================================================
// TrDBEdit
constructor TrDBEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FDBEditLabel.StoreVisible;
  // buttons
  FCanvas:=TControlCanvas.Create;
  FCanvas.Control:=Self;
  FImageChangeLink:=TChangeLink.Create;
  FImageChangeLink.OnChange:=ImageListChange;
  FLeftButton:=TrDBEditButton.Create(Self, bpLeft);
  FRightButton:=TrDBEditButton.Create(Self, bpRight);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBEdit.Destroy;
begin
  FreeAndNil(FDataLink);
  FreeAndNil(FCanvas);
  FreeAndNil(FImageChangeLink);
  FreeAndNil(FLeftButton);
  FreeAndNil(FRightButton);
  inherited Destroy;
end;

procedure TrDBEdit.CMHintShow(var Message: TCMHintShow);
var
  hnt: string;
  ACanvas: TControlCanvas;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    ACanvas:=TControlCanvas.Create;
    try
      ACanvas.Control:=Self;
      ACanvas.Font.Assign(Font);
      if ACanvas.TextWidth(hnt) > Self.ClientWidth - 4 then
        Message.HintInfo^.HintStr:=hnt;
    finally
      FreeAndNil(ACanvas);
    end;
  end;
  inherited;
end;

procedure TrDBEdit.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=ADataSource;
  UpdateState(nil);
end;

procedure TrDBEdit.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrDBEdit.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBEdit.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBEdit.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBEdit.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
end;

procedure TrDBEdit.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
end;

procedure TrDBEdit.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FImages then
    begin
      FImages:=nil;
      FLeftButton.UpdateBounds;
      FRightButton.UpdateBounds;
      UpdateEditMargins;
    end
    else
      if (LeftButton <> nil) and (AComponent = LeftButton.DropDownMenu) then
        LeftButton.DropDownMenu:=nil
      else
        if (RightButton <> nil) and (AComponent = RightButton.DropDownMenu) then
          RightButton.DropDownMenu:=nil;
  end;
end;

procedure TrDBEdit.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
end;

procedure TrDBEdit.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
end;

procedure TrDBEdit.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if FDataLink.Editing and not ro then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  // BUG Delete if Autoedit=false
  if (Key = VK_DELETE) and not FDataLink.Edit then
    Key:=0;
  inherited KeyDown(Key, Shift);
end;

procedure TrDBEdit.DoExit;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

function TrDBEdit.GetOnLeftButtonClick: TNotifyEvent;
begin
  Result:=LeftButton.Glyph.OnClick;
end;

function TrDBEdit.GetOnRightButtonClick: TNotifyEvent;
begin
  Result:=RightButton.Glyph.OnClick;
end;

procedure TrDBEdit.ImageListChange(Sender: TObject);
begin
  if HandleAllocated then
  begin
    FLeftButton.UpdateBounds;
    FRightButton.UpdateBounds;
    UpdateEditMargins;
  end;
end;

procedure TrDBEdit.DefaultHandler(var Message);
begin
  inherited;
  case TMessage(Message).Msg of
    CN_CTLCOLOREDIT:
      begin
        FLeftButton.Glyph.Invalidate;
        FRightButton.Glyph.Invalidate;
      end;
    WM_SIZE:
      FRightButton.UpdateBounds;
  end;
end;

procedure TrDBEdit.SetImages(const Value: TImageList);
begin
  if Value <> FImages then
  begin
    if FImages <> nil then
      FImages.UnRegisterChanges(FImageChangeLink);
    FImages:=Value;
    if FImages <> nil then
    begin
      FImages.RegisterChanges(FImageChangeLink);
      FImages.FreeNotification(Self);
    end;
    FLeftButton.UpdateBounds;
    FRightButton.UpdateBounds;
    UpdateEditMargins;
  end;
end;

procedure TrDBEdit.SetLeftButton(const Value: TrDBEditButton);
begin
  FLeftButton.Assign(Value);
end;

procedure TrDBEdit.SetRightButton(const Value: TrDBEditButton);
begin
  FRightButton.Assign(Value);
end;

procedure TrDBEdit.SetOnLeftButtonClick(const Value: TNotifyEvent);
begin
  LeftButton.Glyph.OnClick:=Value;
end;

procedure TrDBEdit.SetOnRightButtonClick(const Value: TNotifyEvent);
begin
  RightButton.Glyph.OnClick:=Value;
end;

procedure TrDBEdit.WMPaint(var Message: TWMPaint);
var
  AAlignment: TAlignment;
begin
  // check if Right button with right alignment is used, if so use own drawn, else inherited
  if Assigned(Field) and Assigned(RightButton) and RightButton.Visible then
  begin
    AAlignment:=Field.Alignment;
    if UseRightToLeftAlignment then
      ChangeBiDiModeAlignment(AAlignment);
    if (AAlignment <> taLeftJustify) and not Focused then
    begin
      DoOwnPaint(Message.DC, AAlignment);
      Exit;
    end;
  end;
  inherited;
end;

procedure TrDBEdit.DoOwnPaint(MsgDC: HDC; Align: TAlignment);
{$IFDEF DX100P}
const
  StyleColor: array [boolean] of TStyleColor = (scEditDisabled, scEdit);
  StyleFontColor: array [boolean] of TStyleFont = (sfEditBoxTextDisabled, sfEditBoxTextNormal);
{$ENDIF}
var
  Left: integer;
  Margins: TRect;
  R: TRect;
  DC: HDC;
  PS: TPaintStruct;
  s: string;
begin
  { draw with correct right margin for button - copied from DBEdit but fix GetTextMargins }
  if FCanvas = nil then
  begin
    FCanvas:=TControlCanvas.Create;
    FCanvas.Control:=Self;
  end;
  DC:=MsgDC;
  if DC = 0 then
    DC:=BeginPaint(Handle, PS);
  FCanvas.Handle:=DC;
  try
    FCanvas.Font:=Font;
    with FCanvas do
    begin
      R:=ClientRect;
      if not(NewStyleControls and Ctl3D) and (BorderStyle = bsSingle) then
      begin
{$IFDEF DX100P}
        Brush.Color:=GetThemeServices.GetSystemColor(clWindowFrame);
{$ELSE}
        Brush.Color:=clWindowFrame;
{$ENDIF}
        BiDi_FrameRect(FCanvas, R);
        InflateRect(R, -1, -1);
      end;
{$IFDEF DX100P}
      if TStyleManager.IsCustomStyleActive and (seClient in StyleElements) then
      begin
        Brush.Color:=StyleServices.GetStyleColor(StyleColor[Enabled]);
        Font.Color:=StyleServices.GetStyleFontColor(StyleFontColor[Enabled]);
      end
      else
{$ENDIF}
      begin
        Brush.Color:=Color;
        if not Enabled then
          Font.Color:=clGrayText;
      end;
      if (csPaintCopy in ControlState) and (Field <> nil) then
      begin
        s:=Field.DisplayText;
        case CharCase of
          ecUpperCase:
            s:=AnsiUpperCase(s);
          ecLowerCase:
            s:=AnsiLowerCase(s);
        end;
      end
      else
        s:=EditText;
      if PasswordChar <> #0 then
        s:=StringOfChar(PasswordChar, Length(s));
      Margins:=GetTextMargins;
      case Align of
        taLeftJustify:
          Left:=Margins.Left;
        taRightJustify:
          Left:=ClientWidth - TextWidth(s) - Margins.Right - 1;
      else
        Left:=(ClientWidth - TextWidth(s)) div 2;
      end;
      if Left < Margins.Left then
        Left:=Margins.Left;
      if SysLocale.MiddleEast then
        UpdateTextFlags;
      TextRect(R, Left, Margins.Top, s);
    end;
  finally
    FCanvas.Handle:=0;
    if MsgDC = 0 then
      EndPaint(Handle, PS);
  end;
end;

function TrDBEdit.GetTextMargins: TRect;
var
  DC: HDC;
  SaveFont: HFont;
  I, X: integer;
  SysMetrics, Metrics: TTextMetric;
begin
  if NewStyleControls then
  begin
    if BorderStyle = bsNone then
      I:=0
    else
      if Ctl3D then
        I:=1
      else
        I:=2;
    X:=SendMessage(Handle, EM_GETMARGINS, 0, 0);
    Result.Left:=X and $0000FFFF + I;
    Result.Right:=X div $FFFF + I;
    Result.Top:=I;
  end
  else
  begin
    if BorderStyle = bsNone then
      I:=0
    else
    begin
      DC:=GetDC(0);
      GetTextMetrics(DC, SysMetrics);
      SaveFont:=SelectObject(DC, Font.Handle);
      GetTextMetrics(DC, Metrics);
      SelectObject(DC, SaveFont);
      ReleaseDC(0, DC);
      I:=SysMetrics.tmHeight;
      if I > Metrics.tmHeight then
        I:=Metrics.tmHeight;
      I:=I div 4;
    end;
    Result.Left:=I;
    Result.Right:=I;
    Result.Top:=I;
  end;
end;

procedure TrDBEdit.UpdateEditMargins;
var
  LMargin, RMargin: integer;
begin
  if HandleAllocated then
  begin
    LMargin:=0;
    RMargin:=0;
    if (Images <> nil) then
    begin
      if LeftButton.Visible then
        LMargin:=Images.Width + 2;
      if RightButton.Visible then
        RMargin:=Images.Width + 2;
    end;
    SendMessage(Handle, EM_SETMARGINS, EC_LEFTMARGIN or EC_RIGHTMARGIN, MakeLong(LMargin, RMargin));
    // AdjustTextHint(LMargin, TextHint);
    Invalidate;
  end;
end;

procedure TrDBEdit.WndProc(var Message: TMessage);
var
  LLeft, LTop: integer;
begin
  case Message.Msg of
    CN_CTLCOLORSTATIC, CN_CTLCOLOREDIT:
      if FImages <> nil then
      begin
        if LeftButton.Visible then
        begin
          LLeft:=LeftButton.Glyph.Left;
          LTop:=LeftButton.Glyph.Top;
          if GetThemeEnabled and Ctl3D then
          begin
            Inc(LLeft);
            Inc(LTop);
          end;
          ExcludeClipRect(Handle, LLeft + 1, LTop + 1, LeftButton.Glyph.Width + LeftButton.Glyph.Left, LeftButton.Glyph.Height);
        end;

        if RightButton.Visible then
        begin
          LTop:=RightButton.Glyph.Top;
          if GetThemeEnabled and Ctl3D then
            Inc(LTop);
          ExcludeClipRect(Handle, RightButton.Glyph.Left, LTop + 1, RightButton.Glyph.Width + RightButton.Glyph.Left, RightButton.Glyph.Height);
        end;
      end;
  end;

  inherited;

  case Message.Msg of
    CM_BORDERCHANGED, CM_CTL3DCHANGED:
      begin
        if not(csLoading in ComponentState) then
        begin
          LeftButton.UpdateBounds;
          RightButton.UpdateBounds;
        end;
      end;
    CM_FONTCHANGED:
      if not(csLoading in ComponentState) then
        UpdateEditMargins;
  end;
end;

procedure TrDBEdit.ValidateEdit;
begin
  inherited;
  if Field is TNumericField then
    Text:=rtool.CleanNumericString(Text)
  else
    if Field is TDateTimeField then
      Text:=rtool.CleanDateTimeString(Text)
    else
      Text:=rtool.CleanLastEOL(Text);
end;

// ==============================================================================
// TrDBEditDate
constructor TrDBEditDate.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width:=84;
  FUpdateAfterChange:=false;
  if not(csDesigning in ComponentState) then
  begin
    // create PopupCalendar - use SELF as Owner
    FPopupCalendar:=TrPopupMonthCalendar.CreateEx(Self, PopupCallbackSelectDate);
  end;
end;

procedure TrDBEditDate.Loaded;
begin
  inherited Loaded;
  // use internal ButtonClick if not used by custom code
  if not(csDesigning in ComponentState) and not Assigned(OnRightButtonClick) then
    OnRightButtonClick:=FRightButtonClick;
end;

procedure TrDBEditDate.UpdateState(Sender: TObject);
begin
  inherited;
  if not FDataLink.Editing and Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
end;

procedure TrDBEditDate.CMCancelMode(var Message: TCMCancelMode);
begin
  // inherited;
  // hide cal after click to window (exclude only self, popup and RightButton)
  if not Assigned(FPopupCalendar) then
    Exit;
  if not FPopupCalendar.IsShown or (Message.Sender = Self) or (Message.Sender = FPopupCalendar) or Message.Sender.Equals(THackEditButton(RightButton).Glyph)
    then
    Exit;
  FPopupCalendar.HideCalendar;
  SetFocus;
end;

procedure TrDBEditDate.WMKillFocus(var Message: TMessage);
begin
  if not Assigned(FPopupCalendar) then
    Exit;
  FPopupCalendar.HideCalendar;
  inherited;
end;

procedure TrDBEditDate.PopupCallbackSelectDate(Sender: TObject);
begin
  if not ReadOnly then
    SetDate(Trunc(FPopupCalendar.Date));
end;

procedure TrDBEditDate.FRightButtonClick(Sender: TObject);
var
  D: TDateTime;
begin
  if FPopupCalendar.IsShown then
    FPopupCalendar.HideCalendar
  else
  begin
    // ensure to have focus for direct click to button
    if not Focused then
      SetFocus;
    if Text = '' then
      D:=Today
    else
      D:=GetDate;
    FPopupCalendar.ShowCalendar(D);
  end;
end;

procedure TrDBEditDate.DoEnter;
begin
  inherited;
  // store original value to restore by ESC
  OrigText:=Text;
end;

procedure TrDBEditDate.CMExit(var Message: TCMExit);
begin
  if Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
  if not CheckDate(false) then
    SetFocus
  else
    inherited;
end;

function TrDBEditDate.DoDropDown: boolean;
begin
  Result:=false;
  if RightButton.Visible and Assigned(OnRightButtonClick) then
  begin
    OnRightButtonClick(Self);
    Result:=true;
  end;
end;

procedure TrDBEditDate.SetDate(DateValue: TDateTime);
begin
  if (csLoading in ComponentState) then
    Exit;
  if ReadOnly or (Assigned(Field) and Field.ReadOnly) or FDataLink.ReadOnly or not FDataLink.Edit then
    Exit;
  if DateValue = 0 then
    Text:=''
  else
    Text:=DateToStr(DateValue);
  CheckDate(true);
  SelectAll;
end;

function TrDBEditDate.GetDate: TDateTime;
begin
  Result:=0;
  if CheckDate(false) and (Text <> '') then
    Result:=StrToDateDef(ClearStr(' ', Text, true), Date);
end;

function TrDBEditDate.CheckDate(UpdateDataLink: boolean): boolean;
var
  aDate: TDateTime;
begin
  Result:=true;
  if (csLoading in ComponentState) or (csDesigning in ComponentState) then
    Exit;
  if Text = '' then
  begin
    if UpdateDataLink then
      UpdateDatalinkValue;
    Exit;
  end
  else
    Result:=StrToDateEx(Text, aDate, FMinDate, FMaxDate);
  if Result then
  begin
    Text:=DateToStr(aDate);
    if UpdateDataLink then
      UpdateDatalinkValue;
  end;
end;

procedure TrDBEditDate.UpdateDatalinkValue;
begin
  if FUpdateAfterChange then
  begin
    PostMessage(Handle, CM_EXIT, 0, 0);
    PostMessage(Handle, CM_ENTER, 0, 0);
  end;
end;

procedure TrDBEditDate.KeyDown(var Key: Word; Shift: TShiftState);
var
  dif: integer;
begin
  if Assigned(FPopupCalendar) and FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYDOWN, Key, 0);
    Key:=0;
  end;
  if not ReadOnly then
  begin
    case Key of
      VK_UP, VK_DOWN:
        begin
          if Key = VK_UP then
            dif:=1
          else
            dif:=-1;
          if Shift = [] then
            SetDate(GetDate + dif);
          if Shift = [ssCtrl] then
            SetDate(IncMonth(GetDate, dif));
          if Shift = [ssCtrl, ssShift] then
            SetDate(IncYear(GetDate, dif));
          if (Shift = [ssAlt]) and (Key = VK_DOWN) then
            DoDropDown;
          Key:=0;
        end;
      VK_F4:
        begin
          DoDropDown;
          Key:=0;
        end;
    end;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrDBEditDate.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if Assigned(FPopupCalendar) and FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYUP, Key, 0);
    Key:=0;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrDBEditDate.KeyPress(var Key: Char);
var
  Valid, Sep: TSysCharSet;
begin
  if not ReadOnly then
  begin
    case Key of
      'T', 't', '*':
        begin
          SetDate(Date);
          Key:=#0;
        end;
      'S', 's':
        begin
          SetDate(DateUtils.StartOfTheMonth(Date));
          Key:=#0;
        end;
      'E', 'e':
        begin
          SetDate(DateUtils.EndOfTheMonth(Date));
          Key:=#0;
        end;
      'P', 'p', 'D', 'd':
        begin
          DoDropDown;
          Key:=#0;
        end;
      chr(VK_ESCAPE):
        begin
          if FPopupCalendar.IsShown then
            FPopupCalendar.HideCalendar
          else
          begin
            // restore orig value
            Text:=OrigText;
            SelectAll;
          end;
          Key:=#0;
        end;
      chr(VK_RETURN):
        begin
          if FPopupCalendar.IsShown then
          begin
            PopupCallbackSelectDate(nil);
            FPopupCalendar.HideCalendar;
          end;
          CheckDate(true);
          SelectAll;
          Key:=#0;
        end;
    end;
  end;
  Sep:=['.', ',', '/', '-'];
  Valid:=['0' .. '9'] + Sep;
  if (not CharInSet(Key, Valid)) and (Key >= #32) then
    Key:=#0;
  if CharInSet(Key, Sep) then
    Key:=GetDateSeparator;
  if Key <> #0 then
    inherited KeyPress(Key);
end;

// ==============================================================================
// TrDBMemo
constructor TrDBMemo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  FReadOnlyColor:=clBtnFace;
  ScrollBars:=ssVertical;
  FOptionsEx:=[moAllowSetFont, moAllowLoadSaveFile];
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FPopupMenu:=TPopupMenu.Create(Self);
  FPopupMenu.Name:='pmSearch';
  DefineEditPopupMenuItems(FPopupMenu, PopupMenuClick);
  PopupMenu:=FPopupMenu;
  FPopupMenu.OnPopup:=PopupMenuOpen;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBMemo.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBMemo.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=DataSource;
  UpdateState(nil);
end;

procedure TrDBMemo.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBMemo.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
end;

function TrDBMemo.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBMemo.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBMemo.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBMemo.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
end;

procedure TrDBMemo.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
end;

procedure TrDBMemo.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
  end;
end;

procedure TrDBMemo.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
end;

procedure TrDBMemo.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
end;

procedure TrDBMemo.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if FDataLink.Editing and not ro then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBMemo.KeyDown(var Key: Word; Shift: TShiftState);
begin
  // BUG Delete if Autoedit=false
  if (Key = VK_DELETE) and not FDataLink.Edit then
    Key:=0;
  inherited KeyDown(Key, Shift);
end;

function TrDBMemo.SearchText(const Search: string = ''; FromStart: boolean = true; IgnoreCase: boolean = true): boolean;
var
  sText: string;
  Index: integer;
begin
  Result:=false;
  FSearchText:=Search;
  if FSearchText = '' then
    if not rdlg.InputStr(sInput, SFind, FSearchText) then
      Exit;
  FSearchIgnoreCase:=IgnoreCase;
  sText:=Lines.Text;
  if FSearchIgnoreCase then
  begin
    FSearchText:=UpperCase(FSearchText);
    sText:=UpperCase(Lines.Text);
  end;
  if FromStart then
    Index:=1
  else
    Index:=SelStart + SelLength + 1;
  Index:=PosEx(FSearchText, sText, Index);
  Result:=Index > 0;
  if Result then
  begin
    SelStart:=Index - 1;
    SelLength:=Length(FSearchText);
    try
      SetFocus;
    except
    end;
  end;
  if not Result then
    DlgI(sNotFound);
end;

function TrDBMemo.SearchNext;
begin
  Result:=SearchText(FSearchText, false, FSearchIgnoreCase);
end;

procedure TrDBMemo.PopupMenuOpen(Sender: TObject);
var
  I: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  for I:=0 to PopupMenu.Items.Count - 1 do
  begin
    case PopupMenu.Items[I].Tag of
      rcEditUndo:
        PopupMenu.Items[I].Enabled:=CanUndo;
      rcEditCut:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditCopy:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditPaste:
        PopupMenu.Items[I].Enabled:=Clipbrd.Clipboard.AsText <> '';
      rcEditDelete:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditSelectAll:
        ;
      rcEditFind:
        ;
      rcEditFindNext:
        PopupMenu.Items[I].Enabled:=FSearchText <> '';
      rcEditSelectFont:
        PopupMenu.Items[I].Enabled:=moAllowSetFont in FOptionsEx;
      rcMemoWordWrap:
        begin
          PopupMenu.Items[I].Enabled:=moAllowWordWrap in FOptionsEx;
          PopupMenu.Items[I].Checked:=WordWrap;
        end;
      rcMemoLoadFile:
        PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
      rcMemoSaveFile:
        PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
    end;
  end;
end;

procedure TrDBMemo.PopupMenuClick(Sender: TObject);
var
  SaveDlg: TrSaveDialog;
  OpenDlg: TrOpenDialog;
begin
  if Sender is TMenuItem then
  begin
    case TMenuItem(Sender).Tag of
      rcEditUndo:
        Undo;
      rcEditCut:
        CutToClipboard;
      rcEditCopy:
        CopyToClipboard;
      rcEditPaste:
        PasteFromClipboard;
      rcEditDelete:
        ClearSelection;
      rcEditSelectAll:
        SelectAll;
      rcEditFind:
        SearchText;
      rcEditFindNext:
        SearchNext;
      rcEditSelectFont:
        rDialogsEx.ShowSelectFontDialog(Font);
      rcMemoWordWrap:
        begin
          WordWrap:=not WordWrap;
          if WordWrap and (ScrollBars = ssBoth) then
            ScrollBars:=ssVertical;
          if not WordWrap and (ScrollBars = ssVertical) then
            ScrollBars:=ssBoth;
        end;
      rcMemoLoadFile:
        begin
          OpenDlg:=TrOpenDialog.Create(Self);
          try
            if OpenDlg.Execute then
              Lines.LoadFromFile(OpenDlg.FileName);
          finally
            OpenDlg.Free;
          end;
        end;
      rcMemoSaveFile:
        begin
          SaveDlg:=TrSaveDialog.Create(Self);
          try
            if SaveDlg.Execute then
              Lines.SaveToFile(SaveDlg.FileName);
          finally
            SaveDlg.Free;
          end;
        end;
    end;
  end;
end;

procedure TrDBMemo.DoExit;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

// ==============================================================================
// TrDBMemoEx
constructor TrDBMemoEx.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  CloseOnCancel:=false;
  FHighlightWords:=TStringList.Create;
  FHighlightWords.OnChange:=HighlightWordsChanged;
  // FHighlightFromStr:='\';
  // FHighlightToStr:=' ';
  FHighlightedFontColor:=Font.Color;
  FHighlightedBrushColor:=clMoneyGreen;

  FHighlightFromToEx:=TStringList.Create;
  FHighlightFromToEx.OnChange:=HighlightWordsChanged;

  FButtonFrm:=TBoundSpeedButton.Create(Self);
  FButtonFrm.FreeNotification(Self);
  FButtonFrm.Name:='SubBtnFrm';
  FButtonFrm.OnClick:=ButtonClick;
  FButtonFrm.Height:=15;
  FButtonFrm.Width:=18;
{$IFDEF DoNotUseGlyphInDesign}
  if (csDesigning in ComponentState) then
    FButtonFrm.Caption:='f'
  else
{$ENDIF}
  begin
    FButtonFrm.Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('FORMAT'), FScaleRatio));
    FButtonFrm.Hint:=sFormatInfo;
  end;
  FShowButtonFrm:=false;

  FButtonEdit:=TBoundSpeedButton.Create(Self);
  FButtonEdit.FreeNotification(Self);
  FButtonEdit.Name:='SubBtnEdit';
  FButtonEdit.OnClick:=ButtonClick;
  FButtonEdit.Height:=15;
  FButtonEdit.Width:=18;
{$IFDEF DoNotUseGlyphInDesign}
  if (csDesigning in ComponentState) then
    FButtonEdit.Caption:='e'
  else
{$ENDIF}
  begin
    FButtonEdit.Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('EDITWIN'), FScaleRatio));
    FButtonEdit.Hint:=sEditInfo;
  end;
  FShowButtonEdit:=true;

  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  // FBoundLabel.Alignment:=taRightJustify;
  FBoundLabel.Position:=lpTopRight;
  FBoundLabel.Margins.Left:=0;
  FBoundLabel.Margins.Right:=40;
  FBoundLabel.Margins.Top:=0;
  FBoundLabel.Margins.Bottom:=0;
  FBoundLabel.AlignWithMargins:=true;

  FStoredLines:=TStringList.Create;
  FFormatPopupMenu:=nil;
  FFormatType:=ftRTF;
end;

destructor TrDBMemoEx.Destroy;
begin
  FreeAndNil(FHighlightWords);
  FreeAndNil(FHighlightFromToEx);
  FreeAndNil(FStoredLines);
  inherited;
end;

procedure TrDBMemoEx.Loaded;
begin
  inherited;
{$IFDEF DoNotUseGlyphInDesign}
  FButtonFrm.Caption:='';
  FButtonEdit.Caption:='';
{$ENDIF}
  {
    // temporary clear saved glyph = only for existing project in design time only
    FButtonFrm.Glyph:=nil;
    FButtonEdit.Glyph:=nil;
    }
  CreateFormatPopupMenu;
end;

procedure TrDBMemoEx.CreateFormatPopupMenu;
begin
  if not Assigned(FFormatPopupMenu) then
  begin
    // create Popup
    FFormatPopupMenu:=TPopupMenu.Create(Self);
    FFormatPopupMenu.Name:=Name + '_Popup';
    FFormatPopupMenu.AutoHotkeys:=maManual;
  end;
  // popup items - info, bold, italic, underline
  FFormatPopupMenu.Items.Clear;
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatInfo, 0, PopupClick));
  FFormatPopupMenu.Items[0].Enabled:=false;
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, '-', 0, PopupClick));
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatBold, cFormatBold, PopupClick));
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatItalic, cFormatItalic, PopupClick));
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatUnderline, cFormatUnderline, PopupClick));
  if FFormatType = ftHTML then
  begin
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatAsUnnumberedList, cFormatAsUnnumberedList, PopupClick));
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatAsNumberedList, cFormatAsNumberedList, PopupClick));
  end;
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, '-', 0, PopupClick));
  if FFormatType = ftRTF then
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatNewpage, cFormatRTFNewPage, PopupClick));
  if FFormatType = ftHTML then
  begin
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatNewLine2, cFormatHTMLNewLine, PopupClick));
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatNewLineAuto, cFormatHTMLNewLineAuto, PopupClick));
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatHorLine, cFormatHTMLHorLine, PopupClick));
    FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatHTMLLink, cFormatHTMLLink, PopupClick));
  end;
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, '-', 0, PopupClick));
  FFormatPopupMenu.Items.Add(CreateMenuItem(Self, sFormatClear, cFormatClear, PopupClick));
end;

procedure TrDBMemoEx.LoadSBPos;
begin
  if HandleAllocated then
  begin
    FRowIncrement:=-GetScrollPos(Handle, SB_VERT);
    FColIncrement:=-GetScrollPos(Handle, SB_HORZ);
  end;
end;

procedure TrDBMemoEx.SetHighlightWords(Value: TStringList);
begin
  FHighlightWords.Assign(Value);
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightWholeWords(Value: boolean);
begin
  FHighlightWholeWords:=Value;
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightFromStr(const Value: string);
begin
  FHighlightFromStr:=Value;
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightToStr(const Value: string);
begin
  FHighlightToStr:=Value;
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightedFontColor(Value: TColor);
begin
  FHighlightedFontColor:=Value;
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightedBrushColor(Value: TColor);
begin
  FHighlightedBrushColor:=Value;
  Refresh;
end;

procedure TrDBMemoEx.SetHighlightFromToEx(Value: TStringList);
begin
  FHighlightFromToEx.Assign(Value);
  Refresh;
end;

procedure TrDBMemoEx.HighlightWordsChanged(Sender: TObject);
begin
  Refresh;
end;

procedure TrDBMemoEx.WMPaint(var Message: TWMPaint);
const
  SepChars: TSysCharSet = [' ', '_', '-', '(', ')', '/', '\', '=', ',', ';', '.', #9 .. #13];
  XorColor = $00FFD8CE;
var
  DC: HDC;
  Canvas: TCanvas;
  iline, I, FirstLine, LastLine, FirstChar, RowHeight: integer;
  CompStr, TextStr: string;
  ContinueFormat: boolean;

  procedure DrawSel(StartPos: integer; Str: string);
  var
    seppos: integer;
    P: Longint;
    PT: TPoint;
    R: TRect;
  begin
    // draw selected
    P:=SendMessage(Handle, EM_POSFROMCHAR, FirstChar + StartPos - 1, 0);
    if P < 0 then
      Exit;
    PT.X:=LoWord(P);
    PT.Y:=HiWord(P);
    if (PT.X >= 0) and (PT.Y >= 0) and (PT.Y < ClientHeight - RowHeight) then
    begin
      seppos:=Pos(#13, Str);
      if seppos > 0 then
        Str:=copy(Str, 1, seppos - 1);
      R.TopLeft:=PT;
      R.Right:=R.Left + Canvas.TextWidth(Str);
      R.Bottom:=R.Top + Canvas.TextHeight(Str);
      Canvas.TextRect(R, PT.X, PT.Y, Str);
    end;
  end;

  procedure HighlightWord(WordStr: string);
  var
    p1: integer;
    sel: string;
  begin
    WordStr:=AnsiUpperCase(WordStr);
    p1:=0;
    repeat
      // find word
      p1:=PosEx(WordStr, CompStr, p1 + 1);
      if p1 = 0 then
        Break;
      if FHighlightWholeWords and not ContainWholeWord(WordStr, TextStr, SepChars, p1) then
        Continue;
      sel:=copy(TextStr, p1, Length(WordStr));
      DrawSel(p1, sel);
    until false;
  end;

  function HighlightFromTo(FromStr, ToStr: string; ContFrm: boolean): boolean;
  var
    p1, p2: integer;
    sel: string;
  begin
    Result:=false;
    FromStr:=AnsiUpperCase(FromStr);
    ToStr:=AnsiUpperCase(ToStr);
    // proc from to selection
    p1:=0;
    repeat
      // find start pos
      if ContFrm then
        p1:=1
      else
        p1:=PosEx(FromStr, CompStr, p1 + 1);
      ContFrm:=false;
      if p1 = 0 then
        Break;
      // find stop pos
      p2:=PosEx(ToStr, CompStr, p1 + Length(FromStr));
      if p2 = 0 then
      begin
        p2:=Length(TextStr);
        Result:=true;
      end;
      sel:=copy(TextStr, p1, p2 - p1 + Length(ToStr));
      DrawSel(p1, sel);
      p1:=p2 + Length(ToStr);
    until false;
  end;

  procedure HighlightFromToEx(Line: string);
  var
    s1, s2, s3: string;
  begin
    s1:=RemoveStrPart(Line, #9, true);
    if s1 = '' then
      Exit;
    if s1 = '#9' then
      s1:=#9;
    s2:=RemoveStrPart(Line, #9, true);
    if s2 = '' then
      Exit;
    if s2 = '#9' then
      s2:=#9;

    s3:=RemoveStrPart(Line, #9, true);
    if s3 = '' then
      Canvas.Font.Color:=FHighlightedFontColor
    else
      if s3 = '-' then
        Canvas.Font.Color:=Font.Color
      else
        Canvas.Font.Color:=rpictool.StringToColor(s3);

    if Line = '' then
      Canvas.Brush.Color:=FHighlightedBrushColor
    else
      if Line = '-' then
        Canvas.Brush.Color:=clNone
      else
        Canvas.Brush.Color:=rpictool.StringToColor(Line);
    if Canvas.Brush.Color = clNone then
      Canvas.Brush.Style:=bsClear
    else
      Canvas.Brush.Style:=bsSolid;

    ContinueFormat:=HighlightFromTo(s1, s2, ContinueFormat);
  end;

  procedure HighlightWordEx(Line: string);
  var
    s1, s2: string;
  begin
    s1:=RemoveStrPart(Line, #9, true);
    if s1 = '' then
      Exit;

    s2:=RemoveStrPart(Line, #9, true);
    if s2 = '' then
      Canvas.Font.Color:=FHighlightedFontColor
    else
      if s2 = '-' then
        Canvas.Font.Color:=Font.Color
      else
        Canvas.Font.Color:=rpictool.StringToColor(s2);

    if Line = '' then
      Canvas.Brush.Color:=FHighlightedBrushColor
    else
      if Line = '-' then
        Canvas.Brush.Color:=clNone
      else
        Canvas.Brush.Color:=rpictool.StringToColor(Line);
    if Canvas.Brush.Color = clNone then
      Canvas.Brush.Style:=bsClear
    else
      Canvas.Brush.Style:=bsSolid;

    HighlightWord(s1);
  end;

  procedure ShowPageBreak(X: integer);
  begin
    Canvas.Pen.Color:=clBtnShadow;
    Canvas.MoveTo(X, 2);
    Canvas.LineTo(X, ClientHeight);
    Canvas.Pen.Color:=clBtnHighlight;
    Canvas.MoveTo(X + 1, 2);
    Canvas.LineTo(X + 1, ClientHeight);
  end;

  procedure ProcPageBreaks;
  var
    ipb, pbpos, W: integer;
    Str: string;
  begin
    if FPageBreaksCharNumber = '' then
      Exit;
    Canvas.Pen.Style:=psSolid;
    Canvas.Pen.Mode:=pmCopy;
    Canvas.Brush.Style:=bsSolid;
    ipb:=0;
    repeat
      Inc(ipb);
      Str:=rtool.FindStrPart(FPageBreaksCharNumber, ';', ipb, true);
      if Str = '' then
        Break;
      pbpos:=StrToIntDef(Str, 0);
      if pbpos > 0 then
      begin
        Str:=StringOfChar('o', pbpos);
        W:=Canvas.TextWidth(Str);
        if W < ClientWidth then
          ShowPageBreak(W);
      end;
    until false;
  end;

  procedure InitFrm;
  begin
    ContinueFormat:=false;
    Canvas.Font.Color:=FHighlightedFontColor;
    Canvas.Brush.Color:=FHighlightedBrushColor;
  end;

  procedure InitLine;
  begin
    FirstChar:=Perform(EM_LINEINDEX, iline, 0);
    TextStr:=Lines[iline];
    CompStr:=AnsiUpperCase(TextStr);
  end;

begin
  inherited;
  LoadSBPos;
  // Get a DC for the canvas
  DC:=GetDC(Handle);
  try
    Canvas:=TCanvas.Create;
    try
      Canvas.Handle:=DC;
      Canvas.Font:=Font;
      RowHeight:=Canvas.TextHeight('Wg');

      // find and proc all visible lines
      FirstLine:=Perform(EM_GETFIRSTVISIBLELINE, 0, 0);
      LastLine:=Perform(EM_CHARFROMPOS, 0, (ClientHeight - 1) shl 16);
      LastLine:=(LastLine and $FFFF0000) shr 16;

      // old simple FromTo format
      if (FHighlightFromStr <> '') and (FHighlightToStr <> '') then
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          ContinueFormat:=HighlightFromTo(FHighlightFromStr, FHighlightToStr, ContinueFormat);
        end;
      end;

      // new FromToEx format
      for I:=0 to FHighlightFromToEx.Count - 1 do
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          HighlightFromToEx(FHighlightFromToEx[I]);
        end;
      end;

      // new Words format
      for I:=0 to FHighlightWords.Count - 1 do
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          HighlightWordEx(FHighlightWords[I]);
        end;
      end;

      ProcPageBreaks;

      Canvas.Handle:=0;
    finally
      FreeAndNil(Canvas);
    end;
  finally
    // Release the DC when finished
    ReleaseDC(Handle, DC);
  end;
end;

procedure TrDBMemoEx.Change;
begin
  inherited Change;
  Refresh;
end;

procedure TrDBMemoEx.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
  Refresh;
end;

procedure TrDBMemoEx.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FButtonFrm) then
    FButtonFrm.Enabled:=Enabled;
  if Assigned(FButtonEdit) then
    FButtonEdit.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBMemoEx.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FButtonFrm) then
    FButtonFrm.Visible:=Visible and FShowButtonFrm;
  if Assigned(FButtonEdit) then
    FButtonEdit.Visible:=Visible and FShowButtonEdit;
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible; // and FShowButtonEdit;
end;

procedure TrDBMemoEx.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FButtonFrm then
      FButtonFrm:=nil;
    if AComponent = FButtonEdit then
      FButtonEdit:=nil;
    if AComponent = FBoundLabel then
      FBoundLabel:=nil;
  end;
end;

procedure TrDBMemoEx.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  RefreshBtPos;
end;

procedure TrDBMemoEx.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FButtonFrm) then
    FButtonFrm.Parent:=AParent;
  if Assigned(FButtonEdit) then
    FButtonEdit.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrDBMemoEx.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if Assigned(FButtonFrm) then
    FButtonFrm.Enabled:=Enabled and not ro;
  if Assigned(FButtonEdit) then
    FButtonEdit.Enabled:=Enabled;
  // if Assigned(FBoundLabel) then FBoundLabel.Enabled:=Enabled;
  RefreshBtState;
end;

function TrDBMemoEx.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

procedure TrDBMemoEx.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
end;

procedure TrDBMemoEx.SetShowButtonFrm(Value: boolean);
begin
  FShowButtonFrm:=Value;
  RefreshBtPos;
end;

procedure TrDBMemoEx.SetShowButtonEdit(Value: boolean);
begin
  FShowButtonEdit:=Value;
  RefreshBtPos;
end;

procedure TrDBMemoEx.SetStoredLines(Value: TStrings);
begin
  FStoredLines.Assign(Value);
end;

procedure TrDBMemoEx.RefreshBtPos;
var
  L: integer;
begin
  if Assigned(FButtonFrm) and Assigned(FButtonEdit) then
  begin
    FButtonEdit.Visible:=Visible and FShowButtonEdit;
    FButtonFrm.Visible:=Visible and FShowButtonFrm;
    L:=Left + Width - FButtonFrm.Width;
    if FShowButtonEdit then
      L:=L - FButtonEdit.Width - 2;
    FButtonFrm.Left:=L;
    FButtonFrm.Top:=Top - FButtonFrm.Height;
    L:=Left + Width - FButtonEdit.Width;
    FButtonEdit.Left:=L;
    FButtonEdit.Top:=Top - FButtonEdit.Height;
    if not FButtonFrm.Visible then
      FButtonFrm.Top:=Top;
    if not FButtonEdit.Visible then
      FButtonEdit.Top:=Top;
  end;
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
end;

procedure TrDBMemoEx.RefreshBtState;
begin
  if Assigned(FButtonFrm) and Assigned(FButtonEdit) then
  begin
    //
  end;
end;

procedure TrDBMemoEx.ButtonClick(Sender: TObject);
var
  bt: TBoundSpeedButton;
  P: TPoint;
begin
  bt:=Sender as TBoundSpeedButton;
  if bt = FButtonFrm then
  begin
    P:=bt.Parent.ClientToScreen(bt.BoundsRect.TopLeft);
    Self.FFormatPopupMenu.Popup(P.X, P.Y + bt.Height);
  end;
  if bt = FButtonEdit then
    ShowEditWin;
end;

procedure TrDBMemoEx.PopupClick(Sender: TObject);
begin
  FormatInt(TMenuItem(Sender).Tag);
end;

procedure TrDBMemoEx.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if CloseOnCancel and (Key = VK_ESCAPE) and (Owner is TForm) then
    try
      TForm(Owner).Close;
    except
    end;

  if ButtonEdit.Visible then
    if Key = VK_F2 then
      FButtonEdit.Click;
  if ButtonFrm.Visible then
  begin
    if (Shift = [ssCtrl]) and CharInSet(chr(Key), ['b', 'B']) then
    begin
      FormatBold;
      Key:=0;
    end;
    if (Shift = [ssCtrl]) and CharInSet(chr(Key), ['i', 'I']) then
    begin
      FormatItalic;
      Key:=0;
    end;
    if (Shift = [ssCtrl]) and CharInSet(chr(Key), ['u', 'U']) then
    begin
      FormatUnderline;
      Key:=0;
    end;
  end;
  inherited;
end;

procedure TrDBMemoEx.FormatInt(Tag: integer);
const
  RTFfrm1: array [1 .. 3] of string = ('\b ', '\i ', '\ul ');
  RTFfrm2: array [1 .. 3] of string = ('\b0 ', '\i0 ', '\ul0 ');
  HTMLfrm1: array [1 .. 3] of string = ('<B>', '<I>', '<U>');
  HTMLfrm2: array [1 .. 3] of string = ('</B>', '</I>', '</U>');
  RTFfrmPage = '\page ';
  HTMLfrmNewLine = '<br>';
  HTMLfrmHorLine = '<hr>';
  HTMLfrmLink = '<A href=...>text</A>';
  HTMLfrmList1: array [1 .. 3] of string = ('<UL>', '<OL>', '<LI>');
  HTMLfrmList2: array [1 .. 3] of string = ('</UL>', '</OL>', '</LI>');

  procedure AddText(txt: string);
  var
    P: integer;
  begin
    P:=SelStart;
    Text:=copy(Text, 1, P) + txt + copy(Text, P + 1, Length(Text));
    SelStart:=P + Length(txt);
  end;

var
  i1, i2, P: integer;
  s, f1, f2: string;
  L: TStrings;
begin
  if not FDataLink.Edit then
    Exit;
  L:=TStringList.Create;
  try
    // ShowTip(sTipFormat,'TipFormat',Data.PSFile);
    case Tag of
      cFormatBold, cFormatItalic, cFormatUnderline:
        begin
          i1:=SelStart;
          i2:=SelLength;
          if FFormatType = ftRTF then
          begin
            f1:=RTFfrm1[Tag];
            f2:=RTFfrm2[Tag];
          end
          else
          begin
            f1:=HTMLfrm1[Tag];
            f2:=HTMLfrm2[Tag];
          end;
          if copy(SelText, i2, 1) = ' ' then
            i2:=i2 - 1;
          Text:=copy(Text, 1, i1) + f1 + copy(Text, i1 + 1, i2) + f2 + copy(Text, i1 + i2 + 1, Length(Text));
          SelStart:=i1 + i2 + Length(f1) + Length(f2);
        end;
      cFormatAsUnnumberedList, cFormatAsNumberedList:
        begin
          i1:=SelStart;
          i2:=SelLength;
          s:=copy(Text, i1 + 1, i2);
          if Pos(#13, s) = 0 then
            s:=copy(Text, 1, i1) + HTMLfrmList1[3] + copy(Text, i1 + 1, i2) + HTMLfrmList2[3] + copy(Text, i1 + i2 + 1, Length(Text))
          else
          begin
            L.Text:=s;
            for P:=0 to L.Count - 1 do
              L[P]:=HTMLfrmList1[3] + L[P] + HTMLfrmList2[3];
            L.Insert(0, HTMLfrmList1[Tag - cFormatAsUnnumberedList + 1]);
            L.Text:=L.Text + HTMLfrmList2[Tag - cFormatAsUnnumberedList + 1];
            s:=copy(Text, 1, i1) + L.Text + copy(Text, i1 + i2 + 1, Length(Text));
          end;
          Text:=s;
        end;
      cFormatRTFNewPage:
        AddText(RTFfrmPage);
      cFormatHTMLNewLine:
        AddText(HTMLfrmNewLine);
      cFormatHTMLNewLineAuto:
        begin
          L.Text:=Text;
          for P:=0 to L.Count - 1 do
          begin
            s:=Trim(L[P]);
            s:=copy(s, Length(s) - Length(HTMLfrmNewLine) + 1, Length(HTMLfrmNewLine));
            if not SameText(HTMLfrmNewLine, s) then
              L[P]:=L[P] + HTMLfrmNewLine;
          end;
          Text:=L.Text;
        end;
      cFormatHTMLHorLine:
        AddText(HTMLfrmHorLine);
      cFormatHTMLLink:
        AddText(HTMLfrmLink);
      cFormatClear:
        begin
          s:=Text;
          if FFormatType = ftRTF then
          begin
            for P:=1 to 3 do
              s:=ClearStr(RTFfrm1[P], s, false);
            for P:=1 to 3 do
              s:=ClearStr(RTFfrm2[P], s, false);
            s:=ClearStr(RTFfrmPage, s, false);
          end
          else
          begin
            for P:=1 to 3 do
            begin
              s:=ClearStr(HTMLfrm1[P], s, false);
              s:=ClearStr(HTMLfrm2[P], s, false);
            end;
            for P:=1 to 3 do
            begin
              s:=ClearStr(HTMLfrmList1[P], s, false);
              s:=ClearStr(HTMLfrmList2[P], s, false);
            end;
            s:=ClearStr(HTMLfrmNewLine, s, false);
            s:=ClearStr(HTMLfrmHorLine, s, false);
            s:=ClearStr(HTMLfrmLink, s, false);
          end;
          Text:=s;
        end;
    end;
  finally
    FreeAndNil(L);
  end;
  SetFocus;
  SelLength:=1;
  SelLength:=0;
  Invalidate;
end;

procedure TrDBMemoEx.FormatBold;
begin
  FormatInt(cFormatBold);
end;

procedure TrDBMemoEx.FormatItalic;
begin
  FormatInt(cFormatItalic);
end;

procedure TrDBMemoEx.FormatUnderline;
begin
  FormatInt(cFormatUnderline);
end;

procedure TrDBMemoEx.FormatHTMLUnnumberedList;
begin
  FormatInt(cFormatAsUnnumberedList);
end;

procedure TrDBMemoEx.FormatHTMLNumberedList;
begin
  FormatInt(cFormatAsNumberedList);
end;

procedure TrDBMemoEx.FormatNewPage;
begin
  FormatInt(cFormatRTFNewPage);
end;

procedure TrDBMemoEx.FormatHTMLNewLine;
begin
  FormatInt(cFormatHTMLNewLine);
end;

procedure TrDBMemoEx.FormatHTMLNewLinesAuto;
begin
  FormatInt(cFormatHTMLNewLineAuto);
end;

procedure TrDBMemoEx.FormatHTMLHorLine;
begin
  FormatInt(cFormatHTMLHorLine);
end;

procedure TrDBMemoEx.FormatHTMLLink;
begin
  FormatInt(cFormatHTMLLink);
end;

procedure TrDBMemoEx.FormatClear;
begin
  FormatInt(cFormatClear);
end;

procedure TrDBMemoEx.ShowEditWin;
var
  F: TForm;
  M: TrDBMemoEx;
  P: TPanel;
  B: TButton;
begin
  // update field value
  Perform(CM_EXIT, 0, 0);
  Application.ProcessMessages;
  // create new form
  F:=TForm.Create(nil);
  try
    F.BorderIcons:=[biSystemMenu, biMaximize];
    F.Position:=poScreenCenter;
    F.Width:=Screen.MonitorFromWindow(Self.Handle).Width - 80;
    F.Height:=Screen.MonitorFromWindow(Self.Handle).Height - 160;
    F.Caption:=Self.DBEditLabel.Caption;

    M:=TrDBMemoEx.Create(F);
    M.Parent:=F;
    M.AlignWithMargins:=true;
    M.Margins.Top:=22;
    M.Align:=alClient;
    M.ScrollBars:=ssVertical;
    M.Text:=Text;
    M.HighlightFromStr:=HighlightFromStr;
    M.HighlightToStr:=HighlightToStr;
    M.HighlightWholeWords:=HighlightWholeWords;
    M.HighlightWords.Assign(HighlightWords);
    M.HighlightedFontColor:=HighlightedFontColor;
    M.HighlightedBrushColor:=HighlightedBrushColor;
    M.ShowButtonEdit:=false;
    M.ShowButtonFrm:=ShowButtonFrm;
    M.ReadOnly:=ReadOnly;
    M.ReadOnlyColor:=ReadOnlyColor;
    M.SetDataSource(DataSource, DataField);
    M.ButtonFrm.Width:=30;
    M.ButtonFrm.Height:=18;
    M.Font.Assign(Font);
    M.CloseOnCancel:=true;
    M.PageBreaksCharNumber:=PageBreaksCharNumber;
    M.FormatType:=FormatType;
    M.WantReturns:=WantReturns;
    M.WantTabs:=WantTabs;
    M.CreateFormatPopupMenu;

    P:=TPanel.Create(F);
    P.Parent:=F;
    P.Align:=alBottom;
    P.Height:=32;
    B:=TButton.Create(F);
    B.Parent:=P;
    B.AlignWithMargins:=true;
    B.Align:=alRight;
    B.Caption:=sOk;
    B.Default:=true;
    B.Cancel:=true;
    B.ModalResult:=mrOk;

    F.ShowModal;
  finally
    FreeAndNil(F);
  end;
  try
    SetFocus;
  except
  end;
end;

// ==============================================================================
// TrDBMemoCB
constructor TrDBMemoCB.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButtonInsert:=TBoundSpeedButton.Create(Self);
  FButtonInsert.FreeNotification(Self);
  FButtonInsert.Name:='SubBtnInsert';
  FButtonInsert.OnClick:=ButtonClick;
  FButtonInsert.GroupIndex:=1111;
  FButtonInsert.AllowAllUp:=true;
  FButtonInsert.Height:=18;
  FButtonInsert.Width:=18;
{$IFDEF DoNotUseGlyphInDesign}
  if (csDesigning in ComponentState) then
    FButtonInsert.Caption:='i'
  else
{$ENDIF}
    FButtonInsert.Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('CBINS'), FScaleRatio));

  FButtonAdd:=TBoundSpeedButton.Create(Self);
  FButtonAdd.FreeNotification(Self);
  FButtonAdd.Name:='SubBtnAdd';
  FButtonAdd.OnClick:=ButtonClick;
  FButtonAdd.GroupIndex:=1111;
  FButtonAdd.AllowAllUp:=true;
  FButtonAdd.Height:=18;
  FButtonAdd.Width:=18;
{$IFDEF DoNotUseGlyphInDesign}
  if (csDesigning in ComponentState) then
    FButtonAdd.Caption:='a'
  else
{$ENDIF}
    FButtonAdd.Glyph.Assign(rGlyphs.BmpFromIndex(rGlyphs.ButtonIndex('CBADD'), FScaleRatio));

  FListBox:=TBoundListBox.Create(Self);
  FListBox.FreeNotification(Self);
  FListBox.Name:='SubListBox';
  FListBox.OnClick:=ListBoxClick;
  FListBox.OnExit:=ListBoxClick;

  FShowButtonInsert:=true;
  FShowButtonAdd:=true;
  FInserting:=false;
end;

destructor TrDBMemoCB.Destroy;
begin
  inherited Destroy;
end;

procedure TrDBMemoCB.Loaded;
begin
  inherited Loaded;
  FListBox.Visible:=false;
{$IFDEF DoNotUseGlyphInDesign}
  FButtonInsert.Caption:='';
  FButtonAdd.Caption:='';
{$ENDIF}
end;

procedure TrDBMemoCB.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FButtonInsert) then
    FButtonInsert.Enabled:=Enabled;
  if Assigned(FButtonAdd) then
    FButtonAdd.Enabled:=Enabled;
end;

procedure TrDBMemoCB.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FButtonInsert) then
    FButtonInsert.Visible:=Visible and FShowButtonInsert;
  if Assigned(FButtonAdd) then
    FButtonAdd.Visible:=Visible and FShowButtonAdd;
end;

procedure TrDBMemoCB.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FButtonInsert then
      FButtonInsert:=nil;
    if AComponent = FButtonAdd then
      FButtonAdd:=nil;
    if AComponent = FListBox then
      FListBox:=nil;
  end;
end;

procedure TrDBMemoCB.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  RefreshBtPos;
end;

procedure TrDBMemoCB.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FButtonInsert) then
    FButtonInsert.Parent:=AParent;
  if Assigned(FButtonAdd) then
    FButtonAdd.Parent:=AParent;
  if Assigned(FListBox) then
    FListBox.Parent:=AParent;
end;

procedure TrDBMemoCB.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if Assigned(FButtonInsert) then
    FButtonInsert.Enabled:=Enabled and not ro;
  if Assigned(FButtonAdd) then
    FButtonAdd.Enabled:=Enabled and not ro;
  RefreshBtState;
end;

procedure TrDBMemoCB.SetShowButtonInsert(Value: boolean);
begin
  FShowButtonInsert:=Value;
  RefreshBtPos;
end;

procedure TrDBMemoCB.SetShowButtonAdd(Value: boolean);
begin
  FShowButtonAdd:=Value;
  RefreshBtPos;
end;

procedure TrDBMemoCB.RefreshBtPos;
var
  L: integer;
begin
  if Assigned(FButtonInsert) and Assigned(FButtonAdd) then
  begin
    FButtonAdd.Visible:=Visible and FShowButtonAdd;
    FButtonInsert.Visible:=Visible and FShowButtonInsert;
    L:=Left + Width - FButtonInsert.Width;
    if FShowButtonAdd then
      L:=L - FButtonAdd.Width - 4;
    FButtonInsert.Left:=L;
    FButtonInsert.Top:=Top - FButtonInsert.Height;
    L:=Left + Width - FButtonAdd.Width;
    FButtonAdd.Left:=L;
    FButtonAdd.Top:=Top - FButtonAdd.Height;
    if not FButtonInsert.Visible then
      FButtonInsert.Top:=Top;
    if not FButtonAdd.Visible then
      FButtonAdd.Top:=Top;
  end;
  if Assigned(FListBox) then
    FListBox.BoundsRect:=Rect(Left, Top, 8, 8);
end;

procedure TrDBMemoCB.RefreshBtState;
begin
  if Assigned(FButtonInsert) and Assigned(FButtonAdd) and Assigned(FListBox) then
  begin
    if not FDataLink.Editing then
      FListBox.Visible:=false;
    FButtonInsert.Down:=FListBox.Visible and FInserting;
    FButtonAdd.Down:=FListBox.Visible and not FInserting;
  end;
end;

procedure TrDBMemoCB.ButtonClick(Sender: TObject);
var
  bt: TBoundSpeedButton;
begin
  if not Assigned(FListBox) then
    Exit;
  bt:=Sender as TBoundSpeedButton;
  if FListBox.Visible and (FInserting <> (bt = FButtonInsert)) then
    FListBox.Visible:=false;
  FInserting:=bt = FButtonInsert;
  FListBox.BoundsRect:=Rect(Left + 16, Top - 1, Left + Width, Top + Height + 18);
  FListBox.ItemIndex:=-1;
  if Assigned(FShowListBox) then
    FShowListBox(Self);
  if FDataLink.Editing or FDataLink.Edit then
  begin
    FListBox.Visible:=not FListBox.Visible;
    if FListBox.Visible then
      FListBox.SetFocus;
  end;
  RefreshBtState;
end;

procedure TrDBMemoCB.ListBoxClick(Sender: TObject);
var
  s: string;
begin
  if FListBox.ItemIndex <> -1 then
  begin
    s:=FListBox.Items[FListBox.ItemIndex];
    FListBox.ItemIndex:=-1;
    if FInserting then
      SelText:=s
    else
      if not FNoDuplicLine or (Lines.IndexOf(s) < 0) then
        Lines.Add(s);
    Field.AsString:=Text;
  end;
  FListBox.Visible:=false;
  SetFocus;
  RefreshBtState;
end;

procedure TrDBMemoCB.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited;
  if Key = VK_F3 then
    FButtonInsert.Click;
  if Key = VK_F4 then
    FButtonAdd.Click;
end;

// ==============================================================================
// TrDBListBox
constructor TrDBListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  FBoundLabel.Position:=lpTopRight;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBListBox.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBListBox.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=ADataSource;
  UpdateState(nil);
end;

procedure TrDBListBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrDBListBox.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBListBox.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBListBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBListBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBListBox.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrDBListBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FBoundLabel then
      FBoundLabel:=nil;
  end;
end;

procedure TrDBListBox.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
end;

procedure TrDBListBox.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrDBListBox.UpdateState(Sender: TObject);
var
  ro: boolean;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if FDataLink.Editing and not ro then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

function TrDBListBox.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
var
  Key: Word;
begin
  Result:=true;
  while ProcMouseWheel(WheelDelta, Key) do
    SendMessage(Handle, WM_KEYDOWN, Key, 0);
end;

procedure TrDBListBox.DoExit;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

// ==============================================================================
// TrDBComboBox
constructor TrDBComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  FBoundLabel.Position:=lpTopRight;
  DropDownCount:=15;
  FDropDownAutoWidth:=true;
  FDropDownMaxWidth:=600;
  FItemHeightEx:=cDBComboBoxItemHeightDefault;
  FUpdateAfterClick:=false;
  WndRecreated:=false;
  FDisplayedItems:=TDisplayedItems.Create(Self, ListDefined);
  IntTimer:=TTimer.Create(Self);
  FUpdateTimerDelay:=0;
  FLastKeyDown:=0;
  FJustClosed:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBComboBox.Destroy;
begin
  FreeAndNil(FDataLink);
  FreeAndNil(FDisplayedItems);
  inherited Destroy;
end;

procedure TrDBComboBox.CreateParams(var Params: TCreateParams);
begin
  inherited;
  if UseOwnDraw then
    Params.Style:=Params.Style or CBS_OWNERDRAWVARIABLE; // CBS_OWNERDRAWFIXED;
end;

procedure TrDBComboBox.EditWndProc(var Message: TMessage);
begin
  if Message.Msg = WM_MOUSEMOVE then
    ComboWndProc(Message, FEditHandle, FDefEditProc)
  else
    inherited;
end;

procedure TrDBComboBox.CMHintShow(var Message: TCMHintShow);
var
  hnt: string;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    if Canvas.TextWidth(hnt) > Self.ClientWidth - 24 then
      Message.HintInfo^.HintStr:=hnt;
  end;
  inherited;
end;

procedure TrDBComboBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.StoreVisible;
  if UseOwnDraw then
  begin
    WndRecreated:=true;
    RecreateWnd;
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBComboBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * FItemHeightEx);
end;

procedure TrDBComboBox.SetItemHeightEx(Value: integer);
begin
  if Value <> FItemHeightEx then
  begin
    FItemHeightEx:=Value;
    ItemHeight:=Trunc(FScaleRatio * FItemHeightEx);
  end;
end;

procedure TrDBComboBox.ListDefined(Sender: TObject);
begin
  if UseOwnDraw and not WndRecreated then
  begin
    WndRecreated:=true;
    RecreateWnd;
  end;
end;

function TrDBComboBox.UseOwnDraw: boolean;
begin
  Result:=Assigned(FGetListItemProps) or (FDisplayedItems.ListItems.Count > 0);
end;

procedure TrDBComboBox.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=ADataSource;
  UpdateState(nil);
end;

function TrDBComboBox.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBComboBox.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBComboBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBComboBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBComboBox.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrDBComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FBoundLabel then
      FBoundLabel:=nil;
  end;
end;

procedure TrDBComboBox.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
  if HandleAllocated and not Focused then
  begin
    try
      // standard DBComboBox draw content selected after resize, so reset it
      SendMessage(Handle, CB_SETEDITSEL, 0, 0);
    except
    end;
  end;
end;

procedure TrDBComboBox.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrDBComboBox.Change;
begin
  if FDataLink.Edit then
    inherited Change;
  if FLastKeyDown <> VK_RETURN then
    ResetTimer(true, FJustClosed);
  FJustClosed:=false;
end;

procedure TrDBComboBox.CloseUp;
begin
  inherited CloseUp;
  FJustClosed:=true;
end;

procedure TrDBComboBox.Click;
begin
  if FDataLink.Edit then
    inherited Click;
  if FUpdateAfterClick then
    PostMessage(Handle, CM_EXIT, 0, 0);
end;

procedure TrDBComboBox.CMExit(var Message: TCMExit);
begin
  // fixed bug (if combobox is opened during closing window)
  if IsEditing(DataSource) then
    inherited
  else
    try
      inherited;
    except
    end;
end;

procedure TrDBComboBox.UpdateState(Sender: TObject);
var
  ro: boolean;
  F: TField;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if FDataLink.Editing and not ro then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);

  if not(csDesigning in ComponentState) and Assigned(FDataLink.DataSet) then
  begin
    F:=FDataLink.DataSet.FindField(DataField);
    if Assigned(F) and (F.DataType in [ftString, ftWideString]) and (MaxLength = 0) then
      MaxLength:=F.Size;
  end;
end;

function TrDBComboBox.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
var
  Key: Word;
begin
  Result:=true;
  while ProcMouseWheel(WheelDelta, Key) do
    SendMessage(Handle, WM_KEYDOWN, Key, 0);
end;

procedure TrDBComboBox.KeyPress(var Key: Char);
var
  ro: boolean;
begin
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if Key <> #27 then
    if ro or not(FDataLink.Editing or FDataLink.Edit) then
      Key:=#0;
  // BUG in standard DBCombo - do not call Change on first key during AutoComplete
  if Key <> #0 then
    Change;
  inherited KeyPress(Key);
end;

procedure TrDBComboBox.KeyDown(var Key: Word; Shift: TShiftState);
var
  ro: boolean;
begin
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if Key in [VK_BACK, VK_DELETE, VK_UP, VK_DOWN, 32 .. 255] then
    if ro or not(FDataLink.Editing or FDataLink.Edit) then
      Key:=0;
  if (FNullValueKey <> 0) and not ro and (FNullValueKey = ShortCut(Key, Shift)) then
  begin
    if FDataLink.Edit then
      Field.Clear;
    Key:=0;
  end
  else
    inherited KeyDown(Key, Shift);
  FLastKeyDown:=Key;
end;

procedure TrDBComboBox.WndProc(var Message: TMessage);
var
  CBRect, LBRect: TRect;
begin
  if Message.Msg = WM_CTLCOLORLISTBOX then
  begin
    GetWindowRect(Handle, CBRect);
    GetWindowRect(Message.LParam, LBRect);
    if LBRect.Right > Screen.DesktopRect.Right then
      MoveWindow(Message.LParam, LBRect.Left - (LBRect.Right - CBRect.Right), LBRect.Top, LBRect.Right - LBRect.Left, LBRect.Bottom - LBRect.Top, true);
  end;
  inherited WndProc(Message);
end;

procedure TrDBComboBox.DropDown;
var
  W: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if FDropDownAutoWidth then
  begin
    W:=Max(Width, Min(FDropDownMaxWidth, FDisplayedItems.GetMaxItemWidth(Self.DropDownCount)));
    SendMessage(Handle, CB_SETDROPPEDWIDTH, W, 0);
  end;
  inherited DropDown;
end;

procedure TrDBComboBox.DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState);
begin
  if UseOwnDraw and not(odComboBoxEdit in State) then
    FDisplayedItems.DrawListItem(Index, State, Rect, FGetListItemProps)
  else
    inherited;
end;

procedure TrDBComboBox.ResetTimer(SetActive, CallEventNow: boolean);
begin
  if Assigned(IntTimer) then
  begin
    IntTimer.Enabled:=false;
    IntTimer.Interval:=FUpdateTimerDelay;
    IntTimer.OnTimer:=IntTimerEvent;
    IntTimer.Enabled:=SetActive;
    if CallEventNow then
      IntTimerEvent(nil);
  end;
end;

procedure TrDBComboBox.IntTimerEvent(Sender: TObject);
var
  s, n: string;
begin
  IntTimer.Enabled:=false;
  if Assigned(FUpdateTimerEvent) then
    FUpdateTimerEvent(Self);

  if Assigned(FOnUpdateLoadItems) then
  begin
    if SelStart > 0 then
      s:=copy(Text, 1, SelStart)
    else
    begin
      s:=Text;
    end;
    n:='';
    Self.Items.BeginUpdate;
    try
      FOnUpdateLoadItems(Self, s, n);
    finally
      Self.Items.EndUpdate;
    end;
    if (n <> '') then
      Text:=n;
    SelStart:=Length(s);
    SelLength:=Length(Text);
    ResetTimer(false, false);
  end;
end;

procedure TrDBComboBox.DoExit;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

// ==============================================================================
// TrDBLookupListBox
constructor TrDBLookupListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  FBoundLabel.Position:=lpTopRight;
  // Self.NullValueKey:=VK_DELETE;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBLookupListBox.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBLookupListBox.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=ADataSource;
  UpdateState(nil);
end;

procedure TrDBLookupListBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrDBLookupListBox.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBLookupListBox.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
  begin
    FDBEditLabel.SetField(DataSource, DataField);
    FDBEditLabel.SetField(ListSource, ListField);
  end;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLookupListBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLookupListBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBLookupListBox.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrDBLookupListBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FBoundLabel then
      FBoundLabel:=nil;
  end;
end;

procedure TrDBLookupListBox.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
end;

procedure TrDBLookupListBox.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrDBLookupListBox.UpdateState(Sender: TObject);
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  if FDataLink.Editing and CanModify then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBLookupListBox.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  if Assigned(FOnMouseClick) then
    FOnMouseClick(Self);
end;

function TrDBLookupListBox.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
var
  Key: Word;
begin
  Result:=true;
  while ProcMouseWheel(WheelDelta, Key) do
    SendMessage(Handle, WM_KEYDOWN, Key, 0);
end;

procedure TrDBLookupListBox.DoExit;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

// ==============================================================================
// TrDBLookupComboBox
function VarEquals(const V1, V2: Variant): boolean;
begin
  Result:=false;
  try
    Result:=V1 = V2;
  except
  end;
end;

function TrDBLookupControlHelper.GetLookupMode: boolean;
begin
{$IFDEF DX101P}
  Result:=TRttiContext.Create.GetType(Self.ClassInfo).GetField('FLookupMode').GetValue(Self).AsBoolean;
{$ELSE}
  Result:=Self.FLookupMode;
{$ENDIF}
end;

function TrDBLookupControlHelper.GetKeyField: TField;
begin
{$IFDEF DX101P}
  Result:=TField(TRttiContext.Create.GetType(Self.ClassInfo).GetField('FKeyField').GetValue(Self).AsObject);
{$ELSE}
  Result:=Self.FKeyField;
{$ENDIF}
end;

function TrDBLookupListBoxHelper.GetRowCount: integer;
begin
{$IFDEF DX101P}
  Result:=TRttiContext.Create.GetType(Self.ClassInfo).GetField('FRowCount').GetValue(Self).AsInteger;
{$ELSE}
  Result:=Self.FRowCount;
{$ENDIF}
end;

function TrDBLookupListBoxHelper.GetKeySelected: boolean;
begin
{$IFDEF DX101P}
  Result:=TRttiContext.Create.GetType(Self.ClassInfo).GetField('FKeySelected').GetValue(Self).AsBoolean;
{$ELSE}
  Result:=Self.FKeySelected;
{$ENDIF}
end;

function TrDBLookupListBoxHelper.GetRecordCount: integer;
begin
{$IFDEF DX101P}
  Result:=TRttiContext.Create.GetType(Self.ClassInfo).GetField('FRecordCount').GetValue(Self).AsInteger;
{$ELSE}
  Result:=Self.FRecordCount;
{$ENDIF}
end;

function TrDBLookupListBoxHelper.GetRecordIndex: integer;
begin
{$IFDEF DX101P}
  Result:=TRttiContext.Create.GetType(Self.ClassInfo).GetField('FRecordIndex').GetValue(Self).AsInteger;
{$ELSE}
  Result:=Self.FRecordIndex;
{$ENDIF}
end;

function TrDBLookupComboBoxHelper.GetPopupList;
begin
{$IFDEF DX101P}
  Result:=TPopupDataList(TRttiContext.Create.GetType(Self.ClassInfo).GetField('FDataList').GetValue(Self).AsObject);
{$ELSE}
  Result:=Self.FDataList;
{$ENDIF}
end;

procedure TrDBLookupComboBoxHelper.SetPopupList(L: TPopupDataList);
begin
{$IFDEF DX101P}
  TRttiContext.Create.GetType(Self.ClassInfo).GetField('FDataList').SetValue(Self, L);
{$ELSE}
  Self.FDataList:=L;
{$ENDIF}
end;

constructor TrPopupDataList.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItemHeight:=0;
  FItemIncreaseHeight:=2;
  FItemIncreaseWidth:=0;
  SetLength(FColWidth, 0);
end;

function TrPopupDataList.GetTextHeight: integer;
begin
  if FItemHeight > 0 then
    Result:=FItemHeight
  else
    Result:= inherited GetTextHeight + FItemIncreaseHeight;
end;

procedure TrPopupDataList.Paint;
var
  LBorder, RowIdx, ColIdx, W, X, Y, TextWidth, TextHeight, RowHeight, LastFieldIndex: integer;
  ItemText: string;
  R: TRect;
  Selected, SepTop, SepBottom: boolean;
  AListField: TField;
  AAlignment: TAlignment;
  RectBmp, ItemBmp: TBitmap;

  procedure InitCanvas(C: TCanvas);
  begin
    // init Canvas for the row
    C.Font:=Font;
    if Enabled then
      C.Font.Color:=Font.Color
    else
      C.Font.Color:=clGrayText;
    C.Brush.Color:=Color;
    if Selected then
    begin
      C.Font.Color:=clHighlightText;
      C.Brush.Color:=clHighlight;
    end;
    { if ColorToRGB(Color) <> ColorToRGB(clBtnFace) then
      C.Pen.Color:=clBtnFace
      else
      C.Pen.Color:=clBtnShadow;
      }
    C.Pen.Color:=clGray;
    ItemBmp.Width:=0;
    SepTop:=false;
    SepBottom:=false;
  end;

begin
  if Parent is TDBLookupControl then
    LBorder:=4
  else
    LBorder:=2;

  // draw to RectBmp to avoid flickering
  RectBmp:=TBitmap.Create;
  try
    ItemBmp:=TBitmap.Create;
    try
      // get text size
      Canvas.Font:=Font;
      TextWidth:=Canvas.TextWidth('0');
      TextHeight:=Canvas.TextHeight('0');
      RowHeight:=GetTextHeight;
      LastFieldIndex:=ListFields.Count - 1;
      // for each item row
      for RowIdx:=0 to GetRowCount - 1 do
      begin
        Selected:=false;
        // Selected:=not GetKeySelected and (RowIdx = 0);
        InitCanvas(Canvas);
        // init Rect
        R.Top:=RowIdx * RowHeight;
        R.Bottom:=R.Top + RowHeight;
        if RowIdx < GetRecordCount then
        begin
          ListLink.ActiveRecord:=RowIdx;
          if not VarIsNull(KeyValue) and VarEquals(GetKeyField.Value, KeyValue) then
            Selected:=true;
          R.Right:=0;
          // for each List Field
          for ColIdx:=0 to LastFieldIndex do
          begin
            AListField:=ListFields[ColIdx];
            ItemText:=AListField.DisplayText;
            // find column width - client for last column, calculated width for AutoWidth else Field display width
            if ColIdx = LastFieldIndex then
              W:=ClientWidth - R.Right
            else
              if (ColIdx < Length(FColWidth)) and (FColWidth[ColIdx] > 0) then
                W:=FColWidth[ColIdx]
              else
                W:=AListField.DisplayWidth * TextWidth + 8;
            // set rect size
            R.Left:=R.Right;
            R.Right:=R.Right + W;
            RectBmp.Width:=0;
            InitCanvas(RectBmp.Canvas);
            RectBmp.Width:=W;
            RectBmp.Height:=R.Bottom - R.Top;
            // get user defined props
            if Assigned(FGetListItemProps) then
              FGetListItemProps(Parent, RectBmp.Canvas, RowIdx, ColIdx, Selected, GetKeyField, AListField, W, RectBmp.Height, ItemText, ItemBmp, SepTop,
                SepBottom);
            // refresh Rect according updated Width by user props event
            RectBmp.Width:=W;
            R.Right:=R.Left + W;
            // draw text if not empty
            if (ItemText <> '') or not Assigned(FGetListItemProps) then
            begin
              AAlignment:=AListField.Alignment;
              if UseRightToLeftAlignment then
                ChangeBiDiModeAlignment(AAlignment);
              case AAlignment of
                taRightJustify:
                  X:=W - RectBmp.Canvas.TextWidth(ItemText) - 4;
                taCenter:
                  X:=(W - RectBmp.Canvas.TextWidth(ItemText)) div 2;
              else
                X:=LBorder;
              end;
              if X < LBorder then
                X:=LBorder;
              if SysLocale.MiddleEast then
                TControlCanvas(Canvas).UpdateTextFlags;
              Y:=(RowHeight - TextHeight) div 2;
              if Assigned(ItemBmp) and (ItemBmp.Width > 0) and (X < ItemBmp.Width + 4) then
                X:=ItemBmp.Width + 4;
              RectBmp.Canvas.TextRect(Rect(0, 0, RectBmp.Width, RectBmp.Height), X, Y, ItemText);
            end;
            // draw bitmap
            if Assigned(ItemBmp) and (ItemBmp.Width > 0) then
            begin
              rpictool.DrawBmpCenter(RectBmp.Canvas, Rect(2, 0, ItemBmp.Width + 2, RectBmp.Height), ItemBmp);
            end;
            // draw sep line
            if SepTop then
            begin
              RectBmp.Canvas.MoveTo(0, 0);
              RectBmp.Canvas.LineTo(RectBmp.Width, 0);
            end;
            if SepBottom then
            begin
              RectBmp.Canvas.MoveTo(0, RectBmp.Height - 1);
              RectBmp.Canvas.LineTo(RectBmp.Width, RectBmp.Height - 1);
            end;
            // draw column line
            if ColIdx < LastFieldIndex then
            begin
              RectBmp.Canvas.MoveTo(RectBmp.Width - 1, 0);
              RectBmp.Canvas.LineTo(RectBmp.Width - 1, RectBmp.Height);
            end;
            // draw rectbmp to canvas
            Canvas.Draw(R.Left, R.Top, RectBmp);
            // break if no more space
            if R.Right >= ClientWidth then
              Break;
          end;
        end;
        // set rect and fill empty rows
        R.Left:=0;
        R.Right:=ClientWidth;
        if RowIdx >= GetRecordCount then
          Canvas.FillRect(R);
        if Selected then
          BiDi_DrawFocusRect(Canvas, R);
      end;
      if GetRecordCount <> 0 then
        ListLink.ActiveRecord:=GetRecordIndex;
    finally
      FreeAndNil(ItemBmp);
    end;
  finally
    FreeAndNil(RectBmp);
  end;
end;

function TrPopupDataList.GetMaxItemWidth: integer;
var
  I, W: integer;
  F: TField;
  Bk: TBookmark;
begin
  Result:=0;
  if not(Assigned(ListLink) and Assigned(ListLink.DataSet) and ListLink.DataSet.Active) then
    Exit;
  // set array size and reset width
  SetLength(FColWidth, ListFields.Count);
  for I:=0 to ListFields.Count - 1 do
    FColWidth[I]:=0;
  Canvas.Font:=Font;
  try
    Bk:=ListLink.DataSet.GetBookmark;
    try
      // store max width of each column for all list records
      ListLink.DataSet.First;
      while not ListLink.DataSet.Eof do
      begin
        for I:=0 to ListFields.Count - 1 do
        begin
          F:=ListFields[I];
          if Assigned(F) then
          begin
            W:=Canvas.TextWidth(F.DisplayText) + FItemIncreaseWidth + 8;
            if W > FColWidth[I] then
              FColWidth[I]:=W;
          end;
        end;
        ListLink.DataSet.Next;
      end;
      // calc summary width including scrollbar
      for I:=0 to ListFields.Count - 1 do
        Inc(Result, FColWidth[I] + 4);
      Inc(Result, GetSystemMetrics(SM_CXVSCROLL));
    finally
      ListLink.DataSet.GotoBookmark(Bk);
    end;
  except
  end;
end;

procedure TrPopupDataList.ProcessSearchKey(Key: Char);
begin
  if Assigned(FProcessSearchKey) then
    FProcessSearchKey(Self, Key);
  if Key <> #0 then
    inherited ProcessSearchKey(Key);
end;

constructor TrDBLookupComboBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  OwnHintWin:=THintWindow.Create(Self);
  FSearchingResetInterval:=2500;
  FSearchingPrevNext:=stFindFirst;
  FTimer:=TTimer.Create(Self);
  FTimer.SetSubComponent(true);
  FTimer.Enabled:=false;
  FTimer.Interval:=FSearchingResetInterval;
  FTimer.OnTimer:=FOnTimer;
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  FBoundLabel.Position:=lpTopRight;
  DropDownRows:=15;
  FEmptyDisplayValue:='';
  FAllowLookupFilter:=false;
  FAutoDropDown:=false;
  // Self.NullValueKey:=VK_DELETE;
  FDropDownAutoWidth:=true;
  FDropDownMaxWidth:=600;
  FDropDownItemHeight:=0;
  FDropDownItemIncreaseHeight:=2;
  FDropDownItemIncreaseWidth:=0;
  FListFieldIndexMax:=0;
  FDropDownButtonWidth:=GetSystemMetrics(SM_CXVSCROLL);
  // free standard PopupDataList created by DBLookupComboBox
  GetPopupList.Free;
  // create own TrPopupDataList and link to DBLookupComboBox
  FDataListEx:=TrPopupDataList.Create(Self);
  FDataListEx.Visible:=false;
  FDataListEx.Parent:=Self;
  FDataListEx.OnMouseUp:=DataListMouseUp;
  FDataListEx.OnKeyDown:=DataListKeyDown;
  FDataListEx.OnProcessSearchKey:=DataListProcessSearchKey;
  SetPopupList(FDataListEx);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBLookupComboBox.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBLookupComboBox.CMHintShow(var Message: TCMHintShow);
var
  hnt: string;
begin
  if Hint = '' then
  begin
    hnt:=Text;
    if Canvas.TextWidth(hnt) > Self.ClientWidth - 24 then
      Message.HintInfo^.HintStr:=hnt;
  end;
  inherited;
end;

procedure TrDBLookupComboBox.SetDataSource(ADataSource: TDataSource; ADataField: string);
begin
  oldColor:=Color;
  if DataSource <> ADataSource then
    DataSource:=ADataSource;
  if DataField <> ADataField then
    DataField:=ADataField;
  FDataLink.DataSource:=ADataSource;
  UpdateState(nil);
end;

procedure TrDBLookupComboBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource, DataField);
  FDBEditLabel.SetLabelField(ListSource, ListField);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrDBLookupComboBox.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBLookupComboBox.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  if Assigned(FDBEditLabel) then
  begin
    FDBEditLabel.SetField(DataSource, DataField);
    FDBEditLabel.SetLabelField(ListSource, ListField);
  end;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLookupComboBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLookupComboBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBLookupComboBox.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrDBLookupComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FBoundLabel then
      FBoundLabel:=nil;
  end;
end;

procedure TrDBLookupComboBox.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
end;

procedure TrDBLookupComboBox.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrDBLookupComboBox.Paint;
var
  I, W, X, TextW: integer;
  R: TRect;
  Text, Sep: string;
  AAlignment: TAlignment;
  Selected: boolean;
  Details: TThemedElementDetails;
begin
  inherited Paint;
  if (FListFieldIndexMax > ListFieldIndex) { and Assigned(Field) and not Field.IsNull } then
  begin
    // paint more field content
    AAlignment:=taLeftJustify;
    if UseRightToLeftAlignment then
      ChangeBiDiModeAlignment(AAlignment);
    W:=ClientWidth - FDropDownButtonWidth;
    X:=2;
    // get rect
    SetRect(R, 1, 1, W - 1, ClientHeight - 1);
    if UseRightToLeftAlignment then
    begin
      Inc(X, FDropDownButtonWidth);
      Inc(R.Left, FDropDownButtonWidth);
      R.Right:=ClientWidth;
    end;
    if SysLocale.MiddleEast then
      TControlCanvas(Canvas).UpdateTextFlags;
    // draw all fields text
    Sep:='  ' + StringOfChar(' ', FDropDownItemIncreaseWidth div 3);
    Text:='';

    if not ListVisible then
    begin
      // store fields values to array, used after list open
      if LocateKey then
      begin
        SetLength(FldValArr, Min(FListFieldIndexMax, ListFields.Count - 1) - ListFieldIndex + 1);
        for I:=0 to High(FldValArr) do
          FldValArr[I]:=TField(ListFields[ListFieldIndex + I]).DisplayText;
      end
      else
      begin
        SetLength(FldValArr, 1);
        if Assigned(Field) and not Field.IsNull then
          FldValArr[0]:='(' + Field.DisplayText + ')'
        else
          FldValArr[0]:=FEmptyDisplayValue;
      end;
    end;

    // show fields text
    for I:=0 to High(FldValArr) do
    begin
      if I > 0 then
        Text:=Text + Sep + Sep;
      Text:=Text + FldValArr[I];
    end;

    TextW:=Canvas.TextWidth(Text);
    case AAlignment of
      taRightJustify:
        X:=W - TextW - 3;
    end;
    Canvas.TextRect(R, X, 2, Text);
    // draw sep lines
    Canvas.Pen.Color:=clGray;
    Text:='';
    for I:=0 to High(FldValArr) - 1 do
    begin
      Text:=Text + FldValArr[I] + Sep;
      TextW:=Canvas.TextWidth(Text);
      Canvas.MoveTo(TextW, 1);
      Canvas.LineTo(TextW, ClientHeight - 1);
      Text:=Text + Sep;
    end;
    // redraw focus rect
    Selected:=HasFocus and not ListVisible and not(csPaintCopy in ControlState);
    if Selected then
      BiDi_DrawFocusRect(Canvas, R);
  end;

  // show combobox arrow disabled for read only combo or field
  if not CanModify then
  begin
    SetRect(R, ClientWidth - FDropDownButtonWidth, 0, ClientWidth, ClientHeight);
    if UseRightToLeftAlignment then
    begin
      R.Left:=0;
      R.Right:=FDropDownButtonWidth;
    end;
    if GetThemeEnabled then
    begin
      Details:=GetThemeServices.GetElementDetails(tcDropDownButtonDisabled);
      GetThemeServices.DrawElement(Canvas.Handle, Details, R);
    end
    else
      DrawFrameControl(Canvas.Handle, R, DFC_SCROLL, DFCS_SCROLLCOMBOBOX or DFCS_INACTIVE);
  end;
end;

function TrDBLookupComboBox.GetSelectedText: string;
var
  I: integer;
  Sep: string;
begin
  Result:='';
  Sep:='  ' + StringOfChar(' ', FDropDownItemIncreaseWidth div 3);
  for I:=0 to High(FldValArr) do
  begin
    if I > 0 then
      Result:=Result + Sep + Sep;
    Result:=Result + FldValArr[I];
  end;
end;

function TrDBLookupComboBox.GetMaxItemWidth: integer;
begin
  Result:=FDataListEx.GetMaxItemWidth;
end;

procedure TrDBLookupComboBox.DropDown;
var
  W, X: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  // init popup ex props
  FDataListEx.ItemHeight:=DropDownItemHeight;
  FDataListEx.ItemIncreaseHeight:=FDropDownItemIncreaseHeight;
  FDataListEx.ItemIncreaseWidth:=FDropDownItemIncreaseWidth;
  FDataListEx.GetListItemProps:=FGetListItemProps;

  if CanModify and (FDataLink.Editing or FDataLink.Edit) then
  begin
    if FDropDownAutoWidth then
    begin
      // reset lookup filter before dropdown
      if FAllowLookupFilter then
        ListLink.DataSet.Filtered:=false;
      // set list fields to allow calculation of widths
      FDataListEx.KeyField:=KeyField;
      FDataListEx.ListField:=ListField;
      FDataListEx.ListSource:=ListLink.DataSource;
      W:=GetMaxItemWidth;
      if W < Width then
        DropDownWidth:=0
      else
      begin
        if W > FDropDownMaxWidth then
          W:=FDropDownMaxWidth;
        DropDownWidth:=W;
      end;
      X:=Parent.ClientToScreen(Self.BoundsRect.TopLeft).X;
      if X + DropDownWidth > Screen.DesktopRect.Right then
      begin
        DropDownWidth:=Screen.DesktopRect.Right - X;
      end;
    end;
    inherited DropDown;
  end;
end;

procedure TrDBLookupComboBox.CloseUp(Accept: boolean);
begin
  inherited CloseUp(Accept);
  StopSearching;
  // reset lookup filter once closed
  if FAllowLookupFilter then
    ListLink.DataSet.Filtered:=false;
end;

procedure TrDBLookupComboBox.SetFocus;
begin
  try
    inherited SetFocus; // fixed bug when close window with dropeddown listbox
  except
  end;
end;

procedure TrDBLookupComboBox.UpdateState(Sender: TObject);
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if (csDesigning in ComponentState) then
    Exit;
  if FDataLink.Editing and CanModify then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

function TrDBLookupComboBox.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
var
  Key: Word;
begin
  Result:=true;
  while ProcMouseWheel(WheelDelta, Key) do
    SendMessage(Handle, WM_KEYDOWN, Key, 0);
end;

procedure TrDBLookupComboBox.WMKeyDown(var Message: TWMKeyDown);
begin
  if (NullValueKey <> 0) and CanModify and (NullValueKey = ShortCut(Message.CharCode, KeyDataToShiftState(Message.KeyData))) then
  begin
    if Assigned(Field) then
    begin
      FDataLink.Edit;
      Field.Clear;
      Click;
    end
    else
      KeyValue:=null;
    Message.CharCode:=0;
  end;
  inherited;
end;

procedure TrDBLookupComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  // use F4 for dropdown
  if Key = VK_F4 then
  begin
    if ListVisible then
      CloseUp(true)
    else
      DropDown;
    Key:=0;
  end;
  // check searching key if drop down is not shown
  if IsSearching then
    DataListKeyDown(Self, Key, Shift);
  // inherited
  inherited KeyDown(Key, Shift);
end;

procedure TrDBLookupComboBox.KeyPress(var Key: Char);
begin
  if FAutoDropDown and (Key >= #32) and not ListVisible then
    DropDown;
  inherited KeyPress(Key);
end;

procedure TrDBLookupComboBox.DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE, VK_RETURN: // , VK_UP, VK_DOWN:
      StopSearching;
    VK_LEFT, VK_RIGHT:
      begin
        if (ssCtrl in Shift) and FAllowLookupFilter then
          SearchLookupFilter(Key = VK_LEFT)
        else
          SearchNext(Key = VK_LEFT);
        Key:=0;
      end;
  end;
end;

procedure TrDBLookupComboBox.DataListProcessSearchKey(Sender: TObject; var Key: Char);
begin
  ProcessSearchKey(Key);
  Key:=#0;
end;

procedure TrDBLookupComboBox.DataListMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  if Button = mbLeft then
    CloseUp(PtInRect(FDataListEx.ClientRect, Point(X, Y)));
end;

procedure TrDBLookupComboBox.ProcessSearchKey(Key: Char);

  procedure MoveListDataset(Next: boolean);
  begin
    if Next then
      ListLink.DataSet.Next
    else
      ListLink.DataSet.Prior;
  end;

  function CanLookupField(ListFld: TField): boolean;
  begin
    Result:=(ListFld <> nil) and (ListFld.FieldKind in [fkData, fkInternalCalc]) and rdbtool.IsSimpleField(ListFld);
  end;

  function TryLocate(s: string): boolean;
  var
    I: integer;
    ListFld: TField;
  begin
    // try locate text in any field in the list (found if field begins with the text)
    Result:=false;
    for I:=0 to ListFields.Count - 1 do
      try
        ListFld:=ListFields[I];
        if CanLookupField(ListFld) then
          Result:=ListLink.DataSet.Locate(ListFld.FieldName, s, [loCaseInsensitive, loPartialKey]);
        if Result then
          Exit;
      except
        // skip locate error
      end;
  end;

  function TryFind(s: string): boolean;
  var
    Bk: TBookmark;
    Next: boolean;
    I: integer;
    ListFld: TField;
  begin
    // try to find text iin any field in the list (any possition of the text)
    Result:=false;
    // init dataset for scan
    ListLink.DataSet.DisableControls;
    Bk:=ListLink.DataSet.GetBookmark;
    try
      Next:=FSearchingPrevNext in [stFindFirst, stFindNext];
      if FSearchingPrevNext = stFindFirst then
        ListLink.DataSet.First
      else
        MoveListDataset(Next);
      while (Next and not ListLink.DataSet.Eof) or (not Next and not ListLink.DataSet.Bof) do
      begin
        // try locate for each list fields
        for I:=0 to ListFields.Count - 1 do
        begin
          ListFld:=ListFields[I];
          if CanLookupField(ListFld) then
            Result:=AnsiContainsText(ListFld.DisplayText, s);
          if Result then
            Exit;
        end;
        MoveListDataset(Next);
      end;
      // select original record if new record not found
      if not Result then
        try
          ListLink.DataSet.GotoBookmark(Bk);
        except
          // skip bookmark error
        end;
    finally
      // finalize dataset scan
      ListLink.DataSet.EnableControls;
      ListLink.DataSet.FreeBookmark(Bk);
    end;
  end;

  procedure ProcHint(s: string; found: boolean);
  var
    Rp, Rs: TRect;
  begin
    SearchText:=s;
    // activate found key value
    if ListVisible then
      FDataListEx.SelectKeyValue(GetKeyField.Value)
    else
      SelectKeyValue(GetKeyField.Value);
    // prepare hint string
    if not RosiCompConfig.rDBComponentsCfg.ShowSimpleLookupHint then
      s:=GetSearchHintText(s, Key, FSearchingPrevNext, found, ListVisible and FAllowLookupFilter, ListLink.DataSet.Filtered);
    if s <> '' then
    begin
      // find hint win possition and show hint
      Rs:=OwnHintWin.CalcHintRect(Width, s, nil);
      if ListVisible then
      begin
        Rp.TopLeft:=FDataListEx.BoundsRect.TopLeft;
        if Rp.Top < Parent.ClientToScreen(Self.BoundsRect.TopLeft).Y then
          Rp.Top:=Rp.Top - (Rs.Bottom - Rs.Top) - 4
        else
          Rp.Top:=Rp.Top + FDataListEx.Height;
      end
      else
      begin
        Rp.TopLeft:=Parent.ClientToScreen(Self.BoundsRect.TopLeft);
        Rp.Top:=Rp.Top + Height;
      end;
      Rp.Right:=Rp.Left + Rs.Right + 4;
      Rp.Bottom:=Rp.Top + Rs.Bottom + 1;
      OwnHintWin.Color:=Application.HintColor;
      OwnHintWin.ActivateHint(Rp, s);
      // reset timer
      FTimer.Enabled:=false;
      FTimer.Interval:=FSearchingResetInterval;
      FTimer.Enabled:=true;
    end;
    if FAllowLookupFilter and found and (Key <> #0) then
      SearchLookupFilter(false);
    Key:=#0;
  end;

begin
  if not ListActive then
    Exit;
  if FSearchingPrevNext <> stFindFirst then
  begin
    // find next or prev
    ProcHint(SearchText, TryFind(SearchText));
    FSearchingPrevNext:=stFindFirst;
  end
  else
  begin
    case Key of
      #0:
        ProcHint(SearchText, TryLocate(SearchText) or TryFind(SearchText));
      #8:
        if Length(SearchText) = 1 then
        begin
          StopSearching;
          ListLink.DataSet.Filtered:=false;
        end
        else
          ProcHint(copy(SearchText, 1, Length(SearchText) - 1), true);
      #27:
        StopSearching;
      #32 .. High(Char):
        if CanModify and (Length(SearchText) < 32) then
        begin
          if TryLocate(SearchText + Key) or TryFind(SearchText + Key) then
            ProcHint(SearchText + Key, true)
          else
            ProcHint(SearchText, false);
        end;
    end;
  end;
end;

procedure TrDBLookupComboBox.ListLinkDataChanged;
begin
  { no inherited = disable original FIX -
    Fix for Defect# 204311, Non-bound control should clear when list datasource changes }
end;

procedure TrDBLookupComboBox.SearchNext(Previous: boolean);
begin
  if Previous then
    FSearchingPrevNext:=stFindPrev
  else
    FSearchingPrevNext:=stFindNext;
  ProcessSearchKey(#0);
end;

function TrDBLookupComboBox.SearchLookupFilter(ResetFilter: boolean): boolean;
var
  I: integer;
  Fld: TField;
  FilterStr, s: string;
  k: Char;

  function CanLookupField(F: TField): boolean;
  begin
    Result:=(F <> nil) and (F.FieldKind in [fkData, fkInternalCalc]) and rdbtool.IsSimpleField(F);
  end;

  function GetFilterString(F: TField): string;
  var
    LikeCh: string;
    v: double;
  begin
    Result:='';
    if not Assigned(F) or not CanLookupField(F) then
      Exit;
    if F is TNumericField then
    begin
      if TryStrToFloat(SearchText, v) then
      begin
        // standard number, just use it
        Result:=GetFilterCond(F, SearchText, false, '');
      end
    end
    else
      if rdbtool.IsApostrField(F) and not rdbtool.IsDateTimeField(F) then
      begin
        LikeCh:=GetFilterWildCardChar(ListLink.DataSet);
        Result:=GetFilterCond(F, LikeCh + SearchText + LikeCh, false, GetFilterLikeOperator(ListLink.DataSet));
      end;
  end;

begin
  Result:=false;
  k:=#0;
  if ResetFilter then
  begin
    ListLink.DataSet.Filtered:=false;
    if IsSearching then
      ProcessSearchKey(k);
    Result:=true;
  end
  else
    if IsSearching then
    begin
      FilterStr:='';
      // try gen filter for each list fields
      for I:=0 to ListFields.Count - 1 do
      begin
        Fld:=ListFields[I];
        s:=GetFilterString(Fld);
        if s <> '' then
        begin
          if FilterStr <> '' then
            FilterStr:=FilterStr + ' OR ';
          FilterStr:=FilterStr + s;
        end;
      end;
      if FilterStr <> '' then
      begin
        Result:=rdbtool.SetFilterString(ListLink.DataSet, FilterStr, false);
        ProcessSearchKey(k);
      end;
    end;
end;

procedure TrDBLookupComboBox.StopSearching;
begin
  // reset search and hide hint window
  FTimer.Enabled:=false;
  SearchText:='';
  ShowWindow(OwnHintWin.Handle, SW_HIDE);
end;

function TrDBLookupComboBox.IsSearching: boolean;
begin
  Result:=SearchText <> '';
end;

procedure TrDBLookupComboBox.FOnTimer(Sender: TObject);
begin
  StopSearching;
end;

function TrDBLookupComboBox.DropDownDefaultItemHeight: integer;
begin
  Result:=FDataListEx.GetTextHeight - FDataListEx.ItemHeight;
end;

procedure TrDBLookupComboBox.DoExit;
begin
  StopSearching;
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, Field);
  inherited DoExit;
end;

// ==============================================================================
// TrDBCheckBox
constructor TrDBCheckBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReadOnlyColor:=clBtnFace;
  FDataLink:=TrDataLink.Create(UpdateState, UpdateState, nil);
  FShowDisabledIfReadOnly:=false;
  FUpdateAfterClick:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBCheckBox.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBCheckBox.SetDataSource(DS: TDataSource);
begin
  DataSource:=DS;
  FDataLink.DataSource:=DataSource;
  UpdateState(nil);
  UpdateCaption;
end;

procedure TrDBCheckBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBCheckBox.SetShowFieldCaption(const Value: boolean);
begin
  FShowFieldCaption:=Value;
  UpdateCaption;
end;

procedure TrDBCheckBox.UpdateCaption;
var
  Field: TField;
  tempCanvas: TCanvas;
  H: HWND;
begin
  if FShowFieldCaption then
  begin
    Field:=nil;
    try
      if Assigned(DataSource) and Assigned(DataSource.DataSet) then
        Field:=DataSource.DataSet.FindField(DataField);
    except
    end;
    if Field = nil then
      Caption:='FieldLabel'
    else
      if Caption <> Field.DisplayLabel then
      begin
        Caption:=Field.DisplayLabel;
        H:=WindowHandle;
        tempCanvas:=TCanvas.Create;
        try
          tempCanvas.Handle:=GetDeviceContext(H);
          if tempCanvas.TextWidth(Caption) + 20 > Width then
            Width:=tempCanvas.TextWidth(Caption) + 20;
        finally
          FreeAndNil(tempCanvas);
        end;
      end;
  end;
end;

function TrDBCheckBox.GetDataField: string;
begin
  Result:= inherited DataField;
end;

procedure TrDBCheckBox.SetDataField(const Value: string);
begin
  inherited DataField:=Value;
  UpdateCaption;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBCheckBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBCheckBox.Click;
begin
  inherited Click;
  Repaint;
  if FUpdateAfterClick then
    PostMessage(Handle, CM_EXIT, 0, 0);
end;

procedure TrDBCheckBox.UpdateState(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  UpdateCaption;
  Repaint;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBCheckBox.WMPaint(var Message: TWMPaint);
var
  ro: boolean;
  H: HWND;
  tempCanvas: TCanvas;
  R: TRect;
  s: string;
begin
  inherited;
  ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
  if Enabled and FShowDisabledIfReadOnly and ro then
  begin
    H:=WindowHandle;
    tempCanvas:=TCanvas.Create;
    try
      tempCanvas.Handle:=GetDeviceContext(H);
      { // draw gray for read only (do not work in themes)
        R:=ClientRect;
        InflateRect(R, 0, -((R.Bottom - R.Top - 16) div 2));
        tempCanvas.Brush.Color:=FReadOnlyColor;
        if Alignment = taLeftJustify then
        PLeft:=R.Right - 8
        else
        PLeft:=R.Left + 5;

        ro:=ReadOnly or (Assigned(Field) and not Field.CanModify);
        if ro or not FDataLink.Editing then
        tempCanvas.FloodFill(PLeft, R.Top + 5, clWhite, fsSurface);
        }
      // draw gray font if readonly
      R:=ClientRect;
      if Alignment = taRightJustify then
        R.Left:=R.Left + 16
      else
        R.Right:=R.Right - 16;
      tempCanvas.Font.Color:=clGrayText;
      tempCanvas.Brush.Style:=bsClear;
      s:=Caption;
      tempCanvas.TextRect(R, s, [tfVerticalCenter, tfSingleLine]);
    finally
      FreeAndNil(tempCanvas);
    end;
  end;
end;

// ==============================================================================
// TCheckBoxOnly
constructor TCheckBoxOnly.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
  Name:='SubCheckBox';
  Caption:='';
end;

procedure TCheckBoxOnly.SetPos(W: TWinControl; ScaleRatio: double);
var
  X: integer;
begin
  X:=Round(DBCheckBoxWidth * ScaleRatio);
  if (csDesigning in ComponentState) and not Self.Checked then
    SetBounds(W.Left - 1, W.Top, 0, W.Height)
  else
    SetBounds(W.Left - X, W.Top, X, W.Height);
end;

procedure TCheckBoxOnly.CMEnter(var Message: TCMEnter);
begin
  inherited;
  Repaint;
end;

procedure TCheckBoxOnly.CMExit(var Message: TCMExit);
begin
  inherited;
  Repaint;
end;

procedure TCheckBoxOnly.SetInteriorColor(C: TColor);
begin
  FInteriorColor:=C;
  Invalidate;
end;

procedure TCheckBoxOnly.Click;
begin
  inherited Click;
  Repaint;
end;

procedure TCheckBoxOnly.WMPaint(var Message: TWMPaint);
var
  H: HWND;
  tempCanvas: TCanvas;
  R: TRect;
begin
  inherited;
  H:=WindowHandle;
  tempCanvas:=TCanvas.Create;
  try
    tempCanvas.Handle:=GetDeviceContext(H);
    // draw client rect
    R:=ClientRect;
    InflateRect(R, 0, -1);
    R.Right:=R.Right - 2;
    if Focused then
      BiDi_DrawFocusRect(tempCanvas, R);
  finally
    FreeAndNil(tempCanvas);
  end;
end;

// ==============================================================================
// TrDBDateTimePicker
constructor TrDBDateTimePicker.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  Width:=88;
  FReadOnlyColor:=clBtnFace;
  FDataAllowNull:=false;
  FEnabledAll:=true;
  FChanging:=false;
  FReadOnly:=false;
  FAutoSelectDayPart:=false;
{$IFDEF DXEP}
  FFormatEmpty:=FormatSettings.ShortDateFormat;
{$ELSE}
  FFormatEmpty:=ShortDateFormat;
{$ENDIF}
  // FFormatEmpty:=ReplaceStr('d','_',FFormatEmpty,true,false);
  // FFormatEmpty:=ReplaceStr('M','_',FFormatEmpty,true,false);
  // FFormatEmpty:=ReplaceStr('y','_',FFormatEmpty,true,false);
  FFormatEmpty:='__.__.__';

  FDataLink:=TFieldDataLink.Create;
  FDataLink.Control:=Self;
  FDataLink.OnActiveChange:=ActiveChange;
  FDataLink.OnDataChange:=DataChange;
  FDataLink.OnEditingChange:=EditingChange;
  FDataLink.OnUpdateData:=UpdateData;

  FCheckBox:=TCheckBoxOnly.Create(Self);
  FCheckBox.FreeNotification(Self);
  FCheckBox.OnClick:=CheckClick;

  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBDateTimePicker.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBDateTimePicker.Loaded;
begin
  inherited Loaded;
  oldFormat:=Format;
  SetDataSource(DataSource);
  FDBEditLabel.StoreVisible;
  FCheckBox.SetPos(Self, FScaleRatio);
  FCheckBox.SetInteriorColor(Color);
  FCheckBox.Visible:=FDataAllowNull;
  FCheckBox.TabOrder:=TabOrder + 1;
  FCheckBox.TabOrder:=TabOrder;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBDateTimePicker.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
end;

procedure TrDBDateTimePicker.SetDataAllowNull(Value: boolean);
begin
  FDataAllowNull:=Value;
  if Assigned(FCheckBox) then
  begin
    if (csDesigning in ComponentState) then
      FCheckBox.Checked:=FDataAllowNull;
    FCheckBox.SetPos(Self, FScaleRatio);
  end;
  SetSubComp;
end;

procedure TrDBDateTimePicker.SetEnabledAll(Value: boolean);
begin
  FEnabledAll:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=FEnabledAll;
  if Assigned(FCheckBox) then
    FCheckBox.Enabled:=FEnabledAll;
  EditingChange(nil);
  DataChange(nil);
end;

procedure TrDBDateTimePicker.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
end;

procedure TrDBDateTimePicker.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if csLoading in ComponentState then
    Exit;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FCheckBox) then
    FCheckBox.Visible:=Visible and FDataAllowNull;
  // EditingChange(Self);
end;

procedure TrDBDateTimePicker.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FDataLink <> nil) and (AComponent = DataSource) then
      DataSource:=nil;
    if AComponent = FCheckBox then
      FCheckBox:=nil;
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
  end;
end;

procedure TrDBDateTimePicker.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FCheckBox) then
    FCheckBox.SetPos(Self, FScaleRatio);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
end;

procedure TrDBDateTimePicker.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FCheckBox) then
    FCheckBox.Parent:=AParent;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
end;

function TrDBDateTimePicker.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBDateTimePicker.SetDataSource(Value: TDataSource);
begin
  oldColor:=Color;
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLink.DataSource:=Value;
    SetSubComp;
    if not(csDesigning in ComponentState) then
    begin
      EditingChange(nil);
      DataChange(nil);
      SetVisibilityFromField(Self, DataSource, DataField);
    end;
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

function TrDBDateTimePicker.GetDataField: string;
begin
  Result:=FDataLink.FieldName;
end;

procedure TrDBDateTimePicker.SetDataField(const Value: string);
begin
  FDataLink.FieldName:=Value;
  SetSubComp;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBDateTimePicker.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

function TrDBDateTimePicker.GetIsChecked: boolean;
begin
  Result:=FCheckBox.Checked;
end;

procedure TrDBDateTimePicker.SetSubComp;
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  if Assigned(FCheckBox) then
  begin
    FCheckBox.Visible:=Visible and FDataAllowNull;
    if not(csDesigning in ComponentState) then
    begin
      FCheckBox.Enabled:=Assigned(DataSource);
      if not FCheckBox.Enabled then
        FCheckBox.Checked:=false;
    end;
  end;
end;

function TrDBDateTimePicker.CanEdit: boolean;
begin
  Result:=not FReadOnly and FDataLink.Edit;
end;

procedure TrDBDateTimePicker.Change;
begin
  if CanEdit then
  begin
    FDataLink.Modified;
    inherited Change;
  end
  else
    DataChange(nil);
end;

procedure TrDBDateTimePicker.SetFocused(Value: boolean);
begin
  if FFocused <> Value then
  begin
    FFocused:=Value;
    // FDataLink.Reset;
  end;
end;

procedure TrDBDateTimePicker.SetIntFormat;
const
  cFrmNothing = '''''';
begin
  if oldFormat = cFrmNothing then
    oldFormat:='';
  if not Assigned(FDataLink.Field) then
  begin
    if (csDesigning in ComponentState) then
      Format:=oldFormat
    else
    begin
      FCheckBox.Checked:=false;
      Format:='''''';
    end;
  end
  else
    if not(FDataLink.Field.IsNull) then
      Format:=oldFormat
    else
      if FFocused then
        Format:=FFormatEmpty
      else
        Format:=cFrmNothing;
end;

procedure TrDBDateTimePicker.CMEnter(var Message: TCMEnter);
begin
  SetFocused(true);
  inherited;
  SetIntFormat;
end;

procedure TrDBDateTimePicker.CMExit(var Message: TCMExit);
begin
  try
    if FDataLink.Editing then
      FDataLink.UpdateRecord;
  except
    SetFocus;
    raise ;
  end;
  SetFocused(false);
  DoExit;
  SetIntFormat;
end;

procedure TrDBDateTimePicker.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if not Assigned(FDataLink.Field) then
    Exit;
  if (Key = VK_DELETE) and CanEdit then
  begin
    FDataLink.Field.Clear;
    FDataLink.UpdateRecord;
    SetIntFormat;
    Key:=0;
  end;
  if (Key in [$30 .. $39, VK_NUMPAD0 .. VK_NUMPAD9, VK_UP, VK_DOWN]) and CanEdit then
  begin
    if FDataLink.Field.IsNull then
    begin
      Date:=SysUtils.Date;
      FDataLink.Field.AsDateTime:=SysUtils.Date;
      // FDatalink.UpdateRecord;
      SetIntFormat;
      // select first part of date
      Perform(WM_KEYDOWN, VK_RIGHT, 0);
      Perform(WM_KEYUP, VK_RIGHT, 0);
    end;
  end;
  inherited KeyDown(Key, Shift);
end;

procedure TrDBDateTimePicker.KeyPress(var Key: Char);
begin
  if CharInSet(Key, ['0' .. '9']) and not CanEdit then
    Key:=#0
  else
    inherited KeyPress(Key);
end;

procedure TrDBDateTimePicker.CNNotify(var Message: TWMNotifyDT);
var
  st: TsystemTime;
begin
  if not Assigned(FDataLink.Field) then
    inherited
  else
    with Message, NMHdr {$IFNDEF CLR}^{$ENDIF} do
    begin
      Result:=0;
      case code of
        DTN_DROPDOWN:
          begin
            if FDataLink.Field.IsNull then
              Date:=SysUtils.Date;
            DateTimeToSystemTime(Date, st);
            MonthCal_SetCurSel(Self.CalendarHandle, st);
            if CanEdit then
            begin
              // FDatalink.UpdateRecord;
              // SetIntFormat;
            end;
            inherited;
          end;
        DTN_CLOSEUP:
          begin
            // suppress nil value warning
            try
              inherited;
            except
            end;
            try
              if CanEdit then
              begin
                if Date = 0 then
                  FDataLink.Field.Clear;
                FDataLink.UpdateRecord;
              end
              else
                if not FDataLink.Field.IsNull then
                  Date:=FDataLink.Field.AsDateTime;
            except
              if not FDataLink.Field.IsNull then
                Date:=FDataLink.Field.AsDateTime;
            end;
            SetIntFormat;
{$IFDEF DEMO}
            ShowDemoMsg;
{$ENDIF}
          end;
      else
        inherited;
      end;
    end;
end;

procedure TrDBDateTimePicker.CheckClick(Sender: TObject);
begin
  if csDesigning in ComponentState then
    Exit;
  Enabled:=FCheckBox.Checked and FEnabledAll;
  if not FCheckBox.Checked then
    DateTime:=Trunc(SysUtils.Date);
  if not FChanging then
    try
      FChanging:=true;
      try
        Change;
      finally
        FChanging:=false;
      end;
      FDataLink.Modified;
      if not FReadOnly and IsEditing(DataSource) then
        FDataLink.UpdateRecord
      else
        DataChange(nil);
      if FCheckBox.Checked then
        SetFocus;
    except
      DataChange(nil);
    end;
  SetIntFormat;
  FDataLink.Reset;
end;

procedure TrDBDateTimePicker.ActiveChange(Sender: TObject);
begin
  SetSubComp;
  SetVisibilityFromField(Self, DataSource, DataField);
  SetIntFormat;
end;

procedure TrDBDateTimePicker.DataChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  if Assigned(FDataLink.Field) and not FChanging then
  begin
    DateTime:=FDataLink.Field.AsDateTime;
    if FDataAllowNull then
    begin
      FChanging:=true;
      try
        FCheckBox.Checked:=not FDataLink.Field.IsNull;
        if not FCheckBox.Checked then
          DateTime:=Trunc(SysUtils.Date);
      finally
        FChanging:=false;
      end;
    end;
    SetIntFormat;
  end;
end;

procedure TrDBDateTimePicker.EditingChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  Enabled:=((not FDataAllowNull) or FCheckBox.Checked) and FEnabledAll;
  // FDataLink.Reset;
  FCheckBox.Visible:=Visible and FDataAllowNull;
  FCheckBox.Enabled:=FEnabledAll and Assigned(FDataLink.Field);
  if not(csLoading in ComponentState) then
  begin
    if FDataLink.Editing then
      Color:=oldColor
    else
      Color:=FReadOnlyColor;
    // if Assigned(FCheckBox) then FCheckBox.SetInteriorColor(Color);
  end;
end;

procedure TrDBDateTimePicker.UpdateData(Sender: TObject);
begin
  if Assigned(FDataLink.Field) then
  begin
    if Kind = dtkDate then
      FDataLink.Field.AsDateTime:=Trunc(DateTime)
    else
      FDataLink.Field.AsDateTime:=DateTime;
    if FDataAllowNull and (not FCheckBox.Checked) then
      FDataLink.Field.Clear;
  end;
end;

procedure TrDBDateTimePicker.DoExit;
var
  orig: TDTDateFormat;
begin
  RosiCompConfig.rDBComponentsCfg.DoExitControl(Self, FDataLink.Field);
  inherited DoExit;
  if FAutoSelectDayPart then
  begin
    orig:=DateFormat;
    if orig = dfShort then
      DateFormat:=dfLong
    else
      DateFormat:=dfShort;
    if orig = dfShort then
      DateFormat:=dfShort
    else
      DateFormat:=dfLong;
  end;
end;

// ==============================================================================
// TrDBImage
constructor TrDBImage.Create(AOwner: TComponent);
var
  M: TMenuItem;
  I: integer;
begin
  inherited Create(AOwner);
  SaveInJPG:=RosiCompConfig.rDBComponentsCfg.JPGAsDefault;
  FShowConvertMenu:=true;
  FPictureIfEmpty:=TPicture.Create;

  FDataLink:=TFieldDataLink.Create;
  FDataLink.Control:=Self;
  FDataLink.OnActiveChange:=ActiveChange;
  FDataLink.OnDataChange:=DataChange;
  FDataLink.OnEditingChange:=nil;
  FDataLink.OnUpdateData:=DataChange;

  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);

  FBoundButtons:=TBoundButtons.Create(Self);
  FBoundButtons.RefreshEvent:=RefBoundButtons;
  FBoundToolBar:=TToolBar.Create(Self);
  FBoundToolBar.FreeNotification(Self);
  FBoundToolBar.SetSubComponent(true);
  FBoundToolBar.Flat:=true;
  FBoundToolBar.Align:=alNone;
  for I:=1 to rDBImageBtnCount do
  begin
    FBtnArr[I]:=TToolButton.Create(Self);
    FBtnArr[I].FreeNotification(Self);
    FBtnArr[I].Tag:=I;
    FBtnArr[I].OnClick:=ClickBoundButtons;
  end;
  FBtnArr[1].Style:=tbsCheck;

  // add convert submenu
  M:=CreateMenuItem(Self, sPicConvert, cPicConvertMenu, PopupClick);
  FPopupMenu.Items.Insert(3, M);
  M.Add(CreateMenuItem(Self, sPicConvertJPG1, sPicConvertToJPG1, PopupClick));
  M.Add(CreateMenuItem(Self, sPicConvertJPGAll, sPicConvertToJPGAll, PopupClick));
  M.Add(CreateMenuItem(Self, '-', cPicNone, PopupClick));
  M.Add(CreateMenuItem(Self, sPicConvertBMP1, sPicConvertToBMP1, PopupClick));
  M.Add(CreateMenuItem(Self, sPicConvertBMPAll, sPicConvertToBMPAll, PopupClick));
  FPopupMenu.Items.Add(CreateMenuItem(Self, '-', cPicNone, PopupClick));

  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBImage.Destroy;
begin
  FreeAndNil(FDataLink);
  FreeAndNil(FPictureIfEmpty);
  inherited Destroy;
end;

procedure TrDBImage.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource);
  FDBEditLabel.StoreVisible;
  if not GetThemeEnabled then
    BoundButtons.FTransparent:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBImage.Notification(AComponent: TComponent; Operation: TOperation);
var
  I: integer;
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FDataLink <> nil) and (AComponent = DataSource) then
      DataSource:=nil;
    if AComponent = FDBEditLabel then
      FDBEditLabel:=nil;
    if AComponent = FBoundToolBar then
      FBoundToolBar:=nil;
    for I:=1 to rDBImageBtnCount do
      if (AComponent = FBtnArr[I]) then
        FBtnArr[I]:=nil;
  end;
end;

procedure TrDBImage.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  RefBoundButtons(Self);
end;

procedure TrDBImage.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FBoundToolBar) then
    FBoundToolBar.Parent:=AParent;
end;

procedure TrDBImage.SetShowConvertMenu(Value: boolean);
begin
  if Value <> FShowConvertMenu then
  begin
    FShowConvertMenu:=Value;
    GetPopupItem(cPicConvertMenu).Visible:=FShowConvertMenu;
  end;
end;

procedure TrDBImage.CMEnabledChanged(var Message: TMessage);
var
  I: integer;
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  for I:=1 to rDBImageBtnCount do
    if Assigned(FBtnArr[I]) then
      FBtnArr[I].Enabled:=Enabled;
end;

procedure TrDBImage.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  RefBoundButtons(Self);
end;

procedure TrDBImage.SetPictureIfEmpty(Value: TPicture);
begin
  FPictureIfEmpty.Assign(Value);
end;

function TrDBImage.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBImage.SetDataSource(Value: TDataSource);
begin
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLink.DataSource:=Value;
    DataChange(nil);
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
  if (csDesigning in ComponentState) then
    Exit;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

function TrDBImage.GetDataField: string;
begin
  Result:=FDataLink.FieldName;
end;

procedure TrDBImage.SetDataField(const Value: string);
begin
  FDataLink.FieldName:=Value;
  DataChange(nil);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBImage.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBImage.RefBoundButtons(Sender: TObject);
var
  I: integer;
  px, py, W, H: integer;
begin
  if not Assigned(FBoundToolBar) or not Assigned(FBoundToolBar.Parent) then
    Exit;
  for I:=1 to rDBImageBtnCount do
    if not Assigned(FBtnArr[I]) then
      Exit;

  // set toolbar
  FBoundToolBar.Visible:=FBoundButtons.Visible and Visible;
  FBoundToolBar.Transparent:=FBoundButtons.Transparent;
  FBoundToolBar.DoubleBuffered:=false;
  FBoundToolBar.Flat:=FBoundButtons.Flat;
  FBoundToolBar.ShowHint:=ShowHint;
  if FBoundButtons.FShowGlyphs then
    FBoundToolBar.Images:=FBoundButtons.Images
  else
    FBoundToolBar.Images:=nil;
  FBoundToolBar.ShowCaptions:=not FBoundButtons.FShowGlyphs;

  // proc tool buttons
  for I:=rDBImageBtnCount downto 1 do
  begin
    FBtnArr[I].Parent:=FBoundToolBar;
    if FBoundButtons.FShowGlyphs then
      FBtnArr[I].ImageIndex:=I - 1
    else
      FBtnArr[I].Caption:=rtool.FindStrPart(FBoundButtons.Captions, ';', I, true);
    case I of
      1:
        FBtnArr[I].Hint:=sZoomAutoFit;
      2:
        FBtnArr[I].Hint:=sZoomIn;
      3:
        FBtnArr[I].Hint:=sZoomOut;
      4:
        FBtnArr[I].Hint:=sShowPicEdit;
    end;
    FBtnArr[I].Enabled:=not LockOriginalZoom;
  end;
  FBtnArr[4].Visible:=(ShowEditor <> ieNone);

  // set possition
  if BoundButtons.FPosition in [bpTopLeft, bpTopRight, bpBottomLeft, bpBottomRight] then
  begin
    W:=FBtnArr[1].Width * FBoundToolBar.ButtonCount + 1;
    H:=FBtnArr[1].Height + 1;
  end
  else
  begin
    W:=FBtnArr[1].Width + 1;
    H:=FBtnArr[1].Height * FBoundToolBar.ButtonCount + 1;
  end;

  if FBoundButtons.FPosition in [bpTopLeft, bpBottomLeft] then
    px:=Left
  else
    if FBoundButtons.FPosition in [bpTopRight, bpBottomRight] then
      px:=Left + Width - W
    else
      if FBoundButtons.FPosition in [bpLeftTop, bpLeftBottom] then
        px:=Left - W
      else
        px:=Left + Width;

  if BoundButtons.FPosition in [bpTopLeft, bpTopRight] then
    py:=Top - H
  else
    if BoundButtons.FPosition in [bpLeftTop, bpRightTop] then
      py:=Top
    else
      if BoundButtons.FPosition in [bpLeftBottom, bpRightBottom] then
        py:=Top + Height - H
      else
        py:=Top + Height;

  FBoundToolBar.AutoSize:=false;
  if not FBoundButtons.Visible then
  begin
    px:=Left;
    py:=Top;
    W:=10;
    H:=10;
  end;
  FBoundToolBar.SetBounds(px + 1, py + 1, W, H);
  if FBoundButtons.Visible then
    FBoundToolBar.AutoSize:=true;

  FBtnArr[1].Down:=ReduceToFit;
end;

procedure TrDBImage.ClickBoundButtons(Sender: TObject);
begin
  case TToolButton(Sender).Tag of
    1:
      PopupClick(GetPopupItem(cPicZoomFit));
    2:
      PopupClick(GetPopupItem(cPicZoomIn));
    3:
      PopupClick(GetPopupItem(cPicZoomOut));
    4:
      PopupClick(GetPopupItem(cPicEdit));
  end;
end;

procedure TrDBImage.ZoomChanged;
begin
  if Assigned(FBtnArr[1]) then
    FBtnArr[1].Down:=ReduceToFit;
  inherited;
end;

function TrDBImage.CanModify: boolean;
begin
  Result:= inherited and Assigned(FDataLink.Field) and not FDataLink.Field.ReadOnly and FDataLink.CanModify and (IsEditing(DataSource) or DataSource.AutoEdit);
end;

function TrDBImage.CanEdit: boolean;
begin
  Result:=CanModify;
  if Result then
    try
      DataSource.Edit;
    except
      Result:=false;
    end;
end;

procedure TrDBImage.ActiveChange(Sender: TObject);
begin
  DataChange(Sender);
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBImage.DataChange(Sender: TObject);
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  Picture:=nil;
  if Assigned(FDataLink) and Assigned(FDataLink.Field) then
  begin
    ImgGraphicType:=ReadPictureFromField(FDataLink.Field, Picture);
    case ImgGraphicType of
      rpictool.gtNone:
        if Assigned(FPictureIfEmpty.Graphic) and (FPictureIfEmpty.Graphic.Width > 0) then
        begin
          BackgroundText:='';
          Picture.Assign(FPictureIfEmpty);
          LockOriginalZoom:=true;
        end
        else
          BackgroundText:=sPictureEmpty;
      gtError:
        BackgroundText:=sPictureCorrupted;
    else
      begin
        BackgroundText:='';
        LockOriginalZoom:=false;
      end;
    end;
  end;
  RefBoundButtons(Self);
end;

procedure TrDBImage.IntSaveToField(FieldType: TrGraphicType);
var
  Max: integer;
begin
  if not CanEdit then
    Exit;
  ImgGraphicType:=FieldType;
  if ImgGraphicType = rpictool.gtNone then
    if SaveInJPG then
      ImgGraphicType:=gtJPG
    else
      ImgGraphicType:=gtBMP;

  if (FMaxImageSize > 0) then
    if (Picture.Width > FMaxImageSize) or (Picture.Height > FMaxImageSize) then
    begin
      Max:=Picture.Width;
      if Picture.Height > Max then
        Max:=Picture.Height;
      ResizePic(FMaxImageSize * 100 / Max);
      Exit;
    end;
  WritePictureToField(FDataLink.Field, Picture, ImgGraphicType);
end;

procedure TrDBImage.OnAbort(Sender: TObject);
begin
  Aborted:=true;
end;

procedure TrDBImage.OnPopupMenu(Sender: TObject);
begin
  inherited;
  GetPopupItem(cPicConvertMenu).Visible:=FShowConvertMenu;
end;

procedure TrDBImage.PopupClick(Sender: TObject);

  procedure ConvertOneRec(TargetFormat: TrGraphicType);
  var
    e: boolean;
  begin
    if not(ImgGraphicType in [rpictool.gtNone, gtError, TargetFormat]) then
    begin
      e:=IsEditing(FDataLink.DataSet);
      if not e then
        FDataLink.DataSet.Edit;
      IntSaveToField(TargetFormat);
      if not e then
        FDataLink.DataSet.Post;
    end;
  end;

  procedure Convert(All: boolean; TargetFormat: TrGraphicType);
  var
    RecIdx, RecCount: integer;
  begin
    if FDataLink.Field.ReadOnly or FDataLink.ReadOnly then
    begin
      DlgE(sEditRecordErr);
      Exit;
    end;
    try
      if All then
      begin
        if DlgQ2(sPicConvertAllQ, SContinueQ) <> mrYes then
          Exit;
        FDataLink.DataSet.First;
        Aborted:=false;
        RecIdx:=0;
        RecCount:=FDataLink.DataSet.RecordCount;
        while not FDataLink.DataSet.Eof do
        begin
          ShowInfoProgressBar(sPicConvertProcess, nil, OnAbort, Round(100 * RecIdx / RecCount));
          ConvertOneRec(TargetFormat);
          FDataLink.DataSet.Next;
          // sleep(1000);
          Application.ProcessMessages;
          if Aborted then
            Break;
          Inc(RecIdx);
        end;
      end
      else
      begin
        WCursor;
        ConvertOneRec(TargetFormat);
      end;
    finally
      ShowInfo('');
      NCursor;
    end;
  end;

begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  inherited;
  case TMenuItem(Sender).Tag of
    cPicSaveInJpgFormat:
      begin
        if SaveInJPG then
          IntSaveToField(gtJPG)
        else
          IntSaveToField(gtBMP);
      end;
    sPicConvertToJPG1:
      Convert(false, gtJPG);
    sPicConvertToBMP1:
      Convert(false, gtBMP);
    sPicConvertToJPGAll:
      Convert(true, gtJPG);
    sPicConvertToBMPAll:
      Convert(true, gtBMP);
  end;
end;

function TrDBImage.CutImageToClipboard: boolean;
begin
  Result:=CanEdit and inherited;
  if Result then
    FDataLink.Field.Clear;
end;

function TrDBImage.PasteImageFromClipboard: boolean;
begin
  Result:=CanEdit and inherited;
  if Result then
    IntSaveToField(ImgGraphicType);
end;

function TrDBImage.LoadFromFile(const FileName: string): boolean;
begin
  Result:=CanEdit and inherited LoadFromFile(FileName);
  if Result then
    IntSaveToField(ImgGraphicType);
end;

function TrDBImage.ResizePic(ratio: double): boolean;
begin
  Result:=CanEdit and inherited;
  if Result then
    IntSaveToField(ImgGraphicType);
end;

function TrDBImage.ShowEditorWindow: boolean;
begin
  Result:=false;
  if rImageEdit.ShowImageEdit(nil, Picture, nil, ShowEditor in [ieShowModal, ieEditModal], (ShowEditor = ieEditModal) and CanEdit) then
  begin
    // update
    Self.Picture.Assign(Picture);
    IntSaveToField(ImgGraphicType);
    Result:=true;
  end;
end;

// ==============================================================================
// TrDBLabel
constructor TrDBLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink:=TFieldDataLink.Create;
  FDataLink.Control:=Self;
  FDataLink.OnActiveChange:=ActiveChange;
  FDataLink.OnDataChange:=DataChange;
  FDataLink.OnEditingChange:=EditingChange;
end;

destructor TrDBLabel.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBLabel.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource);
end;

procedure TrDBLabel.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FDataLink <> nil) and (AComponent = DataSource) then
      DataSource:=nil;
  end;
end;

function TrDBLabel.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBLabel.SetDataSource(Value: TDataSource);
begin
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLink.DataSource:=Value;
    EditingChange(nil);
    DataChange(nil);
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
  if (csDesigning in ComponentState) then
    Exit;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

function TrDBLabel.GetDataField: string;
begin
  Result:=FDataLink.FieldName;
end;

procedure TrDBLabel.SetDataField(const Value: string);
begin
  FDataLink.FieldName:=Value;
  DataChange(nil);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLabel.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBLabel.ActiveChange(Sender: TObject);
begin
  DataChange(Sender);
  SetVisibilityFromField(Self, DataSource, DataField);
end;

procedure TrDBLabel.DataChange(Sender: TObject);
begin
  if Assigned(FDataLink) and Assigned(FDataLink.Field) then
  begin
    if (FDataLink.Field is TMemoField) or (FDataLink.Field is TWideMemoField) then
      Caption:=FDataLink.Field.AsString
    else
      Caption:=FDataLink.Field.DisplayText;
  end
  else
    if (csDesigning in ComponentState) then
      Self.Caption:=Name
    else
      Self.Caption:='';
end;

procedure TrDBLabel.EditingChange(Sender: TObject);
begin
end;

// ==============================================================================
// TrDBCheckListBox
constructor TCheckItem.Create(ItemId: integer; const ItemText: string; IsChecked: boolean);
begin
  Id:=ItemId;
  Text:=ItemText;
  Checked:=IsChecked;
end;

function TCheckItemList.CheckItem(Index: integer): TCheckItem;
begin
  Result:=Items[Index] as TCheckItem;
end;

procedure TCheckItemList.AddCheckItem(ItemId: integer; const ItemText: string; IsChecked: boolean);
begin
  Add(TCheckItem.Create(ItemId, ItemText, IsChecked));
end;

procedure TCheckItemList.SetItemChecked(ItemId: integer; IsChecked: boolean);
var
  I: integer;
begin
  for I:=0 to Count - 1 do
  begin
    if CheckItem(I).Id = ItemId then
      CheckItem(I).Checked:=IsChecked;
  end;
end;

function TCheckItemList.GetItemChecked(ItemId: integer): boolean;
var
  I: integer;
begin
  Result:=false;
  for I:=0 to Count - 1 do
  begin
    if CheckItem(I).Id = ItemId then
    begin
      Result:=CheckItem(I).Checked;
      Exit;
    end;
  end;
end;

procedure TCheckItemList.ResetItemsChecked;
var
  I: integer;
begin
  for I:=0 to Count - 1 do
    CheckItem(I).Checked:=false;
end;

function TCheckItemList.CheckedCount: integer;
var
  I: integer;
begin
  Result:=0;
  for I:=0 to Count - 1 do
    if CheckItem(I).Checked then
      Inc(Result);
end;

procedure TCheckItemList.StoreItemsForDuplication;
var
  I, M: integer;
begin
  SetLength(FDuplicIdArr, CheckedCount);
  M:=0;
  for I:=0 to Count - 1 do
    if CheckItem(I).Checked then
    begin
      FDuplicIdArr[M]:=CheckItem(I).Id;
      Inc(M);
    end;
end;

procedure TCheckItemList.RestoreDuplicatedItems;
var
  I: integer;
begin
  for I:=0 to High(FDuplicIdArr) do
    Self.SetItemChecked(FDuplicIdArr[I], true);
end;

constructor TrDBCheckListBox.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FReadOnlyColor:=clBtnFace;
  FChanging:=false;
  FPostponedSaving:=false;
  FPostponeSaveNewRecord:=false;
  FDataLink:=TFieldDataLink.Create;
  FDataLink.Control:=Self;
  FDataLink.OnActiveChange:=ActiveChange;
  FDataLink.OnDataChange:=DataChange;
  FDataLink.OnEditingChange:=EditingChange;
  FDataLink.OnUpdateData:=UpdateData;

  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);

  FBoundLabel:=rDBComponents.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  FBoundLabel.Name:='SubLabel';
  FBoundLabel.Caption:='';
  FBoundLabel.Position:=lpTopRight;

  FFindEdit:=TBoundButtonedEdit.CreateLinkEvent(Self, ChangeFindText);
  FFindEdit.FreeNotification(Self);
  FFindEditSpace:=4;
  FFindItemsStyle:=fisNone;
  FFindEdit.Visible:=false;
  oldFindText:='';

  FAllItems:=TCheckItemList.Create;
  FAllItems.OwnsObjects:=true;
  FCheckboxUpdateMode:=chmCheckboxClick;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBCheckListBox.Destroy;
begin
  FreeAndNil(FDataLink);
  FreeAndNil(FAllItems);
  inherited Destroy;
end;

procedure TrDBCheckListBox.Loaded;
begin
  inherited Loaded;
  SetDataSource(DataSource);
  FDBEditLabel.StoreVisible;
  FFindEdit.Visible:=FFindItemsStyle <> fisNone;
  FFindEdit.TabOrder:=TabOrder + 1;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBCheckListBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FDataLink <> nil) and (AComponent = DataSource) then
      DataSource:=nil;
    if (AComponent = FDBEditLabel) then
      FDBEditLabel:=nil;
    if (AComponent = FFindEdit) then
      FFindEdit:=nil;
  end;
end;

procedure TrDBCheckListBox.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  if Assigned(FFindEdit) then
    FFindEdit.RefreshPosition(Self, FFindEditSpace);
  if Assigned(FBoundLabel) then
    FBoundLabel.RefreshPosition;
end;

procedure TrDBCheckListBox.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
  if Assigned(FFindEdit) then
    FFindEdit.Parent:=AParent;
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
  Self.DoubleBuffered:=false;
end;

procedure TrDBCheckListBox.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FFindEdit) then
    FFindEdit.Enabled:=Enabled;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrDBCheckListBox.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  if Assigned(FFindEdit) then
    FFindEdit.Visible:=Visible and (FFindItemsStyle <> fisNone);
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrDBCheckListBox.CMExit(var Message: TCMExit);
begin
  ItemIndex:=-1;
end;

function TrDBCheckListBox.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBCheckListBox.SetDataSource(Value: TDataSource);
begin
  // if (csLoading in ComponentState) then
  oldColor:=Color;
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLink.DataSource:=Value;
    if Assigned(FDBEditLabel) then
      FDBEditLabel.SetField(DataSource, DataField);
    EditingChange(nil);
    DataChange(nil);
    // UpdateChecklistState;
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
  if (csDesigning in ComponentState) then
    Exit;
  SetVisibilityFromField(Self, DataSource, DataField);
end;

function TrDBCheckListBox.GetDataField: string;
begin
  Result:=FDataLink.FieldName;
end;

procedure TrDBCheckListBox.SetDataField(const Value: string);
begin
  FDataLink.FieldName:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, DataField);
  DataChange(nil);
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

function TrDBCheckListBox.GetField: TField;
begin
  Result:=FDataLink.Field;
end;

procedure TrDBCheckListBox.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataField);
end;

procedure TrDBCheckListBox.SetReadOnly(Value: boolean);
begin
  FReadOnly:=Value;
  if not(csLoading in ComponentState) then
    UpdateChecklistState;
end;

procedure TrDBCheckListBox.SetCheckItemsStyle(Value: TCheckItemsStyle);
begin
  FCheckItemsStyle:=Value;
  if not(csLoading in ComponentState) then
    FilterItemsList;
end;

procedure TrDBCheckListBox.SetFindEditSpace(Value: integer);
begin
  FFindEditSpace:=Value;
  if Assigned(FFindEdit) then
    FFindEdit.RefreshPosition(Self, FFindEditSpace);
end;

procedure TrDBCheckListBox.SetFindItemsStyle(Value: TFindItemsStyle);
begin
  FFindItemsStyle:=Value;
  FFindEdit.Visible:=Visible and (FFindItemsStyle <> fisNone);
  if Assigned(FFindEdit) then
    FFindEdit.RefreshPosition(Self, FFindEditSpace);
end;

function TrDBCheckListBox.CanEdit: boolean;
begin
  Result:=false;
  if not Assigned(FDataLink.Field) then
    Exit; // or FDataLink.Field.ReadOnly then Exit;
  if IsEditing(DataSource) then
    Result:=true
  else
    if DataSource.AutoEdit then
      try
        DataSource.Edit;
        Result:=true;
      except
      end;
end;

procedure TrDBCheckListBox.UpdateChecklistState;
var
  I: integer;
  ro: boolean;
begin
  if (csDesigning in ComponentState) then
    Exit;
  if (csLoading in ComponentState) then
    Exit;
  ro:=FReadOnly or not Assigned(FDataLink.Field); // or FDataLink.Field.ReadOnly;
  // check dataset state instead of Datalink in case ReadOnly field
  if Assigned(FDataLink.DataSet) and IsEditing(FDataLink.DataSet) and not ro then
    Color:=oldColor
  else
    Color:=FReadOnlyColor;
  for I:=0 to Items.Count - 1 do
    ItemEnabled[I]:=not ro;
end;

function TrDBCheckListBox.CanModifyCheckItem(Index: integer): boolean;
begin
  Result:=(Index >= 0) and (Index < Items.Count) and ItemEnabled[Index];
end;

procedure TrDBCheckListBox.DblClick;
begin
  if (FCheckboxUpdateMode = chmItemDblClick) and CanModifyCheckItem(ItemIndex) then
  begin
    Checked[ItemIndex]:=not Checked[ItemIndex];
    ClickCheck;
  end;
  inherited;
end;

procedure TrDBCheckListBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X: integer; Y: integer);
var
  oldStatus: boolean;
begin
  if (FCheckboxUpdateMode = chmItemClick) and CanModifyCheckItem(ItemIndex) then
  begin
    oldStatus:=Checked[ItemIndex];
    inherited;
    if oldStatus = Checked[ItemIndex] then
    begin
      Checked[ItemIndex]:=not Checked[ItemIndex];
      ClickCheck;
    end;
  end
  else
    inherited;
end;

procedure TrDBCheckListBox.ClickCheck;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if not Assigned(FDataLink.Field) then
  begin
    inherited ClickCheck;
    FAllItems.SetItemChecked(integer(Items.Objects[ItemIndex]), Checked[ItemIndex]);
    Exit;
  end;
  FChanging:=true;
  try
    CanEdit;
    if IsEditing(DataSource) then
    begin
      FDataLink.Modified;
      inherited ClickCheck;
      // set checked to internal list
      FAllItems.SetItemChecked(integer(Items.Objects[ItemIndex]), Checked[ItemIndex]);
    end;
  finally
    FChanging:=false;
  end;
  if not IsEditing(DataSource) then
    DataChange(nil);
end;

procedure TrDBCheckListBox.DoLoadItemList;
begin
  if Assigned(FOnLoadItemList) then
    FOnLoadItemList(Self, FAllItems);
  UpdateChecklistState;
end;

procedure TrDBCheckListBox.DoLoadItemValues;
begin
  if not FPostponedSaving then
    if Assigned(FOnLoadItemValues) then
      FOnLoadItemValues(Self, FAllItems);
end;

procedure TrDBCheckListBox.DoSaveItemValues;
begin
  FPostponedSaving:=false;
  if Assigned(FOnSaveItemValues) then
    FOnSaveItemValues(Self, FAllItems);
end;

procedure TrDBCheckListBox.DoCancelItemValues;
begin
  FPostponedSaving:=false;
  DoLoadItemValues;
end;

procedure TrDBCheckListBox.ActiveChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  EditingChange(nil);
  // DoLoadItemList;
end;

procedure TrDBCheckListBox.DataChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  if not FChanging and Assigned(FDataLink.Field) then
    DoLoadItemValues;
  // editing change is not call automatically if key field is Read only => call if editing changed
  if Assigned(FDataLink.DataSet) and (oldEditState <> IsEditing(FDataLink.DataSet)) then
    EditingChange(nil);
end;

procedure TrDBCheckListBox.EditingChange(Sender: TObject);
begin
  if (csDesigning in ComponentState) then
    Exit;
  oldEditState:=Assigned(FDataLink.DataSet) and IsEditing(FDataLink.DataSet);
  UpdateChecklistState;
end;

procedure TrDBCheckListBox.UpdateData(Sender: TObject);
begin
  if not Assigned(FDataLink.Field) then
    Exit;
  if FDataLink.Field.IsNull or (FPostponeSaveNewRecord and (FDataLink.DataSet.State = dsInsert)) then
    FPostponedSaving:=true
  else
    DoSaveItemValues;
end;

function TrDBCheckListBox.CheckedCount: integer;
begin
  Result:=FAllItems.CheckedCount;
end;

procedure TrDBCheckListBox.FilterItemsList;
var
  I, idx, chcount: integer;
  sh: boolean;
begin
  Items.BeginUpdate;
  try
    // load visible items
    Items.Clear;
    for I:=0 to FAllItems.Count - 1 do
    begin
      sh:=(FFindItemsStyle <> fisFilter) or (FFindEdit.Text = '') or (rtool.AnsiPosEx(FFindEdit.Text, FAllItems.CheckItem(I).Text, 1, false) > 0);
      if sh then
      begin
        idx:=Items.AddObject(FAllItems.CheckItem(I).Text, TObject(FAllItems.CheckItem(I).Id));
        Checked[idx]:=FAllItems.CheckItem(I).Checked;
      end;
    end;

    // reorder visible items
    if FCheckItemsStyle <> chsAllItems then
    begin
      I:=0;
      chcount:=0;
      while I < Items.Count do
      begin
        if FCheckItemsStyle = chsCheckedFirst then
        begin
          if Checked[I] then
          begin
            Items.Move(I, chcount);
            Inc(chcount);
          end;
          Inc(I);
        end
        else
        begin
          if Checked[I] then
            Inc(I)
          else
            Items.Delete(I);
        end;
      end;
    end;
  finally
    Items.EndUpdate;
  end;
end;

procedure TrDBCheckListBox.SelectItemsList;
var
  I, From: integer;
begin
  if FFindEdit.Text <> '' then
  begin
    if oldFindText = FFindEdit.Text then
    begin
      // start from current position
      From:=ItemIndex;
    end
    else
    begin
      // start from beggining
      oldFindText:=FFindEdit.Text;
      From:=-1;
    end;
    // find next item
    I:=From + 1;
    repeat
      if (I < Items.Count) and (rtool.AnsiPosEx(FFindEdit.Text, Items[I], 1, false) > 0) then
      begin
        ItemIndex:=I;
        Click;
        Break;
      end;
      Inc(I);
      if I >= Items.Count then
        I:=0;
      if (I = From) or ((From < 0) and (I = 0)) then
        Break;
    until false;
  end;
end;

procedure TrDBCheckListBox.ChangeFindText(Sender: TObject);
begin
  case FFindItemsStyle of
    fisNone:
      ;
    fisFilter:
      FilterItemsList;
    fisSelect:
      SelectItemsList;
  end;
end;

procedure TrDBCheckListBox.SetItemState(AItemIndex: integer; AChecked: boolean);
begin
  ItemIndex:=AItemIndex;
  Checked[ItemIndex]:=AChecked;
  ClickCheck;
end;

procedure TrDBCheckListBox.SetAllItemsState(AChecked: boolean);
var
  I: integer;
begin
  for I:=0 to Items.Count - 1 do
    SetItemState(I, AChecked);
end;

procedure TrDBCheckListBox.StoreItemsForDuplication;
begin
  FAllItems.StoreItemsForDuplication;
end;

procedure TrDBCheckListBox.RestoreDuplicatedItems;
begin
  if IsEditing(DataSource) then
  begin
    FDataLink.Modified;
    FAllItems.RestoreDuplicatedItems;
    FilterItemsList;
    if FFindItemsStyle = fisSelect then
      SelectItemsList;
  end;
end;

function TrDBCheckListBox.GetCheckedItemsList(Separator: string; UseItemObjectAsInt: boolean = false): string;
var
  I: integer;
  s: string;
begin
  Result:='';
  for I:=0 to Items.Count - 1 do
    if Self.Checked[I] then
    begin
      if UseItemObjectAsInt then
        s:=IntToStr(integer(Items.Objects[I]))
      else
        s:=Items[I];
      rtool.AddToSepList(Result, s, Separator);
      { if Result <> '' then
        Result:=Result + Separator;
        Result:=Result + Items[I];
        }
    end;
end;

// ==============================================================================
// TrDBCheckListBoxEx
constructor TrDBCheckListBoxEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FListLink:=TDataLink.Create;
  FValuesLink:=TDataLink.Create;
end;

destructor TrDBCheckListBoxEx.Destroy;
begin
  FreeAndNil(FListLink);
  FreeAndNil(FValuesLink);
  inherited Destroy;
end;

function TrDBCheckListBoxEx.GetListSource: TDataSource;
begin
  Result:=FListLink.DataSource;
end;

procedure TrDBCheckListBoxEx.SetListSource(Value: TDataSource);
begin
  if not(FListLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FListLink.DataSource:=Value;
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

function TrDBCheckListBoxEx.GetValuesSource: TDataSource;
begin
  Result:=FValuesLink.DataSource;
end;

procedure TrDBCheckListBoxEx.SetValuesSource(Value: TDataSource);
begin
  if not(FValuesLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FValuesLink.DataSource:=Value;
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TrDBCheckListBoxEx.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FListLink <> nil) and (AComponent = ListSource) then
      ListSource:=nil;
    if (FValuesLink <> nil) and (AComponent = ValuesSource) then
      ValuesSource:=nil;
  end;
end;

procedure TrDBCheckListBoxEx.DoLoadItemList;
var
  Id: integer;
  v: string;
begin
  FAllItems.Clear;
  Items.Clear;
  if not FListLink.Active then
    Exit;
  if FListLink.DataSet.FindField(FListKeyFieldName) = nil then
    DatabaseError(Format('Field "%s" not found in ListSource.', [FListKeyFieldName]), Self);
  if not(FListLink.DataSet.FindField(FListKeyFieldName) is TIntegerField) then
    DatabaseError(Format('ListKeyField "%s" must be integer field.', [FListKeyFieldName]), Self);
  // if FListLink.DataSet.FindField(FListDisplayFieldName) = nil then
  // DatabaseError(Format('Field "%s" not found in ListSource.', [FListDisplayFieldName]), Self);
  FListLink.DataSet.First;
  while not FListLink.DataSet.Eof do
  begin
    Id:=FListLink.DataSet.FieldByName(FListKeyFieldName).AsInteger;
    v:=GetFldListValues(FListLink.DataSet, FListDisplayFieldName);
    FAllItems.AddCheckItem(Id, v, false);
    FListLink.DataSet.Next;
  end;
  DoLoadItemValues;
end;

procedure TrDBCheckListBoxEx.DoLoadItemValues;
begin
  if not FValuesLink.Active then
    Exit;
  if FValuesLink.DataSet.FindField(FValuesDataKeyField) = nil then
    DatabaseError(Format('Field "%s" not found in ValuesSource.', [FValuesDataKeyField]), Self);
  if FValuesLink.DataSet.FindField(FValuesListKeyField) = nil then
    DatabaseError(Format('Field "%s" not found in ValuesSource.', [FValuesListKeyField]), Self);
  if not(FValuesLink.DataSet.FindField(FValuesDataKeyField) is TIntegerField) then
    DatabaseError(Format('FValuesDataKey "%s" must be integer field.', [FValuesDataKeyField]), Self);
  if not(FValuesLink.DataSet.FindField(FValuesListKeyField) is TIntegerField) then
    DatabaseError(Format('FValuesListKey "%s" must be integer field.', [FValuesListKeyField]), Self);

  if not FPostponedSaving then
  begin
    if Assigned(FOnLoadItemValues) then
      FOnLoadItemValues(Self, FAllItems)
    else
    begin
      FValuesLink.DataSet.Filter:=Format('%s=%d', [FValuesDataKeyField, Self.FDataLink.Field.AsInteger]);
      FValuesLink.DataSet.Filtered:=true;
    end;

    FAllItems.ResetItemsChecked;
    FValuesLink.DataSet.First;
    while not FValuesLink.DataSet.Eof do
    begin
      FAllItems.SetItemChecked(FValuesLink.DataSet.FieldByName(FValuesListKeyField).AsInteger, true);
      FValuesLink.DataSet.Next;
    end;
    FilterItemsList;
    if FFindItemsStyle = fisSelect then
      SelectItemsList;
  end;
end;

procedure TrDBCheckListBoxEx.DoSaveItemValues;
var
  I, Id: integer;
begin
  if not FValuesLink.Active then
    DatabaseError('ValuesSource is not active.', Self);
  for I:=0 to FAllItems.Count - 1 do
  begin
    Id:=FAllItems.CheckItem(I).Id;
    if FAllItems.CheckItem(I).Checked then
    begin
      if not LocateRecord(FValuesLink.DataSet.FieldByName(FValuesListKeyField), Id) then
      begin
        FValuesLink.DataSet.Append;
        FValuesLink.DataSet.FieldByName(FValuesDataKeyField).AsInteger:=FDataLink.Field.AsInteger;
        FValuesLink.DataSet.FieldByName(FValuesListKeyField).AsInteger:=Id;
        FValuesLink.DataSet.Post;
      end;
    end
    else
    begin
      if LocateRecord(FValuesLink.DataSet.FieldByName(FValuesListKeyField), Id) then
        FValuesLink.DataSet.Delete;
    end;
  end;
  inherited;
end;

// ==============================================================================
// TrDBBlob
constructor TChildLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
  Parent:=AOwner as TrDBBlob;
  Alignment:=taCenter;
  Layout:=tlCenter;
end;

function TChildLabel.GetLabelText: string;
begin
  if Visible then
    Result:= inherited GetLabelText
  else
    Result:='';
end;

constructor TChildBtn.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetSubComponent(true);
  Parent:=AOwner as TrDBBlob;
end;

constructor TrDBBlob.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  BevelOuter:=bvLowered;
  Width:=150;
  Height:=48;
  Caption:=' ';
  FFileFilter:=sFileTypeAll;
  FFileSizeLimitMB:=0;
  FShowProgressWindow:=true;

  FDataLinkBlob:=TFieldDataLink.Create;
  FDataLinkBlob.Control:=Self;
  FDataLinkInfo:=TFieldDataLink.Create;
  FDataLinkInfo.Control:=Self;
  FDataLinkInfo.OnActiveChange:=DataChange;
  FDataLinkInfo.OnDataChange:=DataChange;
  FDataLinkInfo.OnEditingChange:=nil;
  FDataLinkInfo.OnUpdateData:=DataChange;

  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);

  FInfoLabel:=TChildLabel.Create(Self);
  FInfoLabel.FreeNotification(Self);
  FInfoLabel.Name:='TitleLabel';
  FInfoLabel.Caption:='DataFieldInfo';
  FInfoLabel.Align:=alTop;
  FInfoLabel.AlignWithMargins:=true;
  FInfoLabel.EllipsisPosition:=epEndEllipsis;

  FBtnOpen:=TChildBtn.Create(Self);
  FBtnOpen.Name:='BtnOpen';
  FBtnOpen.Tag:=cBlobOpen;
  FBtnOpen.Caption:=sBlobOpen;
  FBtnOpen.OnClick:=Self.ClickBoundButtons;
  FBtnLoad:=TChildBtn.Create(Self);
  FBtnLoad.Name:='BtnLoad';
  FBtnLoad.Tag:=cBlobLoad;
  FBtnLoad.Caption:=sBlobLoad;
  FBtnLoad.OnClick:=Self.ClickBoundButtons;
  FBtnSave:=TChildBtn.Create(Self);
  FBtnSave.Name:='BtnSave';
  FBtnSave.Tag:=cBlobSave;
  FBtnSave.Caption:=sBlobSave;
  FBtnSave.OnClick:=Self.ClickBoundButtons;
  FBtnClear:=TChildBtn.Create(Self);
  FBtnClear.Name:='BtnClear';
  FBtnClear.Tag:=cBlobClear;
  FBtnClear.Caption:=sBlobClear;
  FBtnClear.OnClick:=Self.ClickBoundButtons;

  FPopupMenu:=TPopupMenu.Create(Self);
  FPopupMenu.Name:='BlobPopup';
  FPopupMenu.OnPopup:=OnPopupMenu;
  FPopupMenu.AutoHotkeys:=maManual;
  FInfoLabel.PopupMenu:=FPopupMenu;
  // popup items - info, size and resize
  FPopupMenu.Items.Add(CreateMenuItem(Self, '', cBlobInfo, ClickBoundButtons));
  FPopupMenu.Items.Add(CreateMenuItem(Self, '-', cBlobNone, ClickBoundButtons));
  FPopupMenu.Items.Add(CreateMenuItem(Self, sBlobRename, cBlobRename, ClickBoundButtons));

  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBBlob.Destroy;
begin
  FreeAndNil(FDataLinkBlob);
  FreeAndNil(FDataLinkInfo);
  inherited Destroy;
end;

procedure TrDBBlob.Loaded;
begin
  inherited Loaded;
  SetDataSourceBlob(DataSourceBlob);
  FDBEditLabel.StoreVisible;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBBlob.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (FDataLinkBlob <> nil) and (AComponent = DataSourceBlob) then
      DataSourceBlob:=nil;
    if (FDataLinkInfo <> nil) and (AComponent = DataSourceInfo) then
      DataSourceInfo:=nil;
    if (AComponent = FDBEditLabel) then
      FDBEditLabel:=nil;
    if (AComponent = FInfoLabel) then
      FInfoLabel:=nil;
  end;
end;

procedure TrDBBlob.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
  RefBoundButtons(Self);
end;

procedure TrDBBlob.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
end;

procedure TrDBBlob.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
  if Assigned(FInfoLabel) then
    FInfoLabel.Enabled:=Enabled;
end;

procedure TrDBBlob.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
  RefBoundButtons(Self);
end;

function TrDBBlob.GetDataSourceBlob: TDataSource;
begin
  Result:=FDataLinkBlob.DataSource;
end;

procedure TrDBBlob.SetDataSourceBlob(Value: TDataSource);
begin
  if not(FDataLinkBlob.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLinkBlob.DataSource:=Value;
    DataChange(nil);
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
  if (csDesigning in ComponentState) then
    Exit;
  SetVisibilityFromField(Self, DataSourceBlob, DataFieldBlob);
end;

function TrDBBlob.GetDataFieldBlob: string;
begin
  Result:=FDataLinkBlob.FieldName;
end;

procedure TrDBBlob.SetDataFieldBlob(const Value: string);
begin
  FDataLinkBlob.FieldName:=Value;
  DataChange(nil);
  SetDBCompName(Self, FDefaultPrefix, DataFieldBlob);
end;

procedure TrDBBlob.SetDefaultPrefix(const Value: string);
begin
  FDefaultPrefix:=Value;
  SetDBCompName(Self, FDefaultPrefix, DataFieldBlob);
end;

function TrDBBlob.GetDataSourceInfo: TDataSource;
begin
  Result:=FDataLinkInfo.DataSource;
end;

procedure TrDBBlob.SetDataSourceInfo(Value: TDataSource);
begin
  if not(FDataLinkInfo.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    FDataLinkInfo.DataSource:=Value;
    DataChange(nil);
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

function TrDBBlob.GetDataFieldInfo: string;
begin
  Result:=FDataLinkInfo.FieldName;
end;

procedure TrDBBlob.SetDataFieldInfo(const Value: string);
begin
  FDataLinkInfo.FieldName:=Value;
  DataChange(nil);
end;

procedure TrDBBlob.SetBtnMargin(Value: integer);
begin
  FBtnMargin:=Value;
  RefBoundButtons(nil);
end;

procedure TrDBBlob.SetBtnSpace(Value: integer);
begin
  FBtnSpace:=Value;
  RefBoundButtons(nil);
end;

procedure TrDBBlob.RefBoundButtons(Sender: TObject);
var
  px, py, W, H: integer;
begin
  if Assigned(FBtnOpen) and Assigned(FBtnLoad) and Assigned(FBtnSave) and Assigned(FBtnClear) then
  begin
    W:=(Width - 2 * FBtnMargin - 3 * FBtnSpace) div 4;
    px:=FBtnMargin;
    py:=FInfoLabel.Top + FInfoLabel.Height + FInfoLabel.Margins.Bottom + FBtnMargin;
    H:=Height - py - FBtnMargin;
    FBtnOpen.SetBounds(px, py, W, H);
    px:=px + W + FBtnSpace;
    FBtnLoad.SetBounds(px, py, W, H);
    px:=px + W + FBtnSpace;
    FBtnSave.SetBounds(px, py, W, H);
    px:=px + W + FBtnSpace;
    FBtnClear.SetBounds(px, py, W, H);
  end;
end;

function TrDBBlob.IsLinked: boolean;
begin
  if csDesigning in ComponentState then
    Result:=true
  else
    Result:=Assigned(FDataLinkInfo.Field) and Assigned(FDataLinkBlob.Field);
end;

function TrDBBlob.CanEdit: boolean;
begin
  Result:=false;
  if not IsLinked then
    Exit;
  if FDataLinkInfo.Field.ReadOnly or FDataLinkBlob.Field.ReadOnly then
    Exit;
  if IsEditing(DataSourceInfo) then
    Result:=true
  else
    if DataSourceInfo.AutoEdit then
      try
        DataSourceInfo.Edit;
        Result:=true;
      except
      end;
end;

procedure TrDBBlob.DataChange(Sender: TObject);
begin
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSourceBlob, DataFieldBlob);
  if not IsLinked then
  begin
    BtnOpen.Enabled:=false;
    BtnLoad.Enabled:=false;
    BtnSave.Enabled:=false;
    BtnClear.Enabled:=false;
    GetPopupItem(cBlobRename).Enabled:=false;
    Exit;
  end;
  if Assigned(FDataLinkInfo) and Assigned(FDataLinkInfo.Field) then
  begin
    if FDataLinkInfo.Field.IsNull then
      FInfoLabel.Caption:=sPictureEmpty
    else
      FInfoLabel.Caption:=FDataLinkInfo.Field.AsString;
    BtnOpen.Enabled:=not FDataLinkInfo.Field.IsNull;
    BtnLoad.Enabled:=true;
    BtnSave.Enabled:=BtnOpen.Enabled;
    BtnClear.Enabled:=BtnOpen.Enabled;
    GetPopupItem(cBlobRename).Enabled:=BtnOpen.Enabled;
  end
  else
    if DataFieldInfo <> '' then
      FInfoLabel.Caption:=DataFieldInfo
    else
      if (csDesigning in ComponentState) then
        FInfoLabel.Caption:='DataFieldInfo'
      else
        FInfoLabel.Caption:='';
  FInfoLabel.Hint:='';
  if FInfoLabel.Canvas.TextWidth(FInfoLabel.Caption) > FInfoLabel.Width then
    FInfoLabel.Hint:=FInfoLabel.Caption;
end;

procedure TrDBBlob.ClickBoundButtons(Sender: TObject);
begin
  case TComponent(Sender).Tag of
    cBlobOpen:
      OpenBlob;
    cBlobLoad:
      LoadFromFile('');
    cBlobSave:
      SaveAsFile('');
    cBlobClear:
      ClearBlob;
    cBlobRename:
      RenameBlob;
  end;
end;

procedure TrDBBlob.OpenBlob;
var
  fn, param: string;
begin
  if not IsLinked then
    Exit;
  fn:=FOpenTempDir;
  if fn = '' then
    fn:=GetTempDir;
  fn:=rtool.SetEndOfPath(fn) + FDataLinkInfo.Field.AsString;
  if not SaveAsFile(fn) then
    Exit;
  if Assigned(FOnOpenBlob) then
    FOnOpenBlob(Self, fn, rfiletool.GetFileSize(fn));
  if FOpenByApp = '' then
    RunFileSimple(fn)
  else
  begin
    param:=Format(FOpenParams, [fn]);
    RunFile(FOpenByApp, param, '', SW_SHOW);
  end;
end;

procedure TrDBBlob.ClearBlob;
begin
  if CanEdit then
  begin
    if Assigned(FPrepareBlobDataset) then
      FPrepareBlobDataset(Self);
    if not FDataLinkBlob.Editing then
      if not FDataLinkBlob.Edit then
        Abort;
    FDataLinkBlob.Field.Clear;
    FDataLinkInfo.Field.Clear;
  end;
end;

procedure TrDBBlob.RenameBlob;
var
  fn, ext: string;
begin
  if CanEdit then
  begin
    fn:=FDataLinkInfo.Field.AsString;
    ext:=ExtractFileExt(fn);
    if rdlg.InputStr(sInput, sBlobRename + ':', fn) then
    begin
      if ExtractFileExt(fn) = '' then
        fn:=fn + ext;
      if not SameText(ext, ExtractFileExt(fn)) then
        if DlgQ(sBlobChangeExtQ) <> mrYes then
          Exit;
      FDataLinkInfo.Field.AsString:=fn;
    end;
  end;
end;

function TrDBBlob.SaveAsFile(const FileName: string; ProgressInfo: string = ''; InfoStayVisible: boolean = false): boolean;
const
  CopySize = 512000;
var
  D: TrSaveDialog;
  bs: TStream;
  fs: TFileStream;
  Copied: Int64;
  s: string;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  Result:=false;
  if not IsLinked then
    Exit;
  D:=TrSaveDialog.Create(nil);
  try
    D.Filter:=FFileFilter;
    D.FilterIndex:=0;
    D.FileName:=FileName;
    try
      if D.FileName = '' then
      begin
        // select file
        D.FileName:=FDataLinkInfo.Field.AsString;
        if not D.Execute then
          Exit;
      end;
      WCursor;
      if Assigned(FPrepareBlobDataset) then
        FPrepareBlobDataset(Self);
      // save blob to file
      bs:=FDataLinkBlob.DataSet.CreateBlobStream(FDataLinkBlob.Field, bmRead);
      try
        fs:=TFileStream.Create(D.FileName, fmCreate);
        try
          Aborted:=false;
          repeat
            Copied:=Min(CopySize, bs.Size - bs.Position);
            fs.CopyFrom(bs, Copied);
            if FShowProgressWindow and (bs.Size > 0) then
            begin
              s:=sLoading + sOfData;
              if ProgressInfo <> '' then
                s:=ProgressInfo + #13#10 + s;
              ShowInfoProgressBar(s, nil, OnAbort, Round(100 * bs.Position / bs.Size));
            end;
            Application.ProcessMessages;
            if Aborted then
              Break;
          until Copied < CopySize;
        finally
          FreeAndNil(fs);
          if not InfoStayVisible then
            ShowInfo('');
        end;
      finally
        FreeAndNil(bs);
      end;
      if Aborted then
        DeleteFile(D.FileName)
      else
        Result:=true;
    finally
      NCursor;
    end;
  finally
    FreeAndNil(D);
  end;
end;

function TrDBBlob.LoadFromFile(const FileName: string; ProgressInfo: string = ''; InfoStayVisible: boolean = false): boolean;
const
  CopySize = 512000;
var
  D: TrOpenDialog;
  bs: TStream;
  fs: TFileStream;
  Copied: Int64;
  s: string;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  Result:=false;
  if not CanEdit then
    Exit;
  D:=TrOpenDialog.Create(nil);
  D.Filter:=FFileFilter;
  D.FilterIndex:=0;
  D.FileName:=FileName;
  try
    if D.FileName = '' then
    begin
      // select file
      if not D.Execute then
        Exit;
    end;
    WCursor;
    if Assigned(FPrepareBlobDataset) then
      FPrepareBlobDataset(Self);
    if not FDataLinkBlob.Editing then
      if not FDataLinkBlob.Edit then
        Abort;

    fs:=TFileStream.Create(D.FileName, fmOpenRead or fmShareDenyWrite);
    try
      if Assigned(FOnLoadBlob) then
        FOnLoadBlob(Self, D.FileName, fs.Size)
      else
        if (FFileSizeLimitMB > 0) and (fs.Size > FFileSizeLimitMB * 1000000) then
          if DlgQ(Format(sBlobFileOverLimitQ, [fs.Size / 1000000, FFileSizeLimitMB]) + cStrEOL + SContinueQ) <> mrYes then
            Exit;
      // load file to blob
      bs:=FDataLinkBlob.DataSet.CreateBlobStream(FDataLinkBlob.Field, bmWrite);
      try
        Aborted:=false;
        repeat
          Copied:=Min(CopySize, fs.Size - fs.Position);
          bs.CopyFrom(fs, Copied);
          if FShowProgressWindow and (fs.Size > 1000) then
          begin
            s:=sLoading + sOfData;
            if ProgressInfo <> '' then
              s:=ProgressInfo + #13#10 + s;
            ShowInfoProgressBar(s, nil, OnAbort, Round(100 * fs.Position / fs.Size));
          end;
          Application.ProcessMessages;
          if Aborted then
            Break;
        until Copied < CopySize;
      finally
        if not InfoStayVisible then
          ShowInfo('');
        FreeAndNil(bs);
      end;
    finally
      FreeAndNil(fs);
    end;

    if Aborted then
    begin
      FDataLinkBlob.Field.Clear;
      FDataLinkInfo.Field.Clear;
    end
    else
    begin
      // update file info
      FDataLinkInfo.Field.AsString:=ExtractFileName(D.FileName);
      Result:=true;
    end;
    DataChange(nil);
  finally
    FreeAndNil(D);
    NCursor;
  end;
end;

function TrDBBlob.BlobInfo: string;
var
  fs: Int64;
  fn, desc: string;
  bs: TStream;
begin
  Result:=sPictureEmpty;
  if not IsLinked then
    Exit;
  fn:=FDataLinkInfo.Field.AsString;
  if fn = '' then
    Exit;
  fn:=ExtractFileExt(fn);
  desc:=GetGenericFileType(fn);
  fn:=UpperCase(copy(fn, 2, 255));

  WCursor;
  try
    if Assigned(FPrepareBlobDataset) then
      FPrepareBlobDataset(Self);
    bs:=FDataLinkBlob.DataSet.CreateBlobStream(FDataLinkBlob.Field, bmRead);
    try
      fs:=bs.Size;
    finally
      FreeAndNil(bs);
    end;
    Result:=Format('%s  (%f MB)', [desc, fs / 1000000]);
  finally
    NCursor;
  end;
end;

procedure TrDBBlob.OnAbort(Sender: TObject);
begin
  Aborted:=true;
end;

function TrDBBlob.GetPopupItem(Tag: integer): TMenuItem;
var
  I: integer;
begin
  Result:=nil;
  for I:=0 to FPopupMenu.Items.Count - 1 do
    if FPopupMenu.Items[I].Tag = Tag then
    begin
      Result:=FPopupMenu.Items[I];
      Break;
    end;
end;

procedure TrDBBlob.OnPopupMenu(Sender: TObject);
begin
  GetPopupItem(cBlobInfo).Caption:=BlobInfo;
end;

// ==============================================================================
// TrDBRecordSelection
constructor TrDBRecordSelection.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataLink:=TrDataLink.Create(ActiveChange, EditingChange, DataChange);
  FDBEditLabel:=TDBBoundLabel.Create(Self);
  FDBEditLabel.FreeNotification(Self);
  FDBEditLabel.ShowRequiredMark:=false;
  // Style:=csDropDownList;
  AutoDropDown:=true;
  DropDownCount:=15;
  FLoadListEvent:=leAfterOpen;
  FEditingState:=esDisable;
  FIsLoadingItems:=false;
  FItemsLoaded:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBRecordSelection.Destroy;
begin
  FreeAndNil(FDataLink);
  inherited Destroy;
end;

procedure TrDBRecordSelection.Loaded;
begin
  inherited Loaded;
  ActiveChange(nil);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, FRecIdField);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBRecordSelection.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Parent:=AParent;
end;

procedure TrDBRecordSelection.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (AComponent = FDBEditLabel) then
      FDBEditLabel:=nil;
  end;
end;

procedure TrDBRecordSelection.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.Enabled:=Enabled;
end;

procedure TrDBRecordSelection.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  case ord(Key) of
    VK_RETURN, VK_ESCAPE:
      LoadActiveRec;
  end;
end;

procedure TrDBRecordSelection.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.ChangeVisible(Visible);
end;

procedure TrDBRecordSelection.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FDBEditLabel) then
    FDBEditLabel.RefreshPosition;
end;

procedure TrDBRecordSelection.DropDown;
begin
  if ((FLoadListEvent = leAfterDropDownFirst) and not FItemsLoaded) or (FLoadListEvent = leAfterDropDownAlways) then
  begin
    LoadList;
    LoadActiveRec;
  end;
  inherited DropDown;
end;

procedure TrDBRecordSelection.Click;
var
  oldId: integer;
begin
  try
    if FDataLink.Editing then
    begin
      // store new selected RecID (ItemIndex is changed after post)
      if ItemIndex >= 0 then
        oldId:=integer(Items.Objects[ItemIndex])
      else
        oldId:=0;
      case FEditingState of
        esAskBeforeMove:
          begin
            case ShowDlg(dtQ, btYesNoCancel, [sSaveChangesQ]) of
              mrYes:
                FDataLink.DataSet.Post;
              mrNo:
                FDataLink.DataSet.Cancel;
            else
              Abort;
            end;
          end;
        esPostBeforeMove:
          FDataLink.DataSet.Post;
        esReadOnly:
          Abort;
      end;
      // restore selected RecID
      ItemIndex:=Items.IndexOfObject(TObject(oldId));
    end;
    // try to change record
    SetActiveRec;
  finally
    // load active record (needed in case that scroll was aborted)
    LoadActiveRec;
  end;
  inherited Click;
end;

function TrDBRecordSelection.GetDataSource: TDataSource;
begin
  Result:=FDataLink.DataSource;
end;

procedure TrDBRecordSelection.SetDataSource(Value: TDataSource);
begin
  if not(FDataLink.DataSourceFixed and (csLoading in ComponentState)) then
  begin
    if Assigned(Value) then
      if Assigned(Value.DataSet) then
        if Value.DataSet.IsUnidirectional then
          DatabaseError(SDataSetUnidirectional);
    FDataLink.DataSource:=Value;
    ActiveChange(nil);
    if Assigned(FDBEditLabel) then
      FDBEditLabel.SetField(DataSource, FRecIdField);
  end;
  if Value <> nil then
    Value.FreeNotification(Self);
end;

procedure TrDBRecordSelection.ActiveChange(Sender: TObject);
begin
  Enabled:=FDataLink.Active;
  if FLoadListEvent = leAfterOpen then
    LoadList;
  if Enabled then
    LoadActiveRec;
end;

procedure TrDBRecordSelection.DataChange(Sender: TObject);
begin
  if not FIsLoadingItems then
    LoadActiveRec;
end;

procedure TrDBRecordSelection.EditingChange(Sender: TObject);
begin
  if FEditingState = esDisable then
    Enabled:=not FDataLink.Editing;
end;

procedure TrDBRecordSelection.SetRecIdField(const Value: string);
begin
  FRecIdField:=Value;
  if Assigned(FDBEditLabel) then
    FDBEditLabel.SetField(DataSource, FRecIdField);
end;

procedure TrDBRecordSelection.SetListFields(const Value: string);
begin
  FListFields:=Value;
end;

function TrDBRecordSelection.GetRecordListId: integer;
begin
  Result:=0;
  if FRecIdField = '' then
    Exit;
  try
    Result:=FDataLink.DataSet.FieldByName(FRecIdField).AsInteger;
  except
    on e: Exception do
    begin
      ShowErr(Name, e);
      FRecIdField:='';
    end;
  end;
end;

function TrDBRecordSelection.GetRecordListText: string;
const
  StrSep = '"';
  FldSep = ';';
var
  I, J, strcount: integer;
  s, cs: string;
  wasstr: boolean;
begin
  Result:='';
  if FListFields = '' then
    Exit;
  try
    I:=1;
    wasstr:=false;
    cs:='';
    repeat
      // extract part to field separator
      s:=rtool.FindStrPart(FListFields, FldSep, I, true);
      if s = '' then
        Break;
      s:=cs + s;
      // count number of StrSep
      strcount:=0;
      for J:=1 to Length(s) do
        if copy(s, J, 1) = StrSep then
          Inc(strcount);
      if (strcount mod 2) = 0 then
      begin
        // 0 or 2 separators = Fieldname or constatnt
        // reset prev const string
        cs:='';
        if strcount > 0 then
        begin
          // str const used = not use space as default field separator
          wasstr:=true;
          s:=rtool.FindStrFromTo(StrSep, StrSep, s, true, true);
          Result:=Result + s;
        end
        else
        begin
          // add space as default field separator
          if not wasstr and (I > 1) then
            Result:=Result + ' ';
          Result:=Result + FDataLink.DataSet.FieldByName(s).DisplayText;
        end
      end
      else
        // store const string part for next run
        cs:=s + FldSep;
      Inc(I);
    until false;
  except
    on e: Exception do
    begin
      ShowErr(Name, e);
      FListFields:='';
    end;
  end;
end;

function TrDBRecordSelection.LoadList: integer;
var
  L: TStringList;
  Bk: TBookmark;
begin
  Items.Clear;
  Result:=0;
  if Assigned(FDataLink.DataSet) and FDataLink.DataSet.Active then
  begin
    FIsLoadingItems:=true;
    L:=TStringList.Create;
    try
      L.Clear;
      Bk:=FDataLink.DataSet.GetBookmark;
      FDataLink.DataSet.DisableControls;
      try
        FDataLink.DataSet.First;
        while not FDataLink.DataSet.Eof do
        begin
          L.AddObject(GetRecordListText, TObject(GetRecordListId));
          FDataLink.DataSet.Next;
        end;
      finally
        FDataLink.DataSet.GotoBookmark(Bk);
        FDataLink.DataSet.EnableControls;
      end;
      Items.Assign(L);
      Result:=L.Count;
      FItemsLoaded:=true;
    finally
      FreeAndNil(L);
      FIsLoadingItems:=false;
    end;
  end;
end;

procedure TrDBRecordSelection.Resort;
var
  I: integer;
begin
  if Sorted then
  begin
    // resort items
    Sorted:=false;
    Sorted:=true;
  end;
  // select active item
  I:=Items.IndexOfObject(TObject(GetRecordListId));
  if I >= 0 then
    ItemIndex:=I;
end;

procedure TrDBRecordSelection.LoadActiveRec;
var
  I: integer;
  s: string;
begin
  I:=Items.IndexOfObject(TObject(GetRecordListId));
  s:=GetRecordListText;
  if I >= 0 then
  begin
    ItemIndex:=I;
    if Items[I] <> s then
    begin
      // update item text if was changed
      Items[I]:=s;
      Resort;
    end;
  end
  else
    if not FDataLink.Editing then
    begin
      // add new item
      Items.AddObject(s, TObject(GetRecordListId));
      Resort;
    end
    else
      ItemIndex:=I;
  Repaint;
end;

procedure TrDBRecordSelection.SetActiveRec;
var
  Id: integer;
begin
  if ItemIndex < 0 then
    Exit;
  Id:=integer(Items.Objects[ItemIndex]);
  FDataLink.DataSet.Locate(FRecIdField, Id, []);
end;

{
  GetListValuesCB = procedure(Field: TField; List: TStrings) of object;
  procedure CreateDBControls(DataSource: TDataSource; ScrollBox: TScrollBox; OnlyOneCol, DelExistComps: boolean; GetListValuesCallBack: GetListValuesCB);

  // ==============================================================================
  // function for creating components in forms (in ScrollBox)
  // according Datasource and Tag value of fields
  // set Tag=2 for showing components, Tag=4 for showing in new line
  procedure CreateDBControls(DataSource: TDataSource; ScrollBox: TScrollBox; OnlyOneCol, DelExistComps: boolean; GetListValuesCallBack: GetListValuesCB);
  const
  lsep = 10;
  rsep = 20;
  wmultiple = 8;

  var
  i, X, Y, dy, temp: integer;
  F: TField;
  e: TrDBEdit;
  B: TrDBCheckBox;
  M: TrDBMemo;
  cb: TrDBComboBox;
  lcb: TrDBLookupComboBox;
  dt: TrDBDateTimePicker;
  L: TStrings;

  procedure SetControl(C: TControl; newline: boolean; dleft, dtop, sumheight: integer);
  begin
  C.Parent:=ScrollBox;
  // limit width to parent width
  if C.Width > ScrollBox.ClientWidth - lsep - rsep - dleft then
  C.Width:=ScrollBox.ClientWidth - lsep - rsep - dleft;
  // new line?
  if newline or OnlyOneCol or (F.Tag and 4 <> 0) or (ScrollBox.ClientWidth - rsep - X < C.Width + dleft) then
  begin
  // move line
  Y:=Y + dy;
  dy:=0;
  // set left
  X:=lsep;
  end;
  C.Left:=X + dleft;
  X:=X + dleft + C.Width + lsep;
  C.Top:=Y + dtop;
  if dy < sumheight then
  dy:=sumheight;
  end;

  begin
  if DelExistComps then
  while ScrollBox.ControlCount > 0 do
  ScrollBox.Controls[0].Free;

  L:=TStringList.Create;
  try
  X:=lsep;
  Y:=12;
  dy:=0;
  for i:=0 to DataSource.DataSet.FieldCount - 1 do
  begin
  F:=DataSource.DataSet.Fields[i];
  if F.Tag and 2 = 0 then
  Continue;

  L.Clear;
  if Assigned(GetListValuesCallBack) then
  GetListValuesCallBack(F, L);
  if L.Count > 0 then
  begin
  cb:=TrDBComboBox.Create(ScrollBox.Owner);
  cb.Width:=F.DisplayWidth * wmultiple;
  SetControl(cb, false, 0, 14, 38);
  cb.Items.Assign(L);
  cb.SetDataSource(DataSource, F.FieldName);
  cb.DropDownCount:=15;
  end
  else
  if F.FieldKind = fkLookup then
  begin
  lcb:=TrDBLookupComboBox.Create(ScrollBox.Owner);
  lcb.Width:=F.DisplayWidth * wmultiple;
  SetControl(lcb, false, 0, 14, 38);
  lcb.SetDataSource(DataSource, F.FieldName);
  lcb.DropDownRows:=15;
  end
  else
  if F is TMemoField then
  begin
  M:=TrDBMemo.Create(ScrollBox.Owner);
  M.Width:=600;
  M.Height:=48;
  if i = DataSource.DataSet.FieldCount - 1 then
  M.Height:=ScrollBox.Height - Y - dy - 24;
  if M.Height < 48 then
  M.Height:=48;
  SetControl(M, true, 0, 14, 68);
  M.SetDataSource(DataSource, F.FieldName);
  end
  else
  if F is TBooleanField then
  begin
  B:=TrDBCheckBox.Create(ScrollBox.Owner);
  B.Caption:=F.DisplayLabel;
  B.Width:=Length(B.Caption) * 5 + 25;
  SetControl(B, false, 0, 4, 24);
  B.SetDataSource(DataSource);
  B.DataField:=F.FieldName;
  end
  else
  if IsDateTimeField(F) then
  begin
  dt:=TrDBDateTimePicker.Create(ScrollBox.Owner);
  if F.Required then
  temp:=0
  else
  temp:=DBCheckBoxWidth;
  SetControl(dt, false, temp, 14, 38);
  dt.DataAllowNull:=not F.Required;
  dt.DataSource:=DataSource;
  dt.DataField:=F.FieldName;
  end
  else
  begin
  e:=TrDBEdit.Create(ScrollBox.Owner);
  e.Width:=F.DisplayWidth * wmultiple;
  SetControl(e, false, 0, 14, 38);
  e.SetDataSource(DataSource, F.FieldName);
  end;
  end;
  finally
  L.Free;
  end;
  end;
  }
end.
