unit rDBGrid;

{<
  rDBGrid unit contain TrDBGrid - enhanced TDBGrid component which provides many new functions
  (show checkboxes, graphics and memo fields, build-in editors,
  pictures in title row or in data cells, multiple line titles and rows,
  sorting functions, fiter buttons, grouping fields etc).

  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2006)
  @lastmod(Feb 2021)
}

interface

{$INCLUDE _rosi.inc}
// FPopupCalendar is used instead of TDateTimePicker if compiler switch TrPopupMonthCalendar is defined
{$DEFINE TrPopupMonthCalendar}
{$DEFINE no_TrDBGridDrawColumnCellOnEnd}

uses
  Windows, Messages, SysUtils, Graphics, Classes, Controls, Menus, ImgList,
  Grids, DBGrids, DB, DBCtrls, Forms, Dialogs, StdCtrls, ComCtrls, ExtCtrls,
  Types, Math, IniFiles, StrUtils, Clipbrd, Variants, GraphUtil, DateUtils,
{$IFDEF DXE2P} UITypes, {$ENDIF} Themes, Contnrs,
  PropSaveMain, rImprovedComps, rDBComponents, RosiCompGlobalCfg;

const
  GroupHeaderSep = '::';
  FieldListSep = ';';
  DataImagesFieldCfg = 'Cfg_Align';

  { constants used by BeforeScanDataset and AfterScanDataset events to indicate reason of dataset scan }
  grrColumnWidth = 1;
  grrSummaryCalc = 2;
  grrQuickLookup = 3;
  grrLoadPickList = 4;
  grrCopyData = 5;
  grrLoadFilterList = 6;

  { constans used as TAG for summary text in title content popup and for footer row }
  cSummaryTag_Prefix = '%';
  cSummaryTag_Sum = 'SUM';
  cSummaryTag_Min = 'MIN';
  cSummaryTag_Max = 'MAX';
  cSummaryTag_AVG = 'AVG';
  cSummaryTag_CountVal = 'COUNTVAL';
  cSummaryTag_CountNull = 'COUNTNULL';
  cSummaryTag_CountAll = 'COUNTALL';
  cSummaryTag_SelSum = 'SELSUM';
  cSummaryTag_SelMin = 'SELMIN';
  cSummaryTag_SelMax = 'SELMAX';
  cSummaryTag_SelAVG = 'SELAVG';
  cSummaryTag_SelCountVal = 'SELCOUNTVAL';
  cSummaryTag_SelCountNull = 'SELCOUNTNULL';
  cSummaryTag_SelCountAll = 'SELCOUNTALL';
  cSummaryTag_DistinctValues = 'DISTINCTVALUES';
  cSummaryTag_RecNo = 'RECNO';
  cSummary_DistinctValuesMax = 25;

type
  TrDBGrid = class;
  TrDBGridCustomSorter = class;
{$IFDEF UseObsoleteOption}
  { list of OptionEx properties, deprecated, use TOptionsEx2 instead }
  TOptionEx = (dgTitleExBtn, dgTitleBtn, dgFixColBtn, dgBool, dgMemo, dgMemoShowText, dgGraphic, dgGraphicShowImg, dgDateTimePicker, dgDrawRowFocus,
    dgDrawColoredRow, dgOneClickCheckbox, dgAutoSaveCheckBox, dgHideHScrollBar, dgNotProcDblClick, dgNotProcMemoDblClick, dgNotProcGraphicDblClick,
    dgOnlyPickListValues, dgAllowReadOnlyButton, dgNotHighlightMultiselect, dgStandardTabKey, dgDragCell, dgUseTitlePopup, dgAutoPickListWidth,
    dgUseCellParamsConditions, dgFilterButtons, dgMultiFieldFilter, dgFieldGroups, dgAlwaysUseFieldGroupColors, dgAllowPasteMultipleCells, dgRotateTitleText,
    dgBlockInsertByKey);
  { set of OptionEx properties }
  TOptionsEx = set of TOptionEx;
{$ENDIF}

  { setting for TrDBGrid fixed cells, part of TOptionsEx2 }
  TOptionsFixedCell = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FTitleExBtn: boolean;
    FTitleExBtnSelectAll: boolean;
    FTitleExBtnHintRecCount: boolean;
    FTitleBtn: boolean;
    FFixedColBtn: boolean;
    FTitlePopup: boolean;
    FTitlePopupDistinctValues: boolean;
    FTitleImageSpace: integer;
    FTitleTextMinWidth: integer;
    FTitleTextSeparators: string;
    FTitleForceWrapSeparators: string;
    FTitleTextCaseSeparator: boolean;
    FRotateTitleText: boolean;
    FFixedCols: integer;
    FFixedColColor: TColor;
    FFixedColFont: TFont;
    FFixedColParentFont: boolean;
    procedure SetTitleExBtn(Value: boolean);
    procedure SetTitleBtn(Value: boolean);
    procedure SetFixedColBtn(Value: boolean);
    procedure SetRotateTitleText(Value: boolean);
    procedure SetTitleImageSpace(Value: integer);
    procedure SetTitleTextMinWidth(Value: integer);
    procedure SetTitleTextSeparators(Value: string);
    procedure SetTitleForceWrapSeparators(Value: string);
    procedure SetTitleTextCaseSeparator(Value: boolean);
    procedure SetFixedCols(Value: integer);
    function GetFixedCols: integer;
    procedure SetFixedColColor(Value: TColor);
    procedure SetFixedColFont(Value: TFont);
    procedure SetFixedColParentFont(Value: boolean);
    procedure FontChanged(Sender: TObject);
    function IsFontStored: boolean;
    function IsTitleTextSeparatorsStored: boolean;
  public
    constructor Create(Owner: TrDBGrid);
    destructor Destroy; override;
  published
    { use Cell(0,0) as button }
    property TitleExBtn: boolean read FTitleExBtn write SetTitleExBtn default true;
    { use Cell(0,0) button for selecting of all records (used by rDBGrid_MS) }
    property TitleExBtnSelectAll: boolean read FTitleExBtnSelectAll write FTitleExBtnSelectAll default true;
    { show Recordcount hint for Cell(0,0) }
    property TitleExBtnHintRecCount: boolean read FTitleExBtnHintRecCount write FTitleExBtnHintRecCount default false;
    { use title row as button }
    property TitleBtn: boolean read FTitleBtn write SetTitleBtn default true;
    { use fixed column as button }
    property FixedColBtn: boolean read FFixedColBtn write SetFixedColBtn default true;
    { use summary popup for title row }
    property TitlePopup: boolean read FTitlePopup write FTitlePopup default true;
    { add distinct values to summary popup for title row }
    property TitlePopupDistinctValues: boolean read FTitlePopupDistinctValues write FTitlePopupDistinctValues default false;
    { space between image and text in title row }
    property TitleImageSpace: integer read FTitleImageSpace write SetTitleImageSpace default 3;
    { minimal width for text in title row, then the image and glyph is hidden to get space for text }
    property TitleTextMinWidth: integer read FTitleTextMinWidth write SetTitleTextMinWidth default 16;
    { define list of separators for title row text wrapping }
    property TitleTextSeparators: string read FTitleTextSeparators write SetTitleTextSeparators stored IsTitleTextSeparatorsStored;
    { define list of separators for title row text force wrapping (separator is replaced by end of line) }
    property TitleForceWrapSeparators: string read FTitleForceWrapSeparators write SetTitleForceWrapSeparators;
    { define if Case letter is used as separator for title row text wrapping }
    property TitleTextCaseSeparator: boolean read FTitleTextCaseSeparator write SetTitleTextCaseSeparator default true;
    { text in title row is rotated 90 }
    property RotateTitleText: boolean read FRotateTitleText write SetRotateTitleText default false;
    { number of fixed columns (besides indicator) }
    property FixedCols: integer read GetFixedCols write SetFixedCols default 0;
    { color of fixed columns }
    property FixedColColor: TColor read FFixedColColor write SetFixedColColor default clNone;
    { font of fixed columns }
    property FixedColFont: TFont read FFixedColFont write SetFixedColFont stored IsFontStored;
    { parent font of fixed columns }
    property FixedColParentFont: boolean read FFixedColParentFont write SetFixedColParentFont default true;
  end;

  { blob content appearance type for TrDBGrid - standard, glyph or content }
  TOptionsBlobType = (obtNone, obtGlyph, obtContent);

  { setting for TrDBGrid editors, part of TOptionsEx2 }
  TOptionsEditors = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FBoolGlyphs: boolean;
    FBoolOneClickChange: boolean;
    FBoolAutoSave: boolean;
    FMemo: TOptionsBlobType;
    FGraphic: TOptionsBlobType;
    FDateTimePicker: boolean;
    DateTimePickerMinWidth: integer;
    FOnlyPickListValues: boolean;
    FTextAsMemoSize: integer;

    FNotProcDblClick: boolean;
    FNotProcMemoDblClick: boolean;
    FNotProcGraphicDblClick: boolean;

    FAllowPasteMultipleCells: boolean;
    FAllowReadOnlyButton: boolean;
    FAutoPickListWidth: boolean;
    FAllowLookupFilter: boolean;
    FAutoDropDown: boolean;
    FBlockInsertByKey: boolean;
    procedure SetBoolGlyphs(Value: boolean);
    procedure SetMemo(Value: TOptionsBlobType);
    procedure SetGraphic(Value: TOptionsBlobType);
    procedure SetDateTimePicker(Value: boolean);
  public
    constructor Create(Owner: TrDBGrid);
    function ShowMemoGlyph: boolean;
    function ShowMemoContent: boolean;
    function ShowGraphicGlyph: boolean;
    function ShowGraphicContent: boolean;
  published
    { show checkbox for boolean fields }
    property BoolGlyphs: boolean read FBoolGlyphs write SetBoolGlyphs default true;
    { change bool value after simple click }
    property BoolOneClickChange: boolean read FBoolOneClickChange write FBoolOneClickChange default false;
    { record is automatically saved after bool update if was in browse mode before change }
    property BoolAutoSave: boolean read FBoolAutoSave write FBoolAutoSave default false;
    { cell type for memo fields (none, glyph or content) }
    property Memo: TOptionsBlobType read FMemo write SetMemo default obtGlyph;
    { cell type for graphic fields (none, glyph or content) }
    property Graphic: TOptionsBlobType read FGraphic write SetGraphic default obtGlyph;
    { use datetime picker for date field }
    property DateTimePicker: boolean read FDateTimePicker write SetDateTimePicker default true;
    { limits picklist fields to the list values only }
    property OnlyPickListValues: boolean read FOnlyPickListValues write FOnlyPickListValues default false;
    { size of text field to show as memo }
    property TextAsMemoSize: integer read FTextAsMemoSize write FTextAsMemoSize default 0;
    { block dbl-click event for all fields }
    property NotProcDblClick: boolean read FNotProcDblClick write FNotProcDblClick default false;
    { block dbl-click event for memo fields = do not show memo dialog }
    property NotProcMemoDblClick: boolean read FNotProcMemoDblClick write FNotProcMemoDblClick default false;
    { block dbl-click event for graphic fields = do not show graphic dialog }
    property NotProcGraphicDblClick: boolean read FNotProcGraphicDblClick write FNotProcGraphicDblClick default false;

    { Ctrl+V paste clipboard as multiple cell/records if possible }
    property AllowPasteMultipleCells: boolean read FAllowPasteMultipleCells write FAllowPasteMultipleCells default false;
    { edit button in the cell is enabled also for read-only cells }
    property AllowReadOnlyButton: boolean read FAllowReadOnlyButton write FAllowReadOnlyButton default false;
    { allow filtering of data for the lookup fields }
    property AllowLookupFilter: boolean read FAllowLookupFilter write FAllowLookupFilter default false;
    { allow automatic dropdown after key press }
    property AutoDropDown: boolean read FAutoDropDown write FAutoDropDown default false;
    { width of dropdown picklist is set accroding items length }
    property AutoPickListWidth: boolean read FAutoPickListWidth write FAutoPickListWidth default true;
    { block inserting of new record by insert or down key }
    property BlockInsertByKey: boolean read FBlockInsertByKey write FBlockInsertByKey default false;
  end;

  { setting for TrDBGrid appearance, part of TOptionsEx2 }
  TOptionsAppearance = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FSkipCellHighlight: boolean;
    FDrawRowFocus: boolean;
    FDrawColoredRow: boolean;
    FHighlightMultiselect: boolean;
    FHideHScrollBar: boolean;
    FStandardTabKey: boolean;
    FReturnAsTabKey: boolean;
    FDragCell: boolean;
    FDragToScrollData: boolean;
    FHintMaxLength: integer;
    FColumnMinWidth: integer;
    FColumnMaxWidth: integer;
    FLookupResetDelay: cardinal;
    FDataRowLineColor: TColor;
    FDataRowLineStyle: TPenStyle;
    FDataColLineColor: TColor;
    FDataColLineStyle: TPenStyle;
    FTitleRowLineColor: TColor;
    FTitleRowLineStyle: TPenStyle;
    FTitleColLineColor: TColor;
    FTitleColInGroupLineColor: TColor;
    FTitleColLineStyle: TPenStyle;
    FDataRowInFixedColsAsTitle: boolean;
    FColumnWidthMaxRecordsSearch: integer;
    FColumnFontStyleColor: boolean;
    FTitleFontStyleColor: boolean;
    procedure SetDrawRowFocus(Value: boolean);
    procedure SetDrawColoredRow(Value: boolean);
    procedure SetHighlightMultiselect(Value: boolean);
    procedure SetHideHScrollBar(Value: boolean);
    procedure SetDataRowLineColor(Value: TColor);
    procedure SetDataRowLineStyle(Value: TPenStyle);
    procedure SetDataColLineColor(Value: TColor);
    procedure SetDataColLineStyle(Value: TPenStyle);
    procedure SetTitleRowLineColor(Value: TColor);
    procedure SetTitleRowLineStyle(Value: TPenStyle);
    procedure SetTitleColLineColor(Value: TColor);
    procedure SetTitleColInGroupLineColor(Value: TColor);
    procedure SetTitleColLineStyle(Value: TPenStyle);
    procedure SetDataRowInFixedColsAsTitle(Value: boolean);
    procedure SetColumnFontStyleColor(Value: boolean);
    procedure SetTitleFontStyleColor(Value: boolean);
  public
    constructor Create(Owner: TrDBGrid);
  published
    { disable cell highlight }
    property SkipCellHighlight: boolean read FSkipCellHighlight write FSkipCellHighlight default false;
    { show focus for all active row }
    property DrawRowFocus: boolean read FDrawRowFocus write SetDrawRowFocus default false;
    { each second row is colored with color defined by ColorOddRow }
    property DrawColoredRow: boolean read FDrawColoredRow write SetDrawColoredRow default false;
    { mutiselected rows are highlighted or colored }
    property HighlightMultiselect: boolean read FHighlightMultiselect write SetHighlightMultiselect default true;
    { hide horizontal scrollbar }
    property HideHScrollBar: boolean read FHideHScrollBar write SetHideHScrollBar default false;
    { tab key in not edit mode select next control }
    property StandardTabKey: boolean read FStandardTabKey write FStandardTabKey default false;
    { return key in not edit mode select next column or row (as tab key) }
    property ReturnAsTabKey: boolean read FReturnAsTabKey write FReturnAsTabKey default false;
    { enable dragging of cell value to other controls }
    property DragCell: boolean read FDragCell write FDragCell default false;
    { enable using drag and move to scroll data in the grid }
    property DragToScrollData: boolean read FDragToScrollData write FDragToScrollData default false;
    { maximal number of chars in the hint message }
    property HintMaxLength: integer read FHintMaxLength write FHintMaxLength default 2000;
    { minimal column width used for AutoFit and SetColumnWidth }
    property ColumnMinWidth: integer read FColumnMinWidth write FColumnMinWidth default 40;
    { miximal column width used for AutoFit and SetColumnWidth }
    property ColumnMaxWidth: integer read FColumnMaxWidth write FColumnMaxWidth default 200;
    { define maximal count of records searched to find data column width }
    property ColumnWidthMaxRecordsSearch: integer read FColumnWidthMaxRecordsSearch write FColumnWidthMaxRecordsSearch default 2000;
    { define delay from key press to lookup reset }
    property LookupResetDelay: cardinal read FLookupResetDelay write FLookupResetDelay default 2500;
    { define own color of row line in data grid, use clNone to use standard line }
    property DataRowLineColor: TColor read FDataRowLineColor write SetDataRowLineColor default clNone;
    { define own style of row line in data grid }
    property DataRowLineStyle: TPenStyle read FDataRowLineStyle write SetDataRowLineStyle default psSolid;
    { define own color of column line in data grid, use clNone to use standard line }
    property DataColLineColor: TColor read FDataColLineColor write SetDataColLineColor default clNone;
    { define own style of column line in data grid }
    property DataColLineStyle: TPenStyle read FDataColLineStyle write SetDataColLineStyle default psSolid;
    { define if row line for fixed columns ar the same as in title }
    property DataRowInFixedColsAsTitle: boolean read FDataRowInFixedColsAsTitle write SetDataRowInFixedColsAsTitle default false;

    { define own color of row line in title/fixed part of grid, use clNone to use standard line }
    property TitleRowLineColor: TColor read FTitleRowLineColor write SetTitleRowLineColor default clNone;
    { define own style of row line in title/fixed part of grid }
    property TitleRowLineStyle: TPenStyle read FTitleRowLineStyle write SetTitleRowLineStyle default psSolid;
    { define own color of column line in title/fixed part of grid, use clNone to use standard line }
    property TitleColLineColor: TColor read FTitleColLineColor write SetTitleColLineColor default clNone;
    { define own color of column line in title of grid inside the defined group, use clNone to use standard line }
    property TitleColInGroupLineColor: TColor read FTitleColInGroupLineColor write SetTitleColInGroupLineColor default clNone;
    { define own style of column line in title/fixed part of grid }
    property TitleColLineStyle: TPenStyle read FTitleColLineStyle write SetTitleColLineStyle default psSolid;

    { define if custom font color is replaced by style font color for columns }
    property ColumnFontStyleColor: boolean read FColumnFontStyleColor write SetColumnFontStyleColor default false;
    { define if custom font color is replaced by style font color for titles }
    property TitleFontStyleColor: boolean read FTitleFontStyleColor write SetTitleFontStyleColor default false;
  end;

  { text filter type for TrDBGrid filter bar }
  TOptionTextFilterType = (oftAnyPos, oftStartPos, oftExactValue);

  { setting for TrDBGrid filters, part of TOptionsEx2 }
  TOptionsFilters = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FFunnelBtn: boolean;
    FFunnelBtnActiveOnly: boolean;
    FMultiFieldFilter: boolean;
    FTextBar: boolean;
    FTextBarTopLayout: boolean;
    TextBarHeight: integer;
    FTextMarginX: integer;
    FTextMarginY: integer;
    FTextBkColor: TColor;
    FTextFilterType: TOptionTextFilterType;
    FTextFilterLikeChar: char;
    FTextFilterLikeOperator: string;
    FTextListAutoLoad: boolean;
    FTextListAddFilter: boolean;
    FTextComboStyle: TComboBoxStyle;
    FUseDatePartAsRange: boolean;
    FUseNumberPartAsRange: boolean;
    FNumberPartWildChar: char;
    EditColumnIndex: integer;
    EditValues: TStrings;
    EditStatuses: TStrings;
    EditChanged: boolean;
    EditValueList: TMemIniFile;
    FSetFilterDelay: cardinal;
    procedure SetFunnelBtn(Value: boolean);
    procedure SetTextBar(Value: boolean);
    procedure SetTextBarTopLayout(Value: boolean);
    procedure SetTextMarginX(Value: integer);
    procedure SetTextMarginY(Value: integer);
    procedure SetTextBkColor(Value: TColor);
    procedure SetTextComboStyle(Value: TComboBoxStyle);
    function TextBarTopHeight: integer;
    function TextBarBootomHeight: integer;
    function GetEditField: TField;
  public
    constructor Create(Owner: TrDBGrid);
    destructor Destroy; override;
    function GetTextFilterLikeChar: char;
    function GetTextFilterLikeOperator: string;
  published
    { show funnel buttons in the title row }
    property FunnelBtn: boolean read FFunnelBtn write SetFunnelBtn default false;
    { show funnel glyphs for cell with active filter or hottrack title cell }
    property FunnelBtnActiveOnly: boolean read FFunnelBtnActiveOnly write FFunnelBtnActiveOnly default false;
    { enable combination of funnel filter for more coluumns (AND condition) }
    property MultiFieldFilter: boolean read FMultiFieldFilter write FMultiFieldFilter default false;
    { show text edit to enter filter text }
    property TextBar: boolean read FTextBar write SetTextBar default false;
    { show text edit above title text }
    property TextBarTopLayout: boolean read FTextBarTopLayout write SetTextBarTopLayout default false;
    { free space on left and right border }
    property TextMarginX: integer read FTextMarginX write SetTextMarginX default 2;
    { free space on top and bottom border }
    property TextMarginY: integer read FTextMarginY write SetTextMarginY default 1;
    { background color of text edit }
    property TextBkColor: TColor read FTextBkColor write SetTextBkColor default clWindow;
    { load filter picklist from the dataset values }
    property TextListAutoLoad: boolean read FTextListAutoLoad write FTextListAutoLoad default false;
    { add entered value for filter to the picklist }
    property TextListAddFilter: boolean read FTextListAddFilter write FTextListAddFilter default true;
    { comparison type for text filter }
    property TextFilterType: TOptionTextFilterType read FTextFilterType write FTextFilterType default oftAnyPos;
    { define used LIKE char for filter }
    property TextFilterLikeChar: char read FTextFilterLikeChar write FTextFilterLikeChar default #0;
    { define used LIKE operator for filter }
    property TextFilterLikeOperator: string read FTextFilterLikeOperator write FTextFilterLikeOperator;
    { define style of text filter combo }
    property TextComboStyle: TComboBoxStyle read FTextComboStyle write SetTextComboStyle default csDropDown;
    { enable range filtering for date field by entering of year or month+year only }
    property UseDatePartAsRange: boolean read FUseDatePartAsRange write FUseDatePartAsRange default true;
    { enable range filtering for numeric field by entering of numer with wild char on last digits (120* =>  x>=1200 and x<1201 }
    property UseNumberPartAsRange: boolean read FUseNumberPartAsRange write FUseNumberPartAsRange default true;
    { define used wild char for UseNumberPartAsRange }
    property NumberPartWildChar: char read FNumberPartWildChar write FNumberPartWildChar default '*';
    { delay for automatic activation of filter after text edit change }
    property SetFilterDelay: cardinal read FSetFilterDelay write FSetFilterDelay default 1000;
  end;

  { setting for TrDBGrid groups, part of TOptionsEx2 }
  TOptionsGroups = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FActive: boolean;
    FAlwaysUseGroupColors: boolean;
    FGroupHeightPixels: integer;
    FCollapseBtnSize: integer;
    FGroupLineWidth: byte;
    procedure SetActive(Value: boolean);
    procedure SetAlwaysUseGroupColors(Value: boolean);
    procedure SetGroupHeightPixels(Value: integer);
    procedure SetCollapseBtnSize(Value: integer);
    procedure SetGroupLineWidth(Value: byte);
    function GroupHeight: integer;
  public
    constructor Create(Owner: TrDBGrid);
  published
    { activate Groups usage }
    property Active: boolean read FActive write SetActive default false;
    { use group color for title row when group is not active }
    property AlwaysUseGroupColors: boolean read FAlwaysUseGroupColors write SetAlwaysUseGroupColors default false;
    { height of free space for group buttons and lines }
    property GroupHeightPixels: integer read FGroupHeightPixels write SetGroupHeightPixels default 12;
    { size of group collapse / expand button }
    property CollapseBtnSize: integer read FCollapseBtnSize write SetCollapseBtnSize default 10;
    { width of group lines }
    property GroupLineWidth: byte read FGroupLineWidth write SetGroupLineWidth default 1;
  end;

  { setting for TrDBGrid cell condition, part of TOptionsEx2 }
  TOptionsCellCondition = class(TPersistent)
  private
    FOwner: TrDBGrid;
    FActive: boolean;
    procedure SetActive(Value: boolean);
  public
    constructor Create(Owner: TrDBGrid);
  published
    { activate cell conditions }
    property Active: boolean read FActive write SetActive default false;
  end;

  { setting for TrDBGrid new functions, replacement of previous TOptionsEx }
  TOptionsEx2 = class(TPersistent)
  private
    FOwner: TrDBGrid;
{$IFDEF UseObsoleteOption}
    FActivateOptionsEx2: boolean;
{$ENDIF}
    FFixedCell: TOptionsFixedCell;
    FEditors: TOptionsEditors;
    FAppearance: TOptionsAppearance;
    FFilters: TOptionsFilters;
    FGroups: TOptionsGroups;
    FCellCondition: TOptionsCellCondition;
  public
    constructor Create(Owner: TrDBGrid);
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override; {< @exclude }
  published
{$IFDEF UseObsoleteOption}
    { activate usage of OptionsEx2 instead of depracated OptionsEx }
    property ActivateOptionsEx2: boolean read FActivateOptionsEx2 write FActivateOptionsEx2 default false;
{$ENDIF}
    { setting for rDBGrid fixed cells }
    property FixedCell: TOptionsFixedCell read FFixedCell write FFixedCell;
    { setting for rDBGrid editors }
    property Editors: TOptionsEditors read FEditors write FEditors;
    { setting for rDBGrid appearance }
    property Appearance: TOptionsAppearance read FAppearance write FAppearance;
    { setting for rDBGrid filters }
    property Filters: TOptionsFilters read FFilters write FFilters;
    { setting for rDBGrid groups }
    property Groups: TOptionsGroups read FGroups write FGroups;
    { setting for rDBGrid cell condition }
    property CellCondition: TOptionsCellCondition read FCellCondition write FCellCondition;
  end;

  { list of possible AutoLookup conditions }
  TLookupCondition = (lcNever, lcBrowse, lcNotEdit, lcBrowseFilter, lcNotEditFilter);
  { list of possible column width types }
  TColumnWidth = (cwNone, cwDefaultWidth, cwSameWidth, cwAutoWidth, cwDataWidth, cwLabelWidth, cwDataLabelWidth, cwHideEmptyCol);

  { record definition for storing of default column width and actual filter }
  TColumnStoredProp = record
    FieldName: string;
    Width: integer;
    Visible: boolean;
    FilterString: string;
    // FilterList: string;
  end;

  { array type for storing of TColumnStoredProp for all columns }
  TColumnStoredProps = array of TColumnStoredProp;

  { data image record used by rDBGrid }
  TDataImageDef = record
    UseDataImage: boolean;
    ImageIndex: integer;
    ImageAlignment: string;
    ShowText: string;
  end;

  { define record for distinct values for TSummaryResults }
  TDistinctValues = record
    Value: string;
    Count: integer;
  end;

  { define record for calculated summary }
  TSummaryResults = record
  private
    HList: THashedStringList;
  public
    Field: TField;
    Sum, Min, Max, Avg: double;
    CountValues, CountEmpty: integer;
    SelSum, SelMin, SelMax, SelAvg: double;
    SelCountValues, SelCountEmpty: integer;
    DistinctValueCount: integer;
    DistinctValues: array of TDistinctValues;
  end;

  { array of field names used for sorting }
  TSortFieldNames = array of string;
  { array of directions used for sorting }
  TSortDirections = array of boolean;

  { event is called after click on title row to provide sorting if Sorter component is not used;
    if "Accept" remains true, "SortFieldName" and "SortDesc" is changed
    and sort marker is shown, in other case nothing is changed;
    obsolete, use ChangeSortEventEx for multifield sorting }
  TChangeSortEvent = procedure(Sender: TObject; DataSet: TDataSet; SortFieldName: string; SortDesc: boolean; var Accept: boolean) of object;

  { Enhanced event with MultiField sorting supprt;
    event is called after click on title row to provide sorting if Sorter component is not used;
    if "Accept" remains true, "SortFieldName" and "SortDirections" is changed
    and sort marker is shown, in other case nothing is changed }
  TChangeSortEventEx = procedure(Sender: TObject; DataSet: TDataSet; SortFieldNames: TSortFieldNames; SortDirections: TSortDirections; var Accept: boolean)
    of object;

  { event is called to load picklist or textbar-filter list for particular field - usually by aggregated query }
  TLoadPickListEvent = procedure(Sender: TObject; DS: TDataSet; FieldName: string; PickList: TStrings) of object;
  { callback procedure used by TLoadFunnelFilterEvent }
  TFunnelFilterAddToList = procedure(DisplayText, ValueText: string; Checked: boolean) of object;
  { event is called to load funnel-filter list for particular field - usually by aggregated query }
  TLoadFunnelFilterEvent = procedure(Sender: TObject; DS: TDataSet; FieldName: string; AddToListProc: TFunnelFilterAddToList; var IsLoaded: boolean) of object;

  { event is called before showing of hint window to get user defined hint string
    HintFieldValue contain value of field defined by HintFieldName from record under mouse }
  TGetHintStringEventEx = procedure(Sender: TObject; UnderMouseColumn: TColumn; TitleRow, FooterRow: boolean; FieldValues: TStrings; CellText: string;
    var HintString: string) of object;
  { event is called to get title line text or draw title line directly }
  TDrawTitleLineEvent = procedure(Sender: TObject; const Rect: TRect; var Text: string) of object;
  { event is called to get font and text of fixed column cell }
  TGetFixedColStateEvent = procedure(Sender: TObject; ActiveRow: boolean; AFont: TFont; var Text: string; var CheckBoxChecked: boolean) of object;
  { event is called to get font and text of fixed row cell (title cell) }
  TGetFixedRowStateEvent = procedure(Sender: TObject; DataCol: integer; Column: TColumn; AFont: TFont; var Text: string; var RotateText: boolean) of object;
  { event is called to get params of data cell }
  TGetCellParamsEvent = procedure(Sender: TObject; Field: TField; AFont: TFont; var Background: TColor; ActiveRow, Highlight: boolean) of object;
  { event is called to get extended params of data cell }
  TGetCellParamsEventEx = procedure(Sender: TObject; Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef;
    ActiveRow, Highlight: boolean) of object;
  { event is called to set all calculated values for one column (field), usually by some SQL query instead of record scan }
  TRecalculateSummaryResultEvent = procedure(Sender: TObject; F: TField; var SummaryResults: TSummaryResults) of object;
  { event is called to update or format one particular calculated value of field (sum, min, max, avg ...) }
  TGetSummaryResultEvent = procedure(Sender: TObject; F: TField; SummaryResults: TSummaryResults; TagStr: string; ForFooter: boolean; var DisplayText: string)
    of object;

  { event is called to get footer simpletext or draw footer rect directly }
  TDrawFooterRowEvent = procedure(Sender: TObject; FooterRect: TRect; var FooterSimpleText: string; var DefaultDrawing: boolean) of object;
  { event is called to get footer cell text or draw footer cell directly }
  TDrawFooterCellEvent = procedure(Sender: TObject; FooterCellRect: TRect; ColIndex: integer; Field: TField; var FooterCellText: string;
    var DefaultDrawing: boolean) of object;
  { event is called after click to footer cell }
  TFooterCellClick = procedure(Sender: TObject; ACol: integer; Column: TColumn) of object;

  { event is called after setting filter for any column }
  TSetColumnFilterEvent = procedure(Sender: TObject; DataSet: TDataSet; Field: TField; FilterString, FilterValue: string; FilterForm: TForm;
    var Result: boolean) of object;
  { event is called before and after dataset scan }
  TScanDatasetEvent = procedure(Sender: TObject; DataSet: TDataSet; Reason: integer) of object;
  { event is called after changing of column width }
  TColWidthChangedEvent = procedure(Sender: TObject; Column: TColumn; ManualSizing: boolean) of object;
  { event is called to define format of Ellipsis button in the cell }
  TGetEllipsisButtonFormat = procedure(Sender: TObject; ColIndex: Longint; Field: TField; var ButtonWidth: integer; var Glyph: TBitmap) of object;
  { event is called before editing of cell, CanEdit define possibility to editation }
  TCanEditEvent = procedure(Sender: TObject; Field: TField; Column: TColumn; var CanEdit: boolean) of object;

  { general event type used for TGeneralEvent }
  TGeneralEventType = (geChangeScaleBefore, geChangeScaleAfter, geIsBool, geIsMemo, geIsMemoGlyph, geIsMemoText, geIsRichEdit, geIsGraphic, geIsFieldAsDate,
    geIsDate, geIsFieldAsTime, geIsTime, geIsColor, geIsLink, geIsPassword, geIsFilter, geIsDrawDropDown, geIsDataBar, geIsStatusBar, geIsDataImage,
    geIsFixedColumn, geIsReadOnlyColumn, geIsLookupFilterColumn, geGetTitleImageIndex, geShowMemoDialog, geShowGraphicDialog, geSetColumnWidth, geContextPopup,
    geContextTitlePopup, geGetDataBarParam, geGetStatusBarParam, geGetFooterRowParam, geGetDataImageIndex, geGetDataImageAlignment);

  { general event record used as parameter for TGeneralEvent }
  TGeneralEventRec = record
    Column: TColumn;
    Field: TField;
    ParamObject: TObject;
    ParamValue: variant;
    FieldValue: variant;
  end;

  { event is called in many events to add posibility of any custom code and to skip rest of the standard event }
  TGeneralEvent = procedure(Sender: TObject; EventType: TGeneralEventType; var EventParam: TGeneralEventRec; var ContinueEvent: boolean) of object;

  { define inplace editor used by TrDBGrid }
  TrDBGridInplaceEditEx = class(TInplaceEditList)
  private
    FDataList: TrPopupDataList;
    FUseDataList: boolean;
    FLookupSource: TDatasource;
    oldFieldColorString: string;
    procedure PickListDrawItem(Control: TWinControl; Index: integer; aRect: TRect; aState: TOwnerDrawState);
    procedure PickListMeasureItem(Control: TWinControl; Index: integer; var NewHeight: integer);
    function CheckSelectUserDefinedColor(F: TField): boolean;
    procedure PopupCallbackSelectDate(Sender: TObject);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
    procedure WMPaste(var Message: TMessage); message WM_PASTE;
  protected
    FPopupCalendar: TrPopupMonthCalendar;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    procedure CloseUp(Accept: boolean); override;
    procedure DoEditButtonClick; override;
    procedure DropDown; override;
    procedure DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DataListProcessSearchKey(Sender: TObject; var Key: char);
    procedure UpdateContents; override;
    procedure DblClick; override;
    procedure SetListWidth; virtual;
    procedure PaintWindow(DC: HDC); override;
    function ProcDateTimeHotKey(Key: Word; Shift: TShiftState): boolean; virtual;
  public
    constructor Create(Owner: TComponent); override; {< @exclude }
    property DataList: TrPopupDataList read FDataList; {< @exclude }
  end;

  { define object for setting of text in title row used by TrDBGrid }
  TTitleLineText = class(TPersistent)
  private
    AGrid: TrDBGrid;
    PosLeft, PosRight: integer;
    LineHeight, SumHeight: integer;
    FFixHeight: integer;
    FText: string;
    FFont: TFont;
    FParentFont: boolean;
    FColor: TColor;
    FTextAlign: TAlignment;
    FShowDesignTime: boolean;
    procedure SetFixHeight(Value: integer);
    procedure SetText(const Value: string);
    procedure SetFont(Value: TFont);
    procedure SetColor(Value: TColor);
    procedure SetTextAlign(Value: TAlignment);
    procedure SetShowDesignTime(Value: boolean);
    procedure SetParentFont(Value: boolean);
    function IsFontStored: boolean;
    procedure FontChanged(Sender: TObject);
    function GetLineCount: integer;
    function GetLineText(LineIdx: integer): string;
  public
    constructor Create(Owner: TrDBGrid); dynamic; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Assign(Source: TPersistent); override; {< @exclude }
  published
    { define fixed title line height, if 0 then height is calculated from text }
    property FixHeight: integer read FFixHeight write SetFixHeight default 0;
    { define background color for title line }
    property Color: TColor read FColor write SetColor default clBtnFace;
    { define text in title line, use | as line separator }
    property Text: string read FText write SetText;
    { define using of parent font instead of defined font }
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    { define font for text in title line }
    property Font: TFont read FFont write SetFont stored IsFontStored;
    { define alignment for text in title line }
    property TextAlign: TAlignment read FTextAlign write SetTextAlign default taLeftJustify;
    { deprecated - not used }
    property ShowDesignTime: boolean read FShowDesignTime write SetShowDesignTime stored false default false;
  end;

  { define object for setting of text in fixed column for TrDBGrid }
  TFixedColText = class(TPersistent)
  private
    AGrid: TrDBGrid;
    FFixedColWidth: integer;
    FTitleText: string;
    FTextMask: string;
    FFont: TFont;
    FParentFont: boolean;
    FShowIndicator: boolean;
    FShowCheckbox: boolean;
    FTextAlign: TAlignment;
    procedure SetFixedColWidth(Value: integer);
    procedure SetTitleText(const Value: string);
    procedure SetTextMask(const Value: string);
    procedure SetFont(Value: TFont);
    procedure SetTextAlign(Value: TAlignment);
    procedure SetParentFont(Value: boolean);
    procedure SetShowIndicator(Value: boolean);
    procedure SetShowCheckbox(Value: boolean);
    function IsFontStored: boolean;
    procedure FontChanged(Sender: TObject);
    function IsFixedColUsed: boolean;
  public
    constructor Create(Owner: TrDBGrid); dynamic; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Assign(Source: TPersistent); override; {< @exclude }
  published
    { define width of fixed column }
    property FixedColWidth: integer read FFixedColWidth write SetFixedColWidth default IndicatorWidth;
    { define showing of indicator image }
    property ShowIndicator: boolean read FShowIndicator write SetShowIndicator default true;
    { define showing of checkbox in fixed column }
    property ShowCheckbox: boolean read FShowCheckbox write SetShowCheckbox default false;
    { define text in title row of fixed col }
    property TitleText: string read FTitleText write SetTitleText;
    { define text mask in data rows of fixed col, use <<FieldName>> to add field value }
    property TextMask: string read FTextMask write SetTextMask;
    { define using of parent font instead of defined font }
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    { define font for text in fixed col }
    property Font: TFont read FFont write SetFont stored IsFontStored;
    { define alignment for text in fixed col }
    property TextAlign: TAlignment read FTextAlign write SetTextAlign default taLeftJustify;
  end;

  { define object for setting of text in footer row used by TrDBGrid }
  TFooterRow = class(TPersistent)
  private
    AGrid: TrDBGrid;
    FFooterVisible: boolean;
    FFooterHeight: integer;
    FFont: TFont;
    FParentFont: boolean;
    FBoldText: boolean;
    FColor: TColor;
    FShowCellLines: boolean;
    FCellLinesColor: TColor;
    FFixedColText: string;
    FSimpleText: string;
    FSimpleTextAlign: TAlignment;
    FFieldFooterDefs: TStrings;
    FRecalculateAfterOpen: boolean;
    FRecalculateAfterFilter: boolean;
    FRefreshAfterScroll: boolean;
    procedure SetFooterVisible(Value: boolean);
    procedure SetFooterHeight(Value: integer);
    procedure SetFont(Value: TFont);
    procedure SetParentFont(Value: boolean);
    function IsFontStored: boolean;
    procedure FontChanged(Sender: TObject);
    procedure FieldFooterDefChanged(Sender: TObject);
    procedure SetBoldText(Value: boolean);
    procedure SetShowCellLines(Value: boolean);
    procedure SetCellLinesColor(Value: TColor);
    procedure SetColor(Value: TColor);
    procedure SetFixedColText(Value: string);
    procedure SetSimpleText(Value: string);
    procedure SetSimpleTextAlign(Value: TAlignment);
    procedure SetFieldFooterDefs(Value: TStrings);
    function GetFooterRowHeight: integer;
  public
    constructor Create(Owner: TrDBGrid); dynamic; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Assign(Source: TPersistent); override; {< @exclude }
    { return rectangle of all footer row }
    function FooterRect: TRect;
    { return rectangle of one cell of footer row }
    function FooterCellRect(ACol: integer): TRect;
    { indicate if footer text is defined for the field }
    function IsFooterDefined(Field: TField): boolean;
    { return footer text for the field }
    function GetFooterText(Field: TField): string;
    { return footer parameters for the field as string }
    // function GetFooterParams(FieldName: string): string; overload;
    { return footer parameters for the field as string }
    // function GetFooterParams(Field: TField): string; overload;
    { set footer parameters for the field }
    // procedure SetFooterParamsEx(FieldName: string; TextMask, Params: string);
    { set alignment and font style according params string }
    // procedure UseFooterParams(Params: string; var Align: TAlignmentDefault; AFont: TFont; var NormalStyleDefined: boolean);
    { redraw footer content (redraw only - not recalculate) }
    procedure Refresh;
  published
    { switch on/off showing of footer inside grid }
    property FooterVisible: boolean read FFooterVisible write SetFooterVisible default false;
    { define specific height of footer row, 0 means current data row height }
    property FooterHeight: integer read FFooterHeight write SetFooterHeight default 0;
    { define using of parent font instead of defined font }
    property ParentFont: boolean read FParentFont write SetParentFont default true;
    { define font for text in footer }
    property Font: TFont read FFont write SetFont stored IsFontStored;
    { define that footer use bold text although parent font is used }
    property BoldText: boolean read FBoldText write SetBoldText default false;
    { define background color of footer }
    property Color: TColor read FColor write SetColor default clNone;
    { define showing of cells frame }
    property ShowCellLines: boolean read FShowCellLines write SetShowCellLines default true;
    { define color of cell lines }
    property CellLinesColor: TColor read FCellLinesColor write SetCellLinesColor default clNone;
    { define footer text in fixed col }
    property FixedColText: string read FFixedColText write SetFixedColText;
    { define footer simple text }
    property SimpleText: string read FSimpleText write SetSimpleText;
    { define alignment of footer text }
    property SimpleTextAlign: TAlignment read FSimpleTextAlign write SetSimpleTextAlign default taLeftJustify;
    { define content of footer for each field in following format:
      FieldName1=footertext;param
      FieldName2=footertext;param
      where footertext can contain tags for calculated values %SUM, %MIN, %MAX, %AVG, %COUNTVAL (see cSummaryTag_xxx for tag list) }
    property FieldFooterDefs: TStrings read FFieldFooterDefs write SetFieldFooterDefs;
    { switch on/off automatic footer recalculation when dataset active change }
    property RecalculateAfterOpen: boolean read FRecalculateAfterOpen write FRecalculateAfterOpen default false;
    { switch on/off automatic footer recalculation when grid filter change }
    property RecalculateAfterFilter: boolean read FRecalculateAfterFilter write FRecalculateAfterFilter default false;
    { switch on/off automatic footer redraw after scroll (needed only if RecNo is used as summary content) }
    property RefreshAfterScroll: boolean read FRefreshAfterScroll write FRefreshAfterScroll default false;
  end;

  { list of operators for cell highlighting conditions }
  TCellParamsCondOperators = (cpoEqual, cpoLess, cpoLessOrEq, cpoGreater, cpoGreaterOrEq, cpoNotEqual, cpoStart, cpoNotStart, cpoContain, cpoNotContain,
    cpoAllways);
  { list of logical operators for cell highlighting conditions }
  TCellParamsCondActions = (cpaSet, cpaAND, cpaOR);

  { define object for storing of one condition for cell highlighting used by TrDBGrid }
  TCellParamsCond = class
  public
    FieldName: string;
    CondOperator: TCellParamsCondOperators;
    Value: string;
    Action: TCellParamsCondActions;
    FontColor: TColor;
    BkColor: TColor;
    Bold: boolean;
    WholeRow: boolean;
    ImageDef: TDataImageDef;
    StopIfTrue: boolean;
    Deactive: boolean;
    ErrorDesc: string;
    LastRecNo: integer;
    LastResult: boolean;
  end;

  { define list of conditions for cell highlighting used by TrDBGrid }
  TCellParamsCondList = class(TList)
  private
    Gr: TrDBGrid;
    DS: TDataSet;
    procedure ProcIniContent(IniContent: TStrings; GridKey: string; Load: boolean);
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;
  public
    procedure LinkGrid(Grid: TrDBGrid);
    procedure LinkDataset(DataSet: TDataSet);
    procedure ResetCond;
    procedure LoadFromIniContent(IniContent: TStrings; GridKey: string = '');
    procedure StoreToIniContent(IniContent: TStrings; GridKey: string = '');
    procedure GetFieldCond(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef; ActiveRow, Highlight, RefreshCond: boolean);
      virtual;
  end;

  { define record for grouped field used by TrDBGrid }
  TGroupInfo = record
    Field: TField;
    IsGrouped: boolean;
    GroupID: integer;
    GroupTitle: string;
    TitleFontColor: TColor;
    GroupLineColor: TColor;
    // FieldSetting: string;
    IsMaster: boolean;
    MasterField: TField;
    ContentCollapsed: string;
    GroupLocked: boolean;
  end;

  { special extension of TrComboBoxEx for using in filter bar }
  TrComboBoxExFilter = class(TrComboBoxEx)
  private
    Gr: TrDBGrid;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CancelMode;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  end;

  { enhanced TDBGrid component which provides many new functions
    (show checkboxes, glyph for graphics and memo fields, build-in editors,
    pictures in title row or in data cells, multiple line titles,
    sorting functions etc) }
  TrDBGrid = class(TDBGrid)
  private
    TempBmp: TBitmap;
    isPushing: boolean;
    isDataScrolling: boolean;
    isDblClick: boolean;
    isColMoved: boolean;
    isLink: boolean;
    isActiveRow, isHotTrackRow: boolean;
    isReturnAsTabKeyActive: boolean;
    ShowHintRecCount: boolean;
    ShiftState: TShiftState;
    MMoveCoord, MDownCoord: TGridCoord;
    oldWidth: integer;
    oldRowPos, oldRecNo: integer;
    oldAfterScroll, oldBeforeScroll: TDataSetNotifyEvent;
    oldDataChange: TDataChangeEvent;
    ExUpdateLock: boolean;
    FHintFieldValues: TStrings;
    oldHintPause: integer;
    FDropDownRowsCount: cardinal;
    FFilterCombo: TrComboBoxExFilter;
    FFilterComboFontColor: TColor;
    OwnHintWin: THintWindow;
    FSearchText: string;
    FSearchingPrevNext: TSearchingType;
    FTimer: TTimer;
    FSorter: TrDBGridCustomSorter;
    FLastColWidthChangedIdx: integer;
    FColumnStoredProps: TColumnStoredProps;
    FPrevActiveBookmark: TBookmark;
    FActiveBookmark: TBookmark;
    ActBmp: integer;
    FCustomGlyphs: TrCustomGlyphs;
    FGlyphLookStyle: TGlyphLookStyles;
    FGridLookStyle: TGridLookStyles;
    FTitleLineText: TTitleLineText;
    FFixedColText: TFixedColText;
    FFooterRow: TFooterRow;
    FTitleImageList: TImageList;
    FTitleImages: TStrings;
    FDataImageList: TImageList;
    FDataImages: TStringList;
    FDataImagesIni: TMemIniFile;
    FFieldsGroups: TStringList;
    FFieldsGroupsDefault: TStringList;
    FFieldsGroupsIni: TMemIniFile;
    FTitleLines: integer;
    FRowFixHeight: integer;
    FRowLines: integer;
    FRowVerticalAlign: TVerticalAlignment;
    FTitleHints: TStrings;
{$IFDEF UseObsoleteOption}
    FOptionsEx_old: TOptionsEx;
{$ENDIF}
    FOptionsEx2: TOptionsEx2;
    FLookupCondition: TLookupCondition;
    FSortFieldName: string;
    FSortDesc: boolean;
    FSortDirections: string;
    FTextAsMemoSize_old: integer;
    FPropSaveFile: TPropSaveFile;
    FColumnWidth: TColumnWidth;
    FColumnWidthAfterDblClick: TColumnWidth;
    FColorOddRow: TColor;
    FColorActiveRow: TColor;
    FColorHotTrackRow: TColor;
    FColorRequiredField: TColor;
    FPasswordChar: char;
    FFieldsAsBoolean: TStrings;
    FFieldsAsRichText: TStrings;
    FFieldsAsColor: TStrings;
    FFieldsAsDate: TStrings;
    FFieldsAsTime: TStrings;
    FFieldsAsGraphics: TStrings;
    FFieldsAsLink: TStrings;
    FFieldsAsPassword: TStrings;
    FFieldsAsDataBar: TStrings;
    FFieldsAsStatusBar: TStrings;
    FFieldsAutoPickList: TStrings;
    FFieldsDrawDropDown: TStrings;
    FFieldsFilterButton: TStrings;
    FFieldsFilterMapping: TStrings;
    FFieldsFixedColumnWidth: TStrings;
    FTitlePopup: TPopupMenu;
    FTitlePopupItems: TStrings;
    FTitleExPopupMenu: TPopupMenu;
    FFontColorFiltered: TColor;
    FFontColorLink: TColor;
    FFontColorReadOnly: TColor;
    FCellParamsCondList: TCellParamsCondList;
    FSummaryResultsArr: array of TSummaryResults;
    // lookup list
    FLookupListItemHeight: integer;
    FLookupListItemIncreaseHeight: integer;
    FLookupListItemIncreaseWidth: integer;
    FGetLookupListItemHeight: TGetLookupListItemHeight;
    FGetLookupListItemProps: TGetLookupListItemProps;
{$IFNDEF TrPopupMonthCalendar}
    FDateTimePicker: TWinControl;
    FDateTimePickerCreate: TDateTimePickerCreate;
    FDateTimePickerEvent: TDateTimePickerEvent;
{$ENDIF}
    // events
    FOnCanEdit: TCanEditEvent;
    FOnChangeSort: TChangeSortEvent;
    FOnChangeSortEx: TChangeSortEventEx;
    FOnLinkClick: TLinkClickEvent;
    FOnBoolClick: TLinkClickEvent;
    FOnTitleExClick: TNotifyEvent;
    FOnFixColClick: TNotifyEvent;
    FOnFooterCellClick: TFooterCellClick;
    FOnGetFixedColState: TGetFixedColStateEvent;
    FOnGetFixedRowState: TGetFixedRowStateEvent;
    FOnDrawActiveRowFocus: TDrawColumnCellEvent;
    FOnDrawColumnCellEnd: TDrawColumnCellEvent;
    FOnDrawTitleCellEnd: TDrawCellEvent;
    FOnDrawCellBackground: TDrawCellEvent;
    FOnRefreshData: TNotifyEvent;
    FOnGetCellParams: TGetCellParamsEvent;
    FOnGetCellParamsEx: TGetCellParamsEventEx;
    FOnLoadPickList: TLoadPickListEvent;
    FOnTopLeftChanged: TNotifyEvent;
    FOnDataScrolled: TNotifyEvent;
    FOnGetHintStringEx: TGetHintStringEventEx;
    FOnRecalculateSummaryResult: TRecalculateSummaryResultEvent;
    FOnGetSummaryResult: TGetSummaryResultEvent;
    FOnPickListDownItem: TDrawItemEvent;
    FOnPickListMeasureItem: TMeasureItemEvent;
    FOnPickListSetSize: Controls.TCanResizeEvent;
    FOnGetEllipsisButtonFormat: TGetEllipsisButtonFormat;
    FOnDrawTitleLine: TDrawTitleLineEvent;
    FOnSetColumnFilter: TSetColumnFilterEvent;
    FOnColWidthChanged: TColWidthChangedEvent;
    FOnGeneralEvent: TGeneralEvent;
    FGeneralEventRec: TGeneralEventRec;
    FBeforeScanDataset: TScanDatasetEvent;
    FAfterScanDataset: TScanDatasetEvent;
    // Footer Row
    FCalculateDistinctValues: boolean;
    FUseCalculatedSummaryForPopup: boolean;
    FOnDrawFooterRow: TDrawFooterRowEvent;
    FOnDrawFooterCell: TDrawFooterCellEvent;
    // Footer DBGrid
    FooterGridUpdating: boolean;
    FFooterGrid: TrDBGrid;
    FFooterGridFieldPrefix: string;
    // inform master grid after TopLeftChanged
    NotifyTopLeftChanged: TNotifyEvent;
    // FilterDlg
    FFilterDlg: TForm;
    FFilterDlgLastColIdx: integer;
    FFilteredFieldNames: string;
    FColumnFilters: TColumnStoredProps;
    FFilterChanged: TNotifyEvent;
    FMouseOverRow: integer;
    FOnLoadFilterList: TLoadPickListEvent;
    FOnLoadFunnelFilter: TLoadFunnelFilterEvent;
    // Callback vars
    CBScan_Int: integer;
    CBScan_IntArr: array of integer;
    CBScan_Str: string;
    CBScan_Field: TField;
    procedure SetCustomGlyphs(const Value: TrCustomGlyphs);
    procedure SetGlyphLookStyle(const Value: TGlyphLookStyles);
    procedure SetGridLookStyle(const Value: TGridLookStyles);
    procedure SetSortFieldName(const Value: string);
    procedure SetSortDesc(const Value: boolean);
    procedure SetSortDirections(const Value: string);
{$IFDEF UseObsoleteOption}
    procedure SetOptionsEx(const Value: TOptionsEx);
{$ENDIF}
    procedure SetFColumnWidth(const Value: TColumnWidth);
    procedure SetTitleLines(const Value: integer);
    procedure SetRowFixHeight(const Value: integer);
    procedure SetRowLines(const Value: integer);
    procedure SetRowVerticalAlign(const Value: TVerticalAlignment);
    procedure SetFieldsAsLink(Value: TStrings);
    procedure SetFieldsAsBoolean(Value: TStrings);
    procedure SetFieldsAsRichText(Value: TStrings);
    procedure SetFieldsAsGraphics(Value: TStrings);
    procedure SetFieldsAsDate(Value: TStrings);
    procedure SetFieldsAsTime(Value: TStrings);
    procedure SetFieldsAsColor(Value: TStrings);
    procedure SetFieldsAsPassword(Value: TStrings);
    procedure SetFieldsAsDataBar(Value: TStrings);
    procedure SetFieldsAsStatusBar(Value: TStrings);
    procedure SetFieldsAutoPickList(Value: TStrings);
    procedure SetFieldsDrawDropDown(Value: TStrings);
    procedure SetFieldsFilterButton(Value: TStrings);
    procedure SetFieldsFilterMapping(Value: TStrings);
    procedure SetFieldsFixedColumnWidth(Value: TStrings);
    procedure SetTitlePopupItems(Value: TStrings);
    procedure SetTitleImages(Value: TStrings);
    procedure SetTitleHints(Value: TStrings);
    procedure SetDataImages(Value: TStringList);
    procedure DataImagesChanged(Sender: TObject);
    procedure SetFieldsGroups(Value: TStringList);
    procedure FieldsGroupsChanged(Sender: TObject);
    procedure SetPropSaveFile(Value: TPropSaveFile);
    procedure SetTitleExPopupMenu(Value: TPopupMenu);
    procedure SetTitleImageList(Value: TImageList);
    procedure SetDataImageList(Value: TImageList);
    procedure SetColorOddRow(Value: TColor);
    procedure SetColorActiveRow(Value: TColor);
    procedure SetColorHotTrackRow(Value: TColor);
    procedure SetFontColorFiltered(Value: TColor);
    procedure SetFilteredFieldNames(Value: string);
    procedure SetSorter(Value: TrDBGridCustomSorter);
    procedure SetMouseOverRow(const Value: integer);
    function SetFixedCols(Value: integer): integer;
    function GetFixedCols: integer;
{$IFDEF UseObsoleteOption}
    procedure CopyOptionsEx_oldToOptionsEx2;
{$ENDIF}
    procedure FixCellRectHiddenColumns(var Rect: TRect; ACol: integer);
    procedure FixCellRectColRowLines(var Rect: TRect);
    function ClientRectToParentRect(R: TRect): TRect;
    function TitleRectToTextRect(RCell: TRect): TRect;
    procedure InvalidateRowEx(ARow: Longint);
    procedure InvalidateHotTrackCell;
    procedure ResetHotTrackCell;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMHintShow(var Message: TCMHintShow); message CM_HINTSHOW;
    procedure CMMouseLeave(var AMsg: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure FOnTimer(Sender: TObject);
    function CheckCoord(Coord1, Coord2: TGridCoord): boolean;
    function CheckCoordXY(Coord: TGridCoord; X, Y: integer): boolean;
    function HasParentNormalState: boolean;
    procedure CheckSelectedCol;
    function CanEdit: boolean;
    function GetMouseOverCellHint(var C: TColumn; var Text: string; var AlwaysShow: boolean): boolean;
    function IsColumnReadOnly(C: TColumn): boolean;
    function IsLookupFilterColumn(C: TColumn): boolean;
    function IsBoolEx(F: TField): boolean;
    function IsMemoEx(F: TField; TextAsMemo: boolean): boolean;
    function IsMemoAsGlyph(F: TField; TextAsMemo: boolean = false): boolean;
    function IsMemoAsText(F: TField; TextAsMemo: boolean = false): boolean;
    function IsRichTextEx(F: TField): boolean;
    function IsGraphicEx(F: TField): boolean;
    function IsFieldAsDateEx(F: TField): boolean;
    function IsDateEx(F: TField): boolean;
    function IsFieldAsTimeEx(F: TField): boolean;
    function IsTimeEx(F: TField): boolean;
    function IsColorEx(F: TField): boolean;
    function IsLinkEx(F: TField): boolean;
    function IsPasswordEx(F: TField): boolean;
    function IsDataBarEx(F: TField): boolean;
    function IsStatusBarEx(F: TField): boolean;
    function IsDrawDropDownEx(F: TField): boolean;
    function IsDataImageEx(F: TField): boolean;
    function GetDataImageDef(F: TField; var DataImageDef: TDataImageDef): boolean;
    function IsTitleImageEx(F: TField): boolean;
    function GetTitleImageIndex(F: TField; var ImgIdx: integer): boolean;
    function IsFixedColumnWidth(F: TField; FieldName: string): boolean;
    function ChangeBool(reset: boolean): boolean;
    function GetDateValueEx(F: TField): TDateTime;
    procedure SetDateValueEx(F: TField; Value: TDateTime);
    function GetTimeValueEx(F: TField): TDateTime;
    procedure SetTimeValueEx(F: TField; Value: TDateTime);
    // Procedures return dynamic array of sorting FieldName, Directions
    function GetSortDirections(const DirectionList: string): TSortDirections;
    function GetSortFieldNames(const FieldList: string): TSortFieldNames;
    function TrimSortFields(const ASortFieldNames: string): string;
    function TrimSortDirections(const ASortDirections: string): string;
    procedure CheckFormatSortDirection(var ASortFieldNames, ASortDirections: string; FieldToRemove: string);
    function ChangeSortDirection(const FieldName: string): string;
    // footer
    procedure FooterScrolled(Sender: TObject);
    procedure UpdateFooterGrid;
    // grouping func.
    function GetGroupID(GroupDef: string): integer;
    function GetFldGroupInfo(Fld: TField; SkipCheckFldGroupOption: boolean = false): TGroupInfo;
    function GetColGroupInfo(ColIdx: integer; FindFirstVisibleCol: boolean): TGroupInfo;
    function IsGroupMasterExpanded(ColIdx: integer): boolean;
    procedure MoveGroupedCol(FromIndex, ToIndex: Longint);
    procedure IntExpandCollapseGroup(GroupID: integer; Expand: boolean);
    procedure IntExpandCollapseAll(Expand: boolean);
    function GetGroupBtnRect(var RCell: TRect): TPoint;
    function PointInGroupBtnRect(RCell: TRect; Point: TPoint; ColIdx: integer; var GroupInfo: TGroupInfo): boolean;
    function GetGroupMasterTitleContent(Fld: TField; var Text: string): boolean;
    function GetGroupMasterCellContent(Fld: TField; var Text: string): boolean;
    function GetLineWidth(const s: string; AFont: TFont = nil): integer;
    function GetLine(var s: string; W: integer; IgnoreWidth: boolean = false): string;
    function GetTitleWidth(Column: TColumn): integer;
    function GetLimitColumnWidth(W: integer; LimitMin, LimitMax: boolean): integer;
    function CheckDropDownEdit: boolean;
    procedure IntSetColumnWidth(Column: TColumn; NewWidth: integer);
    // summary
    procedure InitSummaryResults(var R: TSummaryResults);
    function GetSummaryResults(F: TField; Recalculate: boolean): TSummaryResults;
    function ReplaceSummaryTags(var s: string; F: TField; R: TSummaryResults; ForFooter: boolean): string;
    // filter func
    function GetFilterPropsIdx(Fld: TField): integer;
    function IsFilterColumn(ColIdx: integer; AllowLookupField: boolean): boolean;
    function IsFunnelButtonColumn(ColIdx: integer): boolean;
    function IsFiterTextBarColumn(ColIdx: integer): boolean;
    function IsFilterFieldActive(FieldName: string): boolean;
    function CheckColWidthForFunnelButton(W: integer): boolean;
    function GetFilterComboRect(RCell: TRect; UseMargin: boolean): TRect;
    function PointInFilterComboRect(RCell: TRect; Point: TPoint; DataCol: integer; UseMargin: boolean): boolean;
    function PointInFilterFunnelRect(RCell: TRect; Point: TPoint; DataCol: integer): boolean;
    procedure FilterComboRefreshBounds;
    procedure FilterComboRefreshColor(Status: string);
    procedure FilterComboMoveToNextCol(Backward: boolean);
    procedure FilterComboShow(REdit: TRect; ColumnIndex: integer);
    procedure FilterComboHide(ClearAllValues, ClearAllFilters: boolean);
    procedure FilterComboExit(Sender: TObject);
    procedure FilterComboChange(Sender: TObject);
    procedure FilterComboKeyPress(Sender: TObject; var Key: char);
    procedure FilterComboKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FilterComboUpdateTimer(Sender: TObject);
    procedure FilterListLoadValues;
  protected
    FScaleRatio: double;
    FRichText: TrRichEdit;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    function TitleRow: integer;
    function FixCol: integer;
    procedure Loaded; override;
    procedure SetDefaultTitlePopup;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure SetParent(AParent: TWinControl); override;
{$IFNDEF TrPopupMonthCalendar}
    function CreateDateTimePicker: TWinControl; virtual;
    procedure SetDateTimePickerDate(ADateTime: TDateTime); virtual;
    function GetDateTimePickerDate: TDateTime; virtual;
    procedure ShowDateTimePicker(Rect: TRect); virtual;
    procedure DateTimePickerClose(Sender: TObject);
    procedure DateTimePickerChange(Sender: TObject);
{$ENDIF}
    function CreateEditor: TInplaceEdit; override;
    function GetEditStyle(ACol, ARow: Longint): TEditStyle; override;
    function CanEditModify: boolean; override;
    function CanEditShow: boolean; override;
    procedure SetEditText(ACol, ARow: Longint; const Value: string); override;
    function GetEditText(ACol, ARow: Longint): string; override;
    function GetClientRect: TRect; override;
    procedure LayoutChanged; override;
    procedure SetColumnAttributes; override;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure ColEnter; override;
    procedure ColExit; override;
    procedure CalcSizingState(X, Y: integer; var State: TGridState; var Index: Longint; var SizingPos, SizingOfs: integer; var FixedInfo: TGridDrawInfo);
      override;
    procedure ColWidthsChanged; override;
    procedure ColumnMoved(FromIndex, ToIndex: Longint); override;
    function SelectCell(ACol, ARow: Longint): boolean; override;
    function HighlightCell(DataCol, DataRow: integer; const Value: string; aState: TGridDrawState): boolean; override;
    procedure Paint; override;
    procedure DrawRectBidi(Rect: TRect);
    function DrawTextAligned(Rect: TRect; Text: string; Format: integer; VertAlignment: TVerticalAlignment = taAlignTop; TruncatedText: string = '';
      FixedWordWrap: boolean = false): boolean;
    procedure DrawCell(ACol, ARow: Longint; aRect: TRect; aState: TGridDrawState); override;
    procedure DrawCellBackground(const aRect: TRect; AColor: TColor; aState: TGridDrawState; ACol, ARow: integer); override;
    procedure DrawColumnCell(const Rect: TRect; DataCol: integer; Column: TColumn; State: TGridDrawState); override;
    procedure DrawTitleGridLines(ACol, ARow: Longint; aRect: TRect; aState: TGridDrawState; InsideGroup: boolean);
    procedure DrawDataGridLines(const Rect: TRect; DataCol: integer; Column: TColumn; State: TGridDrawState);
    procedure DrawRichText(aRect: TRect; Text: string; BkColor: TColor); virtual;
    procedure CellClick(Column: TColumn); override;
    procedure TitleClick(Column: TColumn); override;
    procedure TitleExClick(Button: TMouseButton; Shift: TShiftState); virtual;
    procedure TitlePopupClick(Sender: TObject);
    procedure FixColClick(Button: TMouseButton; Shift: TShiftState); virtual;
    procedure Click; override;
    procedure DblClick; override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: char); override;
    procedure ProcessGridSearchKey(var Key: char); virtual;
    procedure ProcessLookupSearchKey(var Key: char); virtual;
    procedure ProcessPicklistSearchKey(var Key: char); virtual;
    function SearchNext(Previous: boolean): boolean;
    function SearchGetFilterString(Fld: TField): string;
    function SearchGridFilter(ResetFilter: boolean): boolean;
    function SearchLookupFilter(ResetFilter: boolean): boolean;
    procedure StopSearching;
    function IsSearching: boolean;
    procedure ShowIntHint(Hint: string; TopLeft: TPoint; UseDelay: boolean);
    procedure Scroll(Distance: integer); override;
    procedure UpdateScrollBar; override;
    procedure DoContextPopup(MousePos: TPoint; var Handled: boolean); override;
    procedure LinkActive(Value: boolean); override;
    procedure EditButtonClick; override;
    procedure TopLeftChanged; override;
    function RowIsMultiselected(FromActiveRecord: boolean; FromDataRow: integer): boolean; virtual;
    procedure LoadColorPickList;
    function IntCopyData(ToClipboard, AllRec, AllFields, IncTitles, IncHidden: boolean; Fields: TStrings): string; virtual;
    procedure IntPasteData(FromCol: integer; AddAsNewRecs: boolean); virtual;
    procedure IntShowFilterDlg(ACol: integer); virtual;
    function CheckColumnDrag(var Origin, Destination: integer; const MousePt: TPoint): boolean; override;
    property PrevActiveBookmark: TBookmark read FPrevActiveBookmark;
    procedure DrawFooterRow;
    function MouseCoordEx(X, Y: integer): TGridCoord;
    property MouseOverRow: integer read FMouseOverRow write SetMouseOverRow;
    procedure DoRefreshFilteredFieldNames;
    procedure DoFilterChange;
    function IntSetFilterForField(Fld: TField; const FilterString: string; MultiFieldFilter: boolean; SetFilter: boolean = true; ShowError: boolean = true)
      : boolean; virtual;
    function DoGeneralEvent(EventType: TGeneralEventType): boolean;
    function DoGeneralEventField(EventType: TGeneralEventType; F: TField): boolean;
    function DoGeneralEventFieldBool(EventType: TGeneralEventType; F: TField; var BoolValue: boolean): boolean;
    function DoGeneralEventParam(EventType: TGeneralEventType; Column: TColumn; Field: TField; ParamObject: TObject; ParamValue: variant): boolean;
    // scan all proc and callback proc
    procedure ScanAllRecords(CallBackProc: TScanDatasetEvent; CallBackObject: TObject; ScanReason: integer; ProcAllRec: boolean = true;
      MaxRecords: integer = 0);
    procedure CBProc_OneColumnWidth(Sender: TObject; DataSet: TDataSet; RecNum: integer);
    procedure CBProc_ColumnWidth(Sender: TObject; DataSet: TDataSet; RecNum: integer);
    procedure CBProc_RecalculateSummaryResults(Sender: TObject; DataSet: TDataSet; RecNum: integer);
    procedure CBProc_LoadAutoPickList(Sender: TObject; DataSet: TDataSet; RecNum: integer);
    procedure CBProc_FilterListLoadValues(Sender: TObject; DataSet: TDataSet; RecNum: integer);
    procedure CBProc_IntCopyData(Sender: TObject; DataSet: TDataSet; RecNum: integer);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure Invalidate; override; {< @exclude }
    procedure Assign(Source: TPersistent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { indicates whether Grid is linked to any Dataset }
    function IsDatasetLinked: boolean;
    { indicates whether linked Dataset is active }
    function IsActive: boolean;
    { indicates whether linked Dataset has active filter }
    function IsFiltered: boolean;
    { indicates whether Dataset is empty }
    function IsEmpty: boolean;
    { indicates whether Dataset is not empty }
    function IsNotEmpty: boolean; deprecated 'Use IsEmpty instead.';
    { indicates whether Field is shown as boolean field }
    function IsFieldAsBoolean(F: TField): boolean;
    { return boolean value of field }
    function FieldAsBooleanValue(F: TField): boolean;
    { force LayoutChange event }
    procedure RefreshLayout;
    { returns grid coord of last Mouse down (click or dblclick) }
    property MouseDownCoord: TGridCoord read MDownCoord;
    { indicates current search text for Grid search or Lookup or PickList search }
    property SearchText: string read FSearchText write FSearchText;

    { manually resort grid - calls Sorter DoSort procedure or OnChangeSortEx or OnChangeSort event }
    function ReSort: boolean;
    { set new sort properties and call Sorter DoSort procedure or OnChangeSort procedure,
      obsolete, use ChangeSortEx for multifield sorting }
    function ChangeSort(const NewSortFieldName: string; NewSortDesc: boolean): boolean;
    { set new sort properties and call Sorter DoSort procedure or OnChangeSortEx procedure }
    function ChangeSortEx(const NewSortFieldName, NewSortDirections: string): boolean;
    { returns position of SelFldName in SortFieldName (0=not found) }
    function PosInSortField(const FieldName: string): integer;
    { returns sort direction for field (true=DESC, false=ASC) }
    function GetSortDirection(const FieldName: string): boolean;
    { returns array of sort fields }
    function SortFieldNamesArr: TSortFieldNames;
    { returns array of sort directions }
    function SortDirectionsArr: TSortDirections;

    { change all column widths according ColumnWidth parameter and increase width by Additional space }
    procedure SetColumnWidth(ColumnWidth: TColumnWidth; AdditionalSpace: integer = 0);
    { changes one column width according ColumnWidth parameter }
    procedure SetOneColumnWidth(ColumnWidth: TColumnWidth; Column: TColumn; AdditionalSpace: integer = 0);
    { changes all columns position, width and visibility according original loaded state }
    procedure RestoreColumnsWidth(RestoreVisibility: boolean = true; RestorePosition: boolean = true);

    { save internal position of selected row in grid, use RestoreRowPos to restore the same possition,
      if SetExUpdateLock=true then all drawing is blocked till RestoreRowPos is called }
    procedure SaveRowPos(SetExUpdateLock: boolean = false);
    { moved grid to same row with same dataset record as was saved by SaveRowPos }
    procedure RestoreRowPos;
    { scroll data inside the grid with keeping of active record }
    function ScrollData(ScrollBy: integer): integer;
    { store and reset all dataset and datasource events for quick movement throught all records }
    procedure SaveResetDatasetEvents(Reason: integer = 0);
    { restore all dataset and datasource events stored by SaveResetDatasetEvents }
    procedure RestoreDatasetEvents(Reason: integer = 0);
    { link another grid as footer DBGrid and set all properties }
    procedure SetFooterDBGrid(FooterDBGrid: TrDBGrid; const FooterFieldPrefix: string; ShowHSBarMaster, ShowHSBarFooter: boolean);

    { returns pointer to InplaceEditor }
    function GetInplaceEditor: TrDBGridInplaceEditEx;
    { hide InplaceEditor }
    procedure HideInplaceEditor;
    { stop or start updating of grid, used mainly for internal purposes }
    procedure StopUpdate(stop: boolean);
    { load PickList values for all AutoPickList fields,
      use OnLoadPickList event for loading by used defined query }
    procedure LoadAutoPickList;

    { return column index for specific field name }
    function GetFieldNameColumnIndex(const FieldName: string): integer;
    { returns index of left column }
    function GetLeftCol: integer;
    { set left column of grid (move grids scrollbar to right) }
    procedure SetLeftCol(ALeftCol: integer);
    { returns rect of active cell, can be used in some events }
    function GetActiveCellRect: TRect;
    { draw bitmap in cell rectangle }
    procedure DrawBMPInCell(Rect: TRect; Bmp: TBitmap);

    { show memo editor for the field }
    procedure ShowMemo(Field: TField);
    { show graphics editor for the field }
    procedure ShowGraphic(Field: TField);

    { Defines list of conditions defining cell colors }
    property CellParamsCondList: TCellParamsCondList read FCellParamsCondList;
    { evaluate conditions for selected field and active record and returns Font, Background and image definition }
    procedure GetCellParams(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef; ActiveRow, Highlight: boolean); virtual;
    { refresh grid colors after changing of CellParamCondList }
    procedure RefreshCellParamsCond;
    { return summary results for Field }
    function SummaryResults(Fld: TField; Recalculate: boolean = true): TSummaryResults; overload;
    { return summary results for Column }
    function SummaryResults(Column: TColumn; Recalculate: boolean = true): TSummaryResults; overload;
    { recalculate summary results for all field, paramater define that calculated result is used also for popup summary }
    procedure RecalculateSummaryResults(UseCalculatedSummaryForPopup: boolean = false);

    { get displayed title from the column with the field }
    function GetTitleText(Column: TColumn): string;
    { get displayed value from the column with the field }
    function GetCellValue(Fld: TField; UseDisplayText: boolean = true): string;
    { copy data from selected cell to clipboard }
    function CopyCellData(CopyToClipboard: boolean = true): string;
    { copy data from selected fields and all records to clipboard }
    procedure CopyFieldsData(FieldNames: TStrings; IncludeTitles: boolean = false);
    { copy data from active field and all records to clipboard }
    procedure CopyActiveFieldData(IncludeTitles: boolean = false);
    { copy data from active record and all fields to clipboard }
    procedure CopyActiveRecordData(IncludeTitles: boolean = false; IncludeHidden: boolean = false);
    { copy data from dataset to clipboard }
    procedure CopyAllData(IncludeTitles: boolean = false; IncludeHidden: boolean = false);
    { paste data from clipboard to dataset }
    procedure PasteData(AddAsNewRecs: boolean);
    { copy data from selected cell to next/previous record }
    function FillCellData(FillUp: boolean = false; SpecificFieldNames: string = ''): boolean;

    { returns GroupInfo record for field }
    function GetFieldGroupInfo(Fld: TField): TGroupInfo;
    { indicates whether master field of the same group as Fld is visible }
    function IsFieldGroupMasterVisible(Fld: TField): boolean;
    { indicates whether master field of the same group as Fld is expanded }
    function IsFieldGroupMasterExpanded(Fld: TField): boolean;
    { rearrange all grouped columns to be together = should be called after user defined column moving }
    procedure FixedGroupedCol(SwitchGroupOn: boolean; RestoreDefaults: boolean = false);
    { collapse one group defined by GroupID }
    procedure CollapseGroup(GroupID: integer); overload;
    { collapse one group defined by Field }
    procedure CollapseGroup(Fld: TField); overload;
    { expand one group defined by GroupID }
    procedure ExpandGroup(GroupID: integer); overload;
    { expand one group defined by Field }
    procedure ExpandGroup(Fld: TField); overload;
    { collapse all groups }
    procedure CollapseAll;
    { expand all groups }
    procedure ExpandAll;
    { define number of rows for picklist }
    property DropDownRowsCount: cardinal read FDropDownRowsCount write FDropDownRowsCount;

    { show funnel filter dialog for active or given column (deprecated 'Use ShowFunnelFilter instead.') }
    procedure ShowFilterDlg(Column: TColumn = nil); deprecated 'Use ShowFunnelFilter instead.';
    { show funnel filter dialog for active or given column }
    procedure ShowFunnelFilter(Column: TColumn = nil);
    { show text filter box for active or given column }
    procedure ShowTextFilter(Column: TColumn = nil);
    { show text filter box for active or given column and set filter text }
    procedure SetTextFilter(Column: TColumn; FilterText: string; ActivateFilter: boolean = true);
    { Returns TrComboBoxEx for setting of their props }
    property FilterCombo: TrComboBoxExFilter read FFilterCombo;
    { array of column filter }
    property ColumnFilters: TColumnStoredProps read FColumnFilters;
    { return actual filter string defined by ColumnFilters }
    function GetColumnFilterString: string;
    { set filter string for particular field }
    function SetFilterForField(Fld: TField; const FilterString: string; MultiFieldFilter: boolean; SetFilter: boolean = true; ShowError: boolean = true)
      : boolean;
    { define list of field which is used in filter = show active filter glyph }
    property FilteredFieldNames: string read FFilteredFieldNames write SetFilteredFieldNames;
    { reset filter and filter bar }
    procedure ResetFilter;
    { refresh list of field which is used in filter - loaded directly from filter string }
    procedure RefreshFilteredFieldNames; deprecated 'Use FilterChangedNotification instead.';
    { notify rDBGrid that filter was changed externally }
    procedure FilterChangedNotification;

    { return DataImages settings for one field }
    function DataImages_GetSettings(FieldName: string; ConditionList: TStrings; var AddImageAlign: string): integer;
    { set or update DataImages settings for one field }
    procedure DataImages_SetSettings(FieldName: string; ConditionList: TStrings; AddImageAlign: string = '');
    { return list of group definitions from FieldGroup }
    function Fields_GetGroupList(GroupList: TStrings): integer;
    { return group definition from FieldGroup for one GroupID }
    function Fields_GetGroupDefinition(GroupID: integer): string;
    { return FieldGroup settings for one GroupDefinition }
    function Fields_GetGroupSettings(GroupDefinition: string; var GroupID: integer; var GroupTitle, ContentCollapsed: string;
      var TitleFontColor, GroupLineColor: TColor; var GroupLocked: boolean; FieldList: TStrings): boolean;
    { set or update FieldGroup settings for one GroupID }
    function Fields_SetGroupSettings(GroupID: integer; GroupTitle, ContentCollapsed: string; TitleFontColor, GroupLineColor: TColor; GroupLocked: boolean;
      FieldList: TStrings): string;
    { return FieldsAsLink settings for one field }
    function Fields_GetLinkType(FieldName: string): TrLinkType;
    { set or update FieldsAsLink settings for one field }
    procedure Fields_SetLinkType(FieldName: string; LinkType: TrLinkType);
    { return FieldsAsBoolean settings for one field }
    function Fields_GetBoolValues(FieldName: string): string;
    { set or update FieldsAsBoolean settings for one field }
    procedure Fields_SetBoolValues(FieldName: string; BoolValues: string);
    { return FieldsAsDataBar settings for one field }
    function Fields_GetDataBarSettings(FieldName: string; var MinValue, MaxValue: double; var ValueType: TrDataBarText; var ColorFrom, ColorTo: TColor)
      : boolean;
    { set or update FieldsAsDataBar settings for one field }
    function Fields_SetDataBarSettings(FieldName: string; MinValue, MaxValue: double; ValueType: TrDataBarText = dbtPercent; ColorFrom: TColor = clSkyBlue;
      ColorTo: TColor = clWhite): string;
    { return FieldsAsStatusBar settings for one field }
    function Fields_GetStatusBarSettings(FieldName: string; var Steps: integer; var BarType: TrStatusBarType; var ColorAll, ColorDone: TColor): boolean;
    { set or update FieldsAsStatusBar settings for one field }
    function Fields_SetStatusBarSettings(FieldName: string; Steps: integer; BarType: TrStatusBarType = sbtProgressBar; ColorAll: TColor = clNone;
      ColorDone: TColor = clNone): string;
    { return TitleImages settings for one field }
    function Title_GetImageIndex(FieldName: string): integer;
    { set or update TitleImages settings for one field }
    procedure Title_SetImageIndex(FieldName: string; ImageIndex: integer);
    { return TitleHint settings for one field }
    function Title_GetHint(FieldName: string): string;
    { set or update TitleHint settings for one field }
    procedure Title_SetHint(FieldName: string; Hint: string);
    { return FooterRow.FieldFooterDefs settings for one field }
    function FooterRow_GetSettings(FieldName: string; var ContentMask: string; var TextAlignment: TAlignmentDefault; var FontStyles: TFontStylesDefault;
      var FontColor: TColor; var FontSize: integer; AFont: TFont = nil): boolean;
    { set or update FooterRow.FieldFooterDefs settings for one field }
    function FooterRow_SetSettings(FieldName: string; ContentMask: string; TextAlignment: TAlignmentDefault = tadDefault; FontStyles: TFontStylesDefault = [];
      FontColor: TColor = clNone; FontSize: integer = 0): string;
  published
{$IFDEF UseObsoleteOption}
    { list of new options }
    property OptionsEx: TOptionsEx read FOptionsEx_old write SetOptionsEx default[dgTitleExBtn, dgTitleBtn, dgFixColBtn, dgBool, dgMemo, dgGraphic,
      dgDateTimePicker, dgUseTitlePopup, dgAutoPickListWidth];
{$ENDIF}
    { new version of new options setting, replace the obsolete OptionsEx }
    property OptionsEx2: TOptionsEx2 read FOptionsEx2 write FOptionsEx2;

    { define link to PropSaveFile which is used for saving of possition of dialogs }
    property PropSaveFile: TPropSaveFile read FPropSaveFile write SetPropSaveFile;
    { defines link to rCustomGlyphs component which can define all glyphs }
    property CustomGlyphs: TrCustomGlyphs read FCustomGlyphs write SetCustomGlyphs;
    { Defines look of glyphs for memo and graphics fields }
    property GlyphLookStyle: TGlyphLookStyles read FGlyphLookStyle write SetGlyphLookStyle default glsStandard;
    { Defines used color theme (gradient, backgrounf, fonts) }
    property GridLookStyle: TGridLookStyles read FGridLookStyle write SetGridLookStyle default grs_Default;
    { Defines color of each odd row, works only for bidirectional dataset }
    property ColorOddRow: TColor read FColorOddRow write SetColorOddRow default clMoneyGreen;
    { Defines color of active row }
    property ColorActiveRow: TColor read FColorActiveRow write SetColorActiveRow default clInfoBk;
    { Defines color of hot track (mouse over) row }
    property ColorHotTrackRow: TColor read FColorHotTrackRow write SetColorHotTrackRow default clNone;
    { Defines color of required field (for active or all rows) }
    property ColorRequiredField: TColor read FColorRequiredField write FColorRequiredField default clNone;
    { Defines text color when dataset is filtered }
    property FontColorFiltered: TColor read FFontColorFiltered write SetFontColorFiltered default clNavy;
    { Defines text color of link cells (http, mail) }
    property FontColorLink: TColor read FFontColorLink write FFontColorLink default clHighlight;
    { Defines text color when field or column is readonly during editing }
    property FontColorReadOnly: TColor read FFontColorReadOnly write FFontColorReadOnly default clGrayText;

    { Defines which columns has sort marker, supports MultiField (e.g. FirtsName,LastName) }
    property SortFieldName: string read FSortFieldName write SetSortFieldName;
    { Defines direction of sort marker }
    property SortDesc: boolean read FSortDesc write SetSortDesc default false;
    { Defines directions in case of multifield, A=Asc, D=Desc, e.g. A,D,A }
    property SortDirections: string read FSortDirections write SetSortDirections;
    { Defines sorter component - descendant of TrDBGridCustomSorter }
    property Sorter: TrDBGridCustomSorter read FSorter write SetSorter;

    { Defines type of automatic changing of column width }
    property ColumnWidth: TColumnWidth read FColumnWidth write SetFColumnWidth default cwNone;
    { Defines type of automatic changing of column width after double click on column border in title row }
    property ColumnWidthAfterDblClick: TColumnWidth read FColumnWidthAfterDblClick write FColumnWidthAfterDblClick default cwDataLabelWidth;

    { Defines minimum size of string field for using a memo dialog for editation }
    property TextAsMemoSize: integer read FTextAsMemoSize_old write FTextAsMemoSize_old default 0;
    { Defines in which situation lookup is started after keypress }
    property LookupCondition: TLookupCondition read FLookupCondition write FLookupCondition default lcBrowse;
    { Defines number of lines in title row }
    property TitleLines: integer read FTitleLines write SetTitleLines default 1;
    { Defines fixed height of data rows }
    property RowFixHeight: integer read FRowFixHeight write SetRowFixHeight default 0;
    { Defines number of lines in data row }
    property RowLines: integer read FRowLines write SetRowLines default 1;
    { Defines vertical alignment for data row - used only if RowLines>1 }
    property RowVerticalAlign: TVerticalAlignment read FRowVerticalAlign write SetRowVerticalAlign default taAlignTop;
    { define properties for text in title line, see TTitleLineText }
    property TitleLineText: TTitleLineText read FTitleLineText write FTitleLineText;
    { define properties for text in fixed column, see TFixedColText }
    property FixedColText: TFixedColText read FFixedColText write FFixedColText;
    { define properties for footer row, see TFooterRow }
    property FooterRow: TFooterRow read FFooterRow write FFooterRow;
    { Defines char used as mask for password fields }
    property PasswordChar: char read FPasswordChar write FPasswordChar default '*';
    { Defines list of fields as link with link type, can be set easily by design time dialog }
    property FieldsAsLink: TStrings read FFieldsAsLink write SetFieldsAsLink;
    { Defines list of fields as boolean fields with TRUE FALSE values, can be set easily by design time dialog }
    property FieldsAsBoolean: TStrings read FFieldsAsBoolean write SetFieldsAsBoolean;
    { Defines list of fields as boolean fields with TRUE FALSE values, can be set easily by design time dialog }
    property FieldsAsRichText: TStrings read FFieldsAsRichText write SetFieldsAsRichText;
    { Defines list of fields as graphics fields, can be set easily by design time dialog }
    property FieldsAsGraphics: TStrings read FFieldsAsGraphics write SetFieldsAsGraphics;
    { Defines list of fields displayed as date fields, can be set easily by design time dialog }
    property FieldsAsDate: TStrings read FFieldsAsDate write SetFieldsAsDate;
    { Defines list of fields displayed as time fields, can be set easily by design time dialog }
    property FieldsAsTime: TStrings read FFieldsAsTime write SetFieldsAsTime;
    { Defines list of fields displayed as color fields, can be set easily by design time dialog }
    property FieldsAsColor: TStrings read FFieldsAsColor write SetFieldsAsColor;
    { Defines list of fields displayed and edited as password = hide chars }
    property FieldsAsPassword: TStrings read FFieldsAsPassword write SetFieldsAsPassword;
    { Defines list of fields displayed as data bar }
    property FieldsAsDataBar: TStrings read FFieldsAsDataBar write SetFieldsAsDataBar;
    { Defines list of fields displayed as status bar }
    property FieldsAsStatusBar: TStrings read FFieldsAsStatusBar write SetFieldsAsStatusBar;
    { Defines list of fields with autopicklist function, can be set easily by design time dialog }
    property FieldsAutoPickList: TStrings read FFieldsAutoPickList write SetFieldsAutoPickList;
    { Defines list of fields with custom draw event of dropdown list }
    property FieldsDrawDropDown: TStrings read FFieldsDrawDropDown write SetFieldsDrawDropDown;
    { Defines list of fields with filter button in Title Row (if dgFilterButtons is true) }
    property FieldsFilterButton: TStrings read FFieldsFilterButton write SetFieldsFilterButton;
    { Defines mapping of the visible field with text filter (e.g. lookup field) to another field used for the filter condition }
    property FieldsFilterMapping: TStrings read FFieldsFilterMapping write SetFieldsFilterMapping;
    { Defines list of fields with fixed column width, width is not changed by ColumnWidth property }
    property FieldsFixedColumnWidth: TStrings read FFieldsFixedColumnWidth write SetFieldsFixedColumnWidth;

    { Defines Popup Items for calculation of numeric fields, see rstring.pas for default values }
    property TitlePopupItems: TStrings read FTitlePopupItems write SetTitlePopupItems;
    { Defines Popup Items for left top cell }
    property TitleExPopupMenu: TPopupMenu read FTitleExPopupMenu write SetTitleExPopupMenu;
    { Defines title hints for each field (column), can be set easily by design time dialog }
    property TitleHints: TStrings read FTitleHints write SetTitleHints;
    { link to ImageList for images in title rows }
    property TitleImageList: TImageList read FTitleImageList write SetTitleImageList;
    { defines images in title row, can be set easily by design time dialog }
    property TitleImages: TStrings read FTitleImages write SetTitleImages;
    { link to ImageList for images in data cells }
    property DataImageList: TImageList read FDataImageList write SetDataImageList;
    { defines images in data cells, can be set easily by design time dialog }
    property DataImages: TStringList read FDataImages write SetDataImages;
    { defines field groups, can be set easily by design time dialog }
    property FieldsGroups: TStringList read FFieldsGroups write SetFieldsGroups;

    { defines absolute item height of lookup list, default is used if LookupItemHeight=0 }
    property LookupListItemHeight: integer read FLookupListItemHeight write FLookupListItemHeight default 0;
    { defines increase of standard item height of dropdown list }
    property LookupListItemIncreaseHeight: integer read FLookupListItemIncreaseHeight write FLookupListItemIncreaseHeight default 2;
    { defines increase of item column width }
    property LookupListItemIncreaseWidth: integer read FLookupListItemIncreaseWidth write FLookupListItemIncreaseWidth default 0;
{$IFNDEF TrPopupMonthCalendar}
    { Returns TDateTimePicker for setting of their props }
    property DateTimePicker: TWinControl read FDateTimePicker;
    { event type for creating of DateTimePicker }
    property DateTimePickerCreate: TDateTimePickerCreate read FDateTimePickerCreate write FDateTimePickerCreate;
    { event type for DateTimePicker action }
    property DateTimePickerEvent: TDateTimePickerEvent read FDateTimePickerEvent write FDateTimePickerEvent;
{$ENDIF}
    { event is called before editing of cell, CanEdit define possibility to editation }
    property OnCanEdit: TCanEditEvent read FOnCanEdit write FOnCanEdit;
    { event is called after click on title row to provide sorting if Sorter component is not used;
      if "Accept" remains true, "SortFieldName" and "SortDesc" is changed
      and sort marker is shown, in other case nothing is changed;
      obsolete, use ChangeSortEventEx for multifield sorting }
    property OnChangeSort: TChangeSortEvent read FOnChangeSort write FOnChangeSort;
    { Enhanced event with MultiField sorting supprt;
      event is called after click on title row to provide sorting if Sorter component is not used;
      if "Accept" remains true, "SortFieldName" and "SortDirections" is changed
      and sort marker is shown, in other case nothing is changed }
    property OnChangeSortEx: TChangeSortEventEx read FOnChangeSortEx write FOnChangeSortEx;
    { event is called after click to user_defined hyperlink in the cell }
    property OnLinkClick: TLinkClickEvent read FOnLinkClick write FOnLinkClick;
    { event is called after click to checkbox in then cell }
    property OnBoolClick: TLinkClickEvent read FOnBoolClick write FOnBoolClick;
    { event is called after click to title ex (cell 0,0) }
    property OnTitleExClick: TNotifyEvent read FOnTitleExClick write FOnTitleExClick;
    { event is called after click to fixed col cell (column 0) }
    property OnFixColClick: TNotifyEvent read FOnFixColClick write FOnFixColClick;
    { event is called after click to footer cell }
    property OnFooterCellClick: TFooterCellClick read FOnFooterCellClick write FOnFooterCellClick;
    { event is called to get font and text of fixed column cell }
    property OnGetFixedColState: TGetFixedColStateEvent read FOnGetFixedColState write FOnGetFixedColState;
    { event is called to get font and text of fixed row cell (title cell) }
    property OnGetFixedRowState: TGetFixedRowStateEvent read FOnGetFixedRowState write FOnGetFixedRowState;
    { event is called to get title line text or custom draw of title line directly }
    property OnDrawTitleLine: TDrawTitleLineEvent read FOnDrawTitleLine write FOnDrawTitleLine;
    { event is called to draw cell background directly }
    property OnDrawCellBackground: TDrawCellEvent read FOnDrawCellBackground write FOnDrawCellBackground;
    { event is called to add any custom draw of title cell directly }
    property OnDrawTitleCellEnd: TDrawCellEvent read FOnDrawTitleCellEnd write FOnDrawTitleCellEnd;
    { event is called to add any custom draw of data cell directly }
    property OnDrawColumnCellEnd: TDrawColumnCellEvent read FOnDrawColumnCellEnd write FOnDrawColumnCellEnd;
    { event is called after cell draw to custom draw of focus rectange }
    property OnDrawActiveRowFocus: TDrawColumnCellEvent read FOnDrawActiveRowFocus write FOnDrawActiveRowFocus;
    { event is called to get params of data cell }
    property OnGetCellParams: TGetCellParamsEvent read FOnGetCellParams write FOnGetCellParams;
    { event is called to get params of data cell }
    property OnGetCellParamsEx: TGetCellParamsEventEx read FOnGetCellParamsEx write FOnGetCellParamsEx;
    { event is called to draw whole footer row }
    property OnDrawFooterRow: TDrawFooterRowEvent read FOnDrawFooterRow write FOnDrawFooterRow;
    { event is called to draw one cell of footer row }
    property OnDrawFooterCell: TDrawFooterCellEvent read FOnDrawFooterCell write FOnDrawFooterCell;
    { event is called to load picklist for particular field - usually by aggregated query }
    property OnLoadPickList: TLoadPickListEvent read FOnLoadPickList write FOnLoadPickList;
    { event is called to load filter bar list for particular field - usually by aggregated query }
    property OnLoadFilterList: TLoadPickListEvent read FOnLoadFilterList write FOnLoadFilterList;
    { event is called to load funnel-filter list for particular field - usually by aggregated query }
    property OnLoadFunnelFilter: TLoadFunnelFilterEvent read FOnLoadFunnelFilter write FOnLoadFunnelFilter;
    { event is called before showing of hint window to get user defined hint string }
    property OnGetHintStringEx: TGetHintStringEventEx read FOnGetHintStringEx write FOnGetHintStringEx;
    { event is called to set all calculated values for one column (field), usually by some SQL query instead of record scan }
    property OnRecalculateSummaryResult: TRecalculateSummaryResultEvent read FOnRecalculateSummaryResult write FOnRecalculateSummaryResult;
    { event is called to update or format one particular calculated value of field (sum, min, max, avg ...) }
    property OnGetSummaryResult: TGetSummaryResultEvent read FOnGetSummaryResult write FOnGetSummaryResult;
    { event is called to custom draw of PickList dropdowm item, FieldsDrawDropDown defines fields with dropdowm custom drawing }
    property OnPickListDrawItem: TDrawItemEvent read FOnPickListDownItem write FOnPickListDownItem;
    { event is called to get PickList dropdowm item height, FieldsDrawDropDown defines fields with dropdowm custom drawing }
    property OnPickListMeasureItem: TMeasureItemEvent read FOnPickListMeasureItem write FOnPickListMeasureItem;
    { event is called to set size of PickList dropdown box }
    property OnPickListSetSize: Controls.TCanResizeEvent read FOnPickListSetSize write FOnPickListSetSize;
    { event is called to set Lookup dropdown item height }
    property OnGetLookupListItemHeight: TGetLookupListItemHeight read FGetLookupListItemHeight write FGetLookupListItemHeight;
    { event is called to set Lookup dropdown item style or custom draw item }
    property OnGetLookupListItemProps: TGetLookupListItemProps read FGetLookupListItemProps write FGetLookupListItemProps;
    { event is called to define format of Ellipsis button in the cell }
    property OnGetEllipsisButtonFormat: TGetEllipsisButtonFormat read FOnGetEllipsisButtonFormat write FOnGetEllipsisButtonFormat;
    { event is called after horizonal scrolling (first visible column is changed) }
    property OnTopLeftChanged: TNotifyEvent read FOnTopLeftChanged write FOnTopLeftChanged;
    { event is called after user scroll to another record }
    property OnDataScrolled: TNotifyEvent read FOnDataScrolled write FOnDataScrolled;
    { event is called before dataset scan }
    property BeforeScanDataset: TScanDatasetEvent read FBeforeScanDataset write FBeforeScanDataset;
    { event is called after dataset scan }
    property AfterScanDataset: TScanDatasetEvent read FAfterScanDataset write FAfterScanDataset;
    { event is called after changing of column width }
    property OnColWidthChanged: TColWidthChangedEvent read FOnColWidthChanged write FOnColWidthChanged;
    { event is called to custom setting of text filter for the column }
    property OnSetColumnFilter: TSetColumnFilterEvent read FOnSetColumnFilter write FOnSetColumnFilter;
    { event is called when dataset filter is activated, deactivated or changed,
      working only with using of funnel filter or rDBAction filter,
      setting filter directly to dataset do not fire this event }
    property OnFilterChanged: TNotifyEvent read FFilterChanged write FFilterChanged;
    { event is not called by TrDBGrid, it is used by rDBAction only }
    property OnRefreshData: TNotifyEvent read FOnRefreshData write FOnRefreshData;
    { event is called in many events to add posibility of any custom code and to skip rest of the standard event }
    property OnGeneralEvent: TGeneralEvent read FOnGeneralEvent write FOnGeneralEvent;
  end;

  { define base Sorter component for dataset sorting;
    must be used as based class for any other dataset sorter }
  TrDBGridCustomSorter = class(TComponent)
  private
    FGoToFirstAfterSort: boolean;
    FGoToFirstAfterRefresh: boolean;
  protected
    SortFieldSeparator: string;
    SortStringSeparator: string;
    DescendingString: string;
    procedure BracketFieldName(var FieldName: string); virtual;
    function GetIndexNameHash(const IndexName: string; MaxLen: integer = 30): string;
    { virtual procedure called to set sorting properties in dataset, MUST be override to provide sorting }
    function SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string): boolean; virtual;
    { virtual procedure called to refresh dataset, can be override }
    procedure DatasetRequery(DataSet: TDataSet); virtual;
  public
    constructor Create(Owner: TComponent); override; {< @exclude }
    { virtual procedure provide sorting, calls SetSort_CreateIndex to sort dataset }
    function DoSort(Sender: TrDBGrid; DataSet: TDataSet; SortFieldNames: TSortFieldNames; SortDirections: TSortDirections): boolean; virtual;
    { virtual procedure provide refreshing of dataset, calls DatasetRequery }
    procedure DoRefresh(Sender: TrDBGrid; DataSet: TDataSet); virtual;
  published
    { define selecting of first record after sorting }
    property GoToFirstAfterSort: boolean read FGoToFirstAfterSort write FGoToFirstAfterSort default false;
    { define selecting of first record after refresh }
    property GoToFirstAfterRefresh: boolean read FGoToFirstAfterRefresh write FGoToFirstAfterRefresh default false;
  end;

  { find all rDBGrids on the form and call Resort procedure for each of them }
procedure ResortAllGrids(F: TForm; ToFirst: boolean); overload;
function GetFldListrDBGrid(rDBGrid: TrDBGrid; List: TStrings; FldsDef: TGetFldDefs): integer;

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rdlg, rtool, rdbtool, rpictool, rMessageDigest,
  rDBGrid_MemoForm, rDBGrid_GraphicForm, rDBGrid_FilterForm, PropSaveGrids;

const
  cFieldSep = ',';
  cFooterRowYCoord = -2;
  cTitleTextSeparators = ' ._-/\@#$&|';

type
  THackCustomGrid = class(TCustomGrid)
  end;

  THackGridDataLink = class(TGridDataLink)
  end;

  THackCustomListBox = class(TCustomListBox)
  end;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 30 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi DB', [TrDBGrid]);
end;

procedure ResortAllGrids(F: TForm; ToFirst: boolean);
var
  I: integer;
  G: TrDBGrid;
begin
  for I:=0 to F.ComponentCount - 1 do
    if F.Components[I] is TrDBGrid then
    begin
      G:=F.Components[I] as TrDBGrid;
      G.ReSort;
      if ToFirst and G.IsActive then
        G.DataLink.DataSet.First;
    end;
end;

function GetFldListrDBGrid(rDBGrid: TrDBGrid; List: TStrings; FldsDef: TGetFldDefs): integer;
var
  I: integer;
begin
  List.BeginUpdate;
  try
    List.Clear;
    with rDBGrid do
      for I:=0 to Columns.Count - 1 do
        if ((not(fdVisibleFieldOnly in FldsDef)) or (Columns[I].Visible)) and (not(fdDataFieldOnly in FldsDef) or IsDataField(Columns[I].Field)) and
          ((fdAllField in FldsDef) or ((fdSimpleField in FldsDef) and IsSimpleField(Columns[I].Field)) or
            ((fdMemoField in FldsDef) and IsMemoField(Columns[I].Field)) or ((fdGraphicField in FldsDef) and IsGraphicField(Columns[I].Field))) then
          if fdGetLabels in FldsDef then
            List.Add(rDBGrid.GetTitleText(Columns[I]))
          else
            List.Add(Columns[I].FieldName);
    Result:=List.Count;
  finally
    List.EndUpdate;
  end;
end;

function GetCellAlignment(HorzAign: TAlignment; VertAlign: TVerticalAlignment; SingleLine: boolean): integer;
const
  AlignHFlags: array [TAlignment] of integer = (DT_LEFT, DT_RIGHT, DT_CENTER);
  AlignVFlags: array [TVerticalAlignment] of integer = (0, DT_BOTTOM, DT_VCENTER);
  AlignWordBreakFlags: array [boolean] of integer = (DT_WORDBREAK, DT_SINGLELINE);
begin
  Result:=DT_EXPANDTABS or DT_NOPREFIX or AlignHFlags[HorzAign] or AlignVFlags[VertAlign] or AlignWordBreakFlags[SingleLine];
end;

// ==============================================================================
// TrDBGrid Options

constructor TOptionsFixedCell.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  FTitleExBtn:=true;
  FTitleExBtnSelectAll:=true;
  FTitleBtn:=true;
  FFixedColBtn:=true;
  FTitlePopup:=true;
  FTitlePopupDistinctValues:=false;
  FTitleImageSpace:=3;
  FTitleTextMinWidth:=16;
  FFixedColColor:=clNone;
  FFixedColFont:=TFont.Create;
  FFixedColFont.OnChange:=FontChanged;
  FFixedColParentFont:=true;
  FTitleTextSeparators:=cTitleTextSeparators;
  FTitleForceWrapSeparators:='';
  FTitleTextCaseSeparator:=true;
end;

destructor TOptionsFixedCell.Destroy;
begin
  FreeAndNil(FFixedColFont);
  inherited;
end;

procedure TOptionsFixedCell.SetTitleExBtn(Value: boolean);
begin
  FTitleExBtn:=Value;
  if FTitleBtn or FTitleExBtn then
    FOwner.Options:=FOwner.Options + [dgTitleClick]
  else
    FOwner.Options:=FOwner.Options - [dgTitleClick];
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetTitleBtn(Value: boolean);
begin
  FTitleBtn:=Value;
  if FTitleBtn or FTitleExBtn then
    FOwner.Options:=FOwner.Options + [dgTitleClick]
  else
    FOwner.Options:=FOwner.Options - [dgTitleClick];
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetFixedColBtn(Value: boolean);
var
  NewGridOptions: TGridOptions;
begin
  FFixedColBtn:=Value;
  // add original options to basic class, DBGrid do not support it
  NewGridOptions:=THackCustomGrid(FOwner).Options;
  if FFixedColBtn then
    Include(NewGridOptions, goFixedColClick)
  else
    Exclude(NewGridOptions, goFixedColClick);
  THackCustomGrid(FOwner).Options:=NewGridOptions;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetRotateTitleText(Value: boolean);
begin
  FRotateTitleText:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetTitleImageSpace(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 100 then
    Value:=100;
  FTitleImageSpace:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetTitleTextMinWidth(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 200 then
    Value:=200;
  FTitleTextMinWidth:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetTitleTextSeparators(Value: string);
begin
  FTitleTextSeparators:=Value;
  FOwner.Invalidate;
end;

function TOptionsFixedCell.IsTitleTextSeparatorsStored: boolean;
begin
  Result:=FTitleTextSeparators <> cTitleTextSeparators;
end;

procedure TOptionsFixedCell.SetTitleForceWrapSeparators(Value: string);
begin
  FTitleForceWrapSeparators:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetTitleTextCaseSeparator(Value: boolean);
begin
  FTitleTextCaseSeparator:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetFixedCols(Value: integer);
begin
  FFixedCols:=FOwner.SetFixedCols(Value);
end;

function TOptionsFixedCell.GetFixedCols: integer;
begin
  Result:=FOwner.GetFixedCols;
end;

procedure TOptionsFixedCell.SetFixedColColor(Value: TColor);
begin
  FFixedColColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFixedCell.SetFixedColFont(Value: TFont);
begin
  FFixedColFont.Assign(Value);
  FFixedColParentFont:=false;
  FOwner.RefreshLayout;
end;

procedure TOptionsFixedCell.SetFixedColParentFont(Value: boolean);
begin
  FFixedColParentFont:=Value;
  FOwner.RefreshLayout;
end;

function TOptionsFixedCell.IsFontStored: boolean;
begin
  Result:=not FFixedColParentFont;
end;

procedure TOptionsFixedCell.FontChanged(Sender: TObject);
begin
  FFixedColParentFont:=false;
  FOwner.RefreshLayout;
end;

constructor TOptionsEditors.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  FBoolGlyphs:=true;
  FMemo:=obtGlyph;
  FGraphic:=obtGlyph;
  FDateTimePicker:=true;
  FAutoPickListWidth:=true;
  FAllowLookupFilter:=false;
  FAutoDropDown:=false;
  DateTimePickerMinWidth:=84;
  FTextAsMemoSize:=0;
end;

procedure TOptionsEditors.SetBoolGlyphs(Value: boolean);
begin
  FBoolGlyphs:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsEditors.SetMemo(Value: TOptionsBlobType);
begin
  FMemo:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsEditors.SetGraphic(Value: TOptionsBlobType);
begin
  FGraphic:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsEditors.SetDateTimePicker(Value: boolean);
var
  I: integer;
  F: TField;
begin
  FDateTimePicker:=Value;
  for I:=0 to FOwner.Columns.Count - 1 do
  begin
    F:=FOwner.Columns[I].Field;
    if Assigned(F) and (FOwner.IsFieldAsDateEx(F) or IsDateField(F)) then
      if FDateTimePicker then
        FOwner.Columns[I].ButtonStyle:=cbsEllipsis
      else
        FOwner.Columns[I].ButtonStyle:=cbsAuto;
  end;
end;

function TOptionsEditors.ShowMemoGlyph: boolean;
begin
  Result:=FMemo = obtGlyph;
end;

function TOptionsEditors.ShowMemoContent: boolean;
begin
  Result:=FMemo = obtContent;
end;

function TOptionsEditors.ShowGraphicGlyph: boolean;
begin
  Result:=FGraphic = obtGlyph;
end;

function TOptionsEditors.ShowGraphicContent: boolean;
begin
  Result:=FGraphic = obtContent;
end;

constructor TOptionsAppearance.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  FSkipCellHighlight:=false;
  FHighlightMultiselect:=true;
  FHintMaxLength:=2000;
  FColumnMinWidth:=40;
  FColumnMaxWidth:=200;
  FLookupResetDelay:=2500;
  FDataRowLineColor:=clNone;
  FDataRowLineStyle:=psSolid;
  FDataColLineColor:=clNone;
  FDataColLineStyle:=psSolid;
  FTitleRowLineColor:=clNone;
  FTitleRowLineStyle:=psSolid;
  FTitleColLineColor:=clNone;
  FTitleColInGroupLineColor:=clNone;
  FTitleColLineStyle:=psSolid;
  FDataRowInFixedColsAsTitle:=false;
  FColumnWidthMaxRecordsSearch:=2000;
  FColumnFontStyleColor:=false;
  FTitleFontStyleColor:=false;
end;

procedure TOptionsAppearance.SetDrawRowFocus(Value: boolean);
begin
  FDrawRowFocus:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDrawColoredRow(Value: boolean);
begin
  FDrawColoredRow:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetHighlightMultiselect(Value: boolean);
begin
  FHighlightMultiselect:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetHideHScrollBar(Value: boolean);
begin
  FHideHScrollBar:=Value;
  if HideHScrollBar then
    FOwner.ScrollBars:=ssNone
  else
    FOwner.ScrollBars:=ssHorizontal;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDataRowLineColor(Value: TColor);
begin
  FDataRowLineColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDataRowLineStyle(Value: TPenStyle);
begin
  FDataRowLineStyle:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDataColLineColor(Value: TColor);
begin
  FDataColLineColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDataColLineStyle(Value: TPenStyle);
begin
  FDataColLineStyle:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleRowLineColor(Value: TColor);
begin
  FTitleRowLineColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleRowLineStyle(Value: TPenStyle);
begin
  FTitleRowLineStyle:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleColLineColor(Value: TColor);
begin
  FTitleColLineColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleColInGroupLineColor(Value: TColor);
begin
  FTitleColInGroupLineColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleColLineStyle(Value: TPenStyle);
begin
  FTitleColLineStyle:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetDataRowInFixedColsAsTitle(Value: boolean);
begin
  FDataRowInFixedColsAsTitle:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetColumnFontStyleColor(Value: boolean);
begin
  FColumnFontStyleColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsAppearance.SetTitleFontStyleColor(Value: boolean);
begin
  FTitleFontStyleColor:=Value;
  FOwner.Invalidate;
end;

constructor TOptionsFilters.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  EditValues:=TStringList.Create;
  EditStatuses:=TStringList.Create;
  EditValueList:=TMemIniFile.Create('');
  FTextBar:=false;
  FTextBarTopLayout:=false;
  TextBarHeight:=21;
  FTextMarginX:=2;
  FTextMarginY:=1;
  FTextBkColor:=clWindow; // clInfoBk;
  FTextListAutoLoad:=false;
  FTextListAddFilter:=true;
  FTextFilterType:=oftAnyPos;
  FTextComboStyle:=csDropDown;
  FTextFilterLikeChar:=#0;
  FTextFilterLikeOperator:='';
  FUseDatePartAsRange:=true;
  FUseNumberPartAsRange:=true;
  FNumberPartWildChar:='*';
  FSetFilterDelay:=1000;
  EditColumnIndex:=-1;
  EditChanged:=false;
end;

destructor TOptionsFilters.Destroy;
begin
  FreeAndNil(EditValues);
  FreeAndNil(EditStatuses);
  FreeAndNil(EditValueList);
  inherited;
end;

procedure TOptionsFilters.SetFunnelBtn(Value: boolean);
begin
  FFunnelBtn:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFilters.SetTextBar(Value: boolean);
begin
  FTextBar:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsFilters.SetTextBarTopLayout(Value: boolean);
begin
  FTextBarTopLayout:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsFilters.SetTextMarginX(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 20 then
    Value:=20;
  FTextMarginX:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsFilters.SetTextMarginY(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 20 then
    Value:=20;
  FTextMarginY:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsFilters.SetTextBkColor(Value: TColor);
begin
  FTextBkColor:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsFilters.SetTextComboStyle(Value: TComboBoxStyle);
begin
  FTextComboStyle:=Value;
  if Assigned(FOwner.FFilterCombo) then
    FOwner.FFilterCombo.Style:=Value;
end;

function TOptionsFilters.TextBarTopHeight: integer;
begin
  if FTextBar and FTextBarTopLayout then
    Result:=TextBarHeight + FOwner.UseScale(2 * FTextMarginY)
  else
    Result:=0;
end;

function TOptionsFilters.TextBarBootomHeight: integer;
begin
  if FTextBar and not FTextBarTopLayout then
    Result:=TextBarHeight + FOwner.UseScale(2 * FTextMarginY)
  else
    Result:=0;
end;

function TOptionsFilters.GetEditField: TField;
begin
  Result:=nil;
  if EditColumnIndex >= 0 then
    Result:=FOwner.Columns[EditColumnIndex].Field;
end;

function TOptionsFilters.GetTextFilterLikeChar: char;
var
  DS: TDataSet;
begin
  if Assigned(FOwner) and FOwner.IsDatasetLinked then
    DS:=FOwner.DataLink.DataSet
  else
    DS:=nil;
  Result:=FTextFilterLikeChar;
  if Result = #0 then
    Result:=GetFilterWildCardChar(DS);
end;

function TOptionsFilters.GetTextFilterLikeOperator: string;
var
  DS: TDataSet;
begin
  if Assigned(FOwner) and FOwner.IsDatasetLinked then
    DS:=FOwner.DataLink.DataSet
  else
    DS:=nil;
  Result:=FTextFilterLikeOperator;
  if Result = '' then
    Result:=GetFilterLikeOperator(DS);
end;

constructor TOptionsGroups.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  FGroupHeightPixels:=12;
  FCollapseBtnSize:=10;
  FGroupLineWidth:=1;
end;

procedure TOptionsGroups.SetActive(Value: boolean);
begin
  FActive:=Value;
  FOwner.RefreshLayout;
  if FActive then
    FOwner.FixedGroupedCol(false);
end;

procedure TOptionsGroups.SetAlwaysUseGroupColors(Value: boolean);
begin
  FAlwaysUseGroupColors:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsGroups.SetGroupHeightPixels(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 100 then
    Value:=100;
  FGroupHeightPixels:=Value;
  FOwner.RefreshLayout;
end;

procedure TOptionsGroups.SetCollapseBtnSize(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  if Value > 20 then
    Value:=20;
  FCollapseBtnSize:=Value;
  FOwner.Invalidate;
end;

procedure TOptionsGroups.SetGroupLineWidth(Value: byte);
begin
  if Value < 1 then
    Value:=1;
  if Value > 10 then
    Value:=10;
  FGroupLineWidth:=Value;
  FOwner.Invalidate;
end;

function TOptionsGroups.GroupHeight: integer;
begin
  if FActive then
    Result:=FOwner.UseScale(FGroupHeightPixels)
  else
    Result:=0;
end;

constructor TOptionsCellCondition.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
end;

procedure TOptionsCellCondition.SetActive(Value: boolean);
begin
  FActive:=Value;
  FOwner.Invalidate;
end;

constructor TOptionsEx2.Create(Owner: TrDBGrid);
begin
  FOwner:=Owner;
  FFixedCell:=TOptionsFixedCell.Create(FOwner);
  FEditors:=TOptionsEditors.Create(FOwner);
  FAppearance:=TOptionsAppearance.Create(FOwner);
  FFilters:=TOptionsFilters.Create(FOwner);
  FGroups:=TOptionsGroups.Create(FOwner);
  FCellCondition:=TOptionsCellCondition.Create(FOwner);
end;

destructor TOptionsEx2.Destroy;
begin
  FreeAndNil(FFixedCell);
  FreeAndNil(FEditors);
  FreeAndNil(FAppearance);
  FreeAndNil(FFilters);
  FreeAndNil(FGroups);
  FreeAndNil(FCellCondition);
  inherited;
end;

procedure TOptionsEx2.Assign(Source: TPersistent);
var
  sr: TOptionsEx2;
begin
  if Source is TOptionsEx2 then
  begin
    sr:=Source as TOptionsEx2;
    FFixedCell.TitleExBtn:=sr.FFixedCell.FTitleExBtn;
    FFixedCell.TitleExBtnSelectAll:=sr.FFixedCell.FTitleExBtnSelectAll;
    FFixedCell.TitleBtn:=sr.FFixedCell.FTitleBtn;
    FFixedCell.FixedColBtn:=sr.FFixedCell.FFixedColBtn;
    FFixedCell.TitlePopup:=sr.FFixedCell.FTitlePopup;
    FFixedCell.FTitlePopupDistinctValues:=sr.FFixedCell.FTitlePopupDistinctValues;
    FFixedCell.TitleImageSpace:=sr.FFixedCell.FTitleImageSpace;
    FFixedCell.TitleTextMinWidth:=sr.FFixedCell.FTitleTextMinWidth;
    FFixedCell.TitleTextSeparators:=sr.FFixedCell.FTitleTextSeparators;
    FFixedCell.TitleForceWrapSeparators:=sr.FFixedCell.FTitleForceWrapSeparators;
    FFixedCell.TitleTextCaseSeparator:=sr.FFixedCell.FTitleTextCaseSeparator;
    FFixedCell.RotateTitleText:=sr.FFixedCell.FRotateTitleText;
    FFixedCell.FixedCols:=sr.FFixedCell.FFixedCols;
    FFixedCell.FixedColColor:=sr.FFixedCell.FFixedColColor;
    FFixedCell.FixedColFont:=sr.FFixedCell.FFixedColFont;
    FFixedCell.FixedColParentFont:=sr.FFixedCell.FFixedColParentFont;

    FEditors.BoolGlyphs:=sr.FEditors.FBoolGlyphs;
    FEditors.BoolOneClickChange:=sr.FEditors.FBoolOneClickChange;
    FEditors.BoolAutoSave:=sr.FEditors.FBoolAutoSave;
    FEditors.Memo:=sr.FEditors.FMemo;
    FEditors.Graphic:=sr.FEditors.FGraphic;
    FEditors.DateTimePicker:=sr.FEditors.FDateTimePicker;
    FEditors.DateTimePickerMinWidth:=sr.FEditors.DateTimePickerMinWidth;
    FEditors.TextAsMemoSize:=sr.FEditors.FTextAsMemoSize;
    FEditors.OnlyPickListValues:=sr.FEditors.FOnlyPickListValues;
    FEditors.NotProcDblClick:=sr.FEditors.FNotProcDblClick;
    FEditors.NotProcMemoDblClick:=sr.FEditors.FNotProcMemoDblClick;
    FEditors.NotProcGraphicDblClick:=sr.FEditors.FNotProcGraphicDblClick;
    FEditors.AllowPasteMultipleCells:=sr.FEditors.FAllowPasteMultipleCells;
    FEditors.AllowReadOnlyButton:=sr.FEditors.FAllowReadOnlyButton;
    FEditors.AutoPickListWidth:=sr.FEditors.FAutoPickListWidth;
    FEditors.BlockInsertByKey:=sr.FEditors.FBlockInsertByKey;

    FAppearance.DrawRowFocus:=sr.FAppearance.FDrawRowFocus;
    FAppearance.DrawColoredRow:=sr.FAppearance.FDrawColoredRow;
    FAppearance.HighlightMultiselect:=sr.FAppearance.FHighlightMultiselect;
    FAppearance.HideHScrollBar:=sr.FAppearance.FHideHScrollBar;
    FAppearance.StandardTabKey:=sr.FAppearance.FStandardTabKey;
    FAppearance.ReturnAsTabKey:=sr.FAppearance.FReturnAsTabKey;
    FAppearance.DragCell:=sr.FAppearance.FDragCell;
    FAppearance.HintMaxLength:=sr.FAppearance.FHintMaxLength;
    FAppearance.ColumnMinWidth:=sr.FAppearance.FColumnMinWidth;
    FAppearance.ColumnMaxWidth:=sr.FAppearance.FColumnMaxWidth;
    FAppearance.FDataRowLineColor:=sr.FAppearance.FDataRowLineColor;
    FAppearance.FDataRowLineStyle:=sr.FAppearance.FDataRowLineStyle;
    FAppearance.FDataColLineColor:=sr.FAppearance.FDataColLineColor;
    FAppearance.FDataColLineStyle:=sr.FAppearance.FDataColLineStyle;
    FAppearance.FTitleRowLineColor:=sr.FAppearance.FTitleRowLineColor;
    FAppearance.FTitleRowLineStyle:=sr.FAppearance.FTitleRowLineStyle;
    FAppearance.FTitleColLineColor:=sr.FAppearance.FTitleColLineColor;
    FAppearance.FTitleColInGroupLineColor:=sr.FAppearance.FTitleColInGroupLineColor;
    FAppearance.FTitleColLineStyle:=sr.FAppearance.FTitleColLineStyle;
    FAppearance.FDataRowInFixedColsAsTitle:=sr.FAppearance.FDataRowInFixedColsAsTitle;

    FFilters.FunnelBtn:=sr.FFilters.FFunnelBtn;
    FFilters.FunnelBtnActiveOnly:=sr.FFilters.FFunnelBtnActiveOnly;
    FFilters.MultiFieldFilter:=sr.FFilters.FMultiFieldFilter;
    FFilters.TextBar:=sr.FFilters.FTextBar;
    FFilters.TextBarTopLayout:=sr.FFilters.FTextBarTopLayout;
    // FFilters.TextBarHeight:=sr.FFilters.TextBarHeight;
    FFilters.TextMarginX:=sr.FFilters.FTextMarginX;
    FFilters.TextMarginY:=sr.FFilters.FTextMarginY;
    FFilters.TextBkColor:=sr.FFilters.FTextBkColor;
    FFilters.TextListAutoLoad:=sr.FFilters.FTextListAutoLoad;
    FFilters.TextListAddFilter:=sr.FFilters.TextListAddFilter;
    FFilters.SetFilterDelay:=sr.FFilters.FSetFilterDelay;
    FFilters.TextFilterType:=sr.FFilters.FTextFilterType;
    FFilters.TextFilterLikeChar:=sr.FFilters.FTextFilterLikeChar;
    FFilters.TextFilterLikeOperator:=sr.FFilters.FTextFilterLikeOperator;
    FFilters.UseDatePartAsRange:=sr.FFilters.FUseDatePartAsRange;
    FFilters.UseNumberPartAsRange:=sr.FFilters.FUseNumberPartAsRange;

    FGroups.Active:=sr.FGroups.FActive;
    FGroups.AlwaysUseGroupColors:=sr.FGroups.FAlwaysUseGroupColors;
    FGroups.GroupHeightPixels:=sr.FGroups.FGroupHeightPixels;
    FGroups.CollapseBtnSize:=sr.FGroups.FCollapseBtnSize;
    FGroups.GroupLineWidth:=sr.FGroups.FGroupLineWidth;

    FCellCondition.Active:=sr.FCellCondition.FActive;
  end;
end;

// ==============================================================================
// TrDBGridInplaceEditEx functions
constructor TrDBGridInplaceEditEx.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  FLookupSource:=TDatasource.Create(self);
  if not(csDesigning in ComponentState) then
  begin
    // create PopupCalendar - use SELF as Owner
    FPopupCalendar:=TrPopupMonthCalendar.CreateEx(self, PopupCallbackSelectDate);
  end;
end;

procedure TrDBGridInplaceEditEx.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if PasswordChar <> #0 then
    Params.Style:=Params.Style - ES_MULTILINE;
end;

procedure TrDBGridInplaceEditEx.UpdateContents;
var
  Column: TColumn;
begin
  inherited UpdateContents;
  if FUseDataList then
  begin
    if FDataList = nil then
    begin
      FDataList:=TrPopupDataList.Create(self);
      FDataList.Visible:=false;
      FDataList.Parent:=self;
      FDataList.OnMouseUp:=ListMouseUp;
      FDataList.OnKeyDown:=DataListKeyDown;
      FDataList.OnProcessSearchKey:=DataListProcessSearchKey;
    end;
    if EditStyle = esPickList then
      ActiveList:=FDataList;
  end;
  Column:=TrDBGrid(Grid).Columns[TrDBGrid(Grid).SelectedIndex];
  if TrDBGrid(Grid).IsPasswordEx(Column.Field) then
    PasswordChar:=TrDBGrid(Grid).PasswordChar
  else
    PasswordChar:=#0;
  ReadOnly:=Column.ReadOnly;
  Font.Assign(Column.Font);
  // added - fix
  // Color:=Column.Color;
  ImeMode:=Column.ImeMode;
  ImeName:=Column.ImeName;
end;

procedure TrDBGridInplaceEditEx.PopupCallbackSelectDate(Sender: TObject);
begin
  with TrDBGrid(Grid) do
  begin
    SetDateValueEx(SelectedField, Trunc(FPopupCalendar.Date));
    Text:=GetEditText(Col, Row);
  end;
end;

procedure TrDBGridInplaceEditEx.DblClick;
begin
  if (EditStyle = esPickList) then
    DropDown
  else
    inherited DblClick;
end;

function TrDBGridInplaceEditEx.ProcDateTimeHotKey(Key: Word; Shift: TShiftState): boolean;
var
  D: TDateTime;
  dif, step: integer;
begin
  Result:=false;
  with TrDBGrid(Grid) do
  begin
    if IsDateEx(SelectedField) then
    begin
      if not CanEdit then
        Exit;
      if SelectedField.IsNull then
        D:=Date
      else
        D:=GetDateValueEx(SelectedField);
      case Key of
        ord('*'), ord(';'), ord('t'), ord('T'):
          D:=Trunc(Date);
        ord('s'), ord('S'):
          D:=DateUtils.StartOfTheMonth(D);
        ord('e'), ord('E'):
          D:=DateUtils.EndOfTheMonth(D);
      else
        begin
          if SelectedField.IsNull then
            Exit;
          if Key = VK_UP then
            dif:=1
          else
            dif:=-1;
          if Shift = [ssCtrl] then
            D:=DateUtils.IncDay(D, dif)
          else
            D:=IncMonth(D, dif);
        end;
      end;
      SetDateValueEx(SelectedField, Trunc(D));
      SelectAll;
      Result:=true;
    end
    else
      if IsTimeEx(SelectedField) then
      begin
        if not CanEdit then
          Exit;
        if Key = ord('*') then
          D:=Frac(Time)
        else
        begin
          if SelectedField.IsNull then
            Exit;
          D:=GetTimeValueEx(SelectedField);
          if Key = VK_UP then
            dif:=1
          else
            dif:=-1;
          step:=StrToIntDef(FFieldsAsTime.Values[SelectedField.FieldName], 1);
          if Shift = [ssCtrl] then
            D:=DateUtils.IncMinute(D, dif * step)
          else
            D:=DateUtils.IncHour(D, dif);
        end;
        SetTimeValueEx(SelectedField, D);
        SelectAll;
        Result:=true;
      end;
  end;
end;

procedure KillMessage(Wnd: HWnd; Msg: integer);
// Delete the requested message from the queue, but throw back
// any WM_QUIT msgs that PeekMessage may also return
var
  M: TMsg;
begin
  M.Message:=0;
  if PeekMessage(M, Wnd, Msg, Msg, pm_Remove) and (M.Message = WM_QUIT) then
    PostQuitMessage(M.wparam);
end;

procedure TrDBGridInplaceEditEx.KeyDown(var Key: Word; Shift: TShiftState);

  procedure SendToParent;
  begin
    TrDBGrid(Grid).KeyDown(Key, Shift);
    Key:=0;
  end;

  function Selection: TSelection;
  begin
    SendMessage(Handle, EM_GETSEL, Longint(@Result.StartPos), Longint(@Result.EndPos));
  end;

  function CaretPos: integer;
  var
    P: TPoint;
  begin
    Windows.GetCaretPos(P);
    Result:=SendMessage(Handle, EM_CHARFROMPOS, 0, MakeLong(P.X, P.Y));
  end;

  function RightSide: boolean;
  begin
    with Selection do
      Result:=(CaretPos = GetTextLen) and ((StartPos = 0) or (EndPos = StartPos)) and (EndPos = GetTextLen);
  end;

  function LeftSide: boolean;
  begin
    with Selection do
      Result:=(CaretPos = 0) and (StartPos = 0) and ((EndPos = 0) or (EndPos = GetTextLen));
  end;

begin
  if FPopupCalendar.IsShown then
  begin
    if (Key in rPopupMonthCalendarKeys) then
    begin
      // send key to cal
      PostMessage(FPopupCalendar.Handle, WM_KEYDOWN, Key, 0);
      Key:=0;
    end;
    if Key = VK_ESCAPE then
    begin
      FPopupCalendar.HideCalendar;
      Key:=0;
    end;
  end;

  // for lookup editor without dropdown use Left+Right to SearchNext only (do not move cursor)
  if (Key in [VK_LEFT, VK_RIGHT]) and TrDBGrid(Grid).IsSearching and FUseDataList and not ListVisible then
  begin
    SendToParent;
    Key:=0;
  end;

  // for lookup editor without dropdown use Delete to clear field
  if (Key in [VK_DELETE]) and FUseDataList and not ListVisible then
  begin
    SendToParent;
    Key:=0;
  end;

  // bug fix for Ctrl+Enter for Ellipsis
  // KillMessage has to be called before DoEditButtonClick in case of any modal dialog
  if (EditStyle = esEllipsis) and (Key = VK_RETURN) and (Shift = [ssCtrl]) then
  begin
    KillMessage(Handle, WM_CHAR);
    DoEditButtonClick;
    Key:=0;
  end;

  // use F4 for dropdown
  if (Key = VK_F4) and (Shift = []) and (EditStyle in [esEllipsis, esPickList]) then
  begin
    if EditStyle = esEllipsis then
      TrDBGrid(Grid).EditButtonClick
    else
      DropDown;
    Key:=0;
  end;

  // specific action for date or time field
  if (Key in [VK_UP, VK_DOWN]) and ((Shift = [ssCtrl]) or (Shift = [ssCtrl, ssShift])) then
  begin
    if ProcDateTimeHotKey(Key, Shift) then
      Key:=0;
  end;

  // use moving to another cell it Left or Right is used in the border of editor
  case Key of
    VK_LEFT:
      if LeftSide then
        SendToParent;
    VK_RIGHT:
      if RightSide then
        SendToParent;
  end;

  inherited KeyDown(Key, Shift);
end;

procedure TrDBGridInplaceEditEx.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYUP, Key, 0);
    Key:=0;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrDBGridInplaceEditEx.KeyPress(var Key: char);
begin
  if FPopupCalendar.IsShown then
  begin
    case ord(Key) of
      VK_RETURN:
        begin
          PopupCallbackSelectDate(nil);
          FPopupCalendar.HideCalendar;
          Key:=#0;
        end;
    else
      FPopupCalendar.HideCalendar;
    end;
  end;

  // for lookup editor without dropdown use Left+Right to SearchNext only (do not move cursor)
  if TrDBGrid(Grid).FOptionsEx2.FEditors.FAutoDropDown and (Key >= #32) and FUseDataList and not ListVisible then
    DropDown;

  case Key of
    '*', ';', 't', 'T', 's', 'S', 'e', 'E':
      if ProcDateTimeHotKey(ord(Key), []) then
        Key:=#0;
  end;
  if Key <> #0 then
    inherited KeyPress(Key);
end;

procedure TrDBGridInplaceEditEx.DataListKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  // use rDBGrid enhanced search for datalist
  with TrDBGrid(Grid) do
  begin
    case Key of
      VK_ESCAPE, VK_RETURN: // , VK_UP, VK_DOWN:
        StopSearching;
      VK_LEFT, VK_RIGHT:
        begin
          if (ssCtrl in Shift) and IsLookupFilterColumn(Columns[SelectedIndex]) then
            SearchLookupFilter(Key = VK_LEFT)
          else
            SearchNext(Key = VK_LEFT);
          Key:=0;
        end;
    end;
  end;
end;

procedure TrDBGridInplaceEditEx.DataListProcessSearchKey(Sender: TObject; var Key: char);
begin
  // use rDBGrid enhanced search for datalist
  TrDBGrid(Grid).ProcessLookupSearchKey(Key);
  Key:=#0;
end;

procedure TrDBGridInplaceEditEx.DoEditButtonClick;
begin
  TrDBGrid(Grid).EditButtonClick;
end;

procedure TrDBGridInplaceEditEx.DropDown;
var
  Column: TColumn;
  UpdateWidth, ListOwnDraw: boolean;
  H: integer;
begin
  if not Assigned(ActiveList) then
    Exit;
  UpdateWidth:=not ListVisible;
  ListOwnDraw:=false;
  if not ListVisible then
  begin
    with TrDBGrid(Grid) do
      Column:=Columns[SelectedIndex];
    if ActiveList = FDataList then
      with Column.Field do
      begin
        // reset lookup filter before dropdown
        if TrDBGrid(Grid).IsLookupFilterColumn(Column) then
          LookupDataSet.Filtered:=false;
        // set datalist
        FDataList.Font:=Font;
        FDataList.Color:=GetStyleBkColor_ListBox(Color);
        FDataList.Font.Color:=GetStyleFontColor_ListBox(FDataList.Font.Color);
        FLookupSource.DataSet:=LookupDataSet;
        FDataList.KeyField:=LookupKeyFields;
        FDataList.ListField:=LookupResultField;
        FDataList.ListSource:=FLookupSource;
        FDataList.KeyValue:=DataSet.FieldByName(KeyFields).Value;
        // set lookup list props ex
        H:=TrDBGrid(Grid).FLookupListItemHeight;
        if Assigned(TrDBGrid(Grid).FGetLookupListItemHeight) then
          TrDBGrid(Grid).FGetLookupListItemHeight(Grid, FDataList.Canvas, FDataList.GetKeyField, LookupResultField, H);
        FDataList.ItemHeight:=H;
        FDataList.ItemIncreaseHeight:=TrDBGrid(Grid).FLookupListItemIncreaseHeight;
        FDataList.ItemIncreaseWidth:=TrDBGrid(Grid).FLookupListItemIncreaseWidth;
        FDataList.GetListItemProps:=TrDBGrid(Grid).FGetLookupListItemProps;
        // set rowcount
        if LookupDataSet.Active and (LookupDataSet.RecordCount < integer(Column.DropDownRows)) then
          FDataList.RowCount:=LookupDataSet.RecordCount
        else
          FDataList.RowCount:=Column.DropDownRows;
      end
      else
        if ActiveList = PickList then
        begin
          PickList.Items.Assign(Column.PickList);
          with TrDBGrid(Grid), THackCustomListBox(PickList) do
          begin
            OnDrawItem:=nil;
            OnMeasureItem:=nil;
            if IsColorEx(SelectedField) or IsDrawDropDownEx(SelectedField) then
            begin
              OnDrawItem:=PickListDrawItem;
              OnMeasureItem:=PickListMeasureItem;
              ListOwnDraw:=true;
              // set OwnerDraw after inherited DropDown to keep common setting of list size and avoid flickering
            end;
          end;
          DropDownRows:=Column.DropDownRows;
        end;
  end;
  inherited DropDown;
  if ActiveList = PickList then
  begin
    if ListOwnDraw then
      THackCustomListBox(PickList).Style:=lbOwnerDrawVariable
    else
      THackCustomListBox(PickList).Style:=lbStandard;
    // PickList.Visible:=true;
  end;
  if ListVisible and UpdateWidth then
    SetListWidth;
end;

procedure TrDBGridInplaceEditEx.CloseUp(Accept: boolean);
var
  Column: TColumn;
  MasterField: TField;
  ListValue: variant;
begin
  if ListVisible then
  begin
    with TrDBGrid(Grid) do
      Column:=Columns[SelectedIndex];
    if GetCapture <> 0 then
      SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    if ActiveList = DataList then
      ListValue:=DataList.KeyValue
    else
      if PickList.ItemIndex <> -1 then
        ListValue:=PickList.Items[PickList.ItemIndex];
    ShowScrollBar(ActiveList.Handle, SB_BOTH, false);
    SetWindowPos(ActiveList.Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_HIDEWINDOW);
    ListVisible:=false;
    if Assigned(FDataList) then
      FDataList.ListSource:=nil;
    FLookupSource.DataSet:=nil;
    Invalidate;
    if Accept then
    begin
      if ActiveList = DataList then
      begin
        MasterField:=Column.Field.DataSet.FieldByName(Column.Field.KeyFields);
        if MasterField.CanModify and TrDBGrid(Grid).DataLink.Edit then
          MasterField.Value:=ListValue;
      end
      else
      begin
        if (not VarIsNull(ListValue)) and EditCanModify then
          with TrDBGrid(Grid) do
          begin
            SetEditText(Row, Col, ListValue);
            Columns[SelectedIndex].Field.Text:=ListValue;
            InvalidateEditor;
            // show ColorDialog if UserDefined color was selected
            if IsColorEx(SelectedField) then
              CheckSelectUserDefinedColor(Columns[SelectedIndex].Field);
          end;
      end;
    end;
    // reset lookup filter after close
    if TrDBGrid(Grid).IsLookupFilterColumn(Column) then
      Column.Field.LookupDataSet.Filtered:=false;
  end;
end;

procedure TrDBGridInplaceEditEx.SetListWidth;
var
  I, W, MaxW: integer;
  P: TPoint;
  NewWidth, NewHeight: integer;
  DoResize: boolean;
  rGrid: TrDBGrid;
begin
  NewWidth:=ActiveList.Width;
  NewHeight:=ActiveList.Height;
  DoResize:=false;
  rGrid:=TrDBGrid(Grid);
  if rGrid.FOptionsEx2.FEditors.FAutoPickListWidth then
  begin
    if ActiveList = FDataList then
    begin
      W:=FDataList.GetMaxItemWidth;
      if W > NewWidth then
      begin
        NewWidth:=W;
        DoResize:=true;
      end;
      // if W > FDropDownMaxWidth then  W:=FDropDownMaxWidth;
    end
    else
      if ActiveList = PickList then
      begin
        if rGrid.IsColorEx(rGrid.SelectedField) then
        begin
          // add place for color rect - calculate max width of items and set Newwidth
          MaxW:=0;
          for I:=0 to PickList.Items.Count - 1 do
          begin
            W:=PickList.Canvas.TextWidth(PickList.Items[I]);
            if W > MaxW then
              MaxW:=W;
          end;
          Inc(MaxW, 12 + Height + GetSystemMetrics(SM_CXVSCROLL));
          if NewWidth < MaxW then
          begin
            NewWidth:=MaxW;
            DoResize:=true;
          end
        end
        else
          if NewWidth <> Width then
          begin
            Inc(NewWidth, 12);
            DoResize:=true;
          end;
      end;
  end;
  if Assigned(rGrid.OnPickListSetSize) then
  begin
    rGrid.OnPickListSetSize(ActiveList, NewWidth, NewHeight, DoResize);
  end;
  if DoResize then
  begin
    // limit width according columns width and grid width
    P:=Parent.ClientToScreen(Point(Left, Top));
    if P.X + NewWidth > Screen.DesktopRect.Right then
      NewWidth:=Screen.DesktopRect.Right - P.X;
    // set top possition according new height and work area
    I:=P.Y + Height;
    if I + NewHeight > Screen.DesktopRect.Bottom then
      I:=P.Y - NewHeight;
    // set new position and size
    SetWindowPos(ActiveList.Handle, HWND_TOP, P.X, I, NewWidth, NewHeight, SWP_NOACTIVATE or SWP_SHOWWINDOW);
  end;
end;

procedure TrDBGridInplaceEditEx.PaintWindow(DC: HDC);
var
  R: TRect;
  Flags: integer;
  W, X, Y: integer;
  B: TBitmap;
  Gr: TrDBGrid;
  OwnDraw, DrawBtn: boolean;
  Details: TThemedElementDetails;
begin
  if Assigned(Grid) and (Grid is TrDBGrid) and (EditStyle = esEllipsis) then
  begin
    OwnDraw:=false;
    Gr:=TrDBGrid(Grid);
    B:=rGlyphs.BmpFromIndex(rGlyphs.EllipsisIndex, Gr.ScaleRatio, Gr.FCustomGlyphs);
    R:=ButtonRect;
    W:=ButtonWidth;
    DrawBtn:=true;
    Flags:=0;
    if Assigned(Gr.FOnGetEllipsisButtonFormat) then
    begin
      OwnDraw:=true;
      Gr.FOnGetEllipsisButtonFormat(Gr, Gr.SelectedIndex, Gr.SelectedField, W, B);
      DrawBtn:=W > 0;
      if not DrawBtn then
      begin
        if Assigned(B) and (B.Width > 0) then
          W:=B.Width + 4
        else
          W:=ButtonWidth;
      end;
      R.Left:=R.Right - W;
    end;
    if OwnDraw or (Assigned(B) and (B.Width > 0)) then
    begin
      if DrawBtn then
      begin
        // show button
        if GetThemeEnabled then
        begin
          if Pressed then
            Details:=GetThemeServices.GetElementDetails(tbPushButtonPressed)
          else
            // if MouseInControl then
            // Details:=GetThemeServices.GetElementDetails(tbPushButtonHot)
            // else
            Details:=GetThemeServices.GetElementDetails(tbPushButtonNormal);
          GetThemeServices.DrawElement(DC, Details, R);
        end
        else
        begin
          if Pressed then
            Flags:=BF_FLAT;
          DrawEdge(DC, R, EDGE_RAISED, BF_RECT or BF_MIDDLE or Flags);
        end;
      end;
      if Assigned(B) and (B.Width > 0) then
      begin
        rpictool.DrawBmpCenter(DC, R, B, true);
      end
      else
      begin
        X:=R.Left + ((R.Right - R.Left) shr 1) - 1 + ord(Pressed);
        Y:=R.Top + ((R.Bottom - R.Top) shr 1) - 1 + ord(Pressed);
        W:=W shr 3;
        if W = 0 then
          W:=1;
        PatBlt(DC, X, Y, W, W, BLACKNESS);
        PatBlt(DC, X - (W * 2), Y, W, W, BLACKNESS);
        PatBlt(DC, X + (W * 2), Y, W, W, BLACKNESS);
      end;
      ExcludeClipRect(DC, R.Left, R.Top, R.Right, R.Bottom);
    end;
  end;
  inherited PaintWindow(DC);
end;

procedure TrDBGridInplaceEditEx.PickListMeasureItem(Control: TWinControl; Index: integer; var NewHeight: integer);
var
  lb: TCustomListBox;
begin
  with TrDBGrid(Grid) do
  begin
    if IsColorEx(SelectedField) then
    begin
      lb:=(Control as TCustomListBox);
      NewHeight:=lb.Canvas.TextHeight('Wg') + 2;
    end;
    if IsDrawDropDownEx(SelectedField) and Assigned(OnPickListMeasureItem) then
      OnPickListMeasureItem(Control, Index, NewHeight);
  end;
end;

procedure TrDBGridInplaceEditEx.PickListDrawItem(Control: TWinControl; Index: integer; aRect: TRect; aState: TOwnerDrawState);
var
  lb: TCustomListBox;
  s: string;
  C: TColor;
  R: TRect;
  oldB: TBrushRecall;
  idx: integer;
begin
  with TrDBGrid(Grid) do
  begin
    if IsColorEx(SelectedField) then
    begin
      oldFieldColorString:=SelectedField.AsString;
      // show color and text
      lb:=(Control as TCustomListBox);
      s:=lb.Items[Index];
      idx:=integer(lb.Items.Objects[Index]);
      C:=rpictool.GetColorValue(idx);
      if (idx = ColorDefCustomIdx) and (copy(oldFieldColorString, 1, 1) = '$') then
        C:=StringToColor(oldFieldColorString);

      R:=aRect;
      R.Right:=R.Left + R.Bottom - R.Top;
      InflateRect(R, -1, -1);
      oldB:=TBrushRecall.Create(lb.Canvas.Brush);
      try
        BiDi_FrameRect(lb.Canvas, R, clBlack);
        InflateRect(R, -1, -1);
        lb.Canvas.Brush.Color:=C;
        if C <> clNone then
          lb.Canvas.FillRect(R);
      finally
        FreeAndNil(oldB);
      end;
      R:=aRect;
      Inc(R.Left, R.Bottom - R.Top);
      lb.Canvas.TextRect(R, R.Left + 2, R.Top + 1, s);
    end;
    if IsDrawDropDownEx(SelectedField) and Assigned(OnPickListDrawItem) then
      OnPickListDrawItem(Control, Index, aRect, aState);
  end;
end;

function TrDBGridInplaceEditEx.CheckSelectUserDefinedColor(F: TField): boolean;
var
  Dlg: TColorDialog;
begin
  Result:=false;
  if rpictool.GetColorIdx(F.AsString) = ColorDefCustomIdx then
  begin
    Dlg:=TColorDialog.Create(self);
    Dlg.Color:=rpictool.StringToColor(oldFieldColorString);
    Dlg.Options:=[cdFullOpen, cdAnyColor];
    try
      if Dlg.Execute(Handle) then
      begin
        F.AsString:='$' + IntToHex(ColorToRGB(Dlg.Color), 8);
        Result:=true;
      end
      else
        F.AsString:=oldFieldColorString;
    finally
      FreeAndNil(Dlg);
    end;
  end;
end;

procedure TrDBGridInplaceEditEx.CMCancelMode(var Message: TCMCancelMode);
begin
  inherited;
  if not Assigned(FPopupCalendar) then
    Exit;
  if not FPopupCalendar.IsShown or (Message.Sender = self) or (Message.Sender = FPopupCalendar) then
    Exit;
  FPopupCalendar.HideCalendar;
end;

procedure TrDBGridInplaceEditEx.WMKillFocus(var Message: TMessage);
begin
  inherited;
  if Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
end;

procedure TrDBGridInplaceEditEx.WMPaste(var Message: TMessage);
var
  s: string;
begin
  inherited;
  with TrDBGrid(Grid) do
  begin
    s:=EditText;
    if Columns[SelectedIndex].Field is TNumericField then
      s:=rtool.CleanNumericString(s)
    else
      if Columns[SelectedIndex].Field is TDateTimeField then
        s:=rtool.CleanDateTimeString(s)
      else
        s:=rtool.CleanLastEOL(s);
    if s <> EditText then
      try
        EditText:=s;
        SetCursor(Length(s));
        SetEditText(0, 0, s);
      except
      end;
  end;
end;

// ==============================================================================
// TFixedColText functions
constructor TFixedColText.Create(Owner: TrDBGrid);
begin
  AGrid:=Owner;
  FShowIndicator:=true;
  FShowCheckbox:=false;
  FFont:=TFont.Create;
  FFont.OnChange:=FontChanged;
  FParentFont:=true;
  FFixedColWidth:=IndicatorWidth;
end;

destructor TFixedColText.Destroy;
begin
  FreeAndNil(FFont);
  inherited;
end;

procedure TFixedColText.Assign(Source: TPersistent);
var
  sr: TFixedColText;
begin
  if Source is TFixedColText then
  begin
    sr:=Source as TFixedColText;
    FixedColWidth:=sr.FFixedColWidth;
    TitleText:=sr.FTitleText;
    TextMask:=sr.FTextMask;
    Font.Assign(sr.FFont);
    ParentFont:=sr.FParentFont;
    ShowIndicator:=sr.FShowIndicator;
    ShowCheckbox:=sr.FShowCheckbox;
    TextAlign:=sr.FTextAlign;
  end;
end;

procedure TFixedColText.SetShowIndicator(Value: boolean);
begin
  FShowIndicator:=Value;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetShowCheckbox(Value: boolean);
begin
  FShowCheckbox:=Value;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetParentFont(Value: boolean);
begin
  if Value then
    FFont.Assign(AGrid.Font);
  FParentFont:=Value;
end;

function TFixedColText.IsFontStored: boolean;
begin
  Result:=not FParentFont;
end;

procedure TFixedColText.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetFixedColWidth(Value: integer);
begin
  if Value < 8 then
    Value:=0;
  if FFixedColWidth = Value then
    Exit;
  FFixedColWidth:=Value;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetTitleText(const Value: string);
begin
  if FTitleText = Value then
    Exit;
  FTitleText:=Value;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetTextMask(const Value: string);
begin
  if FTextMask = Value then
    Exit;
  FTextMask:=Value;
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  AGrid.RefreshLayout;
end;

procedure TFixedColText.SetTextAlign(Value: TAlignment);
begin
  FTextAlign:=Value;
  AGrid.RefreshLayout;
end;

function TFixedColText.IsFixedColUsed: boolean;
begin
  Result:=FShowCheckbox or (FTextMask <> '');
end;

// ==============================================================================
// TTitleLineText functions
constructor TTitleLineText.Create(Owner: TrDBGrid);
begin
  AGrid:=Owner;
  Color:=clBtnFace;
  FFont:=TFont.Create;
  FFont.OnChange:=FontChanged;
  FParentFont:=true;
  SumHeight:=0;
  LineHeight:=0;
  FixHeight:=0;
end;

destructor TTitleLineText.Destroy;
begin
  FreeAndNil(FFont);
  inherited;
end;

procedure TTitleLineText.Assign(Source: TPersistent);
var
  sr: TTitleLineText;
begin
  if Source is TTitleLineText then
  begin
    sr:=Source as TTitleLineText;
    FixHeight:=sr.FFixHeight;
    Text:=sr.FText;
    Font.Assign(sr.FFont);
    ParentFont:=sr.FParentFont;
    Color:=sr.FColor;
    TextAlign:=sr.FTextAlign;
    ShowDesignTime:=sr.FShowDesignTime;
  end;
end;

procedure TTitleLineText.SetFixHeight(Value: integer);
begin
  if Value < 4 then
    Value:=0;
  if FFixHeight <> Value then
  begin
    FFixHeight:=Value;
    AGrid.RefreshLayout;
  end;
end;

procedure TTitleLineText.SetParentFont(Value: boolean);
begin
  if Value then
    FFont.Assign(AGrid.Font);
  FParentFont:=Value;
end;

function TTitleLineText.IsFontStored: boolean;
begin
  Result:=not FParentFont;
end;

procedure TTitleLineText.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  AGrid.RefreshLayout;
end;

procedure TTitleLineText.SetText(const Value: string);
begin
  if FText = Value then
    Exit;
  FText:=Value;
  AGrid.RefreshLayout;
end;

procedure TTitleLineText.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  AGrid.RefreshLayout;
end;

procedure TTitleLineText.SetColor(Value: TColor);
begin
  FColor:=Value;
  AGrid.RefreshLayout;
end;

procedure TTitleLineText.SetTextAlign(Value: TAlignment);
begin
  FTextAlign:=Value;
  AGrid.RefreshLayout;
end;

procedure TTitleLineText.SetShowDesignTime(Value: boolean);
begin
  FShowDesignTime:=Value;
  AGrid.RefreshLayout;
end;

function TTitleLineText.GetLineCount: integer;
begin
  Result:=0;
  while GetLineText(Result + 1) <> '' do
    Inc(Result);
end;

function TTitleLineText.GetLineText(LineIdx: integer): string;
begin
  Result:=rtool.FindStrPart(FText, '|', LineIdx, true)
end;

// ==============================================================================
// TFooterRow functions
constructor TFooterRow.Create(Owner: TrDBGrid);
begin
  AGrid:=Owner;
  FFooterVisible:=false;
  FFooterHeight:=0;
  FFont:=TFont.Create;
  FFont.OnChange:=FontChanged;
  FParentFont:=true;
  FShowCellLines:=true;
  FCellLinesColor:=clNone;
  FColor:=clNone;
  FSimpleText:='';
  FSimpleTextAlign:=taLeftJustify;
  FBoldText:=false;
  FRecalculateAfterOpen:=false;
  FRecalculateAfterFilter:=false;
  FRefreshAfterScroll:=false;
  FFieldFooterDefs:=TStringList.Create;
  TStringList(FFieldFooterDefs).OnChange:=FieldFooterDefChanged;
end;

destructor TFooterRow.Destroy;
begin
  FreeAndNil(FFieldFooterDefs);
  FreeAndNil(FFont);
  inherited;
end;

procedure TFooterRow.Assign(Source: TPersistent);
var
  sr: TFooterRow;
begin
  if Source is TFooterRow then
  begin
    sr:=Source as TFooterRow;
    FooterVisible:=sr.FFooterVisible;
    FooterHeight:=sr.FFooterHeight;
    Font.Assign(sr.FFont);
    BoldText:=sr.FBoldText;
    BoldText:=sr.FBoldText;
    Color:=sr.FColor;
    ShowCellLines:=sr.FShowCellLines;
    CellLinesColor:=sr.FCellLinesColor;
    FixedColText:=sr.FFixedColText;
    SimpleText:=sr.FSimpleText;
    SimpleTextAlign:=sr.FSimpleTextAlign;
    FieldFooterDefs.Assign(sr.FFieldFooterDefs);
  end;
end;

procedure TFooterRow.SetFooterVisible(Value: boolean);
begin
  FFooterVisible:=Value;
  AGrid.RefreshLayout;
end;

procedure TFooterRow.SetFooterHeight(Value: integer);
begin
  if Value < 8 then
    Value:=0;
  if FFooterHeight = Value then
    Exit;
  FFooterHeight:=Value;
  AGrid.RefreshLayout;
end;

procedure TFooterRow.SetParentFont(Value: boolean);
begin
  if Value then
    FFont.Assign(AGrid.Font);
  FParentFont:=Value;
end;

function TFooterRow.IsFontStored: boolean;
begin
  Result:=not FParentFont;
end;

procedure TFooterRow.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  AGrid.RefreshLayout;
end;

procedure TFooterRow.FieldFooterDefChanged(Sender: TObject);
begin
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetFont(Value: TFont);
begin
  FFont.Assign(Value);
  AGrid.RefreshLayout;
end;

procedure TFooterRow.SetBoldText(Value: boolean);
begin
  FBoldText:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetShowCellLines(Value: boolean);
begin
  FShowCellLines:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetCellLinesColor(Value: TColor);
begin
  FCellLinesColor:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetColor(Value: TColor);
begin
  FColor:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetFixedColText(Value: string);
begin
  FFixedColText:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetSimpleText(Value: string);
begin
  FSimpleText:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetSimpleTextAlign(Value: TAlignment);
begin
  FSimpleTextAlign:=Value;
  AGrid.DrawFooterRow;
end;

procedure TFooterRow.SetFieldFooterDefs(Value: TStrings);
begin
  FFieldFooterDefs.Assign(Value);
  AGrid.DrawFooterRow;
end;

function TFooterRow.GetFooterRowHeight: integer;
begin
  Result:=0;
  if FooterVisible then
  begin
    if FFooterHeight = 0 then
      Result:=AGrid.RowHeights[1] + 2
    else
      Result:=FFooterHeight;
  end;
end;

function TFooterRow.FooterRect: TRect;
begin
  Result:=AGrid.ClientRect;
  Result.Top:=Result.Bottom;
  Result.Bottom:=Result.Top + GetFooterRowHeight;
end;

function TFooterRow.FooterCellRect(ACol: integer): TRect;
var
  R: TRect;
begin
  R:=FooterRect;
  Result:=AGrid.CellRect(ACol, 1);
  Result.Top:=R.Top + 1;
  Result.Bottom:=R.Bottom - 1;
  if (Result.Left = 0) and (Result.Right = 0) then
    Exit;
  InflateRect(Result, 1, 1);
end;

function TFooterRow.IsFooterDefined(Field: TField): boolean;
begin
  Result:=Assigned(Field) and (Trim(FFieldFooterDefs.Values[Field.FieldName]) <> '');
end;

function TFooterRow.GetFooterText(Field: TField): string;
var
  s: string;
begin
  Result:='';
  if not Assigned(Field) then
    Exit;
  s:=Trim(FFieldFooterDefs.Values[Field.FieldName]);
  if s = '' then
    Exit;
  Result:=rtool.RemoveStrPart(s, FieldListSep);
  if not(csDesigning in AGrid.ComponentState) then
    AGrid.ReplaceSummaryTags(Result, Field, AGrid.GetSummaryResults(Field, false), true);
end;

{
  function TFooterRow.GetFooterParams(FieldName: string): string;
  var
  s: string;
  begin
  s:=Trim(FFieldFooterDefs.Values[FieldName]);
  RemoveStrPart(s, FieldListSep);
  Result:=s;
  end;

  function TFooterRow.GetFooterParams(Field: TField): string;
  begin
  if Assigned(Field) then
  Result:=GetFooterParams(Field.FieldName)
  else
  Result:='';
  end;

  procedure TFooterRow.SetFooterParamsEx(FieldName: string; TextMask, Params: string);
  var
  s: string;
  begin
  s:=TextMask + FieldListSep + Params;
  FFieldFooterDefs.Values[FieldName]:=s;
  end;

  procedure TFooterRow.UseFooterParams(Params: string; var Align: TAlignmentDefault; AFont: TFont; var NormalStyleDefined: boolean);
  var
  cfg: string;
  begin
  NormalStyleDefined:=false;
  while Params <> '' do
  begin
  cfg:=AnsiUpperCase(Trim(rtool.RemoveStrPart(Params, FieldListSep)));
  if cfg = 'L' then
  Align:=tadLeft
  else
  if cfg = 'R' then
  Align:=tadRight
  else
  if cfg = 'C' then
  Align:=tadCenter
  else
  UseFontFormatParams(AFont, cfg, NormalStyleDefined);
  end;
  end;
}

procedure TFooterRow.Refresh;
begin
  AGrid.DrawFooterRow;
end;

// ==============================================================================
// TrComboBoxExFilter functions
constructor TrComboBoxExFilter.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Gr:=TrDBGrid(AOwner);
  SetSubComponent(true);
  Ctl3D:=false;
  Visible:=false;
  DropDownCount:=18;
  Height:=21;
end;

procedure TrComboBoxExFilter.CMCancelMode(var Message: TCMCancelMode);
begin
  inherited;
  if Message.Sender <> self then
    Gr.FilterComboHide(false, false);
end;

// ==============================================================================
// ==============================================================================
// TrDBGrid functions
constructor TrDBGrid.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  TempBmp:=TBitmap.Create;
  FUpdateFields:=false;
  ExUpdateLock:=false;
  ShowHintRecCount:=false;
  oldRowPos:=0;
  oldRecNo:=0;
  FPrevActiveBookmark:=nil;
  FActiveBookmark:=nil;
  FDropDownRowsCount:=15;
  FooterGridUpdating:=false;
  // Init OptionsEx
  Options:=Options + [dgTitleHotTrack];
  FGlyphLookStyle:=RosiCompConfig.rDBGridCfg.DefaultGlyphLookStyle;
  FGridLookStyle:=RosiCompConfig.rDBGridCfg.DefaultGridLookStyle;
{$IFDEF UseObsoleteOption}
  FOptionsEx_old:=[dgTitleExBtn, dgTitleBtn, dgFixColBtn, dgBool, dgMemo, dgGraphic, dgDateTimePicker, dgUseTitlePopup, dgAutoPickListWidth];
{$ENDIF}
  FOptionsEx2:=TOptionsEx2.Create(self);
{$IFDEF UseObsoleteOption}
  FOptionsEx2.FActivateOptionsEx2:=false;
{$ENDIF}
  FLookupCondition:=lcBrowse;
  FColorOddRow:=clMoneyGreen;
  FColorActiveRow:=clInfoBk;
  FColorHotTrackRow:=clNone;
  FColorRequiredField:=clNone;
  FColumnWidthAfterDblClick:=cwDataLabelWidth;
  FFontColorFiltered:=clNavy;
  FFontColorLink:=clHighlight;
  FFontColorReadOnly:=clGrayText;
  DrawingStyle:=RosiCompConfig.rDBGridCfg.DefaultDrawingStyle;
  FLastColWidthChangedIdx:=-1;
  FFilterDlg:=nil;
  FFilterDlgLastColIdx:=-1;
  FMouseOverRow:=-1;

  FTitleLines:=1;
  FRowFixHeight:=0;
  FRowLines:=1;
  FRowVerticalAlign:=taAlignTop;
  FFixedColText:=TFixedColText.Create(self);
  FTitleLineText:=TTitleLineText.Create(self);
  FFooterRow:=TFooterRow.Create(self);
  FUseCalculatedSummaryForPopup:=false;
  FPasswordChar:='*';
  FLookupListItemHeight:=0;
  FLookupListItemIncreaseHeight:=2;
  FLookupListItemIncreaseWidth:=0;

  FCellParamsCondList:=TCellParamsCondList.Create;
  FCellParamsCondList.LinkGrid(self);

  // Init strings
  FHintFieldValues:=TStringList.Create;
  FFieldsAsLink:=TStringList.Create;
  FFieldsAsBoolean:=TStringList.Create;
  FFieldsAsRichText:=TStringList.Create;
  FFieldsAsGraphics:=TStringList.Create;
  FFieldsAsDate:=TStringList.Create;
  FFieldsAsTime:=TStringList.Create;
  FFieldsAsColor:=TStringList.Create;
  FFieldsAsPassword:=TStringList.Create;
  FFieldsAsDataBar:=TStringList.Create;
  FFieldsAsStatusBar:=TStringList.Create;
  FFieldsAutoPickList:=TStringList.Create;
  FFieldsDrawDropDown:=TStringList.Create;
  FFieldsFilterButton:=TStringList.Create;
  FFieldsFilterMapping:=TStringList.Create;
  FFieldsFixedColumnWidth:=TStringList.Create;

  FTitleHints:=TStringList.Create;
  FTitleImages:=TStringList.Create;
  FDataImages:=TStringList.Create;
  FDataImages.OnChange:=DataImagesChanged;
  FDataImagesIni:=TMemIniFile.Create('');
  FFieldsGroups:=TStringList.Create;
  FFieldsGroups.OnChange:=FieldsGroupsChanged;
  FFieldsGroupsIni:=TMemIniFile.Create('');
  FFieldsGroupsDefault:=TStringList.Create;

  FTitlePopupItems:=TStringList.Create;
{$IFNDEF TrPopupMonthCalendar}
  FDateTimePicker:=nil;
{$ENDIF}
  FSearchingPrevNext:=stFindFirst;
  OwnHintWin:=THintWindow.Create(self);

  if not(csDesigning in ComponentState) then
  begin
    // switch off DefaultDrawing
    DefaultDrawing:=false;
    // Init search values
    FSearchText:='';
    // init timer
    FTimer:=TTimer.Create(self);
    FTimer.SetSubComponent(true);
    FTimer.Enabled:=false;
    FTimer.OnTimer:=FOnTimer;
    oldHintPause:=Application.HintPause;
    // title popup
    FTitlePopup:=TPopupMenu.Create(self);
    FTitlePopup.AutoHotkeys:=maManual;
    // create filtercombo subcomp
    FFilterCombo:=TrComboBoxExFilter.Create(self);
    FFilterCombo.OnExit:=FilterComboExit;
    FFilterCombo.OnChange:=FilterComboChange;
    FFilterCombo.OnKeyPress:=FilterComboKeyPress;
    FFilterCombo.OnKeyDown:=FilterComboKeyDown;
    FFilterCombo.OnUpdateTimer:=FilterComboUpdateTimer;
    FFilterCombo.Style:=FOptionsEx2.FFilters.FTextComboStyle;
    FOptionsEx2.FFilters.TextBarHeight:=FFilterCombo.Height;
    // rich text
    FRichText:=TrRichEdit.Create(self);
    FRichText.Visible:=false;
    FRichText.Parent:=self;
    FRichText.BorderStyle:=bsNone;
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(self, rDBComps_ComponentCreatedEventID);
end;

destructor TrDBGrid.Destroy;
begin
  FreeAndNil(FOptionsEx2);

  FreeAndNil(FCellParamsCondList);
  FreeAndNil(FHintFieldValues);
  FreeAndNil(FFieldsAsLink);
  FreeAndNil(FFieldsAsBoolean);
  FreeAndNil(FFieldsAsRichText);
  FreeAndNil(FFieldsAsGraphics);
  FreeAndNil(FFieldsAsDate);
  FreeAndNil(FFieldsAsTime);
  FreeAndNil(FFieldsAsColor);
  FreeAndNil(FFieldsAsPassword);
  FreeAndNil(FFieldsAsDataBar);
  FreeAndNil(FFieldsAsStatusBar);
  FreeAndNil(FFieldsAutoPickList);
  FreeAndNil(FFieldsDrawDropDown);
  FreeAndNil(FFieldsFilterButton);
  FreeAndNil(FFieldsFilterMapping);
  FreeAndNil(FFieldsFixedColumnWidth);

  FreeAndNil(FTitleHints);
  FreeAndNil(FTitleImages);
  FreeAndNil(FDataImages);
  FreeAndNil(FDataImagesIni);
  FreeAndNil(FFieldsGroups);
  FreeAndNil(FFieldsGroupsIni);
  FreeAndNil(FFieldsGroupsDefault);

  FreeAndNil(FTitlePopupItems);
  FreeAndNil(FFixedColText);
  FreeAndNil(FTitleLineText);
  FreeAndNil(FFooterRow);

  FreeAndNil(TempBmp);
  inherited Destroy;
end;

procedure TrDBGrid.Loaded;
var
  I: integer;
begin
  inherited Loaded;
  { // temporary clear saved values = developing time only
    FTitlePopupItems.Clear;
    FOptionsEx:=FOptionsEx+[dgUseTitlePopup];
    }
{$IFDEF UseObsoleteOption}
  if not FOptionsEx2.FActivateOptionsEx2 then
  begin
    // copy old options to new structure
    CopyOptionsEx_oldToOptionsEx2;
    // FOptionsEx2.FActivateOptionsEx2:=true;
  end;
{$ENDIF}
  if RosiCompConfig.rDBGridCfg.FilterMultiFieldForcedOn then
    FOptionsEx2.FFilters.FMultiFieldFilter:=true;

  if not(csDesigning in ComponentState) then
  begin
{$IFNDEF TrPopupMonthCalendar}
    // create datetime subcomp
    FDateTimePicker:=CreateDateTimePicker;
{$ENDIF}
    // switch off DefaultDrawing
    DefaultDrawing:=false;
    SetDefaultTitlePopup;
    // load stored columns props
    SetLength(FColumnStoredProps, Columns.Count);
    for I:=0 to Columns.Count - 1 do
    begin
      FColumnStoredProps[I].FieldName:=Columns[I].FieldName;
      FColumnStoredProps[I].Width:=Columns[I].Width;
      FColumnStoredProps[I].Visible:=Columns[I].Visible;
    end;
    SetLength(FColumnFilters, 0);
  end;
  // call paint to set default colors to avoid reading of undefined column color if grid is hidden and not painted
  try
    Paint;
  except
  end;
  // rearrange grouped column to be sure that columns are together
  FixedGroupedCol(false);
  FFieldsGroupsDefault.Assign(FFieldsGroups);
  RosiCompConfig.DoRosiCompGeneralEventSimple(self, rDBComps_ComponentLoadedEventID);
end;

procedure TrDBGrid.SetDefaultTitlePopup;
begin
  // define default popup if empty
  if FTitlePopupItems.Count = 0 then
  begin
    FTitlePopupItems.Add(sPMSum);
    FTitlePopupItems.Add(sPMMin);
    FTitlePopupItems.Add(sPMMax);
    FTitlePopupItems.Add(sPMAvg);
    FTitlePopupItems.Add('-');
    FTitlePopupItems.Add(sPMCountVal);
    FTitlePopupItems.Add(sPMCountNull);
    FTitlePopupItems.Add(sPMCountAll);
    FTitlePopupItems.Add(sPMDistinctValues);
  end;
end;

procedure TrDBGrid.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: boolean {$ENDIF});
{$IFNDEF DX102P}
var
  I: integer;
  C: TColumn;
  Warr: array of integer;
{$ENDIF}
begin
  if not DoGeneralEvent(geChangeScaleBefore) then
    Exit;
  StopUpdate(true);
  try
{$IFNDEF DX102P}
    // store column width before scale
    SetLength(Warr, Columns.Count);
    for I:=0 to Columns.Count - 1 do
      Warr[I]:=Columns[I].Width;
{$ENDIF}
    inherited;
    FScaleRatio:=FScaleRatio * M / D;

    // update spec size
    FOptionsEx2.FFilters.TextBarHeight:=MulDiv(FOptionsEx2.FFilters.TextBarHeight, M, D);
    FFixedColText.FFixedColWidth:=MulDiv(FFixedColText.FFixedColWidth, M, D);
{$IFNDEF DX102P}
    // update column width and column font if stored
    for I:=Columns.Count - 1 downto 0 do
    begin
      C:=Columns[I];
      if C.Visible then
        C.Width:=MulDiv(Warr[I], M, D);
      if (cvFont in C.AssignedValues) then
        rtool.DPI_ScaleFontHeight(C.Font, M, D);
      if (cvTitleFont in C.AssignedValues) then
        rtool.DPI_ScaleFontHeight(C.Title.Font, M, D);
    end;
    if not ParentFont then
      rtool.DPI_ScaleFontHeight(TitleFont, M, D);
{$ENDIF}
    // update other fonts
    if not FTitleLineText.FParentFont then
      rtool.DPI_ScaleFontHeight(FTitleLineText.Font, M, D);
    if not FFixedColText.FParentFont then
      rtool.DPI_ScaleFontHeight(FFixedColText.Font, M, D);
    if not FFooterRow.FParentFont then
      rtool.DPI_ScaleFontHeight(FFooterRow.Font, M, D);
  finally
    StopUpdate(false);
  end;
  if not DoGeneralEvent(geChangeScaleAfter) then
    Exit;
end;

function TrDBGrid.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

procedure TrDBGrid.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if (AComponent = FPropSaveFile) then
      FPropSaveFile:=nil;
    if (AComponent = FTitleExPopupMenu) then
      FTitleExPopupMenu:=nil;
    if (AComponent = FTitleImageList) then
      FTitleImageList:=nil;
    if (AComponent = FDataImageList) then
      FDataImageList:=nil;
    if (AComponent = FSorter) then
      FSorter:=nil;
    if (AComponent = FCustomGlyphs) then
    begin
      FCustomGlyphs:=nil;
      Invalidate;
    end;
  end;
end;

procedure TrDBGrid.SetParent(AParent: TWinControl);
begin
  inherited;
{$IFNDEF TrPopupMonthCalendar}
  if Assigned(AParent) and Assigned(FDateTimePicker) then
  begin
    FDateTimePicker.Parent:=self; // self.Parent  - BUG for new record, DO NOT use;
    FDateTimePicker.BoundsRect:=Rect(0, 0, 0, 0);
  end;
{$ENDIF}
  if Assigned(AParent) and Assigned(FFilterCombo) then
  begin
    FFilterCombo.Parent:=Parent;
    FFilterCombo.BoundsRect:=Rect(0, 0, 0, 0);
    // refresh filrer bar height
    FFilterCombo.Height:=21;
    SendMessage(FFilterCombo.Handle, CM_FONTCHANGED, 0, 0);
    // refresh FTextBarHeight
    FOptionsEx2.FFilters.TextBarHeight:=FFilterCombo.Height;
  end;
end;

procedure TrDBGrid.Invalidate;
begin
  if Assigned(FCellParamsCondList) and not(csDestroying in ComponentState) then
    FCellParamsCondList.ResetCond;
  inherited;
end;
{$IFDEF UseObsoleteOption}

procedure TrDBGrid.CopyOptionsEx_oldToOptionsEx2;
var
  oldOptions: TOptionsEx;
begin
  oldOptions:=FOptionsEx_old;
  // FOptionsEx2.FActivateOptionsEx2:=true;
  FOptionsEx2.FFixedCell.FTitleExBtn:=dgTitleExBtn in oldOptions;
  FOptionsEx2.FFixedCell.FTitleBtn:=dgTitleBtn in oldOptions;
  FOptionsEx2.FFixedCell.FFixedColBtn:=dgFixColBtn in oldOptions;
  FOptionsEx2.FFixedCell.FTitlePopup:=dgUseTitlePopup in oldOptions;
  FOptionsEx2.FFixedCell.FRotateTitleText:=dgRotateTitleText in oldOptions;

  FOptionsEx2.FEditors.FBoolGlyphs:=dgBool in oldOptions;
  FOptionsEx2.FEditors.FBoolOneClickChange:=dgOneClickCheckbox in oldOptions;
  FOptionsEx2.FEditors.FBoolAutoSave:=dgAutoSaveCheckBox in oldOptions;
  if dgMemoShowText in oldOptions then
    FOptionsEx2.FEditors.FMemo:=obtContent
  else
    if dgMemo in oldOptions then
      FOptionsEx2.FEditors.FMemo:=obtGlyph
    else
      FOptionsEx2.FEditors.FMemo:=obtNone;
  if dgGraphicShowImg in oldOptions then
    FOptionsEx2.FEditors.FGraphic:=obtContent
  else
    if dgGraphic in oldOptions then
      FOptionsEx2.FEditors.FGraphic:=obtGlyph
    else
      FOptionsEx2.FEditors.FGraphic:=obtNone;
  FOptionsEx2.FEditors.FDateTimePicker:=dgDateTimePicker in oldOptions;
  FOptionsEx2.FEditors.FOnlyPickListValues:=dgOnlyPickListValues in oldOptions;
  FOptionsEx2.FEditors.FTextAsMemoSize:=FTextAsMemoSize_old;
  FOptionsEx2.FEditors.FNotProcDblClick:=dgNotProcDblClick in oldOptions;
  FOptionsEx2.FEditors.FNotProcMemoDblClick:=dgNotProcMemoDblClick in oldOptions;
  FOptionsEx2.FEditors.FNotProcGraphicDblClick:=dgNotProcGraphicDblClick in oldOptions;
  FOptionsEx2.FEditors.FAllowPasteMultipleCells:=dgAllowPasteMultipleCells in oldOptions;
  FOptionsEx2.FEditors.FAllowReadOnlyButton:=dgAllowReadOnlyButton in oldOptions;
  FOptionsEx2.FEditors.FAutoPickListWidth:=dgAutoPickListWidth in oldOptions;
  FOptionsEx2.FEditors.FBlockInsertByKey:=dgBlockInsertByKey in oldOptions;

  FOptionsEx2.FAppearance.FDrawRowFocus:=dgDrawRowFocus in oldOptions;
  FOptionsEx2.FAppearance.FDrawColoredRow:=dgDrawColoredRow in oldOptions;
  FOptionsEx2.FAppearance.FHighlightMultiselect:=not(dgNotHighlightMultiselect in oldOptions);
  FOptionsEx2.FAppearance.FHideHScrollBar:=dgHideHScrollBar in oldOptions;
  FOptionsEx2.FAppearance.FStandardTabKey:=dgStandardTabKey in oldOptions;
  FOptionsEx2.FAppearance.FDragCell:=dgDragCell in oldOptions;

  FOptionsEx2.FFilters.FFunnelBtn:=dgFilterButtons in oldOptions;
  FOptionsEx2.FFilters.FMultiFieldFilter:=dgMultiFieldFilter in oldOptions;

  FOptionsEx2.FGroups.FActive:=dgFieldGroups in oldOptions;
  FOptionsEx2.FGroups.FAlwaysUseGroupColors:=dgAlwaysUseFieldGroupColors in oldOptions;

  FOptionsEx2.FCellCondition.FActive:=dgUseCellParamsConditions in oldOptions;
end;
{$ENDIF}

procedure TrDBGrid.Assign(Source: TPersistent);
var
  sgr: TDBGrid;
  srgr: TrDBGrid;
begin
  if Source is TDBGrid then
  begin
    sgr:=Source as TDBGrid;
    Color:=sgr.Color;
    Columns.Assign(sgr.Columns);
    DataSource:=sgr.DataSource;
    DrawingStyle:=sgr.DrawingStyle;
    FixedColor:=sgr.FixedColor;
    GradientEndColor:=sgr.GradientEndColor;
    GradientStartColor:=sgr.GradientStartColor;
    Font.Assign(sgr.Font);
    Options:=sgr.Options;
    TitleFont.Assign(sgr.TitleFont);
    // set internal color (by default set in Paint, but is not called if visible=false
    THackCustomGrid(self).FInternalColor:=sgr.Color;
  end;
  if Source is TrDBGrid then
  begin
    srgr:=Source as TrDBGrid;
    ColorActiveRow:=srgr.ColorActiveRow;
    ColorHotTrackRow:=srgr.ColorHotTrackRow;
    ColorOddRow:=srgr.ColorOddRow;
    ColumnWidth:=srgr.ColumnWidth;
    ColumnWidthAfterDblClick:=srgr.ColumnWidthAfterDblClick;
    CustomGlyphs:=srgr.CustomGlyphs;

    DataImageList:=srgr.DataImageList;
    DataImages.Assign(srgr.DataImages);

    FieldsAsBoolean.Assign(srgr.FieldsAsBoolean);
    FieldsAsColor.Assign(srgr.FieldsAsColor);
    FieldsAsDate.Assign(srgr.FieldsAsDate);
    FieldsAsGraphics.Assign(srgr.FieldsAsGraphics);
    FieldsAsLink.Assign(srgr.FieldsAsLink);
    FieldsAsPassword.Assign(srgr.FieldsAsPassword);
    FieldsAsDataBar.Assign(srgr.FieldsAsDataBar);
    FieldsAsStatusBar.Assign(srgr.FieldsAsStatusBar);
    FieldsFixedColumnWidth.Assign(srgr.FieldsFixedColumnWidth);
    FieldsAutoPickList.Assign(srgr.FieldsAutoPickList);
    FieldsDrawDropDown.Assign(srgr.FieldsDrawDropDown);
    FieldsFilterButton.Assign(srgr.FieldsFilterButton);
    FieldsGroups.Assign(srgr.FieldsGroups);

    FixedColText.Assign(srgr.FixedColText);
    FontColorFiltered:=srgr.FontColorFiltered;
    FontColorLink:=srgr.FontColorLink;
    FontColorReadOnly:=srgr.FontColorReadOnly;

    FooterRow.Assign(srgr.FooterRow);
    GlyphLookStyle:=srgr.GlyphLookStyle;
    GridLookStyle:=srgr.GridLookStyle;

    LookupCondition:=srgr.LookupCondition;
    LookupListItemHeight:=srgr.LookupListItemHeight;
    LookupListItemIncreaseHeight:=srgr.LookupListItemIncreaseHeight;
    LookupListItemIncreaseWidth:=srgr.LookupListItemIncreaseWidth;
{$IFDEF UseObsoleteOption}
    OptionsEx:=srgr.OptionsEx;
{$ENDIF}
    OptionsEx2.Assign(srgr.OptionsEx2);

    PropSaveFile:=srgr.PropSaveFile;
    TitleLines:=srgr.TitleLines;
    RowFixHeight:=srgr.RowFixHeight;
    RowLines:=srgr.RowLines;
    RowVerticalAlign:=srgr.RowVerticalAlign;
    Sorter:=srgr.Sorter;
    TitleExPopupMenu:=srgr.TitleExPopupMenu;
    TitleFont.Assign(srgr.TitleFont);
    TitleHints.Assign(srgr.TitleHints);
    TitleImageList:=srgr.TitleImageList;
    TitleImages.Assign(srgr.TitleImages);
    TitleLines:=srgr.TitleLines;
    TitleLineText.Assign(srgr.TitleLineText);
    TitlePopupItems.Assign(srgr.TitlePopupItems);

    CellParamsCondList.Assign(srgr.CellParamsCondList);
  end;
end;

// ==============================================================================
// SET Props funct block
{$IFDEF UseObsoleteOption}

procedure TrDBGrid.SetOptionsEx(const Value: TOptionsEx);
begin
  FOptionsEx_old:=Value;
  if not FOptionsEx2.FActivateOptionsEx2 and not(csLoading in ComponentState) then
    CopyOptionsEx_oldToOptionsEx2;
  RefreshLayout;
  // Invalidate;
end;
{$ENDIF}

procedure TrDBGrid.SetCustomGlyphs(const Value: TrCustomGlyphs);
begin
  FCustomGlyphs:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
  Invalidate;
end;

procedure TrDBGrid.SetGlyphLookStyle(const Value: TGlyphLookStyles);
begin
  FGlyphLookStyle:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetGridLookStyle(const Value: TGridLookStyles);
var
  I: integer;
  GLS: TGridLookStyleDef;

  function GetDefColor(NewColor, DefColor: TColor): TColor;
  begin
    if NewColor = clNone then
      Result:=DefColor
    else
      Result:=NewColor;
  end;

begin
  if (FGridLookStyle = Value) or (csLoading in ComponentState) then
    Exit;
  // update all colors according scheme
  FGridLookStyle:=Value;
  I:=ord(FGridLookStyle);
  if I <= High(GridLookStyleDefs) then
  begin
    SaveRowPos(true);
    try
      GLS:=GridLookStyleDefs[I];
      GradientStartColor:=GetDefColor(GLS.GradientStartColor, clWhite);
      GradientEndColor:=GetDefColor(GLS.GradientEndColor, $00D6D6D6);
      TitleFont.Color:=GetDefColor(GLS.TitleFontColor, clWindowText);
      Font.Color:=GetDefColor(GLS.TextFontColor, clWindowText);
      Color:=GetDefColor(GLS.TextBkColor, clWindow);
      ColorActiveRow:=GetDefColor(GLS.TextBkColorActiveRow, clInfoBk);
      ColorOddRow:=GetDefColor(GLS.TextBkColorOddRow, clMoneyGreen);
      FOptionsEx2.FAppearance.FDrawColoredRow:=GLS.TextBkColorOddRow <> clNone;
      TitleLineText.Color:=GradientStartColor;
      TitleLineText.Font.Color:=TitleFont.Color;
    finally
      RestoreRowPos;
    end;
  end;
  Invalidate;
end;

procedure TrDBGrid.SetSortFieldName(const Value: string);
begin
  FSortFieldName:=TrimSortFields(Value);
  Invalidate;
end;

procedure TrDBGrid.SetSortDesc(const Value: boolean);
begin
  FSortDesc:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetSortDirections(const Value: string);
begin
  FSortDirections:=TrimSortDirections(Value);
  Invalidate;
end;

procedure TrDBGrid.SetFColumnWidth(const Value: TColumnWidth);
begin
  SetColumnWidth(Value);
  if Value in [cwDefaultWidth, cwDataWidth, cwDataLabelWidth] then
    FColumnWidth:=cwNone
  else
    FColumnWidth:=Value;
end;

procedure TrDBGrid.SetColorOddRow(Value: TColor);
begin
  FColorOddRow:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetColorActiveRow(Value: TColor);
begin
  FColorActiveRow:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetColorHotTrackRow(Value: TColor);
begin
  FColorHotTrackRow:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetFontColorFiltered(Value: TColor);
begin
  FFontColorFiltered:=Value;
  Invalidate;
end;

procedure TrDBGrid.SetTitleLines(const Value: integer);
begin
  if (Value >= 1) then
  begin
    FTitleLines:=Value;
    LayoutChanged;
  end;
end;

procedure TrDBGrid.SetRowFixHeight(const Value: integer);
begin
  if (Value <> FRowFixHeight) then
  begin
    FRowFixHeight:=Value;
    if FRowFixHeight < 5 then
      FRowFixHeight:=0;
    LayoutChanged;
  end;
end;

procedure TrDBGrid.SetRowLines(const Value: integer);
begin
  if (Value >= 1) then
  begin
    FRowLines:=Value;
    LayoutChanged;
  end;
end;

procedure TrDBGrid.SetRowVerticalAlign(const Value: TVerticalAlignment);
begin
  if (Value <> FRowVerticalAlign) then
  begin
    FRowVerticalAlign:=Value;
    LayoutChanged;
  end;
end;

procedure TrDBGrid.SetPropSaveFile(Value: TPropSaveFile);
begin
  FPropSaveFile:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

procedure TrDBGrid.SetTitleExPopupMenu(Value: TPopupMenu);
begin
  FTitleExPopupMenu:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

procedure TrDBGrid.SetFieldsAsLink(Value: TStrings);
begin
  FFieldsAsLink.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsBoolean(Value: TStrings);
begin
  FFieldsAsBoolean.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsRichText(Value: TStrings);
begin
  FFieldsAsRichText.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsGraphics(Value: TStrings);
begin
  FFieldsAsGraphics.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsDate(Value: TStrings);
begin
  FFieldsAsDate.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsTime(Value: TStrings);
begin
  FFieldsAsTime.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsColor(Value: TStrings);
begin
  FFieldsAsColor.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsPassword(Value: TStrings);
begin
  FFieldsAsPassword.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsDataBar(Value: TStrings);
begin
  FFieldsAsDataBar.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAsStatusBar(Value: TStrings);
begin
  FFieldsAsStatusBar.Assign(Value);
end;

procedure TrDBGrid.SetFieldsAutoPickList(Value: TStrings);
begin
  FFieldsAutoPickList.Assign(Value);
end;

procedure TrDBGrid.SetFieldsDrawDropDown(Value: TStrings);
begin
  FFieldsDrawDropDown.Assign(Value);
end;

procedure TrDBGrid.SetFieldsFilterButton(Value: TStrings);
begin
  FFieldsFilterButton.Assign(Value);
  Invalidate;
end;

procedure TrDBGrid.SetFieldsFilterMapping(Value: TStrings);
begin
  FFieldsFilterMapping.Assign(Value);
  Invalidate;
end;

procedure TrDBGrid.SetFieldsFixedColumnWidth(Value: TStrings);
begin
  FFieldsFixedColumnWidth.Assign(Value);
  Invalidate;
end;

procedure TrDBGrid.SetTitlePopupItems(Value: TStrings);
begin
  FTitlePopupItems.Assign(Value);
end;

procedure TrDBGrid.SetTitleImageList(Value: TImageList);
begin
  FTitleImageList:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
  Invalidate;
end;

procedure TrDBGrid.SetTitleHints(Value: TStrings);
begin
  FTitleHints.Assign(Value);
end;

procedure TrDBGrid.SetTitleImages(Value: TStrings);
begin
  FTitleImages.Assign(Value);
  Invalidate;
end;

procedure TrDBGrid.SetDataImageList(Value: TImageList);
begin
  FDataImageList:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
  Invalidate;
end;

procedure TrDBGrid.SetDataImages(Value: TStringList);
begin
  FDataImages.Assign(Value);
end;

procedure TrDBGrid.DataImagesChanged(Sender: TObject);
begin
  FDataImagesIni.SetStrings(FDataImages);
  Invalidate;
end;

procedure TrDBGrid.SetFieldsGroups(Value: TStringList);
begin
  FFieldsGroups.Assign(Value);
end;

procedure TrDBGrid.FieldsGroupsChanged(Sender: TObject);
begin
  FFieldsGroupsIni.SetStrings(FFieldsGroups);
  Invalidate;
  FixedGroupedCol(false);
end;

procedure TrDBGrid.SetFilteredFieldNames(Value: string);
begin
  if FFilteredFieldNames <> Value then
  begin
    FFilteredFieldNames:=Value;
    Invalidate;
  end;
  FFilterDlgLastColIdx:=-1;
end;

procedure TrDBGrid.SetSorter(Value: TrDBGridCustomSorter);
begin
  FSorter:=Value;
  if Value <> nil then
    Value.FreeNotification(self);
end;

function TrDBGrid.SetFixedCols(Value: integer): integer;
var
  FixCount, intRow: integer;
begin
  FixCount:=Max(Value, 0) + IndicatorOffset;
  if IsActive and not(csLoading in ComponentState) and (ColCount > IndicatorOffset + 1) then
  begin
    FixCount:=Min(FixCount, ColCount - 1);
    intRow:=Row;
    FixedCols:=FixCount;
    Row:=intRow;
  end;
  Result:=FixCount - IndicatorOffset;
end;

function TrDBGrid.GetFixedCols: integer;
begin
  if IsActive then
    Result:=FixedCols - IndicatorOffset
  else
    Result:=FOptionsEx2.FFixedCell.FFixedCols;
end;

procedure TrDBGrid.SetMouseOverRow(const Value: integer);
begin
  if not(csDesigning in ComponentState) and (FColorHotTrackRow <> clNone) then
  begin
    if FMouseOverRow <> Value then
    begin
      if FMouseOverRow >= 0 then
        InvalidateRowEx(FMouseOverRow);
      FMouseOverRow:=Value;
      if FMouseOverRow > TitleRow then
        InvalidateRowEx(FMouseOverRow);
    end;
  end;
end;

function TrDBGrid.ClientRectToParentRect(R: TRect): TRect;
var
  D: integer;
begin
  Result:=R;
  D:=Left;
  if BorderStyle = bsSingle then
  begin
    if Ctl3D then
      Inc(D, GetSystemMetrics(SM_CXEDGE))
    else
      Inc(D, GetSystemMetrics(SM_CXBORDER));
  end;

  if UseRightToLeftAlignment and (ScrollBars <> ssNone) then
  begin
    Inc(D, GetSystemMetrics(SM_CXVSCROLL));
  end;
  Inc(Result.Left, D);
  Inc(Result.Right, D);
  D:=Top;
  if BorderStyle = bsSingle then
  begin
    if Ctl3D then
      Inc(D, GetSystemMetrics(SM_CYEDGE))
    else
      Inc(D, GetSystemMetrics(SM_CYBORDER));
  end;
  Inc(Result.Top, D);
  Inc(Result.Bottom, D);
end;

function TrDBGrid.TitleRectToTextRect(RCell: TRect): TRect;
begin
  Result:=RCell;
  // add space for title line, filterbar and groups
  Inc(Result.Top, FTitleLineText.SumHeight);
  Inc(Result.Top, FOptionsEx2.FFilters.TextBarTopHeight);
  Inc(Result.Top, FOptionsEx2.FGroups.GroupHeight);
  Dec(Result.Bottom, FOptionsEx2.FFilters.TextBarBootomHeight);
end;

procedure TrDBGrid.InvalidateRowEx(ARow: integer);
var
  I: integer;
begin
  if not HandleAllocated then
    Exit;
  for I:=0 to ColCount - 1 do
    self.InvalidateCell(I, ARow);
end;

procedure TrDBGrid.InvalidateHotTrackCell;
begin
  if (GetHotTrackCoord(FHotTrackCell).X > -1) or (GetHotTrackCoord(FHotTrackCell).Y > -1) then
    InvalidateCell(GetHotTrackCoord(FHotTrackCell).X, GetHotTrackCoord(FHotTrackCell).Y);
end;

procedure TrDBGrid.ResetHotTrackCell;
begin
  SetHotTrackCoord(FHotTrackCell, -1, -1);
end;

// ==============================================================================
// DateTimePicker block
// FPopupCalendar is used instead of TDateTimePicker if compiler switch TrPopupMonthCalendar is defined
{$IFNDEF TrPopupMonthCalendar}

function TrDBGrid.CreateDateTimePicker: TWinControl;
begin
  // Init DateTimePicker
  Result:=nil;
  if Assigned(FDateTimePickerCreate) then
    FDateTimePickerCreate(self, Result, DateTimePickerChange, DateTimePickerClose)
  else
  begin
    Result:=TDateTimePicker.Create(self);
    TDateTimePicker(Result).OnChange:=DateTimePickerChange;
    TDateTimePicker(Result).OnCloseUp:=DateTimePickerClose;
  end;
  if Assigned(Result) then
  begin
    Result.SetSubComponent(true);
    Result.Visible:=false;
    Result.Parent:=self; // self.Parent;
    Result.BoundsRect:=Rect(0, 0, 0, 0);
  end;
end;

procedure TrDBGrid.SetDateTimePickerDate(ADateTime: TDateTime);
var
  Done: boolean;
begin
  Done:=false;
  if Assigned(FDateTimePickerEvent) then
    FDateTimePickerEvent(self, FDateTimePicker, dpaSetDate, ADateTime, Done);
  if (FDateTimePicker is TDateTimePicker) and not Done then
    try
      if ADateTime = 0 then
        TDateTimePicker(FDateTimePicker).Date:=Date
      else
        TDateTimePicker(FDateTimePicker).Date:=ADateTime;
    except
      Beep;
    end;
end;

function TrDBGrid.GetDateTimePickerDate: TDateTime;
var
  Done: boolean;
begin
  Done:=false;
  Result:=Date;
  if Assigned(FDateTimePickerEvent) then
    FDateTimePickerEvent(self, FDateTimePicker, dpaGetDate, Result, Done);
  if (FDateTimePicker is TDateTimePicker) and not Done then
    Result:=TDateTimePicker(FDateTimePicker).Date;
end;

procedure TrDBGrid.ShowDateTimePicker(Rect: TRect);
var
  dt: TDateTime;
  Done: boolean;
begin
  if FDateTimePicker = nil then
    Exit;
  Done:=false;
  dt:=Date;
  if Rect.Right - Rect.Left < UseScale(FOptionsEx2.FEditors.DateTimePickerMinWidth) then
    Rect.Right:=Rect.Left + UseScale(FOptionsEx2.FEditors.DateTimePickerMinWidth);
  FDateTimePicker.BoundsRect:=Rect; // ClientRectToParentRect(Rect);
  if Assigned(FDateTimePickerEvent) then
    FDateTimePickerEvent(self, FDateTimePicker, dpaShow, dt, Done);
  if not Done then
  begin
    FDateTimePicker.Show;
    try
      FDateTimePicker.SetFocus;
      // open picker
{$IFDEF DXE2P}
      if TStyleManager.IsCustomStyleActive then
      begin
        // refresh height of picker to get correct calendar size for style
        SendMessage(FDateTimePicker.Handle, CM_FONTCHANGED, 0, 0);
        SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
        // resize back to cell size
        FDateTimePicker.BoundsRect:=Rect;
      end
      else
{$ENDIF}
        SendMessage(FDateTimePicker.Handle, WM_SYSKEYDOWN, VK_DOWN, 0);
    except
      FDateTimePicker.Hide;
    end;
  end;
end;

procedure TrDBGrid.DateTimePickerChange(Sender: TObject);
var
  F: TField;
begin
  F:=Columns[SelectedIndex].Field;
  try
    SetDateValueEx(F, Trunc(GetDateTimePickerDate));
    if Assigned(InplaceEditor) then
      InplaceEditor.Text:=GetEditText(Col, Row);
  except
  end;
end;

procedure TrDBGrid.DateTimePickerClose(Sender: TObject);
var
  dt: TDateTime;
  Done: boolean;
begin
  Done:=false;
  dt:=Date;
  if FDateTimePicker <> nil then
  begin
    if Assigned(FDateTimePickerEvent) then
      FDateTimePickerEvent(self, FDateTimePicker, dpaHide, dt, Done);
    if not Done then
      FDateTimePicker.Hide;
  end;
  try
    SetFocus;
    if Assigned(InplaceEditor) then
    begin
      InplaceEditor.SetFocus;
      InplaceEditor.SelectAll;
    end;
  except
  end;
end;
{$ENDIF}

// ==============================================================================
// Inplace Editor block
function TrDBGrid.CreateEditor: TInplaceEdit;
begin
  Result:=TrDBGridInplaceEditEx.Create(self);
end;

function TrDBGrid.GetInplaceEditor: TrDBGridInplaceEditEx;
begin
  Result:=InplaceEditor as TrDBGridInplaceEditEx;
end;

procedure TrDBGrid.HideInplaceEditor;
begin
  HideEdit;
end;

function TrDBGrid.GetEditStyle(ACol, ARow: Longint): TEditStyle;
begin
  if IsDateEx(SelectedField) then
    Result:=esEllipsis
  else
    Result:= inherited GetEditStyle(ACol, ARow);
end;

// ==============================================================================
// Aux funct block
function TrDBGrid.IsDatasetLinked: boolean;
begin
  Result:=Assigned(DataLink) and Assigned(DataLink.DataSet);
end;

function TrDBGrid.IsActive: boolean;
begin
  Result:=IsDatasetLinked and DataLink.Active;
end;

function TrDBGrid.IsFiltered: boolean;
begin
  Result:=IsActive and DataLink.DataSet.Filtered and (Trim(DataLink.DataSet.Filter) <> '');
end;

function TrDBGrid.IsEmpty: boolean;
begin
  Result:=not IsActive or DataLink.DataSet.IsEmpty;
end;

function TrDBGrid.IsNotEmpty: boolean;
begin
  Result:=IsActive and not DataLink.DataSet.IsEmpty;
end;

function TrDBGrid.IsFieldAsBoolean(F: TField): boolean;
begin
  Result:=IsBoolEx(F);
end;

function TrDBGrid.FieldAsBooleanValue(F: TField): boolean;
var
  FldText: string;
begin
  if F is TBooleanField then
    Result:=F.AsBoolean
  else
  begin
    // find index and bool value
    FldText:=FFieldsAsBoolean.Values[F.FieldName];
    if FldText = '' then
      FldText:=rDefBoolValue;
    FldText:=Trim(rtool.FindStrPart(FldText, FieldListSep, 1, false));
    Result:=CompareText(F.AsString, FldText) = 0;
  end;
end;

function TrDBGrid.TitleRow: integer;
begin
  if dgTitles in Options then
    Result:=0
  else
    Result:=-1;
end;

function TrDBGrid.FixCol: integer;
begin
  Result:=IndicatorOffset - 1;
end;

function TrDBGrid.CheckCoord(Coord1, Coord2: TGridCoord): boolean;
begin
  Result:=(Coord1.X = Coord2.X) and (Coord1.Y = Coord2.Y);
end;

function TrDBGrid.CheckCoordXY(Coord: TGridCoord; X, Y: integer): boolean;
begin
  Result:=(Coord.X = X) and (Coord.Y = Y);
end;

function TrDBGrid.IsBoolEx(F: TField): boolean;
begin
  Result:=Assigned(F) and FOptionsEx2.FEditors.FBoolGlyphs and ((F is TBooleanField) or (FFieldsAsBoolean.IndexOfName(F.FieldName) > -1));
  DoGeneralEventFieldBool(geIsBool, F, Result);
end;

function TrDBGrid.IsMemoEx(F: TField; TextAsMemo: boolean): boolean;
begin
  Result:=Assigned(F) and (FOptionsEx2.FEditors.ShowMemoGlyph or FOptionsEx2.FEditors.ShowMemoContent) and IsMemoField(F);
  if TextAsMemo and (F is TStringField) or (F is TWideStringField) then
    if (FOptionsEx2.FEditors.FTextAsMemoSize > 0) and (FOptionsEx2.FEditors.FTextAsMemoSize < F.Size) then
      Result:=true;
  DoGeneralEventFieldBool(geIsMemo, F, Result);
end;

function TrDBGrid.IsRichTextEx(F: TField): boolean;
begin
  Result:=IsMemoEx(F, false) and (FFieldsAsRichText.IndexOf(F.FieldName) > -1);
  DoGeneralEventFieldBool(geIsRichEdit, F, Result);
end;

function TrDBGrid.IsMemoAsGlyph(F: TField; TextAsMemo: boolean = false): boolean;
begin
  Result:=IsMemoEx(F, TextAsMemo) and FOptionsEx2.FEditors.ShowMemoGlyph;
  DoGeneralEventFieldBool(geIsMemoGlyph, F, Result);
end;

function TrDBGrid.IsMemoAsText(F: TField; TextAsMemo: boolean = false): boolean;
begin
  Result:=IsMemoEx(F, TextAsMemo) and FOptionsEx2.FEditors.ShowMemoContent;
  DoGeneralEventFieldBool(geIsMemoText, F, Result);
end;

function TrDBGrid.IsGraphicEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (FOptionsEx2.FEditors.ShowGraphicGlyph or FOptionsEx2.FEditors.ShowGraphicContent) and
    (IsGraphicField(F) or (FFieldsAsGraphics.IndexOf(F.FieldName) > -1));
  DoGeneralEventFieldBool(geIsGraphic, F, Result);
end;

function TrDBGrid.IsFieldAsDateEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (FFieldsAsDate.IndexOf(F.FieldName) >= 0);
  DoGeneralEventFieldBool(geIsFieldAsDate, F, Result);
end;

function TrDBGrid.IsDateEx(F: TField): boolean;
begin
  Result:=Assigned(F) and FOptionsEx2.FEditors.FDateTimePicker and (IsFieldAsDateEx(F) or IsDateField(F));
  DoGeneralEventFieldBool(geIsDate, F, Result);
end;

function TrDBGrid.IsFieldAsTimeEx(F: TField): boolean;
begin
  Result:=Assigned(F) and ((FFieldsAsTime.IndexOf(F.FieldName) >= 0) or (FFieldsAsTime.IndexOfName(F.FieldName) >= 0));
  DoGeneralEventFieldBool(geIsFieldAsTime, F, Result);
end;

function TrDBGrid.IsTimeEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (IsFieldAsTimeEx(F) or IsTimeField(F));
  DoGeneralEventFieldBool(geIsTime, F, Result);
end;

function TrDBGrid.IsColorEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (FFieldsAsColor.IndexOf(F.FieldName) >= 0);
  DoGeneralEventFieldBool(geIsColor, F, Result);
end;

function TrDBGrid.IsLinkEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (FFieldsAsLink.IndexOfName(F.FieldName) >= 0);
  DoGeneralEventFieldBool(geIsLink, F, Result);
end;

function TrDBGrid.IsPasswordEx(F: TField): boolean;
begin
  Result:=Assigned(F) and ((FFieldsAsPassword.IndexOf(F.FieldName) >= 0) or (FFieldsAsPassword.IndexOfName(F.FieldName) >= 0));
  DoGeneralEventFieldBool(geIsPassword, F, Result);
end;

function TrDBGrid.IsDataBarEx(F: TField): boolean;
begin
  Result:=Assigned(F) and ((FFieldsAsDataBar.IndexOf(F.FieldName) >= 0) or (FFieldsAsDataBar.IndexOfName(F.FieldName) >= 0));
  DoGeneralEventFieldBool(geIsDataBar, F, Result);
end;

function TrDBGrid.IsStatusBarEx(F: TField): boolean;
begin
  Result:=Assigned(F) and ((FFieldsAsStatusBar.IndexOf(F.FieldName) >= 0) or (FFieldsAsStatusBar.IndexOfName(F.FieldName) >= 0));
  DoGeneralEventFieldBool(geIsStatusBar, F, Result);
end;

function TrDBGrid.IsFixedColumnWidth(F: TField; FieldName: string): boolean;
begin
  // Result:=Assigned(F) and (FFieldsFixedColumnWidth.IndexOf(F.FieldName) >= 0);
  Result:=FFieldsFixedColumnWidth.IndexOf(FieldName) >= 0;
  DoGeneralEventFieldBool(geIsFixedColumn, F, Result);
end;

function TrDBGrid.IsDrawDropDownEx(F: TField): boolean;
begin
  Result:=Assigned(F) and (FFieldsDrawDropDown.IndexOf(F.FieldName) >= 0);
  DoGeneralEventFieldBool(geIsDrawDropDown, F, Result);
end;

function TrDBGrid.IsDataImageEx(F: TField): boolean;
begin
  Result:=Assigned(F) and FDataImagesIni.SectionExists(F.FieldName);
  DoGeneralEventFieldBool(geIsDataImage, F, Result);
end;

function TrDBGrid.GetDataImageDef(F: TField; var DataImageDef: TDataImageDef): boolean;

  function CheckNumCond(Val: double): string;
  var
    L: TStrings;
    I: integer;
    s, cond: string;
    v: double;
    D: TDateTime;
    found: boolean;
  begin
    Result:='';
    L:=TStringList.Create;
    try
      // load section values
      FDataImagesIni.ReadSectionValues(F.FieldName, L);
      for I:=0 to L.Count - 1 do
      begin
        s:=L.Names[I];
        cond:='';
        // extract cond
        while Pos(copy(s, 1, 1), '<>~') > 0 do
        begin
          cond:=cond + copy(s, 1, 1);
          s:=copy(s, 2, Length(s));
        end;
        try
          if rdbtool.IsDateTimeField(F) then
          begin
            found:=TryStrToDate(s, D);
            v:=D;
          end
          else
            found:=TryStrToFloat(s, v);
          if found then
            found:=((cond = '<') and (Val < v)) or ((cond = '>') and (Val > v)) or ((cond = '<>') and (Val <> v)) or ((cond = '<~') and (Val <= v)) or
              ((cond = '>~') and (Val >= v));
          if found then
          begin
            Result:=L.ValueFromIndex[I];
            Break;
          end;
        except
          // nothing
        end;
      end;
    finally
      FreeAndNil(L);
    end;
  end;

var
  IniText, FldText: string;
  ImgIndex: integer;
begin
  Result:=false;
  if not Assigned(F) then
    Exit;
  FldText:=F.AsString;
  // try to find exact value
  IniText:=FDataImagesIni.ReadString(F.FieldName, FldText, '');
  if (IniText = '') and ((F is TNumericField) or rdbtool.IsDateTimeField(F)) then
    // check condition
    IniText:=CheckNumCond(F.AsFloat);
  if IniText = '' then
    // find default value
    IniText:=FDataImagesIni.ReadString(F.FieldName, '*', '');
  DoGeneralEventParam(geGetDataImageIndex, nil, F, nil, IniText);
  IniText:=FGeneralEventRec.ParamValue;
  if IniText <> '' then
  begin
    DataImageDef.UseDataImage:=true;
    ImgIndex:=StrToIntDef(IniText, -2);
    DataImageDef.ImageIndex:=ImgIndex;
    DataImageDef.ImageAlignment:=FDataImagesIni.ReadString(F.FieldName, DataImagesFieldCfg, '');
    DoGeneralEventParam(geGetDataImageAlignment, nil, F, nil, DataImageDef.ImageAlignment);
    DataImageDef.ImageAlignment:=FGeneralEventRec.ParamValue;
    if ImgIndex = -2 then
      DataImageDef.ShowText:=IniText
    else
      if DataImageDef.ImageAlignment <> '' then
        DataImageDef.ShowText:=F.DisplayText;
  end;
  Result:=DataImageDef.UseDataImage;
end;

function TrDBGrid.Fields_GetLinkType(FieldName: string): TrLinkType;
var
  s: string;
begin
  s:=FFieldsAsLink.Values[FieldName];
  Result:=ltNone;
  if s <> '' then
    Result:=ltUserDefined;
  if s = rRunHTTP then
    Result:=ltHTTP;
  if s = rRunMailTo then
    Result:=ltMailTo;
end;

procedure TrDBGrid.Fields_SetLinkType(FieldName: string; LinkType: TrLinkType);
var
  s: string;
begin
  if Trim(FieldName) = '' then
    Exit;
  case LinkType of
    ltNone:
      s:='';
    ltUserDefined:
      s:='user_defined';
    ltHTTP:
      s:=rRunHTTP;
    ltMailTo:
      s:=rRunMailTo;
  end;
  FFieldsAsLink.Values[FieldName]:=s;
end;

function TrDBGrid.Fields_GetBoolValues(FieldName: string): string;
begin
  Result:=FFieldsAsBoolean.Values[FieldName];
end;

procedure TrDBGrid.Fields_SetBoolValues(FieldName: string; BoolValues: string);
begin
  if Trim(FieldName) = '' then
    Exit;
  FFieldsAsBoolean.Values[FieldName]:=BoolValues;
end;

function TrDBGrid.Fields_GetDataBarSettings(FieldName: string; var MinValue, MaxValue: double; var ValueType: TrDataBarText; var ColorFrom, ColorTo: TColor)
  : boolean;
var
  s: string;
  I: integer;
begin
  Result:=false;
  if copy(FieldName, 1, 1) = '*' then
    s:=copy(FieldName, 2, Length(FieldName))
  else
    s:=FFieldsAsDataBar.Values[FieldName];
  DoGeneralEventParam(geGetDataBarParam, nil, DataLink.DataSet.FindField(FieldName), nil, s);
  s:=FGeneralEventRec.ParamValue;
  if s = '' then
    Exit;
  MinValue:=StrToFloatDef(UpdateToDecimalSeparator(rtool.RemoveStrPart(s, FieldListSep)), 0);
  MaxValue:=StrToFloatDef(UpdateToDecimalSeparator(rtool.RemoveStrPart(s, FieldListSep)), 100);
  I:=StrToIntDef(rtool.RemoveStrPart(s, FieldListSep), 0);
  if I > ord( High(TrDataBarText)) then
    ValueType:= High(TrDataBarText)
  else
    ValueType:=TrDataBarText(I);
  ColorFrom:=StringToColor(rtool.RemoveStrPart(s, FieldListSep));
  ColorTo:=StringToColor(rtool.RemoveStrPart(s, FieldListSep));
  Result:=true;
end;

function TrDBGrid.Fields_SetDataBarSettings(FieldName: string; MinValue, MaxValue: double; ValueType: TrDataBarText; ColorFrom, ColorTo: TColor): string;
var
  res: string;

  procedure AddProps(s: string); overload;
  begin
    if res <> '' then
      res:=res + FieldListSep;
    res:=res + s;
  end;
  procedure AddProps(v: double); overload;
  begin
    AddProps(FloatToStr(v));
  end;

begin
  res:='';
  if (MinValue <> 0) or (MaxValue <> 0) then
  begin
    AddProps(MinValue);
    AddProps(MaxValue);
    AddProps(ord(ValueType));
    if ColorFrom <> clNone then
    begin
      AddProps(rpictool.ColorToString(ColorFrom, false));
      if ColorTo <> clNone then
        AddProps(rpictool.ColorToString(ColorTo, false));
    end;
  end;
  if FieldName <> '' then
  begin
    FFieldsAsDataBar.Values[FieldName]:=res;
    Invalidate;
  end;
  Result:=res;
end;

function TrDBGrid.Fields_GetStatusBarSettings(FieldName: string; var Steps: integer; var BarType: TrStatusBarType; var ColorAll, ColorDone: TColor): boolean;
var
  s: string;
  I: integer;
begin
  Result:=false;
  if copy(FieldName, 1, 1) = '*' then
    s:=copy(FieldName, 2, Length(FieldName))
  else
    s:=FFieldsAsStatusBar.Values[FieldName];
  DoGeneralEventParam(geGetStatusBarParam, nil, DataLink.DataSet.FindField(FieldName), nil, s);
  s:=FGeneralEventRec.ParamValue;
  if s = '' then
    Exit;
  Steps:=StrToIntDef(rtool.RemoveStrPart(s, FieldListSep), 5);
  I:=StrToIntDef(rtool.RemoveStrPart(s, FieldListSep), 0);
  if I > ord( High(TrStatusBarType)) then
    BarType:= High(TrStatusBarType)
  else
    BarType:=TrStatusBarType(I);
  ColorAll:=StringToColor(rtool.RemoveStrPart(s, FieldListSep));
  ColorDone:=StringToColor(rtool.RemoveStrPart(s, FieldListSep));
  Result:=true;
end;

function TrDBGrid.Fields_SetStatusBarSettings(FieldName: string; Steps: integer; BarType: TrStatusBarType; ColorAll, ColorDone: TColor): string;
var
  res: string;

  procedure AddProps(s: string); overload;
  begin
    if res <> '' then
      res:=res + FieldListSep;
    res:=res + s;
  end;
  procedure AddProps(I: integer); overload;
  begin
    AddProps(IntToStr(I));
  end;

begin
  res:='';
  if Steps > 0 then
  begin
    AddProps(Steps);
    AddProps(ord(BarType));
    if ColorAll <> clNone then
    begin
      AddProps(rpictool.ColorToString(ColorAll, false));
      if ColorDone <> clNone then
        AddProps(rpictool.ColorToString(ColorDone, false));
    end;
  end;
  if FieldName <> '' then
  begin
    FFieldsAsStatusBar.Values[FieldName]:=res;
    Invalidate;
  end;
  Result:=res;
end;

function TrDBGrid.FooterRow_GetSettings(FieldName: string; var ContentMask: string; var TextAlignment: TAlignmentDefault; var FontStyles: TFontStylesDefault;
  var FontColor: TColor; var FontSize: integer; AFont: TFont = nil): boolean;
var
  s, cfg: string;
  F: TFont;
  NormalStyleDef: boolean;
begin
  Result:=false;
  if copy(FieldName, 1, 1) = '*' then
    s:=copy(FieldName, 2, Length(FieldName))
  else
    s:=FFooterRow.FFieldFooterDefs.Values[FieldName];
  DoGeneralEventParam(geGetFooterRowParam, nil, DataLink.DataSet.FindField(FieldName), nil, s);
  s:=FGeneralEventRec.ParamValue;
  if s = '' then
    Exit;
  F:=TFont.Create;
  try
    F.Style:=[];
    F.Size:=0;
    F.Color:=clNone;
    ContentMask:=RemoveStrPart(s, FieldListSep);
    NormalStyleDef:=false;
    TextAlignment:=tadDefault;
    while s <> '' do
    begin
      cfg:=AnsiUpperCase(Trim(rtool.RemoveStrPart(s, FieldListSep)));
      if cfg = 'L' then
        TextAlignment:=tadLeft
      else
        if cfg = 'R' then
          TextAlignment:=tadRight
        else
          if cfg = 'C' then
            TextAlignment:=tadCenter
          else
          begin
            UseFontFormatParams(F, cfg, NormalStyleDef);
            if Assigned(AFont) then
              UseFontFormatParams(AFont, cfg);
          end;
    end;
    FontStyles:=[];
    if NormalStyleDef then
      Include(FontStyles, fsdNormal);
    if fsBold in F.Style then
      Include(FontStyles, fsdBold);
    if fsItalic in F.Style then
      Include(FontStyles, fsdItalic);
    if fsUnderline in F.Style then
      Include(FontStyles, fsdUnderline);
    FontColor:=F.Color;
    FontSize:=F.Size;
  finally
    FreeAndNil(F);
  end;
  Result:=true;
end;

function TrDBGrid.FooterRow_SetSettings(FieldName: string; ContentMask: string; TextAlignment: TAlignmentDefault; FontStyles: TFontStylesDefault;
  FontColor: TColor; FontSize: integer): string;
var
  res: string;

  procedure AddProps(s: string); overload;
  begin
    if res <> '' then
      res:=res + FieldListSep;
    res:=res + s;
  end;
  procedure AddProps(I: integer); overload;
  begin
    AddProps(IntToStr(I));
  end;

begin
  res:='';
  if ContentMask <> '' then
  begin
    AddProps(ContentMask);
    if fsdNormal in FontStyles then
      AddProps('N');
    if fsdBold in FontStyles then
      AddProps('B');
    if fsdItalic in FontStyles then
      AddProps('I');
    if fsdUnderline in FontStyles then
      AddProps('U');
    case TextAlignment of
      tadLeft:
        AddProps('L');
      tadCenter:
        AddProps('C');
      tadRight:
        AddProps('R');
    end;
    if FontColor <> clNone then
      AddProps(rpictool.ColorToString(FontColor, false));
    if FontSize > 0 then
      AddProps(FontSize);
  end;
  if FieldName <> '' then
  begin
    FFooterRow.FFieldFooterDefs.Values[FieldName]:=res;
    Invalidate;
  end;
  Result:=res;
end;

function TrDBGrid.Title_GetImageIndex(FieldName: string): integer;
begin
  Result:=StrToIntDef(FTitleImages.Values[FieldName], -1);
end;

procedure TrDBGrid.Title_SetImageIndex(FieldName: string; ImageIndex: integer);
begin
  if Trim(FieldName) = '' then
    Exit;
  if ImageIndex < 0 then
    FTitleImages.Values[FieldName]:=''
  else
    FTitleImages.Values[FieldName]:=IntToStr(ImageIndex);
end;

function TrDBGrid.Title_GetHint(FieldName: string): string;
begin
  Result:=FTitleHints.Values[FieldName];
end;

procedure TrDBGrid.Title_SetHint(FieldName: string; Hint: string);
begin
  if Trim(FieldName) = '' then
    Exit;
  FTitleHints.Values[FieldName]:=Hint;
end;

function TrDBGrid.DataImages_GetSettings(FieldName: string; ConditionList: TStrings; var AddImageAlign: string): integer;
var
  I: integer;
begin
  FDataImagesIni.ReadSectionValues(FieldName, ConditionList);
  I:=ConditionList.IndexOfName(DataImagesFieldCfg);
  if I >= 0 then
  begin
    AddImageAlign:=ConditionList.ValueFromIndex[I];
    ConditionList.Delete(I);
  end
  else
    AddImageAlign:='';
  Result:=ConditionList.Count;
end;

procedure TrDBGrid.DataImages_SetSettings(FieldName: string; ConditionList: TStrings; AddImageAlign: string);
var
  I: integer;
  L: TStrings;
begin
  FDataImagesIni.EraseSection(FieldName);
  if Assigned(ConditionList) then
  begin
    for I:=0 to ConditionList.Count - 1 do
      FDataImagesIni.WriteString(FieldName, ConditionList.Names[I], ConditionList.ValueFromIndex[I]);
    if AddImageAlign <> '' then
      FDataImagesIni.WriteString(FieldName, DataImagesFieldCfg, AddImageAlign);
  end;
  L:=TStringList.Create;
  try
    FDataImagesIni.GetStrings(L);
    FDataImages.Clear;
    FDataImages.Assign(L);
  finally
    FreeAndNil(L);
  end;
end;

function TrDBGrid.IsTitleImageEx(F: TField): boolean;
var
  I: integer;
begin
  Result:=Assigned(F) and GetTitleImageIndex(F, I);
  // DoGeneralEventFieldBool(geIsTitleImage, F, Result);
end;

function TrDBGrid.GetTitleImageIndex(F: TField; var ImgIdx: integer): boolean;
begin
  ImgIdx:=-1;
  if Assigned(F) and Assigned(FTitleImageList) and (FTitleImages.IndexOfName(F.FieldName) >= 0) then
  begin
    ImgIdx:=StrToIntDef(FTitleImages.Values[F.FieldName], -1);
    if ImgIdx >= FTitleImageList.Count then
      ImgIdx:=-1;
  end;
  DoGeneralEventParam(geGetTitleImageIndex, nil, F, nil, ImgIdx);
  ImgIdx:=FGeneralEventRec.ParamValue;
  Result:=ImgIdx >= 0;
end;

function TrDBGrid.ChangeBool(reset: boolean): boolean;
var
  v, v1, v2: string;
  wasEditing: boolean;
begin
  Result:=false;
  wasEditing:=DataLink.Editing;
  if CanEdit then
    try
      if SelectedField is TBooleanField then
      begin
        if reset then
          SelectedField.AsBoolean:=false
        else
          SelectedField.AsBoolean:=not SelectedField.AsBoolean;
      end
      else
      begin
        // find index and bool value
        v:=FFieldsAsBoolean.Values[SelectedField.FieldName];
        if v = '' then
          v:=rDefBoolValue;
        v1:=Trim(rtool.RemoveStrPart(v, FieldListSep, true));
        v2:=Trim(rtool.RemoveStrPart(v, FieldListSep, true));
        if reset then
          SelectedField.AsString:=v2
        else
          if CompareText(SelectedField.AsString, v1) = 0 then
            SelectedField.AsString:=v2
          else
            SelectedField.AsString:=v1;
      end;
      Result:=true;
      if Assigned(FOnBoolClick) then
        FOnBoolClick(self, DataLink.DataSet, Columns[SelectedIndex], SelectedField, SelectedField.AsVariant);
      if FOptionsEx2.FEditors.FBoolAutoSave and not wasEditing then
        DataLink.DataSet.Post;
    except
    end;
end;

function TrDBGrid.GetDateValueEx(F: TField): TDateTime;
begin
  if IsFieldAsDateEx(F) then
  begin
    if F is TNumericField then
      Result:=F.AsFloat
    else
      if F.AsString = '' then
        Result:=0
      else
        Result:=StrToDate(F.AsString);
  end
  else
    Result:=F.AsDateTime;
end;

procedure TrDBGrid.SetDateValueEx(F: TField; Value: TDateTime);
begin
  if IsFieldAsDateEx(F) then
  begin
    if F is TNumericField then
      F.AsFloat:=Value
    else
      F.AsString:=DateToStr(Value)
  end
  else
    F.AsDateTime:=Value;
end;

function TrDBGrid.GetTimeValueEx(F: TField): TDateTime;
var
  s: string;
begin
  if IsFieldAsTimeEx(F) then
  begin
    if F is TNumericField then
      Result:=F.AsFloat
    else
      if F.AsString = '' then
        Result:=0
      else
      begin
        s:=copy(F.AsString, 1, 8);
        Result:=StrToTime(s);
      end;
  end
  else
    Result:=F.AsDateTime;
end;

procedure TrDBGrid.SetTimeValueEx(F: TField; Value: TDateTime);
begin
  if IsFieldAsTimeEx(F) then
  begin
    if F is TNumericField then
      F.AsFloat:=Value
    else
      F.AsString:=TimeToStr(Value)
  end
  else
    F.AsDateTime:=Value;
end;

procedure TrDBGrid.ShowMemo(Field: TField);
var
  n: string;
begin
  if not DoGeneralEventField(geShowMemoDialog, Field) then
    Exit;
  // Create and show Memo Form
  n:=Parent.Name + '_' + Name + '_MemoForm';
  SelectedField:=Field;
  with TrDBGridMemoForm.Create(self) do
    try
      TempBmp.Assign(rGlyphs.BmpFromIndex(rGlyphs.MemoIndex(false, ord(FGlyphLookStyle)), FScaleRatio, FCustomGlyphs));
      ConvertBmpToIco(TempBmp, Icon, 16);
      Name:=n;
      Position:=poDesigned;
      Caption:=Field.DisplayLabel; // Columns[SelectedIndex].Title.Caption;
      IsEdit:=CanEdit;
      IsRichEdit:=IsRichTextEx(Field);
      Text:=Field.AsString;
      Left:=ClientOrigin.X + Width - Width - 10;
      Top:=ClientOrigin.Y + Height - Height - 10;
      DlgPropSave.PropSaveFile:=FPropSaveFile;
      if ShowModal = mrOk then
      begin
        Field.AsString:=Text;
        InvalidateEditor;
      end;
    finally
      Free;
    end;
end;

procedure TrDBGrid.ShowGraphic(Field: TField);
var
  n: string;
begin
  if not DoGeneralEventField(geShowGraphicDialog, Field) then
    Exit;
  // Create and show Graphic form
  n:=Parent.Name + '_' + Name + '_GraphicForm';
  SelectedField:=Field;
  with TrDBGridGraphicForm.Create(self) do
  begin
    try
      TempBmp.Assign(rGlyphs.BmpFromIndex(rGlyphs.PictureIndex(false, ord(FGlyphLookStyle)), FScaleRatio, FCustomGlyphs));
      ConvertBmpToIco(TempBmp, Icon, 16);
      Name:=n;
      Position:=poDesigned;
      Caption:=Field.DisplayLabel; // Columns[SelectedIndex].Title.Caption;
      DBImg.DataSource:=DataSource;
      DBImg.DataField:=Field.FieldName;
      IsEdit:=CanEdit;
      Left:=ClientOrigin.X + Width - Width - 10;
      Top:=ClientOrigin.Y + Height - Height - 10;
      DlgPropSave.PropSaveFile:=FPropSaveFile;
      ShowModal;
    except
      on E: Exception do
        ShowMessage(E.Message);
    end;
    Free;
  end;
end;

function TrDBGrid.GetMouseOverCellHint(var C: TColumn; var Text: string; var AlwaysShow: boolean): boolean;

  function GetPictureHint(F: TField): string;
  var
    P: TPicture;
    GrType: TrGraphicType;
  begin
    P:=TPicture.Create;
    try
      GrType:=ReadPictureFromField(F, P);
      case GrType of
        gtNone:
          Result:=sPictureEmpty;
        gtError:
          Result:=sPictureCorrupted;
      else
        begin
          Result:=GraphicTypeLabel[GrType] + ': ' + Format('%d x %d', [P.Width, P.Height]);
          if Assigned(RosiCompConfig.Global_HTMLHintPicture) then
          begin
            RosiCompConfig.Global_HTMLHintPicture.Assign(P);
            Result:=Result + rHTMLHintImageTag;
          end;
        end;
      end;
    finally
      FreeAndNil(P);
    end;
  end;

var
  OldRec, M: integer;
  R, I: integer;
  s: string;
begin
  // Return full text of cell under mouse point
  C:=nil;
  Text:='';
  AlwaysShow:=false;
  FHintFieldValues.Clear;
  if IsActive then
  begin
    try
      // titleex show record count as hint (ShowHintRecCount=true)
      if CheckCoordXY(MMoveCoord, FixCol, TitleRow) and (FOptionsEx2.FFixedCell.FTitleExBtnHintRecCount or ShowHintRecCount) then
      begin
        Text:=Format(sPrintRecCount, [DataLink.DataSet.RecordCount]);
        ShowHintRecCount:=false;
        AlwaysShow:=true;
      end;
      R:=RawToDataColumn(MMoveCoord.X);
      if (R >= 0) then
        C:=Columns[R];
      if Assigned(C) then
      begin
        if MMoveCoord.Y = cFooterRowYCoord then
        begin
          // footer row
          Text:=FFooterRow.GetFooterText(C.Field);
          Text:=ReplaceStr('|', cStrEOL, Text);
        end
        else
          if MMoveCoord.Y = TitleRow then
            Text:=C.Title.Caption
          else
          begin
            OldRec:=DataLink.ActiveRecord;
            try
              DataLink.ActiveRecord:=MMoveCoord.Y - TitleRow - 1;
              if Assigned(C.Field) then
                if IsGraphicEx(C.Field) then
                begin
                  Text:=GetPictureHint(C.Field);
                  AlwaysShow:=true;
                end
                else
                  if Assigned(FRichText) and IsRichTextEx(C.Field) then
                  begin
                    Text:=C.Field.AsString;
                    if Text <> '' then
                    begin
                      if Assigned(RosiCompGlobalCfg.RosiCompConfig.Global_HTMLHintComponent) then
                        Text:=Text + RosiCompGlobalCfg.rHTMLHintRichTextTag
                      else
                      begin
                        FRichText.RtfText:=Text;
                        Text:=FRichText.Lines.Text;
                      end;
                      AlwaysShow:=true;
                    end;
                  end
                  else
                    if IsMemoEx(C.Field, false) then
                    begin
                      Text:=C.Field.AsString;
                      AlwaysShow:=true;
                    end
                    else
                      if not(FOptionsEx2.FGroups.FActive and GetGroupMasterCellContent(C.Field, Text)) then
                        Text:=C.Field.DisplayText;
              if Assigned(FOnGetHintStringEx) then
              begin
                // store all field values from record under mouse
                for I:=0 to DataLink.DataSet.FieldCount - 1 do
                  try
                    if IsMemoField(DataLink.DataSet.Fields[I]) then
                      s:=DataLink.DataSet.Fields[I].AsString
                    else
                      s:=DataLink.DataSet.Fields[I].DisplayText;
                    s:=ReplaceStr(cStrEOL, RosiCompConfig.rDBGridCfg.CellSingleLineEOLSeparator, s);
                    FHintFieldValues.Values[DataLink.DataSet.Fields[I].FieldName]:=s;
                  except
                    // Beep;
                  end;
              end;
            finally
              DataLink.ActiveRecord:=OldRec;
            end;
          end;
      end;
    except
      Text:='';
    end;
  end;
  M:=FOptionsEx2.FAppearance.FHintMaxLength;
  if (M > 0) and (Length(Text) > M) then
  begin
    Text:=copy(Text, 1, M) + '...';
    AlwaysShow:=true;
  end;
  Result:=Text <> '';
end;

procedure TrDBGrid.ScanAllRecords(CallBackProc: TScanDatasetEvent; CallBackObject: TObject; ScanReason: integer; ProcAllRec: boolean = true;
  MaxRecords: integer = 0);
var
  rec: integer;
  Bk: TBookmark;
begin
  WCursor;
  try
    Bk:=DataLink.DataSet.GetBookmark;
    try
      DataLink.DataSet.DisableControls;
      SaveResetDatasetEvents(ScanReason);
      try
        rec:=0;
        if ProcAllRec then
          DataLink.DataSet.First;
        while not DataLink.DataSet.Eof do
        begin
          // proc
          CallBackProc(CallBackObject, DataLink.DataSet, rec);
          Inc(rec);
          if (MaxRecords > 0) and (rec > MaxRecords) then
            Break;
          if ProcAllRec then
            DataLink.DataSet.Next
          else
            Break;
        end;
        DataLink.DataSet.GotoBookmark(Bk);
      finally
        RestoreDatasetEvents(ScanReason);
        DataLink.DataSet.EnableControls;
      end;
    finally
      DataLink.DataSet.FreeBookmark(Bk);
    end;
  finally
    NCursor;
  end;
end;

function TrDBGrid.DoGeneralEvent(EventType: TGeneralEventType): boolean;
begin
  Result:=DoGeneralEventField(EventType, nil);
end;

function TrDBGrid.DoGeneralEventField(EventType: TGeneralEventType; F: TField): boolean;
var
  B: boolean;
begin
  B:=false;
  Result:=DoGeneralEventFieldBool(EventType, F, B);
end;

function TrDBGrid.DoGeneralEventFieldBool(EventType: TGeneralEventType; F: TField; var BoolValue: boolean): boolean;
begin
  Result:=DoGeneralEventParam(EventType, nil, F, nil, BoolValue);
  BoolValue:=FGeneralEventRec.ParamValue;
end;

function TrDBGrid.DoGeneralEventParam(EventType: TGeneralEventType; Column: TColumn; Field: TField; ParamObject: TObject; ParamValue: variant): boolean;
begin
  Result:=true;
  // FGeneralEventRec.Rect:=R;
  FGeneralEventRec.Column:=Column;
  FGeneralEventRec.Field:=Field;
  FGeneralEventRec.ParamObject:=ParamObject;
  FGeneralEventRec.ParamValue:=ParamValue;
  if Assigned(Field) and Field.DataSet.Active then
    FGeneralEventRec.FieldValue:=Field.AsVariant;
  if Assigned(FOnGeneralEvent) then
    FOnGeneralEvent(self, EventType, FGeneralEventRec, Result);
end;

// ==============================================================================
// SORT funct block
function TrDBGrid.TrimSortFields(const ASortFieldNames: string): string;
// check and format SortFieldNames
var
  s, F: string;
begin
  Result:='';
  s:=ASortFieldNames;
  repeat
    F:=Trim(rtool.RemoveStrPart(s, cFieldSep));
    if F <> '' then
    begin
      if (Result <> '') then
        Result:=Result + cFieldSep;
      Result:=Result + F;
    end;
    if s = '' then
      Break;
  until false;
end;

function TrDBGrid.TrimSortDirections(const ASortDirections: string): string;
// check and format SortDirections
var
  s, F: string;
begin
  Result:='';
  s:=ASortDirections;
  repeat
    F:=Trim(rtool.RemoveStrPart(s, cFieldSep));
    if F <> '' then
    begin
      if F <> 'D' then
        F:='A';
      if (Result <> '') then
        Result:=Result + cFieldSep;
      Result:=Result + F;
    end;
    if s = '' then
      Break;
  until false;
end;

function TrDBGrid.PosInSortField(const FieldName: string): integer;
// return index of field in SortFieldNames (0=not found)
var
  s, F: string;
  I: integer;
begin
  Result:=0;
  I:=1;
  s:=FSortFieldName;
  repeat
    F:=Trim(rtool.RemoveStrPart(s, cFieldSep));
    if AnsiSameText(F, Trim(FieldName)) then
      Result:=I;
    if s = '' then
      Break;
    Inc(I);
  until false;
end;

function TrDBGrid.GetSortDirection(const FieldName: string): boolean;
// return direction for field SelFldName
var
  I: integer;
  s: string;
begin
  Result:=FSortDesc;
  I:=PosInSortField(FieldName);
  if I = 0 then
    Exit;
  s:=rtool.FindStrPart(FSortDirections, cFieldSep, I, true);
  Result:=s = 'D';
end;

procedure TrDBGrid.CheckFormatSortDirection(var ASortFieldNames, ASortDirections: string; FieldToRemove: string);
var
  fList, dList, F, D: string;
begin
  if ASortFieldNames = '' then
    Exit;
  fList:=TrimSortFields(ASortFieldNames);
  dList:=UpperCase(TrimSortDirections(ASortDirections));
  ASortFieldNames:='';
  ASortDirections:='';
  repeat
    F:=rtool.RemoveStrPart(fList, cFieldSep);
    D:=rtool.RemoveStrPart(dList, cFieldSep);
    if F = '' then
      Break;
    if not AnsiSameText(F, FieldToRemove) then
    begin
      rtool.AddToSepList(ASortFieldNames, F, cFieldSep);
      if not(D = 'D') then
        D:='A';
      rtool.AddToSepList(ASortDirections, D, cFieldSep);
    end;
  until false;
end;

function TrDBGrid.ChangeSortDirection(const FieldName: string): string;
var
  I, flxidx: integer;
  s: string;
begin
  flxidx:=PosInSortField(FieldName);
  if flxidx = 0 then
    Result:=''
  else
  begin
    // change dir
    if GetSortDirection(FieldName) then
      s:='A'
    else
      s:='D';
    I:=(flxidx - 1) * 2 + 1;
    Result:=copy(FSortDirections, 1, I - 1) + s + copy(FSortDirections, I + 1, Length(FSortDirections));
  end;
end;

function TrDBGrid.GetSortFieldNames(const FieldList: string): TSortFieldNames;
var
  I: integer;
  s: string;
begin
  SetLength(Result, 0);
  I:=0;
  repeat
    Inc(I);
    s:=rtool.FindStrPart(FieldList, cFieldSep, I, true);
    if s = '' then
      Break;
    SetLength(Result, I);
    Result[I - 1]:=s;
  until false;
end;

function TrDBGrid.GetSortDirections(const DirectionList: string): TSortDirections;
var
  I: integer;
  s: string;
begin
  SetLength(Result, 0);
  I:=0;
  repeat
    Inc(I);
    s:=rtool.FindStrPart(DirectionList, cFieldSep, I, true);
    if s = '' then
      Break;
    SetLength(Result, I);
    Result[I - 1]:=s = 'D';
  until false;
end;

function TrDBGrid.SortFieldNamesArr: TSortFieldNames;
begin
  Result:=GetSortFieldNames(FSortFieldName);
end;

function TrDBGrid.SortDirectionsArr: TSortDirections;
begin
  Result:=GetSortDirections(FSortDirections);
end;

function TrDBGrid.ReSort: boolean;
begin
  if Assigned(FSorter) then
    Result:=FSorter.DoSort(self, DataLink.DataSet, GetSortFieldNames(FSortFieldName), GetSortDirections(FSortDirections))
  else
    if Assigned(FOnChangeSortEx) then
      Result:=ChangeSortEx(FSortFieldName, FSortDirections)
    else
      Result:=ChangeSort(FSortFieldName, FSortDesc);
end;

function TrDBGrid.ChangeSort(const NewSortFieldName: string; NewSortDesc: boolean): boolean;
var
  SortEx: string;
begin
  Result:=IsDatasetLinked and (Assigned(FSorter) or Assigned(FOnChangeSort));
  if Result then
  begin
    if NewSortDesc then
      SortEx:='D'
    else
      SortEx:='A';
    // Call external sort procedure
    try
      if Assigned(FSorter) then
        Result:=FSorter.DoSort(self, DataLink.DataSet, GetSortFieldNames(NewSortFieldName), GetSortDirections(SortEx))
      else
        if Assigned(FOnChangeSort) then
          FOnChangeSort(self, DataLink.DataSet, NewSortFieldName, NewSortDesc, Result);
    except
      Result:=false;
    end;
    if Result then
    begin
      FSortDesc:=NewSortDesc;
      SortFieldName:=NewSortFieldName;
    end;
  end
  else
  begin
    Result:=ChangeSortEx(NewSortFieldName, SortEx);
  end;
end;

function TrDBGrid.ChangeSortEx(const NewSortFieldName, NewSortDirections: string): boolean;
begin
  Result:=IsDatasetLinked and (Assigned(FSorter) or Assigned(FOnChangeSortEx));
  if Result then
  begin
    // Call external sort procedure
    try
      if Assigned(FSorter) then
        Result:=FSorter.DoSort(self, DataLink.DataSet, GetSortFieldNames(NewSortFieldName), GetSortDirections(NewSortDirections))
      else
        if Assigned(FOnChangeSortEx) then
          FOnChangeSortEx(self, DataLink.DataSet, GetSortFieldNames(NewSortFieldName), GetSortDirections(NewSortDirections), Result);
    except
      Result:=false;
    end;
    if Result then
    begin
      FSortDirections:=NewSortDirections;
      SortFieldName:=NewSortFieldName;
    end;
  end;
end;

// ==============================================================================
// Columns Width block
function TrDBGrid.GetTitleWidth(Column: TColumn): integer;
var
  W, LineCount, maxLineW: integer;
  s: string;
begin
  s:=GetTitleText(Column);
  W:=GetLineWidth(s, Column.Title.Font) + 4;
  if (W > UseScale(FOptionsEx2.FAppearance.FColumnMinWidth)) and (TitleLines > 1) then
  begin
    // get number of words and max word width
    maxLineW:=0;
    LineCount:=0;
    while s <> '' do
    begin
      Inc(LineCount);
      maxLineW:=Max(maxLineW, GetLineWidth(GetLine(s, 0, true), Column.Title.Font));
    end;
    // estimate width according total length, number of lines and number of words
    W:=Round(W * (1 + (LineCount - TitleLines) * 0.2) / TitleLines);
    // compare with max word length
    W:=Max(W, maxLineW + 4);
  end;
  if IsTitleImageEx(Column.Field) then
    Inc(W, FTitleImageList.Width + 4);
  if IsFunnelButtonColumn(Column.Index) then
    Inc(W, rGlyphs.FilterWidth(FScaleRatio) + 6);
  if IsSimpleField(Column.Field) and (Assigned(Sorter) or Assigned(FOnChangeSort) or Assigned(FOnChangeSortEx)) then
    Inc(W, rGlyphs.SortWidth(FScaleRatio) + 6);
  Result:=GetLimitColumnWidth(W, true, true);
end;

procedure TrDBGrid.IntSetColumnWidth(Column: TColumn; NewWidth: integer);
begin
  if Column.Width <> NewWidth then
  begin
    if not DoGeneralEventParam(geSetColumnWidth, Column, nil, nil, NewWidth) then
      Exit;
    NewWidth:=FGeneralEventRec.ParamValue;
    FLastColWidthChangedIdx:=DataToRawColumn(Column.Index);
    Column.Width:=NewWidth;
  end;
end;

function TrDBGrid.GetLimitColumnWidth(W: integer; LimitMin, LimitMax: boolean): integer;
begin
  if LimitMin and (W < UseScale(FOptionsEx2.FAppearance.FColumnMinWidth)) then
    Result:=UseScale(FOptionsEx2.FAppearance.FColumnMinWidth)
  else
    if LimitMax and (W > UseScale(FOptionsEx2.FAppearance.FColumnMaxWidth)) then
      Result:=UseScale(FOptionsEx2.FAppearance.FColumnMaxWidth)
    else
      Result:=W;
end;

procedure TrDBGrid.CBProc_OneColumnWidth(Sender: TObject; DataSet: TDataSet; RecNum: integer);
var
  W: integer;
  s: string;
  C: TColumn;
begin
  C:=TColumn(Sender);
  s:='';
  if IsSimpleField(C.Field) then
  begin
    if not GetGroupMasterCellContent(C.Field, s) then
      s:=C.Field.DisplayText;
  end
  else
    if FOptionsEx2.FEditors.ShowMemoContent and (C.Field.DataType in [ftMemo, ftFmtMemo, ftWideMemo]) then
      s:=C.Field.AsString;
  W:=GetLineWidth(s, C.Font) + 8;
  if W > CBScan_Int then
    CBScan_Int:=W;
end;

procedure TrDBGrid.SetOneColumnWidth(ColumnWidth: TColumnWidth; Column: TColumn; AdditionalSpace: integer = 0);
var
  W: integer;
begin
  if not Assigned(Column) then
    Exit;
  // Set width according ColumnWidth
  W:=-1;
  case ColumnWidth of
    cwNone:
      W:=Column.Width;
    cwDefaultWidth:
      W:=Column.DefaultWidth;
    cwDataWidth, cwDataLabelWidth:
      if IsActive then
      begin
        if ColumnWidth = cwDataLabelWidth then
          W:=GetTitleWidth(Column);
        CBScan_Int:=0;
        ScanAllRecords(CBProc_OneColumnWidth, Column, grrColumnWidth, true, FOptionsEx2.FAppearance.FColumnWidthMaxRecordsSearch);
        if (CBScan_Int > 0) then
        begin
          if IsDateEx(Column.Field) then
            CBScan_Int:=CBScan_Int + DefaultRowHeight;
          W:=Max(W, CBScan_Int);
        end;
      end;
    cwLabelWidth:
      W:=GetTitleWidth(Column);
  end;
  if W > 0 then
    IntSetColumnWidth(Column, GetLimitColumnWidth(W, true, true) + AdditionalSpace);
end;

procedure TrDBGrid.CBProc_ColumnWidth(Sender: TObject; DataSet: TDataSet; RecNum: integer);
var
  W, I: integer;
  s: string;
  C: TColumn;
begin
  for I:=0 to Columns.Count - 1 do
  begin
    s:='';
    C:=Columns[I];
    if IsSimpleField(C.Field) then
    begin
      if not GetGroupMasterCellContent(C.Field, s) then
        s:=C.Field.DisplayText;
    end
    else
      if FOptionsEx2.FEditors.ShowMemoContent and IsMemoField(C.Field) then
        s:=C.Field.AsString;
    if s = '' then
      W:=0
    else
      W:=GetLineWidth(s, C.Font) + 8;
    if W > CBScan_IntArr[I] then
      CBScan_IntArr[I]:=W;
  end;
end;

procedure TrDBGrid.SetColumnWidth(ColumnWidth: TColumnWidth; AdditionalSpace: integer = 0);
var
  I, Cols, GridCW, SumWidth, SumFixedWidth, MinColW, MaxColW: integer;
  R: double;
  lim: boolean;
begin
  if (ColumnWidth = cwNone) or (Columns.Count = 0) then
    Exit;
  // Compute number and width of visible columns
  Cols:=0;
  SumWidth:=0;
  SumFixedWidth:=0;
  MinColW:=UseScale(FOptionsEx2.FAppearance.FColumnMinWidth);
  MaxColW:=Width;
  for I:=0 to Columns.Count - 1 do
    if Columns[I].Visible then
    begin
      // limit max width
      if Columns[I].Width > MaxColW then
        Columns[I].Width:=MaxColW;
      if IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
      begin
        Inc(SumFixedWidth, Columns[I].Width);
        if dgColLines in Options then
          Inc(SumFixedWidth);
      end
      else
      begin
        Inc(Cols);
        Inc(SumWidth, Columns[I].Width);
      end;
    end;
  // any non-fixed col ?
  if (Cols = 0) or (SumWidth = 0) then
    Exit;
  // Compute Grid Client Width
  GridCW:=Width - GetSystemMetrics(SM_CXVSCROLL) - 2;
  if dgIndicator in Options then
  begin
    if Assigned(FFixedColText) then
      Dec(GridCW, FFixedColText.FFixedColWidth)
    else
      Dec(GridCW, IndicatorWidth);
  end;
  if BorderStyle = bsSingle then
  begin
    if Ctl3D then
      Dec(GridCW, 4)
    else
      Dec(GridCW, 2);
  end;
  if dgColLines in Options then
    Dec(GridCW, Cols);
  // Set width according ColumnWidth
  case ColumnWidth of
    cwDefaultWidth, cwLabelWidth:
      for I:=0 to Columns.Count - 1 do
        if Columns[I].Visible and not IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
          SetOneColumnWidth(ColumnWidth, Columns[I], AdditionalSpace);
    cwSameWidth:
      for I:=0 to Columns.Count - 1 do
        if Columns[I].Visible and not IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
          IntSetColumnWidth(Columns[I], GetLimitColumnWidth((GridCW - SumFixedWidth) div Cols, true, false));
    cwAutoWidth:
      begin
        // calc new width ratio
        R:=(GridCW - SumFixedWidth) / SumWidth;
        lim:=false;
        if R < 1 then
        begin
          // reduce width - check MinColW
          for I:=0 to Columns.Count - 1 do
            if Columns[I].Visible and not IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
              if Trunc(Columns.Items[I].Width * R) < MinColW then
              begin
                lim:=true;
                Break;
              end;
        end;
        if lim then
        begin
          // calc ratio after usage of MinColW
          if SumWidth > Cols * MinColW then
            R:=(GridCW - Cols * MinColW - SumFixedWidth) / (SumWidth - Cols * MinColW)
          else
            R:=0;
        end
        else
          // no MinColW used, just use ratio
          MinColW:=0;
        // change width
        for I:=0 to Columns.Count - 1 do
          if Columns[I].Visible and not IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
            IntSetColumnWidth(Columns[I], GetLimitColumnWidth(Trunc((Columns.Items[I].Width - MinColW) * R) + MinColW, true, false));
      end;
    cwDataWidth, cwDataLabelWidth, cwHideEmptyCol:
      if IsActive then
      begin
        // init
        SetLength(CBScan_IntArr, Columns.Count);
        for I:=0 to Columns.Count - 1 do
        begin
          CBScan_IntArr[I]:=0;
          if ColumnWidth = cwDataLabelWidth then
            CBScan_IntArr[I]:=GetTitleWidth(Columns[I]);
        end;
        // scan
        ScanAllRecords(CBProc_ColumnWidth, nil, grrColumnWidth, true, FOptionsEx2.FAppearance.FColumnWidthMaxRecordsSearch);
        // set
        for I:=0 to Columns.Count - 1 do
        begin
          if (ColumnWidth = cwHideEmptyCol) then
          begin
            if CBScan_IntArr[I] = 0 then
              Columns[I].Visible:=false;
          end
          else
          begin
            if (Columns.Count > 4) and (CBScan_IntArr[I] > Width div 4) then
              CBScan_IntArr[I]:=Width div 4;
            if (CBScan_IntArr[I] = 0) then
              CBScan_IntArr[I]:=8;
            if IsDateEx(Columns[I].Field) then
              CBScan_IntArr[I]:=CBScan_IntArr[I] + DefaultRowHeight;
            if not IsFixedColumnWidth(Columns[I].Field, Columns[I].FieldName) then
              IntSetColumnWidth(Columns[I], GetLimitColumnWidth(CBScan_IntArr[I], true, true) + AdditionalSpace);
          end;
        end;
        Finalize(CBScan_IntArr);
      end;
  end;
end;

procedure TrDBGrid.RestoreColumnsWidth(RestoreVisibility: boolean = true; RestorePosition: boolean = true);
var
  I, C, cidx: integer;
begin
  for I:=0 to High(FColumnStoredProps) do
  begin
    // find current column possition according FieldName
    cidx:=-1;
    for C:=0 to Columns.Count - 1 do
      if AnsiSameText(FColumnStoredProps[I].FieldName, Columns[C].FieldName) then
      begin
        cidx:=C;
        Break;
      end;
    // restore and move to original pos.
    if cidx >= 0 then
    begin
      Columns[cidx].Width:=FColumnStoredProps[I].Width;
      if RestoreVisibility then
        Columns[cidx].Visible:=FColumnStoredProps[I].Visible;
      if RestorePosition then
        Columns[cidx].Index:=I;
    end;
  end;
end;

// ==============================================================================
// Aux event block
function TrDBGrid.HasParentNormalState: boolean;
begin
  Result:=Assigned(Parent) and (([csLoading, csDestroying] * ComponentState) = []);
end;

function TrDBGrid.IsColumnReadOnly(C: TColumn): boolean;
begin
  Result:=ReadOnly or not IsActive or DataLink.ReadOnly or C.ReadOnly or not Assigned(C.Field) or C.Field.ReadOnly;
  DoGeneralEventFieldBool(geIsReadOnlyColumn, C.Field, Result);
end;

function TrDBGrid.IsLookupFilterColumn(C: TColumn): boolean;
begin
  Result:=FOptionsEx2.FEditors.FAllowLookupFilter and Assigned(C) and Assigned(C.Field) and Assigned(C.Field.LookupDataSet) and C.Field.LookupDataSet.Active;
  DoGeneralEventFieldBool(geIsLookupFilterColumn, C.Field, Result);
end;

function TrDBGrid.CanEdit: boolean;
var
  C: TColumn;
begin
  // try to start EDITING for any field and return true if editing started
  Result:=false;
  C:=Columns[SelectedIndex];
  if (dgEditing in Options) and not IsColumnReadOnly(C) then
  begin
    Result:=true;
    if Assigned(FOnCanEdit) then
      FOnCanEdit(self, C.Field, C, Result);
    if Result then
    begin
      DataLink.Edit;
      Result:=DataLink.Editing;
      // if Result then DataLink.Modified;
    end;
  end;
end;

function TrDBGrid.CanEditModify: boolean;
var
  C: TColumn;
begin
  Result:=false;
  C:=Columns[SelectedIndex];
  if IsColumnReadOnly(C) then
    Exit;
  // allow edit for MemoAsText
  if C.Field.CanModify and (not(C.Field.DataType in ftNonTextTypes + [ftWideMemo]) or Assigned(C.Field.OnSetText) or IsMemoAsText(C.Field)) then
  begin
    DataLink.Edit;
    Result:=DataLink.Editing;
    if Result then
      DataLink.Modified;
  end;
end;

function TrDBGrid.CanEditShow: boolean;
begin
  Result:= inherited CanEditShow;
  if (LayoutLock <> 0) then
    Exit;
  try
    if IsActive and Assigned(SelectedField) then
    begin
      if Assigned(FOnCanEdit) then
        FOnCanEdit(self, SelectedField, Columns[SelectedIndex], Result);
      if FOptionsEx2.FEditors.FAllowReadOnlyButton and (Columns[SelectedIndex].ButtonStyle = cbsEllipsis) then
        Exit;
      if ReadOnly or SelectedField.ReadOnly or Columns[SelectedIndex].ReadOnly or (not DataLink.DataSet.CanModify) or IsBoolEx(SelectedField) or IsMemoAsGlyph
        (SelectedField) or IsRichTextEx(SelectedField) or IsGraphicEx(SelectedField) then
        EditorMode:=false;
    end
    else
      Result:=false;
  except
  end;
end;

procedure TrDBGrid.SetEditText(ACol, ARow: Longint; const Value: string);
var
  s: string;
  maxl, v: integer;
begin
  s:=Value;
  maxl:=GetEditLimit;
  if maxl > 0 then
    s:=copy(s, 1, maxl);

  if Assigned(SelectedField) and IsDateEx(SelectedField) then
  begin
    // replace separators
    s:=UpdateToDateSeparator(s);
    if Length(s) < 3 then
    begin
      if TryStrToInt(s, v) then
        try
          // use entered number as the day and use current month and year
          s:=DateToStr(DateUtils.RecodeDay(Date, v));
        except
          // skip exception
        end;
    end;
  end;
  if Assigned(SelectedField) and IsTimeEx(SelectedField) then
  begin
    // replace separators
    s:=UpdateToTimeSeparator(s);
    if Length(s) < 3 then
    begin
      if TryStrToInt(s, v) then
        try
          // use entered number as the hour and use 00 minute
          s:=IntToStr(v) + GetTimeSeparator + '00';
        except
          // skip exception
        end;
    end;
  end;

  inherited SetEditText(ACol, ARow, s);
end;

function TrDBGrid.GetEditText(ACol, ARow: Longint): string;
begin
  Result:= inherited GetEditText(ACol, ARow);
  if IsActive then
    with Columns[RawToDataColumn(ACol)] do
      if Assigned(Field) and IsMemoAsText(Field) then
      begin
        // result content of memo field instead of text "Memo"
        Result:=Field.AsString;
      end;
end;

function TrDBGrid.GetClientRect: TRect;
begin
  Result:= inherited GetClientRect;
  if Assigned(FFooterRow) then
    Result.Bottom:=Result.Bottom - FFooterRow.GetFooterRowHeight;
end;

procedure TrDBGrid.CheckSelectedCol;
begin
  // check if selected col is inside fixed cols, then move to first non fixed col
  if IsActive and (Col < FixedCols) then
    Col:=FixedCols;
end;

procedure TrDBGrid.LayoutChanged;
var
  I, H, PixelsPerRow, PixelsTitle: integer;
  RestoreCanvas: boolean;
begin
  inherited LayoutChanged;
  if LayoutLock = 0 then
    Exit;
  if not HasParentNormalState then
    Exit;

  CheckSelectedCol;

  // calculate and set rows height
  RestoreCanvas:=not HandleAllocated;
  if RestoreCanvas then
    Canvas.Handle:=GetDC(0);
  try
    // find data row height
    Canvas.Font.Assign(Font);
    if FRowFixHeight > 0 then
      PixelsPerRow:=FRowFixHeight
    else
    begin
      PixelsPerRow:=Canvas.TextHeight('Wg') * FRowLines + 3;
      if dgRowLines in Options then
        Inc(PixelsPerRow, GridLineWidth)
      else
        Inc(PixelsPerRow);
    end;

    // find title row height
    if (dgTitles in Options) then
    begin
      Canvas.Font.Assign(TitleFont);
      PixelsTitle:=Canvas.TextHeight('Wg');
      for I:=0 to Columns.Count - 1 do
      begin
        Canvas.Font.Assign(Columns[I].Title.Font);
        H:=Canvas.TextHeight('Wg');
        if H > PixelsTitle then
          PixelsTitle:=H;
      end;
      // find height of titletext
      Canvas.Font.Assign(FTitleLineText.Font);
      FTitleLineText.LineHeight:=Canvas.TextHeight('Wg');
      if FTitleLineText.FFixHeight > 0 then
        FTitleLineText.SumHeight:=FTitleLineText.FFixHeight
      else
      begin
        FTitleLineText.SumHeight:=0;
        if FTitleLineText.Text <> '' then
        begin
          FTitleLineText.SumHeight:=FTitleLineText.LineHeight * FTitleLineText.GetLineCount;
          Inc(FTitleLineText.SumHeight, 1);
        end;
      end;

      // finalize full title row height
      PixelsTitle:=PixelsTitle * TitleLines + FTitleLineText.SumHeight + 4;

      // add space for Grouping
      Inc(PixelsTitle, FOptionsEx2.FGroups.GroupHeight);
      // add space for FilterBar
      Inc(PixelsTitle, FOptionsEx2.FFilters.TextBarTopHeight + FOptionsEx2.FFilters.TextBarBootomHeight);
    end
    else
      PixelsTitle:=PixelsPerRow;

    // set number of rows
    DefaultRowHeight:=PixelsPerRow;
    RowCount:=1 + (Height - PixelsTitle) div (PixelsPerRow);
    // set row height
    RowHeights[0]:=PixelsTitle;
    for I:=1 to RowCount - 1 do
      RowHeights[I]:=PixelsPerRow;

  finally
    if RestoreCanvas then
    begin
      ReleaseDC(0, Canvas.Handle);
      Canvas.Handle:=0;
    end;
  end;
  // set TopRow if already scrolled - MUST in other case top row will be hidden!!
  TopRow:=FixedRows;

  // send a WM_SIZE message to let the base component recompute the visible rows in the private UpdateRowCount method
  // WM_SIZE replaced by DataChanged !
  // PostMessage(Handle, WM_SIZE, 0, MakeLong(Width, Height));

  // call private DataChanged procedure of grid via DataLink to update rowcount, scrollbars, activerec and editor
  if Assigned(DataLink) then
    THackGridDataLink(DataLink).DataSetChanged;

  FTitleLineText.PosRight:=0;
  // update EditStyle for DateTime
  FOptionsEx2.FEditors.DateTimePicker:=FOptionsEx2.FEditors.FDateTimePicker;
  // update footer grid
  UpdateFooterGrid;
end;

procedure TrDBGrid.RefreshLayout;
begin
  LayoutChanged;
end;

procedure TrDBGrid.SetColumnAttributes;
begin
  inherited;
  if (dgIndicator in Options) and Assigned(FFixedColText) and (FFixedColText.FFixedColWidth <> IndicatorWidth) then
    ColWidths[0]:=FFixedColText.FFixedColWidth;
  SetFixedCols(FOptionsEx2.FFixedCell.FFixedCols);
end;

procedure TrDBGrid.DoEnter;
begin
  inherited;
  FFilterDlgLastColIdx:=-1;
end;

procedure TrDBGrid.DoExit;
begin
  inherited;
  FFilterDlgLastColIdx:=-1;
  StopSearching;
end;

procedure TrDBGrid.ColEnter;
var
  I: integer;
begin
  inherited;
  I:=RawToDataColumn(Col);
  if (I >= 0) and (I < Columns.Count) and (Columns[I].DropDownRows < FDropDownRowsCount) then
    Columns[I].DropDownRows:=FDropDownRowsCount;
end;

procedure TrDBGrid.ColExit;
begin
  inherited;
  // reset temporary lookup value
  FOnTimer(self);
end;

procedure TrDBGrid.CalcSizingState(X, Y: integer; var State: TGridState; var Index, SizingPos, SizingOfs: integer; var FixedInfo: TGridDrawInfo);
begin
  inherited CalcSizingState(X, Y, State, Index, SizingPos, SizingOfs, FixedInfo);
  FLastColWidthChangedIdx:=Index;
end;

procedure TrDBGrid.ColWidthsChanged;
var
  I: integer;
begin
  inherited;
  if Assigned(FOnColWidthChanged) and (FLastColWidthChangedIdx >= 0) then
  begin
    I:=RawToDataColumn(FLastColWidthChangedIdx);
    if (I < Columns.Count) and (Columns[I].Visible) then
      FOnColWidthChanged(self, Columns[I], FGridState = gsColSizing);
    FLastColWidthChangedIdx:=-1;
  end;
end;

procedure TrDBGrid.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  inherited ColumnMoved(FromIndex, ToIndex);
  isColMoved:=true;
  MoveGroupedCol(FromIndex, ToIndex);
end;

procedure TrDBGrid.Scroll(Distance: integer);
var
  OldRect: TRect;
begin
  if not HandleAllocated then
    Exit;
  FPrevActiveBookmark:=FActiveBookmark;
  if IsActive then
    FActiveBookmark:=DataLink.DataSet.Bookmark;
  // fix bug if columns width=grid width, BoxRect returns only width of first column
  OldRect:=BoxRect(0, Row, ColCount - 1, Row);
  BeginUpdate;
  try
    inherited;
  finally
    EndUpdate;
  end;
  if OldRect.Right = ColWidths[0] then
  begin
    Invalidate;
    DrawFooterRow;
  end
  else
    if FFooterRow.FRefreshAfterScroll then
      DrawFooterRow;
  // reset timer
  if not(csDesigning in ComponentState) then
    FOnTimer(self);
  if Assigned(FOnDataScrolled) then
    FOnDataScrolled(self);
end;

procedure TrDBGrid.UpdateScrollBar;
begin
  // ShowScrollBar(Handle, SB_VERT, false);
  // exit;
  inherited;
  // allways SHOW VERT scrollbar - BUG in standard Grid if HORZ scrollbar is not shown
  if HasParentNormalState then
  begin
    ShowScrollBar(Handle, SB_VERT, true);
  end;
end;

procedure TrDBGrid.InitSummaryResults(var R: TSummaryResults);
begin
  R.Field:=nil;
  R.Sum:=0;
  R.Min:=0;
  R.Max:=0;
  R.Avg:=0;
  R.CountValues:=0;
  R.CountEmpty:=0;
  R.SelSum:=0;
  R.SelMin:=0;
  R.SelMax:=0;
  R.SelAvg:=0;
  R.SelCountValues:=0;
  R.SelCountEmpty:=0;
  R.DistinctValueCount:=0;
  SetLength(R.DistinctValues, 0);
end;

procedure TrDBGrid.CBProc_RecalculateSummaryResults(Sender: TObject; DataSet: TDataSet; RecNum: integer);
var
  I, R: integer;
  F: TField;
  sel: boolean;
begin
  sel:=RowIsMultiselected(true, 0);
  for I:=0 to Columns.Count - 1 do
  begin
    F:=FSummaryResultsArr[I].Field;
    if not Assigned(F) then
      Exit;
    if F.IsNull then
    begin
      Inc(FSummaryResultsArr[I].CountEmpty);
      if sel then
        Inc(FSummaryResultsArr[I].SelCountEmpty);
    end
    else
    begin
      if (F is TNumericField) or IsDateTimeField(F) then
      begin
        if FSummaryResultsArr[I].CountValues = 0 then
        begin
          FSummaryResultsArr[I].Min:=F.AsFloat;
          FSummaryResultsArr[I].Max:=F.AsFloat;
        end;
        if FSummaryResultsArr[I].Min > F.AsFloat then
          FSummaryResultsArr[I].Min:=F.AsFloat;
        if FSummaryResultsArr[I].Max < F.AsFloat then
          FSummaryResultsArr[I].Max:=F.AsFloat;
        FSummaryResultsArr[I].Sum:=FSummaryResultsArr[I].Sum + F.AsFloat;
        if sel then
        begin
          if FSummaryResultsArr[I].SelCountValues = 0 then
          begin
            FSummaryResultsArr[I].SelMin:=F.AsFloat;
            FSummaryResultsArr[I].SelMax:=F.AsFloat;
          end;
          if FSummaryResultsArr[I].SelMin > F.AsFloat then
            FSummaryResultsArr[I].SelMin:=F.AsFloat;
          if FSummaryResultsArr[I].SelMax < F.AsFloat then
            FSummaryResultsArr[I].SelMax:=F.AsFloat;
          FSummaryResultsArr[I].SelSum:=FSummaryResultsArr[I].SelSum + F.AsFloat;
        end;
      end
      else
        if F is TBooleanField then
        begin
          if F.AsBoolean then
          begin
            FSummaryResultsArr[I].Sum:=FSummaryResultsArr[I].Sum + 1;
            if sel then
              FSummaryResultsArr[I].SelSum:=FSummaryResultsArr[I].SelSum + 1;
          end;
        end;
      Inc(FSummaryResultsArr[I].CountValues);
      if sel then
        Inc(FSummaryResultsArr[I].SelCountValues);
      // add distinct values
      if FCalculateDistinctValues then
      begin
        R:=FSummaryResultsArr[I].HList.Add(F.DisplayText);
        FSummaryResultsArr[I].HList.Objects[R]:=TObject(integer(FSummaryResultsArr[I].HList.Objects[R]) + 1);
      end;
    end;
  end;
end;

function ListSortObjectCompare(List: TStringList; Index1, Index2: integer): integer;
begin
  if integer(List.Objects[Index1]) < integer(List.Objects[Index2]) then
    Result:=1
  else
    if integer(List.Objects[Index1]) > integer(List.Objects[Index2]) then
      Result:=-1
    else
      Result:=CompareText(List[Index1], List[Index2]);
end;

procedure TrDBGrid.RecalculateSummaryResults(UseCalculatedSummaryForPopup: boolean = false);
  function DoRound(v: double): double;
  var
    E: integer;
  begin
    Result:=v;
    if v = 0 then
      Exit;
    E:=0;
    while (Abs(v) < Power(10, E + 5)) and (E > -15) do
      Dec(E);
    Result:=RoundTo(v, E);
  end;

var
  I, DC, di: integer;
begin
  FUseCalculatedSummaryForPopup:=UseCalculatedSummaryForPopup;
  FCalculateDistinctValues:=FOptionsEx2.FFixedCell.FTitlePopupDistinctValues and ContainsText
    (FTitlePopupItems.Text, cSummaryTag_Prefix + cSummaryTag_DistinctValues);
  // init sum array
  SetLength(FSummaryResultsArr, Columns.Count);
  for I:=0 to Columns.Count - 1 do
  begin
    InitSummaryResults(FSummaryResultsArr[I]);
    FSummaryResultsArr[I].Field:=Columns[I].Field;
    FSummaryResultsArr[I].HList:=THashedStringList.Create;
    FSummaryResultsArr[I].HList.Sorted:=true;
    FSummaryResultsArr[I].HList.Duplicates:=dupIgnore;
  end;
  try
    if Assigned(FOnRecalculateSummaryResult) then
    begin
      // call event to get summary result for each field
      for I:=0 to Columns.Count - 1 do
        FOnRecalculateSummaryResult(self, Columns[I].Field, FSummaryResultsArr[I])
    end
    else
    begin
      // calculate summary for all fields
      ScanAllRecords(CBProc_RecalculateSummaryResults, nil, grrSummaryCalc);
      for I:=0 to Columns.Count - 1 do
      begin
        // calc avg
        if FSummaryResultsArr[I].CountValues > 0 then
          FSummaryResultsArr[I].Avg:=DoRound(FSummaryResultsArr[I].Sum / FSummaryResultsArr[I].CountValues);
        if FSummaryResultsArr[I].SelCountValues > 0 then
          FSummaryResultsArr[I].SelAvg:=DoRound(FSummaryResultsArr[I].SelSum / FSummaryResultsArr[I].SelCountValues);
        // add distinct values sorted
        FSummaryResultsArr[I].DistinctValueCount:=FSummaryResultsArr[I].HList.Count;
        FSummaryResultsArr[I].HList.Sorted:=false;
        FSummaryResultsArr[I].HList.CustomSort(ListSortObjectCompare);
        DC:=Min(FSummaryResultsArr[I].HList.Count, cSummary_DistinctValuesMax);
        SetLength(FSummaryResultsArr[I].DistinctValues, DC);
        for di:=0 to DC - 1 do
        begin
          FSummaryResultsArr[I].DistinctValues[di].Value:=FSummaryResultsArr[I].HList[di];
          FSummaryResultsArr[I].DistinctValues[di].Count:=integer(FSummaryResultsArr[I].HList.Objects[di]);
        end;
      end;
    end;
  finally
    for I:=0 to Columns.Count - 1 do
      FreeAndNil(FSummaryResultsArr[I].HList);
  end;
  // refresh footer
  DrawFooterRow;
end;

function TrDBGrid.GetSummaryResults(F: TField; Recalculate: boolean): TSummaryResults;
var
  I: integer;
begin
  InitSummaryResults(Result);
  if Recalculate then
    RecalculateSummaryResults;
  for I:=0 to High(FSummaryResultsArr) do
  begin
    if FSummaryResultsArr[I].Field = F then
    begin
      Result:=FSummaryResultsArr[I];
      Break;
    end;
  end;
end;

function TrDBGrid.ReplaceSummaryTags(var s: string; F: TField; R: TSummaryResults; ForFooter: boolean): string;
  procedure ReplaceTags(tag: string; Val: double; AsInt: boolean = false; AsSum: boolean = false);
  var
    v: string;
  begin
    tag:=cSummaryTag_Prefix + tag;
    if AnsiPosEx(tag, s) > 0 then
    begin
      if AsInt then
        v:=IntToStr(Round(Val))
      else
      begin
        if (R.Field is TNumericField) then
          v:=FormatFloat((R.Field as TNumericField).DisplayFormat, Val)
        else
          if (R.Field is TDateTimeField) and not AsSum then
            v:=FormatDateTime(TDateTimeField(R.Field).DisplayFormat, Val)
          else
            if (R.Field is TSQLTimeStampField) and not AsSum then
              v:=FormatDateTime(TSQLTimeStampField(R.Field).DisplayFormat, Val)
            else
              if (R.Field is TBooleanField) then
              begin
                if AsSum or (Val = 0) then
                  v:=IntToStr(Round(Val))
                else
                  v:=FormatFloat('0.000', Val);
              end
              else
                v:='';
      end;
      if Assigned(FOnGetSummaryResult) then
        FOnGetSummaryResult(self, F, R, tag, ForFooter, v);
      s:=ReplaceStr(tag, v, s, true, false);
      if Result <> '' then
        Result:=Result + '; ';
      Result:=Result + v;
    end;
  end;

var
  recno: integer;
begin
  Result:='';
  if IsActive then
    recno:=DataLink.DataSet.recno
  else
    recno:=0;
  ReplaceTags(cSummaryTag_RecNo, recno, true);
  ReplaceTags(cSummaryTag_Sum, R.Sum, false, true);
  ReplaceTags(cSummaryTag_Min, R.Min);
  ReplaceTags(cSummaryTag_Max, R.Max);
  ReplaceTags(cSummaryTag_AVG, R.Avg);
  ReplaceTags(cSummaryTag_CountVal, R.CountValues, true);
  ReplaceTags(cSummaryTag_CountNull, R.CountEmpty, true);
  ReplaceTags(cSummaryTag_CountAll, R.CountValues + R.CountEmpty, true);

  ReplaceTags(cSummaryTag_SelSum, R.SelSum, false, true);
  ReplaceTags(cSummaryTag_SelMin, R.SelMin);
  ReplaceTags(cSummaryTag_SelMax, R.SelMax);
  ReplaceTags(cSummaryTag_SelAVG, R.SelAvg);
  ReplaceTags(cSummaryTag_SelCountVal, R.SelCountValues, true);
  ReplaceTags(cSummaryTag_SelCountNull, R.SelCountEmpty, true);
  ReplaceTags(cSummaryTag_SelCountAll, R.SelCountValues + R.SelCountEmpty, true);

  ReplaceTags(cSummaryTag_DistinctValues, R.DistinctValueCount, true);
end;

function TrDBGrid.SummaryResults(Fld: TField; Recalculate: boolean = true): TSummaryResults;
begin
  Result:=GetSummaryResults(Fld, Recalculate);
end;

function TrDBGrid.SummaryResults(Column: TColumn; Recalculate: boolean = true): TSummaryResults;
begin
  Result:=SummaryResults(Column.Field, Recalculate);
end;

procedure TrDBGrid.DoContextPopup(MousePos: TPoint; var Handled: boolean);
var
  I, j: integer;
  XY, s: string;
  C: TColumn;
  dv: boolean;
  NewItem, NewItemDV: TMenuItem;
  SumRes: TSummaryResults;
  PopupCoord: TGridCoord;
begin
  try
    XY:=Format('%d;%d', [MousePos.X, MousePos.Y]);
    if not DoGeneralEventParam(geContextPopup, nil, nil, nil, XY) then
      Exit;
    PopupCoord:=MouseCoordEx(MousePos.X, MousePos.Y);
    if PopupCoord.Y = TitleRow then
    begin
      I:=RawToDataColumn(PopupCoord.X);
      if I < 0 then
      begin
        if Assigned(FTitleExPopupMenu) then
          FTitleExPopupMenu.Popup(ClientToScreen(MousePos).X, ClientToScreen(MousePos).Y);
      end
      else
      begin
        C:=Columns[I];
        // clear popupmenu
        for I:=FTitlePopup.Items.Count - 1 downto 0 do
          FTitlePopup.Items[I].Free;
        FTitlePopup.Items.Clear;
        SetDefaultTitlePopup;
        if FOptionsEx2.FFixedCell.FTitlePopup and (FTitlePopupItems.Count > 0) and (Trim(FTitlePopupItems.Text) <> '') then
        begin
          SumRes:=GetSummaryResults(C.Field, not FUseCalculatedSummaryForPopup);
          // fill PopupMenu - add title
          NewItem:=TMenuItem.Create(FTitlePopup);
          NewItem.Caption:=C.Title.Caption + ':';
          NewItem.Hint:=C.FieldName;
          NewItem.tag:=9;
          NewItem.OnClick:=TitlePopupClick;
          NewItem.Enabled:=true; // false;
          FTitlePopup.Items.Add(NewItem);
          NewItem:=TMenuItem.Create(FTitlePopup);
          NewItem.Caption:='-';
          FTitlePopup.Items.Add(NewItem);
          // add items
          for I:=0 to FTitlePopupItems.Count - 1 do
          begin
            s:=Trim(FTitlePopupItems[I]);
            // item for distinct values?
            dv:=ContainsText(s, cSummaryTag_Prefix + cSummaryTag_DistinctValues);
            if dv and not FOptionsEx2.FFixedCell.FTitlePopupDistinctValues then
              Continue;
            NewItem:=TMenuItem.Create(FTitlePopup);
            if dv then
            begin
              // s:=ReplaceStr(cSummaryTag_Prefix + cSummaryTag_DistinctValues, IntToStr(Length(SumRes.DistinctValues)), s, true, false);
              // s:=ClearStr(cSummaryTag_Prefix + cSummaryTag_DistinctValues, s);
              NewItem.tag:=1;
            end
            else
            begin
              NewItem.tag:=2;
            end;
            NewItem.Hint:=ReplaceSummaryTags(s, C.Field, SumRes, false);
            NewItem.Caption:=s;
            NewItem.OnClick:=TitlePopupClick;
            FTitlePopup.Items.Add(NewItem);
            if dv then
              // add distinct values
              for j:=0 to High(SumRes.DistinctValues) do
              begin
                NewItemDV:=TMenuItem.Create(FTitlePopup);
                s:=Format('%s (%d)', [SumRes.DistinctValues[j].Value, SumRes.DistinctValues[j].Count]);
                NewItemDV.Caption:=s;
                NewItemDV.tag:=2;
                NewItemDV.OnClick:=TitlePopupClick;
                NewItem.Add(NewItemDV);
              end;
          end;

          if not DoGeneralEventParam(geContextTitlePopup, C, nil, FTitlePopup, XY) then
            Exit;
          FTitlePopup.Popup(ClientToScreen(MousePos).X, ClientToScreen(MousePos).Y);
        end;
      end;
      Handled:=true;
    end;
  finally
    inherited DoContextPopup(MousePos, Handled);
  end;
end;

procedure TrDBGrid.LinkActive(Value: boolean);
begin
  HideEditor;
  inherited LinkActive(Value);
  if Value then
    LoadAutoPickList;
  if Value then
    LoadColorPickList;
  if Value then
    SetColumnWidth(FColumnWidth);
  if Value and FFooterRow.FFooterVisible and FFooterRow.FRecalculateAfterOpen then
    RecalculateSummaryResults(true);
end;

procedure TrDBGrid.StopUpdate(stop: boolean);
begin
  if stop then
    BeginUpdate
  else
    EndUpdate;
end;

procedure TrDBGrid.EditButtonClick;
var
  F: TField;
  D: TDateTime;
begin
  F:=Columns[SelectedIndex].Field;
  // save edited value
  if IsActive and DataLink.Editing then
    THackGridDataLink(DataLink).UpdateData;

  if IsDateEx(F) then
  begin
    if not CanEdit then
      Exit;
    try
      D:=GetDateValueEx(F);
      if D = 0 then
        D:=Date;
    except
      D:=Date;
    end;
{$IFDEF TrPopupMonthCalendar}
    GetInplaceEditor.FPopupCalendar.ShowCalendar(D);
{$ELSE}
    SetDateTimePickerDate(D);
    ShowDateTimePicker(CellRect(Col, Row));
{$ENDIF}
  end
  else
    inherited;
end;

// ==============================================================================
// Save/Restore Event and Pos block
procedure TrDBGrid.SaveRowPos(SetExUpdateLock: boolean = false);
begin
  if IsActive and HandleAllocated then
  begin
    oldRowPos:=DataLink.ActiveRecord;
    oldRecNo:=DataLink.DataSet.recno;
  end;
  if SetExUpdateLock then
    ExUpdateLock:=true;
end;

procedure TrDBGrid.RestoreRowPos;
var
  ActRec, dif, MoveCount: integer;
begin
  ExUpdateLock:=false;
  if IsActive and HandleAllocated then
    try
      if DataLink.DataSet.recno <> oldRecNo then
        Exit;
      BeginUpdate;
      try
        ActRec:=DataLink.ActiveRecord;
        dif:=ActRec - oldRowPos;
        if dif > 0 then
          MoveCount:=dif + VisibleRowCount - ActRec - 1
        else
          if dif < 0 then
            MoveCount:=dif - ActRec
          else
            Exit;
        DataLink.DataSet.DisableControls;
        try
          DataLink.DataSet.MoveBy(MoveCount);
          DataLink.DataSet.MoveBy(-MoveCount);
        finally
          DataLink.DataSet.EnableControls;
        end;
      finally
        EndUpdate;
      end;
    except
    end;
end;

function TrDBGrid.ScrollData(ScrollBy: integer): integer;
var
  di: integer;
begin
  Result:=0;
  if ScrollBy = 0 then
    Exit;
  if ScrollBy < 0 then
    di:=DataLink.ActiveRecord
  else
    di:=-(DataLink.BufferCount - DataLink.ActiveRecord - 1);
  BeginUpdate;
  try
    DataLink.DataSet.DisableControls;
    try
      Result:=DataLink.DataSet.MoveBy(ScrollBy - di);
      if Result <> 0 then
        DataLink.DataSet.MoveBy(-Result);
    finally
      DataLink.DataSet.EnableControls;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TrDBGrid.SaveResetDatasetEvents(Reason: integer = 0);
begin
  oldAfterScroll:=nil;
  oldBeforeScroll:=nil;
  oldDataChange:=nil;
  if not IsDatasetLinked then
    Exit;
  if Assigned(FBeforeScanDataset) then
    FBeforeScanDataset(self, DataLink.DataSet, Reason);
  oldDataChange:=DataLink.DataSource.OnDataChange;
  DataLink.DataSource.OnDataChange:=nil;
  if not Assigned(DataLink.DataSet) then
    Exit;
  oldBeforeScroll:=DataLink.DataSet.BeforeScroll;
  oldAfterScroll:=DataLink.DataSet.AfterScroll;
  DataLink.DataSet.BeforeScroll:=nil;
  DataLink.DataSet.AfterScroll:=nil;
end;

procedure TrDBGrid.RestoreDatasetEvents(Reason: integer = 0);
begin
  if not IsDatasetLinked then
    Exit;
  DataLink.DataSource.OnDataChange:=oldDataChange;
  if not Assigned(DataLink.DataSource.DataSet) then
    Exit;
  DataLink.DataSet.BeforeScroll:=oldBeforeScroll;
  DataLink.DataSet.AfterScroll:=oldAfterScroll;
  if Assigned(FAfterScanDataset) then
    FAfterScanDataset(self, DataLink.DataSet, Reason);
end;

// ==============================================================================
// DRAW block
function TrDBGrid.GetActiveCellRect: TRect;
begin
  Result:=CellRect(Col, Row);
end;

function TrDBGrid.RowIsMultiselected(FromActiveRecord: boolean; FromDataRow: integer): boolean;
var
  OldActive, Index: integer;
begin
  if FromActiveRecord then
    Result:=IsActive and SelectedRows.Find(DataLink.DataSet.Bookmark, Index)
  else
  begin
    Result:=false;
    if (FromDataRow < 0) or not IsActive or not(dgMultiSelect in Options) then
      Exit;
    OldActive:=DataLink.ActiveRecord;
    try
      DataLink.ActiveRecord:=FromDataRow;
      Result:=SelectedRows.Find(DataLink.DataSet.Bookmark, Index);
    finally
      DataLink.ActiveRecord:=OldActive;
    end;
  end;
end;

function TrDBGrid.GetLineWidth(const s: string; AFont: TFont = nil): integer;
var
  oldFont: TFontRecall;
begin
  oldFont:=TFontRecall.Create(Canvas.Font);
  try
    if Assigned(AFont) then
      Canvas.Font.Assign(AFont);
    Result:=Canvas.TextWidth(s);
    if UseRightToLeftAlignment then
      Inc(Result, 2);
  finally
    FreeAndNil(oldFont);
  end;
end;

function TrDBGrid.GetLine(var s: string; W: integer; IgnoreWidth: boolean = false): string;
var
  I, DivPos, Skip: integer;
  NotCapPrev, Cap, Force: boolean;
begin
  Result:=s;
  if (s = '') or ((W < 3) and not IgnoreWidth) then
    Exit;
  DivPos:=-1;
  NotCapPrev:=false;
  Skip:=0;
  Force:=false;
  I:=1;
  if FOptionsEx2.FFixedCell.FTitleForceWrapSeparators <> '' then
  begin
    while I <= Length(s) do
    begin
      if Pos(s[I], FOptionsEx2.FFixedCell.FTitleForceWrapSeparators) > 0 then
      begin
        DivPos:=I;
        Force:=true;
        Skip:=1;
        Break;
      end;
      Inc(I);
    end;
  end;
  if not Force then
  begin
    I:=1;
    while I <= Length(s) do
    begin
      if Pos(s[I], FOptionsEx2.FFixedCell.FTitleTextSeparators) > 0 then
        DivPos:=I;
      if FOptionsEx2.FFixedCell.FTitleTextCaseSeparator then
      begin
        Cap:=CharInSet(s[I], ['A' .. 'Z']);
        if NotCapPrev and Cap then
          DivPos:=I - 1;
        NotCapPrev:=CharInSet(s[I], ['a' .. 'z']);
      end;
      if IgnoreWidth then
      begin
        if DivPos > 0 then
          Break;
      end
      else
        if GetLineWidth(copy(s, 1, I)) > W then
          Break;
      Inc(I);
    end;
  end;
  if (DivPos = -1) then
    DivPos:=I - 1;
  if (I - 1 = Length(s)) then
    DivPos:=I;
  Result:=copy(s, 1, DivPos - Skip);
  s:=copy(s, DivPos + 1, Length(s));
end;

procedure TrDBGrid.FixCellRectHiddenColumns(var Rect: TRect; ACol: integer);
var
  I, W: integer;
begin
  // fixing BUG
  // if dgColLines is not used then not visible columns with Width=-1 descrease left possition of next visible column
  // possible solution with minimum influences is to decrese width of previous column
  if (ACol > FixCol) and not(dgColLines in Options) then
  begin
    W:=0;
    for I:=RawToDataColumn(ACol) + 1 to Columns.Count - 1 do
      if Columns[I].Visible then
        Break
      else
        Inc(W);
    Dec(Rect.Right, W);
  end;
end;

procedure TrDBGrid.FixCellRectColRowLines(var Rect: TRect);
begin
  if not(dgColLines in Options) then
    Dec(Rect.Right);
  if not(dgRowLines in Options) then
    Dec(Rect.Bottom);
end;

procedure TrDBGrid.Paint;
var
  R, RBk: TRect;
begin
  inherited Paint;
  // draw backgroung for edited cell
  // BUG - in standard Grid is not drawn and causes black background and flickering
  if CanEditShow then
  begin
    R:=CellRect(Col, Row);
    RBk:=R;
    if dgColLines in Options then
      RBk.Right:=RBk.Right - 1;
    if dgRowLines in Options then
      RBk.Bottom:=RBk.Bottom - 1;
    DrawCellBackground(RBk, Color, [], Col, Row);
    // FixCellRectColRowLines(R);
    DrawDataGridLines(R, RawToDataColumn(Col), nil, []);
  end;
  DrawFooterRow;
end;

function TrDBGrid.SelectCell(ACol, ARow: Longint): boolean;
begin
  Result:= inherited;
  // block selecting of fixed col
  if Result and (ACol < FixedCols) then
    Result:=false;
end;

function TrDBGrid.HighlightCell(DataCol, DataRow: integer; const Value: string; aState: TGridDrawState): boolean;
begin
  Result:= inherited HighlightCell(DataCol, DataRow, Value, aState);
  if not FOptionsEx2.FAppearance.FHighlightMultiselect then
    Result:=(gdSelected in aState) and ((dgAlwaysShowSelection in Options) or Focused)
    { updatelock eliminates flicker when tabbing between rows }
    and ((UpdateLock = 0) or (dgRowSelect in Options))
  else
    Result:=Result or RowIsMultiselected(true, DataRow);
end;

procedure TrDBGrid.DrawRectBidi(Rect: TRect);
begin
  if UseRightToLeftAlignment then
  begin
    Dec(Rect.Left);
    Dec(Rect.Right);
  end;
  Canvas.Rectangle(Rect);
end;

function TrDBGrid.DrawTextAligned(Rect: TRect; Text: string; Format: integer; VertAlignment: TVerticalAlignment = taAlignTop; TruncatedText: string = '';
  FixedWordWrap: boolean = false): boolean;
var
  rc: TRect;
  L: TStringList;
  I, t, H: integer;
begin
  // calculate text rect and align
  BiDi_ModifyRect(Canvas, Rect);
  rc:=Rect;
  if UseRightToLeftAlignment then
  begin
    Format:=DrawTextBiDiModeFlags(Format); // or DT_RTLREADING;
    // Format:= Format and not DT_RIGHT;
  end;
  if FixedWordWrap and ((DT_WORDBREAK and Format) = DT_WORDBREAK) and (Pos(cStrEOL, Text) > 0) then
  begin
    // fixed EOL with WordBreak - draw line by line
    Result:=true;
    L:=TStringList.Create;
    try
      // get lines
      while Text <> '' do
        L.Add(rtool.RemoveStrPart(Text, cStrEOL));
      H:=Canvas.TextHeight('Wq');
      t:=0;
      case VertAlignment of
        taVerticalCenter:
          t:=(Rect.Bottom - Rect.Top - H * L.Count) div 2;
        taAlignBottom:
          t:=(Rect.Bottom - Rect.Top - H * L.Count);
      end;
      if t < 0 then
        t:=0;
      rc:=Rect;
      rc.Top:=rc.Top + t;
      rc.Bottom:=rc.Top + H;
      for I:=0 to L.Count - 1 do
      begin
        Text:=L[I];
        DrawText(Canvas.Handle, PChar(Text), Length(Text), rc, Format or DT_SINGLELINE or DT_END_ELLIPSIS);
        rc.Top:=rc.Top + H;
        rc.Bottom:=rc.Top + H;
      end;
    finally
      L.Free;
    end;
  end
  else
  begin
    DrawText(Canvas.Handle, PChar(Text), Length(Text), rc, DT_CALCRECT or Format);
    Result:=(rc.Right <= Rect.Right) and (rc.Bottom <= Rect.Bottom);
    if rc.Bottom <= Rect.Bottom then
    begin
      case VertAlignment of
        taVerticalCenter:
          Inc(Rect.Top, (Rect.Bottom - rc.Bottom) div 2);
        taAlignBottom:
          Inc(Rect.Top, Rect.Bottom - rc.Bottom);
      end;
    end;
    if not Result and (TruncatedText <> '') then
      Text:=TruncatedText;
    DrawText(Canvas.Handle, PChar(Text), Length(Text), Rect, Format);
  end;
end;

procedure TrDBGrid.DrawCell(ACol, ARow: Longint; aRect: TRect; aState: TGridDrawState);
var
  aColumn: TColumn;
  oldPen: TPenRecall;
  oldFont: TFontRecall;
  ImgBmp: TBitmap;
  CellR, RSort, RFilter, RImg, RText, RGroup: TRect;
  ColIdx, OldActive, SortIdx, BmpIdx, I, maxlines: integer;
  ts, temps: string;
  aLeft, rnum, tempL, TempR: integer;
  GrExpanded, FixColChecked, ColSortDesc, FilterActive, ShowTitleImg, ShowFilterGlyph, ShowSortGlyph, InsideGroup: boolean;
  GroupInf, GroupInfR: TGroupInfo;
  RGroupCenter: TPoint;
  FixColor, FixFontColor: TColor;

  function DrawTitleTextAlign(R: TRect; AText: string; Align: TAlignment): integer;
  var
    a, aLeft, aWidth, aTop, MaxWidth, tHeight: integer;
    Lines: TStringList;
    oldBrush: TBrushRecall;
    RotTitle: boolean;
  begin
    Result:=R.Left;
    // draw text
    if Assigned(aColumn) and (FOptionsEx2.FGroups.FActive or FOptionsEx2.FGroups.FAlwaysUseGroupColors) then
    begin
      with GetFldGroupInfo(aColumn.Field, FOptionsEx2.FGroups.FAlwaysUseGroupColors) do
        if IsGrouped then
        begin
          if IsMaster and (GroupTitle <> '') and not IsGroupMasterExpanded(ACol) then
            AText:=GroupTitle;
          if TitleFontColor <> clNone then
            Canvas.Font.Color:=TitleFontColor;
        end;
    end;
    RotTitle:=FOptionsEx2.FFixedCell.FRotateTitleText;
    if Assigned(FOnGetFixedRowState) then
      FOnGetFixedRowState(self, RawToDataColumn(ACol), aColumn, Canvas.Font, AText, RotTitle);
    if not RotTitle then
      MaxWidth:=R.Right - R.Left
    else
      MaxWidth:=R.Bottom - R.Top;
    tHeight:=Canvas.TextHeight('Wg');
    Lines:=TStringList.Create;
    try
      // Wrap lines
      if not RotTitle then
      begin
        for a:=1 to FTitleLines - 1 do
          Lines.Add(GetLine(AText, MaxWidth));
        Lines.Add(AText);
      end
      else
      begin
        a:=1;
        while (AText <> '') and (a < 10) do
        begin
          Lines.Add(GetLine(AText, MaxWidth));
          Inc(a);
        end;
      end;
      // clean lines
      for a:=Lines.Count - 1 downto 0 do
      begin
        Lines[a]:=Trim(Lines[a]);
        if Lines[a] = '' then
          Lines.Delete(a);
      end;
      // rotate font
      if RotTitle then
        FontGetRotated(Canvas.Font, 900);
      // Write Lines
      oldBrush:=TBrushRecall.Create(Canvas.Brush);
      try
        Canvas.Brush.Style:=bsClear;
        aLeft:=R.Left;
        for a:=0 to Lines.Count - 1 do
        begin
          if not RotTitle then
          begin
            aWidth:=GetLineWidth(Lines[a]);
            aLeft:=R.Left + 1;
            case Align of
              taCenter:
                aLeft:=R.Left + (MaxWidth - aWidth) div 2;
              taRightJustify:
                aLeft:=R.Right - aWidth;
            end;
            if aLeft <= R.Left then
              aLeft:=R.Left + 1;
            aTop:=R.Top + (R.Bottom - R.Top - tHeight * Lines.Count) div 2 + tHeight * a;
            Canvas.TextRect(R, aLeft, aTop, Lines[a]);
            if a = 0 then
              Result:=aLeft
            else
              if aLeft < Result then
                Result:=aLeft;
          end
          else
          begin
            // rotate text
            if a = 0 then
            begin
              aLeft:=R.Left + (R.Right - R.Left - tHeight * Lines.Count) div 2 + tHeight * a;
              if aLeft <= R.Left then
                aLeft:=R.Left + 1;
              Result:=aLeft;
            end
            else
              Inc(aLeft, tHeight);
            aWidth:=GetLineWidth(Lines[a]);
            aTop:=R.Bottom - 2;
            case Align of
              taCenter:
                aTop:=R.Bottom - (MaxWidth - aWidth) div 2;
              taRightJustify:
                aTop:=R.Top + aWidth;
            end;
            Canvas.TextRect(R, aLeft, aTop, Lines[a]);
          end;
        end;
      finally
        FreeAndNil(oldBrush);
      end;
    finally
      FreeAndNil(Lines);
    end;
  end;

begin
  if ExUpdateLock then
    Exit;
  isActiveRow:=false;
  isHotTrackRow:=false;
  InsideGroup:=false;
  ColIdx:=RawToDataColumn(ACol);
  if IsActive then
  begin
    isActiveRow:=DataLink.ActiveRecord = ARow - 1 - TitleRow;
    isHotTrackRow:=ARow = FMouseOverRow;
  end;

  if (ARow > TitleRow) and (ACol > FixCol) then
  begin
    // draw data cell
    inherited DrawCell(ACol, ARow, aRect, aState);
    Exit;
  end;

  // draw fixed cell
  if (ACol > FixCol) then
  begin
    aColumn:=Columns[ColIdx];
    FixColor:=aColumn.Title.Color;
    FixFontColor:=aColumn.Title.Font.Color;
    if FOptionsEx2.FAppearance.FTitleFontStyleColor or not(cvTitleFont in aColumn.AssignedValues) then
      FixFontColor:=GetStyleFontColor_GridText(FixFontColor, true, gdHotTrack in aState, gdPressed in aState);
  end
  else
  begin
    aColumn:=nil;
    FixColor:=FixedColor;
    FixFontColor:=GetStyleFontColor_GridText(TitleFont.Color, true);
  end;

  FixCellRectHiddenColumns(aRect, ACol);
  CellR:=aRect;
  if (dgColLines in Options) or (dgRowLines in Options) then
    InflateRect(aRect, -1, -1)
  else
    if (ARow > TitleRow) and (ACol = FixCol) then
      Inc(aRect.Top);
  if not(dgColLines in Options) then
    Inc(aRect.Left);

  if CheckHotTrackCoord(FHotTrackCell, ACol, ARow) then
    aState:=aState + [gdHotTrack];
  RText:=aRect;
  if (ARow = TitleRow) and (ACol > FixCol) and (FTitleLineText.SumHeight > 0) then
    RText.Top:=RText.Top + FTitleLineText.SumHeight;
  DrawCellBackground(RText, FixColor, aState, ACol, ARow);

  // TitleEx Btn
  if (ARow = TitleRow) and (ACol = FixCol) then
  begin
    // add space for title line, filterbar and groups
    RText:=TitleRectToTextRect(aRect);
    // Dec(RText.Bottom, FOptionsEx2.FFilters.TextBarBootomHeight);
    InflateRect(RText, -1, -1);

    if (gdPressed in aState) then
    begin
      Inc(RText.Top);
      Inc(RText.Right);
    end;

    aLeft:=RText.Left;
    // draw fix col title text
    if FFixedColText.FTitleText <> '' then
    begin
      Inc(RText.Left, IndicatorWidth);
      Canvas.Font.Assign(TitleFont);
      Canvas.Font.Color:=FixFontColor;
      ts:=FFixedColText.FTitleText;
      DrawTitleTextAlign(RText, ts, FFixedColText.FTextAlign);
    end;
    if FOptionsEx2.FFixedCell.FTitleExBtn then
    begin
      // draw sum picture
      RText.Left:=aLeft;
      if FFixedColText.FTitleText <> '' then
        RText.Right:=RText.Left + IndicatorWidth;
      BitmapChangeColor(rGlyphs.BmpFromIndex(rGlyphs.TitleExIndex(IsFiltered), FScaleRatio, FCustomGlyphs), TempBmp, FixFontColor);
      DrawBmpCenter(Canvas, RText, TempBmp);
    end;
    if FOptionsEx2.FFilters.FTextBar then
    begin
      // draw filter bar glyph
      RText:=GetFilterComboRect(aRect, false);
      Inc(RText.Left);
      if (gdPressed in aState) then
      begin
        Inc(RText.Top);
        Inc(RText.Right);
      end;
      ImgBmp:=rGlyphs.BmpFromIndex(rGlyphs.FilterBarIndex(IsFiltered), FScaleRatio, FCustomGlyphs);
      if Assigned(ImgBmp) and (FFixedColText.FixedColWidth > 10) then
      begin
        BitmapChangeColor(ImgBmp, TempBmp, FixFontColor);
        DrawBmpCenter(Canvas, RText, TempBmp);
      end
      else
        DrawTitleTextAlign(RText, 'X', taCenter);
    end;
  end;

  // FixCol Btn
  if (ARow > TitleRow) and (ACol = FixCol) and IsActive then
  begin
    // draw fix col text
    if FFixedColText.IsFixedColUsed and not IsEmpty then
    begin
      OldActive:=DataLink.ActiveRecord;
      try
        RText:=aRect;
        InflateRect(RText, -1, -1);
        if (gdPressed in aState) then
        begin
          // Inc(RText.Top);
          Inc(RText.Left);
        end;
        if FFixedColText.ShowIndicator then
          Inc(RText.Left, IndicatorWidth);
        // locate record and load state
        DataLink.ActiveRecord:=ARow - 1 - TitleRow;
        // get fix col text
        ts:=FFixedColText.TextMask;
        if Pos('<<#>>', ts) > 0 then
        begin
          rnum:=DataLink.DataSet.recno;
          if rnum > 0 then
            temps:=IntToStr(rnum)
          else
            temps:='';
          ts:=rtool.ReplaceStr('<<#>>', temps, ts, true, true);
        end;
        ts:=FillTextMask(ts, DataLink.DataSet);

        if FFixedColText.ParentFont then
        begin
          Canvas.Font.Assign(Font);
          Canvas.Font.Color:=FixFontColor;
        end
        else
          Canvas.Font.Assign(FFixedColText.FFont);
        FixColChecked:=RowIsMultiselected(true, 0);
        if Assigned(FOnGetFixedColState) then
          FOnGetFixedColState(self, isActiveRow, Canvas.Font, ts, FixColChecked);

        // show checkbox ?
        if FFixedColText.ShowCheckbox then
        begin
          TempBmp.Assign(rGlyphs.BmpFromIndex(rGlyphs.CheckIndex(FixColChecked), FScaleRatio, FCustomGlyphs));
          if (RText.Right - RText.Left + 1 >= TempBmp.Width) then
          begin
            if FFixedColText.TextMask <> '' then
              RText.Right:=RText.Left + TempBmp.Width;
            DrawBmpCenter(Canvas, RText, TempBmp);
            RText.Left:=RText.Right + 3;
            RText.Right:=aRect.Right;
          end;
        end;
        // show text
        if (FFixedColText.TextMask <> '') and (RText.Right - RText.Left > 2) then
        begin
          RText.Right:=aRect.Right - 1;
          DrawTextAligned(RText, ts, GetCellAlignment(FFixedColText.FTextAlign, FRowVerticalAlign, (FRowLines = 1) or
                (GetLineWidth(ts) < RText.Right - RText.Left)));
        end;
      finally
        DataLink.ActiveRecord:=OldActive;
      end;
    end;

    // draw fix col RecMarker
    if FFixedColText.ShowIndicator then
    begin
      RText:=aRect;
      // align left if another fixed col content used
      if FFixedColText.IsFixedColUsed then
        RText.Right:=RText.Left + IndicatorWidth;
      ActBmp:=rGlyphs.RecordIndicatorIndex(ARow - 1 - TitleRow = DataLink.ActiveRecord, not FFixedColText.ShowCheckbox and RowIsMultiselected
          (false, ARow - 1 - TitleRow), ReadOnly or (not(dgEditing in Options)) or not DataLink.DataSet.CanModify, DataLink.DataSet.State);
      if ActBmp <> -1 then
      begin
        BitmapChangeColor(rGlyphs.BmpFromIndex(ActBmp, FScaleRatio, FCustomGlyphs), TempBmp, FixFontColor);
        DrawBmpCenter(Canvas, RText, TempBmp);
      end;
    end;
  end;

  // Title Text
  if (ARow = TitleRow) and (ACol > FixCol) and (FTitleLineText.SumHeight > 0) then
  begin
    // save left pos of grid title
    if ColIdx = 0 then
    begin
      FTitleLineText.PosLeft:=aRect.Left;
    end;
    // save right pos of grid title
    if CellR.Right > FTitleLineText.PosRight then
      FTitleLineText.PosRight:=CellR.Right;
    if FTitleLineText.PosRight > ClientRect.Right then
      FTitleLineText.PosRight:=ClientRect.Right;

    oldPen:=TPenRecall.Create(Canvas.Pen);
    oldFont:=TFontRecall.Create(Canvas.Font);
    try
      if FTitleLineText.FColor = clNone then
        Canvas.Brush.Color:=GetStyleBkColor_GradientBase(Canvas.Brush.Color)
      else
        Canvas.Brush.Color:=FTitleLineText.FColor;
      Canvas.Font.Assign(FTitleLineText.FFont);
      if FTitleLineText.ParentFont then
        Canvas.Font.Color:=FixFontColor;

      // set rect
      RText:=Rect(FTitleLineText.PosLeft - 1, 0, FTitleLineText.PosRight, FTitleLineText.SumHeight - 1);
      Canvas.FillRect(RText);
      ts:=FTitleLineText.FText;
      if Assigned(FOnDrawTitleLine) then
        FOnDrawTitleLine(self, RText, ts);
      if ts <> '' then
      begin
        maxlines:=FTitleLineText.GetLineCount;
        for I:=1 to maxlines do
        // proc all lines
        begin
          ts:=FTitleLineText.GetLineText(I);
          if IsActive and (Pos('%d', ts) > 0) then
            ts:=Format(ts, [DataLink.DataSet.RecordCount]);
          RText.Top:=((RText.Bottom - FTitleLineText.LineHeight * maxlines) div (maxlines + 1)) * I + FTitleLineText.LineHeight * (I - 1);
          case FTitleLineText.FTextAlign of
            taLeftJustify:
              aLeft:=RText.Left + 2;
            taCenter:
              aLeft:=(RText.Right - RText.Left - GetLineWidth(ts)) div 2 + RText.Left;
          else
            aLeft:=RText.Right - GetLineWidth(ts) - 2;
          end;
          Canvas.TextRect(RText, aLeft, RText.Top, ts);
        end;
        Canvas.Pen.Color:=clGray; // clBlack;
        Canvas.MoveTo(RText.Left, FTitleLineText.SumHeight - 1);
        Canvas.LineTo(RText.Right, FTitleLineText.SumHeight - 1);
      end;
    finally
      FreeAndNil(oldFont);
      FreeAndNil(oldPen);
    end;
  end;

  // Title Row
  if (ARow = TitleRow) and (ACol > FixCol) then
  begin
    Canvas.Font.Assign(aColumn.Title.Font);
    Canvas.Font.Color:=FixFontColor;
    Canvas.Brush.Color:=aColumn.Title.Color;
    // init text rect
    RText:=TitleRectToTextRect(aRect);
    if (gdPressed in aState) then
    begin
      Inc(RText.Top);
      Inc(RText.Right);
    end;
    RImg:=RText;
    RFilter:=RText;
    InflateRect(RFilter, -1, -1);
    RFilter.Left:=RFilter.Right;
    RSort:=RFilter;

    // set filter button
    ShowFilterGlyph:=IsFunnelButtonColumn(ColIdx) and CheckColWidthForFunnelButton(RText.Right - RText.Left);
    if ShowFilterGlyph then
    begin
      RFilter.Left:=RFilter.Right - (rGlyphs.FilterWidth(FScaleRatio) + 2);
      RSort.Right:=RFilter.Left - 1;
      RSort.Left:=RSort.Right;
      RText.Right:=RFilter.Left - 2;
    end;
    // set sort marker
    SortIdx:=PosInSortField(aColumn.FieldName);
    ShowSortGlyph:=(SortIdx > 0) and ((RText.Right - RText.Left) - (rGlyphs.SortWidth(FScaleRatio) + 2) > UseScale(FOptionsEx2.FFixedCell.FTitleTextMinWidth));
    if ShowSortGlyph then
    begin
      RSort.Left:=RSort.Right - rGlyphs.SortWidth(FScaleRatio) - 2;
      RText.Right:=RSort.Left - 2;
    end;
    // set title image
    BmpIdx:=-1;
    ShowTitleImg:=GetTitleImageIndex(aColumn.Field, BmpIdx) and
      ((RText.Right - RText.Left) - (FTitleImageList.Width + UseScale(FOptionsEx2.FFixedCell.FTitleImageSpace)) > UseScale
        (FOptionsEx2.FFixedCell.FTitleTextMinWidth));
    if ShowTitleImg then
      Inc(RText.Left, FTitleImageList.Width + UseScale(FOptionsEx2.FFixedCell.FTitleImageSpace));
    // draw text
    RText.Left:=DrawTitleTextAlign(RText, aColumn.Title.Caption, aColumn.Title.Alignment);
    Canvas.Font.Assign(aColumn.Title.Font);
    Canvas.Font.Color:=FixFontColor;
    // draw title image
    if ShowTitleImg and (BmpIdx >= 0) then
    begin
      RImg.Right:=RText.Left - UseScale(FOptionsEx2.FFixedCell.FTitleImageSpace);
      RImg.Left:=RImg.Right - FTitleImageList.Width;
      ImgBmp:=TBitmap.Create;
      try
        ImgBmp.Canvas.Brush.Color:=clBtnFace;
        ImgBmp.Canvas.FillRect(ImgBmp.Canvas.ClipRect);
        ImgBmp.Transparent:=FTitleImageList.tag = 0;
        FTitleImageList.GetBitmap(BmpIdx, ImgBmp);
        DrawBmpCenter(Canvas, RImg, ImgBmp);
      finally
        FreeAndNil(ImgBmp);
      end;
    end;
    // draw sort marker
    if ShowSortGlyph then
    begin
      if Assigned(FSorter) or Assigned(FOnChangeSortEx) or ((csDesigning in ComponentState) and (SortDirections <> '')) then
        ColSortDesc:=GetSortDirection(aColumn.FieldName)
      else
        ColSortDesc:=FSortDesc;
      DrawBmpCenter(Canvas, RSort, rGlyphs.BmpFromIndex(rGlyphs.SortIndex(ColSortDesc), FScaleRatio, FCustomGlyphs));
      if SortIdx > 1 then
      begin
        // Draw points if more fields sorting
        oldPen:=TPenRecall.Create(Canvas.Pen);
        try
          Canvas.Pen.Color:=Canvas.Font.Color;
          RSort.Top:=RSort.Top + (RSort.Bottom - RSort.Top - rGlyphs.SortWidth(FScaleRatio)) div 2 - 1;
          RSort.Bottom:=RSort.Top + 2;
          RSort.Right:=RSort.Right - 2;
          RSort.Left:=RSort.Right - 2;
          while SortIdx > 1 do
          begin
            Canvas.Rectangle(RSort);
            RSort.Left:=RSort.Left - 3;
            RSort.Right:=RSort.Right - 3;
            Dec(SortIdx);
          end;
        finally
          FreeAndNil(oldPen);
        end;
      end;
    end;
    // draw filter button
    if ShowFilterGlyph then
    begin
      FilterActive:=IsFilterFieldActive(aColumn.FieldName);
      if not FOptionsEx2.FFilters.FFunnelBtnActiveOnly or FilterActive or CheckHotTrackCoord(FHotTrackCell, ACol, ARow) then
        DrawBmpCenter(Canvas, RFilter, rGlyphs.BmpFromIndex(rGlyphs.FilterIndex(FilterActive), FScaleRatio, FCustomGlyphs));
    end;
    // draw FilterBar edits
    if IsFiterTextBarColumn(ColIdx) and (ColIdx <> FOptionsEx2.FFilters.EditColumnIndex) then
    begin
      RFilter:=GetFilterComboRect(CellR, true);
      Canvas.Brush.Style:=bsClear;
      Canvas.Brush.Color:=FOptionsEx2.FFilters.FTextBkColor;
      if Canvas.Brush.Color = clWindow then
        Canvas.Brush.Color:=GetStyleBkColor_ComboBox(Canvas.Brush.Color);
      Canvas.Pen.Color:=clWindowFrame;
      Canvas.Rectangle(RFilter);
      Canvas.Brush.Style:=bsClear;
      InflateRect(RFilter, -3, 0);
      Dec(RFilter.Bottom, 1);
      ts:=FOptionsEx2.FFilters.EditStatuses.Values[aColumn.FieldName];
      if ts = 'E' then
        Canvas.Font.Color:=clRed;
      ts:=FOptionsEx2.FFilters.EditValues.Values[aColumn.FieldName];
      DrawTextAligned(RFilter, ts, DT_SINGLELINE + DT_VCENTER + DT_END_ELLIPSIS + DT_NOPREFIX);
      Canvas.Font.Color:=FixFontColor;
    end;
    // draw Title Row Group indicator
    if FOptionsEx2.FGroups.FActive then
    begin
      GroupInf:=GetColGroupInfo(ACol, false);
      if GroupInf.IsGrouped then
      begin
        GroupInfR:=GetColGroupInfo(ACol + 1, true);
        GrExpanded:=IsGroupMasterExpanded(ACol);
        // get btn rect
        RGroup:=CellR;
        RGroupCenter:=GetGroupBtnRect(RGroup);
        oldPen:=TPenRecall.Create(Canvas.Pen);
        try
          if GroupInf.GroupLineColor = clNone then
            Canvas.Pen.Color:=Canvas.Font.Color // aColumn.Title.Font.Color
          else
            Canvas.Pen.Color:=GroupInf.GroupLineColor;
          Canvas.Pen.Width:=UseScale(FOptionsEx2.FGroups.FGroupLineWidth);
          tempL:=2 + (Canvas.Pen.Width - 1) div 2;
          TempR:=2;
          if Canvas.Pen.Width > 1 then
            TempR:=TempR + (Canvas.Pen.Width) div 1;
          // draw group lines
          if GroupInf.IsMaster then
          begin
            Canvas.Brush.Style:=bsClear;
            DrawRectBidi(RGroup);
            // draw -
            Canvas.MoveTo(RGroup.Left + tempL, RGroupCenter.Y);
            Canvas.LineTo(RGroup.Right - TempR, RGroupCenter.Y);
            if GrExpanded then
            begin
              // draw line to right
              Canvas.MoveTo(RGroup.Right, RGroupCenter.Y);
              Canvas.LineTo(CellR.Right, RGroupCenter.Y);
              InsideGroup:=true;
            end
            else
            begin
              // draw +
              Canvas.MoveTo(RGroupCenter.X, RGroup.Top + tempL);
              Canvas.LineTo(RGroupCenter.X, RGroup.Bottom - TempR);
            end;
          end
          else
          begin
            Canvas.MoveTo(CellR.Left, RGroupCenter.Y);
            if GroupInfR.GroupID = GroupInf.GroupID then
            begin
              // not last in group, draw line to right
              Canvas.LineTo(CellR.Right, RGroupCenter.Y);
              InsideGroup:=true;
            end
            else
              // last in group, draw line to center
              Canvas.LineTo(RGroupCenter.X, RGroupCenter.Y);
            // draw slave line
            Canvas.MoveTo(RGroupCenter.X, RGroupCenter.Y);
            Canvas.LineTo(RGroupCenter.X, RGroup.Bottom);
          end;
        finally
          FreeAndNil(oldPen);
        end;
      end;
    end;
  end;

  // exclude TitleLine height
  if (ARow = TitleRow) and (ACol > FixCol) then
    Inc(CellR.Top, FTitleLineText.SumHeight);

  Canvas.Brush.Color:=FixColor;
  DrawTitleGridLines(ACol, ARow, CellR, aState, InsideGroup);
  // reduce border
  FixCellRectColRowLines(CellR);
  // call events after drawing
  if Assigned(FOnDrawTitleCellEnd) then
    FOnDrawTitleCellEnd(self, ACol, ARow, CellR, aState);
end;

procedure TrDBGrid.DrawCellBackground(const aRect: TRect; AColor: TColor; aState: TGridDrawState; ACol, ARow: integer);
begin
  inherited;
  if Assigned(FOnDrawCellBackground) then
    FOnDrawCellBackground(self, ACol, ARow, aRect, aState);
end;

procedure TrDBGrid.DrawColumnCell(const Rect: TRect; DataCol: integer; Column: TColumn; State: TGridDrawState);
type
  TContentType = (ctText, ctGlyph, ctImage, ctDataImage);
var
  CellR, DrawR: TRect;
  Highlighted, AsLink, IsColored, GrEmpty, ShowValue: boolean;
  ContentType: TContentType;
  FldText: string;
  Background, oldFontColor, AColor, StyleFontColor: TColor;
  Bmp: TBitmap;
  oldBrush: TBrushRecall;
  DataImageDef: TDataImageDef;

  procedure DrawCellTextAlign(R: TRect; Text: string; IsMemo: boolean; HAlign: TAlignment);
  begin
    InflateRect(R, -2, -2);
    R.Right:=R.Right - 1;
    // disable next code to allow to have manual line feed
    // if not IsMemo and (Canvas.TextWidth(Text) < R.Right - R.Left) then sl:=true;
    if (FRowLines = 1) and not IsMemo then
      Text:=ReplaceStr(cStrEOL, RosiCompConfig.rDBGridCfg.CellSingleLineEOLSeparator, Text);
    if FRowLines = 1 then
    begin
      if IsDateTimeField(Column.Field) then
      begin
        if Canvas.TextWidth(Text) > R.Right - R.Left then
          HAlign:=taLeftJustify;
      end;
      DrawTextAligned(R, Text, GetCellAlignment(HAlign, FRowVerticalAlign, true));
    end
    else
      DrawTextAligned(R, Text, GetCellAlignment(HAlign, FRowVerticalAlign, false), FRowVerticalAlign, '', not IsMemo);
  end;

  procedure DrawCellText(R: TRect; Text: string; IsMemo: boolean);
  begin
    DrawCellTextAlign(R, Text, IsMemo, Column.Alignment);
  end;

  procedure DrawFieldText;
  begin
    FldText:='';
    if Assigned(Column.Field) then
      FldText:=Column.Field.DisplayText;
    DrawCellText(DrawR, FldText, false);
  end;

  procedure DrawCellBmp(R: TRect; B: TBitmap; Resize: boolean; AlignCfg: string = '');
  var
    H: integer;
    Transparent: boolean;
  begin
    // InflateRect(R, -1, -1);
    if Resize then
    begin
      Transparent:=B.Transparent;
      rpictool.SmoothResizeBitmap(B, R.Right - R.Left, R.Bottom - R.Top, false);
      B.Transparent:=Transparent;
    end;
    if AlignCfg <> '' then
    begin
      if SameText(copy(AlignCfg, 1, 1), 'L') then
        R.Right:=R.Left + B.Width;
      if SameText(copy(AlignCfg, 1, 1), 'R') then
        R.Left:=R.Right - B.Width;
    end;
    H:=R.Bottom - R.Top;
    if ((FRowLines > 1) or (FRowFixHeight > 0)) and (FRowVerticalAlign <> taVerticalCenter) and (B.Height + 4 < H) then
    begin
      // align Bmp pos
      case FRowVerticalAlign of
        taAlignTop:
          R.Bottom:=R.Top + B.Height + 4;
        taAlignBottom:
          R.Top:=R.Bottom - B.Height - 4;
      end;
    end;
    DrawBmpCenter(Canvas, R, B);
  end;

  procedure ShowDataImage;
  var
    PicAlign: string;
  begin
    PicAlign:=DataImageDef.ImageAlignment;
    if DataImageDef.ShowText <> '' then
    begin
      // display text value instead of graphics
      if DataImageDef.ShowText = '+' then
        DataImageDef.ShowText:=Column.Field.DisplayText // draw field value
      else
        if DataImageDef.ShowText[1] = '''' then
          DataImageDef.ShowText:=copy(DataImageDef.ShowText, 2, 255); // remove first '
      DrawCellText(CellR, DataImageDef.ShowText, false);
      // automatic alignment of picture with text
      if SameText(copy(PicAlign, 1, 1), 'A') then
      begin
        case Column.Alignment of
          taLeftJustify:
            PicAlign:='R';
          taRightJustify, taCenter:
            PicAlign:='L';
        end;
      end;
    end;
    if Assigned(FDataImageList) and DataImageDef.UseDataImage and (DataImageDef.ImageIndex >= 0) and (DataImageDef.ImageIndex < FDataImageList.Count) then
    begin
      // display image
      Bmp:=TBitmap.Create;
      try
        if Highlighted then
          Bmp.Canvas.Brush.Color:=clHighlight
        else
          Bmp.Canvas.Brush.Color:=Canvas.Brush.Color;
        Bmp.Canvas.FillRect(Bmp.Canvas.ClipRect);
        Bmp.Transparent:=FDataImageList.tag = 0;
        FDataImageList.GetBitmap(DataImageDef.ImageIndex, Bmp);
        DrawCellBmp(CellR, Bmp, true, PicAlign);
      finally
        FreeAndNil(Bmp);
      end;
    end;
  end;

  procedure ShowMemoAsText;
  var
    I: integer;
  begin
    FldText:=Column.Field.AsString;
    if FRowLines = 1 then
    begin
      FldText:=ReplaceStr(cStrEOL, ' | ', FldText, true, false);
      for I:=1 to Length(FldText) do
      begin
        if GetLineWidth(copy(FldText, 1, I) + '...') > CellR.Right - CellR.Left - 4 then
        begin
          FldText:=copy(FldText, 1, I - 1) + '...';
          Break;
        end;
      end;
    end;
    DrawCellText(CellR, FldText, FRowLines > 1);
  end;

  procedure DrawColorRect(R: TRect);
  begin
    // try to find color from string
    AColor:=rpictool.StringToColor(Column.Field.AsString);
    if AColor <> clNone then
    begin
      oldBrush:=TBrushRecall.Create(Canvas.Brush);
      try
        R.Right:=R.Left + R.Bottom - R.Top;
        InflateRect(R, -2, -2);
        Canvas.Brush.Style:=bsSolid;
        Canvas.Brush.Color:=AColor;
        Canvas.FillRect(R);
      finally
        FreeAndNil(oldBrush);
      end;
    end;
  end;

  procedure ShowDataBar;
  var
    vmin, vmax, v, perc: double;
    valType: TrDataBarText;
    CFrom, CTo: TColor;
  begin
    if Column.Field.IsNull then
      Exit;
    if Fields_GetDataBarSettings(Column.Field.FieldName, vmin, vmax, valType, CFrom, CTo) then
    begin
      try
        if VarIsNumeric(FGeneralEventRec.FieldValue) then
          try
            v:=FGeneralEventRec.FieldValue;
          except
            v:=Column.Field.AsFloat;
          end
        else
          v:=Column.Field.AsFloat;
      except
        v:=vmin;
      end;
      perc:=DrawDataBar(Canvas, CellR, v, vmin, vmax, CFrom, CTo);
      Canvas.Brush.Style:=bsClear;
      ShowValue:=ShowValue and (valType = dbtValue);
      if valType = dbtPercent then
        DrawCellTextAlign(DrawR, FormatFloat('0', perc * 100) + '%', false, taCenter);
    end;
  end;

  procedure ShowsStatusBar;
  var
    v, Steps: integer;
    BarType: TrStatusBarType;
    CAll, CDone: TColor;
  begin
    if Column.Field.IsNull then
      Exit;
    if Fields_GetStatusBarSettings(Column.Field.FieldName, Steps, BarType, CAll, CDone) then
    begin
      try
        if VarIsNumeric(FGeneralEventRec.FieldValue) then
          try
            v:=FGeneralEventRec.FieldValue;
          except
            v:=Column.Field.AsInteger;
          end
        else
          v:=Column.Field.AsInteger;
      except
        v:=0;
      end;
      DrawStatusBar(Canvas, CellR, v, Steps, BarType, CAll, CDone);
      Canvas.Brush.Style:=bsClear;
    end;
    ShowValue:=false;
  end;

begin
  CellR:=Rect;
  FixCellRectHiddenColumns(CellR, DataToRawColumn(DataCol));
  Background:=Column.Color;
  if not FOptionsEx2.FAppearance.FColumnFontStyleColor and (cvFont in Column.AssignedValues) then
    Canvas.Font.Color:=Column.Font.Color;

  if DataToRawColumn(DataCol) < FixedCols then
  begin
    // fixed cols - revert rect size
    if (gdFixed in State) and ([dgRowLines, dgColLines] * Options = [dgRowLines, dgColLines]) then
    begin
      InflateRect(CellR, 1, 1);
    end;
    // set default or special font
    if FOptionsEx2.FixedCell.FFixedColParentFont then
    begin
      Canvas.Font.Assign(Column.Font);
      Canvas.Font.Color:=GetStyleFontColor_GridText(Canvas.Font.Color, true, gdHotTrack in State, gdPressed in State);
    end
    else
      Canvas.Font.Assign(FOptionsEx2.FixedCell.FFixedColFont);
    if FOptionsEx2.FixedCell.FFixedColColor <> clNone then
      Background:=FOptionsEx2.FixedCell.FFixedColColor;
  end;

  oldFontColor:=Canvas.Font.Color;
  Highlighted:=HighlightCell(DataCol, 0, '', State);
  DataImageDef.UseDataImage:=false;
  DataImageDef.ImageIndex:=-1;
  DataImageDef.ImageAlignment:='';
  DataImageDef.ShowText:='';

  // get current font and background for cell
  GetCellParams(Column.Field, Canvas.Font, Background, DataImageDef, isActiveRow, Highlighted);
  if Background = clHighlight then
  begin
    Highlighted:=true;
    // if (FInternalDrawingStyle <> gdsClassic) then
    // State:=[gdSelected,gdFocused];
  end;

  if FOptionsEx2.FAppearance.FSkipCellHighlight then
    Highlighted:=false;

  Canvas.Brush.Color:=Background;
  if IsActive then
  begin
    // show Link ?
    AsLink:=IsLinkEx(Column.Field);
    // show colored odd Row?
    IsColored:=FOptionsEx2.FAppearance.FDrawColoredRow and (not Highlighted) and (DataLink.DataSet.recno mod 2 = 0) and (Canvas.Brush.Color = Column.Color);
    // show readonly columns gray if edit
    if not AsLink and isActiveRow and (not RosiCompConfig.rDBGridCfg.FontColorReadOnly_EditingOnly or DataLink.Editing) and (FFontColorReadOnly <> clNone) then
      if Column.ReadOnly or (Assigned(Column.Field) and Column.Field.ReadOnly) then
        Canvas.Font.Color:=FFontColorReadOnly;

    if AsLink then
    begin
      Canvas.Font.Style:=Canvas.Font.Style + [fsUnderline];
      if not Highlighted and (FFontColorLink <> clNone) then
        Canvas.Font.Color:=FFontColorLink;
    end;
    if IsColored then
      Canvas.Brush.Color:=FColorOddRow;
    if isActiveRow and (not Highlighted) and (FColorActiveRow <> clNone) and (Background = Column.Color) then
      Canvas.Brush.Color:=FColorActiveRow;
    if isActiveRow and (not Highlighted) and (not RosiCompConfig.rDBGridCfg.ColorRequiredField_EditingOnly or DataLink.Editing) and
      (FColorRequiredField <> clNone) and (Background = Column.Color) then
      if (Assigned(Column.Field) and Column.Field.Required) then
        Canvas.Brush.Color:=FColorRequiredField;
    if isHotTrackRow and (FColorHotTrackRow <> clNone) then
      Canvas.Brush.Color:=FColorHotTrackRow;

    // change color for filtered
    if IsFiltered then
    begin
      if (oldFontColor = Canvas.Font.Color) and not Highlighted and (FFontColorFiltered <> clNone) then
        Canvas.Font.Color:=FFontColorFiltered;
    end
    else
    begin
      if FFilteredFieldNames <> '' then
        FilteredFieldNames:='';
    end;
  end;
{$IFNDEF TrDBGridDrawColumnCellOnEnd}
  inherited DrawColumnCell(CellR, DataCol, Column, State);
{$ENDIF}
  // fill background
  if Highlighted then
  begin
    DrawCellHighlight(CellR, State, Col, Row);
    Canvas.Brush.Style:=bsClear;
  end
  else
    if Assigned(FOnDrawCellBackground) then
      FOnDrawCellBackground(self, Col, Row, CellR, State)
    else
      Canvas.FillRect(CellR);

  if IsActive then
  begin
    ShowValue:=true;
    // use data image condition or design time defs
    if IsEmpty then
      DataImageDef.UseDataImage:=false
    else
      if not DataImageDef.UseDataImage and IsDataImageEx(Column.Field) then
        GetDataImageDef(Column.Field, DataImageDef);
    // find content type of cell - text, glyph, image
    ContentType:=ctText;
    if DataImageDef.UseDataImage then
      ContentType:=ctDataImage
    else
      // Boolean field
      if IsBoolEx(Column.Field) then
      begin
        ContentType:=ctGlyph;
        ActBmp:=rGlyphs.CheckIndex(FieldAsBooleanValue(Column.Field));
      end
      else
        // Graphic field
        if IsGraphicEx(Column.Field) then
        begin
          ContentType:=ctGlyph;
          try
            GrEmpty:=Column.Field.IsNull;
          except
            GrEmpty:=false;
          end;
          ActBmp:=rGlyphs.PictureIndex(GrEmpty, ord(FGlyphLookStyle));
          if FOptionsEx2.FEditors.ShowGraphicContent then
            ContentType:=ctImage;
        end
        else
          // Memo field
          if IsMemoAsGlyph(Column.Field) then
          begin
            ContentType:=ctGlyph;
            ActBmp:=rGlyphs.MemoIndex(Column.Field.AsString = '', ord(FGlyphLookStyle));
          end;

    // draw cell according type
    DrawR:=CellR;
    case ContentType of
      ctGlyph:
        begin
          // show glyph
          StyleFontColor:=GetStyleFontColor_GridText(Canvas.Font.Color);
          TempBmp.Assign(rGlyphs.BmpFromIndex(ActBmp, FScaleRatio, FCustomGlyphs));
          if not((gdSelected in State) or (dgMultiSelect in Options) or (ColorToRGB(StyleFontColor) = ColorToRGB(clBlack)) or
              (ColorToRGB(StyleFontColor) = ColorToRGB(clWhite)) or IsSameColor(StyleFontColor, clWhite, 85)) then
          begin
            // change color of selected image to font color
            BitmapChangeColor(rGlyphs.BmpFromIndex(ActBmp, FScaleRatio, FCustomGlyphs), TempBmp, StyleFontColor);
          end;
          DrawCellBmp(CellR, TempBmp, false);
        end;
      ctImage:
        begin
          // show blob image
          Bmp:=TBitmap.Create;
          Bmp.Transparent:=false;
          try
            if IsGraphicTypeValid(rpictool.ReadBitmapFromField(Column.Field, Bmp)) then
              DrawCellBmp(CellR, Bmp, true);
          finally
            FreeAndNil(Bmp);
          end;
        end;
      ctDataImage:
        ShowDataImage;
    else
      begin
        if Assigned(FRichText) and IsRichTextEx(Column.Field) then
          DrawRichText(CellR, Column.Field.AsString, clNone)
        else
          // memo as text
          if IsMemoAsText(Column.Field) then
            ShowMemoAsText
          else
          begin
            // draw master content instead of field content
            if FOptionsEx2.FGroups.FActive and GetGroupMasterCellContent(Column.Field, FldText) then
            begin
              DrawCellText(CellR, FldText, false);
              ShowValue:=false;
            end;
            if IsPasswordEx(Column.Field) then
            begin
              // use password mask
              FldText:=FFieldsAsPassword.Values[Column.Field.FieldName];
              if FldText = '' then
                FldText:='***';
              DrawCellText(DrawR, FldText, false);
              ShowValue:=false;
            end;
            // text as color - show color square and keep space for color square
            if IsColorEx(Column.Field) then
            begin
              DrawColorRect(DrawR);
              DrawR.Left:=DrawR.Left + DrawR.Bottom - DrawR.Top;
            end;
            if IsDataBarEx(Column.Field) then
              ShowDataBar;
            if IsStatusBarEx(Column.Field) then
              ShowsStatusBar;
            if ShowValue then
            begin
              if (FRowLines = 1) and (FRowVerticalAlign = taAlignTop) and not IsDateTimeField(Column.Field) then
                // use default drawing
                DefaultDrawColumnCell(DrawR, DataCol, Column, State)
              else
                // use own alignment
                DrawFieldText;
            end;
          end;
      end;
    end;

    // Draw focus rect (row or cell)
    Canvas.Brush.Style:=bsSolid;
    if isActiveRow and FOptionsEx2.FAppearance.FDrawRowFocus then
    begin
      if Assigned(FOnDrawActiveRowFocus) then
        FOnDrawActiveRowFocus(self, CellR, DataCol, Column, State)
      else
      begin
        BiDi_DrawFocusRect(Canvas, CellR);
      end;
    end
    else
      if (ValidParentForm(self).ActiveControl = self) and (gdSelected in State) and ((dgAlwaysShowSelection in Options) or Focused) and not
        (dgRowSelect in Options) and not(csDesigning in ComponentState) and not FOptionsEx2.FAppearance.FSkipCellHighlight then
        BiDi_DrawFocusRect(Canvas, CellR);
  end;

  DrawDataGridLines(CellR, DataCol, Column, State);
  // reduce border
  FixCellRectColRowLines(CellR);
  // call events after drawing
{$IFDEF TrDBGridDrawColumnCellOnEnd}
  inherited DrawColumnCell(CellR, DataCol, Column, State);
{$ELSE}
  if Assigned(FOnDrawColumnCellEnd) then
    FOnDrawColumnCellEnd(self, CellR, DataCol, Column, State);
{$ENDIF}
end;

procedure TrDBGrid.DrawTitleGridLines(ACol: integer; ARow: integer; aRect: TRect; aState: TGridDrawState; InsideGroup: boolean);
var
  Pen: TPenRecall;
begin
  // draw edge for classic style
  if Ctl3D and ([dgRowLines, dgColLines] * Options = [dgRowLines, dgColLines]) and (FOptionsEx2.FAppearance.FTitleRowLineColor = clNone) and
    (FOptionsEx2.FAppearance.FTitleColLineColor = clNone)
{$IFDEF D2010P}
  and (FInternalDrawingStyle = gdsClassic) and not(gdPressed in aState)
{$ENDIF} then
  begin
    DrawEdge(Canvas.Handle, aRect, BDR_RAISEDINNER, BF_BOTTOMRIGHT);
    DrawEdge(Canvas.Handle, aRect, BDR_RAISEDINNER, BF_TOPLEFT);
  end;

  Pen:=TPenRecall.Create(Canvas.Pen);
  try
    if (dgRowLines in Options) and (FOptionsEx2.FAppearance.FTitleRowLineColor <> clNone) then
    begin
      Canvas.Pen.Color:=FOptionsEx2.FAppearance.FTitleRowLineColor;
      Canvas.Pen.Style:=FOptionsEx2.FAppearance.FTitleRowLineStyle;
      Canvas.MoveTo(aRect.Left, aRect.Bottom);
      Canvas.LineTo(aRect.Right, aRect.Bottom);
    end;
    if (dgColLines in Options) and (FOptionsEx2.FAppearance.FTitleColLineColor <> clNone) then
    begin
      if InsideGroup and (FOptionsEx2.FAppearance.FTitleColInGroupLineColor <> clNone) then
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FTitleColInGroupLineColor
      else
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FTitleColLineColor;
      Canvas.Pen.Style:=FOptionsEx2.FAppearance.FTitleColLineStyle;
      Canvas.MoveTo(aRect.Right, aRect.Top);
      Canvas.LineTo(aRect.Right, aRect.Bottom);
    end;
  finally
    FreeAndNil(Pen);
  end;
end;

procedure TrDBGrid.DrawDataGridLines(const Rect: TRect; DataCol: integer; Column: TColumn; State: TGridDrawState);
var
  Pen: TPenRecall;
begin
  Pen:=TPenRecall.Create(Canvas.Pen);
  try
    if (dgRowLines in Options) and (FOptionsEx2.FAppearance.FDataRowLineColor <> clNone) then
    begin
      if (DataToRawColumn(DataCol) < FixedCols) and FOptionsEx2.FAppearance.FDataRowInFixedColsAsTitle then
      begin
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FTitleRowLineColor;
        Canvas.Pen.Style:=FOptionsEx2.FAppearance.FTitleRowLineStyle;
      end
      else
      begin
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FDataRowLineColor;
        Canvas.Pen.Style:=FOptionsEx2.FAppearance.FDataRowLineStyle;
      end;
      if Canvas.Pen.Color <> clNone then
      begin
        Canvas.MoveTo(Rect.Left, Rect.Bottom);
        Canvas.LineTo(Rect.Right, Rect.Bottom);
      end;
    end;
    if (dgColLines in Options) and (FOptionsEx2.FAppearance.FDataColLineColor <> clNone) then
    begin
      if DataToRawColumn(DataCol) < FixedCols then
      begin
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FTitleColLineColor;
        Canvas.Pen.Style:=FOptionsEx2.FAppearance.FTitleColLineStyle;
      end
      else
      begin
        Canvas.Pen.Color:=FOptionsEx2.FAppearance.FDataColLineColor;
        Canvas.Pen.Style:=FOptionsEx2.FAppearance.FDataColLineStyle;
      end;
      Canvas.MoveTo(Rect.Right, Rect.Top);
      Canvas.LineTo(Rect.Right, Rect.Bottom);
    end;
  finally
    FreeAndNil(Pen);
  end;
end;

procedure TrDBGrid.DrawBMPInCell(Rect: TRect; Bmp: TBitmap);
begin
  rpictool.DrawBmpCenter(Canvas, Rect, Bmp);
end;

procedure TrDBGrid.DrawRichText(aRect: TRect; Text: string; BkColor: TColor);
var
  R: TRect;
  topm: integer;
begin
  if Assigned(FRichText) then
  begin
    if BkColor <> clNone then
      FRichText.Color:=BkColor
    else
      FRichText.Color:=Canvas.Brush.Color;
    FRichText.RtfText:=Text;
    InflateRect(aRect, -1, -1);
    topm:=0;
    if FRowVerticalAlign <> taAlignTop then
    begin
      // calc rect
      R:=aRect;
      FRichText.CalculateTextHeight(R);
      if R.Bottom < aRect.Bottom then
      begin
        case FRowVerticalAlign of
          taAlignBottom:
            topm:=aRect.Bottom - aRect.Top - (R.Bottom - R.Top);
          taVerticalCenter:
            topm:=(aRect.Bottom - aRect.Top - (R.Bottom - R.Top)) div 2;
        end;
      end;
    end;
    FRichText.DrawToCanvas(aRect, 1, topm, Canvas);
  end;
end;

procedure TrDBGrid.RefreshCellParamsCond;
begin
  Invalidate;
end;

procedure TrDBGrid.GetCellParams(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef; ActiveRow, Highlight: boolean);
begin
  FCellParamsCondList.GetFieldCond(Field, AFont, Background, DataImageDef, ActiveRow, Highlight, false);
  if Assigned(FOnGetCellParams) then
    FOnGetCellParams(self, Field, AFont, Background, ActiveRow, Highlight);
  if Assigned(FOnGetCellParamsEx) then
    FOnGetCellParamsEx(self, Field, AFont, Background, DataImageDef, ActiveRow, Highlight);
end;

// ==============================================================================
// Messages event block
procedure TrDBGrid.WMSize(var Message: TWMSize);
begin
  inherited;
  // Change Columns width after resize
  if oldWidth <> Message.Width then
    SetColumnWidth(FColumnWidth);
  oldWidth:=Message.Width;
  FilterComboRefreshBounds;
  if (csDesigning in ComponentState) then
    Invalidate;
end;

procedure TrDBGrid.CMMouseLeave(var AMsg: TMessage);
begin
  Screen.Cursor:=Cursor;
  MouseOverRow:=-1;
  InvalidateHotTrackCell;
  ResetHotTrackCell;
  inherited;
end;

procedure TrDBGrid.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if FTitleLineText.FParentFont then
  begin
    FTitleLineText.FFont.Assign(Font);
    FTitleLineText.FParentFont:=true;
  end;
  if FFooterRow.FParentFont then
  begin
    FFooterRow.FFont.Assign(Font);
    FFooterRow.FParentFont:=true;
  end;
end;

procedure TrDBGrid.CMHintShow(var Message: TCMHintShow);
var
  C: TColumn;
  txt, hnt, s: string;
  AlwaysShow: boolean;
  W: integer;
begin
  // Show string on mouse possition for Hint
  if not GetMouseOverCellHint(C, txt, AlwaysShow) then
    Exit;
  if AlwaysShow or not Assigned(C) then
    hnt:=txt
  else
  begin
    try
      // check hint width
      if (MMoveCoord.Y = TitleRow) then
      begin
        // find spec. hint for title line
        s:=FTitleHints.Values[C.FieldName];
        if s <> '' then
          hnt:=ReplaceStr('|', #13, s, true, true)
        else
        begin
          // check title line
          Canvas.Font.Assign(C.Title.Font);
          W:=0;
          if IsTitleImageEx(C.Field) then
            W:=FTitleImageList.Width + 1;
          if GetLineWidth(txt, C.Title.Font) + W > C.Width - 2 then
            hnt:=txt;
        end;
      end
      else
        // check data row
        if GetLineWidth(txt, C.Font) > C.Width - 2 then
          hnt:=txt;
    except
      hnt:='';
    end;
  end;
  // proc hint event
  if Assigned(FOnGetHintStringEx) then
    FOnGetHintStringEx(self, C, MMoveCoord.Y = TitleRow, MMoveCoord.Y = cFooterRowYCoord, FHintFieldValues, txt, hnt);
  // Set hint string in message
  Message.HintInfo^.HintStr:=hnt;
  inherited;
end;

procedure TrDBGrid.FOnTimer(Sender: TObject);
begin
  StopSearching;
end;

// ==============================================================================
// CLICK and MOUSE event block
procedure TrDBGrid.CellClick(Column: TColumn);
var
  s: string;
begin
  inherited CellClick(Column);
  if Assigned(Column.Field) then
  begin
    // click to link ?
    if isLink and IsLinkEx(Column.Field) then
    begin
      case Fields_GetLinkType(Column.Field.FieldName) of
        ltHTTP, ltMailTo:
          begin
            if Column.Field.AsString <> '' then
            begin
              s:=FFieldsAsLink.Values[Column.Field.FieldName];
              if AnsiPosEx(s, Column.Field.AsString, 1, false) = 0 then
                s:=s + Column.Field.AsString
              else
                s:=Column.Field.AsString;
              RunFileSimple(s);
            end;
          end;
        ltUserDefined:
          if Assigned(FOnLinkClick) then
            FOnLinkClick(self, DataLink.DataSet, Column, Column.Field, Column.Field.AsVariant);
      end;
    end;
    // click to bool ?
    if (MDownCoord.X > FixCol) and (MDownCoord.Y > TitleRow) then
      if FOptionsEx2.FEditors.FBoolOneClickChange and IsBoolEx(Column.Field) then
        ChangeBool(false);
  end;
  if FOptionsEx2.FAppearance.FDragCell then
    BeginDrag(false);
end;

procedure TrDBGrid.TitleClick(Column: TColumn);
var
  FldExists: boolean;
  SelFld: string;
  SortFld: string;
  SortDirs: string;
  SelDesc: boolean;
begin
  if isPushing and (not isColMoved) then // and (not isBlockedClick)
  begin
    inherited TitleClick(Column);
    if not IsActive then
      Exit;
    // Change sort
    SelFld:=AnsiUpperCase(Column.FieldName);
    FldExists:=PosInSortField(SelFld) > 0;
    if (ssShift in ShiftState) then
    begin
      SortFld:=FSortFieldName;
      // selected next field for sorting
      if not FldExists then
        rtool.AddToSepList(SortFld, SelFld, cFieldSep);
    end
    else
      SortFld:=SelFld;
    if Assigned(FSorter) or Assigned(FOnChangeSortEx) then
    begin
      if FldExists then
        SortDirs:=ChangeSortDirection(SelFld)
      else
        SortDirs:=FSortDirections;
      if not(FldExists and (ssCtrl in ShiftState)) then
        SelFld:='';
      CheckFormatSortDirection(SortFld, SortDirs, SelFld);
      ChangeSortEx(SortFld, SortDirs);
    end
    else
    begin
      if FldExists then
        SelDesc:=not FSortDesc
      else
        SelDesc:=FSortDesc;
      ChangeSort(SortFld, SelDesc);
    end;
  end;
  isColMoved:=false;
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
end;

procedure TrDBGrid.TitleExClick(Button: TMouseButton; Shift: TShiftState);
begin
  if Assigned(OnTitleExClick) then
    OnTitleExClick(self)
  else
  begin
    ShowHintRecCount:=true;
    Application.HintPause:=0;
    SendMessage(Handle, WM_MOUSEMOVE, 0, 0);
    Application.ActivateHint(Mouse.CursorPos);
    // if IsActive then
    // rdbtool.ShowRecordCount(Self);
  end;
end;

procedure TrDBGrid.TitlePopupClick(Sender: TObject);
begin
  if not(Sender is TMenuItem) then
    Exit;
  if TMenuItem(Sender).tag > 1 then
    Clipboard.AsText:=TMenuItem(Sender).Hint;
end;

procedure TrDBGrid.FixColClick(Button: TMouseButton; Shift: TShiftState);
begin
  if Assigned(OnFixColClick) then
    FOnFixColClick(self);
end;

function TrDBGrid.CheckDropDownEdit: boolean;
begin
  Result:=false;
  if Columns[SelectedIndex].ButtonStyle = cbsEllipsis then
  begin
    ShowEditor;
    EditButtonClick;
    Result:=true;
  end
  else
    if (Columns[SelectedIndex].ButtonStyle = cbsAuto) and
      ((Columns[SelectedIndex].PickList.Count > 0) or ((SelectedField <> nil) and (SelectedField.FieldKind = fkLookup))) and CanEdit then
    begin
      ShowEditor;
      if Assigned(InplaceEditor) then
        GetInplaceEditor.DropDown;
      Result:=true;
    end;
end;

procedure TrDBGrid.Click;
begin
  inherited Click;
end;

procedure TrDBGrid.DblClick;
var
  Inh: boolean;
begin
  Inh:=true;
  if IsActive and not FOptionsEx2.FEditors.FNotProcDblClick then
  begin
    isDblClick:=true;
    // DblClick to Datacell
    if MDownCoord.Y = TitleRow then
      Inh:=false;
    if (MDownCoord.X > FixCol) and (MDownCoord.Y > TitleRow) then
    begin
      if (not FOptionsEx2.FEditors.FBoolOneClickChange) and IsBoolEx(SelectedField) then
        Inh:=not ChangeBool(false);
      if IsGraphicEx(SelectedField) and not FOptionsEx2.FEditors.FNotProcGraphicDblClick then
      begin
        ShowGraphic(SelectedField);
        Inh:=false;
{$IFDEF DEMO}
        ShowDemoMsg;
{$ENDIF}
      end
      else
        if IsMemoEx(SelectedField, true) and not FOptionsEx2.FEditors.FNotProcMemoDblClick then
        begin
          ShowMemo(SelectedField);
          Inh:=false;
{$IFDEF DEMO}
          ShowDemoMsg;
{$ENDIF}
        end;
    end;
  end;
  if Inh then
  begin
    inherited DblClick;
    if (MDownCoord.X > FixCol) and (MDownCoord.Y > TitleRow) then
      if not IsColumnReadOnly(Columns[SelectedIndex]) and not CheckDropDownEdit then
        ShowEditor;
  end;
end;

procedure TrDBGrid.MouseMove(Shift: TShiftState; X, Y: integer);
var
  oldCoord, NewCoord: TGridCoord;
  Reshow, AlwaysShow: boolean;
  C: TColumn;
  R: TRect;
  s: string;
  ci, W: integer;
begin
  NewCoord:=MouseCoordEx(X, Y);
  Reshow:=not CheckCoord(MMoveCoord, NewCoord);
  oldCoord:=MMoveCoord;
  MMoveCoord:=NewCoord;
  // fix bug in MouseMove - eliminate flickering of title row and add highlight fixed col
  case FGridState of
    gsSelecting, gsColMoving, gsRowMoving, gsRowSizing, gsColSizing:
      inherited MouseMove(Shift, X, Y);
  else
    begin
      if (csDesigning in ComponentState) then
        Exit;
      if not CheckHotTrackCoord(FHotTrackCell, NewCoord) then
      begin
        // invalidate previous hot cell
        InvalidateHotTrackCell;
        // Highlight "fixed" cell
        if (FOptionsEx2.FFixedCell.FTitleBtn and (NewCoord.X > FixCol) and (NewCoord.Y = TitleRow) and (Y > FTitleLineText.SumHeight)) or
          (FOptionsEx2.FFixedCell.FFixedColBtn and (NewCoord.Y > TitleRow) and (NewCoord.X = FixCol)) or
          (FOptionsEx2.FFixedCell.FTitleExBtn and CheckCoordXY(NewCoord, FixCol, TitleRow)) then
        begin
          // save new hot cell and invalidate
          SetHotTrackCoord(FHotTrackCell, NewCoord);
          InvalidateHotTrackCell;
        end
        else
          ResetHotTrackCell;
        FHotTrackCell.Pressed:=false;
      end
      else
        // check move to titletext area
        if (FOptionsEx2.FFixedCell.FTitleBtn and (NewCoord.X > FixCol) and (NewCoord.Y = TitleRow) and (Y <= FTitleLineText.SumHeight)) then
        begin
          InvalidateHotTrackCell;
          ResetHotTrackCell;
        end;

      // scroll data
      if isDataScrolling and (NewCoord.Y > 0) then
        ScrollData(oldCoord.Y - NewCoord.Y);

      // call standard OnMouseMove because inherited is not called
      if Assigned(OnMouseMove) then
        OnMouseMove(self, Shift, X, Y);
    end;
  end;

  if not(csDesigning in ComponentState) and (FColorHotTrackRow <> clNone) then
    MouseOverRow:=NewCoord.Y;

  // Activate hint if mouse enter new cell
  if Reshow then
    Application.ActivateHint(Mouse.CursorPos);

  if not(csDesigning in ComponentState) then
    try
      isLink:=false;
      // above a Link ?
      ci:=RawToDataColumn(NewCoord.X);
      if ci < 0 then
        Exit;
      C:=Columns[ci];
      R:=CellRect(NewCoord.X, NewCoord.Y);
      InflateRect(R, -3, -3);
      OffsetRect(R, 0, 2);
      if not IsEmpty and IsLinkEx(C.Field) and (NewCoord.Y <> TitleRow) then
      begin
        if FDataImagesIni.SectionExists(C.FieldName) then
        begin
          R.Left:=(R.Right - R.Left - FDataImageList.Width) div 2 + R.Left;
          R.Right:=R.Left + FDataImageList.Width;
          isLink:=PtInRect(R, Point(X, Y));
        end
        else
          if GetMouseOverCellHint(C, s, AlwaysShow) then
          begin
            W:=GetLineWidth(s, C.Font);
            case C.Alignment of
              taLeftJustify:
                R.Right:=R.Left + W;
              taRightJustify:
                R.Left:=R.Right - W;
              taCenter:
                begin
                  R.Left:=(R.Right - R.Left - W) div 2 + R.Left;
                  R.Right:=R.Left + W;
                end;
            end;
            isLink:=PtInRect(R, Point(X, Y));
          end;
        if isLink then
          Screen.Cursor:=crHandPoint
        else
          Screen.Cursor:=Cursor;
      end
      else
        Screen.Cursor:=Cursor;
    except
    end;
end;

procedure TrDBGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  CellR: TRect;
  DataCol: integer;
  GroupInf: TGroupInfo;
  SkipInheritedBug, ClickToIntBtn: boolean;
begin
  // save keyb and mouse state
  ShiftState:=Shift;
  isPushing:=false;
  MDownCoord:=MouseCoordEx(X, Y);
  CellR:=CellRect(MDownCoord.X, MDownCoord.Y);
  DataCol:=RawToDataColumn(MDownCoord.X);
  ClickToIntBtn:=false;
  // titleEx click
  if CheckCoordXY(MDownCoord, FixCol, TitleRow) then
  begin
    ClickToIntBtn:=PointInFilterComboRect(CellR, Point(X, Y), -999, false);
  end;
  // title row click
  if (MDownCoord.X > FixCol) and (MDownCoord.Y = TitleRow) then
  begin
    ClickToIntBtn:=(Y < FTitleLineText.SumHeight) or PointInFilterComboRect(CellR, Point(X, Y), DataCol, true) or PointInFilterFunnelRect
      (CellR, Point(X, Y), DataCol) or PointInGroupBtnRect(CellR, Point(X, Y), MDownCoord.X, GroupInf);
  end;
  if ClickToIntBtn then
  begin
    SetFocus;
    Exit;
  end;

  // fix the bug in standard DBGrid
  SkipInheritedBug:=false;
  if Focused and not(ssDouble in Shift) and not Sizing(X, Y) then
  begin
    if (MDownCoord.X < 0) and (MDownCoord.Y < 0) then
    begin
      if IsActive and DataLink.Editing then
        THackGridDataLink(DataLink).UpdateData;
      SkipInheritedBug:=true;
    end;
  end;
  if not SkipInheritedBug then
    inherited MouseDown(Button, Shift, X, Y);

  // left button
  if (ssLeft in Shift) and (not isDblClick) then
  begin
    // TitleEx Btn
    if FOptionsEx2.FFixedCell.FTitleExBtn and CheckCoordXY(MDownCoord, FixCol, TitleRow) then
      isPushing:=true;
    // Title Btn
    if FOptionsEx2.FFixedCell.FTitleBtn and (MDownCoord.Y = TitleRow) and (MDownCoord.X > FixCol) then
      isPushing:=true;
    // FixCel Btn
    if FOptionsEx2.FFixedCell.FFixedColBtn and (MDownCoord.Y > TitleRow) and (MDownCoord.X = FixCol) then
    begin
      FHotTrackCell.Pressed:=true;
      FHotTrackCell.Button:=Button;
      InvalidateHotTrackCell;
      isPushing:=true;
    end;
    // data scrolling
    if FOptionsEx2.FAppearance.FDragToScrollData then
      isDataScrolling:=not isPushing;
    // drag and drop cell
    if FOptionsEx2.FAppearance.FDragCell and (MDownCoord.Y > TitleRow) and (MDownCoord.X > FixCol) then
      BeginDrag(false);
  end;

  // column auto width after dbl click on column border
  if isDblClick and (MDownCoord.Y = TitleRow) and (MDownCoord.X > FixCol) then
  begin
    DataCol:=-1;
    if X >= CellR.Right - 3 then
      if UseRightToLeftAlignment then
        DataCol:=RawToDataColumn(MDownCoord.X - 1)
      else
        DataCol:=RawToDataColumn(MDownCoord.X);
    if X <= CellR.Left + 3 then
      if UseRightToLeftAlignment then
        DataCol:=RawToDataColumn(MDownCoord.X)
      else
        DataCol:=RawToDataColumn(MDownCoord.X - 1);

    while DataCol >= 0 do
    begin
      if Columns[DataCol].Visible then
        Break;
      Dec(DataCol);
    end;

    if DataCol >= 0 then
      SetOneColumnWidth(FColumnWidthAfterDblClick, Columns[DataCol], 0);
  end;
end;

procedure TrDBGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  CellR: TRect;
  UpCoord: TGridCoord;
  GroupInf: TGroupInfo;
  ShowFilter: boolean;
  DataCol, W: integer;
begin
  isDataScrolling:=false;
  isColMoved:=false;
  ShowFilter:=false;
  UpCoord:=MouseCoordEx(X, Y);

  // fix resize of first column if width>grid client
  if (FGridState = gsColSizing) and (Columns.Count > 0) then
  begin
    if FixCol = 0 then
      W:=ColWidths[0]
    else
      W:=0;
    if (W + Columns[0].Width > ClientWidth) then
      Columns[0].Width:=Max(16, X - W - Left);
  end;

  if not isDblClick and not((FGridState = gsRowSizing) or (FGridState = gsColSizing)) and CheckCoord(UpCoord, MDownCoord) then
  begin
    // Actual Cell Rect
    CellR:=CellRect(MDownCoord.X, MDownCoord.Y);
    DataCol:=RawToDataColumn(UpCoord.X);

    // titleEx click
    if CheckCoordXY(UpCoord, FixCol, TitleRow) then
    begin
      if PointInFilterComboRect(CellR, Point(X, Y), -999, false) then
        FilterComboHide(true, true);
    end;

    // title row click
    if (UpCoord.X > FixCol) and (UpCoord.Y = TitleRow) then
    begin
      // filter bar click
      if PointInFilterComboRect(CellR, Point(X, Y), DataCol, true) then
        FilterComboShow(GetFilterComboRect(CellR, true), DataCol);
      // funnel filter click
      if PointInFilterFunnelRect(CellR, Point(X, Y), DataCol) then
        ShowFilter:=(FFilterDlgLastColIdx <> UpCoord.X);
      // expand or collapse
      if PointInGroupBtnRect(CellR, Point(X, Y), UpCoord.X, GroupInf) then
        IntExpandCollapseGroup(GroupInf.GroupID, not IsGroupMasterExpanded(UpCoord.X));
    end;

    if (UpCoord.Y = cFooterRowYCoord) and Assigned(FOnFooterCellClick) then
    begin
      // footer cell click
      if (DataCol >= 0) and (DataCol < Columns.Count) then
        FOnFooterCellClick(self, UpCoord.X, Columns[DataCol])
      else
        FOnFooterCellClick(self, UpCoord.X, nil);
    end;
  end;

  inherited MouseUp(Button, Shift, X, Y);

  if ShowFilter then
    IntShowFilterDlg(UpCoord.X)
  else
    FFilterDlgLastColIdx:=-1;

  if isPushing then // and not isBlockedClick then
  begin
    isPushing:=false;
    // Invalidate;
    // Release button on the same cell as pressed
    if CheckCoord(UpCoord, MDownCoord) then
    begin
      // TitleEx Btn
      if CheckCoordXY(UpCoord, FixCol, TitleRow) then
        TitleExClick(Button, Shift);
      // FixCol Btn
      if (UpCoord.X = FixCol) and (UpCoord.Y > TitleRow) then
        FixColClick(Button, Shift);
    end;
  end;

  if FOptionsEx2.FAppearance.FDragCell and Dragging then
    EndDrag(false);

  isDblClick:=false;
end;

function TrDBGrid.MouseCoordEx(X, Y: integer): TGridCoord;
var
  I: integer;
  R: TRect;
begin
  Result:=MouseCoord(X, Y);
  if FFooterRow.FFooterVisible then
  begin
    R:=FFooterRow.FooterRect;
    if PtInRect(R, Point(X, Y)) then
    begin
      Result.Y:=cFooterRowYCoord;
      for I:=0 to ColCount do
        if PtInRect(FFooterRow.FooterCellRect(I), Point(X, Y)) then
        begin
          Result.X:=I;
          Break;
        end;
    end;
  end;
end;

function TrDBGrid.DoMouseWheel(Shift: TShiftState; WheelDelta: integer; MousePos: TPoint): boolean;
var
  Key: Word;
begin
  if Assigned(InplaceEditor) and GetInplaceEditor.ListVisible then
  begin
    while ProcMouseWheel(WheelDelta, Key) do
      SendMessage(GetInplaceEditor.ActiveList.Handle, WM_KEYDOWN, Key, 0);
    Result:=true;
  end
  else
    Result:= inherited DoMouseWheel(Shift, WheelDelta, MousePos);
end;

// ==============================================================================
// KEY event block
procedure TrDBGrid.KeyDown(var Key: Word; Shift: TShiftState);
var
  C: TWinControl;
  s: string;
  DoCancel: boolean;

  procedure DoDelete;
  begin
    if CanEdit then
      try
        if SelectedField.FieldKind = fkLookup then
        begin
          SelectedField.DataSet.FieldByName(SelectedField.KeyFields).Clear;
          if EditorMode then
            InplaceEditor.Clear;
          Invalidate;
          // Key:=0;
        end
        else
        begin
          SelectedField.Clear;
        end;
        if IsBoolEx(SelectedField) then
          ChangeBool(true);
      except
      end;
  end;

begin
  // search functions
  if (Key in [VK_ESCAPE, VK_DELETE, VK_RETURN]) then
    StopSearching;

  if (ssCtrl in Shift) and (Key in [VK_LEFT, VK_RIGHT]) then
  begin
    if FLookupCondition in [lcBrowseFilter, lcNotEditFilter] then
      if SearchGridFilter(Key = VK_LEFT) then
        Key:=0;
  end;
  // if (Key in [VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT]) then
  if (Key in [VK_UP, VK_DOWN]) then
    if SearchNext(Key in [VK_UP, VK_LEFT]) then
      Key:=0;

  // Tab key - if edit select next field else select next control
  if (Key = VK_TAB) and not FOptionsEx2.FAppearance.FStandardTabKey then
    try
      C:=self;
      if IsActive and DataLink.Editing then
        C:=nil;
      while Assigned(C) and not(C is TForm) do
        C:=C.Parent;
      if Assigned(C) then
      begin
        SendMessage(TForm(C).Handle, WM_NEXTDLGCTL, ord(Shift = [ssShift]), 0);
        Key:=0;
      end;
    except
    end;

  // ReturnAsTabKey
  isReturnAsTabKeyActive:=(Key = VK_RETURN) and (Shift = []) and FOptionsEx2.FAppearance.FReturnAsTabKey and not EditorMode;
  if isReturnAsTabKeyActive then
    Key:=VK_TAB;

  // Ctrl+C or Ctrl+Insert - copy to clipboard
  if (CharInSet(chr(Key), ['c', 'C']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssCtrl])) then
  begin
    CopyCellData;
    Key:=0;
  end;
  // Ctrl+X - Cut to clipboard
  if (Shift = [ssCtrl]) and CharInSet(chr(Key), ['x', 'X']) then
  begin
    CopyCellData;
    DoDelete;
    Key:=0;
  end;
  // Ctrl+V or Shift+Insert- paste from clipboard
  if (CharInSet(chr(Key), ['v', 'V']) and (Shift = [ssCtrl])) or ((Key = VK_INSERT) and (Shift = [ssShift])) then
  begin
    if CanEdit then
    begin
      s:=Clipboard.AsText;
      s:=rtool.CleanLastEOL(s);
      if FOptionsEx2.FEditors.FAllowPasteMultipleCells and ((Pos(cStrEOL, s) > 0) or (Pos(#9, s) > 0)) then
        IntPasteData(SelectedIndex, false)
      else
      begin
        try
          if Assigned(SelectedField) and (SelectedField is TNumericField) then
            s:=rtool.CleanNumericString(s);
          if Assigned(SelectedField) and (SelectedField is TDateTimeField) then
            s:=rtool.CleanDateTimeString(s);
          SelectedField.AsString:=s;
          ShowEditor;
          Key:=0;
        except
        end;
      end;
    end;
  end;
  // Del key - delete field
  if (Key = VK_DELETE) and (Shift = []) then
    DoDelete;

  // F4 - click to edditbutton
  if (Key = VK_F4) then
    CheckDropDownEdit;

  // block record insert and append by insertkey or arrowkey
  DoCancel:=false;
  if IsActive and FOptionsEx2.FEditors.FBlockInsertByKey then
  begin
    if (Key = VK_INSERT) then
      Key:=0;
    if (Key = VK_DOWN) and not(ssCtrl in Shift) and (DataLink.DataSet.State = dsBrowse) then
      DoCancel:=true
  end;

  inherited KeyDown(Key, Shift);

  // cancel new record
  if DoCancel and (DataLink.DataSet.State = dsInsert) then
    DataLink.DataSet.Cancel;
end;

procedure TrDBGrid.KeyPress(var Key: char);
var
  isPickList: boolean;
begin
  // ReturnAsTabKey
  if (Key = chr(VK_RETURN)) and FOptionsEx2.FAppearance.FReturnAsTabKey and not EditorMode and isReturnAsTabKeyActive then
    Key:=chr(VK_TAB);
  isReturnAsTabKeyActive:=false;

  // Space key
  if Key = #32 then
  begin
    if IsBoolEx(SelectedField) then
    begin
      if ChangeBool(false) then
        Key:=#0;
    end;
  end;

  // Enter key
  if Key = chr(VK_RETURN) then
  begin
    // reset temporary lookup value
    StopSearching;
    // Show memo or graphics
    if IsGraphicEx(SelectedField) then
    begin
      ShowGraphic(SelectedField);
      Key:=#0;
    end
    else
      if IsMemoAsGlyph(SelectedField, true) then
      begin
        ShowMemo(SelectedField);
        Key:=#0;
      end;
  end;

  // Lookup grid record (if not edit) or Select value for lookup field
  if Assigned(SelectedField) and ((Key >= #32) or (CharInSet(Key, [#8, #27]) and IsSearching)) then
  begin
    if ((FLookupCondition in [lcNotEdit, lcNotEditFilter]) and not DataLink.Editing) or
      ((FLookupCondition in [lcBrowse, lcBrowseFilter]) and not(dgEditing in Options)) then
      ProcessGridSearchKey(Key)
    else
      if not(Columns[SelectedIndex].ReadOnly or SelectedField.ReadOnly) then
      begin
        if (SelectedField.FieldKind = fkLookup) then
        begin
          if CanEdit then
            ProcessLookupSearchKey(Key);
        end
        else
        begin
          isPickList:=(Columns[SelectedIndex].PickList.Count > 0);
          if isPickList and FOptionsEx2.FEditors.FOnlyPickListValues then
          begin
            Key:=#0;
            isPickList:=false;
          end;
          if isPickList and CanEdit then
            ProcessPicklistSearchKey(Key);
        end;
      end;
  end;
  inherited KeyPress(Key);
end;

procedure TrDBGrid.ProcessGridSearchKey(var Key: char);

  procedure MoveListDataset(Next: boolean);
  begin
    if Next then
      DataLink.DataSet.Next
    else
      DataLink.DataSet.Prior;
  end;

  function TryLocate(s: string): boolean;
  begin
    // try locate text in selected field (found if field begins with the text)
    Result:=false;
    if DataLink.DataSet.Filtered then
      Exit;
    if (SelectedField is TStringField) and (SelectedField.FieldKind in [fkData, fkCalculated]) then
      try
        Result:=DataLink.DataSet.Locate(SelectedField.FieldName, s, [loCaseInsensitive, loPartialKey]);
        // try C as CH
        if not Result and CharInSet(Key, ['c', 'C']) then
          Result:=DataLink.DataSet.Locate(SelectedField.FieldName, s + 'h', [loCaseInsensitive, loPartialKey]);
      except
        // skip locate error
      end;
  end;

  function TryFind(s: string): boolean;
  var
    Bk: TBookmark;
    Next: boolean;
  begin
    // try to find text in selected field (any possition of the text)
    Result:=false;
    // init dataset for scan
    SelectedField.DataSet.DisableControls;
    Bk:=SelectedField.DataSet.GetBookmark;
    SaveResetDatasetEvents(grrQuickLookup);
    if Assigned(oldBeforeScroll) then
      oldBeforeScroll(SelectedField.DataSet);
    try
      // select first record
      Next:=FSearchingPrevNext in [stFindFirst, stFindNext];
      if FSearchingPrevNext = stFindFirst then
        DataLink.DataSet.First
      else
        MoveListDataset(Next);
      // scan all rec
      while (Next and not DataLink.DataSet.Eof) or (not Next and not DataLink.DataSet.Bof) do
      begin
        // try find text for selected field
        if SelectedField is TBooleanField then
        begin
          if Length(s) = 1 then
          begin
            if ContainsText('1AYT', s) then
              Result:=SelectedField.AsBoolean;
            if ContainsText('0NF', s) then
              Result:=not SelectedField.AsBoolean;
          end;
        end
        else
          Result:=AnsiContainsText(SelectedField.AsString, s);
        if Result then
          Break;
        MoveListDataset(Next);
      end;
      // select original record if new record not found
      if not Result then
        try
          DataLink.DataSet.GotoBookmark(Bk);
        except
          // skip bookmark error
        end;
    finally
      // finalize dataset scan
      if Assigned(oldAfterScroll) then
        oldAfterScroll(SelectedField.DataSet);
      RestoreDatasetEvents(grrQuickLookup);
      SelectedField.DataSet.EnableControls;
      SelectedField.DataSet.FreeBookmark(Bk);
    end;
  end;

  procedure ProcHint(s: string; found: boolean);
  var
    P: TPoint;
  begin
    FSearchText:=s;
    // prepare hint string
    s:=GetSearchHintText(s, Key, FSearchingPrevNext, found, FLookupCondition in [lcBrowseFilter, lcNotEditFilter], IsFiltered);
    if s <> '' then
    begin
      // find hint win possition and show hint
      P:=GetActiveCellRect.TopLeft;
      P.Y:=P.Y + DefaultRowHeight + 8;
      P.X:=P.X + 8;
      ShowIntHint(s, P, true);
    end;
    Key:=#0;
  end;

begin
  if FSearchingPrevNext <> stFindFirst then
  begin
    // find next or prev
    ProcHint(FSearchText, TryFind(FSearchText));
    FSearchingPrevNext:=stFindFirst;
  end
  else
  begin
    case Key of
      #0:
        ProcHint(FSearchText, TryLocate(FSearchText) or TryFind(FSearchText));
      #8:
        if Length(FSearchText) = 1 then
          StopSearching
        else
          ProcHint(copy(FSearchText, 1, Length(FSearchText) - 1), true);
      #27:
        StopSearching;
      #32 .. High(char):
        if TryLocate(FSearchText + Key) or TryFind(FSearchText + Key) then
          ProcHint(FSearchText + Key, true)
        else
          ProcHint(FSearchText, false);
    end;
  end;
end;

procedure TrDBGrid.ProcessLookupSearchKey(var Key: char);
var
  SelAll: boolean;

  procedure MoveListDataset(Next: boolean);
  begin
    if Next then
      SelectedField.LookupDataSet.Next
    else
      SelectedField.LookupDataSet.Prior;
  end;

  function TryLocate(s: string): boolean;
  begin
    // try locate text in selected field (found if field begins with the text)
    Result:=false;
    if SelectedField.LookupDataSet.Filtered then
      Exit;
    try
      Result:=SelectedField.LookupDataSet.Locate(SelectedField.LookupResultField, s, [loCaseInsensitive, loPartialKey]);
      // try C as CH
      if not Result and CharInSet(Key, ['c', 'C']) then
        Result:=SelectedField.LookupDataSet.Locate(SelectedField.LookupResultField, s + 'h', [loCaseInsensitive, loPartialKey]);
      SelAll:=not Result;
    except
      Result:=false;
      // skip locate error
    end;
  end;

  function TryFind(s: string): boolean;
  var
    Bk: TBookmark;
    Next: boolean;
  begin
    // try to find text in selected field (any possition of the text)
    Result:=false;
    // init dataset for scan
    SelectedField.LookupDataSet.DisableControls;
    Bk:=SelectedField.LookupDataSet.GetBookmark;
    try
      // select first record
      Next:=FSearchingPrevNext in [stFindFirst, stFindNext];
      if FSearchingPrevNext = stFindFirst then
        SelectedField.LookupDataSet.First
      else
        MoveListDataset(Next);
      // scan all rec
      while (Next and not SelectedField.LookupDataSet.Eof) or (not Next and not SelectedField.LookupDataSet.Bof) do
      begin
        // try find text in lookup result field
        Result:=AnsiContainsText(SelectedField.LookupDataSet.FieldByName(SelectedField.LookupResultField).AsString, s);
        if Result then
          Break;
        MoveListDataset(Next);
      end;
      // select original record if new record not found
      if not Result then
        try
          SelectedField.LookupDataSet.GotoBookmark(Bk);
        except
          // skip bookmark error
        end;
    finally
      // finalize dataset scan
      SelectedField.LookupDataSet.EnableControls;
      SelectedField.LookupDataSet.FreeBookmark(Bk);
    end;
  end;

  procedure ProcHint(s: string; found: boolean);
  var
    P: TPoint;
  begin
    FSearchText:=s;
    // update field and Editor
    if GetInplaceEditor.ListVisible then
    begin
      GetInplaceEditor.FDataList.KeyValue:=SelectedField.LookupDataSet.FieldByName(SelectedField.LookupKeyFields).AsVariant;
    end
    else
    begin
      SelectedField.DataSet.FieldByName(SelectedField.KeyFields).AsVariant:=SelectedField.LookupDataSet.FieldByName(SelectedField.LookupKeyFields).AsVariant;
      if SelAll then
        InplaceEditor.SelectAll
      else
      begin
        InplaceEditor.SelStart:=Length(s);
        InplaceEditor.SelLength:=InplaceEditor.GetTextLen;
      end;
    end;
    // prepare hint string
    if not RosiCompConfig.rDBComponentsCfg.ShowSimpleLookupHint then
      s:=GetSearchHintText(s, Key, FSearchingPrevNext, found, GetInplaceEditor.ListVisible and IsLookupFilterColumn(Columns[SelectedIndex]),
        SelectedField.LookupDataSet.Filtered);
    if s <> '' then
    begin
      // find hint win possition and show hint
      if GetInplaceEditor.ListVisible then
      begin
        P:=ScreenToClient(GetInplaceEditor.DataList.BoundsRect.TopLeft);
        if P.Y < GetActiveCellRect.TopLeft.Y then
          P.Y:=P.Y - 24
        else
          P.Y:=P.Y + GetInplaceEditor.DataList.Height + 8;
      end
      else
      begin
        P:=GetActiveCellRect.TopLeft;
        P.Y:=P.Y + DefaultRowHeight + 8;
      end;
      P.X:=P.X + 4;
      ShowIntHint(s, P, false);
    end;
    if IsLookupFilterColumn(Columns[SelectedIndex]) and found and (Key <> #0) then
      SearchLookupFilter(false);
    Key:=#0;
  end;

begin
  SelAll:=true;
  if FSearchingPrevNext <> stFindFirst then
  begin
    // find next or prev
    ProcHint(FSearchText, TryFind(FSearchText));
    FSearchingPrevNext:=stFindFirst;
  end
  else
  begin
    case Key of
      #0:
        ProcHint(FSearchText, TryLocate(FSearchText) or TryFind(FSearchText));
      #8:
        if Length(FSearchText) = 1 then
        begin
          StopSearching;
          SelectedField.LookupDataSet.Filtered:=false;
        end
        else
          ProcHint(copy(FSearchText, 1, Length(FSearchText) - 1), true);
      #27:
        StopSearching;
      #32 .. High(char):
        if TryLocate(FSearchText + Key) or TryFind(FSearchText + Key) then
          ProcHint(FSearchText + Key, true)
        else
          ProcHint(FSearchText, false);
    end;
  end;
end;

procedure TrDBGrid.ProcessPicklistSearchKey(var Key: char);

  procedure SetInplaceEditText(s: string);
  begin
    if not CanEditModify then
      Exit;
    InplaceEditor.Text:=s;
    SetEditText(Row, Col, s);
    InplaceEditor.SelStart:=Length(FSearchText);
    InplaceEditor.SelLength:=InplaceEditor.GetTextLen;
    Key:=#0;
  end;

  function TryFind(s: string): boolean;
  var
    I: integer;
  begin
    Result:=false;
    for I:=0 to Columns[SelectedIndex].PickList.Count - 1 do
      if AnsiPosEx(s, Columns[SelectedIndex].PickList.Strings[I], 1, false) = 1 then
      begin
        SetInplaceEditText(Columns[SelectedIndex].PickList.Strings[I]);
        Result:=true;
        Break;
      end;
  end;

begin
  if (InplaceEditor.SelStart <> Length(FSearchText)) or (InplaceEditor.SelStart + InplaceEditor.SelLength <> Length(InplaceEditor.Text)) then
  begin
    FSearchText:='';
    Exit;
  end;
  case Key of
    #8:
      begin
        FSearchText:=copy(FSearchText, 1, Length(FSearchText) - 1);
        SetInplaceEditText(FSearchText);
      end;
    #32 .. High(char):
      begin
        FSearchText:=FSearchText + Key;
        if not TryFind(FSearchText) then
          ;
        // SetInplaceEditText(FSearchText);
      end;
  end;
end;

function TrDBGrid.SearchNext(Previous: boolean): boolean;
var
  k: char;
begin
  if not IsSearching then
    Result:=false
  else
  begin
    Result:=true;
    if Previous then
      FSearchingPrevNext:=stFindPrev
    else
      FSearchingPrevNext:=stFindNext;
    k:=#0;
    if not EditorMode then
      ProcessGridSearchKey(k)
    else
      if GetInplaceEditor.FUseDataList then
        ProcessLookupSearchKey(k)
      else
        Result:=false;
  end;
end;

function TrDBGrid.SearchGetFilterString(Fld: TField): string;
var
  LikeCh: string;
begin
  Result:='';
  if not Assigned(Fld) then
    Exit;
  if Fld is TBooleanField then
    // just use cell value
    Result:=GetFilterCond(Fld, GetBooleanFieldTextValue(Fld, Fld.AsBoolean), Fld.IsNull, '')
  else
    if (Fld is TNumericField) or rdbtool.IsDateTimeField(Fld) then
      // just use cell value
      Result:=GetFilterCond(Fld, Fld.AsString, false, '')
    else
      if rdbtool.IsApostrField(Fld) then
      begin
        LikeCh:=FOptionsEx2.FFilters.GetTextFilterLikeChar;
        Result:=GetFilterCond(Fld, LikeCh + FSearchText + LikeCh, false, FOptionsEx2.FFilters.GetTextFilterLikeOperator);
      end;
end;

function TrDBGrid.SearchGridFilter(ResetFilter: boolean): boolean;
var
  FilterStr: string;
  k: char;
begin
  Result:=false;
  if not IsSearching then
    Exit;
  k:=#0;
  if ResetFilter then
  begin
    DataLink.DataSet.Filtered:=false;
    ProcessGridSearchKey(k);
    Result:=true;
    FilterChangedNotification;
  end
  else
  begin
    FilterStr:=SearchGetFilterString(SelectedField);
    if FilterStr <> '' then
    begin
      Result:=rdbtool.SetFilterString(DataLink.DataSet, FilterStr, false);
      ProcessGridSearchKey(k);
      FilterChangedNotification;
    end;
  end;
end;

function TrDBGrid.SearchLookupFilter(ResetFilter: boolean): boolean;
var
  FilterStr: string;
  k: char;
begin
  Result:=false;
  k:=#0;
  if ResetFilter then
  begin
    SelectedField.LookupDataSet.Filtered:=false;
    if IsSearching then
      ProcessLookupSearchKey(k);
    Result:=true;
  end
  else
    if IsSearching then
    begin
      FilterStr:=SearchGetFilterString(SelectedField.LookupDataSet.FindField(SelectedField.LookupResultField));
      if FilterStr <> '' then
      begin
        Result:=rdbtool.SetFilterString(SelectedField.LookupDataSet, FilterStr, false);
        ProcessLookupSearchKey(k);
      end;
    end;
end;

procedure TrDBGrid.StopSearching;
begin
  // stop timer, reset search and hide hint
  FTimer.Enabled:=false;
  FSearchText:='';
  ShowWindow(OwnHintWin.Handle, SW_HIDE);
  Application.HintPause:=oldHintPause;
  Application.HideHint;
  NCursor;
  try
    if Assigned(InplaceEditor) and InplaceEditor.Visible then
    begin
      InplaceEditor.SelStart:=0;
      InplaceEditor.SelLength:=1000;
    end;
  except
  end;
end;

function TrDBGrid.IsSearching: boolean;
begin
  Result:=SearchText <> '';
end;

procedure TrDBGrid.ShowIntHint(Hint: string; TopLeft: TPoint; UseDelay: boolean);
var
  Rp, Rs: TRect;
begin
  // update hint win possition and show hint
  Rs:=OwnHintWin.CalcHintRect(Width, Hint, nil);
  Rp.TopLeft:=Parent.ClientToScreen(TopLeft);
  Rp.Top:=Rp.Top + Top;
  Rp.Right:=Rp.Left + Rs.Right + 4;
  Rp.Bottom:=Rp.Top + Rs.Bottom + 1;
  OwnHintWin.Color:=Application.HintColor;
  OwnHintWin.ActivateHint(Rp, Hint);
  // reset timer
  FTimer.Enabled:=false;
  if UseDelay then
  begin
    FTimer.Interval:=FOptionsEx2.FAppearance.FLookupResetDelay;
    FTimer.Enabled:=true;
  end;
end;

// ==============================================================================
// ListPicker Event block
procedure TrDBGrid.CBProc_LoadAutoPickList(Sender: TObject; DataSet: TDataSet; RecNum: integer);
var
  I: integer;
begin
  for I:=0 to Columns.Count - 1 do
  begin
    if FFieldsAutoPickList.IndexOf(Columns[I].FieldName) >= 0 then
      if Columns[I].PickList.IndexOf(Columns[I].Field.AsString) = -1 then
        Columns[I].PickList.Add(Columns[I].Field.AsString);
  end;
end;

procedure TrDBGrid.LoadAutoPickList;
var
  I: integer;
  L: TStringList;
begin
  if (csDesigning in ComponentState) then
    Exit;
  if not IsActive then
    Exit;
  // set picklist for autopicklist fields
  if FFieldsAutoPickList.Count > 0 then
  begin
    if Assigned(FOnLoadPickList) then
    begin
      // call external proc.
      for I:=0 to Columns.Count - 1 do
        if FFieldsAutoPickList.IndexOf(Columns[I].FieldName) >= 0 then
          FOnLoadPickList(self, DataLink.DataSet, Columns[I].FieldName, Columns[I].PickList);
    end
    else
    begin
      // load picklist directly from dataset
      // reset picklist
      for I:=0 to Columns.Count - 1 do
        if FFieldsAutoPickList.IndexOf(Columns[I].FieldName) >= 0 then
          Columns[I].PickList.Clear;
      // load picklist
      ScanAllRecords(CBProc_LoadAutoPickList, nil, grrLoadPickList);
      // sort picklist
      for I:=0 to Columns.Count - 1 do
        if FFieldsAutoPickList.IndexOf(Columns[I].FieldName) >= 0 then
        begin
          L:=TStringList.Create;
          try
            L.Assign(Columns[I].PickList);
            if IsDateTimeField(Columns[I].Field) then
              L.CustomSort(SortDate)
            else
              if Columns[I].Field is TNumericField then
                L.CustomSort(SortNum)
              else
                L.Sort;
            Columns[I].PickList.Assign(L);
          finally
            FreeAndNil(L);
          end;
        end;
    end;
  end;
end;

procedure TrDBGrid.LoadColorPickList;
var
  I: integer;
begin
  if (csDesigning in ComponentState) then
    Exit;
  // set picklist for colorpicklist fields
  if FFieldsAsColor.Count > 0 then
  begin
    for I:=0 to Columns.Count - 1 do
      if FFieldsAsColor.IndexOf(Columns[I].FieldName) >= 0 then
        rpictool.GetColorList(Columns[I].PickList, true, true, true);
  end;
end;

// ==============================================================================
// FILTER form block
procedure TrDBGrid.DoFilterChange;
begin
  if FFooterRow.FFooterVisible and FFooterRow.FRecalculateAfterFilter then
    RecalculateSummaryResults(true);
  if Assigned(FFilterChanged) then
    FFilterChanged(self);
end;

procedure TrDBGrid.ShowFilterDlg(Column: TColumn = nil);
begin
  ShowFunnelFilter(Column);
end;

procedure TrDBGrid.ShowFunnelFilter(Column: TColumn = nil);
var
  C: integer;
begin
  if Column = nil then
    C:=RawToDataColumn(Col)
  else
    C:=Column.Index;
  if IsFunnelButtonColumn(C) then
    IntShowFilterDlg(DataToRawColumn(C));
end;

procedure TrDBGrid.ShowTextFilter(Column: TColumn = nil);
var
  C: integer;
  R: TRect;
begin
  if Column = nil then
    C:=RawToDataColumn(Col)
  else
    C:=Column.Index;
  if IsFiterTextBarColumn(C) then
  begin
    R:=CellRect(DataToRawColumn(C), 0);
    R:=GetFilterComboRect(R, true);
    FilterComboShow(R, C);
  end;
end;

procedure TrDBGrid.SetTextFilter(Column: TColumn; FilterText: string; ActivateFilter: boolean = true);
begin
  ShowTextFilter(Column);
  if Assigned(FFilterCombo) and FFilterCombo.Visible then
  begin
    FFilterCombo.Text:=FilterText;
    FFilterCombo.OnChange(nil);
    if ActivateFilter then
      SetFocus
    else
      FFilterCombo.SelStart:=Length(FilterText);
    // FFilterCombo.ResetTimer(not SkipDelay, SkipDelay);
  end;
end;

procedure TrDBGrid.IntShowFilterDlg(ACol: integer);
var
  CRect, FormRect: TRect;
  I: integer;
begin
  if not IsActive then
    Exit;
  FFilterDlgLastColIdx:=ACol;
  CreateF(TrDBGridFilterForm, FFilterDlg);
  FFilterDlg.BiDiMode:=BiDiMode;
  CRect:=CellRect(ACol, TitleRow + 1);
  FormRect.TopLeft:=ClientToScreen(CRect.TopLeft);
  FormRect.Bottom:=FormRect.Top + Max(ClientHeight - CRect.Top - 8, FFilterDlg.Constraints.MinHeight);
  FormRect.Right:=FormRect.Left + Max(CRect.Right - CRect.Left, FFilterDlg.Constraints.MinWidth);
  I:=Parent.ClientToScreen(BoundsRect.BottomRight).X - 16;
  if FormRect.Right > I then
  begin
    FormRect.Right:=I;
    FormRect.Left:=FormRect.Right - FFilterDlg.Constraints.MinWidth;
  end;
  with (FFilterDlg as TrDBGridFilterForm) do
    try
      BoundsRect:=FormRect;
      LoadList(self, Columns[RawToDataColumn(ACol)]);
      chAnd.Visible:=FOptionsEx2.FFilters.FMultiFieldFilter;
      I:=CheckBoxList.Top + CheckBoxList.Items.Count * CheckBoxList.ItemHeight + pBtn.Height + 18;
      if chAnd.Visible then
        Inc(I, chAnd.Height);
      if I < Height then
        Height:=I;
      Show;
    except
      on E: Exception do
        ShowErr(sNotSupported, E);
    end;
end;

function TrDBGrid.GetColumnFilterString: string;
var
  I: integer;
  s: string;
begin
  // produce new filter string
  Result:='';
  FFilteredFieldNames:='';
  for I:=0 to High(FColumnFilters) do
  begin
    s:=Trim(FColumnFilters[I].FilterString);
    if s <> '' then
    begin
      if Result <> '' then
        Result:=Result + ' AND ';
      s:='(' + s + ')';
      Result:=Result + s;
      FFilteredFieldNames:=FFilteredFieldNames + FColumnFilters[I].FieldName + FieldListSep;
    end;
  end;
end;

function TrDBGrid.GetFilterPropsIdx(Fld: TField): integer;
var
  I: integer;
begin
  Result:=-1;
  if Assigned(Fld) then
  begin
    // try to find FColumnFilters index for the field
    for I:=0 to High(FColumnFilters) do
    begin
      if AnsiSameText(FColumnFilters[I].FieldName, Fld.FieldName) then
        Result:=I;
    end;
    if Result < 0 then
    begin
      // create new item
      Result:= High(FColumnFilters) + 1;
      SetLength(FColumnFilters, Result + 1);
      // set new filter for the field
      FColumnFilters[Result].FieldName:=Fld.FieldName;
      FColumnFilters[Result].FilterString:='';
    end;
  end;
end;

function TrDBGrid.IntSetFilterForField(Fld: TField; const FilterString: string; MultiFieldFilter: boolean; SetFilter: boolean = true; ShowError: boolean = true)
  : boolean;
var
  I, idx: integer;
  FltStr: string;
begin
  Result:=true;
  if Assigned(Fld) then
  begin
    idx:=GetFilterPropsIdx(Fld);
    // set new filter for the field
    FColumnFilters[idx].FilterString:=Trim(FilterString);

    if not MultiFieldFilter then
    begin
      // reset all other filters
      for I:=0 to High(FColumnFilters) do
        if (I <> idx) then
        begin
          FColumnFilters[I].FilterString:='';
          FOptionsEx2.FFilters.EditValues.Values[FColumnFilters[I].FieldName]:='';
          FOptionsEx2.FFilters.EditStatuses.Values[FColumnFilters[I].FieldName]:='';
        end;
    end;

    if SetFilter then
    begin
      // produce new filter string
      FltStr:=GetColumnFilterString;
      // set filter
      if FltStr = '' then
        ClearFilter(DataLink.DataSet)
      else
      begin
        Result:=SetFilterString(DataLink.DataSet, FltStr, false);
        if not Result then
        begin
          if MultiFieldFilter and (PosEx('ADO', DataLink.DataSet.ClassName) > 0) then
          begin
            if ShowError then
            begin
              DlgW('ADO dataset does not support multi-field filter. Simple field filter is used.');
              if Assigned(FFilterDlg) then
                TrDBGridFilterForm(FFilterDlg).chAnd.Checked:=false;
              Result:=SetFilterForField(Fld, FilterString, false, SetFilter, ShowError);
            end;
          end
          else
            Result:=SetFilterString(DataLink.DataSet, FltStr, ShowError);
        end;
      end;
    end;
  end
  else
  begin
    if SetFilter then
      ClearFilter(DataLink.DataSet);
    SetLength(FColumnFilters, 0);
  end;

  // update funnel glyphs
  FFilterDlgLastColIdx:=-1;
  Invalidate;
  if SetFilter then
    DoFilterChange;
end;

function TrDBGrid.SetFilterForField(Fld: TField; const FilterString: string; MultiFieldFilter: boolean; SetFilter: boolean = true; ShowError: boolean = true)
  : boolean;
begin
  Result:=IntSetFilterForField(Fld, FilterString, MultiFieldFilter, SetFilter, ShowError);
end;

procedure TrDBGrid.DoRefreshFilteredFieldNames;
const
  SepChars: TSysCharSet = [' ', '=', '+', '-', '*', '/', '<', '>', '(', ')', '[', ']', #9 .. #13];
var
  I: integer;
  FldName, FltStr: string;
begin
  FFilteredFieldNames:='';
  FFilterDlgLastColIdx:=-1;
  if IsFiltered then
  begin
    FltStr:=DataLink.DataSet.Filter;
    // extract constat string
    FltStr:=rtool.ReplaceStrFromTo('''', '''', '', FltStr);
    for I:=0 to DataLink.DataSet.FieldCount - 1 do
    begin
      FldName:=DataLink.DataSet.Fields[I].FieldName;
      if ContainWholeWord(FldName, FltStr, SepChars) then
        FFilteredFieldNames:=FFilteredFieldNames + FldName + FieldListSep;
    end;
  end;
  Invalidate;
end;

procedure TrDBGrid.FilterChangedNotification;
begin
  DoRefreshFilteredFieldNames;
  DoFilterChange;
end;

procedure TrDBGrid.RefreshFilteredFieldNames;
begin
  FilterChangedNotification;
end;

// ==============================================================================
// FilterBar func group
function TrDBGrid.IsFilterColumn(ColIdx: integer; AllowLookupField: boolean): boolean;
var
  Fld: TField;
  res: boolean;
begin
  Result:=ColIdx = -999;
  if (ColIdx < 0) or (ColIdx >= Columns.Count) then
    Exit;
  Fld:=Columns[ColIdx].Field;
  if Assigned(Fld) then
  begin
    res:=(not Fld.Calculated) and (AllowLookupField or not Fld.Lookup) and (rdbtool.IsSimpleField(Fld) or rdbtool.IsMemoField(Fld)) and
      ((FieldsFilterButton.Count = 0) or (FieldsFilterButton.IndexOf(Columns[ColIdx].FieldName) >= 0));
    Result:=res or (FieldsFilterMapping.IndexOfName(Fld.FieldName) >= 0);
  end;
  DoGeneralEventFieldBool(geIsFilter, Fld, Result);
end;

function TrDBGrid.IsFunnelButtonColumn(ColIdx: integer): boolean;
begin
  Result:=FOptionsEx2.FFilters.FFunnelBtn and IsFilterColumn(ColIdx, true);
end;

function TrDBGrid.IsFiterTextBarColumn(ColIdx: integer): boolean;
begin
  Result:=FOptionsEx2.FFilters.FTextBar and IsFilterColumn(ColIdx, false);
end;

function TrDBGrid.IsFilterFieldActive(FieldName: string): boolean;
begin
  Result:=ContainWholeWord(FieldName, FFilteredFieldNames, [FieldListSep]);
end;

function TrDBGrid.CheckColWidthForFunnelButton(W: integer): boolean;
begin
  Result:=(W - (rGlyphs.FilterWidth(FScaleRatio) + 2) > UseScale(FOptionsEx2.FFixedCell.FTitleTextMinWidth));
end;

function TrDBGrid.GetFilterComboRect(RCell: TRect; UseMargin: boolean): TRect;
begin
  Result:=RCell;
  Inc(Result.Top, FTitleLineText.SumHeight);
  if FOptionsEx2.FFilters.FTextBarTopLayout then
    Result.Bottom:=Result.Top + FOptionsEx2.FFilters.TextBarTopHeight
  else
    Result.Top:=Result.Bottom - FOptionsEx2.FFilters.TextBarBootomHeight;
  if UseMargin then
  begin
    InflateRect(Result, UseScale(-FOptionsEx2.FFilters.FTextMarginX), UseScale(-FOptionsEx2.FFilters.FTextMarginY));
    if Result.Right < Result.Left + 24 then
      Result.Right:=Result.Left;
  end;
end;

function TrDBGrid.PointInFilterComboRect(RCell: TRect; Point: TPoint; DataCol: integer; UseMargin: boolean): boolean;
begin
  Result:=false;
  if IsFiterTextBarColumn(DataCol) then
  begin
    RCell:=GetFilterComboRect(RCell, UseMargin);
    Result:=PtInRect(RCell, Point);
  end;
end;

function TrDBGrid.PointInFilterFunnelRect(RCell: TRect; Point: TPoint; DataCol: integer): boolean;
begin
  Result:=false;
  RCell:=TitleRectToTextRect(RCell);
  InflateRect(RCell, -1, -1);
  if IsFunnelButtonColumn(DataCol) and CheckColWidthForFunnelButton(RCell.Right - RCell.Left) then
  begin
    InflateRect(RCell, -2, -2);
    if UseRightToLeftAlignment then
      RCell.Right:=RCell.Left + rGlyphs.FilterWidth(FScaleRatio)
    else
      RCell.Left:=RCell.Right - rGlyphs.FilterWidth(FScaleRatio);
    Result:=PtInRect(RCell, Point);
  end;
end;

procedure TrDBGrid.FilterComboRefreshBounds;
var
  R: TRect;
begin
  if not Assigned(FFilterCombo) or not FFilterCombo.Visible or (FOptionsEx2.FFilters.EditColumnIndex < 0) or (TitleRow < 0) then
    Exit;
  R:=CellRect(DataToRawColumn(FOptionsEx2.FFilters.EditColumnIndex), 0);
  R:=GetFilterComboRect(R, true);
  FFilterCombo.BoundsRect:=ClientRectToParentRect(R);
end;

procedure TrDBGrid.FilterComboRefreshColor(Status: string);
begin
  if Assigned(FFilterCombo) then
  begin
    if FFilterCombo.Font.Color <> clRed then
      FFilterComboFontColor:=FFilterCombo.Font.Color;
    if Status = 'E' then
      FFilterCombo.Font.Color:=clRed
    else
      FFilterCombo.Font.Color:=FFilterComboFontColor;
  end;
end;

procedure TrDBGrid.FilterComboMoveToNextCol(Backward: boolean);
var
  PrevCol, NewCol: integer;
begin
  PrevCol:=FOptionsEx2.FFilters.EditColumnIndex;
  NewCol:=PrevCol;
  repeat
    if Backward then
      Dec(NewCol)
    else
      Inc(NewCol);
    if NewCol < 0 then
      NewCol:=Columns.Count - 1;
    if NewCol > Columns.Count - 1 then
      NewCol:=0;
    if Columns[NewCol].Visible and IsFiterTextBarColumn(NewCol) then
      Break;
  until NewCol = PrevCol;
  if NewCol <> PrevCol then
    ShowTextFilter(Columns[NewCol]);
end;

procedure TrDBGrid.FilterComboShow(REdit: TRect; ColumnIndex: integer);
begin
  SetFocus;
  if Assigned(Columns[ColumnIndex].Field) then
  begin
    // store active col+field
    FOptionsEx2.FFilters.EditColumnIndex:=ColumnIndex;
    FOptionsEx2.FFilters.EditChanged:=false;
    // set combo props
    FFilterCombo.BoundsRect:=ClientRectToParentRect(REdit);
    FFilterCombo.Color:=FOptionsEx2.FFilters.FTextBkColor;
    FFilterCombo.UpdateTimerDelay:=FOptionsEx2.FFilters.FSetFilterDelay;
    // set combo values
    FFilterCombo.Text:=FOptionsEx2.FFilters.EditValues.Values[Columns[ColumnIndex].FieldName];
    FilterComboRefreshColor(FOptionsEx2.FFilters.EditStatuses.Values[Columns[ColumnIndex].FieldName]);
    if Assigned(FOnLoadFilterList) then
      // call external proc.
      FOnLoadFilterList(self, DataLink.DataSet, FOptionsEx2.FFilters.GetEditField.FieldName, FFilterCombo.Items)
    else
      if FOptionsEx2.FFilters.FTextListAutoLoad then
        FilterListLoadValues
      else
        FOptionsEx2.FFilters.EditValueList.ReadSection(Columns[ColumnIndex].FieldName, FFilterCombo.Items);
    FFilterCombo.Sorted:=true;
    // show combo
    FFilterCombo.BiDiMode:=BiDiMode;
    FFilterCombo.Visible:=true;
    FFilterCombo.SetFocus;

    // refresh TextBarHeight size
    FOptionsEx2.FFilters.TextBarHeight:=FFilterCombo.Height;
  end;
end;

procedure TrDBGrid.FilterComboHide(ClearAllValues, ClearAllFilters: boolean);
begin
  // hide combo and invalidate title cell
  FFilterCombo.Visible:=false;
  if ClearAllValues then
  begin
    FOptionsEx2.FFilters.EditValues.Clear;
    InvalidateTitles;
  end
  else
    InvalidateCell(DataToRawColumn(FOptionsEx2.FFilters.EditColumnIndex), 0);
  if ClearAllFilters then
  begin
    SetFilterForField(nil, '', true);
  end;
  // reset active col+field
  FOptionsEx2.FFilters.EditColumnIndex:=-1;
  FOptionsEx2.FFilters.EditChanged:=false;
end;

procedure TrDBGrid.FilterComboExit(Sender: TObject);
var
  FltField: TField;
begin
  FFilterCombo.ResetTimer(false, true);
  // store combo values
  FltField:=FOptionsEx2.FFilters.GetEditField;
  if Assigned(FltField) then
    FOptionsEx2.FFilters.EditValues.Values[FltField.FieldName]:=FFilterCombo.Text;
  FilterComboHide(false, false);
end;

procedure TrDBGrid.FilterComboChange(Sender: TObject);
begin
  FOptionsEx2.FFilters.EditChanged:=true;
end;

procedure TrDBGrid.FilterComboKeyPress(Sender: TObject; var Key: char);
begin
  if Key = #13 then
  begin
    // use filter
    FOptionsEx2.FFilters.EditChanged:=true;
    FFilterCombo.ResetTimer(false, true);
    Key:=#0;
  end;
  if Key = #27 then
  begin
    // reset filter
    FOptionsEx2.FFilters.EditChanged:=false;
    FilterComboHide(false, false);
    FFilterCombo.ResetTimer(false, false);
    Key:=#0;
    // select grid
    SetFocus;
  end;
end;

procedure TrDBGrid.FilterComboKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Key in [VK_DOWN, VK_UP] then
    if FFilterCombo.Style = csSimple then
      SetFocus;
  if (Key in [VK_TAB]) and (ssCtrl in Shift) then
    FilterComboMoveToNextCol(false);
  if (Key in [VK_LEFT, VK_RIGHT]) and (ssAlt in Shift) then
    FilterComboMoveToNextCol(Key = VK_LEFT);
end;

procedure TrDBGrid.FilterComboUpdateTimer(Sender: TObject);
var
  FltField: TField;
  cond, FltStr1, FltStr2, FltStrFinal: string;

  procedure SetDateCond(CondText: string);
  var
    D: TDateTime;
    I, Y, M: integer;
  begin
    if TryStrToDate(CondText, D) then
    begin
      // standard date, just use it
      if cond = '' then
      begin
        // use range for whole date to ignore time value if exists
        FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, '>=');
        FltStr2:=GetFilterCond(FltField, GetDateFieldFilterValue(D + 1), false, '<');
      end
      else
        // use entered Cond
        FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, cond);
    end
    else
      if FOptionsEx2.FFilters.FUseDatePartAsRange then
      begin
        // use text as part of date
        // find separator
        M:=0;
        Y:=StrToIntDef(CondText, 0);
        if Y = 0 then
        begin
          // try to find some separator
          for I:=1 to Length(CondText) do
            if Pos(CondText[I], '/-.,') > 0 then
            begin
              M:=StrToIntDef(copy(CondText, 1, I - 1), 0);
              Y:=StrToIntDef(copy(CondText, I + 1, Length(CondText)), 0);
              if M > Y then
              begin
                // swap
                Y:=StrToIntDef(copy(CondText, 1, I - 1), 0);
                M:=StrToIntDef(copy(CondText, I + 1, Length(CondText)), 0);
              end;
            end;
        end;
        try
          if (M < 1) or (M > 12) then
          begin
            // only year entered
            D:=EncodeDate(Y, 1, 1);
            if cond = '' then
            begin
              // set range
              FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, '>=');
              D:=EncodeDate(Y + 1, 1, 1);
              FltStr2:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, '<');
            end
            else
            begin
              // use entered Cond, update date in condition
              if cond = '>' then
                D:=EncodeDate(Y, 12, 31);
              if cond = '<=' then
                D:=EncodeDate(Y, 12, 31);
              FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, cond);
            end;
          end
          else
          begin
            // year + month
            D:=EncodeDate(Y, M, 1);
            if cond = '' then
            begin
              // set range
              FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, '>=');
              if M < 12 then
                D:=EncodeDate(Y, M + 1, 1)
              else
                D:=EncodeDate(Y + 1, 1, 1);
              FltStr2:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, '<');
            end
            else
            begin
              // use entered Cond, update date in condition
              if cond = '>' then
                D:=EncodeDate(Y, M + 1, 1) - 1;
              if cond = '<=' then
                D:=EncodeDate(Y, M + 1, 1) - 1;
              FltStr1:=GetFilterCond(FltField, GetDateFieldFilterValue(D), false, cond);
            end;
          end;
        except
          // invalid, ignore
        end;
      end;
  end;

  procedure GetNumCond(CondText: string);
  var
    I: integer;
    v: double;
    stemp: string;
  begin
    if TryStrToFloat(CondText, v) then
    begin
      // standard number, just use it
      FltStr1:=GetFilterCond(FltField, CondText, false, cond);
    end
    else
      if FOptionsEx2.FFilters.FUseNumberPartAsRange and (FOptionsEx2.FFilters.FNumberPartWildChar <> #0) then
      begin
        I:=Pos(FOptionsEx2.FFilters.FNumberPartWildChar, CondText);
        if I > 1 then
        begin
          try
            // check if decimal separator is just before *
            if copy(CondText, I - 1, 1) = rtool.GetDecimalSeparator then
              Dec(I);
            // get numeric part
            v:=StrToFloat(copy(CondText, 1, I - 1));
            // get rest of wild chars and replace by 0
            stemp:=copy(CondText, I, Length(CondText));
            stemp:=rtool.ReplaceStr(FOptionsEx2.FFilters.FNumberPartWildChar, '0', stemp);
            FltStr1:=GetFilterCond(FltField, GetNumericFieldFilterValue(v) + stemp, false, '>=');
            FltStr2:=GetFilterCond(FltField, GetNumericFieldFilterValue(v + 1) + stemp, false, '<');
          except
            // invalid, ignore
          end;
        end;
      end;
  end;

  procedure GetTextCond(CondText: string);
  var
    LikeCh: char;
  begin
    // add like char for text field
    if (cond = '') and rdbtool.IsApostrField(FltField) and not rdbtool.IsDateTimeField(FltField) then
    begin
      LikeCh:=FOptionsEx2.FFilters.GetTextFilterLikeChar;
      case FOptionsEx2.FFilters.FTextFilterType of
        oftAnyPos:
          CondText:=LikeCh + CondText + LikeCh;
        oftStartPos:
          CondText:=CondText + LikeCh;
        oftExactValue:
          ;
      end;
      cond:=FOptionsEx2.FFilters.GetTextFilterLikeOperator;
    end;
    FltStr1:=GetFilterCond(FltField, CondText, false, cond);
  end;

  procedure ProcOneCond(CondText: string);
  begin
    // extract condition
    cond:='';
    while Pos(copy(CondText, 1, 1), '<>=') > 0 do
    begin
      cond:=cond + copy(CondText, 1, 1);
      CondText:=copy(CondText, 2, Length(CondText));
    end;
    if CondText = '' then
      Exit;
    if cond = '=' then
      cond:='';
    // datetime field
    if rdbtool.IsDateTimeField(FltField) then
      SetDateCond(CondText)
    else
      // numeric field
      if (cond = '') and (FltField is TNumericField) then
        GetNumCond(CondText)
      else
        // text field
        GetTextCond(CondText);
  end;

  procedure ProcAllCond(CondText: string);
  const
    sepAnd = '&';
    sepOr = '|';
  var
    sepIdxAnd, sepIdxOr: integer;
    sep, cond1: string;
    CondAnd: boolean;
  begin
    CondAnd:=false;
    repeat
      sepIdxAnd:=Pos(sepAnd, CondText);
      sepIdxOr:=Pos(sepOr, CondText);
      sep:=sepAnd;
      if (sepIdxOr > 0) and ((sepIdxAnd = 0) or (sepIdxOr < sepIdxAnd)) then
        sep:=sepOr;

      cond1:=Trim(rtool.RemoveStrPart(CondText, sep));
      FltStr1:='';
      FltStr2:='';
      ProcOneCond(cond1);

      if FltStr1 <> '' then
      begin
        if FltStr2 <> '' then
          FltStr1:=FltStr1 + ' AND ' + FltStr2;
        if FltStrFinal <> '' then
        begin
          if CondAnd then
            FltStrFinal:=rdbtool.MakeSQLCondAND([FltStrFinal, FltStr1])
          else
            FltStrFinal:=rdbtool.MakeSQLCondOR([FltStrFinal, FltStr1]);
        end
        else
          FltStrFinal:=FltStr1;

        CondAnd:=sep = sepAnd;
      end;
    until CondText = '';
  end;

var
  FltFieldName, EditTxt: string;
  I: integer;
  FltResult: boolean;
begin
  if not FOptionsEx2.FFilters.EditChanged or (FOptionsEx2.FFilters.EditColumnIndex < 0) then
    Exit;

  FOptionsEx2.FFilters.EditChanged:=false;
  FltField:=Columns[FOptionsEx2.FFilters.EditColumnIndex].Field;
  // check mapping to another field
  FltFieldName:=FieldsFilterMapping.Values[FltField.FieldName];
  if (FltFieldName <> '') and IsDatasetLinked then
    FltField:=DataLink.DataSet.FindField(FltFieldName);
  if not Assigned(FltField) or not Assigned(FltField.DataSet) then
    Exit;

  FltFieldName:=FltField.FieldName;
  EditTxt:=Trim(FFilterCombo.Text);
  FltStrFinal:='';
  FltResult:=true;
  if EditTxt <> '' then
  begin
    // null value
    if SameText(EditTxt, '(null)') or SameText(EditTxt, 'is null') or SameText(EditTxt, '=null') then
      FltStrFinal:=GetFilterCond(FltField, EditTxt, true)
    else
      ProcAllCond(EditTxt);
    FltResult:=FltStrFinal <> '';
  end;
  // activate filter
  if Assigned(FOnSetColumnFilter) then
    FOnSetColumnFilter(self, DataLink.DataSet, FltField, FltStrFinal, FFilterCombo.Text, nil, FltResult)
  else
    if FltResult then
      FltResult:=SetFilterForField(FltField, FltStrFinal, true, true, false);
  if FltResult then
    EditTxt:=''
  else
    EditTxt:='E';
  FOptionsEx2.FFilters.EditStatuses.Values[FltField.FieldName]:=EditTxt;
  FilterComboRefreshColor(EditTxt);
  if FltResult then
    if (FltStrFinal <> '') and FOptionsEx2.FFilters.FTextListAddFilter and not FOptionsEx2.FFilters.FTextListAutoLoad and Assigned(FltField) then
    begin
      if (Length(FFilterCombo.Text) > 2) and not FOptionsEx2.FFilters.EditValueList.ValueExists(FltFieldName, FFilterCombo.Text) then
      begin
        // add value to the list
        I:=FFilterCombo.SelStart;
        FOptionsEx2.FFilters.EditValueList.WriteString(FltFieldName, FFilterCombo.Text, '1');
        FOptionsEx2.FFilters.EditValueList.ReadSection(FltFieldName, FFilterCombo.Items);
        FFilterCombo.SelStart:=I;
      end;
    end;
end;

procedure TrDBGrid.CBProc_FilterListLoadValues(Sender: TObject; DataSet: TDataSet; RecNum: integer);
begin
  if TStringList(Sender).IndexOf(CBScan_Field.AsString) = -1 then
    TStringList(Sender).Add(CBScan_Field.AsString);
end;

procedure TrDBGrid.FilterListLoadValues;
var
  L: TStringList;
begin
  // load picklist directly from dataset
  if (csDesigning in ComponentState) then
    Exit;
  if not IsActive then
    Exit;
  CBScan_Field:=FOptionsEx2.FFilters.GetEditField;
  if not Assigned(CBScan_Field) then
    Exit;
  L:=TStringList.Create;
  try
    // reset picklist
    L.Clear;
    // load picklist
    ScanAllRecords(CBProc_FilterListLoadValues, L, grrLoadPickList);
    // sort picklist
    if IsDateTimeField(CBScan_Field) then
      L.CustomSort(SortDate)
    else
      if CBScan_Field is TNumericField then
        L.CustomSort(SortNum)
      else
        L.Sort;
    FFilterCombo.Items.Assign(L);
  finally
    FreeAndNil(L);
  end;
end;

procedure TrDBGrid.ResetFilter;
begin
  FilterComboHide(true, true);
end;

// ==============================================================================
// GroupFields func block
function TrDBGrid.Fields_GetGroupList(GroupList: TStrings): integer;
var
  I, id: integer;
begin
  FFieldsGroupsIni.ReadSections(GroupList);
  for I:=0 to GroupList.Count - 1 do
  begin
    id:=GetGroupID(GroupList[I]);
    if id < 0 then
      id:=0;
    GroupList.Objects[I]:=TObject(id)
  end;
  Result:=GroupList.Count;
end;

function TrDBGrid.GetGroupID(GroupDef: string): integer;
begin
  Result:=StrToIntDef(rtool.FindStrPart(GroupDef, GroupHeaderSep, 1, true), -1)
end;

function TrDBGrid.Fields_GetGroupDefinition(GroupID: integer): string;
var
  I: integer;
  L: TStringList;
begin
  Result:='';
  L:=TStringList.Create;
  try
    Fields_GetGroupList(L);
    for I:=0 to L.Count - 1 do
    begin
      if GroupID = integer(L.Objects[I]) then
      begin
        Result:=L[I];
        Break;
      end;
    end;
  finally
    FreeAndNil(L);
  end;
end;

function TrDBGrid.Fields_GetGroupSettings(GroupDefinition: string; var GroupID: integer; var GroupTitle, ContentCollapsed: string; var TitleFontColor,
  GroupLineColor: TColor; var GroupLocked: boolean; FieldList: TStrings): boolean;
var
  I: integer;
  L: TStrings;
  s: string;
begin
  // extract group def
  s:=GroupDefinition;
  GroupID:=StrToIntDef(rtool.RemoveStrPart(s, GroupHeaderSep, true), -1);
  GroupTitle:=rtool.RemoveStrPart(s, GroupHeaderSep, true);
  TitleFontColor:=rpictool.StringToColor(rtool.RemoveStrPart(s, GroupHeaderSep, true));
  GroupLineColor:=rpictool.StringToColor(rtool.RemoveStrPart(s, GroupHeaderSep, true));
  ContentCollapsed:=rtool.RemoveStrPart(s, GroupHeaderSep, true);
  GroupLocked:=rtool.RemoveStrPart(s, GroupHeaderSep, true) = '1';
  // load group fields
  if Assigned(FieldList) then
  begin
    L:=TStringList.Create;
    try
      FFieldsGroupsIni.ReadSectionValues(GroupDefinition, L);
      FieldList.Clear;
      for I:=0 to L.Count - 1 do
        if SameText(L.ValueFromIndex[I], 'M') then
          FieldList.Insert(0, L.Names[I])
        else
          FieldList.Add(L.Names[I]);
    finally
      FreeAndNil(L);
    end;
  end;
  Result:=GroupID > 0;
end;

function TrDBGrid.Fields_SetGroupSettings(GroupID: integer; GroupTitle, ContentCollapsed: string; TitleFontColor, GroupLineColor: TColor; GroupLocked: boolean;
  FieldList: TStrings): string;
var
  I: integer;
  res, old: string;
  L: TStrings;

  procedure AddProp(s: string); overload;
  begin
    if res <> '' then
      res:=res + GroupHeaderSep;
    res:=res + ReplaceStr(GroupHeaderSep, ': :', s, true, true);
  end;
  procedure AddProp(I: integer); overload;
  begin
    AddProp(IntToStr(I));
  end;

begin
  // generate section name
  res:='';
  AddProp(GroupID);
  AddProp(GroupTitle);
  if TitleFontColor = clNone then
    AddProp('')
  else
    AddProp(rpictool.ColorToString(TitleFontColor, false));
  if GroupLineColor = clNone then
    AddProp('')
  else
    AddProp(rpictool.ColorToString(GroupLineColor, false));
  AddProp(ContentCollapsed);
  if GroupLocked then
    AddProp('1')
  else
    AddProp('0');
  if (GroupID > 0) then
  begin
    // delete current section name for the same GroupID
    old:=Fields_GetGroupDefinition(GroupID);
    FFieldsGroupsIni.EraseSection(old);
    if Assigned(FieldList) then
    begin
      for I:=0 to FieldList.Count - 1 do
      begin
        if I = 0 then
          old:='M'
        else
          old:='0';
        FFieldsGroupsIni.WriteString(res, FieldList[I], old);
      end;
    end;
    // copy Ini to FFieldsGroups property
    L:=TStringList.Create;
    try
      FFieldsGroupsIni.GetStrings(L);
      FFieldsGroups.Assign(L); ;
    finally
      FreeAndNil(L);
    end;
    FixedGroupedCol(false);
  end;
  Result:=res;
end;

function TrDBGrid.GetFldGroupInfo(Fld: TField; SkipCheckFldGroupOption: boolean): TGroupInfo;
var
  I: integer;
  L: TStrings;
  s: string;
begin
  // reset info
  Result.Field:=Fld;
  Result.IsGrouped:=false;
  Result.GroupID:=-1;
  Result.GroupTitle:='';
  Result.TitleFontColor:=clNone;
  Result.GroupLineColor:=clNone;
  Result.IsMaster:=false;
  Result.MasterField:=nil;
  Result.ContentCollapsed:='';
  if (FOptionsEx2.FGroups.FActive or SkipCheckFldGroupOption) and Assigned(Fld) and IsDatasetLinked then
  begin
    L:=TStringList.Create;
    try
      Fields_GetGroupList(L);
      // load each section and try to find field
      for I:=0 to L.Count - 1 do
      begin
        if FFieldsGroupsIni.ValueExists(L[I], Fld.FieldName) then
        begin
          // load section setting - 5 parts separated by  ::
          s:=L[I];
          Result.IsGrouped:=Fields_GetGroupSettings(L[I], Result.GroupID, Result.GroupTitle, Result.ContentCollapsed, Result.TitleFontColor,
            Result.GroupLineColor, Result.GroupLocked, L);
          // Result.FieldSetting:=L.Values[Fld.FieldName];
          if L.Count > 0 then
          begin
            Result.IsMaster:=AnsiSameText(L[0], Fld.FieldName);
            if Result.IsMaster then
              Result.MasterField:=Fld
            else
              Result.MasterField:=DataLink.DataSet.FindField(L[0]);
          end;
          Exit;
        end;
      end;
    finally
      FreeAndNil(L);
    end;
  end;
end;

function TrDBGrid.GetColGroupInfo(ColIdx: integer; FindFirstVisibleCol: boolean): TGroupInfo;
var
  C: integer;
begin
  C:=RawToDataColumn(ColIdx);
  if FindFirstVisibleCol then
  begin
    while (C >= 0) and (C < Columns.Count) do
    begin
      if Columns[C].Visible then
        Break;
      Inc(ColIdx);
      C:=RawToDataColumn(ColIdx);
    end;
  end;
  if (C >= 0) and (C < Columns.Count) then
    Result:=GetFldGroupInfo(Columns[C].Field)
  else
    Result:=GetFldGroupInfo(nil);
end;

function TrDBGrid.IsGroupMasterExpanded(ColIdx: integer): boolean;
var
  InfoColL, InfoColR: TGroupInfo;
begin
  Result:=false;
  if FOptionsEx2.FGroups.FActive then
  begin
    InfoColL:=GetColGroupInfo(ColIdx, false);
    if InfoColL.IsGrouped and InfoColL.IsMaster then
    begin
      // find first visible col on right
      InfoColR:=GetColGroupInfo(ColIdx + 1, true);
      Result:=InfoColR.IsGrouped and (InfoColL.GroupID = InfoColR.GroupID);
    end;
  end;
end;

procedure TrDBGrid.FixedGroupedCol(SwitchGroupOn: boolean; RestoreDefaults: boolean = false);
var
  GroupInfo1, GroupInfo2: TGroupInfo;
  c1, c2, LastColIdx, step: integer;
  MasterVisible: boolean;
begin
  if RestoreDefaults then
    FFieldsGroups.Assign(FFieldsGroupsDefault);
  if SwitchGroupOn then
    FOptionsEx2.FGroups.Active:=true;
  if not FOptionsEx2.FGroups.FActive then
    Exit;
  c1:=0;
  while c1 < Columns.Count do
  begin
    if c1 < 0 then
      c1:=0;
    step:=1;
    // find master col
    GroupInfo1:=GetFldGroupInfo(Columns[c1].Field);
    if GroupInfo1.IsGrouped and GroupInfo1.IsMaster then
    begin
      MasterVisible:=Columns[c1].Visible;
      LastColIdx:=c1;
      // moved group cols from Left to Master+1
      for c2:=c1 - 1 downto 0 do
      begin
        GroupInfo2:=GetFldGroupInfo(Columns[c2].Field);
        if (GroupInfo2.GroupID = GroupInfo1.GroupID) and not GroupInfo2.IsMaster then
        begin
          if not MasterVisible then
            Columns[c2].Visible:=false;
          Columns[c2].Index:=Columns[c1].Index + 0;
          // go back to previous Col and proc again
          step:=-1;
          Break;
        end;
      end;
      if step > 0 then
        // moved group cols from Right to LastGroupCol+1
        for c2:=c1 + 1 to Columns.Count - 1 do
        begin
          GroupInfo2:=GetFldGroupInfo(Columns[c2].Field);
          if GroupInfo2.GroupID = GroupInfo1.GroupID then
          begin
            if not MasterVisible then
              Columns[c2].Visible:=false;
            if c2 > LastColIdx + 1 then
            begin
              Columns[c2].Index:=Columns[LastColIdx + 1].Index;
              Inc(LastColIdx);
            end
            else
              LastColIdx:=c2;
          end;
        end;
    end;
    Inc(c1, step);
  end;
end;

procedure TrDBGrid.MoveGroupedCol(FromIndex, ToIndex: Longint);
var
  GroupInfo1, GroupInfo2: TGroupInfo;
  ColIdx, I, ColDelta, TargetDelta: integer;
  ColArr: array of integer;
begin
  if not FOptionsEx2.FGroups.FActive then
    Exit;
  // column already MOVED - use ToIndex for moved col
  GroupInfo1:=GetColGroupInfo(ToIndex, false);
  if GroupInfo1.IsGrouped and GroupInfo1.IsMaster then
  begin
    SetLength(ColArr, Columns.Count);
    for I:=0 to High(ColArr) do
      ColArr[I]:=-1;
    // find and save ColIdx for all col in the same Group
    I:=0;
    ColIdx:=FromIndex;
    while (ColIdx <> ToIndex) and (RawToDataColumn(ColIdx) < Columns.Count) do
    begin
      GroupInfo2:=GetColGroupInfo(ColIdx, false);
      if GroupInfo2.GroupID = GroupInfo1.GroupID then
      begin
        ColArr[I]:=ColIdx;
        Inc(I);
      end;
      Inc(ColIdx);
    end;
    // move found cols in group
    ColDelta:=0;
    TargetDelta:=0;
    for I:=0 to High(ColArr) do
    begin
      if ColArr[I] < 0 then
        Break;
      if ColArr[I] + ColDelta > ToIndex then
        Inc(TargetDelta);
      Columns[RawToDataColumn(ColArr[I] + ColDelta)].Index:=RawToDataColumn(ToIndex + TargetDelta);
      if ColArr[I] + ColDelta < ToIndex then
        Dec(ColDelta);
    end;
  end;
end;

function TrDBGrid.CheckColumnDrag(var Origin: integer; var Destination: integer; const MousePt: TPoint): boolean;
var
  GiOrig, GiDestLeft, GiDestRight: TGroupInfo;
  IntDest, C: integer;
begin
  // DlgI(Format('%d-%d',[Origin,Destination]));
  Result:=true;
  if FOptionsEx2.FGroups.FActive then
  begin
    // same pos ?
    if Origin = Destination then
      Exit;
    IntDest:=Destination;
    // move dest for moving to right
    if IntDest > Origin then
      Inc(IntDest);
    // get group info - Orig, DestLeft
    GiOrig:=GetColGroupInfo(Origin, false);
    GiDestLeft:=GetColGroupInfo(IntDest - 1, false);

    if (IntDest > Origin) and GiDestLeft.IsGrouped then
    begin
      // skip hidden=collapsed column
      C:=RawToDataColumn(IntDest);
      while (C < Columns.Count) and not Columns[C].Visible do
      begin
        Inc(IntDest);
        C:=RawToDataColumn(IntDest);
      end;
      Destination:=IntDest - 1;
    end;

    // get group info - DestRight
    GiDestRight:=GetColGroupInfo(IntDest, false);
    if not GiOrig.IsGrouped and not GiDestRight.IsGrouped then
      Exit;

    if GiOrig.IsGrouped then
    begin
      // move inside own group ?
      if (GiOrig.GroupID = GiDestLeft.GroupID) or (GiOrig.GroupID = GiDestRight.GroupID) then
      begin
        // not moved MASTER col
        if GiOrig.IsMaster then
          Result:=false;
        // not moved TO MASTER
        if (GiOrig.GroupID = GiDestRight.GroupID) and GiDestRight.IsMaster then
          Result:=false;
        Exit;
      end
      else
      begin
        // move outside group - not move NOT MASTER
        if not GiOrig.IsMaster then
          Result:=false;
      end;
    end;
    if GiDestRight.IsGrouped then
    begin
      // not moved inside DEST group
      if not GiDestRight.IsMaster then
        Result:=false
    end;
  end;
end;

function TrDBGrid.GetGroupMasterTitleContent(Fld: TField; var Text: string): boolean;
var
  GroupInfo: TGroupInfo;
begin
  Result:=false;
  GroupInfo:=GetFldGroupInfo(Fld);
  if GroupInfo.IsMaster and (GroupInfo.GroupTitle <> '') and not IsFieldGroupMasterExpanded(Fld) then
  begin
    Result:=true;
    Text:=GroupInfo.GroupTitle;
  end;
end;

function TrDBGrid.GetGroupMasterCellContent(Fld: TField; var Text: string): boolean;
var
  GroupInfo: TGroupInfo;
begin
  Result:=false;
  GroupInfo:=GetFldGroupInfo(Fld);
  if GroupInfo.IsMaster and (GroupInfo.ContentCollapsed <> '') and not IsFieldGroupMasterExpanded(Fld) then
  begin
    Result:=true;
    if IsEmpty then
      Text:=''
    else
      Text:=FillTextMask(GroupInfo.ContentCollapsed, DataLink.DataSet);
  end;
end;

function TrDBGrid.GetGroupBtnRect(var RCell: TRect): TPoint;
var
  bs: integer;
begin
  if FOptionsEx2.FGroups.FActive then
  begin
    bs:=Round(ScaleRatio * FOptionsEx2.FGroups.FCollapseBtnSize);
    Inc(RCell.Top, FTitleLineText.SumHeight);
    Inc(RCell.Top, FOptionsEx2.FFilters.TextBarTopHeight);
    RCell.Bottom:=RCell.Top + UseScale(FOptionsEx2.FGroups.FGroupHeightPixels);
    Result.X:=RCell.Left + (RCell.Right - RCell.Left) div 2;
    Result.Y:=RCell.Top + (RCell.Bottom - RCell.Top) div 2;
    RCell.Left:=Result.X - bs div 2;
    RCell.Right:=RCell.Left + bs + 1;
    RCell.Top:=Result.Y - bs div 2;
    RCell.Bottom:=RCell.Top + bs + 1;
  end
  else
  begin
    RCell.Left:=0;
    RCell.Right:=0;
    RCell.Top:=0;
    RCell.Bottom:=0;
  end;
end;

function TrDBGrid.PointInGroupBtnRect(RCell: TRect; Point: TPoint; ColIdx: integer; var GroupInfo: TGroupInfo): boolean;
begin
  Result:=false;
  if FOptionsEx2.FGroups.FActive then
  begin
    GroupInfo:=GetColGroupInfo(ColIdx, false);
    if GroupInfo.IsGrouped and GroupInfo.IsMaster then
    begin
      GetGroupBtnRect(RCell);
      // increase btn rect to simplify click to group btn
      InflateRect(RCell, 2, 2);
      Result:=PtInRect(RCell, Point);
    end;
  end;
end;

function TrDBGrid.GetFieldGroupInfo(Fld: TField): TGroupInfo;
begin
  Result:=GetFldGroupInfo(Fld);
end;

function TrDBGrid.IsFieldGroupMasterVisible(Fld: TField): boolean;
var
  C: integer;
  GroupInfo: TGroupInfo;
begin
  Result:=false;
  GroupInfo:=GetFldGroupInfo(Fld);
  if GroupInfo.IsGrouped then
  begin
    for C:=0 to Columns.Count - 1 do
      if GroupInfo.MasterField = Columns[C].Field then
      begin
        Result:=Columns[C].Visible;
        Exit;
      end;
  end;
end;

function TrDBGrid.IsFieldGroupMasterExpanded(Fld: TField): boolean;
var
  I: integer;
begin
  Result:=false;
  for I:=0 to Columns.Count - 1 do
    if Columns[I].Field = Fld then
      Result:=IsGroupMasterExpanded(DataToRawColumn(I));
end;

procedure TrDBGrid.IntExpandCollapseGroup(GroupID: integer; Expand: boolean);
var
  C: integer;
  GroupInfo1: TGroupInfo;
begin
  if GroupID < 0 then
    Exit;
  for C:=0 to Columns.Count - 1 do
  begin
    GroupInfo1:=GetFldGroupInfo(Columns[C].Field);
    if GroupInfo1.IsGrouped and (GroupInfo1.GroupID = GroupID) and not GroupInfo1.IsMaster then
    begin
      Columns[C].Visible:=Expand;
      // limit Column Width if not stored (first expand)
      if Expand and (not(cvWidth in Columns[C].AssignedValues) or (Columns[C].Width <= 0)) then
        SetOneColumnWidth(cwDefaultWidth, Columns[C]);
    end;
  end;
  if FColumnWidth in [cwSameWidth, cwAutoWidth] then
    SetColumnWidth(FColumnWidth);
end;

procedure TrDBGrid.CollapseGroup(GroupID: integer);
begin
  IntExpandCollapseGroup(GroupID, false);
end;

procedure TrDBGrid.CollapseGroup(Fld: TField);
begin
  with GetFldGroupInfo(Fld) do
    if IsMaster then
      IntExpandCollapseGroup(GroupID, false);
end;

procedure TrDBGrid.ExpandGroup(GroupID: integer);
begin
  IntExpandCollapseGroup(GroupID, true);
end;

procedure TrDBGrid.ExpandGroup(Fld: TField);
begin
  with GetFldGroupInfo(Fld) do
    if IsMaster then
      IntExpandCollapseGroup(GroupID, true);
end;

procedure TrDBGrid.IntExpandCollapseAll(Expand: boolean);
var
  I, GroupID: integer;
  L: TStringList;
begin
  L:=TStringList.Create;
  try
    Fields_GetGroupList(L);
    for I:=0 to L.Count - 1 do
    begin
      GroupID:=integer(L.Objects[I]);
      if GroupID >= 0 then
      begin
        if Expand then
          ExpandGroup(GroupID)
        else
          CollapseGroup(GroupID);
      end;
    end;
  finally
    FreeAndNil(L);
  end;
end;

procedure TrDBGrid.CollapseAll;
begin
  IntExpandCollapseAll(false);
end;

procedure TrDBGrid.ExpandAll;
begin
  IntExpandCollapseAll(true);
end;

// ==============================================================================
// FooterRow group
procedure TrDBGrid.DrawFooterRow;
var
  I, j, W: integer;
  RF, rc: TRect;
  LastTextRight: integer;
  BkColor, LineColor: TColor;
  Frm2: integer;
  DefaultDrawing: boolean;

  C: TColumn;
  Fld: TField;
  sSimpleText, sCellText: string;
  CMask: string;
  Align: TAlignment;
  AlignDef: TAlignmentDefault;
  FontStylesDef: TFontStylesDefault;
  FColor: TColor;
  FSize: integer;

  procedure SetFrm2(a: TAlignment);
  begin
    case a of
      taRightJustify:
        Frm2:=DT_RIGHT;
      taCenter:
        Frm2:=DT_CENTER;
    else
      Frm2:=0;
    end;
  end;

begin
  // check grid visibility
  if not Visible then
    Exit;
  // check footer visibility
  if not Assigned(FFooterRow) or not FFooterRow.FooterVisible then
    Exit;

  // fill footer rect (is not filled by paint because reduced ClientRect)
  RF:=FFooterRow.FooterRect;
  Canvas.Brush.Color:=GetStyleBkColor_Grid(Color);
  Canvas.FillRect(RF);

  // call event
  DefaultDrawing:=true;
  sSimpleText:=FFooterRow.SimpleText;
  if Assigned(FOnDrawFooterRow) then
    FOnDrawFooterRow(self, RF, sSimpleText, DefaultDrawing);
  if not DefaultDrawing then
    Exit;

  // set line and bk color
  if FFooterRow.FCellLinesColor <> clNone then
    LineColor:=FFooterRow.FCellLinesColor
  else
  begin
    case FInternalDrawingStyle of
      gdsThemed:
        LineColor:=$F0F0F0;
      gdsGradient:
        LineColor:=$F0F0F0;
    else
      LineColor:=clSilver;
    end;
    LineColor:=GetStyleColor_GridCellBorder(LineColor);
  end;

  if FFooterRow.FColor = clNone then
    BkColor:=GetStyleBkColor_Grid(Color)
  else
    BkColor:=FFooterRow.FColor;

  // draw bk and lines for all columns
  for I:=0 to ColCount - 1 do
  begin
    // get column rect
    rc:=FFooterRow.FooterCellRect(I);
    // skip invisible columns
    if (rc.Right - rc.Left) < 3 then
      Continue;
    // draw background
    if I = FixCol then
    begin
      InflateRect(rc, -1, -1);
      DrawCellBackground(rc, FixedColor, [gdFixed], 0, 1);
      InflateRect(rc, 1, 1);
    end
    else
    begin
      Canvas.Brush.Color:=BkColor;
      Canvas.FillRect(rc);
    end;
    // draw cell frame
    if FFooterRow.FShowCellLines then
    begin
      BiDi_FrameRect(Canvas, rc, LineColor);
    end;
  end;

  // draw content for all columns
  LastTextRight:=0;
  for I:=0 to ColCount - 1 do
  begin
    // get column rect
    rc:=FFooterRow.FooterCellRect(I);
    // skip invisible columns
    if (rc.Right - rc.Left) < 3 then
      Continue;
    if I = FixCol then
      LastTextRight:=rc.Right + 3;

    Canvas.Font.Assign(FFooterRow.Font);
    // get cell props and content
    sCellText:='';
    if I = FixCol then
    begin
      C:=nil;
      Fld:=nil;
      if FFooterRow.ParentFont then
        Canvas.Font.Assign(FFixedColText.FFont);
      Align:=FFixedColText.FTextAlign;
      sCellText:=FFooterRow.FixedColText;
    end
    else
    begin
      C:=Columns[RawToDataColumn(I)];
      Fld:=C.Field;
      if FFooterRow.ParentFont then
        Canvas.Font.Assign(C.Font);
      Align:=C.Alignment;
      sCellText:=FFooterRow.GetFooterText(Fld);
    end;
    if Canvas.Font.Color = clWindowText then
      Canvas.Font.Color:=GetStyleFontColor_GridText(Canvas.Font.Color);
    if FFooterRow.FBoldText then
      Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
    AlignDef:=tadDefault;
    if Assigned(Fld) and FooterRow_GetSettings(Fld.FieldName, CMask, AlignDef, FontStylesDef, FColor, FSize, Canvas.Font) then
    begin
      UpdateAlignmentFromAlignmentDefault(Align, AlignDef);
    end;

    SetFrm2(Align);
    // call event
    DefaultDrawing:=true;
    if Assigned(FOnDrawFooterCell) then
      FOnDrawFooterCell(self, rc, I, Fld, sCellText, DefaultDrawing);

    if sCellText = '' then
      Continue;
    if DefaultDrawing then
    begin
      InflateRect(rc, -3, -1);
      // check text height
      if Canvas.TextHeight('Wg') * 2 > rc.Bottom - rc.Top then
      begin
        // use single line footer
        // check text width, use another right or left cells if needed
        W:=GetLineWidth(sCellText);
        if Assigned(C) and (W > rc.Right - rc.Left) then
        begin
          if Align = taRightJustify then
          begin
            // use left empty cell
            if rc.Left - LastTextRight > 3 then
              rc.Left:=LastTextRight;
            if W < rc.Right - rc.Left then
              rc.Left:=rc.Right - W - 2;
          end
          else
          begin
            // use right empty cell
            LastTextRight:=rc.Right;
            for j:=I + 1 to ColCount - 1 do
            begin
              if FFooterRow.IsFooterDefined(Columns[RawToDataColumn(j)].Field) then
                Break
              else
                LastTextRight:=Max(LastTextRight, FFooterRow.FooterCellRect(j).Right - 3);
            end;
            rc.Right:=Min(LastTextRight, RF.Right);
            if W < rc.Right - rc.Left then
              rc.Right:=rc.Left + W + 2;
          end;
          // fill text area to remove grid lines
          if FFooterRow.FShowCellLines then
          begin
            Canvas.Brush.Color:=BkColor;
            Canvas.FillRect(rc);
          end
        end;
        // draw cell text
        Canvas.Brush.Style:=bsClear;
        DrawTextAligned(rc, sCellText, DT_SINGLELINE + DT_VCENTER + DT_END_ELLIPSIS + Frm2);
      end
      else
      begin
        // more lines = word wrap
        // draw cell text
        sCellText:=ReplaceStr('|', cStrEOL, sCellText);
        Canvas.Brush.Style:=bsClear;
        DrawTextAligned(rc, sCellText, DT_WORDBREAK + Frm2, taVerticalCenter, '###');
      end;
    end;
    LastTextRight:=rc.Right + 3;
  end;

  // draw simple text
  if sSimpleText <> '' then
  begin
    // skip fixed col
    if FixCol = 0 then
      if UseRightToLeftAlignment then
        Dec(RF.Right, ColWidths[0])
      else
        Inc(RF.Left, ColWidths[0]);
    // check right of last cell
    rc:=FFooterRow.FooterCellRect(ColCount - 1);
    if (rc.Left > 0) then
      if UseRightToLeftAlignment then
        RF.Left:=Max(RF.Left, rc.Left)
      else
        RF.Right:=Min(RF.Right, rc.Right);
    InflateRect(RF, -2, -1);
    Canvas.Font.Assign(FFooterRow.Font);
    if Canvas.Font.Color = clWindowText then
      Canvas.Font.Color:=GetStyleFontColor_GridText(Canvas.Font.Color);
    if FFooterRow.FBoldText then
      Canvas.Font.Style:=Canvas.Font.Style + [fsBold];
    Canvas.Brush.Style:=bsClear;
    SetFrm2(FFooterRow.FSimpleTextAlign);
    DrawTextAligned(RF, sSimpleText, DT_SINGLELINE + DT_VCENTER + DT_END_ELLIPSIS + Frm2);
  end;
end;

// ==============================================================================
// FooterGrid Event group
function TrDBGrid.GetFieldNameColumnIndex(const FieldName: string): integer;
var
  I: integer;
begin
  Result:=-1;
  for I:=0 to Columns.Count - 1 do
    if AnsiSameText(Columns[I].FieldName, FieldName) then
    begin
      Result:=I;
      Break;
    end;
end;

function TrDBGrid.GetLeftCol: integer;
begin
  Result:=LeftCol;
end;

procedure TrDBGrid.SetLeftCol(ALeftCol: integer);
begin
  LeftCol:=ALeftCol;
end;

procedure TrDBGrid.TopLeftChanged;
begin
  inherited TopLeftChanged;
  FilterComboRefreshBounds;
  DrawFooterRow;
  FTitleLineText.PosRight:=0;
  if Assigned(FOnTopLeftChanged) then
    FOnTopLeftChanged(self);
  // move footer grid
  if Assigned(FFooterGrid) then
    FFooterGrid.SetLeftCol(LeftCol);
  // notify master grid
  if Assigned(NotifyTopLeftChanged) then
    NotifyTopLeftChanged(self);
end;

procedure TrDBGrid.SetFooterDBGrid(FooterDBGrid: TrDBGrid; const FooterFieldPrefix: string; ShowHSBarMaster, ShowHSBarFooter: boolean);
begin
  FFooterGrid:=FooterDBGrid;
  FFooterGridFieldPrefix:=FooterFieldPrefix;
  FooterDBGrid.NotifyTopLeftChanged:=FooterScrolled;
  if ShowHSBarMaster then
    ScrollBars:=ssHorizontal
  else
    ScrollBars:=ssNone;
  if ShowHSBarFooter then
    FooterDBGrid.ScrollBars:=ssHorizontal
  else
    FooterDBGrid.ScrollBars:=ssNone;
  FFooterGrid.Options:=FFooterGrid.Options - [dgColumnResize];
  UpdateFooterGrid;
end;

procedure TrDBGrid.FooterScrolled(Sender: TObject);
begin
  if (not FooterGridUpdating) and Assigned(FFooterGrid) then
    LeftCol:=FFooterGrid.GetLeftCol;
end;

procedure TrDBGrid.UpdateFooterGrid;
var
  I, j, found: integer;
begin
  if not Assigned(FFooterGrid) then
    Exit;
  // move columns in footer grid according master grid
  FooterGridUpdating:=true;
  try
    for I:=0 to Columns.Count - 1 do
    begin
      // try to find same column or empty column
      found:=-1;
      j:=I;
      while j < FFooterGrid.Columns.Count do
      begin
        if AnsiSameText(FFooterGridFieldPrefix + Columns[I].FieldName, FFooterGrid.Columns[j].FieldName) then
        begin
          found:=j;
          Break;
        end;
        if FFooterGrid.Columns[j].FieldName = '' then
          found:=j;
        Inc(j);
      end;
      if found = -1 then
      begin
        // add empty column
        FFooterGrid.Columns.Add;
        found:=FFooterGrid.Columns.Count - 1;
      end;
      // move column and set props
      if I <> found then
        FFooterGrid.Columns[found].Index:=I;
      FFooterGrid.Columns[I].Visible:=Columns[I].Visible;
      FFooterGrid.Columns[I].Width:=Columns[I].Width;
    end;
    FFooterGrid.SetLeftCol(LeftCol);
  finally
    FooterGridUpdating:=false;
  end;
end;

// ==============================================================================
// COPY funct group
procedure TrDBGrid.CBProc_IntCopyData(Sender: TObject; DataSet: TDataSet; RecNum: integer);
var
  I: integer;
  sr, sc: string;
begin
  sr:='';
  for I:=0 to High(CBScan_IntArr) do
  begin
    if RecNum = -1 then
      sc:=Columns[CBScan_IntArr[I]].Title.Caption
    else
      sc:=GetCellValue(Columns[CBScan_IntArr[I]].Field);
    if (Pos(#10, sc) > 0) or (Pos('"', sc) > 0) or (Pos(#9, sc) > 0) then
    begin
      sc:=ReplaceStr(cStrEOL, #10, sc, true, true);
      sc:=ReplaceStr('"', '""', sc, true, true);
      sc:='"' + sc + '"';
    end;
    if I > 0 then
      sr:=sr + #9;
    sr:=sr + sc;
  end;
  CBScan_Str:=CBScan_Str + sr + cStrEOL;
end;

function TrDBGrid.IntCopyData(ToClipboard, AllRec, AllFields, IncTitles, IncHidden: boolean; Fields: TStrings): string;
var
  I, C: integer;
begin
  // fill array of columns to copy
  C:=0;
  SetLength(CBScan_IntArr, 0);
  for I:=0 to Columns.Count - 1 do
  begin
    if (AllFields and (IncHidden or Columns[I].Visible)) or (Assigned(Fields) and (Fields.IndexOf(Columns[I].FieldName) >= 0)) then
    begin
      SetLength(CBScan_IntArr, C + 1);
      CBScan_IntArr[C]:=I;
      Inc(C);
    end;
  end;
  // read data
  CBScan_Str:='';
  if IncTitles then
    CBProc_IntCopyData(self, nil, -1);
  ScanAllRecords(CBProc_IntCopyData, nil, grrCopyData, AllRec);
  Finalize(CBScan_IntArr);
  // use result
  Result:=CBScan_Str;
  if ToClipboard then
    Clipboard.AsText:=CBScan_Str;
end;

function TrDBGrid.GetTitleText(Column: TColumn): string;
var
  FldText: string;
begin
  if FOptionsEx2.FGroups.FActive and GetGroupMasterTitleContent(Column.Field, FldText) then
    Result:=FldText
  else
    Result:=Column.Title.Caption;
end;

function TrDBGrid.GetCellValue(Fld: TField; UseDisplayText: boolean = true): string;
var
  FldText: string;
begin
  if Assigned(Fld) then
  begin
    if FOptionsEx2.FGroups.FActive and GetGroupMasterCellContent(Fld, FldText) then
      Result:=FldText
    else
      if rdbtool.IsGraphicField(Fld) then
        Result:=Fld.DisplayText
      else
        if rdbtool.IsMemoField(Fld) then
          Result:=Fld.AsString
        else
          if UseDisplayText then
            Result:=Fld.DisplayText
          else
            Result:=Fld.AsString;
  end
  else
    Result:='';
end;

function TrDBGrid.CopyCellData(CopyToClipboard: boolean = true): string;
begin
  try
    Result:=GetCellValue(SelectedField);
    if CopyToClipboard then
      Clipboard.AsText:=Result;
  except
  end;
end;

procedure TrDBGrid.CopyFieldsData(FieldNames: TStrings; IncludeTitles: boolean = false);
begin
  IntCopyData(true, true, false, IncludeTitles, true, FieldNames);
end;

procedure TrDBGrid.CopyActiveFieldData(IncludeTitles: boolean = false);
var
  Fld: TStrings;
begin
  Fld:=TStringList.Create;
  try
    if Assigned(SelectedField) then
      Fld.Add(SelectedField.FieldName);
    IntCopyData(true, true, false, IncludeTitles, true, Fld);
  finally
    FreeAndNil(Fld);
  end;
end;

procedure TrDBGrid.CopyActiveRecordData(IncludeTitles: boolean = false; IncludeHidden: boolean = false);
begin
  IntCopyData(true, false, true, IncludeTitles, IncludeHidden, nil);
end;

procedure TrDBGrid.CopyAllData(IncludeTitles: boolean = false; IncludeHidden: boolean = false);
begin
  IntCopyData(true, true, true, IncludeTitles, IncludeHidden, nil);
end;

procedure TrDBGrid.IntPasteData(FromCol: integer; AddAsNewRecs: boolean);
var
  C, pr, pc: integer;
  SelRangeText, sr, sc, srrest, screst, err: string;
  multiplerec: boolean;
begin
  try
    SelRangeText:=Clipboard.AsText;
    if copy(SelRangeText, Length(SelRangeText) - 1, 2) = cStrEOL then
      SelRangeText:=copy(SelRangeText, 1, Length(SelRangeText) - 2);
  except
    Exit;
  end;
  if not IsActive or (FromCol > Columns.Count - 1) or (SelRangeText = '') then
    Exit;

  err:='';
  srrest:=SelRangeText;
  multiplerec:=Pos(cStrEOL, srrest) > 0;
  repeat
    if AddAsNewRecs then
      DataLink.DataSet.Append
    else
      DataLink.DataSet.Edit;
    if not IsEditing(DataLink.DataSet) then
      Exit;

    // extract row text
    pr:=Pos(cStrEOL, srrest);
    if pr > 0 then
    begin
      sr:=copy(srrest, 1, pr - 1);
      srrest:=copy(srrest, pr + 2, Length(srrest));
    end
    else
    begin
      sr:=srrest;
      srrest:='';
    end;
    screst:=sr;
    C:=FromCol;
    repeat
      // extract cell text
      pc:=Pos(#9, screst);
      if pc > 0 then
      begin
        sc:=copy(screst, 1, pc - 1);
        screst:=copy(screst, pc + 1, Length(screst));
      end
      else
      begin
        sc:=screst;
        screst:='';
      end;
      // proc special chars in cell text
      if (copy(sc, 1, 1) = '"') and (copy(sc, Length(sc), 1) = '"') then
      begin
        sc:=copy(sc, 2, Length(sc) - 2);
        sc:=ReplaceStr('""', '"', sc, true, false);
        sc:=ReplaceStr(#10, cStrEOL, sc, true, false);
      end;

      // update column field
      try
        if Columns[C].Field is TNumericField then
        begin
          // remove spaces and non-braking spaces
          sc:=rtool.RemoveThousandSeparator(sc);
          sc:=rtool.CleanNumericString(sc);
        end;
        if Columns[C].Field is TDateTimeField then
        begin
          // remove spaces and non-braking spaces
          sc:=rtool.CleanDateTimeString(sc);
        end;
        Columns[C].Field.AsString:=sc;
      except
        on E: Exception do
          err:=err + E.Message + '; ';
      end;

      // find next visible column
      repeat
        Inc(C);
      until (C > Columns.Count - 1) or Columns[C].Visible;
      // skip rest data if last column
      if C > Columns.Count - 1 then
        Break;
    until screst = '';
    try
      if multiplerec then
        DataLink.DataSet.Post;
    except
      on E: Exception do
      begin
        if DlgQ2(sSaveRecordErr + #13 + sContinueQ, sErrorDescription + #13 + E.Message) = mrYes then
          DataLink.DataSet.Cancel
        else
          Break;
      end;
    end;
    if multiplerec and not AddAsNewRecs and (srrest <> '') then
      DataLink.DataSet.Next;
  until srrest = '';
end;

procedure TrDBGrid.PasteData(AddAsNewRecs: boolean);
begin
  IntPasteData(SelectedIndex, AddAsNewRecs);
end;

function TrDBGrid.FillCellData(FillUp: boolean = false; SpecificFieldNames: string = ''): boolean;
var
  I, SelCol: integer;
  F: string;
  FldArr: array of TField;
  ValArr: array of variant;
begin
  Result:=false;
  // store selected column
  SelCol:=SelectedIndex;
  // post if editing
  if DataLink.Editing then
    DataLink.DataSet.Post;
  // store fields to copy
  SetLength(FldArr, Columns.Count);
  SetLength(ValArr, Columns.Count);
  for I:=0 to High(FldArr) do
    FldArr[I]:=nil;
  if SpecificFieldNames = '' then
    FldArr[0]:=SelectedField
  else
  begin
    I:=0;
    repeat
      F:=Trim(rtool.RemoveStrPart(SpecificFieldNames, ','));
      FldArr[I]:=DataLink.DataSet.FindField(F);
      if FldArr[I] <> nil then
        Inc(I);
    until (I > High(FldArr)) or (SpecificFieldNames = '');
  end;
  // store values
  for I:=0 to High(FldArr) do
    if FldArr[I] <> nil then
      ValArr[I]:=FldArr[I].AsVariant;
  // move record
  if FillUp then
    DataLink.DataSet.Prior
  else
    DataLink.DataSet.Next;
  if DataLink.DataSet.Eof or DataLink.DataSet.Bof then
  begin
    if FillUp then
      DataLink.DataSet.Insert
    else
      DataLink.DataSet.Append;
  end
  else
    DataLink.DataSet.Edit;
  // restore selected column
  SelectedIndex:=SelCol;
  if DataLink.Editing then
  begin
    // paste copied values
    for I:=0 to High(FldArr) do
      if FldArr[I] <> nil then
        FldArr[I].AsVariant:=ValArr[I];
    Result:=true;
  end;
end;

// ==============================================================================
// TCellParamsCondList func group
procedure TCellParamsCondList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  if (Action = lnDeleted) then
    TCellParamsCond(Ptr).Free;
  inherited Notify(Ptr, Action);
end;

procedure TCellParamsCondList.LinkGrid(Grid: TrDBGrid);
begin
  Gr:=Grid;
  DS:=nil;
end;

procedure TCellParamsCondList.LinkDataset(DataSet: TDataSet);
begin
  Gr:=nil;
  DS:=DataSet;
end;

procedure TCellParamsCondList.ResetCond;
var
  I: integer;
  cp: TCellParamsCond;
begin
  for I:=0 to Count - 1 do
  begin
    cp:=Items[I];
    cp.LastRecNo:=-1;
  end;
end;

procedure TCellParamsCondList.ProcIniContent(IniContent: TStrings; GridKey: string; Load: boolean);
var
  tempPSF: TPropSaveFile;
  tempGPS: TrDBGridsPropSave;
begin
  if GridKey = '' then
    GridKey:=Gr.Owner.Name + '.' + Gr.Name;
  tempPSF:=TPropSaveFile.Create(Gr);
  try
    tempGPS:=TrDBGridsPropSave.Create(Gr);
    try
      tempPSF.SetStrings(IniContent);
      if Load then
      begin
        tempGPS.ReadCellParamsCondList(tempPSF, GridKey, self);
        if Count > 0 then
          Gr.FOptionsEx2.FCellCondition.FActive:=true;
      end
      else
      begin
        tempGPS.SaveCellParamsCondList(tempPSF, GridKey, self);
        IniContent.Clear;
        tempPSF.GetStrings(IniContent);
      end;
    finally
      tempGPS.Free;
    end;
    tempPSF.DisableAll:=true;
    Gr.Invalidate;
  finally
    tempPSF.Free;
  end;
end;

procedure TCellParamsCondList.LoadFromIniContent(IniContent: TStrings; GridKey: string);
begin
  ProcIniContent(IniContent, GridKey, true);
end;

procedure TCellParamsCondList.StoreToIniContent(IniContent: TStrings; GridKey: string);
begin
  ProcIniContent(IniContent, GridKey, false);
end;

procedure TCellParamsCondList.GetFieldCond(Field: TField; AFont: TFont; var Background: TColor; var DataImageDef: TDataImageDef;
  ActiveRow, Highlight, RefreshCond: boolean);
var
  I: integer;
  isEna: boolean;
  cp: TCellParamsCond;
  FldComp, FldVal: TField;
  PrevAct: TCellParamsCondActions;
  eq, eqprev: boolean;
  bval: boolean;
  sval, s: string;
  fval: extended;
  vr: TValueRelationship;

  function GetSummaryRes(calc, tag: string; F: TField; Val: double): boolean;
  begin
    Result:=SameText(tag, calc);
    if Result then
    begin
      fval:=Val;
      sval:=FloatToStr(Val);
    end;
  end;

  function CheckSummaryCalc(sumstr: string): boolean;
  var
    FldName, t: string;
    SumRes: TSummaryResults;
  begin
    Result:=false;
    // extract summary field
    FldName:=rtool.FindStrFromTo('(', ')', sumstr, false, true);
    if FldName = '' then
      Exit;
    // get summary result
    SumRes:=Gr.GetSummaryResults(DS.FieldByName(FldName), false);
    // recalculate
    if not Assigned(SumRes.Field) then
      SumRes:=Gr.GetSummaryResults(DS.FieldByName(FldName), true);
    // extract summary type
    t:=rtool.FindStrPart(sumstr, '(', 1);
    if Assigned(SumRes.Field) then
      Result:=(GetSummaryRes(t, cSummaryTag_Sum, FldVal, SumRes.Sum) or GetSummaryRes(t, cSummaryTag_Min, FldVal, SumRes.Min) or GetSummaryRes
          (t, cSummaryTag_Max, FldVal, SumRes.Max) or GetSummaryRes(t, cSummaryTag_AVG, FldVal, SumRes.Avg) or GetSummaryRes
          (t, cSummaryTag_CountVal, FldVal, SumRes.CountValues) or GetSummaryRes(t, cSummaryTag_CountNull, FldVal, SumRes.CountEmpty) or GetSummaryRes
          (t, cSummaryTag_CountAll, FldVal, SumRes.CountValues + SumRes.CountEmpty));
  end;

begin
  PrevAct:=cpaSet;
  eqprev:=true;
  isEna:=true;
  if Assigned(Gr) then
  begin
    if Gr.IsDatasetLinked then
      DS:=Gr.DataLink.DataSet
    else
      DS:=nil;
    isEna:=Gr.FOptionsEx2.FCellCondition.FActive;
  end;
  isEna:=isEna and Assigned(DS) and DS.Active;
  if isEna then
  begin
    // set cell font and Bk color according user setting in CellParamsCondList
    for I:=0 to Count - 1 do
    begin
      cp:=Items[I];
      if cp.Deactive then
        Continue;
      try
        FldComp:=DS.FieldByName(cp.FieldName);
        eq:=cp.CondOperator = cpoAllways;
        if not eq then
          if (not RefreshCond) and Assigned(Gr) and (Gr.DataLink.ActiveRecord = cp.LastRecNo) then
            eq:=cp.LastResult
          else
          begin
            // save compared value to fval, sval and bval
            fval:=0;
            bval:=false;
            if copy(cp.Value, 1, 1) = '=' then
            begin
              // compare with current date or field value
              s:=copy(cp.Value, 2, 255);
              if SameText(s, rActualDateAbr) then
              begin
                fval:=Date;
                sval:=DateToStr(fval);
              end
              else
                if not CheckSummaryCalc(s) then
                begin
                  FldVal:=DS.FieldByName(s);
                  try
                    if IsDateTimeField(FldVal) then
                      fval:=FldVal.AsDateTime
                    else
                      if (FldVal is TBooleanField) then
                        bval:=FldVal.AsBoolean
                      else
                        if (FldVal is TNumericField) then
                          fval:=FldVal.AsFloat;
                  except
                  end;
                  sval:=FldVal.AsString;
                end;
            end
            else
            begin
              try
                if IsDateTimeField(FldComp) then
                  fval:=StrToDateDef(cp.Value, 0)
                else
                  fval:=StrToFloatDef(cp.Value, 0);
              except
              end;
              sval:=cp.Value;
              if FldComp is TBooleanField then
                bval:=SameText(sval, '1') or SameText(sval, copy(sYes, 1, Length(sval)));
            end;
            if FldComp.IsNull or (sval = '') then
            begin
              if (cp.CondOperator = cpoEqual) then
                eq:=(sval = '') and FldComp.IsNull
              else
                if (cp.CondOperator = cpoNotEqual) then
                  eq:=(sval <> '') or not FldComp.IsNull
                else
                  eq:=false;
            end
            else
              // compare field with fval or sval or bval
              if (FldComp is TNumericField) or IsDateTimeField(FldComp) then
              begin
                vr:=CompareValue(FldComp.AsFloat, fval, 1E-15);
                case cp.CondOperator of
                  cpoEqual:
                    eq:=vr = EqualsValue;
                  cpoLess:
                    eq:=vr < EqualsValue;
                  cpoLessOrEq:
                    eq:=vr <= EqualsValue;
                  cpoGreater:
                    eq:=vr > EqualsValue;
                  cpoGreaterOrEq:
                    eq:=vr >= EqualsValue;
                  cpoNotEqual:
                    eq:=vr <> EqualsValue;
                end;
              end
              else
                if FldComp is TBooleanField then
                begin
                  if cp.CondOperator = cpoEqual then
                    eq:=FldComp.AsBoolean = bval
                  else
                    eq:=FldComp.AsBoolean <> bval;
                end
                else
                begin
                  case cp.CondOperator of
                    cpoEqual:
                      eq:=SameText(FldComp.AsString, sval);
                    cpoLess:
                      eq:=FldComp.AsString < sval;
                    cpoLessOrEq:
                      eq:=FldComp.AsString <= sval;
                    cpoGreater:
                      eq:=FldComp.AsString > sval;
                    cpoGreaterOrEq:
                      eq:=FldComp.AsString >= sval;
                    cpoNotEqual:
                      eq:=not SameText(FldComp.AsString, sval);
                    cpoStart:
                      eq:=AnsiStartsText(sval, FldComp.AsString);
                    cpoNotStart:
                      eq:=not AnsiStartsText(sval, FldComp.AsString);
                    cpoContain:
                      eq:=AnsiContainsText(FldComp.AsString, sval);
                    cpoNotContain:
                      eq:=not AnsiContainsText(FldComp.AsString, sval);
                  end;
                end;
            if Assigned(Gr) then
              cp.LastRecNo:=Gr.DataLink.ActiveRecord
            else
              cp.LastRecNo:=-1;
            cp.LastResult:=eq;
          end;

        // use result of comparison
        case PrevAct of
          cpaAND:
            eq:=eqprev and eq;
          cpaOR:
            eq:=eqprev or eq;
        end;
        PrevAct:=cp.Action;
        eqprev:=eq;
        if (cp.Action = cpaSet) and eq and (cp.WholeRow or (FldComp = Field)) then
        begin
          if not Highlight then
          begin
            if cp.FontColor <> clNone then
              AFont.Color:=cp.FontColor;
            if cp.BkColor <> clNone then
              Background:=cp.BkColor;
          end;
          if cp.Bold then
            AFont.Style:=AFont.Style + [fsBold];
          DataImageDef:=cp.ImageDef;
        end;
        if cp.StopIfTrue and (cp.Action = cpaSet) and eq then
          Break;
      except
        on E: Exception do
        begin
          // deactive errored condition and save error description
          cp.Deactive:=true;
          cp.ErrorDesc:=E.Message;
          ShowErr(Format(sCellParamsCondErr, [cp.FieldName]), E);
        end;
      end;
    end;
  end;
end;

// ==============================================================================
// TrDBGridCustomSorter func group
constructor TrDBGridCustomSorter.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  SortFieldSeparator:='; ';
  SortStringSeparator:=', ';
  DescendingString:='DESC';
  FGoToFirstAfterSort:=false;
  FGoToFirstAfterRefresh:=false;
end;

procedure TrDBGridCustomSorter.BracketFieldName(var FieldName: string);
begin
  if Pos(' ', FieldName) > 0 then
    FieldName:='[' + FieldName + ']';
end;

function TrDBGridCustomSorter.GetIndexNameHash(const IndexName: string; MaxLen: integer = 30): string;
var
  MDigest: IMD5;
begin
  // indexname is too long - use hash from fields names
  if Length(IndexName) > MaxLen then
  begin
    MDigest:=GetMD5;
    MDigest.Update(IndexName);
    Result:=copy(MDigest.AsString, 1, MaxLen);
  end
  else
    Result:=IndexName;
end;

function TrDBGridCustomSorter.SetSort_CreateIndex(DataSet: TDataSet; const SortString: string; const SortFields: string; const DescendingFields: string)
  : boolean;
begin
  // must be override
  Result:=false;
end;

function TrDBGridCustomSorter.DoSort(Sender: TrDBGrid; DataSet: TDataSet; SortFieldNames: TSortFieldNames; SortDirections: TSortDirections): boolean;
var
  B: TBookmark;
  F: TField;
  oldCursor: TCursor;
  aFieldName, SortString, SortFields, DescFields: string;
  I: integer;
begin
  Result:=Assigned(DataSet) and DataSet.Active and (not IsEditing(DataSet));
  if not Result then
    Exit;
  oldCursor:=SQLCursor;
  try
    try
      SortString:='';
      SortFields:='';
      DescFields:='';
      for I:=0 to High(SortFieldNames) do
      begin
        aFieldName:=SortFieldNames[I];
        F:=DataSet.FindField(aFieldName);
        Result:=Assigned(F) and IsSimpleField(F) and not F.Calculated;
        if not Result then
          Exit;
        if (F.FieldKind = fkLookup) then
          aFieldName:=F.KeyFields;
        BracketFieldName(aFieldName);

        if SortString <> '' then
          SortString:=SortString + SortStringSeparator;
        SortString:=SortString + aFieldName;
        if SortFields <> '' then
          SortFields:=SortFields + SortFieldSeparator;
        SortFields:=SortFields + aFieldName;
        if (I <= High(SortDirections)) and SortDirections[I] then
        begin
          SortString:=SortString + ' ' + DescendingString;
          if DescFields <> '' then
            DescFields:=DescFields + SortFieldSeparator;
          DescFields:=DescFields + aFieldName;
        end;
      end;
      B:=DataSet.GetBookmark;
      try
        Result:=SetSort_CreateIndex(DataSet, SortString, SortFields, DescFields);
        if FGoToFirstAfterSort then
          DataSet.First
        else
          DataSet.GotoBookmark(B);
      finally
        DataSet.FreeBookmark(B);
      end;
    except
      on E: Exception do
      begin
        Result:=false;
        ShowErr('', E);
      end;
    end;
  finally
    ChCursor(oldCursor);
  end;
end;

procedure TrDBGridCustomSorter.DatasetRequery(DataSet: TDataSet);
begin
  DataSet.Close;
  DataSet.Open;
end;

procedure TrDBGridCustomSorter.DoRefresh(Sender: TrDBGrid; DataSet: TDataSet);
var
  B: TBookmark;
  oldCursor: TCursor;
begin
  if Assigned(DataSet) and DataSet.Active then
  begin
    oldCursor:=SQLCursor;
    try
      B:=DataSet.GetBookmark;
      try
        DatasetRequery(DataSet);
        if DataSet.BookmarkValid(B) then
          try
            if FGoToFirstAfterRefresh then
              DataSet.First
            else
              DataSet.GotoBookmark(B);
          except
            DataSet.Last;
          end;
      finally
        DataSet.FreeBookmark(B);
      end;
    finally
      ChCursor(oldCursor);
    end;
  end;
end;

end.
