unit rImprovedComps;

{<
  rImprovedComps unit contains set of enhanced components,
  usually introducing many new functions and fixing some bugs (eg. high DPI):
  TrGroupBox, TrImageEx, TrMemoEx, TrComboBoxEx, TrButton,
  TrStringsHolder, TrEditDate, TrEditNum, TrImageListEx,
  TrTabControlEx + TrPageControlEx.

  TrListBox, TrColorBox, TrColorListBox, TrCheckListBox - fix for high DPI

  TrGroupBox - provide new type of groupbox with following functions:
  - headline caption with definable colors and possition;
  - 4 types of frames with definable colors;
  - min max buttons with automatical and animated functions;
  - gradient background or background image;
  - support dropping of files to box;
  - support automatic resizing of children components

  TrImageEx - enhanced TImage with following functions:
  - bound label with definable possition;
  - drawing support for basic shapes (rect, elipse, halfelipse)

  TrMemoEx - enhanced Memo component which provides
  highlighting of words in text according list of words or
  highlighting of text between start and stop tag

  TrComboBoxEx - enhanced ComboBox component which provides
  automatic setting of dropdown listbox width according length of items text,
  multicolumn dropdown, and internal timer for delayed event after editing

  TrButton - enhaced TButton with following functions:
  - new property ImagesDisabled which is used for disable state of button
  instead of DisableImageIndex, so the same image list as
  for toolbar and menus can be used

  TrStringsHolder - non visual component containing Strings object
  for editing in design time and storing in application

  TrEditDate -  enhanced TEdit component which provides functionality
  for editing of date value

  TrEditNum -  enhanced TEdit component which provides functionality
  for editing of integer or float number

  TrImageListEx - enhanced ImageList component which provides
  storing of images in external compressed resource file in design and runtime
  and can block storing images in DFM file for descendant modules, so list can be used in ihnerited forms

  TrTabControlEx + TrPageControlEx - enhanced TabControl and PageControl
  - provides colored tabs, hottrack fonts, color and border;
  - close button in tabs;
  - function for reading active tab caption and changing of active tab
  - supports Themes, use rStyleHooks unit to add Styles support for XE2 and higher


  Updated version compatible with Delphi2010+ and 64bit platform.
  Splash screen in demo version.
  @author(Rosinsky Software - <http://www.rosinsky.cz/delphi>)
  @created(Oct 2010)
  @lastmod(Jan 2021)
}

interface

{$INCLUDE _rosi.inc}

uses Messages, Windows, SysUtils, ComCtrls, Controls, Forms, Classes, Dialogs,
  Menus, Graphics, StdCtrls, ExtCtrls, Buttons, Math, StrUtils, Clipbrd, RichEdit,
{$IFDEF DXE2P} UITypes, Types, {$ENDIF} ImgList, ShellApi, ShlObj, GraphUtil,
  StdActns, CommCtrl, Consts, rHTMLLabel, UxTheme, CheckLst, DateUtils,
{$IFDEF DX101P} Rtti, {$ENDIF}
  RosiCompGlobalCfg;

const
  cListBoxItemHeightDefault = 16;
  cComboBoxItemHeightDefault = 16;
  cDBComboBoxItemHeightDefault = 15;

type
  { list of min max button types, used by TrGroupBox }
  TMinMaxBtnType = (mmbNone, mmbRound, mmbRect, mmbPicture);
  { list of background picture types, used by TrGroupBox }
  TBackgroundPictureType = (bkpNone, bkpCenter, bkpStretch, bkpTile);
  { list of caption possition types, used by TrGroupBox }
  TCaptionPosition = (cpTop, cpBottom, cpLeft, cpRight);
  { list of actions use by procedure SetAllChildren of TrGroupBox }
  TSetChildrenValueTypes = (chvSetVisibility, chvSetEnability, chvSetColor, chvSetDataSource);

type
  { enhanced TListBox with high DPI bug fix }
  TrListBox = class(TListBox)
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  end;

  { enhanced TColorBox with high DPI bug fix }
  TrColorBox = class(TColorBox)
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  end;

  { enhanced TColorListBox with high DPI bug fix }
  TrColorListBox = class(TColorListBox)
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  end;

  { @exclude }
  TCheckListBoxHelper = class helper for TCheckListBox
  public
    procedure SetCheckSize(size: integer);
  end;

  { enhanced TCheckListBox with high DPI bug fix }
  TrCheckListBox = class(TCheckListBox)
  private
    FCheckboxUpdateMode: TCheckboxUpdateMode;
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function CanModifyCheckItem(Index: integer): Boolean; virtual;
    procedure DblClick; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { indicates count of checked items }
    function CheckedCount: integer;
  published
    { define mode for checkbox update }
    property CheckboxUpdateMode: TCheckboxUpdateMode read FCheckboxUpdateMode write FCheckboxUpdateMode default chmCheckboxClick;
  end;

  { enhanced TStatusBar with high DPI bug fix }
  TrStatusBar = class(TStatusBar)
  protected
    FScaleRatio: double;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
  end;

  { enhanced ToolBar with with high DPI bug fix }
  TrToolBar = class(TToolBar)
  private
    FScaleAutoButtonSize: Boolean;
  protected
    FScaleRatio: double;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    procedure RefreshSize;
  published
    property ScaleAutoButtonSize: Boolean read FScaleAutoButtonSize write FScaleAutoButtonSize default false;
  end;

  // ==============================================================================
  // TrGroupBox
type
  { list of possible vertical alignment for child component used by TrGroupBox }
  TrGroupBoxVerticalAlignment = (gbvaNothing, gbvaColumn, gbvaColumnLeft, gbvaColumnRight, gbvaLeft, gbvaRight, gbvaExpand);
  { list of possible horizontal alignment for child component used by TrGroupBox }
  TrGroupBoxHorizontalAlignment = (gbhaNothing, gbhaExpand, gbhaBottom);

const
  { list of string abbreviation for possible vertical alignment for child component used by TrGroupBox }
  rGroupBoxVerticalAlignmentStr: array [TrGroupBoxVerticalAlignment] of string = ('0', 'C', 'CL', 'CR', 'L', 'R', 'E');
  { list of string abbreviation for possible horizontal alignment for child component used by TrGroupBox }
  rGroupBoxHorizontalAlignmentStr: array [TrGroupBoxHorizontalAlignment] of string = ('0', 'E', 'B');

type
  TrGroupBox = class;

  { event type for custom drawing used by TrGroupBox }
  TDrawControl = procedure(WinControl: TWinControl; const Rect: TRect) of object;
  { event type for dropping of files used by TrGroupBox }
  TDropFilesEvent = procedure(Sender: TObject; X, Y: integer) of object;

  { subcomponent for autosize functionality for the TrGroupBox inside parent TrGroupBox;
    parent AutoResizeChild property define columns and rows matrix,
    AutoResizePos properties define possition of child rGroupBox }
  TAutoResizePos = class(TPersistent)
  private
    FOwner: TrGroupBox;
    FColumnIndexFrom: integer;
    FColumnIndexTo: integer;
    FRowIndexFrom: integer;
    FRowIndexTo: integer;
    FLastColumn: Boolean;
    FLastRow: Boolean;
    procedure SetColumnIndexFrom(Value: integer);
    procedure SetColumnIndexTo(Value: integer);
    procedure SetRowIndexFrom(Value: integer);
    procedure SetRowIndexTo(Value: integer);
    procedure SetLastColumn(Value: Boolean);
    procedure SetLastRow(Value: Boolean);
    procedure UpdatePos;
  public
    constructor Create(AOwner: TrGroupBox); {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    { define first column of parent groupbox used by this groupbox }
    property ColumnIndexFrom: integer read FColumnIndexFrom write SetColumnIndexFrom default 0;
    { define last column of parent groupbox used by this groupbox }
    property ColumnIndexTo: integer read FColumnIndexTo write SetColumnIndexTo default 0;
    { define first row of parent groupbox used by this groupbox }
    property RowIndexFrom: integer read FRowIndexFrom write SetRowIndexFrom default 0;
    { define last row of parent groupbox used by this groupbox }
    property RowIndexTo: integer read FRowIndexTo write SetRowIndexTo default 0;
    { define whether groupbox is expanded to right border of parent }
    property LastColumn: Boolean read FLastColumn write SetLastColumn default false;
    { define whether groupbox is expanded to bottom border of parent }
    property LastRow: Boolean read FLastRow write SetLastRow default false;
  end;

  { subcomponent for autosize functionality of child components used by TrGroupBox;
    1/ TrGroupBox child controls are resized according AutoResizePos properties of child control;
    2/ other child controls are resized according AutoResizeChild.ChildCompsPos property }
  TAutoResizeChild = class(TPersistent)
  private
    FOwner: TrGroupBox;
    FAutoResizeActive: Boolean;
    FAutoResizeDelay: integer;
    FColSpacing: integer;
    FRowSpacing: integer;
    FBorderSpacing: integer;
    FBorderLeftAddSpacing: integer;
    FColWidthsPercent: TStrings;
    FRowHeightsPercent: TStrings;
    FChildCompsPos: TStrings;
    FAutoSetTabOrder: Boolean;
    procedure SetAutoResizeActive(Value: Boolean);
    procedure SetColSpacing(Value: integer);
    procedure SetRowSpacing(Value: integer);
    procedure SetBorderSpacing(Value: integer);
    procedure SetBorderLeftAddSpacing(Value: integer);
    procedure SetColWidthsPercent(Value: TStrings);
    procedure SetRowHeightsPercent(Value: TStrings);
    procedure SetChildCompsPos(Value: TStrings);
  public
    constructor Create(AOwner: TrGroupBox); {< @exclude }
    destructor Destroy; override; {< @exclude }
    { get percentage width of selected column }
    function GetColWidthPercent(ColIndex: integer): integer;
    { get percentage height of selected row }
    function GetRowHeightPercent(RowIndex: integer): integer;
    { refresh possition of child components }
    procedure UpdateChildComp;
  published
    { switch on/off autoresize function }
    property AutoResizeActive: Boolean read FAutoResizeActive write SetAutoResizeActive default false;
    { define delay for autoresize after Groupbox Resize }
    property AutoResizeDelay: integer read FAutoResizeDelay write FAutoResizeDelay default 100;
    { define whether child components tab order is set according order of components in ChildCompsPos }
    property AutoSetTabOrder: Boolean read FAutoSetTabOrder write FAutoSetTabOrder default true;
    { define space between columns }
    property ColSpacing: integer read FColSpacing write SetColSpacing default 4;
    { define space between rows }
    property RowSpacing: integer read FRowSpacing write SetRowSpacing default 4;
    { define space between border and first/last column and rows }
    property BorderSpacing: integer read FBorderSpacing write SetBorderSpacing default 8;
    { define additional space between border and first column }
    property BorderLeftAddSpacing: integer read FBorderLeftAddSpacing write SetBorderLeftAddSpacing default 0;
    { define list of percentage width for each column of child rGroupBox }
    property ColWidthsPercent: TStrings read FColWidthsPercent write SetColWidthsPercent;
    { define list of percentage height for each row of child rGroupBox }
    property RowHeightsPercent: TStrings read FRowHeightsPercent write SetRowHeightsPercent;
    { define setting for each child component, use design time dialog to set in IDE;
      each line has following content:
      CompName=VerAlignStyle;ColIndex;ColCount;ReduceWidth;HorAlignStyle;ReduceHeight
      VerAlignStyle:
      0 .. no auto alignment;
      C .. use column definition;
      CL .. keep original size and move to left border of column;
      CR .. keep original size and move to right border of column;
      L .. keep original size and move to left border of parent;
      R .. keep original size and move to right border of parent;
      E .. keep left possition and expand to right border of parent;
      HorAlignStyle:
      0 .. no auto alignment;
      E .. keep top possition and expand to bottom border of parent;
      B .. keep original size and move to bottom border of parent
      if AlignWithMargins of child control is set to true, calculated size is
      modified according Child.Margins values }
    property ChildCompsPos: TStrings read FChildCompsPos write SetChildCompsPos;
  end;

  { enhanced TGroupBox which provides new type of groupbox with following functions:
    - headline caption with definable colors and possition;
    - 4 types of frames with definable colors;
    - min max buttons with automatical and animated functions;
    - gradient background or background image;
    - support dropping of files to box;
    - support automatic resizing of children components }
  TrGroupBox = class(TGroupBox)
  private
    LastTickCount: cardinal;
    FFrameType: TFrameType;
    FFrameColor: TColor;
    FCaptionHeight: byte;
    FScaledCaptionHeight: byte;
    FCaptionColor: TColor;
    FCaptionFontColor: TColor;
    FCaptionDisableFontColor: TColor;
    FCaptionFont: TFont;
    FCaptionParentFont: Boolean;
    FCaptionAlignment: TAlignment;
    FCaptionPosition: TCaptionPosition;
    FGradientColorFrom: TColor;
    FGradientColorTo: TColor;
    FGradientType: TGradientType;
    FMinMaxBtnType: TMinMaxBtnType;
    FMinBtnPicture: TPicture;
    FMaxBtnPicture: TPicture;
    FBackgroundPicture: TPicture;
    FBackgroundPictureType: TBackgroundPictureType;
    FUseTabColor: Boolean;
    FMinimizeAnimate: Boolean;
    FShowSizeGrip: Boolean;
    FBtnRect: TRect;
    FMinimized: Boolean;
    FMinimizing: Boolean;
    FOnDrawControl: TDrawControl;
    FOnMinimizeChange: TNotifyEvent;
    FOriginSize: integer;
    FControlsVisibility: array of Boolean;
    FAutoResizePos: TAutoResizePos;
    FAutoResizeChild: TAutoResizeChild;
    FEnableDropFile: Boolean;
    FDropPoint: TPoint;
    FDropFilesList: TStringList;
    FOnDropFiles: TDropFilesEvent;
    FAutoMaximizeTime: cardinal;
    FAutoMinimizeTime: cardinal;
    FAutoEnableChildren: Boolean;
    FClientRectTop: integer;
    FAutoHint: Boolean;
    FChangeScale: TChangeScaleEvent;
    FOnRefreshAutoPosition: TNotifyEvent;
    IntTimer: TTimer;
    procedure SetFrameType(Value: TFrameType);
    procedure SetFrameColor(Value: TColor);
    procedure SetCaptionHeight(Value: byte);
    procedure SetCaptionColor(Value: TColor);
    procedure SetCaptionFontColor(Value: TColor);
    procedure SetCaptionDisableFontColor(Value: TColor);
    procedure SetCaptionFont(Value: TFont);
    procedure SetCaptionParentFont(Value: Boolean);
    procedure SetCaptionAlignment(Value: TAlignment);
    procedure SetCaptionPosition(Value: TCaptionPosition);
    procedure SetGradientColorFrom(Value: TColor);
    procedure SetGradientColorTo(Value: TColor);
    procedure SetGradientType(Value: TGradientType);
    procedure SetMinMaxBtnType(Value: TMinMaxBtnType);
    procedure SetMinBtnPicture(Value: TPicture);
    procedure SetMaxBtnPicture(Value: TPicture);
    procedure SetBackgroundPicture(Value: TPicture);
    procedure SetBackgroundPictureType(Value: TBackgroundPictureType);
    procedure SetUseTabColor(Value: Boolean);
    procedure SetMinimized(Value: Boolean);
    procedure SetShowSizeGrip(Value: Boolean);
    function IsFontStored: Boolean;
    procedure FontChanged(Sender: TObject);
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure DoMinimize;
    procedure DoRestore;
    procedure PictureChanged(Sender: TObject);
    function StoreCaptionHeight: Boolean;
    function StoreCaptionColor: Boolean;
    function StoreCaptionFontColor: Boolean;

    function GetDropFilesCount: integer;
    function GetDropFileNames(I: integer): string;
    function GetDropControl: TControl;
    procedure SetEnableDropFile(Value: Boolean);
    procedure SetClientRectTop(Value: integer);
    function ParentIsTab(C: TrGroupBox): Boolean;
    procedure IntTimerEvent(Sender: TObject);
  protected
    FScaleRatio: double;
    FRefPosImmediately: Boolean;
    procedure CreateWnd; override; {< @exclude }
    procedure Loaded; override; {< @exclude }
    procedure AdjustClientRect(var Rect: TRect); override; {< @exclude }
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure Paint; override; {< @exclude }
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override; {< @exclude }
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override; {< @exclude }
    procedure Resize; override; {< @exclude }
    procedure CMEnabledChanged(var Msg: TMessage); message CM_ENABLEDCHANGED; {< @exclude }
    { File drop message handler - used to record info about dropped files }
    procedure WMDropFiles(var Msg: TMessage); message WM_DROPFILES;
    { Triggers OnDropFiles event }
    procedure DropFiles; dynamic;
    procedure SetNewBounds(Bounds: TRect);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    function ExecuteAction(Action: TBasicAction): Boolean; override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { The number of files dropped }
    property DropFilesCount: integer read GetDropFilesCount;
    { The names of the dropped files }
    property DropFileNames[I: integer]: string read GetDropFileNames;
    { The mouse coordinates where the last files were dropped }
    property DropPoint: TPoint read FDropPoint;
    { The child control (if any) under the mouse when files are dropped -
      returns nil if no control under mouse }
    property DropControl: TControl read GetDropControl;
    { realign possition of all child controls }
    procedure RefreshAutoPosition;
    { check time and do minimize on Time, should be called by user in OnIdle event }
    procedure CheckAutoMinimize;
    { set value to SetChildrenValueType for all children components }
    procedure SetAllChildren(SetChildrenValueType: TSetChildrenValueTypes; Value: variant; P: Pointer; ExcludeTag: integer);
    { set size of not minimized box (stored automatically before minimize) }
    property OriginSize: integer read FOriginSize write FOriginSize default 0;
    { set auto resize setting for particular component }
    function SetAutoResizeChildComponentPos(CompName: string; VerticalAlign: TrGroupBoxVerticalAlignment; ColumnIndexFrom, ColumnIndexTo, ReduceWidth: integer;
      HorizontalAlign: TrGroupBoxHorizontalAlignment; ReduceHeight: integer): string;
    { get auto resize setting for particular component }
    function GetAutoResizeChildComponentPos(CompName: string; var VerticalAlign: TrGroupBoxVerticalAlignment; var ColumnIndexFrom, ColumnIndexTo,
      ReduceWidth: integer; var HorizontalAlign: TrGroupBoxHorizontalAlignment; var ReduceHeight: integer): Boolean;
  published
    { switch on/off automatic enabling of all children control according enable state of the rGroupBox }
    property AutoEnableChildren: Boolean read FAutoEnableChildren write FAutoEnableChildren default true;
    { switch on/off showing of system hint in Caption, used for rGroupBox as status bar }
    property AutoHint: Boolean read FAutoHint write FAutoHint default false;
    { define top of the client rect }
    property ClientRectTop: integer read FClientRectTop write SetClientRectTop default - 1;
    { define type of the frame }
    property FrameType: TFrameType read FFrameType write SetFrameType default ft3D;
    { define color of the frame }
    property FrameColor: TColor read FFrameColor write SetFrameColor default clWindowFrame;
    { define height of the caption rectangle }
    property CaptionHeight: byte read FCaptionHeight write SetCaptionHeight stored StoreCaptionHeight;
    { define background color of the caption rectangle }
    property CaptionColor: TColor read FCaptionColor write SetCaptionColor stored StoreCaptionColor;
    { define font color of the caption rectangle }
    property CaptionFontColor: TColor read FCaptionFontColor write SetCaptionFontColor stored StoreCaptionFontColor;
    { define font color of the caption rectangle in disable state }
    property CaptionDisableFontColor: TColor read FCaptionDisableFontColor write SetCaptionDisableFontColor default clBtnShadow;
    { define font of the caption rectangle }
    property CaptionFont: TFont read FCaptionFont write SetCaptionFont stored IsFontStored;
    { define if the caption uses parent font }
    property CaptionParentFont: Boolean read FCaptionParentFont write SetCaptionParentFont default true;
    { define alignment of the caption text }
    property CaptionAlignment: TAlignment read FCaptionAlignment write SetCaptionAlignment default taCenter;
    { define position of the caption rectangle }
    property CaptionPosition: TCaptionPosition read FCaptionPosition write SetCaptionPosition default cpTop;
    { define start color of the background gradient }
    property GradientColorFrom: TColor read FGradientColorFrom write SetGradientColorFrom default clBtnFace;
    { define end color of the background gradient }
    property GradientColorTo: TColor read FGradientColorTo write SetGradientColorTo default clBtnFace;
    { define type of the background gradient }
    property GradientType: TGradientType read FGradientType write SetGradientType default gtNone;
    { define type of the minimize button in caption rectangle }
    property MinMaxBtnType: TMinMaxBtnType read FMinMaxBtnType write SetMinMaxBtnType default mmbNone;
    { define glyph for the minimize button }
    property MinBtnPicture: TPicture read FMinBtnPicture write SetMinBtnPicture;
    { define glyph for the maximize button }
    property MaxBtnPicture: TPicture read FMaxBtnPicture write SetMaxBtnPicture;
    { define background picture for the box }
    property BackgroundPicture: TPicture read FBackgroundPicture write SetBackgroundPicture;
    { define type of background picture for the box }
    property BackgroundPictureType: TBackgroundPictureType read FBackgroundPictureType write SetBackgroundPictureType default bkpNone;
    { define thet background has the same color as Tab background color }
    property UseTabColor: Boolean read FUseTabColor write SetUseTabColor default false;
    { define whether box is minimized }
    property Minimized: Boolean read FMinimized write SetMinimized default false;
    { define whether minimizing is animated }
    property MinimizeAnimate: Boolean read FMinimizeAnimate write FMinimizeAnimate default false;
    { define minimize step delay if minimizing is animated }
    property AutoMinimizeTime: cardinal read FAutoMinimizeTime write FAutoMinimizeTime default 0;
    { define maximize step delay if minimizing is animated }
    property AutoMaximizeTime: cardinal read FAutoMaximizeTime write FAutoMaximizeTime default 0;
    { define whether size grip is shown at right bottom corner }
    property ShowSizeGrip: Boolean read FShowSizeGrip write SetShowSizeGrip default false;
    { subcomponent define autoresize properties for the box }
    property AutoResizePos: TAutoResizePos read FAutoResizePos write FAutoResizePos;
    { subcomponent define autoresize properties for the children controls }
    property AutoResizeChild: TAutoResizeChild read FAutoResizeChild write FAutoResizeChild;
    property OnResize; {< @exclude }
    property OnDrawControl: TDrawControl read FOnDrawControl write FOnDrawControl; {< @exclude }
    { event is called after AutoPosition refreshing }
    property OnRefreshAutoPosition: TNotifyEvent read FOnRefreshAutoPosition write FOnRefreshAutoPosition;
    { event is called after minimize state change }
    property OnMinimizeChange: TNotifyEvent read FOnMinimizeChange write FOnMinimizeChange;
    { switch on/off dropping of files }
    property EnableDropFile: Boolean read FEnableDropFile write SetEnableDropFile default false;
    { event called when files are dropped on the control; it identify sending
      control and the X and Y co-oridnates where files were dropped }
    property OnDropFiles: TDropFilesEvent read FOnDropFiles write FOnDropFiles;
    { event is called when Scale is changed }
    property OnChangeScale: TChangeScaleEvent read FChangeScale write FChangeScale;
  end;

type
  // ==============================================================================
  // TrImageEx

  TrImageEx = class;
  { list of shape types used by TrImageEx }
  TShapeType = (stRect, stRoundRect, stElipse, stHalfElipseLeft, stHalfElipseTop, stHalfElipseRight, stHalfElipseBottom);
  { list of boundlabel positions used by TrImageEx }
  TLabelPos = (lpInside, lpTopLeft, lpTopCenter, lpTopRight, lpBottomLeft, lpBottomCenter, lpBottomRight);

  { Label subcomponent used by other components }
  TBoundLabel = class(TCustomLabel)
  private
    FImage: TrImageEx;
    FPosition: TLabelPos;
    procedure SetPosition(Value: TLabelPos);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure AdjustBounds; override;
    procedure UpdatePos;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  published
    { define possition of bound label }
    property Position: TLabelPos read FPosition write SetPosition default lpInside;
    property Caption;
    property Color;
    property Font;
    property ParentColor;
    property ParentFont;
    property ShowAccelChar;
    property Transparent;
    property Visible;
    property OnClick;
    property OnDblClick;
  end;

  { enhanced TImage supporting label and drawing of shapes }
  TrImageEx = class(TImage)
  private
    FBoundLabel: rImprovedComps.TBoundLabel;
  protected
    procedure Loaded; override;
    procedure SetParent(AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure CMVisibleChanged(var Message: TMessage); message CM_VISIBLECHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure Paint; override;
  public
    SaveAsBMP: Boolean;
    SaveAsShape: Boolean;
    ShapeType: rImprovedComps.TShapeType;
    PenColor, FillColor, BkColor: TColor;
    PenWidth: integer;
    LastFileName: string;
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    procedure SetBounds(ALeft: integer; ATop: integer; AWidth: integer; AHeight: integer); override; {< @exclude }
    { draw defined shape }
    procedure DrawShape;
    { define shape type }
    procedure SetShape(AShapeType: rImprovedComps.TShapeType; APenColor, AFillColor, ABkColor: TColor; APenWidth: integer);
    { assign shape type from another ImageEx }
    procedure AssignShape(Source: TrImageEx);
    { assign another picture }
    procedure AssignPicture(P: TPicture; AsBMP: Boolean);
  published
    property BoundLabel: rImprovedComps.TBoundLabel read FBoundLabel; {< @exclude }
  end;

  // ==============================================================================
  // TrMemoEx

  { list of options (popup menus) used by TrMemoEx }
  TrMemoOption = (moAllowSetFont, moAllowWordWrap, moAllowLoadSaveFile);
  { set of options (popup menus) used by TrMemoEx }
  TrMemoOptions = set of TrMemoOption;

  { enhanced TMemo component which provides
    highlighting of words in text according list of words or
    highlighting of text between start and stop tag }
  TrMemoEx = class(TMemo)
  private
    FColIncrement: integer;
    FRowIncrement: integer;
    FOptionsEx: TrMemoOptions;
    FHighlightWords: TStringList;
    FHighlightWholeWords: Boolean;
    FHighlightFromStr: string;
    FHighlightToStr: string;
    FHighlightedFontColor: TColor;
    FHighlightedBrushColor: TColor;
    FHighlightFromToEx: TStringList;
    FPageBreaksCharNumber: string;
    FPopupMenu: TPopupMenu;
    procedure SetHighlightWords(Value: TStringList);
    procedure SetHighlightWholeWords(Value: Boolean);
    procedure SetHighlightFromStr(const Value: string);
    procedure SetHighlightToStr(const Value: string);
    procedure SetHighlightedFontColor(Value: TColor);
    procedure SetHighlightedBrushColor(Value: TColor);
    procedure SetHighlightFromToEx(Value: TStringList);
    procedure HighlightWordsChanged(Sender: TObject);
    procedure PopupMenuOpen(Sender: TObject);
    procedure PopupMenuClick(Sender: TObject);
    procedure LoadSBPos;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    // procedure HandleCopy(var Message: TWMPaint); message WM_COPY;
  protected
    { Protected declarations }
    FSearchText: string;
    FSearchIgnoreCase: Boolean;
    procedure Loaded; override;
    procedure Change; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { search and select text if found }
    function SearchText(const Search: string = ''; FromStart: Boolean = true; IgnoreCase: Boolean = true): Boolean;
    { search and select next text occurrence }
    function SearchNext: Boolean;
  published
    { list of hightlighted words, font color and background color can be defined in line on 2nd and 3rd place with TAB separator }
    property HighlightWords: TStringList read FHighlightWords write SetHighlightWords;
    { define whether only whole words is highlighted }
    property HighlightWholeWords: Boolean read FHighlightWholeWords write SetHighlightWholeWords default false;
    { define start tag for hightlighting,
      text between start and stop tag is highlighted }
    property HighlightFromStr: string read FHighlightFromStr write SetHighlightFromStr;
    { define stop tag for hightlighting,
      text between start and stop tag is highlighted }
    property HighlightToStr: string read FHighlightToStr write SetHighlightToStr;
    { define font color of highlighted text }
    property HighlightedFontColor: TColor read FHighlightedFontColor write SetHighlightedFontColor default clWindowText;
    { define background color of highlighted text }
    property HighlightedBrushColor: TColor read FHighlightedBrushColor write SetHighlightedBrushColor default clMoneyGreen;
    { enhanced from-to hightlighting, list of FromStr + ToStr + font color + background color
      items is defined in line with TAB separator, use #9 instead of tab and #EOL# instead on end of line }
    property HighlightFromToEx: TStringList read FHighlightFromToEx write SetHighlightFromToEx;
    { define number of chars used for showing of pagebreaks inside of edit area }
    property PageBreaksCharNumber: string read FPageBreaksCharNumber write FPageBreaksCharNumber;
    { define additional options }
    property OptionsEx: TrMemoOptions read FOptionsEx write FOptionsEx default[moAllowSetFont, moAllowLoadSaveFile];
  end;

  // ==============================================================================
  // TrComboBoxEx

type
  { event type for getting of item properties used by TrComboBoxEx and TrDBComboBox }
  TGetListItemProps = procedure(Sender: TObject; Canvas: TCanvas; Index: integer; State: TOwnerDrawState; var Rect: TRect; var Text: string; ShowBmp: TBitmap;
    var DrawSeparatorTop: Boolean; var DrawSeparatorBottom: Boolean) of object;

  { dropdown items properties used for TrComboBoxEx and TrDBComboBox }
  TDisplayedItems = class(TPersistent)
  private
    AComboBox: TCustomComboBox;
    AItemSpace: integer;
    FListItems: TStrings;
    FAddItemAsFirstColumn: Boolean;
    FColumnSeparator: string;
    FColumnsWidth: array of integer;
    FDrawColumnLine: Boolean;
    FOnListDefined: TNotifyEvent;
    ColumnsWidthCalculated: Boolean;
    procedure SetDisplayedItems(Value: TStrings);
    procedure SetAddItemAsFirstColumn(Value: Boolean);
    procedure SetColumnSeparator(Value: string);
    procedure ItemUpdated(Sender: TObject);
    function GetItemDisplayedText(idx: integer): string;
    procedure CalcColumnsWidth;
  public
    constructor Create(AOwner: TCustomComboBox; OnListDefined: TNotifyEvent); {< @exclude }
    destructor Destroy; override; {< @exclude }
    { return maximal width of all items }
    function GetMaxItemWidth(DropDownCount: integer): integer;
    { draw one item of combo dropdown }
    procedure DrawListItem(Index: integer; State: TOwnerDrawState; Rect: TRect; GetListItemProps: TGetListItemProps);
  published
    { define additional items text }
    property ListItems: TStrings read FListItems write SetDisplayedItems;
    { define whether ListItem is shown at begining }
    property AddItemAsFirstColumn: Boolean read FAddItemAsFirstColumn write SetAddItemAsFirstColumn default false;
    { define separator used in ListItem }
    property ColumnSeparator: string read FColumnSeparator write SetColumnSeparator;
    { define whether column line is drawn }
    property DrawColumnLine: Boolean read FDrawColumnLine write FDrawColumnLine default true;
  end;

  { event type used by TrComboBox }
  TComboLoadItemsEvent = procedure(Sender: TObject; CurText: string; var NewText: string) of object;

  { enhanced TComboBox component which provides additional item properties and
    automatic setting of dropdown listbox width according length of items text }
  TrComboBoxEx = class(TComboBox)
  private
    FDropDownAutoWidth: Boolean;
    FDropDownMaxWidth: integer;
    FDisplayedItems: TDisplayedItems;
    FGetListItemProps: TGetListItemProps;
    WndRecreated: Boolean;
    IntTimer: TTimer;
    FUpdateTimerEvent: TNotifyEvent;
    FUpdateTimerDelay: cardinal;
    FOnUpdateLoadItems: TComboLoadItemsEvent;
    FLastKeyDown: Word;
    FJustClosed: Boolean;
    FItemHeightEx: integer;
    procedure ListDefined(Sender: TObject);
    function UseOwnDraw: Boolean;
    procedure IntTimerEvent(Sender: TObject);
    procedure SetItemHeightEx(Value: integer);
    function GetSelectedText: string;
  protected
    FScaleRatio: double;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Message: TMessage); override;
    procedure Loaded; override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure DropDown; override;
    procedure DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState); override;
    procedure Change; override;
    procedure CloseUp; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { reset after change timer }
    procedure ResetTimer(SetActive, CallEventNow: Boolean);
    { add active od specified text to Items }
    procedure AddTextToItems(AText: string = '');
    { remove active od specified text from Items }
    procedure RemoveTextFromItems(AText: string = '');
    { return seleced items as text }
    property SelectedText: string read GetSelectedText;
  published
    { define whether drop down width is set according maximal item width }
    property DropDownAutoWidth: Boolean read FDropDownAutoWidth write FDropDownAutoWidth default true;
    { define maximal width if DropDownAutoWidth is used }
    property DropDownMaxWidth: integer read FDropDownMaxWidth write FDropDownMaxWidth default 600;
    { define additional item columns }
    property DisplayedItems: TDisplayedItems read FDisplayedItems write FDisplayedItems;
    { event for user definition of item properties }
    property OnGetListItemProps: TGetListItemProps read FGetListItemProps write FGetListItemProps;
    { define after update timer interval }
    property UpdateTimerDelay: cardinal read FUpdateTimerDelay write FUpdateTimerDelay default 0;
    { event called with defined delay after update }
    property OnUpdateTimer: TNotifyEvent read FUpdateTimerEvent write FUpdateTimerEvent;
    { event called to load items during OnUpdateTimer event }
    property OnUpdateLoadItems: TComboLoadItemsEvent read FOnUpdateLoadItems write FOnUpdateLoadItems;
    { contain last pressed key }
    property LastKeyDown: Word read FLastKeyDown;
    { define item height for fixed draw before scaling }
    property ItemHeightEx: integer read FItemHeightEx write SetItemHeightEx default cComboBoxItemHeightDefault;
  end;

  // ==============================================================================
  // TrButton

  { enhanced TButton with ImagesDisabled property
    which is used for disable state of button instead of DisableImageIndex }
  TrButton = class(TButton)
  private
    FImagesEnabled: TCustomImageList;
    FImagesDisabled: TCustomImageList;
    procedure SetImagesDisabled(const Value: TCustomImageList);
  protected
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
  public
  published
    { define image list used for disable state }
    property ImagesDisabled: TCustomImageList read FImagesDisabled write SetImagesDisabled;
  end;

  // ==============================================================================
  // TrPopupMonthCalendar
const
  rPopupMonthCalendarKeys: set of byte = [VK_UP, VK_DOWN, VK_LEFT, VK_RIGHT];

type
  { Popup version of TMonthCalendar, used by TrEditDate, TrDBEditDate, TrDBGrid and TrStringGridEd }
  TrPopupMonthCalendar = class(TMonthCalendar)
  private
    SelectDate: TNotifyEvent;
    FIsShown: Boolean;
    procedure WMMouseActivate(var Message: TWMMouseActivate); message WM_MOUSEACTIVATE;
    procedure CNNotify(var Message: TWMNotifyMC); message CN_NOTIFY;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor CreateEx(AOwner: TComponent; OnSelectDate: TNotifyEvent);
    procedure ShowCalendar(ADate: TDateTime);
    procedure HideCalendar;
    property IsShown: Boolean read FIsShown;
  end;

  // ==============================================================================
  // TrEditDate + TrEditNum

  { enhanced TButtonedEdit allows fast entering of the date value by keyboard or by Month calendar;
    supported input formats - dd, dd/mm, dd/mm/yy, dd/mm/yyyy
    shortcuts - arrows Up/Down (optionaly Ctrl+Shift) to move date/month/year
    t = today, s = start of month, e = end of month, d,p = dropdown month calendar (if right button is visible) }
  TrEditDate = class(TButtonedEdit)
  private
    // FLastDateValue: TDateTime;
    FMinDate: TDateTime;
    FMaxDate: TDateTime;
    FNowIfEmpty: Boolean;
    FNowOnStart: Boolean;
    OrigText: string;
    procedure SetDate(ADateValue: TDateTime);
    function GetDate: TDateTime;
    function GetDateValue(var ADateValue: TDateTime): Boolean;
    function CheckDate: Boolean;
    procedure FRightButtonClick(Sender: TObject);
    procedure PopupCallbackSelectDate(Sender: TObject);
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
    procedure WMKillFocus(var Message: TMessage); message WM_KillFocus;
  protected
    FPopupCalendar: TrPopupMonthCalendar;
    procedure Loaded; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyUp(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure DoEnter; override;
    procedure DoExit; override;
    procedure EMSetReadOnly(var Message: TMessage); message EM_SETREADONLY;
    function DoDropDown: Boolean; virtual;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  published
    { Minimal value which can be entered }
    property MinDate: TDateTime read FMinDate write FMinDate;
    { Maximal value which can be entered }
    property MaxDate: TDateTime read FMaxDate write FMaxDate;
    { Defines whether Now is used if Text is empty }
    property NowIfEmpty: Boolean read FNowIfEmpty write FNowIfEmpty default false;
    { Defines whether Now is entered on startup }
    property NowOnStart: Boolean read FNowOnStart write FNowOnStart default false;
    { Get or Set actual Date }
    property DateValue: TDateTime read GetDate write SetDate;
    property Text stored false; {< @exclude }
  end;

  { list of numeric styles used by TrEditNum }
  enStyle = (enInteger, enFloat);

  { enhanced TButtonedEdit component which provides functionality
    for editing of integer or float number }
  TrEditNum = class(TButtonedEdit)
  private
    FStyle: enStyle;
    FMinValue: extended;
    FMaxValue: extended;
    FStepValue: extended;
    FStepOnly: Boolean;
    FFormat: string;
    procedure SetValue(Value: extended);
    function GetValue: extended;
    procedure SetAsInteger(Value: integer);
    function GetAsInteger: integer;
    procedure SetFormat(const Value: string);
    function CheckValue: Boolean;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { increment value by step value }
    procedure StepUp;
    { decrement value by step value }
    procedure StepDown;
  published
    { Defines type of edit value - Integer or Float }
    property Style: enStyle read FStyle write FStyle default enInteger;
    { Minimal value which can be entered }
    property MinValue: extended read FMinValue write FMinValue;
    { Maximal value which can be entered }
    property MaxValue: extended read FMaxValue write FMaxValue;
    { if user click to arrow UP or DOWN value is inc/de crement by step value }
    property StepValue: extended read FStepValue write FStepValue;
    { Defines whether only values correspond with step value can be entered }
    property StepOnly: Boolean read FStepOnly write FStepOnly default false;
    { Defines format of number }
    property Format: string read FFormat write SetFormat;
    { Get or Set number value }
    property Value: extended read GetValue write SetValue;
    { Get or Set integre value }
    property AsInteger: integer read GetAsInteger write SetAsInteger stored false;
    property Text stored false; {< @exclude }
  end;

  // ==============================================================================
  // TrCheckTreeView

  { enhanced TTreeView component which provides checkboxes
    and child/parent node auto check functions }
  TrCheckTreeView = class(TTreeView)
  private
    FShowCheckBox: Boolean;
    FAutoCheckChildren: Boolean;
    FAutoCheckParents: Boolean;
    FCheckBoxClick: TTVChangedEvent;
    procedure SetShowCheckBox(Value: Boolean);
    function IntGetItemsCount(Node: TTreeNode; Siblink, Checked: Boolean): integer;
  protected
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
    procedure DoCheckBoxClick(Node: TTreeNode); virtual;
    procedure ItemsRecreated; virtual;
  public
    { return check state for defined node }
    function GetCheckState(Node: TTreeNode): Boolean;
    { set check state for defined node }
    procedure SetCheckState(Node: TTreeNode; Checked: Boolean);
    { set check state for defined node and all children }
    procedure SetAllChildrenCheckState(Node: TTreeNode; Checked: Boolean);
    { set check state for defined node and all parents }
    procedure SetAllParentsCheckState(Node: TTreeNode; AutoChecked, Checked: Boolean);
    { set check state for all nodes }
    procedure SetAllItemsCheckState(Checked: Boolean);
    { return number of all child nodes }
    function GetChildrenCount(Node: TTreeNode): integer;
    { return number of checked child nodes }
    function GetChildrenCheckedCount(Node: TTreeNode): integer;
    { return number of all siblink nodes }
    function GetSiblinkCount(Node: TTreeNode): integer;
    { return number of checked siblink nodes }
    function GetSiblinkCheckedCount(Node: TTreeNode): integer;
  published
    { define whether checkboxes are shown }
    property ShowCheckBox: Boolean read FShowCheckBox write SetShowCheckBox default false;
    { define whether children nodes are automatically set according parent node }
    property AutoCheckChildren: Boolean read FAutoCheckChildren write FAutoCheckChildren default false;
    { define whether parent nodes are automatically set according children nodes }
    property AutoCheckParents: Boolean read FAutoCheckParents write FAutoCheckParents default false;
    { event called after checkbox click }
    property OnCheckBoxClick: TTVChangedEvent read FCheckBoxClick write FCheckBoxClick;
  end;

  // ==============================================================================
  // TrStringsHolder

  { non visual component containing Strings object
    for editing in design time and storing in application }
  TrStringsHolder = class(TComponent)
  private
    FStrings: TStrings;
  protected
    procedure SetStrings(Value: TStrings);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
  published
    { define stored string list }
    property Strings: TStrings read FStrings write SetStrings;
  end;

  // ==============================================================================
  // TrImageListEx

  { enhanced TImageList component which provides storing of images
    in external compressed resource file in design and runtime
    and can block storing images in DFM file for descendant modules,
    so list can be used in ihnerited forms }
  TrImageListEx = class(TImageList)
  private
    FStoreToDFM: Boolean;
    FStoreToExternalFile: Boolean;
    FExternalFileName: string;
    FExternalFileDesignTimeDir: string;
    FOriginalImages: TCustomImageList;
    FMinScaleToResize: double;
    function GetDefaultDesignTimeDir: string;
    function IsStoredDesignTimeDir: Boolean;
    function IsMinScaleToResizeStored: Boolean;
  protected
    procedure Loaded; override;
    procedure DefineProperties(Filer: TFiler); override;
    function GetResourceFileName(ShowErrMsg: Boolean): string;
    procedure ReadExternalFile(Reader: TReader);
    procedure WriteExternalFile(Writer: TWriter);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { stores current images to internal image list wich is used as source of images during resizing, is called automatically durong loading }
    procedure StoreAsOriginalImages;
    { change Width and Height of image list and resize images }
    procedure ResizeImages(ScaleRatio: double);
  published
    { define whether images are stored in DFM }
    property StoreToDFM: Boolean read FStoreToDFM write FStoreToDFM default true;
    { define whether images are stored in external file }
    property StoreToExternalFile: Boolean read FStoreToExternalFile write FStoreToExternalFile default false;
    { define file name of external file if StoreToExternalFile=true }
    property ExternalFileName: string read FExternalFileName write FExternalFileName;
    { define directory of external file for design time if StoreToExternalFile=true }
    property ExternalFileDesignTimeDir: string read FExternalFileDesignTimeDir write FExternalFileDesignTimeDir stored IsStoredDesignTimeDir;
    { define minimal scale ration when images are streched, in ther case images are only centered }
    property MinScaleToResize: double read FMinScaleToResize write FMinScaleToResize stored IsMinScaleToResizeStored;
  end;

  // ==============================================================================
  // TrPageControlEx + TrTabControlEx

  { @exclude }
  THackCustomTabControl = class(TCustomTabControl)
  end;

  { event type for getting of tab props used by TTabPropsEx }
  TTabGetPropsEvent = procedure(Sender: TObject; TabIndex: integer; Font: TFont; var BkColor: TColor; var CanClose: Boolean) of object;
  { event type for tab closing notification used by TTabPropsEx }
  TTabCloseEvent = procedure(Sender: TObject; TabIndex: integer; var CanClose: Boolean) of object;
  { event type for getting of tab popup menu used by TTabPropsEx }
  TTabGetPopupMenu = procedure(Sender: TObject; TabIndex: integer; var PopupMenu: TPopupMenu) of object;
  { event type for tab custom drawing used by TTabPropsEx }
  TTabDrawTabEvent = procedure(Sender: TCustomTabControl; TabIndex: integer; PaneRect, TabRect: TRect; BkColor: TColor; Selected, Hot, CloseBtn: Boolean)
    of object;

  { subcomponent defines all props and events for TrTabControlEx and TrPageControlEx }
  TTabPropsEx = class(TComponent)
  private
    FOwner: TCustomTabControl;
    TC: THackCustomTabControl;
    FScaleRatio: double;
    FHotTabIndex: integer;
    FHotBtnIndex: integer;

    FDrawThemed: Boolean;
    FWordWrap: Boolean;
    FTabColor: TColor;
    FSelectedTabColor: TColor;
    FHotTabColor: TColor;

    FParentFont: Boolean;
    FTabFont: TFont;
    FSelectedTabFont: TFont;
    FHotTabFont: TFont;

    FSelectedBorderColor: TColor;
    FHotBorderColor: TColor;

    FShowCloseBtn: Boolean;
    FCloseBtnSize: integer;
    FCloseBtn: TPicture;
    FHotCloseBtn: TPicture;

    FOnTabGetProps: TTabGetPropsEvent;
    FOnTabClose: TTabCloseEvent;
    FOnTabGetPopupMenu: TTabGetPopupMenu;
    FOnDrawTabBackground: TTabDrawTabEvent;
    FOnDrawTabContent: TTabDrawTabEvent;
    FOnDrawTabCloseBtn: TTabDrawTabEvent;
    FOnDrawTabFinalize: TTabDrawTabEvent;

    procedure SetDrawThemed(Value: Boolean);
    procedure SetWordWrap(Value: Boolean);
    procedure SetTabColor(Value: TColor);
    procedure SetSelectedTabColor(Value: TColor);
    procedure SetHotTabColor(Value: TColor);

    procedure SetParentFont(Value: Boolean);
    procedure SetTabFont(Value: TFont);
    procedure SetSelectedTabFont(Value: TFont);
    procedure SetHotTabFont(Value: TFont);
    function IsFontStored: Boolean;
    procedure FontChanged(Sender: TObject);

    procedure SetSelectedBorderColor(Value: TColor);
    procedure SetHotBorderColor(Value: TColor);

    procedure SetShowCloseBtn(Value: Boolean);
    procedure SetCloseBtnSize(Value: integer);
    procedure SetCloseBtn(Value: TPicture);
    procedure SetHotCloseBtn(Value: TPicture);
  protected
    procedure ResetHotState(DoInvalidate: Boolean);
    procedure ResetFont(Font: TFont);
    function CheckRightBorder(TabIndex: integer; var R: TRect; AddBorder: integer): Boolean;
    procedure DrawItemMessage(var Message: TWMDrawItem);
    procedure DrawTabEx(TabIndex: integer; ItemRect: TRect; ItemState: cardinal);
    procedure DrawBkThemed(TabIndex: integer; PaneRect, TabRect: TRect; Selected, Hot: Boolean); virtual;
    procedure DrawBkStandard(TabIndex: integer; PaneRect, TabRect: TRect; BkColor: TColor; Selected, Hot: Boolean); virtual;
    procedure DrawTabContent(TabIndex: integer; TabRect: TRect; Selected, Hot, CloseBtn: Boolean); virtual;
    procedure DrawCloseBtn(TabIndex: integer; BtnRect: TRect; Hot: Boolean); virtual;
    procedure UpdateScale(ScaleRatio: double; M, D: integer);
    function UseScale(I: integer): integer;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { return whether close button is shown for defined tab }
    function CloseBtnShown(TabIndex: integer): Boolean;
    { return rectangle of close button for defined tab }
    function CloseBtnRect(TabIndex: integer): TRect;
    { check whether click of mouse is on close button }
    procedure CheckMouseClick(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
    { check whether mouse is over close button }
    procedure CheckMouseMove(Shift: TShiftState; X, Y: integer);
    { define index of hot tab }
    property HotTabIndex: integer read FHotTabIndex;
    { define index of hot button }
    property HotBtnIndex: integer read FHotBtnIndex;
  published
    { define whether themes is used for tabs }
    property DrawThemed: Boolean read FDrawThemed write SetDrawThemed default true;
    { define whether long tab text is wrapped }
    property WordWrap: Boolean read FWordWrap write SetWordWrap default false;
    { define background color of tab }
    property TabColor: TColor read FTabColor write SetTabColor default clNone;
    { define background color of selected tab }
    property SelectedTabColor: TColor read FSelectedTabColor write SetSelectedTabColor default clNone;
    { define background color of hot tab }
    property HotTabColor: TColor read FHotTabColor write SetHotTabColor default clNone;

    { define whether parent font is used }
    property ParentFont: Boolean read FParentFont write SetParentFont default true;
    { define font of tab }
    property TabFont: TFont read FTabFont write SetTabFont stored IsFontStored;
    { define font of selected tab }
    property SelectedTabFont: TFont read FSelectedTabFont write SetSelectedTabFont stored IsFontStored;
    { define font of hot tab }
    property HotTabFont: TFont read FHotTabFont write SetHotTabFont stored IsFontStored;

    { define border color of selected tab }
    property SelectedBorderColor: TColor read FSelectedBorderColor write SetSelectedBorderColor default $000080FF;
    { define border color of hot tab }
    property HotBorderColor: TColor read FHotBorderColor write SetHotBorderColor default $000080FF;

    { define whether close button is shown }
    property ShowCloseBtn: Boolean read FShowCloseBtn write SetShowCloseBtn default false;
    { define size of close button in tab }
    property CloseBtnSize: integer read FCloseBtnSize write SetCloseBtnSize default 16;
    { define close button picture }
    property CloseBtn: TPicture read FCloseBtn write SetCloseBtn;
    { define close button picture in hot state }
    property HotCloseBtn: TPicture read FHotCloseBtn write SetHotCloseBtn;

    { event for user defined state of tab }
    property OnTabGetProps: TTabGetPropsEvent read FOnTabGetProps write FOnTabGetProps;
    { event called after close button click }
    property OnTabClose: TTabCloseEvent read FOnTabClose write FOnTabClose;
    { event used for definition of tab popup }
    property OnTabGetPopupMenu: TTabGetPopupMenu read FOnTabGetPopupMenu write FOnTabGetPopupMenu;
    { event for user drawing of tab background }
    property OnDrawTabBackground: TTabDrawTabEvent read FOnDrawTabBackground write FOnDrawTabBackground;
    { event for user drawing of tab content }
    property OnDrawTabContent: TTabDrawTabEvent read FOnDrawTabContent write FOnDrawTabContent;
    { event for user drawing of tab close button }
    property OnDrawTabCloseBtn: TTabDrawTabEvent read FOnDrawTabCloseBtn write FOnDrawTabCloseBtn;
    { event for any additional user drawing after tab is drawn }
    property OnDrawTabFinalize: TTabDrawTabEvent read FOnDrawTabFinalize write FOnDrawTabFinalize;
  end;

  { enhanced TPageControl which provides colored tabs, hottrack fonts, color and border,
    close button in tabs, function for reading active tab caption and changing of active tab;
    OwnerDraw must be set to TRUE to use this functions }
  TrPageControlEx = class(TPageControl)
  private
    FTabPropsEx: TTabPropsEx;
    FTabRightButton: TrButton;
    FTabRightButtonVisible: Boolean;
    FTabRightLabel: TrHTMLLabel;
    FTabRightLabelVisible: Boolean;
    procedure SetTabRightButtonVisible(Value: Boolean);
    procedure SetTabRightLabelVisible(Value: Boolean);
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  protected
    FScaleRatio: double;
    procedure Loaded; override;
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { add spaces to all tabs to make place for close buttons }
    procedure AddSpaceForCloseBtn(Center: Boolean = false);
    { select defined tab including calling CanChange and OnChange event }
    function SetActiveTabEx(Index: integer): Boolean;
    { select defined page including calling CanChange and OnChange event }
    function SetActivePageEx(TS: TTabSheet): Boolean;
    { return caption of active tab }
    function ActiveTabCaption: string;
    { return object of active tab }
    function ActiveTabObject: TObject;
    { return number of visible tabs }
    function VisiblePageCount: integer;
  published
    { subcomponent define all new tab properties }
    property TabPropsEx: TTabPropsEx read FTabPropsEx write FTabPropsEx;
    { subcomponent define additional button on right border }
    property TabRightButton: TrButton read FTabRightButton write FTabRightButton;
    { define visibility of additional button on right border }
    property TabRightButtonVisible: Boolean read FTabRightButtonVisible write SetTabRightButtonVisible default false;
    { subcomponent define additional label on right border }
    property TabRightLabel: TrHTMLLabel read FTabRightLabel write FTabRightLabel;
    { define visibility of additional label on right border }
    property TabRightLabelVisible: Boolean read FTabRightLabelVisible write SetTabRightLabelVisible default false;
  end;

  { enhanced TTabControl which provides colored tabs, hottrack fonts, color and border,
    close button in tabs, function for reading active tab caption and changing of active tab;
    OwnerDraw must be set to TRUE to use this functions }
  TrTabControlEx = class(TTabControl)
  private
    FTabPropsEx: TTabPropsEx;
    FTabRightButton: TrButton;
    FTabRightButtonVisible: Boolean;
    FTabRightLabel: TrHTMLLabel;
    FTabRightLabelVisible: Boolean;
    procedure SetTabRightButtonVisible(Value: Boolean);
    procedure SetTabRightLabelVisible(Value: Boolean);
    procedure CNDrawItem(var Message: TWMDrawItem); message CN_DRAWITEM;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
  protected
    FScaleRatio: double;
    procedure Loaded; override;
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure ChangeScale(M, D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF}); override;
    function UseScale(I: integer): integer;
    procedure MouseMove(Shift: TShiftState; X, Y: integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer); override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    destructor Destroy; override; {< @exclude }
    { define actual scale ratio relative to original designed size }
    property ScaleRatio: double read FScaleRatio;
    { relace all & by && - allows to show & character in tab caption }
    procedure ReplaceAccelChar;
    { add spaces to all tabs to make place for close buttons }
    procedure AddSpaceForCloseBtn(Center: Boolean = false);
    { select defined tab including calling CanChange and OnChange event }
    function SetActiveTabEx(Index: integer): Boolean;
    { return caption of active tab }
    function ActiveTabCaption: string;
    { return object of active tab }
    function ActiveTabObject: TObject;
  published
    { subcomponent define all new tab properties }
    property TabPropsEx: TTabPropsEx read FTabPropsEx write FTabPropsEx;
    { subcomponent define additional button on right border }
    property TabRightButton: TrButton read FTabRightButton write FTabRightButton;
    { define visibility of additional button on right border }
    property TabRightButtonVisible: Boolean read FTabRightButtonVisible write SetTabRightButtonVisible default false;
    { subcomponent define additional label on right border }
    property TabRightLabel: TrHTMLLabel read FTabRightLabel write FTabRightLabel;
    { define visibility of additional label on right border }
    property TabRightLabelVisible: Boolean read FTabRightLabelVisible write SetTabRightLabelVisible default false;
  end;

  // ==============================================================================
  // TrRichEdit
  { enhanced TRichEdit with internal Popup menu, extended formatting, search function and support for drawing to Canvas }
  TrRichEdit = class(TRichEdit)
  private
    FPopupMenu: TPopupMenu;
    FSupportURLLink: Boolean;
    function GetRtfText: string;
    procedure SetRtfText(AValue: string);
    function PixelsToTwips(pixels: integer): integer;
    procedure DrawToCanvasInt(var R: TRect; OffsetX, OffsetY: integer; ACanvas: TCanvas; CalcSize, ShowEllipsis: Boolean);
    function GetNumberingEx: integer;
    procedure SetNumberingEx(Value: integer);
    procedure SetSupportURLLink(Value: Boolean);
  protected
    // FScaleRatio: double;
    FSearchText: string;
    FSearchIgnoreCase: Boolean;
    procedure CreateWnd; override;
    procedure WndProc(var Message: TMessage); override;
    procedure InitParaFormat2(var Para2: TParaFormat2);
    procedure GetParaFormat2(var Para2: TParaFormat2);
    procedure SetParaFormat2(var Para2: TParaFormat2);
    procedure PopupMenuOpen(Sender: TObject);
    procedure OnPopupClick(Sender: TObject);
    procedure PopupCommand(CmdTag: integer);
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
    { keep Rect width and set Rect height according to the RTF content }
    procedure CalculateTextHeight(var R: TRect);
    { draw RTF content to the canvas }
    procedure DrawToCanvas(R: TRect; OffsetX, OffsetY: integer; ACanvas: TCanvas; ShowEllipsis: Boolean = true);
    { add text to the end and select it }
    procedure AddSelectText(AText: string);
    { insert text to the current position and select it (can replace current selection, insert before of after) }
    procedure InsertSelectText(AText: string; ReplaceSelected: Boolean = false; BeforeSelected: Boolean = false);
    { switch on/off font style for selection }
    procedure SwitchFontStyle(AStyle: TFontStyle);
    { increase font size }
    function IncreaseFontSize(Step: integer = 0): Boolean;
    { decrease font size }
    function DecreaseFontSize(Step: integer = 0): Boolean;
    { open font selection dialog }
    function SelectFont: Boolean;
    { increase left indent }
    function IncreaseIndent(Step: integer = 10): Boolean;
    { decrease left indent }
    function DecreaseIndent(Step: integer = 10): Boolean;
    { get or set numbering style }
    property NumberingEx: integer read GetNumberingEx write SetNumberingEx;
    { switch on/off bullet point numbering }
    function SwitchBulletPoints: Boolean;
    { switch on/off number numbering }
    function SwitchBulletNumbers: Boolean;
    { switch on/off letter numbering }
    function SwitchBulletLetters: Boolean;
    { search and select text if found }
    function SearchText(const Search: string = ''; FromStart: Boolean = true; IgnoreCase: Boolean = true): Boolean;
    { search and select next text occurrence }
    function SearchNext: Boolean;
    { set or get content as rtf format }
    property RtfText: string read GetRtfText write SetRtfText;
  published
    property SupportURLLink: Boolean read FSupportURLLink write SetSupportURLLink default false;
  end;
{$IFDEF CategoryPanelGroupEx}
  // ==============================================================================
  // TrCategoryPanelGroupEx

  { fix the bug with subcomponents saved into DFM }
  TrCategoryPanelEx = class(TCategoryPanel)
  public
    procedure GetChildren(Proc: TGetChildProc; Root: TComponent); override; {< @exclude }
  end;

  { fix the bug with subcomponents saved into DFM }
  TrCategoryPanelGroupEx = class(TCategoryPanelGroup)
  protected
    function GetCategoryPanelClass: TCategoryPanelClass; override;
  public
    constructor Create(AOwner: TComponent); override; {< @exclude }
  end;
{$ENDIF}

procedure Register; {< @exclude }

implementation

{$R *.res}

uses rstring, rtool, rdlg, rpictool,
  zlib_interface, AppDirCfg, rDialogsEx;

const
  BoundLabelSpacing = 1;

type
  { @exclude }
  THackControl = class(TControl)
  end;

  THackEditButton = class(TEditButton)
  end;
{$IFDEF DEMO}

var
  DemoCounter: integer;

procedure ShowDemoMsg;
const
  DemoMsgText = 'You are using DEMO version of Rosinsky Components!' + #13 + 'More info on http://www.rosinsky.cz/delphi';
begin
  Inc(DemoCounter);
  if DemoCounter mod 30 = 0 then
    ShowMessage(DemoMsgText);
end;
{$ENDIF}

procedure Register;
begin
  RegisterComponents('Rosi', [TrListBox, TrColorBox, TrColorListBox, TrCheckListBox, TrStatusBar, TrToolBar, TrGroupBox, TrImageEx, TrButton, TrPageControlEx,
    TrTabControlEx, TrEditDate, TrEditNum, TrMemoEx, TrComboBoxEx,
{$IFDEF CategoryPanelGroupEx} TrCategoryPanelGroupEx, {$ENDIF}
    TrStringsHolder, TrImageListEx, TrCheckTreeView, TrRichEdit]);
end;

constructor TrListBox.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
end;

procedure TrListBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  // lbStandard Style has bug with ItemHeight and scrollbars, so change it to lbOwnerDrawFixed
  if (Style = lbStandard) then
  begin
    Style:=lbOwnerDrawFixed;
    ItemHeight:=cListBoxItemHeightDefault;
    OnDrawItem:=nil;
  end;
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * cListBoxItemHeightDefault);
end;

constructor TrColorBox.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
end;

procedure TrColorBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * cComboBoxItemHeightDefault);
end;

constructor TrColorListBox.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
end;

procedure TrColorListBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * cListBoxItemHeightDefault);
end;

procedure TCheckListBoxHelper.SetCheckSize(size: integer);
begin
{$IFDEF DX101P}
  // FCheckWidth is class item and cannot be accessed
  // TRttiContext.Create.GetType(Self.ClassInfo).GetField('FCheckWidth').SetValue(Self, size);
  // TRttiContext.Create.GetType(Self.ClassInfo).GetField('FCheckHeight').SetValue(Self, size);
{$ELSE}
{$IFDEF DXE2P}
  Self.FCheckWidth:=size;
  Self.FCheckHeight:=size;
{$ELSE}
  // nothing - cannot be changed
{$ENDIF}
{$ENDIF}
end;

constructor TrCheckListBox.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  FCheckboxUpdateMode:=chmCheckboxClick;
end;

procedure TrCheckListBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  if (Style = lbStandard) then
  begin
    Style:=lbOwnerDrawFixed;
    ItemHeight:=cListBoxItemHeightDefault;
    OnDrawItem:=nil;
  end;
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * cListBoxItemHeightDefault);
  // change checkbox size
  SetCheckSize(Round(ItemHeight * 0.8));
end;

function TrCheckListBox.CheckedCount: integer;
var
  I: integer;
begin
  Result:=0;
  for I:=0 to Items.Count - 1 do
    if Checked[I] then
      Inc(Result);
end;

function TrCheckListBox.CanModifyCheckItem(Index: integer): Boolean;
begin
  Result:=(Index >= 0) and (Index < Items.Count) and ItemEnabled[Index];
end;

procedure TrCheckListBox.DblClick;
begin
  inherited;
  if (FCheckboxUpdateMode = chmItemDblClick) and CanModifyCheckItem(ItemIndex) then
  begin
    Checked[ItemIndex]:=not Checked[ItemIndex];
    ClickCheck;
    Click;
  end;
end;

procedure TrCheckListBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X: integer; Y: integer);
var
  oldStatus: Boolean;
begin
  if (FCheckboxUpdateMode = chmItemClick) and (Shift = [ssLeft]) and CanModifyCheckItem(ItemIndex) then
  begin
    oldStatus:=Checked[ItemIndex];
    inherited;
    if oldStatus = Checked[ItemIndex] then
    begin
      Checked[ItemIndex]:=not Checked[ItemIndex];
      ClickCheck;
    end;
  end
  else
    inherited;
end;

constructor TrStatusBar.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
end;

procedure TrStatusBar.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
var
  I: integer;
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  for I:=0 to Panels.Count - 1 do
    Panels[I].Width:=MulDiv(Panels[I].Width, M, D);
end;

constructor TrToolBar.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  FScaleAutoButtonSize:=false;
end;

procedure TrToolBar.Loaded;
begin
  inherited;
  // RefreshSize;
end;

procedure TrToolBar.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  if FScaleAutoButtonSize then
    RefreshSize
  else
    inherited;
end;

procedure TrToolBar.RefreshSize;
begin
  ButtonHeight:=Height;
  ButtonWidth:=ButtonHeight + 1;
  AdjustSize;
end;

// ==============================================================================
// TrGroupBox AutoresizePos subcomp
constructor TAutoResizePos.Create(AOwner: TrGroupBox);
begin
  FOwner:=AOwner;
end;

destructor TAutoResizePos.Destroy;
begin
  inherited;
end;

procedure TAutoResizePos.UpdatePos;
begin
  if csLoading in Self.FOwner.ComponentState then
    Exit;
  if Assigned(FOwner) and Assigned(FOwner.Parent) and (FOwner.Parent is TrGroupBox) then
    TrGroupBox(FOwner.Parent).RefreshAutoPosition;
end;

procedure TAutoResizePos.SetColumnIndexFrom(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FColumnIndexFrom:=Value;
  if FColumnIndexTo < Value then
    FColumnIndexTo:=Value;
  UpdatePos;
end;

procedure TAutoResizePos.SetColumnIndexTo(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FColumnIndexTo:=Value;
  if Value < FColumnIndexFrom then
    FColumnIndexFrom:=Value;
  UpdatePos;
end;

procedure TAutoResizePos.SetRowIndexFrom(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FRowIndexFrom:=Value;
  if FRowIndexTo < Value then
    FRowIndexTo:=Value;
  UpdatePos;
end;

procedure TAutoResizePos.SetRowIndexTo(Value: integer);
begin
  if Value < 0 then
    Value:=0;
  FRowIndexTo:=Value;
  if Value < FRowIndexFrom then
    FRowIndexFrom:=Value;
  UpdatePos;
end;

procedure TAutoResizePos.SetLastColumn(Value: Boolean);
begin
  FLastColumn:=Value;
  UpdatePos;
end;

procedure TAutoResizePos.SetLastRow(Value: Boolean);
begin
  FLastRow:=Value;
  UpdatePos;
end;

// ==============================================================================
// TrGroupBox AutoresizeChild subcomp
constructor TAutoResizeChild.Create(AOwner: TrGroupBox);
begin
  FOwner:=AOwner;
  FColWidthsPercent:=TStringList.Create;
  FRowHeightsPercent:=TStringList.Create;
  FChildCompsPos:=TStringList.Create;
  FColSpacing:=4;
  FRowSpacing:=4;
  FBorderSpacing:=8;
  FBorderLeftAddSpacing:=0;
  AutoSetTabOrder:=true;
  FAutoResizeDelay:=100;
end;

destructor TAutoResizeChild.Destroy;
begin
  FreeAndNil(FColWidthsPercent);
  FreeAndNil(FRowHeightsPercent);
  FreeAndNil(FChildCompsPos);
  inherited;
end;

procedure TAutoResizeChild.UpdateChildComp;
begin
  if FAutoResizeActive then
    if Assigned(FOwner) then
      FOwner.RefreshAutoPosition;
end;

procedure TAutoResizeChild.SetAutoResizeActive(Value: Boolean);
begin
  FAutoResizeActive:=Value;
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetColSpacing(Value: integer);
begin
  FColSpacing:=Value;
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetRowSpacing(Value: integer);
begin
  FRowSpacing:=Value;
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetBorderSpacing(Value: integer);
begin
  FBorderSpacing:=Value;
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetBorderLeftAddSpacing(Value: integer);
begin
  FBorderLeftAddSpacing:=Value;
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetColWidthsPercent(Value: TStrings);
begin
  FColWidthsPercent.Assign(Value);
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetRowHeightsPercent(Value: TStrings);
begin
  FRowHeightsPercent.Assign(Value);
  UpdateChildComp;
  if Assigned(FOwner) then
    FOwner.Invalidate;
end;

procedure TAutoResizeChild.SetChildCompsPos(Value: TStrings);
begin
  FChildCompsPos.Assign(Value);
  UpdateChildComp;
end;

function TAutoResizeChild.GetColWidthPercent(ColIndex: integer): integer;
var
  s: string;
begin
  Result:=0;
  if ColIndex <= FColWidthsPercent.Count then
  begin
    s:=FColWidthsPercent[ColIndex - 1];
    try
      Result:=StrToInt(s);
    except
    end;
  end;
end;

function TAutoResizeChild.GetRowHeightPercent(RowIndex: integer): integer;
var
  s: string;
begin
  Result:=0;
  if RowIndex <= FRowHeightsPercent.Count then
  begin
    s:=FRowHeightsPercent[RowIndex - 1];
    try
      Result:=StrToInt(s);
    except
    end;
  end;
end;

// ==============================================================================
// TrGroupBox
constructor TrGroupBox.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  FFrameType:=rGroupBoxDefFrame;
  FFrameColor:=clWindowFrame;
  FCaptionHeight:=RosiCompConfig.rGroupBoxCfg.DefaultCaptionHeight;
  FScaledCaptionHeight:=FCaptionHeight;
  FCaptionFont:=TFont.Create;
  FCaptionFont.OnChange:=FontChanged;
  FCaptionParentFont:=true;
  FCaptionColor:=RosiCompConfig.rGroupBoxCfg.DefaultCaptionColor;
  FCaptionFontColor:=RosiCompConfig.rGroupBoxCfg.DefaultFontColor;
  FCaptionDisableFontColor:=clBtnShadow;
  FCaptionAlignment:=taCenter;
  FCaptionPosition:=cpTop;
  FGradientColorFrom:=clBtnFace;
  FGradientColorTo:=clBtnFace;
  FUseTabColor:=false;
  FMinimized:=false;
  FMinimizing:=false;
  FMinimizeAnimate:=false;
  FShowSizeGrip:=false;
  FMinBtnPicture:=TPicture.Create;
  FMaxBtnPicture:=TPicture.Create;
  FBackgroundPicture:=TPicture.Create;
  FMinBtnPicture.OnChange:=PictureChanged;
  FMaxBtnPicture.OnChange:=PictureChanged;
  FBackgroundPicture.OnChange:=PictureChanged;
  FAutoResizePos:=TAutoResizePos.Create(Self);
  FAutoResizeChild:=TAutoResizeChild.Create(Self);
  FDropFilesList:=TStringList.Create;
  FAutoEnableChildren:=true;
  FClientRectTop:=-1;
  FAutoHint:=false;
  FOriginSize:=0;
  IntTimer:=TTimer.Create(Self);
  IntTimer.Interval:=100;
  IntTimer.OnTimer:=IntTimerEvent;

  { Make control accept child controls }
  ControlStyle:=ControlStyle + [csAcceptsControls];
  // SetEnableDropFile(false);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrGroupBox.Destroy;
begin
  FreeAndNil(FCaptionFont);
  FreeAndNil(FMinBtnPicture);
  FreeAndNil(FMaxBtnPicture);
  FreeAndNil(FBackgroundPicture);
  FreeAndNil(FAutoResizePos);
  FreeAndNil(FAutoResizeChild);
  FreeAndNil(FDropFilesList);
  FreeAndNil(IntTimer);
  inherited;
end;

procedure TrGroupBox.Loaded;
begin
  inherited;
  if not(csDesigning in ComponentState) then
  begin
    if RosiCompConfig.rGroupBoxCfg.UseSimpleTabColor and ParentIsTab(Self) then
      UseTabColor:=true;
    if FUseTabColor then
      Color:=RosiCompConfig.rGroupBoxCfg.DefaultBkColor;
    // if not FCaptionParentFont then rtool.DPI_ScaleFontHeight(FCaptionFont, FScaleRatio);
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrGroupBox.CreateWnd;
begin
  inherited CreateWnd;
  { registers that window can accept dropped files }
  DragAcceptFiles(Handle, FEnableDropFile and Enabled);
end;

procedure TrGroupBox.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  FScaledCaptionHeight:=UseScale(FCaptionHeight);
  if not FCaptionParentFont then
    rtool.DPI_ScaleFontHeight(FCaptionFont, M, D);
  if Assigned(FChangeScale) then
    FChangeScale(M, D, {$IFDEF DX102P} isDpiChange, {$ELSE} false, {$ENDIF} FScaleRatio);
end;

function TrGroupBox.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

function TrGroupBox.ParentIsTab(C: TrGroupBox): Boolean;
begin
  if C.Parent is TrGroupBox then
    Result:=ParentIsTab(C.Parent as TrGroupBox)
  else
    Result:=(C.Parent is TTabSheet) or (C.Parent is TTabControl);
end;

procedure TrGroupBox.CMEnabledChanged(var Msg: TMessage);

  procedure SetEna(C: TControl);
  var
    I: integer;
  begin
    if Assigned(C) then
    begin
      if C is TPageControl then
      begin
        for I:=0 to TPageControl(C).PageCount - 1 do
          SetEna(TPageControl(C).Pages[I]);
      end
      else
      begin
        if (C is TWinControl) and (csAcceptsControls in C.ControlStyle) then
          for I:=0 to TWinControl(C).ControlCount - 1 do
            SetEna(TWinControl(C).Controls[I]);
      end;
      C.Enabled:=Enabled;
    end;
  end;

begin
  inherited;
  DragAcceptFiles(Handle, FEnableDropFile and Enabled);
  if FAutoEnableChildren then
    SetEna(Self);
  Invalidate;
end;

procedure TrGroupBox.SetEnableDropFile(Value: Boolean);
begin
  FEnableDropFile:=Value;
  DragAcceptFiles(Handle, FEnableDropFile and Enabled);
end;

procedure TrGroupBox.SetClientRectTop(Value: integer);
begin
  FClientRectTop:=Value;
  Realign;
end;

function TrGroupBox.StoreCaptionHeight: Boolean;
begin
  Result:=FCaptionHeight <> RosiCompConfig.rGroupBoxCfg.DefaultCaptionHeight;
end;

function TrGroupBox.StoreCaptionColor: Boolean;
begin
  Result:=FCaptionColor <> RosiCompConfig.rGroupBoxCfg.DefaultCaptionColor;
end;

function TrGroupBox.StoreCaptionFontColor: Boolean;
begin
  Result:=FCaptionFontColor <> RosiCompConfig.rGroupBoxCfg.DefaultFontColor;
end;

procedure TrGroupBox.SetFrameType(Value: TFrameType);
begin
  FFrameType:=Value;
  // adjust client rect by changing of size
  Width:=Width + 1;
  Width:=Width - 1;
  Invalidate;
end;

procedure TrGroupBox.SetFrameColor(Value: TColor);
begin
  FFrameColor:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionHeight(Value: byte);
begin
  FCaptionHeight:=Value;
  FScaledCaptionHeight:=UseScale(FCaptionHeight);
  // adjust client rect by changing of size
  Width:=Width + 1;
  Width:=Width - 1;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionColor(Value: TColor);
begin
  FCaptionColor:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionFontColor(Value: TColor);
begin
  FCaptionFontColor:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionDisableFontColor(Value: TColor);
begin
  FCaptionDisableFontColor:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionFont(Value: TFont);
begin
  FCaptionFont.Assign(Value);
end;

procedure TrGroupBox.SetCaptionParentFont(Value: Boolean);
begin
  if Value then
    FCaptionFont.Assign(Font);
  FCaptionParentFont:=Value;
end;

procedure TrGroupBox.SetCaptionAlignment(Value: TAlignment);
begin
  FCaptionAlignment:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetCaptionPosition(Value: TCaptionPosition);
begin
  FCaptionPosition:=Value;
  // adjust client rect by changing of size
  Width:=Width + 1;
  Width:=Width - 1;
  Invalidate;
end;

procedure TrGroupBox.FontChanged(Sender: TObject);
begin
  FCaptionParentFont:=false;
  Invalidate;
end;

function TrGroupBox.IsFontStored: Boolean;
begin
  Result:=not FCaptionParentFont;
end;

procedure TrGroupBox.CMFontChanged(var Message: TMessage);
begin
  inherited;
  if FCaptionParentFont then
  begin
    FCaptionFont.Assign(Font);
    FCaptionParentFont:=true;
  end;
end;

procedure TrGroupBox.SetGradientColorFrom(Value: TColor);
begin
  FGradientColorFrom:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetGradientColorTo(Value: TColor);
begin
  FGradientColorTo:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetGradientType(Value: TGradientType);
begin
  FGradientType:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetMinMaxBtnType(Value: TMinMaxBtnType);
begin
  FMinMaxBtnType:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetMinBtnPicture(Value: TPicture);
begin
  FMinBtnPicture.Assign(Value);
end;

procedure TrGroupBox.SetMaxBtnPicture(Value: TPicture);
begin
  FMaxBtnPicture.Assign(Value);
end;

procedure TrGroupBox.SetBackgroundPicture(Value: TPicture);
begin
  FBackgroundPicture.Assign(Value);
end;

procedure TrGroupBox.SetBackgroundPictureType(Value: TBackgroundPictureType);
begin
  FBackgroundPictureType:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetUseTabColor(Value: Boolean);
begin
  FUseTabColor:=Value;
  if FUseTabColor then
  begin
    ParentBackground:=false;
    if not(csDesigning in ComponentState) and FUseTabColor then
      Color:=RosiCompConfig.rGroupBoxCfg.DefaultBkColor;
  end;
end;

procedure TrGroupBox.SetShowSizeGrip(Value: Boolean);
begin
  FShowSizeGrip:=Value;
  Invalidate;
end;

procedure TrGroupBox.SetMinimized(Value: Boolean);
begin
  if FMinimized <> Value then
  begin
    if Value then
      DoMinimize
    else
      DoRestore;
  end;
end;

procedure TrGroupBox.SetAllChildren(SetChildrenValueType: TSetChildrenValueTypes; Value: variant; P: Pointer; ExcludeTag: integer);
  procedure ProcWinControl(WCont: TWinControl);
  var
    I: integer;
    C: TControl;
  begin
    for I:=0 to WCont.ControlCount - 1 do
    begin
      C:=WCont.Controls[I];
      if C.Tag = ExcludeTag then
        Continue;
      case SetChildrenValueType of
        chvSetVisibility:
          C.Visible:=Value;
        chvSetEnability:
          C.Enabled:=Value;
        chvSetColor:
          THackControl(C).Color:=Value;
        { chvSetDataSource:
          begin
          if C is TrDBMoveButton then
          TrDBMoveButton(C).DataSource:=P;
          if C is TrDBToolButton then
          TrDBToolButton(C).DataSource:=P;
          end;
          chvSetSpecEnability:
          begin
          if C is TrDBMoveButton then
          TrDBMoveButton(C).EnabledAll:=Value;
          if C is TrDBToolButton then
          TrDBToolButton(C).EnabledAll:=Value;
          end;
          }
      end;
      if C is TWinControl then
        ProcWinControl(C as TWinControl);
    end;
  end;

begin
  ProcWinControl(Self);
end;

procedure TrGroupBox.DoMinimize;
var
  I, s: integer;
  vert: Boolean;
  R: TRect;
begin
  if Align = alClient then
    Exit;
  FMinimizing:=true;
  try
    // save visibility of controls and hide all
    SetLength(FControlsVisibility, ControlCount);
    for I:=0 to ControlCount - 1 do
    begin
      FControlsVisibility[I]:=Controls[I].Visible;
      if Controls[I].Align <> alNone then
        Controls[I].Visible:=false;
    end;
    vert:=(FCaptionPosition in [cpLeft, cpRight]) or (Align in [alLeft, alRight]);
    // save original size
    if vert then
      FOriginSize:=Round(Width / FScaleRatio)
    else
      FOriginSize:=Round(Height / FScaleRatio);
    // change size to caption size
    s:=FScaledCaptionHeight + 1;
    if FFrameType <> ftNone then
      Inc(s, 2);
    if (FFrameType in [ftDoubleLine, ft3D]) then
      Inc(s, 2);

    if vert then
      R:=Rect(Left, Top, Left + s, Top + Height)
    else
      R:=Rect(Left, Top, Left + Width, Top + s);
    if FMinimizeAnimate then
      AnimateResize(Self, R)
    else
      if vert then
        Width:=s
      else
        Height:=s;
    FMinimized:=true;
    for I:=0 to ControlCount - 1 do
      if vert or (Controls[I].Align <> alNone) or (Controls[I].Top + Controls[I].Height > s) then
        Controls[I].Visible:=false;
    Invalidate;
  finally
    FMinimizing:=false;
  end;
  if Assigned(FOnMinimizeChange) then
    FOnMinimizeChange(Self);
end;

procedure TrGroupBox.DoRestore;
var
  I, s, smin: integer;
  vert: Boolean;
  R: TRect;
begin
  if Align = alClient then
    Exit;
  FMinimizing:=true;
  try
    vert:=(FCaptionPosition in [cpLeft, cpRight]) or (Align in [alLeft, alRight]);
    if FOriginSize < FScaledCaptionHeight + 4 then
    begin
      // original size is not known, set size according owned controls
      smin:=FScaledCaptionHeight + 16;
      for I:=0 to Self.ControlCount - 1 do
      begin
        if vert then
          s:=Controls[I].Left + Controls[I].Width
        else
          s:=Controls[I].Top + Controls[I].Height;
        if s > smin then
          smin:=s;
      end;
      FOriginSize:=smin + 8;
    end;
    FMinimized:=false;
    if vert then
      R:=Rect(Left, Top, Left + UseScale(FOriginSize), Top + Height)
    else
      R:=Rect(Left, Top, Left + Width, Top + UseScale(FOriginSize));
    if FMinimizeAnimate then
      AnimateResize(Self, R)
    else
      if vert then
        Width:=UseScale(FOriginSize)
      else
        Height:=UseScale(FOriginSize);
    // restore controls visibility
    for I:=0 to ControlCount - 1 do
    begin
      if I > High(FControlsVisibility) then
        Break;
      Controls[I].Visible:=FControlsVisibility[I];
    end;
  finally
    FMinimizing:=false;
  end;
  if Assigned(FOnMinimizeChange) then
    FOnMinimizeChange(Self);
  RefreshAutoPosition;
end;

procedure TrGroupBox.PictureChanged(Sender: TObject);
begin
  Invalidate;
end;

procedure TrGroupBox.AdjustClientRect(var Rect: TRect);
begin
  inherited AdjustClientRect(Rect);
  if Ctl3d then
    InflateRect(Rect, 1, 1);
  Rect.Top:=1 + Padding.Top;
  if FScaledCaptionHeight > 0 then
  begin
    case FCaptionPosition of
      cpTop:
        Inc(Rect.Top, FScaledCaptionHeight + 1);
      cpBottom:
        Dec(Rect.Bottom, FScaledCaptionHeight + 1);
      cpLeft:
        Inc(Rect.Left, FScaledCaptionHeight + 1);
      cpRight:
        Dec(Rect.Right, FScaledCaptionHeight + 1);
    end;
  end;
  case FFrameType of
    ftNone:
      InflateRect(Rect, 1, 1);
    ftDoubleLine, ft3D:
      InflateRect(Rect, -1, -1);
  end;
  if FClientRectTop >= 0 then
    Rect.Top:=FClientRectTop;
end;

procedure TrGroupBox.Paint;
var
  Rw, Rclient, Rcapt, Rcaptlabel, RTemp: TRect;
  P: TPoint;
  Flags: Longint;
  I, w, h, cw, ch, sc3, sc4: integer;
begin
  // recalculate rectangles - whole, client and caption rect
  // if Parent = nil then  Exit;
  sc3:=UseScale(3);
  sc4:=UseScale(4);

  Rw:=Rect(0, 0, Width, Height);
  Rclient:=Rw;
  if FFrameType <> ftNone then
    InflateRect(Rclient, -1, -1);
  if FFrameType in [ftDoubleLine, ft3D] then
    InflateRect(Rclient, -1, -1);
  Rcapt:=Rclient;

  if not FMinimized then
  begin
    case FCaptionPosition of
      cpTop:
        begin
          Rclient.Top:=Rclient.Top + FScaledCaptionHeight;
          Rcapt.Bottom:=Rcapt.Top + FScaledCaptionHeight + 1;
        end;
      cpBottom:
        begin
          Rclient.Bottom:=Rclient.Bottom - FScaledCaptionHeight;
          Rcapt.Top:=Rcapt.Bottom - (FScaledCaptionHeight + 1);
        end;
      cpLeft:
        begin
          Rclient.Left:=Rclient.Left + FScaledCaptionHeight;
          Rcapt.Right:=Rcapt.Left + FScaledCaptionHeight + 1;
        end;
      cpRight:
        begin
          Rclient.Right:=Rclient.Right - FScaledCaptionHeight;
          Rcapt.Left:=Rcapt.Right - (FScaledCaptionHeight + 1);
        end;
    end;

    // draw background gradient
    if FGradientType <> TGradientType.gtNone then
      if (FBackgroundPictureType in [bkpNone, bkpCenter]) or (FBackgroundPicture.Width = 0) then
        DrawGradient(Canvas, Rclient, FGradientColorFrom, FGradientColorTo, FGradientType = gtVertical, cGradientStandardSteps);
    // draw background picture
    case FBackgroundPictureType of
      bkpCenter, bkpStretch:
        rpictool.DrawPictureCenter(Canvas, Rclient, FBackgroundPicture, FBackgroundPictureType = bkpStretch, false);
      bkpTile:
        rpictool.DrawPictureTiles(Canvas, Rclient, FBackgroundPicture, false);
    end;
  end;

  with Canvas do
  begin
    // init font and colors
    Font.Assign(FCaptionFont);
    if FCaptionFontColor <> clNone then
      Font.Color:=FCaptionFontColor;
    if not Enabled then
      Font.Color:=FCaptionDisableFontColor;
    Brush.Color:=FCaptionColor;
    Pen.Color:=FCaptionColor;
    if FScaledCaptionHeight > 0 then
    begin
      // draw caption rect
      Rectangle(Rcapt);
      // draw caption text
      Pen.Color:=FCaptionFont.Color;
      if (FCaptionPosition in [cpLeft, cpRight]) or (FMinimized and (Align in [alLeft, alRight])) then
      begin
        // rotate text
        w:=TextWidth(Text) + 1;
        h:=TextHeight(Text) + 1;
        FontGetRotated(Font, 900);
        // find possition
        P.X:=Rcapt.Left + (Rcapt.Right - Rcapt.Left - h) div 2 - 1;
        P.Y:=Rcapt.Bottom - sc4;
        case FCaptionAlignment of
          taRightJustify:
            begin
              P.Y:=Rcapt.Top + w + sc4;
              if FMinMaxBtnType <> mmbNone then
                Inc(P.Y, FScaledCaptionHeight);
            end;
          taCenter:
            P.Y:=Rcapt.Top + (Rcapt.Bottom - Rcapt.Top - w) div 2 + w;
        end;
        if P.Y > Rcapt.Bottom - sc4 then
          P.Y:=Rcapt.Bottom - sc4;
        TextOut(P.X, P.Y, Text);
        // Font.Assign(FCaptionFont);
      end
      else
      begin
        Rcaptlabel:=Rcapt;
        Flags:=DT_VCENTER + DT_SINGLELINE;
        case FCaptionAlignment of
          taRightJustify:
            begin
              Flags:=Flags + DT_RIGHT;
              if FMinMaxBtnType <> mmbNone then
                Dec(Rcaptlabel.Right, FScaledCaptionHeight);
            end;
          taCenter:
            Flags:=Flags + DT_CENTER;
        end;
        Flags:=DrawTextBiDiModeFlags(Flags);
        InflateRect(Rcaptlabel, -sc4, 0);
        DrawText(Handle, PChar(Text), Length(Text), Rcaptlabel, Flags);
      end;

      if FMinMaxBtnType <> mmbNone then
      begin
        // calculate minmax rect
        Rcapt.Left:=Rcapt.Right - FScaledCaptionHeight - 1;
        Rcapt.Bottom:=Rcapt.Top + FScaledCaptionHeight + 1;
        // OffsetRect(Rcapt,-+,0);
        if FMinMaxBtnType <> mmbPicture then
          while Rcapt.Right - Rcapt.Left > UseScale(15) do
            if (Rcapt.Bottom - Rcapt.Top) mod 2 = 0 then
            begin
              Rcapt.Bottom:=Rcapt.Bottom - 1;
              Rcapt.Right:=Rcapt.Right - 1;
            end
            else
            begin
              Rcapt.Top:=Rcapt.Top + 1;
              Rcapt.Left:=Rcapt.Left + 1;
            end;

        // draw minmax button
        FBtnRect:=Rcapt;
        Pen.Color:=FCaptionColor;
        Brush.Color:=Font.Color;
        P.X:=Rcapt.Left + (Rcapt.Right - Rcapt.Left) div 2;
        P.Y:=Rcapt.Top + (Rcapt.Bottom - Rcapt.Top) div 2;

        case FMinMaxBtnType of
          mmbRect:
            begin
              Rectangle(Rcapt);
              InflateRect(Rcapt, -sc3, -sc3);
              BiDi_FrameRect(Canvas, Rcapt, FCaptionColor);
              if FMinimized then
                Rectangle(Rect(Rcapt.Left, Rcapt.Top + sc4, Rcapt.Right, Rcapt.Bottom - 1))
              else
                Rectangle(Rect(Rcapt.Left, Rcapt.Top + 1, Rcapt.Right, Rcapt.Bottom - sc4));
            end;
          mmbRound:
            begin
              Ellipse(Rcapt);
              Pen.Width:=2;
              if not FMinimized then
              begin
                MoveTo(P.X - sc3, P.Y - 1);
                LineTo(P.X, P.Y - sc4);
                LineTo(P.X + sc3, P.Y - 1);
                MoveTo(P.X - sc3, P.Y + sc3);
                LineTo(P.X, P.Y - sc3 + sc3);
                LineTo(P.X + sc3, P.Y + sc3);
              end
              else
              begin
                MoveTo(P.X - sc3, P.Y - sc3);
                LineTo(P.X, P.Y);
                LineTo(P.X + sc3, P.Y - sc3);
                MoveTo(P.X - sc3, P.Y + 1);
                LineTo(P.X, P.Y + sc4);
                LineTo(P.X + sc3, P.Y + 1);
              end;
            end;
          mmbPicture:
            try
              if FMinimized then
              begin
                DrawPictureCenter(Canvas, Rcapt, FMinBtnPicture, true, true);
              end
              else
              begin
                DrawPictureCenter(Canvas, Rcapt, FMaxBtnPicture, true, true);
              end;
            except
            end;
        end;
      end;
    end
    else
      if Caption <> '' then
      begin
        // caption in panel
        Rcaptlabel:=Rclient;
        Flags:=DT_VCENTER + DT_SINGLELINE;
        case FCaptionAlignment of
          taRightJustify:
            Flags:=Flags + DT_RIGHT;
          taCenter:
            Flags:=Flags + DT_CENTER;
        end;
        Flags:=DrawTextBiDiModeFlags(Flags);
        InflateRect(Rcaptlabel, -sc4, 0);
        DrawText(Handle, PChar(Text), Length(Text), Rcaptlabel, Flags);
      end;

    // show size grip
    if FShowSizeGrip and not FMinimized and (Owner is TForm) and (TForm(Owner).WindowState <> wsMaximized) then
    begin
      RTemp:=Rclient;
      RTemp.Left:=RTemp.Right - UseScale(12);
      RTemp.Top:=RTemp.Bottom - UseScale(12);
      {
        Pen.Style:=psSolid;
        Pen.Color:=Self.FFrameColor;
        RTemp.Left:=RTemp.Right - UseScale(12);
        RTemp.Top:=RTemp.Bottom - UseScale(12);
        Pen.Width:=1;
        MoveTo(RTemp.Left, RTemp.Bottom);
        LineTo(RTemp.Right, RTemp.Top);
        MoveTo(RTemp.Left + sc4, RTemp.Bottom);
        LineTo(RTemp.Right, RTemp.Top + sc4);
        MoveTo(RTemp.Left + 2 * sc4, RTemp.Bottom);
        LineTo(RTemp.Right, RTemp.Top + 2 * sc4);
        }
      Brush.Color:=ColorAdjustLuma(FFrameColor, 80, false);
      for h:=1 to 3 do
        for w:=1 to h do
        begin
          RTemp.Left:=Rclient.Right - w * UseScale(3);
          RTemp.Right:=RTemp.Left + UseScale(2);
          RTemp.Top:=Rclient.Bottom - (4 - h) * UseScale(3);
          RTemp.Bottom:=RTemp.Top + UseScale(2);
          FillRect(RTemp);
        end;
    end;

    // draw frame
    Brush.Color:=FFrameColor;
    Pen.Width:=1;
    case FFrameType of
      ftSingleLine:
        BiDi_FrameRect(Canvas, Rw);
      ftDoubleLine:
        begin
          BiDi_FrameRect(Canvas, Rw);
          InflateRect(Rw, -1, -1);
          BiDi_FrameRect(Canvas, Rw);
        end;
      ft3D:
        begin
          BiDi_FrameRect(Canvas, Rw, clBtnHighlight);
          Inc(Rw.Left);
          Inc(Rw.Top);
          BiDi_FrameRect(Canvas, Rw);
          OffsetRect(Rw, -1, -1);
          BiDi_FrameRect(Canvas, Rw, clBtnShadow);
        end;
    end;

    // draw aux line
    if (csDesigning in ComponentState) and FAutoResizeChild.FAutoResizeActive and not FMinimized then
    begin
      cw:=Rclient.Right - Rclient.Left - 2 * FAutoResizeChild.FBorderSpacing - FAutoResizeChild.FBorderLeftAddSpacing -
        (FAutoResizeChild.FColWidthsPercent.Count - 1) * FAutoResizeChild.FColSpacing;
      w:=Rclient.Left + FAutoResizeChild.FBorderSpacing + FAutoResizeChild.FBorderLeftAddSpacing + FAutoResizeChild.ColSpacing div 2;
      for I:=1 to FAutoResizeChild.FColWidthsPercent.Count - 1 do
      begin
        w:=w + FAutoResizeChild.GetColWidthPercent(I) * cw div 100;
        if (w > 0) and (w < Width) then
        begin
          Brush.Color:=clWhite;
          Pen.Width:=1;
          Pen.Color:=clMedGray;
          Pen.Style:=psDot;
          MoveTo(w, Rclient.Bottom);
          LineTo(w, Rclient.Top);
        end;
        w:=w + FAutoResizeChild.ColSpacing;
      end;

      ch:=Rclient.Bottom - Rclient.Top - 2 * FAutoResizeChild.FBorderSpacing - (FAutoResizeChild.FRowHeightsPercent.Count - 1) * FAutoResizeChild.FRowSpacing;
      h:=Rclient.Top + FAutoResizeChild.FBorderSpacing + FAutoResizeChild.RowSpacing div 2;
      for I:=1 to FAutoResizeChild.FRowHeightsPercent.Count - 1 do
      begin
        h:=h + FAutoResizeChild.GetRowHeightPercent(I) * ch div 100;
        if (h > 0) and (h < Height) then
        begin
          Brush.Color:=clWhite;
          Pen.Width:=1;
          Pen.Color:=clMedGray;
          Pen.Style:=psDot;
          MoveTo(Rclient.Left, h);
          LineTo(Rclient.Right, h);
        end;
        h:=h + FAutoResizeChild.RowSpacing;
      end;
    end;
  end;

  if Assigned(FOnDrawControl) then
    FOnDrawControl(Self, Rect(0, 0, Width, Height));
end;

procedure TrGroupBox.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  Change: Boolean;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  inherited;
  if ssDouble in Shift then
    Change:=PtInRect(Rect(0, 0, Width, FScaledCaptionHeight), Point(X, Y))
  else
    Change:=PtInRect(FBtnRect, Point(X, Y));
  if Change and (FMinMaxBtnType <> mmbNone) then
    if FMinimized then
      DoRestore
    else
      DoMinimize;
  // self.Color:=RosiCompConfig.rGroupBoxCfg.DefaultBkColor;
end;

procedure TrGroupBox.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
end;

procedure TrGroupBox.Resize;
begin
  if FRefPosImmediately or (FAutoResizeChild.FAutoResizeDelay <= 0) then
    RefreshAutoPosition
  else
  begin
    IntTimer.Enabled:=false;
    IntTimer.Interval:=FAutoResizeChild.FAutoResizeDelay;
    IntTimer.Enabled:=true;
  end;
  inherited;
end;

procedure TrGroupBox.IntTimerEvent(Sender: TObject);
begin
  IntTimer.Enabled:=false;
  RefreshAutoPosition;
end;

function TrGroupBox.SetAutoResizeChildComponentPos(CompName: string; VerticalAlign: TrGroupBoxVerticalAlignment;
  ColumnIndexFrom, ColumnIndexTo, ReduceWidth: integer; HorizontalAlign: TrGroupBoxHorizontalAlignment; ReduceHeight: integer): string;
var
  props: string;
  procedure AddProps(s: string); overload;
  begin
    if props <> '' then
      props:=props + rGroupBoxDefSeparator;
    props:=props + s;
  end;
  procedure AddProps(I: integer); overload;
  begin
    AddProps(IntToStr(I));
  end;

begin
  AddProps(rGroupBoxVerticalAlignmentStr[VerticalAlign]);
  AddProps(ColumnIndexFrom);
  AddProps(ColumnIndexTo);
  AddProps(ReduceWidth);
  AddProps(rGroupBoxHorizontalAlignmentStr[HorizontalAlign]);
  AddProps(ReduceHeight);
  if CompName <> '' then
  begin
    FAutoResizeChild.FChildCompsPos.Values[CompName]:=props;
    RefreshAutoPosition;
  end;
  Result:=props;
end;

function TrGroupBox.GetAutoResizeChildComponentPos(CompName: string; var VerticalAlign: TrGroupBoxVerticalAlignment; var ColumnIndexFrom, ColumnIndexTo,
  ReduceWidth: integer; var HorizontalAlign: TrGroupBoxHorizontalAlignment; var ReduceHeight: integer): Boolean;
var
  props, s: string;
  va: TrGroupBoxVerticalAlignment;
  ha: TrGroupBoxHorizontalAlignment;
begin
  Result:=false;
  // if compname starts with * then extract values from the input, in other case from FChildCompsPos property
  if copy(CompName, 1, 1) = '*' then
    props:=copy(CompName, 2, Length(CompName))
  else
    props:=FAutoResizeChild.FChildCompsPos.Values[CompName];
  if props = '' then
    Exit;
  VerticalAlign:=gbvaNothing;
  s:=rtool.FindStrPart(props, rGroupBoxDefSeparator, 1, true);
  for va:= Low(TrGroupBoxVerticalAlignment) to High(TrGroupBoxVerticalAlignment) do
    if SameText(rGroupBoxVerticalAlignmentStr[va], s) then
      VerticalAlign:=va;
  ColumnIndexFrom:=StrToIntDef(rtool.FindStrPart(props, rGroupBoxDefSeparator, 2, true), 0);
  ColumnIndexTo:=StrToIntDef(rtool.FindStrPart(props, rGroupBoxDefSeparator, 3, true), 0);
  ReduceWidth:=StrToIntDef(rtool.FindStrPart(props, rGroupBoxDefSeparator, 4, true), 0);
  HorizontalAlign:=gbhaNothing;
  s:=rtool.FindStrPart(props, rGroupBoxDefSeparator, 5, true);
  for ha:= Low(TrGroupBoxHorizontalAlignment) to High(TrGroupBoxHorizontalAlignment) do
    if SameText(rGroupBoxHorizontalAlignmentStr[ha], s) then
      HorizontalAlign:=ha;
  ReduceHeight:=StrToIntDef(rtool.FindStrPart(props, rGroupBoxDefSeparator, 6, true), 0);
  Result:=true;
end;

procedure TrGroupBox.SetNewBounds(Bounds: TRect);
begin
  FRefPosImmediately:=true;
  BoundsRect:=Bounds;
end;

procedure TrGroupBox.RefreshAutoPosition;
var
  I, j, cw, ch, col1, col2, cols, row1, row2, rows, resw, resh: integer;
  CR, R: TRect;
  GB: TrGroupBox;
  C: TControl;
  cn: string;
  va: TrGroupBoxVerticalAlignment;
  ha: TrGroupBoxHorizontalAlignment;
  vert: Boolean;

  function CheckMargin(Margin: integer): integer;
  begin
    // margin 1020 means -20
    if Margin > 1000 then
      Result:=1000 - Margin
    else
      Result:=Margin
  end;

begin
  FRefPosImmediately:=false;
  IntTimer.Enabled:=false;
  if not FAutoResizeChild.FAutoResizeActive then
    Exit;
  cols:=FAutoResizeChild.FColWidthsPercent.Count;
  rows:=FAutoResizeChild.FRowHeightsPercent.Count;
  CR:=GetClientRect;
  AdjustClientRect(CR);

  // align child GroupBox according their AutoresizePos properties
  for I:=0 to ControlCount - 1 do
    if Controls[I] is TrGroupBox then
    begin
      GB:=Controls[I] as TrGroupBox;
      col1:=GB.FAutoResizePos.FColumnIndexFrom;
      col2:=GB.FAutoResizePos.FColumnIndexTo;
      if col2 > cols then
        col2:=cols;
      row1:=GB.FAutoResizePos.FRowIndexFrom;
      row2:=GB.FAutoResizePos.FRowIndexTo;
      if (row1 > 0) and (row2 > rows) then
        row2:=rows;

      R:=GB.BoundsRect;
      if (col1 in [1 .. cols]) and (FAutoResizeChild.GetColWidthPercent(col1) > 0) then
      begin
        cw:=CR.Right - CR.Left - 2 * UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(FAutoResizeChild.FBorderLeftAddSpacing) - (cols - 1) * UseScale
          (FAutoResizeChild.FColSpacing);
        // calc left
        R.Left:=CR.Left + UseScale(FAutoResizeChild.FBorderSpacing) + UseScale(FAutoResizeChild.FBorderLeftAddSpacing) + (col1 - 1) * UseScale
          (FAutoResizeChild.ColSpacing);
        for j:=1 to col1 - 1 do
          R.Left:=R.Left + FAutoResizeChild.GetColWidthPercent(j) * cw div 100;
        // calc right
        R.Right:=R.Left;
        if col2 > col1 then
          R.Right:=R.Right + (col2 - col1) * UseScale(FAutoResizeChild.ColSpacing);
        for j:=col1 to col2 do
          R.Right:=R.Right + FAutoResizeChild.GetColWidthPercent(j) * cw div 100;
      end;
      if GB.FAutoResizePos.LastColumn then
      begin
        // expand to right, use ColumnIndexTo as Border
        R.Right:=Max(R.Left + 10, CR.Right - UseScale(FAutoResizeChild.FBorderSpacing) - col2);
      end
      else
        if (col1 = 0) and (col2 > cols) then
        begin
          // move to right
          cw:=R.Right - R.Left;
          R.Right:=CR.Right - UseScale(FAutoResizeChild.FBorderSpacing);
          R.Left:=R.Right - cw;
        end;

      if (row1 in [1 .. rows]) and (FAutoResizeChild.GetRowHeightPercent(row1) > 0) then
      begin
        ch:=CR.Bottom - CR.Top - 2 * UseScale(FAutoResizeChild.FBorderSpacing) - (rows - 1) * UseScale(FAutoResizeChild.FRowSpacing);
        // calc top
        R.Top:=CR.Top + UseScale(FAutoResizeChild.FBorderSpacing) + (row1 - 1) * UseScale(FAutoResizeChild.RowSpacing);
        for j:=1 to row1 - 1 do
          R.Top:=R.Top + FAutoResizeChild.GetRowHeightPercent(j) * ch div 100;
        // calc bottom
        R.Bottom:=R.Top;
        if row2 > row1 then
          R.Bottom:=R.Bottom + (row2 - row1) * UseScale(FAutoResizeChild.RowSpacing);
        for j:=row1 to row2 do
          R.Bottom:=R.Bottom + FAutoResizeChild.GetRowHeightPercent(j) * ch div 100;
      end;

      if GB.FAutoResizePos.LastRow then
      begin
        // expand to bottom, use RowIndexTo as Border
        R.Bottom:=Max(R.Top + 10, CR.Bottom - UseScale(FAutoResizeChild.FBorderSpacing) - row2);
      end
      else
        if (row1 = 0) and (row2 > rows) then
        begin
          // move to bottom
          ch:=R.Bottom - R.Top;
          R.Bottom:=CR.Bottom - UseScale(FAutoResizeChild.FBorderSpacing);
          R.Top:=R.Bottom - ch;
        end;

      if GB.Minimized then
      begin
        // only set original size and resize one size (vert or horiz)
        vert:=Align in [alLeft, alRight];
        if vert then
        begin
          GB.FOriginSize:=R.Right - R.Left;
          GB.Top:=R.Top;
          GB.Height:=R.Bottom - R.Top;
        end
        else
        begin
          GB.FOriginSize:=R.Bottom - R.Top;
          GB.Left:=R.Left;
          GB.Width:=R.Right - R.Left;
        end;
      end
      else
        GB.SetNewBounds(R);
    end;

  // align other child components
  { CompName=VerAlignStyle;ColIndex;ColCount;ReduceWidth;HorAlignStyle;ReduceHeight

    VerAlignStyle:
    0 .. no auto alignment
    C .. use column definition
    E .. keep left possition and expand to right border of parent
    L .. keep original size and move to left border of parent
    R .. keep original size and move to right border of parent

    HorAlignStyle:
    E .. keep top possition and expand to bottom border of parent
    B .. keep original size and move to bottom border of parent
    }
  if not FMinimized and not FMinimizing then
    for I:=0 to FAutoResizeChild.FChildCompsPos.Count - 1 do
    begin
      cn:=FAutoResizeChild.FChildCompsPos.Names[I];
      C:=FindChildControl(cn);
      if not Assigned(C) then
      begin
        for j:=0 to ControlCount - 1 do
          if CompareText(Controls[j].Name, cn) = 0 then
          begin
            C:=Controls[j];
            Break;
          end;
      end;
      if Assigned(C) then
      begin
        if FAutoResizeChild.FAutoSetTabOrder and (C is TWinControl) then
        begin
          if TWinControl(C).TabOrder <> I then
            TWinControl(C).TabOrder:=I;
        end;
        GetAutoResizeChildComponentPos(cn, va, col1, col2, resw, ha, resh);
        R:=C.BoundsRect;
        case va of
          gbvaNothing:
            ;
          gbvaColumn, gbvaColumnLeft, gbvaColumnRight:
            begin
              if (col1 = 0) or (col2 = 0) or (col1 > col2) then
                Continue;
              cw:=CR.Right - CR.Left - 2 * UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(FAutoResizeChild.FBorderLeftAddSpacing) - (col2 - 1) * UseScale
                (FAutoResizeChild.FColSpacing);
              R.Left:=CR.Left + UseScale(FAutoResizeChild.FBorderSpacing) + UseScale(FAutoResizeChild.FBorderLeftAddSpacing) + (col1 - 1) * UseScale
                (FAutoResizeChild.ColSpacing);
              R.Left:=R.Left + cw * (col1 - 1) div col2;
              R.Right:=R.Left + cw div col2 - UseScale(resw);
              if va = gbvaColumnLeft then
                R.Right:=R.Left + C.Width;
              if va = gbvaColumnRight then
                R.Left:=R.Right - C.Width;
            end;
          gbvaExpand:
            begin
              R.Right:=CR.Right - UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(resw)
            end;
          gbvaLeft:
            begin
              R.Left:=CR.Left + UseScale(FAutoResizeChild.FBorderSpacing) + UseScale(FAutoResizeChild.FBorderLeftAddSpacing);
              R.Right:=R.Left + C.Width;
            end;
          gbvaRight:
            begin
              R.Left:=CR.Right - UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(resw) - C.Width;
              R.Right:=R.Left + C.Width;
            end;
        end;

        case ha of
          gbhaNothing:
            ;
          gbhaExpand:
            begin
              R.Bottom:=CR.Bottom - UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(resh);
            end;
          gbhaBottom:
            begin
              R.Top:=CR.Bottom - UseScale(FAutoResizeChild.FBorderSpacing) - UseScale(resh) - C.Height;
              R.Bottom:=R.Top + C.Height;
            end;
        end;

        if C.AlignWithMargins then
        begin
          if va <> gbvaNothing then
          begin
            R.Left:=R.Left + CheckMargin(C.Margins.Left);
            if va in [gbvaLeft, gbvaColumnLeft] then
              R.Right:=R.Right + CheckMargin(C.Margins.Left);
            R.Right:=R.Right - CheckMargin(C.Margins.Right);
            if va in [gbvaRight, gbvaColumnRight] then
              R.Left:=R.Left - CheckMargin(C.Margins.Right);
          end;
          if ha <> gbhaNothing then
            R.Bottom:=R.Bottom - CheckMargin(C.Margins.Bottom);
          if ha = gbhaBottom then
            R.Top:=R.Top + CheckMargin(C.Margins.Top);
        end;
        if (R.Bottom > R.Top) and (R.Right > R.Left) then
          C.BoundsRect:=R;
      end;
    end;
  if Assigned(FOnRefreshAutoPosition) then
    FOnRefreshAutoPosition(Self);
end;

procedure TrGroupBox.WMDropFiles(var Msg: TMessage);
var
  DropH: THandle; { Saves handle used for dragging and dropping }
  NameLen: Word; { Length of buffer required for file name }
  NumDropped: Word; { Number of files dropped }
  PFileName: array [0 .. 255] of Char; { PString to hold file name }
  I: integer; { Loop control for scanning file names }
begin
  DropH:=Msg.WParam;
  try
    FDropFilesList.Clear;
    NumDropped:=DragQueryFile(DropH, cardinal(-1), nil, 0);
    for I:=0 to NumDropped - 1 do
    begin
      NameLen:=DragQueryFile(DropH, I, nil, 0);
      DragQueryFile(DropH, I, PFileName, NameLen + 1);
      FDropFilesList.Add(StrPas(PFileName));
    end;
    DragQueryPoint(DropH, FDropPoint);
  finally
    DragFinish(DropH);
  end;
  DropFiles;
end;

procedure TrGroupBox.DropFiles;
begin
  if Assigned(FOnDropFiles) then
    FOnDropFiles(Self, FDropPoint.X, FDropPoint.Y);
end;

function TrGroupBox.GetDropFilesCount: integer;
begin
  Result:=FDropFilesList.Count;
end;

function TrGroupBox.GetDropFileNames(I: integer): string;
begin
  Result:=FDropFilesList[I];
end;

function TrGroupBox.GetDropControl: TControl;
var
  I: integer;
begin
  Result:=nil;
  for I:=0 to ControlCount - 1 do
  begin
    { Check if current control was under mouse }
    if PtInRect(Controls[I].BoundsRect, Point(FDropPoint.X, FDropPoint.Y)) then
    begin
      Result:=Controls[I];
      Break;
    end;
  end;
end;

function IsMouseOverControl(const ctrl: TControl): Boolean;
var
  sr: TRect;
  p1, p2: TPoint;
begin
  p1:=ctrl.ClientToScreen(Point(0, 0));
  p2:=ctrl.ClientToScreen(Point(ctrl.Width, ctrl.Height));
  sr:=Rect(p1.X, p1.Y, p2.X, p2.Y);
  Result:=PtInRect(sr, Mouse.CursorPos);
end;

procedure TrGroupBox.CheckAutoMinimize;
var
  MouseIn: Boolean;
  delay: cardinal;
begin
  if (FAutoMinimizeTime = 0) and (FAutoMaximizeTime = 0) then
    Exit;
  MouseIn:=IsMouseOverControl(Self);
  if (FMinimized and MouseIn) or (not FMinimized and not MouseIn) then
  begin
    if LastTickCount = 0 then
      LastTickCount:=GetTickCount
    else
    begin
      if FMinimized then
        delay:=FAutoMaximizeTime
      else
        delay:=FAutoMinimizeTime;
      if GetTickCount - LastTickCount > delay then
      begin
        Minimized:=not FMinimized;
        LastTickCount:=0;
      end;
    end;
  end
  else
    LastTickCount:=0;
end;

function TrGroupBox.ExecuteAction(Action: TBasicAction): Boolean;
begin
  if FAutoHint and not(csDesigning in ComponentState) and (Action is THintAction) then
  begin
    Caption:=THintAction(Action).Hint;
    Result:=true;
  end
  else
    Result:= inherited ExecuteAction(Action);
end;

// ==============================================================================
// TrImageEx Boundlabel
constructor TBoundLabel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FImage:=TrImageEx(AOwner);
  SetSubComponent(true);
  Name:='SubLabel';
  Transparent:=true;
  Alignment:=taCenter;
  Layout:=tlCenter;
  // FBoundLabel.Name:=Self.Name+'_BLabel';
  Caption:='';
end;

procedure TBoundLabel.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  if Assigned(FImage.OnMouseDown) then
    FImage.OnMouseDown(Owner, Button, Shift, X, Y);
end;

procedure TBoundLabel.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  if Assigned(FImage.OnMouseUp) then
    FImage.OnMouseUp(Owner, Button, Shift, X, Y);
end;

procedure TBoundLabel.MouseMove(Shift: TShiftState; X, Y: integer);
begin
  if Assigned(FImage.OnMouseMove) then
    FImage.OnMouseMove(Owner, Shift, X, Y);
end;

procedure TBoundLabel.SetPosition(Value: TLabelPos);
begin
  FPosition:=Value;
  UpdatePos;
end;

procedure TBoundLabel.AdjustBounds;
begin
  inherited;
  UpdatePos;
end;

procedure TBoundLabel.UpdatePos;
begin
  case FPosition of
    lpInside:
      SetBounds((FImage.Width - Width) div 2 + FImage.Left, (FImage.Height - Height) div 2 + FImage.Top, Width, Height);
    lpTopLeft:
      SetBounds(FImage.Left, FImage.Top - Height - BoundLabelSpacing, Width, Height);
    lpTopCenter:
      SetBounds((FImage.Width - Width) div 2 + FImage.Left, FImage.Top - Height - BoundLabelSpacing, Width, Height);
    lpTopRight:
      SetBounds(FImage.Left + FImage.Width - Width, FImage.Top - Height - BoundLabelSpacing, Width, Height);
    lpBottomLeft:
      SetBounds(FImage.Left, FImage.Top + FImage.Height + BoundLabelSpacing, Width, Height);
    lpBottomCenter:
      SetBounds((FImage.Width - Width) div 2 + FImage.Left, FImage.Top + FImage.Height + BoundLabelSpacing, Width, Height);
    lpBottomRight:
      SetBounds(FImage.Left + FImage.Width - Width, FImage.Top + FImage.Height + BoundLabelSpacing, Width, Height);
  end;
end;

// ==============================================================================
// TrImageEx
constructor TrImageEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FBoundLabel:=rImprovedComps.TBoundLabel.Create(Self);
  FBoundLabel.FreeNotification(Self);
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrImageEx.Destroy;
begin
  inherited Destroy;
end;

procedure TrImageEx.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrImageEx.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FBoundLabel) then
    FBoundLabel.Enabled:=Enabled;
end;

procedure TrImageEx.CMVisibleChanged(var Message: TMessage);
begin
  inherited;
  if Assigned(FBoundLabel) then
    FBoundLabel.Visible:=Visible;
end;

procedure TrImageEx.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = FBoundLabel) and (Operation = opRemove) then
    FBoundLabel:=nil;
end;

procedure TrImageEx.SetBounds(ALeft, ATop, AWidth, AHeight: integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  if Assigned(FBoundLabel) then
    FBoundLabel.UpdatePos;
  if SaveAsShape then
    DrawShape;
end;

procedure TrImageEx.SetParent(AParent: TWinControl);
begin
  inherited SetParent(AParent);
  if Assigned(FBoundLabel) then
    FBoundLabel.Parent:=AParent;
end;

procedure TrImageEx.Paint;
begin
  inherited;
  // if csDesigning in ComponentState then
end;

procedure TrImageEx.DrawShape;
var
  B: TBitmap;
  R: TRect;
begin
  B:=TBitmap.Create;
  try
    B.Width:=Width;
    B.Height:=Height;
    B.Transparent:=true;
    with B.Canvas do
    begin
      R:=Rect(0, 0, B.Width, B.Height);
      Brush.Color:=BkColor;
      B.TransparentColor:=BkColor;
      FillRect(R);

      Pen.Color:=PenColor;
      Pen.Width:=PenWidth;
      if PenWidth < 1 then
        Pen.Color:=FillColor;
      Brush.Color:=FillColor;

      InflateRect(R, -(PenWidth + 1) div 2, -(PenWidth + 1) div 2);
      case ShapeType of
        stRect:
          Rectangle(R);
        stRoundRect:
          RoundRect(R.Left, R.Top, R.Right, R.Bottom, B.Width div 3, B.Height div 3);
        stElipse:
          Ellipse(R);
        stHalfElipseLeft:
          begin
            Ellipse(R.Left, R.Top, R.Left + (R.Right - R.Top) * 2, R.Bottom);
            MoveTo(R.Right, R.Top + 1);
            LineTo(R.Right, R.Bottom - 1);
          end;
        stHalfElipseRight:
          begin
            Ellipse(R.Left - (R.Right - R.Top), R.Top, R.Right, R.Bottom);
            MoveTo(R.Left, R.Top + 1);
            LineTo(R.Left, R.Bottom - 1);
          end;
        stHalfElipseTop:
          begin
            Ellipse(R.Left, R.Top, R.Right, R.Top + (R.Bottom - R.Top) * 2);
            MoveTo(R.Left + 1, R.Bottom);
            LineTo(R.Right - 1, R.Bottom);
          end;
        stHalfElipseBottom:
          begin
            Ellipse(R.Left, R.Top - (R.Bottom - R.Top), R.Right, R.Bottom);
            MoveTo(R.Left + 1, R.Top);
            LineTo(R.Right - 1, R.Top);
          end;
      end;
    end;

    Picture.Assign(B);
    SaveAsShape:=true;
  finally
    FreeAndNil(B);
  end;
end;

procedure TrImageEx.SetShape(AShapeType: rImprovedComps.TShapeType; APenColor, AFillColor, ABkColor: TColor; APenWidth: integer);
begin
  ShapeType:=AShapeType;
  PenColor:=APenColor;
  FillColor:=AFillColor;
  BkColor:=ABkColor;
  PenWidth:=APenWidth;
  DrawShape;
end;

procedure TrImageEx.AssignShape(Source: TrImageEx);
begin
  if Assigned(Source) then
    SetShape(Source.ShapeType, Source.PenColor, Source.FillColor, Source.BkColor, Source.PenWidth);
end;

procedure TrImageEx.AssignPicture(P: TPicture; AsBMP: Boolean);
begin
  if Assigned(P) then
    Self.Picture.Assign(P);
  SaveAsShape:=false;
  SaveAsBMP:=AsBMP;
end;

// ==============================================================================
// TrMemoEx
constructor TrMemoEx.Create(AOwner: TComponent);
begin
  inherited;
  FOptionsEx:=[moAllowSetFont, moAllowLoadSaveFile];
  FHighlightWords:=TStringList.Create;
  FHighlightWords.OnChange:=HighlightWordsChanged;
  // FHighlightFromStr:='\';
  // FHighlightToStr:=' ';
  FHighlightedFontColor:=Font.Color;
  FHighlightedBrushColor:=clMoneyGreen;
  FHighlightFromToEx:=TStringList.Create;
  FHighlightFromToEx.OnChange:=HighlightWordsChanged;
  FPopupMenu:=TPopupMenu.Create(Self);
  FPopupMenu.Name:='pmSearch';
  DefineEditPopupMenuItems(FPopupMenu, PopupMenuClick);
  PopupMenu:=FPopupMenu;
  FPopupMenu.OnPopup:=PopupMenuOpen;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrMemoEx.Destroy;
begin
  FreeAndNil(FHighlightWords);
  FreeAndNil(FHighlightFromToEx);
  inherited;
end;

procedure TrMemoEx.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrMemoEx.LoadSBPos;
begin
  if HandleAllocated then
  begin
    FRowIncrement:=-GetScrollPos(Handle, SB_VERT);
    FColIncrement:=-GetScrollPos(Handle, SB_HORZ);
  end;
end;

procedure TrMemoEx.SetHighlightWords(Value: TStringList);
begin
  FHighlightWords.Assign(Value);
  Refresh;
end;

procedure TrMemoEx.SetHighlightWholeWords(Value: Boolean);
begin
  FHighlightWholeWords:=Value;
  Refresh;
end;

procedure TrMemoEx.SetHighlightFromStr(const Value: string);
begin
  FHighlightFromStr:=Value;
  Refresh;
end;

procedure TrMemoEx.SetHighlightToStr(const Value: string);
begin
  FHighlightToStr:=Value;
  Refresh;
end;

procedure TrMemoEx.SetHighlightedFontColor(Value: TColor);
begin
  FHighlightedFontColor:=Value;
  Refresh;
end;

procedure TrMemoEx.SetHighlightedBrushColor(Value: TColor);
begin
  FHighlightedBrushColor:=Value;
  Refresh;
end;

procedure TrMemoEx.SetHighlightFromToEx(Value: TStringList);
begin
  FHighlightFromToEx.Assign(Value);
  Refresh;
end;

procedure TrMemoEx.HighlightWordsChanged(Sender: TObject);
begin
  Refresh;
end;

procedure TrMemoEx.WMPaint(var Message: TWMPaint);
const
  SepChars: TSysCharSet = [' ', '_', '-', '(', ')', '/', '\', '=', ',', ';', '.', #9 .. #13];
  XorColor = $00FFD8CE;
var
  DC: HDC;
  Canvas: TCanvas;
  iline, I, FirstLine, LastLine, FirstChar, RowHeight: integer;
  CompStr, TextStr: string;
  ContinueFormat: Boolean;

  procedure DrawSel(StartPos: integer; Str: string);
  var
    seppos: integer;
    P: Longint;
    PT: TPoint;
    R: TRect;
  begin
    // draw selected
    P:=SendMessage(Handle, EM_POSFROMCHAR, FirstChar + StartPos - 1, 0);
    if P < 0 then
      Exit;
    PT.X:=LoWord(P);
    PT.Y:=HiWord(P);
    if (PT.X >= 0) and (PT.Y >= 0) and (PT.Y < ClientHeight - RowHeight) then
    begin
      seppos:=Pos(#13, Str);
      if seppos > 0 then
        Str:=copy(Str, 1, seppos - 1);
      R.TopLeft:=PT;
      R.Right:=R.Left + Canvas.TextWidth(Str);
      R.Bottom:=R.Top + Canvas.TextHeight(Str);
      Canvas.TextRect(R, PT.X, PT.Y, Str);
    end;
  end;

  procedure HighlightWord(WordStr: string);
  var
    p1: integer;
    sel: string;
  begin
    WordStr:=AnsiUpperCase(WordStr);
    p1:=0;
    repeat
      // find word
      p1:=PosEx(WordStr, CompStr, p1 + 1);
      if p1 = 0 then
        Break;
      if FHighlightWholeWords and not ContainWholeWord(WordStr, TextStr, SepChars, p1) then
        Continue;
      sel:=copy(TextStr, p1, Length(WordStr));
      DrawSel(p1, sel);
    until false;
  end;

  function HighlightFromTo(FromStr, ToStr: string; ContFrm: Boolean): Boolean;
  var
    p1, p2: integer;
    sel: string;
  begin
    Result:=false;
    FromStr:=AnsiUpperCase(FromStr);
    ToStr:=AnsiUpperCase(ToStr);
    // proc from to selection
    p1:=0;
    repeat
      // find start pos
      if ContFrm then
        p1:=1
      else
        p1:=PosEx(FromStr, CompStr, p1 + 1);
      ContFrm:=false;
      if p1 = 0 then
        Break;
      // find stop pos
      p2:=PosEx(ToStr, CompStr, p1 + Length(FromStr));
      if p2 = 0 then
      begin
        p2:=Length(TextStr);
        Result:=ToStr <> '#EOL#';
      end;
      sel:=copy(TextStr, p1, p2 - p1 + Length(ToStr));
      DrawSel(p1, sel);
      p1:=p2 + Length(ToStr);
    until false;
  end;

  procedure HighlightFromToEx(Line: string);
  var
    s1, s2, s3: string;
  begin
    s1:=RemoveStrPart(Line, #9, true);
    if s1 = '' then
      Exit;
    if s1 = '#9' then
      s1:=#9;
    s2:=RemoveStrPart(Line, #9, true);
    if s2 = '' then
      Exit;
    if s2 = '#9' then
      s2:=#9;

    s3:=RemoveStrPart(Line, #9, true);
    if s3 = '' then
      Canvas.Font.Color:=FHighlightedFontColor
    else
      if s3 = '-' then
        Canvas.Font.Color:=Font.Color
      else
        Canvas.Font.Color:=rpictool.StringToColor(s3);

    if Line = '' then
      Canvas.Brush.Color:=FHighlightedBrushColor
    else
      if Line = '-' then
        Canvas.Brush.Color:=clNone
      else
        Canvas.Brush.Color:=rpictool.StringToColor(Line);
    if Canvas.Brush.Color = clNone then
      Canvas.Brush.Style:=bsClear
    else
      Canvas.Brush.Style:=bsSolid;

    ContinueFormat:=HighlightFromTo(s1, s2, ContinueFormat);
  end;

  procedure HighlightWordEx(Line: string);
  var
    s1, s2: string;
  begin
    s1:=RemoveStrPart(Line, #9, true);
    if s1 = '' then
      Exit;

    s2:=RemoveStrPart(Line, #9, true);
    if s2 = '' then
      Canvas.Font.Color:=FHighlightedFontColor
    else
      if s2 = '-' then
        Canvas.Font.Color:=Font.Color
      else
        Canvas.Font.Color:=rpictool.StringToColor(s2);

    if Line = '' then
      Canvas.Brush.Color:=FHighlightedBrushColor
    else
      if Line = '-' then
        Canvas.Brush.Color:=clNone
      else
        Canvas.Brush.Color:=rpictool.StringToColor(Line);
    if Canvas.Brush.Color = clNone then
      Canvas.Brush.Style:=bsClear
    else
      Canvas.Brush.Style:=bsSolid;

    HighlightWord(s1);
  end;

  procedure ShowPageBreak(X: integer);
  begin
    Canvas.Pen.Color:=clBtnShadow;
    Canvas.MoveTo(X, 2);
    Canvas.LineTo(X, ClientHeight);
    Canvas.Pen.Color:=clBtnHighlight;
    Canvas.MoveTo(X + 1, 2);
    Canvas.LineTo(X + 1, ClientHeight);
  end;

  procedure ProcPageBreaks;
  var
    ipb, pbpos, w: integer;
    Str: string;
  begin
    if FPageBreaksCharNumber = '' then
      Exit;
    Canvas.Pen.Style:=psSolid;
    Canvas.Pen.Mode:=pmCopy;
    Canvas.Brush.Style:=bsSolid;
    ipb:=0;
    repeat
      Inc(ipb);
      Str:=rtool.FindStrPart(FPageBreaksCharNumber, ';', ipb, true);
      if Str = '' then
        Break;
      pbpos:=StrToIntDef(Str, 0);
      if pbpos > 0 then
      begin
        Str:=StringOfChar('o', pbpos);
        w:=Canvas.TextWidth(Str);
        if w < ClientWidth then
          ShowPageBreak(w);
      end;
    until false;
  end;

  procedure InitFrm;
  begin
    ContinueFormat:=false;
    Canvas.Font.Color:=FHighlightedFontColor;
    Canvas.Brush.Color:=FHighlightedBrushColor;
  end;

  procedure InitLine;
  begin
    FirstChar:=Perform(EM_LINEINDEX, iline, 0);
    TextStr:=Lines[iline];
    CompStr:=AnsiUpperCase(TextStr);
  end;

begin
  inherited;
  LoadSBPos;
  // Get a DC for the canvas
  DC:=GetDC(Handle);
  try
    Canvas:=TCanvas.Create;
    try
      Canvas.Handle:=DC;
      Canvas.Font:=Font;
      RowHeight:=Canvas.TextHeight('Wg');

      // find and proc all visible lines
      FirstLine:=Perform(EM_GETFIRSTVISIBLELINE, 0, 0);
      LastLine:=Perform(EM_CHARFROMPOS, 0, (ClientHeight - 1) shl 16);
      LastLine:=(LastLine and $FFFF0000) shr 16;

      // old simple FromTo format
      if (FHighlightFromStr <> '') and (FHighlightToStr <> '') then
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          ContinueFormat:=HighlightFromTo(FHighlightFromStr, FHighlightToStr, ContinueFormat);
        end;
      end;

      // new FromToEx format
      for I:=0 to FHighlightFromToEx.Count - 1 do
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          HighlightFromToEx(FHighlightFromToEx[I]);
        end;
      end;

      // new Words format
      for I:=0 to FHighlightWords.Count - 1 do
      begin
        InitFrm;
        for iline:=FirstLine to LastLine do
        begin
          InitLine;
          HighlightWordEx(FHighlightWords[I]);
        end;
      end;

      ProcPageBreaks;

      Canvas.Handle:=0;
    finally
      FreeAndNil(Canvas);
    end;
  finally
    // Release the DC when finished
    ReleaseDC(Handle, DC);
  end;
end;

procedure TrMemoEx.Change;
begin
  inherited Change;
  Refresh;
end;

procedure TrMemoEx.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited;
  Refresh;
end;

function TrMemoEx.SearchText(const Search: string = ''; FromStart: Boolean = true; IgnoreCase: Boolean = true): Boolean;
var
  sText: string;
  Index: integer;
begin
  Result:=false;
  FSearchText:=Search;
  if FSearchText = '' then
    if not rdlg.InputStr(sInput, SFind, FSearchText) then
      Exit;
  FSearchIgnoreCase:=IgnoreCase;
  sText:=Lines.Text;
  if FSearchIgnoreCase then
  begin
    FSearchText:=UpperCase(FSearchText);
    sText:=UpperCase(Lines.Text);
  end;
  if FromStart then
    Index:=1
  else
    Index:=SelStart + SelLength + 1;
  Index:=PosEx(FSearchText, sText, Index);
  Result:=Index > 0;
  if Result then
  begin
    SelStart:=Index - 1;
    SelLength:=Length(FSearchText);
    try
      SetFocus;
    except
    end;
  end;
  if not Result then
    DlgI(sNotFound);
end;

function TrMemoEx.SearchNext;
begin
  Result:=SearchText(FSearchText, false, FSearchIgnoreCase);
end;

procedure TrMemoEx.PopupMenuOpen(Sender: TObject);
var
  I: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  for I:=0 to PopupMenu.Items.Count - 1 do
  begin
    case PopupMenu.Items[I].Tag of
      rcEditUndo:
        PopupMenu.Items[I].Enabled:=CanUndo;
      rcEditCut:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditCopy:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditPaste:
        PopupMenu.Items[I].Enabled:=Clipbrd.Clipboard.AsText <> '';
      rcEditDelete:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditSelectAll:
        ;
      rcEditFind:
        ;
      rcEditFindNext:
        PopupMenu.Items[I].Enabled:=FSearchText <> '';
      rcEditSelectFont:
        PopupMenu.Items[I].Enabled:=moAllowSetFont in FOptionsEx;
      rcMemoWordWrap:
        begin
          PopupMenu.Items[I].Enabled:=moAllowWordWrap in FOptionsEx;
          PopupMenu.Items[I].Checked:=WordWrap;
        end;
      rcMemoLoadFile:
        PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
      rcMemoSaveFile:
        PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
    end;
  end;
end;

procedure TrMemoEx.PopupMenuClick(Sender: TObject);
var
  SaveDlg: TrSaveDialog;
  OpenDlg: TrOpenDialog;
begin
  if Sender is TMenuItem then
  begin
    case (Sender as TMenuItem).Tag of
      rcEditUndo:
        Undo;
      rcEditCut:
        CutToClipboard;
      rcEditCopy:
        CopyToClipboard;
      rcEditPaste:
        PasteFromClipboard;
      rcEditDelete:
        ClearSelection;
      rcEditSelectAll:
        SelectAll;
      rcEditFind:
        SearchText;
      rcEditFindNext:
        SearchNext;
      rcEditSelectFont:
        rDialogsEx.ShowSelectFontDialog(Font);
      rcMemoWordWrap:
        begin
          WordWrap:=not WordWrap;
          if WordWrap and (ScrollBars = ssBoth) then
            ScrollBars:=ssVertical;
          if not WordWrap and (ScrollBars = ssVertical) then
            ScrollBars:=ssBoth;
        end;
      rcMemoLoadFile:
        begin
          OpenDlg:=TrOpenDialog.Create(Self);
          try
            if OpenDlg.Execute then
              Lines.LoadFromFile(OpenDlg.FileName);
          finally
            OpenDlg.Free;
          end;
        end;
      rcMemoSaveFile:
        begin
          SaveDlg:=TrSaveDialog.Create(Self);
          try
            if SaveDlg.Execute then
              Lines.SaveToFile(SaveDlg.FileName);
          finally
            SaveDlg.Free;
          end;
        end;
    end;
  end;
end;

// ==============================================================================
// TrComboBoxEx
constructor TDisplayedItems.Create(AOwner: TCustomComboBox; OnListDefined: TNotifyEvent);
begin
  inherited Create;
  FListItems:=TStringList.Create;
  TStringList(FListItems).OnChange:=ItemUpdated;
  AComboBox:=AOwner;
  AItemSpace:=4;
  FAddItemAsFirstColumn:=false;
  FDrawColumnLine:=true;
  ColumnsWidthCalculated:=false;
  FOnListDefined:=OnListDefined;
end;

destructor TDisplayedItems.Destroy;
begin
  FreeAndNil(FListItems);
  inherited;
end;

procedure TDisplayedItems.SetDisplayedItems(Value: TStrings);
begin
  FListItems.Assign(Value);
  ColumnsWidthCalculated:=false;
  if Assigned(FOnListDefined) then
    FOnListDefined(Self);
end;

procedure TDisplayedItems.SetAddItemAsFirstColumn(Value: Boolean);
begin
  FAddItemAsFirstColumn:=Value;
  ColumnsWidthCalculated:=false;
end;

procedure TDisplayedItems.SetColumnSeparator(Value: string);
begin
  FColumnSeparator:=Value;
  ColumnsWidthCalculated:=false;
end;

procedure TDisplayedItems.ItemUpdated(Sender: TObject);
begin
  ColumnsWidthCalculated:=false;
end;

function TDisplayedItems.GetItemDisplayedText(idx: integer): string;
begin
  Result:='';
  if (idx >= 0) and (idx < AComboBox.Items.Count) then
  begin
    if idx < FListItems.Count then
    begin
      Result:=FListItems[idx];
      if FAddItemAsFirstColumn then
        Result:=AComboBox.Items[idx] + FColumnSeparator + Result;
    end
    else
      Result:=AComboBox.Items[idx];
  end;
end;

procedure TDisplayedItems.CalcColumnsWidth;
var
  ItemIdx, ColIdx, w: integer;
  ItemText, ColText: string;
begin
  SetLength(FColumnsWidth, 1);
  FColumnsWidth[0]:=0;
  if (FColumnSeparator = '#9') or (FColumnSeparator = '') then
    FColumnSeparator:=#9;
  for ItemIdx:=0 to AComboBox.Items.Count - 1 do
  begin
    ColIdx:=0;
    ItemText:=GetItemDisplayedText(ItemIdx);
    repeat
      if ColIdx > High(FColumnsWidth) then
      begin
        SetLength(FColumnsWidth, ColIdx + 1);
        FColumnsWidth[ColIdx]:=0;
      end;
      ColText:=rtool.RemoveStrPart(ItemText, FColumnSeparator);
      w:=AComboBox.Canvas.TextWidth(ColText) + AItemSpace;
      if FColumnsWidth[ColIdx] < w then
        FColumnsWidth[ColIdx]:=w;
      Inc(ColIdx);
    until ItemText = '';
  end;
  ColumnsWidthCalculated:=true;
end;

function TDisplayedItems.GetMaxItemWidth(DropDownCount: integer): integer;
var
  ColIdx, w: integer;
begin
  if not ColumnsWidthCalculated then
  begin
    TComboBox(AComboBox).Canvas.Font.Assign(TComboBox(AComboBox).Font);
    CalcColumnsWidth;
  end;
  Result:=AComboBox.Width;
  w:=0;
  for ColIdx:=0 to High(FColumnsWidth) do
    Inc(w, FColumnsWidth[ColIdx] + AItemSpace);
  if (AComboBox.Items.Count > DropDownCount) then
    Inc(w, GetSystemMetrics(SM_CXVSCROLL));
  if (w > Result) then
    Result:=w;
end;

procedure TDisplayedItems.DrawListItem(Index: integer; State: TOwnerDrawState; Rect: TRect; GetListItemProps: TGetListItemProps);
var
  SepTop, SepBottom: Boolean;
  Bmp: TBitmap;
  R, RTemp: TRect;
  ItemText, ColText: string;
  ColIdx: integer;
begin
  if not ColumnsWidthCalculated then
    CalcColumnsWidth;
  SepTop:=false;
  SepBottom:=false;
  R:=Rect;
  ItemText:=GetItemDisplayedText(Index);
  Bmp:=TBitmap.Create;
  try
    // init Pen color for separator lines
    AComboBox.Canvas.Pen.Color:=clGray;
    // get user defined props
    if Assigned(GetListItemProps) then
      GetListItemProps(AComboBox, AComboBox.Canvas, Index, State, R, ItemText, Bmp, SepTop, SepBottom);
    if (ItemText <> '') or not Assigned(GetListItemProps) then
    begin
      AComboBox.Canvas.FillRect(Rect);
      RTemp:=R;
      RTemp.Left:=RTemp.Left + 2;
      if Assigned(Bmp) then
        RTemp.Left:=RTemp.Left + Bmp.Width + 2;
      TControlCanvas(AComboBox.Canvas).UpdateTextFlags;
      for ColIdx:=0 to High(FColumnsWidth) do
      begin
        ColText:=rtool.RemoveStrPart(ItemText, FColumnSeparator);
        AComboBox.Canvas.TextOut(RTemp.Left, RTemp.Top, ColText);
        RTemp.Left:=RTemp.Left + FColumnsWidth[ColIdx];
        if FDrawColumnLine and (ColIdx < High(FColumnsWidth)) then
        begin
          // AComboBox.Canvas.Pen.Color:=clGray;
          AComboBox.Canvas.MoveTo(RTemp.Left, RTemp.Top);
          AComboBox.Canvas.LineTo(RTemp.Left, RTemp.Bottom);
        end;
        RTemp.Left:=RTemp.Left + AItemSpace;
      end;
    end;
    if Assigned(Bmp) and (Bmp.Width > 0) then
    begin
      RTemp:=R;
      RTemp.Left:=2;
      RTemp.Right:=RTemp.Left + Bmp.Width;
      rpictool.DrawBmpCenter(AComboBox.Canvas, RTemp, Bmp);
    end;
    if SepTop then
    begin
      // AComboBox.Canvas.Pen.Color:=clGray;
      AComboBox.Canvas.MoveTo(Rect.Left, Rect.Top);
      AComboBox.Canvas.LineTo(Rect.Right, Rect.Top);
    end;
    if SepBottom then
    begin
      // AComboBox.Canvas.Pen.Color:=clGray;
      AComboBox.Canvas.MoveTo(Rect.Left, Rect.Bottom - 2);
      AComboBox.Canvas.LineTo(Rect.Right, Rect.Bottom - 2);
    end;
  finally
    FreeAndNil(Bmp);
  end;
end;

constructor TrComboBoxEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScaleRatio:=1;
  FDropDownAutoWidth:=true;
  FDropDownMaxWidth:=600;
  FItemHeightEx:=cComboBoxItemHeightDefault;
  WndRecreated:=false;
  FDisplayedItems:=TDisplayedItems.Create(Self, ListDefined);
  IntTimer:=TTimer.Create(Self);
  FUpdateTimerDelay:=0;
  FLastKeyDown:=0;
  FJustClosed:=false;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrComboBoxEx.Destroy;
begin
  FreeAndNil(FDisplayedItems);
  inherited;
end;

procedure TrComboBoxEx.CreateParams(var Params: TCreateParams);
begin
  inherited;
  if UseOwnDraw then
    Params.Style:=Params.Style or CBS_OWNERDRAWFIXED;
end;

procedure TrComboBoxEx.WndProc(var Message: TMessage);
var
  CBRect, LBRect: TRect;
begin
  if Message.Msg = WM_CTLCOLORLISTBOX then
  begin
    GetWindowRect(Handle, CBRect);
    GetWindowRect(Message.LParam, LBRect);
    if LBRect.Right > Screen.DesktopRect.Right then
      MoveWindow(Message.LParam, LBRect.Left - (LBRect.Right - CBRect.Right), LBRect.Top, LBRect.Right - LBRect.Left, LBRect.Bottom - LBRect.Top, true);
  end;
  inherited WndProc(Message);
end;

procedure TrComboBoxEx.Loaded;
begin
  inherited Loaded;
  if UseOwnDraw then
  begin
    WndRecreated:=true;
    RecreateWnd;
  end;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrComboBoxEx.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  ItemHeight:=Trunc(FScaleRatio * FItemHeightEx);
end;

function TrComboBoxEx.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

function TrComboBoxEx.GetSelectedText: string;
begin
  if Style = csDropDownList then
  begin
    if Self.ItemIndex < 0 then
      Result:=''
    else
      Result:=Items[ItemIndex];
  end
  else
    Result:=Text;
end;

procedure TrComboBoxEx.SetItemHeightEx(Value: integer);
begin
  if Value <> FItemHeightEx then
  begin
    FItemHeightEx:=Value;
    ItemHeight:=Trunc(FScaleRatio * FItemHeightEx);
  end;
end;

procedure TrComboBoxEx.ListDefined(Sender: TObject);
begin
  if UseOwnDraw and not WndRecreated then
  begin
    WndRecreated:=true;
    RecreateWnd;
  end;
end;

function TrComboBoxEx.UseOwnDraw: Boolean;
begin
  Result:=Assigned(FGetListItemProps) or (FDisplayedItems.FListItems.Count > 0);
end;

procedure TrComboBoxEx.DropDown;
var
  w: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  if FDropDownAutoWidth then
  begin
    w:=Max(Width, Min(FDropDownMaxWidth, FDisplayedItems.GetMaxItemWidth(Self.DropDownCount)));
    SendMessage(Handle, CB_SETDROPPEDWIDTH, w, 0);
  end;
  inherited DropDown;
end;

procedure TrComboBoxEx.DrawItem(Index: integer; Rect: TRect; State: TOwnerDrawState);
begin
  if UseOwnDraw and not(odComboBoxEdit in State) then
    FDisplayedItems.DrawListItem(Index, State, Rect, FGetListItemProps)
  else
    inherited;
end;

procedure TrComboBoxEx.Change;
begin
  inherited Change;
  if FLastKeyDown <> VK_RETURN then
    ResetTimer(true, FJustClosed);
  FJustClosed:=false;
end;

procedure TrComboBoxEx.CloseUp;
begin
  inherited CloseUp;
  FJustClosed:=true;
end;

procedure TrComboBoxEx.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  FLastKeyDown:=Key;
end;

procedure TrComboBoxEx.ResetTimer(SetActive, CallEventNow: Boolean);
begin
  if Assigned(IntTimer) then
  begin
    IntTimer.Enabled:=false;
    IntTimer.Interval:=FUpdateTimerDelay;
    IntTimer.OnTimer:=IntTimerEvent;
    IntTimer.Enabled:=SetActive;
    if CallEventNow then
      IntTimerEvent(nil);
  end;
end;

procedure TrComboBoxEx.IntTimerEvent(Sender: TObject);
var
  s, n: string;
begin
  IntTimer.Enabled:=false;
  if Assigned(FUpdateTimerEvent) then
    FUpdateTimerEvent(Self);

  if Assigned(FOnUpdateLoadItems) then
  begin
    if SelStart > 0 then
      s:=copy(Text, 1, SelStart)
    else
    begin
      s:=Text;
    end;
    n:='';
    Self.Items.BeginUpdate;
    try
      FOnUpdateLoadItems(Self, s, n);
    finally
      Self.Items.EndUpdate;
    end;
    if (n <> '') then
      Text:=n;
    SelStart:=Length(s);
    SelLength:=Length(Text);
    ResetTimer(false, false);
  end;
end;

procedure TrComboBoxEx.AddTextToItems(AText: string = '');
begin
  if AText = '' then
    AText:=Text;
  AText:=Trim(AText);
  if (Length(AText) > 1) and (Items.IndexOf(AText) < 0) then
    Items.Add(AText);
end;

procedure TrComboBoxEx.RemoveTextFromItems(AText: string = '');
var
  I: integer;
begin
  if AText = '' then
    AText:=Text;
  AText:=Trim(AText);
  I:=Items.IndexOf(AText);
  if I >= 0 then
    Items.Delete(I);
end;

// ==============================================================================
// TrPopupMonthCalendar
constructor TrPopupMonthCalendar.CreateEx(AOwner: TComponent; OnSelectDate: TNotifyEvent);
begin
  inherited Create(AOwner);
  Visible:=false;
  AutoSize:=true;
  Parent:=TWinControl(AOwner);
  SelectDate:=OnSelectDate;
end;

procedure TrPopupMonthCalendar.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    Style:=WS_POPUP or WS_BORDER;
    ExStyle:=WS_EX_TOOLWINDOW;
    AddBiDiModeExStyle(ExStyle);
    WindowClass.Style:=CS_SAVEBITS;
  end;
end;

procedure TrPopupMonthCalendar.WMMouseActivate(var Message: TWMMouseActivate);
begin
  Message.Result:=MA_NOACTIVATE;
end;

procedure TrPopupMonthCalendar.CNNotify(var Message: TWMNotifyMC);
begin
  inherited;
  with Message, NMHdr {$IFNDEF CLR}^{$ENDIF} do
  begin
    case code of
      MCN_SELECT:
        begin
          // select date from calendar - call callback of parent and hide cal
          SelectDate(Self);
          HideCalendar;
        end;
    end;
  end;
end;

procedure TrPopupMonthCalendar.ShowCalendar(ADate: TDateTime);
var
  P: TPoint;
  Y: integer;
begin
  if FIsShown then
    Exit;
  AutoSize:=false;
  Width:=Width + 10;
  AutoSize:=true;
  Date:=ADate;
  // calc pos
  P:=Parent.Parent.ClientToScreen(Point(Parent.Left, Parent.Top));
  Y:=P.Y + Parent.Height;
  if Y + Height > Screen.Height then
    Y:=P.Y - Height;
  // show cal
  SetWindowPos(Handle, HWND_TOP, P.X, Y, 0, 0, SWP_NOSIZE or SWP_NOACTIVATE or SWP_SHOWWINDOW);

  // set autosize after show to fix bug with styles
  AutoSize:=false;
  Width:=Width + 1;
  AutoSize:=true;

  FIsShown:=true;
end;

procedure TrPopupMonthCalendar.HideCalendar;
begin
  if FIsShown then
  begin
    if GetCapture <> 0 then
      SendMessage(GetCapture, WM_CANCELMODE, 0, 0);
    SetWindowPos(Handle, 0, 0, 0, 0, 0, SWP_NOZORDER or SWP_NOMOVE or SWP_NOSIZE or SWP_NOACTIVATE or SWP_HIDEWINDOW);
    FIsShown:=false;
  end;
end;

// ==============================================================================
// TrEditDate
constructor TrEditDate.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width:=84;
  if not(csDesigning in ComponentState) then
  begin
    // create PopupCalendar - use SELF as Owner
    FPopupCalendar:=TrPopupMonthCalendar.CreateEx(Self, PopupCallbackSelectDate);
  end;
end;

procedure TrEditDate.Loaded;
begin
  inherited Loaded;
  RightButton.Enabled:=not ReadOnly;
  if FNowOnStart then
    Text:=DateToStr(Date);
  // use internal ButtonClick if not used by custom code
  if not(csDesigning in ComponentState) and not Assigned(OnRightButtonClick) then
    OnRightButtonClick:=FRightButtonClick;
end;

procedure TrEditDate.CMCancelMode(var Message: TCMCancelMode);
begin
  // hide cal after click to window (exclude only self, popup and RightButton)
  if not Assigned(FPopupCalendar) then
    Exit;
  if not FPopupCalendar.IsShown or (Message.Sender = Self) or (Message.Sender = FPopupCalendar) or Message.Sender.Equals(THackEditButton(RightButton).Glyph)
    then
    Exit;
  FPopupCalendar.HideCalendar;
  SetFocus;
end;

procedure TrEditDate.WMKillFocus(var Message: TMessage);
begin
  if Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
  inherited;
end;

procedure TrEditDate.PopupCallbackSelectDate(Sender: TObject);
begin
  if not ReadOnly then
    SetDate(Trunc(FPopupCalendar.Date));
end;

procedure TrEditDate.FRightButtonClick(Sender: TObject);
var
  D: TDateTime;
begin
  if Assigned(FPopupCalendar) and FPopupCalendar.IsShown then
    FPopupCalendar.HideCalendar
  else
  begin
    // ensure to have focus for direct click to button
    if not Focused then
      SetFocus;
    if Text = '' then
      D:=Today
    else
      D:=GetDate;
    FPopupCalendar.ShowCalendar(D);
  end;
end;

procedure TrEditDate.DoEnter;
begin
  inherited;
  // store original value to restore by ESC
  OrigText:=Text;
end;

procedure TrEditDate.DoExit;
begin
  inherited;
  if Assigned(FPopupCalendar) then
    FPopupCalendar.HideCalendar;
  if not CheckDate then
    SetFocus;
end;

procedure TrEditDate.EMSetReadOnly(var Message: TMessage);
begin
  inherited;
  RightButton.Enabled:=not ReadOnly;
end;

function TrEditDate.DoDropDown: Boolean;
begin
  Result:=false;
  if RightButton.Visible and Assigned(OnRightButtonClick) then
  begin
    OnRightButtonClick(Self);
    Result:=true;
  end;
end;

procedure TrEditDate.SetDate(ADateValue: TDateTime);
begin
  if ADateValue = 0 then
    Text:=''
  else
    try
      Text:=DateToStr(ADateValue);
    except
      Text:=DateToStr(Date);
    end;
  if (csLoading in ComponentState) then
    Exit;
  CheckDate;
  SelectAll;
end;

function TrEditDate.GetDate: TDateTime;
begin
  Result:=0;
  GetDateValue(Result);
  // if CheckDate and (Text <> '') then
  // Result:=StrToDateDef(ClearStr(' ', Text, true), Date);
end;

function TrEditDate.GetDateValue(var ADateValue: TDateTime): Boolean;
begin
  Result:=true;
  // if (csLoading in ComponentState) or (csDesigning in ComponentState) then
  // Exit;
  if Text = '' then
  begin
    if FNowIfEmpty then
      ADateValue:=Date
    else
      ADateValue:=0;
  end
  else
    Result:=StrToDateEx(Text, ADateValue, FMinDate, FMaxDate);
end;

function TrEditDate.CheckDate: Boolean;
var
  ADate: TDateTime;
begin
  ADate:=0;
  Result:=GetDateValue(ADate);
  if Result then
    Text:=DateToStr(ADate);
end;

procedure TrEditDate.KeyDown(var Key: Word; Shift: TShiftState);
var
  dif: integer;
begin
  if FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYDOWN, Key, 0);
    Key:=0;
  end;
  if not ReadOnly then
  begin
    case Key of
      VK_UP, VK_DOWN:
        begin
          if Key = VK_UP then
            dif:=1
          else
            dif:=-1;
          if Shift = [] then
            SetDate(GetDate + dif);
          if Shift = [ssCtrl] then
            SetDate(IncMonth(GetDate, dif));
          if Shift = [ssCtrl, ssShift] then
            SetDate(IncYear(GetDate, dif));
          if (Shift = [ssAlt]) and (Key = VK_DOWN) then
            DoDropDown;
          Key:=0;
        end;
      VK_F4:
        begin
          DoDropDown;
          Key:=0;
        end;
    end;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrEditDate.KeyUp(var Key: Word; Shift: TShiftState);
begin
  if FPopupCalendar.IsShown and (Key in rPopupMonthCalendarKeys) then
  begin
    // send key to cal
    PostMessage(FPopupCalendar.Handle, WM_KEYUP, Key, 0);
    Key:=0;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrEditDate.KeyPress(var Key: Char);
var
  Valid, Sep: TSysCharSet;
begin
  if not ReadOnly then
  begin
    case Key of
      'T', 't', '*':
        begin
          SetDate(Date);
          Key:=#0;
        end;
      'S', 's':
        begin
          SetDate(DateUtils.StartOfTheMonth(Date));
          Key:=#0;
        end;
      'E', 'e':
        begin
          SetDate(DateUtils.EndOfTheMonth(Date));
          Key:=#0;
        end;
      'P', 'p', 'D', 'd':
        begin
          DoDropDown;
          Key:=#0;
        end;
      chr(VK_ESCAPE):
        begin
          if not FPopupCalendar.IsShown then
          begin
            // restore orig value
            Text:=OrigText;
            SelectAll;
          end;
          Key:=#0;
        end;
      chr(VK_RETURN):
        begin
          if FPopupCalendar.IsShown then
            PopupCallbackSelectDate(nil)
          else
          begin
            CheckDate;
            SelectAll;
          end;
          Key:=#0;
        end;
    end;
    if Key = #0 then
      FPopupCalendar.HideCalendar;
  end;
  Sep:=['.', ',', '/', '-'];
  Valid:=['0' .. '9'] + Sep;
  if (not CharInSet(Key, Valid)) and (Key >= #32) then
    Key:=#0;
  if CharInSet(Key, Sep) then
    Key:=GetDateSeparator;
  if Key <> #0 then
  begin
    FPopupCalendar.HideCalendar;
    inherited KeyPress(Key);
  end;
end;

// ==============================================================================
// TrEditNum
constructor TrEditNum.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width:=48;
  Text:='0';
end;

procedure TrEditNum.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_UP:
      begin
        StepUp;
        Key:=0;
      end;
    VK_DOWN:
      begin
        StepDown;
        Key:=0;
      end;
  end;
  if Key <> 0 then
    inherited;
end;

procedure TrEditNum.KeyPress(var Key: Char);
var
  Valid: TSysCharSet;
begin
  if FStyle = enInteger then
    Valid:=['0' .. '9', '-', '+']
  else
    Valid:=['0' .. '9', '-', '+', 'e', 'E', GetDecimalSeparator];
  if (not CharInSet(Key, Valid)) and (Key >= #32) then
    Key:=#0;
  if ord(Key) = VK_RETURN then
  begin
    CheckValue;
    SelectAll;
    Key:=#0;
  end;
  if Key <> #0 then
    inherited KeyPress(Key);
end;

procedure TrEditNum.CMExit(var Message: TCMExit);
begin
  inherited;
  if not CheckValue then
    SetFocus;
end;

procedure TrEditNum.SetValue(Value: extended);
begin
  Text:=FormatFloat(FFormat, Value);
  CheckValue;
  if not(csDesigning in ComponentState) then
    SelectAll;
end;

function TrEditNum.GetValue: extended;
begin
  Result:=0;
  try
    Result:=StrToFloat(Text);
  except
  end;
end;

procedure TrEditNum.SetAsInteger(Value: integer);
begin
  Text:=FormatFloat(FFormat, Value);
  CheckValue;
  if not(csDesigning in ComponentState) then
    SelectAll;
end;

function TrEditNum.GetAsInteger: integer;
begin
  Result:=0;
  if CheckValue then
    Result:=Round(StrToFloat(Text));
end;

procedure TrEditNum.SetFormat(const Value: string);
begin
  FFormat:=Value;
  CheckValue;
end;

function TrEditNum.CheckValue: Boolean;
var
  f: extended;
begin
  Result:=false;
  if not Assigned(Self) then
    Exit;
  Result:=true;
  try
    if Text = '' then
      f:=0
    else
      f:=StrToFloat(Text);
    if (FMinValue <> 0) or (FMaxValue <> 0) then
    begin
      if f > FMaxValue then
        f:=FMaxValue;
      if f < FMinValue then
        f:=FMinValue;
    end;
    if StepOnly and (StepValue <> 0) then
      f:=StepValue * Round(f / StepValue);
    Text:=FormatFloat(FFormat, f);
  except
    Result:=false;
  end;
end;

procedure TrEditNum.StepUp;
begin
  SetValue(GetValue + StepValue);
end;

procedure TrEditNum.StepDown;
begin
  SetValue(GetValue - StepValue);
end;

// ==============================================================================
// TrCheckTreeView

procedure TrCheckTreeView.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  if FShowCheckBox then
    Params.Style:=Params.Style or TVS_CHECKBOXES;
end;

procedure TrCheckTreeView.SetShowCheckBox(Value: Boolean);
begin
  if FShowCheckBox <> Value then
  begin
    FShowCheckBox:=Value;
    ItemsRecreated;
    RecreateWnd;
  end;
end;

procedure TrCheckTreeView.ItemsRecreated;
begin
end;

procedure TrCheckTreeView.MouseDown(Button: TMouseButton; Shift: TShiftState; X: integer; Y: integer);
var
  HitTest: TTVHitTestInfo;
  Node: TTreeNode;
begin
  inherited;
  HitTest.PT.X:=X;
  HitTest.PT.Y:=Y;
  TreeView_HitTest(Handle, HitTest);
  // click to checkbox ?
  if (HitTest.Flags and TVHT_ONITEMSTATEICON) <> 0 then
  begin
    Node:=GetNodeAt(X, Y);
    Select(Node);
    DoCheckBoxClick(Node);
  end;
end;

procedure TrCheckTreeView.DoCheckBoxClick(Node: TTreeNode);
begin
  if FAutoCheckChildren then
    SetAllChildrenCheckState(Node, GetCheckState(Node));
  if FAutoCheckParents then
    SetAllParentsCheckState(Node, true, false);
  if Assigned(FCheckBoxClick) then
    FCheckBoxClick(Self, Node);
end;

function TrCheckTreeView.GetCheckState(Node: TTreeNode): Boolean;
begin
  Result:=TreeView_GetCheckState(Handle, Node.ItemId) = 1;
end;

procedure TrCheckTreeView.SetCheckState(Node: TTreeNode; Checked: Boolean);
var
  LState: UINT;
begin
  // BUG - TreeView_SetCheckState do not work, use SetItemState instead
  // TreeView_SetCheckState(Handle,Node.ItemId,Checked);
  if Checked then
    LState:=1
  else
    LState:=0;
  LState:=IndexToStateImageMask(LState + 1);
  TreeView_SetItemState(Handle, Node.ItemId, LState, TVIS_STATEIMAGEMASK);
end;

procedure TrCheckTreeView.SetAllChildrenCheckState(Node: TTreeNode; Checked: Boolean);
  procedure ProcOne(n: TTreeNode; ch: Boolean);
  begin
    while n <> nil do
    begin
      SetCheckState(n, ch);
      ProcOne(n.getFirstChild, ch);
      n:=n.getNextSibling;
    end;
  end;

begin
  if Assigned(Node) then
    ProcOne(Node.getFirstChild, Checked);
end;

procedure TrCheckTreeView.SetAllParentsCheckState(Node: TTreeNode; AutoChecked, Checked: Boolean);
  procedure ProcOne(n: TTreeNode; autoch, ch: Boolean);
  begin
    if n = nil then
      Exit;
    if autoch then
      ch:=GetChildrenCheckedCount(n) > 0;
    SetCheckState(n, ch);
    ProcOne(n.Parent, autoch, ch);
  end;

begin
  if Assigned(Node) then
    ProcOne(Node.Parent, AutoChecked, Checked);
end;

procedure TrCheckTreeView.SetAllItemsCheckState(Checked: Boolean);
var
  I: integer;
begin
  for I:=0 to Items.Count - 1 do
    SetCheckState(Items[I], Checked);
end;

function TrCheckTreeView.IntGetItemsCount(Node: TTreeNode; Siblink, Checked: Boolean): integer;
var
  n: TTreeNode;
begin
  Result:=0;
  if Node = nil then
    Exit;
  if Siblink then
  begin
    if Node.Parent <> nil then
      n:=Node.Parent.getFirstChild
    else
      n:=Self.Items.GetFirstNode;
  end
  else
    n:=Node.getFirstChild;
  while n <> nil do
  begin
    if not Checked or (GetCheckState(n)) then
      Inc(Result);
    n:=n.getNextSibling;
  end;
end;

function TrCheckTreeView.GetChildrenCount(Node: TTreeNode): integer;
begin
  Result:=IntGetItemsCount(Node, false, false);
end;

function TrCheckTreeView.GetChildrenCheckedCount(Node: TTreeNode): integer;
begin
  Result:=IntGetItemsCount(Node, false, true);
end;

function TrCheckTreeView.GetSiblinkCount(Node: TTreeNode): integer;
begin
  Result:=IntGetItemsCount(Node, true, false);
end;

function TrCheckTreeView.GetSiblinkCheckedCount(Node: TTreeNode): integer;
begin
  Result:=IntGetItemsCount(Node, true, true);
end;

// ==============================================================================
// TrStringsHolder
constructor TrStringsHolder.Create(AOwner: TComponent);
begin
  inherited;
  FStrings:=TStringList.Create;
end;

destructor TrStringsHolder.Destroy;
begin
  FreeAndNil(FStrings);
  inherited;
end;

procedure TrStringsHolder.SetStrings(Value: TStrings);
begin
  FStrings.Assign(Value);
end;

// ==============================================================================
// TrImageListEx

constructor TrImageListEx.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FOriginalImages:=TCustomImageList.Create(Self);
  FStoreToDFM:=true;
  FStoreToExternalFile:=false;
  FExternalFileDesignTimeDir:=GetDefaultDesignTimeDir;
  FMinScaleToResize:=1.75;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrImageListEx.Destroy;
begin
  FreeAndNil(FOriginalImages);
  inherited Destroy;
end;

procedure TrImageListEx.Loaded;
begin
  inherited;
  if FStoreToExternalFile then
    ReadExternalFile(nil);
  StoreAsOriginalImages;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

function TrImageListEx.IsMinScaleToResizeStored: Boolean;
begin
  Result:=FMinScaleToResize <> 1.75;
end;

procedure TrImageListEx.StoreAsOriginalImages;
begin
  FOriginalImages.Clear;
  FOriginalImages.SetSize(Self.Width, Self.Height);
  FOriginalImages.ColorDepth:=Self.ColorDepth;
  FOriginalImages.BkColor:=Self.BkColor;
  FOriginalImages.BlendColor:=Self.BlendColor;
  FOriginalImages.Masked:=Self.Masked;
  FOriginalImages.AddImages(Self);
end;

procedure TrImageListEx.ResizeImages(ScaleRatio: double);
const
  ColorDepthToPixelFormat: array [ Low(TColorDepth) .. High(TColorDepth)] of TPixelFormat = (pf4bit, pfDevice, pf4bit, pf8bit, pf16bit, pf24bit, pf32bit);
var
  I, newW, newH: integer;
  ImgO, MaskO, ImgT, MaskT: TBitmap;
  RectO, RectT: TRect;
  Strech: Boolean;
begin
  // calc new size
  newW:=Round(FOriginalImages.Width * ScaleRatio);
  newH:=Round(FOriginalImages.Height * ScaleRatio);
  if (newW = Width) and (newH = Height) then
    Exit;

  Strech:=ScaleRatio >= FMinScaleToResize;
  RectO:=Rect(0, 0, FOriginalImages.Width, FOriginalImages.Height);
  RectT:=Rect(0, 0, newW, newH);

  WCursor;
  BeginUpdate;
  try
    // update size
    SetSize(newW, newH);
    ImgO:=TBitmap.Create;
    try
      ImgO.PixelFormat:=ColorDepthToPixelFormat[ColorDepth];
      ImgO.SetSize(FOriginalImages.Width, FOriginalImages.Height);
      MaskO:=TBitmap.Create;
      try
        MaskO.Monochrome:=true;
        MaskO.SetSize(FOriginalImages.Width, FOriginalImages.Height);
        ImgT:=TBitmap.Create;
        try
          ImgT.PixelFormat:=ColorDepthToPixelFormat[ColorDepth];
          ImgT.Transparent:=false;
          ImgT.SetSize(Width, Height);
          MaskT:=TBitmap.Create;
          try
            MaskT.Monochrome:=true;
            MaskT.Transparent:=false;
            MaskT.SetSize(Width, Height);
            for I:=0 to FOriginalImages.Count - 1 do
            begin
              // load original image + mask
              ImgO.Canvas.FillRect(RectO);
              ImageList_Draw(FOriginalImages.Handle, I, ImgO.Canvas.Handle, 0, 0, ILD_NORMAL);
              MaskO.Canvas.FillRect(RectO);
              ImageList_Draw(FOriginalImages.Handle, I, MaskO.Canvas.Handle, 0, 0, ILD_MASK);
              // center or resize to new image + mask
              ImgT.Canvas.FillRect(RectT);
              MaskT.Canvas.FillRect(RectT);
              if Strech then
              begin
                rpictool.SmoothResizeToBitmap(ImgO, ImgT);
                rpictool.SmoothResizeToBitmap(MaskO, MaskT)
              end
              else
              begin
                rpictool.DrawBmpCenter(ImgT.Canvas, RectT, ImgO);
                rpictool.DrawBmpCenter(MaskT.Canvas, RectT, MaskO);
              end;
              // add to imagelist
              Add(ImgT, MaskT);
            end;
          finally
            FreeAndNil(MaskT);
          end;
        finally
          FreeAndNil(ImgT);
        end;
      finally
        FreeAndNil(MaskO);
      end;
    finally
      FreeAndNil(ImgO);
    end;
  finally
    EndUpdate;
    NCursor;
  end;
end;

procedure TrImageListEx.DefineProperties(Filer: TFiler);
begin
  if FStoreToDFM then
    inherited DefineProperties(Filer);
  Filer.DefineProperty('ExternalFile', ReadExternalFile, WriteExternalFile, FStoreToExternalFile and (Count > 0));
end;

function TrImageListEx.GetDefaultDesignTimeDir: string;
begin
  Result:=GetSystemDir(CSIDL_PERSONAL);
end;

function TrImageListEx.IsStoredDesignTimeDir: Boolean;
begin
  Result:=FExternalFileDesignTimeDir <> GetDefaultDesignTimeDir;
end;

function TrImageListEx.GetResourceFileName(ShowErrMsg: Boolean): string;
var
  dir, fn: string;
begin
  if csDesigning in ComponentState then
  begin
    dir:=FExternalFileDesignTimeDir;
    if dir = '' then
      dir:=GetDefaultDesignTimeDir;
  end
  else
    dir:=GetAppDir;
  fn:=FExternalFileName;
  if fn = '' then
    fn:=Self.Owner.Name + '_' + Self.Name + '.rsc';
  Result:=SetEndOfPath(dir) + fn;
  dir:=ExtractFilePath(Result);
  if not ForceDirectories(dir) and ShowErrMsg then
    raise EWriteError.Create(SImageWriteFail);
end;

procedure TrImageListEx.ReadExternalFile(Reader: TReader);
var
  SA: TStreamAdapter;
  ResStream: TMemoryStream;
  FileStream: TFileStream;
  fn: string;
begin
  if Assigned(Reader) then
    Reader.ReadString;
  fn:=GetResourceFileName(false);
  if not FileExists(fn) then
    Exit;
  FileStream:=TFileStream.Create(fn, fmOpenRead + fmShareDenyWrite);
  try
    ResStream:=TMemoryStream.Create;
    try
      if DeCompressStream(FileStream, ResStream) > 0 then
      begin
        ResStream.Position:=0;
        SA:=TStreamAdapter.Create(ResStream);
        try
          Handle:=ImageList_Read(SA);
          if Handle = 0 then
            raise EReadError.Create(SImageWriteFail);
        finally
          FreeAndNil(SA);
        end;
      end;
    finally
      FreeAndNil(ResStream);
    end;
  finally
    FreeAndNil(FileStream);
  end;
end;

procedure TrImageListEx.WriteExternalFile(Writer: TWriter);
var
  SA: TStreamAdapter;
  ResStream: TMemoryStream;
  FileStream: TFileStream;
  fn: string;
begin
  fn:=GetResourceFileName(true);
  Writer.WriteString(fn);
  FileStream:=TFileStream.Create(fn, fmCreate);
  try
    ResStream:=TMemoryStream.Create;
    try
      SA:=TStreamAdapter.Create(ResStream);
      try
        if not ImageList_Write(Handle, SA) then
          raise EWriteError.Create(SImageWriteFail);
      finally
        FreeAndNil(SA);
      end;
      CompressStream(ResStream, FileStream);
    finally
      FreeAndNil(ResStream);
    end;
  finally
    FreeAndNil(FileStream);
  end;
end;

// ==============================================================================
// TrButton
procedure TrButton.Loaded;
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  FImagesEnabled:=Images;
  if Enabled then
    Images:=FImagesEnabled
  else
    Images:=FImagesDisabled;
end;

procedure TrButton.SetImagesDisabled(const Value: TCustomImageList);
begin
  if Value <> FImagesDisabled then
  begin
    FImagesDisabled:=Value;
    if FImagesDisabled <> nil then
      FImagesDisabled.FreeNotification(Self);
  end;
end;

procedure TrButton.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if Operation = opRemove then
  begin
    if AComponent = FImagesDisabled then
      FImagesDisabled:=nil;
  end;
end;

procedure TrButton.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  if (csDesigning in ComponentState) then
    Exit;
  if Assigned(FImagesEnabled) then
  begin
    if Enabled then
      Images:=FImagesEnabled
    else
      Images:=FImagesDisabled;
  end;
  Invalidate;
end;

procedure TrButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  P: TPoint;
begin
  if (Button = mbLeft) and (Style <> bsSplitButton) and Assigned(DropDownMenu) then
  begin
    // show popup menu
    P:=ClientToScreen(Point(0, 0));
    // DropDownMenu.Alignment:=paLeft;
    DropDownMenu.Popup(P.X - 1, P.Y + Height);
  end;
  inherited;
end;

// ==============================================================================
// TrCategoryPanelGroupEx
{$IFDEF CategoryPanelGroupEx}

procedure TrCategoryPanelEx.GetChildren(Proc: TGetChildProc; Root: TComponent);
var
  I: integer;
  Control: TControl;
begin
  for I:=0 to FPanelSurface.ControlCount - 1 do
  begin
    Control:=FPanelSurface.Controls[I];
    if Control.Owner = Root then
      Proc(Control);
  end;
end;

var
  Registered: Boolean = false;

constructor TrCategoryPanelGroupEx.Create(AOwner: TComponent);
begin
  inherited;
  if not Registered then
  begin
    Classes.RegisterClass(TrCategoryPanelEx);
    Registered:=true;
  end;
end;

function TrCategoryPanelGroupEx.GetCategoryPanelClass: TCategoryPanelClass;
begin
  Result:=TrCategoryPanelEx;
end;
{$ENDIF}

// ==============================================================================
// TrTabProps subcomponent
constructor TTabPropsEx.Create(AOwner: TComponent);
begin
  inherited;
  SetSubComponent(true);
  Name:=AOwner.Name + 'TabPropsEx';
  FOwner:=TCustomTabControl(AOwner);
  TC:=THackCustomTabControl(AOwner);
  FScaleRatio:=1;
  FHotTabIndex:=-1;
  FHotBtnIndex:=-1;

  FDrawThemed:=true;
  FWordWrap:=false;
  FTabColor:=clNone;
  FSelectedTabColor:=clNone;
  FHotTabColor:=clNone;

  FParentFont:=true;
  FTabFont:=TFont.Create;
  FTabFont.OnChange:=FontChanged;
  FSelectedTabFont:=TFont.Create;
  FSelectedTabFont.OnChange:=FontChanged;
  FHotTabFont:=TFont.Create;
  FHotTabFont.OnChange:=FontChanged;

  FSelectedBorderColor:=$000080FF;
  FHotBorderColor:=$000080FF;

  FShowCloseBtn:=false;
  FCloseBtnSize:=16;
  FCloseBtn:=TPicture.Create;
  FHotCloseBtn:=TPicture.Create;
end;

destructor TTabPropsEx.Destroy;
begin
  FreeAndNil(FTabFont);
  FreeAndNil(FSelectedTabFont);
  FreeAndNil(FHotTabFont);
  FreeAndNil(FCloseBtn);
  FreeAndNil(FHotCloseBtn);
  inherited;
end;

procedure TTabPropsEx.UpdateScale(ScaleRatio: double; M, D: integer);
begin
  FScaleRatio:=ScaleRatio;
  if not FParentFont then
  begin
    rtool.DPI_ScaleFontHeight(FTabFont, M, D);
    rtool.DPI_ScaleFontHeight(FSelectedTabFont, M, D);
    rtool.DPI_ScaleFontHeight(FHotTabFont, M, D);
  end;
end;

function TTabPropsEx.UseScale(I: integer): integer;
begin
  Result:=Round(FScaleRatio * I);
end;

procedure TTabPropsEx.SetDrawThemed(Value: Boolean);
begin
  if Value <> FDrawThemed then
  begin
    FDrawThemed:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetWordWrap(Value: Boolean);
begin
  if Value <> FWordWrap then
  begin
    FWordWrap:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetTabColor(Value: TColor);
begin
  if Value <> FTabColor then
  begin
    FTabColor:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetSelectedTabColor(Value: TColor);
begin
  if Value <> FSelectedTabColor then
  begin
    FSelectedTabColor:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetHotTabColor(Value: TColor);
begin
  if Value <> FHotTabColor then
  begin
    FHotTabColor:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetParentFont(Value: Boolean);
begin
  if Value then
  begin
    FTabFont.Assign(TC.Font);
    FSelectedTabFont.Assign(TC.Font);
    FHotTabFont.Assign(TC.Font);
  end;
  FParentFont:=Value;
end;

procedure TTabPropsEx.SetTabFont(Value: TFont);
begin
  FTabFont.Assign(Value);
end;

procedure TTabPropsEx.SetSelectedTabFont(Value: TFont);
begin
  FSelectedTabFont.Assign(Value);
end;

procedure TTabPropsEx.SetHotTabFont(Value: TFont);
begin
  FHotTabFont.Assign(Value);
end;

procedure TTabPropsEx.SetSelectedBorderColor(Value: TColor);
begin
  if Value <> FSelectedBorderColor then
  begin
    FSelectedBorderColor:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetHotBorderColor(Value: TColor);
begin
  if Value <> FHotBorderColor then
  begin
    FHotBorderColor:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetShowCloseBtn(Value: Boolean);
begin
  if Value <> FShowCloseBtn then
  begin
    FShowCloseBtn:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetCloseBtnSize(Value: integer);
begin
  if Value <> FCloseBtnSize then
  begin
    FCloseBtnSize:=Value;
    TC.Invalidate;
  end;
end;

procedure TTabPropsEx.SetCloseBtn(Value: TPicture);
begin
  FCloseBtn.Assign(Value);
end;

procedure TTabPropsEx.SetHotCloseBtn(Value: TPicture);
begin
  FHotCloseBtn.Assign(Value);
end;

function TTabPropsEx.IsFontStored: Boolean;
begin
  Result:=not FParentFont;
end;

procedure TTabPropsEx.FontChanged(Sender: TObject);
begin
  FParentFont:=false;
  TC.Invalidate;
end;

procedure TTabPropsEx.ResetHotState(DoInvalidate: Boolean);
begin
  if (FHotTabIndex >= 0) or (FHotBtnIndex >= 0) then
  begin
    if DoInvalidate then
    begin
      if FHotTabIndex < 0 then
        TC.Invalidate
      else
        InvalidateRect(TC.Handle, TC.TabRect(FHotTabIndex), false);
    end;
    FHotTabIndex:=-1;
    FHotBtnIndex:=-1;
  end;
end;

procedure TTabPropsEx.ResetFont(Font: TFont);
begin
  if FParentFont then
  begin
    FTabFont.Assign(Font);
    FSelectedTabFont.Assign(Font);
    FHotTabFont.Assign(Font);
    FParentFont:=true;
  end;
end;

function TTabPropsEx.CheckRightBorder(TabIndex: integer; var R: TRect; AddBorder: integer): Boolean;
var
  B: integer;
begin
  B:=TC.ClientWidth;
  if TabIndex < TC.Tabs.Count - 1 then
    Dec(B, UseScale(38));
  if R.Right + AddBorder >= B then
    R.Right:=TC.ClientWidth - UseScale(38) - AddBorder;
  Result:=R.Right > R.Left;
end;

function TTabPropsEx.CloseBtnShown(TabIndex: integer): Boolean;
var
  BkColor: TColor;
begin
  Result:=FShowCloseBtn;
  BkColor:=clNone;
  if Assigned(FOnTabGetProps) then
    FOnTabGetProps(FOwner, TabIndex, TC.Canvas.Font, BkColor, Result);
end;

function TTabPropsEx.CloseBtnRect(TabIndex: integer): TRect;
var
  w, h: integer;
begin
  Result:=TC.TabRect(TabIndex);
  if TabIndex = TC.TabIndex then
    case TC.TabPosition of
      tpTop:
        Dec(Result.Bottom, 2);
      tpBottom:
        Inc(Result.Top, 2);
    end;
  w:=UseScale(FCloseBtnSize);
  h:=w;
  if FCloseBtn.Width > 0 then
  begin
    w:=FCloseBtn.Width;
    h:=FCloseBtn.Height;
  end;
  Result.Left:=Result.Right - w - UseScale(rPageExSpace);
  Result.Top:=(Result.Bottom - Result.Top - h) div 2 + Result.Top;
  Result.Right:=Result.Left + w;
  Result.Bottom:=Result.Top + h;
end;

procedure TTabPropsEx.DrawItemMessage(var Message: TWMDrawItem);
var
  SaveIndex: integer;
  Rgn: HRGN;
begin
  // reset clipped in the passed rectangle
  SelectClipRgn(Message.DrawItemStruct.HDC, 0);
  // copied from original PageControl but call own DrawTab
  with Message.DrawItemStruct {$IFNDEF CLR}^{$ENDIF} do
  begin
    SaveIndex:=SaveDC(HDC);
    TC.Canvas.Lock;
    try
      TC.Canvas.Handle:=HDC;
      TC.Canvas.Font:=TC.Font;
      TC.Canvas.Brush:=TC.Brush;
      // call own draw function
      DrawTabEx(ItemId, rcItem, ItemState);
    finally
      TC.Canvas.Handle:=0;
      TC.Canvas.Unlock;
      RestoreDC(HDC, SaveIndex);
    end;
  end;
  // clip the DC so that the borders to be drawn are out of region
  Rgn:=CreateRectRgn(0, 0, 0, 0);
  SelectClipRgn(Message.DrawItemStruct.HDC, Rgn);
  DeleteObject(Rgn);

  Message.Result:=1;
  // inherited;
end;

procedure TTabPropsEx.DrawTabEx(TabIndex: integer; ItemRect: TRect; ItemState: cardinal);
var
  oldBrush: TBrushRecall;
  oldFont: TFontRecall;
  oldPen: TPenRecall;
  TabRect, PaneRect: TRect;
  Themed, Selected, Hot, CloseBtn: Boolean;
  BkColor: TColor;
begin
  Themed:=GetThemeEnabled and FDrawThemed;
  Selected:=(ItemState and ODS_SELECTED) <> 0;
  Hot:=TC.HotTrack and (TabIndex = FHotTabIndex);
  CloseBtn:=FShowCloseBtn;

  if Assigned(TC.OnDrawTab) then
  begin
    // call user defined draw event
    TC.OnDrawTab(TC, TabIndex, ItemRect, Selected);
    Exit;
  end;

  oldBrush:=TBrushRecall.Create(TC.Canvas.Brush);
  oldFont:=TFontRecall.Create(TC.Canvas.Font);
  oldPen:=TPenRecall.Create(TC.Canvas.Pen);
  try
    // set font
    TC.Canvas.Font.Assign(TC.Font);
    if not FParentFont then
    begin
      TC.Canvas.Font.Assign(FTabFont);
      if Selected then
        TC.Canvas.Font.Assign(FSelectedTabFont)
      else
        if Hot then
          TC.Canvas.Font.Assign(FHotTabFont);
    end;
    // rtool.DPI_SetFontHeight(TC.Canvas.Font);
    // set bk color
    BkColor:=TC.Color;
    if FTabColor <> clNone then
      BkColor:=FTabColor;
    if Selected then
    begin
      if (FSelectedTabColor <> clNone) then
        BkColor:=FSelectedTabColor;
    end
    else
      if Hot and (FHotTabColor <> clNone) then
        BkColor:=FHotTabColor;
    // call props event
    if Assigned(FOnTabGetProps) then
      FOnTabGetProps(FOwner, TabIndex, TC.Canvas.Font, BkColor, CloseBtn);

    // init rect
    PaneRect:=TC.ClientRect;
    TabRect:=TC.TabRect(TabIndex);
    case TC.TabPosition of
      tpTop:
        PaneRect.Top:=TabRect.Bottom;
      tpBottom:
        PaneRect.Bottom:=TabRect.Top;
    end;

    if Selected then
      InflateRect(TabRect, 2, 2);
    if TC.TabPosition = tpBottom then
    begin
      OffsetRect(TabRect, 0, -1);
      if Selected then
        Inc(TabRect.Top);
    end;

    // draw background
    if Assigned(FOnDrawTabBackground) then
      FOnDrawTabBackground(TC, TabIndex, PaneRect, TabRect, BkColor, Selected, Hot, CloseBtn)
    else
      if Themed then
        DrawBkThemed(TabIndex, PaneRect, TabRect, Selected, Hot)
      else
        DrawBkStandard(TabIndex, PaneRect, TabRect, BkColor, Selected, Hot);
    // draw content
    if Assigned(FOnDrawTabContent) then
      FOnDrawTabContent(TC, TabIndex, PaneRect, TabRect, BkColor, Selected, Hot, CloseBtn)
    else
      DrawTabContent(TabIndex, TabRect, Selected, Hot, CloseBtn);
    // draw button
    if Assigned(FOnDrawTabCloseBtn) then
      FOnDrawTabCloseBtn(TC, TabIndex, PaneRect, TabRect, BkColor, Selected, Hot, CloseBtn)
    else
      if CloseBtn then
        DrawCloseBtn(TabIndex, CloseBtnRect(TabIndex), TabIndex = FHotBtnIndex);
    // finalize
    if Assigned(FOnDrawTabFinalize) then
      FOnDrawTabFinalize(TC, TabIndex, PaneRect, TabRect, BkColor, Selected, Hot, CloseBtn);
  finally
    FreeAndNil(oldBrush);
    FreeAndNil(oldFont);
    FreeAndNil(oldPen);
  end;

  // draw focus rect for active tab
  if Selected and TC.Focused then
  begin
    InflateRect(ItemRect, -2, -2);
    if TC.TabPosition = tpBottom then
      Dec(ItemRect.Bottom);
    BiDi_DrawFocusRect(TC.Canvas, ItemRect);
  end;
end;

procedure TTabPropsEx.DrawBkThemed(TabIndex: integer; PaneRect: TRect; TabRect: TRect; Selected, Hot: Boolean);
var
  ThemeData: HTHEME;
  ThemeStateId: integer;
begin
  ThemeData:=OpenThemeData(TC.Handle, VSCLASS_TAB);
  try
    if Selected then
      ThemeStateId:=TIS_SELECTED
    else
      if Hot then
        ThemeStateId:=TIS_HOT
      else
        ThemeStateId:=TIS_NORMAL;
    // Pane
    DrawThemeBackground(ThemeData, TC.Canvas.Handle, TABP_PANE, 0, PaneRect, nil);
    // Tab + frame
    if not CheckRightBorder(TabIndex, TabRect, 0) then
      Exit;
    DrawThemeBackground(ThemeData, TC.Canvas.Handle, TABP_TABITEM, ThemeStateId, TabRect, nil);
    if TC.TabPosition = tpTop then
    begin
      if TC.TabIndex = 0 then
        DrawThemeBackground(ThemeData, TC.Canvas.Handle, TABP_TABITEMLEFTEDGE, ThemeStateId, TabRect, nil);
      Inc(TabRect.Right, 2);
      if TabRect.Right = TC.ClientWidth then
        DrawThemeBackground(ThemeData, TC.Canvas.Handle, TABP_TABITEMRIGHTEDGE, ThemeStateId, TabRect, nil);
    end;
    if TC.TabPosition = tpBottom then
    begin
      // Inc(TabRect.Right,2);
      // DrawThemeBackground(ThemeData, tc.Canvas.Handle, TABP_TABITEMBOTHEDGE, aStateId, TabRect, nil);
      // DrawThemeBackground(ThemeData, tc.Canvas.Handle, TABP_TABITEMRIGHTEDGE, aStateId, TabRect, nil);
      Dec(TabRect.Bottom);
      // draw bottom line missing in standard class
      TC.Canvas.Pen.Color:=clGray;
      TC.Canvas.MoveTo(TabRect.Left, TabRect.Bottom);
      TC.Canvas.LineTo(TabRect.Right, TabRect.Bottom);
      { // remove part of old line in first page
        TC.Canvas.Pen.Color:=TC.Color;
        TC.Canvas.MoveTo(TabRect.Left + 1, TabRect.Top);
        TC.Canvas.LineTo(TabRect.Left + 5, TabRect.Top);
        }
    end;
  finally
    CloseThemeData(ThemeData);
  end;
end;

procedure TTabPropsEx.DrawBkStandard(TabIndex: integer; PaneRect: TRect; TabRect: TRect; BkColor: TColor; Selected, Hot: Boolean);
var
  R: TRect;
  y1, y2: integer;
begin
  // Pane
  TC.Canvas.Brush.Style:=bsSolid;
  Frame3D(TC.Canvas, PaneRect, clBtnHighlight, cl3DDkShadow, 1);

  // Tab frame
  if not CheckRightBorder(TabIndex, TabRect, 0) then
    Exit;
  TC.Canvas.Pen.Style:=psSolid;
  TC.Canvas.Pen.Width:=1;
  R:=TabRect;

  if TC.TabPosition = tpBottom then
  begin
    y1:=R.Bottom - 1;
    y2:=R.Top;
  end
  else
  begin
    y1:=R.Top;
    y2:=R.Bottom;
  end;
  TC.Canvas.Pen.Color:=clWindow;
  TC.Canvas.MoveTo(R.Left, y1);
  TC.Canvas.LineTo(R.Right, y1);
  TC.Canvas.MoveTo(R.Left, y1);
  TC.Canvas.LineTo(R.Left, y2);

  TC.Canvas.Pen.Color:=cl3DDkShadow;
  TC.Canvas.MoveTo(R.Right - 1, y1);
  TC.Canvas.LineTo(R.Right - 1, y2);

  if TC.TabPosition = tpBottom then
    Dec(y1)
  else
    Inc(y1);
  TC.Canvas.Pen.Color:=cl3DLight;
  TC.Canvas.MoveTo(R.Left + 1, y1);
  TC.Canvas.LineTo(R.Right - 1, y1);
  TC.Canvas.MoveTo(R.Left + 1, y1);
  TC.Canvas.LineTo(R.Left + 1, y2);
  TC.Canvas.MoveTo(R.Right - 2, y1);
  TC.Canvas.LineTo(R.Right - 2, y2);

  // tab bk
  TC.Canvas.Brush.Color:=BkColor;
  InflateRect(R, -2, 0);
  if TC.TabPosition = tpBottom then
    Dec(R.Bottom, 2)
  else
    Inc(R.Top, 2);
  TC.Canvas.FillRect(R);

  // tab hotborder
  BkColor:=clNone;
  if Selected then
    BkColor:=FSelectedBorderColor
  else
    if Hot then
      BkColor:=FHotBorderColor;
  if (BkColor <> clNone) then
  begin
    R:=TabRect;
    InflateRect(R, -1, 0);
    if not Selected then
      InflateRect(R, -1, 0);
    TC.Canvas.Brush.Color:=BkColor;
    case TC.TabPosition of
      tpTop:
        R.Bottom:=R.Top + 2;
      tpBottom:
        R.Top:=R.Bottom - 2;
      tpLeft:
        R.Right:=R.Left + 2;
      tpRight:
        R.Left:=R.Right - 2;
    end;
    TC.Canvas.FillRect(R);
  end;
end;

procedure TTabPropsEx.DrawTabContent(TabIndex: integer; TabRect: TRect; Selected, Hot, CloseBtn: Boolean);
var
  TabText: string;
  TextFormat: TTextFormat;
  txtw, imgidx, sumw, dx, dy: integer;
  ShowImg: Boolean;
  R: TRect;
begin
  TabText:=Trim(TC.Tabs[TabIndex]);
  if not CheckRightBorder(TabIndex, TabRect, 0) then
    Exit;
  if Selected then
    case TC.TabPosition of
      tpTop:
        Dec(TabRect.Bottom, 2);
      tpBottom:
        Inc(TabRect.Bottom, 2);
    end;
  // calc pos
  txtw:=TC.Canvas.TextWidth(TabText);
  imgidx:=TC.GetImageIndex(TabIndex);
  ShowImg:=Assigned(TC.Images) and (imgidx >= 0) and (imgidx < TC.Images.Count);
  sumw:=txtw;
  if ShowImg then
    Inc(sumw, TC.Images.Width + UseScale(rPageExSpace));
  if CloseBtn then
    Inc(sumw, UseScale(CloseBtnSize) + UseScale(rPageExSpace));
  dx:=(TabRect.Right - TabRect.Left - sumw) div 2;
  if dx < 4 then
    dx:=4;

  // show picture
  if ShowImg and (TabRect.Left + dx + TC.Images.Width < TabRect.Right) then
  begin
    dy:=(TabRect.Bottom - TabRect.Top - TC.Images.Height) div 2;
    if dy < 0 then
      dy:=0;
    TC.Images.Draw(TC.Canvas, TabRect.Left + dx, TabRect.Top + dy, imgidx);
    Inc(dx, TC.Images.Width + UseScale(rPageExSpace));
  end;

  // show text
  R:=TabRect;
  R.Left:=TabRect.Left + dx;
  if CloseBtn then
    R.Right:=Min(CloseBtnRect(TabIndex).Left - UseScale(rPageExSpace), TabRect.Right)
  else
    R.Right:=TabRect.Right - UseScale(4);

  if FWordWrap and (R.Right - R.Left < txtw) and (Pos(' ', TabText) > 0) then
  begin
    TextFormat:=[tfCenter, tfWordBreak];
    Inc(R.Top, 2);
  end
  else
    TextFormat:=[tfVerticalCenter, tfSingleLine];
  TC.Canvas.Brush.Style:=bsClear;
  TC.Canvas.TextRect(R, TabText, TextFormat);
end;

procedure TTabPropsEx.DrawCloseBtn(TabIndex: integer; BtnRect: TRect; Hot: Boolean);
var
  ThemeData: HTHEME;
  w, sp: integer;
begin
  w:=BtnRect.Right - BtnRect.Left;
  if not CheckRightBorder(TabIndex, BtnRect, UseScale(rPageExSpace + 2)) then
    Exit;
  if BtnRect.Right - BtnRect.Left < w then
    Exit;
  if FCloseBtn.Width > 0 then
  begin
    if (FHotCloseBtn.Width > 0) and Hot then
      TC.Canvas.Draw(BtnRect.Left, BtnRect.Top, FHotCloseBtn.Graphic)
    else
      TC.Canvas.Draw(BtnRect.Left, BtnRect.Top, FCloseBtn.Graphic);
  end
  else
  begin
    ThemeData:=OpenThemeData(TC.Handle, VSCLASS_BUTTON);
    try
      if Hot then
        DrawThemeBackground(ThemeData, TC.Canvas.Handle, BP_PUSHBUTTON, PBS_HOT, BtnRect, nil)
      else
        DrawThemeBackground(ThemeData, TC.Canvas.Handle, BP_PUSHBUTTON, PBS_NORMAL, BtnRect, nil);
    finally
      CloseThemeData(ThemeData);
    end;
    // draw X in button
    TC.Canvas.Pen.Color:=clGray;
    TC.Canvas.Pen.Width:=2;
    sp:=(BtnRect.Bottom - BtnRect.Top) div 3;
    TC.Canvas.MoveTo(BtnRect.Left + sp, BtnRect.Top + sp);
    TC.Canvas.LineTo(BtnRect.Right - sp, BtnRect.Bottom - sp);
    TC.Canvas.MoveTo(BtnRect.Left + sp, BtnRect.Bottom - sp);
    TC.Canvas.LineTo(BtnRect.Right - sp, BtnRect.Top + sp);
    TC.Canvas.Pen.Width:=1;
  end;
end;

procedure TTabPropsEx.CheckMouseMove(Shift: TShiftState; X, Y: integer);
var
  I: integer;
  actHotIndex, actHotBtnIndex: integer;
begin
  if (csDesigning in ComponentState) or not TC.OwnerDraw or not TC.HotTrack then
    Exit;
  actHotIndex:=-1;
  actHotBtnIndex:=-1;
  for I:=0 to TC.Tabs.Count - 1 do
  begin
    if PtInRect(TC.TabRect(I), Point(X, Y)) then
    begin
      if (TC.TabIndex <> I) then
        actHotIndex:=I;
      if PtInRect(CloseBtnRect(I), Point(X, Y)) then
        actHotBtnIndex:=I;
      Break;
    end;
  end;
  // invalidate old and new hot tabs
  if (FHotTabIndex <> actHotIndex) then
    try
      if (FHotTabIndex >= 0) then
        InvalidateRect(TC.Handle, TC.TabRect(FHotTabIndex), false);
      FHotTabIndex:=actHotIndex;
      FHotBtnIndex:=actHotBtnIndex;
      if (FHotTabIndex >= 0) then
        InvalidateRect(TC.Handle, TC.TabRect(FHotTabIndex), false);
    except
      TC.Invalidate;
    end;
  // invalidate old and new buttons
  if (FHotBtnIndex <> actHotBtnIndex) then
    try
      if (FHotBtnIndex >= 0) then
        InvalidateRect(TC.Handle, CloseBtnRect(FHotBtnIndex), false);
      FHotBtnIndex:=actHotBtnIndex;
      if (FHotBtnIndex >= 0) then
        InvalidateRect(TC.Handle, CloseBtnRect(FHotBtnIndex), false);
    except
      TC.Invalidate;
    end;
end;

procedure TTabPropsEx.CheckMouseClick(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
var
  I: integer;
  CanClose: Boolean;
  pm: TPopupMenu;
  P: TPoint;
begin
  if csDesigning in ComponentState then
    Exit;
  I:=FOwner.IndexOfTabAt(X, Y);
  if I < 0 then
    Exit;
  // check CloseBtn
  if (Button = mbLeft) and CloseBtnShown(I) and PtInRect(CloseBtnRect(I), Point(X, Y)) then
  begin
{$IFDEF DEMO}
    ShowDemoMsg;
{$ENDIF}
    CanClose:=true;
    if Assigned(FOnTabClose) then
      FOnTabClose(FOwner, I, CanClose);
    if CanClose then
    begin
      if FOwner is TPageControl then
        with (FOwner as TPageControl) do
        begin
          ActivePage.TabVisible:=false;
          SelectNextPage(false);
        end;
      if FOwner is TTabControl then
        with (FOwner as TTabControl) do
        begin
          Tabs.Delete(I);
          if I > 0 then
            Dec(I);
          if I < Tabs.Count then
            TabIndex:=I;
        end;
    end;
  end;
  // popup menu
  if (Button = mbRight) then
  begin
    pm:=TC.PopupMenu;
    if Assigned(FOnTabGetPopupMenu) then
      FOnTabGetPopupMenu(FOwner, I, pm);
    P.X:=X;
    P.Y:=Y;
    P:=FOwner.ClientToScreen(P);
    if Assigned(pm) then
    begin
      if FOwner is TrPageControlEx then
        TrPageControlEx(FOwner).SetActiveTabEx(I);
      if FOwner is TrTabControlEx then
        TrTabControlEx(FOwner).SetActiveTabEx(I);
      pm.Popup(P.X, P.Y);
    end;
  end;
end;

// ==============================================================================
// TrPageControlEx
constructor TrPageControlEx.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  FTabPropsEx:=TTabPropsEx.Create(Self);
  FTabRightButtonVisible:=false;
  FTabRightLabelVisible:=false;

  FTabRightButton:=TrButton.Create(Self);
  FTabRightButton.SetSubComponent(true);
  FTabRightButton.Name:=Name + 'RightButton';
  FTabRightButton.Parent:=Self;
  FTabRightButton.Caption:='';
  FTabRightButton.Height:=18;
  FTabRightButton.Width:=20;
  FTabRightButton.Top:=0;
  FTabRightButton.Left:=0;

  FTabRightLabel:=TrHTMLLabel.Create(Self);
  FTabRightLabel.SetSubComponent(true);
  FTabRightLabel.Name:=Name + 'RightLabel';
  FTabRightLabel.Parent:=Self;
  FTabRightLabel.Caption:='';
  FTabRightLabel.Top:=0;
  FTabRightLabel.Left:=0;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrPageControlEx.Destroy;
begin
  FreeAndNil(FTabPropsEx);
  inherited;
end;

procedure TrPageControlEx.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrPageControlEx.SetTabRightButtonVisible(Value: Boolean);
begin
  FTabRightButtonVisible:=Value;
  FTabRightButton.Visible:=FTabRightButtonVisible;
end;

procedure TrPageControlEx.SetTabRightLabelVisible(Value: Boolean);
begin
  FTabRightLabelVisible:=Value;
  FTabRightLabel.Visible:=FTabRightLabelVisible;
end;

procedure TrPageControlEx.AdjustClientRect(var Rect: TRect);
var
  maxW, maxH, us4: integer;
  R: TRect;
begin
  inherited AdjustClientRect(Rect);
  us4:=UseScale(4);
  if Tabs.Count = 0 then
  begin
    InflateRect(Rect, us4, us4);
    if TabPosition = tpTop then
      Dec(Rect.Top, 2);
  end;
  // align horizontaly right button and label
  maxW:=Width - us4;
  FTabRightButton.Left:=maxW - FTabRightButton.Width;
  if (csDesigning in ComponentState) or FTabRightButtonVisible then
    maxW:=FTabRightButton.Left - us4;
  FTabRightLabel.Left:=maxW - FTabRightLabel.Width;

  // get tab size
  maxW:=0;
  maxH:=0;
  if Tabs.Count > 0 then
  begin
    R:=TabRect(Tabs.Count - 1);
    maxW:=R.Right;
    maxH:=R.Bottom - R.Top;
  end;
  // align verticaly right button and label
  if maxH > 0 then
  begin
    FTabRightButton.Top:=R.Top + (maxH - 1 - FTabRightButton.Height) div 2;
    FTabRightLabel.Top:=R.Top + (maxH - 1 - FTabRightLabel.Height) div 2;
  end;
  // hide right button and label if small width
  if maxW > FTabRightLabel.Left - UseScale(rPageExSpace) then
    FTabRightLabel.Visible:=false
  else
    FTabRightLabel.Visible:=Self.FTabRightLabelVisible;
  if maxW > FTabRightButton.Left - UseScale(rPageExSpace) then
    FTabRightButton.Visible:=false
  else
    FTabRightButton.Visible:=Self.FTabRightButtonVisible;
end;

procedure TrPageControlEx.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  FTabPropsEx.UpdateScale(FScaleRatio, M, D);
end;

function TrPageControlEx.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

procedure TrPageControlEx.CNDrawItem(var Message: TWMDrawItem);
begin
  FTabPropsEx.DrawItemMessage(Message);
end;

procedure TrPageControlEx.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetHotState(false);
end;

procedure TrPageControlEx.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetHotState(true);
end;

procedure TrPageControlEx.MouseMove(Shift: TShiftState; X, Y: integer);
begin
  inherited MouseMove(Shift, X, Y);
  FTabPropsEx.CheckMouseMove(Shift, X, Y);
end;

procedure TrPageControlEx.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  FTabPropsEx.CheckMouseClick(Button, Shift, X, Y);
end;

procedure TrPageControlEx.CMFontChanged(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetFont(Font);
end;

procedure TrPageControlEx.AddSpaceForCloseBtn(Center: Boolean = false);
var
  I: integer;
  s: string;
begin
  for I:=0 to Self.PageCount - 1 do
  begin
    s:=Pages[I].Caption;
    if Center then
      s:='  ' + Trim(s) + '  '
    else
      s:=Trim(s) + '    ';
    Pages[I].Caption:=s;
  end;
end;

function TrPageControlEx.SetActiveTabEx(Index: integer): Boolean;
begin
  Result:=false;
  if (Index <> TabIndex) and CanChange then
  begin
    Self.TabIndex:=Index;
    Change;
    Result:=true;
  end;
end;

function TrPageControlEx.SetActivePageEx(TS: TTabSheet): Boolean;
begin
  Result:=false;
  if (TS <> Self.ActivePage) and CanChange then
  begin
    Self.ActivePage:=TS;
    Change;
    Result:=true;
  end;
end;

function TrPageControlEx.ActiveTabCaption: string;
begin
  if Self.Tabs.Count > 0 then
    Result:=Tabs[TabIndex]
  else
    Result:='';
end;

function TrPageControlEx.ActiveTabObject: TObject;
begin
  if Self.Tabs.Count > 0 then
    Result:=Tabs.Objects[TabIndex]
  else
    Result:=nil;
end;

function TrPageControlEx.VisiblePageCount: integer;
var
  I: integer;
begin
  Result:=0;
  for I:=0 to PageCount - 1 do
    if Pages[I].TabVisible then
      Inc(Result);
end;

// ==============================================================================
// TrTabControlEx
constructor TrTabControlEx.Create(AOwner: TComponent);
begin
  inherited;
  FScaleRatio:=1;
  FTabPropsEx:=TTabPropsEx.Create(Self);

  FTabRightButtonVisible:=false;
  FTabRightLabelVisible:=false;

  FTabRightButton:=TrButton.Create(Self);
  FTabRightButton.SetSubComponent(true);
  FTabRightButton.Name:=Name + 'RightButton';
  FTabRightButton.Parent:=Self;
  FTabRightButton.Caption:='';
  FTabRightButton.Height:=18;
  FTabRightButton.Width:=20;
  FTabRightButton.Top:=0;
  FTabRightButton.Left:=0;

  FTabRightLabel:=TrHTMLLabel.Create(Self);
  FTabRightLabel.SetSubComponent(true);
  FTabRightLabel.Name:=Name + 'RightLabel';
  FTabRightLabel.Parent:=Self;
  FTabRightLabel.Caption:='';
  FTabRightLabel.Top:=0;
  FTabRightLabel.Left:=0;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentCreatedEventID);
end;

destructor TrTabControlEx.Destroy;
begin
  FreeAndNil(FTabPropsEx);
  inherited;
end;

procedure TrTabControlEx.Loaded;
begin
  inherited;
  RosiCompConfig.DoRosiCompGeneralEventSimple(Self, rDBComps_ComponentLoadedEventID);
end;

procedure TrTabControlEx.AdjustClientRect(var Rect: TRect);
var
  maxW, maxH, us4: integer;
  R: TRect;
begin
  inherited AdjustClientRect(Rect);
  us4:=UseScale(4);
  if Tabs.Count = 0 then
  begin
    InflateRect(Rect, us4, us4);
    if TabPosition = tpTop then
      Dec(Rect.Top, 2);
  end;
  // align horizontaly right button and label
  maxW:=Width - us4;
  FTabRightButton.Left:=maxW - FTabRightButton.Width;
  if (csDesigning in ComponentState) or FTabRightButtonVisible then
    maxW:=FTabRightButton.Left - us4;
  FTabRightLabel.Left:=maxW - FTabRightLabel.Width;

  // get tab size
  maxW:=0;
  maxH:=0;
  if Tabs.Count > 0 then
  begin
    R:=TabRect(Tabs.Count - 1);
    maxW:=R.Right;
    maxH:=R.Bottom - R.Top;
  end;
  // align verticaly right button and label
  if maxH > 0 then
  begin
    FTabRightButton.Top:=R.Top + (maxH - 1 - FTabRightButton.Height) div 2;
    FTabRightLabel.Top:=R.Top + (maxH - 1 - FTabRightLabel.Height) div 2;
  end;
  // hide right button and label if small width
  if maxW > FTabRightLabel.Left - UseScale(rPageExSpace) then
    FTabRightLabel.Visible:=false
  else
    FTabRightLabel.Visible:=Self.FTabRightLabelVisible;
  if maxW > FTabRightButton.Left - UseScale(rPageExSpace) then
    FTabRightButton.Visible:=false
  else
    FTabRightButton.Visible:=Self.FTabRightButtonVisible;
end;

procedure TrTabControlEx.ChangeScale(M: integer; D: integer {$IFDEF DX102P}; isDpiChange: Boolean {$ENDIF});
begin
  inherited;
  FScaleRatio:=FScaleRatio * M / D;
  FTabPropsEx.UpdateScale(FScaleRatio, M, D);
end;

function TrTabControlEx.UseScale(I: integer): integer;
begin
  Result:=Round(I * FScaleRatio);
end;

procedure TrTabControlEx.SetTabRightButtonVisible(Value: Boolean);
begin
  FTabRightButtonVisible:=Value;
  FTabRightButton.Visible:=FTabRightButtonVisible;
end;

procedure TrTabControlEx.SetTabRightLabelVisible(Value: Boolean);
begin
  FTabRightLabelVisible:=Value;
  FTabRightLabel.Visible:=FTabRightLabelVisible;
end;

procedure TrTabControlEx.CNDrawItem(var Message: TWMDrawItem);
begin
  FTabPropsEx.DrawItemMessage(Message);
end;

procedure TrTabControlEx.CMMouseEnter(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetHotState(false);
end;

procedure TrTabControlEx.CMMouseLeave(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetHotState(true);
end;

procedure TrTabControlEx.CMFontChanged(var Message: TMessage);
begin
  inherited;
  FTabPropsEx.ResetFont(Font);
end;

procedure TrTabControlEx.MouseMove(Shift: TShiftState; X, Y: integer);
begin
  inherited MouseMove(Shift, X, Y);
  FTabPropsEx.CheckMouseMove(Shift, X, Y);
end;

procedure TrTabControlEx.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  FTabPropsEx.CheckMouseClick(Button, Shift, X, Y);
end;

procedure TrTabControlEx.AddSpaceForCloseBtn(Center: Boolean = false);
var
  I: integer;
  s: string;
begin
  for I:=0 to Self.Tabs.Count - 1 do
  begin
    s:=Tabs[I];
    if Center then
      s:='  ' + Trim(s) + '  '
    else
      s:=Trim(s) + '    ';
    Tabs[I]:=s;
  end;
end;

procedure TrTabControlEx.ReplaceAccelChar;
var
  I: integer;
begin
  for I:=0 to Self.Tabs.Count - 1 do
    Tabs[I]:=ReplaceStr('&', '&&', Tabs[I]);
end;

function TrTabControlEx.SetActiveTabEx(Index: integer): Boolean;
begin
  Result:=false;
  if (TabIndex <> Index) and CanChange then
  begin
    Self.TabIndex:=Index;
    Change;
    Result:=true;
  end;
end;

function TrTabControlEx.ActiveTabCaption: string;
begin
  if Self.Tabs.Count > 0 then
    Result:=Tabs[TabIndex]
  else
    Result:='';
end;

function TrTabControlEx.ActiveTabObject: TObject;
begin
  if Self.Tabs.Count > 0 then
    Result:=Tabs.Objects[TabIndex]
  else
    Result:=nil;
end;

// ==============================================================================
// TrRichEdit
const
  cRichMenuPopupColorArr: array [1 .. 9] of integer = (0, 9, 1, 5, 2, 12, 14, 7, 3);

constructor TrRichEdit.Create(AOwner: TComponent);
var
  mi: TMenuItem;
  I: integer;
begin
  inherited Create(AOwner);
  FSupportURLLink:=false;
  FPopupMenu:=TPopupMenu.Create(Self);
  PopupMenu:=FPopupMenu;
  FPopupMenu.OnPopup:=PopupMenuOpen;

  // Font menu and submenus
  mi:=rtool.CreateMenuItem(FPopupMenu, sFont, 0, OnPopupClick);
  FPopupMenu.Items.Add(mi);
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFormatBold, TextToShortCut('Ctrl+B'), 101, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFormatItalic, TextToShortCut('Ctrl+I'), 102, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFormatUnderline, TextToShortCut('Ctrl+U'), 103, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFormatStrikeOut, TextToShortCut('Ctrl+Shift+S'), 104, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, '-', 0, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFontSmaller, TextToShortCut('Ctrl+Shift+,'), 111, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sFontBigger, TextToShortCut('Ctrl+Shift+.'), 112, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, sChangeFont + '...', TextToShortCut('Ctrl+Shift+F'), 113, OnPopupClick));

  // Color menu and submenus
  mi:=rtool.CreateMenuItem(FPopupMenu, sFontColor, 0, OnPopupClick);
  FPopupMenu.Items.Add(mi);
  for I:=1 to High(cRichMenuPopupColorArr) do
  begin
    mi.Add(rtool.CreateMenuItem(FPopupMenu, GetColorName(cRichMenuPopupColorArr[I], true), TextToShortCut('Ctrl+Shift+' + IntToStr(I)), 200 + I, OnPopupClick));
  end;

  // Paragraph menu and submenus
  mi:=rtool.CreateMenuItem(FPopupMenu, sParagraph, 0, OnPopupClick);
  FPopupMenu.Items.Add(mi);
  mi.Add(rtool.CreateMenuItem(Self, sIndentDecrease, TextToShortCut('Ctrl+Shift+U'), 301, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sIndentIncrease, TextToShortCut('Ctrl+Shift+I'), 302, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, '-', 0, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sTextAlignLeft, TextToShortCut('Ctrl+L'), 311, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sTextAlignCenter, TextToShortCut('Ctrl+E'), 312, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sTextAlignRight, TextToShortCut('Ctrl+R'), 313, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(FPopupMenu, '-', 0, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sBulletPoints, TextToShortCut('Ctrl+Shift+B'), 321, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sBulletNumbers, TextToShortCut('Ctrl+Shift+N'), 322, OnPopupClick));
  mi.Add(rtool.CreateMenuItem(Self, sBulletLetters, TextToShortCut('Ctrl+Shift+L'), 323, OnPopupClick));

  // add sep
  FPopupMenu.Items.Add(rtool.CreateMenuItem(FPopupMenu, '-', 0, OnPopupClick));
  // standard Edit menu
  DefineEditPopupMenuItems(FPopupMenu, OnPopupClick);
end;

procedure TrRichEdit.CreateWnd;
var
  Mask: DWORD;
begin
  inherited;
  if FSupportURLLink then
  begin
    Mask:=SendMessage(Handle, EM_GETEVENTMASK, 0, 0);
    SendMessage(Handle, EM_SETEVENTMASK, 0, Mask or ENM_LINK);
    SendMessage(Handle, EM_AUTOURLDETECT, 1, 0);
  end;
end;

procedure TrRichEdit.WndProc(var Message: TMessage);
type
  PENLINK = ^ENLINK;
var
  tr: TEXTRANGE;
  Str: string;
  P: PENLINK;
begin
  inherited WndProc(Message);
  case Message.Msg of
    CN_NOTIFY:
      begin
        if FSupportURLLink and (TWMNotify(Message).NMHdr.code = EN_LINK) then
        begin
          P:=PENLINK(Message.LParam);
          if P.Msg = WM_LBUTTONUP then
          begin
            SetLength(Str, P.chrg.cpMax - P.chrg.cpMin);
            tr.chrg:=P.chrg;
            tr.lpstrText:=PChar(Str);
            SendMessage(Handle, EM_GETTEXTRANGE, 0, LParam(@tr));
            ShellExecute(Handle, nil, PChar(Str), nil, nil, SW_SHOWDEFAULT);
          end;
        end;
      end;
  end;
end;

procedure TrRichEdit.SetSupportURLLink(Value: Boolean);
begin
  if Value<>FSupportURLLink then
  begin
    FSupportURLLink:=Value;
    RecreateWnd;
  end;
end;

procedure TrRichEdit.PopupMenuOpen(Sender: TObject);
var
  I: integer;
begin
{$IFDEF DEMO}
  ShowDemoMsg;
{$ENDIF}
  for I:=0 to PopupMenu.Items.Count - 1 do
  begin
    case PopupMenu.Items[I].Tag of
      rcEditUndo:
        PopupMenu.Items[I].Enabled:=CanUndo;
      rcEditCut:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditCopy:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditPaste:
        PopupMenu.Items[I].Enabled:=Clipbrd.Clipboard.AsText <> '';
      rcEditDelete:
        PopupMenu.Items[I].Enabled:=SelLength > 0;
      rcEditSelectAll:
        ;
      rcEditFind:
        ;
      rcEditFindNext:
        PopupMenu.Items[I].Enabled:=FSearchText <> '';
      rcEditSelectFont:
        PopupMenu.Items[I].Visible:=false;
      rcMemoWordWrap:
        begin
          PopupMenu.Items[I].Checked:=WordWrap;
        end;
      rcMemoLoadFile:
        ; // PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
      rcMemoSaveFile:
        ; // PopupMenu.Items[I].Enabled:=moAllowLoadSaveFile in FOptionsEx;
    end;
  end;
end;

procedure TrRichEdit.OnPopupClick(Sender: TObject);
begin
  PopupCommand(TMenuItem(Sender).Tag);
end;

procedure TrRichEdit.PopupCommand(CmdTag: integer);
var
  SaveDlg: TrSaveDialog;
  OpenDlg: TrOpenDialog;
begin
  case CmdTag of
    rcEditUndo:
      Undo;
    rcEditCut:
      CutToClipboard;
    rcEditCopy:
      CopyToClipboard;
    rcEditPaste:
      PasteFromClipboard;
    rcEditDelete:
      ClearSelection;
    rcEditSelectAll:
      SelectAll;
    rcEditFind:
      SearchText;
    rcEditFindNext:
      SearchNext;
    rcEditSelectFont:
      rDialogsEx.ShowSelectFontDialog(Font);
    rcMemoWordWrap:
      begin
        WordWrap:=not WordWrap;
        if WordWrap and (ScrollBars = ssBoth) then
          ScrollBars:=ssVertical;
        if not WordWrap and (ScrollBars = ssVertical) then
          ScrollBars:=ssBoth;
      end;
    rcMemoLoadFile:
      begin
        OpenDlg:=TrOpenDialog.Create(Self);
        try
          if OpenDlg.Execute then
            Lines.LoadFromFile(OpenDlg.FileName);
        finally
          OpenDlg.Free;
        end;
      end;
    rcMemoSaveFile:
      begin
        SaveDlg:=TrSaveDialog.Create(Self);
        try
          if SaveDlg.Execute then
            Lines.SaveToFile(SaveDlg.FileName);
        finally
          SaveDlg.Free;
        end;
      end;

    101:
      SwitchFontStyle(fsBold);
    102:
      SwitchFontStyle(fsItalic);
    103:
      SwitchFontStyle(fsUnderline);
    104:
      SwitchFontStyle(fsStrikeOut);

    111:
      DecreaseFontSize();
    112:
      IncreaseFontSize();
    113:
      SelectFont;

    200 .. 210:
      SelAttributes.Color:=GetColorValue(cRichMenuPopupColorArr[CmdTag - 200]);

    301:
      DecreaseIndent();
    302:
      IncreaseIndent();

    311:
      Paragraph.Alignment:=taLeftJustify;
    312:
      Paragraph.Alignment:=taCenter;
    313:
      Paragraph.Alignment:=taRightJustify;

    321:
      SwitchBulletPoints;
    322:
      SwitchBulletNumbers;
    323:
      SwitchBulletLetters;
  end;
end;

function TrRichEdit.PixelsToTwips(pixels: integer): integer;
begin
  Result:=pixels * 1440 div Screen.PixelsPerInch;
end;

procedure TrRichEdit.DrawToCanvasInt(var R: TRect; OffsetX, OffsetY: integer; ACanvas: TCanvas; CalcSize, ShowEllipsis: Boolean);
var
  FBmp: TBitmap;
  range: TFormatRange;
  ox, oy, h: integer;
  aR: TRect;
begin
  FBmp:=TBitmap.Create;
  try
    // init bmp with rect with and maximal height
    FBmp.Width:=R.Right - R.Left;
    FBmp.Height:=Screen.Height;

    // init RtfEdit range
    ox:=PixelsToTwips(OffsetX);
    oy:=PixelsToTwips(OffsetY);
    FillChar(range, SizeOf(TFormatRange), 0);
    range.HDC:=FBmp.Canvas.Handle;
    range.hdcTarget:=FBmp.Canvas.Handle;
    range.rc.Left:=ox;
    range.rc.Top:=oy;
    range.rc.Right:=PixelsToTwips(FBmp.Width);
    range.rc.Bottom:=PixelsToTwips(FBmp.Height);
    range.chrg.cpMin:=0;
    range.chrg.cpMax:=-1;
    // calc size
    SendMessage(Handle, EM_FORMATRANGE, 0, Longint(@range));
    h:=((range.rc.Bottom - range.rc.Top) * Screen.PixelsPerInch) div 1440;

    if CalcSize then
    begin
      R.Bottom:=R.Top + h + 4;
    end
    else
    begin
      // set required size and draw
      FBmp.Height:=R.Bottom - R.Top;
      FBmp.Canvas.Brush.Color:=Color;
      FBmp.Canvas.FillRect(Rect(0, 0, FBmp.Width, FBmp.Height));
      range.HDC:=FBmp.Canvas.Handle;
      range.hdcTarget:=FBmp.Canvas.Handle;
      range.rc.Bottom:=PixelsToTwips(FBmp.Height);
      SendMessage(Handle, EM_FORMATRANGE, 1, Longint(@range));
      // show trim glyphs
      if ShowEllipsis and (h > FBmp.Height) and (Lines.Text <> '') then
      begin
        FBmp.Canvas.Brush.Style:=bsSolid;
        FBmp.Canvas.Brush.Color:=Font.Color;
        aR.Top:=FBmp.Height - 8;
        aR.Bottom:=aR.Top + 3;
        for h:=0 to 2 do
        begin
          aR.Left:=FBmp.Width - 20 + h * 5;
          aR.Right:=aR.Left + 3;
          FBmp.Canvas.FillRect(aR);
        end;
      end;
      rpictool.DrawBmpCenter(ACanvas, R, FBmp);
    end;
    // free mem
    SendMessage(Handle, EM_FORMATRANGE, 0, 0);
  finally
    FBmp.Free;
  end;
end;

procedure TrRichEdit.CalculateTextHeight(var R: TRect);
begin
  if Lines.Text <> '' then
    DrawToCanvasInt(R, 0, 0, nil, true, true);
end;

procedure TrRichEdit.DrawToCanvas(R: TRect; OffsetX, OffsetY: integer; ACanvas: TCanvas; ShowEllipsis: Boolean = true);
begin
  DrawToCanvasInt(R, OffsetX, OffsetY, ACanvas, false, ShowEllipsis);
end;

procedure TrRichEdit.AddSelectText(AText: string);
var
  P: integer;
begin
  SelectAll;
  P:=SelLength - 1;
  SelStart:=P;
  SelLength:=0;
  SetSelText(AText);
  SelStart:=P;
  SelLength:=Length(AText);
end;

procedure TrRichEdit.InsertSelectText(AText: string; ReplaceSelected: Boolean = false; BeforeSelected: Boolean = false);
var
  P: integer;
begin
  if not ReplaceSelected then
  begin
    if not BeforeSelected then
      SelStart:=SelStart + SelLength;
    SelLength:=0;
  end;
  P:=SelStart;
  SetSelText(AText);
  SelStart:=P;
  SelLength:=Length(AText);
end;

procedure TrRichEdit.InitParaFormat2(var Para2: TParaFormat2);
begin
{$IF DEFINED(CLR)}
  Paragraph.cbSize:=Marshal.SizeOf(TypeOf(TParaFormat2));
{$ELSE}
  FillChar(Para2, SizeOf(Para2), 0);
  Para2.cbSize:=SizeOf(Para2);
{$IFEND}
end;

procedure TrRichEdit.GetParaFormat2(var Para2: TParaFormat2);
begin
  InitParaFormat2(Para2);
  if HandleAllocated then
    SendGetStructMessage(Handle, EM_GETPARAFORMAT, 0, Para2, true);
end;

procedure TrRichEdit.SetParaFormat2(var Para2: TParaFormat2);
begin
  HandleNeeded; { copied from ComCtrls.TParaAttributes }
  if HandleAllocated then
  begin
    if UseRightToLeftAlignment then
      if Para2.wAlignment = PFA_LEFT then
        Para2.wAlignment:=PFA_RIGHT
      else
        if Para2.wAlignment = PFA_RIGHT then
          Para2.wAlignment:=PFA_LEFT;
    SendStructMessage(Handle, EM_SETPARAFORMAT, 0, Para2);
  end;
end;

function TrRichEdit.GetNumberingEx: integer;
var
  Para2: TParaFormat2;
begin
  GetParaFormat2(Para2);
  Result:=(Para2.wNumbering);
end;

procedure TrRichEdit.SetNumberingEx(Value: integer);
var
  Para2: TParaFormat2;
begin
  if Value = 0 then
  begin
    // reset indent if no bullets
    Paragraph.LeftIndent:=0;
    Paragraph.FirstIndent:=0;
  end
  else
    if Min(Paragraph.LeftIndent, Paragraph.FirstIndent) < 10 then
    begin
      Paragraph.LeftIndent:=10;
      Paragraph.FirstIndent:=10;
    end;
  InitParaFormat2(Para2);
  with Para2 do
  begin
    dwMask:=PFM_NUMBERING or PFM_NUMBERINGSTYLE or PFM_NUMBERINGSTART or PFM_NUMBERINGTAB;
    wNumbering:=(Value);
    if Value = 2 then
      wNumberingStyle:=$200
    else
      wNumberingStyle:=0;
    wNumberingStart:=1;
    wNumberingTab:=1440 div 4;
  end;
  SetParaFormat2(Para2);
end;

{
  // The PARAFORMAT2 structure is used to set the numbering style.
  // This is done through the following structure members:
  fmt.dwMask
  // Set the following values (bitwise-or them together) to identify
  // which of the remaining structure members are valid:
  // PFM_NUMBERING, PFM_NUMBERINGSTART, PFM_NUMBERINGSTYLE, and PFM_NUMBERINGTAB
  fmt.wNumbering
  //0 no numbering or bullets
  //1 (PFN_BULLET) uses bullet character
  //2 Uses Arabic numbers (1, 2, 3, ...).
  //3 Uses lowercase letters (a, b, c, ...).
  //4 Uses uppercase letters (A, B, C, ...).
  //5 Uses lowercase Roman numerals (i, ii, iii, ...).
  //6 Uses uppercase Roman numerals (I, II, III, ...).
  //7 Uses a sequence of characters beginning with the Unicode
  //  character specified by the wNumberingStart member.
  fmt.wNumberingStart := 1;
  //  Starting value for numbering.
  fmt.wNumberingStyle := $200;
  // Styles for numbering:
  // 0 : Follows the number with a right parenthesis.  1)
  // $100 : Encloses the number in parentheses.       (1)
  // $200 : Follows the number with a period.          1.
  // $300 : Displays only the number.                  1
  // $400 : Continues a numbered list without applying the next number or bullet.
  // $8000 : Starts a new number with wNumberingStart.
  fmt.wNumberingTab := 1440 div 4;
  // Minimum space between a paragraph number and the paragraph text, in twips
  end;
}

function TrRichEdit.GetRtfText: string;
var
  s: TStringStream;
begin
  s:=TStringStream.Create;
  try
    Lines.SaveToStream(s);
    Result:=rtool.CleanString(s.DataString, #0);
  finally
    s.Free;
  end;
end;

procedure TrRichEdit.SetRtfText(AValue: string);
var
  s: TStringStream;
begin
  s:=TStringStream.Create(AValue);
  try
    Lines.LoadFromStream(s);
  finally
    s.Free;
  end;
end;

function TrRichEdit.SearchText(const Search: string = ''; FromStart: Boolean = true; IgnoreCase: Boolean = true): Boolean;
var
  sText: string;
  Index: integer;
begin
  Result:=false;
  FSearchText:=Search;
  if FSearchText = '' then
    if not rdlg.InputStr(sInput, SFind, FSearchText) then
      Exit;
  FSearchIgnoreCase:=IgnoreCase;
  sText:=Lines.Text;
  if FSearchIgnoreCase then
  begin
    FSearchText:=UpperCase(FSearchText);
    sText:=UpperCase(Lines.Text);
  end;
  if FromStart then
    Index:=0
  else
    Index:=SelStart + SelLength + 1;
  Index:=FindText(FSearchText, Index, GetTextLen, []);
  Result:=Index >= 0;
  if Result then
  begin
    SelStart:=Index;
    SelLength:=Length(FSearchText);
    try
      SetFocus;
    except
    end;
  end;
  if not Result then
    DlgI(sNotFound);
end;

function TrRichEdit.SearchNext;
begin
  Result:=SearchText(FSearchText, false, FSearchIgnoreCase);
end;

procedure TrRichEdit.SwitchFontStyle(AStyle: TFontStyle);
begin
  if AStyle in SelAttributes.Style then
    SelAttributes.Style:=SelAttributes.Style - [AStyle]
  else
    SelAttributes.Style:=SelAttributes.Style + [AStyle];
end;

function TrRichEdit.IncreaseFontSize(Step: integer = 0): Boolean;
begin
  Result:=SelAttributes.size < 120;
  if Result then
  begin
    if Step = 0 then
      Step:=1 + SelAttributes.size div 12;
    SelAttributes.size:=SelAttributes.size + Step;
  end;
end;

function TrRichEdit.DecreaseFontSize(Step: integer = 0): Boolean;
begin
  Result:=SelAttributes.size > 5;
  if Result then
  begin
    if Step = 0 then
      Step:=1 + SelAttributes.size div 12;
    SelAttributes.size:=SelAttributes.size - Step;
  end;
end;

function TrRichEdit.SelectFont: Boolean;
var
  f: TFont;
begin
  f:=TFont.Create;
  try
    f.Assign(SelAttributes);
    Result:=rDialogsEx.ShowSelectFontDialog(f);
    if Result then
      SelAttributes.Assign(f);
  finally
    f.Free;
  end;
end;

function TrRichEdit.IncreaseIndent(Step: integer = 10): Boolean;
begin
  Result:=Paragraph.LeftIndent < 1200;
  if Result then
  begin
    Paragraph.LeftIndent:=Paragraph.LeftIndent + Step;
    Paragraph.FirstIndent:=Paragraph.LeftIndent;
  end;
end;

function TrRichEdit.DecreaseIndent(Step: integer = 10): Boolean;
begin
  Result:=Max(Paragraph.LeftIndent, Paragraph.FirstIndent) > 0;
  if Result then
  begin
    Paragraph.LeftIndent:=Paragraph.LeftIndent - Step;
    Paragraph.FirstIndent:=Paragraph.LeftIndent;
  end;
end;

function TrRichEdit.SwitchBulletPoints: Boolean;
begin
  Result:=NumberingEx <> 1;
  if Result then
    NumberingEx:=1
  else
    NumberingEx:=0;
end;

function TrRichEdit.SwitchBulletNumbers: Boolean;
begin
  Result:=NumberingEx <> 2;
  if Result then
    NumberingEx:=2
  else
    NumberingEx:=0;
end;

function TrRichEdit.SwitchBulletLetters: Boolean;
begin
  Result:=NumberingEx <> 3;
  if Result then
    NumberingEx:=3
  else
    NumberingEx:=0;
end;

end.
